#include <catch2/catch_session.hpp>

#include <SFML/Main.hpp>

////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
    return Catch::Session().run(argc, argv);
}
