#include <SFML/Config.hpp>

#include <catch2/catch_test_macros.hpp>

TEST_CASE("[System] SFML/Config.hpp")
{
    SECTION("Version macros")
    {
        STATIC_CHECK(SFML_VERSION_MAJOR == EXPECTED_SFML_VERSION_MAJOR);
        STATIC_CHECK(SFML_VERSION_MINOR == EXPECTED_SFML_VERSION_MINOR);
        STATIC_CHECK(SFML_VERSION_PATCH == EXPECTED_SFML_VERSION_PATCH);
        STATIC_CHECK(SFML_VERSION_IS_RELEASE == EXPECTED_SFML_VERSION_IS_RELEASE);
    }
}
