/* Override search/search.css rules to adjust search box placement */
#MSearchBox {
    position: relative;
    display: block;
    border-radius: 0;
    box-shadow: none;
    border-width: 1px;
    border-color: #BBB;
    border-style: solid;
}
.right #MSearchClose {
    margin: 0;
    right: auto;
}
#MSearchResultsWindow {
    background-color: #EAF5DB;
    border-color: #999;
}
#MSearchResults > .SRPage {
    background-color: #EAF5DB;
}
a.SRSymbol {
    color: rgb(70, 100, 30);
}
a.SRSymbol:hover {
    text-decoration: underline;
}
a.SRScope {
    color: rgb(70, 100, 30);
}
a.SRScope:hover {
    text-decoration: underline;
}
#MSearchSelectWindow {
    background-color: #EAF5DB;
}
#MSearchSelectWindow > a.SelectItem {
    color: #333;
    text-decoration: none;
}
#MSearchSelectWindow > a.SelectItem:hover {
    color: white;
    background-color: #8CC841;
    text-decoration: none;
}
