
#pragma once

#include <cstdlib>

namespace unicode {

const char32_t START_OF_HEADING = 0x0001;
const char32_t START_OF_TEXT = 0x0002;
const char32_t END_OF_TEXT = 0x0003;
const char32_t END_OF_TRANSMISSION = 0x0004;
const char32_t ENQUIRY = 0x0005;
const char32_t ACKNOWLEDGE = 0x0006;
const char32_t ALERT = 0x0007;
const char32_t BACKSPACE = 0x0008;
const char32_t CHARACTER_TABULATION = 0x0009;
const char32_t LINE_FEED = 0x000A;
const char32_t LINE_TABULATION = 0x000B;
const char32_t FORM_FEED = 0x000C;
const char32_t CARRIAGE_RETURN = 0x000D;
const char32_t SHIFT_OUT = 0x000E;
const char32_t SHIFT_IN = 0x000F;
const char32_t DATA_LINK_ESCAPE = 0x0010;
const char32_t DEVICE_CONTROL_ONE = 0x0011;
const char32_t DEVICE_CONTROL_TWO = 0x0012;
const char32_t DEVICE_CONTROL_THREE = 0x0013;
const char32_t DEVICE_CONTROL_FOUR = 0x0014;
const char32_t NEGATIVE_ACKNOWLEDGE = 0x0015;
const char32_t SYNCHRONOUS_IDLE = 0x0016;
const char32_t END_OF_TRANSMISSION_BLOCK = 0x0017;
const char32_t CANCEL = 0x0018;
const char32_t END_OF_MEDIUM = 0x0019;
const char32_t SUBSTITUTE = 0x001A;
const char32_t ESCAPE = 0x001B;
const char32_t INFORMATION_SEPARATOR_FOUR = 0x001C;
const char32_t INFORMATION_SEPARATOR_THREE = 0x001D;
const char32_t INFORMATION_SEPARATOR_TWO = 0x001E;
const char32_t INFORMATION_SEPARATOR_ONE = 0x001F;
const char32_t SPACE = 0x0020;
const char32_t EXCLAMATION_MARK = 0x0021;
const char32_t QUOTATION_MARK = 0x0022;
const char32_t NUMBER_SIGN = 0x0023;
const char32_t DOLLAR_SIGN = 0x0024;
const char32_t PERCENT_SIGN = 0x0025;
const char32_t AMPERSAND = 0x0026;
const char32_t APOSTROPHE = 0x0027;
const char32_t LEFT_PARENTHESIS = 0x0028;
const char32_t RIGHT_PARENTHESIS = 0x0029;
const char32_t ASTERISK = 0x002A;
const char32_t PLUS_SIGN = 0x002B;
const char32_t COMMA = 0x002C;
const char32_t HYPHEN_MINUS = 0x002D;
const char32_t FULL_STOP = 0x002E;
const char32_t SOLIDUS = 0x002F;
const char32_t DIGIT_ZERO = 0x0030;
const char32_t DIGIT_ONE = 0x0031;
const char32_t DIGIT_TWO = 0x0032;
const char32_t DIGIT_THREE = 0x0033;
const char32_t DIGIT_FOUR = 0x0034;
const char32_t DIGIT_FIVE = 0x0035;
const char32_t DIGIT_SIX = 0x0036;
const char32_t DIGIT_SEVEN = 0x0037;
const char32_t DIGIT_EIGHT = 0x0038;
const char32_t DIGIT_NINE = 0x0039;
const char32_t COLON = 0x003A;
const char32_t SEMICOLON = 0x003B;
const char32_t LESS_THAN_SIGN = 0x003C;
const char32_t EQUALS_SIGN = 0x003D;
const char32_t GREATER_THAN_SIGN = 0x003E;
const char32_t QUESTION_MARK = 0x003F;
const char32_t COMMERCIAL_AT = 0x0040;
const char32_t LATIN_CAPITAL_LETTER_A = 0x0041;
const char32_t LATIN_CAPITAL_LETTER_B = 0x0042;
const char32_t LATIN_CAPITAL_LETTER_C = 0x0043;
const char32_t LATIN_CAPITAL_LETTER_D = 0x0044;
const char32_t LATIN_CAPITAL_LETTER_E = 0x0045;
const char32_t LATIN_CAPITAL_LETTER_F = 0x0046;
const char32_t LATIN_CAPITAL_LETTER_G = 0x0047;
const char32_t LATIN_CAPITAL_LETTER_H = 0x0048;
const char32_t LATIN_CAPITAL_LETTER_I = 0x0049;
const char32_t LATIN_CAPITAL_LETTER_J = 0x004A;
const char32_t LATIN_CAPITAL_LETTER_K = 0x004B;
const char32_t LATIN_CAPITAL_LETTER_L = 0x004C;
const char32_t LATIN_CAPITAL_LETTER_M = 0x004D;
const char32_t LATIN_CAPITAL_LETTER_N = 0x004E;
const char32_t LATIN_CAPITAL_LETTER_O = 0x004F;
const char32_t LATIN_CAPITAL_LETTER_P = 0x0050;
const char32_t LATIN_CAPITAL_LETTER_Q = 0x0051;
const char32_t LATIN_CAPITAL_LETTER_R = 0x0052;
const char32_t LATIN_CAPITAL_LETTER_S = 0x0053;
const char32_t LATIN_CAPITAL_LETTER_T = 0x0054;
const char32_t LATIN_CAPITAL_LETTER_U = 0x0055;
const char32_t LATIN_CAPITAL_LETTER_V = 0x0056;
const char32_t LATIN_CAPITAL_LETTER_W = 0x0057;
const char32_t LATIN_CAPITAL_LETTER_X = 0x0058;
const char32_t LATIN_CAPITAL_LETTER_Y = 0x0059;
const char32_t LATIN_CAPITAL_LETTER_Z = 0x005A;
const char32_t LEFT_SQUARE_BRACKET = 0x005B;
const char32_t REVERSE_SOLIDUS = 0x005C;
const char32_t RIGHT_SQUARE_BRACKET = 0x005D;
const char32_t CIRCUMFLEX_ACCENT = 0x005E;
const char32_t LOW_LINE = 0x005F;
const char32_t GRAVE_ACCENT = 0x0060;
const char32_t LATIN_SMALL_LETTER_A = 0x0061;
const char32_t LATIN_SMALL_LETTER_B = 0x0062;
const char32_t LATIN_SMALL_LETTER_C = 0x0063;
const char32_t LATIN_SMALL_LETTER_D = 0x0064;
const char32_t LATIN_SMALL_LETTER_E = 0x0065;
const char32_t LATIN_SMALL_LETTER_F = 0x0066;
const char32_t LATIN_SMALL_LETTER_G = 0x0067;
const char32_t LATIN_SMALL_LETTER_H = 0x0068;
const char32_t LATIN_SMALL_LETTER_I = 0x0069;
const char32_t LATIN_SMALL_LETTER_J = 0x006A;
const char32_t LATIN_SMALL_LETTER_K = 0x006B;
const char32_t LATIN_SMALL_LETTER_L = 0x006C;
const char32_t LATIN_SMALL_LETTER_M = 0x006D;
const char32_t LATIN_SMALL_LETTER_N = 0x006E;
const char32_t LATIN_SMALL_LETTER_O = 0x006F;
const char32_t LATIN_SMALL_LETTER_P = 0x0070;
const char32_t LATIN_SMALL_LETTER_Q = 0x0071;
const char32_t LATIN_SMALL_LETTER_R = 0x0072;
const char32_t LATIN_SMALL_LETTER_S = 0x0073;
const char32_t LATIN_SMALL_LETTER_T = 0x0074;
const char32_t LATIN_SMALL_LETTER_U = 0x0075;
const char32_t LATIN_SMALL_LETTER_V = 0x0076;
const char32_t LATIN_SMALL_LETTER_W = 0x0077;
const char32_t LATIN_SMALL_LETTER_X = 0x0078;
const char32_t LATIN_SMALL_LETTER_Y = 0x0079;
const char32_t LATIN_SMALL_LETTER_Z = 0x007A;
const char32_t LEFT_CURLY_BRACKET = 0x007B;
const char32_t VERTICAL_LINE = 0x007C;
const char32_t RIGHT_CURLY_BRACKET = 0x007D;
const char32_t TILDE = 0x007E;
const char32_t DELETE = 0x007F;
const char32_t PADDING_CHARACTER = 0x0080;
const char32_t HIGH_OCTET_PRESET = 0x0081;
const char32_t BREAK_PERMITTED_HERE = 0x0082;
const char32_t NO_BREAK_HERE = 0x0083;
const char32_t INDEX = 0x0084;
const char32_t NEXT_LINE = 0x0085;
const char32_t START_OF_SELECTED_AREA = 0x0086;
const char32_t END_OF_SELECTED_AREA = 0x0087;
const char32_t CHARACTER_TABULATION_SET = 0x0088;
const char32_t CHARACTER_TABULATION_WITH_JUSTIFICATION = 0x0089;
const char32_t LINE_TABULATION_SET = 0x008A;
const char32_t PARTIAL_LINE_FORWARD = 0x008B;
const char32_t PARTIAL_LINE_BACKWARD = 0x008C;
const char32_t REVERSE_LINE_FEED = 0x008D;
const char32_t SINGLE_SHIFT_TWO = 0x008E;
const char32_t SINGLE_SHIFT_THREE = 0x008F;
const char32_t DEVICE_CONTROL_STRING = 0x0090;
const char32_t PRIVATE_USE_ONE = 0x0091;
const char32_t PRIVATE_USE_TWO = 0x0092;
const char32_t SET_TRANSMIT_STATE = 0x0093;
const char32_t CANCEL_CHARACTER = 0x0094;
const char32_t MESSAGE_WAITING = 0x0095;
const char32_t START_OF_GUARDED_AREA = 0x0096;
const char32_t END_OF_GUARDED_AREA = 0x0097;
const char32_t START_OF_STRING = 0x0098;
const char32_t SINGLE_GRAPHIC_CHARACTER_INTRODUCER = 0x0099;
const char32_t SINGLE_CHARACTER_INTRODUCER = 0x009A;
const char32_t CONTROL_SEQUENCE_INTRODUCER = 0x009B;
const char32_t STRING_TERMINATOR = 0x009C;
const char32_t OPERATING_SYSTEM_COMMAND = 0x009D;
const char32_t PRIVACY_MESSAGE = 0x009E;
const char32_t APPLICATION_PROGRAM_COMMAND = 0x009F;
const char32_t NO_BREAK_SPACE = 0x00A0;
const char32_t INVERTED_EXCLAMATION_MARK = 0x00A1;
const char32_t CENT_SIGN = 0x00A2;
const char32_t POUND_SIGN = 0x00A3;
const char32_t CURRENCY_SIGN = 0x00A4;
const char32_t YEN_SIGN = 0x00A5;
const char32_t BROKEN_BAR = 0x00A6;
const char32_t SECTION_SIGN = 0x00A7;
const char32_t DIAERESIS = 0x00A8;
const char32_t COPYRIGHT_SIGN = 0x00A9;
const char32_t FEMININE_ORDINAL_INDICATOR = 0x00AA;
const char32_t LEFT_POINTING_DOUBLE_ANGLE_QUOTATION_MARK = 0x00AB;
const char32_t NOT_SIGN = 0x00AC;
const char32_t SOFT_HYPHEN = 0x00AD;
const char32_t REGISTERED_SIGN = 0x00AE;
const char32_t MACRON = 0x00AF;
const char32_t DEGREE_SIGN = 0x00B0;
const char32_t PLUS_MINUS_SIGN = 0x00B1;
const char32_t SUPERSCRIPT_TWO = 0x00B2;
const char32_t SUPERSCRIPT_THREE = 0x00B3;
const char32_t ACUTE_ACCENT = 0x00B4;
const char32_t MICRO_SIGN = 0x00B5;
const char32_t PILCROW_SIGN = 0x00B6;
const char32_t MIDDLE_DOT = 0x00B7;
const char32_t CEDILLA = 0x00B8;
const char32_t SUPERSCRIPT_ONE = 0x00B9;
const char32_t MASCULINE_ORDINAL_INDICATOR = 0x00BA;
const char32_t RIGHT_POINTING_DOUBLE_ANGLE_QUOTATION_MARK = 0x00BB;
const char32_t VULGAR_FRACTION_ONE_QUARTER = 0x00BC;
const char32_t VULGAR_FRACTION_ONE_HALF = 0x00BD;
const char32_t VULGAR_FRACTION_THREE_QUARTERS = 0x00BE;
const char32_t INVERTED_QUESTION_MARK = 0x00BF;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_GRAVE = 0x00C0;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_ACUTE = 0x00C1;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX = 0x00C2;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_TILDE = 0x00C3;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_DIAERESIS = 0x00C4;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_RING_ABOVE = 0x00C5;
const char32_t LATIN_CAPITAL_LETTER_AE = 0x00C6;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_CEDILLA = 0x00C7;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_GRAVE = 0x00C8;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_ACUTE = 0x00C9;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX = 0x00CA;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_DIAERESIS = 0x00CB;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_GRAVE = 0x00CC;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_ACUTE = 0x00CD;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_CIRCUMFLEX = 0x00CE;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_DIAERESIS = 0x00CF;
const char32_t LATIN_CAPITAL_LETTER_ETH = 0x00D0;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_TILDE = 0x00D1;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_GRAVE = 0x00D2;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_ACUTE = 0x00D3;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX = 0x00D4;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_TILDE = 0x00D5;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_DIAERESIS = 0x00D6;
const char32_t MULTIPLICATION_SIGN = 0x00D7;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_STROKE = 0x00D8;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_GRAVE = 0x00D9;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_ACUTE = 0x00DA;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_CIRCUMFLEX = 0x00DB;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS = 0x00DC;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_ACUTE = 0x00DD;
const char32_t LATIN_CAPITAL_LETTER_THORN = 0x00DE;
const char32_t LATIN_SMALL_LETTER_SHARP_S = 0x00DF;
const char32_t LATIN_SMALL_LETTER_A_WITH_GRAVE = 0x00E0;
const char32_t LATIN_SMALL_LETTER_A_WITH_ACUTE = 0x00E1;
const char32_t LATIN_SMALL_LETTER_A_WITH_CIRCUMFLEX = 0x00E2;
const char32_t LATIN_SMALL_LETTER_A_WITH_TILDE = 0x00E3;
const char32_t LATIN_SMALL_LETTER_A_WITH_DIAERESIS = 0x00E4;
const char32_t LATIN_SMALL_LETTER_A_WITH_RING_ABOVE = 0x00E5;
const char32_t LATIN_SMALL_LETTER_AE = 0x00E6;
const char32_t LATIN_SMALL_LETTER_C_WITH_CEDILLA = 0x00E7;
const char32_t LATIN_SMALL_LETTER_E_WITH_GRAVE = 0x00E8;
const char32_t LATIN_SMALL_LETTER_E_WITH_ACUTE = 0x00E9;
const char32_t LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX = 0x00EA;
const char32_t LATIN_SMALL_LETTER_E_WITH_DIAERESIS = 0x00EB;
const char32_t LATIN_SMALL_LETTER_I_WITH_GRAVE = 0x00EC;
const char32_t LATIN_SMALL_LETTER_I_WITH_ACUTE = 0x00ED;
const char32_t LATIN_SMALL_LETTER_I_WITH_CIRCUMFLEX = 0x00EE;
const char32_t LATIN_SMALL_LETTER_I_WITH_DIAERESIS = 0x00EF;
const char32_t LATIN_SMALL_LETTER_ETH = 0x00F0;
const char32_t LATIN_SMALL_LETTER_N_WITH_TILDE = 0x00F1;
const char32_t LATIN_SMALL_LETTER_O_WITH_GRAVE = 0x00F2;
const char32_t LATIN_SMALL_LETTER_O_WITH_ACUTE = 0x00F3;
const char32_t LATIN_SMALL_LETTER_O_WITH_CIRCUMFLEX = 0x00F4;
const char32_t LATIN_SMALL_LETTER_O_WITH_TILDE = 0x00F5;
const char32_t LATIN_SMALL_LETTER_O_WITH_DIAERESIS = 0x00F6;
const char32_t DIVISION_SIGN = 0x00F7;
const char32_t LATIN_SMALL_LETTER_O_WITH_STROKE = 0x00F8;
const char32_t LATIN_SMALL_LETTER_U_WITH_GRAVE = 0x00F9;
const char32_t LATIN_SMALL_LETTER_U_WITH_ACUTE = 0x00FA;
const char32_t LATIN_SMALL_LETTER_U_WITH_CIRCUMFLEX = 0x00FB;
const char32_t LATIN_SMALL_LETTER_U_WITH_DIAERESIS = 0x00FC;
const char32_t LATIN_SMALL_LETTER_Y_WITH_ACUTE = 0x00FD;
const char32_t LATIN_SMALL_LETTER_THORN = 0x00FE;
const char32_t LATIN_SMALL_LETTER_Y_WITH_DIAERESIS = 0x00FF;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_MACRON = 0x0100;
const char32_t LATIN_SMALL_LETTER_A_WITH_MACRON = 0x0101;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_BREVE = 0x0102;
const char32_t LATIN_SMALL_LETTER_A_WITH_BREVE = 0x0103;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_OGONEK = 0x0104;
const char32_t LATIN_SMALL_LETTER_A_WITH_OGONEK = 0x0105;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_ACUTE = 0x0106;
const char32_t LATIN_SMALL_LETTER_C_WITH_ACUTE = 0x0107;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_CIRCUMFLEX = 0x0108;
const char32_t LATIN_SMALL_LETTER_C_WITH_CIRCUMFLEX = 0x0109;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_DOT_ABOVE = 0x010A;
const char32_t LATIN_SMALL_LETTER_C_WITH_DOT_ABOVE = 0x010B;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_CARON = 0x010C;
const char32_t LATIN_SMALL_LETTER_C_WITH_CARON = 0x010D;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_CARON = 0x010E;
const char32_t LATIN_SMALL_LETTER_D_WITH_CARON = 0x010F;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_STROKE = 0x0110;
const char32_t LATIN_SMALL_LETTER_D_WITH_STROKE = 0x0111;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_MACRON = 0x0112;
const char32_t LATIN_SMALL_LETTER_E_WITH_MACRON = 0x0113;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_BREVE = 0x0114;
const char32_t LATIN_SMALL_LETTER_E_WITH_BREVE = 0x0115;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_DOT_ABOVE = 0x0116;
const char32_t LATIN_SMALL_LETTER_E_WITH_DOT_ABOVE = 0x0117;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_OGONEK = 0x0118;
const char32_t LATIN_SMALL_LETTER_E_WITH_OGONEK = 0x0119;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CARON = 0x011A;
const char32_t LATIN_SMALL_LETTER_E_WITH_CARON = 0x011B;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_CIRCUMFLEX = 0x011C;
const char32_t LATIN_SMALL_LETTER_G_WITH_CIRCUMFLEX = 0x011D;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_BREVE = 0x011E;
const char32_t LATIN_SMALL_LETTER_G_WITH_BREVE = 0x011F;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_DOT_ABOVE = 0x0120;
const char32_t LATIN_SMALL_LETTER_G_WITH_DOT_ABOVE = 0x0121;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_CEDILLA = 0x0122;
const char32_t LATIN_SMALL_LETTER_G_WITH_CEDILLA = 0x0123;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_CIRCUMFLEX = 0x0124;
const char32_t LATIN_SMALL_LETTER_H_WITH_CIRCUMFLEX = 0x0125;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_STROKE = 0x0126;
const char32_t LATIN_SMALL_LETTER_H_WITH_STROKE = 0x0127;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_TILDE = 0x0128;
const char32_t LATIN_SMALL_LETTER_I_WITH_TILDE = 0x0129;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_MACRON = 0x012A;
const char32_t LATIN_SMALL_LETTER_I_WITH_MACRON = 0x012B;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_BREVE = 0x012C;
const char32_t LATIN_SMALL_LETTER_I_WITH_BREVE = 0x012D;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_OGONEK = 0x012E;
const char32_t LATIN_SMALL_LETTER_I_WITH_OGONEK = 0x012F;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_DOT_ABOVE = 0x0130;
const char32_t LATIN_SMALL_LETTER_DOTLESS_I = 0x0131;
const char32_t LATIN_CAPITAL_LIGATURE_IJ = 0x0132;
const char32_t LATIN_SMALL_LIGATURE_IJ = 0x0133;
const char32_t LATIN_CAPITAL_LETTER_J_WITH_CIRCUMFLEX = 0x0134;
const char32_t LATIN_SMALL_LETTER_J_WITH_CIRCUMFLEX = 0x0135;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_CEDILLA = 0x0136;
const char32_t LATIN_SMALL_LETTER_K_WITH_CEDILLA = 0x0137;
const char32_t LATIN_SMALL_LETTER_KRA = 0x0138;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_ACUTE = 0x0139;
const char32_t LATIN_SMALL_LETTER_L_WITH_ACUTE = 0x013A;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_CEDILLA = 0x013B;
const char32_t LATIN_SMALL_LETTER_L_WITH_CEDILLA = 0x013C;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_CARON = 0x013D;
const char32_t LATIN_SMALL_LETTER_L_WITH_CARON = 0x013E;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_MIDDLE_DOT = 0x013F;
const char32_t LATIN_SMALL_LETTER_L_WITH_MIDDLE_DOT = 0x0140;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_STROKE = 0x0141;
const char32_t LATIN_SMALL_LETTER_L_WITH_STROKE = 0x0142;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_ACUTE = 0x0143;
const char32_t LATIN_SMALL_LETTER_N_WITH_ACUTE = 0x0144;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_CEDILLA = 0x0145;
const char32_t LATIN_SMALL_LETTER_N_WITH_CEDILLA = 0x0146;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_CARON = 0x0147;
const char32_t LATIN_SMALL_LETTER_N_WITH_CARON = 0x0148;
const char32_t LATIN_SMALL_LETTER_N_PRECEDED_BY_APOSTROPHE = 0x0149;
const char32_t LATIN_CAPITAL_LETTER_ENG = 0x014A;
const char32_t LATIN_SMALL_LETTER_ENG = 0x014B;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_MACRON = 0x014C;
const char32_t LATIN_SMALL_LETTER_O_WITH_MACRON = 0x014D;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_BREVE = 0x014E;
const char32_t LATIN_SMALL_LETTER_O_WITH_BREVE = 0x014F;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_DOUBLE_ACUTE = 0x0150;
const char32_t LATIN_SMALL_LETTER_O_WITH_DOUBLE_ACUTE = 0x0151;
const char32_t LATIN_CAPITAL_LIGATURE_OE = 0x0152;
const char32_t LATIN_SMALL_LIGATURE_OE = 0x0153;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_ACUTE = 0x0154;
const char32_t LATIN_SMALL_LETTER_R_WITH_ACUTE = 0x0155;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_CEDILLA = 0x0156;
const char32_t LATIN_SMALL_LETTER_R_WITH_CEDILLA = 0x0157;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_CARON = 0x0158;
const char32_t LATIN_SMALL_LETTER_R_WITH_CARON = 0x0159;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_ACUTE = 0x015A;
const char32_t LATIN_SMALL_LETTER_S_WITH_ACUTE = 0x015B;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_CIRCUMFLEX = 0x015C;
const char32_t LATIN_SMALL_LETTER_S_WITH_CIRCUMFLEX = 0x015D;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_CEDILLA = 0x015E;
const char32_t LATIN_SMALL_LETTER_S_WITH_CEDILLA = 0x015F;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_CARON = 0x0160;
const char32_t LATIN_SMALL_LETTER_S_WITH_CARON = 0x0161;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_CEDILLA = 0x0162;
const char32_t LATIN_SMALL_LETTER_T_WITH_CEDILLA = 0x0163;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_CARON = 0x0164;
const char32_t LATIN_SMALL_LETTER_T_WITH_CARON = 0x0165;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_STROKE = 0x0166;
const char32_t LATIN_SMALL_LETTER_T_WITH_STROKE = 0x0167;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_TILDE = 0x0168;
const char32_t LATIN_SMALL_LETTER_U_WITH_TILDE = 0x0169;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_MACRON = 0x016A;
const char32_t LATIN_SMALL_LETTER_U_WITH_MACRON = 0x016B;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_BREVE = 0x016C;
const char32_t LATIN_SMALL_LETTER_U_WITH_BREVE = 0x016D;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_RING_ABOVE = 0x016E;
const char32_t LATIN_SMALL_LETTER_U_WITH_RING_ABOVE = 0x016F;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DOUBLE_ACUTE = 0x0170;
const char32_t LATIN_SMALL_LETTER_U_WITH_DOUBLE_ACUTE = 0x0171;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_OGONEK = 0x0172;
const char32_t LATIN_SMALL_LETTER_U_WITH_OGONEK = 0x0173;
const char32_t LATIN_CAPITAL_LETTER_W_WITH_CIRCUMFLEX = 0x0174;
const char32_t LATIN_SMALL_LETTER_W_WITH_CIRCUMFLEX = 0x0175;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_CIRCUMFLEX = 0x0176;
const char32_t LATIN_SMALL_LETTER_Y_WITH_CIRCUMFLEX = 0x0177;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_DIAERESIS = 0x0178;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_ACUTE = 0x0179;
const char32_t LATIN_SMALL_LETTER_Z_WITH_ACUTE = 0x017A;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_DOT_ABOVE = 0x017B;
const char32_t LATIN_SMALL_LETTER_Z_WITH_DOT_ABOVE = 0x017C;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_CARON = 0x017D;
const char32_t LATIN_SMALL_LETTER_Z_WITH_CARON = 0x017E;
const char32_t LATIN_SMALL_LETTER_LONG_S = 0x017F;
const char32_t LATIN_SMALL_LETTER_B_WITH_STROKE = 0x0180;
const char32_t LATIN_CAPITAL_LETTER_B_WITH_HOOK = 0x0181;
const char32_t LATIN_CAPITAL_LETTER_B_WITH_TOPBAR = 0x0182;
const char32_t LATIN_SMALL_LETTER_B_WITH_TOPBAR = 0x0183;
const char32_t LATIN_CAPITAL_LETTER_TONE_SIX = 0x0184;
const char32_t LATIN_SMALL_LETTER_TONE_SIX = 0x0185;
const char32_t LATIN_CAPITAL_LETTER_OPEN_O = 0x0186;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_HOOK = 0x0187;
const char32_t LATIN_SMALL_LETTER_C_WITH_HOOK = 0x0188;
const char32_t LATIN_CAPITAL_LETTER_AFRICAN_D = 0x0189;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_HOOK = 0x018A;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_TOPBAR = 0x018B;
const char32_t LATIN_SMALL_LETTER_D_WITH_TOPBAR = 0x018C;
const char32_t LATIN_SMALL_LETTER_TURNED_DELTA = 0x018D;
const char32_t LATIN_CAPITAL_LETTER_REVERSED_E = 0x018E;
const char32_t LATIN_CAPITAL_LETTER_SCHWA = 0x018F;
const char32_t LATIN_CAPITAL_LETTER_OPEN_E = 0x0190;
const char32_t LATIN_CAPITAL_LETTER_F_WITH_HOOK = 0x0191;
const char32_t LATIN_SMALL_LETTER_F_WITH_HOOK = 0x0192;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_HOOK = 0x0193;
const char32_t LATIN_CAPITAL_LETTER_GAMMA = 0x0194;
const char32_t LATIN_SMALL_LETTER_HV = 0x0195;
const char32_t LATIN_CAPITAL_LETTER_IOTA = 0x0196;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_STROKE = 0x0197;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_HOOK = 0x0198;
const char32_t LATIN_SMALL_LETTER_K_WITH_HOOK = 0x0199;
const char32_t LATIN_SMALL_LETTER_L_WITH_BAR = 0x019A;
const char32_t LATIN_SMALL_LETTER_LAMBDA_WITH_STROKE = 0x019B;
const char32_t LATIN_CAPITAL_LETTER_TURNED_M = 0x019C;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_LEFT_HOOK = 0x019D;
const char32_t LATIN_SMALL_LETTER_N_WITH_LONG_RIGHT_LEG = 0x019E;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_MIDDLE_TILDE = 0x019F;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_HORN = 0x01A0;
const char32_t LATIN_SMALL_LETTER_O_WITH_HORN = 0x01A1;
const char32_t LATIN_CAPITAL_LETTER_OI = 0x01A2;
const char32_t LATIN_SMALL_LETTER_OI = 0x01A3;
const char32_t LATIN_CAPITAL_LETTER_P_WITH_HOOK = 0x01A4;
const char32_t LATIN_SMALL_LETTER_P_WITH_HOOK = 0x01A5;
const char32_t LATIN_LETTER_YR = 0x01A6;
const char32_t LATIN_CAPITAL_LETTER_TONE_TWO = 0x01A7;
const char32_t LATIN_SMALL_LETTER_TONE_TWO = 0x01A8;
const char32_t LATIN_CAPITAL_LETTER_ESH = 0x01A9;
const char32_t LATIN_LETTER_REVERSED_ESH_LOOP = 0x01AA;
const char32_t LATIN_SMALL_LETTER_T_WITH_PALATAL_HOOK = 0x01AB;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_HOOK = 0x01AC;
const char32_t LATIN_SMALL_LETTER_T_WITH_HOOK = 0x01AD;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_RETROFLEX_HOOK = 0x01AE;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_HORN = 0x01AF;
const char32_t LATIN_SMALL_LETTER_U_WITH_HORN = 0x01B0;
const char32_t LATIN_CAPITAL_LETTER_UPSILON = 0x01B1;
const char32_t LATIN_CAPITAL_LETTER_V_WITH_HOOK = 0x01B2;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_HOOK = 0x01B3;
const char32_t LATIN_SMALL_LETTER_Y_WITH_HOOK = 0x01B4;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_STROKE = 0x01B5;
const char32_t LATIN_SMALL_LETTER_Z_WITH_STROKE = 0x01B6;
const char32_t LATIN_CAPITAL_LETTER_EZH = 0x01B7;
const char32_t LATIN_CAPITAL_LETTER_EZH_REVERSED = 0x01B8;
const char32_t LATIN_SMALL_LETTER_EZH_REVERSED = 0x01B9;
const char32_t LATIN_SMALL_LETTER_EZH_WITH_TAIL = 0x01BA;
const char32_t LATIN_LETTER_TWO_WITH_STROKE = 0x01BB;
const char32_t LATIN_CAPITAL_LETTER_TONE_FIVE = 0x01BC;
const char32_t LATIN_SMALL_LETTER_TONE_FIVE = 0x01BD;
const char32_t LATIN_LETTER_INVERTED_GLOTTAL_STOP_WITH_STROKE = 0x01BE;
const char32_t LATIN_LETTER_WYNN = 0x01BF;
const char32_t LATIN_LETTER_DENTAL_CLICK = 0x01C0;
const char32_t LATIN_LETTER_LATERAL_CLICK = 0x01C1;
const char32_t LATIN_LETTER_ALVEOLAR_CLICK = 0x01C2;
const char32_t LATIN_LETTER_RETROFLEX_CLICK = 0x01C3;
const char32_t LATIN_CAPITAL_LETTER_DZ_WITH_CARON = 0x01C4;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_SMALL_LETTER_Z_WITH_CARON = 0x01C5;
const char32_t LATIN_SMALL_LETTER_DZ_WITH_CARON = 0x01C6;
const char32_t LATIN_CAPITAL_LETTER_LJ = 0x01C7;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_SMALL_LETTER_J = 0x01C8;
const char32_t LATIN_SMALL_LETTER_LJ = 0x01C9;
const char32_t LATIN_CAPITAL_LETTER_NJ = 0x01CA;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_SMALL_LETTER_J = 0x01CB;
const char32_t LATIN_SMALL_LETTER_NJ = 0x01CC;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_CARON = 0x01CD;
const char32_t LATIN_SMALL_LETTER_A_WITH_CARON = 0x01CE;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_CARON = 0x01CF;
const char32_t LATIN_SMALL_LETTER_I_WITH_CARON = 0x01D0;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_CARON = 0x01D1;
const char32_t LATIN_SMALL_LETTER_O_WITH_CARON = 0x01D2;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_CARON = 0x01D3;
const char32_t LATIN_SMALL_LETTER_U_WITH_CARON = 0x01D4;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS_AND_MACRON = 0x01D5;
const char32_t LATIN_SMALL_LETTER_U_WITH_DIAERESIS_AND_MACRON = 0x01D6;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS_AND_ACUTE = 0x01D7;
const char32_t LATIN_SMALL_LETTER_U_WITH_DIAERESIS_AND_ACUTE = 0x01D8;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS_AND_CARON = 0x01D9;
const char32_t LATIN_SMALL_LETTER_U_WITH_DIAERESIS_AND_CARON = 0x01DA;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS_AND_GRAVE = 0x01DB;
const char32_t LATIN_SMALL_LETTER_U_WITH_DIAERESIS_AND_GRAVE = 0x01DC;
const char32_t LATIN_SMALL_LETTER_TURNED_E = 0x01DD;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_DIAERESIS_AND_MACRON = 0x01DE;
const char32_t LATIN_SMALL_LETTER_A_WITH_DIAERESIS_AND_MACRON = 0x01DF;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_DOT_ABOVE_AND_MACRON = 0x01E0;
const char32_t LATIN_SMALL_LETTER_A_WITH_DOT_ABOVE_AND_MACRON = 0x01E1;
const char32_t LATIN_CAPITAL_LETTER_AE_WITH_MACRON = 0x01E2;
const char32_t LATIN_SMALL_LETTER_AE_WITH_MACRON = 0x01E3;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_STROKE = 0x01E4;
const char32_t LATIN_SMALL_LETTER_G_WITH_STROKE = 0x01E5;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_CARON = 0x01E6;
const char32_t LATIN_SMALL_LETTER_G_WITH_CARON = 0x01E7;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_CARON = 0x01E8;
const char32_t LATIN_SMALL_LETTER_K_WITH_CARON = 0x01E9;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_OGONEK = 0x01EA;
const char32_t LATIN_SMALL_LETTER_O_WITH_OGONEK = 0x01EB;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_OGONEK_AND_MACRON = 0x01EC;
const char32_t LATIN_SMALL_LETTER_O_WITH_OGONEK_AND_MACRON = 0x01ED;
const char32_t LATIN_CAPITAL_LETTER_EZH_WITH_CARON = 0x01EE;
const char32_t LATIN_SMALL_LETTER_EZH_WITH_CARON = 0x01EF;
const char32_t LATIN_SMALL_LETTER_J_WITH_CARON = 0x01F0;
const char32_t LATIN_CAPITAL_LETTER_DZ = 0x01F1;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_SMALL_LETTER_Z = 0x01F2;
const char32_t LATIN_SMALL_LETTER_DZ = 0x01F3;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_ACUTE = 0x01F4;
const char32_t LATIN_SMALL_LETTER_G_WITH_ACUTE = 0x01F5;
const char32_t LATIN_CAPITAL_LETTER_HWAIR = 0x01F6;
const char32_t LATIN_CAPITAL_LETTER_WYNN = 0x01F7;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_GRAVE = 0x01F8;
const char32_t LATIN_SMALL_LETTER_N_WITH_GRAVE = 0x01F9;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_RING_ABOVE_AND_ACUTE = 0x01FA;
const char32_t LATIN_SMALL_LETTER_A_WITH_RING_ABOVE_AND_ACUTE = 0x01FB;
const char32_t LATIN_CAPITAL_LETTER_AE_WITH_ACUTE = 0x01FC;
const char32_t LATIN_SMALL_LETTER_AE_WITH_ACUTE = 0x01FD;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_STROKE_AND_ACUTE = 0x01FE;
const char32_t LATIN_SMALL_LETTER_O_WITH_STROKE_AND_ACUTE = 0x01FF;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_DOUBLE_GRAVE = 0x0200;
const char32_t LATIN_SMALL_LETTER_A_WITH_DOUBLE_GRAVE = 0x0201;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_INVERTED_BREVE = 0x0202;
const char32_t LATIN_SMALL_LETTER_A_WITH_INVERTED_BREVE = 0x0203;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_DOUBLE_GRAVE = 0x0204;
const char32_t LATIN_SMALL_LETTER_E_WITH_DOUBLE_GRAVE = 0x0205;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_INVERTED_BREVE = 0x0206;
const char32_t LATIN_SMALL_LETTER_E_WITH_INVERTED_BREVE = 0x0207;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_DOUBLE_GRAVE = 0x0208;
const char32_t LATIN_SMALL_LETTER_I_WITH_DOUBLE_GRAVE = 0x0209;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_INVERTED_BREVE = 0x020A;
const char32_t LATIN_SMALL_LETTER_I_WITH_INVERTED_BREVE = 0x020B;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_DOUBLE_GRAVE = 0x020C;
const char32_t LATIN_SMALL_LETTER_O_WITH_DOUBLE_GRAVE = 0x020D;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_INVERTED_BREVE = 0x020E;
const char32_t LATIN_SMALL_LETTER_O_WITH_INVERTED_BREVE = 0x020F;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_DOUBLE_GRAVE = 0x0210;
const char32_t LATIN_SMALL_LETTER_R_WITH_DOUBLE_GRAVE = 0x0211;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_INVERTED_BREVE = 0x0212;
const char32_t LATIN_SMALL_LETTER_R_WITH_INVERTED_BREVE = 0x0213;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DOUBLE_GRAVE = 0x0214;
const char32_t LATIN_SMALL_LETTER_U_WITH_DOUBLE_GRAVE = 0x0215;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_INVERTED_BREVE = 0x0216;
const char32_t LATIN_SMALL_LETTER_U_WITH_INVERTED_BREVE = 0x0217;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_COMMA_BELOW = 0x0218;
const char32_t LATIN_SMALL_LETTER_S_WITH_COMMA_BELOW = 0x0219;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_COMMA_BELOW = 0x021A;
const char32_t LATIN_SMALL_LETTER_T_WITH_COMMA_BELOW = 0x021B;
const char32_t LATIN_CAPITAL_LETTER_YOGH = 0x021C;
const char32_t LATIN_SMALL_LETTER_YOGH = 0x021D;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_CARON = 0x021E;
const char32_t LATIN_SMALL_LETTER_H_WITH_CARON = 0x021F;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_LONG_RIGHT_LEG = 0x0220;
const char32_t LATIN_SMALL_LETTER_D_WITH_CURL = 0x0221;
const char32_t LATIN_CAPITAL_LETTER_OU = 0x0222;
const char32_t LATIN_SMALL_LETTER_OU = 0x0223;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_HOOK = 0x0224;
const char32_t LATIN_SMALL_LETTER_Z_WITH_HOOK = 0x0225;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_DOT_ABOVE = 0x0226;
const char32_t LATIN_SMALL_LETTER_A_WITH_DOT_ABOVE = 0x0227;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CEDILLA = 0x0228;
const char32_t LATIN_SMALL_LETTER_E_WITH_CEDILLA = 0x0229;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_DIAERESIS_AND_MACRON = 0x022A;
const char32_t LATIN_SMALL_LETTER_O_WITH_DIAERESIS_AND_MACRON = 0x022B;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_TILDE_AND_MACRON = 0x022C;
const char32_t LATIN_SMALL_LETTER_O_WITH_TILDE_AND_MACRON = 0x022D;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_DOT_ABOVE = 0x022E;
const char32_t LATIN_SMALL_LETTER_O_WITH_DOT_ABOVE = 0x022F;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_DOT_ABOVE_AND_MACRON = 0x0230;
const char32_t LATIN_SMALL_LETTER_O_WITH_DOT_ABOVE_AND_MACRON = 0x0231;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_MACRON = 0x0232;
const char32_t LATIN_SMALL_LETTER_Y_WITH_MACRON = 0x0233;
const char32_t LATIN_SMALL_LETTER_L_WITH_CURL = 0x0234;
const char32_t LATIN_SMALL_LETTER_N_WITH_CURL = 0x0235;
const char32_t LATIN_SMALL_LETTER_T_WITH_CURL = 0x0236;
const char32_t LATIN_SMALL_LETTER_DOTLESS_J = 0x0237;
const char32_t LATIN_SMALL_LETTER_DB_DIGRAPH = 0x0238;
const char32_t LATIN_SMALL_LETTER_QP_DIGRAPH = 0x0239;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_STROKE = 0x023A;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_STROKE = 0x023B;
const char32_t LATIN_SMALL_LETTER_C_WITH_STROKE = 0x023C;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_BAR = 0x023D;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_DIAGONAL_STROKE = 0x023E;
const char32_t LATIN_SMALL_LETTER_S_WITH_SWASH_TAIL = 0x023F;
const char32_t LATIN_SMALL_LETTER_Z_WITH_SWASH_TAIL = 0x0240;
const char32_t LATIN_CAPITAL_LETTER_GLOTTAL_STOP = 0x0241;
const char32_t LATIN_SMALL_LETTER_GLOTTAL_STOP = 0x0242;
const char32_t LATIN_CAPITAL_LETTER_B_WITH_STROKE = 0x0243;
const char32_t LATIN_CAPITAL_LETTER_U_BAR = 0x0244;
const char32_t LATIN_CAPITAL_LETTER_TURNED_V = 0x0245;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_STROKE = 0x0246;
const char32_t LATIN_SMALL_LETTER_E_WITH_STROKE = 0x0247;
const char32_t LATIN_CAPITAL_LETTER_J_WITH_STROKE = 0x0248;
const char32_t LATIN_SMALL_LETTER_J_WITH_STROKE = 0x0249;
const char32_t LATIN_CAPITAL_LETTER_SMALL_Q_WITH_HOOK_TAIL = 0x024A;
const char32_t LATIN_SMALL_LETTER_Q_WITH_HOOK_TAIL = 0x024B;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_STROKE = 0x024C;
const char32_t LATIN_SMALL_LETTER_R_WITH_STROKE = 0x024D;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_STROKE = 0x024E;
const char32_t LATIN_SMALL_LETTER_Y_WITH_STROKE = 0x024F;
const char32_t LATIN_SMALL_LETTER_TURNED_A = 0x0250;
const char32_t LATIN_SMALL_LETTER_ALPHA = 0x0251;
const char32_t LATIN_SMALL_LETTER_TURNED_ALPHA = 0x0252;
const char32_t LATIN_SMALL_LETTER_B_WITH_HOOK = 0x0253;
const char32_t LATIN_SMALL_LETTER_OPEN_O = 0x0254;
const char32_t LATIN_SMALL_LETTER_C_WITH_CURL = 0x0255;
const char32_t LATIN_SMALL_LETTER_D_WITH_TAIL = 0x0256;
const char32_t LATIN_SMALL_LETTER_D_WITH_HOOK = 0x0257;
const char32_t LATIN_SMALL_LETTER_REVERSED_E = 0x0258;
const char32_t LATIN_SMALL_LETTER_SCHWA = 0x0259;
const char32_t LATIN_SMALL_LETTER_SCHWA_WITH_HOOK = 0x025A;
const char32_t LATIN_SMALL_LETTER_OPEN_E = 0x025B;
const char32_t LATIN_SMALL_LETTER_REVERSED_OPEN_E = 0x025C;
const char32_t LATIN_SMALL_LETTER_REVERSED_OPEN_E_WITH_HOOK = 0x025D;
const char32_t LATIN_SMALL_LETTER_CLOSED_REVERSED_OPEN_E = 0x025E;
const char32_t LATIN_SMALL_LETTER_DOTLESS_J_WITH_STROKE = 0x025F;
const char32_t LATIN_SMALL_LETTER_G_WITH_HOOK = 0x0260;
const char32_t LATIN_SMALL_LETTER_SCRIPT_G = 0x0261;
const char32_t LATIN_LETTER_SMALL_CAPITAL_G = 0x0262;
const char32_t LATIN_SMALL_LETTER_GAMMA = 0x0263;
const char32_t LATIN_SMALL_LETTER_RAMS_HORN = 0x0264;
const char32_t LATIN_SMALL_LETTER_TURNED_H = 0x0265;
const char32_t LATIN_SMALL_LETTER_H_WITH_HOOK = 0x0266;
const char32_t LATIN_SMALL_LETTER_HENG_WITH_HOOK = 0x0267;
const char32_t LATIN_SMALL_LETTER_I_WITH_STROKE = 0x0268;
const char32_t LATIN_SMALL_LETTER_IOTA = 0x0269;
const char32_t LATIN_LETTER_SMALL_CAPITAL_I = 0x026A;
const char32_t LATIN_SMALL_LETTER_L_WITH_MIDDLE_TILDE = 0x026B;
const char32_t LATIN_SMALL_LETTER_L_WITH_BELT = 0x026C;
const char32_t LATIN_SMALL_LETTER_L_WITH_RETROFLEX_HOOK = 0x026D;
const char32_t LATIN_SMALL_LETTER_LEZH = 0x026E;
const char32_t LATIN_SMALL_LETTER_TURNED_M = 0x026F;
const char32_t LATIN_SMALL_LETTER_TURNED_M_WITH_LONG_LEG = 0x0270;
const char32_t LATIN_SMALL_LETTER_M_WITH_HOOK = 0x0271;
const char32_t LATIN_SMALL_LETTER_N_WITH_LEFT_HOOK = 0x0272;
const char32_t LATIN_SMALL_LETTER_N_WITH_RETROFLEX_HOOK = 0x0273;
const char32_t LATIN_LETTER_SMALL_CAPITAL_N = 0x0274;
const char32_t LATIN_SMALL_LETTER_BARRED_O = 0x0275;
const char32_t LATIN_LETTER_SMALL_CAPITAL_OE = 0x0276;
const char32_t LATIN_SMALL_LETTER_CLOSED_OMEGA = 0x0277;
const char32_t LATIN_SMALL_LETTER_PHI = 0x0278;
const char32_t LATIN_SMALL_LETTER_TURNED_R = 0x0279;
const char32_t LATIN_SMALL_LETTER_TURNED_R_WITH_LONG_LEG = 0x027A;
const char32_t LATIN_SMALL_LETTER_TURNED_R_WITH_HOOK = 0x027B;
const char32_t LATIN_SMALL_LETTER_R_WITH_LONG_LEG = 0x027C;
const char32_t LATIN_SMALL_LETTER_R_WITH_TAIL = 0x027D;
const char32_t LATIN_SMALL_LETTER_R_WITH_FISHHOOK = 0x027E;
const char32_t LATIN_SMALL_LETTER_REVERSED_R_WITH_FISHHOOK = 0x027F;
const char32_t LATIN_LETTER_SMALL_CAPITAL_R = 0x0280;
const char32_t LATIN_LETTER_SMALL_CAPITAL_INVERTED_R = 0x0281;
const char32_t LATIN_SMALL_LETTER_S_WITH_HOOK = 0x0282;
const char32_t LATIN_SMALL_LETTER_ESH = 0x0283;
const char32_t LATIN_SMALL_LETTER_DOTLESS_J_WITH_STROKE_AND_HOOK = 0x0284;
const char32_t LATIN_SMALL_LETTER_SQUAT_REVERSED_ESH = 0x0285;
const char32_t LATIN_SMALL_LETTER_ESH_WITH_CURL = 0x0286;
const char32_t LATIN_SMALL_LETTER_TURNED_T = 0x0287;
const char32_t LATIN_SMALL_LETTER_T_WITH_RETROFLEX_HOOK = 0x0288;
const char32_t LATIN_SMALL_LETTER_U_BAR = 0x0289;
const char32_t LATIN_SMALL_LETTER_UPSILON = 0x028A;
const char32_t LATIN_SMALL_LETTER_V_WITH_HOOK = 0x028B;
const char32_t LATIN_SMALL_LETTER_TURNED_V = 0x028C;
const char32_t LATIN_SMALL_LETTER_TURNED_W = 0x028D;
const char32_t LATIN_SMALL_LETTER_TURNED_Y = 0x028E;
const char32_t LATIN_LETTER_SMALL_CAPITAL_Y = 0x028F;
const char32_t LATIN_SMALL_LETTER_Z_WITH_RETROFLEX_HOOK = 0x0290;
const char32_t LATIN_SMALL_LETTER_Z_WITH_CURL = 0x0291;
const char32_t LATIN_SMALL_LETTER_EZH = 0x0292;
const char32_t LATIN_SMALL_LETTER_EZH_WITH_CURL = 0x0293;
const char32_t LATIN_LETTER_GLOTTAL_STOP = 0x0294;
const char32_t LATIN_LETTER_PHARYNGEAL_VOICED_FRICATIVE = 0x0295;
const char32_t LATIN_LETTER_INVERTED_GLOTTAL_STOP = 0x0296;
const char32_t LATIN_LETTER_STRETCHED_C = 0x0297;
const char32_t LATIN_LETTER_BILABIAL_CLICK = 0x0298;
const char32_t LATIN_LETTER_SMALL_CAPITAL_B = 0x0299;
const char32_t LATIN_SMALL_LETTER_CLOSED_OPEN_E = 0x029A;
const char32_t LATIN_LETTER_SMALL_CAPITAL_G_WITH_HOOK = 0x029B;
const char32_t LATIN_LETTER_SMALL_CAPITAL_H = 0x029C;
const char32_t LATIN_SMALL_LETTER_J_WITH_CROSSED_TAIL = 0x029D;
const char32_t LATIN_SMALL_LETTER_TURNED_K = 0x029E;
const char32_t LATIN_LETTER_SMALL_CAPITAL_L = 0x029F;
const char32_t LATIN_SMALL_LETTER_Q_WITH_HOOK = 0x02A0;
const char32_t LATIN_LETTER_GLOTTAL_STOP_WITH_STROKE = 0x02A1;
const char32_t LATIN_LETTER_REVERSED_GLOTTAL_STOP_WITH_STROKE = 0x02A2;
const char32_t LATIN_SMALL_LETTER_DZ_DIGRAPH = 0x02A3;
const char32_t LATIN_SMALL_LETTER_DEZH_DIGRAPH = 0x02A4;
const char32_t LATIN_SMALL_LETTER_DZ_DIGRAPH_WITH_CURL = 0x02A5;
const char32_t LATIN_SMALL_LETTER_TS_DIGRAPH = 0x02A6;
const char32_t LATIN_SMALL_LETTER_TESH_DIGRAPH = 0x02A7;
const char32_t LATIN_SMALL_LETTER_TC_DIGRAPH_WITH_CURL = 0x02A8;
const char32_t LATIN_SMALL_LETTER_FENG_DIGRAPH = 0x02A9;
const char32_t LATIN_SMALL_LETTER_LS_DIGRAPH = 0x02AA;
const char32_t LATIN_SMALL_LETTER_LZ_DIGRAPH = 0x02AB;
const char32_t LATIN_LETTER_BILABIAL_PERCUSSIVE = 0x02AC;
const char32_t LATIN_LETTER_BIDENTAL_PERCUSSIVE = 0x02AD;
const char32_t LATIN_SMALL_LETTER_TURNED_H_WITH_FISHHOOK = 0x02AE;
const char32_t LATIN_SMALL_LETTER_TURNED_H_WITH_FISHHOOK_AND_TAIL = 0x02AF;
const char32_t MODIFIER_LETTER_SMALL_H = 0x02B0;
const char32_t MODIFIER_LETTER_SMALL_H_WITH_HOOK = 0x02B1;
const char32_t MODIFIER_LETTER_SMALL_J = 0x02B2;
const char32_t MODIFIER_LETTER_SMALL_R = 0x02B3;
const char32_t MODIFIER_LETTER_SMALL_TURNED_R = 0x02B4;
const char32_t MODIFIER_LETTER_SMALL_TURNED_R_WITH_HOOK = 0x02B5;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_INVERTED_R = 0x02B6;
const char32_t MODIFIER_LETTER_SMALL_W = 0x02B7;
const char32_t MODIFIER_LETTER_SMALL_Y = 0x02B8;
const char32_t MODIFIER_LETTER_PRIME = 0x02B9;
const char32_t MODIFIER_LETTER_DOUBLE_PRIME = 0x02BA;
const char32_t MODIFIER_LETTER_TURNED_COMMA = 0x02BB;
const char32_t MODIFIER_LETTER_APOSTROPHE = 0x02BC;
const char32_t MODIFIER_LETTER_REVERSED_COMMA = 0x02BD;
const char32_t MODIFIER_LETTER_RIGHT_HALF_RING = 0x02BE;
const char32_t MODIFIER_LETTER_LEFT_HALF_RING = 0x02BF;
const char32_t MODIFIER_LETTER_GLOTTAL_STOP = 0x02C0;
const char32_t MODIFIER_LETTER_REVERSED_GLOTTAL_STOP = 0x02C1;
const char32_t MODIFIER_LETTER_LEFT_ARROWHEAD = 0x02C2;
const char32_t MODIFIER_LETTER_RIGHT_ARROWHEAD = 0x02C3;
const char32_t MODIFIER_LETTER_UP_ARROWHEAD = 0x02C4;
const char32_t MODIFIER_LETTER_DOWN_ARROWHEAD = 0x02C5;
const char32_t MODIFIER_LETTER_CIRCUMFLEX_ACCENT = 0x02C6;
const char32_t CARON = 0x02C7;
const char32_t MODIFIER_LETTER_VERTICAL_LINE = 0x02C8;
const char32_t MODIFIER_LETTER_MACRON = 0x02C9;
const char32_t MODIFIER_LETTER_ACUTE_ACCENT = 0x02CA;
const char32_t MODIFIER_LETTER_GRAVE_ACCENT = 0x02CB;
const char32_t MODIFIER_LETTER_LOW_VERTICAL_LINE = 0x02CC;
const char32_t MODIFIER_LETTER_LOW_MACRON = 0x02CD;
const char32_t MODIFIER_LETTER_LOW_GRAVE_ACCENT = 0x02CE;
const char32_t MODIFIER_LETTER_LOW_ACUTE_ACCENT = 0x02CF;
const char32_t MODIFIER_LETTER_TRIANGULAR_COLON = 0x02D0;
const char32_t MODIFIER_LETTER_HALF_TRIANGULAR_COLON = 0x02D1;
const char32_t MODIFIER_LETTER_CENTRED_RIGHT_HALF_RING = 0x02D2;
const char32_t MODIFIER_LETTER_CENTRED_LEFT_HALF_RING = 0x02D3;
const char32_t MODIFIER_LETTER_UP_TACK = 0x02D4;
const char32_t MODIFIER_LETTER_DOWN_TACK = 0x02D5;
const char32_t MODIFIER_LETTER_PLUS_SIGN = 0x02D6;
const char32_t MODIFIER_LETTER_MINUS_SIGN = 0x02D7;
const char32_t BREVE = 0x02D8;
const char32_t DOT_ABOVE = 0x02D9;
const char32_t RING_ABOVE = 0x02DA;
const char32_t OGONEK = 0x02DB;
const char32_t SMALL_TILDE = 0x02DC;
const char32_t DOUBLE_ACUTE_ACCENT = 0x02DD;
const char32_t MODIFIER_LETTER_RHOTIC_HOOK = 0x02DE;
const char32_t MODIFIER_LETTER_CROSS_ACCENT = 0x02DF;
const char32_t MODIFIER_LETTER_SMALL_GAMMA = 0x02E0;
const char32_t MODIFIER_LETTER_SMALL_L = 0x02E1;
const char32_t MODIFIER_LETTER_SMALL_S = 0x02E2;
const char32_t MODIFIER_LETTER_SMALL_X = 0x02E3;
const char32_t MODIFIER_LETTER_SMALL_REVERSED_GLOTTAL_STOP = 0x02E4;
const char32_t MODIFIER_LETTER_EXTRA_HIGH_TONE_BAR = 0x02E5;
const char32_t MODIFIER_LETTER_HIGH_TONE_BAR = 0x02E6;
const char32_t MODIFIER_LETTER_MID_TONE_BAR = 0x02E7;
const char32_t MODIFIER_LETTER_LOW_TONE_BAR = 0x02E8;
const char32_t MODIFIER_LETTER_EXTRA_LOW_TONE_BAR = 0x02E9;
const char32_t MODIFIER_LETTER_YIN_DEPARTING_TONE_MARK = 0x02EA;
const char32_t MODIFIER_LETTER_YANG_DEPARTING_TONE_MARK = 0x02EB;
const char32_t MODIFIER_LETTER_VOICING = 0x02EC;
const char32_t MODIFIER_LETTER_UNASPIRATED = 0x02ED;
const char32_t MODIFIER_LETTER_DOUBLE_APOSTROPHE = 0x02EE;
const char32_t MODIFIER_LETTER_LOW_DOWN_ARROWHEAD = 0x02EF;
const char32_t MODIFIER_LETTER_LOW_UP_ARROWHEAD = 0x02F0;
const char32_t MODIFIER_LETTER_LOW_LEFT_ARROWHEAD = 0x02F1;
const char32_t MODIFIER_LETTER_LOW_RIGHT_ARROWHEAD = 0x02F2;
const char32_t MODIFIER_LETTER_LOW_RING = 0x02F3;
const char32_t MODIFIER_LETTER_MIDDLE_GRAVE_ACCENT = 0x02F4;
const char32_t MODIFIER_LETTER_MIDDLE_DOUBLE_GRAVE_ACCENT = 0x02F5;
const char32_t MODIFIER_LETTER_MIDDLE_DOUBLE_ACUTE_ACCENT = 0x02F6;
const char32_t MODIFIER_LETTER_LOW_TILDE = 0x02F7;
const char32_t MODIFIER_LETTER_RAISED_COLON = 0x02F8;
const char32_t MODIFIER_LETTER_BEGIN_HIGH_TONE = 0x02F9;
const char32_t MODIFIER_LETTER_END_HIGH_TONE = 0x02FA;
const char32_t MODIFIER_LETTER_BEGIN_LOW_TONE = 0x02FB;
const char32_t MODIFIER_LETTER_END_LOW_TONE = 0x02FC;
const char32_t MODIFIER_LETTER_SHELF = 0x02FD;
const char32_t MODIFIER_LETTER_OPEN_SHELF = 0x02FE;
const char32_t MODIFIER_LETTER_LOW_LEFT_ARROW = 0x02FF;
const char32_t COMBINING_GRAVE_ACCENT = 0x0300;
const char32_t COMBINING_ACUTE_ACCENT = 0x0301;
const char32_t COMBINING_CIRCUMFLEX_ACCENT = 0x0302;
const char32_t COMBINING_TILDE = 0x0303;
const char32_t COMBINING_MACRON = 0x0304;
const char32_t COMBINING_OVERLINE = 0x0305;
const char32_t COMBINING_BREVE = 0x0306;
const char32_t COMBINING_DOT_ABOVE = 0x0307;
const char32_t COMBINING_DIAERESIS = 0x0308;
const char32_t COMBINING_HOOK_ABOVE = 0x0309;
const char32_t COMBINING_RING_ABOVE = 0x030A;
const char32_t COMBINING_DOUBLE_ACUTE_ACCENT = 0x030B;
const char32_t COMBINING_CARON = 0x030C;
const char32_t COMBINING_VERTICAL_LINE_ABOVE = 0x030D;
const char32_t COMBINING_DOUBLE_VERTICAL_LINE_ABOVE = 0x030E;
const char32_t COMBINING_DOUBLE_GRAVE_ACCENT = 0x030F;
const char32_t COMBINING_CANDRABINDU = 0x0310;
const char32_t COMBINING_INVERTED_BREVE = 0x0311;
const char32_t COMBINING_TURNED_COMMA_ABOVE = 0x0312;
const char32_t COMBINING_COMMA_ABOVE = 0x0313;
const char32_t COMBINING_REVERSED_COMMA_ABOVE = 0x0314;
const char32_t COMBINING_COMMA_ABOVE_RIGHT = 0x0315;
const char32_t COMBINING_GRAVE_ACCENT_BELOW = 0x0316;
const char32_t COMBINING_ACUTE_ACCENT_BELOW = 0x0317;
const char32_t COMBINING_LEFT_TACK_BELOW = 0x0318;
const char32_t COMBINING_RIGHT_TACK_BELOW = 0x0319;
const char32_t COMBINING_LEFT_ANGLE_ABOVE = 0x031A;
const char32_t COMBINING_HORN = 0x031B;
const char32_t COMBINING_LEFT_HALF_RING_BELOW = 0x031C;
const char32_t COMBINING_UP_TACK_BELOW = 0x031D;
const char32_t COMBINING_DOWN_TACK_BELOW = 0x031E;
const char32_t COMBINING_PLUS_SIGN_BELOW = 0x031F;
const char32_t COMBINING_MINUS_SIGN_BELOW = 0x0320;
const char32_t COMBINING_PALATALIZED_HOOK_BELOW = 0x0321;
const char32_t COMBINING_RETROFLEX_HOOK_BELOW = 0x0322;
const char32_t COMBINING_DOT_BELOW = 0x0323;
const char32_t COMBINING_DIAERESIS_BELOW = 0x0324;
const char32_t COMBINING_RING_BELOW = 0x0325;
const char32_t COMBINING_COMMA_BELOW = 0x0326;
const char32_t COMBINING_CEDILLA = 0x0327;
const char32_t COMBINING_OGONEK = 0x0328;
const char32_t COMBINING_VERTICAL_LINE_BELOW = 0x0329;
const char32_t COMBINING_BRIDGE_BELOW = 0x032A;
const char32_t COMBINING_INVERTED_DOUBLE_ARCH_BELOW = 0x032B;
const char32_t COMBINING_CARON_BELOW = 0x032C;
const char32_t COMBINING_CIRCUMFLEX_ACCENT_BELOW = 0x032D;
const char32_t COMBINING_BREVE_BELOW = 0x032E;
const char32_t COMBINING_INVERTED_BREVE_BELOW = 0x032F;
const char32_t COMBINING_TILDE_BELOW = 0x0330;
const char32_t COMBINING_MACRON_BELOW = 0x0331;
const char32_t COMBINING_LOW_LINE = 0x0332;
const char32_t COMBINING_DOUBLE_LOW_LINE = 0x0333;
const char32_t COMBINING_TILDE_OVERLAY = 0x0334;
const char32_t COMBINING_SHORT_STROKE_OVERLAY = 0x0335;
const char32_t COMBINING_LONG_STROKE_OVERLAY = 0x0336;
const char32_t COMBINING_SHORT_SOLIDUS_OVERLAY = 0x0337;
const char32_t COMBINING_LONG_SOLIDUS_OVERLAY = 0x0338;
const char32_t COMBINING_RIGHT_HALF_RING_BELOW = 0x0339;
const char32_t COMBINING_INVERTED_BRIDGE_BELOW = 0x033A;
const char32_t COMBINING_SQUARE_BELOW = 0x033B;
const char32_t COMBINING_SEAGULL_BELOW = 0x033C;
const char32_t COMBINING_X_ABOVE = 0x033D;
const char32_t COMBINING_VERTICAL_TILDE = 0x033E;
const char32_t COMBINING_DOUBLE_OVERLINE = 0x033F;
const char32_t COMBINING_GRAVE_TONE_MARK = 0x0340;
const char32_t COMBINING_ACUTE_TONE_MARK = 0x0341;
const char32_t COMBINING_GREEK_PERISPOMENI = 0x0342;
const char32_t COMBINING_GREEK_KORONIS = 0x0343;
const char32_t COMBINING_GREEK_DIALYTIKA_TONOS = 0x0344;
const char32_t COMBINING_GREEK_YPOGEGRAMMENI = 0x0345;
const char32_t COMBINING_BRIDGE_ABOVE = 0x0346;
const char32_t COMBINING_EQUALS_SIGN_BELOW = 0x0347;
const char32_t COMBINING_DOUBLE_VERTICAL_LINE_BELOW = 0x0348;
const char32_t COMBINING_LEFT_ANGLE_BELOW = 0x0349;
const char32_t COMBINING_NOT_TILDE_ABOVE = 0x034A;
const char32_t COMBINING_HOMOTHETIC_ABOVE = 0x034B;
const char32_t COMBINING_ALMOST_EQUAL_TO_ABOVE = 0x034C;
const char32_t COMBINING_LEFT_RIGHT_ARROW_BELOW = 0x034D;
const char32_t COMBINING_UPWARDS_ARROW_BELOW = 0x034E;
const char32_t COMBINING_GRAPHEME_JOINER = 0x034F;
const char32_t COMBINING_RIGHT_ARROWHEAD_ABOVE = 0x0350;
const char32_t COMBINING_LEFT_HALF_RING_ABOVE = 0x0351;
const char32_t COMBINING_FERMATA = 0x0352;
const char32_t COMBINING_X_BELOW = 0x0353;
const char32_t COMBINING_LEFT_ARROWHEAD_BELOW = 0x0354;
const char32_t COMBINING_RIGHT_ARROWHEAD_BELOW = 0x0355;
const char32_t COMBINING_RIGHT_ARROWHEAD_AND_UP_ARROWHEAD_BELOW = 0x0356;
const char32_t COMBINING_RIGHT_HALF_RING_ABOVE = 0x0357;
const char32_t COMBINING_DOT_ABOVE_RIGHT = 0x0358;
const char32_t COMBINING_ASTERISK_BELOW = 0x0359;
const char32_t COMBINING_DOUBLE_RING_BELOW = 0x035A;
const char32_t COMBINING_ZIGZAG_ABOVE = 0x035B;
const char32_t COMBINING_DOUBLE_BREVE_BELOW = 0x035C;
const char32_t COMBINING_DOUBLE_BREVE = 0x035D;
const char32_t COMBINING_DOUBLE_MACRON = 0x035E;
const char32_t COMBINING_DOUBLE_MACRON_BELOW = 0x035F;
const char32_t COMBINING_DOUBLE_TILDE = 0x0360;
const char32_t COMBINING_DOUBLE_INVERTED_BREVE = 0x0361;
const char32_t COMBINING_DOUBLE_RIGHTWARDS_ARROW_BELOW = 0x0362;
const char32_t COMBINING_LATIN_SMALL_LETTER_A = 0x0363;
const char32_t COMBINING_LATIN_SMALL_LETTER_E = 0x0364;
const char32_t COMBINING_LATIN_SMALL_LETTER_I = 0x0365;
const char32_t COMBINING_LATIN_SMALL_LETTER_O = 0x0366;
const char32_t COMBINING_LATIN_SMALL_LETTER_U = 0x0367;
const char32_t COMBINING_LATIN_SMALL_LETTER_C = 0x0368;
const char32_t COMBINING_LATIN_SMALL_LETTER_D = 0x0369;
const char32_t COMBINING_LATIN_SMALL_LETTER_H = 0x036A;
const char32_t COMBINING_LATIN_SMALL_LETTER_M = 0x036B;
const char32_t COMBINING_LATIN_SMALL_LETTER_R = 0x036C;
const char32_t COMBINING_LATIN_SMALL_LETTER_T = 0x036D;
const char32_t COMBINING_LATIN_SMALL_LETTER_V = 0x036E;
const char32_t COMBINING_LATIN_SMALL_LETTER_X = 0x036F;
const char32_t GREEK_CAPITAL_LETTER_HETA = 0x0370;
const char32_t GREEK_SMALL_LETTER_HETA = 0x0371;
const char32_t GREEK_CAPITAL_LETTER_ARCHAIC_SAMPI = 0x0372;
const char32_t GREEK_SMALL_LETTER_ARCHAIC_SAMPI = 0x0373;
const char32_t GREEK_NUMERAL_SIGN = 0x0374;
const char32_t GREEK_LOWER_NUMERAL_SIGN = 0x0375;
const char32_t GREEK_CAPITAL_LETTER_PAMPHYLIAN_DIGAMMA = 0x0376;
const char32_t GREEK_SMALL_LETTER_PAMPHYLIAN_DIGAMMA = 0x0377;
const char32_t GREEK_YPOGEGRAMMENI = 0x037A;
const char32_t GREEK_SMALL_REVERSED_LUNATE_SIGMA_SYMBOL = 0x037B;
const char32_t GREEK_SMALL_DOTTED_LUNATE_SIGMA_SYMBOL = 0x037C;
const char32_t GREEK_SMALL_REVERSED_DOTTED_LUNATE_SIGMA_SYMBOL = 0x037D;
const char32_t GREEK_QUESTION_MARK = 0x037E;
const char32_t GREEK_CAPITAL_LETTER_YOT = 0x037F;
const char32_t GREEK_TONOS = 0x0384;
const char32_t GREEK_DIALYTIKA_TONOS = 0x0385;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_TONOS = 0x0386;
const char32_t GREEK_ANO_TELEIA = 0x0387;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_TONOS = 0x0388;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_TONOS = 0x0389;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_TONOS = 0x038A;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_TONOS = 0x038C;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_TONOS = 0x038E;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_TONOS = 0x038F;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DIALYTIKA_AND_TONOS = 0x0390;
const char32_t GREEK_CAPITAL_LETTER_ALPHA = 0x0391;
const char32_t GREEK_CAPITAL_LETTER_BETA = 0x0392;
const char32_t GREEK_CAPITAL_LETTER_GAMMA = 0x0393;
const char32_t GREEK_CAPITAL_LETTER_DELTA = 0x0394;
const char32_t GREEK_CAPITAL_LETTER_EPSILON = 0x0395;
const char32_t GREEK_CAPITAL_LETTER_ZETA = 0x0396;
const char32_t GREEK_CAPITAL_LETTER_ETA = 0x0397;
const char32_t GREEK_CAPITAL_LETTER_THETA = 0x0398;
const char32_t GREEK_CAPITAL_LETTER_IOTA = 0x0399;
const char32_t GREEK_CAPITAL_LETTER_KAPPA = 0x039A;
const char32_t GREEK_CAPITAL_LETTER_LAMDA = 0x039B;
const char32_t GREEK_CAPITAL_LETTER_MU = 0x039C;
const char32_t GREEK_CAPITAL_LETTER_NU = 0x039D;
const char32_t GREEK_CAPITAL_LETTER_XI = 0x039E;
const char32_t GREEK_CAPITAL_LETTER_OMICRON = 0x039F;
const char32_t GREEK_CAPITAL_LETTER_PI = 0x03A0;
const char32_t GREEK_CAPITAL_LETTER_RHO = 0x03A1;
const char32_t GREEK_CAPITAL_LETTER_SIGMA = 0x03A3;
const char32_t GREEK_CAPITAL_LETTER_TAU = 0x03A4;
const char32_t GREEK_CAPITAL_LETTER_UPSILON = 0x03A5;
const char32_t GREEK_CAPITAL_LETTER_PHI = 0x03A6;
const char32_t GREEK_CAPITAL_LETTER_CHI = 0x03A7;
const char32_t GREEK_CAPITAL_LETTER_PSI = 0x03A8;
const char32_t GREEK_CAPITAL_LETTER_OMEGA = 0x03A9;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_DIALYTIKA = 0x03AA;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_DIALYTIKA = 0x03AB;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_TONOS = 0x03AC;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_TONOS = 0x03AD;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_TONOS = 0x03AE;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_TONOS = 0x03AF;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DIALYTIKA_AND_TONOS = 0x03B0;
const char32_t GREEK_SMALL_LETTER_ALPHA = 0x03B1;
const char32_t GREEK_SMALL_LETTER_BETA = 0x03B2;
const char32_t GREEK_SMALL_LETTER_GAMMA = 0x03B3;
const char32_t GREEK_SMALL_LETTER_DELTA = 0x03B4;
const char32_t GREEK_SMALL_LETTER_EPSILON = 0x03B5;
const char32_t GREEK_SMALL_LETTER_ZETA = 0x03B6;
const char32_t GREEK_SMALL_LETTER_ETA = 0x03B7;
const char32_t GREEK_SMALL_LETTER_THETA = 0x03B8;
const char32_t GREEK_SMALL_LETTER_IOTA = 0x03B9;
const char32_t GREEK_SMALL_LETTER_KAPPA = 0x03BA;
const char32_t GREEK_SMALL_LETTER_LAMDA = 0x03BB;
const char32_t GREEK_SMALL_LETTER_MU = 0x03BC;
const char32_t GREEK_SMALL_LETTER_NU = 0x03BD;
const char32_t GREEK_SMALL_LETTER_XI = 0x03BE;
const char32_t GREEK_SMALL_LETTER_OMICRON = 0x03BF;
const char32_t GREEK_SMALL_LETTER_PI = 0x03C0;
const char32_t GREEK_SMALL_LETTER_RHO = 0x03C1;
const char32_t GREEK_SMALL_LETTER_FINAL_SIGMA = 0x03C2;
const char32_t GREEK_SMALL_LETTER_SIGMA = 0x03C3;
const char32_t GREEK_SMALL_LETTER_TAU = 0x03C4;
const char32_t GREEK_SMALL_LETTER_UPSILON = 0x03C5;
const char32_t GREEK_SMALL_LETTER_PHI = 0x03C6;
const char32_t GREEK_SMALL_LETTER_CHI = 0x03C7;
const char32_t GREEK_SMALL_LETTER_PSI = 0x03C8;
const char32_t GREEK_SMALL_LETTER_OMEGA = 0x03C9;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DIALYTIKA = 0x03CA;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DIALYTIKA = 0x03CB;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_TONOS = 0x03CC;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_TONOS = 0x03CD;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_TONOS = 0x03CE;
const char32_t GREEK_CAPITAL_KAI_SYMBOL = 0x03CF;
const char32_t GREEK_BETA_SYMBOL = 0x03D0;
const char32_t GREEK_THETA_SYMBOL = 0x03D1;
const char32_t GREEK_UPSILON_WITH_HOOK_SYMBOL = 0x03D2;
const char32_t GREEK_UPSILON_WITH_ACUTE_AND_HOOK_SYMBOL = 0x03D3;
const char32_t GREEK_UPSILON_WITH_DIAERESIS_AND_HOOK_SYMBOL = 0x03D4;
const char32_t GREEK_PHI_SYMBOL = 0x03D5;
const char32_t GREEK_PI_SYMBOL = 0x03D6;
const char32_t GREEK_KAI_SYMBOL = 0x03D7;
const char32_t GREEK_LETTER_ARCHAIC_KOPPA = 0x03D8;
const char32_t GREEK_SMALL_LETTER_ARCHAIC_KOPPA = 0x03D9;
const char32_t GREEK_LETTER_STIGMA = 0x03DA;
const char32_t GREEK_SMALL_LETTER_STIGMA = 0x03DB;
const char32_t GREEK_LETTER_DIGAMMA = 0x03DC;
const char32_t GREEK_SMALL_LETTER_DIGAMMA = 0x03DD;
const char32_t GREEK_LETTER_KOPPA = 0x03DE;
const char32_t GREEK_SMALL_LETTER_KOPPA = 0x03DF;
const char32_t GREEK_LETTER_SAMPI = 0x03E0;
const char32_t GREEK_SMALL_LETTER_SAMPI = 0x03E1;
const char32_t COPTIC_CAPITAL_LETTER_SHEI = 0x03E2;
const char32_t COPTIC_SMALL_LETTER_SHEI = 0x03E3;
const char32_t COPTIC_CAPITAL_LETTER_FEI = 0x03E4;
const char32_t COPTIC_SMALL_LETTER_FEI = 0x03E5;
const char32_t COPTIC_CAPITAL_LETTER_KHEI = 0x03E6;
const char32_t COPTIC_SMALL_LETTER_KHEI = 0x03E7;
const char32_t COPTIC_CAPITAL_LETTER_HORI = 0x03E8;
const char32_t COPTIC_SMALL_LETTER_HORI = 0x03E9;
const char32_t COPTIC_CAPITAL_LETTER_GANGIA = 0x03EA;
const char32_t COPTIC_SMALL_LETTER_GANGIA = 0x03EB;
const char32_t COPTIC_CAPITAL_LETTER_SHIMA = 0x03EC;
const char32_t COPTIC_SMALL_LETTER_SHIMA = 0x03ED;
const char32_t COPTIC_CAPITAL_LETTER_DEI = 0x03EE;
const char32_t COPTIC_SMALL_LETTER_DEI = 0x03EF;
const char32_t GREEK_KAPPA_SYMBOL = 0x03F0;
const char32_t GREEK_RHO_SYMBOL = 0x03F1;
const char32_t GREEK_LUNATE_SIGMA_SYMBOL = 0x03F2;
const char32_t GREEK_LETTER_YOT = 0x03F3;
const char32_t GREEK_CAPITAL_THETA_SYMBOL = 0x03F4;
const char32_t GREEK_LUNATE_EPSILON_SYMBOL = 0x03F5;
const char32_t GREEK_REVERSED_LUNATE_EPSILON_SYMBOL = 0x03F6;
const char32_t GREEK_CAPITAL_LETTER_SHO = 0x03F7;
const char32_t GREEK_SMALL_LETTER_SHO = 0x03F8;
const char32_t GREEK_CAPITAL_LUNATE_SIGMA_SYMBOL = 0x03F9;
const char32_t GREEK_CAPITAL_LETTER_SAN = 0x03FA;
const char32_t GREEK_SMALL_LETTER_SAN = 0x03FB;
const char32_t GREEK_RHO_WITH_STROKE_SYMBOL = 0x03FC;
const char32_t GREEK_CAPITAL_REVERSED_LUNATE_SIGMA_SYMBOL = 0x03FD;
const char32_t GREEK_CAPITAL_DOTTED_LUNATE_SIGMA_SYMBOL = 0x03FE;
const char32_t GREEK_CAPITAL_REVERSED_DOTTED_LUNATE_SIGMA_SYMBOL = 0x03FF;
const char32_t CYRILLIC_CAPITAL_LETTER_IE_WITH_GRAVE = 0x0400;
const char32_t CYRILLIC_CAPITAL_LETTER_IO = 0x0401;
const char32_t CYRILLIC_CAPITAL_LETTER_DJE = 0x0402;
const char32_t CYRILLIC_CAPITAL_LETTER_GJE = 0x0403;
const char32_t CYRILLIC_CAPITAL_LETTER_UKRAINIAN_IE = 0x0404;
const char32_t CYRILLIC_CAPITAL_LETTER_DZE = 0x0405;
const char32_t CYRILLIC_CAPITAL_LETTER_BYELORUSSIAN_UKRAINIAN_I = 0x0406;
const char32_t CYRILLIC_CAPITAL_LETTER_YI = 0x0407;
const char32_t CYRILLIC_CAPITAL_LETTER_JE = 0x0408;
const char32_t CYRILLIC_CAPITAL_LETTER_LJE = 0x0409;
const char32_t CYRILLIC_CAPITAL_LETTER_NJE = 0x040A;
const char32_t CYRILLIC_CAPITAL_LETTER_TSHE = 0x040B;
const char32_t CYRILLIC_CAPITAL_LETTER_KJE = 0x040C;
const char32_t CYRILLIC_CAPITAL_LETTER_I_WITH_GRAVE = 0x040D;
const char32_t CYRILLIC_CAPITAL_LETTER_SHORT_U = 0x040E;
const char32_t CYRILLIC_CAPITAL_LETTER_DZHE = 0x040F;
const char32_t CYRILLIC_CAPITAL_LETTER_A = 0x0410;
const char32_t CYRILLIC_CAPITAL_LETTER_BE = 0x0411;
const char32_t CYRILLIC_CAPITAL_LETTER_VE = 0x0412;
const char32_t CYRILLIC_CAPITAL_LETTER_GHE = 0x0413;
const char32_t CYRILLIC_CAPITAL_LETTER_DE = 0x0414;
const char32_t CYRILLIC_CAPITAL_LETTER_IE = 0x0415;
const char32_t CYRILLIC_CAPITAL_LETTER_ZHE = 0x0416;
const char32_t CYRILLIC_CAPITAL_LETTER_ZE = 0x0417;
const char32_t CYRILLIC_CAPITAL_LETTER_I = 0x0418;
const char32_t CYRILLIC_CAPITAL_LETTER_SHORT_I = 0x0419;
const char32_t CYRILLIC_CAPITAL_LETTER_KA = 0x041A;
const char32_t CYRILLIC_CAPITAL_LETTER_EL = 0x041B;
const char32_t CYRILLIC_CAPITAL_LETTER_EM = 0x041C;
const char32_t CYRILLIC_CAPITAL_LETTER_EN = 0x041D;
const char32_t CYRILLIC_CAPITAL_LETTER_O = 0x041E;
const char32_t CYRILLIC_CAPITAL_LETTER_PE = 0x041F;
const char32_t CYRILLIC_CAPITAL_LETTER_ER = 0x0420;
const char32_t CYRILLIC_CAPITAL_LETTER_ES = 0x0421;
const char32_t CYRILLIC_CAPITAL_LETTER_TE = 0x0422;
const char32_t CYRILLIC_CAPITAL_LETTER_U = 0x0423;
const char32_t CYRILLIC_CAPITAL_LETTER_EF = 0x0424;
const char32_t CYRILLIC_CAPITAL_LETTER_HA = 0x0425;
const char32_t CYRILLIC_CAPITAL_LETTER_TSE = 0x0426;
const char32_t CYRILLIC_CAPITAL_LETTER_CHE = 0x0427;
const char32_t CYRILLIC_CAPITAL_LETTER_SHA = 0x0428;
const char32_t CYRILLIC_CAPITAL_LETTER_SHCHA = 0x0429;
const char32_t CYRILLIC_CAPITAL_LETTER_HARD_SIGN = 0x042A;
const char32_t CYRILLIC_CAPITAL_LETTER_YERU = 0x042B;
const char32_t CYRILLIC_CAPITAL_LETTER_SOFT_SIGN = 0x042C;
const char32_t CYRILLIC_CAPITAL_LETTER_E = 0x042D;
const char32_t CYRILLIC_CAPITAL_LETTER_YU = 0x042E;
const char32_t CYRILLIC_CAPITAL_LETTER_YA = 0x042F;
const char32_t CYRILLIC_SMALL_LETTER_A = 0x0430;
const char32_t CYRILLIC_SMALL_LETTER_BE = 0x0431;
const char32_t CYRILLIC_SMALL_LETTER_VE = 0x0432;
const char32_t CYRILLIC_SMALL_LETTER_GHE = 0x0433;
const char32_t CYRILLIC_SMALL_LETTER_DE = 0x0434;
const char32_t CYRILLIC_SMALL_LETTER_IE = 0x0435;
const char32_t CYRILLIC_SMALL_LETTER_ZHE = 0x0436;
const char32_t CYRILLIC_SMALL_LETTER_ZE = 0x0437;
const char32_t CYRILLIC_SMALL_LETTER_I = 0x0438;
const char32_t CYRILLIC_SMALL_LETTER_SHORT_I = 0x0439;
const char32_t CYRILLIC_SMALL_LETTER_KA = 0x043A;
const char32_t CYRILLIC_SMALL_LETTER_EL = 0x043B;
const char32_t CYRILLIC_SMALL_LETTER_EM = 0x043C;
const char32_t CYRILLIC_SMALL_LETTER_EN = 0x043D;
const char32_t CYRILLIC_SMALL_LETTER_O = 0x043E;
const char32_t CYRILLIC_SMALL_LETTER_PE = 0x043F;
const char32_t CYRILLIC_SMALL_LETTER_ER = 0x0440;
const char32_t CYRILLIC_SMALL_LETTER_ES = 0x0441;
const char32_t CYRILLIC_SMALL_LETTER_TE = 0x0442;
const char32_t CYRILLIC_SMALL_LETTER_U = 0x0443;
const char32_t CYRILLIC_SMALL_LETTER_EF = 0x0444;
const char32_t CYRILLIC_SMALL_LETTER_HA = 0x0445;
const char32_t CYRILLIC_SMALL_LETTER_TSE = 0x0446;
const char32_t CYRILLIC_SMALL_LETTER_CHE = 0x0447;
const char32_t CYRILLIC_SMALL_LETTER_SHA = 0x0448;
const char32_t CYRILLIC_SMALL_LETTER_SHCHA = 0x0449;
const char32_t CYRILLIC_SMALL_LETTER_HARD_SIGN = 0x044A;
const char32_t CYRILLIC_SMALL_LETTER_YERU = 0x044B;
const char32_t CYRILLIC_SMALL_LETTER_SOFT_SIGN = 0x044C;
const char32_t CYRILLIC_SMALL_LETTER_E = 0x044D;
const char32_t CYRILLIC_SMALL_LETTER_YU = 0x044E;
const char32_t CYRILLIC_SMALL_LETTER_YA = 0x044F;
const char32_t CYRILLIC_SMALL_LETTER_IE_WITH_GRAVE = 0x0450;
const char32_t CYRILLIC_SMALL_LETTER_IO = 0x0451;
const char32_t CYRILLIC_SMALL_LETTER_DJE = 0x0452;
const char32_t CYRILLIC_SMALL_LETTER_GJE = 0x0453;
const char32_t CYRILLIC_SMALL_LETTER_UKRAINIAN_IE = 0x0454;
const char32_t CYRILLIC_SMALL_LETTER_DZE = 0x0455;
const char32_t CYRILLIC_SMALL_LETTER_BYELORUSSIAN_UKRAINIAN_I = 0x0456;
const char32_t CYRILLIC_SMALL_LETTER_YI = 0x0457;
const char32_t CYRILLIC_SMALL_LETTER_JE = 0x0458;
const char32_t CYRILLIC_SMALL_LETTER_LJE = 0x0459;
const char32_t CYRILLIC_SMALL_LETTER_NJE = 0x045A;
const char32_t CYRILLIC_SMALL_LETTER_TSHE = 0x045B;
const char32_t CYRILLIC_SMALL_LETTER_KJE = 0x045C;
const char32_t CYRILLIC_SMALL_LETTER_I_WITH_GRAVE = 0x045D;
const char32_t CYRILLIC_SMALL_LETTER_SHORT_U = 0x045E;
const char32_t CYRILLIC_SMALL_LETTER_DZHE = 0x045F;
const char32_t CYRILLIC_CAPITAL_LETTER_OMEGA = 0x0460;
const char32_t CYRILLIC_SMALL_LETTER_OMEGA = 0x0461;
const char32_t CYRILLIC_CAPITAL_LETTER_YAT = 0x0462;
const char32_t CYRILLIC_SMALL_LETTER_YAT = 0x0463;
const char32_t CYRILLIC_CAPITAL_LETTER_IOTIFIED_E = 0x0464;
const char32_t CYRILLIC_SMALL_LETTER_IOTIFIED_E = 0x0465;
const char32_t CYRILLIC_CAPITAL_LETTER_LITTLE_YUS = 0x0466;
const char32_t CYRILLIC_SMALL_LETTER_LITTLE_YUS = 0x0467;
const char32_t CYRILLIC_CAPITAL_LETTER_IOTIFIED_LITTLE_YUS = 0x0468;
const char32_t CYRILLIC_SMALL_LETTER_IOTIFIED_LITTLE_YUS = 0x0469;
const char32_t CYRILLIC_CAPITAL_LETTER_BIG_YUS = 0x046A;
const char32_t CYRILLIC_SMALL_LETTER_BIG_YUS = 0x046B;
const char32_t CYRILLIC_CAPITAL_LETTER_IOTIFIED_BIG_YUS = 0x046C;
const char32_t CYRILLIC_SMALL_LETTER_IOTIFIED_BIG_YUS = 0x046D;
const char32_t CYRILLIC_CAPITAL_LETTER_KSI = 0x046E;
const char32_t CYRILLIC_SMALL_LETTER_KSI = 0x046F;
const char32_t CYRILLIC_CAPITAL_LETTER_PSI = 0x0470;
const char32_t CYRILLIC_SMALL_LETTER_PSI = 0x0471;
const char32_t CYRILLIC_CAPITAL_LETTER_FITA = 0x0472;
const char32_t CYRILLIC_SMALL_LETTER_FITA = 0x0473;
const char32_t CYRILLIC_CAPITAL_LETTER_IZHITSA = 0x0474;
const char32_t CYRILLIC_SMALL_LETTER_IZHITSA = 0x0475;
const char32_t CYRILLIC_CAPITAL_LETTER_IZHITSA_WITH_DOUBLE_GRAVE_ACCENT = 0x0476;
const char32_t CYRILLIC_SMALL_LETTER_IZHITSA_WITH_DOUBLE_GRAVE_ACCENT = 0x0477;
const char32_t CYRILLIC_CAPITAL_LETTER_UK = 0x0478;
const char32_t CYRILLIC_SMALL_LETTER_UK = 0x0479;
const char32_t CYRILLIC_CAPITAL_LETTER_ROUND_OMEGA = 0x047A;
const char32_t CYRILLIC_SMALL_LETTER_ROUND_OMEGA = 0x047B;
const char32_t CYRILLIC_CAPITAL_LETTER_OMEGA_WITH_TITLO = 0x047C;
const char32_t CYRILLIC_SMALL_LETTER_OMEGA_WITH_TITLO = 0x047D;
const char32_t CYRILLIC_CAPITAL_LETTER_OT = 0x047E;
const char32_t CYRILLIC_SMALL_LETTER_OT = 0x047F;
const char32_t CYRILLIC_CAPITAL_LETTER_KOPPA = 0x0480;
const char32_t CYRILLIC_SMALL_LETTER_KOPPA = 0x0481;
const char32_t CYRILLIC_THOUSANDS_SIGN = 0x0482;
const char32_t COMBINING_CYRILLIC_TITLO = 0x0483;
const char32_t COMBINING_CYRILLIC_PALATALIZATION = 0x0484;
const char32_t COMBINING_CYRILLIC_DASIA_PNEUMATA = 0x0485;
const char32_t COMBINING_CYRILLIC_PSILI_PNEUMATA = 0x0486;
const char32_t COMBINING_CYRILLIC_POKRYTIE = 0x0487;
const char32_t COMBINING_CYRILLIC_HUNDRED_THOUSANDS_SIGN = 0x0488;
const char32_t COMBINING_CYRILLIC_MILLIONS_SIGN = 0x0489;
const char32_t CYRILLIC_CAPITAL_LETTER_SHORT_I_WITH_TAIL = 0x048A;
const char32_t CYRILLIC_SMALL_LETTER_SHORT_I_WITH_TAIL = 0x048B;
const char32_t CYRILLIC_CAPITAL_LETTER_SEMISOFT_SIGN = 0x048C;
const char32_t CYRILLIC_SMALL_LETTER_SEMISOFT_SIGN = 0x048D;
const char32_t CYRILLIC_CAPITAL_LETTER_ER_WITH_TICK = 0x048E;
const char32_t CYRILLIC_SMALL_LETTER_ER_WITH_TICK = 0x048F;
const char32_t CYRILLIC_CAPITAL_LETTER_GHE_WITH_UPTURN = 0x0490;
const char32_t CYRILLIC_SMALL_LETTER_GHE_WITH_UPTURN = 0x0491;
const char32_t CYRILLIC_CAPITAL_LETTER_GHE_WITH_STROKE = 0x0492;
const char32_t CYRILLIC_SMALL_LETTER_GHE_WITH_STROKE = 0x0493;
const char32_t CYRILLIC_CAPITAL_LETTER_GHE_WITH_MIDDLE_HOOK = 0x0494;
const char32_t CYRILLIC_SMALL_LETTER_GHE_WITH_MIDDLE_HOOK = 0x0495;
const char32_t CYRILLIC_CAPITAL_LETTER_ZHE_WITH_DESCENDER = 0x0496;
const char32_t CYRILLIC_SMALL_LETTER_ZHE_WITH_DESCENDER = 0x0497;
const char32_t CYRILLIC_CAPITAL_LETTER_ZE_WITH_DESCENDER = 0x0498;
const char32_t CYRILLIC_SMALL_LETTER_ZE_WITH_DESCENDER = 0x0499;
const char32_t CYRILLIC_CAPITAL_LETTER_KA_WITH_DESCENDER = 0x049A;
const char32_t CYRILLIC_SMALL_LETTER_KA_WITH_DESCENDER = 0x049B;
const char32_t CYRILLIC_CAPITAL_LETTER_KA_WITH_VERTICAL_STROKE = 0x049C;
const char32_t CYRILLIC_SMALL_LETTER_KA_WITH_VERTICAL_STROKE = 0x049D;
const char32_t CYRILLIC_CAPITAL_LETTER_KA_WITH_STROKE = 0x049E;
const char32_t CYRILLIC_SMALL_LETTER_KA_WITH_STROKE = 0x049F;
const char32_t CYRILLIC_CAPITAL_LETTER_BASHKIR_KA = 0x04A0;
const char32_t CYRILLIC_SMALL_LETTER_BASHKIR_KA = 0x04A1;
const char32_t CYRILLIC_CAPITAL_LETTER_EN_WITH_DESCENDER = 0x04A2;
const char32_t CYRILLIC_SMALL_LETTER_EN_WITH_DESCENDER = 0x04A3;
const char32_t CYRILLIC_CAPITAL_LIGATURE_EN_GHE = 0x04A4;
const char32_t CYRILLIC_SMALL_LIGATURE_EN_GHE = 0x04A5;
const char32_t CYRILLIC_CAPITAL_LETTER_PE_WITH_MIDDLE_HOOK = 0x04A6;
const char32_t CYRILLIC_SMALL_LETTER_PE_WITH_MIDDLE_HOOK = 0x04A7;
const char32_t CYRILLIC_CAPITAL_LETTER_ABKHASIAN_HA = 0x04A8;
const char32_t CYRILLIC_SMALL_LETTER_ABKHASIAN_HA = 0x04A9;
const char32_t CYRILLIC_CAPITAL_LETTER_ES_WITH_DESCENDER = 0x04AA;
const char32_t CYRILLIC_SMALL_LETTER_ES_WITH_DESCENDER = 0x04AB;
const char32_t CYRILLIC_CAPITAL_LETTER_TE_WITH_DESCENDER = 0x04AC;
const char32_t CYRILLIC_SMALL_LETTER_TE_WITH_DESCENDER = 0x04AD;
const char32_t CYRILLIC_CAPITAL_LETTER_STRAIGHT_U = 0x04AE;
const char32_t CYRILLIC_SMALL_LETTER_STRAIGHT_U = 0x04AF;
const char32_t CYRILLIC_CAPITAL_LETTER_STRAIGHT_U_WITH_STROKE = 0x04B0;
const char32_t CYRILLIC_SMALL_LETTER_STRAIGHT_U_WITH_STROKE = 0x04B1;
const char32_t CYRILLIC_CAPITAL_LETTER_HA_WITH_DESCENDER = 0x04B2;
const char32_t CYRILLIC_SMALL_LETTER_HA_WITH_DESCENDER = 0x04B3;
const char32_t CYRILLIC_CAPITAL_LIGATURE_TE_TSE = 0x04B4;
const char32_t CYRILLIC_SMALL_LIGATURE_TE_TSE = 0x04B5;
const char32_t CYRILLIC_CAPITAL_LETTER_CHE_WITH_DESCENDER = 0x04B6;
const char32_t CYRILLIC_SMALL_LETTER_CHE_WITH_DESCENDER = 0x04B7;
const char32_t CYRILLIC_CAPITAL_LETTER_CHE_WITH_VERTICAL_STROKE = 0x04B8;
const char32_t CYRILLIC_SMALL_LETTER_CHE_WITH_VERTICAL_STROKE = 0x04B9;
const char32_t CYRILLIC_CAPITAL_LETTER_SHHA = 0x04BA;
const char32_t CYRILLIC_SMALL_LETTER_SHHA = 0x04BB;
const char32_t CYRILLIC_CAPITAL_LETTER_ABKHASIAN_CHE = 0x04BC;
const char32_t CYRILLIC_SMALL_LETTER_ABKHASIAN_CHE = 0x04BD;
const char32_t CYRILLIC_CAPITAL_LETTER_ABKHASIAN_CHE_WITH_DESCENDER = 0x04BE;
const char32_t CYRILLIC_SMALL_LETTER_ABKHASIAN_CHE_WITH_DESCENDER = 0x04BF;
const char32_t CYRILLIC_LETTER_PALOCHKA = 0x04C0;
const char32_t CYRILLIC_CAPITAL_LETTER_ZHE_WITH_BREVE = 0x04C1;
const char32_t CYRILLIC_SMALL_LETTER_ZHE_WITH_BREVE = 0x04C2;
const char32_t CYRILLIC_CAPITAL_LETTER_KA_WITH_HOOK = 0x04C3;
const char32_t CYRILLIC_SMALL_LETTER_KA_WITH_HOOK = 0x04C4;
const char32_t CYRILLIC_CAPITAL_LETTER_EL_WITH_TAIL = 0x04C5;
const char32_t CYRILLIC_SMALL_LETTER_EL_WITH_TAIL = 0x04C6;
const char32_t CYRILLIC_CAPITAL_LETTER_EN_WITH_HOOK = 0x04C7;
const char32_t CYRILLIC_SMALL_LETTER_EN_WITH_HOOK = 0x04C8;
const char32_t CYRILLIC_CAPITAL_LETTER_EN_WITH_TAIL = 0x04C9;
const char32_t CYRILLIC_SMALL_LETTER_EN_WITH_TAIL = 0x04CA;
const char32_t CYRILLIC_CAPITAL_LETTER_KHAKASSIAN_CHE = 0x04CB;
const char32_t CYRILLIC_SMALL_LETTER_KHAKASSIAN_CHE = 0x04CC;
const char32_t CYRILLIC_CAPITAL_LETTER_EM_WITH_TAIL = 0x04CD;
const char32_t CYRILLIC_SMALL_LETTER_EM_WITH_TAIL = 0x04CE;
const char32_t CYRILLIC_SMALL_LETTER_PALOCHKA = 0x04CF;
const char32_t CYRILLIC_CAPITAL_LETTER_A_WITH_BREVE = 0x04D0;
const char32_t CYRILLIC_SMALL_LETTER_A_WITH_BREVE = 0x04D1;
const char32_t CYRILLIC_CAPITAL_LETTER_A_WITH_DIAERESIS = 0x04D2;
const char32_t CYRILLIC_SMALL_LETTER_A_WITH_DIAERESIS = 0x04D3;
const char32_t CYRILLIC_CAPITAL_LIGATURE_A_IE = 0x04D4;
const char32_t CYRILLIC_SMALL_LIGATURE_A_IE = 0x04D5;
const char32_t CYRILLIC_CAPITAL_LETTER_IE_WITH_BREVE = 0x04D6;
const char32_t CYRILLIC_SMALL_LETTER_IE_WITH_BREVE = 0x04D7;
const char32_t CYRILLIC_CAPITAL_LETTER_SCHWA = 0x04D8;
const char32_t CYRILLIC_SMALL_LETTER_SCHWA = 0x04D9;
const char32_t CYRILLIC_CAPITAL_LETTER_SCHWA_WITH_DIAERESIS = 0x04DA;
const char32_t CYRILLIC_SMALL_LETTER_SCHWA_WITH_DIAERESIS = 0x04DB;
const char32_t CYRILLIC_CAPITAL_LETTER_ZHE_WITH_DIAERESIS = 0x04DC;
const char32_t CYRILLIC_SMALL_LETTER_ZHE_WITH_DIAERESIS = 0x04DD;
const char32_t CYRILLIC_CAPITAL_LETTER_ZE_WITH_DIAERESIS = 0x04DE;
const char32_t CYRILLIC_SMALL_LETTER_ZE_WITH_DIAERESIS = 0x04DF;
const char32_t CYRILLIC_CAPITAL_LETTER_ABKHASIAN_DZE = 0x04E0;
const char32_t CYRILLIC_SMALL_LETTER_ABKHASIAN_DZE = 0x04E1;
const char32_t CYRILLIC_CAPITAL_LETTER_I_WITH_MACRON = 0x04E2;
const char32_t CYRILLIC_SMALL_LETTER_I_WITH_MACRON = 0x04E3;
const char32_t CYRILLIC_CAPITAL_LETTER_I_WITH_DIAERESIS = 0x04E4;
const char32_t CYRILLIC_SMALL_LETTER_I_WITH_DIAERESIS = 0x04E5;
const char32_t CYRILLIC_CAPITAL_LETTER_O_WITH_DIAERESIS = 0x04E6;
const char32_t CYRILLIC_SMALL_LETTER_O_WITH_DIAERESIS = 0x04E7;
const char32_t CYRILLIC_CAPITAL_LETTER_BARRED_O = 0x04E8;
const char32_t CYRILLIC_SMALL_LETTER_BARRED_O = 0x04E9;
const char32_t CYRILLIC_CAPITAL_LETTER_BARRED_O_WITH_DIAERESIS = 0x04EA;
const char32_t CYRILLIC_SMALL_LETTER_BARRED_O_WITH_DIAERESIS = 0x04EB;
const char32_t CYRILLIC_CAPITAL_LETTER_E_WITH_DIAERESIS = 0x04EC;
const char32_t CYRILLIC_SMALL_LETTER_E_WITH_DIAERESIS = 0x04ED;
const char32_t CYRILLIC_CAPITAL_LETTER_U_WITH_MACRON = 0x04EE;
const char32_t CYRILLIC_SMALL_LETTER_U_WITH_MACRON = 0x04EF;
const char32_t CYRILLIC_CAPITAL_LETTER_U_WITH_DIAERESIS = 0x04F0;
const char32_t CYRILLIC_SMALL_LETTER_U_WITH_DIAERESIS = 0x04F1;
const char32_t CYRILLIC_CAPITAL_LETTER_U_WITH_DOUBLE_ACUTE = 0x04F2;
const char32_t CYRILLIC_SMALL_LETTER_U_WITH_DOUBLE_ACUTE = 0x04F3;
const char32_t CYRILLIC_CAPITAL_LETTER_CHE_WITH_DIAERESIS = 0x04F4;
const char32_t CYRILLIC_SMALL_LETTER_CHE_WITH_DIAERESIS = 0x04F5;
const char32_t CYRILLIC_CAPITAL_LETTER_GHE_WITH_DESCENDER = 0x04F6;
const char32_t CYRILLIC_SMALL_LETTER_GHE_WITH_DESCENDER = 0x04F7;
const char32_t CYRILLIC_CAPITAL_LETTER_YERU_WITH_DIAERESIS = 0x04F8;
const char32_t CYRILLIC_SMALL_LETTER_YERU_WITH_DIAERESIS = 0x04F9;
const char32_t CYRILLIC_CAPITAL_LETTER_GHE_WITH_STROKE_AND_HOOK = 0x04FA;
const char32_t CYRILLIC_SMALL_LETTER_GHE_WITH_STROKE_AND_HOOK = 0x04FB;
const char32_t CYRILLIC_CAPITAL_LETTER_HA_WITH_HOOK = 0x04FC;
const char32_t CYRILLIC_SMALL_LETTER_HA_WITH_HOOK = 0x04FD;
const char32_t CYRILLIC_CAPITAL_LETTER_HA_WITH_STROKE = 0x04FE;
const char32_t CYRILLIC_SMALL_LETTER_HA_WITH_STROKE = 0x04FF;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_DE = 0x0500;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_DE = 0x0501;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_DJE = 0x0502;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_DJE = 0x0503;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_ZJE = 0x0504;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_ZJE = 0x0505;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_DZJE = 0x0506;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_DZJE = 0x0507;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_LJE = 0x0508;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_LJE = 0x0509;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_NJE = 0x050A;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_NJE = 0x050B;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_SJE = 0x050C;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_SJE = 0x050D;
const char32_t CYRILLIC_CAPITAL_LETTER_KOMI_TJE = 0x050E;
const char32_t CYRILLIC_SMALL_LETTER_KOMI_TJE = 0x050F;
const char32_t CYRILLIC_CAPITAL_LETTER_REVERSED_ZE = 0x0510;
const char32_t CYRILLIC_SMALL_LETTER_REVERSED_ZE = 0x0511;
const char32_t CYRILLIC_CAPITAL_LETTER_EL_WITH_HOOK = 0x0512;
const char32_t CYRILLIC_SMALL_LETTER_EL_WITH_HOOK = 0x0513;
const char32_t CYRILLIC_CAPITAL_LETTER_LHA = 0x0514;
const char32_t CYRILLIC_SMALL_LETTER_LHA = 0x0515;
const char32_t CYRILLIC_CAPITAL_LETTER_RHA = 0x0516;
const char32_t CYRILLIC_SMALL_LETTER_RHA = 0x0517;
const char32_t CYRILLIC_CAPITAL_LETTER_YAE = 0x0518;
const char32_t CYRILLIC_SMALL_LETTER_YAE = 0x0519;
const char32_t CYRILLIC_CAPITAL_LETTER_QA = 0x051A;
const char32_t CYRILLIC_SMALL_LETTER_QA = 0x051B;
const char32_t CYRILLIC_CAPITAL_LETTER_WE = 0x051C;
const char32_t CYRILLIC_SMALL_LETTER_WE = 0x051D;
const char32_t CYRILLIC_CAPITAL_LETTER_ALEUT_KA = 0x051E;
const char32_t CYRILLIC_SMALL_LETTER_ALEUT_KA = 0x051F;
const char32_t CYRILLIC_CAPITAL_LETTER_EL_WITH_MIDDLE_HOOK = 0x0520;
const char32_t CYRILLIC_SMALL_LETTER_EL_WITH_MIDDLE_HOOK = 0x0521;
const char32_t CYRILLIC_CAPITAL_LETTER_EN_WITH_MIDDLE_HOOK = 0x0522;
const char32_t CYRILLIC_SMALL_LETTER_EN_WITH_MIDDLE_HOOK = 0x0523;
const char32_t CYRILLIC_CAPITAL_LETTER_PE_WITH_DESCENDER = 0x0524;
const char32_t CYRILLIC_SMALL_LETTER_PE_WITH_DESCENDER = 0x0525;
const char32_t CYRILLIC_CAPITAL_LETTER_SHHA_WITH_DESCENDER = 0x0526;
const char32_t CYRILLIC_SMALL_LETTER_SHHA_WITH_DESCENDER = 0x0527;
const char32_t CYRILLIC_CAPITAL_LETTER_EN_WITH_LEFT_HOOK = 0x0528;
const char32_t CYRILLIC_SMALL_LETTER_EN_WITH_LEFT_HOOK = 0x0529;
const char32_t CYRILLIC_CAPITAL_LETTER_DZZHE = 0x052A;
const char32_t CYRILLIC_SMALL_LETTER_DZZHE = 0x052B;
const char32_t CYRILLIC_CAPITAL_LETTER_DCHE = 0x052C;
const char32_t CYRILLIC_SMALL_LETTER_DCHE = 0x052D;
const char32_t CYRILLIC_CAPITAL_LETTER_EL_WITH_DESCENDER = 0x052E;
const char32_t CYRILLIC_SMALL_LETTER_EL_WITH_DESCENDER = 0x052F;
const char32_t ARMENIAN_CAPITAL_LETTER_AYB = 0x0531;
const char32_t ARMENIAN_CAPITAL_LETTER_BEN = 0x0532;
const char32_t ARMENIAN_CAPITAL_LETTER_GIM = 0x0533;
const char32_t ARMENIAN_CAPITAL_LETTER_DA = 0x0534;
const char32_t ARMENIAN_CAPITAL_LETTER_ECH = 0x0535;
const char32_t ARMENIAN_CAPITAL_LETTER_ZA = 0x0536;
const char32_t ARMENIAN_CAPITAL_LETTER_EH = 0x0537;
const char32_t ARMENIAN_CAPITAL_LETTER_ET = 0x0538;
const char32_t ARMENIAN_CAPITAL_LETTER_TO = 0x0539;
const char32_t ARMENIAN_CAPITAL_LETTER_ZHE = 0x053A;
const char32_t ARMENIAN_CAPITAL_LETTER_INI = 0x053B;
const char32_t ARMENIAN_CAPITAL_LETTER_LIWN = 0x053C;
const char32_t ARMENIAN_CAPITAL_LETTER_XEH = 0x053D;
const char32_t ARMENIAN_CAPITAL_LETTER_CA = 0x053E;
const char32_t ARMENIAN_CAPITAL_LETTER_KEN = 0x053F;
const char32_t ARMENIAN_CAPITAL_LETTER_HO = 0x0540;
const char32_t ARMENIAN_CAPITAL_LETTER_JA = 0x0541;
const char32_t ARMENIAN_CAPITAL_LETTER_GHAD = 0x0542;
const char32_t ARMENIAN_CAPITAL_LETTER_CHEH = 0x0543;
const char32_t ARMENIAN_CAPITAL_LETTER_MEN = 0x0544;
const char32_t ARMENIAN_CAPITAL_LETTER_YI = 0x0545;
const char32_t ARMENIAN_CAPITAL_LETTER_NOW = 0x0546;
const char32_t ARMENIAN_CAPITAL_LETTER_SHA = 0x0547;
const char32_t ARMENIAN_CAPITAL_LETTER_VO = 0x0548;
const char32_t ARMENIAN_CAPITAL_LETTER_CHA = 0x0549;
const char32_t ARMENIAN_CAPITAL_LETTER_PEH = 0x054A;
const char32_t ARMENIAN_CAPITAL_LETTER_JHEH = 0x054B;
const char32_t ARMENIAN_CAPITAL_LETTER_RA = 0x054C;
const char32_t ARMENIAN_CAPITAL_LETTER_SEH = 0x054D;
const char32_t ARMENIAN_CAPITAL_LETTER_VEW = 0x054E;
const char32_t ARMENIAN_CAPITAL_LETTER_TIWN = 0x054F;
const char32_t ARMENIAN_CAPITAL_LETTER_REH = 0x0550;
const char32_t ARMENIAN_CAPITAL_LETTER_CO = 0x0551;
const char32_t ARMENIAN_CAPITAL_LETTER_YIWN = 0x0552;
const char32_t ARMENIAN_CAPITAL_LETTER_PIWR = 0x0553;
const char32_t ARMENIAN_CAPITAL_LETTER_KEH = 0x0554;
const char32_t ARMENIAN_CAPITAL_LETTER_OH = 0x0555;
const char32_t ARMENIAN_CAPITAL_LETTER_FEH = 0x0556;
const char32_t ARMENIAN_MODIFIER_LETTER_LEFT_HALF_RING = 0x0559;
const char32_t ARMENIAN_APOSTROPHE = 0x055A;
const char32_t ARMENIAN_EMPHASIS_MARK = 0x055B;
const char32_t ARMENIAN_EXCLAMATION_MARK = 0x055C;
const char32_t ARMENIAN_COMMA = 0x055D;
const char32_t ARMENIAN_QUESTION_MARK = 0x055E;
const char32_t ARMENIAN_ABBREVIATION_MARK = 0x055F;
const char32_t ARMENIAN_SMALL_LETTER_TURNED_AYB = 0x0560;
const char32_t ARMENIAN_SMALL_LETTER_AYB = 0x0561;
const char32_t ARMENIAN_SMALL_LETTER_BEN = 0x0562;
const char32_t ARMENIAN_SMALL_LETTER_GIM = 0x0563;
const char32_t ARMENIAN_SMALL_LETTER_DA = 0x0564;
const char32_t ARMENIAN_SMALL_LETTER_ECH = 0x0565;
const char32_t ARMENIAN_SMALL_LETTER_ZA = 0x0566;
const char32_t ARMENIAN_SMALL_LETTER_EH = 0x0567;
const char32_t ARMENIAN_SMALL_LETTER_ET = 0x0568;
const char32_t ARMENIAN_SMALL_LETTER_TO = 0x0569;
const char32_t ARMENIAN_SMALL_LETTER_ZHE = 0x056A;
const char32_t ARMENIAN_SMALL_LETTER_INI = 0x056B;
const char32_t ARMENIAN_SMALL_LETTER_LIWN = 0x056C;
const char32_t ARMENIAN_SMALL_LETTER_XEH = 0x056D;
const char32_t ARMENIAN_SMALL_LETTER_CA = 0x056E;
const char32_t ARMENIAN_SMALL_LETTER_KEN = 0x056F;
const char32_t ARMENIAN_SMALL_LETTER_HO = 0x0570;
const char32_t ARMENIAN_SMALL_LETTER_JA = 0x0571;
const char32_t ARMENIAN_SMALL_LETTER_GHAD = 0x0572;
const char32_t ARMENIAN_SMALL_LETTER_CHEH = 0x0573;
const char32_t ARMENIAN_SMALL_LETTER_MEN = 0x0574;
const char32_t ARMENIAN_SMALL_LETTER_YI = 0x0575;
const char32_t ARMENIAN_SMALL_LETTER_NOW = 0x0576;
const char32_t ARMENIAN_SMALL_LETTER_SHA = 0x0577;
const char32_t ARMENIAN_SMALL_LETTER_VO = 0x0578;
const char32_t ARMENIAN_SMALL_LETTER_CHA = 0x0579;
const char32_t ARMENIAN_SMALL_LETTER_PEH = 0x057A;
const char32_t ARMENIAN_SMALL_LETTER_JHEH = 0x057B;
const char32_t ARMENIAN_SMALL_LETTER_RA = 0x057C;
const char32_t ARMENIAN_SMALL_LETTER_SEH = 0x057D;
const char32_t ARMENIAN_SMALL_LETTER_VEW = 0x057E;
const char32_t ARMENIAN_SMALL_LETTER_TIWN = 0x057F;
const char32_t ARMENIAN_SMALL_LETTER_REH = 0x0580;
const char32_t ARMENIAN_SMALL_LETTER_CO = 0x0581;
const char32_t ARMENIAN_SMALL_LETTER_YIWN = 0x0582;
const char32_t ARMENIAN_SMALL_LETTER_PIWR = 0x0583;
const char32_t ARMENIAN_SMALL_LETTER_KEH = 0x0584;
const char32_t ARMENIAN_SMALL_LETTER_OH = 0x0585;
const char32_t ARMENIAN_SMALL_LETTER_FEH = 0x0586;
const char32_t ARMENIAN_SMALL_LIGATURE_ECH_YIWN = 0x0587;
const char32_t ARMENIAN_SMALL_LETTER_YI_WITH_STROKE = 0x0588;
const char32_t ARMENIAN_FULL_STOP = 0x0589;
const char32_t ARMENIAN_HYPHEN = 0x058A;
const char32_t RIGHT_FACING_ARMENIAN_ETERNITY_SIGN = 0x058D;
const char32_t LEFT_FACING_ARMENIAN_ETERNITY_SIGN = 0x058E;
const char32_t ARMENIAN_DRAM_SIGN = 0x058F;
const char32_t HEBREW_ACCENT_ETNAHTA = 0x0591;
const char32_t HEBREW_ACCENT_SEGOL = 0x0592;
const char32_t HEBREW_ACCENT_SHALSHELET = 0x0593;
const char32_t HEBREW_ACCENT_ZAQEF_QATAN = 0x0594;
const char32_t HEBREW_ACCENT_ZAQEF_GADOL = 0x0595;
const char32_t HEBREW_ACCENT_TIPEHA = 0x0596;
const char32_t HEBREW_ACCENT_REVIA = 0x0597;
const char32_t HEBREW_ACCENT_ZARQA = 0x0598;
const char32_t HEBREW_ACCENT_PASHTA = 0x0599;
const char32_t HEBREW_ACCENT_YETIV = 0x059A;
const char32_t HEBREW_ACCENT_TEVIR = 0x059B;
const char32_t HEBREW_ACCENT_GERESH = 0x059C;
const char32_t HEBREW_ACCENT_GERESH_MUQDAM = 0x059D;
const char32_t HEBREW_ACCENT_GERSHAYIM = 0x059E;
const char32_t HEBREW_ACCENT_QARNEY_PARA = 0x059F;
const char32_t HEBREW_ACCENT_TELISHA_GEDOLA = 0x05A0;
const char32_t HEBREW_ACCENT_PAZER = 0x05A1;
const char32_t HEBREW_ACCENT_ATNAH_HAFUKH = 0x05A2;
const char32_t HEBREW_ACCENT_MUNAH = 0x05A3;
const char32_t HEBREW_ACCENT_MAHAPAKH = 0x05A4;
const char32_t HEBREW_ACCENT_MERKHA = 0x05A5;
const char32_t HEBREW_ACCENT_MERKHA_KEFULA = 0x05A6;
const char32_t HEBREW_ACCENT_DARGA = 0x05A7;
const char32_t HEBREW_ACCENT_QADMA = 0x05A8;
const char32_t HEBREW_ACCENT_TELISHA_QETANA = 0x05A9;
const char32_t HEBREW_ACCENT_YERAH_BEN_YOMO = 0x05AA;
const char32_t HEBREW_ACCENT_OLE = 0x05AB;
const char32_t HEBREW_ACCENT_ILUY = 0x05AC;
const char32_t HEBREW_ACCENT_DEHI = 0x05AD;
const char32_t HEBREW_ACCENT_ZINOR = 0x05AE;
const char32_t HEBREW_MARK_MASORA_CIRCLE = 0x05AF;
const char32_t HEBREW_POINT_SHEVA = 0x05B0;
const char32_t HEBREW_POINT_HATAF_SEGOL = 0x05B1;
const char32_t HEBREW_POINT_HATAF_PATAH = 0x05B2;
const char32_t HEBREW_POINT_HATAF_QAMATS = 0x05B3;
const char32_t HEBREW_POINT_HIRIQ = 0x05B4;
const char32_t HEBREW_POINT_TSERE = 0x05B5;
const char32_t HEBREW_POINT_SEGOL = 0x05B6;
const char32_t HEBREW_POINT_PATAH = 0x05B7;
const char32_t HEBREW_POINT_QAMATS = 0x05B8;
const char32_t HEBREW_POINT_HOLAM = 0x05B9;
const char32_t HEBREW_POINT_HOLAM_HASER_FOR_VAV = 0x05BA;
const char32_t HEBREW_POINT_QUBUTS = 0x05BB;
const char32_t HEBREW_POINT_DAGESH_OR_MAPIQ = 0x05BC;
const char32_t HEBREW_POINT_METEG = 0x05BD;
const char32_t HEBREW_PUNCTUATION_MAQAF = 0x05BE;
const char32_t HEBREW_POINT_RAFE = 0x05BF;
const char32_t HEBREW_PUNCTUATION_PASEQ = 0x05C0;
const char32_t HEBREW_POINT_SHIN_DOT = 0x05C1;
const char32_t HEBREW_POINT_SIN_DOT = 0x05C2;
const char32_t HEBREW_PUNCTUATION_SOF_PASUQ = 0x05C3;
const char32_t HEBREW_MARK_UPPER_DOT = 0x05C4;
const char32_t HEBREW_MARK_LOWER_DOT = 0x05C5;
const char32_t HEBREW_PUNCTUATION_NUN_HAFUKHA = 0x05C6;
const char32_t HEBREW_POINT_QAMATS_QATAN = 0x05C7;
const char32_t HEBREW_LETTER_ALEF = 0x05D0;
const char32_t HEBREW_LETTER_BET = 0x05D1;
const char32_t HEBREW_LETTER_GIMEL = 0x05D2;
const char32_t HEBREW_LETTER_DALET = 0x05D3;
const char32_t HEBREW_LETTER_HE = 0x05D4;
const char32_t HEBREW_LETTER_VAV = 0x05D5;
const char32_t HEBREW_LETTER_ZAYIN = 0x05D6;
const char32_t HEBREW_LETTER_HET = 0x05D7;
const char32_t HEBREW_LETTER_TET = 0x05D8;
const char32_t HEBREW_LETTER_YOD = 0x05D9;
const char32_t HEBREW_LETTER_FINAL_KAF = 0x05DA;
const char32_t HEBREW_LETTER_KAF = 0x05DB;
const char32_t HEBREW_LETTER_LAMED = 0x05DC;
const char32_t HEBREW_LETTER_FINAL_MEM = 0x05DD;
const char32_t HEBREW_LETTER_MEM = 0x05DE;
const char32_t HEBREW_LETTER_FINAL_NUN = 0x05DF;
const char32_t HEBREW_LETTER_NUN = 0x05E0;
const char32_t HEBREW_LETTER_SAMEKH = 0x05E1;
const char32_t HEBREW_LETTER_AYIN = 0x05E2;
const char32_t HEBREW_LETTER_FINAL_PE = 0x05E3;
const char32_t HEBREW_LETTER_PE = 0x05E4;
const char32_t HEBREW_LETTER_FINAL_TSADI = 0x05E5;
const char32_t HEBREW_LETTER_TSADI = 0x05E6;
const char32_t HEBREW_LETTER_QOF = 0x05E7;
const char32_t HEBREW_LETTER_RESH = 0x05E8;
const char32_t HEBREW_LETTER_SHIN = 0x05E9;
const char32_t HEBREW_LETTER_TAV = 0x05EA;
const char32_t HEBREW_YOD_TRIANGLE = 0x05EF;
const char32_t HEBREW_LIGATURE_YIDDISH_DOUBLE_VAV = 0x05F0;
const char32_t HEBREW_LIGATURE_YIDDISH_VAV_YOD = 0x05F1;
const char32_t HEBREW_LIGATURE_YIDDISH_DOUBLE_YOD = 0x05F2;
const char32_t HEBREW_PUNCTUATION_GERESH = 0x05F3;
const char32_t HEBREW_PUNCTUATION_GERSHAYIM = 0x05F4;
const char32_t ARABIC_NUMBER_SIGN = 0x0600;
const char32_t ARABIC_SIGN_SANAH = 0x0601;
const char32_t ARABIC_FOOTNOTE_MARKER = 0x0602;
const char32_t ARABIC_SIGN_SAFHA = 0x0603;
const char32_t ARABIC_SIGN_SAMVAT = 0x0604;
const char32_t ARABIC_NUMBER_MARK_ABOVE = 0x0605;
const char32_t ARABIC_INDIC_CUBE_ROOT = 0x0606;
const char32_t ARABIC_INDIC_FOURTH_ROOT = 0x0607;
const char32_t ARABIC_RAY = 0x0608;
const char32_t ARABIC_INDIC_PER_MILLE_SIGN = 0x0609;
const char32_t ARABIC_INDIC_PER_TEN_THOUSAND_SIGN = 0x060A;
const char32_t AFGHANI_SIGN = 0x060B;
const char32_t ARABIC_COMMA = 0x060C;
const char32_t ARABIC_DATE_SEPARATOR = 0x060D;
const char32_t ARABIC_POETIC_VERSE_SIGN = 0x060E;
const char32_t ARABIC_SIGN_MISRA = 0x060F;
const char32_t ARABIC_SIGN_SALLALLAHOU_ALAYHE_WASSALLAM = 0x0610;
const char32_t ARABIC_SIGN_ALAYHE_ASSALLAM = 0x0611;
const char32_t ARABIC_SIGN_RAHMATULLAH_ALAYHE = 0x0612;
const char32_t ARABIC_SIGN_RADI_ALLAHOU_ANHU = 0x0613;
const char32_t ARABIC_SIGN_TAKHALLUS = 0x0614;
const char32_t ARABIC_SMALL_HIGH_TAH = 0x0615;
const char32_t ARABIC_SMALL_HIGH_LIGATURE_ALEF_WITH_LAM_WITH_YEH = 0x0616;
const char32_t ARABIC_SMALL_HIGH_ZAIN = 0x0617;
const char32_t ARABIC_SMALL_FATHA = 0x0618;
const char32_t ARABIC_SMALL_DAMMA = 0x0619;
const char32_t ARABIC_SMALL_KASRA = 0x061A;
const char32_t ARABIC_SEMICOLON = 0x061B;
const char32_t ARABIC_LETTER_MARK = 0x061C;
const char32_t ARABIC_END_OF_TEXT_MARK = 0x061D;
const char32_t ARABIC_TRIPLE_DOT_PUNCTUATION_MARK = 0x061E;
const char32_t ARABIC_QUESTION_MARK = 0x061F;
const char32_t ARABIC_LETTER_KASHMIRI_YEH = 0x0620;
const char32_t ARABIC_LETTER_HAMZA = 0x0621;
const char32_t ARABIC_LETTER_ALEF_WITH_MADDA_ABOVE = 0x0622;
const char32_t ARABIC_LETTER_ALEF_WITH_HAMZA_ABOVE = 0x0623;
const char32_t ARABIC_LETTER_WAW_WITH_HAMZA_ABOVE = 0x0624;
const char32_t ARABIC_LETTER_ALEF_WITH_HAMZA_BELOW = 0x0625;
const char32_t ARABIC_LETTER_YEH_WITH_HAMZA_ABOVE = 0x0626;
const char32_t ARABIC_LETTER_ALEF = 0x0627;
const char32_t ARABIC_LETTER_BEH = 0x0628;
const char32_t ARABIC_LETTER_TEH_MARBUTA = 0x0629;
const char32_t ARABIC_LETTER_TEH = 0x062A;
const char32_t ARABIC_LETTER_THEH = 0x062B;
const char32_t ARABIC_LETTER_JEEM = 0x062C;
const char32_t ARABIC_LETTER_HAH = 0x062D;
const char32_t ARABIC_LETTER_KHAH = 0x062E;
const char32_t ARABIC_LETTER_DAL = 0x062F;
const char32_t ARABIC_LETTER_THAL = 0x0630;
const char32_t ARABIC_LETTER_REH = 0x0631;
const char32_t ARABIC_LETTER_ZAIN = 0x0632;
const char32_t ARABIC_LETTER_SEEN = 0x0633;
const char32_t ARABIC_LETTER_SHEEN = 0x0634;
const char32_t ARABIC_LETTER_SAD = 0x0635;
const char32_t ARABIC_LETTER_DAD = 0x0636;
const char32_t ARABIC_LETTER_TAH = 0x0637;
const char32_t ARABIC_LETTER_ZAH = 0x0638;
const char32_t ARABIC_LETTER_AIN = 0x0639;
const char32_t ARABIC_LETTER_GHAIN = 0x063A;
const char32_t ARABIC_LETTER_KEHEH_WITH_TWO_DOTS_ABOVE = 0x063B;
const char32_t ARABIC_LETTER_KEHEH_WITH_THREE_DOTS_BELOW = 0x063C;
const char32_t ARABIC_LETTER_FARSI_YEH_WITH_INVERTED_V = 0x063D;
const char32_t ARABIC_LETTER_FARSI_YEH_WITH_TWO_DOTS_ABOVE = 0x063E;
const char32_t ARABIC_LETTER_FARSI_YEH_WITH_THREE_DOTS_ABOVE = 0x063F;
const char32_t ARABIC_TATWEEL = 0x0640;
const char32_t ARABIC_LETTER_FEH = 0x0641;
const char32_t ARABIC_LETTER_QAF = 0x0642;
const char32_t ARABIC_LETTER_KAF = 0x0643;
const char32_t ARABIC_LETTER_LAM = 0x0644;
const char32_t ARABIC_LETTER_MEEM = 0x0645;
const char32_t ARABIC_LETTER_NOON = 0x0646;
const char32_t ARABIC_LETTER_HEH = 0x0647;
const char32_t ARABIC_LETTER_WAW = 0x0648;
const char32_t ARABIC_LETTER_ALEF_MAKSURA = 0x0649;
const char32_t ARABIC_LETTER_YEH = 0x064A;
const char32_t ARABIC_FATHATAN = 0x064B;
const char32_t ARABIC_DAMMATAN = 0x064C;
const char32_t ARABIC_KASRATAN = 0x064D;
const char32_t ARABIC_FATHA = 0x064E;
const char32_t ARABIC_DAMMA = 0x064F;
const char32_t ARABIC_KASRA = 0x0650;
const char32_t ARABIC_SHADDA = 0x0651;
const char32_t ARABIC_SUKUN = 0x0652;
const char32_t ARABIC_MADDAH_ABOVE = 0x0653;
const char32_t ARABIC_HAMZA_ABOVE = 0x0654;
const char32_t ARABIC_HAMZA_BELOW = 0x0655;
const char32_t ARABIC_SUBSCRIPT_ALEF = 0x0656;
const char32_t ARABIC_INVERTED_DAMMA = 0x0657;
const char32_t ARABIC_MARK_NOON_GHUNNA = 0x0658;
const char32_t ARABIC_ZWARAKAY = 0x0659;
const char32_t ARABIC_VOWEL_SIGN_SMALL_V_ABOVE = 0x065A;
const char32_t ARABIC_VOWEL_SIGN_INVERTED_SMALL_V_ABOVE = 0x065B;
const char32_t ARABIC_VOWEL_SIGN_DOT_BELOW = 0x065C;
const char32_t ARABIC_REVERSED_DAMMA = 0x065D;
const char32_t ARABIC_FATHA_WITH_TWO_DOTS = 0x065E;
const char32_t ARABIC_WAVY_HAMZA_BELOW = 0x065F;
const char32_t ARABIC_INDIC_DIGIT_ZERO = 0x0660;
const char32_t ARABIC_INDIC_DIGIT_ONE = 0x0661;
const char32_t ARABIC_INDIC_DIGIT_TWO = 0x0662;
const char32_t ARABIC_INDIC_DIGIT_THREE = 0x0663;
const char32_t ARABIC_INDIC_DIGIT_FOUR = 0x0664;
const char32_t ARABIC_INDIC_DIGIT_FIVE = 0x0665;
const char32_t ARABIC_INDIC_DIGIT_SIX = 0x0666;
const char32_t ARABIC_INDIC_DIGIT_SEVEN = 0x0667;
const char32_t ARABIC_INDIC_DIGIT_EIGHT = 0x0668;
const char32_t ARABIC_INDIC_DIGIT_NINE = 0x0669;
const char32_t ARABIC_PERCENT_SIGN = 0x066A;
const char32_t ARABIC_DECIMAL_SEPARATOR = 0x066B;
const char32_t ARABIC_THOUSANDS_SEPARATOR = 0x066C;
const char32_t ARABIC_FIVE_POINTED_STAR = 0x066D;
const char32_t ARABIC_LETTER_DOTLESS_BEH = 0x066E;
const char32_t ARABIC_LETTER_DOTLESS_QAF = 0x066F;
const char32_t ARABIC_LETTER_SUPERSCRIPT_ALEF = 0x0670;
const char32_t ARABIC_LETTER_ALEF_WASLA = 0x0671;
const char32_t ARABIC_LETTER_ALEF_WITH_WAVY_HAMZA_ABOVE = 0x0672;
const char32_t ARABIC_LETTER_ALEF_WITH_WAVY_HAMZA_BELOW = 0x0673;
const char32_t ARABIC_LETTER_HIGH_HAMZA = 0x0674;
const char32_t ARABIC_LETTER_HIGH_HAMZA_ALEF = 0x0675;
const char32_t ARABIC_LETTER_HIGH_HAMZA_WAW = 0x0676;
const char32_t ARABIC_LETTER_U_WITH_HAMZA_ABOVE = 0x0677;
const char32_t ARABIC_LETTER_HIGH_HAMZA_YEH = 0x0678;
const char32_t ARABIC_LETTER_TTEH = 0x0679;
const char32_t ARABIC_LETTER_TTEHEH = 0x067A;
const char32_t ARABIC_LETTER_BEEH = 0x067B;
const char32_t ARABIC_LETTER_TEH_WITH_RING = 0x067C;
const char32_t ARABIC_LETTER_TEH_WITH_THREE_DOTS_ABOVE_DOWNWARDS = 0x067D;
const char32_t ARABIC_LETTER_PEH = 0x067E;
const char32_t ARABIC_LETTER_TEHEH = 0x067F;
const char32_t ARABIC_LETTER_BEHEH = 0x0680;
const char32_t ARABIC_LETTER_HAH_WITH_HAMZA_ABOVE = 0x0681;
const char32_t ARABIC_LETTER_HAH_WITH_TWO_DOTS_VERTICAL_ABOVE = 0x0682;
const char32_t ARABIC_LETTER_NYEH = 0x0683;
const char32_t ARABIC_LETTER_DYEH = 0x0684;
const char32_t ARABIC_LETTER_HAH_WITH_THREE_DOTS_ABOVE = 0x0685;
const char32_t ARABIC_LETTER_TCHEH = 0x0686;
const char32_t ARABIC_LETTER_TCHEHEH = 0x0687;
const char32_t ARABIC_LETTER_DDAL = 0x0688;
const char32_t ARABIC_LETTER_DAL_WITH_RING = 0x0689;
const char32_t ARABIC_LETTER_DAL_WITH_DOT_BELOW = 0x068A;
const char32_t ARABIC_LETTER_DAL_WITH_DOT_BELOW_AND_SMALL_TAH = 0x068B;
const char32_t ARABIC_LETTER_DAHAL = 0x068C;
const char32_t ARABIC_LETTER_DDAHAL = 0x068D;
const char32_t ARABIC_LETTER_DUL = 0x068E;
const char32_t ARABIC_LETTER_DAL_WITH_THREE_DOTS_ABOVE_DOWNWARDS = 0x068F;
const char32_t ARABIC_LETTER_DAL_WITH_FOUR_DOTS_ABOVE = 0x0690;
const char32_t ARABIC_LETTER_RREH = 0x0691;
const char32_t ARABIC_LETTER_REH_WITH_SMALL_V = 0x0692;
const char32_t ARABIC_LETTER_REH_WITH_RING = 0x0693;
const char32_t ARABIC_LETTER_REH_WITH_DOT_BELOW = 0x0694;
const char32_t ARABIC_LETTER_REH_WITH_SMALL_V_BELOW = 0x0695;
const char32_t ARABIC_LETTER_REH_WITH_DOT_BELOW_AND_DOT_ABOVE = 0x0696;
const char32_t ARABIC_LETTER_REH_WITH_TWO_DOTS_ABOVE = 0x0697;
const char32_t ARABIC_LETTER_JEH = 0x0698;
const char32_t ARABIC_LETTER_REH_WITH_FOUR_DOTS_ABOVE = 0x0699;
const char32_t ARABIC_LETTER_SEEN_WITH_DOT_BELOW_AND_DOT_ABOVE = 0x069A;
const char32_t ARABIC_LETTER_SEEN_WITH_THREE_DOTS_BELOW = 0x069B;
const char32_t ARABIC_LETTER_SEEN_WITH_THREE_DOTS_BELOW_AND_THREE_DOTS_ABOVE = 0x069C;
const char32_t ARABIC_LETTER_SAD_WITH_TWO_DOTS_BELOW = 0x069D;
const char32_t ARABIC_LETTER_SAD_WITH_THREE_DOTS_ABOVE = 0x069E;
const char32_t ARABIC_LETTER_TAH_WITH_THREE_DOTS_ABOVE = 0x069F;
const char32_t ARABIC_LETTER_AIN_WITH_THREE_DOTS_ABOVE = 0x06A0;
const char32_t ARABIC_LETTER_DOTLESS_FEH = 0x06A1;
const char32_t ARABIC_LETTER_FEH_WITH_DOT_MOVED_BELOW = 0x06A2;
const char32_t ARABIC_LETTER_FEH_WITH_DOT_BELOW = 0x06A3;
const char32_t ARABIC_LETTER_VEH = 0x06A4;
const char32_t ARABIC_LETTER_FEH_WITH_THREE_DOTS_BELOW = 0x06A5;
const char32_t ARABIC_LETTER_PEHEH = 0x06A6;
const char32_t ARABIC_LETTER_QAF_WITH_DOT_ABOVE = 0x06A7;
const char32_t ARABIC_LETTER_QAF_WITH_THREE_DOTS_ABOVE = 0x06A8;
const char32_t ARABIC_LETTER_KEHEH = 0x06A9;
const char32_t ARABIC_LETTER_SWASH_KAF = 0x06AA;
const char32_t ARABIC_LETTER_KAF_WITH_RING = 0x06AB;
const char32_t ARABIC_LETTER_KAF_WITH_DOT_ABOVE = 0x06AC;
const char32_t ARABIC_LETTER_NG = 0x06AD;
const char32_t ARABIC_LETTER_KAF_WITH_THREE_DOTS_BELOW = 0x06AE;
const char32_t ARABIC_LETTER_GAF = 0x06AF;
const char32_t ARABIC_LETTER_GAF_WITH_RING = 0x06B0;
const char32_t ARABIC_LETTER_NGOEH = 0x06B1;
const char32_t ARABIC_LETTER_GAF_WITH_TWO_DOTS_BELOW = 0x06B2;
const char32_t ARABIC_LETTER_GUEH = 0x06B3;
const char32_t ARABIC_LETTER_GAF_WITH_THREE_DOTS_ABOVE = 0x06B4;
const char32_t ARABIC_LETTER_LAM_WITH_SMALL_V = 0x06B5;
const char32_t ARABIC_LETTER_LAM_WITH_DOT_ABOVE = 0x06B6;
const char32_t ARABIC_LETTER_LAM_WITH_THREE_DOTS_ABOVE = 0x06B7;
const char32_t ARABIC_LETTER_LAM_WITH_THREE_DOTS_BELOW = 0x06B8;
const char32_t ARABIC_LETTER_NOON_WITH_DOT_BELOW = 0x06B9;
const char32_t ARABIC_LETTER_NOON_GHUNNA = 0x06BA;
const char32_t ARABIC_LETTER_RNOON = 0x06BB;
const char32_t ARABIC_LETTER_NOON_WITH_RING = 0x06BC;
const char32_t ARABIC_LETTER_NOON_WITH_THREE_DOTS_ABOVE = 0x06BD;
const char32_t ARABIC_LETTER_HEH_DOACHASHMEE = 0x06BE;
const char32_t ARABIC_LETTER_TCHEH_WITH_DOT_ABOVE = 0x06BF;
const char32_t ARABIC_LETTER_HEH_WITH_YEH_ABOVE = 0x06C0;
const char32_t ARABIC_LETTER_HEH_GOAL = 0x06C1;
const char32_t ARABIC_LETTER_HEH_GOAL_WITH_HAMZA_ABOVE = 0x06C2;
const char32_t ARABIC_LETTER_TEH_MARBUTA_GOAL = 0x06C3;
const char32_t ARABIC_LETTER_WAW_WITH_RING = 0x06C4;
const char32_t ARABIC_LETTER_KIRGHIZ_OE = 0x06C5;
const char32_t ARABIC_LETTER_OE = 0x06C6;
const char32_t ARABIC_LETTER_U = 0x06C7;
const char32_t ARABIC_LETTER_YU = 0x06C8;
const char32_t ARABIC_LETTER_KIRGHIZ_YU = 0x06C9;
const char32_t ARABIC_LETTER_WAW_WITH_TWO_DOTS_ABOVE = 0x06CA;
const char32_t ARABIC_LETTER_VE = 0x06CB;
const char32_t ARABIC_LETTER_FARSI_YEH = 0x06CC;
const char32_t ARABIC_LETTER_YEH_WITH_TAIL = 0x06CD;
const char32_t ARABIC_LETTER_YEH_WITH_SMALL_V = 0x06CE;
const char32_t ARABIC_LETTER_WAW_WITH_DOT_ABOVE = 0x06CF;
const char32_t ARABIC_LETTER_E = 0x06D0;
const char32_t ARABIC_LETTER_YEH_WITH_THREE_DOTS_BELOW = 0x06D1;
const char32_t ARABIC_LETTER_YEH_BARREE = 0x06D2;
const char32_t ARABIC_LETTER_YEH_BARREE_WITH_HAMZA_ABOVE = 0x06D3;
const char32_t ARABIC_FULL_STOP = 0x06D4;
const char32_t ARABIC_LETTER_AE = 0x06D5;
const char32_t ARABIC_SMALL_HIGH_LIGATURE_SAD_WITH_LAM_WITH_ALEF_MAKSURA = 0x06D6;
const char32_t ARABIC_SMALL_HIGH_LIGATURE_QAF_WITH_LAM_WITH_ALEF_MAKSURA = 0x06D7;
const char32_t ARABIC_SMALL_HIGH_MEEM_INITIAL_FORM = 0x06D8;
const char32_t ARABIC_SMALL_HIGH_LAM_ALEF = 0x06D9;
const char32_t ARABIC_SMALL_HIGH_JEEM = 0x06DA;
const char32_t ARABIC_SMALL_HIGH_THREE_DOTS = 0x06DB;
const char32_t ARABIC_SMALL_HIGH_SEEN = 0x06DC;
const char32_t ARABIC_END_OF_AYAH = 0x06DD;
const char32_t ARABIC_START_OF_RUB_EL_HIZB = 0x06DE;
const char32_t ARABIC_SMALL_HIGH_ROUNDED_ZERO = 0x06DF;
const char32_t ARABIC_SMALL_HIGH_UPRIGHT_RECTANGULAR_ZERO = 0x06E0;
const char32_t ARABIC_SMALL_HIGH_DOTLESS_HEAD_OF_KHAH = 0x06E1;
const char32_t ARABIC_SMALL_HIGH_MEEM_ISOLATED_FORM = 0x06E2;
const char32_t ARABIC_SMALL_LOW_SEEN = 0x06E3;
const char32_t ARABIC_SMALL_HIGH_MADDA = 0x06E4;
const char32_t ARABIC_SMALL_WAW = 0x06E5;
const char32_t ARABIC_SMALL_YEH = 0x06E6;
const char32_t ARABIC_SMALL_HIGH_YEH = 0x06E7;
const char32_t ARABIC_SMALL_HIGH_NOON = 0x06E8;
const char32_t ARABIC_PLACE_OF_SAJDAH = 0x06E9;
const char32_t ARABIC_EMPTY_CENTRE_LOW_STOP = 0x06EA;
const char32_t ARABIC_EMPTY_CENTRE_HIGH_STOP = 0x06EB;
const char32_t ARABIC_ROUNDED_HIGH_STOP_WITH_FILLED_CENTRE = 0x06EC;
const char32_t ARABIC_SMALL_LOW_MEEM = 0x06ED;
const char32_t ARABIC_LETTER_DAL_WITH_INVERTED_V = 0x06EE;
const char32_t ARABIC_LETTER_REH_WITH_INVERTED_V = 0x06EF;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_ZERO = 0x06F0;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_ONE = 0x06F1;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_TWO = 0x06F2;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_THREE = 0x06F3;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_FOUR = 0x06F4;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_FIVE = 0x06F5;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_SIX = 0x06F6;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_SEVEN = 0x06F7;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_EIGHT = 0x06F8;
const char32_t EXTENDED_ARABIC_INDIC_DIGIT_NINE = 0x06F9;
const char32_t ARABIC_LETTER_SHEEN_WITH_DOT_BELOW = 0x06FA;
const char32_t ARABIC_LETTER_DAD_WITH_DOT_BELOW = 0x06FB;
const char32_t ARABIC_LETTER_GHAIN_WITH_DOT_BELOW = 0x06FC;
const char32_t ARABIC_SIGN_SINDHI_AMPERSAND = 0x06FD;
const char32_t ARABIC_SIGN_SINDHI_POSTPOSITION_MEN = 0x06FE;
const char32_t ARABIC_LETTER_HEH_WITH_INVERTED_V = 0x06FF;
const char32_t SYRIAC_END_OF_PARAGRAPH = 0x0700;
const char32_t SYRIAC_SUPRALINEAR_FULL_STOP = 0x0701;
const char32_t SYRIAC_SUBLINEAR_FULL_STOP = 0x0702;
const char32_t SYRIAC_SUPRALINEAR_COLON = 0x0703;
const char32_t SYRIAC_SUBLINEAR_COLON = 0x0704;
const char32_t SYRIAC_HORIZONTAL_COLON = 0x0705;
const char32_t SYRIAC_COLON_SKEWED_LEFT = 0x0706;
const char32_t SYRIAC_COLON_SKEWED_RIGHT = 0x0707;
const char32_t SYRIAC_SUPRALINEAR_COLON_SKEWED_LEFT = 0x0708;
const char32_t SYRIAC_SUBLINEAR_COLON_SKEWED_RIGHT = 0x0709;
const char32_t SYRIAC_CONTRACTION = 0x070A;
const char32_t SYRIAC_HARKLEAN_OBELUS = 0x070B;
const char32_t SYRIAC_HARKLEAN_METOBELUS = 0x070C;
const char32_t SYRIAC_HARKLEAN_ASTERISCUS = 0x070D;
const char32_t SYRIAC_ABBREVIATION_MARK = 0x070F;
const char32_t SYRIAC_LETTER_ALAPH = 0x0710;
const char32_t SYRIAC_LETTER_SUPERSCRIPT_ALAPH = 0x0711;
const char32_t SYRIAC_LETTER_BETH = 0x0712;
const char32_t SYRIAC_LETTER_GAMAL = 0x0713;
const char32_t SYRIAC_LETTER_GAMAL_GARSHUNI = 0x0714;
const char32_t SYRIAC_LETTER_DALATH = 0x0715;
const char32_t SYRIAC_LETTER_DOTLESS_DALATH_RISH = 0x0716;
const char32_t SYRIAC_LETTER_HE = 0x0717;
const char32_t SYRIAC_LETTER_WAW = 0x0718;
const char32_t SYRIAC_LETTER_ZAIN = 0x0719;
const char32_t SYRIAC_LETTER_HETH = 0x071A;
const char32_t SYRIAC_LETTER_TETH = 0x071B;
const char32_t SYRIAC_LETTER_TETH_GARSHUNI = 0x071C;
const char32_t SYRIAC_LETTER_YUDH = 0x071D;
const char32_t SYRIAC_LETTER_YUDH_HE = 0x071E;
const char32_t SYRIAC_LETTER_KAPH = 0x071F;
const char32_t SYRIAC_LETTER_LAMADH = 0x0720;
const char32_t SYRIAC_LETTER_MIM = 0x0721;
const char32_t SYRIAC_LETTER_NUN = 0x0722;
const char32_t SYRIAC_LETTER_SEMKATH = 0x0723;
const char32_t SYRIAC_LETTER_FINAL_SEMKATH = 0x0724;
const char32_t SYRIAC_LETTER_E = 0x0725;
const char32_t SYRIAC_LETTER_PE = 0x0726;
const char32_t SYRIAC_LETTER_REVERSED_PE = 0x0727;
const char32_t SYRIAC_LETTER_SADHE = 0x0728;
const char32_t SYRIAC_LETTER_QAPH = 0x0729;
const char32_t SYRIAC_LETTER_RISH = 0x072A;
const char32_t SYRIAC_LETTER_SHIN = 0x072B;
const char32_t SYRIAC_LETTER_TAW = 0x072C;
const char32_t SYRIAC_LETTER_PERSIAN_BHETH = 0x072D;
const char32_t SYRIAC_LETTER_PERSIAN_GHAMAL = 0x072E;
const char32_t SYRIAC_LETTER_PERSIAN_DHALATH = 0x072F;
const char32_t SYRIAC_PTHAHA_ABOVE = 0x0730;
const char32_t SYRIAC_PTHAHA_BELOW = 0x0731;
const char32_t SYRIAC_PTHAHA_DOTTED = 0x0732;
const char32_t SYRIAC_ZQAPHA_ABOVE = 0x0733;
const char32_t SYRIAC_ZQAPHA_BELOW = 0x0734;
const char32_t SYRIAC_ZQAPHA_DOTTED = 0x0735;
const char32_t SYRIAC_RBASA_ABOVE = 0x0736;
const char32_t SYRIAC_RBASA_BELOW = 0x0737;
const char32_t SYRIAC_DOTTED_ZLAMA_HORIZONTAL = 0x0738;
const char32_t SYRIAC_DOTTED_ZLAMA_ANGULAR = 0x0739;
const char32_t SYRIAC_HBASA_ABOVE = 0x073A;
const char32_t SYRIAC_HBASA_BELOW = 0x073B;
const char32_t SYRIAC_HBASA_ESASA_DOTTED = 0x073C;
const char32_t SYRIAC_ESASA_ABOVE = 0x073D;
const char32_t SYRIAC_ESASA_BELOW = 0x073E;
const char32_t SYRIAC_RWAHA = 0x073F;
const char32_t SYRIAC_FEMININE_DOT = 0x0740;
const char32_t SYRIAC_QUSHSHAYA = 0x0741;
const char32_t SYRIAC_RUKKAKHA = 0x0742;
const char32_t SYRIAC_TWO_VERTICAL_DOTS_ABOVE = 0x0743;
const char32_t SYRIAC_TWO_VERTICAL_DOTS_BELOW = 0x0744;
const char32_t SYRIAC_THREE_DOTS_ABOVE = 0x0745;
const char32_t SYRIAC_THREE_DOTS_BELOW = 0x0746;
const char32_t SYRIAC_OBLIQUE_LINE_ABOVE = 0x0747;
const char32_t SYRIAC_OBLIQUE_LINE_BELOW = 0x0748;
const char32_t SYRIAC_MUSIC = 0x0749;
const char32_t SYRIAC_BARREKH = 0x074A;
const char32_t SYRIAC_LETTER_SOGDIAN_ZHAIN = 0x074D;
const char32_t SYRIAC_LETTER_SOGDIAN_KHAPH = 0x074E;
const char32_t SYRIAC_LETTER_SOGDIAN_FE = 0x074F;
const char32_t ARABIC_LETTER_BEH_WITH_THREE_DOTS_HORIZONTALLY_BELOW = 0x0750;
const char32_t ARABIC_LETTER_BEH_WITH_DOT_BELOW_AND_THREE_DOTS_ABOVE = 0x0751;
const char32_t ARABIC_LETTER_BEH_WITH_THREE_DOTS_POINTING_UPWARDS_BELOW = 0x0752;
const char32_t ARABIC_LETTER_BEH_WITH_THREE_DOTS_POINTING_UPWARDS_BELOW_AND_TWO_DOTS_ABOVE = 0x0753;
const char32_t ARABIC_LETTER_BEH_WITH_TWO_DOTS_BELOW_AND_DOT_ABOVE = 0x0754;
const char32_t ARABIC_LETTER_BEH_WITH_INVERTED_SMALL_V_BELOW = 0x0755;
const char32_t ARABIC_LETTER_BEH_WITH_SMALL_V = 0x0756;
const char32_t ARABIC_LETTER_HAH_WITH_TWO_DOTS_ABOVE = 0x0757;
const char32_t ARABIC_LETTER_HAH_WITH_THREE_DOTS_POINTING_UPWARDS_BELOW = 0x0758;
const char32_t ARABIC_LETTER_DAL_WITH_TWO_DOTS_VERTICALLY_BELOW_AND_SMALL_TAH = 0x0759;
const char32_t ARABIC_LETTER_DAL_WITH_INVERTED_SMALL_V_BELOW = 0x075A;
const char32_t ARABIC_LETTER_REH_WITH_STROKE = 0x075B;
const char32_t ARABIC_LETTER_SEEN_WITH_FOUR_DOTS_ABOVE = 0x075C;
const char32_t ARABIC_LETTER_AIN_WITH_TWO_DOTS_ABOVE = 0x075D;
const char32_t ARABIC_LETTER_AIN_WITH_THREE_DOTS_POINTING_DOWNWARDS_ABOVE = 0x075E;
const char32_t ARABIC_LETTER_AIN_WITH_TWO_DOTS_VERTICALLY_ABOVE = 0x075F;
const char32_t ARABIC_LETTER_FEH_WITH_TWO_DOTS_BELOW = 0x0760;
const char32_t ARABIC_LETTER_FEH_WITH_THREE_DOTS_POINTING_UPWARDS_BELOW = 0x0761;
const char32_t ARABIC_LETTER_KEHEH_WITH_DOT_ABOVE = 0x0762;
const char32_t ARABIC_LETTER_KEHEH_WITH_THREE_DOTS_ABOVE = 0x0763;
const char32_t ARABIC_LETTER_KEHEH_WITH_THREE_DOTS_POINTING_UPWARDS_BELOW = 0x0764;
const char32_t ARABIC_LETTER_MEEM_WITH_DOT_ABOVE = 0x0765;
const char32_t ARABIC_LETTER_MEEM_WITH_DOT_BELOW = 0x0766;
const char32_t ARABIC_LETTER_NOON_WITH_TWO_DOTS_BELOW = 0x0767;
const char32_t ARABIC_LETTER_NOON_WITH_SMALL_TAH = 0x0768;
const char32_t ARABIC_LETTER_NOON_WITH_SMALL_V = 0x0769;
const char32_t ARABIC_LETTER_LAM_WITH_BAR = 0x076A;
const char32_t ARABIC_LETTER_REH_WITH_TWO_DOTS_VERTICALLY_ABOVE = 0x076B;
const char32_t ARABIC_LETTER_REH_WITH_HAMZA_ABOVE = 0x076C;
const char32_t ARABIC_LETTER_SEEN_WITH_TWO_DOTS_VERTICALLY_ABOVE = 0x076D;
const char32_t ARABIC_LETTER_HAH_WITH_SMALL_ARABIC_LETTER_TAH_BELOW = 0x076E;
const char32_t ARABIC_LETTER_HAH_WITH_SMALL_ARABIC_LETTER_TAH_AND_TWO_DOTS = 0x076F;
const char32_t ARABIC_LETTER_SEEN_WITH_SMALL_ARABIC_LETTER_TAH_AND_TWO_DOTS = 0x0770;
const char32_t ARABIC_LETTER_REH_WITH_SMALL_ARABIC_LETTER_TAH_AND_TWO_DOTS = 0x0771;
const char32_t ARABIC_LETTER_HAH_WITH_SMALL_ARABIC_LETTER_TAH_ABOVE = 0x0772;
const char32_t ARABIC_LETTER_ALEF_WITH_EXTENDED_ARABIC_INDIC_DIGIT_TWO_ABOVE = 0x0773;
const char32_t ARABIC_LETTER_ALEF_WITH_EXTENDED_ARABIC_INDIC_DIGIT_THREE_ABOVE = 0x0774;
const char32_t ARABIC_LETTER_FARSI_YEH_WITH_EXTENDED_ARABIC_INDIC_DIGIT_TWO_ABOVE = 0x0775;
const char32_t ARABIC_LETTER_FARSI_YEH_WITH_EXTENDED_ARABIC_INDIC_DIGIT_THREE_ABOVE = 0x0776;
const char32_t ARABIC_LETTER_FARSI_YEH_WITH_EXTENDED_ARABIC_INDIC_DIGIT_FOUR_BELOW = 0x0777;
const char32_t ARABIC_LETTER_WAW_WITH_EXTENDED_ARABIC_INDIC_DIGIT_TWO_ABOVE = 0x0778;
const char32_t ARABIC_LETTER_WAW_WITH_EXTENDED_ARABIC_INDIC_DIGIT_THREE_ABOVE = 0x0779;
const char32_t ARABIC_LETTER_YEH_BARREE_WITH_EXTENDED_ARABIC_INDIC_DIGIT_TWO_ABOVE = 0x077A;
const char32_t ARABIC_LETTER_YEH_BARREE_WITH_EXTENDED_ARABIC_INDIC_DIGIT_THREE_ABOVE = 0x077B;
const char32_t ARABIC_LETTER_HAH_WITH_EXTENDED_ARABIC_INDIC_DIGIT_FOUR_BELOW = 0x077C;
const char32_t ARABIC_LETTER_SEEN_WITH_EXTENDED_ARABIC_INDIC_DIGIT_FOUR_ABOVE = 0x077D;
const char32_t ARABIC_LETTER_SEEN_WITH_INVERTED_V = 0x077E;
const char32_t ARABIC_LETTER_KAF_WITH_TWO_DOTS_ABOVE = 0x077F;
const char32_t THAANA_LETTER_HAA = 0x0780;
const char32_t THAANA_LETTER_SHAVIYANI = 0x0781;
const char32_t THAANA_LETTER_NOONU = 0x0782;
const char32_t THAANA_LETTER_RAA = 0x0783;
const char32_t THAANA_LETTER_BAA = 0x0784;
const char32_t THAANA_LETTER_LHAVIYANI = 0x0785;
const char32_t THAANA_LETTER_KAAFU = 0x0786;
const char32_t THAANA_LETTER_ALIFU = 0x0787;
const char32_t THAANA_LETTER_VAAVU = 0x0788;
const char32_t THAANA_LETTER_MEEMU = 0x0789;
const char32_t THAANA_LETTER_FAAFU = 0x078A;
const char32_t THAANA_LETTER_DHAALU = 0x078B;
const char32_t THAANA_LETTER_THAA = 0x078C;
const char32_t THAANA_LETTER_LAAMU = 0x078D;
const char32_t THAANA_LETTER_GAAFU = 0x078E;
const char32_t THAANA_LETTER_GNAVIYANI = 0x078F;
const char32_t THAANA_LETTER_SEENU = 0x0790;
const char32_t THAANA_LETTER_DAVIYANI = 0x0791;
const char32_t THAANA_LETTER_ZAVIYANI = 0x0792;
const char32_t THAANA_LETTER_TAVIYANI = 0x0793;
const char32_t THAANA_LETTER_YAA = 0x0794;
const char32_t THAANA_LETTER_PAVIYANI = 0x0795;
const char32_t THAANA_LETTER_JAVIYANI = 0x0796;
const char32_t THAANA_LETTER_CHAVIYANI = 0x0797;
const char32_t THAANA_LETTER_TTAA = 0x0798;
const char32_t THAANA_LETTER_HHAA = 0x0799;
const char32_t THAANA_LETTER_KHAA = 0x079A;
const char32_t THAANA_LETTER_THAALU = 0x079B;
const char32_t THAANA_LETTER_ZAA = 0x079C;
const char32_t THAANA_LETTER_SHEENU = 0x079D;
const char32_t THAANA_LETTER_SAADHU = 0x079E;
const char32_t THAANA_LETTER_DAADHU = 0x079F;
const char32_t THAANA_LETTER_TO = 0x07A0;
const char32_t THAANA_LETTER_ZO = 0x07A1;
const char32_t THAANA_LETTER_AINU = 0x07A2;
const char32_t THAANA_LETTER_GHAINU = 0x07A3;
const char32_t THAANA_LETTER_QAAFU = 0x07A4;
const char32_t THAANA_LETTER_WAAVU = 0x07A5;
const char32_t THAANA_ABAFILI = 0x07A6;
const char32_t THAANA_AABAAFILI = 0x07A7;
const char32_t THAANA_IBIFILI = 0x07A8;
const char32_t THAANA_EEBEEFILI = 0x07A9;
const char32_t THAANA_UBUFILI = 0x07AA;
const char32_t THAANA_OOBOOFILI = 0x07AB;
const char32_t THAANA_EBEFILI = 0x07AC;
const char32_t THAANA_EYBEYFILI = 0x07AD;
const char32_t THAANA_OBOFILI = 0x07AE;
const char32_t THAANA_OABOAFILI = 0x07AF;
const char32_t THAANA_SUKUN = 0x07B0;
const char32_t THAANA_LETTER_NAA = 0x07B1;
const char32_t NKO_DIGIT_ZERO = 0x07C0;
const char32_t NKO_DIGIT_ONE = 0x07C1;
const char32_t NKO_DIGIT_TWO = 0x07C2;
const char32_t NKO_DIGIT_THREE = 0x07C3;
const char32_t NKO_DIGIT_FOUR = 0x07C4;
const char32_t NKO_DIGIT_FIVE = 0x07C5;
const char32_t NKO_DIGIT_SIX = 0x07C6;
const char32_t NKO_DIGIT_SEVEN = 0x07C7;
const char32_t NKO_DIGIT_EIGHT = 0x07C8;
const char32_t NKO_DIGIT_NINE = 0x07C9;
const char32_t NKO_LETTER_A = 0x07CA;
const char32_t NKO_LETTER_EE = 0x07CB;
const char32_t NKO_LETTER_I = 0x07CC;
const char32_t NKO_LETTER_E = 0x07CD;
const char32_t NKO_LETTER_U = 0x07CE;
const char32_t NKO_LETTER_OO = 0x07CF;
const char32_t NKO_LETTER_O = 0x07D0;
const char32_t NKO_LETTER_DAGBASINNA = 0x07D1;
const char32_t NKO_LETTER_N = 0x07D2;
const char32_t NKO_LETTER_BA = 0x07D3;
const char32_t NKO_LETTER_PA = 0x07D4;
const char32_t NKO_LETTER_TA = 0x07D5;
const char32_t NKO_LETTER_JA = 0x07D6;
const char32_t NKO_LETTER_CHA = 0x07D7;
const char32_t NKO_LETTER_DA = 0x07D8;
const char32_t NKO_LETTER_RA = 0x07D9;
const char32_t NKO_LETTER_RRA = 0x07DA;
const char32_t NKO_LETTER_SA = 0x07DB;
const char32_t NKO_LETTER_GBA = 0x07DC;
const char32_t NKO_LETTER_FA = 0x07DD;
const char32_t NKO_LETTER_KA = 0x07DE;
const char32_t NKO_LETTER_LA = 0x07DF;
const char32_t NKO_LETTER_NA_WOLOSO = 0x07E0;
const char32_t NKO_LETTER_MA = 0x07E1;
const char32_t NKO_LETTER_NYA = 0x07E2;
const char32_t NKO_LETTER_NA = 0x07E3;
const char32_t NKO_LETTER_HA = 0x07E4;
const char32_t NKO_LETTER_WA = 0x07E5;
const char32_t NKO_LETTER_YA = 0x07E6;
const char32_t NKO_LETTER_NYA_WOLOSO = 0x07E7;
const char32_t NKO_LETTER_JONA_JA = 0x07E8;
const char32_t NKO_LETTER_JONA_CHA = 0x07E9;
const char32_t NKO_LETTER_JONA_RA = 0x07EA;
const char32_t NKO_COMBINING_SHORT_HIGH_TONE = 0x07EB;
const char32_t NKO_COMBINING_SHORT_LOW_TONE = 0x07EC;
const char32_t NKO_COMBINING_SHORT_RISING_TONE = 0x07ED;
const char32_t NKO_COMBINING_LONG_DESCENDING_TONE = 0x07EE;
const char32_t NKO_COMBINING_LONG_HIGH_TONE = 0x07EF;
const char32_t NKO_COMBINING_LONG_LOW_TONE = 0x07F0;
const char32_t NKO_COMBINING_LONG_RISING_TONE = 0x07F1;
const char32_t NKO_COMBINING_NASALIZATION_MARK = 0x07F2;
const char32_t NKO_COMBINING_DOUBLE_DOT_ABOVE = 0x07F3;
const char32_t NKO_HIGH_TONE_APOSTROPHE = 0x07F4;
const char32_t NKO_LOW_TONE_APOSTROPHE = 0x07F5;
const char32_t NKO_SYMBOL_OO_DENNEN = 0x07F6;
const char32_t NKO_SYMBOL_GBAKURUNEN = 0x07F7;
const char32_t NKO_COMMA = 0x07F8;
const char32_t NKO_EXCLAMATION_MARK = 0x07F9;
const char32_t NKO_LAJANYALAN = 0x07FA;
const char32_t NKO_DANTAYALAN = 0x07FD;
const char32_t NKO_DOROME_SIGN = 0x07FE;
const char32_t NKO_TAMAN_SIGN = 0x07FF;
const char32_t SAMARITAN_LETTER_ALAF = 0x0800;
const char32_t SAMARITAN_LETTER_BIT = 0x0801;
const char32_t SAMARITAN_LETTER_GAMAN = 0x0802;
const char32_t SAMARITAN_LETTER_DALAT = 0x0803;
const char32_t SAMARITAN_LETTER_IY = 0x0804;
const char32_t SAMARITAN_LETTER_BAA = 0x0805;
const char32_t SAMARITAN_LETTER_ZEN = 0x0806;
const char32_t SAMARITAN_LETTER_IT = 0x0807;
const char32_t SAMARITAN_LETTER_TIT = 0x0808;
const char32_t SAMARITAN_LETTER_YUT = 0x0809;
const char32_t SAMARITAN_LETTER_KAAF = 0x080A;
const char32_t SAMARITAN_LETTER_LABAT = 0x080B;
const char32_t SAMARITAN_LETTER_MIM = 0x080C;
const char32_t SAMARITAN_LETTER_NUN = 0x080D;
const char32_t SAMARITAN_LETTER_SINGAAT = 0x080E;
const char32_t SAMARITAN_LETTER_IN = 0x080F;
const char32_t SAMARITAN_LETTER_FI = 0x0810;
const char32_t SAMARITAN_LETTER_TSAADIY = 0x0811;
const char32_t SAMARITAN_LETTER_QUF = 0x0812;
const char32_t SAMARITAN_LETTER_RISH = 0x0813;
const char32_t SAMARITAN_LETTER_SHAN = 0x0814;
const char32_t SAMARITAN_LETTER_TAAF = 0x0815;
const char32_t SAMARITAN_MARK_IN = 0x0816;
const char32_t SAMARITAN_MARK_IN_ALAF = 0x0817;
const char32_t SAMARITAN_MARK_OCCLUSION = 0x0818;
const char32_t SAMARITAN_MARK_DAGESH = 0x0819;
const char32_t SAMARITAN_MODIFIER_LETTER_EPENTHETIC_YUT = 0x081A;
const char32_t SAMARITAN_MARK_EPENTHETIC_YUT = 0x081B;
const char32_t SAMARITAN_VOWEL_SIGN_LONG_E = 0x081C;
const char32_t SAMARITAN_VOWEL_SIGN_E = 0x081D;
const char32_t SAMARITAN_VOWEL_SIGN_OVERLONG_AA = 0x081E;
const char32_t SAMARITAN_VOWEL_SIGN_LONG_AA = 0x081F;
const char32_t SAMARITAN_VOWEL_SIGN_AA = 0x0820;
const char32_t SAMARITAN_VOWEL_SIGN_OVERLONG_A = 0x0821;
const char32_t SAMARITAN_VOWEL_SIGN_LONG_A = 0x0822;
const char32_t SAMARITAN_VOWEL_SIGN_A = 0x0823;
const char32_t SAMARITAN_MODIFIER_LETTER_SHORT_A = 0x0824;
const char32_t SAMARITAN_VOWEL_SIGN_SHORT_A = 0x0825;
const char32_t SAMARITAN_VOWEL_SIGN_LONG_U = 0x0826;
const char32_t SAMARITAN_VOWEL_SIGN_U = 0x0827;
const char32_t SAMARITAN_MODIFIER_LETTER_I = 0x0828;
const char32_t SAMARITAN_VOWEL_SIGN_LONG_I = 0x0829;
const char32_t SAMARITAN_VOWEL_SIGN_I = 0x082A;
const char32_t SAMARITAN_VOWEL_SIGN_O = 0x082B;
const char32_t SAMARITAN_VOWEL_SIGN_SUKUN = 0x082C;
const char32_t SAMARITAN_MARK_NEQUDAA = 0x082D;
const char32_t SAMARITAN_PUNCTUATION_NEQUDAA = 0x0830;
const char32_t SAMARITAN_PUNCTUATION_AFSAAQ = 0x0831;
const char32_t SAMARITAN_PUNCTUATION_ANGED = 0x0832;
const char32_t SAMARITAN_PUNCTUATION_BAU = 0x0833;
const char32_t SAMARITAN_PUNCTUATION_ATMAAU = 0x0834;
const char32_t SAMARITAN_PUNCTUATION_SHIYYAALAA = 0x0835;
const char32_t SAMARITAN_ABBREVIATION_MARK = 0x0836;
const char32_t SAMARITAN_PUNCTUATION_MELODIC_QITSA = 0x0837;
const char32_t SAMARITAN_PUNCTUATION_ZIQAA = 0x0838;
const char32_t SAMARITAN_PUNCTUATION_QITSA = 0x0839;
const char32_t SAMARITAN_PUNCTUATION_ZAEF = 0x083A;
const char32_t SAMARITAN_PUNCTUATION_TURU = 0x083B;
const char32_t SAMARITAN_PUNCTUATION_ARKAANU = 0x083C;
const char32_t SAMARITAN_PUNCTUATION_SOF_MASHFAAT = 0x083D;
const char32_t SAMARITAN_PUNCTUATION_ANNAAU = 0x083E;
const char32_t MANDAIC_LETTER_HALQA = 0x0840;
const char32_t MANDAIC_LETTER_AB = 0x0841;
const char32_t MANDAIC_LETTER_AG = 0x0842;
const char32_t MANDAIC_LETTER_AD = 0x0843;
const char32_t MANDAIC_LETTER_AH = 0x0844;
const char32_t MANDAIC_LETTER_USHENNA = 0x0845;
const char32_t MANDAIC_LETTER_AZ = 0x0846;
const char32_t MANDAIC_LETTER_IT = 0x0847;
const char32_t MANDAIC_LETTER_ATT = 0x0848;
const char32_t MANDAIC_LETTER_AKSA = 0x0849;
const char32_t MANDAIC_LETTER_AK = 0x084A;
const char32_t MANDAIC_LETTER_AL = 0x084B;
const char32_t MANDAIC_LETTER_AM = 0x084C;
const char32_t MANDAIC_LETTER_AN = 0x084D;
const char32_t MANDAIC_LETTER_AS = 0x084E;
const char32_t MANDAIC_LETTER_IN = 0x084F;
const char32_t MANDAIC_LETTER_AP = 0x0850;
const char32_t MANDAIC_LETTER_ASZ = 0x0851;
const char32_t MANDAIC_LETTER_AQ = 0x0852;
const char32_t MANDAIC_LETTER_AR = 0x0853;
const char32_t MANDAIC_LETTER_ASH = 0x0854;
const char32_t MANDAIC_LETTER_AT = 0x0855;
const char32_t MANDAIC_LETTER_DUSHENNA = 0x0856;
const char32_t MANDAIC_LETTER_KAD = 0x0857;
const char32_t MANDAIC_LETTER_AIN = 0x0858;
const char32_t MANDAIC_AFFRICATION_MARK = 0x0859;
const char32_t MANDAIC_VOCALIZATION_MARK = 0x085A;
const char32_t MANDAIC_GEMINATION_MARK = 0x085B;
const char32_t MANDAIC_PUNCTUATION = 0x085E;
const char32_t SYRIAC_LETTER_MALAYALAM_NGA = 0x0860;
const char32_t SYRIAC_LETTER_MALAYALAM_JA = 0x0861;
const char32_t SYRIAC_LETTER_MALAYALAM_NYA = 0x0862;
const char32_t SYRIAC_LETTER_MALAYALAM_TTA = 0x0863;
const char32_t SYRIAC_LETTER_MALAYALAM_NNA = 0x0864;
const char32_t SYRIAC_LETTER_MALAYALAM_NNNA = 0x0865;
const char32_t SYRIAC_LETTER_MALAYALAM_BHA = 0x0866;
const char32_t SYRIAC_LETTER_MALAYALAM_RA = 0x0867;
const char32_t SYRIAC_LETTER_MALAYALAM_LLA = 0x0868;
const char32_t SYRIAC_LETTER_MALAYALAM_LLLA = 0x0869;
const char32_t SYRIAC_LETTER_MALAYALAM_SSA = 0x086A;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_FATHA = 0x0870;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_TOP_RIGHT_FATHA = 0x0871;
const char32_t ARABIC_LETTER_ALEF_WITH_RIGHT_MIDDLE_STROKE = 0x0872;
const char32_t ARABIC_LETTER_ALEF_WITH_LEFT_MIDDLE_STROKE = 0x0873;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_KASRA = 0x0874;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_BOTTOM_RIGHT_KASRA = 0x0875;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_ROUND_DOT_ABOVE = 0x0876;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_RIGHT_ROUND_DOT = 0x0877;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_LEFT_ROUND_DOT = 0x0878;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_ROUND_DOT_BELOW = 0x0879;
const char32_t ARABIC_LETTER_ALEF_WITH_DOT_ABOVE = 0x087A;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_TOP_RIGHT_FATHA_AND_DOT_ABOVE = 0x087B;
const char32_t ARABIC_LETTER_ALEF_WITH_RIGHT_MIDDLE_STROKE_AND_DOT_ABOVE = 0x087C;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_BOTTOM_RIGHT_KASRA_AND_DOT_ABOVE = 0x087D;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_TOP_RIGHT_FATHA_AND_LEFT_RING = 0x087E;
const char32_t ARABIC_LETTER_ALEF_WITH_RIGHT_MIDDLE_STROKE_AND_LEFT_RING = 0x087F;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_BOTTOM_RIGHT_KASRA_AND_LEFT_RING = 0x0880;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_RIGHT_HAMZA = 0x0881;
const char32_t ARABIC_LETTER_ALEF_WITH_ATTACHED_LEFT_HAMZA = 0x0882;
const char32_t ARABIC_TATWEEL_WITH_OVERSTRUCK_HAMZA = 0x0883;
const char32_t ARABIC_TATWEEL_WITH_OVERSTRUCK_WAW = 0x0884;
const char32_t ARABIC_TATWEEL_WITH_TWO_DOTS_BELOW = 0x0885;
const char32_t ARABIC_LETTER_THIN_YEH = 0x0886;
const char32_t ARABIC_BASELINE_ROUND_DOT = 0x0887;
const char32_t ARABIC_RAISED_ROUND_DOT = 0x0888;
const char32_t ARABIC_LETTER_NOON_WITH_INVERTED_SMALL_V = 0x0889;
const char32_t ARABIC_LETTER_HAH_WITH_INVERTED_SMALL_V_BELOW = 0x088A;
const char32_t ARABIC_LETTER_TAH_WITH_DOT_BELOW = 0x088B;
const char32_t ARABIC_LETTER_TAH_WITH_THREE_DOTS_BELOW = 0x088C;
const char32_t ARABIC_LETTER_KEHEH_WITH_TWO_DOTS_VERTICALLY_BELOW = 0x088D;
const char32_t ARABIC_VERTICAL_TAIL = 0x088E;
const char32_t ARABIC_LETTER_NOON_WITH_RING_ABOVE = 0x088F;
const char32_t ARABIC_POUND_MARK_ABOVE = 0x0890;
const char32_t ARABIC_PIASTRE_MARK_ABOVE = 0x0891;
const char32_t ARABIC_PEPET = 0x0897;
const char32_t ARABIC_SMALL_HIGH_WORD_AL_JUZ = 0x0898;
const char32_t ARABIC_SMALL_LOW_WORD_ISHMAAM = 0x0899;
const char32_t ARABIC_SMALL_LOW_WORD_IMAALA = 0x089A;
const char32_t ARABIC_SMALL_LOW_WORD_TASHEEL = 0x089B;
const char32_t ARABIC_MADDA_WAAJIB = 0x089C;
const char32_t ARABIC_SUPERSCRIPT_ALEF_MOKHASSAS = 0x089D;
const char32_t ARABIC_DOUBLED_MADDA = 0x089E;
const char32_t ARABIC_HALF_MADDA_OVER_MADDA = 0x089F;
const char32_t ARABIC_LETTER_BEH_WITH_SMALL_V_BELOW = 0x08A0;
const char32_t ARABIC_LETTER_BEH_WITH_HAMZA_ABOVE = 0x08A1;
const char32_t ARABIC_LETTER_JEEM_WITH_TWO_DOTS_ABOVE = 0x08A2;
const char32_t ARABIC_LETTER_TAH_WITH_TWO_DOTS_ABOVE = 0x08A3;
const char32_t ARABIC_LETTER_FEH_WITH_DOT_BELOW_AND_THREE_DOTS_ABOVE = 0x08A4;
const char32_t ARABIC_LETTER_QAF_WITH_DOT_BELOW = 0x08A5;
const char32_t ARABIC_LETTER_LAM_WITH_DOUBLE_BAR = 0x08A6;
const char32_t ARABIC_LETTER_MEEM_WITH_THREE_DOTS_ABOVE = 0x08A7;
const char32_t ARABIC_LETTER_YEH_WITH_TWO_DOTS_BELOW_AND_HAMZA_ABOVE = 0x08A8;
const char32_t ARABIC_LETTER_YEH_WITH_TWO_DOTS_BELOW_AND_DOT_ABOVE = 0x08A9;
const char32_t ARABIC_LETTER_REH_WITH_LOOP = 0x08AA;
const char32_t ARABIC_LETTER_WAW_WITH_DOT_WITHIN = 0x08AB;
const char32_t ARABIC_LETTER_ROHINGYA_YEH = 0x08AC;
const char32_t ARABIC_LETTER_LOW_ALEF = 0x08AD;
const char32_t ARABIC_LETTER_DAL_WITH_THREE_DOTS_BELOW = 0x08AE;
const char32_t ARABIC_LETTER_SAD_WITH_THREE_DOTS_BELOW = 0x08AF;
const char32_t ARABIC_LETTER_GAF_WITH_INVERTED_STROKE = 0x08B0;
const char32_t ARABIC_LETTER_STRAIGHT_WAW = 0x08B1;
const char32_t ARABIC_LETTER_ZAIN_WITH_INVERTED_V_ABOVE = 0x08B2;
const char32_t ARABIC_LETTER_AIN_WITH_THREE_DOTS_BELOW = 0x08B3;
const char32_t ARABIC_LETTER_KAF_WITH_DOT_BELOW = 0x08B4;
const char32_t ARABIC_LETTER_QAF_WITH_DOT_BELOW_AND_NO_DOTS_ABOVE = 0x08B5;
const char32_t ARABIC_LETTER_BEH_WITH_SMALL_MEEM_ABOVE = 0x08B6;
const char32_t ARABIC_LETTER_PEH_WITH_SMALL_MEEM_ABOVE = 0x08B7;
const char32_t ARABIC_LETTER_TEH_WITH_SMALL_TEH_ABOVE = 0x08B8;
const char32_t ARABIC_LETTER_REH_WITH_SMALL_NOON_ABOVE = 0x08B9;
const char32_t ARABIC_LETTER_YEH_WITH_TWO_DOTS_BELOW_AND_SMALL_NOON_ABOVE = 0x08BA;
const char32_t ARABIC_LETTER_AFRICAN_FEH = 0x08BB;
const char32_t ARABIC_LETTER_AFRICAN_QAF = 0x08BC;
const char32_t ARABIC_LETTER_AFRICAN_NOON = 0x08BD;
const char32_t ARABIC_LETTER_PEH_WITH_SMALL_V = 0x08BE;
const char32_t ARABIC_LETTER_TEH_WITH_SMALL_V = 0x08BF;
const char32_t ARABIC_LETTER_TTEH_WITH_SMALL_V = 0x08C0;
const char32_t ARABIC_LETTER_TCHEH_WITH_SMALL_V = 0x08C1;
const char32_t ARABIC_LETTER_KEHEH_WITH_SMALL_V = 0x08C2;
const char32_t ARABIC_LETTER_GHAIN_WITH_THREE_DOTS_ABOVE = 0x08C3;
const char32_t ARABIC_LETTER_AFRICAN_QAF_WITH_THREE_DOTS_ABOVE = 0x08C4;
const char32_t ARABIC_LETTER_JEEM_WITH_THREE_DOTS_ABOVE = 0x08C5;
const char32_t ARABIC_LETTER_JEEM_WITH_THREE_DOTS_BELOW = 0x08C6;
const char32_t ARABIC_LETTER_LAM_WITH_SMALL_ARABIC_LETTER_TAH_ABOVE = 0x08C7;
const char32_t ARABIC_LETTER_GRAF = 0x08C8;
const char32_t ARABIC_SMALL_FARSI_YEH = 0x08C9;
const char32_t ARABIC_SMALL_HIGH_FARSI_YEH = 0x08CA;
const char32_t ARABIC_SMALL_HIGH_YEH_BARREE_WITH_TWO_DOTS_BELOW = 0x08CB;
const char32_t ARABIC_SMALL_HIGH_WORD_SAH = 0x08CC;
const char32_t ARABIC_SMALL_HIGH_ZAH = 0x08CD;
const char32_t ARABIC_LARGE_ROUND_DOT_ABOVE = 0x08CE;
const char32_t ARABIC_LARGE_ROUND_DOT_BELOW = 0x08CF;
const char32_t ARABIC_SUKUN_BELOW = 0x08D0;
const char32_t ARABIC_LARGE_CIRCLE_BELOW = 0x08D1;
const char32_t ARABIC_LARGE_ROUND_DOT_INSIDE_CIRCLE_BELOW = 0x08D2;
const char32_t ARABIC_SMALL_LOW_WAW = 0x08D3;
const char32_t ARABIC_SMALL_HIGH_WORD_AR_RUB = 0x08D4;
const char32_t ARABIC_SMALL_HIGH_SAD = 0x08D5;
const char32_t ARABIC_SMALL_HIGH_AIN = 0x08D6;
const char32_t ARABIC_SMALL_HIGH_QAF = 0x08D7;
const char32_t ARABIC_SMALL_HIGH_NOON_WITH_KASRA = 0x08D8;
const char32_t ARABIC_SMALL_LOW_NOON_WITH_KASRA = 0x08D9;
const char32_t ARABIC_SMALL_HIGH_WORD_ATH_THALATHA = 0x08DA;
const char32_t ARABIC_SMALL_HIGH_WORD_AS_SAJDA = 0x08DB;
const char32_t ARABIC_SMALL_HIGH_WORD_AN_NISF = 0x08DC;
const char32_t ARABIC_SMALL_HIGH_WORD_SAKTA = 0x08DD;
const char32_t ARABIC_SMALL_HIGH_WORD_QIF = 0x08DE;
const char32_t ARABIC_SMALL_HIGH_WORD_WAQFA = 0x08DF;
const char32_t ARABIC_SMALL_HIGH_FOOTNOTE_MARKER = 0x08E0;
const char32_t ARABIC_SMALL_HIGH_SIGN_SAFHA = 0x08E1;
const char32_t ARABIC_DISPUTED_END_OF_AYAH = 0x08E2;
const char32_t ARABIC_TURNED_DAMMA_BELOW = 0x08E3;
const char32_t ARABIC_CURLY_FATHA = 0x08E4;
const char32_t ARABIC_CURLY_DAMMA = 0x08E5;
const char32_t ARABIC_CURLY_KASRA = 0x08E6;
const char32_t ARABIC_CURLY_FATHATAN = 0x08E7;
const char32_t ARABIC_CURLY_DAMMATAN = 0x08E8;
const char32_t ARABIC_CURLY_KASRATAN = 0x08E9;
const char32_t ARABIC_TONE_ONE_DOT_ABOVE = 0x08EA;
const char32_t ARABIC_TONE_TWO_DOTS_ABOVE = 0x08EB;
const char32_t ARABIC_TONE_LOOP_ABOVE = 0x08EC;
const char32_t ARABIC_TONE_ONE_DOT_BELOW = 0x08ED;
const char32_t ARABIC_TONE_TWO_DOTS_BELOW = 0x08EE;
const char32_t ARABIC_TONE_LOOP_BELOW = 0x08EF;
const char32_t ARABIC_OPEN_FATHATAN = 0x08F0;
const char32_t ARABIC_OPEN_DAMMATAN = 0x08F1;
const char32_t ARABIC_OPEN_KASRATAN = 0x08F2;
const char32_t ARABIC_SMALL_HIGH_WAW = 0x08F3;
const char32_t ARABIC_FATHA_WITH_RING = 0x08F4;
const char32_t ARABIC_FATHA_WITH_DOT_ABOVE = 0x08F5;
const char32_t ARABIC_KASRA_WITH_DOT_BELOW = 0x08F6;
const char32_t ARABIC_LEFT_ARROWHEAD_ABOVE = 0x08F7;
const char32_t ARABIC_RIGHT_ARROWHEAD_ABOVE = 0x08F8;
const char32_t ARABIC_LEFT_ARROWHEAD_BELOW = 0x08F9;
const char32_t ARABIC_RIGHT_ARROWHEAD_BELOW = 0x08FA;
const char32_t ARABIC_DOUBLE_RIGHT_ARROWHEAD_ABOVE = 0x08FB;
const char32_t ARABIC_DOUBLE_RIGHT_ARROWHEAD_ABOVE_WITH_DOT = 0x08FC;
const char32_t ARABIC_RIGHT_ARROWHEAD_ABOVE_WITH_DOT = 0x08FD;
const char32_t ARABIC_DAMMA_WITH_DOT = 0x08FE;
const char32_t ARABIC_MARK_SIDEWAYS_NOON_GHUNNA = 0x08FF;
const char32_t DEVANAGARI_SIGN_INVERTED_CANDRABINDU = 0x0900;
const char32_t DEVANAGARI_SIGN_CANDRABINDU = 0x0901;
const char32_t DEVANAGARI_SIGN_ANUSVARA = 0x0902;
const char32_t DEVANAGARI_SIGN_VISARGA = 0x0903;
const char32_t DEVANAGARI_LETTER_SHORT_A = 0x0904;
const char32_t DEVANAGARI_LETTER_A = 0x0905;
const char32_t DEVANAGARI_LETTER_AA = 0x0906;
const char32_t DEVANAGARI_LETTER_I = 0x0907;
const char32_t DEVANAGARI_LETTER_II = 0x0908;
const char32_t DEVANAGARI_LETTER_U = 0x0909;
const char32_t DEVANAGARI_LETTER_UU = 0x090A;
const char32_t DEVANAGARI_LETTER_VOCALIC_R = 0x090B;
const char32_t DEVANAGARI_LETTER_VOCALIC_L = 0x090C;
const char32_t DEVANAGARI_LETTER_CANDRA_E = 0x090D;
const char32_t DEVANAGARI_LETTER_SHORT_E = 0x090E;
const char32_t DEVANAGARI_LETTER_E = 0x090F;
const char32_t DEVANAGARI_LETTER_AI = 0x0910;
const char32_t DEVANAGARI_LETTER_CANDRA_O = 0x0911;
const char32_t DEVANAGARI_LETTER_SHORT_O = 0x0912;
const char32_t DEVANAGARI_LETTER_O = 0x0913;
const char32_t DEVANAGARI_LETTER_AU = 0x0914;
const char32_t DEVANAGARI_LETTER_KA = 0x0915;
const char32_t DEVANAGARI_LETTER_KHA = 0x0916;
const char32_t DEVANAGARI_LETTER_GA = 0x0917;
const char32_t DEVANAGARI_LETTER_GHA = 0x0918;
const char32_t DEVANAGARI_LETTER_NGA = 0x0919;
const char32_t DEVANAGARI_LETTER_CA = 0x091A;
const char32_t DEVANAGARI_LETTER_CHA = 0x091B;
const char32_t DEVANAGARI_LETTER_JA = 0x091C;
const char32_t DEVANAGARI_LETTER_JHA = 0x091D;
const char32_t DEVANAGARI_LETTER_NYA = 0x091E;
const char32_t DEVANAGARI_LETTER_TTA = 0x091F;
const char32_t DEVANAGARI_LETTER_TTHA = 0x0920;
const char32_t DEVANAGARI_LETTER_DDA = 0x0921;
const char32_t DEVANAGARI_LETTER_DDHA = 0x0922;
const char32_t DEVANAGARI_LETTER_NNA = 0x0923;
const char32_t DEVANAGARI_LETTER_TA = 0x0924;
const char32_t DEVANAGARI_LETTER_THA = 0x0925;
const char32_t DEVANAGARI_LETTER_DA = 0x0926;
const char32_t DEVANAGARI_LETTER_DHA = 0x0927;
const char32_t DEVANAGARI_LETTER_NA = 0x0928;
const char32_t DEVANAGARI_LETTER_NNNA = 0x0929;
const char32_t DEVANAGARI_LETTER_PA = 0x092A;
const char32_t DEVANAGARI_LETTER_PHA = 0x092B;
const char32_t DEVANAGARI_LETTER_BA = 0x092C;
const char32_t DEVANAGARI_LETTER_BHA = 0x092D;
const char32_t DEVANAGARI_LETTER_MA = 0x092E;
const char32_t DEVANAGARI_LETTER_YA = 0x092F;
const char32_t DEVANAGARI_LETTER_RA = 0x0930;
const char32_t DEVANAGARI_LETTER_RRA = 0x0931;
const char32_t DEVANAGARI_LETTER_LA = 0x0932;
const char32_t DEVANAGARI_LETTER_LLA = 0x0933;
const char32_t DEVANAGARI_LETTER_LLLA = 0x0934;
const char32_t DEVANAGARI_LETTER_VA = 0x0935;
const char32_t DEVANAGARI_LETTER_SHA = 0x0936;
const char32_t DEVANAGARI_LETTER_SSA = 0x0937;
const char32_t DEVANAGARI_LETTER_SA = 0x0938;
const char32_t DEVANAGARI_LETTER_HA = 0x0939;
const char32_t DEVANAGARI_VOWEL_SIGN_OE = 0x093A;
const char32_t DEVANAGARI_VOWEL_SIGN_OOE = 0x093B;
const char32_t DEVANAGARI_SIGN_NUKTA = 0x093C;
const char32_t DEVANAGARI_SIGN_AVAGRAHA = 0x093D;
const char32_t DEVANAGARI_VOWEL_SIGN_AA = 0x093E;
const char32_t DEVANAGARI_VOWEL_SIGN_I = 0x093F;
const char32_t DEVANAGARI_VOWEL_SIGN_II = 0x0940;
const char32_t DEVANAGARI_VOWEL_SIGN_U = 0x0941;
const char32_t DEVANAGARI_VOWEL_SIGN_UU = 0x0942;
const char32_t DEVANAGARI_VOWEL_SIGN_VOCALIC_R = 0x0943;
const char32_t DEVANAGARI_VOWEL_SIGN_VOCALIC_RR = 0x0944;
const char32_t DEVANAGARI_VOWEL_SIGN_CANDRA_E = 0x0945;
const char32_t DEVANAGARI_VOWEL_SIGN_SHORT_E = 0x0946;
const char32_t DEVANAGARI_VOWEL_SIGN_E = 0x0947;
const char32_t DEVANAGARI_VOWEL_SIGN_AI = 0x0948;
const char32_t DEVANAGARI_VOWEL_SIGN_CANDRA_O = 0x0949;
const char32_t DEVANAGARI_VOWEL_SIGN_SHORT_O = 0x094A;
const char32_t DEVANAGARI_VOWEL_SIGN_O = 0x094B;
const char32_t DEVANAGARI_VOWEL_SIGN_AU = 0x094C;
const char32_t DEVANAGARI_SIGN_VIRAMA = 0x094D;
const char32_t DEVANAGARI_VOWEL_SIGN_PRISHTHAMATRA_E = 0x094E;
const char32_t DEVANAGARI_VOWEL_SIGN_AW = 0x094F;
const char32_t DEVANAGARI_OM = 0x0950;
const char32_t DEVANAGARI_STRESS_SIGN_UDATTA = 0x0951;
const char32_t DEVANAGARI_STRESS_SIGN_ANUDATTA = 0x0952;
const char32_t DEVANAGARI_GRAVE_ACCENT = 0x0953;
const char32_t DEVANAGARI_ACUTE_ACCENT = 0x0954;
const char32_t DEVANAGARI_VOWEL_SIGN_CANDRA_LONG_E = 0x0955;
const char32_t DEVANAGARI_VOWEL_SIGN_UE = 0x0956;
const char32_t DEVANAGARI_VOWEL_SIGN_UUE = 0x0957;
const char32_t DEVANAGARI_LETTER_QA = 0x0958;
const char32_t DEVANAGARI_LETTER_KHHA = 0x0959;
const char32_t DEVANAGARI_LETTER_GHHA = 0x095A;
const char32_t DEVANAGARI_LETTER_ZA = 0x095B;
const char32_t DEVANAGARI_LETTER_DDDHA = 0x095C;
const char32_t DEVANAGARI_LETTER_RHA = 0x095D;
const char32_t DEVANAGARI_LETTER_FA = 0x095E;
const char32_t DEVANAGARI_LETTER_YYA = 0x095F;
const char32_t DEVANAGARI_LETTER_VOCALIC_RR = 0x0960;
const char32_t DEVANAGARI_LETTER_VOCALIC_LL = 0x0961;
const char32_t DEVANAGARI_VOWEL_SIGN_VOCALIC_L = 0x0962;
const char32_t DEVANAGARI_VOWEL_SIGN_VOCALIC_LL = 0x0963;
const char32_t DEVANAGARI_DANDA = 0x0964;
const char32_t DEVANAGARI_DOUBLE_DANDA = 0x0965;
const char32_t DEVANAGARI_DIGIT_ZERO = 0x0966;
const char32_t DEVANAGARI_DIGIT_ONE = 0x0967;
const char32_t DEVANAGARI_DIGIT_TWO = 0x0968;
const char32_t DEVANAGARI_DIGIT_THREE = 0x0969;
const char32_t DEVANAGARI_DIGIT_FOUR = 0x096A;
const char32_t DEVANAGARI_DIGIT_FIVE = 0x096B;
const char32_t DEVANAGARI_DIGIT_SIX = 0x096C;
const char32_t DEVANAGARI_DIGIT_SEVEN = 0x096D;
const char32_t DEVANAGARI_DIGIT_EIGHT = 0x096E;
const char32_t DEVANAGARI_DIGIT_NINE = 0x096F;
const char32_t DEVANAGARI_ABBREVIATION_SIGN = 0x0970;
const char32_t DEVANAGARI_SIGN_HIGH_SPACING_DOT = 0x0971;
const char32_t DEVANAGARI_LETTER_CANDRA_A = 0x0972;
const char32_t DEVANAGARI_LETTER_OE = 0x0973;
const char32_t DEVANAGARI_LETTER_OOE = 0x0974;
const char32_t DEVANAGARI_LETTER_AW = 0x0975;
const char32_t DEVANAGARI_LETTER_UE = 0x0976;
const char32_t DEVANAGARI_LETTER_UUE = 0x0977;
const char32_t DEVANAGARI_LETTER_MARWARI_DDA = 0x0978;
const char32_t DEVANAGARI_LETTER_ZHA = 0x0979;
const char32_t DEVANAGARI_LETTER_HEAVY_YA = 0x097A;
const char32_t DEVANAGARI_LETTER_GGA = 0x097B;
const char32_t DEVANAGARI_LETTER_JJA = 0x097C;
const char32_t DEVANAGARI_LETTER_GLOTTAL_STOP = 0x097D;
const char32_t DEVANAGARI_LETTER_DDDA = 0x097E;
const char32_t DEVANAGARI_LETTER_BBA = 0x097F;
const char32_t BENGALI_ANJI = 0x0980;
const char32_t BENGALI_SIGN_CANDRABINDU = 0x0981;
const char32_t BENGALI_SIGN_ANUSVARA = 0x0982;
const char32_t BENGALI_SIGN_VISARGA = 0x0983;
const char32_t BENGALI_LETTER_A = 0x0985;
const char32_t BENGALI_LETTER_AA = 0x0986;
const char32_t BENGALI_LETTER_I = 0x0987;
const char32_t BENGALI_LETTER_II = 0x0988;
const char32_t BENGALI_LETTER_U = 0x0989;
const char32_t BENGALI_LETTER_UU = 0x098A;
const char32_t BENGALI_LETTER_VOCALIC_R = 0x098B;
const char32_t BENGALI_LETTER_VOCALIC_L = 0x098C;
const char32_t BENGALI_LETTER_E = 0x098F;
const char32_t BENGALI_LETTER_AI = 0x0990;
const char32_t BENGALI_LETTER_O = 0x0993;
const char32_t BENGALI_LETTER_AU = 0x0994;
const char32_t BENGALI_LETTER_KA = 0x0995;
const char32_t BENGALI_LETTER_KHA = 0x0996;
const char32_t BENGALI_LETTER_GA = 0x0997;
const char32_t BENGALI_LETTER_GHA = 0x0998;
const char32_t BENGALI_LETTER_NGA = 0x0999;
const char32_t BENGALI_LETTER_CA = 0x099A;
const char32_t BENGALI_LETTER_CHA = 0x099B;
const char32_t BENGALI_LETTER_JA = 0x099C;
const char32_t BENGALI_LETTER_JHA = 0x099D;
const char32_t BENGALI_LETTER_NYA = 0x099E;
const char32_t BENGALI_LETTER_TTA = 0x099F;
const char32_t BENGALI_LETTER_TTHA = 0x09A0;
const char32_t BENGALI_LETTER_DDA = 0x09A1;
const char32_t BENGALI_LETTER_DDHA = 0x09A2;
const char32_t BENGALI_LETTER_NNA = 0x09A3;
const char32_t BENGALI_LETTER_TA = 0x09A4;
const char32_t BENGALI_LETTER_THA = 0x09A5;
const char32_t BENGALI_LETTER_DA = 0x09A6;
const char32_t BENGALI_LETTER_DHA = 0x09A7;
const char32_t BENGALI_LETTER_NA = 0x09A8;
const char32_t BENGALI_LETTER_PA = 0x09AA;
const char32_t BENGALI_LETTER_PHA = 0x09AB;
const char32_t BENGALI_LETTER_BA = 0x09AC;
const char32_t BENGALI_LETTER_BHA = 0x09AD;
const char32_t BENGALI_LETTER_MA = 0x09AE;
const char32_t BENGALI_LETTER_YA = 0x09AF;
const char32_t BENGALI_LETTER_RA = 0x09B0;
const char32_t BENGALI_LETTER_LA = 0x09B2;
const char32_t BENGALI_LETTER_SHA = 0x09B6;
const char32_t BENGALI_LETTER_SSA = 0x09B7;
const char32_t BENGALI_LETTER_SA = 0x09B8;
const char32_t BENGALI_LETTER_HA = 0x09B9;
const char32_t BENGALI_SIGN_NUKTA = 0x09BC;
const char32_t BENGALI_SIGN_AVAGRAHA = 0x09BD;
const char32_t BENGALI_VOWEL_SIGN_AA = 0x09BE;
const char32_t BENGALI_VOWEL_SIGN_I = 0x09BF;
const char32_t BENGALI_VOWEL_SIGN_II = 0x09C0;
const char32_t BENGALI_VOWEL_SIGN_U = 0x09C1;
const char32_t BENGALI_VOWEL_SIGN_UU = 0x09C2;
const char32_t BENGALI_VOWEL_SIGN_VOCALIC_R = 0x09C3;
const char32_t BENGALI_VOWEL_SIGN_VOCALIC_RR = 0x09C4;
const char32_t BENGALI_VOWEL_SIGN_E = 0x09C7;
const char32_t BENGALI_VOWEL_SIGN_AI = 0x09C8;
const char32_t BENGALI_VOWEL_SIGN_O = 0x09CB;
const char32_t BENGALI_VOWEL_SIGN_AU = 0x09CC;
const char32_t BENGALI_SIGN_VIRAMA = 0x09CD;
const char32_t BENGALI_LETTER_KHANDA_TA = 0x09CE;
const char32_t BENGALI_AU_LENGTH_MARK = 0x09D7;
const char32_t BENGALI_LETTER_RRA = 0x09DC;
const char32_t BENGALI_LETTER_RHA = 0x09DD;
const char32_t BENGALI_LETTER_YYA = 0x09DF;
const char32_t BENGALI_LETTER_VOCALIC_RR = 0x09E0;
const char32_t BENGALI_LETTER_VOCALIC_LL = 0x09E1;
const char32_t BENGALI_VOWEL_SIGN_VOCALIC_L = 0x09E2;
const char32_t BENGALI_VOWEL_SIGN_VOCALIC_LL = 0x09E3;
const char32_t BENGALI_DIGIT_ZERO = 0x09E6;
const char32_t BENGALI_DIGIT_ONE = 0x09E7;
const char32_t BENGALI_DIGIT_TWO = 0x09E8;
const char32_t BENGALI_DIGIT_THREE = 0x09E9;
const char32_t BENGALI_DIGIT_FOUR = 0x09EA;
const char32_t BENGALI_DIGIT_FIVE = 0x09EB;
const char32_t BENGALI_DIGIT_SIX = 0x09EC;
const char32_t BENGALI_DIGIT_SEVEN = 0x09ED;
const char32_t BENGALI_DIGIT_EIGHT = 0x09EE;
const char32_t BENGALI_DIGIT_NINE = 0x09EF;
const char32_t BENGALI_LETTER_RA_WITH_MIDDLE_DIAGONAL = 0x09F0;
const char32_t BENGALI_LETTER_RA_WITH_LOWER_DIAGONAL = 0x09F1;
const char32_t BENGALI_RUPEE_MARK = 0x09F2;
const char32_t BENGALI_RUPEE_SIGN = 0x09F3;
const char32_t BENGALI_CURRENCY_NUMERATOR_ONE = 0x09F4;
const char32_t BENGALI_CURRENCY_NUMERATOR_TWO = 0x09F5;
const char32_t BENGALI_CURRENCY_NUMERATOR_THREE = 0x09F6;
const char32_t BENGALI_CURRENCY_NUMERATOR_FOUR = 0x09F7;
const char32_t BENGALI_CURRENCY_NUMERATOR_ONE_LESS_THAN_THE_DENOMINATOR = 0x09F8;
const char32_t BENGALI_CURRENCY_DENOMINATOR_SIXTEEN = 0x09F9;
const char32_t BENGALI_ISSHAR = 0x09FA;
const char32_t BENGALI_GANDA_MARK = 0x09FB;
const char32_t BENGALI_LETTER_VEDIC_ANUSVARA = 0x09FC;
const char32_t BENGALI_ABBREVIATION_SIGN = 0x09FD;
const char32_t BENGALI_SANDHI_MARK = 0x09FE;
const char32_t GURMUKHI_SIGN_ADAK_BINDI = 0x0A01;
const char32_t GURMUKHI_SIGN_BINDI = 0x0A02;
const char32_t GURMUKHI_SIGN_VISARGA = 0x0A03;
const char32_t GURMUKHI_LETTER_A = 0x0A05;
const char32_t GURMUKHI_LETTER_AA = 0x0A06;
const char32_t GURMUKHI_LETTER_I = 0x0A07;
const char32_t GURMUKHI_LETTER_II = 0x0A08;
const char32_t GURMUKHI_LETTER_U = 0x0A09;
const char32_t GURMUKHI_LETTER_UU = 0x0A0A;
const char32_t GURMUKHI_LETTER_EE = 0x0A0F;
const char32_t GURMUKHI_LETTER_AI = 0x0A10;
const char32_t GURMUKHI_LETTER_OO = 0x0A13;
const char32_t GURMUKHI_LETTER_AU = 0x0A14;
const char32_t GURMUKHI_LETTER_KA = 0x0A15;
const char32_t GURMUKHI_LETTER_KHA = 0x0A16;
const char32_t GURMUKHI_LETTER_GA = 0x0A17;
const char32_t GURMUKHI_LETTER_GHA = 0x0A18;
const char32_t GURMUKHI_LETTER_NGA = 0x0A19;
const char32_t GURMUKHI_LETTER_CA = 0x0A1A;
const char32_t GURMUKHI_LETTER_CHA = 0x0A1B;
const char32_t GURMUKHI_LETTER_JA = 0x0A1C;
const char32_t GURMUKHI_LETTER_JHA = 0x0A1D;
const char32_t GURMUKHI_LETTER_NYA = 0x0A1E;
const char32_t GURMUKHI_LETTER_TTA = 0x0A1F;
const char32_t GURMUKHI_LETTER_TTHA = 0x0A20;
const char32_t GURMUKHI_LETTER_DDA = 0x0A21;
const char32_t GURMUKHI_LETTER_DDHA = 0x0A22;
const char32_t GURMUKHI_LETTER_NNA = 0x0A23;
const char32_t GURMUKHI_LETTER_TA = 0x0A24;
const char32_t GURMUKHI_LETTER_THA = 0x0A25;
const char32_t GURMUKHI_LETTER_DA = 0x0A26;
const char32_t GURMUKHI_LETTER_DHA = 0x0A27;
const char32_t GURMUKHI_LETTER_NA = 0x0A28;
const char32_t GURMUKHI_LETTER_PA = 0x0A2A;
const char32_t GURMUKHI_LETTER_PHA = 0x0A2B;
const char32_t GURMUKHI_LETTER_BA = 0x0A2C;
const char32_t GURMUKHI_LETTER_BHA = 0x0A2D;
const char32_t GURMUKHI_LETTER_MA = 0x0A2E;
const char32_t GURMUKHI_LETTER_YA = 0x0A2F;
const char32_t GURMUKHI_LETTER_RA = 0x0A30;
const char32_t GURMUKHI_LETTER_LA = 0x0A32;
const char32_t GURMUKHI_LETTER_LLA = 0x0A33;
const char32_t GURMUKHI_LETTER_VA = 0x0A35;
const char32_t GURMUKHI_LETTER_SHA = 0x0A36;
const char32_t GURMUKHI_LETTER_SA = 0x0A38;
const char32_t GURMUKHI_LETTER_HA = 0x0A39;
const char32_t GURMUKHI_SIGN_NUKTA = 0x0A3C;
const char32_t GURMUKHI_VOWEL_SIGN_AA = 0x0A3E;
const char32_t GURMUKHI_VOWEL_SIGN_I = 0x0A3F;
const char32_t GURMUKHI_VOWEL_SIGN_II = 0x0A40;
const char32_t GURMUKHI_VOWEL_SIGN_U = 0x0A41;
const char32_t GURMUKHI_VOWEL_SIGN_UU = 0x0A42;
const char32_t GURMUKHI_VOWEL_SIGN_EE = 0x0A47;
const char32_t GURMUKHI_VOWEL_SIGN_AI = 0x0A48;
const char32_t GURMUKHI_VOWEL_SIGN_OO = 0x0A4B;
const char32_t GURMUKHI_VOWEL_SIGN_AU = 0x0A4C;
const char32_t GURMUKHI_SIGN_VIRAMA = 0x0A4D;
const char32_t GURMUKHI_SIGN_UDAAT = 0x0A51;
const char32_t GURMUKHI_LETTER_KHHA = 0x0A59;
const char32_t GURMUKHI_LETTER_GHHA = 0x0A5A;
const char32_t GURMUKHI_LETTER_ZA = 0x0A5B;
const char32_t GURMUKHI_LETTER_RRA = 0x0A5C;
const char32_t GURMUKHI_LETTER_FA = 0x0A5E;
const char32_t GURMUKHI_DIGIT_ZERO = 0x0A66;
const char32_t GURMUKHI_DIGIT_ONE = 0x0A67;
const char32_t GURMUKHI_DIGIT_TWO = 0x0A68;
const char32_t GURMUKHI_DIGIT_THREE = 0x0A69;
const char32_t GURMUKHI_DIGIT_FOUR = 0x0A6A;
const char32_t GURMUKHI_DIGIT_FIVE = 0x0A6B;
const char32_t GURMUKHI_DIGIT_SIX = 0x0A6C;
const char32_t GURMUKHI_DIGIT_SEVEN = 0x0A6D;
const char32_t GURMUKHI_DIGIT_EIGHT = 0x0A6E;
const char32_t GURMUKHI_DIGIT_NINE = 0x0A6F;
const char32_t GURMUKHI_TIPPI = 0x0A70;
const char32_t GURMUKHI_ADDAK = 0x0A71;
const char32_t GURMUKHI_IRI = 0x0A72;
const char32_t GURMUKHI_URA = 0x0A73;
const char32_t GURMUKHI_EK_ONKAR = 0x0A74;
const char32_t GURMUKHI_SIGN_YAKASH = 0x0A75;
const char32_t GURMUKHI_ABBREVIATION_SIGN = 0x0A76;
const char32_t GUJARATI_SIGN_CANDRABINDU = 0x0A81;
const char32_t GUJARATI_SIGN_ANUSVARA = 0x0A82;
const char32_t GUJARATI_SIGN_VISARGA = 0x0A83;
const char32_t GUJARATI_LETTER_A = 0x0A85;
const char32_t GUJARATI_LETTER_AA = 0x0A86;
const char32_t GUJARATI_LETTER_I = 0x0A87;
const char32_t GUJARATI_LETTER_II = 0x0A88;
const char32_t GUJARATI_LETTER_U = 0x0A89;
const char32_t GUJARATI_LETTER_UU = 0x0A8A;
const char32_t GUJARATI_LETTER_VOCALIC_R = 0x0A8B;
const char32_t GUJARATI_LETTER_VOCALIC_L = 0x0A8C;
const char32_t GUJARATI_VOWEL_CANDRA_E = 0x0A8D;
const char32_t GUJARATI_LETTER_E = 0x0A8F;
const char32_t GUJARATI_LETTER_AI = 0x0A90;
const char32_t GUJARATI_VOWEL_CANDRA_O = 0x0A91;
const char32_t GUJARATI_LETTER_O = 0x0A93;
const char32_t GUJARATI_LETTER_AU = 0x0A94;
const char32_t GUJARATI_LETTER_KA = 0x0A95;
const char32_t GUJARATI_LETTER_KHA = 0x0A96;
const char32_t GUJARATI_LETTER_GA = 0x0A97;
const char32_t GUJARATI_LETTER_GHA = 0x0A98;
const char32_t GUJARATI_LETTER_NGA = 0x0A99;
const char32_t GUJARATI_LETTER_CA = 0x0A9A;
const char32_t GUJARATI_LETTER_CHA = 0x0A9B;
const char32_t GUJARATI_LETTER_JA = 0x0A9C;
const char32_t GUJARATI_LETTER_JHA = 0x0A9D;
const char32_t GUJARATI_LETTER_NYA = 0x0A9E;
const char32_t GUJARATI_LETTER_TTA = 0x0A9F;
const char32_t GUJARATI_LETTER_TTHA = 0x0AA0;
const char32_t GUJARATI_LETTER_DDA = 0x0AA1;
const char32_t GUJARATI_LETTER_DDHA = 0x0AA2;
const char32_t GUJARATI_LETTER_NNA = 0x0AA3;
const char32_t GUJARATI_LETTER_TA = 0x0AA4;
const char32_t GUJARATI_LETTER_THA = 0x0AA5;
const char32_t GUJARATI_LETTER_DA = 0x0AA6;
const char32_t GUJARATI_LETTER_DHA = 0x0AA7;
const char32_t GUJARATI_LETTER_NA = 0x0AA8;
const char32_t GUJARATI_LETTER_PA = 0x0AAA;
const char32_t GUJARATI_LETTER_PHA = 0x0AAB;
const char32_t GUJARATI_LETTER_BA = 0x0AAC;
const char32_t GUJARATI_LETTER_BHA = 0x0AAD;
const char32_t GUJARATI_LETTER_MA = 0x0AAE;
const char32_t GUJARATI_LETTER_YA = 0x0AAF;
const char32_t GUJARATI_LETTER_RA = 0x0AB0;
const char32_t GUJARATI_LETTER_LA = 0x0AB2;
const char32_t GUJARATI_LETTER_LLA = 0x0AB3;
const char32_t GUJARATI_LETTER_VA = 0x0AB5;
const char32_t GUJARATI_LETTER_SHA = 0x0AB6;
const char32_t GUJARATI_LETTER_SSA = 0x0AB7;
const char32_t GUJARATI_LETTER_SA = 0x0AB8;
const char32_t GUJARATI_LETTER_HA = 0x0AB9;
const char32_t GUJARATI_SIGN_NUKTA = 0x0ABC;
const char32_t GUJARATI_SIGN_AVAGRAHA = 0x0ABD;
const char32_t GUJARATI_VOWEL_SIGN_AA = 0x0ABE;
const char32_t GUJARATI_VOWEL_SIGN_I = 0x0ABF;
const char32_t GUJARATI_VOWEL_SIGN_II = 0x0AC0;
const char32_t GUJARATI_VOWEL_SIGN_U = 0x0AC1;
const char32_t GUJARATI_VOWEL_SIGN_UU = 0x0AC2;
const char32_t GUJARATI_VOWEL_SIGN_VOCALIC_R = 0x0AC3;
const char32_t GUJARATI_VOWEL_SIGN_VOCALIC_RR = 0x0AC4;
const char32_t GUJARATI_VOWEL_SIGN_CANDRA_E = 0x0AC5;
const char32_t GUJARATI_VOWEL_SIGN_E = 0x0AC7;
const char32_t GUJARATI_VOWEL_SIGN_AI = 0x0AC8;
const char32_t GUJARATI_VOWEL_SIGN_CANDRA_O = 0x0AC9;
const char32_t GUJARATI_VOWEL_SIGN_O = 0x0ACB;
const char32_t GUJARATI_VOWEL_SIGN_AU = 0x0ACC;
const char32_t GUJARATI_SIGN_VIRAMA = 0x0ACD;
const char32_t GUJARATI_OM = 0x0AD0;
const char32_t GUJARATI_LETTER_VOCALIC_RR = 0x0AE0;
const char32_t GUJARATI_LETTER_VOCALIC_LL = 0x0AE1;
const char32_t GUJARATI_VOWEL_SIGN_VOCALIC_L = 0x0AE2;
const char32_t GUJARATI_VOWEL_SIGN_VOCALIC_LL = 0x0AE3;
const char32_t GUJARATI_DIGIT_ZERO = 0x0AE6;
const char32_t GUJARATI_DIGIT_ONE = 0x0AE7;
const char32_t GUJARATI_DIGIT_TWO = 0x0AE8;
const char32_t GUJARATI_DIGIT_THREE = 0x0AE9;
const char32_t GUJARATI_DIGIT_FOUR = 0x0AEA;
const char32_t GUJARATI_DIGIT_FIVE = 0x0AEB;
const char32_t GUJARATI_DIGIT_SIX = 0x0AEC;
const char32_t GUJARATI_DIGIT_SEVEN = 0x0AED;
const char32_t GUJARATI_DIGIT_EIGHT = 0x0AEE;
const char32_t GUJARATI_DIGIT_NINE = 0x0AEF;
const char32_t GUJARATI_ABBREVIATION_SIGN = 0x0AF0;
const char32_t GUJARATI_RUPEE_SIGN = 0x0AF1;
const char32_t GUJARATI_LETTER_ZHA = 0x0AF9;
const char32_t GUJARATI_SIGN_SUKUN = 0x0AFA;
const char32_t GUJARATI_SIGN_SHADDA = 0x0AFB;
const char32_t GUJARATI_SIGN_MADDAH = 0x0AFC;
const char32_t GUJARATI_SIGN_THREE_DOT_NUKTA_ABOVE = 0x0AFD;
const char32_t GUJARATI_SIGN_CIRCLE_NUKTA_ABOVE = 0x0AFE;
const char32_t GUJARATI_SIGN_TWO_CIRCLE_NUKTA_ABOVE = 0x0AFF;
const char32_t ORIYA_SIGN_CANDRABINDU = 0x0B01;
const char32_t ORIYA_SIGN_ANUSVARA = 0x0B02;
const char32_t ORIYA_SIGN_VISARGA = 0x0B03;
const char32_t ORIYA_LETTER_A = 0x0B05;
const char32_t ORIYA_LETTER_AA = 0x0B06;
const char32_t ORIYA_LETTER_I = 0x0B07;
const char32_t ORIYA_LETTER_II = 0x0B08;
const char32_t ORIYA_LETTER_U = 0x0B09;
const char32_t ORIYA_LETTER_UU = 0x0B0A;
const char32_t ORIYA_LETTER_VOCALIC_R = 0x0B0B;
const char32_t ORIYA_LETTER_VOCALIC_L = 0x0B0C;
const char32_t ORIYA_LETTER_E = 0x0B0F;
const char32_t ORIYA_LETTER_AI = 0x0B10;
const char32_t ORIYA_LETTER_O = 0x0B13;
const char32_t ORIYA_LETTER_AU = 0x0B14;
const char32_t ORIYA_LETTER_KA = 0x0B15;
const char32_t ORIYA_LETTER_KHA = 0x0B16;
const char32_t ORIYA_LETTER_GA = 0x0B17;
const char32_t ORIYA_LETTER_GHA = 0x0B18;
const char32_t ORIYA_LETTER_NGA = 0x0B19;
const char32_t ORIYA_LETTER_CA = 0x0B1A;
const char32_t ORIYA_LETTER_CHA = 0x0B1B;
const char32_t ORIYA_LETTER_JA = 0x0B1C;
const char32_t ORIYA_LETTER_JHA = 0x0B1D;
const char32_t ORIYA_LETTER_NYA = 0x0B1E;
const char32_t ORIYA_LETTER_TTA = 0x0B1F;
const char32_t ORIYA_LETTER_TTHA = 0x0B20;
const char32_t ORIYA_LETTER_DDA = 0x0B21;
const char32_t ORIYA_LETTER_DDHA = 0x0B22;
const char32_t ORIYA_LETTER_NNA = 0x0B23;
const char32_t ORIYA_LETTER_TA = 0x0B24;
const char32_t ORIYA_LETTER_THA = 0x0B25;
const char32_t ORIYA_LETTER_DA = 0x0B26;
const char32_t ORIYA_LETTER_DHA = 0x0B27;
const char32_t ORIYA_LETTER_NA = 0x0B28;
const char32_t ORIYA_LETTER_PA = 0x0B2A;
const char32_t ORIYA_LETTER_PHA = 0x0B2B;
const char32_t ORIYA_LETTER_BA = 0x0B2C;
const char32_t ORIYA_LETTER_BHA = 0x0B2D;
const char32_t ORIYA_LETTER_MA = 0x0B2E;
const char32_t ORIYA_LETTER_YA = 0x0B2F;
const char32_t ORIYA_LETTER_RA = 0x0B30;
const char32_t ORIYA_LETTER_LA = 0x0B32;
const char32_t ORIYA_LETTER_LLA = 0x0B33;
const char32_t ORIYA_LETTER_VA = 0x0B35;
const char32_t ORIYA_LETTER_SHA = 0x0B36;
const char32_t ORIYA_LETTER_SSA = 0x0B37;
const char32_t ORIYA_LETTER_SA = 0x0B38;
const char32_t ORIYA_LETTER_HA = 0x0B39;
const char32_t ORIYA_SIGN_NUKTA = 0x0B3C;
const char32_t ORIYA_SIGN_AVAGRAHA = 0x0B3D;
const char32_t ORIYA_VOWEL_SIGN_AA = 0x0B3E;
const char32_t ORIYA_VOWEL_SIGN_I = 0x0B3F;
const char32_t ORIYA_VOWEL_SIGN_II = 0x0B40;
const char32_t ORIYA_VOWEL_SIGN_U = 0x0B41;
const char32_t ORIYA_VOWEL_SIGN_UU = 0x0B42;
const char32_t ORIYA_VOWEL_SIGN_VOCALIC_R = 0x0B43;
const char32_t ORIYA_VOWEL_SIGN_VOCALIC_RR = 0x0B44;
const char32_t ORIYA_VOWEL_SIGN_E = 0x0B47;
const char32_t ORIYA_VOWEL_SIGN_AI = 0x0B48;
const char32_t ORIYA_VOWEL_SIGN_O = 0x0B4B;
const char32_t ORIYA_VOWEL_SIGN_AU = 0x0B4C;
const char32_t ORIYA_SIGN_VIRAMA = 0x0B4D;
const char32_t ORIYA_SIGN_OVERLINE = 0x0B55;
const char32_t ORIYA_AI_LENGTH_MARK = 0x0B56;
const char32_t ORIYA_AU_LENGTH_MARK = 0x0B57;
const char32_t ORIYA_LETTER_RRA = 0x0B5C;
const char32_t ORIYA_LETTER_RHA = 0x0B5D;
const char32_t ORIYA_LETTER_YYA = 0x0B5F;
const char32_t ORIYA_LETTER_VOCALIC_RR = 0x0B60;
const char32_t ORIYA_LETTER_VOCALIC_LL = 0x0B61;
const char32_t ORIYA_VOWEL_SIGN_VOCALIC_L = 0x0B62;
const char32_t ORIYA_VOWEL_SIGN_VOCALIC_LL = 0x0B63;
const char32_t ORIYA_DIGIT_ZERO = 0x0B66;
const char32_t ORIYA_DIGIT_ONE = 0x0B67;
const char32_t ORIYA_DIGIT_TWO = 0x0B68;
const char32_t ORIYA_DIGIT_THREE = 0x0B69;
const char32_t ORIYA_DIGIT_FOUR = 0x0B6A;
const char32_t ORIYA_DIGIT_FIVE = 0x0B6B;
const char32_t ORIYA_DIGIT_SIX = 0x0B6C;
const char32_t ORIYA_DIGIT_SEVEN = 0x0B6D;
const char32_t ORIYA_DIGIT_EIGHT = 0x0B6E;
const char32_t ORIYA_DIGIT_NINE = 0x0B6F;
const char32_t ORIYA_ISSHAR = 0x0B70;
const char32_t ORIYA_LETTER_WA = 0x0B71;
const char32_t ORIYA_FRACTION_ONE_QUARTER = 0x0B72;
const char32_t ORIYA_FRACTION_ONE_HALF = 0x0B73;
const char32_t ORIYA_FRACTION_THREE_QUARTERS = 0x0B74;
const char32_t ORIYA_FRACTION_ONE_SIXTEENTH = 0x0B75;
const char32_t ORIYA_FRACTION_ONE_EIGHTH = 0x0B76;
const char32_t ORIYA_FRACTION_THREE_SIXTEENTHS = 0x0B77;
const char32_t TAMIL_SIGN_ANUSVARA = 0x0B82;
const char32_t TAMIL_SIGN_VISARGA = 0x0B83;
const char32_t TAMIL_LETTER_A = 0x0B85;
const char32_t TAMIL_LETTER_AA = 0x0B86;
const char32_t TAMIL_LETTER_I = 0x0B87;
const char32_t TAMIL_LETTER_II = 0x0B88;
const char32_t TAMIL_LETTER_U = 0x0B89;
const char32_t TAMIL_LETTER_UU = 0x0B8A;
const char32_t TAMIL_LETTER_E = 0x0B8E;
const char32_t TAMIL_LETTER_EE = 0x0B8F;
const char32_t TAMIL_LETTER_AI = 0x0B90;
const char32_t TAMIL_LETTER_O = 0x0B92;
const char32_t TAMIL_LETTER_OO = 0x0B93;
const char32_t TAMIL_LETTER_AU = 0x0B94;
const char32_t TAMIL_LETTER_KA = 0x0B95;
const char32_t TAMIL_LETTER_NGA = 0x0B99;
const char32_t TAMIL_LETTER_CA = 0x0B9A;
const char32_t TAMIL_LETTER_JA = 0x0B9C;
const char32_t TAMIL_LETTER_NYA = 0x0B9E;
const char32_t TAMIL_LETTER_TTA = 0x0B9F;
const char32_t TAMIL_LETTER_NNA = 0x0BA3;
const char32_t TAMIL_LETTER_TA = 0x0BA4;
const char32_t TAMIL_LETTER_NA = 0x0BA8;
const char32_t TAMIL_LETTER_NNNA = 0x0BA9;
const char32_t TAMIL_LETTER_PA = 0x0BAA;
const char32_t TAMIL_LETTER_MA = 0x0BAE;
const char32_t TAMIL_LETTER_YA = 0x0BAF;
const char32_t TAMIL_LETTER_RA = 0x0BB0;
const char32_t TAMIL_LETTER_RRA = 0x0BB1;
const char32_t TAMIL_LETTER_LA = 0x0BB2;
const char32_t TAMIL_LETTER_LLA = 0x0BB3;
const char32_t TAMIL_LETTER_LLLA = 0x0BB4;
const char32_t TAMIL_LETTER_VA = 0x0BB5;
const char32_t TAMIL_LETTER_SHA = 0x0BB6;
const char32_t TAMIL_LETTER_SSA = 0x0BB7;
const char32_t TAMIL_LETTER_SA = 0x0BB8;
const char32_t TAMIL_LETTER_HA = 0x0BB9;
const char32_t TAMIL_VOWEL_SIGN_AA = 0x0BBE;
const char32_t TAMIL_VOWEL_SIGN_I = 0x0BBF;
const char32_t TAMIL_VOWEL_SIGN_II = 0x0BC0;
const char32_t TAMIL_VOWEL_SIGN_U = 0x0BC1;
const char32_t TAMIL_VOWEL_SIGN_UU = 0x0BC2;
const char32_t TAMIL_VOWEL_SIGN_E = 0x0BC6;
const char32_t TAMIL_VOWEL_SIGN_EE = 0x0BC7;
const char32_t TAMIL_VOWEL_SIGN_AI = 0x0BC8;
const char32_t TAMIL_VOWEL_SIGN_O = 0x0BCA;
const char32_t TAMIL_VOWEL_SIGN_OO = 0x0BCB;
const char32_t TAMIL_VOWEL_SIGN_AU = 0x0BCC;
const char32_t TAMIL_SIGN_VIRAMA = 0x0BCD;
const char32_t TAMIL_OM = 0x0BD0;
const char32_t TAMIL_AU_LENGTH_MARK = 0x0BD7;
const char32_t TAMIL_DIGIT_ZERO = 0x0BE6;
const char32_t TAMIL_DIGIT_ONE = 0x0BE7;
const char32_t TAMIL_DIGIT_TWO = 0x0BE8;
const char32_t TAMIL_DIGIT_THREE = 0x0BE9;
const char32_t TAMIL_DIGIT_FOUR = 0x0BEA;
const char32_t TAMIL_DIGIT_FIVE = 0x0BEB;
const char32_t TAMIL_DIGIT_SIX = 0x0BEC;
const char32_t TAMIL_DIGIT_SEVEN = 0x0BED;
const char32_t TAMIL_DIGIT_EIGHT = 0x0BEE;
const char32_t TAMIL_DIGIT_NINE = 0x0BEF;
const char32_t TAMIL_NUMBER_TEN = 0x0BF0;
const char32_t TAMIL_NUMBER_ONE_HUNDRED = 0x0BF1;
const char32_t TAMIL_NUMBER_ONE_THOUSAND = 0x0BF2;
const char32_t TAMIL_DAY_SIGN = 0x0BF3;
const char32_t TAMIL_MONTH_SIGN = 0x0BF4;
const char32_t TAMIL_YEAR_SIGN = 0x0BF5;
const char32_t TAMIL_DEBIT_SIGN = 0x0BF6;
const char32_t TAMIL_CREDIT_SIGN = 0x0BF7;
const char32_t TAMIL_AS_ABOVE_SIGN = 0x0BF8;
const char32_t TAMIL_RUPEE_SIGN = 0x0BF9;
const char32_t TAMIL_NUMBER_SIGN = 0x0BFA;
const char32_t TELUGU_SIGN_COMBINING_CANDRABINDU_ABOVE = 0x0C00;
const char32_t TELUGU_SIGN_CANDRABINDU = 0x0C01;
const char32_t TELUGU_SIGN_ANUSVARA = 0x0C02;
const char32_t TELUGU_SIGN_VISARGA = 0x0C03;
const char32_t TELUGU_SIGN_COMBINING_ANUSVARA_ABOVE = 0x0C04;
const char32_t TELUGU_LETTER_A = 0x0C05;
const char32_t TELUGU_LETTER_AA = 0x0C06;
const char32_t TELUGU_LETTER_I = 0x0C07;
const char32_t TELUGU_LETTER_II = 0x0C08;
const char32_t TELUGU_LETTER_U = 0x0C09;
const char32_t TELUGU_LETTER_UU = 0x0C0A;
const char32_t TELUGU_LETTER_VOCALIC_R = 0x0C0B;
const char32_t TELUGU_LETTER_VOCALIC_L = 0x0C0C;
const char32_t TELUGU_LETTER_E = 0x0C0E;
const char32_t TELUGU_LETTER_EE = 0x0C0F;
const char32_t TELUGU_LETTER_AI = 0x0C10;
const char32_t TELUGU_LETTER_O = 0x0C12;
const char32_t TELUGU_LETTER_OO = 0x0C13;
const char32_t TELUGU_LETTER_AU = 0x0C14;
const char32_t TELUGU_LETTER_KA = 0x0C15;
const char32_t TELUGU_LETTER_KHA = 0x0C16;
const char32_t TELUGU_LETTER_GA = 0x0C17;
const char32_t TELUGU_LETTER_GHA = 0x0C18;
const char32_t TELUGU_LETTER_NGA = 0x0C19;
const char32_t TELUGU_LETTER_CA = 0x0C1A;
const char32_t TELUGU_LETTER_CHA = 0x0C1B;
const char32_t TELUGU_LETTER_JA = 0x0C1C;
const char32_t TELUGU_LETTER_JHA = 0x0C1D;
const char32_t TELUGU_LETTER_NYA = 0x0C1E;
const char32_t TELUGU_LETTER_TTA = 0x0C1F;
const char32_t TELUGU_LETTER_TTHA = 0x0C20;
const char32_t TELUGU_LETTER_DDA = 0x0C21;
const char32_t TELUGU_LETTER_DDHA = 0x0C22;
const char32_t TELUGU_LETTER_NNA = 0x0C23;
const char32_t TELUGU_LETTER_TA = 0x0C24;
const char32_t TELUGU_LETTER_THA = 0x0C25;
const char32_t TELUGU_LETTER_DA = 0x0C26;
const char32_t TELUGU_LETTER_DHA = 0x0C27;
const char32_t TELUGU_LETTER_NA = 0x0C28;
const char32_t TELUGU_LETTER_PA = 0x0C2A;
const char32_t TELUGU_LETTER_PHA = 0x0C2B;
const char32_t TELUGU_LETTER_BA = 0x0C2C;
const char32_t TELUGU_LETTER_BHA = 0x0C2D;
const char32_t TELUGU_LETTER_MA = 0x0C2E;
const char32_t TELUGU_LETTER_YA = 0x0C2F;
const char32_t TELUGU_LETTER_RA = 0x0C30;
const char32_t TELUGU_LETTER_RRA = 0x0C31;
const char32_t TELUGU_LETTER_LA = 0x0C32;
const char32_t TELUGU_LETTER_LLA = 0x0C33;
const char32_t TELUGU_LETTER_LLLA = 0x0C34;
const char32_t TELUGU_LETTER_VA = 0x0C35;
const char32_t TELUGU_LETTER_SHA = 0x0C36;
const char32_t TELUGU_LETTER_SSA = 0x0C37;
const char32_t TELUGU_LETTER_SA = 0x0C38;
const char32_t TELUGU_LETTER_HA = 0x0C39;
const char32_t TELUGU_SIGN_NUKTA = 0x0C3C;
const char32_t TELUGU_SIGN_AVAGRAHA = 0x0C3D;
const char32_t TELUGU_VOWEL_SIGN_AA = 0x0C3E;
const char32_t TELUGU_VOWEL_SIGN_I = 0x0C3F;
const char32_t TELUGU_VOWEL_SIGN_II = 0x0C40;
const char32_t TELUGU_VOWEL_SIGN_U = 0x0C41;
const char32_t TELUGU_VOWEL_SIGN_UU = 0x0C42;
const char32_t TELUGU_VOWEL_SIGN_VOCALIC_R = 0x0C43;
const char32_t TELUGU_VOWEL_SIGN_VOCALIC_RR = 0x0C44;
const char32_t TELUGU_VOWEL_SIGN_E = 0x0C46;
const char32_t TELUGU_VOWEL_SIGN_EE = 0x0C47;
const char32_t TELUGU_VOWEL_SIGN_AI = 0x0C48;
const char32_t TELUGU_VOWEL_SIGN_O = 0x0C4A;
const char32_t TELUGU_VOWEL_SIGN_OO = 0x0C4B;
const char32_t TELUGU_VOWEL_SIGN_AU = 0x0C4C;
const char32_t TELUGU_SIGN_VIRAMA = 0x0C4D;
const char32_t TELUGU_LENGTH_MARK = 0x0C55;
const char32_t TELUGU_AI_LENGTH_MARK = 0x0C56;
const char32_t TELUGU_LETTER_TSA = 0x0C58;
const char32_t TELUGU_LETTER_DZA = 0x0C59;
const char32_t TELUGU_LETTER_RRRA = 0x0C5A;
const char32_t TELUGU_ARCHAIC_SHRII = 0x0C5C;
const char32_t TELUGU_LETTER_NAKAARA_POLLU = 0x0C5D;
const char32_t TELUGU_LETTER_VOCALIC_RR = 0x0C60;
const char32_t TELUGU_LETTER_VOCALIC_LL = 0x0C61;
const char32_t TELUGU_VOWEL_SIGN_VOCALIC_L = 0x0C62;
const char32_t TELUGU_VOWEL_SIGN_VOCALIC_LL = 0x0C63;
const char32_t TELUGU_DIGIT_ZERO = 0x0C66;
const char32_t TELUGU_DIGIT_ONE = 0x0C67;
const char32_t TELUGU_DIGIT_TWO = 0x0C68;
const char32_t TELUGU_DIGIT_THREE = 0x0C69;
const char32_t TELUGU_DIGIT_FOUR = 0x0C6A;
const char32_t TELUGU_DIGIT_FIVE = 0x0C6B;
const char32_t TELUGU_DIGIT_SIX = 0x0C6C;
const char32_t TELUGU_DIGIT_SEVEN = 0x0C6D;
const char32_t TELUGU_DIGIT_EIGHT = 0x0C6E;
const char32_t TELUGU_DIGIT_NINE = 0x0C6F;
const char32_t TELUGU_SIGN_SIDDHAM = 0x0C77;
const char32_t TELUGU_FRACTION_DIGIT_ZERO_FOR_ODD_POWERS_OF_FOUR = 0x0C78;
const char32_t TELUGU_FRACTION_DIGIT_ONE_FOR_ODD_POWERS_OF_FOUR = 0x0C79;
const char32_t TELUGU_FRACTION_DIGIT_TWO_FOR_ODD_POWERS_OF_FOUR = 0x0C7A;
const char32_t TELUGU_FRACTION_DIGIT_THREE_FOR_ODD_POWERS_OF_FOUR = 0x0C7B;
const char32_t TELUGU_FRACTION_DIGIT_ONE_FOR_EVEN_POWERS_OF_FOUR = 0x0C7C;
const char32_t TELUGU_FRACTION_DIGIT_TWO_FOR_EVEN_POWERS_OF_FOUR = 0x0C7D;
const char32_t TELUGU_FRACTION_DIGIT_THREE_FOR_EVEN_POWERS_OF_FOUR = 0x0C7E;
const char32_t TELUGU_SIGN_TUUMU = 0x0C7F;
const char32_t KANNADA_SIGN_SPACING_CANDRABINDU = 0x0C80;
const char32_t KANNADA_SIGN_CANDRABINDU = 0x0C81;
const char32_t KANNADA_SIGN_ANUSVARA = 0x0C82;
const char32_t KANNADA_SIGN_VISARGA = 0x0C83;
const char32_t KANNADA_SIGN_SIDDHAM = 0x0C84;
const char32_t KANNADA_LETTER_A = 0x0C85;
const char32_t KANNADA_LETTER_AA = 0x0C86;
const char32_t KANNADA_LETTER_I = 0x0C87;
const char32_t KANNADA_LETTER_II = 0x0C88;
const char32_t KANNADA_LETTER_U = 0x0C89;
const char32_t KANNADA_LETTER_UU = 0x0C8A;
const char32_t KANNADA_LETTER_VOCALIC_R = 0x0C8B;
const char32_t KANNADA_LETTER_VOCALIC_L = 0x0C8C;
const char32_t KANNADA_LETTER_E = 0x0C8E;
const char32_t KANNADA_LETTER_EE = 0x0C8F;
const char32_t KANNADA_LETTER_AI = 0x0C90;
const char32_t KANNADA_LETTER_O = 0x0C92;
const char32_t KANNADA_LETTER_OO = 0x0C93;
const char32_t KANNADA_LETTER_AU = 0x0C94;
const char32_t KANNADA_LETTER_KA = 0x0C95;
const char32_t KANNADA_LETTER_KHA = 0x0C96;
const char32_t KANNADA_LETTER_GA = 0x0C97;
const char32_t KANNADA_LETTER_GHA = 0x0C98;
const char32_t KANNADA_LETTER_NGA = 0x0C99;
const char32_t KANNADA_LETTER_CA = 0x0C9A;
const char32_t KANNADA_LETTER_CHA = 0x0C9B;
const char32_t KANNADA_LETTER_JA = 0x0C9C;
const char32_t KANNADA_LETTER_JHA = 0x0C9D;
const char32_t KANNADA_LETTER_NYA = 0x0C9E;
const char32_t KANNADA_LETTER_TTA = 0x0C9F;
const char32_t KANNADA_LETTER_TTHA = 0x0CA0;
const char32_t KANNADA_LETTER_DDA = 0x0CA1;
const char32_t KANNADA_LETTER_DDHA = 0x0CA2;
const char32_t KANNADA_LETTER_NNA = 0x0CA3;
const char32_t KANNADA_LETTER_TA = 0x0CA4;
const char32_t KANNADA_LETTER_THA = 0x0CA5;
const char32_t KANNADA_LETTER_DA = 0x0CA6;
const char32_t KANNADA_LETTER_DHA = 0x0CA7;
const char32_t KANNADA_LETTER_NA = 0x0CA8;
const char32_t KANNADA_LETTER_PA = 0x0CAA;
const char32_t KANNADA_LETTER_PHA = 0x0CAB;
const char32_t KANNADA_LETTER_BA = 0x0CAC;
const char32_t KANNADA_LETTER_BHA = 0x0CAD;
const char32_t KANNADA_LETTER_MA = 0x0CAE;
const char32_t KANNADA_LETTER_YA = 0x0CAF;
const char32_t KANNADA_LETTER_RA = 0x0CB0;
const char32_t KANNADA_LETTER_RRA = 0x0CB1;
const char32_t KANNADA_LETTER_LA = 0x0CB2;
const char32_t KANNADA_LETTER_LLA = 0x0CB3;
const char32_t KANNADA_LETTER_VA = 0x0CB5;
const char32_t KANNADA_LETTER_SHA = 0x0CB6;
const char32_t KANNADA_LETTER_SSA = 0x0CB7;
const char32_t KANNADA_LETTER_SA = 0x0CB8;
const char32_t KANNADA_LETTER_HA = 0x0CB9;
const char32_t KANNADA_SIGN_NUKTA = 0x0CBC;
const char32_t KANNADA_SIGN_AVAGRAHA = 0x0CBD;
const char32_t KANNADA_VOWEL_SIGN_AA = 0x0CBE;
const char32_t KANNADA_VOWEL_SIGN_I = 0x0CBF;
const char32_t KANNADA_VOWEL_SIGN_II = 0x0CC0;
const char32_t KANNADA_VOWEL_SIGN_U = 0x0CC1;
const char32_t KANNADA_VOWEL_SIGN_UU = 0x0CC2;
const char32_t KANNADA_VOWEL_SIGN_VOCALIC_R = 0x0CC3;
const char32_t KANNADA_VOWEL_SIGN_VOCALIC_RR = 0x0CC4;
const char32_t KANNADA_VOWEL_SIGN_E = 0x0CC6;
const char32_t KANNADA_VOWEL_SIGN_EE = 0x0CC7;
const char32_t KANNADA_VOWEL_SIGN_AI = 0x0CC8;
const char32_t KANNADA_VOWEL_SIGN_O = 0x0CCA;
const char32_t KANNADA_VOWEL_SIGN_OO = 0x0CCB;
const char32_t KANNADA_VOWEL_SIGN_AU = 0x0CCC;
const char32_t KANNADA_SIGN_VIRAMA = 0x0CCD;
const char32_t KANNADA_LENGTH_MARK = 0x0CD5;
const char32_t KANNADA_AI_LENGTH_MARK = 0x0CD6;
const char32_t KANNADA_ARCHAIC_SHRII = 0x0CDC;
const char32_t KANNADA_LETTER_NAKAARA_POLLU = 0x0CDD;
const char32_t KANNADA_LETTER_FA = 0x0CDE;
const char32_t KANNADA_LETTER_VOCALIC_RR = 0x0CE0;
const char32_t KANNADA_LETTER_VOCALIC_LL = 0x0CE1;
const char32_t KANNADA_VOWEL_SIGN_VOCALIC_L = 0x0CE2;
const char32_t KANNADA_VOWEL_SIGN_VOCALIC_LL = 0x0CE3;
const char32_t KANNADA_DIGIT_ZERO = 0x0CE6;
const char32_t KANNADA_DIGIT_ONE = 0x0CE7;
const char32_t KANNADA_DIGIT_TWO = 0x0CE8;
const char32_t KANNADA_DIGIT_THREE = 0x0CE9;
const char32_t KANNADA_DIGIT_FOUR = 0x0CEA;
const char32_t KANNADA_DIGIT_FIVE = 0x0CEB;
const char32_t KANNADA_DIGIT_SIX = 0x0CEC;
const char32_t KANNADA_DIGIT_SEVEN = 0x0CED;
const char32_t KANNADA_DIGIT_EIGHT = 0x0CEE;
const char32_t KANNADA_DIGIT_NINE = 0x0CEF;
const char32_t KANNADA_SIGN_JIHVAMULIYA = 0x0CF1;
const char32_t KANNADA_SIGN_UPADHMANIYA = 0x0CF2;
const char32_t KANNADA_SIGN_COMBINING_ANUSVARA_ABOVE_RIGHT = 0x0CF3;
const char32_t MALAYALAM_SIGN_COMBINING_ANUSVARA_ABOVE = 0x0D00;
const char32_t MALAYALAM_SIGN_CANDRABINDU = 0x0D01;
const char32_t MALAYALAM_SIGN_ANUSVARA = 0x0D02;
const char32_t MALAYALAM_SIGN_VISARGA = 0x0D03;
const char32_t MALAYALAM_LETTER_VEDIC_ANUSVARA = 0x0D04;
const char32_t MALAYALAM_LETTER_A = 0x0D05;
const char32_t MALAYALAM_LETTER_AA = 0x0D06;
const char32_t MALAYALAM_LETTER_I = 0x0D07;
const char32_t MALAYALAM_LETTER_II = 0x0D08;
const char32_t MALAYALAM_LETTER_U = 0x0D09;
const char32_t MALAYALAM_LETTER_UU = 0x0D0A;
const char32_t MALAYALAM_LETTER_VOCALIC_R = 0x0D0B;
const char32_t MALAYALAM_LETTER_VOCALIC_L = 0x0D0C;
const char32_t MALAYALAM_LETTER_E = 0x0D0E;
const char32_t MALAYALAM_LETTER_EE = 0x0D0F;
const char32_t MALAYALAM_LETTER_AI = 0x0D10;
const char32_t MALAYALAM_LETTER_O = 0x0D12;
const char32_t MALAYALAM_LETTER_OO = 0x0D13;
const char32_t MALAYALAM_LETTER_AU = 0x0D14;
const char32_t MALAYALAM_LETTER_KA = 0x0D15;
const char32_t MALAYALAM_LETTER_KHA = 0x0D16;
const char32_t MALAYALAM_LETTER_GA = 0x0D17;
const char32_t MALAYALAM_LETTER_GHA = 0x0D18;
const char32_t MALAYALAM_LETTER_NGA = 0x0D19;
const char32_t MALAYALAM_LETTER_CA = 0x0D1A;
const char32_t MALAYALAM_LETTER_CHA = 0x0D1B;
const char32_t MALAYALAM_LETTER_JA = 0x0D1C;
const char32_t MALAYALAM_LETTER_JHA = 0x0D1D;
const char32_t MALAYALAM_LETTER_NYA = 0x0D1E;
const char32_t MALAYALAM_LETTER_TTA = 0x0D1F;
const char32_t MALAYALAM_LETTER_TTHA = 0x0D20;
const char32_t MALAYALAM_LETTER_DDA = 0x0D21;
const char32_t MALAYALAM_LETTER_DDHA = 0x0D22;
const char32_t MALAYALAM_LETTER_NNA = 0x0D23;
const char32_t MALAYALAM_LETTER_TA = 0x0D24;
const char32_t MALAYALAM_LETTER_THA = 0x0D25;
const char32_t MALAYALAM_LETTER_DA = 0x0D26;
const char32_t MALAYALAM_LETTER_DHA = 0x0D27;
const char32_t MALAYALAM_LETTER_NA = 0x0D28;
const char32_t MALAYALAM_LETTER_NNNA = 0x0D29;
const char32_t MALAYALAM_LETTER_PA = 0x0D2A;
const char32_t MALAYALAM_LETTER_PHA = 0x0D2B;
const char32_t MALAYALAM_LETTER_BA = 0x0D2C;
const char32_t MALAYALAM_LETTER_BHA = 0x0D2D;
const char32_t MALAYALAM_LETTER_MA = 0x0D2E;
const char32_t MALAYALAM_LETTER_YA = 0x0D2F;
const char32_t MALAYALAM_LETTER_RA = 0x0D30;
const char32_t MALAYALAM_LETTER_RRA = 0x0D31;
const char32_t MALAYALAM_LETTER_LA = 0x0D32;
const char32_t MALAYALAM_LETTER_LLA = 0x0D33;
const char32_t MALAYALAM_LETTER_LLLA = 0x0D34;
const char32_t MALAYALAM_LETTER_VA = 0x0D35;
const char32_t MALAYALAM_LETTER_SHA = 0x0D36;
const char32_t MALAYALAM_LETTER_SSA = 0x0D37;
const char32_t MALAYALAM_LETTER_SA = 0x0D38;
const char32_t MALAYALAM_LETTER_HA = 0x0D39;
const char32_t MALAYALAM_LETTER_TTTA = 0x0D3A;
const char32_t MALAYALAM_SIGN_VERTICAL_BAR_VIRAMA = 0x0D3B;
const char32_t MALAYALAM_SIGN_CIRCULAR_VIRAMA = 0x0D3C;
const char32_t MALAYALAM_SIGN_AVAGRAHA = 0x0D3D;
const char32_t MALAYALAM_VOWEL_SIGN_AA = 0x0D3E;
const char32_t MALAYALAM_VOWEL_SIGN_I = 0x0D3F;
const char32_t MALAYALAM_VOWEL_SIGN_II = 0x0D40;
const char32_t MALAYALAM_VOWEL_SIGN_U = 0x0D41;
const char32_t MALAYALAM_VOWEL_SIGN_UU = 0x0D42;
const char32_t MALAYALAM_VOWEL_SIGN_VOCALIC_R = 0x0D43;
const char32_t MALAYALAM_VOWEL_SIGN_VOCALIC_RR = 0x0D44;
const char32_t MALAYALAM_VOWEL_SIGN_E = 0x0D46;
const char32_t MALAYALAM_VOWEL_SIGN_EE = 0x0D47;
const char32_t MALAYALAM_VOWEL_SIGN_AI = 0x0D48;
const char32_t MALAYALAM_VOWEL_SIGN_O = 0x0D4A;
const char32_t MALAYALAM_VOWEL_SIGN_OO = 0x0D4B;
const char32_t MALAYALAM_VOWEL_SIGN_AU = 0x0D4C;
const char32_t MALAYALAM_SIGN_VIRAMA = 0x0D4D;
const char32_t MALAYALAM_LETTER_DOT_REPH = 0x0D4E;
const char32_t MALAYALAM_SIGN_PARA = 0x0D4F;
const char32_t MALAYALAM_LETTER_CHILLU_M = 0x0D54;
const char32_t MALAYALAM_LETTER_CHILLU_Y = 0x0D55;
const char32_t MALAYALAM_LETTER_CHILLU_LLL = 0x0D56;
const char32_t MALAYALAM_AU_LENGTH_MARK = 0x0D57;
const char32_t MALAYALAM_FRACTION_ONE_ONE_HUNDRED_AND_SIXTIETH = 0x0D58;
const char32_t MALAYALAM_FRACTION_ONE_FORTIETH = 0x0D59;
const char32_t MALAYALAM_FRACTION_THREE_EIGHTIETHS = 0x0D5A;
const char32_t MALAYALAM_FRACTION_ONE_TWENTIETH = 0x0D5B;
const char32_t MALAYALAM_FRACTION_ONE_TENTH = 0x0D5C;
const char32_t MALAYALAM_FRACTION_THREE_TWENTIETHS = 0x0D5D;
const char32_t MALAYALAM_FRACTION_ONE_FIFTH = 0x0D5E;
const char32_t MALAYALAM_LETTER_ARCHAIC_II = 0x0D5F;
const char32_t MALAYALAM_LETTER_VOCALIC_RR = 0x0D60;
const char32_t MALAYALAM_LETTER_VOCALIC_LL = 0x0D61;
const char32_t MALAYALAM_VOWEL_SIGN_VOCALIC_L = 0x0D62;
const char32_t MALAYALAM_VOWEL_SIGN_VOCALIC_LL = 0x0D63;
const char32_t MALAYALAM_DIGIT_ZERO = 0x0D66;
const char32_t MALAYALAM_DIGIT_ONE = 0x0D67;
const char32_t MALAYALAM_DIGIT_TWO = 0x0D68;
const char32_t MALAYALAM_DIGIT_THREE = 0x0D69;
const char32_t MALAYALAM_DIGIT_FOUR = 0x0D6A;
const char32_t MALAYALAM_DIGIT_FIVE = 0x0D6B;
const char32_t MALAYALAM_DIGIT_SIX = 0x0D6C;
const char32_t MALAYALAM_DIGIT_SEVEN = 0x0D6D;
const char32_t MALAYALAM_DIGIT_EIGHT = 0x0D6E;
const char32_t MALAYALAM_DIGIT_NINE = 0x0D6F;
const char32_t MALAYALAM_NUMBER_TEN = 0x0D70;
const char32_t MALAYALAM_NUMBER_ONE_HUNDRED = 0x0D71;
const char32_t MALAYALAM_NUMBER_ONE_THOUSAND = 0x0D72;
const char32_t MALAYALAM_FRACTION_ONE_QUARTER = 0x0D73;
const char32_t MALAYALAM_FRACTION_ONE_HALF = 0x0D74;
const char32_t MALAYALAM_FRACTION_THREE_QUARTERS = 0x0D75;
const char32_t MALAYALAM_FRACTION_ONE_SIXTEENTH = 0x0D76;
const char32_t MALAYALAM_FRACTION_ONE_EIGHTH = 0x0D77;
const char32_t MALAYALAM_FRACTION_THREE_SIXTEENTHS = 0x0D78;
const char32_t MALAYALAM_DATE_MARK = 0x0D79;
const char32_t MALAYALAM_LETTER_CHILLU_NN = 0x0D7A;
const char32_t MALAYALAM_LETTER_CHILLU_N = 0x0D7B;
const char32_t MALAYALAM_LETTER_CHILLU_RR = 0x0D7C;
const char32_t MALAYALAM_LETTER_CHILLU_L = 0x0D7D;
const char32_t MALAYALAM_LETTER_CHILLU_LL = 0x0D7E;
const char32_t MALAYALAM_LETTER_CHILLU_K = 0x0D7F;
const char32_t SINHALA_SIGN_CANDRABINDU = 0x0D81;
const char32_t SINHALA_SIGN_ANUSVARAYA = 0x0D82;
const char32_t SINHALA_SIGN_VISARGAYA = 0x0D83;
const char32_t SINHALA_LETTER_AYANNA = 0x0D85;
const char32_t SINHALA_LETTER_AAYANNA = 0x0D86;
const char32_t SINHALA_LETTER_AEYANNA = 0x0D87;
const char32_t SINHALA_LETTER_AEEYANNA = 0x0D88;
const char32_t SINHALA_LETTER_IYANNA = 0x0D89;
const char32_t SINHALA_LETTER_IIYANNA = 0x0D8A;
const char32_t SINHALA_LETTER_UYANNA = 0x0D8B;
const char32_t SINHALA_LETTER_UUYANNA = 0x0D8C;
const char32_t SINHALA_LETTER_IRUYANNA = 0x0D8D;
const char32_t SINHALA_LETTER_IRUUYANNA = 0x0D8E;
const char32_t SINHALA_LETTER_ILUYANNA = 0x0D8F;
const char32_t SINHALA_LETTER_ILUUYANNA = 0x0D90;
const char32_t SINHALA_LETTER_EYANNA = 0x0D91;
const char32_t SINHALA_LETTER_EEYANNA = 0x0D92;
const char32_t SINHALA_LETTER_AIYANNA = 0x0D93;
const char32_t SINHALA_LETTER_OYANNA = 0x0D94;
const char32_t SINHALA_LETTER_OOYANNA = 0x0D95;
const char32_t SINHALA_LETTER_AUYANNA = 0x0D96;
const char32_t SINHALA_LETTER_ALPAPRAANA_KAYANNA = 0x0D9A;
const char32_t SINHALA_LETTER_MAHAAPRAANA_KAYANNA = 0x0D9B;
const char32_t SINHALA_LETTER_ALPAPRAANA_GAYANNA = 0x0D9C;
const char32_t SINHALA_LETTER_MAHAAPRAANA_GAYANNA = 0x0D9D;
const char32_t SINHALA_LETTER_KANTAJA_NAASIKYAYA = 0x0D9E;
const char32_t SINHALA_LETTER_SANYAKA_GAYANNA = 0x0D9F;
const char32_t SINHALA_LETTER_ALPAPRAANA_CAYANNA = 0x0DA0;
const char32_t SINHALA_LETTER_MAHAAPRAANA_CAYANNA = 0x0DA1;
const char32_t SINHALA_LETTER_ALPAPRAANA_JAYANNA = 0x0DA2;
const char32_t SINHALA_LETTER_MAHAAPRAANA_JAYANNA = 0x0DA3;
const char32_t SINHALA_LETTER_TAALUJA_NAASIKYAYA = 0x0DA4;
const char32_t SINHALA_LETTER_TAALUJA_SANYOOGA_NAAKSIKYAYA = 0x0DA5;
const char32_t SINHALA_LETTER_SANYAKA_JAYANNA = 0x0DA6;
const char32_t SINHALA_LETTER_ALPAPRAANA_TTAYANNA = 0x0DA7;
const char32_t SINHALA_LETTER_MAHAAPRAANA_TTAYANNA = 0x0DA8;
const char32_t SINHALA_LETTER_ALPAPRAANA_DDAYANNA = 0x0DA9;
const char32_t SINHALA_LETTER_MAHAAPRAANA_DDAYANNA = 0x0DAA;
const char32_t SINHALA_LETTER_MUURDHAJA_NAYANNA = 0x0DAB;
const char32_t SINHALA_LETTER_SANYAKA_DDAYANNA = 0x0DAC;
const char32_t SINHALA_LETTER_ALPAPRAANA_TAYANNA = 0x0DAD;
const char32_t SINHALA_LETTER_MAHAAPRAANA_TAYANNA = 0x0DAE;
const char32_t SINHALA_LETTER_ALPAPRAANA_DAYANNA = 0x0DAF;
const char32_t SINHALA_LETTER_MAHAAPRAANA_DAYANNA = 0x0DB0;
const char32_t SINHALA_LETTER_DANTAJA_NAYANNA = 0x0DB1;
const char32_t SINHALA_LETTER_SANYAKA_DAYANNA = 0x0DB3;
const char32_t SINHALA_LETTER_ALPAPRAANA_PAYANNA = 0x0DB4;
const char32_t SINHALA_LETTER_MAHAAPRAANA_PAYANNA = 0x0DB5;
const char32_t SINHALA_LETTER_ALPAPRAANA_BAYANNA = 0x0DB6;
const char32_t SINHALA_LETTER_MAHAAPRAANA_BAYANNA = 0x0DB7;
const char32_t SINHALA_LETTER_MAYANNA = 0x0DB8;
const char32_t SINHALA_LETTER_AMBA_BAYANNA = 0x0DB9;
const char32_t SINHALA_LETTER_YAYANNA = 0x0DBA;
const char32_t SINHALA_LETTER_RAYANNA = 0x0DBB;
const char32_t SINHALA_LETTER_DANTAJA_LAYANNA = 0x0DBD;
const char32_t SINHALA_LETTER_VAYANNA = 0x0DC0;
const char32_t SINHALA_LETTER_TAALUJA_SAYANNA = 0x0DC1;
const char32_t SINHALA_LETTER_MUURDHAJA_SAYANNA = 0x0DC2;
const char32_t SINHALA_LETTER_DANTAJA_SAYANNA = 0x0DC3;
const char32_t SINHALA_LETTER_HAYANNA = 0x0DC4;
const char32_t SINHALA_LETTER_MUURDHAJA_LAYANNA = 0x0DC5;
const char32_t SINHALA_LETTER_FAYANNA = 0x0DC6;
const char32_t SINHALA_SIGN_AL_LAKUNA = 0x0DCA;
const char32_t SINHALA_VOWEL_SIGN_AELA_PILLA = 0x0DCF;
const char32_t SINHALA_VOWEL_SIGN_KETTI_AEDA_PILLA = 0x0DD0;
const char32_t SINHALA_VOWEL_SIGN_DIGA_AEDA_PILLA = 0x0DD1;
const char32_t SINHALA_VOWEL_SIGN_KETTI_IS_PILLA = 0x0DD2;
const char32_t SINHALA_VOWEL_SIGN_DIGA_IS_PILLA = 0x0DD3;
const char32_t SINHALA_VOWEL_SIGN_KETTI_PAA_PILLA = 0x0DD4;
const char32_t SINHALA_VOWEL_SIGN_DIGA_PAA_PILLA = 0x0DD6;
const char32_t SINHALA_VOWEL_SIGN_GAETTA_PILLA = 0x0DD8;
const char32_t SINHALA_VOWEL_SIGN_KOMBUVA = 0x0DD9;
const char32_t SINHALA_VOWEL_SIGN_DIGA_KOMBUVA = 0x0DDA;
const char32_t SINHALA_VOWEL_SIGN_KOMBU_DEKA = 0x0DDB;
const char32_t SINHALA_VOWEL_SIGN_KOMBUVA_HAA_AELA_PILLA = 0x0DDC;
const char32_t SINHALA_VOWEL_SIGN_KOMBUVA_HAA_DIGA_AELA_PILLA = 0x0DDD;
const char32_t SINHALA_VOWEL_SIGN_KOMBUVA_HAA_GAYANUKITTA = 0x0DDE;
const char32_t SINHALA_VOWEL_SIGN_GAYANUKITTA = 0x0DDF;
const char32_t SINHALA_LITH_DIGIT_ZERO = 0x0DE6;
const char32_t SINHALA_LITH_DIGIT_ONE = 0x0DE7;
const char32_t SINHALA_LITH_DIGIT_TWO = 0x0DE8;
const char32_t SINHALA_LITH_DIGIT_THREE = 0x0DE9;
const char32_t SINHALA_LITH_DIGIT_FOUR = 0x0DEA;
const char32_t SINHALA_LITH_DIGIT_FIVE = 0x0DEB;
const char32_t SINHALA_LITH_DIGIT_SIX = 0x0DEC;
const char32_t SINHALA_LITH_DIGIT_SEVEN = 0x0DED;
const char32_t SINHALA_LITH_DIGIT_EIGHT = 0x0DEE;
const char32_t SINHALA_LITH_DIGIT_NINE = 0x0DEF;
const char32_t SINHALA_VOWEL_SIGN_DIGA_GAETTA_PILLA = 0x0DF2;
const char32_t SINHALA_VOWEL_SIGN_DIGA_GAYANUKITTA = 0x0DF3;
const char32_t SINHALA_PUNCTUATION_KUNDDALIYA = 0x0DF4;
const char32_t THAI_CHARACTER_KO_KAI = 0x0E01;
const char32_t THAI_CHARACTER_KHO_KHAI = 0x0E02;
const char32_t THAI_CHARACTER_KHO_KHUAT = 0x0E03;
const char32_t THAI_CHARACTER_KHO_KHWAI = 0x0E04;
const char32_t THAI_CHARACTER_KHO_KHON = 0x0E05;
const char32_t THAI_CHARACTER_KHO_RAKHANG = 0x0E06;
const char32_t THAI_CHARACTER_NGO_NGU = 0x0E07;
const char32_t THAI_CHARACTER_CHO_CHAN = 0x0E08;
const char32_t THAI_CHARACTER_CHO_CHING = 0x0E09;
const char32_t THAI_CHARACTER_CHO_CHANG = 0x0E0A;
const char32_t THAI_CHARACTER_SO_SO = 0x0E0B;
const char32_t THAI_CHARACTER_CHO_CHOE = 0x0E0C;
const char32_t THAI_CHARACTER_YO_YING = 0x0E0D;
const char32_t THAI_CHARACTER_DO_CHADA = 0x0E0E;
const char32_t THAI_CHARACTER_TO_PATAK = 0x0E0F;
const char32_t THAI_CHARACTER_THO_THAN = 0x0E10;
const char32_t THAI_CHARACTER_THO_NANGMONTHO = 0x0E11;
const char32_t THAI_CHARACTER_THO_PHUTHAO = 0x0E12;
const char32_t THAI_CHARACTER_NO_NEN = 0x0E13;
const char32_t THAI_CHARACTER_DO_DEK = 0x0E14;
const char32_t THAI_CHARACTER_TO_TAO = 0x0E15;
const char32_t THAI_CHARACTER_THO_THUNG = 0x0E16;
const char32_t THAI_CHARACTER_THO_THAHAN = 0x0E17;
const char32_t THAI_CHARACTER_THO_THONG = 0x0E18;
const char32_t THAI_CHARACTER_NO_NU = 0x0E19;
const char32_t THAI_CHARACTER_BO_BAIMAI = 0x0E1A;
const char32_t THAI_CHARACTER_PO_PLA = 0x0E1B;
const char32_t THAI_CHARACTER_PHO_PHUNG = 0x0E1C;
const char32_t THAI_CHARACTER_FO_FA = 0x0E1D;
const char32_t THAI_CHARACTER_PHO_PHAN = 0x0E1E;
const char32_t THAI_CHARACTER_FO_FAN = 0x0E1F;
const char32_t THAI_CHARACTER_PHO_SAMPHAO = 0x0E20;
const char32_t THAI_CHARACTER_MO_MA = 0x0E21;
const char32_t THAI_CHARACTER_YO_YAK = 0x0E22;
const char32_t THAI_CHARACTER_RO_RUA = 0x0E23;
const char32_t THAI_CHARACTER_RU = 0x0E24;
const char32_t THAI_CHARACTER_LO_LING = 0x0E25;
const char32_t THAI_CHARACTER_LU = 0x0E26;
const char32_t THAI_CHARACTER_WO_WAEN = 0x0E27;
const char32_t THAI_CHARACTER_SO_SALA = 0x0E28;
const char32_t THAI_CHARACTER_SO_RUSI = 0x0E29;
const char32_t THAI_CHARACTER_SO_SUA = 0x0E2A;
const char32_t THAI_CHARACTER_HO_HIP = 0x0E2B;
const char32_t THAI_CHARACTER_LO_CHULA = 0x0E2C;
const char32_t THAI_CHARACTER_O_ANG = 0x0E2D;
const char32_t THAI_CHARACTER_HO_NOKHUK = 0x0E2E;
const char32_t THAI_CHARACTER_PAIYANNOI = 0x0E2F;
const char32_t THAI_CHARACTER_SARA_A = 0x0E30;
const char32_t THAI_CHARACTER_MAI_HAN_AKAT = 0x0E31;
const char32_t THAI_CHARACTER_SARA_AA = 0x0E32;
const char32_t THAI_CHARACTER_SARA_AM = 0x0E33;
const char32_t THAI_CHARACTER_SARA_I = 0x0E34;
const char32_t THAI_CHARACTER_SARA_II = 0x0E35;
const char32_t THAI_CHARACTER_SARA_UE = 0x0E36;
const char32_t THAI_CHARACTER_SARA_UEE = 0x0E37;
const char32_t THAI_CHARACTER_SARA_U = 0x0E38;
const char32_t THAI_CHARACTER_SARA_UU = 0x0E39;
const char32_t THAI_CHARACTER_PHINTHU = 0x0E3A;
const char32_t THAI_CURRENCY_SYMBOL_BAHT = 0x0E3F;
const char32_t THAI_CHARACTER_SARA_E = 0x0E40;
const char32_t THAI_CHARACTER_SARA_AE = 0x0E41;
const char32_t THAI_CHARACTER_SARA_O = 0x0E42;
const char32_t THAI_CHARACTER_SARA_AI_MAIMUAN = 0x0E43;
const char32_t THAI_CHARACTER_SARA_AI_MAIMALAI = 0x0E44;
const char32_t THAI_CHARACTER_LAKKHANGYAO = 0x0E45;
const char32_t THAI_CHARACTER_MAIYAMOK = 0x0E46;
const char32_t THAI_CHARACTER_MAITAIKHU = 0x0E47;
const char32_t THAI_CHARACTER_MAI_EK = 0x0E48;
const char32_t THAI_CHARACTER_MAI_THO = 0x0E49;
const char32_t THAI_CHARACTER_MAI_TRI = 0x0E4A;
const char32_t THAI_CHARACTER_MAI_CHATTAWA = 0x0E4B;
const char32_t THAI_CHARACTER_THANTHAKHAT = 0x0E4C;
const char32_t THAI_CHARACTER_NIKHAHIT = 0x0E4D;
const char32_t THAI_CHARACTER_YAMAKKAN = 0x0E4E;
const char32_t THAI_CHARACTER_FONGMAN = 0x0E4F;
const char32_t THAI_DIGIT_ZERO = 0x0E50;
const char32_t THAI_DIGIT_ONE = 0x0E51;
const char32_t THAI_DIGIT_TWO = 0x0E52;
const char32_t THAI_DIGIT_THREE = 0x0E53;
const char32_t THAI_DIGIT_FOUR = 0x0E54;
const char32_t THAI_DIGIT_FIVE = 0x0E55;
const char32_t THAI_DIGIT_SIX = 0x0E56;
const char32_t THAI_DIGIT_SEVEN = 0x0E57;
const char32_t THAI_DIGIT_EIGHT = 0x0E58;
const char32_t THAI_DIGIT_NINE = 0x0E59;
const char32_t THAI_CHARACTER_ANGKHANKHU = 0x0E5A;
const char32_t THAI_CHARACTER_KHOMUT = 0x0E5B;
const char32_t LAO_LETTER_KO = 0x0E81;
const char32_t LAO_LETTER_KHO_SUNG = 0x0E82;
const char32_t LAO_LETTER_KHO_TAM = 0x0E84;
const char32_t LAO_LETTER_PALI_GHA = 0x0E86;
const char32_t LAO_LETTER_NGO = 0x0E87;
const char32_t LAO_LETTER_CO = 0x0E88;
const char32_t LAO_LETTER_PALI_CHA = 0x0E89;
const char32_t LAO_LETTER_SO_TAM = 0x0E8A;
const char32_t LAO_LETTER_PALI_JHA = 0x0E8C;
const char32_t LAO_LETTER_NYO = 0x0E8D;
const char32_t LAO_LETTER_PALI_NYA = 0x0E8E;
const char32_t LAO_LETTER_PALI_TTA = 0x0E8F;
const char32_t LAO_LETTER_PALI_TTHA = 0x0E90;
const char32_t LAO_LETTER_PALI_DDA = 0x0E91;
const char32_t LAO_LETTER_PALI_DDHA = 0x0E92;
const char32_t LAO_LETTER_PALI_NNA = 0x0E93;
const char32_t LAO_LETTER_DO = 0x0E94;
const char32_t LAO_LETTER_TO = 0x0E95;
const char32_t LAO_LETTER_THO_SUNG = 0x0E96;
const char32_t LAO_LETTER_THO_TAM = 0x0E97;
const char32_t LAO_LETTER_PALI_DHA = 0x0E98;
const char32_t LAO_LETTER_NO = 0x0E99;
const char32_t LAO_LETTER_BO = 0x0E9A;
const char32_t LAO_LETTER_PO = 0x0E9B;
const char32_t LAO_LETTER_PHO_SUNG = 0x0E9C;
const char32_t LAO_LETTER_FO_TAM = 0x0E9D;
const char32_t LAO_LETTER_PHO_TAM = 0x0E9E;
const char32_t LAO_LETTER_FO_SUNG = 0x0E9F;
const char32_t LAO_LETTER_PALI_BHA = 0x0EA0;
const char32_t LAO_LETTER_MO = 0x0EA1;
const char32_t LAO_LETTER_YO = 0x0EA2;
const char32_t LAO_LETTER_LO_LING = 0x0EA3;
const char32_t LAO_LETTER_LO_LOOT = 0x0EA5;
const char32_t LAO_LETTER_WO = 0x0EA7;
const char32_t LAO_LETTER_SANSKRIT_SHA = 0x0EA8;
const char32_t LAO_LETTER_SANSKRIT_SSA = 0x0EA9;
const char32_t LAO_LETTER_SO_SUNG = 0x0EAA;
const char32_t LAO_LETTER_HO_SUNG = 0x0EAB;
const char32_t LAO_LETTER_PALI_LLA = 0x0EAC;
const char32_t LAO_LETTER_O = 0x0EAD;
const char32_t LAO_LETTER_HO_TAM = 0x0EAE;
const char32_t LAO_ELLIPSIS = 0x0EAF;
const char32_t LAO_VOWEL_SIGN_A = 0x0EB0;
const char32_t LAO_VOWEL_SIGN_MAI_KAN = 0x0EB1;
const char32_t LAO_VOWEL_SIGN_AA = 0x0EB2;
const char32_t LAO_VOWEL_SIGN_AM = 0x0EB3;
const char32_t LAO_VOWEL_SIGN_I = 0x0EB4;
const char32_t LAO_VOWEL_SIGN_II = 0x0EB5;
const char32_t LAO_VOWEL_SIGN_Y = 0x0EB6;
const char32_t LAO_VOWEL_SIGN_YY = 0x0EB7;
const char32_t LAO_VOWEL_SIGN_U = 0x0EB8;
const char32_t LAO_VOWEL_SIGN_UU = 0x0EB9;
const char32_t LAO_SIGN_PALI_VIRAMA = 0x0EBA;
const char32_t LAO_VOWEL_SIGN_MAI_KON = 0x0EBB;
const char32_t LAO_SEMIVOWEL_SIGN_LO = 0x0EBC;
const char32_t LAO_SEMIVOWEL_SIGN_NYO = 0x0EBD;
const char32_t LAO_VOWEL_SIGN_E = 0x0EC0;
const char32_t LAO_VOWEL_SIGN_EI = 0x0EC1;
const char32_t LAO_VOWEL_SIGN_O = 0x0EC2;
const char32_t LAO_VOWEL_SIGN_AY = 0x0EC3;
const char32_t LAO_VOWEL_SIGN_AI = 0x0EC4;
const char32_t LAO_KO_LA = 0x0EC6;
const char32_t LAO_TONE_MAI_EK = 0x0EC8;
const char32_t LAO_TONE_MAI_THO = 0x0EC9;
const char32_t LAO_TONE_MAI_TI = 0x0ECA;
const char32_t LAO_TONE_MAI_CATAWA = 0x0ECB;
const char32_t LAO_CANCELLATION_MARK = 0x0ECC;
const char32_t LAO_NIGGAHITA = 0x0ECD;
const char32_t LAO_YAMAKKAN = 0x0ECE;
const char32_t LAO_DIGIT_ZERO = 0x0ED0;
const char32_t LAO_DIGIT_ONE = 0x0ED1;
const char32_t LAO_DIGIT_TWO = 0x0ED2;
const char32_t LAO_DIGIT_THREE = 0x0ED3;
const char32_t LAO_DIGIT_FOUR = 0x0ED4;
const char32_t LAO_DIGIT_FIVE = 0x0ED5;
const char32_t LAO_DIGIT_SIX = 0x0ED6;
const char32_t LAO_DIGIT_SEVEN = 0x0ED7;
const char32_t LAO_DIGIT_EIGHT = 0x0ED8;
const char32_t LAO_DIGIT_NINE = 0x0ED9;
const char32_t LAO_HO_NO = 0x0EDC;
const char32_t LAO_HO_MO = 0x0EDD;
const char32_t LAO_LETTER_KHMU_GO = 0x0EDE;
const char32_t LAO_LETTER_KHMU_NYO = 0x0EDF;
const char32_t TIBETAN_SYLLABLE_OM = 0x0F00;
const char32_t TIBETAN_MARK_GTER_YIG_MGO_TRUNCATED_A = 0x0F01;
const char32_t TIBETAN_MARK_GTER_YIG_MGO__UM_RNAM_BCAD_MA = 0x0F02;
const char32_t TIBETAN_MARK_GTER_YIG_MGO__UM_GTER_TSHEG_MA = 0x0F03;
const char32_t TIBETAN_MARK_INITIAL_YIG_MGO_MDUN_MA = 0x0F04;
const char32_t TIBETAN_MARK_CLOSING_YIG_MGO_SGAB_MA = 0x0F05;
const char32_t TIBETAN_MARK_CARET_YIG_MGO_PHUR_SHAD_MA = 0x0F06;
const char32_t TIBETAN_MARK_YIG_MGO_TSHEG_SHAD_MA = 0x0F07;
const char32_t TIBETAN_MARK_SBRUL_SHAD = 0x0F08;
const char32_t TIBETAN_MARK_BSKUR_YIG_MGO = 0x0F09;
const char32_t TIBETAN_MARK_BKA__SHOG_YIG_MGO = 0x0F0A;
const char32_t TIBETAN_MARK_INTERSYLLABIC_TSHEG = 0x0F0B;
const char32_t TIBETAN_MARK_DELIMITER_TSHEG_BSTAR = 0x0F0C;
const char32_t TIBETAN_MARK_SHAD = 0x0F0D;
const char32_t TIBETAN_MARK_NYIS_SHAD = 0x0F0E;
const char32_t TIBETAN_MARK_TSHEG_SHAD = 0x0F0F;
const char32_t TIBETAN_MARK_NYIS_TSHEG_SHAD = 0x0F10;
const char32_t TIBETAN_MARK_RIN_CHEN_SPUNGS_SHAD = 0x0F11;
const char32_t TIBETAN_MARK_RGYA_GRAM_SHAD = 0x0F12;
const char32_t TIBETAN_MARK_CARET__DZUD_RTAGS_ME_LONG_CAN = 0x0F13;
const char32_t TIBETAN_MARK_GTER_TSHEG = 0x0F14;
const char32_t TIBETAN_LOGOTYPE_SIGN_CHAD_RTAGS = 0x0F15;
const char32_t TIBETAN_LOGOTYPE_SIGN_LHAG_RTAGS = 0x0F16;
const char32_t TIBETAN_ASTROLOGICAL_SIGN_SGRA_GCAN__CHAR_RTAGS = 0x0F17;
const char32_t TIBETAN_ASTROLOGICAL_SIGN__KHYUD_PA = 0x0F18;
const char32_t TIBETAN_ASTROLOGICAL_SIGN_SDONG_TSHUGS = 0x0F19;
const char32_t TIBETAN_SIGN_RDEL_DKAR_GCIG = 0x0F1A;
const char32_t TIBETAN_SIGN_RDEL_DKAR_GNYIS = 0x0F1B;
const char32_t TIBETAN_SIGN_RDEL_DKAR_GSUM = 0x0F1C;
const char32_t TIBETAN_SIGN_RDEL_NAG_GCIG = 0x0F1D;
const char32_t TIBETAN_SIGN_RDEL_NAG_GNYIS = 0x0F1E;
const char32_t TIBETAN_SIGN_RDEL_DKAR_RDEL_NAG = 0x0F1F;
const char32_t TIBETAN_DIGIT_ZERO = 0x0F20;
const char32_t TIBETAN_DIGIT_ONE = 0x0F21;
const char32_t TIBETAN_DIGIT_TWO = 0x0F22;
const char32_t TIBETAN_DIGIT_THREE = 0x0F23;
const char32_t TIBETAN_DIGIT_FOUR = 0x0F24;
const char32_t TIBETAN_DIGIT_FIVE = 0x0F25;
const char32_t TIBETAN_DIGIT_SIX = 0x0F26;
const char32_t TIBETAN_DIGIT_SEVEN = 0x0F27;
const char32_t TIBETAN_DIGIT_EIGHT = 0x0F28;
const char32_t TIBETAN_DIGIT_NINE = 0x0F29;
const char32_t TIBETAN_DIGIT_HALF_ONE = 0x0F2A;
const char32_t TIBETAN_DIGIT_HALF_TWO = 0x0F2B;
const char32_t TIBETAN_DIGIT_HALF_THREE = 0x0F2C;
const char32_t TIBETAN_DIGIT_HALF_FOUR = 0x0F2D;
const char32_t TIBETAN_DIGIT_HALF_FIVE = 0x0F2E;
const char32_t TIBETAN_DIGIT_HALF_SIX = 0x0F2F;
const char32_t TIBETAN_DIGIT_HALF_SEVEN = 0x0F30;
const char32_t TIBETAN_DIGIT_HALF_EIGHT = 0x0F31;
const char32_t TIBETAN_DIGIT_HALF_NINE = 0x0F32;
const char32_t TIBETAN_DIGIT_HALF_ZERO = 0x0F33;
const char32_t TIBETAN_MARK_BSDUS_RTAGS = 0x0F34;
const char32_t TIBETAN_MARK_NGAS_BZUNG_NYI_ZLA = 0x0F35;
const char32_t TIBETAN_MARK_CARET__DZUD_RTAGS_BZHI_MIG_CAN = 0x0F36;
const char32_t TIBETAN_MARK_NGAS_BZUNG_SGOR_RTAGS = 0x0F37;
const char32_t TIBETAN_MARK_CHE_MGO = 0x0F38;
const char32_t TIBETAN_MARK_TSA__PHRU = 0x0F39;
const char32_t TIBETAN_MARK_GUG_RTAGS_GYON = 0x0F3A;
const char32_t TIBETAN_MARK_GUG_RTAGS_GYAS = 0x0F3B;
const char32_t TIBETAN_MARK_ANG_KHANG_GYON = 0x0F3C;
const char32_t TIBETAN_MARK_ANG_KHANG_GYAS = 0x0F3D;
const char32_t TIBETAN_SIGN_YAR_TSHES = 0x0F3E;
const char32_t TIBETAN_SIGN_MAR_TSHES = 0x0F3F;
const char32_t TIBETAN_LETTER_KA = 0x0F40;
const char32_t TIBETAN_LETTER_KHA = 0x0F41;
const char32_t TIBETAN_LETTER_GA = 0x0F42;
const char32_t TIBETAN_LETTER_GHA = 0x0F43;
const char32_t TIBETAN_LETTER_NGA = 0x0F44;
const char32_t TIBETAN_LETTER_CA = 0x0F45;
const char32_t TIBETAN_LETTER_CHA = 0x0F46;
const char32_t TIBETAN_LETTER_JA = 0x0F47;
const char32_t TIBETAN_LETTER_NYA = 0x0F49;
const char32_t TIBETAN_LETTER_TTA = 0x0F4A;
const char32_t TIBETAN_LETTER_TTHA = 0x0F4B;
const char32_t TIBETAN_LETTER_DDA = 0x0F4C;
const char32_t TIBETAN_LETTER_DDHA = 0x0F4D;
const char32_t TIBETAN_LETTER_NNA = 0x0F4E;
const char32_t TIBETAN_LETTER_TA = 0x0F4F;
const char32_t TIBETAN_LETTER_THA = 0x0F50;
const char32_t TIBETAN_LETTER_DA = 0x0F51;
const char32_t TIBETAN_LETTER_DHA = 0x0F52;
const char32_t TIBETAN_LETTER_NA = 0x0F53;
const char32_t TIBETAN_LETTER_PA = 0x0F54;
const char32_t TIBETAN_LETTER_PHA = 0x0F55;
const char32_t TIBETAN_LETTER_BA = 0x0F56;
const char32_t TIBETAN_LETTER_BHA = 0x0F57;
const char32_t TIBETAN_LETTER_MA = 0x0F58;
const char32_t TIBETAN_LETTER_TSA = 0x0F59;
const char32_t TIBETAN_LETTER_TSHA = 0x0F5A;
const char32_t TIBETAN_LETTER_DZA = 0x0F5B;
const char32_t TIBETAN_LETTER_DZHA = 0x0F5C;
const char32_t TIBETAN_LETTER_WA = 0x0F5D;
const char32_t TIBETAN_LETTER_ZHA = 0x0F5E;
const char32_t TIBETAN_LETTER_ZA = 0x0F5F;
const char32_t TIBETAN_LETTER__A = 0x0F60;
const char32_t TIBETAN_LETTER_YA = 0x0F61;
const char32_t TIBETAN_LETTER_RA = 0x0F62;
const char32_t TIBETAN_LETTER_LA = 0x0F63;
const char32_t TIBETAN_LETTER_SHA = 0x0F64;
const char32_t TIBETAN_LETTER_SSA = 0x0F65;
const char32_t TIBETAN_LETTER_SA = 0x0F66;
const char32_t TIBETAN_LETTER_HA = 0x0F67;
const char32_t TIBETAN_LETTER_A = 0x0F68;
const char32_t TIBETAN_LETTER_KSSA = 0x0F69;
const char32_t TIBETAN_LETTER_FIXED_FORM_RA = 0x0F6A;
const char32_t TIBETAN_LETTER_KKA = 0x0F6B;
const char32_t TIBETAN_LETTER_RRA = 0x0F6C;
const char32_t TIBETAN_VOWEL_SIGN_AA = 0x0F71;
const char32_t TIBETAN_VOWEL_SIGN_I = 0x0F72;
const char32_t TIBETAN_VOWEL_SIGN_II = 0x0F73;
const char32_t TIBETAN_VOWEL_SIGN_U = 0x0F74;
const char32_t TIBETAN_VOWEL_SIGN_UU = 0x0F75;
const char32_t TIBETAN_VOWEL_SIGN_VOCALIC_R = 0x0F76;
const char32_t TIBETAN_VOWEL_SIGN_VOCALIC_RR = 0x0F77;
const char32_t TIBETAN_VOWEL_SIGN_VOCALIC_L = 0x0F78;
const char32_t TIBETAN_VOWEL_SIGN_VOCALIC_LL = 0x0F79;
const char32_t TIBETAN_VOWEL_SIGN_E = 0x0F7A;
const char32_t TIBETAN_VOWEL_SIGN_EE = 0x0F7B;
const char32_t TIBETAN_VOWEL_SIGN_O = 0x0F7C;
const char32_t TIBETAN_VOWEL_SIGN_OO = 0x0F7D;
const char32_t TIBETAN_SIGN_RJES_SU_NGA_RO = 0x0F7E;
const char32_t TIBETAN_SIGN_RNAM_BCAD = 0x0F7F;
const char32_t TIBETAN_VOWEL_SIGN_REVERSED_I = 0x0F80;
const char32_t TIBETAN_VOWEL_SIGN_REVERSED_II = 0x0F81;
const char32_t TIBETAN_SIGN_NYI_ZLA_NAA_DA = 0x0F82;
const char32_t TIBETAN_SIGN_SNA_LDAN = 0x0F83;
const char32_t TIBETAN_MARK_HALANTA = 0x0F84;
const char32_t TIBETAN_MARK_PALUTA = 0x0F85;
const char32_t TIBETAN_SIGN_LCI_RTAGS = 0x0F86;
const char32_t TIBETAN_SIGN_YANG_RTAGS = 0x0F87;
const char32_t TIBETAN_SIGN_LCE_TSA_CAN = 0x0F88;
const char32_t TIBETAN_SIGN_MCHU_CAN = 0x0F89;
const char32_t TIBETAN_SIGN_GRU_CAN_RGYINGS = 0x0F8A;
const char32_t TIBETAN_SIGN_GRU_MED_RGYINGS = 0x0F8B;
const char32_t TIBETAN_SIGN_INVERTED_MCHU_CAN = 0x0F8C;
const char32_t TIBETAN_SUBJOINED_SIGN_LCE_TSA_CAN = 0x0F8D;
const char32_t TIBETAN_SUBJOINED_SIGN_MCHU_CAN = 0x0F8E;
const char32_t TIBETAN_SUBJOINED_SIGN_INVERTED_MCHU_CAN = 0x0F8F;
const char32_t TIBETAN_SUBJOINED_LETTER_KA = 0x0F90;
const char32_t TIBETAN_SUBJOINED_LETTER_KHA = 0x0F91;
const char32_t TIBETAN_SUBJOINED_LETTER_GA = 0x0F92;
const char32_t TIBETAN_SUBJOINED_LETTER_GHA = 0x0F93;
const char32_t TIBETAN_SUBJOINED_LETTER_NGA = 0x0F94;
const char32_t TIBETAN_SUBJOINED_LETTER_CA = 0x0F95;
const char32_t TIBETAN_SUBJOINED_LETTER_CHA = 0x0F96;
const char32_t TIBETAN_SUBJOINED_LETTER_JA = 0x0F97;
const char32_t TIBETAN_SUBJOINED_LETTER_NYA = 0x0F99;
const char32_t TIBETAN_SUBJOINED_LETTER_TTA = 0x0F9A;
const char32_t TIBETAN_SUBJOINED_LETTER_TTHA = 0x0F9B;
const char32_t TIBETAN_SUBJOINED_LETTER_DDA = 0x0F9C;
const char32_t TIBETAN_SUBJOINED_LETTER_DDHA = 0x0F9D;
const char32_t TIBETAN_SUBJOINED_LETTER_NNA = 0x0F9E;
const char32_t TIBETAN_SUBJOINED_LETTER_TA = 0x0F9F;
const char32_t TIBETAN_SUBJOINED_LETTER_THA = 0x0FA0;
const char32_t TIBETAN_SUBJOINED_LETTER_DA = 0x0FA1;
const char32_t TIBETAN_SUBJOINED_LETTER_DHA = 0x0FA2;
const char32_t TIBETAN_SUBJOINED_LETTER_NA = 0x0FA3;
const char32_t TIBETAN_SUBJOINED_LETTER_PA = 0x0FA4;
const char32_t TIBETAN_SUBJOINED_LETTER_PHA = 0x0FA5;
const char32_t TIBETAN_SUBJOINED_LETTER_BA = 0x0FA6;
const char32_t TIBETAN_SUBJOINED_LETTER_BHA = 0x0FA7;
const char32_t TIBETAN_SUBJOINED_LETTER_MA = 0x0FA8;
const char32_t TIBETAN_SUBJOINED_LETTER_TSA = 0x0FA9;
const char32_t TIBETAN_SUBJOINED_LETTER_TSHA = 0x0FAA;
const char32_t TIBETAN_SUBJOINED_LETTER_DZA = 0x0FAB;
const char32_t TIBETAN_SUBJOINED_LETTER_DZHA = 0x0FAC;
const char32_t TIBETAN_SUBJOINED_LETTER_WA = 0x0FAD;
const char32_t TIBETAN_SUBJOINED_LETTER_ZHA = 0x0FAE;
const char32_t TIBETAN_SUBJOINED_LETTER_ZA = 0x0FAF;
const char32_t TIBETAN_SUBJOINED_LETTER__A = 0x0FB0;
const char32_t TIBETAN_SUBJOINED_LETTER_YA = 0x0FB1;
const char32_t TIBETAN_SUBJOINED_LETTER_RA = 0x0FB2;
const char32_t TIBETAN_SUBJOINED_LETTER_LA = 0x0FB3;
const char32_t TIBETAN_SUBJOINED_LETTER_SHA = 0x0FB4;
const char32_t TIBETAN_SUBJOINED_LETTER_SSA = 0x0FB5;
const char32_t TIBETAN_SUBJOINED_LETTER_SA = 0x0FB6;
const char32_t TIBETAN_SUBJOINED_LETTER_HA = 0x0FB7;
const char32_t TIBETAN_SUBJOINED_LETTER_A = 0x0FB8;
const char32_t TIBETAN_SUBJOINED_LETTER_KSSA = 0x0FB9;
const char32_t TIBETAN_SUBJOINED_LETTER_FIXED_FORM_WA = 0x0FBA;
const char32_t TIBETAN_SUBJOINED_LETTER_FIXED_FORM_YA = 0x0FBB;
const char32_t TIBETAN_SUBJOINED_LETTER_FIXED_FORM_RA = 0x0FBC;
const char32_t TIBETAN_KU_RU_KHA = 0x0FBE;
const char32_t TIBETAN_KU_RU_KHA_BZHI_MIG_CAN = 0x0FBF;
const char32_t TIBETAN_CANTILLATION_SIGN_HEAVY_BEAT = 0x0FC0;
const char32_t TIBETAN_CANTILLATION_SIGN_LIGHT_BEAT = 0x0FC1;
const char32_t TIBETAN_CANTILLATION_SIGN_CANG_TE_U = 0x0FC2;
const char32_t TIBETAN_CANTILLATION_SIGN_SBUB__CHAL = 0x0FC3;
const char32_t TIBETAN_SYMBOL_DRIL_BU = 0x0FC4;
const char32_t TIBETAN_SYMBOL_RDO_RJE = 0x0FC5;
const char32_t TIBETAN_SYMBOL_PADMA_GDAN = 0x0FC6;
const char32_t TIBETAN_SYMBOL_RDO_RJE_RGYA_GRAM = 0x0FC7;
const char32_t TIBETAN_SYMBOL_PHUR_PA = 0x0FC8;
const char32_t TIBETAN_SYMBOL_NOR_BU = 0x0FC9;
const char32_t TIBETAN_SYMBOL_NOR_BU_NYIS__KHYIL = 0x0FCA;
const char32_t TIBETAN_SYMBOL_NOR_BU_GSUM__KHYIL = 0x0FCB;
const char32_t TIBETAN_SYMBOL_NOR_BU_BZHI__KHYIL = 0x0FCC;
const char32_t TIBETAN_SIGN_RDEL_NAG_RDEL_DKAR = 0x0FCE;
const char32_t TIBETAN_SIGN_RDEL_NAG_GSUM = 0x0FCF;
const char32_t TIBETAN_MARK_BSKA__SHOG_GI_MGO_RGYAN = 0x0FD0;
const char32_t TIBETAN_MARK_MNYAM_YIG_GI_MGO_RGYAN = 0x0FD1;
const char32_t TIBETAN_MARK_NYIS_TSHEG = 0x0FD2;
const char32_t TIBETAN_MARK_INITIAL_BRDA_RNYING_YIG_MGO_MDUN_MA = 0x0FD3;
const char32_t TIBETAN_MARK_CLOSING_BRDA_RNYING_YIG_MGO_SGAB_MA = 0x0FD4;
const char32_t RIGHT_FACING_SVASTI_SIGN = 0x0FD5;
const char32_t LEFT_FACING_SVASTI_SIGN = 0x0FD6;
const char32_t RIGHT_FACING_SVASTI_SIGN_WITH_DOTS = 0x0FD7;
const char32_t LEFT_FACING_SVASTI_SIGN_WITH_DOTS = 0x0FD8;
const char32_t TIBETAN_MARK_LEADING_MCHAN_RTAGS = 0x0FD9;
const char32_t TIBETAN_MARK_TRAILING_MCHAN_RTAGS = 0x0FDA;
const char32_t MYANMAR_LETTER_KA = 0x1000;
const char32_t MYANMAR_LETTER_KHA = 0x1001;
const char32_t MYANMAR_LETTER_GA = 0x1002;
const char32_t MYANMAR_LETTER_GHA = 0x1003;
const char32_t MYANMAR_LETTER_NGA = 0x1004;
const char32_t MYANMAR_LETTER_CA = 0x1005;
const char32_t MYANMAR_LETTER_CHA = 0x1006;
const char32_t MYANMAR_LETTER_JA = 0x1007;
const char32_t MYANMAR_LETTER_JHA = 0x1008;
const char32_t MYANMAR_LETTER_NYA = 0x1009;
const char32_t MYANMAR_LETTER_NNYA = 0x100A;
const char32_t MYANMAR_LETTER_TTA = 0x100B;
const char32_t MYANMAR_LETTER_TTHA = 0x100C;
const char32_t MYANMAR_LETTER_DDA = 0x100D;
const char32_t MYANMAR_LETTER_DDHA = 0x100E;
const char32_t MYANMAR_LETTER_NNA = 0x100F;
const char32_t MYANMAR_LETTER_TA = 0x1010;
const char32_t MYANMAR_LETTER_THA = 0x1011;
const char32_t MYANMAR_LETTER_DA = 0x1012;
const char32_t MYANMAR_LETTER_DHA = 0x1013;
const char32_t MYANMAR_LETTER_NA = 0x1014;
const char32_t MYANMAR_LETTER_PA = 0x1015;
const char32_t MYANMAR_LETTER_PHA = 0x1016;
const char32_t MYANMAR_LETTER_BA = 0x1017;
const char32_t MYANMAR_LETTER_BHA = 0x1018;
const char32_t MYANMAR_LETTER_MA = 0x1019;
const char32_t MYANMAR_LETTER_YA = 0x101A;
const char32_t MYANMAR_LETTER_RA = 0x101B;
const char32_t MYANMAR_LETTER_LA = 0x101C;
const char32_t MYANMAR_LETTER_WA = 0x101D;
const char32_t MYANMAR_LETTER_SA = 0x101E;
const char32_t MYANMAR_LETTER_HA = 0x101F;
const char32_t MYANMAR_LETTER_LLA = 0x1020;
const char32_t MYANMAR_LETTER_A = 0x1021;
const char32_t MYANMAR_LETTER_SHAN_A = 0x1022;
const char32_t MYANMAR_LETTER_I = 0x1023;
const char32_t MYANMAR_LETTER_II = 0x1024;
const char32_t MYANMAR_LETTER_U = 0x1025;
const char32_t MYANMAR_LETTER_UU = 0x1026;
const char32_t MYANMAR_LETTER_E = 0x1027;
const char32_t MYANMAR_LETTER_MON_E = 0x1028;
const char32_t MYANMAR_LETTER_O = 0x1029;
const char32_t MYANMAR_LETTER_AU = 0x102A;
const char32_t MYANMAR_VOWEL_SIGN_TALL_AA = 0x102B;
const char32_t MYANMAR_VOWEL_SIGN_AA = 0x102C;
const char32_t MYANMAR_VOWEL_SIGN_I = 0x102D;
const char32_t MYANMAR_VOWEL_SIGN_II = 0x102E;
const char32_t MYANMAR_VOWEL_SIGN_U = 0x102F;
const char32_t MYANMAR_VOWEL_SIGN_UU = 0x1030;
const char32_t MYANMAR_VOWEL_SIGN_E = 0x1031;
const char32_t MYANMAR_VOWEL_SIGN_AI = 0x1032;
const char32_t MYANMAR_VOWEL_SIGN_MON_II = 0x1033;
const char32_t MYANMAR_VOWEL_SIGN_MON_O = 0x1034;
const char32_t MYANMAR_VOWEL_SIGN_E_ABOVE = 0x1035;
const char32_t MYANMAR_SIGN_ANUSVARA = 0x1036;
const char32_t MYANMAR_SIGN_DOT_BELOW = 0x1037;
const char32_t MYANMAR_SIGN_VISARGA = 0x1038;
const char32_t MYANMAR_SIGN_VIRAMA = 0x1039;
const char32_t MYANMAR_SIGN_ASAT = 0x103A;
const char32_t MYANMAR_CONSONANT_SIGN_MEDIAL_YA = 0x103B;
const char32_t MYANMAR_CONSONANT_SIGN_MEDIAL_RA = 0x103C;
const char32_t MYANMAR_CONSONANT_SIGN_MEDIAL_WA = 0x103D;
const char32_t MYANMAR_CONSONANT_SIGN_MEDIAL_HA = 0x103E;
const char32_t MYANMAR_LETTER_GREAT_SA = 0x103F;
const char32_t MYANMAR_DIGIT_ZERO = 0x1040;
const char32_t MYANMAR_DIGIT_ONE = 0x1041;
const char32_t MYANMAR_DIGIT_TWO = 0x1042;
const char32_t MYANMAR_DIGIT_THREE = 0x1043;
const char32_t MYANMAR_DIGIT_FOUR = 0x1044;
const char32_t MYANMAR_DIGIT_FIVE = 0x1045;
const char32_t MYANMAR_DIGIT_SIX = 0x1046;
const char32_t MYANMAR_DIGIT_SEVEN = 0x1047;
const char32_t MYANMAR_DIGIT_EIGHT = 0x1048;
const char32_t MYANMAR_DIGIT_NINE = 0x1049;
const char32_t MYANMAR_SIGN_LITTLE_SECTION = 0x104A;
const char32_t MYANMAR_SIGN_SECTION = 0x104B;
const char32_t MYANMAR_SYMBOL_LOCATIVE = 0x104C;
const char32_t MYANMAR_SYMBOL_COMPLETED = 0x104D;
const char32_t MYANMAR_SYMBOL_AFOREMENTIONED = 0x104E;
const char32_t MYANMAR_SYMBOL_GENITIVE = 0x104F;
const char32_t MYANMAR_LETTER_SHA = 0x1050;
const char32_t MYANMAR_LETTER_SSA = 0x1051;
const char32_t MYANMAR_LETTER_VOCALIC_R = 0x1052;
const char32_t MYANMAR_LETTER_VOCALIC_RR = 0x1053;
const char32_t MYANMAR_LETTER_VOCALIC_L = 0x1054;
const char32_t MYANMAR_LETTER_VOCALIC_LL = 0x1055;
const char32_t MYANMAR_VOWEL_SIGN_VOCALIC_R = 0x1056;
const char32_t MYANMAR_VOWEL_SIGN_VOCALIC_RR = 0x1057;
const char32_t MYANMAR_VOWEL_SIGN_VOCALIC_L = 0x1058;
const char32_t MYANMAR_VOWEL_SIGN_VOCALIC_LL = 0x1059;
const char32_t MYANMAR_LETTER_MON_NGA = 0x105A;
const char32_t MYANMAR_LETTER_MON_JHA = 0x105B;
const char32_t MYANMAR_LETTER_MON_BBA = 0x105C;
const char32_t MYANMAR_LETTER_MON_BBE = 0x105D;
const char32_t MYANMAR_CONSONANT_SIGN_MON_MEDIAL_NA = 0x105E;
const char32_t MYANMAR_CONSONANT_SIGN_MON_MEDIAL_MA = 0x105F;
const char32_t MYANMAR_CONSONANT_SIGN_MON_MEDIAL_LA = 0x1060;
const char32_t MYANMAR_LETTER_SGAW_KAREN_SHA = 0x1061;
const char32_t MYANMAR_VOWEL_SIGN_SGAW_KAREN_EU = 0x1062;
const char32_t MYANMAR_TONE_MARK_SGAW_KAREN_HATHI = 0x1063;
const char32_t MYANMAR_TONE_MARK_SGAW_KAREN_KE_PHO = 0x1064;
const char32_t MYANMAR_LETTER_WESTERN_PWO_KAREN_THA = 0x1065;
const char32_t MYANMAR_LETTER_WESTERN_PWO_KAREN_PWA = 0x1066;
const char32_t MYANMAR_VOWEL_SIGN_WESTERN_PWO_KAREN_EU = 0x1067;
const char32_t MYANMAR_VOWEL_SIGN_WESTERN_PWO_KAREN_UE = 0x1068;
const char32_t MYANMAR_SIGN_WESTERN_PWO_KAREN_TONE_1 = 0x1069;
const char32_t MYANMAR_SIGN_WESTERN_PWO_KAREN_TONE_2 = 0x106A;
const char32_t MYANMAR_SIGN_WESTERN_PWO_KAREN_TONE_3 = 0x106B;
const char32_t MYANMAR_SIGN_WESTERN_PWO_KAREN_TONE_4 = 0x106C;
const char32_t MYANMAR_SIGN_WESTERN_PWO_KAREN_TONE_5 = 0x106D;
const char32_t MYANMAR_LETTER_EASTERN_PWO_KAREN_NNA = 0x106E;
const char32_t MYANMAR_LETTER_EASTERN_PWO_KAREN_YWA = 0x106F;
const char32_t MYANMAR_LETTER_EASTERN_PWO_KAREN_GHWA = 0x1070;
const char32_t MYANMAR_VOWEL_SIGN_GEBA_KAREN_I = 0x1071;
const char32_t MYANMAR_VOWEL_SIGN_KAYAH_OE = 0x1072;
const char32_t MYANMAR_VOWEL_SIGN_KAYAH_U = 0x1073;
const char32_t MYANMAR_VOWEL_SIGN_KAYAH_EE = 0x1074;
const char32_t MYANMAR_LETTER_SHAN_KA = 0x1075;
const char32_t MYANMAR_LETTER_SHAN_KHA = 0x1076;
const char32_t MYANMAR_LETTER_SHAN_GA = 0x1077;
const char32_t MYANMAR_LETTER_SHAN_CA = 0x1078;
const char32_t MYANMAR_LETTER_SHAN_ZA = 0x1079;
const char32_t MYANMAR_LETTER_SHAN_NYA = 0x107A;
const char32_t MYANMAR_LETTER_SHAN_DA = 0x107B;
const char32_t MYANMAR_LETTER_SHAN_NA = 0x107C;
const char32_t MYANMAR_LETTER_SHAN_PHA = 0x107D;
const char32_t MYANMAR_LETTER_SHAN_FA = 0x107E;
const char32_t MYANMAR_LETTER_SHAN_BA = 0x107F;
const char32_t MYANMAR_LETTER_SHAN_THA = 0x1080;
const char32_t MYANMAR_LETTER_SHAN_HA = 0x1081;
const char32_t MYANMAR_CONSONANT_SIGN_SHAN_MEDIAL_WA = 0x1082;
const char32_t MYANMAR_VOWEL_SIGN_SHAN_AA = 0x1083;
const char32_t MYANMAR_VOWEL_SIGN_SHAN_E = 0x1084;
const char32_t MYANMAR_VOWEL_SIGN_SHAN_E_ABOVE = 0x1085;
const char32_t MYANMAR_VOWEL_SIGN_SHAN_FINAL_Y = 0x1086;
const char32_t MYANMAR_SIGN_SHAN_TONE_2 = 0x1087;
const char32_t MYANMAR_SIGN_SHAN_TONE_3 = 0x1088;
const char32_t MYANMAR_SIGN_SHAN_TONE_5 = 0x1089;
const char32_t MYANMAR_SIGN_SHAN_TONE_6 = 0x108A;
const char32_t MYANMAR_SIGN_SHAN_COUNCIL_TONE_2 = 0x108B;
const char32_t MYANMAR_SIGN_SHAN_COUNCIL_TONE_3 = 0x108C;
const char32_t MYANMAR_SIGN_SHAN_COUNCIL_EMPHATIC_TONE = 0x108D;
const char32_t MYANMAR_LETTER_RUMAI_PALAUNG_FA = 0x108E;
const char32_t MYANMAR_SIGN_RUMAI_PALAUNG_TONE_5 = 0x108F;
const char32_t MYANMAR_SHAN_DIGIT_ZERO = 0x1090;
const char32_t MYANMAR_SHAN_DIGIT_ONE = 0x1091;
const char32_t MYANMAR_SHAN_DIGIT_TWO = 0x1092;
const char32_t MYANMAR_SHAN_DIGIT_THREE = 0x1093;
const char32_t MYANMAR_SHAN_DIGIT_FOUR = 0x1094;
const char32_t MYANMAR_SHAN_DIGIT_FIVE = 0x1095;
const char32_t MYANMAR_SHAN_DIGIT_SIX = 0x1096;
const char32_t MYANMAR_SHAN_DIGIT_SEVEN = 0x1097;
const char32_t MYANMAR_SHAN_DIGIT_EIGHT = 0x1098;
const char32_t MYANMAR_SHAN_DIGIT_NINE = 0x1099;
const char32_t MYANMAR_SIGN_KHAMTI_TONE_1 = 0x109A;
const char32_t MYANMAR_SIGN_KHAMTI_TONE_3 = 0x109B;
const char32_t MYANMAR_VOWEL_SIGN_AITON_A = 0x109C;
const char32_t MYANMAR_VOWEL_SIGN_AITON_AI = 0x109D;
const char32_t MYANMAR_SYMBOL_SHAN_ONE = 0x109E;
const char32_t MYANMAR_SYMBOL_SHAN_EXCLAMATION = 0x109F;
const char32_t GEORGIAN_CAPITAL_LETTER_AN = 0x10A0;
const char32_t GEORGIAN_CAPITAL_LETTER_BAN = 0x10A1;
const char32_t GEORGIAN_CAPITAL_LETTER_GAN = 0x10A2;
const char32_t GEORGIAN_CAPITAL_LETTER_DON = 0x10A3;
const char32_t GEORGIAN_CAPITAL_LETTER_EN = 0x10A4;
const char32_t GEORGIAN_CAPITAL_LETTER_VIN = 0x10A5;
const char32_t GEORGIAN_CAPITAL_LETTER_ZEN = 0x10A6;
const char32_t GEORGIAN_CAPITAL_LETTER_TAN = 0x10A7;
const char32_t GEORGIAN_CAPITAL_LETTER_IN = 0x10A8;
const char32_t GEORGIAN_CAPITAL_LETTER_KAN = 0x10A9;
const char32_t GEORGIAN_CAPITAL_LETTER_LAS = 0x10AA;
const char32_t GEORGIAN_CAPITAL_LETTER_MAN = 0x10AB;
const char32_t GEORGIAN_CAPITAL_LETTER_NAR = 0x10AC;
const char32_t GEORGIAN_CAPITAL_LETTER_ON = 0x10AD;
const char32_t GEORGIAN_CAPITAL_LETTER_PAR = 0x10AE;
const char32_t GEORGIAN_CAPITAL_LETTER_ZHAR = 0x10AF;
const char32_t GEORGIAN_CAPITAL_LETTER_RAE = 0x10B0;
const char32_t GEORGIAN_CAPITAL_LETTER_SAN = 0x10B1;
const char32_t GEORGIAN_CAPITAL_LETTER_TAR = 0x10B2;
const char32_t GEORGIAN_CAPITAL_LETTER_UN = 0x10B3;
const char32_t GEORGIAN_CAPITAL_LETTER_PHAR = 0x10B4;
const char32_t GEORGIAN_CAPITAL_LETTER_KHAR = 0x10B5;
const char32_t GEORGIAN_CAPITAL_LETTER_GHAN = 0x10B6;
const char32_t GEORGIAN_CAPITAL_LETTER_QAR = 0x10B7;
const char32_t GEORGIAN_CAPITAL_LETTER_SHIN = 0x10B8;
const char32_t GEORGIAN_CAPITAL_LETTER_CHIN = 0x10B9;
const char32_t GEORGIAN_CAPITAL_LETTER_CAN = 0x10BA;
const char32_t GEORGIAN_CAPITAL_LETTER_JIL = 0x10BB;
const char32_t GEORGIAN_CAPITAL_LETTER_CIL = 0x10BC;
const char32_t GEORGIAN_CAPITAL_LETTER_CHAR = 0x10BD;
const char32_t GEORGIAN_CAPITAL_LETTER_XAN = 0x10BE;
const char32_t GEORGIAN_CAPITAL_LETTER_JHAN = 0x10BF;
const char32_t GEORGIAN_CAPITAL_LETTER_HAE = 0x10C0;
const char32_t GEORGIAN_CAPITAL_LETTER_HE = 0x10C1;
const char32_t GEORGIAN_CAPITAL_LETTER_HIE = 0x10C2;
const char32_t GEORGIAN_CAPITAL_LETTER_WE = 0x10C3;
const char32_t GEORGIAN_CAPITAL_LETTER_HAR = 0x10C4;
const char32_t GEORGIAN_CAPITAL_LETTER_HOE = 0x10C5;
const char32_t GEORGIAN_CAPITAL_LETTER_YN = 0x10C7;
const char32_t GEORGIAN_CAPITAL_LETTER_AEN = 0x10CD;
const char32_t GEORGIAN_LETTER_AN = 0x10D0;
const char32_t GEORGIAN_LETTER_BAN = 0x10D1;
const char32_t GEORGIAN_LETTER_GAN = 0x10D2;
const char32_t GEORGIAN_LETTER_DON = 0x10D3;
const char32_t GEORGIAN_LETTER_EN = 0x10D4;
const char32_t GEORGIAN_LETTER_VIN = 0x10D5;
const char32_t GEORGIAN_LETTER_ZEN = 0x10D6;
const char32_t GEORGIAN_LETTER_TAN = 0x10D7;
const char32_t GEORGIAN_LETTER_IN = 0x10D8;
const char32_t GEORGIAN_LETTER_KAN = 0x10D9;
const char32_t GEORGIAN_LETTER_LAS = 0x10DA;
const char32_t GEORGIAN_LETTER_MAN = 0x10DB;
const char32_t GEORGIAN_LETTER_NAR = 0x10DC;
const char32_t GEORGIAN_LETTER_ON = 0x10DD;
const char32_t GEORGIAN_LETTER_PAR = 0x10DE;
const char32_t GEORGIAN_LETTER_ZHAR = 0x10DF;
const char32_t GEORGIAN_LETTER_RAE = 0x10E0;
const char32_t GEORGIAN_LETTER_SAN = 0x10E1;
const char32_t GEORGIAN_LETTER_TAR = 0x10E2;
const char32_t GEORGIAN_LETTER_UN = 0x10E3;
const char32_t GEORGIAN_LETTER_PHAR = 0x10E4;
const char32_t GEORGIAN_LETTER_KHAR = 0x10E5;
const char32_t GEORGIAN_LETTER_GHAN = 0x10E6;
const char32_t GEORGIAN_LETTER_QAR = 0x10E7;
const char32_t GEORGIAN_LETTER_SHIN = 0x10E8;
const char32_t GEORGIAN_LETTER_CHIN = 0x10E9;
const char32_t GEORGIAN_LETTER_CAN = 0x10EA;
const char32_t GEORGIAN_LETTER_JIL = 0x10EB;
const char32_t GEORGIAN_LETTER_CIL = 0x10EC;
const char32_t GEORGIAN_LETTER_CHAR = 0x10ED;
const char32_t GEORGIAN_LETTER_XAN = 0x10EE;
const char32_t GEORGIAN_LETTER_JHAN = 0x10EF;
const char32_t GEORGIAN_LETTER_HAE = 0x10F0;
const char32_t GEORGIAN_LETTER_HE = 0x10F1;
const char32_t GEORGIAN_LETTER_HIE = 0x10F2;
const char32_t GEORGIAN_LETTER_WE = 0x10F3;
const char32_t GEORGIAN_LETTER_HAR = 0x10F4;
const char32_t GEORGIAN_LETTER_HOE = 0x10F5;
const char32_t GEORGIAN_LETTER_FI = 0x10F6;
const char32_t GEORGIAN_LETTER_YN = 0x10F7;
const char32_t GEORGIAN_LETTER_ELIFI = 0x10F8;
const char32_t GEORGIAN_LETTER_TURNED_GAN = 0x10F9;
const char32_t GEORGIAN_LETTER_AIN = 0x10FA;
const char32_t GEORGIAN_PARAGRAPH_SEPARATOR = 0x10FB;
const char32_t MODIFIER_LETTER_GEORGIAN_NAR = 0x10FC;
const char32_t GEORGIAN_LETTER_AEN = 0x10FD;
const char32_t GEORGIAN_LETTER_HARD_SIGN = 0x10FE;
const char32_t GEORGIAN_LETTER_LABIAL_SIGN = 0x10FF;
const char32_t HANGUL_CHOSEONG_KIYEOK = 0x1100;
const char32_t HANGUL_CHOSEONG_SSANGKIYEOK = 0x1101;
const char32_t HANGUL_CHOSEONG_NIEUN = 0x1102;
const char32_t HANGUL_CHOSEONG_TIKEUT = 0x1103;
const char32_t HANGUL_CHOSEONG_SSANGTIKEUT = 0x1104;
const char32_t HANGUL_CHOSEONG_RIEUL = 0x1105;
const char32_t HANGUL_CHOSEONG_MIEUM = 0x1106;
const char32_t HANGUL_CHOSEONG_PIEUP = 0x1107;
const char32_t HANGUL_CHOSEONG_SSANGPIEUP = 0x1108;
const char32_t HANGUL_CHOSEONG_SIOS = 0x1109;
const char32_t HANGUL_CHOSEONG_SSANGSIOS = 0x110A;
const char32_t HANGUL_CHOSEONG_IEUNG = 0x110B;
const char32_t HANGUL_CHOSEONG_CIEUC = 0x110C;
const char32_t HANGUL_CHOSEONG_SSANGCIEUC = 0x110D;
const char32_t HANGUL_CHOSEONG_CHIEUCH = 0x110E;
const char32_t HANGUL_CHOSEONG_KHIEUKH = 0x110F;
const char32_t HANGUL_CHOSEONG_THIEUTH = 0x1110;
const char32_t HANGUL_CHOSEONG_PHIEUPH = 0x1111;
const char32_t HANGUL_CHOSEONG_HIEUH = 0x1112;
const char32_t HANGUL_CHOSEONG_NIEUN_KIYEOK = 0x1113;
const char32_t HANGUL_CHOSEONG_SSANGNIEUN = 0x1114;
const char32_t HANGUL_CHOSEONG_NIEUN_TIKEUT = 0x1115;
const char32_t HANGUL_CHOSEONG_NIEUN_PIEUP = 0x1116;
const char32_t HANGUL_CHOSEONG_TIKEUT_KIYEOK = 0x1117;
const char32_t HANGUL_CHOSEONG_RIEUL_NIEUN = 0x1118;
const char32_t HANGUL_CHOSEONG_SSANGRIEUL = 0x1119;
const char32_t HANGUL_CHOSEONG_RIEUL_HIEUH = 0x111A;
const char32_t HANGUL_CHOSEONG_KAPYEOUNRIEUL = 0x111B;
const char32_t HANGUL_CHOSEONG_MIEUM_PIEUP = 0x111C;
const char32_t HANGUL_CHOSEONG_KAPYEOUNMIEUM = 0x111D;
const char32_t HANGUL_CHOSEONG_PIEUP_KIYEOK = 0x111E;
const char32_t HANGUL_CHOSEONG_PIEUP_NIEUN = 0x111F;
const char32_t HANGUL_CHOSEONG_PIEUP_TIKEUT = 0x1120;
const char32_t HANGUL_CHOSEONG_PIEUP_SIOS = 0x1121;
const char32_t HANGUL_CHOSEONG_PIEUP_SIOS_KIYEOK = 0x1122;
const char32_t HANGUL_CHOSEONG_PIEUP_SIOS_TIKEUT = 0x1123;
const char32_t HANGUL_CHOSEONG_PIEUP_SIOS_PIEUP = 0x1124;
const char32_t HANGUL_CHOSEONG_PIEUP_SSANGSIOS = 0x1125;
const char32_t HANGUL_CHOSEONG_PIEUP_SIOS_CIEUC = 0x1126;
const char32_t HANGUL_CHOSEONG_PIEUP_CIEUC = 0x1127;
const char32_t HANGUL_CHOSEONG_PIEUP_CHIEUCH = 0x1128;
const char32_t HANGUL_CHOSEONG_PIEUP_THIEUTH = 0x1129;
const char32_t HANGUL_CHOSEONG_PIEUP_PHIEUPH = 0x112A;
const char32_t HANGUL_CHOSEONG_KAPYEOUNPIEUP = 0x112B;
const char32_t HANGUL_CHOSEONG_KAPYEOUNSSANGPIEUP = 0x112C;
const char32_t HANGUL_CHOSEONG_SIOS_KIYEOK = 0x112D;
const char32_t HANGUL_CHOSEONG_SIOS_NIEUN = 0x112E;
const char32_t HANGUL_CHOSEONG_SIOS_TIKEUT = 0x112F;
const char32_t HANGUL_CHOSEONG_SIOS_RIEUL = 0x1130;
const char32_t HANGUL_CHOSEONG_SIOS_MIEUM = 0x1131;
const char32_t HANGUL_CHOSEONG_SIOS_PIEUP = 0x1132;
const char32_t HANGUL_CHOSEONG_SIOS_PIEUP_KIYEOK = 0x1133;
const char32_t HANGUL_CHOSEONG_SIOS_SSANGSIOS = 0x1134;
const char32_t HANGUL_CHOSEONG_SIOS_IEUNG = 0x1135;
const char32_t HANGUL_CHOSEONG_SIOS_CIEUC = 0x1136;
const char32_t HANGUL_CHOSEONG_SIOS_CHIEUCH = 0x1137;
const char32_t HANGUL_CHOSEONG_SIOS_KHIEUKH = 0x1138;
const char32_t HANGUL_CHOSEONG_SIOS_THIEUTH = 0x1139;
const char32_t HANGUL_CHOSEONG_SIOS_PHIEUPH = 0x113A;
const char32_t HANGUL_CHOSEONG_SIOS_HIEUH = 0x113B;
const char32_t HANGUL_CHOSEONG_CHITUEUMSIOS = 0x113C;
const char32_t HANGUL_CHOSEONG_CHITUEUMSSANGSIOS = 0x113D;
const char32_t HANGUL_CHOSEONG_CEONGCHIEUMSIOS = 0x113E;
const char32_t HANGUL_CHOSEONG_CEONGCHIEUMSSANGSIOS = 0x113F;
const char32_t HANGUL_CHOSEONG_PANSIOS = 0x1140;
const char32_t HANGUL_CHOSEONG_IEUNG_KIYEOK = 0x1141;
const char32_t HANGUL_CHOSEONG_IEUNG_TIKEUT = 0x1142;
const char32_t HANGUL_CHOSEONG_IEUNG_MIEUM = 0x1143;
const char32_t HANGUL_CHOSEONG_IEUNG_PIEUP = 0x1144;
const char32_t HANGUL_CHOSEONG_IEUNG_SIOS = 0x1145;
const char32_t HANGUL_CHOSEONG_IEUNG_PANSIOS = 0x1146;
const char32_t HANGUL_CHOSEONG_SSANGIEUNG = 0x1147;
const char32_t HANGUL_CHOSEONG_IEUNG_CIEUC = 0x1148;
const char32_t HANGUL_CHOSEONG_IEUNG_CHIEUCH = 0x1149;
const char32_t HANGUL_CHOSEONG_IEUNG_THIEUTH = 0x114A;
const char32_t HANGUL_CHOSEONG_IEUNG_PHIEUPH = 0x114B;
const char32_t HANGUL_CHOSEONG_YESIEUNG = 0x114C;
const char32_t HANGUL_CHOSEONG_CIEUC_IEUNG = 0x114D;
const char32_t HANGUL_CHOSEONG_CHITUEUMCIEUC = 0x114E;
const char32_t HANGUL_CHOSEONG_CHITUEUMSSANGCIEUC = 0x114F;
const char32_t HANGUL_CHOSEONG_CEONGCHIEUMCIEUC = 0x1150;
const char32_t HANGUL_CHOSEONG_CEONGCHIEUMSSANGCIEUC = 0x1151;
const char32_t HANGUL_CHOSEONG_CHIEUCH_KHIEUKH = 0x1152;
const char32_t HANGUL_CHOSEONG_CHIEUCH_HIEUH = 0x1153;
const char32_t HANGUL_CHOSEONG_CHITUEUMCHIEUCH = 0x1154;
const char32_t HANGUL_CHOSEONG_CEONGCHIEUMCHIEUCH = 0x1155;
const char32_t HANGUL_CHOSEONG_PHIEUPH_PIEUP = 0x1156;
const char32_t HANGUL_CHOSEONG_KAPYEOUNPHIEUPH = 0x1157;
const char32_t HANGUL_CHOSEONG_SSANGHIEUH = 0x1158;
const char32_t HANGUL_CHOSEONG_YEORINHIEUH = 0x1159;
const char32_t HANGUL_CHOSEONG_KIYEOK_TIKEUT = 0x115A;
const char32_t HANGUL_CHOSEONG_NIEUN_SIOS = 0x115B;
const char32_t HANGUL_CHOSEONG_NIEUN_CIEUC = 0x115C;
const char32_t HANGUL_CHOSEONG_NIEUN_HIEUH = 0x115D;
const char32_t HANGUL_CHOSEONG_TIKEUT_RIEUL = 0x115E;
const char32_t HANGUL_CHOSEONG_FILLER = 0x115F;
const char32_t HANGUL_JUNGSEONG_FILLER = 0x1160;
const char32_t HANGUL_JUNGSEONG_A = 0x1161;
const char32_t HANGUL_JUNGSEONG_AE = 0x1162;
const char32_t HANGUL_JUNGSEONG_YA = 0x1163;
const char32_t HANGUL_JUNGSEONG_YAE = 0x1164;
const char32_t HANGUL_JUNGSEONG_EO = 0x1165;
const char32_t HANGUL_JUNGSEONG_E = 0x1166;
const char32_t HANGUL_JUNGSEONG_YEO = 0x1167;
const char32_t HANGUL_JUNGSEONG_YE = 0x1168;
const char32_t HANGUL_JUNGSEONG_O = 0x1169;
const char32_t HANGUL_JUNGSEONG_WA = 0x116A;
const char32_t HANGUL_JUNGSEONG_WAE = 0x116B;
const char32_t HANGUL_JUNGSEONG_OE = 0x116C;
const char32_t HANGUL_JUNGSEONG_YO = 0x116D;
const char32_t HANGUL_JUNGSEONG_U = 0x116E;
const char32_t HANGUL_JUNGSEONG_WEO = 0x116F;
const char32_t HANGUL_JUNGSEONG_WE = 0x1170;
const char32_t HANGUL_JUNGSEONG_WI = 0x1171;
const char32_t HANGUL_JUNGSEONG_YU = 0x1172;
const char32_t HANGUL_JUNGSEONG_EU = 0x1173;
const char32_t HANGUL_JUNGSEONG_YI = 0x1174;
const char32_t HANGUL_JUNGSEONG_I = 0x1175;
const char32_t HANGUL_JUNGSEONG_A_O = 0x1176;
const char32_t HANGUL_JUNGSEONG_A_U = 0x1177;
const char32_t HANGUL_JUNGSEONG_YA_O = 0x1178;
const char32_t HANGUL_JUNGSEONG_YA_YO = 0x1179;
const char32_t HANGUL_JUNGSEONG_EO_O = 0x117A;
const char32_t HANGUL_JUNGSEONG_EO_U = 0x117B;
const char32_t HANGUL_JUNGSEONG_EO_EU = 0x117C;
const char32_t HANGUL_JUNGSEONG_YEO_O = 0x117D;
const char32_t HANGUL_JUNGSEONG_YEO_U = 0x117E;
const char32_t HANGUL_JUNGSEONG_O_EO = 0x117F;
const char32_t HANGUL_JUNGSEONG_O_E = 0x1180;
const char32_t HANGUL_JUNGSEONG_O_YE = 0x1181;
const char32_t HANGUL_JUNGSEONG_O_O = 0x1182;
const char32_t HANGUL_JUNGSEONG_O_U = 0x1183;
const char32_t HANGUL_JUNGSEONG_YO_YA = 0x1184;
const char32_t HANGUL_JUNGSEONG_YO_YAE = 0x1185;
const char32_t HANGUL_JUNGSEONG_YO_YEO = 0x1186;
const char32_t HANGUL_JUNGSEONG_YO_O = 0x1187;
const char32_t HANGUL_JUNGSEONG_YO_I = 0x1188;
const char32_t HANGUL_JUNGSEONG_U_A = 0x1189;
const char32_t HANGUL_JUNGSEONG_U_AE = 0x118A;
const char32_t HANGUL_JUNGSEONG_U_EO_EU = 0x118B;
const char32_t HANGUL_JUNGSEONG_U_YE = 0x118C;
const char32_t HANGUL_JUNGSEONG_U_U = 0x118D;
const char32_t HANGUL_JUNGSEONG_YU_A = 0x118E;
const char32_t HANGUL_JUNGSEONG_YU_EO = 0x118F;
const char32_t HANGUL_JUNGSEONG_YU_E = 0x1190;
const char32_t HANGUL_JUNGSEONG_YU_YEO = 0x1191;
const char32_t HANGUL_JUNGSEONG_YU_YE = 0x1192;
const char32_t HANGUL_JUNGSEONG_YU_U = 0x1193;
const char32_t HANGUL_JUNGSEONG_YU_I = 0x1194;
const char32_t HANGUL_JUNGSEONG_EU_U = 0x1195;
const char32_t HANGUL_JUNGSEONG_EU_EU = 0x1196;
const char32_t HANGUL_JUNGSEONG_YI_U = 0x1197;
const char32_t HANGUL_JUNGSEONG_I_A = 0x1198;
const char32_t HANGUL_JUNGSEONG_I_YA = 0x1199;
const char32_t HANGUL_JUNGSEONG_I_O = 0x119A;
const char32_t HANGUL_JUNGSEONG_I_U = 0x119B;
const char32_t HANGUL_JUNGSEONG_I_EU = 0x119C;
const char32_t HANGUL_JUNGSEONG_I_ARAEA = 0x119D;
const char32_t HANGUL_JUNGSEONG_ARAEA = 0x119E;
const char32_t HANGUL_JUNGSEONG_ARAEA_EO = 0x119F;
const char32_t HANGUL_JUNGSEONG_ARAEA_U = 0x11A0;
const char32_t HANGUL_JUNGSEONG_ARAEA_I = 0x11A1;
const char32_t HANGUL_JUNGSEONG_SSANGARAEA = 0x11A2;
const char32_t HANGUL_JUNGSEONG_A_EU = 0x11A3;
const char32_t HANGUL_JUNGSEONG_YA_U = 0x11A4;
const char32_t HANGUL_JUNGSEONG_YEO_YA = 0x11A5;
const char32_t HANGUL_JUNGSEONG_O_YA = 0x11A6;
const char32_t HANGUL_JUNGSEONG_O_YAE = 0x11A7;
const char32_t HANGUL_JONGSEONG_KIYEOK = 0x11A8;
const char32_t HANGUL_JONGSEONG_SSANGKIYEOK = 0x11A9;
const char32_t HANGUL_JONGSEONG_KIYEOK_SIOS = 0x11AA;
const char32_t HANGUL_JONGSEONG_NIEUN = 0x11AB;
const char32_t HANGUL_JONGSEONG_NIEUN_CIEUC = 0x11AC;
const char32_t HANGUL_JONGSEONG_NIEUN_HIEUH = 0x11AD;
const char32_t HANGUL_JONGSEONG_TIKEUT = 0x11AE;
const char32_t HANGUL_JONGSEONG_RIEUL = 0x11AF;
const char32_t HANGUL_JONGSEONG_RIEUL_KIYEOK = 0x11B0;
const char32_t HANGUL_JONGSEONG_RIEUL_MIEUM = 0x11B1;
const char32_t HANGUL_JONGSEONG_RIEUL_PIEUP = 0x11B2;
const char32_t HANGUL_JONGSEONG_RIEUL_SIOS = 0x11B3;
const char32_t HANGUL_JONGSEONG_RIEUL_THIEUTH = 0x11B4;
const char32_t HANGUL_JONGSEONG_RIEUL_PHIEUPH = 0x11B5;
const char32_t HANGUL_JONGSEONG_RIEUL_HIEUH = 0x11B6;
const char32_t HANGUL_JONGSEONG_MIEUM = 0x11B7;
const char32_t HANGUL_JONGSEONG_PIEUP = 0x11B8;
const char32_t HANGUL_JONGSEONG_PIEUP_SIOS = 0x11B9;
const char32_t HANGUL_JONGSEONG_SIOS = 0x11BA;
const char32_t HANGUL_JONGSEONG_SSANGSIOS = 0x11BB;
const char32_t HANGUL_JONGSEONG_IEUNG = 0x11BC;
const char32_t HANGUL_JONGSEONG_CIEUC = 0x11BD;
const char32_t HANGUL_JONGSEONG_CHIEUCH = 0x11BE;
const char32_t HANGUL_JONGSEONG_KHIEUKH = 0x11BF;
const char32_t HANGUL_JONGSEONG_THIEUTH = 0x11C0;
const char32_t HANGUL_JONGSEONG_PHIEUPH = 0x11C1;
const char32_t HANGUL_JONGSEONG_HIEUH = 0x11C2;
const char32_t HANGUL_JONGSEONG_KIYEOK_RIEUL = 0x11C3;
const char32_t HANGUL_JONGSEONG_KIYEOK_SIOS_KIYEOK = 0x11C4;
const char32_t HANGUL_JONGSEONG_NIEUN_KIYEOK = 0x11C5;
const char32_t HANGUL_JONGSEONG_NIEUN_TIKEUT = 0x11C6;
const char32_t HANGUL_JONGSEONG_NIEUN_SIOS = 0x11C7;
const char32_t HANGUL_JONGSEONG_NIEUN_PANSIOS = 0x11C8;
const char32_t HANGUL_JONGSEONG_NIEUN_THIEUTH = 0x11C9;
const char32_t HANGUL_JONGSEONG_TIKEUT_KIYEOK = 0x11CA;
const char32_t HANGUL_JONGSEONG_TIKEUT_RIEUL = 0x11CB;
const char32_t HANGUL_JONGSEONG_RIEUL_KIYEOK_SIOS = 0x11CC;
const char32_t HANGUL_JONGSEONG_RIEUL_NIEUN = 0x11CD;
const char32_t HANGUL_JONGSEONG_RIEUL_TIKEUT = 0x11CE;
const char32_t HANGUL_JONGSEONG_RIEUL_TIKEUT_HIEUH = 0x11CF;
const char32_t HANGUL_JONGSEONG_SSANGRIEUL = 0x11D0;
const char32_t HANGUL_JONGSEONG_RIEUL_MIEUM_KIYEOK = 0x11D1;
const char32_t HANGUL_JONGSEONG_RIEUL_MIEUM_SIOS = 0x11D2;
const char32_t HANGUL_JONGSEONG_RIEUL_PIEUP_SIOS = 0x11D3;
const char32_t HANGUL_JONGSEONG_RIEUL_PIEUP_HIEUH = 0x11D4;
const char32_t HANGUL_JONGSEONG_RIEUL_KAPYEOUNPIEUP = 0x11D5;
const char32_t HANGUL_JONGSEONG_RIEUL_SSANGSIOS = 0x11D6;
const char32_t HANGUL_JONGSEONG_RIEUL_PANSIOS = 0x11D7;
const char32_t HANGUL_JONGSEONG_RIEUL_KHIEUKH = 0x11D8;
const char32_t HANGUL_JONGSEONG_RIEUL_YEORINHIEUH = 0x11D9;
const char32_t HANGUL_JONGSEONG_MIEUM_KIYEOK = 0x11DA;
const char32_t HANGUL_JONGSEONG_MIEUM_RIEUL = 0x11DB;
const char32_t HANGUL_JONGSEONG_MIEUM_PIEUP = 0x11DC;
const char32_t HANGUL_JONGSEONG_MIEUM_SIOS = 0x11DD;
const char32_t HANGUL_JONGSEONG_MIEUM_SSANGSIOS = 0x11DE;
const char32_t HANGUL_JONGSEONG_MIEUM_PANSIOS = 0x11DF;
const char32_t HANGUL_JONGSEONG_MIEUM_CHIEUCH = 0x11E0;
const char32_t HANGUL_JONGSEONG_MIEUM_HIEUH = 0x11E1;
const char32_t HANGUL_JONGSEONG_KAPYEOUNMIEUM = 0x11E2;
const char32_t HANGUL_JONGSEONG_PIEUP_RIEUL = 0x11E3;
const char32_t HANGUL_JONGSEONG_PIEUP_PHIEUPH = 0x11E4;
const char32_t HANGUL_JONGSEONG_PIEUP_HIEUH = 0x11E5;
const char32_t HANGUL_JONGSEONG_KAPYEOUNPIEUP = 0x11E6;
const char32_t HANGUL_JONGSEONG_SIOS_KIYEOK = 0x11E7;
const char32_t HANGUL_JONGSEONG_SIOS_TIKEUT = 0x11E8;
const char32_t HANGUL_JONGSEONG_SIOS_RIEUL = 0x11E9;
const char32_t HANGUL_JONGSEONG_SIOS_PIEUP = 0x11EA;
const char32_t HANGUL_JONGSEONG_PANSIOS = 0x11EB;
const char32_t HANGUL_JONGSEONG_IEUNG_KIYEOK = 0x11EC;
const char32_t HANGUL_JONGSEONG_IEUNG_SSANGKIYEOK = 0x11ED;
const char32_t HANGUL_JONGSEONG_SSANGIEUNG = 0x11EE;
const char32_t HANGUL_JONGSEONG_IEUNG_KHIEUKH = 0x11EF;
const char32_t HANGUL_JONGSEONG_YESIEUNG = 0x11F0;
const char32_t HANGUL_JONGSEONG_YESIEUNG_SIOS = 0x11F1;
const char32_t HANGUL_JONGSEONG_YESIEUNG_PANSIOS = 0x11F2;
const char32_t HANGUL_JONGSEONG_PHIEUPH_PIEUP = 0x11F3;
const char32_t HANGUL_JONGSEONG_KAPYEOUNPHIEUPH = 0x11F4;
const char32_t HANGUL_JONGSEONG_HIEUH_NIEUN = 0x11F5;
const char32_t HANGUL_JONGSEONG_HIEUH_RIEUL = 0x11F6;
const char32_t HANGUL_JONGSEONG_HIEUH_MIEUM = 0x11F7;
const char32_t HANGUL_JONGSEONG_HIEUH_PIEUP = 0x11F8;
const char32_t HANGUL_JONGSEONG_YEORINHIEUH = 0x11F9;
const char32_t HANGUL_JONGSEONG_KIYEOK_NIEUN = 0x11FA;
const char32_t HANGUL_JONGSEONG_KIYEOK_PIEUP = 0x11FB;
const char32_t HANGUL_JONGSEONG_KIYEOK_CHIEUCH = 0x11FC;
const char32_t HANGUL_JONGSEONG_KIYEOK_KHIEUKH = 0x11FD;
const char32_t HANGUL_JONGSEONG_KIYEOK_HIEUH = 0x11FE;
const char32_t HANGUL_JONGSEONG_SSANGNIEUN = 0x11FF;
const char32_t ETHIOPIC_SYLLABLE_HA = 0x1200;
const char32_t ETHIOPIC_SYLLABLE_HU = 0x1201;
const char32_t ETHIOPIC_SYLLABLE_HI = 0x1202;
const char32_t ETHIOPIC_SYLLABLE_HAA = 0x1203;
const char32_t ETHIOPIC_SYLLABLE_HEE = 0x1204;
const char32_t ETHIOPIC_SYLLABLE_HE = 0x1205;
const char32_t ETHIOPIC_SYLLABLE_HO = 0x1206;
const char32_t ETHIOPIC_SYLLABLE_HOA = 0x1207;
const char32_t ETHIOPIC_SYLLABLE_LA = 0x1208;
const char32_t ETHIOPIC_SYLLABLE_LU = 0x1209;
const char32_t ETHIOPIC_SYLLABLE_LI = 0x120A;
const char32_t ETHIOPIC_SYLLABLE_LAA = 0x120B;
const char32_t ETHIOPIC_SYLLABLE_LEE = 0x120C;
const char32_t ETHIOPIC_SYLLABLE_LE = 0x120D;
const char32_t ETHIOPIC_SYLLABLE_LO = 0x120E;
const char32_t ETHIOPIC_SYLLABLE_LWA = 0x120F;
const char32_t ETHIOPIC_SYLLABLE_HHA = 0x1210;
const char32_t ETHIOPIC_SYLLABLE_HHU = 0x1211;
const char32_t ETHIOPIC_SYLLABLE_HHI = 0x1212;
const char32_t ETHIOPIC_SYLLABLE_HHAA = 0x1213;
const char32_t ETHIOPIC_SYLLABLE_HHEE = 0x1214;
const char32_t ETHIOPIC_SYLLABLE_HHE = 0x1215;
const char32_t ETHIOPIC_SYLLABLE_HHO = 0x1216;
const char32_t ETHIOPIC_SYLLABLE_HHWA = 0x1217;
const char32_t ETHIOPIC_SYLLABLE_MA = 0x1218;
const char32_t ETHIOPIC_SYLLABLE_MU = 0x1219;
const char32_t ETHIOPIC_SYLLABLE_MI = 0x121A;
const char32_t ETHIOPIC_SYLLABLE_MAA = 0x121B;
const char32_t ETHIOPIC_SYLLABLE_MEE = 0x121C;
const char32_t ETHIOPIC_SYLLABLE_ME = 0x121D;
const char32_t ETHIOPIC_SYLLABLE_MO = 0x121E;
const char32_t ETHIOPIC_SYLLABLE_MWA = 0x121F;
const char32_t ETHIOPIC_SYLLABLE_SZA = 0x1220;
const char32_t ETHIOPIC_SYLLABLE_SZU = 0x1221;
const char32_t ETHIOPIC_SYLLABLE_SZI = 0x1222;
const char32_t ETHIOPIC_SYLLABLE_SZAA = 0x1223;
const char32_t ETHIOPIC_SYLLABLE_SZEE = 0x1224;
const char32_t ETHIOPIC_SYLLABLE_SZE = 0x1225;
const char32_t ETHIOPIC_SYLLABLE_SZO = 0x1226;
const char32_t ETHIOPIC_SYLLABLE_SZWA = 0x1227;
const char32_t ETHIOPIC_SYLLABLE_RA = 0x1228;
const char32_t ETHIOPIC_SYLLABLE_RU = 0x1229;
const char32_t ETHIOPIC_SYLLABLE_RI = 0x122A;
const char32_t ETHIOPIC_SYLLABLE_RAA = 0x122B;
const char32_t ETHIOPIC_SYLLABLE_REE = 0x122C;
const char32_t ETHIOPIC_SYLLABLE_RE = 0x122D;
const char32_t ETHIOPIC_SYLLABLE_RO = 0x122E;
const char32_t ETHIOPIC_SYLLABLE_RWA = 0x122F;
const char32_t ETHIOPIC_SYLLABLE_SA = 0x1230;
const char32_t ETHIOPIC_SYLLABLE_SU = 0x1231;
const char32_t ETHIOPIC_SYLLABLE_SI = 0x1232;
const char32_t ETHIOPIC_SYLLABLE_SAA = 0x1233;
const char32_t ETHIOPIC_SYLLABLE_SEE = 0x1234;
const char32_t ETHIOPIC_SYLLABLE_SE = 0x1235;
const char32_t ETHIOPIC_SYLLABLE_SO = 0x1236;
const char32_t ETHIOPIC_SYLLABLE_SWA = 0x1237;
const char32_t ETHIOPIC_SYLLABLE_SHA = 0x1238;
const char32_t ETHIOPIC_SYLLABLE_SHU = 0x1239;
const char32_t ETHIOPIC_SYLLABLE_SHI = 0x123A;
const char32_t ETHIOPIC_SYLLABLE_SHAA = 0x123B;
const char32_t ETHIOPIC_SYLLABLE_SHEE = 0x123C;
const char32_t ETHIOPIC_SYLLABLE_SHE = 0x123D;
const char32_t ETHIOPIC_SYLLABLE_SHO = 0x123E;
const char32_t ETHIOPIC_SYLLABLE_SHWA = 0x123F;
const char32_t ETHIOPIC_SYLLABLE_QA = 0x1240;
const char32_t ETHIOPIC_SYLLABLE_QU = 0x1241;
const char32_t ETHIOPIC_SYLLABLE_QI = 0x1242;
const char32_t ETHIOPIC_SYLLABLE_QAA = 0x1243;
const char32_t ETHIOPIC_SYLLABLE_QEE = 0x1244;
const char32_t ETHIOPIC_SYLLABLE_QE = 0x1245;
const char32_t ETHIOPIC_SYLLABLE_QO = 0x1246;
const char32_t ETHIOPIC_SYLLABLE_QOA = 0x1247;
const char32_t ETHIOPIC_SYLLABLE_QWA = 0x1248;
const char32_t ETHIOPIC_SYLLABLE_QWI = 0x124A;
const char32_t ETHIOPIC_SYLLABLE_QWAA = 0x124B;
const char32_t ETHIOPIC_SYLLABLE_QWEE = 0x124C;
const char32_t ETHIOPIC_SYLLABLE_QWE = 0x124D;
const char32_t ETHIOPIC_SYLLABLE_QHA = 0x1250;
const char32_t ETHIOPIC_SYLLABLE_QHU = 0x1251;
const char32_t ETHIOPIC_SYLLABLE_QHI = 0x1252;
const char32_t ETHIOPIC_SYLLABLE_QHAA = 0x1253;
const char32_t ETHIOPIC_SYLLABLE_QHEE = 0x1254;
const char32_t ETHIOPIC_SYLLABLE_QHE = 0x1255;
const char32_t ETHIOPIC_SYLLABLE_QHO = 0x1256;
const char32_t ETHIOPIC_SYLLABLE_QHWA = 0x1258;
const char32_t ETHIOPIC_SYLLABLE_QHWI = 0x125A;
const char32_t ETHIOPIC_SYLLABLE_QHWAA = 0x125B;
const char32_t ETHIOPIC_SYLLABLE_QHWEE = 0x125C;
const char32_t ETHIOPIC_SYLLABLE_QHWE = 0x125D;
const char32_t ETHIOPIC_SYLLABLE_BA = 0x1260;
const char32_t ETHIOPIC_SYLLABLE_BU = 0x1261;
const char32_t ETHIOPIC_SYLLABLE_BI = 0x1262;
const char32_t ETHIOPIC_SYLLABLE_BAA = 0x1263;
const char32_t ETHIOPIC_SYLLABLE_BEE = 0x1264;
const char32_t ETHIOPIC_SYLLABLE_BE = 0x1265;
const char32_t ETHIOPIC_SYLLABLE_BO = 0x1266;
const char32_t ETHIOPIC_SYLLABLE_BWA = 0x1267;
const char32_t ETHIOPIC_SYLLABLE_VA = 0x1268;
const char32_t ETHIOPIC_SYLLABLE_VU = 0x1269;
const char32_t ETHIOPIC_SYLLABLE_VI = 0x126A;
const char32_t ETHIOPIC_SYLLABLE_VAA = 0x126B;
const char32_t ETHIOPIC_SYLLABLE_VEE = 0x126C;
const char32_t ETHIOPIC_SYLLABLE_VE = 0x126D;
const char32_t ETHIOPIC_SYLLABLE_VO = 0x126E;
const char32_t ETHIOPIC_SYLLABLE_VWA = 0x126F;
const char32_t ETHIOPIC_SYLLABLE_TA = 0x1270;
const char32_t ETHIOPIC_SYLLABLE_TU = 0x1271;
const char32_t ETHIOPIC_SYLLABLE_TI = 0x1272;
const char32_t ETHIOPIC_SYLLABLE_TAA = 0x1273;
const char32_t ETHIOPIC_SYLLABLE_TEE = 0x1274;
const char32_t ETHIOPIC_SYLLABLE_TE = 0x1275;
const char32_t ETHIOPIC_SYLLABLE_TO = 0x1276;
const char32_t ETHIOPIC_SYLLABLE_TWA = 0x1277;
const char32_t ETHIOPIC_SYLLABLE_CA = 0x1278;
const char32_t ETHIOPIC_SYLLABLE_CU = 0x1279;
const char32_t ETHIOPIC_SYLLABLE_CI = 0x127A;
const char32_t ETHIOPIC_SYLLABLE_CAA = 0x127B;
const char32_t ETHIOPIC_SYLLABLE_CEE = 0x127C;
const char32_t ETHIOPIC_SYLLABLE_CE = 0x127D;
const char32_t ETHIOPIC_SYLLABLE_CO = 0x127E;
const char32_t ETHIOPIC_SYLLABLE_CWA = 0x127F;
const char32_t ETHIOPIC_SYLLABLE_XA = 0x1280;
const char32_t ETHIOPIC_SYLLABLE_XU = 0x1281;
const char32_t ETHIOPIC_SYLLABLE_XI = 0x1282;
const char32_t ETHIOPIC_SYLLABLE_XAA = 0x1283;
const char32_t ETHIOPIC_SYLLABLE_XEE = 0x1284;
const char32_t ETHIOPIC_SYLLABLE_XE = 0x1285;
const char32_t ETHIOPIC_SYLLABLE_XO = 0x1286;
const char32_t ETHIOPIC_SYLLABLE_XOA = 0x1287;
const char32_t ETHIOPIC_SYLLABLE_XWA = 0x1288;
const char32_t ETHIOPIC_SYLLABLE_XWI = 0x128A;
const char32_t ETHIOPIC_SYLLABLE_XWAA = 0x128B;
const char32_t ETHIOPIC_SYLLABLE_XWEE = 0x128C;
const char32_t ETHIOPIC_SYLLABLE_XWE = 0x128D;
const char32_t ETHIOPIC_SYLLABLE_NA = 0x1290;
const char32_t ETHIOPIC_SYLLABLE_NU = 0x1291;
const char32_t ETHIOPIC_SYLLABLE_NI = 0x1292;
const char32_t ETHIOPIC_SYLLABLE_NAA = 0x1293;
const char32_t ETHIOPIC_SYLLABLE_NEE = 0x1294;
const char32_t ETHIOPIC_SYLLABLE_NE = 0x1295;
const char32_t ETHIOPIC_SYLLABLE_NO = 0x1296;
const char32_t ETHIOPIC_SYLLABLE_NWA = 0x1297;
const char32_t ETHIOPIC_SYLLABLE_NYA = 0x1298;
const char32_t ETHIOPIC_SYLLABLE_NYU = 0x1299;
const char32_t ETHIOPIC_SYLLABLE_NYI = 0x129A;
const char32_t ETHIOPIC_SYLLABLE_NYAA = 0x129B;
const char32_t ETHIOPIC_SYLLABLE_NYEE = 0x129C;
const char32_t ETHIOPIC_SYLLABLE_NYE = 0x129D;
const char32_t ETHIOPIC_SYLLABLE_NYO = 0x129E;
const char32_t ETHIOPIC_SYLLABLE_NYWA = 0x129F;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_A = 0x12A0;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_U = 0x12A1;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_I = 0x12A2;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_AA = 0x12A3;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_EE = 0x12A4;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_E = 0x12A5;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_O = 0x12A6;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_WA = 0x12A7;
const char32_t ETHIOPIC_SYLLABLE_KA = 0x12A8;
const char32_t ETHIOPIC_SYLLABLE_KU = 0x12A9;
const char32_t ETHIOPIC_SYLLABLE_KI = 0x12AA;
const char32_t ETHIOPIC_SYLLABLE_KAA = 0x12AB;
const char32_t ETHIOPIC_SYLLABLE_KEE = 0x12AC;
const char32_t ETHIOPIC_SYLLABLE_KE = 0x12AD;
const char32_t ETHIOPIC_SYLLABLE_KO = 0x12AE;
const char32_t ETHIOPIC_SYLLABLE_KOA = 0x12AF;
const char32_t ETHIOPIC_SYLLABLE_KWA = 0x12B0;
const char32_t ETHIOPIC_SYLLABLE_KWI = 0x12B2;
const char32_t ETHIOPIC_SYLLABLE_KWAA = 0x12B3;
const char32_t ETHIOPIC_SYLLABLE_KWEE = 0x12B4;
const char32_t ETHIOPIC_SYLLABLE_KWE = 0x12B5;
const char32_t ETHIOPIC_SYLLABLE_KXA = 0x12B8;
const char32_t ETHIOPIC_SYLLABLE_KXU = 0x12B9;
const char32_t ETHIOPIC_SYLLABLE_KXI = 0x12BA;
const char32_t ETHIOPIC_SYLLABLE_KXAA = 0x12BB;
const char32_t ETHIOPIC_SYLLABLE_KXEE = 0x12BC;
const char32_t ETHIOPIC_SYLLABLE_KXE = 0x12BD;
const char32_t ETHIOPIC_SYLLABLE_KXO = 0x12BE;
const char32_t ETHIOPIC_SYLLABLE_KXWA = 0x12C0;
const char32_t ETHIOPIC_SYLLABLE_KXWI = 0x12C2;
const char32_t ETHIOPIC_SYLLABLE_KXWAA = 0x12C3;
const char32_t ETHIOPIC_SYLLABLE_KXWEE = 0x12C4;
const char32_t ETHIOPIC_SYLLABLE_KXWE = 0x12C5;
const char32_t ETHIOPIC_SYLLABLE_WA = 0x12C8;
const char32_t ETHIOPIC_SYLLABLE_WU = 0x12C9;
const char32_t ETHIOPIC_SYLLABLE_WI = 0x12CA;
const char32_t ETHIOPIC_SYLLABLE_WAA = 0x12CB;
const char32_t ETHIOPIC_SYLLABLE_WEE = 0x12CC;
const char32_t ETHIOPIC_SYLLABLE_WE = 0x12CD;
const char32_t ETHIOPIC_SYLLABLE_WO = 0x12CE;
const char32_t ETHIOPIC_SYLLABLE_WOA = 0x12CF;
const char32_t ETHIOPIC_SYLLABLE_PHARYNGEAL_A = 0x12D0;
const char32_t ETHIOPIC_SYLLABLE_PHARYNGEAL_U = 0x12D1;
const char32_t ETHIOPIC_SYLLABLE_PHARYNGEAL_I = 0x12D2;
const char32_t ETHIOPIC_SYLLABLE_PHARYNGEAL_AA = 0x12D3;
const char32_t ETHIOPIC_SYLLABLE_PHARYNGEAL_EE = 0x12D4;
const char32_t ETHIOPIC_SYLLABLE_PHARYNGEAL_E = 0x12D5;
const char32_t ETHIOPIC_SYLLABLE_PHARYNGEAL_O = 0x12D6;
const char32_t ETHIOPIC_SYLLABLE_ZA = 0x12D8;
const char32_t ETHIOPIC_SYLLABLE_ZU = 0x12D9;
const char32_t ETHIOPIC_SYLLABLE_ZI = 0x12DA;
const char32_t ETHIOPIC_SYLLABLE_ZAA = 0x12DB;
const char32_t ETHIOPIC_SYLLABLE_ZEE = 0x12DC;
const char32_t ETHIOPIC_SYLLABLE_ZE = 0x12DD;
const char32_t ETHIOPIC_SYLLABLE_ZO = 0x12DE;
const char32_t ETHIOPIC_SYLLABLE_ZWA = 0x12DF;
const char32_t ETHIOPIC_SYLLABLE_ZHA = 0x12E0;
const char32_t ETHIOPIC_SYLLABLE_ZHU = 0x12E1;
const char32_t ETHIOPIC_SYLLABLE_ZHI = 0x12E2;
const char32_t ETHIOPIC_SYLLABLE_ZHAA = 0x12E3;
const char32_t ETHIOPIC_SYLLABLE_ZHEE = 0x12E4;
const char32_t ETHIOPIC_SYLLABLE_ZHE = 0x12E5;
const char32_t ETHIOPIC_SYLLABLE_ZHO = 0x12E6;
const char32_t ETHIOPIC_SYLLABLE_ZHWA = 0x12E7;
const char32_t ETHIOPIC_SYLLABLE_YA = 0x12E8;
const char32_t ETHIOPIC_SYLLABLE_YU = 0x12E9;
const char32_t ETHIOPIC_SYLLABLE_YI = 0x12EA;
const char32_t ETHIOPIC_SYLLABLE_YAA = 0x12EB;
const char32_t ETHIOPIC_SYLLABLE_YEE = 0x12EC;
const char32_t ETHIOPIC_SYLLABLE_YE = 0x12ED;
const char32_t ETHIOPIC_SYLLABLE_YO = 0x12EE;
const char32_t ETHIOPIC_SYLLABLE_YOA = 0x12EF;
const char32_t ETHIOPIC_SYLLABLE_DA = 0x12F0;
const char32_t ETHIOPIC_SYLLABLE_DU = 0x12F1;
const char32_t ETHIOPIC_SYLLABLE_DI = 0x12F2;
const char32_t ETHIOPIC_SYLLABLE_DAA = 0x12F3;
const char32_t ETHIOPIC_SYLLABLE_DEE = 0x12F4;
const char32_t ETHIOPIC_SYLLABLE_DE = 0x12F5;
const char32_t ETHIOPIC_SYLLABLE_DO = 0x12F6;
const char32_t ETHIOPIC_SYLLABLE_DWA = 0x12F7;
const char32_t ETHIOPIC_SYLLABLE_DDA = 0x12F8;
const char32_t ETHIOPIC_SYLLABLE_DDU = 0x12F9;
const char32_t ETHIOPIC_SYLLABLE_DDI = 0x12FA;
const char32_t ETHIOPIC_SYLLABLE_DDAA = 0x12FB;
const char32_t ETHIOPIC_SYLLABLE_DDEE = 0x12FC;
const char32_t ETHIOPIC_SYLLABLE_DDE = 0x12FD;
const char32_t ETHIOPIC_SYLLABLE_DDO = 0x12FE;
const char32_t ETHIOPIC_SYLLABLE_DDWA = 0x12FF;
const char32_t ETHIOPIC_SYLLABLE_JA = 0x1300;
const char32_t ETHIOPIC_SYLLABLE_JU = 0x1301;
const char32_t ETHIOPIC_SYLLABLE_JI = 0x1302;
const char32_t ETHIOPIC_SYLLABLE_JAA = 0x1303;
const char32_t ETHIOPIC_SYLLABLE_JEE = 0x1304;
const char32_t ETHIOPIC_SYLLABLE_JE = 0x1305;
const char32_t ETHIOPIC_SYLLABLE_JO = 0x1306;
const char32_t ETHIOPIC_SYLLABLE_JWA = 0x1307;
const char32_t ETHIOPIC_SYLLABLE_GA = 0x1308;
const char32_t ETHIOPIC_SYLLABLE_GU = 0x1309;
const char32_t ETHIOPIC_SYLLABLE_GI = 0x130A;
const char32_t ETHIOPIC_SYLLABLE_GAA = 0x130B;
const char32_t ETHIOPIC_SYLLABLE_GEE = 0x130C;
const char32_t ETHIOPIC_SYLLABLE_GE = 0x130D;
const char32_t ETHIOPIC_SYLLABLE_GO = 0x130E;
const char32_t ETHIOPIC_SYLLABLE_GOA = 0x130F;
const char32_t ETHIOPIC_SYLLABLE_GWA = 0x1310;
const char32_t ETHIOPIC_SYLLABLE_GWI = 0x1312;
const char32_t ETHIOPIC_SYLLABLE_GWAA = 0x1313;
const char32_t ETHIOPIC_SYLLABLE_GWEE = 0x1314;
const char32_t ETHIOPIC_SYLLABLE_GWE = 0x1315;
const char32_t ETHIOPIC_SYLLABLE_GGA = 0x1318;
const char32_t ETHIOPIC_SYLLABLE_GGU = 0x1319;
const char32_t ETHIOPIC_SYLLABLE_GGI = 0x131A;
const char32_t ETHIOPIC_SYLLABLE_GGAA = 0x131B;
const char32_t ETHIOPIC_SYLLABLE_GGEE = 0x131C;
const char32_t ETHIOPIC_SYLLABLE_GGE = 0x131D;
const char32_t ETHIOPIC_SYLLABLE_GGO = 0x131E;
const char32_t ETHIOPIC_SYLLABLE_GGWAA = 0x131F;
const char32_t ETHIOPIC_SYLLABLE_THA = 0x1320;
const char32_t ETHIOPIC_SYLLABLE_THU = 0x1321;
const char32_t ETHIOPIC_SYLLABLE_THI = 0x1322;
const char32_t ETHIOPIC_SYLLABLE_THAA = 0x1323;
const char32_t ETHIOPIC_SYLLABLE_THEE = 0x1324;
const char32_t ETHIOPIC_SYLLABLE_THE = 0x1325;
const char32_t ETHIOPIC_SYLLABLE_THO = 0x1326;
const char32_t ETHIOPIC_SYLLABLE_THWA = 0x1327;
const char32_t ETHIOPIC_SYLLABLE_CHA = 0x1328;
const char32_t ETHIOPIC_SYLLABLE_CHU = 0x1329;
const char32_t ETHIOPIC_SYLLABLE_CHI = 0x132A;
const char32_t ETHIOPIC_SYLLABLE_CHAA = 0x132B;
const char32_t ETHIOPIC_SYLLABLE_CHEE = 0x132C;
const char32_t ETHIOPIC_SYLLABLE_CHE = 0x132D;
const char32_t ETHIOPIC_SYLLABLE_CHO = 0x132E;
const char32_t ETHIOPIC_SYLLABLE_CHWA = 0x132F;
const char32_t ETHIOPIC_SYLLABLE_PHA = 0x1330;
const char32_t ETHIOPIC_SYLLABLE_PHU = 0x1331;
const char32_t ETHIOPIC_SYLLABLE_PHI = 0x1332;
const char32_t ETHIOPIC_SYLLABLE_PHAA = 0x1333;
const char32_t ETHIOPIC_SYLLABLE_PHEE = 0x1334;
const char32_t ETHIOPIC_SYLLABLE_PHE = 0x1335;
const char32_t ETHIOPIC_SYLLABLE_PHO = 0x1336;
const char32_t ETHIOPIC_SYLLABLE_PHWA = 0x1337;
const char32_t ETHIOPIC_SYLLABLE_TSA = 0x1338;
const char32_t ETHIOPIC_SYLLABLE_TSU = 0x1339;
const char32_t ETHIOPIC_SYLLABLE_TSI = 0x133A;
const char32_t ETHIOPIC_SYLLABLE_TSAA = 0x133B;
const char32_t ETHIOPIC_SYLLABLE_TSEE = 0x133C;
const char32_t ETHIOPIC_SYLLABLE_TSE = 0x133D;
const char32_t ETHIOPIC_SYLLABLE_TSO = 0x133E;
const char32_t ETHIOPIC_SYLLABLE_TSWA = 0x133F;
const char32_t ETHIOPIC_SYLLABLE_TZA = 0x1340;
const char32_t ETHIOPIC_SYLLABLE_TZU = 0x1341;
const char32_t ETHIOPIC_SYLLABLE_TZI = 0x1342;
const char32_t ETHIOPIC_SYLLABLE_TZAA = 0x1343;
const char32_t ETHIOPIC_SYLLABLE_TZEE = 0x1344;
const char32_t ETHIOPIC_SYLLABLE_TZE = 0x1345;
const char32_t ETHIOPIC_SYLLABLE_TZO = 0x1346;
const char32_t ETHIOPIC_SYLLABLE_TZOA = 0x1347;
const char32_t ETHIOPIC_SYLLABLE_FA = 0x1348;
const char32_t ETHIOPIC_SYLLABLE_FU = 0x1349;
const char32_t ETHIOPIC_SYLLABLE_FI = 0x134A;
const char32_t ETHIOPIC_SYLLABLE_FAA = 0x134B;
const char32_t ETHIOPIC_SYLLABLE_FEE = 0x134C;
const char32_t ETHIOPIC_SYLLABLE_FE = 0x134D;
const char32_t ETHIOPIC_SYLLABLE_FO = 0x134E;
const char32_t ETHIOPIC_SYLLABLE_FWA = 0x134F;
const char32_t ETHIOPIC_SYLLABLE_PA = 0x1350;
const char32_t ETHIOPIC_SYLLABLE_PU = 0x1351;
const char32_t ETHIOPIC_SYLLABLE_PI = 0x1352;
const char32_t ETHIOPIC_SYLLABLE_PAA = 0x1353;
const char32_t ETHIOPIC_SYLLABLE_PEE = 0x1354;
const char32_t ETHIOPIC_SYLLABLE_PE = 0x1355;
const char32_t ETHIOPIC_SYLLABLE_PO = 0x1356;
const char32_t ETHIOPIC_SYLLABLE_PWA = 0x1357;
const char32_t ETHIOPIC_SYLLABLE_RYA = 0x1358;
const char32_t ETHIOPIC_SYLLABLE_MYA = 0x1359;
const char32_t ETHIOPIC_SYLLABLE_FYA = 0x135A;
const char32_t ETHIOPIC_COMBINING_GEMINATION_AND_VOWEL_LENGTH_MARK = 0x135D;
const char32_t ETHIOPIC_COMBINING_VOWEL_LENGTH_MARK = 0x135E;
const char32_t ETHIOPIC_COMBINING_GEMINATION_MARK = 0x135F;
const char32_t ETHIOPIC_SECTION_MARK = 0x1360;
const char32_t ETHIOPIC_WORDSPACE = 0x1361;
const char32_t ETHIOPIC_FULL_STOP = 0x1362;
const char32_t ETHIOPIC_COMMA = 0x1363;
const char32_t ETHIOPIC_SEMICOLON = 0x1364;
const char32_t ETHIOPIC_COLON = 0x1365;
const char32_t ETHIOPIC_PREFACE_COLON = 0x1366;
const char32_t ETHIOPIC_QUESTION_MARK = 0x1367;
const char32_t ETHIOPIC_PARAGRAPH_SEPARATOR = 0x1368;
const char32_t ETHIOPIC_DIGIT_ONE = 0x1369;
const char32_t ETHIOPIC_DIGIT_TWO = 0x136A;
const char32_t ETHIOPIC_DIGIT_THREE = 0x136B;
const char32_t ETHIOPIC_DIGIT_FOUR = 0x136C;
const char32_t ETHIOPIC_DIGIT_FIVE = 0x136D;
const char32_t ETHIOPIC_DIGIT_SIX = 0x136E;
const char32_t ETHIOPIC_DIGIT_SEVEN = 0x136F;
const char32_t ETHIOPIC_DIGIT_EIGHT = 0x1370;
const char32_t ETHIOPIC_DIGIT_NINE = 0x1371;
const char32_t ETHIOPIC_NUMBER_TEN = 0x1372;
const char32_t ETHIOPIC_NUMBER_TWENTY = 0x1373;
const char32_t ETHIOPIC_NUMBER_THIRTY = 0x1374;
const char32_t ETHIOPIC_NUMBER_FORTY = 0x1375;
const char32_t ETHIOPIC_NUMBER_FIFTY = 0x1376;
const char32_t ETHIOPIC_NUMBER_SIXTY = 0x1377;
const char32_t ETHIOPIC_NUMBER_SEVENTY = 0x1378;
const char32_t ETHIOPIC_NUMBER_EIGHTY = 0x1379;
const char32_t ETHIOPIC_NUMBER_NINETY = 0x137A;
const char32_t ETHIOPIC_NUMBER_HUNDRED = 0x137B;
const char32_t ETHIOPIC_NUMBER_TEN_THOUSAND = 0x137C;
const char32_t ETHIOPIC_SYLLABLE_SEBATBEIT_MWA = 0x1380;
const char32_t ETHIOPIC_SYLLABLE_MWI = 0x1381;
const char32_t ETHIOPIC_SYLLABLE_MWEE = 0x1382;
const char32_t ETHIOPIC_SYLLABLE_MWE = 0x1383;
const char32_t ETHIOPIC_SYLLABLE_SEBATBEIT_BWA = 0x1384;
const char32_t ETHIOPIC_SYLLABLE_BWI = 0x1385;
const char32_t ETHIOPIC_SYLLABLE_BWEE = 0x1386;
const char32_t ETHIOPIC_SYLLABLE_BWE = 0x1387;
const char32_t ETHIOPIC_SYLLABLE_SEBATBEIT_FWA = 0x1388;
const char32_t ETHIOPIC_SYLLABLE_FWI = 0x1389;
const char32_t ETHIOPIC_SYLLABLE_FWEE = 0x138A;
const char32_t ETHIOPIC_SYLLABLE_FWE = 0x138B;
const char32_t ETHIOPIC_SYLLABLE_SEBATBEIT_PWA = 0x138C;
const char32_t ETHIOPIC_SYLLABLE_PWI = 0x138D;
const char32_t ETHIOPIC_SYLLABLE_PWEE = 0x138E;
const char32_t ETHIOPIC_SYLLABLE_PWE = 0x138F;
const char32_t ETHIOPIC_TONAL_MARK_YIZET = 0x1390;
const char32_t ETHIOPIC_TONAL_MARK_DERET = 0x1391;
const char32_t ETHIOPIC_TONAL_MARK_RIKRIK = 0x1392;
const char32_t ETHIOPIC_TONAL_MARK_SHORT_RIKRIK = 0x1393;
const char32_t ETHIOPIC_TONAL_MARK_DIFAT = 0x1394;
const char32_t ETHIOPIC_TONAL_MARK_KENAT = 0x1395;
const char32_t ETHIOPIC_TONAL_MARK_CHIRET = 0x1396;
const char32_t ETHIOPIC_TONAL_MARK_HIDET = 0x1397;
const char32_t ETHIOPIC_TONAL_MARK_DERET_HIDET = 0x1398;
const char32_t ETHIOPIC_TONAL_MARK_KURT = 0x1399;
const char32_t CHEROKEE_LETTER_A = 0x13A0;
const char32_t CHEROKEE_LETTER_E = 0x13A1;
const char32_t CHEROKEE_LETTER_I = 0x13A2;
const char32_t CHEROKEE_LETTER_O = 0x13A3;
const char32_t CHEROKEE_LETTER_U = 0x13A4;
const char32_t CHEROKEE_LETTER_V = 0x13A5;
const char32_t CHEROKEE_LETTER_GA = 0x13A6;
const char32_t CHEROKEE_LETTER_KA = 0x13A7;
const char32_t CHEROKEE_LETTER_GE = 0x13A8;
const char32_t CHEROKEE_LETTER_GI = 0x13A9;
const char32_t CHEROKEE_LETTER_GO = 0x13AA;
const char32_t CHEROKEE_LETTER_GU = 0x13AB;
const char32_t CHEROKEE_LETTER_GV = 0x13AC;
const char32_t CHEROKEE_LETTER_HA = 0x13AD;
const char32_t CHEROKEE_LETTER_HE = 0x13AE;
const char32_t CHEROKEE_LETTER_HI = 0x13AF;
const char32_t CHEROKEE_LETTER_HO = 0x13B0;
const char32_t CHEROKEE_LETTER_HU = 0x13B1;
const char32_t CHEROKEE_LETTER_HV = 0x13B2;
const char32_t CHEROKEE_LETTER_LA = 0x13B3;
const char32_t CHEROKEE_LETTER_LE = 0x13B4;
const char32_t CHEROKEE_LETTER_LI = 0x13B5;
const char32_t CHEROKEE_LETTER_LO = 0x13B6;
const char32_t CHEROKEE_LETTER_LU = 0x13B7;
const char32_t CHEROKEE_LETTER_LV = 0x13B8;
const char32_t CHEROKEE_LETTER_MA = 0x13B9;
const char32_t CHEROKEE_LETTER_ME = 0x13BA;
const char32_t CHEROKEE_LETTER_MI = 0x13BB;
const char32_t CHEROKEE_LETTER_MO = 0x13BC;
const char32_t CHEROKEE_LETTER_MU = 0x13BD;
const char32_t CHEROKEE_LETTER_NA = 0x13BE;
const char32_t CHEROKEE_LETTER_HNA = 0x13BF;
const char32_t CHEROKEE_LETTER_NAH = 0x13C0;
const char32_t CHEROKEE_LETTER_NE = 0x13C1;
const char32_t CHEROKEE_LETTER_NI = 0x13C2;
const char32_t CHEROKEE_LETTER_NO = 0x13C3;
const char32_t CHEROKEE_LETTER_NU = 0x13C4;
const char32_t CHEROKEE_LETTER_NV = 0x13C5;
const char32_t CHEROKEE_LETTER_QUA = 0x13C6;
const char32_t CHEROKEE_LETTER_QUE = 0x13C7;
const char32_t CHEROKEE_LETTER_QUI = 0x13C8;
const char32_t CHEROKEE_LETTER_QUO = 0x13C9;
const char32_t CHEROKEE_LETTER_QUU = 0x13CA;
const char32_t CHEROKEE_LETTER_QUV = 0x13CB;
const char32_t CHEROKEE_LETTER_SA = 0x13CC;
const char32_t CHEROKEE_LETTER_S = 0x13CD;
const char32_t CHEROKEE_LETTER_SE = 0x13CE;
const char32_t CHEROKEE_LETTER_SI = 0x13CF;
const char32_t CHEROKEE_LETTER_SO = 0x13D0;
const char32_t CHEROKEE_LETTER_SU = 0x13D1;
const char32_t CHEROKEE_LETTER_SV = 0x13D2;
const char32_t CHEROKEE_LETTER_DA = 0x13D3;
const char32_t CHEROKEE_LETTER_TA = 0x13D4;
const char32_t CHEROKEE_LETTER_DE = 0x13D5;
const char32_t CHEROKEE_LETTER_TE = 0x13D6;
const char32_t CHEROKEE_LETTER_DI = 0x13D7;
const char32_t CHEROKEE_LETTER_TI = 0x13D8;
const char32_t CHEROKEE_LETTER_DO = 0x13D9;
const char32_t CHEROKEE_LETTER_DU = 0x13DA;
const char32_t CHEROKEE_LETTER_DV = 0x13DB;
const char32_t CHEROKEE_LETTER_DLA = 0x13DC;
const char32_t CHEROKEE_LETTER_TLA = 0x13DD;
const char32_t CHEROKEE_LETTER_TLE = 0x13DE;
const char32_t CHEROKEE_LETTER_TLI = 0x13DF;
const char32_t CHEROKEE_LETTER_TLO = 0x13E0;
const char32_t CHEROKEE_LETTER_TLU = 0x13E1;
const char32_t CHEROKEE_LETTER_TLV = 0x13E2;
const char32_t CHEROKEE_LETTER_TSA = 0x13E3;
const char32_t CHEROKEE_LETTER_TSE = 0x13E4;
const char32_t CHEROKEE_LETTER_TSI = 0x13E5;
const char32_t CHEROKEE_LETTER_TSO = 0x13E6;
const char32_t CHEROKEE_LETTER_TSU = 0x13E7;
const char32_t CHEROKEE_LETTER_TSV = 0x13E8;
const char32_t CHEROKEE_LETTER_WA = 0x13E9;
const char32_t CHEROKEE_LETTER_WE = 0x13EA;
const char32_t CHEROKEE_LETTER_WI = 0x13EB;
const char32_t CHEROKEE_LETTER_WO = 0x13EC;
const char32_t CHEROKEE_LETTER_WU = 0x13ED;
const char32_t CHEROKEE_LETTER_WV = 0x13EE;
const char32_t CHEROKEE_LETTER_YA = 0x13EF;
const char32_t CHEROKEE_LETTER_YE = 0x13F0;
const char32_t CHEROKEE_LETTER_YI = 0x13F1;
const char32_t CHEROKEE_LETTER_YO = 0x13F2;
const char32_t CHEROKEE_LETTER_YU = 0x13F3;
const char32_t CHEROKEE_LETTER_YV = 0x13F4;
const char32_t CHEROKEE_LETTER_MV = 0x13F5;
const char32_t CHEROKEE_SMALL_LETTER_YE = 0x13F8;
const char32_t CHEROKEE_SMALL_LETTER_YI = 0x13F9;
const char32_t CHEROKEE_SMALL_LETTER_YO = 0x13FA;
const char32_t CHEROKEE_SMALL_LETTER_YU = 0x13FB;
const char32_t CHEROKEE_SMALL_LETTER_YV = 0x13FC;
const char32_t CHEROKEE_SMALL_LETTER_MV = 0x13FD;
const char32_t CANADIAN_SYLLABICS_HYPHEN = 0x1400;
const char32_t CANADIAN_SYLLABICS_E = 0x1401;
const char32_t CANADIAN_SYLLABICS_AAI = 0x1402;
const char32_t CANADIAN_SYLLABICS_I = 0x1403;
const char32_t CANADIAN_SYLLABICS_II = 0x1404;
const char32_t CANADIAN_SYLLABICS_O = 0x1405;
const char32_t CANADIAN_SYLLABICS_OO = 0x1406;
const char32_t CANADIAN_SYLLABICS_Y_CREE_OO = 0x1407;
const char32_t CANADIAN_SYLLABICS_CARRIER_EE = 0x1408;
const char32_t CANADIAN_SYLLABICS_CARRIER_I = 0x1409;
const char32_t CANADIAN_SYLLABICS_A = 0x140A;
const char32_t CANADIAN_SYLLABICS_AA = 0x140B;
const char32_t CANADIAN_SYLLABICS_WE = 0x140C;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_WE = 0x140D;
const char32_t CANADIAN_SYLLABICS_WI = 0x140E;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_WI = 0x140F;
const char32_t CANADIAN_SYLLABICS_WII = 0x1410;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_WII = 0x1411;
const char32_t CANADIAN_SYLLABICS_WO = 0x1412;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_WO = 0x1413;
const char32_t CANADIAN_SYLLABICS_WOO = 0x1414;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_WOO = 0x1415;
const char32_t CANADIAN_SYLLABICS_NASKAPI_WOO = 0x1416;
const char32_t CANADIAN_SYLLABICS_WA = 0x1417;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_WA = 0x1418;
const char32_t CANADIAN_SYLLABICS_WAA = 0x1419;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_WAA = 0x141A;
const char32_t CANADIAN_SYLLABICS_NASKAPI_WAA = 0x141B;
const char32_t CANADIAN_SYLLABICS_AI = 0x141C;
const char32_t CANADIAN_SYLLABICS_Y_CREE_W = 0x141D;
const char32_t CANADIAN_SYLLABICS_GLOTTAL_STOP = 0x141E;
const char32_t CANADIAN_SYLLABICS_FINAL_ACUTE = 0x141F;
const char32_t CANADIAN_SYLLABICS_FINAL_GRAVE = 0x1420;
const char32_t CANADIAN_SYLLABICS_FINAL_BOTTOM_HALF_RING = 0x1421;
const char32_t CANADIAN_SYLLABICS_FINAL_TOP_HALF_RING = 0x1422;
const char32_t CANADIAN_SYLLABICS_FINAL_RIGHT_HALF_RING = 0x1423;
const char32_t CANADIAN_SYLLABICS_FINAL_RING = 0x1424;
const char32_t CANADIAN_SYLLABICS_FINAL_DOUBLE_ACUTE = 0x1425;
const char32_t CANADIAN_SYLLABICS_FINAL_DOUBLE_SHORT_VERTICAL_STROKES = 0x1426;
const char32_t CANADIAN_SYLLABICS_FINAL_MIDDLE_DOT = 0x1427;
const char32_t CANADIAN_SYLLABICS_FINAL_SHORT_HORIZONTAL_STROKE = 0x1428;
const char32_t CANADIAN_SYLLABICS_FINAL_PLUS = 0x1429;
const char32_t CANADIAN_SYLLABICS_FINAL_DOWN_TACK = 0x142A;
const char32_t CANADIAN_SYLLABICS_EN = 0x142B;
const char32_t CANADIAN_SYLLABICS_IN = 0x142C;
const char32_t CANADIAN_SYLLABICS_ON = 0x142D;
const char32_t CANADIAN_SYLLABICS_AN = 0x142E;
const char32_t CANADIAN_SYLLABICS_PE = 0x142F;
const char32_t CANADIAN_SYLLABICS_PAAI = 0x1430;
const char32_t CANADIAN_SYLLABICS_PI = 0x1431;
const char32_t CANADIAN_SYLLABICS_PII = 0x1432;
const char32_t CANADIAN_SYLLABICS_PO = 0x1433;
const char32_t CANADIAN_SYLLABICS_POO = 0x1434;
const char32_t CANADIAN_SYLLABICS_Y_CREE_POO = 0x1435;
const char32_t CANADIAN_SYLLABICS_CARRIER_HEE = 0x1436;
const char32_t CANADIAN_SYLLABICS_CARRIER_HI = 0x1437;
const char32_t CANADIAN_SYLLABICS_PA = 0x1438;
const char32_t CANADIAN_SYLLABICS_PAA = 0x1439;
const char32_t CANADIAN_SYLLABICS_PWE = 0x143A;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_PWE = 0x143B;
const char32_t CANADIAN_SYLLABICS_PWI = 0x143C;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_PWI = 0x143D;
const char32_t CANADIAN_SYLLABICS_PWII = 0x143E;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_PWII = 0x143F;
const char32_t CANADIAN_SYLLABICS_PWO = 0x1440;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_PWO = 0x1441;
const char32_t CANADIAN_SYLLABICS_PWOO = 0x1442;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_PWOO = 0x1443;
const char32_t CANADIAN_SYLLABICS_PWA = 0x1444;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_PWA = 0x1445;
const char32_t CANADIAN_SYLLABICS_PWAA = 0x1446;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_PWAA = 0x1447;
const char32_t CANADIAN_SYLLABICS_Y_CREE_PWAA = 0x1448;
const char32_t CANADIAN_SYLLABICS_P = 0x1449;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_P = 0x144A;
const char32_t CANADIAN_SYLLABICS_CARRIER_H = 0x144B;
const char32_t CANADIAN_SYLLABICS_TE = 0x144C;
const char32_t CANADIAN_SYLLABICS_TAAI = 0x144D;
const char32_t CANADIAN_SYLLABICS_TI = 0x144E;
const char32_t CANADIAN_SYLLABICS_TII = 0x144F;
const char32_t CANADIAN_SYLLABICS_TO = 0x1450;
const char32_t CANADIAN_SYLLABICS_TOO = 0x1451;
const char32_t CANADIAN_SYLLABICS_Y_CREE_TOO = 0x1452;
const char32_t CANADIAN_SYLLABICS_CARRIER_DEE = 0x1453;
const char32_t CANADIAN_SYLLABICS_CARRIER_DI = 0x1454;
const char32_t CANADIAN_SYLLABICS_TA = 0x1455;
const char32_t CANADIAN_SYLLABICS_TAA = 0x1456;
const char32_t CANADIAN_SYLLABICS_TWE = 0x1457;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_TWE = 0x1458;
const char32_t CANADIAN_SYLLABICS_TWI = 0x1459;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_TWI = 0x145A;
const char32_t CANADIAN_SYLLABICS_TWII = 0x145B;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_TWII = 0x145C;
const char32_t CANADIAN_SYLLABICS_TWO = 0x145D;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_TWO = 0x145E;
const char32_t CANADIAN_SYLLABICS_TWOO = 0x145F;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_TWOO = 0x1460;
const char32_t CANADIAN_SYLLABICS_TWA = 0x1461;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_TWA = 0x1462;
const char32_t CANADIAN_SYLLABICS_TWAA = 0x1463;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_TWAA = 0x1464;
const char32_t CANADIAN_SYLLABICS_NASKAPI_TWAA = 0x1465;
const char32_t CANADIAN_SYLLABICS_T = 0x1466;
const char32_t CANADIAN_SYLLABICS_TTE = 0x1467;
const char32_t CANADIAN_SYLLABICS_TTI = 0x1468;
const char32_t CANADIAN_SYLLABICS_TTO = 0x1469;
const char32_t CANADIAN_SYLLABICS_TTA = 0x146A;
const char32_t CANADIAN_SYLLABICS_KE = 0x146B;
const char32_t CANADIAN_SYLLABICS_KAAI = 0x146C;
const char32_t CANADIAN_SYLLABICS_KI = 0x146D;
const char32_t CANADIAN_SYLLABICS_KII = 0x146E;
const char32_t CANADIAN_SYLLABICS_KO = 0x146F;
const char32_t CANADIAN_SYLLABICS_KOO = 0x1470;
const char32_t CANADIAN_SYLLABICS_Y_CREE_KOO = 0x1471;
const char32_t CANADIAN_SYLLABICS_KA = 0x1472;
const char32_t CANADIAN_SYLLABICS_KAA = 0x1473;
const char32_t CANADIAN_SYLLABICS_KWE = 0x1474;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_KWE = 0x1475;
const char32_t CANADIAN_SYLLABICS_KWI = 0x1476;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_KWI = 0x1477;
const char32_t CANADIAN_SYLLABICS_KWII = 0x1478;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_KWII = 0x1479;
const char32_t CANADIAN_SYLLABICS_KWO = 0x147A;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_KWO = 0x147B;
const char32_t CANADIAN_SYLLABICS_KWOO = 0x147C;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_KWOO = 0x147D;
const char32_t CANADIAN_SYLLABICS_KWA = 0x147E;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_KWA = 0x147F;
const char32_t CANADIAN_SYLLABICS_KWAA = 0x1480;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_KWAA = 0x1481;
const char32_t CANADIAN_SYLLABICS_NASKAPI_KWAA = 0x1482;
const char32_t CANADIAN_SYLLABICS_K = 0x1483;
const char32_t CANADIAN_SYLLABICS_KW = 0x1484;
const char32_t CANADIAN_SYLLABICS_SOUTH_SLAVEY_KEH = 0x1485;
const char32_t CANADIAN_SYLLABICS_SOUTH_SLAVEY_KIH = 0x1486;
const char32_t CANADIAN_SYLLABICS_SOUTH_SLAVEY_KOH = 0x1487;
const char32_t CANADIAN_SYLLABICS_SOUTH_SLAVEY_KAH = 0x1488;
const char32_t CANADIAN_SYLLABICS_CE = 0x1489;
const char32_t CANADIAN_SYLLABICS_CAAI = 0x148A;
const char32_t CANADIAN_SYLLABICS_CI = 0x148B;
const char32_t CANADIAN_SYLLABICS_CII = 0x148C;
const char32_t CANADIAN_SYLLABICS_CO = 0x148D;
const char32_t CANADIAN_SYLLABICS_COO = 0x148E;
const char32_t CANADIAN_SYLLABICS_Y_CREE_COO = 0x148F;
const char32_t CANADIAN_SYLLABICS_CA = 0x1490;
const char32_t CANADIAN_SYLLABICS_CAA = 0x1491;
const char32_t CANADIAN_SYLLABICS_CWE = 0x1492;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_CWE = 0x1493;
const char32_t CANADIAN_SYLLABICS_CWI = 0x1494;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_CWI = 0x1495;
const char32_t CANADIAN_SYLLABICS_CWII = 0x1496;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_CWII = 0x1497;
const char32_t CANADIAN_SYLLABICS_CWO = 0x1498;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_CWO = 0x1499;
const char32_t CANADIAN_SYLLABICS_CWOO = 0x149A;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_CWOO = 0x149B;
const char32_t CANADIAN_SYLLABICS_CWA = 0x149C;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_CWA = 0x149D;
const char32_t CANADIAN_SYLLABICS_CWAA = 0x149E;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_CWAA = 0x149F;
const char32_t CANADIAN_SYLLABICS_NASKAPI_CWAA = 0x14A0;
const char32_t CANADIAN_SYLLABICS_C = 0x14A1;
const char32_t CANADIAN_SYLLABICS_SAYISI_TH = 0x14A2;
const char32_t CANADIAN_SYLLABICS_ME = 0x14A3;
const char32_t CANADIAN_SYLLABICS_MAAI = 0x14A4;
const char32_t CANADIAN_SYLLABICS_MI = 0x14A5;
const char32_t CANADIAN_SYLLABICS_MII = 0x14A6;
const char32_t CANADIAN_SYLLABICS_MO = 0x14A7;
const char32_t CANADIAN_SYLLABICS_MOO = 0x14A8;
const char32_t CANADIAN_SYLLABICS_Y_CREE_MOO = 0x14A9;
const char32_t CANADIAN_SYLLABICS_MA = 0x14AA;
const char32_t CANADIAN_SYLLABICS_MAA = 0x14AB;
const char32_t CANADIAN_SYLLABICS_MWE = 0x14AC;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_MWE = 0x14AD;
const char32_t CANADIAN_SYLLABICS_MWI = 0x14AE;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_MWI = 0x14AF;
const char32_t CANADIAN_SYLLABICS_MWII = 0x14B0;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_MWII = 0x14B1;
const char32_t CANADIAN_SYLLABICS_MWO = 0x14B2;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_MWO = 0x14B3;
const char32_t CANADIAN_SYLLABICS_MWOO = 0x14B4;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_MWOO = 0x14B5;
const char32_t CANADIAN_SYLLABICS_MWA = 0x14B6;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_MWA = 0x14B7;
const char32_t CANADIAN_SYLLABICS_MWAA = 0x14B8;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_MWAA = 0x14B9;
const char32_t CANADIAN_SYLLABICS_NASKAPI_MWAA = 0x14BA;
const char32_t CANADIAN_SYLLABICS_M = 0x14BB;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_M = 0x14BC;
const char32_t CANADIAN_SYLLABICS_MH = 0x14BD;
const char32_t CANADIAN_SYLLABICS_ATHAPASCAN_M = 0x14BE;
const char32_t CANADIAN_SYLLABICS_SAYISI_M = 0x14BF;
const char32_t CANADIAN_SYLLABICS_NE = 0x14C0;
const char32_t CANADIAN_SYLLABICS_NAAI = 0x14C1;
const char32_t CANADIAN_SYLLABICS_NI = 0x14C2;
const char32_t CANADIAN_SYLLABICS_NII = 0x14C3;
const char32_t CANADIAN_SYLLABICS_NO = 0x14C4;
const char32_t CANADIAN_SYLLABICS_NOO = 0x14C5;
const char32_t CANADIAN_SYLLABICS_Y_CREE_NOO = 0x14C6;
const char32_t CANADIAN_SYLLABICS_NA = 0x14C7;
const char32_t CANADIAN_SYLLABICS_NAA = 0x14C8;
const char32_t CANADIAN_SYLLABICS_NWE = 0x14C9;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_NWE = 0x14CA;
const char32_t CANADIAN_SYLLABICS_NWA = 0x14CB;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_NWA = 0x14CC;
const char32_t CANADIAN_SYLLABICS_NWAA = 0x14CD;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_NWAA = 0x14CE;
const char32_t CANADIAN_SYLLABICS_NASKAPI_NWAA = 0x14CF;
const char32_t CANADIAN_SYLLABICS_N = 0x14D0;
const char32_t CANADIAN_SYLLABICS_CARRIER_NG = 0x14D1;
const char32_t CANADIAN_SYLLABICS_NH = 0x14D2;
const char32_t CANADIAN_SYLLABICS_LE = 0x14D3;
const char32_t CANADIAN_SYLLABICS_LAAI = 0x14D4;
const char32_t CANADIAN_SYLLABICS_LI = 0x14D5;
const char32_t CANADIAN_SYLLABICS_LII = 0x14D6;
const char32_t CANADIAN_SYLLABICS_LO = 0x14D7;
const char32_t CANADIAN_SYLLABICS_LOO = 0x14D8;
const char32_t CANADIAN_SYLLABICS_Y_CREE_LOO = 0x14D9;
const char32_t CANADIAN_SYLLABICS_LA = 0x14DA;
const char32_t CANADIAN_SYLLABICS_LAA = 0x14DB;
const char32_t CANADIAN_SYLLABICS_LWE = 0x14DC;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LWE = 0x14DD;
const char32_t CANADIAN_SYLLABICS_LWI = 0x14DE;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LWI = 0x14DF;
const char32_t CANADIAN_SYLLABICS_LWII = 0x14E0;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LWII = 0x14E1;
const char32_t CANADIAN_SYLLABICS_LWO = 0x14E2;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LWO = 0x14E3;
const char32_t CANADIAN_SYLLABICS_LWOO = 0x14E4;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LWOO = 0x14E5;
const char32_t CANADIAN_SYLLABICS_LWA = 0x14E6;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LWA = 0x14E7;
const char32_t CANADIAN_SYLLABICS_LWAA = 0x14E8;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LWAA = 0x14E9;
const char32_t CANADIAN_SYLLABICS_L = 0x14EA;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_L = 0x14EB;
const char32_t CANADIAN_SYLLABICS_MEDIAL_L = 0x14EC;
const char32_t CANADIAN_SYLLABICS_SE = 0x14ED;
const char32_t CANADIAN_SYLLABICS_SAAI = 0x14EE;
const char32_t CANADIAN_SYLLABICS_SI = 0x14EF;
const char32_t CANADIAN_SYLLABICS_SII = 0x14F0;
const char32_t CANADIAN_SYLLABICS_SO = 0x14F1;
const char32_t CANADIAN_SYLLABICS_SOO = 0x14F2;
const char32_t CANADIAN_SYLLABICS_Y_CREE_SOO = 0x14F3;
const char32_t CANADIAN_SYLLABICS_SA = 0x14F4;
const char32_t CANADIAN_SYLLABICS_SAA = 0x14F5;
const char32_t CANADIAN_SYLLABICS_SWE = 0x14F6;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SWE = 0x14F7;
const char32_t CANADIAN_SYLLABICS_SWI = 0x14F8;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SWI = 0x14F9;
const char32_t CANADIAN_SYLLABICS_SWII = 0x14FA;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SWII = 0x14FB;
const char32_t CANADIAN_SYLLABICS_SWO = 0x14FC;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SWO = 0x14FD;
const char32_t CANADIAN_SYLLABICS_SWOO = 0x14FE;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SWOO = 0x14FF;
const char32_t CANADIAN_SYLLABICS_SWA = 0x1500;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SWA = 0x1501;
const char32_t CANADIAN_SYLLABICS_SWAA = 0x1502;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SWAA = 0x1503;
const char32_t CANADIAN_SYLLABICS_NASKAPI_SWAA = 0x1504;
const char32_t CANADIAN_SYLLABICS_S = 0x1505;
const char32_t CANADIAN_SYLLABICS_ATHAPASCAN_S = 0x1506;
const char32_t CANADIAN_SYLLABICS_SW = 0x1507;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_S = 0x1508;
const char32_t CANADIAN_SYLLABICS_MOOSE_CREE_SK = 0x1509;
const char32_t CANADIAN_SYLLABICS_NASKAPI_SKW = 0x150A;
const char32_t CANADIAN_SYLLABICS_NASKAPI_S_W = 0x150B;
const char32_t CANADIAN_SYLLABICS_NASKAPI_SPWA = 0x150C;
const char32_t CANADIAN_SYLLABICS_NASKAPI_STWA = 0x150D;
const char32_t CANADIAN_SYLLABICS_NASKAPI_SKWA = 0x150E;
const char32_t CANADIAN_SYLLABICS_NASKAPI_SCWA = 0x150F;
const char32_t CANADIAN_SYLLABICS_SHE = 0x1510;
const char32_t CANADIAN_SYLLABICS_SHI = 0x1511;
const char32_t CANADIAN_SYLLABICS_SHII = 0x1512;
const char32_t CANADIAN_SYLLABICS_SHO = 0x1513;
const char32_t CANADIAN_SYLLABICS_SHOO = 0x1514;
const char32_t CANADIAN_SYLLABICS_SHA = 0x1515;
const char32_t CANADIAN_SYLLABICS_SHAA = 0x1516;
const char32_t CANADIAN_SYLLABICS_SHWE = 0x1517;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SHWE = 0x1518;
const char32_t CANADIAN_SYLLABICS_SHWI = 0x1519;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SHWI = 0x151A;
const char32_t CANADIAN_SYLLABICS_SHWII = 0x151B;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SHWII = 0x151C;
const char32_t CANADIAN_SYLLABICS_SHWO = 0x151D;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SHWO = 0x151E;
const char32_t CANADIAN_SYLLABICS_SHWOO = 0x151F;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SHWOO = 0x1520;
const char32_t CANADIAN_SYLLABICS_SHWA = 0x1521;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SHWA = 0x1522;
const char32_t CANADIAN_SYLLABICS_SHWAA = 0x1523;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_SHWAA = 0x1524;
const char32_t CANADIAN_SYLLABICS_SH = 0x1525;
const char32_t CANADIAN_SYLLABICS_YE = 0x1526;
const char32_t CANADIAN_SYLLABICS_YAAI = 0x1527;
const char32_t CANADIAN_SYLLABICS_YI = 0x1528;
const char32_t CANADIAN_SYLLABICS_YII = 0x1529;
const char32_t CANADIAN_SYLLABICS_YO = 0x152A;
const char32_t CANADIAN_SYLLABICS_YOO = 0x152B;
const char32_t CANADIAN_SYLLABICS_Y_CREE_YOO = 0x152C;
const char32_t CANADIAN_SYLLABICS_YA = 0x152D;
const char32_t CANADIAN_SYLLABICS_YAA = 0x152E;
const char32_t CANADIAN_SYLLABICS_YWE = 0x152F;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_YWE = 0x1530;
const char32_t CANADIAN_SYLLABICS_YWI = 0x1531;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_YWI = 0x1532;
const char32_t CANADIAN_SYLLABICS_YWII = 0x1533;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_YWII = 0x1534;
const char32_t CANADIAN_SYLLABICS_YWO = 0x1535;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_YWO = 0x1536;
const char32_t CANADIAN_SYLLABICS_YWOO = 0x1537;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_YWOO = 0x1538;
const char32_t CANADIAN_SYLLABICS_YWA = 0x1539;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_YWA = 0x153A;
const char32_t CANADIAN_SYLLABICS_YWAA = 0x153B;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_YWAA = 0x153C;
const char32_t CANADIAN_SYLLABICS_NASKAPI_YWAA = 0x153D;
const char32_t CANADIAN_SYLLABICS_Y = 0x153E;
const char32_t CANADIAN_SYLLABICS_BIBLE_CREE_Y = 0x153F;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_Y = 0x1540;
const char32_t CANADIAN_SYLLABICS_SAYISI_YI = 0x1541;
const char32_t CANADIAN_SYLLABICS_RE = 0x1542;
const char32_t CANADIAN_SYLLABICS_R_CREE_RE = 0x1543;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LE = 0x1544;
const char32_t CANADIAN_SYLLABICS_RAAI = 0x1545;
const char32_t CANADIAN_SYLLABICS_RI = 0x1546;
const char32_t CANADIAN_SYLLABICS_RII = 0x1547;
const char32_t CANADIAN_SYLLABICS_RO = 0x1548;
const char32_t CANADIAN_SYLLABICS_ROO = 0x1549;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LO = 0x154A;
const char32_t CANADIAN_SYLLABICS_RA = 0x154B;
const char32_t CANADIAN_SYLLABICS_RAA = 0x154C;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LA = 0x154D;
const char32_t CANADIAN_SYLLABICS_RWAA = 0x154E;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_RWAA = 0x154F;
const char32_t CANADIAN_SYLLABICS_R = 0x1550;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_R = 0x1551;
const char32_t CANADIAN_SYLLABICS_MEDIAL_R = 0x1552;
const char32_t CANADIAN_SYLLABICS_FE = 0x1553;
const char32_t CANADIAN_SYLLABICS_FAAI = 0x1554;
const char32_t CANADIAN_SYLLABICS_FI = 0x1555;
const char32_t CANADIAN_SYLLABICS_FII = 0x1556;
const char32_t CANADIAN_SYLLABICS_FO = 0x1557;
const char32_t CANADIAN_SYLLABICS_FOO = 0x1558;
const char32_t CANADIAN_SYLLABICS_FA = 0x1559;
const char32_t CANADIAN_SYLLABICS_FAA = 0x155A;
const char32_t CANADIAN_SYLLABICS_FWAA = 0x155B;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_FWAA = 0x155C;
const char32_t CANADIAN_SYLLABICS_F = 0x155D;
const char32_t CANADIAN_SYLLABICS_THE = 0x155E;
const char32_t CANADIAN_SYLLABICS_N_CREE_THE = 0x155F;
const char32_t CANADIAN_SYLLABICS_THI = 0x1560;
const char32_t CANADIAN_SYLLABICS_N_CREE_THI = 0x1561;
const char32_t CANADIAN_SYLLABICS_THII = 0x1562;
const char32_t CANADIAN_SYLLABICS_N_CREE_THII = 0x1563;
const char32_t CANADIAN_SYLLABICS_THO = 0x1564;
const char32_t CANADIAN_SYLLABICS_THOO = 0x1565;
const char32_t CANADIAN_SYLLABICS_THA = 0x1566;
const char32_t CANADIAN_SYLLABICS_THAA = 0x1567;
const char32_t CANADIAN_SYLLABICS_THWAA = 0x1568;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_THWAA = 0x1569;
const char32_t CANADIAN_SYLLABICS_TH = 0x156A;
const char32_t CANADIAN_SYLLABICS_TTHE = 0x156B;
const char32_t CANADIAN_SYLLABICS_TTHI = 0x156C;
const char32_t CANADIAN_SYLLABICS_TTHO = 0x156D;
const char32_t CANADIAN_SYLLABICS_TTHA = 0x156E;
const char32_t CANADIAN_SYLLABICS_TTH = 0x156F;
const char32_t CANADIAN_SYLLABICS_TYE = 0x1570;
const char32_t CANADIAN_SYLLABICS_TYI = 0x1571;
const char32_t CANADIAN_SYLLABICS_TYO = 0x1572;
const char32_t CANADIAN_SYLLABICS_TYA = 0x1573;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_HE = 0x1574;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_HI = 0x1575;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_HII = 0x1576;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_HO = 0x1577;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_HOO = 0x1578;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_HA = 0x1579;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_HAA = 0x157A;
const char32_t CANADIAN_SYLLABICS_NUNAVIK_H = 0x157B;
const char32_t CANADIAN_SYLLABICS_NUNAVUT_H = 0x157C;
const char32_t CANADIAN_SYLLABICS_HK = 0x157D;
const char32_t CANADIAN_SYLLABICS_QAAI = 0x157E;
const char32_t CANADIAN_SYLLABICS_QI = 0x157F;
const char32_t CANADIAN_SYLLABICS_QII = 0x1580;
const char32_t CANADIAN_SYLLABICS_QO = 0x1581;
const char32_t CANADIAN_SYLLABICS_QOO = 0x1582;
const char32_t CANADIAN_SYLLABICS_QA = 0x1583;
const char32_t CANADIAN_SYLLABICS_QAA = 0x1584;
const char32_t CANADIAN_SYLLABICS_Q = 0x1585;
const char32_t CANADIAN_SYLLABICS_TLHE = 0x1586;
const char32_t CANADIAN_SYLLABICS_TLHI = 0x1587;
const char32_t CANADIAN_SYLLABICS_TLHO = 0x1588;
const char32_t CANADIAN_SYLLABICS_TLHA = 0x1589;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_RE = 0x158A;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_RI = 0x158B;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_RO = 0x158C;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_RA = 0x158D;
const char32_t CANADIAN_SYLLABICS_NGAAI = 0x158E;
const char32_t CANADIAN_SYLLABICS_NGI = 0x158F;
const char32_t CANADIAN_SYLLABICS_NGII = 0x1590;
const char32_t CANADIAN_SYLLABICS_NGO = 0x1591;
const char32_t CANADIAN_SYLLABICS_NGOO = 0x1592;
const char32_t CANADIAN_SYLLABICS_NGA = 0x1593;
const char32_t CANADIAN_SYLLABICS_NGAA = 0x1594;
const char32_t CANADIAN_SYLLABICS_NG = 0x1595;
const char32_t CANADIAN_SYLLABICS_NNG = 0x1596;
const char32_t CANADIAN_SYLLABICS_SAYISI_SHE = 0x1597;
const char32_t CANADIAN_SYLLABICS_SAYISI_SHI = 0x1598;
const char32_t CANADIAN_SYLLABICS_SAYISI_SHO = 0x1599;
const char32_t CANADIAN_SYLLABICS_SAYISI_SHA = 0x159A;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THE = 0x159B;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THI = 0x159C;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THO = 0x159D;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THA = 0x159E;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_TH = 0x159F;
const char32_t CANADIAN_SYLLABICS_LHI = 0x15A0;
const char32_t CANADIAN_SYLLABICS_LHII = 0x15A1;
const char32_t CANADIAN_SYLLABICS_LHO = 0x15A2;
const char32_t CANADIAN_SYLLABICS_LHOO = 0x15A3;
const char32_t CANADIAN_SYLLABICS_LHA = 0x15A4;
const char32_t CANADIAN_SYLLABICS_LHAA = 0x15A5;
const char32_t CANADIAN_SYLLABICS_LH = 0x15A6;
const char32_t CANADIAN_SYLLABICS_TH_CREE_THE = 0x15A7;
const char32_t CANADIAN_SYLLABICS_TH_CREE_THI = 0x15A8;
const char32_t CANADIAN_SYLLABICS_TH_CREE_THII = 0x15A9;
const char32_t CANADIAN_SYLLABICS_TH_CREE_THO = 0x15AA;
const char32_t CANADIAN_SYLLABICS_TH_CREE_THOO = 0x15AB;
const char32_t CANADIAN_SYLLABICS_TH_CREE_THA = 0x15AC;
const char32_t CANADIAN_SYLLABICS_TH_CREE_THAA = 0x15AD;
const char32_t CANADIAN_SYLLABICS_TH_CREE_TH = 0x15AE;
const char32_t CANADIAN_SYLLABICS_AIVILIK_B = 0x15AF;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_E = 0x15B0;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_I = 0x15B1;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_O = 0x15B2;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_A = 0x15B3;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_WE = 0x15B4;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_WI = 0x15B5;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_WO = 0x15B6;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_WA = 0x15B7;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_NE = 0x15B8;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_NI = 0x15B9;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_NO = 0x15BA;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_NA = 0x15BB;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_KE = 0x15BC;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_KI = 0x15BD;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_KO = 0x15BE;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_KA = 0x15BF;
const char32_t CANADIAN_SYLLABICS_SAYISI_HE = 0x15C0;
const char32_t CANADIAN_SYLLABICS_SAYISI_HI = 0x15C1;
const char32_t CANADIAN_SYLLABICS_SAYISI_HO = 0x15C2;
const char32_t CANADIAN_SYLLABICS_SAYISI_HA = 0x15C3;
const char32_t CANADIAN_SYLLABICS_CARRIER_GHU = 0x15C4;
const char32_t CANADIAN_SYLLABICS_CARRIER_GHO = 0x15C5;
const char32_t CANADIAN_SYLLABICS_CARRIER_GHE = 0x15C6;
const char32_t CANADIAN_SYLLABICS_CARRIER_GHEE = 0x15C7;
const char32_t CANADIAN_SYLLABICS_CARRIER_GHI = 0x15C8;
const char32_t CANADIAN_SYLLABICS_CARRIER_GHA = 0x15C9;
const char32_t CANADIAN_SYLLABICS_CARRIER_RU = 0x15CA;
const char32_t CANADIAN_SYLLABICS_CARRIER_RO = 0x15CB;
const char32_t CANADIAN_SYLLABICS_CARRIER_RE = 0x15CC;
const char32_t CANADIAN_SYLLABICS_CARRIER_REE = 0x15CD;
const char32_t CANADIAN_SYLLABICS_CARRIER_RI = 0x15CE;
const char32_t CANADIAN_SYLLABICS_CARRIER_RA = 0x15CF;
const char32_t CANADIAN_SYLLABICS_CARRIER_WU = 0x15D0;
const char32_t CANADIAN_SYLLABICS_CARRIER_WO = 0x15D1;
const char32_t CANADIAN_SYLLABICS_CARRIER_WE = 0x15D2;
const char32_t CANADIAN_SYLLABICS_CARRIER_WEE = 0x15D3;
const char32_t CANADIAN_SYLLABICS_CARRIER_WI = 0x15D4;
const char32_t CANADIAN_SYLLABICS_CARRIER_WA = 0x15D5;
const char32_t CANADIAN_SYLLABICS_CARRIER_HWU = 0x15D6;
const char32_t CANADIAN_SYLLABICS_CARRIER_HWO = 0x15D7;
const char32_t CANADIAN_SYLLABICS_CARRIER_HWE = 0x15D8;
const char32_t CANADIAN_SYLLABICS_CARRIER_HWEE = 0x15D9;
const char32_t CANADIAN_SYLLABICS_CARRIER_HWI = 0x15DA;
const char32_t CANADIAN_SYLLABICS_CARRIER_HWA = 0x15DB;
const char32_t CANADIAN_SYLLABICS_CARRIER_THU = 0x15DC;
const char32_t CANADIAN_SYLLABICS_CARRIER_THO = 0x15DD;
const char32_t CANADIAN_SYLLABICS_CARRIER_THE = 0x15DE;
const char32_t CANADIAN_SYLLABICS_CARRIER_THEE = 0x15DF;
const char32_t CANADIAN_SYLLABICS_CARRIER_THI = 0x15E0;
const char32_t CANADIAN_SYLLABICS_CARRIER_THA = 0x15E1;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTU = 0x15E2;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTO = 0x15E3;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTE = 0x15E4;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTEE = 0x15E5;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTI = 0x15E6;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTA = 0x15E7;
const char32_t CANADIAN_SYLLABICS_CARRIER_PU = 0x15E8;
const char32_t CANADIAN_SYLLABICS_CARRIER_PO = 0x15E9;
const char32_t CANADIAN_SYLLABICS_CARRIER_PE = 0x15EA;
const char32_t CANADIAN_SYLLABICS_CARRIER_PEE = 0x15EB;
const char32_t CANADIAN_SYLLABICS_CARRIER_PI = 0x15EC;
const char32_t CANADIAN_SYLLABICS_CARRIER_PA = 0x15ED;
const char32_t CANADIAN_SYLLABICS_CARRIER_P = 0x15EE;
const char32_t CANADIAN_SYLLABICS_CARRIER_GU = 0x15EF;
const char32_t CANADIAN_SYLLABICS_CARRIER_GO = 0x15F0;
const char32_t CANADIAN_SYLLABICS_CARRIER_GE = 0x15F1;
const char32_t CANADIAN_SYLLABICS_CARRIER_GEE = 0x15F2;
const char32_t CANADIAN_SYLLABICS_CARRIER_GI = 0x15F3;
const char32_t CANADIAN_SYLLABICS_CARRIER_GA = 0x15F4;
const char32_t CANADIAN_SYLLABICS_CARRIER_KHU = 0x15F5;
const char32_t CANADIAN_SYLLABICS_CARRIER_KHO = 0x15F6;
const char32_t CANADIAN_SYLLABICS_CARRIER_KHE = 0x15F7;
const char32_t CANADIAN_SYLLABICS_CARRIER_KHEE = 0x15F8;
const char32_t CANADIAN_SYLLABICS_CARRIER_KHI = 0x15F9;
const char32_t CANADIAN_SYLLABICS_CARRIER_KHA = 0x15FA;
const char32_t CANADIAN_SYLLABICS_CARRIER_KKU = 0x15FB;
const char32_t CANADIAN_SYLLABICS_CARRIER_KKO = 0x15FC;
const char32_t CANADIAN_SYLLABICS_CARRIER_KKE = 0x15FD;
const char32_t CANADIAN_SYLLABICS_CARRIER_KKEE = 0x15FE;
const char32_t CANADIAN_SYLLABICS_CARRIER_KKI = 0x15FF;
const char32_t CANADIAN_SYLLABICS_CARRIER_KKA = 0x1600;
const char32_t CANADIAN_SYLLABICS_CARRIER_KK = 0x1601;
const char32_t CANADIAN_SYLLABICS_CARRIER_NU = 0x1602;
const char32_t CANADIAN_SYLLABICS_CARRIER_NO = 0x1603;
const char32_t CANADIAN_SYLLABICS_CARRIER_NE = 0x1604;
const char32_t CANADIAN_SYLLABICS_CARRIER_NEE = 0x1605;
const char32_t CANADIAN_SYLLABICS_CARRIER_NI = 0x1606;
const char32_t CANADIAN_SYLLABICS_CARRIER_NA = 0x1607;
const char32_t CANADIAN_SYLLABICS_CARRIER_MU = 0x1608;
const char32_t CANADIAN_SYLLABICS_CARRIER_MO = 0x1609;
const char32_t CANADIAN_SYLLABICS_CARRIER_ME = 0x160A;
const char32_t CANADIAN_SYLLABICS_CARRIER_MEE = 0x160B;
const char32_t CANADIAN_SYLLABICS_CARRIER_MI = 0x160C;
const char32_t CANADIAN_SYLLABICS_CARRIER_MA = 0x160D;
const char32_t CANADIAN_SYLLABICS_CARRIER_YU = 0x160E;
const char32_t CANADIAN_SYLLABICS_CARRIER_YO = 0x160F;
const char32_t CANADIAN_SYLLABICS_CARRIER_YE = 0x1610;
const char32_t CANADIAN_SYLLABICS_CARRIER_YEE = 0x1611;
const char32_t CANADIAN_SYLLABICS_CARRIER_YI = 0x1612;
const char32_t CANADIAN_SYLLABICS_CARRIER_YA = 0x1613;
const char32_t CANADIAN_SYLLABICS_CARRIER_JU = 0x1614;
const char32_t CANADIAN_SYLLABICS_SAYISI_JU = 0x1615;
const char32_t CANADIAN_SYLLABICS_CARRIER_JO = 0x1616;
const char32_t CANADIAN_SYLLABICS_CARRIER_JE = 0x1617;
const char32_t CANADIAN_SYLLABICS_CARRIER_JEE = 0x1618;
const char32_t CANADIAN_SYLLABICS_CARRIER_JI = 0x1619;
const char32_t CANADIAN_SYLLABICS_SAYISI_JI = 0x161A;
const char32_t CANADIAN_SYLLABICS_CARRIER_JA = 0x161B;
const char32_t CANADIAN_SYLLABICS_CARRIER_JJU = 0x161C;
const char32_t CANADIAN_SYLLABICS_CARRIER_JJO = 0x161D;
const char32_t CANADIAN_SYLLABICS_CARRIER_JJE = 0x161E;
const char32_t CANADIAN_SYLLABICS_CARRIER_JJEE = 0x161F;
const char32_t CANADIAN_SYLLABICS_CARRIER_JJI = 0x1620;
const char32_t CANADIAN_SYLLABICS_CARRIER_JJA = 0x1621;
const char32_t CANADIAN_SYLLABICS_CARRIER_LU = 0x1622;
const char32_t CANADIAN_SYLLABICS_CARRIER_LO = 0x1623;
const char32_t CANADIAN_SYLLABICS_CARRIER_LE = 0x1624;
const char32_t CANADIAN_SYLLABICS_CARRIER_LEE = 0x1625;
const char32_t CANADIAN_SYLLABICS_CARRIER_LI = 0x1626;
const char32_t CANADIAN_SYLLABICS_CARRIER_LA = 0x1627;
const char32_t CANADIAN_SYLLABICS_CARRIER_DLU = 0x1628;
const char32_t CANADIAN_SYLLABICS_CARRIER_DLO = 0x1629;
const char32_t CANADIAN_SYLLABICS_CARRIER_DLE = 0x162A;
const char32_t CANADIAN_SYLLABICS_CARRIER_DLEE = 0x162B;
const char32_t CANADIAN_SYLLABICS_CARRIER_DLI = 0x162C;
const char32_t CANADIAN_SYLLABICS_CARRIER_DLA = 0x162D;
const char32_t CANADIAN_SYLLABICS_CARRIER_LHU = 0x162E;
const char32_t CANADIAN_SYLLABICS_CARRIER_LHO = 0x162F;
const char32_t CANADIAN_SYLLABICS_CARRIER_LHE = 0x1630;
const char32_t CANADIAN_SYLLABICS_CARRIER_LHEE = 0x1631;
const char32_t CANADIAN_SYLLABICS_CARRIER_LHI = 0x1632;
const char32_t CANADIAN_SYLLABICS_CARRIER_LHA = 0x1633;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLHU = 0x1634;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLHO = 0x1635;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLHE = 0x1636;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLHEE = 0x1637;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLHI = 0x1638;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLHA = 0x1639;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLU = 0x163A;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLO = 0x163B;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLE = 0x163C;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLEE = 0x163D;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLI = 0x163E;
const char32_t CANADIAN_SYLLABICS_CARRIER_TLA = 0x163F;
const char32_t CANADIAN_SYLLABICS_CARRIER_ZU = 0x1640;
const char32_t CANADIAN_SYLLABICS_CARRIER_ZO = 0x1641;
const char32_t CANADIAN_SYLLABICS_CARRIER_ZE = 0x1642;
const char32_t CANADIAN_SYLLABICS_CARRIER_ZEE = 0x1643;
const char32_t CANADIAN_SYLLABICS_CARRIER_ZI = 0x1644;
const char32_t CANADIAN_SYLLABICS_CARRIER_ZA = 0x1645;
const char32_t CANADIAN_SYLLABICS_CARRIER_Z = 0x1646;
const char32_t CANADIAN_SYLLABICS_CARRIER_INITIAL_Z = 0x1647;
const char32_t CANADIAN_SYLLABICS_CARRIER_DZU = 0x1648;
const char32_t CANADIAN_SYLLABICS_CARRIER_DZO = 0x1649;
const char32_t CANADIAN_SYLLABICS_CARRIER_DZE = 0x164A;
const char32_t CANADIAN_SYLLABICS_CARRIER_DZEE = 0x164B;
const char32_t CANADIAN_SYLLABICS_CARRIER_DZI = 0x164C;
const char32_t CANADIAN_SYLLABICS_CARRIER_DZA = 0x164D;
const char32_t CANADIAN_SYLLABICS_CARRIER_SU = 0x164E;
const char32_t CANADIAN_SYLLABICS_CARRIER_SO = 0x164F;
const char32_t CANADIAN_SYLLABICS_CARRIER_SE = 0x1650;
const char32_t CANADIAN_SYLLABICS_CARRIER_SEE = 0x1651;
const char32_t CANADIAN_SYLLABICS_CARRIER_SI = 0x1652;
const char32_t CANADIAN_SYLLABICS_CARRIER_SA = 0x1653;
const char32_t CANADIAN_SYLLABICS_CARRIER_SHU = 0x1654;
const char32_t CANADIAN_SYLLABICS_CARRIER_SHO = 0x1655;
const char32_t CANADIAN_SYLLABICS_CARRIER_SHE = 0x1656;
const char32_t CANADIAN_SYLLABICS_CARRIER_SHEE = 0x1657;
const char32_t CANADIAN_SYLLABICS_CARRIER_SHI = 0x1658;
const char32_t CANADIAN_SYLLABICS_CARRIER_SHA = 0x1659;
const char32_t CANADIAN_SYLLABICS_CARRIER_SH = 0x165A;
const char32_t CANADIAN_SYLLABICS_CARRIER_TSU = 0x165B;
const char32_t CANADIAN_SYLLABICS_CARRIER_TSO = 0x165C;
const char32_t CANADIAN_SYLLABICS_CARRIER_TSE = 0x165D;
const char32_t CANADIAN_SYLLABICS_CARRIER_TSEE = 0x165E;
const char32_t CANADIAN_SYLLABICS_CARRIER_TSI = 0x165F;
const char32_t CANADIAN_SYLLABICS_CARRIER_TSA = 0x1660;
const char32_t CANADIAN_SYLLABICS_CARRIER_CHU = 0x1661;
const char32_t CANADIAN_SYLLABICS_CARRIER_CHO = 0x1662;
const char32_t CANADIAN_SYLLABICS_CARRIER_CHE = 0x1663;
const char32_t CANADIAN_SYLLABICS_CARRIER_CHEE = 0x1664;
const char32_t CANADIAN_SYLLABICS_CARRIER_CHI = 0x1665;
const char32_t CANADIAN_SYLLABICS_CARRIER_CHA = 0x1666;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTSU = 0x1667;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTSO = 0x1668;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTSE = 0x1669;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTSEE = 0x166A;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTSI = 0x166B;
const char32_t CANADIAN_SYLLABICS_CARRIER_TTSA = 0x166C;
const char32_t CANADIAN_SYLLABICS_CHI_SIGN = 0x166D;
const char32_t CANADIAN_SYLLABICS_FULL_STOP = 0x166E;
const char32_t CANADIAN_SYLLABICS_QAI = 0x166F;
const char32_t CANADIAN_SYLLABICS_NGAI = 0x1670;
const char32_t CANADIAN_SYLLABICS_NNGI = 0x1671;
const char32_t CANADIAN_SYLLABICS_NNGII = 0x1672;
const char32_t CANADIAN_SYLLABICS_NNGO = 0x1673;
const char32_t CANADIAN_SYLLABICS_NNGOO = 0x1674;
const char32_t CANADIAN_SYLLABICS_NNGA = 0x1675;
const char32_t CANADIAN_SYLLABICS_NNGAA = 0x1676;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THWEE = 0x1677;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THWI = 0x1678;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THWII = 0x1679;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THWO = 0x167A;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THWOO = 0x167B;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THWA = 0x167C;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_THWAA = 0x167D;
const char32_t CANADIAN_SYLLABICS_WOODS_CREE_FINAL_TH = 0x167E;
const char32_t CANADIAN_SYLLABICS_BLACKFOOT_W = 0x167F;
const char32_t OGHAM_SPACE_MARK = 0x1680;
const char32_t OGHAM_LETTER_BEITH = 0x1681;
const char32_t OGHAM_LETTER_LUIS = 0x1682;
const char32_t OGHAM_LETTER_FEARN = 0x1683;
const char32_t OGHAM_LETTER_SAIL = 0x1684;
const char32_t OGHAM_LETTER_NION = 0x1685;
const char32_t OGHAM_LETTER_UATH = 0x1686;
const char32_t OGHAM_LETTER_DAIR = 0x1687;
const char32_t OGHAM_LETTER_TINNE = 0x1688;
const char32_t OGHAM_LETTER_COLL = 0x1689;
const char32_t OGHAM_LETTER_CEIRT = 0x168A;
const char32_t OGHAM_LETTER_MUIN = 0x168B;
const char32_t OGHAM_LETTER_GORT = 0x168C;
const char32_t OGHAM_LETTER_NGEADAL = 0x168D;
const char32_t OGHAM_LETTER_STRAIF = 0x168E;
const char32_t OGHAM_LETTER_RUIS = 0x168F;
const char32_t OGHAM_LETTER_AILM = 0x1690;
const char32_t OGHAM_LETTER_ONN = 0x1691;
const char32_t OGHAM_LETTER_UR = 0x1692;
const char32_t OGHAM_LETTER_EADHADH = 0x1693;
const char32_t OGHAM_LETTER_IODHADH = 0x1694;
const char32_t OGHAM_LETTER_EABHADH = 0x1695;
const char32_t OGHAM_LETTER_OR = 0x1696;
const char32_t OGHAM_LETTER_UILLEANN = 0x1697;
const char32_t OGHAM_LETTER_IFIN = 0x1698;
const char32_t OGHAM_LETTER_EAMHANCHOLL = 0x1699;
const char32_t OGHAM_LETTER_PEITH = 0x169A;
const char32_t OGHAM_FEATHER_MARK = 0x169B;
const char32_t OGHAM_REVERSED_FEATHER_MARK = 0x169C;
const char32_t RUNIC_LETTER_FEHU_FEOH_FE_F = 0x16A0;
const char32_t RUNIC_LETTER_V = 0x16A1;
const char32_t RUNIC_LETTER_URUZ_UR_U = 0x16A2;
const char32_t RUNIC_LETTER_YR = 0x16A3;
const char32_t RUNIC_LETTER_Y = 0x16A4;
const char32_t RUNIC_LETTER_W = 0x16A5;
const char32_t RUNIC_LETTER_THURISAZ_THURS_THORN = 0x16A6;
const char32_t RUNIC_LETTER_ETH = 0x16A7;
const char32_t RUNIC_LETTER_ANSUZ_A = 0x16A8;
const char32_t RUNIC_LETTER_OS_O = 0x16A9;
const char32_t RUNIC_LETTER_AC_A = 0x16AA;
const char32_t RUNIC_LETTER_AESC = 0x16AB;
const char32_t RUNIC_LETTER_LONG_BRANCH_OSS_O = 0x16AC;
const char32_t RUNIC_LETTER_SHORT_TWIG_OSS_O = 0x16AD;
const char32_t RUNIC_LETTER_O = 0x16AE;
const char32_t RUNIC_LETTER_OE = 0x16AF;
const char32_t RUNIC_LETTER_ON = 0x16B0;
const char32_t RUNIC_LETTER_RAIDO_RAD_REID_R = 0x16B1;
const char32_t RUNIC_LETTER_KAUNA = 0x16B2;
const char32_t RUNIC_LETTER_CEN = 0x16B3;
const char32_t RUNIC_LETTER_KAUN_K = 0x16B4;
const char32_t RUNIC_LETTER_G = 0x16B5;
const char32_t RUNIC_LETTER_ENG = 0x16B6;
const char32_t RUNIC_LETTER_GEBO_GYFU_G = 0x16B7;
const char32_t RUNIC_LETTER_GAR = 0x16B8;
const char32_t RUNIC_LETTER_WUNJO_WYNN_W = 0x16B9;
const char32_t RUNIC_LETTER_HAGLAZ_H = 0x16BA;
const char32_t RUNIC_LETTER_HAEGL_H = 0x16BB;
const char32_t RUNIC_LETTER_LONG_BRANCH_HAGALL_H = 0x16BC;
const char32_t RUNIC_LETTER_SHORT_TWIG_HAGALL_H = 0x16BD;
const char32_t RUNIC_LETTER_NAUDIZ_NYD_NAUD_N = 0x16BE;
const char32_t RUNIC_LETTER_SHORT_TWIG_NAUD_N = 0x16BF;
const char32_t RUNIC_LETTER_DOTTED_N = 0x16C0;
const char32_t RUNIC_LETTER_ISAZ_IS_ISS_I = 0x16C1;
const char32_t RUNIC_LETTER_E = 0x16C2;
const char32_t RUNIC_LETTER_JERAN_J = 0x16C3;
const char32_t RUNIC_LETTER_GER = 0x16C4;
const char32_t RUNIC_LETTER_LONG_BRANCH_AR_AE = 0x16C5;
const char32_t RUNIC_LETTER_SHORT_TWIG_AR_A = 0x16C6;
const char32_t RUNIC_LETTER_IWAZ_EOH = 0x16C7;
const char32_t RUNIC_LETTER_PERTHO_PEORTH_P = 0x16C8;
const char32_t RUNIC_LETTER_ALGIZ_EOLHX = 0x16C9;
const char32_t RUNIC_LETTER_SOWILO_S = 0x16CA;
const char32_t RUNIC_LETTER_SIGEL_LONG_BRANCH_SOL_S = 0x16CB;
const char32_t RUNIC_LETTER_SHORT_TWIG_SOL_S = 0x16CC;
const char32_t RUNIC_LETTER_C = 0x16CD;
const char32_t RUNIC_LETTER_Z = 0x16CE;
const char32_t RUNIC_LETTER_TIWAZ_TIR_TYR_T = 0x16CF;
const char32_t RUNIC_LETTER_SHORT_TWIG_TYR_T = 0x16D0;
const char32_t RUNIC_LETTER_D = 0x16D1;
const char32_t RUNIC_LETTER_BERKANAN_BEORC_BJARKAN_B = 0x16D2;
const char32_t RUNIC_LETTER_SHORT_TWIG_BJARKAN_B = 0x16D3;
const char32_t RUNIC_LETTER_DOTTED_P = 0x16D4;
const char32_t RUNIC_LETTER_OPEN_P = 0x16D5;
const char32_t RUNIC_LETTER_EHWAZ_EH_E = 0x16D6;
const char32_t RUNIC_LETTER_MANNAZ_MAN_M = 0x16D7;
const char32_t RUNIC_LETTER_LONG_BRANCH_MADR_M = 0x16D8;
const char32_t RUNIC_LETTER_SHORT_TWIG_MADR_M = 0x16D9;
const char32_t RUNIC_LETTER_LAUKAZ_LAGU_LOGR_L = 0x16DA;
const char32_t RUNIC_LETTER_DOTTED_L = 0x16DB;
const char32_t RUNIC_LETTER_INGWAZ = 0x16DC;
const char32_t RUNIC_LETTER_ING = 0x16DD;
const char32_t RUNIC_LETTER_DAGAZ_DAEG_D = 0x16DE;
const char32_t RUNIC_LETTER_OTHALAN_ETHEL_O = 0x16DF;
const char32_t RUNIC_LETTER_EAR = 0x16E0;
const char32_t RUNIC_LETTER_IOR = 0x16E1;
const char32_t RUNIC_LETTER_CWEORTH = 0x16E2;
const char32_t RUNIC_LETTER_CALC = 0x16E3;
const char32_t RUNIC_LETTER_CEALC = 0x16E4;
const char32_t RUNIC_LETTER_STAN = 0x16E5;
const char32_t RUNIC_LETTER_LONG_BRANCH_YR = 0x16E6;
const char32_t RUNIC_LETTER_SHORT_TWIG_YR = 0x16E7;
const char32_t RUNIC_LETTER_ICELANDIC_YR = 0x16E8;
const char32_t RUNIC_LETTER_Q = 0x16E9;
const char32_t RUNIC_LETTER_X = 0x16EA;
const char32_t RUNIC_SINGLE_PUNCTUATION = 0x16EB;
const char32_t RUNIC_MULTIPLE_PUNCTUATION = 0x16EC;
const char32_t RUNIC_CROSS_PUNCTUATION = 0x16ED;
const char32_t RUNIC_ARLAUG_SYMBOL = 0x16EE;
const char32_t RUNIC_TVIMADUR_SYMBOL = 0x16EF;
const char32_t RUNIC_BELGTHOR_SYMBOL = 0x16F0;
const char32_t RUNIC_LETTER_K = 0x16F1;
const char32_t RUNIC_LETTER_SH = 0x16F2;
const char32_t RUNIC_LETTER_OO = 0x16F3;
const char32_t RUNIC_LETTER_FRANKS_CASKET_OS = 0x16F4;
const char32_t RUNIC_LETTER_FRANKS_CASKET_IS = 0x16F5;
const char32_t RUNIC_LETTER_FRANKS_CASKET_EH = 0x16F6;
const char32_t RUNIC_LETTER_FRANKS_CASKET_AC = 0x16F7;
const char32_t RUNIC_LETTER_FRANKS_CASKET_AESC = 0x16F8;
const char32_t TAGALOG_LETTER_A = 0x1700;
const char32_t TAGALOG_LETTER_I = 0x1701;
const char32_t TAGALOG_LETTER_U = 0x1702;
const char32_t TAGALOG_LETTER_KA = 0x1703;
const char32_t TAGALOG_LETTER_GA = 0x1704;
const char32_t TAGALOG_LETTER_NGA = 0x1705;
const char32_t TAGALOG_LETTER_TA = 0x1706;
const char32_t TAGALOG_LETTER_DA = 0x1707;
const char32_t TAGALOG_LETTER_NA = 0x1708;
const char32_t TAGALOG_LETTER_PA = 0x1709;
const char32_t TAGALOG_LETTER_BA = 0x170A;
const char32_t TAGALOG_LETTER_MA = 0x170B;
const char32_t TAGALOG_LETTER_YA = 0x170C;
const char32_t TAGALOG_LETTER_RA = 0x170D;
const char32_t TAGALOG_LETTER_LA = 0x170E;
const char32_t TAGALOG_LETTER_WA = 0x170F;
const char32_t TAGALOG_LETTER_SA = 0x1710;
const char32_t TAGALOG_LETTER_HA = 0x1711;
const char32_t TAGALOG_VOWEL_SIGN_I = 0x1712;
const char32_t TAGALOG_VOWEL_SIGN_U = 0x1713;
const char32_t TAGALOG_SIGN_VIRAMA = 0x1714;
const char32_t TAGALOG_SIGN_PAMUDPOD = 0x1715;
const char32_t TAGALOG_LETTER_ARCHAIC_RA = 0x171F;
const char32_t HANUNOO_LETTER_A = 0x1720;
const char32_t HANUNOO_LETTER_I = 0x1721;
const char32_t HANUNOO_LETTER_U = 0x1722;
const char32_t HANUNOO_LETTER_KA = 0x1723;
const char32_t HANUNOO_LETTER_GA = 0x1724;
const char32_t HANUNOO_LETTER_NGA = 0x1725;
const char32_t HANUNOO_LETTER_TA = 0x1726;
const char32_t HANUNOO_LETTER_DA = 0x1727;
const char32_t HANUNOO_LETTER_NA = 0x1728;
const char32_t HANUNOO_LETTER_PA = 0x1729;
const char32_t HANUNOO_LETTER_BA = 0x172A;
const char32_t HANUNOO_LETTER_MA = 0x172B;
const char32_t HANUNOO_LETTER_YA = 0x172C;
const char32_t HANUNOO_LETTER_RA = 0x172D;
const char32_t HANUNOO_LETTER_LA = 0x172E;
const char32_t HANUNOO_LETTER_WA = 0x172F;
const char32_t HANUNOO_LETTER_SA = 0x1730;
const char32_t HANUNOO_LETTER_HA = 0x1731;
const char32_t HANUNOO_VOWEL_SIGN_I = 0x1732;
const char32_t HANUNOO_VOWEL_SIGN_U = 0x1733;
const char32_t HANUNOO_SIGN_PAMUDPOD = 0x1734;
const char32_t PHILIPPINE_SINGLE_PUNCTUATION = 0x1735;
const char32_t PHILIPPINE_DOUBLE_PUNCTUATION = 0x1736;
const char32_t BUHID_LETTER_A = 0x1740;
const char32_t BUHID_LETTER_I = 0x1741;
const char32_t BUHID_LETTER_U = 0x1742;
const char32_t BUHID_LETTER_KA = 0x1743;
const char32_t BUHID_LETTER_GA = 0x1744;
const char32_t BUHID_LETTER_NGA = 0x1745;
const char32_t BUHID_LETTER_TA = 0x1746;
const char32_t BUHID_LETTER_DA = 0x1747;
const char32_t BUHID_LETTER_NA = 0x1748;
const char32_t BUHID_LETTER_PA = 0x1749;
const char32_t BUHID_LETTER_BA = 0x174A;
const char32_t BUHID_LETTER_MA = 0x174B;
const char32_t BUHID_LETTER_YA = 0x174C;
const char32_t BUHID_LETTER_RA = 0x174D;
const char32_t BUHID_LETTER_LA = 0x174E;
const char32_t BUHID_LETTER_WA = 0x174F;
const char32_t BUHID_LETTER_SA = 0x1750;
const char32_t BUHID_LETTER_HA = 0x1751;
const char32_t BUHID_VOWEL_SIGN_I = 0x1752;
const char32_t BUHID_VOWEL_SIGN_U = 0x1753;
const char32_t TAGBANWA_LETTER_A = 0x1760;
const char32_t TAGBANWA_LETTER_I = 0x1761;
const char32_t TAGBANWA_LETTER_U = 0x1762;
const char32_t TAGBANWA_LETTER_KA = 0x1763;
const char32_t TAGBANWA_LETTER_GA = 0x1764;
const char32_t TAGBANWA_LETTER_NGA = 0x1765;
const char32_t TAGBANWA_LETTER_TA = 0x1766;
const char32_t TAGBANWA_LETTER_DA = 0x1767;
const char32_t TAGBANWA_LETTER_NA = 0x1768;
const char32_t TAGBANWA_LETTER_PA = 0x1769;
const char32_t TAGBANWA_LETTER_BA = 0x176A;
const char32_t TAGBANWA_LETTER_MA = 0x176B;
const char32_t TAGBANWA_LETTER_YA = 0x176C;
const char32_t TAGBANWA_LETTER_LA = 0x176E;
const char32_t TAGBANWA_LETTER_WA = 0x176F;
const char32_t TAGBANWA_LETTER_SA = 0x1770;
const char32_t TAGBANWA_VOWEL_SIGN_I = 0x1772;
const char32_t TAGBANWA_VOWEL_SIGN_U = 0x1773;
const char32_t KHMER_LETTER_KA = 0x1780;
const char32_t KHMER_LETTER_KHA = 0x1781;
const char32_t KHMER_LETTER_KO = 0x1782;
const char32_t KHMER_LETTER_KHO = 0x1783;
const char32_t KHMER_LETTER_NGO = 0x1784;
const char32_t KHMER_LETTER_CA = 0x1785;
const char32_t KHMER_LETTER_CHA = 0x1786;
const char32_t KHMER_LETTER_CO = 0x1787;
const char32_t KHMER_LETTER_CHO = 0x1788;
const char32_t KHMER_LETTER_NYO = 0x1789;
const char32_t KHMER_LETTER_DA = 0x178A;
const char32_t KHMER_LETTER_TTHA = 0x178B;
const char32_t KHMER_LETTER_DO = 0x178C;
const char32_t KHMER_LETTER_TTHO = 0x178D;
const char32_t KHMER_LETTER_NNO = 0x178E;
const char32_t KHMER_LETTER_TA = 0x178F;
const char32_t KHMER_LETTER_THA = 0x1790;
const char32_t KHMER_LETTER_TO = 0x1791;
const char32_t KHMER_LETTER_THO = 0x1792;
const char32_t KHMER_LETTER_NO = 0x1793;
const char32_t KHMER_LETTER_BA = 0x1794;
const char32_t KHMER_LETTER_PHA = 0x1795;
const char32_t KHMER_LETTER_PO = 0x1796;
const char32_t KHMER_LETTER_PHO = 0x1797;
const char32_t KHMER_LETTER_MO = 0x1798;
const char32_t KHMER_LETTER_YO = 0x1799;
const char32_t KHMER_LETTER_RO = 0x179A;
const char32_t KHMER_LETTER_LO = 0x179B;
const char32_t KHMER_LETTER_VO = 0x179C;
const char32_t KHMER_LETTER_SHA = 0x179D;
const char32_t KHMER_LETTER_SSO = 0x179E;
const char32_t KHMER_LETTER_SA = 0x179F;
const char32_t KHMER_LETTER_HA = 0x17A0;
const char32_t KHMER_LETTER_LA = 0x17A1;
const char32_t KHMER_LETTER_QA = 0x17A2;
const char32_t KHMER_INDEPENDENT_VOWEL_QAQ = 0x17A3;
const char32_t KHMER_INDEPENDENT_VOWEL_QAA = 0x17A4;
const char32_t KHMER_INDEPENDENT_VOWEL_QI = 0x17A5;
const char32_t KHMER_INDEPENDENT_VOWEL_QII = 0x17A6;
const char32_t KHMER_INDEPENDENT_VOWEL_QU = 0x17A7;
const char32_t KHMER_INDEPENDENT_VOWEL_QUK = 0x17A8;
const char32_t KHMER_INDEPENDENT_VOWEL_QUU = 0x17A9;
const char32_t KHMER_INDEPENDENT_VOWEL_QUUV = 0x17AA;
const char32_t KHMER_INDEPENDENT_VOWEL_RY = 0x17AB;
const char32_t KHMER_INDEPENDENT_VOWEL_RYY = 0x17AC;
const char32_t KHMER_INDEPENDENT_VOWEL_LY = 0x17AD;
const char32_t KHMER_INDEPENDENT_VOWEL_LYY = 0x17AE;
const char32_t KHMER_INDEPENDENT_VOWEL_QE = 0x17AF;
const char32_t KHMER_INDEPENDENT_VOWEL_QAI = 0x17B0;
const char32_t KHMER_INDEPENDENT_VOWEL_QOO_TYPE_ONE = 0x17B1;
const char32_t KHMER_INDEPENDENT_VOWEL_QOO_TYPE_TWO = 0x17B2;
const char32_t KHMER_INDEPENDENT_VOWEL_QAU = 0x17B3;
const char32_t KHMER_VOWEL_INHERENT_AQ = 0x17B4;
const char32_t KHMER_VOWEL_INHERENT_AA = 0x17B5;
const char32_t KHMER_VOWEL_SIGN_AA = 0x17B6;
const char32_t KHMER_VOWEL_SIGN_I = 0x17B7;
const char32_t KHMER_VOWEL_SIGN_II = 0x17B8;
const char32_t KHMER_VOWEL_SIGN_Y = 0x17B9;
const char32_t KHMER_VOWEL_SIGN_YY = 0x17BA;
const char32_t KHMER_VOWEL_SIGN_U = 0x17BB;
const char32_t KHMER_VOWEL_SIGN_UU = 0x17BC;
const char32_t KHMER_VOWEL_SIGN_UA = 0x17BD;
const char32_t KHMER_VOWEL_SIGN_OE = 0x17BE;
const char32_t KHMER_VOWEL_SIGN_YA = 0x17BF;
const char32_t KHMER_VOWEL_SIGN_IE = 0x17C0;
const char32_t KHMER_VOWEL_SIGN_E = 0x17C1;
const char32_t KHMER_VOWEL_SIGN_AE = 0x17C2;
const char32_t KHMER_VOWEL_SIGN_AI = 0x17C3;
const char32_t KHMER_VOWEL_SIGN_OO = 0x17C4;
const char32_t KHMER_VOWEL_SIGN_AU = 0x17C5;
const char32_t KHMER_SIGN_NIKAHIT = 0x17C6;
const char32_t KHMER_SIGN_REAHMUK = 0x17C7;
const char32_t KHMER_SIGN_YUUKALEAPINTU = 0x17C8;
const char32_t KHMER_SIGN_MUUSIKATOAN = 0x17C9;
const char32_t KHMER_SIGN_TRIISAP = 0x17CA;
const char32_t KHMER_SIGN_BANTOC = 0x17CB;
const char32_t KHMER_SIGN_ROBAT = 0x17CC;
const char32_t KHMER_SIGN_TOANDAKHIAT = 0x17CD;
const char32_t KHMER_SIGN_KAKABAT = 0x17CE;
const char32_t KHMER_SIGN_AHSDA = 0x17CF;
const char32_t KHMER_SIGN_SAMYOK_SANNYA = 0x17D0;
const char32_t KHMER_SIGN_VIRIAM = 0x17D1;
const char32_t KHMER_SIGN_COENG = 0x17D2;
const char32_t KHMER_SIGN_BATHAMASAT = 0x17D3;
const char32_t KHMER_SIGN_KHAN = 0x17D4;
const char32_t KHMER_SIGN_BARIYOOSAN = 0x17D5;
const char32_t KHMER_SIGN_CAMNUC_PII_KUUH = 0x17D6;
const char32_t KHMER_SIGN_LEK_TOO = 0x17D7;
const char32_t KHMER_SIGN_BEYYAL = 0x17D8;
const char32_t KHMER_SIGN_PHNAEK_MUAN = 0x17D9;
const char32_t KHMER_SIGN_KOOMUUT = 0x17DA;
const char32_t KHMER_CURRENCY_SYMBOL_RIEL = 0x17DB;
const char32_t KHMER_SIGN_AVAKRAHASANYA = 0x17DC;
const char32_t KHMER_SIGN_ATTHACAN = 0x17DD;
const char32_t KHMER_DIGIT_ZERO = 0x17E0;
const char32_t KHMER_DIGIT_ONE = 0x17E1;
const char32_t KHMER_DIGIT_TWO = 0x17E2;
const char32_t KHMER_DIGIT_THREE = 0x17E3;
const char32_t KHMER_DIGIT_FOUR = 0x17E4;
const char32_t KHMER_DIGIT_FIVE = 0x17E5;
const char32_t KHMER_DIGIT_SIX = 0x17E6;
const char32_t KHMER_DIGIT_SEVEN = 0x17E7;
const char32_t KHMER_DIGIT_EIGHT = 0x17E8;
const char32_t KHMER_DIGIT_NINE = 0x17E9;
const char32_t KHMER_SYMBOL_LEK_ATTAK_SON = 0x17F0;
const char32_t KHMER_SYMBOL_LEK_ATTAK_MUOY = 0x17F1;
const char32_t KHMER_SYMBOL_LEK_ATTAK_PII = 0x17F2;
const char32_t KHMER_SYMBOL_LEK_ATTAK_BEI = 0x17F3;
const char32_t KHMER_SYMBOL_LEK_ATTAK_BUON = 0x17F4;
const char32_t KHMER_SYMBOL_LEK_ATTAK_PRAM = 0x17F5;
const char32_t KHMER_SYMBOL_LEK_ATTAK_PRAM_MUOY = 0x17F6;
const char32_t KHMER_SYMBOL_LEK_ATTAK_PRAM_PII = 0x17F7;
const char32_t KHMER_SYMBOL_LEK_ATTAK_PRAM_BEI = 0x17F8;
const char32_t KHMER_SYMBOL_LEK_ATTAK_PRAM_BUON = 0x17F9;
const char32_t MONGOLIAN_BIRGA = 0x1800;
const char32_t MONGOLIAN_ELLIPSIS = 0x1801;
const char32_t MONGOLIAN_COMMA = 0x1802;
const char32_t MONGOLIAN_FULL_STOP = 0x1803;
const char32_t MONGOLIAN_COLON = 0x1804;
const char32_t MONGOLIAN_FOUR_DOTS = 0x1805;
const char32_t MONGOLIAN_TODO_SOFT_HYPHEN = 0x1806;
const char32_t MONGOLIAN_SIBE_SYLLABLE_BOUNDARY_MARKER = 0x1807;
const char32_t MONGOLIAN_MANCHU_COMMA = 0x1808;
const char32_t MONGOLIAN_MANCHU_FULL_STOP = 0x1809;
const char32_t MONGOLIAN_NIRUGU = 0x180A;
const char32_t MONGOLIAN_FREE_VARIATION_SELECTOR_ONE = 0x180B;
const char32_t MONGOLIAN_FREE_VARIATION_SELECTOR_TWO = 0x180C;
const char32_t MONGOLIAN_FREE_VARIATION_SELECTOR_THREE = 0x180D;
const char32_t MONGOLIAN_VOWEL_SEPARATOR = 0x180E;
const char32_t MONGOLIAN_FREE_VARIATION_SELECTOR_FOUR = 0x180F;
const char32_t MONGOLIAN_DIGIT_ZERO = 0x1810;
const char32_t MONGOLIAN_DIGIT_ONE = 0x1811;
const char32_t MONGOLIAN_DIGIT_TWO = 0x1812;
const char32_t MONGOLIAN_DIGIT_THREE = 0x1813;
const char32_t MONGOLIAN_DIGIT_FOUR = 0x1814;
const char32_t MONGOLIAN_DIGIT_FIVE = 0x1815;
const char32_t MONGOLIAN_DIGIT_SIX = 0x1816;
const char32_t MONGOLIAN_DIGIT_SEVEN = 0x1817;
const char32_t MONGOLIAN_DIGIT_EIGHT = 0x1818;
const char32_t MONGOLIAN_DIGIT_NINE = 0x1819;
const char32_t MONGOLIAN_LETTER_A = 0x1820;
const char32_t MONGOLIAN_LETTER_E = 0x1821;
const char32_t MONGOLIAN_LETTER_I = 0x1822;
const char32_t MONGOLIAN_LETTER_O = 0x1823;
const char32_t MONGOLIAN_LETTER_U = 0x1824;
const char32_t MONGOLIAN_LETTER_OE = 0x1825;
const char32_t MONGOLIAN_LETTER_UE = 0x1826;
const char32_t MONGOLIAN_LETTER_EE = 0x1827;
const char32_t MONGOLIAN_LETTER_NA = 0x1828;
const char32_t MONGOLIAN_LETTER_ANG = 0x1829;
const char32_t MONGOLIAN_LETTER_BA = 0x182A;
const char32_t MONGOLIAN_LETTER_PA = 0x182B;
const char32_t MONGOLIAN_LETTER_QA = 0x182C;
const char32_t MONGOLIAN_LETTER_GA = 0x182D;
const char32_t MONGOLIAN_LETTER_MA = 0x182E;
const char32_t MONGOLIAN_LETTER_LA = 0x182F;
const char32_t MONGOLIAN_LETTER_SA = 0x1830;
const char32_t MONGOLIAN_LETTER_SHA = 0x1831;
const char32_t MONGOLIAN_LETTER_TA = 0x1832;
const char32_t MONGOLIAN_LETTER_DA = 0x1833;
const char32_t MONGOLIAN_LETTER_CHA = 0x1834;
const char32_t MONGOLIAN_LETTER_JA = 0x1835;
const char32_t MONGOLIAN_LETTER_YA = 0x1836;
const char32_t MONGOLIAN_LETTER_RA = 0x1837;
const char32_t MONGOLIAN_LETTER_WA = 0x1838;
const char32_t MONGOLIAN_LETTER_FA = 0x1839;
const char32_t MONGOLIAN_LETTER_KA = 0x183A;
const char32_t MONGOLIAN_LETTER_KHA = 0x183B;
const char32_t MONGOLIAN_LETTER_TSA = 0x183C;
const char32_t MONGOLIAN_LETTER_ZA = 0x183D;
const char32_t MONGOLIAN_LETTER_HAA = 0x183E;
const char32_t MONGOLIAN_LETTER_ZRA = 0x183F;
const char32_t MONGOLIAN_LETTER_LHA = 0x1840;
const char32_t MONGOLIAN_LETTER_ZHI = 0x1841;
const char32_t MONGOLIAN_LETTER_CHI = 0x1842;
const char32_t MONGOLIAN_LETTER_TODO_LONG_VOWEL_SIGN = 0x1843;
const char32_t MONGOLIAN_LETTER_TODO_E = 0x1844;
const char32_t MONGOLIAN_LETTER_TODO_I = 0x1845;
const char32_t MONGOLIAN_LETTER_TODO_O = 0x1846;
const char32_t MONGOLIAN_LETTER_TODO_U = 0x1847;
const char32_t MONGOLIAN_LETTER_TODO_OE = 0x1848;
const char32_t MONGOLIAN_LETTER_TODO_UE = 0x1849;
const char32_t MONGOLIAN_LETTER_TODO_ANG = 0x184A;
const char32_t MONGOLIAN_LETTER_TODO_BA = 0x184B;
const char32_t MONGOLIAN_LETTER_TODO_PA = 0x184C;
const char32_t MONGOLIAN_LETTER_TODO_QA = 0x184D;
const char32_t MONGOLIAN_LETTER_TODO_GA = 0x184E;
const char32_t MONGOLIAN_LETTER_TODO_MA = 0x184F;
const char32_t MONGOLIAN_LETTER_TODO_TA = 0x1850;
const char32_t MONGOLIAN_LETTER_TODO_DA = 0x1851;
const char32_t MONGOLIAN_LETTER_TODO_CHA = 0x1852;
const char32_t MONGOLIAN_LETTER_TODO_JA = 0x1853;
const char32_t MONGOLIAN_LETTER_TODO_TSA = 0x1854;
const char32_t MONGOLIAN_LETTER_TODO_YA = 0x1855;
const char32_t MONGOLIAN_LETTER_TODO_WA = 0x1856;
const char32_t MONGOLIAN_LETTER_TODO_KA = 0x1857;
const char32_t MONGOLIAN_LETTER_TODO_GAA = 0x1858;
const char32_t MONGOLIAN_LETTER_TODO_HAA = 0x1859;
const char32_t MONGOLIAN_LETTER_TODO_JIA = 0x185A;
const char32_t MONGOLIAN_LETTER_TODO_NIA = 0x185B;
const char32_t MONGOLIAN_LETTER_TODO_DZA = 0x185C;
const char32_t MONGOLIAN_LETTER_SIBE_E = 0x185D;
const char32_t MONGOLIAN_LETTER_SIBE_I = 0x185E;
const char32_t MONGOLIAN_LETTER_SIBE_IY = 0x185F;
const char32_t MONGOLIAN_LETTER_SIBE_UE = 0x1860;
const char32_t MONGOLIAN_LETTER_SIBE_U = 0x1861;
const char32_t MONGOLIAN_LETTER_SIBE_ANG = 0x1862;
const char32_t MONGOLIAN_LETTER_SIBE_KA = 0x1863;
const char32_t MONGOLIAN_LETTER_SIBE_GA = 0x1864;
const char32_t MONGOLIAN_LETTER_SIBE_HA = 0x1865;
const char32_t MONGOLIAN_LETTER_SIBE_PA = 0x1866;
const char32_t MONGOLIAN_LETTER_SIBE_SHA = 0x1867;
const char32_t MONGOLIAN_LETTER_SIBE_TA = 0x1868;
const char32_t MONGOLIAN_LETTER_SIBE_DA = 0x1869;
const char32_t MONGOLIAN_LETTER_SIBE_JA = 0x186A;
const char32_t MONGOLIAN_LETTER_SIBE_FA = 0x186B;
const char32_t MONGOLIAN_LETTER_SIBE_GAA = 0x186C;
const char32_t MONGOLIAN_LETTER_SIBE_HAA = 0x186D;
const char32_t MONGOLIAN_LETTER_SIBE_TSA = 0x186E;
const char32_t MONGOLIAN_LETTER_SIBE_ZA = 0x186F;
const char32_t MONGOLIAN_LETTER_SIBE_RAA = 0x1870;
const char32_t MONGOLIAN_LETTER_SIBE_CHA = 0x1871;
const char32_t MONGOLIAN_LETTER_SIBE_ZHA = 0x1872;
const char32_t MONGOLIAN_LETTER_MANCHU_I = 0x1873;
const char32_t MONGOLIAN_LETTER_MANCHU_KA = 0x1874;
const char32_t MONGOLIAN_LETTER_MANCHU_RA = 0x1875;
const char32_t MONGOLIAN_LETTER_MANCHU_FA = 0x1876;
const char32_t MONGOLIAN_LETTER_MANCHU_ZHA = 0x1877;
const char32_t MONGOLIAN_LETTER_CHA_WITH_TWO_DOTS = 0x1878;
const char32_t MONGOLIAN_LETTER_ALI_GALI_ANUSVARA_ONE = 0x1880;
const char32_t MONGOLIAN_LETTER_ALI_GALI_VISARGA_ONE = 0x1881;
const char32_t MONGOLIAN_LETTER_ALI_GALI_DAMARU = 0x1882;
const char32_t MONGOLIAN_LETTER_ALI_GALI_UBADAMA = 0x1883;
const char32_t MONGOLIAN_LETTER_ALI_GALI_INVERTED_UBADAMA = 0x1884;
const char32_t MONGOLIAN_LETTER_ALI_GALI_BALUDA = 0x1885;
const char32_t MONGOLIAN_LETTER_ALI_GALI_THREE_BALUDA = 0x1886;
const char32_t MONGOLIAN_LETTER_ALI_GALI_A = 0x1887;
const char32_t MONGOLIAN_LETTER_ALI_GALI_I = 0x1888;
const char32_t MONGOLIAN_LETTER_ALI_GALI_KA = 0x1889;
const char32_t MONGOLIAN_LETTER_ALI_GALI_NGA = 0x188A;
const char32_t MONGOLIAN_LETTER_ALI_GALI_CA = 0x188B;
const char32_t MONGOLIAN_LETTER_ALI_GALI_TTA = 0x188C;
const char32_t MONGOLIAN_LETTER_ALI_GALI_TTHA = 0x188D;
const char32_t MONGOLIAN_LETTER_ALI_GALI_DDA = 0x188E;
const char32_t MONGOLIAN_LETTER_ALI_GALI_NNA = 0x188F;
const char32_t MONGOLIAN_LETTER_ALI_GALI_TA = 0x1890;
const char32_t MONGOLIAN_LETTER_ALI_GALI_DA = 0x1891;
const char32_t MONGOLIAN_LETTER_ALI_GALI_PA = 0x1892;
const char32_t MONGOLIAN_LETTER_ALI_GALI_PHA = 0x1893;
const char32_t MONGOLIAN_LETTER_ALI_GALI_SSA = 0x1894;
const char32_t MONGOLIAN_LETTER_ALI_GALI_ZHA = 0x1895;
const char32_t MONGOLIAN_LETTER_ALI_GALI_ZA = 0x1896;
const char32_t MONGOLIAN_LETTER_ALI_GALI_AH = 0x1897;
const char32_t MONGOLIAN_LETTER_TODO_ALI_GALI_TA = 0x1898;
const char32_t MONGOLIAN_LETTER_TODO_ALI_GALI_ZHA = 0x1899;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_GHA = 0x189A;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_NGA = 0x189B;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_CA = 0x189C;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_JHA = 0x189D;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_TTA = 0x189E;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_DDHA = 0x189F;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_TA = 0x18A0;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_DHA = 0x18A1;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_SSA = 0x18A2;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_CYA = 0x18A3;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_ZHA = 0x18A4;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_ZA = 0x18A5;
const char32_t MONGOLIAN_LETTER_ALI_GALI_HALF_U = 0x18A6;
const char32_t MONGOLIAN_LETTER_ALI_GALI_HALF_YA = 0x18A7;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_BHA = 0x18A8;
const char32_t MONGOLIAN_LETTER_ALI_GALI_DAGALGA = 0x18A9;
const char32_t MONGOLIAN_LETTER_MANCHU_ALI_GALI_LHA = 0x18AA;
const char32_t CANADIAN_SYLLABICS_OY = 0x18B0;
const char32_t CANADIAN_SYLLABICS_AY = 0x18B1;
const char32_t CANADIAN_SYLLABICS_AAY = 0x18B2;
const char32_t CANADIAN_SYLLABICS_WAY = 0x18B3;
const char32_t CANADIAN_SYLLABICS_POY = 0x18B4;
const char32_t CANADIAN_SYLLABICS_PAY = 0x18B5;
const char32_t CANADIAN_SYLLABICS_PWOY = 0x18B6;
const char32_t CANADIAN_SYLLABICS_TAY = 0x18B7;
const char32_t CANADIAN_SYLLABICS_KAY = 0x18B8;
const char32_t CANADIAN_SYLLABICS_KWAY = 0x18B9;
const char32_t CANADIAN_SYLLABICS_MAY = 0x18BA;
const char32_t CANADIAN_SYLLABICS_NOY = 0x18BB;
const char32_t CANADIAN_SYLLABICS_NAY = 0x18BC;
const char32_t CANADIAN_SYLLABICS_LAY = 0x18BD;
const char32_t CANADIAN_SYLLABICS_SOY = 0x18BE;
const char32_t CANADIAN_SYLLABICS_SAY = 0x18BF;
const char32_t CANADIAN_SYLLABICS_SHOY = 0x18C0;
const char32_t CANADIAN_SYLLABICS_SHAY = 0x18C1;
const char32_t CANADIAN_SYLLABICS_SHWOY = 0x18C2;
const char32_t CANADIAN_SYLLABICS_YOY = 0x18C3;
const char32_t CANADIAN_SYLLABICS_YAY = 0x18C4;
const char32_t CANADIAN_SYLLABICS_RAY = 0x18C5;
const char32_t CANADIAN_SYLLABICS_NWI = 0x18C6;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_NWI = 0x18C7;
const char32_t CANADIAN_SYLLABICS_NWII = 0x18C8;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_NWII = 0x18C9;
const char32_t CANADIAN_SYLLABICS_NWO = 0x18CA;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_NWO = 0x18CB;
const char32_t CANADIAN_SYLLABICS_NWOO = 0x18CC;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_NWOO = 0x18CD;
const char32_t CANADIAN_SYLLABICS_RWEE = 0x18CE;
const char32_t CANADIAN_SYLLABICS_RWI = 0x18CF;
const char32_t CANADIAN_SYLLABICS_RWII = 0x18D0;
const char32_t CANADIAN_SYLLABICS_RWO = 0x18D1;
const char32_t CANADIAN_SYLLABICS_RWOO = 0x18D2;
const char32_t CANADIAN_SYLLABICS_RWA = 0x18D3;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_P = 0x18D4;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_T = 0x18D5;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_K = 0x18D6;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_C = 0x18D7;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_M = 0x18D8;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_N = 0x18D9;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_S = 0x18DA;
const char32_t CANADIAN_SYLLABICS_OJIBWAY_SH = 0x18DB;
const char32_t CANADIAN_SYLLABICS_EASTERN_W = 0x18DC;
const char32_t CANADIAN_SYLLABICS_WESTERN_W = 0x18DD;
const char32_t CANADIAN_SYLLABICS_FINAL_SMALL_RING = 0x18DE;
const char32_t CANADIAN_SYLLABICS_FINAL_RAISED_DOT = 0x18DF;
const char32_t CANADIAN_SYLLABICS_R_CREE_RWE = 0x18E0;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LOO = 0x18E1;
const char32_t CANADIAN_SYLLABICS_WEST_CREE_LAA = 0x18E2;
const char32_t CANADIAN_SYLLABICS_THWE = 0x18E3;
const char32_t CANADIAN_SYLLABICS_THWA = 0x18E4;
const char32_t CANADIAN_SYLLABICS_TTHWE = 0x18E5;
const char32_t CANADIAN_SYLLABICS_TTHOO = 0x18E6;
const char32_t CANADIAN_SYLLABICS_TTHAA = 0x18E7;
const char32_t CANADIAN_SYLLABICS_TLHWE = 0x18E8;
const char32_t CANADIAN_SYLLABICS_TLHOO = 0x18E9;
const char32_t CANADIAN_SYLLABICS_SAYISI_SHWE = 0x18EA;
const char32_t CANADIAN_SYLLABICS_SAYISI_SHOO = 0x18EB;
const char32_t CANADIAN_SYLLABICS_SAYISI_HOO = 0x18EC;
const char32_t CANADIAN_SYLLABICS_CARRIER_GWU = 0x18ED;
const char32_t CANADIAN_SYLLABICS_CARRIER_DENE_GEE = 0x18EE;
const char32_t CANADIAN_SYLLABICS_CARRIER_GAA = 0x18EF;
const char32_t CANADIAN_SYLLABICS_CARRIER_GWA = 0x18F0;
const char32_t CANADIAN_SYLLABICS_SAYISI_JUU = 0x18F1;
const char32_t CANADIAN_SYLLABICS_CARRIER_JWA = 0x18F2;
const char32_t CANADIAN_SYLLABICS_BEAVER_DENE_L = 0x18F3;
const char32_t CANADIAN_SYLLABICS_BEAVER_DENE_R = 0x18F4;
const char32_t CANADIAN_SYLLABICS_CARRIER_DENTAL_S = 0x18F5;
const char32_t LIMBU_VOWEL_CARRIER_LETTER = 0x1900;
const char32_t LIMBU_LETTER_KA = 0x1901;
const char32_t LIMBU_LETTER_KHA = 0x1902;
const char32_t LIMBU_LETTER_GA = 0x1903;
const char32_t LIMBU_LETTER_GHA = 0x1904;
const char32_t LIMBU_LETTER_NGA = 0x1905;
const char32_t LIMBU_LETTER_CA = 0x1906;
const char32_t LIMBU_LETTER_CHA = 0x1907;
const char32_t LIMBU_LETTER_JA = 0x1908;
const char32_t LIMBU_LETTER_JHA = 0x1909;
const char32_t LIMBU_LETTER_YAN = 0x190A;
const char32_t LIMBU_LETTER_TA = 0x190B;
const char32_t LIMBU_LETTER_THA = 0x190C;
const char32_t LIMBU_LETTER_DA = 0x190D;
const char32_t LIMBU_LETTER_DHA = 0x190E;
const char32_t LIMBU_LETTER_NA = 0x190F;
const char32_t LIMBU_LETTER_PA = 0x1910;
const char32_t LIMBU_LETTER_PHA = 0x1911;
const char32_t LIMBU_LETTER_BA = 0x1912;
const char32_t LIMBU_LETTER_BHA = 0x1913;
const char32_t LIMBU_LETTER_MA = 0x1914;
const char32_t LIMBU_LETTER_YA = 0x1915;
const char32_t LIMBU_LETTER_RA = 0x1916;
const char32_t LIMBU_LETTER_LA = 0x1917;
const char32_t LIMBU_LETTER_WA = 0x1918;
const char32_t LIMBU_LETTER_SHA = 0x1919;
const char32_t LIMBU_LETTER_SSA = 0x191A;
const char32_t LIMBU_LETTER_SA = 0x191B;
const char32_t LIMBU_LETTER_HA = 0x191C;
const char32_t LIMBU_LETTER_GYAN = 0x191D;
const char32_t LIMBU_LETTER_TRA = 0x191E;
const char32_t LIMBU_VOWEL_SIGN_A = 0x1920;
const char32_t LIMBU_VOWEL_SIGN_I = 0x1921;
const char32_t LIMBU_VOWEL_SIGN_U = 0x1922;
const char32_t LIMBU_VOWEL_SIGN_EE = 0x1923;
const char32_t LIMBU_VOWEL_SIGN_AI = 0x1924;
const char32_t LIMBU_VOWEL_SIGN_OO = 0x1925;
const char32_t LIMBU_VOWEL_SIGN_AU = 0x1926;
const char32_t LIMBU_VOWEL_SIGN_E = 0x1927;
const char32_t LIMBU_VOWEL_SIGN_O = 0x1928;
const char32_t LIMBU_SUBJOINED_LETTER_YA = 0x1929;
const char32_t LIMBU_SUBJOINED_LETTER_RA = 0x192A;
const char32_t LIMBU_SUBJOINED_LETTER_WA = 0x192B;
const char32_t LIMBU_SMALL_LETTER_KA = 0x1930;
const char32_t LIMBU_SMALL_LETTER_NGA = 0x1931;
const char32_t LIMBU_SMALL_LETTER_ANUSVARA = 0x1932;
const char32_t LIMBU_SMALL_LETTER_TA = 0x1933;
const char32_t LIMBU_SMALL_LETTER_NA = 0x1934;
const char32_t LIMBU_SMALL_LETTER_PA = 0x1935;
const char32_t LIMBU_SMALL_LETTER_MA = 0x1936;
const char32_t LIMBU_SMALL_LETTER_RA = 0x1937;
const char32_t LIMBU_SMALL_LETTER_LA = 0x1938;
const char32_t LIMBU_SIGN_MUKPHRENG = 0x1939;
const char32_t LIMBU_SIGN_KEMPHRENG = 0x193A;
const char32_t LIMBU_SIGN_SA_I = 0x193B;
const char32_t LIMBU_SIGN_LOO = 0x1940;
const char32_t LIMBU_EXCLAMATION_MARK = 0x1944;
const char32_t LIMBU_QUESTION_MARK = 0x1945;
const char32_t LIMBU_DIGIT_ZERO = 0x1946;
const char32_t LIMBU_DIGIT_ONE = 0x1947;
const char32_t LIMBU_DIGIT_TWO = 0x1948;
const char32_t LIMBU_DIGIT_THREE = 0x1949;
const char32_t LIMBU_DIGIT_FOUR = 0x194A;
const char32_t LIMBU_DIGIT_FIVE = 0x194B;
const char32_t LIMBU_DIGIT_SIX = 0x194C;
const char32_t LIMBU_DIGIT_SEVEN = 0x194D;
const char32_t LIMBU_DIGIT_EIGHT = 0x194E;
const char32_t LIMBU_DIGIT_NINE = 0x194F;
const char32_t TAI_LE_LETTER_KA = 0x1950;
const char32_t TAI_LE_LETTER_XA = 0x1951;
const char32_t TAI_LE_LETTER_NGA = 0x1952;
const char32_t TAI_LE_LETTER_TSA = 0x1953;
const char32_t TAI_LE_LETTER_SA = 0x1954;
const char32_t TAI_LE_LETTER_YA = 0x1955;
const char32_t TAI_LE_LETTER_TA = 0x1956;
const char32_t TAI_LE_LETTER_THA = 0x1957;
const char32_t TAI_LE_LETTER_LA = 0x1958;
const char32_t TAI_LE_LETTER_PA = 0x1959;
const char32_t TAI_LE_LETTER_PHA = 0x195A;
const char32_t TAI_LE_LETTER_MA = 0x195B;
const char32_t TAI_LE_LETTER_FA = 0x195C;
const char32_t TAI_LE_LETTER_VA = 0x195D;
const char32_t TAI_LE_LETTER_HA = 0x195E;
const char32_t TAI_LE_LETTER_QA = 0x195F;
const char32_t TAI_LE_LETTER_KHA = 0x1960;
const char32_t TAI_LE_LETTER_TSHA = 0x1961;
const char32_t TAI_LE_LETTER_NA = 0x1962;
const char32_t TAI_LE_LETTER_A = 0x1963;
const char32_t TAI_LE_LETTER_I = 0x1964;
const char32_t TAI_LE_LETTER_EE = 0x1965;
const char32_t TAI_LE_LETTER_EH = 0x1966;
const char32_t TAI_LE_LETTER_U = 0x1967;
const char32_t TAI_LE_LETTER_OO = 0x1968;
const char32_t TAI_LE_LETTER_O = 0x1969;
const char32_t TAI_LE_LETTER_UE = 0x196A;
const char32_t TAI_LE_LETTER_E = 0x196B;
const char32_t TAI_LE_LETTER_AUE = 0x196C;
const char32_t TAI_LE_LETTER_AI = 0x196D;
const char32_t TAI_LE_LETTER_TONE_2 = 0x1970;
const char32_t TAI_LE_LETTER_TONE_3 = 0x1971;
const char32_t TAI_LE_LETTER_TONE_4 = 0x1972;
const char32_t TAI_LE_LETTER_TONE_5 = 0x1973;
const char32_t TAI_LE_LETTER_TONE_6 = 0x1974;
const char32_t NEW_TAI_LUE_LETTER_HIGH_QA = 0x1980;
const char32_t NEW_TAI_LUE_LETTER_LOW_QA = 0x1981;
const char32_t NEW_TAI_LUE_LETTER_HIGH_KA = 0x1982;
const char32_t NEW_TAI_LUE_LETTER_HIGH_XA = 0x1983;
const char32_t NEW_TAI_LUE_LETTER_HIGH_NGA = 0x1984;
const char32_t NEW_TAI_LUE_LETTER_LOW_KA = 0x1985;
const char32_t NEW_TAI_LUE_LETTER_LOW_XA = 0x1986;
const char32_t NEW_TAI_LUE_LETTER_LOW_NGA = 0x1987;
const char32_t NEW_TAI_LUE_LETTER_HIGH_TSA = 0x1988;
const char32_t NEW_TAI_LUE_LETTER_HIGH_SA = 0x1989;
const char32_t NEW_TAI_LUE_LETTER_HIGH_YA = 0x198A;
const char32_t NEW_TAI_LUE_LETTER_LOW_TSA = 0x198B;
const char32_t NEW_TAI_LUE_LETTER_LOW_SA = 0x198C;
const char32_t NEW_TAI_LUE_LETTER_LOW_YA = 0x198D;
const char32_t NEW_TAI_LUE_LETTER_HIGH_TA = 0x198E;
const char32_t NEW_TAI_LUE_LETTER_HIGH_THA = 0x198F;
const char32_t NEW_TAI_LUE_LETTER_HIGH_NA = 0x1990;
const char32_t NEW_TAI_LUE_LETTER_LOW_TA = 0x1991;
const char32_t NEW_TAI_LUE_LETTER_LOW_THA = 0x1992;
const char32_t NEW_TAI_LUE_LETTER_LOW_NA = 0x1993;
const char32_t NEW_TAI_LUE_LETTER_HIGH_PA = 0x1994;
const char32_t NEW_TAI_LUE_LETTER_HIGH_PHA = 0x1995;
const char32_t NEW_TAI_LUE_LETTER_HIGH_MA = 0x1996;
const char32_t NEW_TAI_LUE_LETTER_LOW_PA = 0x1997;
const char32_t NEW_TAI_LUE_LETTER_LOW_PHA = 0x1998;
const char32_t NEW_TAI_LUE_LETTER_LOW_MA = 0x1999;
const char32_t NEW_TAI_LUE_LETTER_HIGH_FA = 0x199A;
const char32_t NEW_TAI_LUE_LETTER_HIGH_VA = 0x199B;
const char32_t NEW_TAI_LUE_LETTER_HIGH_LA = 0x199C;
const char32_t NEW_TAI_LUE_LETTER_LOW_FA = 0x199D;
const char32_t NEW_TAI_LUE_LETTER_LOW_VA = 0x199E;
const char32_t NEW_TAI_LUE_LETTER_LOW_LA = 0x199F;
const char32_t NEW_TAI_LUE_LETTER_HIGH_HA = 0x19A0;
const char32_t NEW_TAI_LUE_LETTER_HIGH_DA = 0x19A1;
const char32_t NEW_TAI_LUE_LETTER_HIGH_BA = 0x19A2;
const char32_t NEW_TAI_LUE_LETTER_LOW_HA = 0x19A3;
const char32_t NEW_TAI_LUE_LETTER_LOW_DA = 0x19A4;
const char32_t NEW_TAI_LUE_LETTER_LOW_BA = 0x19A5;
const char32_t NEW_TAI_LUE_LETTER_HIGH_KVA = 0x19A6;
const char32_t NEW_TAI_LUE_LETTER_HIGH_XVA = 0x19A7;
const char32_t NEW_TAI_LUE_LETTER_LOW_KVA = 0x19A8;
const char32_t NEW_TAI_LUE_LETTER_LOW_XVA = 0x19A9;
const char32_t NEW_TAI_LUE_LETTER_HIGH_SUA = 0x19AA;
const char32_t NEW_TAI_LUE_LETTER_LOW_SUA = 0x19AB;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_VOWEL_SHORTENER = 0x19B0;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_AA = 0x19B1;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_II = 0x19B2;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_U = 0x19B3;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_UU = 0x19B4;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_E = 0x19B5;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_AE = 0x19B6;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_O = 0x19B7;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_OA = 0x19B8;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_UE = 0x19B9;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_AY = 0x19BA;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_AAY = 0x19BB;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_UY = 0x19BC;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_OY = 0x19BD;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_OAY = 0x19BE;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_UEY = 0x19BF;
const char32_t NEW_TAI_LUE_VOWEL_SIGN_IY = 0x19C0;
const char32_t NEW_TAI_LUE_LETTER_FINAL_V = 0x19C1;
const char32_t NEW_TAI_LUE_LETTER_FINAL_NG = 0x19C2;
const char32_t NEW_TAI_LUE_LETTER_FINAL_N = 0x19C3;
const char32_t NEW_TAI_LUE_LETTER_FINAL_M = 0x19C4;
const char32_t NEW_TAI_LUE_LETTER_FINAL_K = 0x19C5;
const char32_t NEW_TAI_LUE_LETTER_FINAL_D = 0x19C6;
const char32_t NEW_TAI_LUE_LETTER_FINAL_B = 0x19C7;
const char32_t NEW_TAI_LUE_TONE_MARK_1 = 0x19C8;
const char32_t NEW_TAI_LUE_TONE_MARK_2 = 0x19C9;
const char32_t NEW_TAI_LUE_DIGIT_ZERO = 0x19D0;
const char32_t NEW_TAI_LUE_DIGIT_ONE = 0x19D1;
const char32_t NEW_TAI_LUE_DIGIT_TWO = 0x19D2;
const char32_t NEW_TAI_LUE_DIGIT_THREE = 0x19D3;
const char32_t NEW_TAI_LUE_DIGIT_FOUR = 0x19D4;
const char32_t NEW_TAI_LUE_DIGIT_FIVE = 0x19D5;
const char32_t NEW_TAI_LUE_DIGIT_SIX = 0x19D6;
const char32_t NEW_TAI_LUE_DIGIT_SEVEN = 0x19D7;
const char32_t NEW_TAI_LUE_DIGIT_EIGHT = 0x19D8;
const char32_t NEW_TAI_LUE_DIGIT_NINE = 0x19D9;
const char32_t NEW_TAI_LUE_THAM_DIGIT_ONE = 0x19DA;
const char32_t NEW_TAI_LUE_SIGN_LAE = 0x19DE;
const char32_t NEW_TAI_LUE_SIGN_LAEV = 0x19DF;
const char32_t KHMER_SYMBOL_PATHAMASAT = 0x19E0;
const char32_t KHMER_SYMBOL_MUOY_KOET = 0x19E1;
const char32_t KHMER_SYMBOL_PII_KOET = 0x19E2;
const char32_t KHMER_SYMBOL_BEI_KOET = 0x19E3;
const char32_t KHMER_SYMBOL_BUON_KOET = 0x19E4;
const char32_t KHMER_SYMBOL_PRAM_KOET = 0x19E5;
const char32_t KHMER_SYMBOL_PRAM_MUOY_KOET = 0x19E6;
const char32_t KHMER_SYMBOL_PRAM_PII_KOET = 0x19E7;
const char32_t KHMER_SYMBOL_PRAM_BEI_KOET = 0x19E8;
const char32_t KHMER_SYMBOL_PRAM_BUON_KOET = 0x19E9;
const char32_t KHMER_SYMBOL_DAP_KOET = 0x19EA;
const char32_t KHMER_SYMBOL_DAP_MUOY_KOET = 0x19EB;
const char32_t KHMER_SYMBOL_DAP_PII_KOET = 0x19EC;
const char32_t KHMER_SYMBOL_DAP_BEI_KOET = 0x19ED;
const char32_t KHMER_SYMBOL_DAP_BUON_KOET = 0x19EE;
const char32_t KHMER_SYMBOL_DAP_PRAM_KOET = 0x19EF;
const char32_t KHMER_SYMBOL_TUTEYASAT = 0x19F0;
const char32_t KHMER_SYMBOL_MUOY_ROC = 0x19F1;
const char32_t KHMER_SYMBOL_PII_ROC = 0x19F2;
const char32_t KHMER_SYMBOL_BEI_ROC = 0x19F3;
const char32_t KHMER_SYMBOL_BUON_ROC = 0x19F4;
const char32_t KHMER_SYMBOL_PRAM_ROC = 0x19F5;
const char32_t KHMER_SYMBOL_PRAM_MUOY_ROC = 0x19F6;
const char32_t KHMER_SYMBOL_PRAM_PII_ROC = 0x19F7;
const char32_t KHMER_SYMBOL_PRAM_BEI_ROC = 0x19F8;
const char32_t KHMER_SYMBOL_PRAM_BUON_ROC = 0x19F9;
const char32_t KHMER_SYMBOL_DAP_ROC = 0x19FA;
const char32_t KHMER_SYMBOL_DAP_MUOY_ROC = 0x19FB;
const char32_t KHMER_SYMBOL_DAP_PII_ROC = 0x19FC;
const char32_t KHMER_SYMBOL_DAP_BEI_ROC = 0x19FD;
const char32_t KHMER_SYMBOL_DAP_BUON_ROC = 0x19FE;
const char32_t KHMER_SYMBOL_DAP_PRAM_ROC = 0x19FF;
const char32_t BUGINESE_LETTER_KA = 0x1A00;
const char32_t BUGINESE_LETTER_GA = 0x1A01;
const char32_t BUGINESE_LETTER_NGA = 0x1A02;
const char32_t BUGINESE_LETTER_NGKA = 0x1A03;
const char32_t BUGINESE_LETTER_PA = 0x1A04;
const char32_t BUGINESE_LETTER_BA = 0x1A05;
const char32_t BUGINESE_LETTER_MA = 0x1A06;
const char32_t BUGINESE_LETTER_MPA = 0x1A07;
const char32_t BUGINESE_LETTER_TA = 0x1A08;
const char32_t BUGINESE_LETTER_DA = 0x1A09;
const char32_t BUGINESE_LETTER_NA = 0x1A0A;
const char32_t BUGINESE_LETTER_NRA = 0x1A0B;
const char32_t BUGINESE_LETTER_CA = 0x1A0C;
const char32_t BUGINESE_LETTER_JA = 0x1A0D;
const char32_t BUGINESE_LETTER_NYA = 0x1A0E;
const char32_t BUGINESE_LETTER_NYCA = 0x1A0F;
const char32_t BUGINESE_LETTER_YA = 0x1A10;
const char32_t BUGINESE_LETTER_RA = 0x1A11;
const char32_t BUGINESE_LETTER_LA = 0x1A12;
const char32_t BUGINESE_LETTER_VA = 0x1A13;
const char32_t BUGINESE_LETTER_SA = 0x1A14;
const char32_t BUGINESE_LETTER_A = 0x1A15;
const char32_t BUGINESE_LETTER_HA = 0x1A16;
const char32_t BUGINESE_VOWEL_SIGN_I = 0x1A17;
const char32_t BUGINESE_VOWEL_SIGN_U = 0x1A18;
const char32_t BUGINESE_VOWEL_SIGN_E = 0x1A19;
const char32_t BUGINESE_VOWEL_SIGN_O = 0x1A1A;
const char32_t BUGINESE_VOWEL_SIGN_AE = 0x1A1B;
const char32_t BUGINESE_PALLAWA = 0x1A1E;
const char32_t BUGINESE_END_OF_SECTION = 0x1A1F;
const char32_t TAI_THAM_LETTER_HIGH_KA = 0x1A20;
const char32_t TAI_THAM_LETTER_HIGH_KHA = 0x1A21;
const char32_t TAI_THAM_LETTER_HIGH_KXA = 0x1A22;
const char32_t TAI_THAM_LETTER_LOW_KA = 0x1A23;
const char32_t TAI_THAM_LETTER_LOW_KXA = 0x1A24;
const char32_t TAI_THAM_LETTER_LOW_KHA = 0x1A25;
const char32_t TAI_THAM_LETTER_NGA = 0x1A26;
const char32_t TAI_THAM_LETTER_HIGH_CA = 0x1A27;
const char32_t TAI_THAM_LETTER_HIGH_CHA = 0x1A28;
const char32_t TAI_THAM_LETTER_LOW_CA = 0x1A29;
const char32_t TAI_THAM_LETTER_LOW_SA = 0x1A2A;
const char32_t TAI_THAM_LETTER_LOW_CHA = 0x1A2B;
const char32_t TAI_THAM_LETTER_NYA = 0x1A2C;
const char32_t TAI_THAM_LETTER_RATA = 0x1A2D;
const char32_t TAI_THAM_LETTER_HIGH_RATHA = 0x1A2E;
const char32_t TAI_THAM_LETTER_DA = 0x1A2F;
const char32_t TAI_THAM_LETTER_LOW_RATHA = 0x1A30;
const char32_t TAI_THAM_LETTER_RANA = 0x1A31;
const char32_t TAI_THAM_LETTER_HIGH_TA = 0x1A32;
const char32_t TAI_THAM_LETTER_HIGH_THA = 0x1A33;
const char32_t TAI_THAM_LETTER_LOW_TA = 0x1A34;
const char32_t TAI_THAM_LETTER_LOW_THA = 0x1A35;
const char32_t TAI_THAM_LETTER_NA = 0x1A36;
const char32_t TAI_THAM_LETTER_BA = 0x1A37;
const char32_t TAI_THAM_LETTER_HIGH_PA = 0x1A38;
const char32_t TAI_THAM_LETTER_HIGH_PHA = 0x1A39;
const char32_t TAI_THAM_LETTER_HIGH_FA = 0x1A3A;
const char32_t TAI_THAM_LETTER_LOW_PA = 0x1A3B;
const char32_t TAI_THAM_LETTER_LOW_FA = 0x1A3C;
const char32_t TAI_THAM_LETTER_LOW_PHA = 0x1A3D;
const char32_t TAI_THAM_LETTER_MA = 0x1A3E;
const char32_t TAI_THAM_LETTER_LOW_YA = 0x1A3F;
const char32_t TAI_THAM_LETTER_HIGH_YA = 0x1A40;
const char32_t TAI_THAM_LETTER_RA = 0x1A41;
const char32_t TAI_THAM_LETTER_RUE = 0x1A42;
const char32_t TAI_THAM_LETTER_LA = 0x1A43;
const char32_t TAI_THAM_LETTER_LUE = 0x1A44;
const char32_t TAI_THAM_LETTER_WA = 0x1A45;
const char32_t TAI_THAM_LETTER_HIGH_SHA = 0x1A46;
const char32_t TAI_THAM_LETTER_HIGH_SSA = 0x1A47;
const char32_t TAI_THAM_LETTER_HIGH_SA = 0x1A48;
const char32_t TAI_THAM_LETTER_HIGH_HA = 0x1A49;
const char32_t TAI_THAM_LETTER_LLA = 0x1A4A;
const char32_t TAI_THAM_LETTER_A = 0x1A4B;
const char32_t TAI_THAM_LETTER_LOW_HA = 0x1A4C;
const char32_t TAI_THAM_LETTER_I = 0x1A4D;
const char32_t TAI_THAM_LETTER_II = 0x1A4E;
const char32_t TAI_THAM_LETTER_U = 0x1A4F;
const char32_t TAI_THAM_LETTER_UU = 0x1A50;
const char32_t TAI_THAM_LETTER_EE = 0x1A51;
const char32_t TAI_THAM_LETTER_OO = 0x1A52;
const char32_t TAI_THAM_LETTER_LAE = 0x1A53;
const char32_t TAI_THAM_LETTER_GREAT_SA = 0x1A54;
const char32_t TAI_THAM_CONSONANT_SIGN_MEDIAL_RA = 0x1A55;
const char32_t TAI_THAM_CONSONANT_SIGN_MEDIAL_LA = 0x1A56;
const char32_t TAI_THAM_CONSONANT_SIGN_LA_TANG_LAI = 0x1A57;
const char32_t TAI_THAM_SIGN_MAI_KANG_LAI = 0x1A58;
const char32_t TAI_THAM_CONSONANT_SIGN_FINAL_NGA = 0x1A59;
const char32_t TAI_THAM_CONSONANT_SIGN_LOW_PA = 0x1A5A;
const char32_t TAI_THAM_CONSONANT_SIGN_HIGH_RATHA_OR_LOW_PA = 0x1A5B;
const char32_t TAI_THAM_CONSONANT_SIGN_MA = 0x1A5C;
const char32_t TAI_THAM_CONSONANT_SIGN_BA = 0x1A5D;
const char32_t TAI_THAM_CONSONANT_SIGN_SA = 0x1A5E;
const char32_t TAI_THAM_SIGN_SAKOT = 0x1A60;
const char32_t TAI_THAM_VOWEL_SIGN_A = 0x1A61;
const char32_t TAI_THAM_VOWEL_SIGN_MAI_SAT = 0x1A62;
const char32_t TAI_THAM_VOWEL_SIGN_AA = 0x1A63;
const char32_t TAI_THAM_VOWEL_SIGN_TALL_AA = 0x1A64;
const char32_t TAI_THAM_VOWEL_SIGN_I = 0x1A65;
const char32_t TAI_THAM_VOWEL_SIGN_II = 0x1A66;
const char32_t TAI_THAM_VOWEL_SIGN_UE = 0x1A67;
const char32_t TAI_THAM_VOWEL_SIGN_UUE = 0x1A68;
const char32_t TAI_THAM_VOWEL_SIGN_U = 0x1A69;
const char32_t TAI_THAM_VOWEL_SIGN_UU = 0x1A6A;
const char32_t TAI_THAM_VOWEL_SIGN_O = 0x1A6B;
const char32_t TAI_THAM_VOWEL_SIGN_OA_BELOW = 0x1A6C;
const char32_t TAI_THAM_VOWEL_SIGN_OY = 0x1A6D;
const char32_t TAI_THAM_VOWEL_SIGN_E = 0x1A6E;
const char32_t TAI_THAM_VOWEL_SIGN_AE = 0x1A6F;
const char32_t TAI_THAM_VOWEL_SIGN_OO = 0x1A70;
const char32_t TAI_THAM_VOWEL_SIGN_AI = 0x1A71;
const char32_t TAI_THAM_VOWEL_SIGN_THAM_AI = 0x1A72;
const char32_t TAI_THAM_VOWEL_SIGN_OA_ABOVE = 0x1A73;
const char32_t TAI_THAM_SIGN_MAI_KANG = 0x1A74;
const char32_t TAI_THAM_SIGN_TONE_1 = 0x1A75;
const char32_t TAI_THAM_SIGN_TONE_2 = 0x1A76;
const char32_t TAI_THAM_SIGN_KHUEN_TONE_3 = 0x1A77;
const char32_t TAI_THAM_SIGN_KHUEN_TONE_4 = 0x1A78;
const char32_t TAI_THAM_SIGN_KHUEN_TONE_5 = 0x1A79;
const char32_t TAI_THAM_SIGN_RA_HAAM = 0x1A7A;
const char32_t TAI_THAM_SIGN_MAI_SAM = 0x1A7B;
const char32_t TAI_THAM_SIGN_KHUEN_LUE_KARAN = 0x1A7C;
const char32_t TAI_THAM_COMBINING_CRYPTOGRAMMIC_DOT = 0x1A7F;
const char32_t TAI_THAM_HORA_DIGIT_ZERO = 0x1A80;
const char32_t TAI_THAM_HORA_DIGIT_ONE = 0x1A81;
const char32_t TAI_THAM_HORA_DIGIT_TWO = 0x1A82;
const char32_t TAI_THAM_HORA_DIGIT_THREE = 0x1A83;
const char32_t TAI_THAM_HORA_DIGIT_FOUR = 0x1A84;
const char32_t TAI_THAM_HORA_DIGIT_FIVE = 0x1A85;
const char32_t TAI_THAM_HORA_DIGIT_SIX = 0x1A86;
const char32_t TAI_THAM_HORA_DIGIT_SEVEN = 0x1A87;
const char32_t TAI_THAM_HORA_DIGIT_EIGHT = 0x1A88;
const char32_t TAI_THAM_HORA_DIGIT_NINE = 0x1A89;
const char32_t TAI_THAM_THAM_DIGIT_ZERO = 0x1A90;
const char32_t TAI_THAM_THAM_DIGIT_ONE = 0x1A91;
const char32_t TAI_THAM_THAM_DIGIT_TWO = 0x1A92;
const char32_t TAI_THAM_THAM_DIGIT_THREE = 0x1A93;
const char32_t TAI_THAM_THAM_DIGIT_FOUR = 0x1A94;
const char32_t TAI_THAM_THAM_DIGIT_FIVE = 0x1A95;
const char32_t TAI_THAM_THAM_DIGIT_SIX = 0x1A96;
const char32_t TAI_THAM_THAM_DIGIT_SEVEN = 0x1A97;
const char32_t TAI_THAM_THAM_DIGIT_EIGHT = 0x1A98;
const char32_t TAI_THAM_THAM_DIGIT_NINE = 0x1A99;
const char32_t TAI_THAM_SIGN_WIANG = 0x1AA0;
const char32_t TAI_THAM_SIGN_WIANGWAAK = 0x1AA1;
const char32_t TAI_THAM_SIGN_SAWAN = 0x1AA2;
const char32_t TAI_THAM_SIGN_KEOW = 0x1AA3;
const char32_t TAI_THAM_SIGN_HOY = 0x1AA4;
const char32_t TAI_THAM_SIGN_DOKMAI = 0x1AA5;
const char32_t TAI_THAM_SIGN_REVERSED_ROTATED_RANA = 0x1AA6;
const char32_t TAI_THAM_SIGN_MAI_YAMOK = 0x1AA7;
const char32_t TAI_THAM_SIGN_KAAN = 0x1AA8;
const char32_t TAI_THAM_SIGN_KAANKUU = 0x1AA9;
const char32_t TAI_THAM_SIGN_SATKAAN = 0x1AAA;
const char32_t TAI_THAM_SIGN_SATKAANKUU = 0x1AAB;
const char32_t TAI_THAM_SIGN_HANG = 0x1AAC;
const char32_t TAI_THAM_SIGN_CAANG = 0x1AAD;
const char32_t COMBINING_DOUBLED_CIRCUMFLEX_ACCENT = 0x1AB0;
const char32_t COMBINING_DIAERESIS_RING = 0x1AB1;
const char32_t COMBINING_INFINITY = 0x1AB2;
const char32_t COMBINING_DOWNWARDS_ARROW = 0x1AB3;
const char32_t COMBINING_TRIPLE_DOT = 0x1AB4;
const char32_t COMBINING_X_X_BELOW = 0x1AB5;
const char32_t COMBINING_WIGGLY_LINE_BELOW = 0x1AB6;
const char32_t COMBINING_OPEN_MARK_BELOW = 0x1AB7;
const char32_t COMBINING_DOUBLE_OPEN_MARK_BELOW = 0x1AB8;
const char32_t COMBINING_LIGHT_CENTRALIZATION_STROKE_BELOW = 0x1AB9;
const char32_t COMBINING_STRONG_CENTRALIZATION_STROKE_BELOW = 0x1ABA;
const char32_t COMBINING_PARENTHESES_ABOVE = 0x1ABB;
const char32_t COMBINING_DOUBLE_PARENTHESES_ABOVE = 0x1ABC;
const char32_t COMBINING_PARENTHESES_BELOW = 0x1ABD;
const char32_t COMBINING_PARENTHESES_OVERLAY = 0x1ABE;
const char32_t COMBINING_LATIN_SMALL_LETTER_W_BELOW = 0x1ABF;
const char32_t COMBINING_LATIN_SMALL_LETTER_TURNED_W_BELOW = 0x1AC0;
const char32_t COMBINING_LEFT_PARENTHESIS_ABOVE_LEFT = 0x1AC1;
const char32_t COMBINING_RIGHT_PARENTHESIS_ABOVE_RIGHT = 0x1AC2;
const char32_t COMBINING_LEFT_PARENTHESIS_BELOW_LEFT = 0x1AC3;
const char32_t COMBINING_RIGHT_PARENTHESIS_BELOW_RIGHT = 0x1AC4;
const char32_t COMBINING_SQUARE_BRACKETS_ABOVE = 0x1AC5;
const char32_t COMBINING_NUMBER_SIGN_ABOVE = 0x1AC6;
const char32_t COMBINING_INVERTED_DOUBLE_ARCH_ABOVE = 0x1AC7;
const char32_t COMBINING_PLUS_SIGN_ABOVE = 0x1AC8;
const char32_t COMBINING_DOUBLE_PLUS_SIGN_ABOVE = 0x1AC9;
const char32_t COMBINING_DOUBLE_PLUS_SIGN_BELOW = 0x1ACA;
const char32_t COMBINING_TRIPLE_ACUTE_ACCENT = 0x1ACB;
const char32_t COMBINING_LATIN_SMALL_LETTER_INSULAR_G = 0x1ACC;
const char32_t COMBINING_LATIN_SMALL_LETTER_INSULAR_R = 0x1ACD;
const char32_t COMBINING_LATIN_SMALL_LETTER_INSULAR_T = 0x1ACE;
const char32_t COMBINING_DOUBLE_CARON = 0x1ACF;
const char32_t COMBINING_VERTICAL_LINE_ACUTE = 0x1AD0;
const char32_t COMBINING_GRAVE_VERTICAL_LINE = 0x1AD1;
const char32_t COMBINING_VERTICAL_LINE_GRAVE = 0x1AD2;
const char32_t COMBINING_ACUTE_VERTICAL_LINE = 0x1AD3;
const char32_t COMBINING_VERTICAL_LINE_MACRON = 0x1AD4;
const char32_t COMBINING_MACRON_VERTICAL_LINE = 0x1AD5;
const char32_t COMBINING_VERTICAL_LINE_ACUTE_GRAVE = 0x1AD6;
const char32_t COMBINING_VERTICAL_LINE_GRAVE_ACUTE = 0x1AD7;
const char32_t COMBINING_MACRON_ACUTE_GRAVE = 0x1AD8;
const char32_t COMBINING_SHARP_SIGN = 0x1AD9;
const char32_t COMBINING_FLAT_SIGN = 0x1ADA;
const char32_t COMBINING_DOWN_TACK_ABOVE = 0x1ADB;
const char32_t COMBINING_DIAERESIS_WITH_RAISED_LEFT_DOT = 0x1ADC;
const char32_t COMBINING_DOT_AND_RING_BELOW = 0x1ADD;
const char32_t COMBINING_LEFT_TACK_ABOVE = 0x1AE0;
const char32_t COMBINING_RIGHT_TACK_ABOVE = 0x1AE1;
const char32_t COMBINING_MINUS_SIGN_ABOVE = 0x1AE2;
const char32_t COMBINING_INVERTED_BRIDGE_ABOVE = 0x1AE3;
const char32_t COMBINING_SQUARE_ABOVE = 0x1AE4;
const char32_t COMBINING_SEAGULL_ABOVE = 0x1AE5;
const char32_t COMBINING_DOUBLE_ARCH_BELOW = 0x1AE6;
const char32_t COMBINING_DOUBLE_ARCH_ABOVE = 0x1AE7;
const char32_t COMBINING_EQUALS_SIGN_ABOVE = 0x1AE8;
const char32_t COMBINING_LEFT_ANGLE_CENTRED_ABOVE = 0x1AE9;
const char32_t COMBINING_UPWARDS_ARROW_ABOVE = 0x1AEA;
const char32_t COMBINING_DOUBLE_RIGHTWARDS_ARROW_ABOVE = 0x1AEB;
const char32_t BALINESE_SIGN_ULU_RICEM = 0x1B00;
const char32_t BALINESE_SIGN_ULU_CANDRA = 0x1B01;
const char32_t BALINESE_SIGN_CECEK = 0x1B02;
const char32_t BALINESE_SIGN_SURANG = 0x1B03;
const char32_t BALINESE_SIGN_BISAH = 0x1B04;
const char32_t BALINESE_LETTER_AKARA = 0x1B05;
const char32_t BALINESE_LETTER_AKARA_TEDUNG = 0x1B06;
const char32_t BALINESE_LETTER_IKARA = 0x1B07;
const char32_t BALINESE_LETTER_IKARA_TEDUNG = 0x1B08;
const char32_t BALINESE_LETTER_UKARA = 0x1B09;
const char32_t BALINESE_LETTER_UKARA_TEDUNG = 0x1B0A;
const char32_t BALINESE_LETTER_RA_REPA = 0x1B0B;
const char32_t BALINESE_LETTER_RA_REPA_TEDUNG = 0x1B0C;
const char32_t BALINESE_LETTER_LA_LENGA = 0x1B0D;
const char32_t BALINESE_LETTER_LA_LENGA_TEDUNG = 0x1B0E;
const char32_t BALINESE_LETTER_EKARA = 0x1B0F;
const char32_t BALINESE_LETTER_AIKARA = 0x1B10;
const char32_t BALINESE_LETTER_OKARA = 0x1B11;
const char32_t BALINESE_LETTER_OKARA_TEDUNG = 0x1B12;
const char32_t BALINESE_LETTER_KA = 0x1B13;
const char32_t BALINESE_LETTER_KA_MAHAPRANA = 0x1B14;
const char32_t BALINESE_LETTER_GA = 0x1B15;
const char32_t BALINESE_LETTER_GA_GORA = 0x1B16;
const char32_t BALINESE_LETTER_NGA = 0x1B17;
const char32_t BALINESE_LETTER_CA = 0x1B18;
const char32_t BALINESE_LETTER_CA_LACA = 0x1B19;
const char32_t BALINESE_LETTER_JA = 0x1B1A;
const char32_t BALINESE_LETTER_JA_JERA = 0x1B1B;
const char32_t BALINESE_LETTER_NYA = 0x1B1C;
const char32_t BALINESE_LETTER_TA_LATIK = 0x1B1D;
const char32_t BALINESE_LETTER_TA_MURDA_MAHAPRANA = 0x1B1E;
const char32_t BALINESE_LETTER_DA_MURDA_ALPAPRANA = 0x1B1F;
const char32_t BALINESE_LETTER_DA_MURDA_MAHAPRANA = 0x1B20;
const char32_t BALINESE_LETTER_NA_RAMBAT = 0x1B21;
const char32_t BALINESE_LETTER_TA = 0x1B22;
const char32_t BALINESE_LETTER_TA_TAWA = 0x1B23;
const char32_t BALINESE_LETTER_DA = 0x1B24;
const char32_t BALINESE_LETTER_DA_MADU = 0x1B25;
const char32_t BALINESE_LETTER_NA = 0x1B26;
const char32_t BALINESE_LETTER_PA = 0x1B27;
const char32_t BALINESE_LETTER_PA_KAPAL = 0x1B28;
const char32_t BALINESE_LETTER_BA = 0x1B29;
const char32_t BALINESE_LETTER_BA_KEMBANG = 0x1B2A;
const char32_t BALINESE_LETTER_MA = 0x1B2B;
const char32_t BALINESE_LETTER_YA = 0x1B2C;
const char32_t BALINESE_LETTER_RA = 0x1B2D;
const char32_t BALINESE_LETTER_LA = 0x1B2E;
const char32_t BALINESE_LETTER_WA = 0x1B2F;
const char32_t BALINESE_LETTER_SA_SAGA = 0x1B30;
const char32_t BALINESE_LETTER_SA_SAPA = 0x1B31;
const char32_t BALINESE_LETTER_SA = 0x1B32;
const char32_t BALINESE_LETTER_HA = 0x1B33;
const char32_t BALINESE_SIGN_REREKAN = 0x1B34;
const char32_t BALINESE_VOWEL_SIGN_TEDUNG = 0x1B35;
const char32_t BALINESE_VOWEL_SIGN_ULU = 0x1B36;
const char32_t BALINESE_VOWEL_SIGN_ULU_SARI = 0x1B37;
const char32_t BALINESE_VOWEL_SIGN_SUKU = 0x1B38;
const char32_t BALINESE_VOWEL_SIGN_SUKU_ILUT = 0x1B39;
const char32_t BALINESE_VOWEL_SIGN_RA_REPA = 0x1B3A;
const char32_t BALINESE_VOWEL_SIGN_RA_REPA_TEDUNG = 0x1B3B;
const char32_t BALINESE_VOWEL_SIGN_LA_LENGA = 0x1B3C;
const char32_t BALINESE_VOWEL_SIGN_LA_LENGA_TEDUNG = 0x1B3D;
const char32_t BALINESE_VOWEL_SIGN_TALING = 0x1B3E;
const char32_t BALINESE_VOWEL_SIGN_TALING_REPA = 0x1B3F;
const char32_t BALINESE_VOWEL_SIGN_TALING_TEDUNG = 0x1B40;
const char32_t BALINESE_VOWEL_SIGN_TALING_REPA_TEDUNG = 0x1B41;
const char32_t BALINESE_VOWEL_SIGN_PEPET = 0x1B42;
const char32_t BALINESE_VOWEL_SIGN_PEPET_TEDUNG = 0x1B43;
const char32_t BALINESE_ADEG_ADEG = 0x1B44;
const char32_t BALINESE_LETTER_KAF_SASAK = 0x1B45;
const char32_t BALINESE_LETTER_KHOT_SASAK = 0x1B46;
const char32_t BALINESE_LETTER_TZIR_SASAK = 0x1B47;
const char32_t BALINESE_LETTER_EF_SASAK = 0x1B48;
const char32_t BALINESE_LETTER_VE_SASAK = 0x1B49;
const char32_t BALINESE_LETTER_ZAL_SASAK = 0x1B4A;
const char32_t BALINESE_LETTER_ASYURA_SASAK = 0x1B4B;
const char32_t BALINESE_LETTER_ARCHAIC_JNYA = 0x1B4C;
const char32_t BALINESE_INVERTED_CARIK_SIKI = 0x1B4E;
const char32_t BALINESE_INVERTED_CARIK_PAREREN = 0x1B4F;
const char32_t BALINESE_DIGIT_ZERO = 0x1B50;
const char32_t BALINESE_DIGIT_ONE = 0x1B51;
const char32_t BALINESE_DIGIT_TWO = 0x1B52;
const char32_t BALINESE_DIGIT_THREE = 0x1B53;
const char32_t BALINESE_DIGIT_FOUR = 0x1B54;
const char32_t BALINESE_DIGIT_FIVE = 0x1B55;
const char32_t BALINESE_DIGIT_SIX = 0x1B56;
const char32_t BALINESE_DIGIT_SEVEN = 0x1B57;
const char32_t BALINESE_DIGIT_EIGHT = 0x1B58;
const char32_t BALINESE_DIGIT_NINE = 0x1B59;
const char32_t BALINESE_PANTI = 0x1B5A;
const char32_t BALINESE_PAMADA = 0x1B5B;
const char32_t BALINESE_WINDU = 0x1B5C;
const char32_t BALINESE_CARIK_PAMUNGKAH = 0x1B5D;
const char32_t BALINESE_CARIK_SIKI = 0x1B5E;
const char32_t BALINESE_CARIK_PAREREN = 0x1B5F;
const char32_t BALINESE_PAMENENG = 0x1B60;
const char32_t BALINESE_MUSICAL_SYMBOL_DONG = 0x1B61;
const char32_t BALINESE_MUSICAL_SYMBOL_DENG = 0x1B62;
const char32_t BALINESE_MUSICAL_SYMBOL_DUNG = 0x1B63;
const char32_t BALINESE_MUSICAL_SYMBOL_DANG = 0x1B64;
const char32_t BALINESE_MUSICAL_SYMBOL_DANG_SURANG = 0x1B65;
const char32_t BALINESE_MUSICAL_SYMBOL_DING = 0x1B66;
const char32_t BALINESE_MUSICAL_SYMBOL_DAENG = 0x1B67;
const char32_t BALINESE_MUSICAL_SYMBOL_DEUNG = 0x1B68;
const char32_t BALINESE_MUSICAL_SYMBOL_DAING = 0x1B69;
const char32_t BALINESE_MUSICAL_SYMBOL_DANG_GEDE = 0x1B6A;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_TEGEH = 0x1B6B;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_ENDEP = 0x1B6C;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_KEMPUL = 0x1B6D;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_KEMPLI = 0x1B6E;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_JEGOGAN = 0x1B6F;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_KEMPUL_WITH_JEGOGAN = 0x1B70;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_KEMPLI_WITH_JEGOGAN = 0x1B71;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_BENDE = 0x1B72;
const char32_t BALINESE_MUSICAL_SYMBOL_COMBINING_GONG = 0x1B73;
const char32_t BALINESE_MUSICAL_SYMBOL_RIGHT_HAND_OPEN_DUG = 0x1B74;
const char32_t BALINESE_MUSICAL_SYMBOL_RIGHT_HAND_OPEN_DAG = 0x1B75;
const char32_t BALINESE_MUSICAL_SYMBOL_RIGHT_HAND_CLOSED_TUK = 0x1B76;
const char32_t BALINESE_MUSICAL_SYMBOL_RIGHT_HAND_CLOSED_TAK = 0x1B77;
const char32_t BALINESE_MUSICAL_SYMBOL_LEFT_HAND_OPEN_PANG = 0x1B78;
const char32_t BALINESE_MUSICAL_SYMBOL_LEFT_HAND_OPEN_PUNG = 0x1B79;
const char32_t BALINESE_MUSICAL_SYMBOL_LEFT_HAND_CLOSED_PLAK = 0x1B7A;
const char32_t BALINESE_MUSICAL_SYMBOL_LEFT_HAND_CLOSED_PLUK = 0x1B7B;
const char32_t BALINESE_MUSICAL_SYMBOL_LEFT_HAND_OPEN_PING = 0x1B7C;
const char32_t BALINESE_PANTI_LANTANG = 0x1B7D;
const char32_t BALINESE_PAMADA_LANTANG = 0x1B7E;
const char32_t BALINESE_PANTI_BAWAK = 0x1B7F;
const char32_t SUNDANESE_SIGN_PANYECEK = 0x1B80;
const char32_t SUNDANESE_SIGN_PANGLAYAR = 0x1B81;
const char32_t SUNDANESE_SIGN_PANGWISAD = 0x1B82;
const char32_t SUNDANESE_LETTER_A = 0x1B83;
const char32_t SUNDANESE_LETTER_I = 0x1B84;
const char32_t SUNDANESE_LETTER_U = 0x1B85;
const char32_t SUNDANESE_LETTER_AE = 0x1B86;
const char32_t SUNDANESE_LETTER_O = 0x1B87;
const char32_t SUNDANESE_LETTER_E = 0x1B88;
const char32_t SUNDANESE_LETTER_EU = 0x1B89;
const char32_t SUNDANESE_LETTER_KA = 0x1B8A;
const char32_t SUNDANESE_LETTER_QA = 0x1B8B;
const char32_t SUNDANESE_LETTER_GA = 0x1B8C;
const char32_t SUNDANESE_LETTER_NGA = 0x1B8D;
const char32_t SUNDANESE_LETTER_CA = 0x1B8E;
const char32_t SUNDANESE_LETTER_JA = 0x1B8F;
const char32_t SUNDANESE_LETTER_ZA = 0x1B90;
const char32_t SUNDANESE_LETTER_NYA = 0x1B91;
const char32_t SUNDANESE_LETTER_TA = 0x1B92;
const char32_t SUNDANESE_LETTER_DA = 0x1B93;
const char32_t SUNDANESE_LETTER_NA = 0x1B94;
const char32_t SUNDANESE_LETTER_PA = 0x1B95;
const char32_t SUNDANESE_LETTER_FA = 0x1B96;
const char32_t SUNDANESE_LETTER_VA = 0x1B97;
const char32_t SUNDANESE_LETTER_BA = 0x1B98;
const char32_t SUNDANESE_LETTER_MA = 0x1B99;
const char32_t SUNDANESE_LETTER_YA = 0x1B9A;
const char32_t SUNDANESE_LETTER_RA = 0x1B9B;
const char32_t SUNDANESE_LETTER_LA = 0x1B9C;
const char32_t SUNDANESE_LETTER_WA = 0x1B9D;
const char32_t SUNDANESE_LETTER_SA = 0x1B9E;
const char32_t SUNDANESE_LETTER_XA = 0x1B9F;
const char32_t SUNDANESE_LETTER_HA = 0x1BA0;
const char32_t SUNDANESE_CONSONANT_SIGN_PAMINGKAL = 0x1BA1;
const char32_t SUNDANESE_CONSONANT_SIGN_PANYAKRA = 0x1BA2;
const char32_t SUNDANESE_CONSONANT_SIGN_PANYIKU = 0x1BA3;
const char32_t SUNDANESE_VOWEL_SIGN_PANGHULU = 0x1BA4;
const char32_t SUNDANESE_VOWEL_SIGN_PANYUKU = 0x1BA5;
const char32_t SUNDANESE_VOWEL_SIGN_PANAELAENG = 0x1BA6;
const char32_t SUNDANESE_VOWEL_SIGN_PANOLONG = 0x1BA7;
const char32_t SUNDANESE_VOWEL_SIGN_PAMEPET = 0x1BA8;
const char32_t SUNDANESE_VOWEL_SIGN_PANEULEUNG = 0x1BA9;
const char32_t SUNDANESE_SIGN_PAMAAEH = 0x1BAA;
const char32_t SUNDANESE_SIGN_VIRAMA = 0x1BAB;
const char32_t SUNDANESE_CONSONANT_SIGN_PASANGAN_MA = 0x1BAC;
const char32_t SUNDANESE_CONSONANT_SIGN_PASANGAN_WA = 0x1BAD;
const char32_t SUNDANESE_LETTER_KHA = 0x1BAE;
const char32_t SUNDANESE_LETTER_SYA = 0x1BAF;
const char32_t SUNDANESE_DIGIT_ZERO = 0x1BB0;
const char32_t SUNDANESE_DIGIT_ONE = 0x1BB1;
const char32_t SUNDANESE_DIGIT_TWO = 0x1BB2;
const char32_t SUNDANESE_DIGIT_THREE = 0x1BB3;
const char32_t SUNDANESE_DIGIT_FOUR = 0x1BB4;
const char32_t SUNDANESE_DIGIT_FIVE = 0x1BB5;
const char32_t SUNDANESE_DIGIT_SIX = 0x1BB6;
const char32_t SUNDANESE_DIGIT_SEVEN = 0x1BB7;
const char32_t SUNDANESE_DIGIT_EIGHT = 0x1BB8;
const char32_t SUNDANESE_DIGIT_NINE = 0x1BB9;
const char32_t SUNDANESE_AVAGRAHA = 0x1BBA;
const char32_t SUNDANESE_LETTER_REU = 0x1BBB;
const char32_t SUNDANESE_LETTER_LEU = 0x1BBC;
const char32_t SUNDANESE_LETTER_BHA = 0x1BBD;
const char32_t SUNDANESE_LETTER_FINAL_K = 0x1BBE;
const char32_t SUNDANESE_LETTER_FINAL_M = 0x1BBF;
const char32_t BATAK_LETTER_A = 0x1BC0;
const char32_t BATAK_LETTER_SIMALUNGUN_A = 0x1BC1;
const char32_t BATAK_LETTER_HA = 0x1BC2;
const char32_t BATAK_LETTER_SIMALUNGUN_HA = 0x1BC3;
const char32_t BATAK_LETTER_MANDAILING_HA = 0x1BC4;
const char32_t BATAK_LETTER_BA = 0x1BC5;
const char32_t BATAK_LETTER_KARO_BA = 0x1BC6;
const char32_t BATAK_LETTER_PA = 0x1BC7;
const char32_t BATAK_LETTER_SIMALUNGUN_PA = 0x1BC8;
const char32_t BATAK_LETTER_NA = 0x1BC9;
const char32_t BATAK_LETTER_MANDAILING_NA = 0x1BCA;
const char32_t BATAK_LETTER_WA = 0x1BCB;
const char32_t BATAK_LETTER_SIMALUNGUN_WA = 0x1BCC;
const char32_t BATAK_LETTER_PAKPAK_WA = 0x1BCD;
const char32_t BATAK_LETTER_GA = 0x1BCE;
const char32_t BATAK_LETTER_SIMALUNGUN_GA = 0x1BCF;
const char32_t BATAK_LETTER_JA = 0x1BD0;
const char32_t BATAK_LETTER_DA = 0x1BD1;
const char32_t BATAK_LETTER_RA = 0x1BD2;
const char32_t BATAK_LETTER_SIMALUNGUN_RA = 0x1BD3;
const char32_t BATAK_LETTER_MA = 0x1BD4;
const char32_t BATAK_LETTER_SIMALUNGUN_MA = 0x1BD5;
const char32_t BATAK_LETTER_SOUTHERN_TA = 0x1BD6;
const char32_t BATAK_LETTER_NORTHERN_TA = 0x1BD7;
const char32_t BATAK_LETTER_SA = 0x1BD8;
const char32_t BATAK_LETTER_SIMALUNGUN_SA = 0x1BD9;
const char32_t BATAK_LETTER_MANDAILING_SA = 0x1BDA;
const char32_t BATAK_LETTER_YA = 0x1BDB;
const char32_t BATAK_LETTER_SIMALUNGUN_YA = 0x1BDC;
const char32_t BATAK_LETTER_NGA = 0x1BDD;
const char32_t BATAK_LETTER_LA = 0x1BDE;
const char32_t BATAK_LETTER_SIMALUNGUN_LA = 0x1BDF;
const char32_t BATAK_LETTER_NYA = 0x1BE0;
const char32_t BATAK_LETTER_CA = 0x1BE1;
const char32_t BATAK_LETTER_NDA = 0x1BE2;
const char32_t BATAK_LETTER_MBA = 0x1BE3;
const char32_t BATAK_LETTER_I = 0x1BE4;
const char32_t BATAK_LETTER_U = 0x1BE5;
const char32_t BATAK_SIGN_TOMPI = 0x1BE6;
const char32_t BATAK_VOWEL_SIGN_E = 0x1BE7;
const char32_t BATAK_VOWEL_SIGN_PAKPAK_E = 0x1BE8;
const char32_t BATAK_VOWEL_SIGN_EE = 0x1BE9;
const char32_t BATAK_VOWEL_SIGN_I = 0x1BEA;
const char32_t BATAK_VOWEL_SIGN_KARO_I = 0x1BEB;
const char32_t BATAK_VOWEL_SIGN_O = 0x1BEC;
const char32_t BATAK_VOWEL_SIGN_KARO_O = 0x1BED;
const char32_t BATAK_VOWEL_SIGN_U = 0x1BEE;
const char32_t BATAK_VOWEL_SIGN_U_FOR_SIMALUNGUN_SA = 0x1BEF;
const char32_t BATAK_CONSONANT_SIGN_NG = 0x1BF0;
const char32_t BATAK_CONSONANT_SIGN_H = 0x1BF1;
const char32_t BATAK_PANGOLAT = 0x1BF2;
const char32_t BATAK_PANONGONAN = 0x1BF3;
const char32_t BATAK_SYMBOL_BINDU_NA_METEK = 0x1BFC;
const char32_t BATAK_SYMBOL_BINDU_PINARBORAS = 0x1BFD;
const char32_t BATAK_SYMBOL_BINDU_JUDUL = 0x1BFE;
const char32_t BATAK_SYMBOL_BINDU_PANGOLAT = 0x1BFF;
const char32_t LEPCHA_LETTER_KA = 0x1C00;
const char32_t LEPCHA_LETTER_KLA = 0x1C01;
const char32_t LEPCHA_LETTER_KHA = 0x1C02;
const char32_t LEPCHA_LETTER_GA = 0x1C03;
const char32_t LEPCHA_LETTER_GLA = 0x1C04;
const char32_t LEPCHA_LETTER_NGA = 0x1C05;
const char32_t LEPCHA_LETTER_CA = 0x1C06;
const char32_t LEPCHA_LETTER_CHA = 0x1C07;
const char32_t LEPCHA_LETTER_JA = 0x1C08;
const char32_t LEPCHA_LETTER_NYA = 0x1C09;
const char32_t LEPCHA_LETTER_TA = 0x1C0A;
const char32_t LEPCHA_LETTER_THA = 0x1C0B;
const char32_t LEPCHA_LETTER_DA = 0x1C0C;
const char32_t LEPCHA_LETTER_NA = 0x1C0D;
const char32_t LEPCHA_LETTER_PA = 0x1C0E;
const char32_t LEPCHA_LETTER_PLA = 0x1C0F;
const char32_t LEPCHA_LETTER_PHA = 0x1C10;
const char32_t LEPCHA_LETTER_FA = 0x1C11;
const char32_t LEPCHA_LETTER_FLA = 0x1C12;
const char32_t LEPCHA_LETTER_BA = 0x1C13;
const char32_t LEPCHA_LETTER_BLA = 0x1C14;
const char32_t LEPCHA_LETTER_MA = 0x1C15;
const char32_t LEPCHA_LETTER_MLA = 0x1C16;
const char32_t LEPCHA_LETTER_TSA = 0x1C17;
const char32_t LEPCHA_LETTER_TSHA = 0x1C18;
const char32_t LEPCHA_LETTER_DZA = 0x1C19;
const char32_t LEPCHA_LETTER_YA = 0x1C1A;
const char32_t LEPCHA_LETTER_RA = 0x1C1B;
const char32_t LEPCHA_LETTER_LA = 0x1C1C;
const char32_t LEPCHA_LETTER_HA = 0x1C1D;
const char32_t LEPCHA_LETTER_HLA = 0x1C1E;
const char32_t LEPCHA_LETTER_VA = 0x1C1F;
const char32_t LEPCHA_LETTER_SA = 0x1C20;
const char32_t LEPCHA_LETTER_SHA = 0x1C21;
const char32_t LEPCHA_LETTER_WA = 0x1C22;
const char32_t LEPCHA_LETTER_A = 0x1C23;
const char32_t LEPCHA_SUBJOINED_LETTER_YA = 0x1C24;
const char32_t LEPCHA_SUBJOINED_LETTER_RA = 0x1C25;
const char32_t LEPCHA_VOWEL_SIGN_AA = 0x1C26;
const char32_t LEPCHA_VOWEL_SIGN_I = 0x1C27;
const char32_t LEPCHA_VOWEL_SIGN_O = 0x1C28;
const char32_t LEPCHA_VOWEL_SIGN_OO = 0x1C29;
const char32_t LEPCHA_VOWEL_SIGN_U = 0x1C2A;
const char32_t LEPCHA_VOWEL_SIGN_UU = 0x1C2B;
const char32_t LEPCHA_VOWEL_SIGN_E = 0x1C2C;
const char32_t LEPCHA_CONSONANT_SIGN_K = 0x1C2D;
const char32_t LEPCHA_CONSONANT_SIGN_M = 0x1C2E;
const char32_t LEPCHA_CONSONANT_SIGN_L = 0x1C2F;
const char32_t LEPCHA_CONSONANT_SIGN_N = 0x1C30;
const char32_t LEPCHA_CONSONANT_SIGN_P = 0x1C31;
const char32_t LEPCHA_CONSONANT_SIGN_R = 0x1C32;
const char32_t LEPCHA_CONSONANT_SIGN_T = 0x1C33;
const char32_t LEPCHA_CONSONANT_SIGN_NYIN_DO = 0x1C34;
const char32_t LEPCHA_CONSONANT_SIGN_KANG = 0x1C35;
const char32_t LEPCHA_SIGN_RAN = 0x1C36;
const char32_t LEPCHA_SIGN_NUKTA = 0x1C37;
const char32_t LEPCHA_PUNCTUATION_TA_ROL = 0x1C3B;
const char32_t LEPCHA_PUNCTUATION_NYET_THYOOM_TA_ROL = 0x1C3C;
const char32_t LEPCHA_PUNCTUATION_CER_WA = 0x1C3D;
const char32_t LEPCHA_PUNCTUATION_TSHOOK_CER_WA = 0x1C3E;
const char32_t LEPCHA_PUNCTUATION_TSHOOK = 0x1C3F;
const char32_t LEPCHA_DIGIT_ZERO = 0x1C40;
const char32_t LEPCHA_DIGIT_ONE = 0x1C41;
const char32_t LEPCHA_DIGIT_TWO = 0x1C42;
const char32_t LEPCHA_DIGIT_THREE = 0x1C43;
const char32_t LEPCHA_DIGIT_FOUR = 0x1C44;
const char32_t LEPCHA_DIGIT_FIVE = 0x1C45;
const char32_t LEPCHA_DIGIT_SIX = 0x1C46;
const char32_t LEPCHA_DIGIT_SEVEN = 0x1C47;
const char32_t LEPCHA_DIGIT_EIGHT = 0x1C48;
const char32_t LEPCHA_DIGIT_NINE = 0x1C49;
const char32_t LEPCHA_LETTER_TTA = 0x1C4D;
const char32_t LEPCHA_LETTER_TTHA = 0x1C4E;
const char32_t LEPCHA_LETTER_DDA = 0x1C4F;
const char32_t OL_CHIKI_DIGIT_ZERO = 0x1C50;
const char32_t OL_CHIKI_DIGIT_ONE = 0x1C51;
const char32_t OL_CHIKI_DIGIT_TWO = 0x1C52;
const char32_t OL_CHIKI_DIGIT_THREE = 0x1C53;
const char32_t OL_CHIKI_DIGIT_FOUR = 0x1C54;
const char32_t OL_CHIKI_DIGIT_FIVE = 0x1C55;
const char32_t OL_CHIKI_DIGIT_SIX = 0x1C56;
const char32_t OL_CHIKI_DIGIT_SEVEN = 0x1C57;
const char32_t OL_CHIKI_DIGIT_EIGHT = 0x1C58;
const char32_t OL_CHIKI_DIGIT_NINE = 0x1C59;
const char32_t OL_CHIKI_LETTER_LA = 0x1C5A;
const char32_t OL_CHIKI_LETTER_AT = 0x1C5B;
const char32_t OL_CHIKI_LETTER_AG = 0x1C5C;
const char32_t OL_CHIKI_LETTER_ANG = 0x1C5D;
const char32_t OL_CHIKI_LETTER_AL = 0x1C5E;
const char32_t OL_CHIKI_LETTER_LAA = 0x1C5F;
const char32_t OL_CHIKI_LETTER_AAK = 0x1C60;
const char32_t OL_CHIKI_LETTER_AAJ = 0x1C61;
const char32_t OL_CHIKI_LETTER_AAM = 0x1C62;
const char32_t OL_CHIKI_LETTER_AAW = 0x1C63;
const char32_t OL_CHIKI_LETTER_LI = 0x1C64;
const char32_t OL_CHIKI_LETTER_IS = 0x1C65;
const char32_t OL_CHIKI_LETTER_IH = 0x1C66;
const char32_t OL_CHIKI_LETTER_INY = 0x1C67;
const char32_t OL_CHIKI_LETTER_IR = 0x1C68;
const char32_t OL_CHIKI_LETTER_LU = 0x1C69;
const char32_t OL_CHIKI_LETTER_UC = 0x1C6A;
const char32_t OL_CHIKI_LETTER_UD = 0x1C6B;
const char32_t OL_CHIKI_LETTER_UNN = 0x1C6C;
const char32_t OL_CHIKI_LETTER_UY = 0x1C6D;
const char32_t OL_CHIKI_LETTER_LE = 0x1C6E;
const char32_t OL_CHIKI_LETTER_EP = 0x1C6F;
const char32_t OL_CHIKI_LETTER_EDD = 0x1C70;
const char32_t OL_CHIKI_LETTER_EN = 0x1C71;
const char32_t OL_CHIKI_LETTER_ERR = 0x1C72;
const char32_t OL_CHIKI_LETTER_LO = 0x1C73;
const char32_t OL_CHIKI_LETTER_OTT = 0x1C74;
const char32_t OL_CHIKI_LETTER_OB = 0x1C75;
const char32_t OL_CHIKI_LETTER_OV = 0x1C76;
const char32_t OL_CHIKI_LETTER_OH = 0x1C77;
const char32_t OL_CHIKI_MU_TTUDDAG = 0x1C78;
const char32_t OL_CHIKI_GAAHLAA_TTUDDAAG = 0x1C79;
const char32_t OL_CHIKI_MU_GAAHLAA_TTUDDAAG = 0x1C7A;
const char32_t OL_CHIKI_RELAA = 0x1C7B;
const char32_t OL_CHIKI_PHAARKAA = 0x1C7C;
const char32_t OL_CHIKI_AHAD = 0x1C7D;
const char32_t OL_CHIKI_PUNCTUATION_MUCAAD = 0x1C7E;
const char32_t OL_CHIKI_PUNCTUATION_DOUBLE_MUCAAD = 0x1C7F;
const char32_t CYRILLIC_SMALL_LETTER_ROUNDED_VE = 0x1C80;
const char32_t CYRILLIC_SMALL_LETTER_LONG_LEGGED_DE = 0x1C81;
const char32_t CYRILLIC_SMALL_LETTER_NARROW_O = 0x1C82;
const char32_t CYRILLIC_SMALL_LETTER_WIDE_ES = 0x1C83;
const char32_t CYRILLIC_SMALL_LETTER_TALL_TE = 0x1C84;
const char32_t CYRILLIC_SMALL_LETTER_THREE_LEGGED_TE = 0x1C85;
const char32_t CYRILLIC_SMALL_LETTER_TALL_HARD_SIGN = 0x1C86;
const char32_t CYRILLIC_SMALL_LETTER_TALL_YAT = 0x1C87;
const char32_t CYRILLIC_SMALL_LETTER_UNBLENDED_UK = 0x1C88;
const char32_t CYRILLIC_CAPITAL_LETTER_TJE = 0x1C89;
const char32_t CYRILLIC_SMALL_LETTER_TJE = 0x1C8A;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_AN = 0x1C90;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_BAN = 0x1C91;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_GAN = 0x1C92;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_DON = 0x1C93;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_EN = 0x1C94;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_VIN = 0x1C95;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_ZEN = 0x1C96;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_TAN = 0x1C97;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_IN = 0x1C98;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_KAN = 0x1C99;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_LAS = 0x1C9A;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_MAN = 0x1C9B;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_NAR = 0x1C9C;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_ON = 0x1C9D;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_PAR = 0x1C9E;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_ZHAR = 0x1C9F;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_RAE = 0x1CA0;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_SAN = 0x1CA1;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_TAR = 0x1CA2;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_UN = 0x1CA3;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_PHAR = 0x1CA4;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_KHAR = 0x1CA5;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_GHAN = 0x1CA6;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_QAR = 0x1CA7;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_SHIN = 0x1CA8;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_CHIN = 0x1CA9;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_CAN = 0x1CAA;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_JIL = 0x1CAB;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_CIL = 0x1CAC;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_CHAR = 0x1CAD;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_XAN = 0x1CAE;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_JHAN = 0x1CAF;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_HAE = 0x1CB0;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_HE = 0x1CB1;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_HIE = 0x1CB2;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_WE = 0x1CB3;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_HAR = 0x1CB4;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_HOE = 0x1CB5;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_FI = 0x1CB6;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_YN = 0x1CB7;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_ELIFI = 0x1CB8;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_TURNED_GAN = 0x1CB9;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_AIN = 0x1CBA;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_AEN = 0x1CBD;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_HARD_SIGN = 0x1CBE;
const char32_t GEORGIAN_MTAVRULI_CAPITAL_LETTER_LABIAL_SIGN = 0x1CBF;
const char32_t SUNDANESE_PUNCTUATION_BINDU_SURYA = 0x1CC0;
const char32_t SUNDANESE_PUNCTUATION_BINDU_PANGLONG = 0x1CC1;
const char32_t SUNDANESE_PUNCTUATION_BINDU_PURNAMA = 0x1CC2;
const char32_t SUNDANESE_PUNCTUATION_BINDU_CAKRA = 0x1CC3;
const char32_t SUNDANESE_PUNCTUATION_BINDU_LEU_SATANGA = 0x1CC4;
const char32_t SUNDANESE_PUNCTUATION_BINDU_KA_SATANGA = 0x1CC5;
const char32_t SUNDANESE_PUNCTUATION_BINDU_DA_SATANGA = 0x1CC6;
const char32_t SUNDANESE_PUNCTUATION_BINDU_BA_SATANGA = 0x1CC7;
const char32_t VEDIC_TONE_KARSHANA = 0x1CD0;
const char32_t VEDIC_TONE_SHARA = 0x1CD1;
const char32_t VEDIC_TONE_PRENKHA = 0x1CD2;
const char32_t VEDIC_SIGN_NIHSHVASA = 0x1CD3;
const char32_t VEDIC_SIGN_YAJURVEDIC_MIDLINE_SVARITA = 0x1CD4;
const char32_t VEDIC_TONE_YAJURVEDIC_AGGRAVATED_INDEPENDENT_SVARITA = 0x1CD5;
const char32_t VEDIC_TONE_YAJURVEDIC_INDEPENDENT_SVARITA = 0x1CD6;
const char32_t VEDIC_TONE_YAJURVEDIC_KATHAKA_INDEPENDENT_SVARITA = 0x1CD7;
const char32_t VEDIC_TONE_CANDRA_BELOW = 0x1CD8;
const char32_t VEDIC_TONE_YAJURVEDIC_KATHAKA_INDEPENDENT_SVARITA_SCHROEDER = 0x1CD9;
const char32_t VEDIC_TONE_DOUBLE_SVARITA = 0x1CDA;
const char32_t VEDIC_TONE_TRIPLE_SVARITA = 0x1CDB;
const char32_t VEDIC_TONE_KATHAKA_ANUDATTA = 0x1CDC;
const char32_t VEDIC_TONE_DOT_BELOW = 0x1CDD;
const char32_t VEDIC_TONE_TWO_DOTS_BELOW = 0x1CDE;
const char32_t VEDIC_TONE_THREE_DOTS_BELOW = 0x1CDF;
const char32_t VEDIC_TONE_RIGVEDIC_KASHMIRI_INDEPENDENT_SVARITA = 0x1CE0;
const char32_t VEDIC_TONE_ATHARVAVEDIC_INDEPENDENT_SVARITA = 0x1CE1;
const char32_t VEDIC_SIGN_VISARGA_SVARITA = 0x1CE2;
const char32_t VEDIC_SIGN_VISARGA_UDATTA = 0x1CE3;
const char32_t VEDIC_SIGN_REVERSED_VISARGA_UDATTA = 0x1CE4;
const char32_t VEDIC_SIGN_VISARGA_ANUDATTA = 0x1CE5;
const char32_t VEDIC_SIGN_REVERSED_VISARGA_ANUDATTA = 0x1CE6;
const char32_t VEDIC_SIGN_VISARGA_UDATTA_WITH_TAIL = 0x1CE7;
const char32_t VEDIC_SIGN_VISARGA_ANUDATTA_WITH_TAIL = 0x1CE8;
const char32_t VEDIC_SIGN_ANUSVARA_ANTARGOMUKHA = 0x1CE9;
const char32_t VEDIC_SIGN_ANUSVARA_BAHIRGOMUKHA = 0x1CEA;
const char32_t VEDIC_SIGN_ANUSVARA_VAMAGOMUKHA = 0x1CEB;
const char32_t VEDIC_SIGN_ANUSVARA_VAMAGOMUKHA_WITH_TAIL = 0x1CEC;
const char32_t VEDIC_SIGN_TIRYAK = 0x1CED;
const char32_t VEDIC_SIGN_HEXIFORM_LONG_ANUSVARA = 0x1CEE;
const char32_t VEDIC_SIGN_LONG_ANUSVARA = 0x1CEF;
const char32_t VEDIC_SIGN_RTHANG_LONG_ANUSVARA = 0x1CF0;
const char32_t VEDIC_SIGN_ANUSVARA_UBHAYATO_MUKHA = 0x1CF1;
const char32_t VEDIC_SIGN_ARDHAVISARGA = 0x1CF2;
const char32_t VEDIC_SIGN_ROTATED_ARDHAVISARGA = 0x1CF3;
const char32_t VEDIC_TONE_CANDRA_ABOVE = 0x1CF4;
const char32_t VEDIC_SIGN_JIHVAMULIYA = 0x1CF5;
const char32_t VEDIC_SIGN_UPADHMANIYA = 0x1CF6;
const char32_t VEDIC_SIGN_ATIKRAMA = 0x1CF7;
const char32_t VEDIC_TONE_RING_ABOVE = 0x1CF8;
const char32_t VEDIC_TONE_DOUBLE_RING_ABOVE = 0x1CF9;
const char32_t VEDIC_SIGN_DOUBLE_ANUSVARA_ANTARGOMUKHA = 0x1CFA;
const char32_t LATIN_LETTER_SMALL_CAPITAL_A = 0x1D00;
const char32_t LATIN_LETTER_SMALL_CAPITAL_AE = 0x1D01;
const char32_t LATIN_SMALL_LETTER_TURNED_AE = 0x1D02;
const char32_t LATIN_LETTER_SMALL_CAPITAL_BARRED_B = 0x1D03;
const char32_t LATIN_LETTER_SMALL_CAPITAL_C = 0x1D04;
const char32_t LATIN_LETTER_SMALL_CAPITAL_D = 0x1D05;
const char32_t LATIN_LETTER_SMALL_CAPITAL_ETH = 0x1D06;
const char32_t LATIN_LETTER_SMALL_CAPITAL_E = 0x1D07;
const char32_t LATIN_SMALL_LETTER_TURNED_OPEN_E = 0x1D08;
const char32_t LATIN_SMALL_LETTER_TURNED_I = 0x1D09;
const char32_t LATIN_LETTER_SMALL_CAPITAL_J = 0x1D0A;
const char32_t LATIN_LETTER_SMALL_CAPITAL_K = 0x1D0B;
const char32_t LATIN_LETTER_SMALL_CAPITAL_L_WITH_STROKE = 0x1D0C;
const char32_t LATIN_LETTER_SMALL_CAPITAL_M = 0x1D0D;
const char32_t LATIN_LETTER_SMALL_CAPITAL_REVERSED_N = 0x1D0E;
const char32_t LATIN_LETTER_SMALL_CAPITAL_O = 0x1D0F;
const char32_t LATIN_LETTER_SMALL_CAPITAL_OPEN_O = 0x1D10;
const char32_t LATIN_SMALL_LETTER_SIDEWAYS_O = 0x1D11;
const char32_t LATIN_SMALL_LETTER_SIDEWAYS_OPEN_O = 0x1D12;
const char32_t LATIN_SMALL_LETTER_SIDEWAYS_O_WITH_STROKE = 0x1D13;
const char32_t LATIN_SMALL_LETTER_TURNED_OE = 0x1D14;
const char32_t LATIN_LETTER_SMALL_CAPITAL_OU = 0x1D15;
const char32_t LATIN_SMALL_LETTER_TOP_HALF_O = 0x1D16;
const char32_t LATIN_SMALL_LETTER_BOTTOM_HALF_O = 0x1D17;
const char32_t LATIN_LETTER_SMALL_CAPITAL_P = 0x1D18;
const char32_t LATIN_LETTER_SMALL_CAPITAL_REVERSED_R = 0x1D19;
const char32_t LATIN_LETTER_SMALL_CAPITAL_TURNED_R = 0x1D1A;
const char32_t LATIN_LETTER_SMALL_CAPITAL_T = 0x1D1B;
const char32_t LATIN_LETTER_SMALL_CAPITAL_U = 0x1D1C;
const char32_t LATIN_SMALL_LETTER_SIDEWAYS_U = 0x1D1D;
const char32_t LATIN_SMALL_LETTER_SIDEWAYS_DIAERESIZED_U = 0x1D1E;
const char32_t LATIN_SMALL_LETTER_SIDEWAYS_TURNED_M = 0x1D1F;
const char32_t LATIN_LETTER_SMALL_CAPITAL_V = 0x1D20;
const char32_t LATIN_LETTER_SMALL_CAPITAL_W = 0x1D21;
const char32_t LATIN_LETTER_SMALL_CAPITAL_Z = 0x1D22;
const char32_t LATIN_LETTER_SMALL_CAPITAL_EZH = 0x1D23;
const char32_t LATIN_LETTER_VOICED_LARYNGEAL_SPIRANT = 0x1D24;
const char32_t LATIN_LETTER_AIN = 0x1D25;
const char32_t GREEK_LETTER_SMALL_CAPITAL_GAMMA = 0x1D26;
const char32_t GREEK_LETTER_SMALL_CAPITAL_LAMDA = 0x1D27;
const char32_t GREEK_LETTER_SMALL_CAPITAL_PI = 0x1D28;
const char32_t GREEK_LETTER_SMALL_CAPITAL_RHO = 0x1D29;
const char32_t GREEK_LETTER_SMALL_CAPITAL_PSI = 0x1D2A;
const char32_t CYRILLIC_LETTER_SMALL_CAPITAL_EL = 0x1D2B;
const char32_t MODIFIER_LETTER_CAPITAL_A = 0x1D2C;
const char32_t MODIFIER_LETTER_CAPITAL_AE = 0x1D2D;
const char32_t MODIFIER_LETTER_CAPITAL_B = 0x1D2E;
const char32_t MODIFIER_LETTER_CAPITAL_BARRED_B = 0x1D2F;
const char32_t MODIFIER_LETTER_CAPITAL_D = 0x1D30;
const char32_t MODIFIER_LETTER_CAPITAL_E = 0x1D31;
const char32_t MODIFIER_LETTER_CAPITAL_REVERSED_E = 0x1D32;
const char32_t MODIFIER_LETTER_CAPITAL_G = 0x1D33;
const char32_t MODIFIER_LETTER_CAPITAL_H = 0x1D34;
const char32_t MODIFIER_LETTER_CAPITAL_I = 0x1D35;
const char32_t MODIFIER_LETTER_CAPITAL_J = 0x1D36;
const char32_t MODIFIER_LETTER_CAPITAL_K = 0x1D37;
const char32_t MODIFIER_LETTER_CAPITAL_L = 0x1D38;
const char32_t MODIFIER_LETTER_CAPITAL_M = 0x1D39;
const char32_t MODIFIER_LETTER_CAPITAL_N = 0x1D3A;
const char32_t MODIFIER_LETTER_CAPITAL_REVERSED_N = 0x1D3B;
const char32_t MODIFIER_LETTER_CAPITAL_O = 0x1D3C;
const char32_t MODIFIER_LETTER_CAPITAL_OU = 0x1D3D;
const char32_t MODIFIER_LETTER_CAPITAL_P = 0x1D3E;
const char32_t MODIFIER_LETTER_CAPITAL_R = 0x1D3F;
const char32_t MODIFIER_LETTER_CAPITAL_T = 0x1D40;
const char32_t MODIFIER_LETTER_CAPITAL_U = 0x1D41;
const char32_t MODIFIER_LETTER_CAPITAL_W = 0x1D42;
const char32_t MODIFIER_LETTER_SMALL_A = 0x1D43;
const char32_t MODIFIER_LETTER_SMALL_TURNED_A = 0x1D44;
const char32_t MODIFIER_LETTER_SMALL_ALPHA = 0x1D45;
const char32_t MODIFIER_LETTER_SMALL_TURNED_AE = 0x1D46;
const char32_t MODIFIER_LETTER_SMALL_B = 0x1D47;
const char32_t MODIFIER_LETTER_SMALL_D = 0x1D48;
const char32_t MODIFIER_LETTER_SMALL_E = 0x1D49;
const char32_t MODIFIER_LETTER_SMALL_SCHWA = 0x1D4A;
const char32_t MODIFIER_LETTER_SMALL_OPEN_E = 0x1D4B;
const char32_t MODIFIER_LETTER_SMALL_TURNED_OPEN_E = 0x1D4C;
const char32_t MODIFIER_LETTER_SMALL_G = 0x1D4D;
const char32_t MODIFIER_LETTER_SMALL_TURNED_I = 0x1D4E;
const char32_t MODIFIER_LETTER_SMALL_K = 0x1D4F;
const char32_t MODIFIER_LETTER_SMALL_M = 0x1D50;
const char32_t MODIFIER_LETTER_SMALL_ENG = 0x1D51;
const char32_t MODIFIER_LETTER_SMALL_O = 0x1D52;
const char32_t MODIFIER_LETTER_SMALL_OPEN_O = 0x1D53;
const char32_t MODIFIER_LETTER_SMALL_TOP_HALF_O = 0x1D54;
const char32_t MODIFIER_LETTER_SMALL_BOTTOM_HALF_O = 0x1D55;
const char32_t MODIFIER_LETTER_SMALL_P = 0x1D56;
const char32_t MODIFIER_LETTER_SMALL_T = 0x1D57;
const char32_t MODIFIER_LETTER_SMALL_U = 0x1D58;
const char32_t MODIFIER_LETTER_SMALL_SIDEWAYS_U = 0x1D59;
const char32_t MODIFIER_LETTER_SMALL_TURNED_M = 0x1D5A;
const char32_t MODIFIER_LETTER_SMALL_V = 0x1D5B;
const char32_t MODIFIER_LETTER_SMALL_AIN = 0x1D5C;
const char32_t MODIFIER_LETTER_SMALL_BETA = 0x1D5D;
const char32_t MODIFIER_LETTER_SMALL_GREEK_GAMMA = 0x1D5E;
const char32_t MODIFIER_LETTER_SMALL_DELTA = 0x1D5F;
const char32_t MODIFIER_LETTER_SMALL_GREEK_PHI = 0x1D60;
const char32_t MODIFIER_LETTER_SMALL_CHI = 0x1D61;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_I = 0x1D62;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_R = 0x1D63;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_U = 0x1D64;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_V = 0x1D65;
const char32_t GREEK_SUBSCRIPT_SMALL_LETTER_BETA = 0x1D66;
const char32_t GREEK_SUBSCRIPT_SMALL_LETTER_GAMMA = 0x1D67;
const char32_t GREEK_SUBSCRIPT_SMALL_LETTER_RHO = 0x1D68;
const char32_t GREEK_SUBSCRIPT_SMALL_LETTER_PHI = 0x1D69;
const char32_t GREEK_SUBSCRIPT_SMALL_LETTER_CHI = 0x1D6A;
const char32_t LATIN_SMALL_LETTER_UE = 0x1D6B;
const char32_t LATIN_SMALL_LETTER_B_WITH_MIDDLE_TILDE = 0x1D6C;
const char32_t LATIN_SMALL_LETTER_D_WITH_MIDDLE_TILDE = 0x1D6D;
const char32_t LATIN_SMALL_LETTER_F_WITH_MIDDLE_TILDE = 0x1D6E;
const char32_t LATIN_SMALL_LETTER_M_WITH_MIDDLE_TILDE = 0x1D6F;
const char32_t LATIN_SMALL_LETTER_N_WITH_MIDDLE_TILDE = 0x1D70;
const char32_t LATIN_SMALL_LETTER_P_WITH_MIDDLE_TILDE = 0x1D71;
const char32_t LATIN_SMALL_LETTER_R_WITH_MIDDLE_TILDE = 0x1D72;
const char32_t LATIN_SMALL_LETTER_R_WITH_FISHHOOK_AND_MIDDLE_TILDE = 0x1D73;
const char32_t LATIN_SMALL_LETTER_S_WITH_MIDDLE_TILDE = 0x1D74;
const char32_t LATIN_SMALL_LETTER_T_WITH_MIDDLE_TILDE = 0x1D75;
const char32_t LATIN_SMALL_LETTER_Z_WITH_MIDDLE_TILDE = 0x1D76;
const char32_t LATIN_SMALL_LETTER_TURNED_G = 0x1D77;
const char32_t MODIFIER_LETTER_CYRILLIC_EN = 0x1D78;
const char32_t LATIN_SMALL_LETTER_INSULAR_G = 0x1D79;
const char32_t LATIN_SMALL_LETTER_TH_WITH_STRIKETHROUGH = 0x1D7A;
const char32_t LATIN_SMALL_CAPITAL_LETTER_I_WITH_STROKE = 0x1D7B;
const char32_t LATIN_SMALL_LETTER_IOTA_WITH_STROKE = 0x1D7C;
const char32_t LATIN_SMALL_LETTER_P_WITH_STROKE = 0x1D7D;
const char32_t LATIN_SMALL_CAPITAL_LETTER_U_WITH_STROKE = 0x1D7E;
const char32_t LATIN_SMALL_LETTER_UPSILON_WITH_STROKE = 0x1D7F;
const char32_t LATIN_SMALL_LETTER_B_WITH_PALATAL_HOOK = 0x1D80;
const char32_t LATIN_SMALL_LETTER_D_WITH_PALATAL_HOOK = 0x1D81;
const char32_t LATIN_SMALL_LETTER_F_WITH_PALATAL_HOOK = 0x1D82;
const char32_t LATIN_SMALL_LETTER_G_WITH_PALATAL_HOOK = 0x1D83;
const char32_t LATIN_SMALL_LETTER_K_WITH_PALATAL_HOOK = 0x1D84;
const char32_t LATIN_SMALL_LETTER_L_WITH_PALATAL_HOOK = 0x1D85;
const char32_t LATIN_SMALL_LETTER_M_WITH_PALATAL_HOOK = 0x1D86;
const char32_t LATIN_SMALL_LETTER_N_WITH_PALATAL_HOOK = 0x1D87;
const char32_t LATIN_SMALL_LETTER_P_WITH_PALATAL_HOOK = 0x1D88;
const char32_t LATIN_SMALL_LETTER_R_WITH_PALATAL_HOOK = 0x1D89;
const char32_t LATIN_SMALL_LETTER_S_WITH_PALATAL_HOOK = 0x1D8A;
const char32_t LATIN_SMALL_LETTER_ESH_WITH_PALATAL_HOOK = 0x1D8B;
const char32_t LATIN_SMALL_LETTER_V_WITH_PALATAL_HOOK = 0x1D8C;
const char32_t LATIN_SMALL_LETTER_X_WITH_PALATAL_HOOK = 0x1D8D;
const char32_t LATIN_SMALL_LETTER_Z_WITH_PALATAL_HOOK = 0x1D8E;
const char32_t LATIN_SMALL_LETTER_A_WITH_RETROFLEX_HOOK = 0x1D8F;
const char32_t LATIN_SMALL_LETTER_ALPHA_WITH_RETROFLEX_HOOK = 0x1D90;
const char32_t LATIN_SMALL_LETTER_D_WITH_HOOK_AND_TAIL = 0x1D91;
const char32_t LATIN_SMALL_LETTER_E_WITH_RETROFLEX_HOOK = 0x1D92;
const char32_t LATIN_SMALL_LETTER_OPEN_E_WITH_RETROFLEX_HOOK = 0x1D93;
const char32_t LATIN_SMALL_LETTER_REVERSED_OPEN_E_WITH_RETROFLEX_HOOK = 0x1D94;
const char32_t LATIN_SMALL_LETTER_SCHWA_WITH_RETROFLEX_HOOK = 0x1D95;
const char32_t LATIN_SMALL_LETTER_I_WITH_RETROFLEX_HOOK = 0x1D96;
const char32_t LATIN_SMALL_LETTER_OPEN_O_WITH_RETROFLEX_HOOK = 0x1D97;
const char32_t LATIN_SMALL_LETTER_ESH_WITH_RETROFLEX_HOOK = 0x1D98;
const char32_t LATIN_SMALL_LETTER_U_WITH_RETROFLEX_HOOK = 0x1D99;
const char32_t LATIN_SMALL_LETTER_EZH_WITH_RETROFLEX_HOOK = 0x1D9A;
const char32_t MODIFIER_LETTER_SMALL_TURNED_ALPHA = 0x1D9B;
const char32_t MODIFIER_LETTER_SMALL_C = 0x1D9C;
const char32_t MODIFIER_LETTER_SMALL_C_WITH_CURL = 0x1D9D;
const char32_t MODIFIER_LETTER_SMALL_ETH = 0x1D9E;
const char32_t MODIFIER_LETTER_SMALL_REVERSED_OPEN_E = 0x1D9F;
const char32_t MODIFIER_LETTER_SMALL_F = 0x1DA0;
const char32_t MODIFIER_LETTER_SMALL_DOTLESS_J_WITH_STROKE = 0x1DA1;
const char32_t MODIFIER_LETTER_SMALL_SCRIPT_G = 0x1DA2;
const char32_t MODIFIER_LETTER_SMALL_TURNED_H = 0x1DA3;
const char32_t MODIFIER_LETTER_SMALL_I_WITH_STROKE = 0x1DA4;
const char32_t MODIFIER_LETTER_SMALL_IOTA = 0x1DA5;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_I = 0x1DA6;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_I_WITH_STROKE = 0x1DA7;
const char32_t MODIFIER_LETTER_SMALL_J_WITH_CROSSED_TAIL = 0x1DA8;
const char32_t MODIFIER_LETTER_SMALL_L_WITH_RETROFLEX_HOOK = 0x1DA9;
const char32_t MODIFIER_LETTER_SMALL_L_WITH_PALATAL_HOOK = 0x1DAA;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_L = 0x1DAB;
const char32_t MODIFIER_LETTER_SMALL_M_WITH_HOOK = 0x1DAC;
const char32_t MODIFIER_LETTER_SMALL_TURNED_M_WITH_LONG_LEG = 0x1DAD;
const char32_t MODIFIER_LETTER_SMALL_N_WITH_LEFT_HOOK = 0x1DAE;
const char32_t MODIFIER_LETTER_SMALL_N_WITH_RETROFLEX_HOOK = 0x1DAF;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_N = 0x1DB0;
const char32_t MODIFIER_LETTER_SMALL_BARRED_O = 0x1DB1;
const char32_t MODIFIER_LETTER_SMALL_PHI = 0x1DB2;
const char32_t MODIFIER_LETTER_SMALL_S_WITH_HOOK = 0x1DB3;
const char32_t MODIFIER_LETTER_SMALL_ESH = 0x1DB4;
const char32_t MODIFIER_LETTER_SMALL_T_WITH_PALATAL_HOOK = 0x1DB5;
const char32_t MODIFIER_LETTER_SMALL_U_BAR = 0x1DB6;
const char32_t MODIFIER_LETTER_SMALL_UPSILON = 0x1DB7;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_U = 0x1DB8;
const char32_t MODIFIER_LETTER_SMALL_V_WITH_HOOK = 0x1DB9;
const char32_t MODIFIER_LETTER_SMALL_TURNED_V = 0x1DBA;
const char32_t MODIFIER_LETTER_SMALL_Z = 0x1DBB;
const char32_t MODIFIER_LETTER_SMALL_Z_WITH_RETROFLEX_HOOK = 0x1DBC;
const char32_t MODIFIER_LETTER_SMALL_Z_WITH_CURL = 0x1DBD;
const char32_t MODIFIER_LETTER_SMALL_EZH = 0x1DBE;
const char32_t MODIFIER_LETTER_SMALL_THETA = 0x1DBF;
const char32_t COMBINING_DOTTED_GRAVE_ACCENT = 0x1DC0;
const char32_t COMBINING_DOTTED_ACUTE_ACCENT = 0x1DC1;
const char32_t COMBINING_SNAKE_BELOW = 0x1DC2;
const char32_t COMBINING_SUSPENSION_MARK = 0x1DC3;
const char32_t COMBINING_MACRON_ACUTE = 0x1DC4;
const char32_t COMBINING_GRAVE_MACRON = 0x1DC5;
const char32_t COMBINING_MACRON_GRAVE = 0x1DC6;
const char32_t COMBINING_ACUTE_MACRON = 0x1DC7;
const char32_t COMBINING_GRAVE_ACUTE_GRAVE = 0x1DC8;
const char32_t COMBINING_ACUTE_GRAVE_ACUTE = 0x1DC9;
const char32_t COMBINING_LATIN_SMALL_LETTER_R_BELOW = 0x1DCA;
const char32_t COMBINING_BREVE_MACRON = 0x1DCB;
const char32_t COMBINING_MACRON_BREVE = 0x1DCC;
const char32_t COMBINING_DOUBLE_CIRCUMFLEX_ABOVE = 0x1DCD;
const char32_t COMBINING_OGONEK_ABOVE = 0x1DCE;
const char32_t COMBINING_ZIGZAG_BELOW = 0x1DCF;
const char32_t COMBINING_IS_BELOW = 0x1DD0;
const char32_t COMBINING_UR_ABOVE = 0x1DD1;
const char32_t COMBINING_US_ABOVE = 0x1DD2;
const char32_t COMBINING_LATIN_SMALL_LETTER_FLATTENED_OPEN_A_ABOVE = 0x1DD3;
const char32_t COMBINING_LATIN_SMALL_LETTER_AE = 0x1DD4;
const char32_t COMBINING_LATIN_SMALL_LETTER_AO = 0x1DD5;
const char32_t COMBINING_LATIN_SMALL_LETTER_AV = 0x1DD6;
const char32_t COMBINING_LATIN_SMALL_LETTER_C_CEDILLA = 0x1DD7;
const char32_t COMBINING_LATIN_SMALL_LETTER_INSULAR_D = 0x1DD8;
const char32_t COMBINING_LATIN_SMALL_LETTER_ETH = 0x1DD9;
const char32_t COMBINING_LATIN_SMALL_LETTER_G = 0x1DDA;
const char32_t COMBINING_LATIN_LETTER_SMALL_CAPITAL_G = 0x1DDB;
const char32_t COMBINING_LATIN_SMALL_LETTER_K = 0x1DDC;
const char32_t COMBINING_LATIN_SMALL_LETTER_L = 0x1DDD;
const char32_t COMBINING_LATIN_LETTER_SMALL_CAPITAL_L = 0x1DDE;
const char32_t COMBINING_LATIN_LETTER_SMALL_CAPITAL_M = 0x1DDF;
const char32_t COMBINING_LATIN_SMALL_LETTER_N = 0x1DE0;
const char32_t COMBINING_LATIN_LETTER_SMALL_CAPITAL_N = 0x1DE1;
const char32_t COMBINING_LATIN_LETTER_SMALL_CAPITAL_R = 0x1DE2;
const char32_t COMBINING_LATIN_SMALL_LETTER_R_ROTUNDA = 0x1DE3;
const char32_t COMBINING_LATIN_SMALL_LETTER_S = 0x1DE4;
const char32_t COMBINING_LATIN_SMALL_LETTER_LONG_S = 0x1DE5;
const char32_t COMBINING_LATIN_SMALL_LETTER_Z = 0x1DE6;
const char32_t COMBINING_LATIN_SMALL_LETTER_ALPHA = 0x1DE7;
const char32_t COMBINING_LATIN_SMALL_LETTER_B = 0x1DE8;
const char32_t COMBINING_LATIN_SMALL_LETTER_BETA = 0x1DE9;
const char32_t COMBINING_LATIN_SMALL_LETTER_SCHWA = 0x1DEA;
const char32_t COMBINING_LATIN_SMALL_LETTER_F = 0x1DEB;
const char32_t COMBINING_LATIN_SMALL_LETTER_L_WITH_DOUBLE_MIDDLE_TILDE = 0x1DEC;
const char32_t COMBINING_LATIN_SMALL_LETTER_O_WITH_LIGHT_CENTRALIZATION_STROKE = 0x1DED;
const char32_t COMBINING_LATIN_SMALL_LETTER_P = 0x1DEE;
const char32_t COMBINING_LATIN_SMALL_LETTER_ESH = 0x1DEF;
const char32_t COMBINING_LATIN_SMALL_LETTER_U_WITH_LIGHT_CENTRALIZATION_STROKE = 0x1DF0;
const char32_t COMBINING_LATIN_SMALL_LETTER_W = 0x1DF1;
const char32_t COMBINING_LATIN_SMALL_LETTER_A_WITH_DIAERESIS = 0x1DF2;
const char32_t COMBINING_LATIN_SMALL_LETTER_O_WITH_DIAERESIS = 0x1DF3;
const char32_t COMBINING_LATIN_SMALL_LETTER_U_WITH_DIAERESIS = 0x1DF4;
const char32_t COMBINING_UP_TACK_ABOVE = 0x1DF5;
const char32_t COMBINING_KAVYKA_ABOVE_RIGHT = 0x1DF6;
const char32_t COMBINING_KAVYKA_ABOVE_LEFT = 0x1DF7;
const char32_t COMBINING_DOT_ABOVE_LEFT = 0x1DF8;
const char32_t COMBINING_WIDE_INVERTED_BRIDGE_BELOW = 0x1DF9;
const char32_t COMBINING_DOT_BELOW_LEFT = 0x1DFA;
const char32_t COMBINING_DELETION_MARK = 0x1DFB;
const char32_t COMBINING_DOUBLE_INVERTED_BREVE_BELOW = 0x1DFC;
const char32_t COMBINING_ALMOST_EQUAL_TO_BELOW = 0x1DFD;
const char32_t COMBINING_LEFT_ARROWHEAD_ABOVE = 0x1DFE;
const char32_t COMBINING_RIGHT_ARROWHEAD_AND_DOWN_ARROWHEAD_BELOW = 0x1DFF;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_RING_BELOW = 0x1E00;
const char32_t LATIN_SMALL_LETTER_A_WITH_RING_BELOW = 0x1E01;
const char32_t LATIN_CAPITAL_LETTER_B_WITH_DOT_ABOVE = 0x1E02;
const char32_t LATIN_SMALL_LETTER_B_WITH_DOT_ABOVE = 0x1E03;
const char32_t LATIN_CAPITAL_LETTER_B_WITH_DOT_BELOW = 0x1E04;
const char32_t LATIN_SMALL_LETTER_B_WITH_DOT_BELOW = 0x1E05;
const char32_t LATIN_CAPITAL_LETTER_B_WITH_LINE_BELOW = 0x1E06;
const char32_t LATIN_SMALL_LETTER_B_WITH_LINE_BELOW = 0x1E07;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_CEDILLA_AND_ACUTE = 0x1E08;
const char32_t LATIN_SMALL_LETTER_C_WITH_CEDILLA_AND_ACUTE = 0x1E09;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_DOT_ABOVE = 0x1E0A;
const char32_t LATIN_SMALL_LETTER_D_WITH_DOT_ABOVE = 0x1E0B;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_DOT_BELOW = 0x1E0C;
const char32_t LATIN_SMALL_LETTER_D_WITH_DOT_BELOW = 0x1E0D;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_LINE_BELOW = 0x1E0E;
const char32_t LATIN_SMALL_LETTER_D_WITH_LINE_BELOW = 0x1E0F;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_CEDILLA = 0x1E10;
const char32_t LATIN_SMALL_LETTER_D_WITH_CEDILLA = 0x1E11;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_CIRCUMFLEX_BELOW = 0x1E12;
const char32_t LATIN_SMALL_LETTER_D_WITH_CIRCUMFLEX_BELOW = 0x1E13;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_MACRON_AND_GRAVE = 0x1E14;
const char32_t LATIN_SMALL_LETTER_E_WITH_MACRON_AND_GRAVE = 0x1E15;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_MACRON_AND_ACUTE = 0x1E16;
const char32_t LATIN_SMALL_LETTER_E_WITH_MACRON_AND_ACUTE = 0x1E17;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX_BELOW = 0x1E18;
const char32_t LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX_BELOW = 0x1E19;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_TILDE_BELOW = 0x1E1A;
const char32_t LATIN_SMALL_LETTER_E_WITH_TILDE_BELOW = 0x1E1B;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CEDILLA_AND_BREVE = 0x1E1C;
const char32_t LATIN_SMALL_LETTER_E_WITH_CEDILLA_AND_BREVE = 0x1E1D;
const char32_t LATIN_CAPITAL_LETTER_F_WITH_DOT_ABOVE = 0x1E1E;
const char32_t LATIN_SMALL_LETTER_F_WITH_DOT_ABOVE = 0x1E1F;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_MACRON = 0x1E20;
const char32_t LATIN_SMALL_LETTER_G_WITH_MACRON = 0x1E21;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_DOT_ABOVE = 0x1E22;
const char32_t LATIN_SMALL_LETTER_H_WITH_DOT_ABOVE = 0x1E23;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_DOT_BELOW = 0x1E24;
const char32_t LATIN_SMALL_LETTER_H_WITH_DOT_BELOW = 0x1E25;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_DIAERESIS = 0x1E26;
const char32_t LATIN_SMALL_LETTER_H_WITH_DIAERESIS = 0x1E27;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_CEDILLA = 0x1E28;
const char32_t LATIN_SMALL_LETTER_H_WITH_CEDILLA = 0x1E29;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_BREVE_BELOW = 0x1E2A;
const char32_t LATIN_SMALL_LETTER_H_WITH_BREVE_BELOW = 0x1E2B;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_TILDE_BELOW = 0x1E2C;
const char32_t LATIN_SMALL_LETTER_I_WITH_TILDE_BELOW = 0x1E2D;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_DIAERESIS_AND_ACUTE = 0x1E2E;
const char32_t LATIN_SMALL_LETTER_I_WITH_DIAERESIS_AND_ACUTE = 0x1E2F;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_ACUTE = 0x1E30;
const char32_t LATIN_SMALL_LETTER_K_WITH_ACUTE = 0x1E31;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_DOT_BELOW = 0x1E32;
const char32_t LATIN_SMALL_LETTER_K_WITH_DOT_BELOW = 0x1E33;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_LINE_BELOW = 0x1E34;
const char32_t LATIN_SMALL_LETTER_K_WITH_LINE_BELOW = 0x1E35;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_DOT_BELOW = 0x1E36;
const char32_t LATIN_SMALL_LETTER_L_WITH_DOT_BELOW = 0x1E37;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_DOT_BELOW_AND_MACRON = 0x1E38;
const char32_t LATIN_SMALL_LETTER_L_WITH_DOT_BELOW_AND_MACRON = 0x1E39;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_LINE_BELOW = 0x1E3A;
const char32_t LATIN_SMALL_LETTER_L_WITH_LINE_BELOW = 0x1E3B;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_CIRCUMFLEX_BELOW = 0x1E3C;
const char32_t LATIN_SMALL_LETTER_L_WITH_CIRCUMFLEX_BELOW = 0x1E3D;
const char32_t LATIN_CAPITAL_LETTER_M_WITH_ACUTE = 0x1E3E;
const char32_t LATIN_SMALL_LETTER_M_WITH_ACUTE = 0x1E3F;
const char32_t LATIN_CAPITAL_LETTER_M_WITH_DOT_ABOVE = 0x1E40;
const char32_t LATIN_SMALL_LETTER_M_WITH_DOT_ABOVE = 0x1E41;
const char32_t LATIN_CAPITAL_LETTER_M_WITH_DOT_BELOW = 0x1E42;
const char32_t LATIN_SMALL_LETTER_M_WITH_DOT_BELOW = 0x1E43;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_DOT_ABOVE = 0x1E44;
const char32_t LATIN_SMALL_LETTER_N_WITH_DOT_ABOVE = 0x1E45;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_DOT_BELOW = 0x1E46;
const char32_t LATIN_SMALL_LETTER_N_WITH_DOT_BELOW = 0x1E47;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_LINE_BELOW = 0x1E48;
const char32_t LATIN_SMALL_LETTER_N_WITH_LINE_BELOW = 0x1E49;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_CIRCUMFLEX_BELOW = 0x1E4A;
const char32_t LATIN_SMALL_LETTER_N_WITH_CIRCUMFLEX_BELOW = 0x1E4B;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_TILDE_AND_ACUTE = 0x1E4C;
const char32_t LATIN_SMALL_LETTER_O_WITH_TILDE_AND_ACUTE = 0x1E4D;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_TILDE_AND_DIAERESIS = 0x1E4E;
const char32_t LATIN_SMALL_LETTER_O_WITH_TILDE_AND_DIAERESIS = 0x1E4F;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_MACRON_AND_GRAVE = 0x1E50;
const char32_t LATIN_SMALL_LETTER_O_WITH_MACRON_AND_GRAVE = 0x1E51;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_MACRON_AND_ACUTE = 0x1E52;
const char32_t LATIN_SMALL_LETTER_O_WITH_MACRON_AND_ACUTE = 0x1E53;
const char32_t LATIN_CAPITAL_LETTER_P_WITH_ACUTE = 0x1E54;
const char32_t LATIN_SMALL_LETTER_P_WITH_ACUTE = 0x1E55;
const char32_t LATIN_CAPITAL_LETTER_P_WITH_DOT_ABOVE = 0x1E56;
const char32_t LATIN_SMALL_LETTER_P_WITH_DOT_ABOVE = 0x1E57;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_DOT_ABOVE = 0x1E58;
const char32_t LATIN_SMALL_LETTER_R_WITH_DOT_ABOVE = 0x1E59;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_DOT_BELOW = 0x1E5A;
const char32_t LATIN_SMALL_LETTER_R_WITH_DOT_BELOW = 0x1E5B;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_DOT_BELOW_AND_MACRON = 0x1E5C;
const char32_t LATIN_SMALL_LETTER_R_WITH_DOT_BELOW_AND_MACRON = 0x1E5D;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_LINE_BELOW = 0x1E5E;
const char32_t LATIN_SMALL_LETTER_R_WITH_LINE_BELOW = 0x1E5F;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_DOT_ABOVE = 0x1E60;
const char32_t LATIN_SMALL_LETTER_S_WITH_DOT_ABOVE = 0x1E61;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_DOT_BELOW = 0x1E62;
const char32_t LATIN_SMALL_LETTER_S_WITH_DOT_BELOW = 0x1E63;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_ACUTE_AND_DOT_ABOVE = 0x1E64;
const char32_t LATIN_SMALL_LETTER_S_WITH_ACUTE_AND_DOT_ABOVE = 0x1E65;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_CARON_AND_DOT_ABOVE = 0x1E66;
const char32_t LATIN_SMALL_LETTER_S_WITH_CARON_AND_DOT_ABOVE = 0x1E67;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_DOT_BELOW_AND_DOT_ABOVE = 0x1E68;
const char32_t LATIN_SMALL_LETTER_S_WITH_DOT_BELOW_AND_DOT_ABOVE = 0x1E69;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_DOT_ABOVE = 0x1E6A;
const char32_t LATIN_SMALL_LETTER_T_WITH_DOT_ABOVE = 0x1E6B;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_DOT_BELOW = 0x1E6C;
const char32_t LATIN_SMALL_LETTER_T_WITH_DOT_BELOW = 0x1E6D;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_LINE_BELOW = 0x1E6E;
const char32_t LATIN_SMALL_LETTER_T_WITH_LINE_BELOW = 0x1E6F;
const char32_t LATIN_CAPITAL_LETTER_T_WITH_CIRCUMFLEX_BELOW = 0x1E70;
const char32_t LATIN_SMALL_LETTER_T_WITH_CIRCUMFLEX_BELOW = 0x1E71;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DIAERESIS_BELOW = 0x1E72;
const char32_t LATIN_SMALL_LETTER_U_WITH_DIAERESIS_BELOW = 0x1E73;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_TILDE_BELOW = 0x1E74;
const char32_t LATIN_SMALL_LETTER_U_WITH_TILDE_BELOW = 0x1E75;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_CIRCUMFLEX_BELOW = 0x1E76;
const char32_t LATIN_SMALL_LETTER_U_WITH_CIRCUMFLEX_BELOW = 0x1E77;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_TILDE_AND_ACUTE = 0x1E78;
const char32_t LATIN_SMALL_LETTER_U_WITH_TILDE_AND_ACUTE = 0x1E79;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_MACRON_AND_DIAERESIS = 0x1E7A;
const char32_t LATIN_SMALL_LETTER_U_WITH_MACRON_AND_DIAERESIS = 0x1E7B;
const char32_t LATIN_CAPITAL_LETTER_V_WITH_TILDE = 0x1E7C;
const char32_t LATIN_SMALL_LETTER_V_WITH_TILDE = 0x1E7D;
const char32_t LATIN_CAPITAL_LETTER_V_WITH_DOT_BELOW = 0x1E7E;
const char32_t LATIN_SMALL_LETTER_V_WITH_DOT_BELOW = 0x1E7F;
const char32_t LATIN_CAPITAL_LETTER_W_WITH_GRAVE = 0x1E80;
const char32_t LATIN_SMALL_LETTER_W_WITH_GRAVE = 0x1E81;
const char32_t LATIN_CAPITAL_LETTER_W_WITH_ACUTE = 0x1E82;
const char32_t LATIN_SMALL_LETTER_W_WITH_ACUTE = 0x1E83;
const char32_t LATIN_CAPITAL_LETTER_W_WITH_DIAERESIS = 0x1E84;
const char32_t LATIN_SMALL_LETTER_W_WITH_DIAERESIS = 0x1E85;
const char32_t LATIN_CAPITAL_LETTER_W_WITH_DOT_ABOVE = 0x1E86;
const char32_t LATIN_SMALL_LETTER_W_WITH_DOT_ABOVE = 0x1E87;
const char32_t LATIN_CAPITAL_LETTER_W_WITH_DOT_BELOW = 0x1E88;
const char32_t LATIN_SMALL_LETTER_W_WITH_DOT_BELOW = 0x1E89;
const char32_t LATIN_CAPITAL_LETTER_X_WITH_DOT_ABOVE = 0x1E8A;
const char32_t LATIN_SMALL_LETTER_X_WITH_DOT_ABOVE = 0x1E8B;
const char32_t LATIN_CAPITAL_LETTER_X_WITH_DIAERESIS = 0x1E8C;
const char32_t LATIN_SMALL_LETTER_X_WITH_DIAERESIS = 0x1E8D;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_DOT_ABOVE = 0x1E8E;
const char32_t LATIN_SMALL_LETTER_Y_WITH_DOT_ABOVE = 0x1E8F;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_CIRCUMFLEX = 0x1E90;
const char32_t LATIN_SMALL_LETTER_Z_WITH_CIRCUMFLEX = 0x1E91;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_DOT_BELOW = 0x1E92;
const char32_t LATIN_SMALL_LETTER_Z_WITH_DOT_BELOW = 0x1E93;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_LINE_BELOW = 0x1E94;
const char32_t LATIN_SMALL_LETTER_Z_WITH_LINE_BELOW = 0x1E95;
const char32_t LATIN_SMALL_LETTER_H_WITH_LINE_BELOW = 0x1E96;
const char32_t LATIN_SMALL_LETTER_T_WITH_DIAERESIS = 0x1E97;
const char32_t LATIN_SMALL_LETTER_W_WITH_RING_ABOVE = 0x1E98;
const char32_t LATIN_SMALL_LETTER_Y_WITH_RING_ABOVE = 0x1E99;
const char32_t LATIN_SMALL_LETTER_A_WITH_RIGHT_HALF_RING = 0x1E9A;
const char32_t LATIN_SMALL_LETTER_LONG_S_WITH_DOT_ABOVE = 0x1E9B;
const char32_t LATIN_SMALL_LETTER_LONG_S_WITH_DIAGONAL_STROKE = 0x1E9C;
const char32_t LATIN_SMALL_LETTER_LONG_S_WITH_HIGH_STROKE = 0x1E9D;
const char32_t LATIN_CAPITAL_LETTER_SHARP_S = 0x1E9E;
const char32_t LATIN_SMALL_LETTER_DELTA = 0x1E9F;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_DOT_BELOW = 0x1EA0;
const char32_t LATIN_SMALL_LETTER_A_WITH_DOT_BELOW = 0x1EA1;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_HOOK_ABOVE = 0x1EA2;
const char32_t LATIN_SMALL_LETTER_A_WITH_HOOK_ABOVE = 0x1EA3;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX_AND_ACUTE = 0x1EA4;
const char32_t LATIN_SMALL_LETTER_A_WITH_CIRCUMFLEX_AND_ACUTE = 0x1EA5;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX_AND_GRAVE = 0x1EA6;
const char32_t LATIN_SMALL_LETTER_A_WITH_CIRCUMFLEX_AND_GRAVE = 0x1EA7;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX_AND_HOOK_ABOVE = 0x1EA8;
const char32_t LATIN_SMALL_LETTER_A_WITH_CIRCUMFLEX_AND_HOOK_ABOVE = 0x1EA9;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX_AND_TILDE = 0x1EAA;
const char32_t LATIN_SMALL_LETTER_A_WITH_CIRCUMFLEX_AND_TILDE = 0x1EAB;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_CIRCUMFLEX_AND_DOT_BELOW = 0x1EAC;
const char32_t LATIN_SMALL_LETTER_A_WITH_CIRCUMFLEX_AND_DOT_BELOW = 0x1EAD;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_BREVE_AND_ACUTE = 0x1EAE;
const char32_t LATIN_SMALL_LETTER_A_WITH_BREVE_AND_ACUTE = 0x1EAF;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_BREVE_AND_GRAVE = 0x1EB0;
const char32_t LATIN_SMALL_LETTER_A_WITH_BREVE_AND_GRAVE = 0x1EB1;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_BREVE_AND_HOOK_ABOVE = 0x1EB2;
const char32_t LATIN_SMALL_LETTER_A_WITH_BREVE_AND_HOOK_ABOVE = 0x1EB3;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_BREVE_AND_TILDE = 0x1EB4;
const char32_t LATIN_SMALL_LETTER_A_WITH_BREVE_AND_TILDE = 0x1EB5;
const char32_t LATIN_CAPITAL_LETTER_A_WITH_BREVE_AND_DOT_BELOW = 0x1EB6;
const char32_t LATIN_SMALL_LETTER_A_WITH_BREVE_AND_DOT_BELOW = 0x1EB7;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_DOT_BELOW = 0x1EB8;
const char32_t LATIN_SMALL_LETTER_E_WITH_DOT_BELOW = 0x1EB9;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_HOOK_ABOVE = 0x1EBA;
const char32_t LATIN_SMALL_LETTER_E_WITH_HOOK_ABOVE = 0x1EBB;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_TILDE = 0x1EBC;
const char32_t LATIN_SMALL_LETTER_E_WITH_TILDE = 0x1EBD;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX_AND_ACUTE = 0x1EBE;
const char32_t LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX_AND_ACUTE = 0x1EBF;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX_AND_GRAVE = 0x1EC0;
const char32_t LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX_AND_GRAVE = 0x1EC1;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX_AND_HOOK_ABOVE = 0x1EC2;
const char32_t LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX_AND_HOOK_ABOVE = 0x1EC3;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX_AND_TILDE = 0x1EC4;
const char32_t LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX_AND_TILDE = 0x1EC5;
const char32_t LATIN_CAPITAL_LETTER_E_WITH_CIRCUMFLEX_AND_DOT_BELOW = 0x1EC6;
const char32_t LATIN_SMALL_LETTER_E_WITH_CIRCUMFLEX_AND_DOT_BELOW = 0x1EC7;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_HOOK_ABOVE = 0x1EC8;
const char32_t LATIN_SMALL_LETTER_I_WITH_HOOK_ABOVE = 0x1EC9;
const char32_t LATIN_CAPITAL_LETTER_I_WITH_DOT_BELOW = 0x1ECA;
const char32_t LATIN_SMALL_LETTER_I_WITH_DOT_BELOW = 0x1ECB;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_DOT_BELOW = 0x1ECC;
const char32_t LATIN_SMALL_LETTER_O_WITH_DOT_BELOW = 0x1ECD;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_HOOK_ABOVE = 0x1ECE;
const char32_t LATIN_SMALL_LETTER_O_WITH_HOOK_ABOVE = 0x1ECF;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX_AND_ACUTE = 0x1ED0;
const char32_t LATIN_SMALL_LETTER_O_WITH_CIRCUMFLEX_AND_ACUTE = 0x1ED1;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX_AND_GRAVE = 0x1ED2;
const char32_t LATIN_SMALL_LETTER_O_WITH_CIRCUMFLEX_AND_GRAVE = 0x1ED3;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX_AND_HOOK_ABOVE = 0x1ED4;
const char32_t LATIN_SMALL_LETTER_O_WITH_CIRCUMFLEX_AND_HOOK_ABOVE = 0x1ED5;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX_AND_TILDE = 0x1ED6;
const char32_t LATIN_SMALL_LETTER_O_WITH_CIRCUMFLEX_AND_TILDE = 0x1ED7;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_CIRCUMFLEX_AND_DOT_BELOW = 0x1ED8;
const char32_t LATIN_SMALL_LETTER_O_WITH_CIRCUMFLEX_AND_DOT_BELOW = 0x1ED9;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_HORN_AND_ACUTE = 0x1EDA;
const char32_t LATIN_SMALL_LETTER_O_WITH_HORN_AND_ACUTE = 0x1EDB;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_HORN_AND_GRAVE = 0x1EDC;
const char32_t LATIN_SMALL_LETTER_O_WITH_HORN_AND_GRAVE = 0x1EDD;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_HORN_AND_HOOK_ABOVE = 0x1EDE;
const char32_t LATIN_SMALL_LETTER_O_WITH_HORN_AND_HOOK_ABOVE = 0x1EDF;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_HORN_AND_TILDE = 0x1EE0;
const char32_t LATIN_SMALL_LETTER_O_WITH_HORN_AND_TILDE = 0x1EE1;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_HORN_AND_DOT_BELOW = 0x1EE2;
const char32_t LATIN_SMALL_LETTER_O_WITH_HORN_AND_DOT_BELOW = 0x1EE3;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_DOT_BELOW = 0x1EE4;
const char32_t LATIN_SMALL_LETTER_U_WITH_DOT_BELOW = 0x1EE5;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_HOOK_ABOVE = 0x1EE6;
const char32_t LATIN_SMALL_LETTER_U_WITH_HOOK_ABOVE = 0x1EE7;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_HORN_AND_ACUTE = 0x1EE8;
const char32_t LATIN_SMALL_LETTER_U_WITH_HORN_AND_ACUTE = 0x1EE9;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_HORN_AND_GRAVE = 0x1EEA;
const char32_t LATIN_SMALL_LETTER_U_WITH_HORN_AND_GRAVE = 0x1EEB;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_HORN_AND_HOOK_ABOVE = 0x1EEC;
const char32_t LATIN_SMALL_LETTER_U_WITH_HORN_AND_HOOK_ABOVE = 0x1EED;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_HORN_AND_TILDE = 0x1EEE;
const char32_t LATIN_SMALL_LETTER_U_WITH_HORN_AND_TILDE = 0x1EEF;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_HORN_AND_DOT_BELOW = 0x1EF0;
const char32_t LATIN_SMALL_LETTER_U_WITH_HORN_AND_DOT_BELOW = 0x1EF1;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_GRAVE = 0x1EF2;
const char32_t LATIN_SMALL_LETTER_Y_WITH_GRAVE = 0x1EF3;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_DOT_BELOW = 0x1EF4;
const char32_t LATIN_SMALL_LETTER_Y_WITH_DOT_BELOW = 0x1EF5;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_HOOK_ABOVE = 0x1EF6;
const char32_t LATIN_SMALL_LETTER_Y_WITH_HOOK_ABOVE = 0x1EF7;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_TILDE = 0x1EF8;
const char32_t LATIN_SMALL_LETTER_Y_WITH_TILDE = 0x1EF9;
const char32_t LATIN_CAPITAL_LETTER_MIDDLE_WELSH_LL = 0x1EFA;
const char32_t LATIN_SMALL_LETTER_MIDDLE_WELSH_LL = 0x1EFB;
const char32_t LATIN_CAPITAL_LETTER_MIDDLE_WELSH_V = 0x1EFC;
const char32_t LATIN_SMALL_LETTER_MIDDLE_WELSH_V = 0x1EFD;
const char32_t LATIN_CAPITAL_LETTER_Y_WITH_LOOP = 0x1EFE;
const char32_t LATIN_SMALL_LETTER_Y_WITH_LOOP = 0x1EFF;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI = 0x1F00;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA = 0x1F01;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI_AND_VARIA = 0x1F02;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA_AND_VARIA = 0x1F03;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI_AND_OXIA = 0x1F04;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA_AND_OXIA = 0x1F05;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI_AND_PERISPOMENI = 0x1F06;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA_AND_PERISPOMENI = 0x1F07;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI = 0x1F08;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA = 0x1F09;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI_AND_VARIA = 0x1F0A;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA_AND_VARIA = 0x1F0B;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI_AND_OXIA = 0x1F0C;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA_AND_OXIA = 0x1F0D;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI_AND_PERISPOMENI = 0x1F0E;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA_AND_PERISPOMENI = 0x1F0F;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_PSILI = 0x1F10;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_DASIA = 0x1F11;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_PSILI_AND_VARIA = 0x1F12;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_DASIA_AND_VARIA = 0x1F13;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_PSILI_AND_OXIA = 0x1F14;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_DASIA_AND_OXIA = 0x1F15;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_PSILI = 0x1F18;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_DASIA = 0x1F19;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_PSILI_AND_VARIA = 0x1F1A;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_DASIA_AND_VARIA = 0x1F1B;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_PSILI_AND_OXIA = 0x1F1C;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_DASIA_AND_OXIA = 0x1F1D;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI = 0x1F20;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA = 0x1F21;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI_AND_VARIA = 0x1F22;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA_AND_VARIA = 0x1F23;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI_AND_OXIA = 0x1F24;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA_AND_OXIA = 0x1F25;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI_AND_PERISPOMENI = 0x1F26;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA_AND_PERISPOMENI = 0x1F27;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI = 0x1F28;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA = 0x1F29;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI_AND_VARIA = 0x1F2A;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA_AND_VARIA = 0x1F2B;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI_AND_OXIA = 0x1F2C;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA_AND_OXIA = 0x1F2D;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI_AND_PERISPOMENI = 0x1F2E;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA_AND_PERISPOMENI = 0x1F2F;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_PSILI = 0x1F30;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DASIA = 0x1F31;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_PSILI_AND_VARIA = 0x1F32;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DASIA_AND_VARIA = 0x1F33;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_PSILI_AND_OXIA = 0x1F34;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DASIA_AND_OXIA = 0x1F35;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_PSILI_AND_PERISPOMENI = 0x1F36;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DASIA_AND_PERISPOMENI = 0x1F37;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_PSILI = 0x1F38;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_DASIA = 0x1F39;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_PSILI_AND_VARIA = 0x1F3A;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_DASIA_AND_VARIA = 0x1F3B;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_PSILI_AND_OXIA = 0x1F3C;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_DASIA_AND_OXIA = 0x1F3D;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_PSILI_AND_PERISPOMENI = 0x1F3E;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_DASIA_AND_PERISPOMENI = 0x1F3F;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_PSILI = 0x1F40;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_DASIA = 0x1F41;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_PSILI_AND_VARIA = 0x1F42;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_DASIA_AND_VARIA = 0x1F43;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_PSILI_AND_OXIA = 0x1F44;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_DASIA_AND_OXIA = 0x1F45;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_PSILI = 0x1F48;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_DASIA = 0x1F49;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_PSILI_AND_VARIA = 0x1F4A;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_DASIA_AND_VARIA = 0x1F4B;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_PSILI_AND_OXIA = 0x1F4C;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_DASIA_AND_OXIA = 0x1F4D;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_PSILI = 0x1F50;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DASIA = 0x1F51;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_PSILI_AND_VARIA = 0x1F52;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DASIA_AND_VARIA = 0x1F53;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_PSILI_AND_OXIA = 0x1F54;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DASIA_AND_OXIA = 0x1F55;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_PSILI_AND_PERISPOMENI = 0x1F56;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DASIA_AND_PERISPOMENI = 0x1F57;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_DASIA = 0x1F59;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_DASIA_AND_VARIA = 0x1F5B;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_DASIA_AND_OXIA = 0x1F5D;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_DASIA_AND_PERISPOMENI = 0x1F5F;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI = 0x1F60;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA = 0x1F61;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI_AND_VARIA = 0x1F62;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA_AND_VARIA = 0x1F63;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI_AND_OXIA = 0x1F64;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA_AND_OXIA = 0x1F65;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI_AND_PERISPOMENI = 0x1F66;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA_AND_PERISPOMENI = 0x1F67;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI = 0x1F68;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA = 0x1F69;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI_AND_VARIA = 0x1F6A;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA_AND_VARIA = 0x1F6B;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI_AND_OXIA = 0x1F6C;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA_AND_OXIA = 0x1F6D;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI_AND_PERISPOMENI = 0x1F6E;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA_AND_PERISPOMENI = 0x1F6F;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_VARIA = 0x1F70;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_OXIA = 0x1F71;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_VARIA = 0x1F72;
const char32_t GREEK_SMALL_LETTER_EPSILON_WITH_OXIA = 0x1F73;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_VARIA = 0x1F74;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_OXIA = 0x1F75;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_VARIA = 0x1F76;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_OXIA = 0x1F77;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_VARIA = 0x1F78;
const char32_t GREEK_SMALL_LETTER_OMICRON_WITH_OXIA = 0x1F79;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_VARIA = 0x1F7A;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_OXIA = 0x1F7B;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_VARIA = 0x1F7C;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_OXIA = 0x1F7D;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI_AND_YPOGEGRAMMENI = 0x1F80;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA_AND_YPOGEGRAMMENI = 0x1F81;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI_AND_VARIA_AND_YPOGEGRAMMENI = 0x1F82;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA_AND_VARIA_AND_YPOGEGRAMMENI = 0x1F83;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI_AND_OXIA_AND_YPOGEGRAMMENI = 0x1F84;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA_AND_OXIA_AND_YPOGEGRAMMENI = 0x1F85;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PSILI_AND_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1F86;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_DASIA_AND_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1F87;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI_AND_PROSGEGRAMMENI = 0x1F88;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA_AND_PROSGEGRAMMENI = 0x1F89;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI_AND_VARIA_AND_PROSGEGRAMMENI = 0x1F8A;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA_AND_VARIA_AND_PROSGEGRAMMENI = 0x1F8B;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI_AND_OXIA_AND_PROSGEGRAMMENI = 0x1F8C;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA_AND_OXIA_AND_PROSGEGRAMMENI = 0x1F8D;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PSILI_AND_PERISPOMENI_AND_PROSGEGRAMMENI = 0x1F8E;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_DASIA_AND_PERISPOMENI_AND_PROSGEGRAMMENI = 0x1F8F;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI_AND_YPOGEGRAMMENI = 0x1F90;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA_AND_YPOGEGRAMMENI = 0x1F91;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI_AND_VARIA_AND_YPOGEGRAMMENI = 0x1F92;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA_AND_VARIA_AND_YPOGEGRAMMENI = 0x1F93;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI_AND_OXIA_AND_YPOGEGRAMMENI = 0x1F94;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA_AND_OXIA_AND_YPOGEGRAMMENI = 0x1F95;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PSILI_AND_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1F96;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_DASIA_AND_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1F97;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI_AND_PROSGEGRAMMENI = 0x1F98;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA_AND_PROSGEGRAMMENI = 0x1F99;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI_AND_VARIA_AND_PROSGEGRAMMENI = 0x1F9A;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA_AND_VARIA_AND_PROSGEGRAMMENI = 0x1F9B;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI_AND_OXIA_AND_PROSGEGRAMMENI = 0x1F9C;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA_AND_OXIA_AND_PROSGEGRAMMENI = 0x1F9D;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PSILI_AND_PERISPOMENI_AND_PROSGEGRAMMENI = 0x1F9E;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_DASIA_AND_PERISPOMENI_AND_PROSGEGRAMMENI = 0x1F9F;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI_AND_YPOGEGRAMMENI = 0x1FA0;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA_AND_YPOGEGRAMMENI = 0x1FA1;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI_AND_VARIA_AND_YPOGEGRAMMENI = 0x1FA2;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA_AND_VARIA_AND_YPOGEGRAMMENI = 0x1FA3;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI_AND_OXIA_AND_YPOGEGRAMMENI = 0x1FA4;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA_AND_OXIA_AND_YPOGEGRAMMENI = 0x1FA5;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PSILI_AND_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1FA6;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_DASIA_AND_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1FA7;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI_AND_PROSGEGRAMMENI = 0x1FA8;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA_AND_PROSGEGRAMMENI = 0x1FA9;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI_AND_VARIA_AND_PROSGEGRAMMENI = 0x1FAA;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA_AND_VARIA_AND_PROSGEGRAMMENI = 0x1FAB;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI_AND_OXIA_AND_PROSGEGRAMMENI = 0x1FAC;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA_AND_OXIA_AND_PROSGEGRAMMENI = 0x1FAD;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PSILI_AND_PERISPOMENI_AND_PROSGEGRAMMENI = 0x1FAE;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_DASIA_AND_PERISPOMENI_AND_PROSGEGRAMMENI = 0x1FAF;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_VRACHY = 0x1FB0;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_MACRON = 0x1FB1;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_VARIA_AND_YPOGEGRAMMENI = 0x1FB2;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_YPOGEGRAMMENI = 0x1FB3;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_OXIA_AND_YPOGEGRAMMENI = 0x1FB4;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PERISPOMENI = 0x1FB6;
const char32_t GREEK_SMALL_LETTER_ALPHA_WITH_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1FB7;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_VRACHY = 0x1FB8;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_MACRON = 0x1FB9;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_VARIA = 0x1FBA;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_OXIA = 0x1FBB;
const char32_t GREEK_CAPITAL_LETTER_ALPHA_WITH_PROSGEGRAMMENI = 0x1FBC;
const char32_t GREEK_KORONIS = 0x1FBD;
const char32_t GREEK_PROSGEGRAMMENI = 0x1FBE;
const char32_t GREEK_PSILI = 0x1FBF;
const char32_t GREEK_PERISPOMENI = 0x1FC0;
const char32_t GREEK_DIALYTIKA_AND_PERISPOMENI = 0x1FC1;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_VARIA_AND_YPOGEGRAMMENI = 0x1FC2;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_YPOGEGRAMMENI = 0x1FC3;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_OXIA_AND_YPOGEGRAMMENI = 0x1FC4;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PERISPOMENI = 0x1FC6;
const char32_t GREEK_SMALL_LETTER_ETA_WITH_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1FC7;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_VARIA = 0x1FC8;
const char32_t GREEK_CAPITAL_LETTER_EPSILON_WITH_OXIA = 0x1FC9;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_VARIA = 0x1FCA;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_OXIA = 0x1FCB;
const char32_t GREEK_CAPITAL_LETTER_ETA_WITH_PROSGEGRAMMENI = 0x1FCC;
const char32_t GREEK_PSILI_AND_VARIA = 0x1FCD;
const char32_t GREEK_PSILI_AND_OXIA = 0x1FCE;
const char32_t GREEK_PSILI_AND_PERISPOMENI = 0x1FCF;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_VRACHY = 0x1FD0;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_MACRON = 0x1FD1;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DIALYTIKA_AND_VARIA = 0x1FD2;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DIALYTIKA_AND_OXIA = 0x1FD3;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_PERISPOMENI = 0x1FD6;
const char32_t GREEK_SMALL_LETTER_IOTA_WITH_DIALYTIKA_AND_PERISPOMENI = 0x1FD7;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_VRACHY = 0x1FD8;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_MACRON = 0x1FD9;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_VARIA = 0x1FDA;
const char32_t GREEK_CAPITAL_LETTER_IOTA_WITH_OXIA = 0x1FDB;
const char32_t GREEK_DASIA_AND_VARIA = 0x1FDD;
const char32_t GREEK_DASIA_AND_OXIA = 0x1FDE;
const char32_t GREEK_DASIA_AND_PERISPOMENI = 0x1FDF;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_VRACHY = 0x1FE0;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_MACRON = 0x1FE1;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DIALYTIKA_AND_VARIA = 0x1FE2;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DIALYTIKA_AND_OXIA = 0x1FE3;
const char32_t GREEK_SMALL_LETTER_RHO_WITH_PSILI = 0x1FE4;
const char32_t GREEK_SMALL_LETTER_RHO_WITH_DASIA = 0x1FE5;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_PERISPOMENI = 0x1FE6;
const char32_t GREEK_SMALL_LETTER_UPSILON_WITH_DIALYTIKA_AND_PERISPOMENI = 0x1FE7;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_VRACHY = 0x1FE8;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_MACRON = 0x1FE9;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_VARIA = 0x1FEA;
const char32_t GREEK_CAPITAL_LETTER_UPSILON_WITH_OXIA = 0x1FEB;
const char32_t GREEK_CAPITAL_LETTER_RHO_WITH_DASIA = 0x1FEC;
const char32_t GREEK_DIALYTIKA_AND_VARIA = 0x1FED;
const char32_t GREEK_DIALYTIKA_AND_OXIA = 0x1FEE;
const char32_t GREEK_VARIA = 0x1FEF;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_VARIA_AND_YPOGEGRAMMENI = 0x1FF2;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_YPOGEGRAMMENI = 0x1FF3;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_OXIA_AND_YPOGEGRAMMENI = 0x1FF4;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PERISPOMENI = 0x1FF6;
const char32_t GREEK_SMALL_LETTER_OMEGA_WITH_PERISPOMENI_AND_YPOGEGRAMMENI = 0x1FF7;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_VARIA = 0x1FF8;
const char32_t GREEK_CAPITAL_LETTER_OMICRON_WITH_OXIA = 0x1FF9;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_VARIA = 0x1FFA;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_OXIA = 0x1FFB;
const char32_t GREEK_CAPITAL_LETTER_OMEGA_WITH_PROSGEGRAMMENI = 0x1FFC;
const char32_t GREEK_OXIA = 0x1FFD;
const char32_t GREEK_DASIA = 0x1FFE;
const char32_t EN_QUAD = 0x2000;
const char32_t EM_QUAD = 0x2001;
const char32_t EN_SPACE = 0x2002;
const char32_t EM_SPACE = 0x2003;
const char32_t THREE_PER_EM_SPACE = 0x2004;
const char32_t FOUR_PER_EM_SPACE = 0x2005;
const char32_t SIX_PER_EM_SPACE = 0x2006;
const char32_t FIGURE_SPACE = 0x2007;
const char32_t PUNCTUATION_SPACE = 0x2008;
const char32_t THIN_SPACE = 0x2009;
const char32_t HAIR_SPACE = 0x200A;
const char32_t ZERO_WIDTH_SPACE = 0x200B;
const char32_t ZERO_WIDTH_NON_JOINER = 0x200C;
const char32_t ZERO_WIDTH_JOINER = 0x200D;
const char32_t LEFT_TO_RIGHT_MARK = 0x200E;
const char32_t RIGHT_TO_LEFT_MARK = 0x200F;
const char32_t HYPHEN = 0x2010;
const char32_t NON_BREAKING_HYPHEN = 0x2011;
const char32_t FIGURE_DASH = 0x2012;
const char32_t EN_DASH = 0x2013;
const char32_t EM_DASH = 0x2014;
const char32_t HORIZONTAL_BAR = 0x2015;
const char32_t DOUBLE_VERTICAL_LINE = 0x2016;
const char32_t DOUBLE_LOW_LINE = 0x2017;
const char32_t LEFT_SINGLE_QUOTATION_MARK = 0x2018;
const char32_t RIGHT_SINGLE_QUOTATION_MARK = 0x2019;
const char32_t SINGLE_LOW_9_QUOTATION_MARK = 0x201A;
const char32_t SINGLE_HIGH_REVERSED_9_QUOTATION_MARK = 0x201B;
const char32_t LEFT_DOUBLE_QUOTATION_MARK = 0x201C;
const char32_t RIGHT_DOUBLE_QUOTATION_MARK = 0x201D;
const char32_t DOUBLE_LOW_9_QUOTATION_MARK = 0x201E;
const char32_t DOUBLE_HIGH_REVERSED_9_QUOTATION_MARK = 0x201F;
const char32_t DAGGER = 0x2020;
const char32_t DOUBLE_DAGGER = 0x2021;
const char32_t BULLET = 0x2022;
const char32_t TRIANGULAR_BULLET = 0x2023;
const char32_t ONE_DOT_LEADER = 0x2024;
const char32_t TWO_DOT_LEADER = 0x2025;
const char32_t HORIZONTAL_ELLIPSIS = 0x2026;
const char32_t HYPHENATION_POINT = 0x2027;
const char32_t LINE_SEPARATOR = 0x2028;
const char32_t PARAGRAPH_SEPARATOR = 0x2029;
const char32_t LEFT_TO_RIGHT_EMBEDDING = 0x202A;
const char32_t RIGHT_TO_LEFT_EMBEDDING = 0x202B;
const char32_t POP_DIRECTIONAL_FORMATTING = 0x202C;
const char32_t LEFT_TO_RIGHT_OVERRIDE = 0x202D;
const char32_t RIGHT_TO_LEFT_OVERRIDE = 0x202E;
const char32_t NARROW_NO_BREAK_SPACE = 0x202F;
const char32_t PER_MILLE_SIGN = 0x2030;
const char32_t PER_TEN_THOUSAND_SIGN = 0x2031;
const char32_t PRIME = 0x2032;
const char32_t DOUBLE_PRIME = 0x2033;
const char32_t TRIPLE_PRIME = 0x2034;
const char32_t REVERSED_PRIME = 0x2035;
const char32_t REVERSED_DOUBLE_PRIME = 0x2036;
const char32_t REVERSED_TRIPLE_PRIME = 0x2037;
const char32_t CARET = 0x2038;
const char32_t SINGLE_LEFT_POINTING_ANGLE_QUOTATION_MARK = 0x2039;
const char32_t SINGLE_RIGHT_POINTING_ANGLE_QUOTATION_MARK = 0x203A;
const char32_t REFERENCE_MARK = 0x203B;
const char32_t DOUBLE_EXCLAMATION_MARK = 0x203C;
const char32_t INTERROBANG = 0x203D;
const char32_t OVERLINE = 0x203E;
const char32_t UNDERTIE = 0x203F;
const char32_t CHARACTER_TIE = 0x2040;
const char32_t CARET_INSERTION_POINT = 0x2041;
const char32_t ASTERISM = 0x2042;
const char32_t HYPHEN_BULLET = 0x2043;
const char32_t FRACTION_SLASH = 0x2044;
const char32_t LEFT_SQUARE_BRACKET_WITH_QUILL = 0x2045;
const char32_t RIGHT_SQUARE_BRACKET_WITH_QUILL = 0x2046;
const char32_t DOUBLE_QUESTION_MARK = 0x2047;
const char32_t QUESTION_EXCLAMATION_MARK = 0x2048;
const char32_t EXCLAMATION_QUESTION_MARK = 0x2049;
const char32_t TIRONIAN_SIGN_ET = 0x204A;
const char32_t REVERSED_PILCROW_SIGN = 0x204B;
const char32_t BLACK_LEFTWARDS_BULLET = 0x204C;
const char32_t BLACK_RIGHTWARDS_BULLET = 0x204D;
const char32_t LOW_ASTERISK = 0x204E;
const char32_t REVERSED_SEMICOLON = 0x204F;
const char32_t CLOSE_UP = 0x2050;
const char32_t TWO_ASTERISKS_ALIGNED_VERTICALLY = 0x2051;
const char32_t COMMERCIAL_MINUS_SIGN = 0x2052;
const char32_t SWUNG_DASH = 0x2053;
const char32_t INVERTED_UNDERTIE = 0x2054;
const char32_t FLOWER_PUNCTUATION_MARK = 0x2055;
const char32_t THREE_DOT_PUNCTUATION = 0x2056;
const char32_t QUADRUPLE_PRIME = 0x2057;
const char32_t FOUR_DOT_PUNCTUATION = 0x2058;
const char32_t FIVE_DOT_PUNCTUATION = 0x2059;
const char32_t TWO_DOT_PUNCTUATION = 0x205A;
const char32_t FOUR_DOT_MARK = 0x205B;
const char32_t DOTTED_CROSS = 0x205C;
const char32_t TRICOLON = 0x205D;
const char32_t VERTICAL_FOUR_DOTS = 0x205E;
const char32_t MEDIUM_MATHEMATICAL_SPACE = 0x205F;
const char32_t WORD_JOINER = 0x2060;
const char32_t FUNCTION_APPLICATION = 0x2061;
const char32_t INVISIBLE_TIMES = 0x2062;
const char32_t INVISIBLE_SEPARATOR = 0x2063;
const char32_t INVISIBLE_PLUS = 0x2064;
const char32_t LEFT_TO_RIGHT_ISOLATE = 0x2066;
const char32_t RIGHT_TO_LEFT_ISOLATE = 0x2067;
const char32_t FIRST_STRONG_ISOLATE = 0x2068;
const char32_t POP_DIRECTIONAL_ISOLATE = 0x2069;
const char32_t INHIBIT_SYMMETRIC_SWAPPING = 0x206A;
const char32_t ACTIVATE_SYMMETRIC_SWAPPING = 0x206B;
const char32_t INHIBIT_ARABIC_FORM_SHAPING = 0x206C;
const char32_t ACTIVATE_ARABIC_FORM_SHAPING = 0x206D;
const char32_t NATIONAL_DIGIT_SHAPES = 0x206E;
const char32_t NOMINAL_DIGIT_SHAPES = 0x206F;
const char32_t SUPERSCRIPT_ZERO = 0x2070;
const char32_t SUPERSCRIPT_LATIN_SMALL_LETTER_I = 0x2071;
const char32_t SUPERSCRIPT_FOUR = 0x2074;
const char32_t SUPERSCRIPT_FIVE = 0x2075;
const char32_t SUPERSCRIPT_SIX = 0x2076;
const char32_t SUPERSCRIPT_SEVEN = 0x2077;
const char32_t SUPERSCRIPT_EIGHT = 0x2078;
const char32_t SUPERSCRIPT_NINE = 0x2079;
const char32_t SUPERSCRIPT_PLUS_SIGN = 0x207A;
const char32_t SUPERSCRIPT_MINUS = 0x207B;
const char32_t SUPERSCRIPT_EQUALS_SIGN = 0x207C;
const char32_t SUPERSCRIPT_LEFT_PARENTHESIS = 0x207D;
const char32_t SUPERSCRIPT_RIGHT_PARENTHESIS = 0x207E;
const char32_t SUPERSCRIPT_LATIN_SMALL_LETTER_N = 0x207F;
const char32_t SUBSCRIPT_ZERO = 0x2080;
const char32_t SUBSCRIPT_ONE = 0x2081;
const char32_t SUBSCRIPT_TWO = 0x2082;
const char32_t SUBSCRIPT_THREE = 0x2083;
const char32_t SUBSCRIPT_FOUR = 0x2084;
const char32_t SUBSCRIPT_FIVE = 0x2085;
const char32_t SUBSCRIPT_SIX = 0x2086;
const char32_t SUBSCRIPT_SEVEN = 0x2087;
const char32_t SUBSCRIPT_EIGHT = 0x2088;
const char32_t SUBSCRIPT_NINE = 0x2089;
const char32_t SUBSCRIPT_PLUS_SIGN = 0x208A;
const char32_t SUBSCRIPT_MINUS = 0x208B;
const char32_t SUBSCRIPT_EQUALS_SIGN = 0x208C;
const char32_t SUBSCRIPT_LEFT_PARENTHESIS = 0x208D;
const char32_t SUBSCRIPT_RIGHT_PARENTHESIS = 0x208E;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_A = 0x2090;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_E = 0x2091;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_O = 0x2092;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_X = 0x2093;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_SCHWA = 0x2094;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_H = 0x2095;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_K = 0x2096;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_L = 0x2097;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_M = 0x2098;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_N = 0x2099;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_P = 0x209A;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_S = 0x209B;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_T = 0x209C;
const char32_t EURO_CURRENCY_SIGN = 0x20A0;
const char32_t COLON_SIGN = 0x20A1;
const char32_t CRUZEIRO_SIGN = 0x20A2;
const char32_t FRENCH_FRANC_SIGN = 0x20A3;
const char32_t LIRA_SIGN = 0x20A4;
const char32_t MILL_SIGN = 0x20A5;
const char32_t NAIRA_SIGN = 0x20A6;
const char32_t PESETA_SIGN = 0x20A7;
const char32_t RUPEE_SIGN = 0x20A8;
const char32_t WON_SIGN = 0x20A9;
const char32_t NEW_SHEQEL_SIGN = 0x20AA;
const char32_t DONG_SIGN = 0x20AB;
const char32_t EURO_SIGN = 0x20AC;
const char32_t KIP_SIGN = 0x20AD;
const char32_t TUGRIK_SIGN = 0x20AE;
const char32_t DRACHMA_SIGN = 0x20AF;
const char32_t GERMAN_PENNY_SIGN = 0x20B0;
const char32_t PESO_SIGN = 0x20B1;
const char32_t GUARANI_SIGN = 0x20B2;
const char32_t AUSTRAL_SIGN = 0x20B3;
const char32_t HRYVNIA_SIGN = 0x20B4;
const char32_t CEDI_SIGN = 0x20B5;
const char32_t LIVRE_TOURNOIS_SIGN = 0x20B6;
const char32_t SPESMILO_SIGN = 0x20B7;
const char32_t TENGE_SIGN = 0x20B8;
const char32_t INDIAN_RUPEE_SIGN = 0x20B9;
const char32_t TURKISH_LIRA_SIGN = 0x20BA;
const char32_t NORDIC_MARK_SIGN = 0x20BB;
const char32_t MANAT_SIGN = 0x20BC;
const char32_t RUBLE_SIGN = 0x20BD;
const char32_t LARI_SIGN = 0x20BE;
const char32_t BITCOIN_SIGN = 0x20BF;
const char32_t SOM_SIGN = 0x20C0;
const char32_t SAUDI_RIYAL_SIGN = 0x20C1;
const char32_t COMBINING_LEFT_HARPOON_ABOVE = 0x20D0;
const char32_t COMBINING_RIGHT_HARPOON_ABOVE = 0x20D1;
const char32_t COMBINING_LONG_VERTICAL_LINE_OVERLAY = 0x20D2;
const char32_t COMBINING_SHORT_VERTICAL_LINE_OVERLAY = 0x20D3;
const char32_t COMBINING_ANTICLOCKWISE_ARROW_ABOVE = 0x20D4;
const char32_t COMBINING_CLOCKWISE_ARROW_ABOVE = 0x20D5;
const char32_t COMBINING_LEFT_ARROW_ABOVE = 0x20D6;
const char32_t COMBINING_RIGHT_ARROW_ABOVE = 0x20D7;
const char32_t COMBINING_RING_OVERLAY = 0x20D8;
const char32_t COMBINING_CLOCKWISE_RING_OVERLAY = 0x20D9;
const char32_t COMBINING_ANTICLOCKWISE_RING_OVERLAY = 0x20DA;
const char32_t COMBINING_THREE_DOTS_ABOVE = 0x20DB;
const char32_t COMBINING_FOUR_DOTS_ABOVE = 0x20DC;
const char32_t COMBINING_ENCLOSING_CIRCLE = 0x20DD;
const char32_t COMBINING_ENCLOSING_SQUARE = 0x20DE;
const char32_t COMBINING_ENCLOSING_DIAMOND = 0x20DF;
const char32_t COMBINING_ENCLOSING_CIRCLE_BACKSLASH = 0x20E0;
const char32_t COMBINING_LEFT_RIGHT_ARROW_ABOVE = 0x20E1;
const char32_t COMBINING_ENCLOSING_SCREEN = 0x20E2;
const char32_t COMBINING_ENCLOSING_KEYCAP = 0x20E3;
const char32_t COMBINING_ENCLOSING_UPWARD_POINTING_TRIANGLE = 0x20E4;
const char32_t COMBINING_REVERSE_SOLIDUS_OVERLAY = 0x20E5;
const char32_t COMBINING_DOUBLE_VERTICAL_STROKE_OVERLAY = 0x20E6;
const char32_t COMBINING_ANNUITY_SYMBOL = 0x20E7;
const char32_t COMBINING_TRIPLE_UNDERDOT = 0x20E8;
const char32_t COMBINING_WIDE_BRIDGE_ABOVE = 0x20E9;
const char32_t COMBINING_LEFTWARDS_ARROW_OVERLAY = 0x20EA;
const char32_t COMBINING_LONG_DOUBLE_SOLIDUS_OVERLAY = 0x20EB;
const char32_t COMBINING_RIGHTWARDS_HARPOON_WITH_BARB_DOWNWARDS = 0x20EC;
const char32_t COMBINING_LEFTWARDS_HARPOON_WITH_BARB_DOWNWARDS = 0x20ED;
const char32_t COMBINING_LEFT_ARROW_BELOW = 0x20EE;
const char32_t COMBINING_RIGHT_ARROW_BELOW = 0x20EF;
const char32_t COMBINING_ASTERISK_ABOVE = 0x20F0;
const char32_t ACCOUNT_OF = 0x2100;
const char32_t ADDRESSED_TO_THE_SUBJECT = 0x2101;
const char32_t DOUBLE_STRUCK_CAPITAL_C = 0x2102;
const char32_t DEGREE_CELSIUS = 0x2103;
const char32_t CENTRE_LINE_SYMBOL = 0x2104;
const char32_t CARE_OF = 0x2105;
const char32_t CADA_UNA = 0x2106;
const char32_t EULER_CONSTANT = 0x2107;
const char32_t SCRUPLE = 0x2108;
const char32_t DEGREE_FAHRENHEIT = 0x2109;
const char32_t SCRIPT_SMALL_G = 0x210A;
const char32_t SCRIPT_CAPITAL_H = 0x210B;
const char32_t BLACK_LETTER_CAPITAL_H = 0x210C;
const char32_t DOUBLE_STRUCK_CAPITAL_H = 0x210D;
const char32_t PLANCK_CONSTANT = 0x210E;
const char32_t PLANCK_CONSTANT_OVER_TWO_PI = 0x210F;
const char32_t SCRIPT_CAPITAL_I = 0x2110;
const char32_t BLACK_LETTER_CAPITAL_I = 0x2111;
const char32_t SCRIPT_CAPITAL_L = 0x2112;
const char32_t SCRIPT_SMALL_L = 0x2113;
const char32_t L_B_BAR_SYMBOL = 0x2114;
const char32_t DOUBLE_STRUCK_CAPITAL_N = 0x2115;
const char32_t NUMERO_SIGN = 0x2116;
const char32_t SOUND_RECORDING_COPYRIGHT = 0x2117;
const char32_t SCRIPT_CAPITAL_P = 0x2118;
const char32_t DOUBLE_STRUCK_CAPITAL_P = 0x2119;
const char32_t DOUBLE_STRUCK_CAPITAL_Q = 0x211A;
const char32_t SCRIPT_CAPITAL_R = 0x211B;
const char32_t BLACK_LETTER_CAPITAL_R = 0x211C;
const char32_t DOUBLE_STRUCK_CAPITAL_R = 0x211D;
const char32_t PRESCRIPTION_TAKE = 0x211E;
const char32_t RESPONSE = 0x211F;
const char32_t SERVICE_MARK = 0x2120;
const char32_t TELEPHONE_SIGN = 0x2121;
const char32_t TRADE_MARK_SIGN = 0x2122;
const char32_t VERSICLE = 0x2123;
const char32_t DOUBLE_STRUCK_CAPITAL_Z = 0x2124;
const char32_t OUNCE_SIGN = 0x2125;
const char32_t OHM_SIGN = 0x2126;
const char32_t INVERTED_OHM_SIGN = 0x2127;
const char32_t BLACK_LETTER_CAPITAL_Z = 0x2128;
const char32_t TURNED_GREEK_SMALL_LETTER_IOTA = 0x2129;
const char32_t KELVIN_SIGN = 0x212A;
const char32_t ANGSTROM_SIGN = 0x212B;
const char32_t SCRIPT_CAPITAL_B = 0x212C;
const char32_t BLACK_LETTER_CAPITAL_C = 0x212D;
const char32_t ESTIMATED_SYMBOL = 0x212E;
const char32_t SCRIPT_SMALL_E = 0x212F;
const char32_t SCRIPT_CAPITAL_E = 0x2130;
const char32_t SCRIPT_CAPITAL_F = 0x2131;
const char32_t TURNED_CAPITAL_F = 0x2132;
const char32_t SCRIPT_CAPITAL_M = 0x2133;
const char32_t SCRIPT_SMALL_O = 0x2134;
const char32_t ALEF_SYMBOL = 0x2135;
const char32_t BET_SYMBOL = 0x2136;
const char32_t GIMEL_SYMBOL = 0x2137;
const char32_t DALET_SYMBOL = 0x2138;
const char32_t INFORMATION_SOURCE = 0x2139;
const char32_t ROTATED_CAPITAL_Q = 0x213A;
const char32_t FACSIMILE_SIGN = 0x213B;
const char32_t DOUBLE_STRUCK_SMALL_PI = 0x213C;
const char32_t DOUBLE_STRUCK_SMALL_GAMMA = 0x213D;
const char32_t DOUBLE_STRUCK_CAPITAL_GAMMA = 0x213E;
const char32_t DOUBLE_STRUCK_CAPITAL_PI = 0x213F;
const char32_t DOUBLE_STRUCK_N_ARY_SUMMATION = 0x2140;
const char32_t TURNED_SANS_SERIF_CAPITAL_G = 0x2141;
const char32_t TURNED_SANS_SERIF_CAPITAL_L = 0x2142;
const char32_t REVERSED_SANS_SERIF_CAPITAL_L = 0x2143;
const char32_t TURNED_SANS_SERIF_CAPITAL_Y = 0x2144;
const char32_t DOUBLE_STRUCK_ITALIC_CAPITAL_D = 0x2145;
const char32_t DOUBLE_STRUCK_ITALIC_SMALL_D = 0x2146;
const char32_t DOUBLE_STRUCK_ITALIC_SMALL_E = 0x2147;
const char32_t DOUBLE_STRUCK_ITALIC_SMALL_I = 0x2148;
const char32_t DOUBLE_STRUCK_ITALIC_SMALL_J = 0x2149;
const char32_t PROPERTY_LINE = 0x214A;
const char32_t TURNED_AMPERSAND = 0x214B;
const char32_t PER_SIGN = 0x214C;
const char32_t AKTIESELSKAB = 0x214D;
const char32_t TURNED_SMALL_F = 0x214E;
const char32_t SYMBOL_FOR_SAMARITAN_SOURCE = 0x214F;
const char32_t VULGAR_FRACTION_ONE_SEVENTH = 0x2150;
const char32_t VULGAR_FRACTION_ONE_NINTH = 0x2151;
const char32_t VULGAR_FRACTION_ONE_TENTH = 0x2152;
const char32_t VULGAR_FRACTION_ONE_THIRD = 0x2153;
const char32_t VULGAR_FRACTION_TWO_THIRDS = 0x2154;
const char32_t VULGAR_FRACTION_ONE_FIFTH = 0x2155;
const char32_t VULGAR_FRACTION_TWO_FIFTHS = 0x2156;
const char32_t VULGAR_FRACTION_THREE_FIFTHS = 0x2157;
const char32_t VULGAR_FRACTION_FOUR_FIFTHS = 0x2158;
const char32_t VULGAR_FRACTION_ONE_SIXTH = 0x2159;
const char32_t VULGAR_FRACTION_FIVE_SIXTHS = 0x215A;
const char32_t VULGAR_FRACTION_ONE_EIGHTH = 0x215B;
const char32_t VULGAR_FRACTION_THREE_EIGHTHS = 0x215C;
const char32_t VULGAR_FRACTION_FIVE_EIGHTHS = 0x215D;
const char32_t VULGAR_FRACTION_SEVEN_EIGHTHS = 0x215E;
const char32_t FRACTION_NUMERATOR_ONE = 0x215F;
const char32_t ROMAN_NUMERAL_ONE = 0x2160;
const char32_t ROMAN_NUMERAL_TWO = 0x2161;
const char32_t ROMAN_NUMERAL_THREE = 0x2162;
const char32_t ROMAN_NUMERAL_FOUR = 0x2163;
const char32_t ROMAN_NUMERAL_FIVE = 0x2164;
const char32_t ROMAN_NUMERAL_SIX = 0x2165;
const char32_t ROMAN_NUMERAL_SEVEN = 0x2166;
const char32_t ROMAN_NUMERAL_EIGHT = 0x2167;
const char32_t ROMAN_NUMERAL_NINE = 0x2168;
const char32_t ROMAN_NUMERAL_TEN = 0x2169;
const char32_t ROMAN_NUMERAL_ELEVEN = 0x216A;
const char32_t ROMAN_NUMERAL_TWELVE = 0x216B;
const char32_t ROMAN_NUMERAL_FIFTY = 0x216C;
const char32_t ROMAN_NUMERAL_ONE_HUNDRED = 0x216D;
const char32_t ROMAN_NUMERAL_FIVE_HUNDRED = 0x216E;
const char32_t ROMAN_NUMERAL_ONE_THOUSAND = 0x216F;
const char32_t SMALL_ROMAN_NUMERAL_ONE = 0x2170;
const char32_t SMALL_ROMAN_NUMERAL_TWO = 0x2171;
const char32_t SMALL_ROMAN_NUMERAL_THREE = 0x2172;
const char32_t SMALL_ROMAN_NUMERAL_FOUR = 0x2173;
const char32_t SMALL_ROMAN_NUMERAL_FIVE = 0x2174;
const char32_t SMALL_ROMAN_NUMERAL_SIX = 0x2175;
const char32_t SMALL_ROMAN_NUMERAL_SEVEN = 0x2176;
const char32_t SMALL_ROMAN_NUMERAL_EIGHT = 0x2177;
const char32_t SMALL_ROMAN_NUMERAL_NINE = 0x2178;
const char32_t SMALL_ROMAN_NUMERAL_TEN = 0x2179;
const char32_t SMALL_ROMAN_NUMERAL_ELEVEN = 0x217A;
const char32_t SMALL_ROMAN_NUMERAL_TWELVE = 0x217B;
const char32_t SMALL_ROMAN_NUMERAL_FIFTY = 0x217C;
const char32_t SMALL_ROMAN_NUMERAL_ONE_HUNDRED = 0x217D;
const char32_t SMALL_ROMAN_NUMERAL_FIVE_HUNDRED = 0x217E;
const char32_t SMALL_ROMAN_NUMERAL_ONE_THOUSAND = 0x217F;
const char32_t ROMAN_NUMERAL_ONE_THOUSAND_C_D = 0x2180;
const char32_t ROMAN_NUMERAL_FIVE_THOUSAND = 0x2181;
const char32_t ROMAN_NUMERAL_TEN_THOUSAND = 0x2182;
const char32_t ROMAN_NUMERAL_REVERSED_ONE_HUNDRED = 0x2183;
const char32_t LATIN_SMALL_LETTER_REVERSED_C = 0x2184;
const char32_t ROMAN_NUMERAL_SIX_LATE_FORM = 0x2185;
const char32_t ROMAN_NUMERAL_FIFTY_EARLY_FORM = 0x2186;
const char32_t ROMAN_NUMERAL_FIFTY_THOUSAND = 0x2187;
const char32_t ROMAN_NUMERAL_ONE_HUNDRED_THOUSAND = 0x2188;
const char32_t VULGAR_FRACTION_ZERO_THIRDS = 0x2189;
const char32_t TURNED_DIGIT_TWO = 0x218A;
const char32_t TURNED_DIGIT_THREE = 0x218B;
const char32_t LEFTWARDS_ARROW = 0x2190;
const char32_t UPWARDS_ARROW = 0x2191;
const char32_t RIGHTWARDS_ARROW = 0x2192;
const char32_t DOWNWARDS_ARROW = 0x2193;
const char32_t LEFT_RIGHT_ARROW = 0x2194;
const char32_t UP_DOWN_ARROW = 0x2195;
const char32_t NORTH_WEST_ARROW = 0x2196;
const char32_t NORTH_EAST_ARROW = 0x2197;
const char32_t SOUTH_EAST_ARROW = 0x2198;
const char32_t SOUTH_WEST_ARROW = 0x2199;
const char32_t LEFTWARDS_ARROW_WITH_STROKE = 0x219A;
const char32_t RIGHTWARDS_ARROW_WITH_STROKE = 0x219B;
const char32_t LEFTWARDS_WAVE_ARROW = 0x219C;
const char32_t RIGHTWARDS_WAVE_ARROW = 0x219D;
const char32_t LEFTWARDS_TWO_HEADED_ARROW = 0x219E;
const char32_t UPWARDS_TWO_HEADED_ARROW = 0x219F;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW = 0x21A0;
const char32_t DOWNWARDS_TWO_HEADED_ARROW = 0x21A1;
const char32_t LEFTWARDS_ARROW_WITH_TAIL = 0x21A2;
const char32_t RIGHTWARDS_ARROW_WITH_TAIL = 0x21A3;
const char32_t LEFTWARDS_ARROW_FROM_BAR = 0x21A4;
const char32_t UPWARDS_ARROW_FROM_BAR = 0x21A5;
const char32_t RIGHTWARDS_ARROW_FROM_BAR = 0x21A6;
const char32_t DOWNWARDS_ARROW_FROM_BAR = 0x21A7;
const char32_t UP_DOWN_ARROW_WITH_BASE = 0x21A8;
const char32_t LEFTWARDS_ARROW_WITH_HOOK = 0x21A9;
const char32_t RIGHTWARDS_ARROW_WITH_HOOK = 0x21AA;
const char32_t LEFTWARDS_ARROW_WITH_LOOP = 0x21AB;
const char32_t RIGHTWARDS_ARROW_WITH_LOOP = 0x21AC;
const char32_t LEFT_RIGHT_WAVE_ARROW = 0x21AD;
const char32_t LEFT_RIGHT_ARROW_WITH_STROKE = 0x21AE;
const char32_t DOWNWARDS_ZIGZAG_ARROW = 0x21AF;
const char32_t UPWARDS_ARROW_WITH_TIP_LEFTWARDS = 0x21B0;
const char32_t UPWARDS_ARROW_WITH_TIP_RIGHTWARDS = 0x21B1;
const char32_t DOWNWARDS_ARROW_WITH_TIP_LEFTWARDS = 0x21B2;
const char32_t DOWNWARDS_ARROW_WITH_TIP_RIGHTWARDS = 0x21B3;
const char32_t RIGHTWARDS_ARROW_WITH_CORNER_DOWNWARDS = 0x21B4;
const char32_t DOWNWARDS_ARROW_WITH_CORNER_LEFTWARDS = 0x21B5;
const char32_t ANTICLOCKWISE_TOP_SEMICIRCLE_ARROW = 0x21B6;
const char32_t CLOCKWISE_TOP_SEMICIRCLE_ARROW = 0x21B7;
const char32_t NORTH_WEST_ARROW_TO_LONG_BAR = 0x21B8;
const char32_t LEFTWARDS_ARROW_TO_BAR_OVER_RIGHTWARDS_ARROW_TO_BAR = 0x21B9;
const char32_t ANTICLOCKWISE_OPEN_CIRCLE_ARROW = 0x21BA;
const char32_t CLOCKWISE_OPEN_CIRCLE_ARROW = 0x21BB;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_UPWARDS = 0x21BC;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_DOWNWARDS = 0x21BD;
const char32_t UPWARDS_HARPOON_WITH_BARB_RIGHTWARDS = 0x21BE;
const char32_t UPWARDS_HARPOON_WITH_BARB_LEFTWARDS = 0x21BF;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_UPWARDS = 0x21C0;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_DOWNWARDS = 0x21C1;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_RIGHTWARDS = 0x21C2;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_LEFTWARDS = 0x21C3;
const char32_t RIGHTWARDS_ARROW_OVER_LEFTWARDS_ARROW = 0x21C4;
const char32_t UPWARDS_ARROW_LEFTWARDS_OF_DOWNWARDS_ARROW = 0x21C5;
const char32_t LEFTWARDS_ARROW_OVER_RIGHTWARDS_ARROW = 0x21C6;
const char32_t LEFTWARDS_PAIRED_ARROWS = 0x21C7;
const char32_t UPWARDS_PAIRED_ARROWS = 0x21C8;
const char32_t RIGHTWARDS_PAIRED_ARROWS = 0x21C9;
const char32_t DOWNWARDS_PAIRED_ARROWS = 0x21CA;
const char32_t LEFTWARDS_HARPOON_OVER_RIGHTWARDS_HARPOON = 0x21CB;
const char32_t RIGHTWARDS_HARPOON_OVER_LEFTWARDS_HARPOON = 0x21CC;
const char32_t LEFTWARDS_DOUBLE_ARROW_WITH_STROKE = 0x21CD;
const char32_t LEFT_RIGHT_DOUBLE_ARROW_WITH_STROKE = 0x21CE;
const char32_t RIGHTWARDS_DOUBLE_ARROW_WITH_STROKE = 0x21CF;
const char32_t LEFTWARDS_DOUBLE_ARROW = 0x21D0;
const char32_t UPWARDS_DOUBLE_ARROW = 0x21D1;
const char32_t RIGHTWARDS_DOUBLE_ARROW = 0x21D2;
const char32_t DOWNWARDS_DOUBLE_ARROW = 0x21D3;
const char32_t LEFT_RIGHT_DOUBLE_ARROW = 0x21D4;
const char32_t UP_DOWN_DOUBLE_ARROW = 0x21D5;
const char32_t NORTH_WEST_DOUBLE_ARROW = 0x21D6;
const char32_t NORTH_EAST_DOUBLE_ARROW = 0x21D7;
const char32_t SOUTH_EAST_DOUBLE_ARROW = 0x21D8;
const char32_t SOUTH_WEST_DOUBLE_ARROW = 0x21D9;
const char32_t LEFTWARDS_TRIPLE_ARROW = 0x21DA;
const char32_t RIGHTWARDS_TRIPLE_ARROW = 0x21DB;
const char32_t LEFTWARDS_SQUIGGLE_ARROW = 0x21DC;
const char32_t RIGHTWARDS_SQUIGGLE_ARROW = 0x21DD;
const char32_t UPWARDS_ARROW_WITH_DOUBLE_STROKE = 0x21DE;
const char32_t DOWNWARDS_ARROW_WITH_DOUBLE_STROKE = 0x21DF;
const char32_t LEFTWARDS_DASHED_ARROW = 0x21E0;
const char32_t UPWARDS_DASHED_ARROW = 0x21E1;
const char32_t RIGHTWARDS_DASHED_ARROW = 0x21E2;
const char32_t DOWNWARDS_DASHED_ARROW = 0x21E3;
const char32_t LEFTWARDS_ARROW_TO_BAR = 0x21E4;
const char32_t RIGHTWARDS_ARROW_TO_BAR = 0x21E5;
const char32_t LEFTWARDS_WHITE_ARROW = 0x21E6;
const char32_t UPWARDS_WHITE_ARROW = 0x21E7;
const char32_t RIGHTWARDS_WHITE_ARROW = 0x21E8;
const char32_t DOWNWARDS_WHITE_ARROW = 0x21E9;
const char32_t UPWARDS_WHITE_ARROW_FROM_BAR = 0x21EA;
const char32_t UPWARDS_WHITE_ARROW_ON_PEDESTAL = 0x21EB;
const char32_t UPWARDS_WHITE_ARROW_ON_PEDESTAL_WITH_HORIZONTAL_BAR = 0x21EC;
const char32_t UPWARDS_WHITE_ARROW_ON_PEDESTAL_WITH_VERTICAL_BAR = 0x21ED;
const char32_t UPWARDS_WHITE_DOUBLE_ARROW = 0x21EE;
const char32_t UPWARDS_WHITE_DOUBLE_ARROW_ON_PEDESTAL = 0x21EF;
const char32_t RIGHTWARDS_WHITE_ARROW_FROM_WALL = 0x21F0;
const char32_t NORTH_WEST_ARROW_TO_CORNER = 0x21F1;
const char32_t SOUTH_EAST_ARROW_TO_CORNER = 0x21F2;
const char32_t UP_DOWN_WHITE_ARROW = 0x21F3;
const char32_t RIGHT_ARROW_WITH_SMALL_CIRCLE = 0x21F4;
const char32_t DOWNWARDS_ARROW_LEFTWARDS_OF_UPWARDS_ARROW = 0x21F5;
const char32_t THREE_RIGHTWARDS_ARROWS = 0x21F6;
const char32_t LEFTWARDS_ARROW_WITH_VERTICAL_STROKE = 0x21F7;
const char32_t RIGHTWARDS_ARROW_WITH_VERTICAL_STROKE = 0x21F8;
const char32_t LEFT_RIGHT_ARROW_WITH_VERTICAL_STROKE = 0x21F9;
const char32_t LEFTWARDS_ARROW_WITH_DOUBLE_VERTICAL_STROKE = 0x21FA;
const char32_t RIGHTWARDS_ARROW_WITH_DOUBLE_VERTICAL_STROKE = 0x21FB;
const char32_t LEFT_RIGHT_ARROW_WITH_DOUBLE_VERTICAL_STROKE = 0x21FC;
const char32_t LEFTWARDS_OPEN_HEADED_ARROW = 0x21FD;
const char32_t RIGHTWARDS_OPEN_HEADED_ARROW = 0x21FE;
const char32_t LEFT_RIGHT_OPEN_HEADED_ARROW = 0x21FF;
const char32_t FOR_ALL = 0x2200;
const char32_t COMPLEMENT = 0x2201;
const char32_t PARTIAL_DIFFERENTIAL = 0x2202;
const char32_t THERE_EXISTS = 0x2203;
const char32_t THERE_DOES_NOT_EXIST = 0x2204;
const char32_t EMPTY_SET = 0x2205;
const char32_t INCREMENT = 0x2206;
const char32_t NABLA = 0x2207;
const char32_t ELEMENT_OF = 0x2208;
const char32_t NOT_AN_ELEMENT_OF = 0x2209;
const char32_t SMALL_ELEMENT_OF = 0x220A;
const char32_t CONTAINS_AS_MEMBER = 0x220B;
const char32_t DOES_NOT_CONTAIN_AS_MEMBER = 0x220C;
const char32_t SMALL_CONTAINS_AS_MEMBER = 0x220D;
const char32_t END_OF_PROOF = 0x220E;
const char32_t N_ARY_PRODUCT = 0x220F;
const char32_t N_ARY_COPRODUCT = 0x2210;
const char32_t N_ARY_SUMMATION = 0x2211;
const char32_t MINUS_SIGN = 0x2212;
const char32_t MINUS_OR_PLUS_SIGN = 0x2213;
const char32_t DOT_PLUS = 0x2214;
const char32_t DIVISION_SLASH = 0x2215;
const char32_t SET_MINUS = 0x2216;
const char32_t ASTERISK_OPERATOR = 0x2217;
const char32_t RING_OPERATOR = 0x2218;
const char32_t BULLET_OPERATOR = 0x2219;
const char32_t SQUARE_ROOT = 0x221A;
const char32_t CUBE_ROOT = 0x221B;
const char32_t FOURTH_ROOT = 0x221C;
const char32_t PROPORTIONAL_TO = 0x221D;
const char32_t INFINITY = 0x221E;
const char32_t RIGHT_ANGLE = 0x221F;
const char32_t ANGLE = 0x2220;
const char32_t MEASURED_ANGLE = 0x2221;
const char32_t SPHERICAL_ANGLE = 0x2222;
const char32_t DIVIDES = 0x2223;
const char32_t DOES_NOT_DIVIDE = 0x2224;
const char32_t PARALLEL_TO = 0x2225;
const char32_t NOT_PARALLEL_TO = 0x2226;
const char32_t LOGICAL_AND = 0x2227;
const char32_t LOGICAL_OR = 0x2228;
const char32_t INTERSECTION = 0x2229;
const char32_t UNION = 0x222A;
const char32_t INTEGRAL = 0x222B;
const char32_t DOUBLE_INTEGRAL = 0x222C;
const char32_t TRIPLE_INTEGRAL = 0x222D;
const char32_t CONTOUR_INTEGRAL = 0x222E;
const char32_t SURFACE_INTEGRAL = 0x222F;
const char32_t VOLUME_INTEGRAL = 0x2230;
const char32_t CLOCKWISE_INTEGRAL = 0x2231;
const char32_t CLOCKWISE_CONTOUR_INTEGRAL = 0x2232;
const char32_t ANTICLOCKWISE_CONTOUR_INTEGRAL = 0x2233;
const char32_t THEREFORE = 0x2234;
const char32_t BECAUSE = 0x2235;
const char32_t RATIO = 0x2236;
const char32_t PROPORTION = 0x2237;
const char32_t DOT_MINUS = 0x2238;
const char32_t EXCESS = 0x2239;
const char32_t GEOMETRIC_PROPORTION = 0x223A;
const char32_t HOMOTHETIC = 0x223B;
const char32_t TILDE_OPERATOR = 0x223C;
const char32_t REVERSED_TILDE = 0x223D;
const char32_t INVERTED_LAZY_S = 0x223E;
const char32_t SINE_WAVE = 0x223F;
const char32_t WREATH_PRODUCT = 0x2240;
const char32_t NOT_TILDE = 0x2241;
const char32_t MINUS_TILDE = 0x2242;
const char32_t ASYMPTOTICALLY_EQUAL_TO = 0x2243;
const char32_t NOT_ASYMPTOTICALLY_EQUAL_TO = 0x2244;
const char32_t APPROXIMATELY_EQUAL_TO = 0x2245;
const char32_t APPROXIMATELY_BUT_NOT_ACTUALLY_EQUAL_TO = 0x2246;
const char32_t NEITHER_APPROXIMATELY_NOR_ACTUALLY_EQUAL_TO = 0x2247;
const char32_t ALMOST_EQUAL_TO = 0x2248;
const char32_t NOT_ALMOST_EQUAL_TO = 0x2249;
const char32_t ALMOST_EQUAL_OR_EQUAL_TO = 0x224A;
const char32_t TRIPLE_TILDE = 0x224B;
const char32_t ALL_EQUAL_TO = 0x224C;
const char32_t EQUIVALENT_TO = 0x224D;
const char32_t GEOMETRICALLY_EQUIVALENT_TO = 0x224E;
const char32_t DIFFERENCE_BETWEEN = 0x224F;
const char32_t APPROACHES_THE_LIMIT = 0x2250;
const char32_t GEOMETRICALLY_EQUAL_TO = 0x2251;
const char32_t APPROXIMATELY_EQUAL_TO_OR_THE_IMAGE_OF = 0x2252;
const char32_t IMAGE_OF_OR_APPROXIMATELY_EQUAL_TO = 0x2253;
const char32_t COLON_EQUALS = 0x2254;
const char32_t EQUALS_COLON = 0x2255;
const char32_t RING_IN_EQUAL_TO = 0x2256;
const char32_t RING_EQUAL_TO = 0x2257;
const char32_t CORRESPONDS_TO = 0x2258;
const char32_t ESTIMATES = 0x2259;
const char32_t EQUIANGULAR_TO = 0x225A;
const char32_t STAR_EQUALS = 0x225B;
const char32_t DELTA_EQUAL_TO = 0x225C;
const char32_t EQUAL_TO_BY_DEFINITION = 0x225D;
const char32_t MEASURED_BY = 0x225E;
const char32_t QUESTIONED_EQUAL_TO = 0x225F;
const char32_t NOT_EQUAL_TO = 0x2260;
const char32_t IDENTICAL_TO = 0x2261;
const char32_t NOT_IDENTICAL_TO = 0x2262;
const char32_t STRICTLY_EQUIVALENT_TO = 0x2263;
const char32_t LESS_THAN_OR_EQUAL_TO = 0x2264;
const char32_t GREATER_THAN_OR_EQUAL_TO = 0x2265;
const char32_t LESS_THAN_OVER_EQUAL_TO = 0x2266;
const char32_t GREATER_THAN_OVER_EQUAL_TO = 0x2267;
const char32_t LESS_THAN_BUT_NOT_EQUAL_TO = 0x2268;
const char32_t GREATER_THAN_BUT_NOT_EQUAL_TO = 0x2269;
const char32_t MUCH_LESS_THAN = 0x226A;
const char32_t MUCH_GREATER_THAN = 0x226B;
const char32_t BETWEEN = 0x226C;
const char32_t NOT_EQUIVALENT_TO = 0x226D;
const char32_t NOT_LESS_THAN = 0x226E;
const char32_t NOT_GREATER_THAN = 0x226F;
const char32_t NEITHER_LESS_THAN_NOR_EQUAL_TO = 0x2270;
const char32_t NEITHER_GREATER_THAN_NOR_EQUAL_TO = 0x2271;
const char32_t LESS_THAN_OR_EQUIVALENT_TO = 0x2272;
const char32_t GREATER_THAN_OR_EQUIVALENT_TO = 0x2273;
const char32_t NEITHER_LESS_THAN_NOR_EQUIVALENT_TO = 0x2274;
const char32_t NEITHER_GREATER_THAN_NOR_EQUIVALENT_TO = 0x2275;
const char32_t LESS_THAN_OR_GREATER_THAN = 0x2276;
const char32_t GREATER_THAN_OR_LESS_THAN = 0x2277;
const char32_t NEITHER_LESS_THAN_NOR_GREATER_THAN = 0x2278;
const char32_t NEITHER_GREATER_THAN_NOR_LESS_THAN = 0x2279;
const char32_t PRECEDES = 0x227A;
const char32_t SUCCEEDS = 0x227B;
const char32_t PRECEDES_OR_EQUAL_TO = 0x227C;
const char32_t SUCCEEDS_OR_EQUAL_TO = 0x227D;
const char32_t PRECEDES_OR_EQUIVALENT_TO = 0x227E;
const char32_t SUCCEEDS_OR_EQUIVALENT_TO = 0x227F;
const char32_t DOES_NOT_PRECEDE = 0x2280;
const char32_t DOES_NOT_SUCCEED = 0x2281;
const char32_t SUBSET_OF = 0x2282;
const char32_t SUPERSET_OF = 0x2283;
const char32_t NOT_A_SUBSET_OF = 0x2284;
const char32_t NOT_A_SUPERSET_OF = 0x2285;
const char32_t SUBSET_OF_OR_EQUAL_TO = 0x2286;
const char32_t SUPERSET_OF_OR_EQUAL_TO = 0x2287;
const char32_t NEITHER_A_SUBSET_OF_NOR_EQUAL_TO = 0x2288;
const char32_t NEITHER_A_SUPERSET_OF_NOR_EQUAL_TO = 0x2289;
const char32_t SUBSET_OF_WITH_NOT_EQUAL_TO = 0x228A;
const char32_t SUPERSET_OF_WITH_NOT_EQUAL_TO = 0x228B;
const char32_t MULTISET = 0x228C;
const char32_t MULTISET_MULTIPLICATION = 0x228D;
const char32_t MULTISET_UNION = 0x228E;
const char32_t SQUARE_IMAGE_OF = 0x228F;
const char32_t SQUARE_ORIGINAL_OF = 0x2290;
const char32_t SQUARE_IMAGE_OF_OR_EQUAL_TO = 0x2291;
const char32_t SQUARE_ORIGINAL_OF_OR_EQUAL_TO = 0x2292;
const char32_t SQUARE_CAP = 0x2293;
const char32_t SQUARE_CUP = 0x2294;
const char32_t CIRCLED_PLUS = 0x2295;
const char32_t CIRCLED_MINUS = 0x2296;
const char32_t CIRCLED_TIMES = 0x2297;
const char32_t CIRCLED_DIVISION_SLASH = 0x2298;
const char32_t CIRCLED_DOT_OPERATOR = 0x2299;
const char32_t CIRCLED_RING_OPERATOR = 0x229A;
const char32_t CIRCLED_ASTERISK_OPERATOR = 0x229B;
const char32_t CIRCLED_EQUALS = 0x229C;
const char32_t CIRCLED_DASH = 0x229D;
const char32_t SQUARED_PLUS = 0x229E;
const char32_t SQUARED_MINUS = 0x229F;
const char32_t SQUARED_TIMES = 0x22A0;
const char32_t SQUARED_DOT_OPERATOR = 0x22A1;
const char32_t RIGHT_TACK = 0x22A2;
const char32_t LEFT_TACK = 0x22A3;
const char32_t DOWN_TACK = 0x22A4;
const char32_t UP_TACK = 0x22A5;
const char32_t ASSERTION = 0x22A6;
const char32_t MODELS = 0x22A7;
const char32_t TRUE = 0x22A8;
const char32_t FORCES = 0x22A9;
const char32_t TRIPLE_VERTICAL_BAR_RIGHT_TURNSTILE = 0x22AA;
const char32_t DOUBLE_VERTICAL_BAR_DOUBLE_RIGHT_TURNSTILE = 0x22AB;
const char32_t DOES_NOT_PROVE = 0x22AC;
const char32_t NOT_TRUE = 0x22AD;
const char32_t DOES_NOT_FORCE = 0x22AE;
const char32_t NEGATED_DOUBLE_VERTICAL_BAR_DOUBLE_RIGHT_TURNSTILE = 0x22AF;
const char32_t PRECEDES_UNDER_RELATION = 0x22B0;
const char32_t SUCCEEDS_UNDER_RELATION = 0x22B1;
const char32_t NORMAL_SUBGROUP_OF = 0x22B2;
const char32_t CONTAINS_AS_NORMAL_SUBGROUP = 0x22B3;
const char32_t NORMAL_SUBGROUP_OF_OR_EQUAL_TO = 0x22B4;
const char32_t CONTAINS_AS_NORMAL_SUBGROUP_OR_EQUAL_TO = 0x22B5;
const char32_t ORIGINAL_OF = 0x22B6;
const char32_t IMAGE_OF = 0x22B7;
const char32_t MULTIMAP = 0x22B8;
const char32_t HERMITIAN_CONJUGATE_MATRIX = 0x22B9;
const char32_t INTERCALATE = 0x22BA;
const char32_t XOR = 0x22BB;
const char32_t NAND = 0x22BC;
const char32_t NOR = 0x22BD;
const char32_t RIGHT_ANGLE_WITH_ARC = 0x22BE;
const char32_t RIGHT_TRIANGLE = 0x22BF;
const char32_t N_ARY_LOGICAL_AND = 0x22C0;
const char32_t N_ARY_LOGICAL_OR = 0x22C1;
const char32_t N_ARY_INTERSECTION = 0x22C2;
const char32_t N_ARY_UNION = 0x22C3;
const char32_t DIAMOND_OPERATOR = 0x22C4;
const char32_t DOT_OPERATOR = 0x22C5;
const char32_t STAR_OPERATOR = 0x22C6;
const char32_t DIVISION_TIMES = 0x22C7;
const char32_t BOWTIE = 0x22C8;
const char32_t LEFT_NORMAL_FACTOR_SEMIDIRECT_PRODUCT = 0x22C9;
const char32_t RIGHT_NORMAL_FACTOR_SEMIDIRECT_PRODUCT = 0x22CA;
const char32_t LEFT_SEMIDIRECT_PRODUCT = 0x22CB;
const char32_t RIGHT_SEMIDIRECT_PRODUCT = 0x22CC;
const char32_t REVERSED_TILDE_EQUALS = 0x22CD;
const char32_t CURLY_LOGICAL_OR = 0x22CE;
const char32_t CURLY_LOGICAL_AND = 0x22CF;
const char32_t DOUBLE_SUBSET = 0x22D0;
const char32_t DOUBLE_SUPERSET = 0x22D1;
const char32_t DOUBLE_INTERSECTION = 0x22D2;
const char32_t DOUBLE_UNION = 0x22D3;
const char32_t PITCHFORK = 0x22D4;
const char32_t EQUAL_AND_PARALLEL_TO = 0x22D5;
const char32_t LESS_THAN_WITH_DOT = 0x22D6;
const char32_t GREATER_THAN_WITH_DOT = 0x22D7;
const char32_t VERY_MUCH_LESS_THAN = 0x22D8;
const char32_t VERY_MUCH_GREATER_THAN = 0x22D9;
const char32_t LESS_THAN_EQUAL_TO_OR_GREATER_THAN = 0x22DA;
const char32_t GREATER_THAN_EQUAL_TO_OR_LESS_THAN = 0x22DB;
const char32_t EQUAL_TO_OR_LESS_THAN = 0x22DC;
const char32_t EQUAL_TO_OR_GREATER_THAN = 0x22DD;
const char32_t EQUAL_TO_OR_PRECEDES = 0x22DE;
const char32_t EQUAL_TO_OR_SUCCEEDS = 0x22DF;
const char32_t DOES_NOT_PRECEDE_OR_EQUAL = 0x22E0;
const char32_t DOES_NOT_SUCCEED_OR_EQUAL = 0x22E1;
const char32_t NOT_SQUARE_IMAGE_OF_OR_EQUAL_TO = 0x22E2;
const char32_t NOT_SQUARE_ORIGINAL_OF_OR_EQUAL_TO = 0x22E3;
const char32_t SQUARE_IMAGE_OF_OR_NOT_EQUAL_TO = 0x22E4;
const char32_t SQUARE_ORIGINAL_OF_OR_NOT_EQUAL_TO = 0x22E5;
const char32_t LESS_THAN_BUT_NOT_EQUIVALENT_TO = 0x22E6;
const char32_t GREATER_THAN_BUT_NOT_EQUIVALENT_TO = 0x22E7;
const char32_t PRECEDES_BUT_NOT_EQUIVALENT_TO = 0x22E8;
const char32_t SUCCEEDS_BUT_NOT_EQUIVALENT_TO = 0x22E9;
const char32_t NOT_NORMAL_SUBGROUP_OF = 0x22EA;
const char32_t DOES_NOT_CONTAIN_AS_NORMAL_SUBGROUP = 0x22EB;
const char32_t NOT_NORMAL_SUBGROUP_OF_OR_EQUAL_TO = 0x22EC;
const char32_t DOES_NOT_CONTAIN_AS_NORMAL_SUBGROUP_OR_EQUAL = 0x22ED;
const char32_t VERTICAL_ELLIPSIS = 0x22EE;
const char32_t MIDLINE_HORIZONTAL_ELLIPSIS = 0x22EF;
const char32_t UP_RIGHT_DIAGONAL_ELLIPSIS = 0x22F0;
const char32_t DOWN_RIGHT_DIAGONAL_ELLIPSIS = 0x22F1;
const char32_t ELEMENT_OF_WITH_LONG_HORIZONTAL_STROKE = 0x22F2;
const char32_t ELEMENT_OF_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = 0x22F3;
const char32_t SMALL_ELEMENT_OF_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = 0x22F4;
const char32_t ELEMENT_OF_WITH_DOT_ABOVE = 0x22F5;
const char32_t ELEMENT_OF_WITH_OVERBAR = 0x22F6;
const char32_t SMALL_ELEMENT_OF_WITH_OVERBAR = 0x22F7;
const char32_t ELEMENT_OF_WITH_UNDERBAR = 0x22F8;
const char32_t ELEMENT_OF_WITH_TWO_HORIZONTAL_STROKES = 0x22F9;
const char32_t CONTAINS_WITH_LONG_HORIZONTAL_STROKE = 0x22FA;
const char32_t CONTAINS_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = 0x22FB;
const char32_t SMALL_CONTAINS_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = 0x22FC;
const char32_t CONTAINS_WITH_OVERBAR = 0x22FD;
const char32_t SMALL_CONTAINS_WITH_OVERBAR = 0x22FE;
const char32_t Z_NOTATION_BAG_MEMBERSHIP = 0x22FF;
const char32_t DIAMETER_SIGN = 0x2300;
const char32_t ELECTRIC_ARROW = 0x2301;
const char32_t HOUSE = 0x2302;
const char32_t UP_ARROWHEAD = 0x2303;
const char32_t DOWN_ARROWHEAD = 0x2304;
const char32_t PROJECTIVE = 0x2305;
const char32_t PERSPECTIVE = 0x2306;
const char32_t WAVY_LINE = 0x2307;
const char32_t LEFT_CEILING = 0x2308;
const char32_t RIGHT_CEILING = 0x2309;
const char32_t LEFT_FLOOR = 0x230A;
const char32_t RIGHT_FLOOR = 0x230B;
const char32_t BOTTOM_RIGHT_CROP = 0x230C;
const char32_t BOTTOM_LEFT_CROP = 0x230D;
const char32_t TOP_RIGHT_CROP = 0x230E;
const char32_t TOP_LEFT_CROP = 0x230F;
const char32_t REVERSED_NOT_SIGN = 0x2310;
const char32_t SQUARE_LOZENGE = 0x2311;
const char32_t ARC = 0x2312;
const char32_t SEGMENT = 0x2313;
const char32_t SECTOR = 0x2314;
const char32_t TELEPHONE_RECORDER = 0x2315;
const char32_t POSITION_INDICATOR = 0x2316;
const char32_t VIEWDATA_SQUARE = 0x2317;
const char32_t PLACE_OF_INTEREST_SIGN = 0x2318;
const char32_t TURNED_NOT_SIGN = 0x2319;
const char32_t WATCH = 0x231A;
const char32_t HOURGLASS = 0x231B;
const char32_t TOP_LEFT_CORNER = 0x231C;
const char32_t TOP_RIGHT_CORNER = 0x231D;
const char32_t BOTTOM_LEFT_CORNER = 0x231E;
const char32_t BOTTOM_RIGHT_CORNER = 0x231F;
const char32_t TOP_HALF_INTEGRAL = 0x2320;
const char32_t BOTTOM_HALF_INTEGRAL = 0x2321;
const char32_t FROWN = 0x2322;
const char32_t SMILE = 0x2323;
const char32_t UP_ARROWHEAD_BETWEEN_TWO_HORIZONTAL_BARS = 0x2324;
const char32_t OPTION_KEY = 0x2325;
const char32_t ERASE_TO_THE_RIGHT = 0x2326;
const char32_t X_IN_A_RECTANGLE_BOX = 0x2327;
const char32_t KEYBOARD = 0x2328;
const char32_t LEFT_POINTING_ANGLE_BRACKET = 0x2329;
const char32_t RIGHT_POINTING_ANGLE_BRACKET = 0x232A;
const char32_t ERASE_TO_THE_LEFT = 0x232B;
const char32_t BENZENE_RING = 0x232C;
const char32_t CYLINDRICITY = 0x232D;
const char32_t ALL_AROUND_PROFILE = 0x232E;
const char32_t SYMMETRY = 0x232F;
const char32_t TOTAL_RUNOUT = 0x2330;
const char32_t DIMENSION_ORIGIN = 0x2331;
const char32_t CONICAL_TAPER = 0x2332;
const char32_t SLOPE = 0x2333;
const char32_t COUNTERBORE = 0x2334;
const char32_t COUNTERSINK = 0x2335;
const char32_t APL_FUNCTIONAL_SYMBOL_I_BEAM = 0x2336;
const char32_t APL_FUNCTIONAL_SYMBOL_SQUISH_QUAD = 0x2337;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_EQUAL = 0x2338;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_DIVIDE = 0x2339;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_DIAMOND = 0x233A;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_JOT = 0x233B;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_CIRCLE = 0x233C;
const char32_t APL_FUNCTIONAL_SYMBOL_CIRCLE_STILE = 0x233D;
const char32_t APL_FUNCTIONAL_SYMBOL_CIRCLE_JOT = 0x233E;
const char32_t APL_FUNCTIONAL_SYMBOL_SLASH_BAR = 0x233F;
const char32_t APL_FUNCTIONAL_SYMBOL_BACKSLASH_BAR = 0x2340;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_SLASH = 0x2341;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_BACKSLASH = 0x2342;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_LESS_THAN = 0x2343;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_GREATER_THAN = 0x2344;
const char32_t APL_FUNCTIONAL_SYMBOL_LEFTWARDS_VANE = 0x2345;
const char32_t APL_FUNCTIONAL_SYMBOL_RIGHTWARDS_VANE = 0x2346;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_LEFTWARDS_ARROW = 0x2347;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_RIGHTWARDS_ARROW = 0x2348;
const char32_t APL_FUNCTIONAL_SYMBOL_CIRCLE_BACKSLASH = 0x2349;
const char32_t APL_FUNCTIONAL_SYMBOL_DOWN_TACK_UNDERBAR = 0x234A;
const char32_t APL_FUNCTIONAL_SYMBOL_DELTA_STILE = 0x234B;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_DOWN_CARET = 0x234C;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_DELTA = 0x234D;
const char32_t APL_FUNCTIONAL_SYMBOL_DOWN_TACK_JOT = 0x234E;
const char32_t APL_FUNCTIONAL_SYMBOL_UPWARDS_VANE = 0x234F;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_UPWARDS_ARROW = 0x2350;
const char32_t APL_FUNCTIONAL_SYMBOL_UP_TACK_OVERBAR = 0x2351;
const char32_t APL_FUNCTIONAL_SYMBOL_DEL_STILE = 0x2352;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_UP_CARET = 0x2353;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_DEL = 0x2354;
const char32_t APL_FUNCTIONAL_SYMBOL_UP_TACK_JOT = 0x2355;
const char32_t APL_FUNCTIONAL_SYMBOL_DOWNWARDS_VANE = 0x2356;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_DOWNWARDS_ARROW = 0x2357;
const char32_t APL_FUNCTIONAL_SYMBOL_QUOTE_UNDERBAR = 0x2358;
const char32_t APL_FUNCTIONAL_SYMBOL_DELTA_UNDERBAR = 0x2359;
const char32_t APL_FUNCTIONAL_SYMBOL_DIAMOND_UNDERBAR = 0x235A;
const char32_t APL_FUNCTIONAL_SYMBOL_JOT_UNDERBAR = 0x235B;
const char32_t APL_FUNCTIONAL_SYMBOL_CIRCLE_UNDERBAR = 0x235C;
const char32_t APL_FUNCTIONAL_SYMBOL_UP_SHOE_JOT = 0x235D;
const char32_t APL_FUNCTIONAL_SYMBOL_QUOTE_QUAD = 0x235E;
const char32_t APL_FUNCTIONAL_SYMBOL_CIRCLE_STAR = 0x235F;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_COLON = 0x2360;
const char32_t APL_FUNCTIONAL_SYMBOL_UP_TACK_DIAERESIS = 0x2361;
const char32_t APL_FUNCTIONAL_SYMBOL_DEL_DIAERESIS = 0x2362;
const char32_t APL_FUNCTIONAL_SYMBOL_STAR_DIAERESIS = 0x2363;
const char32_t APL_FUNCTIONAL_SYMBOL_JOT_DIAERESIS = 0x2364;
const char32_t APL_FUNCTIONAL_SYMBOL_CIRCLE_DIAERESIS = 0x2365;
const char32_t APL_FUNCTIONAL_SYMBOL_DOWN_SHOE_STILE = 0x2366;
const char32_t APL_FUNCTIONAL_SYMBOL_LEFT_SHOE_STILE = 0x2367;
const char32_t APL_FUNCTIONAL_SYMBOL_TILDE_DIAERESIS = 0x2368;
const char32_t APL_FUNCTIONAL_SYMBOL_GREATER_THAN_DIAERESIS = 0x2369;
const char32_t APL_FUNCTIONAL_SYMBOL_COMMA_BAR = 0x236A;
const char32_t APL_FUNCTIONAL_SYMBOL_DEL_TILDE = 0x236B;
const char32_t APL_FUNCTIONAL_SYMBOL_ZILDE = 0x236C;
const char32_t APL_FUNCTIONAL_SYMBOL_STILE_TILDE = 0x236D;
const char32_t APL_FUNCTIONAL_SYMBOL_SEMICOLON_UNDERBAR = 0x236E;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_NOT_EQUAL = 0x236F;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD_QUESTION = 0x2370;
const char32_t APL_FUNCTIONAL_SYMBOL_DOWN_CARET_TILDE = 0x2371;
const char32_t APL_FUNCTIONAL_SYMBOL_UP_CARET_TILDE = 0x2372;
const char32_t APL_FUNCTIONAL_SYMBOL_IOTA = 0x2373;
const char32_t APL_FUNCTIONAL_SYMBOL_RHO = 0x2374;
const char32_t APL_FUNCTIONAL_SYMBOL_OMEGA = 0x2375;
const char32_t APL_FUNCTIONAL_SYMBOL_ALPHA_UNDERBAR = 0x2376;
const char32_t APL_FUNCTIONAL_SYMBOL_EPSILON_UNDERBAR = 0x2377;
const char32_t APL_FUNCTIONAL_SYMBOL_IOTA_UNDERBAR = 0x2378;
const char32_t APL_FUNCTIONAL_SYMBOL_OMEGA_UNDERBAR = 0x2379;
const char32_t APL_FUNCTIONAL_SYMBOL_ALPHA = 0x237A;
const char32_t NOT_CHECK_MARK = 0x237B;
const char32_t RIGHT_ANGLE_WITH_DOWNWARDS_ZIGZAG_ARROW = 0x237C;
const char32_t SHOULDERED_OPEN_BOX = 0x237D;
const char32_t BELL_SYMBOL = 0x237E;
const char32_t VERTICAL_LINE_WITH_MIDDLE_DOT = 0x237F;
const char32_t INSERTION_SYMBOL = 0x2380;
const char32_t CONTINUOUS_UNDERLINE_SYMBOL = 0x2381;
const char32_t DISCONTINUOUS_UNDERLINE_SYMBOL = 0x2382;
const char32_t EMPHASIS_SYMBOL = 0x2383;
const char32_t COMPOSITION_SYMBOL = 0x2384;
const char32_t WHITE_SQUARE_WITH_CENTRE_VERTICAL_LINE = 0x2385;
const char32_t ENTER_SYMBOL = 0x2386;
const char32_t ALTERNATIVE_KEY_SYMBOL = 0x2387;
const char32_t HELM_SYMBOL = 0x2388;
const char32_t CIRCLED_HORIZONTAL_BAR_WITH_NOTCH = 0x2389;
const char32_t CIRCLED_TRIANGLE_DOWN = 0x238A;
const char32_t BROKEN_CIRCLE_WITH_NORTHWEST_ARROW = 0x238B;
const char32_t UNDO_SYMBOL = 0x238C;
const char32_t MONOSTABLE_SYMBOL = 0x238D;
const char32_t HYSTERESIS_SYMBOL = 0x238E;
const char32_t OPEN_CIRCUIT_OUTPUT_H_TYPE_SYMBOL = 0x238F;
const char32_t OPEN_CIRCUIT_OUTPUT_L_TYPE_SYMBOL = 0x2390;
const char32_t PASSIVE_PULL_DOWN_OUTPUT_SYMBOL = 0x2391;
const char32_t PASSIVE_PULL_UP_OUTPUT_SYMBOL = 0x2392;
const char32_t DIRECT_CURRENT_SYMBOL_FORM_TWO = 0x2393;
const char32_t SOFTWARE_FUNCTION_SYMBOL = 0x2394;
const char32_t APL_FUNCTIONAL_SYMBOL_QUAD = 0x2395;
const char32_t DECIMAL_SEPARATOR_KEY_SYMBOL = 0x2396;
const char32_t PREVIOUS_PAGE = 0x2397;
const char32_t NEXT_PAGE = 0x2398;
const char32_t PRINT_SCREEN_SYMBOL = 0x2399;
const char32_t CLEAR_SCREEN_SYMBOL = 0x239A;
const char32_t LEFT_PARENTHESIS_UPPER_HOOK = 0x239B;
const char32_t LEFT_PARENTHESIS_EXTENSION = 0x239C;
const char32_t LEFT_PARENTHESIS_LOWER_HOOK = 0x239D;
const char32_t RIGHT_PARENTHESIS_UPPER_HOOK = 0x239E;
const char32_t RIGHT_PARENTHESIS_EXTENSION = 0x239F;
const char32_t RIGHT_PARENTHESIS_LOWER_HOOK = 0x23A0;
const char32_t LEFT_SQUARE_BRACKET_UPPER_CORNER = 0x23A1;
const char32_t LEFT_SQUARE_BRACKET_EXTENSION = 0x23A2;
const char32_t LEFT_SQUARE_BRACKET_LOWER_CORNER = 0x23A3;
const char32_t RIGHT_SQUARE_BRACKET_UPPER_CORNER = 0x23A4;
const char32_t RIGHT_SQUARE_BRACKET_EXTENSION = 0x23A5;
const char32_t RIGHT_SQUARE_BRACKET_LOWER_CORNER = 0x23A6;
const char32_t LEFT_CURLY_BRACKET_UPPER_HOOK = 0x23A7;
const char32_t LEFT_CURLY_BRACKET_MIDDLE_PIECE = 0x23A8;
const char32_t LEFT_CURLY_BRACKET_LOWER_HOOK = 0x23A9;
const char32_t CURLY_BRACKET_EXTENSION = 0x23AA;
const char32_t RIGHT_CURLY_BRACKET_UPPER_HOOK = 0x23AB;
const char32_t RIGHT_CURLY_BRACKET_MIDDLE_PIECE = 0x23AC;
const char32_t RIGHT_CURLY_BRACKET_LOWER_HOOK = 0x23AD;
const char32_t INTEGRAL_EXTENSION = 0x23AE;
const char32_t HORIZONTAL_LINE_EXTENSION = 0x23AF;
const char32_t UPPER_LEFT_OR_LOWER_RIGHT_CURLY_BRACKET_SECTION = 0x23B0;
const char32_t UPPER_RIGHT_OR_LOWER_LEFT_CURLY_BRACKET_SECTION = 0x23B1;
const char32_t SUMMATION_TOP = 0x23B2;
const char32_t SUMMATION_BOTTOM = 0x23B3;
const char32_t TOP_SQUARE_BRACKET = 0x23B4;
const char32_t BOTTOM_SQUARE_BRACKET = 0x23B5;
const char32_t BOTTOM_SQUARE_BRACKET_OVER_TOP_SQUARE_BRACKET = 0x23B6;
const char32_t RADICAL_SYMBOL_BOTTOM = 0x23B7;
const char32_t LEFT_VERTICAL_BOX_LINE = 0x23B8;
const char32_t RIGHT_VERTICAL_BOX_LINE = 0x23B9;
const char32_t HORIZONTAL_SCAN_LINE_1 = 0x23BA;
const char32_t HORIZONTAL_SCAN_LINE_3 = 0x23BB;
const char32_t HORIZONTAL_SCAN_LINE_7 = 0x23BC;
const char32_t HORIZONTAL_SCAN_LINE_9 = 0x23BD;
const char32_t DENTISTRY_SYMBOL_LIGHT_VERTICAL_AND_TOP_RIGHT = 0x23BE;
const char32_t DENTISTRY_SYMBOL_LIGHT_VERTICAL_AND_BOTTOM_RIGHT = 0x23BF;
const char32_t DENTISTRY_SYMBOL_LIGHT_VERTICAL_WITH_CIRCLE = 0x23C0;
const char32_t DENTISTRY_SYMBOL_LIGHT_DOWN_AND_HORIZONTAL_WITH_CIRCLE = 0x23C1;
const char32_t DENTISTRY_SYMBOL_LIGHT_UP_AND_HORIZONTAL_WITH_CIRCLE = 0x23C2;
const char32_t DENTISTRY_SYMBOL_LIGHT_VERTICAL_WITH_TRIANGLE = 0x23C3;
const char32_t DENTISTRY_SYMBOL_LIGHT_DOWN_AND_HORIZONTAL_WITH_TRIANGLE = 0x23C4;
const char32_t DENTISTRY_SYMBOL_LIGHT_UP_AND_HORIZONTAL_WITH_TRIANGLE = 0x23C5;
const char32_t DENTISTRY_SYMBOL_LIGHT_VERTICAL_AND_WAVE = 0x23C6;
const char32_t DENTISTRY_SYMBOL_LIGHT_DOWN_AND_HORIZONTAL_WITH_WAVE = 0x23C7;
const char32_t DENTISTRY_SYMBOL_LIGHT_UP_AND_HORIZONTAL_WITH_WAVE = 0x23C8;
const char32_t DENTISTRY_SYMBOL_LIGHT_DOWN_AND_HORIZONTAL = 0x23C9;
const char32_t DENTISTRY_SYMBOL_LIGHT_UP_AND_HORIZONTAL = 0x23CA;
const char32_t DENTISTRY_SYMBOL_LIGHT_VERTICAL_AND_TOP_LEFT = 0x23CB;
const char32_t DENTISTRY_SYMBOL_LIGHT_VERTICAL_AND_BOTTOM_LEFT = 0x23CC;
const char32_t SQUARE_FOOT = 0x23CD;
const char32_t RETURN_SYMBOL = 0x23CE;
const char32_t EJECT_SYMBOL = 0x23CF;
const char32_t VERTICAL_LINE_EXTENSION = 0x23D0;
const char32_t METRICAL_BREVE = 0x23D1;
const char32_t METRICAL_LONG_OVER_SHORT = 0x23D2;
const char32_t METRICAL_SHORT_OVER_LONG = 0x23D3;
const char32_t METRICAL_LONG_OVER_TWO_SHORTS = 0x23D4;
const char32_t METRICAL_TWO_SHORTS_OVER_LONG = 0x23D5;
const char32_t METRICAL_TWO_SHORTS_JOINED = 0x23D6;
const char32_t METRICAL_TRISEME = 0x23D7;
const char32_t METRICAL_TETRASEME = 0x23D8;
const char32_t METRICAL_PENTASEME = 0x23D9;
const char32_t EARTH_GROUND = 0x23DA;
const char32_t FUSE = 0x23DB;
const char32_t TOP_PARENTHESIS = 0x23DC;
const char32_t BOTTOM_PARENTHESIS = 0x23DD;
const char32_t TOP_CURLY_BRACKET = 0x23DE;
const char32_t BOTTOM_CURLY_BRACKET = 0x23DF;
const char32_t TOP_TORTOISE_SHELL_BRACKET = 0x23E0;
const char32_t BOTTOM_TORTOISE_SHELL_BRACKET = 0x23E1;
const char32_t WHITE_TRAPEZIUM = 0x23E2;
const char32_t BENZENE_RING_WITH_CIRCLE = 0x23E3;
const char32_t STRAIGHTNESS = 0x23E4;
const char32_t FLATNESS = 0x23E5;
const char32_t AC_CURRENT = 0x23E6;
const char32_t ELECTRICAL_INTERSECTION = 0x23E7;
const char32_t DECIMAL_EXPONENT_SYMBOL = 0x23E8;
const char32_t BLACK_RIGHT_POINTING_DOUBLE_TRIANGLE = 0x23E9;
const char32_t BLACK_LEFT_POINTING_DOUBLE_TRIANGLE = 0x23EA;
const char32_t BLACK_UP_POINTING_DOUBLE_TRIANGLE = 0x23EB;
const char32_t BLACK_DOWN_POINTING_DOUBLE_TRIANGLE = 0x23EC;
const char32_t BLACK_RIGHT_POINTING_DOUBLE_TRIANGLE_WITH_VERTICAL_BAR = 0x23ED;
const char32_t BLACK_LEFT_POINTING_DOUBLE_TRIANGLE_WITH_VERTICAL_BAR = 0x23EE;
const char32_t BLACK_RIGHT_POINTING_TRIANGLE_WITH_DOUBLE_VERTICAL_BAR = 0x23EF;
const char32_t ALARM_CLOCK = 0x23F0;
const char32_t STOPWATCH = 0x23F1;
const char32_t TIMER_CLOCK = 0x23F2;
const char32_t HOURGLASS_WITH_FLOWING_SAND = 0x23F3;
const char32_t BLACK_MEDIUM_LEFT_POINTING_TRIANGLE = 0x23F4;
const char32_t BLACK_MEDIUM_RIGHT_POINTING_TRIANGLE = 0x23F5;
const char32_t BLACK_MEDIUM_UP_POINTING_TRIANGLE = 0x23F6;
const char32_t BLACK_MEDIUM_DOWN_POINTING_TRIANGLE = 0x23F7;
const char32_t DOUBLE_VERTICAL_BAR = 0x23F8;
const char32_t BLACK_SQUARE_FOR_STOP = 0x23F9;
const char32_t BLACK_CIRCLE_FOR_RECORD = 0x23FA;
const char32_t POWER_SYMBOL = 0x23FB;
const char32_t POWER_ON_OFF_SYMBOL = 0x23FC;
const char32_t POWER_ON_SYMBOL = 0x23FD;
const char32_t POWER_SLEEP_SYMBOL = 0x23FE;
const char32_t OBSERVER_EYE_SYMBOL = 0x23FF;
const char32_t SYMBOL_FOR_NULL = 0x2400;
const char32_t SYMBOL_FOR_START_OF_HEADING = 0x2401;
const char32_t SYMBOL_FOR_START_OF_TEXT = 0x2402;
const char32_t SYMBOL_FOR_END_OF_TEXT = 0x2403;
const char32_t SYMBOL_FOR_END_OF_TRANSMISSION = 0x2404;
const char32_t SYMBOL_FOR_ENQUIRY = 0x2405;
const char32_t SYMBOL_FOR_ACKNOWLEDGE = 0x2406;
const char32_t SYMBOL_FOR_BELL = 0x2407;
const char32_t SYMBOL_FOR_BACKSPACE = 0x2408;
const char32_t SYMBOL_FOR_HORIZONTAL_TABULATION = 0x2409;
const char32_t SYMBOL_FOR_LINE_FEED = 0x240A;
const char32_t SYMBOL_FOR_VERTICAL_TABULATION = 0x240B;
const char32_t SYMBOL_FOR_FORM_FEED = 0x240C;
const char32_t SYMBOL_FOR_CARRIAGE_RETURN = 0x240D;
const char32_t SYMBOL_FOR_SHIFT_OUT = 0x240E;
const char32_t SYMBOL_FOR_SHIFT_IN = 0x240F;
const char32_t SYMBOL_FOR_DATA_LINK_ESCAPE = 0x2410;
const char32_t SYMBOL_FOR_DEVICE_CONTROL_ONE = 0x2411;
const char32_t SYMBOL_FOR_DEVICE_CONTROL_TWO = 0x2412;
const char32_t SYMBOL_FOR_DEVICE_CONTROL_THREE = 0x2413;
const char32_t SYMBOL_FOR_DEVICE_CONTROL_FOUR = 0x2414;
const char32_t SYMBOL_FOR_NEGATIVE_ACKNOWLEDGE = 0x2415;
const char32_t SYMBOL_FOR_SYNCHRONOUS_IDLE = 0x2416;
const char32_t SYMBOL_FOR_END_OF_TRANSMISSION_BLOCK = 0x2417;
const char32_t SYMBOL_FOR_CANCEL = 0x2418;
const char32_t SYMBOL_FOR_END_OF_MEDIUM = 0x2419;
const char32_t SYMBOL_FOR_SUBSTITUTE = 0x241A;
const char32_t SYMBOL_FOR_ESCAPE = 0x241B;
const char32_t SYMBOL_FOR_FILE_SEPARATOR = 0x241C;
const char32_t SYMBOL_FOR_GROUP_SEPARATOR = 0x241D;
const char32_t SYMBOL_FOR_RECORD_SEPARATOR = 0x241E;
const char32_t SYMBOL_FOR_UNIT_SEPARATOR = 0x241F;
const char32_t SYMBOL_FOR_SPACE = 0x2420;
const char32_t SYMBOL_FOR_DELETE = 0x2421;
const char32_t BLANK_SYMBOL = 0x2422;
const char32_t OPEN_BOX = 0x2423;
const char32_t SYMBOL_FOR_NEWLINE = 0x2424;
const char32_t SYMBOL_FOR_DELETE_FORM_TWO = 0x2425;
const char32_t SYMBOL_FOR_SUBSTITUTE_FORM_TWO = 0x2426;
const char32_t SYMBOL_FOR_DELETE_SQUARE_CHECKER_BOARD_FORM = 0x2427;
const char32_t SYMBOL_FOR_DELETE_RECTANGULAR_CHECKER_BOARD_FORM = 0x2428;
const char32_t SYMBOL_FOR_DELETE_MEDIUM_SHADE_FORM = 0x2429;
const char32_t OCR_HOOK = 0x2440;
const char32_t OCR_CHAIR = 0x2441;
const char32_t OCR_FORK = 0x2442;
const char32_t OCR_INVERTED_FORK = 0x2443;
const char32_t OCR_BELT_BUCKLE = 0x2444;
const char32_t OCR_BOW_TIE = 0x2445;
const char32_t OCR_BRANCH_BANK_IDENTIFICATION = 0x2446;
const char32_t OCR_AMOUNT_OF_CHECK = 0x2447;
const char32_t OCR_DASH = 0x2448;
const char32_t OCR_CUSTOMER_ACCOUNT_NUMBER = 0x2449;
const char32_t OCR_DOUBLE_BACKSLASH = 0x244A;
const char32_t CIRCLED_DIGIT_ONE = 0x2460;
const char32_t CIRCLED_DIGIT_TWO = 0x2461;
const char32_t CIRCLED_DIGIT_THREE = 0x2462;
const char32_t CIRCLED_DIGIT_FOUR = 0x2463;
const char32_t CIRCLED_DIGIT_FIVE = 0x2464;
const char32_t CIRCLED_DIGIT_SIX = 0x2465;
const char32_t CIRCLED_DIGIT_SEVEN = 0x2466;
const char32_t CIRCLED_DIGIT_EIGHT = 0x2467;
const char32_t CIRCLED_DIGIT_NINE = 0x2468;
const char32_t CIRCLED_NUMBER_TEN = 0x2469;
const char32_t CIRCLED_NUMBER_ELEVEN = 0x246A;
const char32_t CIRCLED_NUMBER_TWELVE = 0x246B;
const char32_t CIRCLED_NUMBER_THIRTEEN = 0x246C;
const char32_t CIRCLED_NUMBER_FOURTEEN = 0x246D;
const char32_t CIRCLED_NUMBER_FIFTEEN = 0x246E;
const char32_t CIRCLED_NUMBER_SIXTEEN = 0x246F;
const char32_t CIRCLED_NUMBER_SEVENTEEN = 0x2470;
const char32_t CIRCLED_NUMBER_EIGHTEEN = 0x2471;
const char32_t CIRCLED_NUMBER_NINETEEN = 0x2472;
const char32_t CIRCLED_NUMBER_TWENTY = 0x2473;
const char32_t PARENTHESIZED_DIGIT_ONE = 0x2474;
const char32_t PARENTHESIZED_DIGIT_TWO = 0x2475;
const char32_t PARENTHESIZED_DIGIT_THREE = 0x2476;
const char32_t PARENTHESIZED_DIGIT_FOUR = 0x2477;
const char32_t PARENTHESIZED_DIGIT_FIVE = 0x2478;
const char32_t PARENTHESIZED_DIGIT_SIX = 0x2479;
const char32_t PARENTHESIZED_DIGIT_SEVEN = 0x247A;
const char32_t PARENTHESIZED_DIGIT_EIGHT = 0x247B;
const char32_t PARENTHESIZED_DIGIT_NINE = 0x247C;
const char32_t PARENTHESIZED_NUMBER_TEN = 0x247D;
const char32_t PARENTHESIZED_NUMBER_ELEVEN = 0x247E;
const char32_t PARENTHESIZED_NUMBER_TWELVE = 0x247F;
const char32_t PARENTHESIZED_NUMBER_THIRTEEN = 0x2480;
const char32_t PARENTHESIZED_NUMBER_FOURTEEN = 0x2481;
const char32_t PARENTHESIZED_NUMBER_FIFTEEN = 0x2482;
const char32_t PARENTHESIZED_NUMBER_SIXTEEN = 0x2483;
const char32_t PARENTHESIZED_NUMBER_SEVENTEEN = 0x2484;
const char32_t PARENTHESIZED_NUMBER_EIGHTEEN = 0x2485;
const char32_t PARENTHESIZED_NUMBER_NINETEEN = 0x2486;
const char32_t PARENTHESIZED_NUMBER_TWENTY = 0x2487;
const char32_t DIGIT_ONE_FULL_STOP = 0x2488;
const char32_t DIGIT_TWO_FULL_STOP = 0x2489;
const char32_t DIGIT_THREE_FULL_STOP = 0x248A;
const char32_t DIGIT_FOUR_FULL_STOP = 0x248B;
const char32_t DIGIT_FIVE_FULL_STOP = 0x248C;
const char32_t DIGIT_SIX_FULL_STOP = 0x248D;
const char32_t DIGIT_SEVEN_FULL_STOP = 0x248E;
const char32_t DIGIT_EIGHT_FULL_STOP = 0x248F;
const char32_t DIGIT_NINE_FULL_STOP = 0x2490;
const char32_t NUMBER_TEN_FULL_STOP = 0x2491;
const char32_t NUMBER_ELEVEN_FULL_STOP = 0x2492;
const char32_t NUMBER_TWELVE_FULL_STOP = 0x2493;
const char32_t NUMBER_THIRTEEN_FULL_STOP = 0x2494;
const char32_t NUMBER_FOURTEEN_FULL_STOP = 0x2495;
const char32_t NUMBER_FIFTEEN_FULL_STOP = 0x2496;
const char32_t NUMBER_SIXTEEN_FULL_STOP = 0x2497;
const char32_t NUMBER_SEVENTEEN_FULL_STOP = 0x2498;
const char32_t NUMBER_EIGHTEEN_FULL_STOP = 0x2499;
const char32_t NUMBER_NINETEEN_FULL_STOP = 0x249A;
const char32_t NUMBER_TWENTY_FULL_STOP = 0x249B;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_A = 0x249C;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_B = 0x249D;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_C = 0x249E;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_D = 0x249F;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_E = 0x24A0;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_F = 0x24A1;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_G = 0x24A2;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_H = 0x24A3;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_I = 0x24A4;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_J = 0x24A5;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_K = 0x24A6;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_L = 0x24A7;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_M = 0x24A8;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_N = 0x24A9;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_O = 0x24AA;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_P = 0x24AB;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_Q = 0x24AC;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_R = 0x24AD;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_S = 0x24AE;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_T = 0x24AF;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_U = 0x24B0;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_V = 0x24B1;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_W = 0x24B2;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_X = 0x24B3;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_Y = 0x24B4;
const char32_t PARENTHESIZED_LATIN_SMALL_LETTER_Z = 0x24B5;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_A = 0x24B6;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_B = 0x24B7;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_C = 0x24B8;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_D = 0x24B9;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_E = 0x24BA;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_F = 0x24BB;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_G = 0x24BC;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_H = 0x24BD;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_I = 0x24BE;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_J = 0x24BF;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_K = 0x24C0;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_L = 0x24C1;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_M = 0x24C2;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_N = 0x24C3;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_O = 0x24C4;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_P = 0x24C5;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_Q = 0x24C6;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_R = 0x24C7;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_S = 0x24C8;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_T = 0x24C9;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_U = 0x24CA;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_V = 0x24CB;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_W = 0x24CC;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_X = 0x24CD;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_Y = 0x24CE;
const char32_t CIRCLED_LATIN_CAPITAL_LETTER_Z = 0x24CF;
const char32_t CIRCLED_LATIN_SMALL_LETTER_A = 0x24D0;
const char32_t CIRCLED_LATIN_SMALL_LETTER_B = 0x24D1;
const char32_t CIRCLED_LATIN_SMALL_LETTER_C = 0x24D2;
const char32_t CIRCLED_LATIN_SMALL_LETTER_D = 0x24D3;
const char32_t CIRCLED_LATIN_SMALL_LETTER_E = 0x24D4;
const char32_t CIRCLED_LATIN_SMALL_LETTER_F = 0x24D5;
const char32_t CIRCLED_LATIN_SMALL_LETTER_G = 0x24D6;
const char32_t CIRCLED_LATIN_SMALL_LETTER_H = 0x24D7;
const char32_t CIRCLED_LATIN_SMALL_LETTER_I = 0x24D8;
const char32_t CIRCLED_LATIN_SMALL_LETTER_J = 0x24D9;
const char32_t CIRCLED_LATIN_SMALL_LETTER_K = 0x24DA;
const char32_t CIRCLED_LATIN_SMALL_LETTER_L = 0x24DB;
const char32_t CIRCLED_LATIN_SMALL_LETTER_M = 0x24DC;
const char32_t CIRCLED_LATIN_SMALL_LETTER_N = 0x24DD;
const char32_t CIRCLED_LATIN_SMALL_LETTER_O = 0x24DE;
const char32_t CIRCLED_LATIN_SMALL_LETTER_P = 0x24DF;
const char32_t CIRCLED_LATIN_SMALL_LETTER_Q = 0x24E0;
const char32_t CIRCLED_LATIN_SMALL_LETTER_R = 0x24E1;
const char32_t CIRCLED_LATIN_SMALL_LETTER_S = 0x24E2;
const char32_t CIRCLED_LATIN_SMALL_LETTER_T = 0x24E3;
const char32_t CIRCLED_LATIN_SMALL_LETTER_U = 0x24E4;
const char32_t CIRCLED_LATIN_SMALL_LETTER_V = 0x24E5;
const char32_t CIRCLED_LATIN_SMALL_LETTER_W = 0x24E6;
const char32_t CIRCLED_LATIN_SMALL_LETTER_X = 0x24E7;
const char32_t CIRCLED_LATIN_SMALL_LETTER_Y = 0x24E8;
const char32_t CIRCLED_LATIN_SMALL_LETTER_Z = 0x24E9;
const char32_t CIRCLED_DIGIT_ZERO = 0x24EA;
const char32_t NEGATIVE_CIRCLED_NUMBER_ELEVEN = 0x24EB;
const char32_t NEGATIVE_CIRCLED_NUMBER_TWELVE = 0x24EC;
const char32_t NEGATIVE_CIRCLED_NUMBER_THIRTEEN = 0x24ED;
const char32_t NEGATIVE_CIRCLED_NUMBER_FOURTEEN = 0x24EE;
const char32_t NEGATIVE_CIRCLED_NUMBER_FIFTEEN = 0x24EF;
const char32_t NEGATIVE_CIRCLED_NUMBER_SIXTEEN = 0x24F0;
const char32_t NEGATIVE_CIRCLED_NUMBER_SEVENTEEN = 0x24F1;
const char32_t NEGATIVE_CIRCLED_NUMBER_EIGHTEEN = 0x24F2;
const char32_t NEGATIVE_CIRCLED_NUMBER_NINETEEN = 0x24F3;
const char32_t NEGATIVE_CIRCLED_NUMBER_TWENTY = 0x24F4;
const char32_t DOUBLE_CIRCLED_DIGIT_ONE = 0x24F5;
const char32_t DOUBLE_CIRCLED_DIGIT_TWO = 0x24F6;
const char32_t DOUBLE_CIRCLED_DIGIT_THREE = 0x24F7;
const char32_t DOUBLE_CIRCLED_DIGIT_FOUR = 0x24F8;
const char32_t DOUBLE_CIRCLED_DIGIT_FIVE = 0x24F9;
const char32_t DOUBLE_CIRCLED_DIGIT_SIX = 0x24FA;
const char32_t DOUBLE_CIRCLED_DIGIT_SEVEN = 0x24FB;
const char32_t DOUBLE_CIRCLED_DIGIT_EIGHT = 0x24FC;
const char32_t DOUBLE_CIRCLED_DIGIT_NINE = 0x24FD;
const char32_t DOUBLE_CIRCLED_NUMBER_TEN = 0x24FE;
const char32_t NEGATIVE_CIRCLED_DIGIT_ZERO = 0x24FF;
const char32_t BOX_DRAWINGS_LIGHT_HORIZONTAL = 0x2500;
const char32_t BOX_DRAWINGS_HEAVY_HORIZONTAL = 0x2501;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL = 0x2502;
const char32_t BOX_DRAWINGS_HEAVY_VERTICAL = 0x2503;
const char32_t BOX_DRAWINGS_LIGHT_TRIPLE_DASH_HORIZONTAL = 0x2504;
const char32_t BOX_DRAWINGS_HEAVY_TRIPLE_DASH_HORIZONTAL = 0x2505;
const char32_t BOX_DRAWINGS_LIGHT_TRIPLE_DASH_VERTICAL = 0x2506;
const char32_t BOX_DRAWINGS_HEAVY_TRIPLE_DASH_VERTICAL = 0x2507;
const char32_t BOX_DRAWINGS_LIGHT_QUADRUPLE_DASH_HORIZONTAL = 0x2508;
const char32_t BOX_DRAWINGS_HEAVY_QUADRUPLE_DASH_HORIZONTAL = 0x2509;
const char32_t BOX_DRAWINGS_LIGHT_QUADRUPLE_DASH_VERTICAL = 0x250A;
const char32_t BOX_DRAWINGS_HEAVY_QUADRUPLE_DASH_VERTICAL = 0x250B;
const char32_t BOX_DRAWINGS_LIGHT_DOWN_AND_RIGHT = 0x250C;
const char32_t BOX_DRAWINGS_DOWN_LIGHT_AND_RIGHT_HEAVY = 0x250D;
const char32_t BOX_DRAWINGS_DOWN_HEAVY_AND_RIGHT_LIGHT = 0x250E;
const char32_t BOX_DRAWINGS_HEAVY_DOWN_AND_RIGHT = 0x250F;
const char32_t BOX_DRAWINGS_LIGHT_DOWN_AND_LEFT = 0x2510;
const char32_t BOX_DRAWINGS_DOWN_LIGHT_AND_LEFT_HEAVY = 0x2511;
const char32_t BOX_DRAWINGS_DOWN_HEAVY_AND_LEFT_LIGHT = 0x2512;
const char32_t BOX_DRAWINGS_HEAVY_DOWN_AND_LEFT = 0x2513;
const char32_t BOX_DRAWINGS_LIGHT_UP_AND_RIGHT = 0x2514;
const char32_t BOX_DRAWINGS_UP_LIGHT_AND_RIGHT_HEAVY = 0x2515;
const char32_t BOX_DRAWINGS_UP_HEAVY_AND_RIGHT_LIGHT = 0x2516;
const char32_t BOX_DRAWINGS_HEAVY_UP_AND_RIGHT = 0x2517;
const char32_t BOX_DRAWINGS_LIGHT_UP_AND_LEFT = 0x2518;
const char32_t BOX_DRAWINGS_UP_LIGHT_AND_LEFT_HEAVY = 0x2519;
const char32_t BOX_DRAWINGS_UP_HEAVY_AND_LEFT_LIGHT = 0x251A;
const char32_t BOX_DRAWINGS_HEAVY_UP_AND_LEFT = 0x251B;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL_AND_RIGHT = 0x251C;
const char32_t BOX_DRAWINGS_VERTICAL_LIGHT_AND_RIGHT_HEAVY = 0x251D;
const char32_t BOX_DRAWINGS_UP_HEAVY_AND_RIGHT_DOWN_LIGHT = 0x251E;
const char32_t BOX_DRAWINGS_DOWN_HEAVY_AND_RIGHT_UP_LIGHT = 0x251F;
const char32_t BOX_DRAWINGS_VERTICAL_HEAVY_AND_RIGHT_LIGHT = 0x2520;
const char32_t BOX_DRAWINGS_DOWN_LIGHT_AND_RIGHT_UP_HEAVY = 0x2521;
const char32_t BOX_DRAWINGS_UP_LIGHT_AND_RIGHT_DOWN_HEAVY = 0x2522;
const char32_t BOX_DRAWINGS_HEAVY_VERTICAL_AND_RIGHT = 0x2523;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL_AND_LEFT = 0x2524;
const char32_t BOX_DRAWINGS_VERTICAL_LIGHT_AND_LEFT_HEAVY = 0x2525;
const char32_t BOX_DRAWINGS_UP_HEAVY_AND_LEFT_DOWN_LIGHT = 0x2526;
const char32_t BOX_DRAWINGS_DOWN_HEAVY_AND_LEFT_UP_LIGHT = 0x2527;
const char32_t BOX_DRAWINGS_VERTICAL_HEAVY_AND_LEFT_LIGHT = 0x2528;
const char32_t BOX_DRAWINGS_DOWN_LIGHT_AND_LEFT_UP_HEAVY = 0x2529;
const char32_t BOX_DRAWINGS_UP_LIGHT_AND_LEFT_DOWN_HEAVY = 0x252A;
const char32_t BOX_DRAWINGS_HEAVY_VERTICAL_AND_LEFT = 0x252B;
const char32_t BOX_DRAWINGS_LIGHT_DOWN_AND_HORIZONTAL = 0x252C;
const char32_t BOX_DRAWINGS_LEFT_HEAVY_AND_RIGHT_DOWN_LIGHT = 0x252D;
const char32_t BOX_DRAWINGS_RIGHT_HEAVY_AND_LEFT_DOWN_LIGHT = 0x252E;
const char32_t BOX_DRAWINGS_DOWN_LIGHT_AND_HORIZONTAL_HEAVY = 0x252F;
const char32_t BOX_DRAWINGS_DOWN_HEAVY_AND_HORIZONTAL_LIGHT = 0x2530;
const char32_t BOX_DRAWINGS_RIGHT_LIGHT_AND_LEFT_DOWN_HEAVY = 0x2531;
const char32_t BOX_DRAWINGS_LEFT_LIGHT_AND_RIGHT_DOWN_HEAVY = 0x2532;
const char32_t BOX_DRAWINGS_HEAVY_DOWN_AND_HORIZONTAL = 0x2533;
const char32_t BOX_DRAWINGS_LIGHT_UP_AND_HORIZONTAL = 0x2534;
const char32_t BOX_DRAWINGS_LEFT_HEAVY_AND_RIGHT_UP_LIGHT = 0x2535;
const char32_t BOX_DRAWINGS_RIGHT_HEAVY_AND_LEFT_UP_LIGHT = 0x2536;
const char32_t BOX_DRAWINGS_UP_LIGHT_AND_HORIZONTAL_HEAVY = 0x2537;
const char32_t BOX_DRAWINGS_UP_HEAVY_AND_HORIZONTAL_LIGHT = 0x2538;
const char32_t BOX_DRAWINGS_RIGHT_LIGHT_AND_LEFT_UP_HEAVY = 0x2539;
const char32_t BOX_DRAWINGS_LEFT_LIGHT_AND_RIGHT_UP_HEAVY = 0x253A;
const char32_t BOX_DRAWINGS_HEAVY_UP_AND_HORIZONTAL = 0x253B;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL_AND_HORIZONTAL = 0x253C;
const char32_t BOX_DRAWINGS_LEFT_HEAVY_AND_RIGHT_VERTICAL_LIGHT = 0x253D;
const char32_t BOX_DRAWINGS_RIGHT_HEAVY_AND_LEFT_VERTICAL_LIGHT = 0x253E;
const char32_t BOX_DRAWINGS_VERTICAL_LIGHT_AND_HORIZONTAL_HEAVY = 0x253F;
const char32_t BOX_DRAWINGS_UP_HEAVY_AND_DOWN_HORIZONTAL_LIGHT = 0x2540;
const char32_t BOX_DRAWINGS_DOWN_HEAVY_AND_UP_HORIZONTAL_LIGHT = 0x2541;
const char32_t BOX_DRAWINGS_VERTICAL_HEAVY_AND_HORIZONTAL_LIGHT = 0x2542;
const char32_t BOX_DRAWINGS_LEFT_UP_HEAVY_AND_RIGHT_DOWN_LIGHT = 0x2543;
const char32_t BOX_DRAWINGS_RIGHT_UP_HEAVY_AND_LEFT_DOWN_LIGHT = 0x2544;
const char32_t BOX_DRAWINGS_LEFT_DOWN_HEAVY_AND_RIGHT_UP_LIGHT = 0x2545;
const char32_t BOX_DRAWINGS_RIGHT_DOWN_HEAVY_AND_LEFT_UP_LIGHT = 0x2546;
const char32_t BOX_DRAWINGS_DOWN_LIGHT_AND_UP_HORIZONTAL_HEAVY = 0x2547;
const char32_t BOX_DRAWINGS_UP_LIGHT_AND_DOWN_HORIZONTAL_HEAVY = 0x2548;
const char32_t BOX_DRAWINGS_RIGHT_LIGHT_AND_LEFT_VERTICAL_HEAVY = 0x2549;
const char32_t BOX_DRAWINGS_LEFT_LIGHT_AND_RIGHT_VERTICAL_HEAVY = 0x254A;
const char32_t BOX_DRAWINGS_HEAVY_VERTICAL_AND_HORIZONTAL = 0x254B;
const char32_t BOX_DRAWINGS_LIGHT_DOUBLE_DASH_HORIZONTAL = 0x254C;
const char32_t BOX_DRAWINGS_HEAVY_DOUBLE_DASH_HORIZONTAL = 0x254D;
const char32_t BOX_DRAWINGS_LIGHT_DOUBLE_DASH_VERTICAL = 0x254E;
const char32_t BOX_DRAWINGS_HEAVY_DOUBLE_DASH_VERTICAL = 0x254F;
const char32_t BOX_DRAWINGS_DOUBLE_HORIZONTAL = 0x2550;
const char32_t BOX_DRAWINGS_DOUBLE_VERTICAL = 0x2551;
const char32_t BOX_DRAWINGS_DOWN_SINGLE_AND_RIGHT_DOUBLE = 0x2552;
const char32_t BOX_DRAWINGS_DOWN_DOUBLE_AND_RIGHT_SINGLE = 0x2553;
const char32_t BOX_DRAWINGS_DOUBLE_DOWN_AND_RIGHT = 0x2554;
const char32_t BOX_DRAWINGS_DOWN_SINGLE_AND_LEFT_DOUBLE = 0x2555;
const char32_t BOX_DRAWINGS_DOWN_DOUBLE_AND_LEFT_SINGLE = 0x2556;
const char32_t BOX_DRAWINGS_DOUBLE_DOWN_AND_LEFT = 0x2557;
const char32_t BOX_DRAWINGS_UP_SINGLE_AND_RIGHT_DOUBLE = 0x2558;
const char32_t BOX_DRAWINGS_UP_DOUBLE_AND_RIGHT_SINGLE = 0x2559;
const char32_t BOX_DRAWINGS_DOUBLE_UP_AND_RIGHT = 0x255A;
const char32_t BOX_DRAWINGS_UP_SINGLE_AND_LEFT_DOUBLE = 0x255B;
const char32_t BOX_DRAWINGS_UP_DOUBLE_AND_LEFT_SINGLE = 0x255C;
const char32_t BOX_DRAWINGS_DOUBLE_UP_AND_LEFT = 0x255D;
const char32_t BOX_DRAWINGS_VERTICAL_SINGLE_AND_RIGHT_DOUBLE = 0x255E;
const char32_t BOX_DRAWINGS_VERTICAL_DOUBLE_AND_RIGHT_SINGLE = 0x255F;
const char32_t BOX_DRAWINGS_DOUBLE_VERTICAL_AND_RIGHT = 0x2560;
const char32_t BOX_DRAWINGS_VERTICAL_SINGLE_AND_LEFT_DOUBLE = 0x2561;
const char32_t BOX_DRAWINGS_VERTICAL_DOUBLE_AND_LEFT_SINGLE = 0x2562;
const char32_t BOX_DRAWINGS_DOUBLE_VERTICAL_AND_LEFT = 0x2563;
const char32_t BOX_DRAWINGS_DOWN_SINGLE_AND_HORIZONTAL_DOUBLE = 0x2564;
const char32_t BOX_DRAWINGS_DOWN_DOUBLE_AND_HORIZONTAL_SINGLE = 0x2565;
const char32_t BOX_DRAWINGS_DOUBLE_DOWN_AND_HORIZONTAL = 0x2566;
const char32_t BOX_DRAWINGS_UP_SINGLE_AND_HORIZONTAL_DOUBLE = 0x2567;
const char32_t BOX_DRAWINGS_UP_DOUBLE_AND_HORIZONTAL_SINGLE = 0x2568;
const char32_t BOX_DRAWINGS_DOUBLE_UP_AND_HORIZONTAL = 0x2569;
const char32_t BOX_DRAWINGS_VERTICAL_SINGLE_AND_HORIZONTAL_DOUBLE = 0x256A;
const char32_t BOX_DRAWINGS_VERTICAL_DOUBLE_AND_HORIZONTAL_SINGLE = 0x256B;
const char32_t BOX_DRAWINGS_DOUBLE_VERTICAL_AND_HORIZONTAL = 0x256C;
const char32_t BOX_DRAWINGS_LIGHT_ARC_DOWN_AND_RIGHT = 0x256D;
const char32_t BOX_DRAWINGS_LIGHT_ARC_DOWN_AND_LEFT = 0x256E;
const char32_t BOX_DRAWINGS_LIGHT_ARC_UP_AND_LEFT = 0x256F;
const char32_t BOX_DRAWINGS_LIGHT_ARC_UP_AND_RIGHT = 0x2570;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_RIGHT_TO_LOWER_LEFT = 0x2571;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_LEFT_TO_LOWER_RIGHT = 0x2572;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_CROSS = 0x2573;
const char32_t BOX_DRAWINGS_LIGHT_LEFT = 0x2574;
const char32_t BOX_DRAWINGS_LIGHT_UP = 0x2575;
const char32_t BOX_DRAWINGS_LIGHT_RIGHT = 0x2576;
const char32_t BOX_DRAWINGS_LIGHT_DOWN = 0x2577;
const char32_t BOX_DRAWINGS_HEAVY_LEFT = 0x2578;
const char32_t BOX_DRAWINGS_HEAVY_UP = 0x2579;
const char32_t BOX_DRAWINGS_HEAVY_RIGHT = 0x257A;
const char32_t BOX_DRAWINGS_HEAVY_DOWN = 0x257B;
const char32_t BOX_DRAWINGS_LIGHT_LEFT_AND_HEAVY_RIGHT = 0x257C;
const char32_t BOX_DRAWINGS_LIGHT_UP_AND_HEAVY_DOWN = 0x257D;
const char32_t BOX_DRAWINGS_HEAVY_LEFT_AND_LIGHT_RIGHT = 0x257E;
const char32_t BOX_DRAWINGS_HEAVY_UP_AND_LIGHT_DOWN = 0x257F;
const char32_t UPPER_HALF_BLOCK = 0x2580;
const char32_t LOWER_ONE_EIGHTH_BLOCK = 0x2581;
const char32_t LOWER_ONE_QUARTER_BLOCK = 0x2582;
const char32_t LOWER_THREE_EIGHTHS_BLOCK = 0x2583;
const char32_t LOWER_HALF_BLOCK = 0x2584;
const char32_t LOWER_FIVE_EIGHTHS_BLOCK = 0x2585;
const char32_t LOWER_THREE_QUARTERS_BLOCK = 0x2586;
const char32_t LOWER_SEVEN_EIGHTHS_BLOCK = 0x2587;
const char32_t FULL_BLOCK = 0x2588;
const char32_t LEFT_SEVEN_EIGHTHS_BLOCK = 0x2589;
const char32_t LEFT_THREE_QUARTERS_BLOCK = 0x258A;
const char32_t LEFT_FIVE_EIGHTHS_BLOCK = 0x258B;
const char32_t LEFT_HALF_BLOCK = 0x258C;
const char32_t LEFT_THREE_EIGHTHS_BLOCK = 0x258D;
const char32_t LEFT_ONE_QUARTER_BLOCK = 0x258E;
const char32_t LEFT_ONE_EIGHTH_BLOCK = 0x258F;
const char32_t RIGHT_HALF_BLOCK = 0x2590;
const char32_t LIGHT_SHADE = 0x2591;
const char32_t MEDIUM_SHADE = 0x2592;
const char32_t DARK_SHADE = 0x2593;
const char32_t UPPER_ONE_EIGHTH_BLOCK = 0x2594;
const char32_t RIGHT_ONE_EIGHTH_BLOCK = 0x2595;
const char32_t QUADRANT_LOWER_LEFT = 0x2596;
const char32_t QUADRANT_LOWER_RIGHT = 0x2597;
const char32_t QUADRANT_UPPER_LEFT = 0x2598;
const char32_t QUADRANT_UPPER_LEFT_AND_LOWER_LEFT_AND_LOWER_RIGHT = 0x2599;
const char32_t QUADRANT_UPPER_LEFT_AND_LOWER_RIGHT = 0x259A;
const char32_t QUADRANT_UPPER_LEFT_AND_UPPER_RIGHT_AND_LOWER_LEFT = 0x259B;
const char32_t QUADRANT_UPPER_LEFT_AND_UPPER_RIGHT_AND_LOWER_RIGHT = 0x259C;
const char32_t QUADRANT_UPPER_RIGHT = 0x259D;
const char32_t QUADRANT_UPPER_RIGHT_AND_LOWER_LEFT = 0x259E;
const char32_t QUADRANT_UPPER_RIGHT_AND_LOWER_LEFT_AND_LOWER_RIGHT = 0x259F;
const char32_t BLACK_SQUARE = 0x25A0;
const char32_t WHITE_SQUARE = 0x25A1;
const char32_t WHITE_SQUARE_WITH_ROUNDED_CORNERS = 0x25A2;
const char32_t WHITE_SQUARE_CONTAINING_BLACK_SMALL_SQUARE = 0x25A3;
const char32_t SQUARE_WITH_HORIZONTAL_FILL = 0x25A4;
const char32_t SQUARE_WITH_VERTICAL_FILL = 0x25A5;
const char32_t SQUARE_WITH_ORTHOGONAL_CROSSHATCH_FILL = 0x25A6;
const char32_t SQUARE_WITH_UPPER_LEFT_TO_LOWER_RIGHT_FILL = 0x25A7;
const char32_t SQUARE_WITH_UPPER_RIGHT_TO_LOWER_LEFT_FILL = 0x25A8;
const char32_t SQUARE_WITH_DIAGONAL_CROSSHATCH_FILL = 0x25A9;
const char32_t BLACK_SMALL_SQUARE = 0x25AA;
const char32_t WHITE_SMALL_SQUARE = 0x25AB;
const char32_t BLACK_RECTANGLE = 0x25AC;
const char32_t WHITE_RECTANGLE = 0x25AD;
const char32_t BLACK_VERTICAL_RECTANGLE = 0x25AE;
const char32_t WHITE_VERTICAL_RECTANGLE = 0x25AF;
const char32_t BLACK_PARALLELOGRAM = 0x25B0;
const char32_t WHITE_PARALLELOGRAM = 0x25B1;
const char32_t BLACK_UP_POINTING_TRIANGLE = 0x25B2;
const char32_t WHITE_UP_POINTING_TRIANGLE = 0x25B3;
const char32_t BLACK_UP_POINTING_SMALL_TRIANGLE = 0x25B4;
const char32_t WHITE_UP_POINTING_SMALL_TRIANGLE = 0x25B5;
const char32_t BLACK_RIGHT_POINTING_TRIANGLE = 0x25B6;
const char32_t WHITE_RIGHT_POINTING_TRIANGLE = 0x25B7;
const char32_t BLACK_RIGHT_POINTING_SMALL_TRIANGLE = 0x25B8;
const char32_t WHITE_RIGHT_POINTING_SMALL_TRIANGLE = 0x25B9;
const char32_t BLACK_RIGHT_POINTING_POINTER = 0x25BA;
const char32_t WHITE_RIGHT_POINTING_POINTER = 0x25BB;
const char32_t BLACK_DOWN_POINTING_TRIANGLE = 0x25BC;
const char32_t WHITE_DOWN_POINTING_TRIANGLE = 0x25BD;
const char32_t BLACK_DOWN_POINTING_SMALL_TRIANGLE = 0x25BE;
const char32_t WHITE_DOWN_POINTING_SMALL_TRIANGLE = 0x25BF;
const char32_t BLACK_LEFT_POINTING_TRIANGLE = 0x25C0;
const char32_t WHITE_LEFT_POINTING_TRIANGLE = 0x25C1;
const char32_t BLACK_LEFT_POINTING_SMALL_TRIANGLE = 0x25C2;
const char32_t WHITE_LEFT_POINTING_SMALL_TRIANGLE = 0x25C3;
const char32_t BLACK_LEFT_POINTING_POINTER = 0x25C4;
const char32_t WHITE_LEFT_POINTING_POINTER = 0x25C5;
const char32_t BLACK_DIAMOND = 0x25C6;
const char32_t WHITE_DIAMOND = 0x25C7;
const char32_t WHITE_DIAMOND_CONTAINING_BLACK_SMALL_DIAMOND = 0x25C8;
const char32_t FISHEYE = 0x25C9;
const char32_t LOZENGE = 0x25CA;
const char32_t WHITE_CIRCLE = 0x25CB;
const char32_t DOTTED_CIRCLE = 0x25CC;
const char32_t CIRCLE_WITH_VERTICAL_FILL = 0x25CD;
const char32_t BULLSEYE = 0x25CE;
const char32_t BLACK_CIRCLE = 0x25CF;
const char32_t CIRCLE_WITH_LEFT_HALF_BLACK = 0x25D0;
const char32_t CIRCLE_WITH_RIGHT_HALF_BLACK = 0x25D1;
const char32_t CIRCLE_WITH_LOWER_HALF_BLACK = 0x25D2;
const char32_t CIRCLE_WITH_UPPER_HALF_BLACK = 0x25D3;
const char32_t CIRCLE_WITH_UPPER_RIGHT_QUADRANT_BLACK = 0x25D4;
const char32_t CIRCLE_WITH_ALL_BUT_UPPER_LEFT_QUADRANT_BLACK = 0x25D5;
const char32_t LEFT_HALF_BLACK_CIRCLE = 0x25D6;
const char32_t RIGHT_HALF_BLACK_CIRCLE = 0x25D7;
const char32_t INVERSE_BULLET = 0x25D8;
const char32_t INVERSE_WHITE_CIRCLE = 0x25D9;
const char32_t UPPER_HALF_INVERSE_WHITE_CIRCLE = 0x25DA;
const char32_t LOWER_HALF_INVERSE_WHITE_CIRCLE = 0x25DB;
const char32_t UPPER_LEFT_QUADRANT_CIRCULAR_ARC = 0x25DC;
const char32_t UPPER_RIGHT_QUADRANT_CIRCULAR_ARC = 0x25DD;
const char32_t LOWER_RIGHT_QUADRANT_CIRCULAR_ARC = 0x25DE;
const char32_t LOWER_LEFT_QUADRANT_CIRCULAR_ARC = 0x25DF;
const char32_t UPPER_HALF_CIRCLE = 0x25E0;
const char32_t LOWER_HALF_CIRCLE = 0x25E1;
const char32_t BLACK_LOWER_RIGHT_TRIANGLE = 0x25E2;
const char32_t BLACK_LOWER_LEFT_TRIANGLE = 0x25E3;
const char32_t BLACK_UPPER_LEFT_TRIANGLE = 0x25E4;
const char32_t BLACK_UPPER_RIGHT_TRIANGLE = 0x25E5;
const char32_t WHITE_BULLET = 0x25E6;
const char32_t SQUARE_WITH_LEFT_HALF_BLACK = 0x25E7;
const char32_t SQUARE_WITH_RIGHT_HALF_BLACK = 0x25E8;
const char32_t SQUARE_WITH_UPPER_LEFT_DIAGONAL_HALF_BLACK = 0x25E9;
const char32_t SQUARE_WITH_LOWER_RIGHT_DIAGONAL_HALF_BLACK = 0x25EA;
const char32_t WHITE_SQUARE_WITH_VERTICAL_BISECTING_LINE = 0x25EB;
const char32_t WHITE_UP_POINTING_TRIANGLE_WITH_DOT = 0x25EC;
const char32_t UP_POINTING_TRIANGLE_WITH_LEFT_HALF_BLACK = 0x25ED;
const char32_t UP_POINTING_TRIANGLE_WITH_RIGHT_HALF_BLACK = 0x25EE;
const char32_t LARGE_CIRCLE = 0x25EF;
const char32_t WHITE_SQUARE_WITH_UPPER_LEFT_QUADRANT = 0x25F0;
const char32_t WHITE_SQUARE_WITH_LOWER_LEFT_QUADRANT = 0x25F1;
const char32_t WHITE_SQUARE_WITH_LOWER_RIGHT_QUADRANT = 0x25F2;
const char32_t WHITE_SQUARE_WITH_UPPER_RIGHT_QUADRANT = 0x25F3;
const char32_t WHITE_CIRCLE_WITH_UPPER_LEFT_QUADRANT = 0x25F4;
const char32_t WHITE_CIRCLE_WITH_LOWER_LEFT_QUADRANT = 0x25F5;
const char32_t WHITE_CIRCLE_WITH_LOWER_RIGHT_QUADRANT = 0x25F6;
const char32_t WHITE_CIRCLE_WITH_UPPER_RIGHT_QUADRANT = 0x25F7;
const char32_t UPPER_LEFT_TRIANGLE = 0x25F8;
const char32_t UPPER_RIGHT_TRIANGLE = 0x25F9;
const char32_t LOWER_LEFT_TRIANGLE = 0x25FA;
const char32_t WHITE_MEDIUM_SQUARE = 0x25FB;
const char32_t BLACK_MEDIUM_SQUARE = 0x25FC;
const char32_t WHITE_MEDIUM_SMALL_SQUARE = 0x25FD;
const char32_t BLACK_MEDIUM_SMALL_SQUARE = 0x25FE;
const char32_t LOWER_RIGHT_TRIANGLE = 0x25FF;
const char32_t BLACK_SUN_WITH_RAYS = 0x2600;
const char32_t CLOUD = 0x2601;
const char32_t UMBRELLA = 0x2602;
const char32_t SNOWMAN = 0x2603;
const char32_t COMET = 0x2604;
const char32_t BLACK_STAR = 0x2605;
const char32_t WHITE_STAR = 0x2606;
const char32_t LIGHTNING = 0x2607;
const char32_t THUNDERSTORM = 0x2608;
const char32_t SUN = 0x2609;
const char32_t ASCENDING_NODE = 0x260A;
const char32_t DESCENDING_NODE = 0x260B;
const char32_t CONJUNCTION = 0x260C;
const char32_t OPPOSITION = 0x260D;
const char32_t BLACK_TELEPHONE = 0x260E;
const char32_t WHITE_TELEPHONE = 0x260F;
const char32_t BALLOT_BOX = 0x2610;
const char32_t BALLOT_BOX_WITH_CHECK = 0x2611;
const char32_t BALLOT_BOX_WITH_X = 0x2612;
const char32_t SALTIRE = 0x2613;
const char32_t UMBRELLA_WITH_RAIN_DROPS = 0x2614;
const char32_t HOT_BEVERAGE = 0x2615;
const char32_t WHITE_SHOGI_PIECE = 0x2616;
const char32_t BLACK_SHOGI_PIECE = 0x2617;
const char32_t SHAMROCK = 0x2618;
const char32_t REVERSED_ROTATED_FLORAL_HEART_BULLET = 0x2619;
const char32_t BLACK_LEFT_POINTING_INDEX = 0x261A;
const char32_t BLACK_RIGHT_POINTING_INDEX = 0x261B;
const char32_t WHITE_LEFT_POINTING_INDEX = 0x261C;
const char32_t WHITE_UP_POINTING_INDEX = 0x261D;
const char32_t WHITE_RIGHT_POINTING_INDEX = 0x261E;
const char32_t WHITE_DOWN_POINTING_INDEX = 0x261F;
const char32_t SKULL_AND_CROSSBONES = 0x2620;
const char32_t CAUTION_SIGN = 0x2621;
const char32_t RADIOACTIVE_SIGN = 0x2622;
const char32_t BIOHAZARD_SIGN = 0x2623;
const char32_t CADUCEUS = 0x2624;
const char32_t ANKH = 0x2625;
const char32_t ORTHODOX_CROSS = 0x2626;
const char32_t CHI_RHO = 0x2627;
const char32_t CROSS_OF_LORRAINE = 0x2628;
const char32_t CROSS_OF_JERUSALEM = 0x2629;
const char32_t STAR_AND_CRESCENT = 0x262A;
const char32_t FARSI_SYMBOL = 0x262B;
const char32_t ADI_SHAKTI = 0x262C;
const char32_t HAMMER_AND_SICKLE = 0x262D;
const char32_t PEACE_SYMBOL = 0x262E;
const char32_t YIN_YANG = 0x262F;
const char32_t TRIGRAM_FOR_HEAVEN = 0x2630;
const char32_t TRIGRAM_FOR_LAKE = 0x2631;
const char32_t TRIGRAM_FOR_FIRE = 0x2632;
const char32_t TRIGRAM_FOR_THUNDER = 0x2633;
const char32_t TRIGRAM_FOR_WIND = 0x2634;
const char32_t TRIGRAM_FOR_WATER = 0x2635;
const char32_t TRIGRAM_FOR_MOUNTAIN = 0x2636;
const char32_t TRIGRAM_FOR_EARTH = 0x2637;
const char32_t WHEEL_OF_DHARMA = 0x2638;
const char32_t WHITE_FROWNING_FACE = 0x2639;
const char32_t WHITE_SMILING_FACE = 0x263A;
const char32_t BLACK_SMILING_FACE = 0x263B;
const char32_t WHITE_SUN_WITH_RAYS = 0x263C;
const char32_t FIRST_QUARTER_MOON = 0x263D;
const char32_t LAST_QUARTER_MOON = 0x263E;
const char32_t MERCURY = 0x263F;
const char32_t FEMALE_SIGN = 0x2640;
const char32_t EARTH = 0x2641;
const char32_t MALE_SIGN = 0x2642;
const char32_t JUPITER = 0x2643;
const char32_t SATURN = 0x2644;
const char32_t URANUS = 0x2645;
const char32_t NEPTUNE = 0x2646;
const char32_t PLUTO = 0x2647;
const char32_t ARIES = 0x2648;
const char32_t TAURUS = 0x2649;
const char32_t GEMINI = 0x264A;
const char32_t CANCER = 0x264B;
const char32_t LEO = 0x264C;
const char32_t VIRGO = 0x264D;
const char32_t LIBRA = 0x264E;
const char32_t SCORPIUS = 0x264F;
const char32_t SAGITTARIUS = 0x2650;
const char32_t CAPRICORN = 0x2651;
const char32_t AQUARIUS = 0x2652;
const char32_t PISCES = 0x2653;
const char32_t WHITE_CHESS_KING = 0x2654;
const char32_t WHITE_CHESS_QUEEN = 0x2655;
const char32_t WHITE_CHESS_ROOK = 0x2656;
const char32_t WHITE_CHESS_BISHOP = 0x2657;
const char32_t WHITE_CHESS_KNIGHT = 0x2658;
const char32_t WHITE_CHESS_PAWN = 0x2659;
const char32_t BLACK_CHESS_KING = 0x265A;
const char32_t BLACK_CHESS_QUEEN = 0x265B;
const char32_t BLACK_CHESS_ROOK = 0x265C;
const char32_t BLACK_CHESS_BISHOP = 0x265D;
const char32_t BLACK_CHESS_KNIGHT = 0x265E;
const char32_t BLACK_CHESS_PAWN = 0x265F;
const char32_t BLACK_SPADE_SUIT = 0x2660;
const char32_t WHITE_HEART_SUIT = 0x2661;
const char32_t WHITE_DIAMOND_SUIT = 0x2662;
const char32_t BLACK_CLUB_SUIT = 0x2663;
const char32_t WHITE_SPADE_SUIT = 0x2664;
const char32_t BLACK_HEART_SUIT = 0x2665;
const char32_t BLACK_DIAMOND_SUIT = 0x2666;
const char32_t WHITE_CLUB_SUIT = 0x2667;
const char32_t HOT_SPRINGS = 0x2668;
const char32_t QUARTER_NOTE = 0x2669;
const char32_t EIGHTH_NOTE = 0x266A;
const char32_t BEAMED_EIGHTH_NOTES = 0x266B;
const char32_t BEAMED_SIXTEENTH_NOTES = 0x266C;
const char32_t MUSIC_FLAT_SIGN = 0x266D;
const char32_t MUSIC_NATURAL_SIGN = 0x266E;
const char32_t MUSIC_SHARP_SIGN = 0x266F;
const char32_t WEST_SYRIAC_CROSS = 0x2670;
const char32_t EAST_SYRIAC_CROSS = 0x2671;
const char32_t UNIVERSAL_RECYCLING_SYMBOL = 0x2672;
const char32_t RECYCLING_SYMBOL_FOR_TYPE_1_PLASTICS = 0x2673;
const char32_t RECYCLING_SYMBOL_FOR_TYPE_2_PLASTICS = 0x2674;
const char32_t RECYCLING_SYMBOL_FOR_TYPE_3_PLASTICS = 0x2675;
const char32_t RECYCLING_SYMBOL_FOR_TYPE_4_PLASTICS = 0x2676;
const char32_t RECYCLING_SYMBOL_FOR_TYPE_5_PLASTICS = 0x2677;
const char32_t RECYCLING_SYMBOL_FOR_TYPE_6_PLASTICS = 0x2678;
const char32_t RECYCLING_SYMBOL_FOR_TYPE_7_PLASTICS = 0x2679;
const char32_t RECYCLING_SYMBOL_FOR_GENERIC_MATERIALS = 0x267A;
const char32_t BLACK_UNIVERSAL_RECYCLING_SYMBOL = 0x267B;
const char32_t RECYCLED_PAPER_SYMBOL = 0x267C;
const char32_t PARTIALLY_RECYCLED_PAPER_SYMBOL = 0x267D;
const char32_t PERMANENT_PAPER_SIGN = 0x267E;
const char32_t WHEELCHAIR_SYMBOL = 0x267F;
const char32_t DIE_FACE_1 = 0x2680;
const char32_t DIE_FACE_2 = 0x2681;
const char32_t DIE_FACE_3 = 0x2682;
const char32_t DIE_FACE_4 = 0x2683;
const char32_t DIE_FACE_5 = 0x2684;
const char32_t DIE_FACE_6 = 0x2685;
const char32_t WHITE_CIRCLE_WITH_DOT_RIGHT = 0x2686;
const char32_t WHITE_CIRCLE_WITH_TWO_DOTS = 0x2687;
const char32_t BLACK_CIRCLE_WITH_WHITE_DOT_RIGHT = 0x2688;
const char32_t BLACK_CIRCLE_WITH_TWO_WHITE_DOTS = 0x2689;
const char32_t MONOGRAM_FOR_YANG = 0x268A;
const char32_t MONOGRAM_FOR_YIN = 0x268B;
const char32_t DIGRAM_FOR_GREATER_YANG = 0x268C;
const char32_t DIGRAM_FOR_LESSER_YIN = 0x268D;
const char32_t DIGRAM_FOR_LESSER_YANG = 0x268E;
const char32_t DIGRAM_FOR_GREATER_YIN = 0x268F;
const char32_t WHITE_FLAG = 0x2690;
const char32_t BLACK_FLAG = 0x2691;
const char32_t HAMMER_AND_PICK = 0x2692;
const char32_t ANCHOR = 0x2693;
const char32_t CROSSED_SWORDS = 0x2694;
const char32_t STAFF_OF_AESCULAPIUS = 0x2695;
const char32_t SCALES = 0x2696;
const char32_t ALEMBIC = 0x2697;
const char32_t FLOWER = 0x2698;
const char32_t GEAR = 0x2699;
const char32_t STAFF_OF_HERMES = 0x269A;
const char32_t ATOM_SYMBOL = 0x269B;
const char32_t FLEUR_DE_LIS = 0x269C;
const char32_t OUTLINED_WHITE_STAR = 0x269D;
const char32_t THREE_LINES_CONVERGING_RIGHT = 0x269E;
const char32_t THREE_LINES_CONVERGING_LEFT = 0x269F;
const char32_t WARNING_SIGN = 0x26A0;
const char32_t HIGH_VOLTAGE_SIGN = 0x26A1;
const char32_t DOUBLED_FEMALE_SIGN = 0x26A2;
const char32_t DOUBLED_MALE_SIGN = 0x26A3;
const char32_t INTERLOCKED_FEMALE_AND_MALE_SIGN = 0x26A4;
const char32_t MALE_AND_FEMALE_SIGN = 0x26A5;
const char32_t MALE_WITH_STROKE_SIGN = 0x26A6;
const char32_t MALE_WITH_STROKE_AND_MALE_AND_FEMALE_SIGN = 0x26A7;
const char32_t VERTICAL_MALE_WITH_STROKE_SIGN = 0x26A8;
const char32_t HORIZONTAL_MALE_WITH_STROKE_SIGN = 0x26A9;
const char32_t MEDIUM_WHITE_CIRCLE = 0x26AA;
const char32_t MEDIUM_BLACK_CIRCLE = 0x26AB;
const char32_t MEDIUM_SMALL_WHITE_CIRCLE = 0x26AC;
const char32_t MARRIAGE_SYMBOL = 0x26AD;
const char32_t DIVORCE_SYMBOL = 0x26AE;
const char32_t UNMARRIED_PARTNERSHIP_SYMBOL = 0x26AF;
const char32_t COFFIN = 0x26B0;
const char32_t FUNERAL_URN = 0x26B1;
const char32_t NEUTER = 0x26B2;
const char32_t CERES = 0x26B3;
const char32_t PALLAS = 0x26B4;
const char32_t JUNO = 0x26B5;
const char32_t VESTA = 0x26B6;
const char32_t CHIRON = 0x26B7;
const char32_t BLACK_MOON_LILITH = 0x26B8;
const char32_t SEXTILE = 0x26B9;
const char32_t SEMISEXTILE = 0x26BA;
const char32_t QUINCUNX = 0x26BB;
const char32_t SESQUIQUADRATE = 0x26BC;
const char32_t SOCCER_BALL = 0x26BD;
const char32_t BASEBALL = 0x26BE;
const char32_t SQUARED_KEY = 0x26BF;
const char32_t WHITE_DRAUGHTS_MAN = 0x26C0;
const char32_t WHITE_DRAUGHTS_KING = 0x26C1;
const char32_t BLACK_DRAUGHTS_MAN = 0x26C2;
const char32_t BLACK_DRAUGHTS_KING = 0x26C3;
const char32_t SNOWMAN_WITHOUT_SNOW = 0x26C4;
const char32_t SUN_BEHIND_CLOUD = 0x26C5;
const char32_t RAIN = 0x26C6;
const char32_t BLACK_SNOWMAN = 0x26C7;
const char32_t THUNDER_CLOUD_AND_RAIN = 0x26C8;
const char32_t TURNED_WHITE_SHOGI_PIECE = 0x26C9;
const char32_t TURNED_BLACK_SHOGI_PIECE = 0x26CA;
const char32_t WHITE_DIAMOND_IN_SQUARE = 0x26CB;
const char32_t CROSSING_LANES = 0x26CC;
const char32_t DISABLED_CAR = 0x26CD;
const char32_t OPHIUCHUS = 0x26CE;
const char32_t PICK = 0x26CF;
const char32_t CAR_SLIDING = 0x26D0;
const char32_t HELMET_WITH_WHITE_CROSS = 0x26D1;
const char32_t CIRCLED_CROSSING_LANES = 0x26D2;
const char32_t CHAINS = 0x26D3;
const char32_t NO_ENTRY = 0x26D4;
const char32_t ALTERNATE_ONE_WAY_LEFT_WAY_TRAFFIC = 0x26D5;
const char32_t BLACK_TWO_WAY_LEFT_WAY_TRAFFIC = 0x26D6;
const char32_t WHITE_TWO_WAY_LEFT_WAY_TRAFFIC = 0x26D7;
const char32_t BLACK_LEFT_LANE_MERGE = 0x26D8;
const char32_t WHITE_LEFT_LANE_MERGE = 0x26D9;
const char32_t DRIVE_SLOW_SIGN = 0x26DA;
const char32_t HEAVY_WHITE_DOWN_POINTING_TRIANGLE = 0x26DB;
const char32_t LEFT_CLOSED_ENTRY = 0x26DC;
const char32_t SQUARED_SALTIRE = 0x26DD;
const char32_t FALLING_DIAGONAL_IN_WHITE_CIRCLE_IN_BLACK_SQUARE = 0x26DE;
const char32_t BLACK_TRUCK = 0x26DF;
const char32_t RESTRICTED_LEFT_ENTRY_1 = 0x26E0;
const char32_t RESTRICTED_LEFT_ENTRY_2 = 0x26E1;
const char32_t ASTRONOMICAL_SYMBOL_FOR_URANUS = 0x26E2;
const char32_t HEAVY_CIRCLE_WITH_STROKE_AND_TWO_DOTS_ABOVE = 0x26E3;
const char32_t PENTAGRAM = 0x26E4;
const char32_t RIGHT_HANDED_INTERLACED_PENTAGRAM = 0x26E5;
const char32_t LEFT_HANDED_INTERLACED_PENTAGRAM = 0x26E6;
const char32_t INVERTED_PENTAGRAM = 0x26E7;
const char32_t BLACK_CROSS_ON_SHIELD = 0x26E8;
const char32_t SHINTO_SHRINE = 0x26E9;
const char32_t CHURCH = 0x26EA;
const char32_t CASTLE = 0x26EB;
const char32_t HISTORIC_SITE = 0x26EC;
const char32_t GEAR_WITHOUT_HUB = 0x26ED;
const char32_t GEAR_WITH_HANDLES = 0x26EE;
const char32_t MAP_SYMBOL_FOR_LIGHTHOUSE = 0x26EF;
const char32_t MOUNTAIN = 0x26F0;
const char32_t UMBRELLA_ON_GROUND = 0x26F1;
const char32_t FOUNTAIN = 0x26F2;
const char32_t FLAG_IN_HOLE = 0x26F3;
const char32_t FERRY = 0x26F4;
const char32_t SAILBOAT = 0x26F5;
const char32_t SQUARE_FOUR_CORNERS = 0x26F6;
const char32_t SKIER = 0x26F7;
const char32_t ICE_SKATE = 0x26F8;
const char32_t PERSON_WITH_BALL = 0x26F9;
const char32_t TENT = 0x26FA;
const char32_t JAPANESE_BANK_SYMBOL = 0x26FB;
const char32_t HEADSTONE_GRAVEYARD_SYMBOL = 0x26FC;
const char32_t FUEL_PUMP = 0x26FD;
const char32_t CUP_ON_BLACK_SQUARE = 0x26FE;
const char32_t WHITE_FLAG_WITH_HORIZONTAL_MIDDLE_BLACK_STRIPE = 0x26FF;
const char32_t BLACK_SAFETY_SCISSORS = 0x2700;
const char32_t UPPER_BLADE_SCISSORS = 0x2701;
const char32_t BLACK_SCISSORS = 0x2702;
const char32_t LOWER_BLADE_SCISSORS = 0x2703;
const char32_t WHITE_SCISSORS = 0x2704;
const char32_t WHITE_HEAVY_CHECK_MARK = 0x2705;
const char32_t TELEPHONE_LOCATION_SIGN = 0x2706;
const char32_t TAPE_DRIVE = 0x2707;
const char32_t AIRPLANE = 0x2708;
const char32_t ENVELOPE = 0x2709;
const char32_t RAISED_FIST = 0x270A;
const char32_t RAISED_HAND = 0x270B;
const char32_t VICTORY_HAND = 0x270C;
const char32_t WRITING_HAND = 0x270D;
const char32_t LOWER_RIGHT_PENCIL = 0x270E;
const char32_t PENCIL = 0x270F;
const char32_t UPPER_RIGHT_PENCIL = 0x2710;
const char32_t WHITE_NIB = 0x2711;
const char32_t BLACK_NIB = 0x2712;
const char32_t CHECK_MARK = 0x2713;
const char32_t HEAVY_CHECK_MARK = 0x2714;
const char32_t MULTIPLICATION_X = 0x2715;
const char32_t HEAVY_MULTIPLICATION_X = 0x2716;
const char32_t BALLOT_X = 0x2717;
const char32_t HEAVY_BALLOT_X = 0x2718;
const char32_t OUTLINED_GREEK_CROSS = 0x2719;
const char32_t HEAVY_GREEK_CROSS = 0x271A;
const char32_t OPEN_CENTRE_CROSS = 0x271B;
const char32_t HEAVY_OPEN_CENTRE_CROSS = 0x271C;
const char32_t LATIN_CROSS = 0x271D;
const char32_t SHADOWED_WHITE_LATIN_CROSS = 0x271E;
const char32_t OUTLINED_LATIN_CROSS = 0x271F;
const char32_t MALTESE_CROSS = 0x2720;
const char32_t STAR_OF_DAVID = 0x2721;
const char32_t FOUR_TEARDROP_SPOKED_ASTERISK = 0x2722;
const char32_t FOUR_BALLOON_SPOKED_ASTERISK = 0x2723;
const char32_t HEAVY_FOUR_BALLOON_SPOKED_ASTERISK = 0x2724;
const char32_t FOUR_CLUB_SPOKED_ASTERISK = 0x2725;
const char32_t BLACK_FOUR_POINTED_STAR = 0x2726;
const char32_t WHITE_FOUR_POINTED_STAR = 0x2727;
const char32_t SPARKLES = 0x2728;
const char32_t STRESS_OUTLINED_WHITE_STAR = 0x2729;
const char32_t CIRCLED_WHITE_STAR = 0x272A;
const char32_t OPEN_CENTRE_BLACK_STAR = 0x272B;
const char32_t BLACK_CENTRE_WHITE_STAR = 0x272C;
const char32_t OUTLINED_BLACK_STAR = 0x272D;
const char32_t HEAVY_OUTLINED_BLACK_STAR = 0x272E;
const char32_t PINWHEEL_STAR = 0x272F;
const char32_t SHADOWED_WHITE_STAR = 0x2730;
const char32_t HEAVY_ASTERISK = 0x2731;
const char32_t OPEN_CENTRE_ASTERISK = 0x2732;
const char32_t EIGHT_SPOKED_ASTERISK = 0x2733;
const char32_t EIGHT_POINTED_BLACK_STAR = 0x2734;
const char32_t EIGHT_POINTED_PINWHEEL_STAR = 0x2735;
const char32_t SIX_POINTED_BLACK_STAR = 0x2736;
const char32_t EIGHT_POINTED_RECTILINEAR_BLACK_STAR = 0x2737;
const char32_t HEAVY_EIGHT_POINTED_RECTILINEAR_BLACK_STAR = 0x2738;
const char32_t TWELVE_POINTED_BLACK_STAR = 0x2739;
const char32_t SIXTEEN_POINTED_ASTERISK = 0x273A;
const char32_t TEARDROP_SPOKED_ASTERISK = 0x273B;
const char32_t OPEN_CENTRE_TEARDROP_SPOKED_ASTERISK = 0x273C;
const char32_t HEAVY_TEARDROP_SPOKED_ASTERISK = 0x273D;
const char32_t SIX_PETALLED_BLACK_AND_WHITE_FLORETTE = 0x273E;
const char32_t BLACK_FLORETTE = 0x273F;
const char32_t WHITE_FLORETTE = 0x2740;
const char32_t EIGHT_PETALLED_OUTLINED_BLACK_FLORETTE = 0x2741;
const char32_t CIRCLED_OPEN_CENTRE_EIGHT_POINTED_STAR = 0x2742;
const char32_t HEAVY_TEARDROP_SPOKED_PINWHEEL_ASTERISK = 0x2743;
const char32_t SNOWFLAKE = 0x2744;
const char32_t TIGHT_TRIFOLIATE_SNOWFLAKE = 0x2745;
const char32_t HEAVY_CHEVRON_SNOWFLAKE = 0x2746;
const char32_t SPARKLE = 0x2747;
const char32_t HEAVY_SPARKLE = 0x2748;
const char32_t BALLOON_SPOKED_ASTERISK = 0x2749;
const char32_t EIGHT_TEARDROP_SPOKED_PROPELLER_ASTERISK = 0x274A;
const char32_t HEAVY_EIGHT_TEARDROP_SPOKED_PROPELLER_ASTERISK = 0x274B;
const char32_t CROSS_MARK = 0x274C;
const char32_t SHADOWED_WHITE_CIRCLE = 0x274D;
const char32_t NEGATIVE_SQUARED_CROSS_MARK = 0x274E;
const char32_t LOWER_RIGHT_DROP_SHADOWED_WHITE_SQUARE = 0x274F;
const char32_t UPPER_RIGHT_DROP_SHADOWED_WHITE_SQUARE = 0x2750;
const char32_t LOWER_RIGHT_SHADOWED_WHITE_SQUARE = 0x2751;
const char32_t UPPER_RIGHT_SHADOWED_WHITE_SQUARE = 0x2752;
const char32_t BLACK_QUESTION_MARK_ORNAMENT = 0x2753;
const char32_t WHITE_QUESTION_MARK_ORNAMENT = 0x2754;
const char32_t WHITE_EXCLAMATION_MARK_ORNAMENT = 0x2755;
const char32_t BLACK_DIAMOND_MINUS_WHITE_X = 0x2756;
const char32_t HEAVY_EXCLAMATION_MARK_SYMBOL = 0x2757;
const char32_t LIGHT_VERTICAL_BAR = 0x2758;
const char32_t MEDIUM_VERTICAL_BAR = 0x2759;
const char32_t HEAVY_VERTICAL_BAR = 0x275A;
const char32_t HEAVY_SINGLE_TURNED_COMMA_QUOTATION_MARK_ORNAMENT = 0x275B;
const char32_t HEAVY_SINGLE_COMMA_QUOTATION_MARK_ORNAMENT = 0x275C;
const char32_t HEAVY_DOUBLE_TURNED_COMMA_QUOTATION_MARK_ORNAMENT = 0x275D;
const char32_t HEAVY_DOUBLE_COMMA_QUOTATION_MARK_ORNAMENT = 0x275E;
const char32_t HEAVY_LOW_SINGLE_COMMA_QUOTATION_MARK_ORNAMENT = 0x275F;
const char32_t HEAVY_LOW_DOUBLE_COMMA_QUOTATION_MARK_ORNAMENT = 0x2760;
const char32_t CURVED_STEM_PARAGRAPH_SIGN_ORNAMENT = 0x2761;
const char32_t HEAVY_EXCLAMATION_MARK_ORNAMENT = 0x2762;
const char32_t HEAVY_HEART_EXCLAMATION_MARK_ORNAMENT = 0x2763;
const char32_t HEAVY_BLACK_HEART = 0x2764;
const char32_t ROTATED_HEAVY_BLACK_HEART_BULLET = 0x2765;
const char32_t FLORAL_HEART = 0x2766;
const char32_t ROTATED_FLORAL_HEART_BULLET = 0x2767;
const char32_t MEDIUM_LEFT_PARENTHESIS_ORNAMENT = 0x2768;
const char32_t MEDIUM_RIGHT_PARENTHESIS_ORNAMENT = 0x2769;
const char32_t MEDIUM_FLATTENED_LEFT_PARENTHESIS_ORNAMENT = 0x276A;
const char32_t MEDIUM_FLATTENED_RIGHT_PARENTHESIS_ORNAMENT = 0x276B;
const char32_t MEDIUM_LEFT_POINTING_ANGLE_BRACKET_ORNAMENT = 0x276C;
const char32_t MEDIUM_RIGHT_POINTING_ANGLE_BRACKET_ORNAMENT = 0x276D;
const char32_t HEAVY_LEFT_POINTING_ANGLE_QUOTATION_MARK_ORNAMENT = 0x276E;
const char32_t HEAVY_RIGHT_POINTING_ANGLE_QUOTATION_MARK_ORNAMENT = 0x276F;
const char32_t HEAVY_LEFT_POINTING_ANGLE_BRACKET_ORNAMENT = 0x2770;
const char32_t HEAVY_RIGHT_POINTING_ANGLE_BRACKET_ORNAMENT = 0x2771;
const char32_t LIGHT_LEFT_TORTOISE_SHELL_BRACKET_ORNAMENT = 0x2772;
const char32_t LIGHT_RIGHT_TORTOISE_SHELL_BRACKET_ORNAMENT = 0x2773;
const char32_t MEDIUM_LEFT_CURLY_BRACKET_ORNAMENT = 0x2774;
const char32_t MEDIUM_RIGHT_CURLY_BRACKET_ORNAMENT = 0x2775;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_ONE = 0x2776;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_TWO = 0x2777;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_THREE = 0x2778;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_FOUR = 0x2779;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_FIVE = 0x277A;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_SIX = 0x277B;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_SEVEN = 0x277C;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_EIGHT = 0x277D;
const char32_t DINGBAT_NEGATIVE_CIRCLED_DIGIT_NINE = 0x277E;
const char32_t DINGBAT_NEGATIVE_CIRCLED_NUMBER_TEN = 0x277F;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_ONE = 0x2780;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_TWO = 0x2781;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_THREE = 0x2782;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_FOUR = 0x2783;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_FIVE = 0x2784;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_SIX = 0x2785;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_SEVEN = 0x2786;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_EIGHT = 0x2787;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_NINE = 0x2788;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_NUMBER_TEN = 0x2789;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_ONE = 0x278A;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_TWO = 0x278B;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_THREE = 0x278C;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_FOUR = 0x278D;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_FIVE = 0x278E;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_SIX = 0x278F;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_SEVEN = 0x2790;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_EIGHT = 0x2791;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_NINE = 0x2792;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_NUMBER_TEN = 0x2793;
const char32_t HEAVY_WIDE_HEADED_RIGHTWARDS_ARROW = 0x2794;
const char32_t HEAVY_PLUS_SIGN = 0x2795;
const char32_t HEAVY_MINUS_SIGN = 0x2796;
const char32_t HEAVY_DIVISION_SIGN = 0x2797;
const char32_t HEAVY_SOUTH_EAST_ARROW = 0x2798;
const char32_t HEAVY_RIGHTWARDS_ARROW = 0x2799;
const char32_t HEAVY_NORTH_EAST_ARROW = 0x279A;
const char32_t DRAFTING_POINT_RIGHTWARDS_ARROW = 0x279B;
const char32_t HEAVY_ROUND_TIPPED_RIGHTWARDS_ARROW = 0x279C;
const char32_t TRIANGLE_HEADED_RIGHTWARDS_ARROW = 0x279D;
const char32_t HEAVY_TRIANGLE_HEADED_RIGHTWARDS_ARROW = 0x279E;
const char32_t DASHED_TRIANGLE_HEADED_RIGHTWARDS_ARROW = 0x279F;
const char32_t HEAVY_DASHED_TRIANGLE_HEADED_RIGHTWARDS_ARROW = 0x27A0;
const char32_t BLACK_RIGHTWARDS_ARROW = 0x27A1;
const char32_t THREE_D_TOP_LIGHTED_RIGHTWARDS_ARROWHEAD = 0x27A2;
const char32_t THREE_D_BOTTOM_LIGHTED_RIGHTWARDS_ARROWHEAD = 0x27A3;
const char32_t BLACK_RIGHTWARDS_ARROWHEAD = 0x27A4;
const char32_t HEAVY_BLACK_CURVED_DOWNWARDS_AND_RIGHTWARDS_ARROW = 0x27A5;
const char32_t HEAVY_BLACK_CURVED_UPWARDS_AND_RIGHTWARDS_ARROW = 0x27A6;
const char32_t SQUAT_BLACK_RIGHTWARDS_ARROW = 0x27A7;
const char32_t HEAVY_CONCAVE_POINTED_BLACK_RIGHTWARDS_ARROW = 0x27A8;
const char32_t RIGHT_SHADED_WHITE_RIGHTWARDS_ARROW = 0x27A9;
const char32_t LEFT_SHADED_WHITE_RIGHTWARDS_ARROW = 0x27AA;
const char32_t BACK_TILTED_SHADOWED_WHITE_RIGHTWARDS_ARROW = 0x27AB;
const char32_t FRONT_TILTED_SHADOWED_WHITE_RIGHTWARDS_ARROW = 0x27AC;
const char32_t HEAVY_LOWER_RIGHT_SHADOWED_WHITE_RIGHTWARDS_ARROW = 0x27AD;
const char32_t HEAVY_UPPER_RIGHT_SHADOWED_WHITE_RIGHTWARDS_ARROW = 0x27AE;
const char32_t NOTCHED_LOWER_RIGHT_SHADOWED_WHITE_RIGHTWARDS_ARROW = 0x27AF;
const char32_t CURLY_LOOP = 0x27B0;
const char32_t NOTCHED_UPPER_RIGHT_SHADOWED_WHITE_RIGHTWARDS_ARROW = 0x27B1;
const char32_t CIRCLED_HEAVY_WHITE_RIGHTWARDS_ARROW = 0x27B2;
const char32_t WHITE_FEATHERED_RIGHTWARDS_ARROW = 0x27B3;
const char32_t BLACK_FEATHERED_SOUTH_EAST_ARROW = 0x27B4;
const char32_t BLACK_FEATHERED_RIGHTWARDS_ARROW = 0x27B5;
const char32_t BLACK_FEATHERED_NORTH_EAST_ARROW = 0x27B6;
const char32_t HEAVY_BLACK_FEATHERED_SOUTH_EAST_ARROW = 0x27B7;
const char32_t HEAVY_BLACK_FEATHERED_RIGHTWARDS_ARROW = 0x27B8;
const char32_t HEAVY_BLACK_FEATHERED_NORTH_EAST_ARROW = 0x27B9;
const char32_t TEARDROP_BARBED_RIGHTWARDS_ARROW = 0x27BA;
const char32_t HEAVY_TEARDROP_SHANKED_RIGHTWARDS_ARROW = 0x27BB;
const char32_t WEDGE_TAILED_RIGHTWARDS_ARROW = 0x27BC;
const char32_t HEAVY_WEDGE_TAILED_RIGHTWARDS_ARROW = 0x27BD;
const char32_t OPEN_OUTLINED_RIGHTWARDS_ARROW = 0x27BE;
const char32_t DOUBLE_CURLY_LOOP = 0x27BF;
const char32_t THREE_DIMENSIONAL_ANGLE = 0x27C0;
const char32_t WHITE_TRIANGLE_CONTAINING_SMALL_WHITE_TRIANGLE = 0x27C1;
const char32_t PERPENDICULAR = 0x27C2;
const char32_t OPEN_SUBSET = 0x27C3;
const char32_t OPEN_SUPERSET = 0x27C4;
const char32_t LEFT_S_SHAPED_BAG_DELIMITER = 0x27C5;
const char32_t RIGHT_S_SHAPED_BAG_DELIMITER = 0x27C6;
const char32_t OR_WITH_DOT_INSIDE = 0x27C7;
const char32_t REVERSE_SOLIDUS_PRECEDING_SUBSET = 0x27C8;
const char32_t SUPERSET_PRECEDING_SOLIDUS = 0x27C9;
const char32_t VERTICAL_BAR_WITH_HORIZONTAL_STROKE = 0x27CA;
const char32_t MATHEMATICAL_RISING_DIAGONAL = 0x27CB;
const char32_t LONG_DIVISION = 0x27CC;
const char32_t MATHEMATICAL_FALLING_DIAGONAL = 0x27CD;
const char32_t SQUARED_LOGICAL_AND = 0x27CE;
const char32_t SQUARED_LOGICAL_OR = 0x27CF;
const char32_t WHITE_DIAMOND_WITH_CENTRED_DOT = 0x27D0;
const char32_t AND_WITH_DOT = 0x27D1;
const char32_t ELEMENT_OF_OPENING_UPWARDS = 0x27D2;
const char32_t LOWER_RIGHT_CORNER_WITH_DOT = 0x27D3;
const char32_t UPPER_LEFT_CORNER_WITH_DOT = 0x27D4;
const char32_t LEFT_OUTER_JOIN = 0x27D5;
const char32_t RIGHT_OUTER_JOIN = 0x27D6;
const char32_t FULL_OUTER_JOIN = 0x27D7;
const char32_t LARGE_UP_TACK = 0x27D8;
const char32_t LARGE_DOWN_TACK = 0x27D9;
const char32_t LEFT_AND_RIGHT_DOUBLE_TURNSTILE = 0x27DA;
const char32_t LEFT_AND_RIGHT_TACK = 0x27DB;
const char32_t LEFT_MULTIMAP = 0x27DC;
const char32_t LONG_RIGHT_TACK = 0x27DD;
const char32_t LONG_LEFT_TACK = 0x27DE;
const char32_t UP_TACK_WITH_CIRCLE_ABOVE = 0x27DF;
const char32_t LOZENGE_DIVIDED_BY_HORIZONTAL_RULE = 0x27E0;
const char32_t WHITE_CONCAVE_SIDED_DIAMOND = 0x27E1;
const char32_t WHITE_CONCAVE_SIDED_DIAMOND_WITH_LEFTWARDS_TICK = 0x27E2;
const char32_t WHITE_CONCAVE_SIDED_DIAMOND_WITH_RIGHTWARDS_TICK = 0x27E3;
const char32_t WHITE_SQUARE_WITH_LEFTWARDS_TICK = 0x27E4;
const char32_t WHITE_SQUARE_WITH_RIGHTWARDS_TICK = 0x27E5;
const char32_t MATHEMATICAL_LEFT_WHITE_SQUARE_BRACKET = 0x27E6;
const char32_t MATHEMATICAL_RIGHT_WHITE_SQUARE_BRACKET = 0x27E7;
const char32_t MATHEMATICAL_LEFT_ANGLE_BRACKET = 0x27E8;
const char32_t MATHEMATICAL_RIGHT_ANGLE_BRACKET = 0x27E9;
const char32_t MATHEMATICAL_LEFT_DOUBLE_ANGLE_BRACKET = 0x27EA;
const char32_t MATHEMATICAL_RIGHT_DOUBLE_ANGLE_BRACKET = 0x27EB;
const char32_t MATHEMATICAL_LEFT_WHITE_TORTOISE_SHELL_BRACKET = 0x27EC;
const char32_t MATHEMATICAL_RIGHT_WHITE_TORTOISE_SHELL_BRACKET = 0x27ED;
const char32_t MATHEMATICAL_LEFT_FLATTENED_PARENTHESIS = 0x27EE;
const char32_t MATHEMATICAL_RIGHT_FLATTENED_PARENTHESIS = 0x27EF;
const char32_t UPWARDS_QUADRUPLE_ARROW = 0x27F0;
const char32_t DOWNWARDS_QUADRUPLE_ARROW = 0x27F1;
const char32_t ANTICLOCKWISE_GAPPED_CIRCLE_ARROW = 0x27F2;
const char32_t CLOCKWISE_GAPPED_CIRCLE_ARROW = 0x27F3;
const char32_t RIGHT_ARROW_WITH_CIRCLED_PLUS = 0x27F4;
const char32_t LONG_LEFTWARDS_ARROW = 0x27F5;
const char32_t LONG_RIGHTWARDS_ARROW = 0x27F6;
const char32_t LONG_LEFT_RIGHT_ARROW = 0x27F7;
const char32_t LONG_LEFTWARDS_DOUBLE_ARROW = 0x27F8;
const char32_t LONG_RIGHTWARDS_DOUBLE_ARROW = 0x27F9;
const char32_t LONG_LEFT_RIGHT_DOUBLE_ARROW = 0x27FA;
const char32_t LONG_LEFTWARDS_ARROW_FROM_BAR = 0x27FB;
const char32_t LONG_RIGHTWARDS_ARROW_FROM_BAR = 0x27FC;
const char32_t LONG_LEFTWARDS_DOUBLE_ARROW_FROM_BAR = 0x27FD;
const char32_t LONG_RIGHTWARDS_DOUBLE_ARROW_FROM_BAR = 0x27FE;
const char32_t LONG_RIGHTWARDS_SQUIGGLE_ARROW = 0x27FF;
const char32_t BRAILLE_PATTERN_BLANK = 0x2800;
const char32_t BRAILLE_PATTERN_DOTS_1 = 0x2801;
const char32_t BRAILLE_PATTERN_DOTS_2 = 0x2802;
const char32_t BRAILLE_PATTERN_DOTS_12 = 0x2803;
const char32_t BRAILLE_PATTERN_DOTS_3 = 0x2804;
const char32_t BRAILLE_PATTERN_DOTS_13 = 0x2805;
const char32_t BRAILLE_PATTERN_DOTS_23 = 0x2806;
const char32_t BRAILLE_PATTERN_DOTS_123 = 0x2807;
const char32_t BRAILLE_PATTERN_DOTS_4 = 0x2808;
const char32_t BRAILLE_PATTERN_DOTS_14 = 0x2809;
const char32_t BRAILLE_PATTERN_DOTS_24 = 0x280A;
const char32_t BRAILLE_PATTERN_DOTS_124 = 0x280B;
const char32_t BRAILLE_PATTERN_DOTS_34 = 0x280C;
const char32_t BRAILLE_PATTERN_DOTS_134 = 0x280D;
const char32_t BRAILLE_PATTERN_DOTS_234 = 0x280E;
const char32_t BRAILLE_PATTERN_DOTS_1234 = 0x280F;
const char32_t BRAILLE_PATTERN_DOTS_5 = 0x2810;
const char32_t BRAILLE_PATTERN_DOTS_15 = 0x2811;
const char32_t BRAILLE_PATTERN_DOTS_25 = 0x2812;
const char32_t BRAILLE_PATTERN_DOTS_125 = 0x2813;
const char32_t BRAILLE_PATTERN_DOTS_35 = 0x2814;
const char32_t BRAILLE_PATTERN_DOTS_135 = 0x2815;
const char32_t BRAILLE_PATTERN_DOTS_235 = 0x2816;
const char32_t BRAILLE_PATTERN_DOTS_1235 = 0x2817;
const char32_t BRAILLE_PATTERN_DOTS_45 = 0x2818;
const char32_t BRAILLE_PATTERN_DOTS_145 = 0x2819;
const char32_t BRAILLE_PATTERN_DOTS_245 = 0x281A;
const char32_t BRAILLE_PATTERN_DOTS_1245 = 0x281B;
const char32_t BRAILLE_PATTERN_DOTS_345 = 0x281C;
const char32_t BRAILLE_PATTERN_DOTS_1345 = 0x281D;
const char32_t BRAILLE_PATTERN_DOTS_2345 = 0x281E;
const char32_t BRAILLE_PATTERN_DOTS_12345 = 0x281F;
const char32_t BRAILLE_PATTERN_DOTS_6 = 0x2820;
const char32_t BRAILLE_PATTERN_DOTS_16 = 0x2821;
const char32_t BRAILLE_PATTERN_DOTS_26 = 0x2822;
const char32_t BRAILLE_PATTERN_DOTS_126 = 0x2823;
const char32_t BRAILLE_PATTERN_DOTS_36 = 0x2824;
const char32_t BRAILLE_PATTERN_DOTS_136 = 0x2825;
const char32_t BRAILLE_PATTERN_DOTS_236 = 0x2826;
const char32_t BRAILLE_PATTERN_DOTS_1236 = 0x2827;
const char32_t BRAILLE_PATTERN_DOTS_46 = 0x2828;
const char32_t BRAILLE_PATTERN_DOTS_146 = 0x2829;
const char32_t BRAILLE_PATTERN_DOTS_246 = 0x282A;
const char32_t BRAILLE_PATTERN_DOTS_1246 = 0x282B;
const char32_t BRAILLE_PATTERN_DOTS_346 = 0x282C;
const char32_t BRAILLE_PATTERN_DOTS_1346 = 0x282D;
const char32_t BRAILLE_PATTERN_DOTS_2346 = 0x282E;
const char32_t BRAILLE_PATTERN_DOTS_12346 = 0x282F;
const char32_t BRAILLE_PATTERN_DOTS_56 = 0x2830;
const char32_t BRAILLE_PATTERN_DOTS_156 = 0x2831;
const char32_t BRAILLE_PATTERN_DOTS_256 = 0x2832;
const char32_t BRAILLE_PATTERN_DOTS_1256 = 0x2833;
const char32_t BRAILLE_PATTERN_DOTS_356 = 0x2834;
const char32_t BRAILLE_PATTERN_DOTS_1356 = 0x2835;
const char32_t BRAILLE_PATTERN_DOTS_2356 = 0x2836;
const char32_t BRAILLE_PATTERN_DOTS_12356 = 0x2837;
const char32_t BRAILLE_PATTERN_DOTS_456 = 0x2838;
const char32_t BRAILLE_PATTERN_DOTS_1456 = 0x2839;
const char32_t BRAILLE_PATTERN_DOTS_2456 = 0x283A;
const char32_t BRAILLE_PATTERN_DOTS_12456 = 0x283B;
const char32_t BRAILLE_PATTERN_DOTS_3456 = 0x283C;
const char32_t BRAILLE_PATTERN_DOTS_13456 = 0x283D;
const char32_t BRAILLE_PATTERN_DOTS_23456 = 0x283E;
const char32_t BRAILLE_PATTERN_DOTS_123456 = 0x283F;
const char32_t BRAILLE_PATTERN_DOTS_7 = 0x2840;
const char32_t BRAILLE_PATTERN_DOTS_17 = 0x2841;
const char32_t BRAILLE_PATTERN_DOTS_27 = 0x2842;
const char32_t BRAILLE_PATTERN_DOTS_127 = 0x2843;
const char32_t BRAILLE_PATTERN_DOTS_37 = 0x2844;
const char32_t BRAILLE_PATTERN_DOTS_137 = 0x2845;
const char32_t BRAILLE_PATTERN_DOTS_237 = 0x2846;
const char32_t BRAILLE_PATTERN_DOTS_1237 = 0x2847;
const char32_t BRAILLE_PATTERN_DOTS_47 = 0x2848;
const char32_t BRAILLE_PATTERN_DOTS_147 = 0x2849;
const char32_t BRAILLE_PATTERN_DOTS_247 = 0x284A;
const char32_t BRAILLE_PATTERN_DOTS_1247 = 0x284B;
const char32_t BRAILLE_PATTERN_DOTS_347 = 0x284C;
const char32_t BRAILLE_PATTERN_DOTS_1347 = 0x284D;
const char32_t BRAILLE_PATTERN_DOTS_2347 = 0x284E;
const char32_t BRAILLE_PATTERN_DOTS_12347 = 0x284F;
const char32_t BRAILLE_PATTERN_DOTS_57 = 0x2850;
const char32_t BRAILLE_PATTERN_DOTS_157 = 0x2851;
const char32_t BRAILLE_PATTERN_DOTS_257 = 0x2852;
const char32_t BRAILLE_PATTERN_DOTS_1257 = 0x2853;
const char32_t BRAILLE_PATTERN_DOTS_357 = 0x2854;
const char32_t BRAILLE_PATTERN_DOTS_1357 = 0x2855;
const char32_t BRAILLE_PATTERN_DOTS_2357 = 0x2856;
const char32_t BRAILLE_PATTERN_DOTS_12357 = 0x2857;
const char32_t BRAILLE_PATTERN_DOTS_457 = 0x2858;
const char32_t BRAILLE_PATTERN_DOTS_1457 = 0x2859;
const char32_t BRAILLE_PATTERN_DOTS_2457 = 0x285A;
const char32_t BRAILLE_PATTERN_DOTS_12457 = 0x285B;
const char32_t BRAILLE_PATTERN_DOTS_3457 = 0x285C;
const char32_t BRAILLE_PATTERN_DOTS_13457 = 0x285D;
const char32_t BRAILLE_PATTERN_DOTS_23457 = 0x285E;
const char32_t BRAILLE_PATTERN_DOTS_123457 = 0x285F;
const char32_t BRAILLE_PATTERN_DOTS_67 = 0x2860;
const char32_t BRAILLE_PATTERN_DOTS_167 = 0x2861;
const char32_t BRAILLE_PATTERN_DOTS_267 = 0x2862;
const char32_t BRAILLE_PATTERN_DOTS_1267 = 0x2863;
const char32_t BRAILLE_PATTERN_DOTS_367 = 0x2864;
const char32_t BRAILLE_PATTERN_DOTS_1367 = 0x2865;
const char32_t BRAILLE_PATTERN_DOTS_2367 = 0x2866;
const char32_t BRAILLE_PATTERN_DOTS_12367 = 0x2867;
const char32_t BRAILLE_PATTERN_DOTS_467 = 0x2868;
const char32_t BRAILLE_PATTERN_DOTS_1467 = 0x2869;
const char32_t BRAILLE_PATTERN_DOTS_2467 = 0x286A;
const char32_t BRAILLE_PATTERN_DOTS_12467 = 0x286B;
const char32_t BRAILLE_PATTERN_DOTS_3467 = 0x286C;
const char32_t BRAILLE_PATTERN_DOTS_13467 = 0x286D;
const char32_t BRAILLE_PATTERN_DOTS_23467 = 0x286E;
const char32_t BRAILLE_PATTERN_DOTS_123467 = 0x286F;
const char32_t BRAILLE_PATTERN_DOTS_567 = 0x2870;
const char32_t BRAILLE_PATTERN_DOTS_1567 = 0x2871;
const char32_t BRAILLE_PATTERN_DOTS_2567 = 0x2872;
const char32_t BRAILLE_PATTERN_DOTS_12567 = 0x2873;
const char32_t BRAILLE_PATTERN_DOTS_3567 = 0x2874;
const char32_t BRAILLE_PATTERN_DOTS_13567 = 0x2875;
const char32_t BRAILLE_PATTERN_DOTS_23567 = 0x2876;
const char32_t BRAILLE_PATTERN_DOTS_123567 = 0x2877;
const char32_t BRAILLE_PATTERN_DOTS_4567 = 0x2878;
const char32_t BRAILLE_PATTERN_DOTS_14567 = 0x2879;
const char32_t BRAILLE_PATTERN_DOTS_24567 = 0x287A;
const char32_t BRAILLE_PATTERN_DOTS_124567 = 0x287B;
const char32_t BRAILLE_PATTERN_DOTS_34567 = 0x287C;
const char32_t BRAILLE_PATTERN_DOTS_134567 = 0x287D;
const char32_t BRAILLE_PATTERN_DOTS_234567 = 0x287E;
const char32_t BRAILLE_PATTERN_DOTS_1234567 = 0x287F;
const char32_t BRAILLE_PATTERN_DOTS_8 = 0x2880;
const char32_t BRAILLE_PATTERN_DOTS_18 = 0x2881;
const char32_t BRAILLE_PATTERN_DOTS_28 = 0x2882;
const char32_t BRAILLE_PATTERN_DOTS_128 = 0x2883;
const char32_t BRAILLE_PATTERN_DOTS_38 = 0x2884;
const char32_t BRAILLE_PATTERN_DOTS_138 = 0x2885;
const char32_t BRAILLE_PATTERN_DOTS_238 = 0x2886;
const char32_t BRAILLE_PATTERN_DOTS_1238 = 0x2887;
const char32_t BRAILLE_PATTERN_DOTS_48 = 0x2888;
const char32_t BRAILLE_PATTERN_DOTS_148 = 0x2889;
const char32_t BRAILLE_PATTERN_DOTS_248 = 0x288A;
const char32_t BRAILLE_PATTERN_DOTS_1248 = 0x288B;
const char32_t BRAILLE_PATTERN_DOTS_348 = 0x288C;
const char32_t BRAILLE_PATTERN_DOTS_1348 = 0x288D;
const char32_t BRAILLE_PATTERN_DOTS_2348 = 0x288E;
const char32_t BRAILLE_PATTERN_DOTS_12348 = 0x288F;
const char32_t BRAILLE_PATTERN_DOTS_58 = 0x2890;
const char32_t BRAILLE_PATTERN_DOTS_158 = 0x2891;
const char32_t BRAILLE_PATTERN_DOTS_258 = 0x2892;
const char32_t BRAILLE_PATTERN_DOTS_1258 = 0x2893;
const char32_t BRAILLE_PATTERN_DOTS_358 = 0x2894;
const char32_t BRAILLE_PATTERN_DOTS_1358 = 0x2895;
const char32_t BRAILLE_PATTERN_DOTS_2358 = 0x2896;
const char32_t BRAILLE_PATTERN_DOTS_12358 = 0x2897;
const char32_t BRAILLE_PATTERN_DOTS_458 = 0x2898;
const char32_t BRAILLE_PATTERN_DOTS_1458 = 0x2899;
const char32_t BRAILLE_PATTERN_DOTS_2458 = 0x289A;
const char32_t BRAILLE_PATTERN_DOTS_12458 = 0x289B;
const char32_t BRAILLE_PATTERN_DOTS_3458 = 0x289C;
const char32_t BRAILLE_PATTERN_DOTS_13458 = 0x289D;
const char32_t BRAILLE_PATTERN_DOTS_23458 = 0x289E;
const char32_t BRAILLE_PATTERN_DOTS_123458 = 0x289F;
const char32_t BRAILLE_PATTERN_DOTS_68 = 0x28A0;
const char32_t BRAILLE_PATTERN_DOTS_168 = 0x28A1;
const char32_t BRAILLE_PATTERN_DOTS_268 = 0x28A2;
const char32_t BRAILLE_PATTERN_DOTS_1268 = 0x28A3;
const char32_t BRAILLE_PATTERN_DOTS_368 = 0x28A4;
const char32_t BRAILLE_PATTERN_DOTS_1368 = 0x28A5;
const char32_t BRAILLE_PATTERN_DOTS_2368 = 0x28A6;
const char32_t BRAILLE_PATTERN_DOTS_12368 = 0x28A7;
const char32_t BRAILLE_PATTERN_DOTS_468 = 0x28A8;
const char32_t BRAILLE_PATTERN_DOTS_1468 = 0x28A9;
const char32_t BRAILLE_PATTERN_DOTS_2468 = 0x28AA;
const char32_t BRAILLE_PATTERN_DOTS_12468 = 0x28AB;
const char32_t BRAILLE_PATTERN_DOTS_3468 = 0x28AC;
const char32_t BRAILLE_PATTERN_DOTS_13468 = 0x28AD;
const char32_t BRAILLE_PATTERN_DOTS_23468 = 0x28AE;
const char32_t BRAILLE_PATTERN_DOTS_123468 = 0x28AF;
const char32_t BRAILLE_PATTERN_DOTS_568 = 0x28B0;
const char32_t BRAILLE_PATTERN_DOTS_1568 = 0x28B1;
const char32_t BRAILLE_PATTERN_DOTS_2568 = 0x28B2;
const char32_t BRAILLE_PATTERN_DOTS_12568 = 0x28B3;
const char32_t BRAILLE_PATTERN_DOTS_3568 = 0x28B4;
const char32_t BRAILLE_PATTERN_DOTS_13568 = 0x28B5;
const char32_t BRAILLE_PATTERN_DOTS_23568 = 0x28B6;
const char32_t BRAILLE_PATTERN_DOTS_123568 = 0x28B7;
const char32_t BRAILLE_PATTERN_DOTS_4568 = 0x28B8;
const char32_t BRAILLE_PATTERN_DOTS_14568 = 0x28B9;
const char32_t BRAILLE_PATTERN_DOTS_24568 = 0x28BA;
const char32_t BRAILLE_PATTERN_DOTS_124568 = 0x28BB;
const char32_t BRAILLE_PATTERN_DOTS_34568 = 0x28BC;
const char32_t BRAILLE_PATTERN_DOTS_134568 = 0x28BD;
const char32_t BRAILLE_PATTERN_DOTS_234568 = 0x28BE;
const char32_t BRAILLE_PATTERN_DOTS_1234568 = 0x28BF;
const char32_t BRAILLE_PATTERN_DOTS_78 = 0x28C0;
const char32_t BRAILLE_PATTERN_DOTS_178 = 0x28C1;
const char32_t BRAILLE_PATTERN_DOTS_278 = 0x28C2;
const char32_t BRAILLE_PATTERN_DOTS_1278 = 0x28C3;
const char32_t BRAILLE_PATTERN_DOTS_378 = 0x28C4;
const char32_t BRAILLE_PATTERN_DOTS_1378 = 0x28C5;
const char32_t BRAILLE_PATTERN_DOTS_2378 = 0x28C6;
const char32_t BRAILLE_PATTERN_DOTS_12378 = 0x28C7;
const char32_t BRAILLE_PATTERN_DOTS_478 = 0x28C8;
const char32_t BRAILLE_PATTERN_DOTS_1478 = 0x28C9;
const char32_t BRAILLE_PATTERN_DOTS_2478 = 0x28CA;
const char32_t BRAILLE_PATTERN_DOTS_12478 = 0x28CB;
const char32_t BRAILLE_PATTERN_DOTS_3478 = 0x28CC;
const char32_t BRAILLE_PATTERN_DOTS_13478 = 0x28CD;
const char32_t BRAILLE_PATTERN_DOTS_23478 = 0x28CE;
const char32_t BRAILLE_PATTERN_DOTS_123478 = 0x28CF;
const char32_t BRAILLE_PATTERN_DOTS_578 = 0x28D0;
const char32_t BRAILLE_PATTERN_DOTS_1578 = 0x28D1;
const char32_t BRAILLE_PATTERN_DOTS_2578 = 0x28D2;
const char32_t BRAILLE_PATTERN_DOTS_12578 = 0x28D3;
const char32_t BRAILLE_PATTERN_DOTS_3578 = 0x28D4;
const char32_t BRAILLE_PATTERN_DOTS_13578 = 0x28D5;
const char32_t BRAILLE_PATTERN_DOTS_23578 = 0x28D6;
const char32_t BRAILLE_PATTERN_DOTS_123578 = 0x28D7;
const char32_t BRAILLE_PATTERN_DOTS_4578 = 0x28D8;
const char32_t BRAILLE_PATTERN_DOTS_14578 = 0x28D9;
const char32_t BRAILLE_PATTERN_DOTS_24578 = 0x28DA;
const char32_t BRAILLE_PATTERN_DOTS_124578 = 0x28DB;
const char32_t BRAILLE_PATTERN_DOTS_34578 = 0x28DC;
const char32_t BRAILLE_PATTERN_DOTS_134578 = 0x28DD;
const char32_t BRAILLE_PATTERN_DOTS_234578 = 0x28DE;
const char32_t BRAILLE_PATTERN_DOTS_1234578 = 0x28DF;
const char32_t BRAILLE_PATTERN_DOTS_678 = 0x28E0;
const char32_t BRAILLE_PATTERN_DOTS_1678 = 0x28E1;
const char32_t BRAILLE_PATTERN_DOTS_2678 = 0x28E2;
const char32_t BRAILLE_PATTERN_DOTS_12678 = 0x28E3;
const char32_t BRAILLE_PATTERN_DOTS_3678 = 0x28E4;
const char32_t BRAILLE_PATTERN_DOTS_13678 = 0x28E5;
const char32_t BRAILLE_PATTERN_DOTS_23678 = 0x28E6;
const char32_t BRAILLE_PATTERN_DOTS_123678 = 0x28E7;
const char32_t BRAILLE_PATTERN_DOTS_4678 = 0x28E8;
const char32_t BRAILLE_PATTERN_DOTS_14678 = 0x28E9;
const char32_t BRAILLE_PATTERN_DOTS_24678 = 0x28EA;
const char32_t BRAILLE_PATTERN_DOTS_124678 = 0x28EB;
const char32_t BRAILLE_PATTERN_DOTS_34678 = 0x28EC;
const char32_t BRAILLE_PATTERN_DOTS_134678 = 0x28ED;
const char32_t BRAILLE_PATTERN_DOTS_234678 = 0x28EE;
const char32_t BRAILLE_PATTERN_DOTS_1234678 = 0x28EF;
const char32_t BRAILLE_PATTERN_DOTS_5678 = 0x28F0;
const char32_t BRAILLE_PATTERN_DOTS_15678 = 0x28F1;
const char32_t BRAILLE_PATTERN_DOTS_25678 = 0x28F2;
const char32_t BRAILLE_PATTERN_DOTS_125678 = 0x28F3;
const char32_t BRAILLE_PATTERN_DOTS_35678 = 0x28F4;
const char32_t BRAILLE_PATTERN_DOTS_135678 = 0x28F5;
const char32_t BRAILLE_PATTERN_DOTS_235678 = 0x28F6;
const char32_t BRAILLE_PATTERN_DOTS_1235678 = 0x28F7;
const char32_t BRAILLE_PATTERN_DOTS_45678 = 0x28F8;
const char32_t BRAILLE_PATTERN_DOTS_145678 = 0x28F9;
const char32_t BRAILLE_PATTERN_DOTS_245678 = 0x28FA;
const char32_t BRAILLE_PATTERN_DOTS_1245678 = 0x28FB;
const char32_t BRAILLE_PATTERN_DOTS_345678 = 0x28FC;
const char32_t BRAILLE_PATTERN_DOTS_1345678 = 0x28FD;
const char32_t BRAILLE_PATTERN_DOTS_2345678 = 0x28FE;
const char32_t BRAILLE_PATTERN_DOTS_12345678 = 0x28FF;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW_WITH_VERTICAL_STROKE = 0x2900;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW_WITH_DOUBLE_VERTICAL_STROKE = 0x2901;
const char32_t LEFTWARDS_DOUBLE_ARROW_WITH_VERTICAL_STROKE = 0x2902;
const char32_t RIGHTWARDS_DOUBLE_ARROW_WITH_VERTICAL_STROKE = 0x2903;
const char32_t LEFT_RIGHT_DOUBLE_ARROW_WITH_VERTICAL_STROKE = 0x2904;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW_FROM_BAR = 0x2905;
const char32_t LEFTWARDS_DOUBLE_ARROW_FROM_BAR = 0x2906;
const char32_t RIGHTWARDS_DOUBLE_ARROW_FROM_BAR = 0x2907;
const char32_t DOWNWARDS_ARROW_WITH_HORIZONTAL_STROKE = 0x2908;
const char32_t UPWARDS_ARROW_WITH_HORIZONTAL_STROKE = 0x2909;
const char32_t UPWARDS_TRIPLE_ARROW = 0x290A;
const char32_t DOWNWARDS_TRIPLE_ARROW = 0x290B;
const char32_t LEFTWARDS_DOUBLE_DASH_ARROW = 0x290C;
const char32_t RIGHTWARDS_DOUBLE_DASH_ARROW = 0x290D;
const char32_t LEFTWARDS_TRIPLE_DASH_ARROW = 0x290E;
const char32_t RIGHTWARDS_TRIPLE_DASH_ARROW = 0x290F;
const char32_t RIGHTWARDS_TWO_HEADED_TRIPLE_DASH_ARROW = 0x2910;
const char32_t RIGHTWARDS_ARROW_WITH_DOTTED_STEM = 0x2911;
const char32_t UPWARDS_ARROW_TO_BAR = 0x2912;
const char32_t DOWNWARDS_ARROW_TO_BAR = 0x2913;
const char32_t RIGHTWARDS_ARROW_WITH_TAIL_WITH_VERTICAL_STROKE = 0x2914;
const char32_t RIGHTWARDS_ARROW_WITH_TAIL_WITH_DOUBLE_VERTICAL_STROKE = 0x2915;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW_WITH_TAIL = 0x2916;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW_WITH_TAIL_WITH_VERTICAL_STROKE = 0x2917;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW_WITH_TAIL_WITH_DOUBLE_VERTICAL_STROKE = 0x2918;
const char32_t LEFTWARDS_ARROW_TAIL = 0x2919;
const char32_t RIGHTWARDS_ARROW_TAIL = 0x291A;
const char32_t LEFTWARDS_DOUBLE_ARROW_TAIL = 0x291B;
const char32_t RIGHTWARDS_DOUBLE_ARROW_TAIL = 0x291C;
const char32_t LEFTWARDS_ARROW_TO_BLACK_DIAMOND = 0x291D;
const char32_t RIGHTWARDS_ARROW_TO_BLACK_DIAMOND = 0x291E;
const char32_t LEFTWARDS_ARROW_FROM_BAR_TO_BLACK_DIAMOND = 0x291F;
const char32_t RIGHTWARDS_ARROW_FROM_BAR_TO_BLACK_DIAMOND = 0x2920;
const char32_t NORTH_WEST_AND_SOUTH_EAST_ARROW = 0x2921;
const char32_t NORTH_EAST_AND_SOUTH_WEST_ARROW = 0x2922;
const char32_t NORTH_WEST_ARROW_WITH_HOOK = 0x2923;
const char32_t NORTH_EAST_ARROW_WITH_HOOK = 0x2924;
const char32_t SOUTH_EAST_ARROW_WITH_HOOK = 0x2925;
const char32_t SOUTH_WEST_ARROW_WITH_HOOK = 0x2926;
const char32_t NORTH_WEST_ARROW_AND_NORTH_EAST_ARROW = 0x2927;
const char32_t NORTH_EAST_ARROW_AND_SOUTH_EAST_ARROW = 0x2928;
const char32_t SOUTH_EAST_ARROW_AND_SOUTH_WEST_ARROW = 0x2929;
const char32_t SOUTH_WEST_ARROW_AND_NORTH_WEST_ARROW = 0x292A;
const char32_t RISING_DIAGONAL_CROSSING_FALLING_DIAGONAL = 0x292B;
const char32_t FALLING_DIAGONAL_CROSSING_RISING_DIAGONAL = 0x292C;
const char32_t SOUTH_EAST_ARROW_CROSSING_NORTH_EAST_ARROW = 0x292D;
const char32_t NORTH_EAST_ARROW_CROSSING_SOUTH_EAST_ARROW = 0x292E;
const char32_t FALLING_DIAGONAL_CROSSING_NORTH_EAST_ARROW = 0x292F;
const char32_t RISING_DIAGONAL_CROSSING_SOUTH_EAST_ARROW = 0x2930;
const char32_t NORTH_EAST_ARROW_CROSSING_NORTH_WEST_ARROW = 0x2931;
const char32_t NORTH_WEST_ARROW_CROSSING_NORTH_EAST_ARROW = 0x2932;
const char32_t WAVE_ARROW_POINTING_DIRECTLY_RIGHT = 0x2933;
const char32_t ARROW_POINTING_RIGHTWARDS_THEN_CURVING_UPWARDS = 0x2934;
const char32_t ARROW_POINTING_RIGHTWARDS_THEN_CURVING_DOWNWARDS = 0x2935;
const char32_t ARROW_POINTING_DOWNWARDS_THEN_CURVING_LEFTWARDS = 0x2936;
const char32_t ARROW_POINTING_DOWNWARDS_THEN_CURVING_RIGHTWARDS = 0x2937;
const char32_t RIGHT_SIDE_ARC_CLOCKWISE_ARROW = 0x2938;
const char32_t LEFT_SIDE_ARC_ANTICLOCKWISE_ARROW = 0x2939;
const char32_t TOP_ARC_ANTICLOCKWISE_ARROW = 0x293A;
const char32_t BOTTOM_ARC_ANTICLOCKWISE_ARROW = 0x293B;
const char32_t TOP_ARC_CLOCKWISE_ARROW_WITH_MINUS = 0x293C;
const char32_t TOP_ARC_ANTICLOCKWISE_ARROW_WITH_PLUS = 0x293D;
const char32_t LOWER_RIGHT_SEMICIRCULAR_CLOCKWISE_ARROW = 0x293E;
const char32_t LOWER_LEFT_SEMICIRCULAR_ANTICLOCKWISE_ARROW = 0x293F;
const char32_t ANTICLOCKWISE_CLOSED_CIRCLE_ARROW = 0x2940;
const char32_t CLOCKWISE_CLOSED_CIRCLE_ARROW = 0x2941;
const char32_t RIGHTWARDS_ARROW_ABOVE_SHORT_LEFTWARDS_ARROW = 0x2942;
const char32_t LEFTWARDS_ARROW_ABOVE_SHORT_RIGHTWARDS_ARROW = 0x2943;
const char32_t SHORT_RIGHTWARDS_ARROW_ABOVE_LEFTWARDS_ARROW = 0x2944;
const char32_t RIGHTWARDS_ARROW_WITH_PLUS_BELOW = 0x2945;
const char32_t LEFTWARDS_ARROW_WITH_PLUS_BELOW = 0x2946;
const char32_t RIGHTWARDS_ARROW_THROUGH_X = 0x2947;
const char32_t LEFT_RIGHT_ARROW_THROUGH_SMALL_CIRCLE = 0x2948;
const char32_t UPWARDS_TWO_HEADED_ARROW_FROM_SMALL_CIRCLE = 0x2949;
const char32_t LEFT_BARB_UP_RIGHT_BARB_DOWN_HARPOON = 0x294A;
const char32_t LEFT_BARB_DOWN_RIGHT_BARB_UP_HARPOON = 0x294B;
const char32_t UP_BARB_RIGHT_DOWN_BARB_LEFT_HARPOON = 0x294C;
const char32_t UP_BARB_LEFT_DOWN_BARB_RIGHT_HARPOON = 0x294D;
const char32_t LEFT_BARB_UP_RIGHT_BARB_UP_HARPOON = 0x294E;
const char32_t UP_BARB_RIGHT_DOWN_BARB_RIGHT_HARPOON = 0x294F;
const char32_t LEFT_BARB_DOWN_RIGHT_BARB_DOWN_HARPOON = 0x2950;
const char32_t UP_BARB_LEFT_DOWN_BARB_LEFT_HARPOON = 0x2951;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_UP_TO_BAR = 0x2952;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_UP_TO_BAR = 0x2953;
const char32_t UPWARDS_HARPOON_WITH_BARB_RIGHT_TO_BAR = 0x2954;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_RIGHT_TO_BAR = 0x2955;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_DOWN_TO_BAR = 0x2956;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_DOWN_TO_BAR = 0x2957;
const char32_t UPWARDS_HARPOON_WITH_BARB_LEFT_TO_BAR = 0x2958;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_LEFT_TO_BAR = 0x2959;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_UP_FROM_BAR = 0x295A;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_UP_FROM_BAR = 0x295B;
const char32_t UPWARDS_HARPOON_WITH_BARB_RIGHT_FROM_BAR = 0x295C;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_RIGHT_FROM_BAR = 0x295D;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_DOWN_FROM_BAR = 0x295E;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_DOWN_FROM_BAR = 0x295F;
const char32_t UPWARDS_HARPOON_WITH_BARB_LEFT_FROM_BAR = 0x2960;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_LEFT_FROM_BAR = 0x2961;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_UP_ABOVE_LEFTWARDS_HARPOON_WITH_BARB_DOWN = 0x2962;
const char32_t UPWARDS_HARPOON_WITH_BARB_LEFT_BESIDE_UPWARDS_HARPOON_WITH_BARB_RIGHT = 0x2963;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_UP_ABOVE_RIGHTWARDS_HARPOON_WITH_BARB_DOWN = 0x2964;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_LEFT_BESIDE_DOWNWARDS_HARPOON_WITH_BARB_RIGHT = 0x2965;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_UP_ABOVE_RIGHTWARDS_HARPOON_WITH_BARB_UP = 0x2966;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_DOWN_ABOVE_RIGHTWARDS_HARPOON_WITH_BARB_DOWN = 0x2967;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_UP_ABOVE_LEFTWARDS_HARPOON_WITH_BARB_UP = 0x2968;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_DOWN_ABOVE_LEFTWARDS_HARPOON_WITH_BARB_DOWN = 0x2969;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_UP_ABOVE_LONG_DASH = 0x296A;
const char32_t LEFTWARDS_HARPOON_WITH_BARB_DOWN_BELOW_LONG_DASH = 0x296B;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_UP_ABOVE_LONG_DASH = 0x296C;
const char32_t RIGHTWARDS_HARPOON_WITH_BARB_DOWN_BELOW_LONG_DASH = 0x296D;
const char32_t UPWARDS_HARPOON_WITH_BARB_LEFT_BESIDE_DOWNWARDS_HARPOON_WITH_BARB_RIGHT = 0x296E;
const char32_t DOWNWARDS_HARPOON_WITH_BARB_LEFT_BESIDE_UPWARDS_HARPOON_WITH_BARB_RIGHT = 0x296F;
const char32_t RIGHT_DOUBLE_ARROW_WITH_ROUNDED_HEAD = 0x2970;
const char32_t EQUALS_SIGN_ABOVE_RIGHTWARDS_ARROW = 0x2971;
const char32_t TILDE_OPERATOR_ABOVE_RIGHTWARDS_ARROW = 0x2972;
const char32_t LEFTWARDS_ARROW_ABOVE_TILDE_OPERATOR = 0x2973;
const char32_t RIGHTWARDS_ARROW_ABOVE_TILDE_OPERATOR = 0x2974;
const char32_t RIGHTWARDS_ARROW_ABOVE_ALMOST_EQUAL_TO = 0x2975;
const char32_t LESS_THAN_ABOVE_LEFTWARDS_ARROW = 0x2976;
const char32_t LEFTWARDS_ARROW_THROUGH_LESS_THAN = 0x2977;
const char32_t GREATER_THAN_ABOVE_RIGHTWARDS_ARROW = 0x2978;
const char32_t SUBSET_ABOVE_RIGHTWARDS_ARROW = 0x2979;
const char32_t LEFTWARDS_ARROW_THROUGH_SUBSET = 0x297A;
const char32_t SUPERSET_ABOVE_LEFTWARDS_ARROW = 0x297B;
const char32_t LEFT_FISH_TAIL = 0x297C;
const char32_t RIGHT_FISH_TAIL = 0x297D;
const char32_t UP_FISH_TAIL = 0x297E;
const char32_t DOWN_FISH_TAIL = 0x297F;
const char32_t TRIPLE_VERTICAL_BAR_DELIMITER = 0x2980;
const char32_t Z_NOTATION_SPOT = 0x2981;
const char32_t Z_NOTATION_TYPE_COLON = 0x2982;
const char32_t LEFT_WHITE_CURLY_BRACKET = 0x2983;
const char32_t RIGHT_WHITE_CURLY_BRACKET = 0x2984;
const char32_t LEFT_WHITE_PARENTHESIS = 0x2985;
const char32_t RIGHT_WHITE_PARENTHESIS = 0x2986;
const char32_t Z_NOTATION_LEFT_IMAGE_BRACKET = 0x2987;
const char32_t Z_NOTATION_RIGHT_IMAGE_BRACKET = 0x2988;
const char32_t Z_NOTATION_LEFT_BINDING_BRACKET = 0x2989;
const char32_t Z_NOTATION_RIGHT_BINDING_BRACKET = 0x298A;
const char32_t LEFT_SQUARE_BRACKET_WITH_UNDERBAR = 0x298B;
const char32_t RIGHT_SQUARE_BRACKET_WITH_UNDERBAR = 0x298C;
const char32_t LEFT_SQUARE_BRACKET_WITH_TICK_IN_TOP_CORNER = 0x298D;
const char32_t RIGHT_SQUARE_BRACKET_WITH_TICK_IN_BOTTOM_CORNER = 0x298E;
const char32_t LEFT_SQUARE_BRACKET_WITH_TICK_IN_BOTTOM_CORNER = 0x298F;
const char32_t RIGHT_SQUARE_BRACKET_WITH_TICK_IN_TOP_CORNER = 0x2990;
const char32_t LEFT_ANGLE_BRACKET_WITH_DOT = 0x2991;
const char32_t RIGHT_ANGLE_BRACKET_WITH_DOT = 0x2992;
const char32_t LEFT_ARC_LESS_THAN_BRACKET = 0x2993;
const char32_t RIGHT_ARC_GREATER_THAN_BRACKET = 0x2994;
const char32_t DOUBLE_LEFT_ARC_GREATER_THAN_BRACKET = 0x2995;
const char32_t DOUBLE_RIGHT_ARC_LESS_THAN_BRACKET = 0x2996;
const char32_t LEFT_BLACK_TORTOISE_SHELL_BRACKET = 0x2997;
const char32_t RIGHT_BLACK_TORTOISE_SHELL_BRACKET = 0x2998;
const char32_t DOTTED_FENCE = 0x2999;
const char32_t VERTICAL_ZIGZAG_LINE = 0x299A;
const char32_t MEASURED_ANGLE_OPENING_LEFT = 0x299B;
const char32_t RIGHT_ANGLE_VARIANT_WITH_SQUARE = 0x299C;
const char32_t MEASURED_RIGHT_ANGLE_WITH_DOT = 0x299D;
const char32_t ANGLE_WITH_S_INSIDE = 0x299E;
const char32_t ACUTE_ANGLE = 0x299F;
const char32_t SPHERICAL_ANGLE_OPENING_LEFT = 0x29A0;
const char32_t SPHERICAL_ANGLE_OPENING_UP = 0x29A1;
const char32_t TURNED_ANGLE = 0x29A2;
const char32_t REVERSED_ANGLE = 0x29A3;
const char32_t ANGLE_WITH_UNDERBAR = 0x29A4;
const char32_t REVERSED_ANGLE_WITH_UNDERBAR = 0x29A5;
const char32_t OBLIQUE_ANGLE_OPENING_UP = 0x29A6;
const char32_t OBLIQUE_ANGLE_OPENING_DOWN = 0x29A7;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_UP_AND_RIGHT = 0x29A8;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_UP_AND_LEFT = 0x29A9;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_DOWN_AND_RIGHT = 0x29AA;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_DOWN_AND_LEFT = 0x29AB;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_RIGHT_AND_UP = 0x29AC;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_LEFT_AND_UP = 0x29AD;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_RIGHT_AND_DOWN = 0x29AE;
const char32_t MEASURED_ANGLE_WITH_OPEN_ARM_ENDING_IN_ARROW_POINTING_LEFT_AND_DOWN = 0x29AF;
const char32_t REVERSED_EMPTY_SET = 0x29B0;
const char32_t EMPTY_SET_WITH_OVERBAR = 0x29B1;
const char32_t EMPTY_SET_WITH_SMALL_CIRCLE_ABOVE = 0x29B2;
const char32_t EMPTY_SET_WITH_RIGHT_ARROW_ABOVE = 0x29B3;
const char32_t EMPTY_SET_WITH_LEFT_ARROW_ABOVE = 0x29B4;
const char32_t CIRCLE_WITH_HORIZONTAL_BAR = 0x29B5;
const char32_t CIRCLED_VERTICAL_BAR = 0x29B6;
const char32_t CIRCLED_PARALLEL = 0x29B7;
const char32_t CIRCLED_REVERSE_SOLIDUS = 0x29B8;
const char32_t CIRCLED_PERPENDICULAR = 0x29B9;
const char32_t CIRCLE_DIVIDED_BY_HORIZONTAL_BAR_AND_TOP_HALF_DIVIDED_BY_VERTICAL_BAR = 0x29BA;
const char32_t CIRCLE_WITH_SUPERIMPOSED_X = 0x29BB;
const char32_t CIRCLED_ANTICLOCKWISE_ROTATED_DIVISION_SIGN = 0x29BC;
const char32_t UP_ARROW_THROUGH_CIRCLE = 0x29BD;
const char32_t CIRCLED_WHITE_BULLET = 0x29BE;
const char32_t CIRCLED_BULLET = 0x29BF;
const char32_t CIRCLED_LESS_THAN = 0x29C0;
const char32_t CIRCLED_GREATER_THAN = 0x29C1;
const char32_t CIRCLE_WITH_SMALL_CIRCLE_TO_THE_RIGHT = 0x29C2;
const char32_t CIRCLE_WITH_TWO_HORIZONTAL_STROKES_TO_THE_RIGHT = 0x29C3;
const char32_t SQUARED_RISING_DIAGONAL_SLASH = 0x29C4;
const char32_t SQUARED_FALLING_DIAGONAL_SLASH = 0x29C5;
const char32_t SQUARED_ASTERISK = 0x29C6;
const char32_t SQUARED_SMALL_CIRCLE = 0x29C7;
const char32_t SQUARED_SQUARE = 0x29C8;
const char32_t TWO_JOINED_SQUARES = 0x29C9;
const char32_t TRIANGLE_WITH_DOT_ABOVE = 0x29CA;
const char32_t TRIANGLE_WITH_UNDERBAR = 0x29CB;
const char32_t S_IN_TRIANGLE = 0x29CC;
const char32_t TRIANGLE_WITH_SERIFS_AT_BOTTOM = 0x29CD;
const char32_t RIGHT_TRIANGLE_ABOVE_LEFT_TRIANGLE = 0x29CE;
const char32_t LEFT_TRIANGLE_BESIDE_VERTICAL_BAR = 0x29CF;
const char32_t VERTICAL_BAR_BESIDE_RIGHT_TRIANGLE = 0x29D0;
const char32_t BOWTIE_WITH_LEFT_HALF_BLACK = 0x29D1;
const char32_t BOWTIE_WITH_RIGHT_HALF_BLACK = 0x29D2;
const char32_t BLACK_BOWTIE = 0x29D3;
const char32_t TIMES_WITH_LEFT_HALF_BLACK = 0x29D4;
const char32_t TIMES_WITH_RIGHT_HALF_BLACK = 0x29D5;
const char32_t WHITE_HOURGLASS = 0x29D6;
const char32_t BLACK_HOURGLASS = 0x29D7;
const char32_t LEFT_WIGGLY_FENCE = 0x29D8;
const char32_t RIGHT_WIGGLY_FENCE = 0x29D9;
const char32_t LEFT_DOUBLE_WIGGLY_FENCE = 0x29DA;
const char32_t RIGHT_DOUBLE_WIGGLY_FENCE = 0x29DB;
const char32_t INCOMPLETE_INFINITY = 0x29DC;
const char32_t TIE_OVER_INFINITY = 0x29DD;
const char32_t INFINITY_NEGATED_WITH_VERTICAL_BAR = 0x29DE;
const char32_t DOUBLE_ENDED_MULTIMAP = 0x29DF;
const char32_t SQUARE_WITH_CONTOURED_OUTLINE = 0x29E0;
const char32_t INCREASES_AS = 0x29E1;
const char32_t SHUFFLE_PRODUCT = 0x29E2;
const char32_t EQUALS_SIGN_AND_SLANTED_PARALLEL = 0x29E3;
const char32_t EQUALS_SIGN_AND_SLANTED_PARALLEL_WITH_TILDE_ABOVE = 0x29E4;
const char32_t IDENTICAL_TO_AND_SLANTED_PARALLEL = 0x29E5;
const char32_t GLEICH_STARK = 0x29E6;
const char32_t THERMODYNAMIC = 0x29E7;
const char32_t DOWN_POINTING_TRIANGLE_WITH_LEFT_HALF_BLACK = 0x29E8;
const char32_t DOWN_POINTING_TRIANGLE_WITH_RIGHT_HALF_BLACK = 0x29E9;
const char32_t BLACK_DIAMOND_WITH_DOWN_ARROW = 0x29EA;
const char32_t BLACK_LOZENGE = 0x29EB;
const char32_t WHITE_CIRCLE_WITH_DOWN_ARROW = 0x29EC;
const char32_t BLACK_CIRCLE_WITH_DOWN_ARROW = 0x29ED;
const char32_t ERROR_BARRED_WHITE_SQUARE = 0x29EE;
const char32_t ERROR_BARRED_BLACK_SQUARE = 0x29EF;
const char32_t ERROR_BARRED_WHITE_DIAMOND = 0x29F0;
const char32_t ERROR_BARRED_BLACK_DIAMOND = 0x29F1;
const char32_t ERROR_BARRED_WHITE_CIRCLE = 0x29F2;
const char32_t ERROR_BARRED_BLACK_CIRCLE = 0x29F3;
const char32_t RULE_DELAYED = 0x29F4;
const char32_t REVERSE_SOLIDUS_OPERATOR = 0x29F5;
const char32_t SOLIDUS_WITH_OVERBAR = 0x29F6;
const char32_t REVERSE_SOLIDUS_WITH_HORIZONTAL_STROKE = 0x29F7;
const char32_t BIG_SOLIDUS = 0x29F8;
const char32_t BIG_REVERSE_SOLIDUS = 0x29F9;
const char32_t DOUBLE_PLUS = 0x29FA;
const char32_t TRIPLE_PLUS = 0x29FB;
const char32_t LEFT_POINTING_CURVED_ANGLE_BRACKET = 0x29FC;
const char32_t RIGHT_POINTING_CURVED_ANGLE_BRACKET = 0x29FD;
const char32_t TINY = 0x29FE;
const char32_t MINY = 0x29FF;
const char32_t N_ARY_CIRCLED_DOT_OPERATOR = 0x2A00;
const char32_t N_ARY_CIRCLED_PLUS_OPERATOR = 0x2A01;
const char32_t N_ARY_CIRCLED_TIMES_OPERATOR = 0x2A02;
const char32_t N_ARY_UNION_OPERATOR_WITH_DOT = 0x2A03;
const char32_t N_ARY_UNION_OPERATOR_WITH_PLUS = 0x2A04;
const char32_t N_ARY_SQUARE_INTERSECTION_OPERATOR = 0x2A05;
const char32_t N_ARY_SQUARE_UNION_OPERATOR = 0x2A06;
const char32_t TWO_LOGICAL_AND_OPERATOR = 0x2A07;
const char32_t TWO_LOGICAL_OR_OPERATOR = 0x2A08;
const char32_t N_ARY_TIMES_OPERATOR = 0x2A09;
const char32_t MODULO_TWO_SUM = 0x2A0A;
const char32_t SUMMATION_WITH_INTEGRAL = 0x2A0B;
const char32_t QUADRUPLE_INTEGRAL_OPERATOR = 0x2A0C;
const char32_t FINITE_PART_INTEGRAL = 0x2A0D;
const char32_t INTEGRAL_WITH_DOUBLE_STROKE = 0x2A0E;
const char32_t INTEGRAL_AVERAGE_WITH_SLASH = 0x2A0F;
const char32_t CIRCULATION_FUNCTION = 0x2A10;
const char32_t ANTICLOCKWISE_INTEGRATION = 0x2A11;
const char32_t LINE_INTEGRATION_WITH_RECTANGULAR_PATH_AROUND_POLE = 0x2A12;
const char32_t LINE_INTEGRATION_WITH_SEMICIRCULAR_PATH_AROUND_POLE = 0x2A13;
const char32_t LINE_INTEGRATION_NOT_INCLUDING_THE_POLE = 0x2A14;
const char32_t INTEGRAL_AROUND_A_POINT_OPERATOR = 0x2A15;
const char32_t QUATERNION_INTEGRAL_OPERATOR = 0x2A16;
const char32_t INTEGRAL_WITH_LEFTWARDS_ARROW_WITH_HOOK = 0x2A17;
const char32_t INTEGRAL_WITH_TIMES_SIGN = 0x2A18;
const char32_t INTEGRAL_WITH_INTERSECTION = 0x2A19;
const char32_t INTEGRAL_WITH_UNION = 0x2A1A;
const char32_t INTEGRAL_WITH_OVERBAR = 0x2A1B;
const char32_t INTEGRAL_WITH_UNDERBAR = 0x2A1C;
const char32_t JOIN = 0x2A1D;
const char32_t LARGE_LEFT_TRIANGLE_OPERATOR = 0x2A1E;
const char32_t Z_NOTATION_SCHEMA_COMPOSITION = 0x2A1F;
const char32_t Z_NOTATION_SCHEMA_PIPING = 0x2A20;
const char32_t Z_NOTATION_SCHEMA_PROJECTION = 0x2A21;
const char32_t PLUS_SIGN_WITH_SMALL_CIRCLE_ABOVE = 0x2A22;
const char32_t PLUS_SIGN_WITH_CIRCUMFLEX_ACCENT_ABOVE = 0x2A23;
const char32_t PLUS_SIGN_WITH_TILDE_ABOVE = 0x2A24;
const char32_t PLUS_SIGN_WITH_DOT_BELOW = 0x2A25;
const char32_t PLUS_SIGN_WITH_TILDE_BELOW = 0x2A26;
const char32_t PLUS_SIGN_WITH_SUBSCRIPT_TWO = 0x2A27;
const char32_t PLUS_SIGN_WITH_BLACK_TRIANGLE = 0x2A28;
const char32_t MINUS_SIGN_WITH_COMMA_ABOVE = 0x2A29;
const char32_t MINUS_SIGN_WITH_DOT_BELOW = 0x2A2A;
const char32_t MINUS_SIGN_WITH_FALLING_DOTS = 0x2A2B;
const char32_t MINUS_SIGN_WITH_RISING_DOTS = 0x2A2C;
const char32_t PLUS_SIGN_IN_LEFT_HALF_CIRCLE = 0x2A2D;
const char32_t PLUS_SIGN_IN_RIGHT_HALF_CIRCLE = 0x2A2E;
const char32_t VECTOR_OR_CROSS_PRODUCT = 0x2A2F;
const char32_t MULTIPLICATION_SIGN_WITH_DOT_ABOVE = 0x2A30;
const char32_t MULTIPLICATION_SIGN_WITH_UNDERBAR = 0x2A31;
const char32_t SEMIDIRECT_PRODUCT_WITH_BOTTOM_CLOSED = 0x2A32;
const char32_t SMASH_PRODUCT = 0x2A33;
const char32_t MULTIPLICATION_SIGN_IN_LEFT_HALF_CIRCLE = 0x2A34;
const char32_t MULTIPLICATION_SIGN_IN_RIGHT_HALF_CIRCLE = 0x2A35;
const char32_t CIRCLED_MULTIPLICATION_SIGN_WITH_CIRCUMFLEX_ACCENT = 0x2A36;
const char32_t MULTIPLICATION_SIGN_IN_DOUBLE_CIRCLE = 0x2A37;
const char32_t CIRCLED_DIVISION_SIGN = 0x2A38;
const char32_t PLUS_SIGN_IN_TRIANGLE = 0x2A39;
const char32_t MINUS_SIGN_IN_TRIANGLE = 0x2A3A;
const char32_t MULTIPLICATION_SIGN_IN_TRIANGLE = 0x2A3B;
const char32_t INTERIOR_PRODUCT = 0x2A3C;
const char32_t RIGHTHAND_INTERIOR_PRODUCT = 0x2A3D;
const char32_t Z_NOTATION_RELATIONAL_COMPOSITION = 0x2A3E;
const char32_t AMALGAMATION_OR_COPRODUCT = 0x2A3F;
const char32_t INTERSECTION_WITH_DOT = 0x2A40;
const char32_t UNION_WITH_MINUS_SIGN = 0x2A41;
const char32_t UNION_WITH_OVERBAR = 0x2A42;
const char32_t INTERSECTION_WITH_OVERBAR = 0x2A43;
const char32_t INTERSECTION_WITH_LOGICAL_AND = 0x2A44;
const char32_t UNION_WITH_LOGICAL_OR = 0x2A45;
const char32_t UNION_ABOVE_INTERSECTION = 0x2A46;
const char32_t INTERSECTION_ABOVE_UNION = 0x2A47;
const char32_t UNION_ABOVE_BAR_ABOVE_INTERSECTION = 0x2A48;
const char32_t INTERSECTION_ABOVE_BAR_ABOVE_UNION = 0x2A49;
const char32_t UNION_BESIDE_AND_JOINED_WITH_UNION = 0x2A4A;
const char32_t INTERSECTION_BESIDE_AND_JOINED_WITH_INTERSECTION = 0x2A4B;
const char32_t CLOSED_UNION_WITH_SERIFS = 0x2A4C;
const char32_t CLOSED_INTERSECTION_WITH_SERIFS = 0x2A4D;
const char32_t DOUBLE_SQUARE_INTERSECTION = 0x2A4E;
const char32_t DOUBLE_SQUARE_UNION = 0x2A4F;
const char32_t CLOSED_UNION_WITH_SERIFS_AND_SMASH_PRODUCT = 0x2A50;
const char32_t LOGICAL_AND_WITH_DOT_ABOVE = 0x2A51;
const char32_t LOGICAL_OR_WITH_DOT_ABOVE = 0x2A52;
const char32_t DOUBLE_LOGICAL_AND = 0x2A53;
const char32_t DOUBLE_LOGICAL_OR = 0x2A54;
const char32_t TWO_INTERSECTING_LOGICAL_AND = 0x2A55;
const char32_t TWO_INTERSECTING_LOGICAL_OR = 0x2A56;
const char32_t SLOPING_LARGE_OR = 0x2A57;
const char32_t SLOPING_LARGE_AND = 0x2A58;
const char32_t LOGICAL_OR_OVERLAPPING_LOGICAL_AND = 0x2A59;
const char32_t LOGICAL_AND_WITH_MIDDLE_STEM = 0x2A5A;
const char32_t LOGICAL_OR_WITH_MIDDLE_STEM = 0x2A5B;
const char32_t LOGICAL_AND_WITH_HORIZONTAL_DASH = 0x2A5C;
const char32_t LOGICAL_OR_WITH_HORIZONTAL_DASH = 0x2A5D;
const char32_t LOGICAL_AND_WITH_DOUBLE_OVERBAR = 0x2A5E;
const char32_t LOGICAL_AND_WITH_UNDERBAR = 0x2A5F;
const char32_t LOGICAL_AND_WITH_DOUBLE_UNDERBAR = 0x2A60;
const char32_t SMALL_VEE_WITH_UNDERBAR = 0x2A61;
const char32_t LOGICAL_OR_WITH_DOUBLE_OVERBAR = 0x2A62;
const char32_t LOGICAL_OR_WITH_DOUBLE_UNDERBAR = 0x2A63;
const char32_t Z_NOTATION_DOMAIN_ANTIRESTRICTION = 0x2A64;
const char32_t Z_NOTATION_RANGE_ANTIRESTRICTION = 0x2A65;
const char32_t EQUALS_SIGN_WITH_DOT_BELOW = 0x2A66;
const char32_t IDENTICAL_WITH_DOT_ABOVE = 0x2A67;
const char32_t TRIPLE_HORIZONTAL_BAR_WITH_DOUBLE_VERTICAL_STROKE = 0x2A68;
const char32_t TRIPLE_HORIZONTAL_BAR_WITH_TRIPLE_VERTICAL_STROKE = 0x2A69;
const char32_t TILDE_OPERATOR_WITH_DOT_ABOVE = 0x2A6A;
const char32_t TILDE_OPERATOR_WITH_RISING_DOTS = 0x2A6B;
const char32_t SIMILAR_MINUS_SIMILAR = 0x2A6C;
const char32_t CONGRUENT_WITH_DOT_ABOVE = 0x2A6D;
const char32_t EQUALS_WITH_ASTERISK = 0x2A6E;
const char32_t ALMOST_EQUAL_TO_WITH_CIRCUMFLEX_ACCENT = 0x2A6F;
const char32_t APPROXIMATELY_EQUAL_OR_EQUAL_TO = 0x2A70;
const char32_t EQUALS_SIGN_ABOVE_PLUS_SIGN = 0x2A71;
const char32_t PLUS_SIGN_ABOVE_EQUALS_SIGN = 0x2A72;
const char32_t EQUALS_SIGN_ABOVE_TILDE_OPERATOR = 0x2A73;
const char32_t DOUBLE_COLON_EQUAL = 0x2A74;
const char32_t TWO_CONSECUTIVE_EQUALS_SIGNS = 0x2A75;
const char32_t THREE_CONSECUTIVE_EQUALS_SIGNS = 0x2A76;
const char32_t EQUALS_SIGN_WITH_TWO_DOTS_ABOVE_AND_TWO_DOTS_BELOW = 0x2A77;
const char32_t EQUIVALENT_WITH_FOUR_DOTS_ABOVE = 0x2A78;
const char32_t LESS_THAN_WITH_CIRCLE_INSIDE = 0x2A79;
const char32_t GREATER_THAN_WITH_CIRCLE_INSIDE = 0x2A7A;
const char32_t LESS_THAN_WITH_QUESTION_MARK_ABOVE = 0x2A7B;
const char32_t GREATER_THAN_WITH_QUESTION_MARK_ABOVE = 0x2A7C;
const char32_t LESS_THAN_OR_SLANTED_EQUAL_TO = 0x2A7D;
const char32_t GREATER_THAN_OR_SLANTED_EQUAL_TO = 0x2A7E;
const char32_t LESS_THAN_OR_SLANTED_EQUAL_TO_WITH_DOT_INSIDE = 0x2A7F;
const char32_t GREATER_THAN_OR_SLANTED_EQUAL_TO_WITH_DOT_INSIDE = 0x2A80;
const char32_t LESS_THAN_OR_SLANTED_EQUAL_TO_WITH_DOT_ABOVE = 0x2A81;
const char32_t GREATER_THAN_OR_SLANTED_EQUAL_TO_WITH_DOT_ABOVE = 0x2A82;
const char32_t LESS_THAN_OR_SLANTED_EQUAL_TO_WITH_DOT_ABOVE_RIGHT = 0x2A83;
const char32_t GREATER_THAN_OR_SLANTED_EQUAL_TO_WITH_DOT_ABOVE_LEFT = 0x2A84;
const char32_t LESS_THAN_OR_APPROXIMATE = 0x2A85;
const char32_t GREATER_THAN_OR_APPROXIMATE = 0x2A86;
const char32_t LESS_THAN_AND_SINGLE_LINE_NOT_EQUAL_TO = 0x2A87;
const char32_t GREATER_THAN_AND_SINGLE_LINE_NOT_EQUAL_TO = 0x2A88;
const char32_t LESS_THAN_AND_NOT_APPROXIMATE = 0x2A89;
const char32_t GREATER_THAN_AND_NOT_APPROXIMATE = 0x2A8A;
const char32_t LESS_THAN_ABOVE_DOUBLE_LINE_EQUAL_ABOVE_GREATER_THAN = 0x2A8B;
const char32_t GREATER_THAN_ABOVE_DOUBLE_LINE_EQUAL_ABOVE_LESS_THAN = 0x2A8C;
const char32_t LESS_THAN_ABOVE_SIMILAR_OR_EQUAL = 0x2A8D;
const char32_t GREATER_THAN_ABOVE_SIMILAR_OR_EQUAL = 0x2A8E;
const char32_t LESS_THAN_ABOVE_SIMILAR_ABOVE_GREATER_THAN = 0x2A8F;
const char32_t GREATER_THAN_ABOVE_SIMILAR_ABOVE_LESS_THAN = 0x2A90;
const char32_t LESS_THAN_ABOVE_GREATER_THAN_ABOVE_DOUBLE_LINE_EQUAL = 0x2A91;
const char32_t GREATER_THAN_ABOVE_LESS_THAN_ABOVE_DOUBLE_LINE_EQUAL = 0x2A92;
const char32_t LESS_THAN_ABOVE_SLANTED_EQUAL_ABOVE_GREATER_THAN_ABOVE_SLANTED_EQUAL = 0x2A93;
const char32_t GREATER_THAN_ABOVE_SLANTED_EQUAL_ABOVE_LESS_THAN_ABOVE_SLANTED_EQUAL = 0x2A94;
const char32_t SLANTED_EQUAL_TO_OR_LESS_THAN = 0x2A95;
const char32_t SLANTED_EQUAL_TO_OR_GREATER_THAN = 0x2A96;
const char32_t SLANTED_EQUAL_TO_OR_LESS_THAN_WITH_DOT_INSIDE = 0x2A97;
const char32_t SLANTED_EQUAL_TO_OR_GREATER_THAN_WITH_DOT_INSIDE = 0x2A98;
const char32_t DOUBLE_LINE_EQUAL_TO_OR_LESS_THAN = 0x2A99;
const char32_t DOUBLE_LINE_EQUAL_TO_OR_GREATER_THAN = 0x2A9A;
const char32_t DOUBLE_LINE_SLANTED_EQUAL_TO_OR_LESS_THAN = 0x2A9B;
const char32_t DOUBLE_LINE_SLANTED_EQUAL_TO_OR_GREATER_THAN = 0x2A9C;
const char32_t SIMILAR_OR_LESS_THAN = 0x2A9D;
const char32_t SIMILAR_OR_GREATER_THAN = 0x2A9E;
const char32_t SIMILAR_ABOVE_LESS_THAN_ABOVE_EQUALS_SIGN = 0x2A9F;
const char32_t SIMILAR_ABOVE_GREATER_THAN_ABOVE_EQUALS_SIGN = 0x2AA0;
const char32_t DOUBLE_NESTED_LESS_THAN = 0x2AA1;
const char32_t DOUBLE_NESTED_GREATER_THAN = 0x2AA2;
const char32_t DOUBLE_NESTED_LESS_THAN_WITH_UNDERBAR = 0x2AA3;
const char32_t GREATER_THAN_OVERLAPPING_LESS_THAN = 0x2AA4;
const char32_t GREATER_THAN_BESIDE_LESS_THAN = 0x2AA5;
const char32_t LESS_THAN_CLOSED_BY_CURVE = 0x2AA6;
const char32_t GREATER_THAN_CLOSED_BY_CURVE = 0x2AA7;
const char32_t LESS_THAN_CLOSED_BY_CURVE_ABOVE_SLANTED_EQUAL = 0x2AA8;
const char32_t GREATER_THAN_CLOSED_BY_CURVE_ABOVE_SLANTED_EQUAL = 0x2AA9;
const char32_t SMALLER_THAN = 0x2AAA;
const char32_t LARGER_THAN = 0x2AAB;
const char32_t SMALLER_THAN_OR_EQUAL_TO = 0x2AAC;
const char32_t LARGER_THAN_OR_EQUAL_TO = 0x2AAD;
const char32_t EQUALS_SIGN_WITH_BUMPY_ABOVE = 0x2AAE;
const char32_t PRECEDES_ABOVE_SINGLE_LINE_EQUALS_SIGN = 0x2AAF;
const char32_t SUCCEEDS_ABOVE_SINGLE_LINE_EQUALS_SIGN = 0x2AB0;
const char32_t PRECEDES_ABOVE_SINGLE_LINE_NOT_EQUAL_TO = 0x2AB1;
const char32_t SUCCEEDS_ABOVE_SINGLE_LINE_NOT_EQUAL_TO = 0x2AB2;
const char32_t PRECEDES_ABOVE_EQUALS_SIGN = 0x2AB3;
const char32_t SUCCEEDS_ABOVE_EQUALS_SIGN = 0x2AB4;
const char32_t PRECEDES_ABOVE_NOT_EQUAL_TO = 0x2AB5;
const char32_t SUCCEEDS_ABOVE_NOT_EQUAL_TO = 0x2AB6;
const char32_t PRECEDES_ABOVE_ALMOST_EQUAL_TO = 0x2AB7;
const char32_t SUCCEEDS_ABOVE_ALMOST_EQUAL_TO = 0x2AB8;
const char32_t PRECEDES_ABOVE_NOT_ALMOST_EQUAL_TO = 0x2AB9;
const char32_t SUCCEEDS_ABOVE_NOT_ALMOST_EQUAL_TO = 0x2ABA;
const char32_t DOUBLE_PRECEDES = 0x2ABB;
const char32_t DOUBLE_SUCCEEDS = 0x2ABC;
const char32_t SUBSET_WITH_DOT = 0x2ABD;
const char32_t SUPERSET_WITH_DOT = 0x2ABE;
const char32_t SUBSET_WITH_PLUS_SIGN_BELOW = 0x2ABF;
const char32_t SUPERSET_WITH_PLUS_SIGN_BELOW = 0x2AC0;
const char32_t SUBSET_WITH_MULTIPLICATION_SIGN_BELOW = 0x2AC1;
const char32_t SUPERSET_WITH_MULTIPLICATION_SIGN_BELOW = 0x2AC2;
const char32_t SUBSET_OF_OR_EQUAL_TO_WITH_DOT_ABOVE = 0x2AC3;
const char32_t SUPERSET_OF_OR_EQUAL_TO_WITH_DOT_ABOVE = 0x2AC4;
const char32_t SUBSET_OF_ABOVE_EQUALS_SIGN = 0x2AC5;
const char32_t SUPERSET_OF_ABOVE_EQUALS_SIGN = 0x2AC6;
const char32_t SUBSET_OF_ABOVE_TILDE_OPERATOR = 0x2AC7;
const char32_t SUPERSET_OF_ABOVE_TILDE_OPERATOR = 0x2AC8;
const char32_t SUBSET_OF_ABOVE_ALMOST_EQUAL_TO = 0x2AC9;
const char32_t SUPERSET_OF_ABOVE_ALMOST_EQUAL_TO = 0x2ACA;
const char32_t SUBSET_OF_ABOVE_NOT_EQUAL_TO = 0x2ACB;
const char32_t SUPERSET_OF_ABOVE_NOT_EQUAL_TO = 0x2ACC;
const char32_t SQUARE_LEFT_OPEN_BOX_OPERATOR = 0x2ACD;
const char32_t SQUARE_RIGHT_OPEN_BOX_OPERATOR = 0x2ACE;
const char32_t CLOSED_SUBSET = 0x2ACF;
const char32_t CLOSED_SUPERSET = 0x2AD0;
const char32_t CLOSED_SUBSET_OR_EQUAL_TO = 0x2AD1;
const char32_t CLOSED_SUPERSET_OR_EQUAL_TO = 0x2AD2;
const char32_t SUBSET_ABOVE_SUPERSET = 0x2AD3;
const char32_t SUPERSET_ABOVE_SUBSET = 0x2AD4;
const char32_t SUBSET_ABOVE_SUBSET = 0x2AD5;
const char32_t SUPERSET_ABOVE_SUPERSET = 0x2AD6;
const char32_t SUPERSET_BESIDE_SUBSET = 0x2AD7;
const char32_t SUPERSET_BESIDE_AND_JOINED_BY_DASH_WITH_SUBSET = 0x2AD8;
const char32_t ELEMENT_OF_OPENING_DOWNWARDS = 0x2AD9;
const char32_t PITCHFORK_WITH_TEE_TOP = 0x2ADA;
const char32_t TRANSVERSAL_INTERSECTION = 0x2ADB;
const char32_t FORKING = 0x2ADC;
const char32_t NONFORKING = 0x2ADD;
const char32_t SHORT_LEFT_TACK = 0x2ADE;
const char32_t SHORT_DOWN_TACK = 0x2ADF;
const char32_t SHORT_UP_TACK = 0x2AE0;
const char32_t PERPENDICULAR_WITH_S = 0x2AE1;
const char32_t VERTICAL_BAR_TRIPLE_RIGHT_TURNSTILE = 0x2AE2;
const char32_t DOUBLE_VERTICAL_BAR_LEFT_TURNSTILE = 0x2AE3;
const char32_t VERTICAL_BAR_DOUBLE_LEFT_TURNSTILE = 0x2AE4;
const char32_t DOUBLE_VERTICAL_BAR_DOUBLE_LEFT_TURNSTILE = 0x2AE5;
const char32_t LONG_DASH_FROM_LEFT_MEMBER_OF_DOUBLE_VERTICAL = 0x2AE6;
const char32_t SHORT_DOWN_TACK_WITH_OVERBAR = 0x2AE7;
const char32_t SHORT_UP_TACK_WITH_UNDERBAR = 0x2AE8;
const char32_t SHORT_UP_TACK_ABOVE_SHORT_DOWN_TACK = 0x2AE9;
const char32_t DOUBLE_DOWN_TACK = 0x2AEA;
const char32_t DOUBLE_UP_TACK = 0x2AEB;
const char32_t DOUBLE_STROKE_NOT_SIGN = 0x2AEC;
const char32_t REVERSED_DOUBLE_STROKE_NOT_SIGN = 0x2AED;
const char32_t DOES_NOT_DIVIDE_WITH_REVERSED_NEGATION_SLASH = 0x2AEE;
const char32_t VERTICAL_LINE_WITH_CIRCLE_ABOVE = 0x2AEF;
const char32_t VERTICAL_LINE_WITH_CIRCLE_BELOW = 0x2AF0;
const char32_t DOWN_TACK_WITH_CIRCLE_BELOW = 0x2AF1;
const char32_t PARALLEL_WITH_HORIZONTAL_STROKE = 0x2AF2;
const char32_t PARALLEL_WITH_TILDE_OPERATOR = 0x2AF3;
const char32_t TRIPLE_VERTICAL_BAR_BINARY_RELATION = 0x2AF4;
const char32_t TRIPLE_VERTICAL_BAR_WITH_HORIZONTAL_STROKE = 0x2AF5;
const char32_t TRIPLE_COLON_OPERATOR = 0x2AF6;
const char32_t TRIPLE_NESTED_LESS_THAN = 0x2AF7;
const char32_t TRIPLE_NESTED_GREATER_THAN = 0x2AF8;
const char32_t DOUBLE_LINE_SLANTED_LESS_THAN_OR_EQUAL_TO = 0x2AF9;
const char32_t DOUBLE_LINE_SLANTED_GREATER_THAN_OR_EQUAL_TO = 0x2AFA;
const char32_t TRIPLE_SOLIDUS_BINARY_RELATION = 0x2AFB;
const char32_t LARGE_TRIPLE_VERTICAL_BAR_OPERATOR = 0x2AFC;
const char32_t DOUBLE_SOLIDUS_OPERATOR = 0x2AFD;
const char32_t WHITE_VERTICAL_BAR = 0x2AFE;
const char32_t N_ARY_WHITE_VERTICAL_BAR = 0x2AFF;
const char32_t NORTH_EAST_WHITE_ARROW = 0x2B00;
const char32_t NORTH_WEST_WHITE_ARROW = 0x2B01;
const char32_t SOUTH_EAST_WHITE_ARROW = 0x2B02;
const char32_t SOUTH_WEST_WHITE_ARROW = 0x2B03;
const char32_t LEFT_RIGHT_WHITE_ARROW = 0x2B04;
const char32_t LEFTWARDS_BLACK_ARROW = 0x2B05;
const char32_t UPWARDS_BLACK_ARROW = 0x2B06;
const char32_t DOWNWARDS_BLACK_ARROW = 0x2B07;
const char32_t NORTH_EAST_BLACK_ARROW = 0x2B08;
const char32_t NORTH_WEST_BLACK_ARROW = 0x2B09;
const char32_t SOUTH_EAST_BLACK_ARROW = 0x2B0A;
const char32_t SOUTH_WEST_BLACK_ARROW = 0x2B0B;
const char32_t LEFT_RIGHT_BLACK_ARROW = 0x2B0C;
const char32_t UP_DOWN_BLACK_ARROW = 0x2B0D;
const char32_t RIGHTWARDS_ARROW_WITH_TIP_DOWNWARDS = 0x2B0E;
const char32_t RIGHTWARDS_ARROW_WITH_TIP_UPWARDS = 0x2B0F;
const char32_t LEFTWARDS_ARROW_WITH_TIP_DOWNWARDS = 0x2B10;
const char32_t LEFTWARDS_ARROW_WITH_TIP_UPWARDS = 0x2B11;
const char32_t SQUARE_WITH_TOP_HALF_BLACK = 0x2B12;
const char32_t SQUARE_WITH_BOTTOM_HALF_BLACK = 0x2B13;
const char32_t SQUARE_WITH_UPPER_RIGHT_DIAGONAL_HALF_BLACK = 0x2B14;
const char32_t SQUARE_WITH_LOWER_LEFT_DIAGONAL_HALF_BLACK = 0x2B15;
const char32_t DIAMOND_WITH_LEFT_HALF_BLACK = 0x2B16;
const char32_t DIAMOND_WITH_RIGHT_HALF_BLACK = 0x2B17;
const char32_t DIAMOND_WITH_TOP_HALF_BLACK = 0x2B18;
const char32_t DIAMOND_WITH_BOTTOM_HALF_BLACK = 0x2B19;
const char32_t DOTTED_SQUARE = 0x2B1A;
const char32_t BLACK_LARGE_SQUARE = 0x2B1B;
const char32_t WHITE_LARGE_SQUARE = 0x2B1C;
const char32_t BLACK_VERY_SMALL_SQUARE = 0x2B1D;
const char32_t WHITE_VERY_SMALL_SQUARE = 0x2B1E;
const char32_t BLACK_PENTAGON = 0x2B1F;
const char32_t WHITE_PENTAGON = 0x2B20;
const char32_t WHITE_HEXAGON = 0x2B21;
const char32_t BLACK_HEXAGON = 0x2B22;
const char32_t HORIZONTAL_BLACK_HEXAGON = 0x2B23;
const char32_t BLACK_LARGE_CIRCLE = 0x2B24;
const char32_t BLACK_MEDIUM_DIAMOND = 0x2B25;
const char32_t WHITE_MEDIUM_DIAMOND = 0x2B26;
const char32_t BLACK_MEDIUM_LOZENGE = 0x2B27;
const char32_t WHITE_MEDIUM_LOZENGE = 0x2B28;
const char32_t BLACK_SMALL_DIAMOND = 0x2B29;
const char32_t BLACK_SMALL_LOZENGE = 0x2B2A;
const char32_t WHITE_SMALL_LOZENGE = 0x2B2B;
const char32_t BLACK_HORIZONTAL_ELLIPSE = 0x2B2C;
const char32_t WHITE_HORIZONTAL_ELLIPSE = 0x2B2D;
const char32_t BLACK_VERTICAL_ELLIPSE = 0x2B2E;
const char32_t WHITE_VERTICAL_ELLIPSE = 0x2B2F;
const char32_t LEFT_ARROW_WITH_SMALL_CIRCLE = 0x2B30;
const char32_t THREE_LEFTWARDS_ARROWS = 0x2B31;
const char32_t LEFT_ARROW_WITH_CIRCLED_PLUS = 0x2B32;
const char32_t LONG_LEFTWARDS_SQUIGGLE_ARROW = 0x2B33;
const char32_t LEFTWARDS_TWO_HEADED_ARROW_WITH_VERTICAL_STROKE = 0x2B34;
const char32_t LEFTWARDS_TWO_HEADED_ARROW_WITH_DOUBLE_VERTICAL_STROKE = 0x2B35;
const char32_t LEFTWARDS_TWO_HEADED_ARROW_FROM_BAR = 0x2B36;
const char32_t LEFTWARDS_TWO_HEADED_TRIPLE_DASH_ARROW = 0x2B37;
const char32_t LEFTWARDS_ARROW_WITH_DOTTED_STEM = 0x2B38;
const char32_t LEFTWARDS_ARROW_WITH_TAIL_WITH_VERTICAL_STROKE = 0x2B39;
const char32_t LEFTWARDS_ARROW_WITH_TAIL_WITH_DOUBLE_VERTICAL_STROKE = 0x2B3A;
const char32_t LEFTWARDS_TWO_HEADED_ARROW_WITH_TAIL = 0x2B3B;
const char32_t LEFTWARDS_TWO_HEADED_ARROW_WITH_TAIL_WITH_VERTICAL_STROKE = 0x2B3C;
const char32_t LEFTWARDS_TWO_HEADED_ARROW_WITH_TAIL_WITH_DOUBLE_VERTICAL_STROKE = 0x2B3D;
const char32_t LEFTWARDS_ARROW_THROUGH_X = 0x2B3E;
const char32_t WAVE_ARROW_POINTING_DIRECTLY_LEFT = 0x2B3F;
const char32_t EQUALS_SIGN_ABOVE_LEFTWARDS_ARROW = 0x2B40;
const char32_t REVERSE_TILDE_OPERATOR_ABOVE_LEFTWARDS_ARROW = 0x2B41;
const char32_t LEFTWARDS_ARROW_ABOVE_REVERSE_ALMOST_EQUAL_TO = 0x2B42;
const char32_t RIGHTWARDS_ARROW_THROUGH_GREATER_THAN = 0x2B43;
const char32_t RIGHTWARDS_ARROW_THROUGH_SUPERSET = 0x2B44;
const char32_t LEFTWARDS_QUADRUPLE_ARROW = 0x2B45;
const char32_t RIGHTWARDS_QUADRUPLE_ARROW = 0x2B46;
const char32_t REVERSE_TILDE_OPERATOR_ABOVE_RIGHTWARDS_ARROW = 0x2B47;
const char32_t RIGHTWARDS_ARROW_ABOVE_REVERSE_ALMOST_EQUAL_TO = 0x2B48;
const char32_t TILDE_OPERATOR_ABOVE_LEFTWARDS_ARROW = 0x2B49;
const char32_t LEFTWARDS_ARROW_ABOVE_ALMOST_EQUAL_TO = 0x2B4A;
const char32_t LEFTWARDS_ARROW_ABOVE_REVERSE_TILDE_OPERATOR = 0x2B4B;
const char32_t RIGHTWARDS_ARROW_ABOVE_REVERSE_TILDE_OPERATOR = 0x2B4C;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ZIGZAG_ARROW = 0x2B4D;
const char32_t SHORT_SLANTED_NORTH_ARROW = 0x2B4E;
const char32_t SHORT_BACKSLANTED_SOUTH_ARROW = 0x2B4F;
const char32_t WHITE_MEDIUM_STAR = 0x2B50;
const char32_t BLACK_SMALL_STAR = 0x2B51;
const char32_t WHITE_SMALL_STAR = 0x2B52;
const char32_t BLACK_RIGHT_POINTING_PENTAGON = 0x2B53;
const char32_t WHITE_RIGHT_POINTING_PENTAGON = 0x2B54;
const char32_t HEAVY_LARGE_CIRCLE = 0x2B55;
const char32_t HEAVY_OVAL_WITH_OVAL_INSIDE = 0x2B56;
const char32_t HEAVY_CIRCLE_WITH_CIRCLE_INSIDE = 0x2B57;
const char32_t HEAVY_CIRCLE = 0x2B58;
const char32_t HEAVY_CIRCLED_SALTIRE = 0x2B59;
const char32_t SLANTED_NORTH_ARROW_WITH_HOOKED_HEAD = 0x2B5A;
const char32_t BACKSLANTED_SOUTH_ARROW_WITH_HOOKED_TAIL = 0x2B5B;
const char32_t SLANTED_NORTH_ARROW_WITH_HORIZONTAL_TAIL = 0x2B5C;
const char32_t BACKSLANTED_SOUTH_ARROW_WITH_HORIZONTAL_TAIL = 0x2B5D;
const char32_t BENT_ARROW_POINTING_DOWNWARDS_THEN_NORTH_EAST = 0x2B5E;
const char32_t SHORT_BENT_ARROW_POINTING_DOWNWARDS_THEN_NORTH_EAST = 0x2B5F;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW = 0x2B60;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW = 0x2B61;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW = 0x2B62;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW = 0x2B63;
const char32_t LEFT_RIGHT_TRIANGLE_HEADED_ARROW = 0x2B64;
const char32_t UP_DOWN_TRIANGLE_HEADED_ARROW = 0x2B65;
const char32_t NORTH_WEST_TRIANGLE_HEADED_ARROW = 0x2B66;
const char32_t NORTH_EAST_TRIANGLE_HEADED_ARROW = 0x2B67;
const char32_t SOUTH_EAST_TRIANGLE_HEADED_ARROW = 0x2B68;
const char32_t SOUTH_WEST_TRIANGLE_HEADED_ARROW = 0x2B69;
const char32_t LEFTWARDS_TRIANGLE_HEADED_DASHED_ARROW = 0x2B6A;
const char32_t UPWARDS_TRIANGLE_HEADED_DASHED_ARROW = 0x2B6B;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_DASHED_ARROW = 0x2B6C;
const char32_t DOWNWARDS_TRIANGLE_HEADED_DASHED_ARROW = 0x2B6D;
const char32_t CLOCKWISE_TRIANGLE_HEADED_OPEN_CIRCLE_ARROW = 0x2B6E;
const char32_t ANTICLOCKWISE_TRIANGLE_HEADED_OPEN_CIRCLE_ARROW = 0x2B6F;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B70;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B71;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B72;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B73;
const char32_t NORTH_WEST_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B76;
const char32_t NORTH_EAST_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B77;
const char32_t SOUTH_EAST_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B78;
const char32_t SOUTH_WEST_TRIANGLE_HEADED_ARROW_TO_BAR = 0x2B79;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_DOUBLE_HORIZONTAL_STROKE = 0x2B7A;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_DOUBLE_HORIZONTAL_STROKE = 0x2B7B;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_DOUBLE_HORIZONTAL_STROKE = 0x2B7C;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_DOUBLE_HORIZONTAL_STROKE = 0x2B7D;
const char32_t HORIZONTAL_TAB_KEY = 0x2B7E;
const char32_t VERTICAL_TAB_KEY = 0x2B7F;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_OVER_RIGHTWARDS_TRIANGLE_HEADED_ARROW = 0x2B80;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_LEFTWARDS_OF_DOWNWARDS_TRIANGLE_HEADED_ARROW = 0x2B81;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_OVER_LEFTWARDS_TRIANGLE_HEADED_ARROW = 0x2B82;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_LEFTWARDS_OF_UPWARDS_TRIANGLE_HEADED_ARROW = 0x2B83;
const char32_t LEFTWARDS_TRIANGLE_HEADED_PAIRED_ARROWS = 0x2B84;
const char32_t UPWARDS_TRIANGLE_HEADED_PAIRED_ARROWS = 0x2B85;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_PAIRED_ARROWS = 0x2B86;
const char32_t DOWNWARDS_TRIANGLE_HEADED_PAIRED_ARROWS = 0x2B87;
const char32_t LEFTWARDS_BLACK_CIRCLED_WHITE_ARROW = 0x2B88;
const char32_t UPWARDS_BLACK_CIRCLED_WHITE_ARROW = 0x2B89;
const char32_t RIGHTWARDS_BLACK_CIRCLED_WHITE_ARROW = 0x2B8A;
const char32_t DOWNWARDS_BLACK_CIRCLED_WHITE_ARROW = 0x2B8B;
const char32_t ANTICLOCKWISE_TRIANGLE_HEADED_RIGHT_U_SHAPED_ARROW = 0x2B8C;
const char32_t ANTICLOCKWISE_TRIANGLE_HEADED_BOTTOM_U_SHAPED_ARROW = 0x2B8D;
const char32_t ANTICLOCKWISE_TRIANGLE_HEADED_LEFT_U_SHAPED_ARROW = 0x2B8E;
const char32_t ANTICLOCKWISE_TRIANGLE_HEADED_TOP_U_SHAPED_ARROW = 0x2B8F;
const char32_t RETURN_LEFT = 0x2B90;
const char32_t RETURN_RIGHT = 0x2B91;
const char32_t NEWLINE_LEFT = 0x2B92;
const char32_t NEWLINE_RIGHT = 0x2B93;
const char32_t FOUR_CORNER_ARROWS_CIRCLING_ANTICLOCKWISE = 0x2B94;
const char32_t RIGHTWARDS_BLACK_ARROW = 0x2B95;
const char32_t EQUALS_SIGN_WITH_INFINITY_ABOVE = 0x2B96;
const char32_t SYMBOL_FOR_TYPE_A_ELECTRONICS = 0x2B97;
const char32_t THREE_D_TOP_LIGHTED_LEFTWARDS_EQUILATERAL_ARROWHEAD = 0x2B98;
const char32_t THREE_D_RIGHT_LIGHTED_UPWARDS_EQUILATERAL_ARROWHEAD = 0x2B99;
const char32_t THREE_D_TOP_LIGHTED_RIGHTWARDS_EQUILATERAL_ARROWHEAD = 0x2B9A;
const char32_t THREE_D_LEFT_LIGHTED_DOWNWARDS_EQUILATERAL_ARROWHEAD = 0x2B9B;
const char32_t BLACK_LEFTWARDS_EQUILATERAL_ARROWHEAD = 0x2B9C;
const char32_t BLACK_UPWARDS_EQUILATERAL_ARROWHEAD = 0x2B9D;
const char32_t BLACK_RIGHTWARDS_EQUILATERAL_ARROWHEAD = 0x2B9E;
const char32_t BLACK_DOWNWARDS_EQUILATERAL_ARROWHEAD = 0x2B9F;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_LEFTWARDS = 0x2BA0;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_RIGHTWARDS = 0x2BA1;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_LEFTWARDS = 0x2BA2;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_RIGHTWARDS = 0x2BA3;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_UPWARDS = 0x2BA4;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_UPWARDS = 0x2BA5;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_DOWNWARDS = 0x2BA6;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_LONG_TIP_DOWNWARDS = 0x2BA7;
const char32_t BLACK_CURVED_DOWNWARDS_AND_LEFTWARDS_ARROW = 0x2BA8;
const char32_t BLACK_CURVED_DOWNWARDS_AND_RIGHTWARDS_ARROW = 0x2BA9;
const char32_t BLACK_CURVED_UPWARDS_AND_LEFTWARDS_ARROW = 0x2BAA;
const char32_t BLACK_CURVED_UPWARDS_AND_RIGHTWARDS_ARROW = 0x2BAB;
const char32_t BLACK_CURVED_LEFTWARDS_AND_UPWARDS_ARROW = 0x2BAC;
const char32_t BLACK_CURVED_RIGHTWARDS_AND_UPWARDS_ARROW = 0x2BAD;
const char32_t BLACK_CURVED_LEFTWARDS_AND_DOWNWARDS_ARROW = 0x2BAE;
const char32_t BLACK_CURVED_RIGHTWARDS_AND_DOWNWARDS_ARROW = 0x2BAF;
const char32_t RIBBON_ARROW_DOWN_LEFT = 0x2BB0;
const char32_t RIBBON_ARROW_DOWN_RIGHT = 0x2BB1;
const char32_t RIBBON_ARROW_UP_LEFT = 0x2BB2;
const char32_t RIBBON_ARROW_UP_RIGHT = 0x2BB3;
const char32_t RIBBON_ARROW_LEFT_UP = 0x2BB4;
const char32_t RIBBON_ARROW_RIGHT_UP = 0x2BB5;
const char32_t RIBBON_ARROW_LEFT_DOWN = 0x2BB6;
const char32_t RIBBON_ARROW_RIGHT_DOWN = 0x2BB7;
const char32_t UPWARDS_WHITE_ARROW_FROM_BAR_WITH_HORIZONTAL_BAR = 0x2BB8;
const char32_t UP_ARROWHEAD_IN_A_RECTANGLE_BOX = 0x2BB9;
const char32_t OVERLAPPING_WHITE_SQUARES = 0x2BBA;
const char32_t OVERLAPPING_WHITE_AND_BLACK_SQUARES = 0x2BBB;
const char32_t OVERLAPPING_BLACK_SQUARES = 0x2BBC;
const char32_t BALLOT_BOX_WITH_LIGHT_X = 0x2BBD;
const char32_t CIRCLED_X = 0x2BBE;
const char32_t CIRCLED_BOLD_X = 0x2BBF;
const char32_t BLACK_SQUARE_CENTRED = 0x2BC0;
const char32_t BLACK_DIAMOND_CENTRED = 0x2BC1;
const char32_t TURNED_BLACK_PENTAGON = 0x2BC2;
const char32_t HORIZONTAL_BLACK_OCTAGON = 0x2BC3;
const char32_t BLACK_OCTAGON = 0x2BC4;
const char32_t BLACK_MEDIUM_UP_POINTING_TRIANGLE_CENTRED = 0x2BC5;
const char32_t BLACK_MEDIUM_DOWN_POINTING_TRIANGLE_CENTRED = 0x2BC6;
const char32_t BLACK_MEDIUM_LEFT_POINTING_TRIANGLE_CENTRED = 0x2BC7;
const char32_t BLACK_MEDIUM_RIGHT_POINTING_TRIANGLE_CENTRED = 0x2BC8;
const char32_t NEPTUNE_FORM_TWO = 0x2BC9;
const char32_t TOP_HALF_BLACK_CIRCLE = 0x2BCA;
const char32_t BOTTOM_HALF_BLACK_CIRCLE = 0x2BCB;
const char32_t LIGHT_FOUR_POINTED_BLACK_CUSP = 0x2BCC;
const char32_t ROTATED_LIGHT_FOUR_POINTED_BLACK_CUSP = 0x2BCD;
const char32_t WHITE_FOUR_POINTED_CUSP = 0x2BCE;
const char32_t ROTATED_WHITE_FOUR_POINTED_CUSP = 0x2BCF;
const char32_t SQUARE_POSITION_INDICATOR = 0x2BD0;
const char32_t UNCERTAINTY_SIGN = 0x2BD1;
const char32_t GROUP_MARK = 0x2BD2;
const char32_t PLUTO_FORM_TWO = 0x2BD3;
const char32_t PLUTO_FORM_THREE = 0x2BD4;
const char32_t PLUTO_FORM_FOUR = 0x2BD5;
const char32_t PLUTO_FORM_FIVE = 0x2BD6;
const char32_t TRANSPLUTO = 0x2BD7;
const char32_t PROSERPINA = 0x2BD8;
const char32_t ASTRAEA = 0x2BD9;
const char32_t HYGIEA = 0x2BDA;
const char32_t PHOLUS = 0x2BDB;
const char32_t NESSUS = 0x2BDC;
const char32_t WHITE_MOON_SELENA = 0x2BDD;
const char32_t BLACK_DIAMOND_ON_CROSS = 0x2BDE;
const char32_t TRUE_LIGHT_MOON_ARTA = 0x2BDF;
const char32_t CUPIDO = 0x2BE0;
const char32_t HADES = 0x2BE1;
const char32_t ZEUS = 0x2BE2;
const char32_t KRONOS = 0x2BE3;
const char32_t APOLLON = 0x2BE4;
const char32_t ADMETOS = 0x2BE5;
const char32_t VULCANUS = 0x2BE6;
const char32_t POSEIDON = 0x2BE7;
const char32_t LEFT_HALF_BLACK_STAR = 0x2BE8;
const char32_t RIGHT_HALF_BLACK_STAR = 0x2BE9;
const char32_t STAR_WITH_LEFT_HALF_BLACK = 0x2BEA;
const char32_t STAR_WITH_RIGHT_HALF_BLACK = 0x2BEB;
const char32_t LEFTWARDS_TWO_HEADED_ARROW_WITH_TRIANGLE_ARROWHEADS = 0x2BEC;
const char32_t UPWARDS_TWO_HEADED_ARROW_WITH_TRIANGLE_ARROWHEADS = 0x2BED;
const char32_t RIGHTWARDS_TWO_HEADED_ARROW_WITH_TRIANGLE_ARROWHEADS = 0x2BEE;
const char32_t DOWNWARDS_TWO_HEADED_ARROW_WITH_TRIANGLE_ARROWHEADS = 0x2BEF;
const char32_t ERIS_FORM_ONE = 0x2BF0;
const char32_t ERIS_FORM_TWO = 0x2BF1;
const char32_t SEDNA = 0x2BF2;
const char32_t RUSSIAN_ASTROLOGICAL_SYMBOL_VIGINTILE = 0x2BF3;
const char32_t RUSSIAN_ASTROLOGICAL_SYMBOL_NOVILE = 0x2BF4;
const char32_t RUSSIAN_ASTROLOGICAL_SYMBOL_QUINTILE = 0x2BF5;
const char32_t RUSSIAN_ASTROLOGICAL_SYMBOL_BINOVILE = 0x2BF6;
const char32_t RUSSIAN_ASTROLOGICAL_SYMBOL_SENTAGON = 0x2BF7;
const char32_t RUSSIAN_ASTROLOGICAL_SYMBOL_TREDECILE = 0x2BF8;
const char32_t EQUALS_SIGN_WITH_INFINITY_BELOW = 0x2BF9;
const char32_t UNITED_SYMBOL = 0x2BFA;
const char32_t SEPARATED_SYMBOL = 0x2BFB;
const char32_t DOUBLED_SYMBOL = 0x2BFC;
const char32_t PASSED_SYMBOL = 0x2BFD;
const char32_t REVERSED_RIGHT_ANGLE = 0x2BFE;
const char32_t HELLSCHREIBER_PAUSE_SYMBOL = 0x2BFF;
const char32_t GLAGOLITIC_CAPITAL_LETTER_AZU = 0x2C00;
const char32_t GLAGOLITIC_CAPITAL_LETTER_BUKY = 0x2C01;
const char32_t GLAGOLITIC_CAPITAL_LETTER_VEDE = 0x2C02;
const char32_t GLAGOLITIC_CAPITAL_LETTER_GLAGOLI = 0x2C03;
const char32_t GLAGOLITIC_CAPITAL_LETTER_DOBRO = 0x2C04;
const char32_t GLAGOLITIC_CAPITAL_LETTER_YESTU = 0x2C05;
const char32_t GLAGOLITIC_CAPITAL_LETTER_ZHIVETE = 0x2C06;
const char32_t GLAGOLITIC_CAPITAL_LETTER_DZELO = 0x2C07;
const char32_t GLAGOLITIC_CAPITAL_LETTER_ZEMLJA = 0x2C08;
const char32_t GLAGOLITIC_CAPITAL_LETTER_IZHE = 0x2C09;
const char32_t GLAGOLITIC_CAPITAL_LETTER_INITIAL_IZHE = 0x2C0A;
const char32_t GLAGOLITIC_CAPITAL_LETTER_I = 0x2C0B;
const char32_t GLAGOLITIC_CAPITAL_LETTER_DJERVI = 0x2C0C;
const char32_t GLAGOLITIC_CAPITAL_LETTER_KAKO = 0x2C0D;
const char32_t GLAGOLITIC_CAPITAL_LETTER_LJUDIJE = 0x2C0E;
const char32_t GLAGOLITIC_CAPITAL_LETTER_MYSLITE = 0x2C0F;
const char32_t GLAGOLITIC_CAPITAL_LETTER_NASHI = 0x2C10;
const char32_t GLAGOLITIC_CAPITAL_LETTER_ONU = 0x2C11;
const char32_t GLAGOLITIC_CAPITAL_LETTER_POKOJI = 0x2C12;
const char32_t GLAGOLITIC_CAPITAL_LETTER_RITSI = 0x2C13;
const char32_t GLAGOLITIC_CAPITAL_LETTER_SLOVO = 0x2C14;
const char32_t GLAGOLITIC_CAPITAL_LETTER_TVRIDO = 0x2C15;
const char32_t GLAGOLITIC_CAPITAL_LETTER_UKU = 0x2C16;
const char32_t GLAGOLITIC_CAPITAL_LETTER_FRITU = 0x2C17;
const char32_t GLAGOLITIC_CAPITAL_LETTER_HERU = 0x2C18;
const char32_t GLAGOLITIC_CAPITAL_LETTER_OTU = 0x2C19;
const char32_t GLAGOLITIC_CAPITAL_LETTER_PE = 0x2C1A;
const char32_t GLAGOLITIC_CAPITAL_LETTER_SHTA = 0x2C1B;
const char32_t GLAGOLITIC_CAPITAL_LETTER_TSI = 0x2C1C;
const char32_t GLAGOLITIC_CAPITAL_LETTER_CHRIVI = 0x2C1D;
const char32_t GLAGOLITIC_CAPITAL_LETTER_SHA = 0x2C1E;
const char32_t GLAGOLITIC_CAPITAL_LETTER_YERU = 0x2C1F;
const char32_t GLAGOLITIC_CAPITAL_LETTER_YERI = 0x2C20;
const char32_t GLAGOLITIC_CAPITAL_LETTER_YATI = 0x2C21;
const char32_t GLAGOLITIC_CAPITAL_LETTER_SPIDERY_HA = 0x2C22;
const char32_t GLAGOLITIC_CAPITAL_LETTER_YU = 0x2C23;
const char32_t GLAGOLITIC_CAPITAL_LETTER_SMALL_YUS = 0x2C24;
const char32_t GLAGOLITIC_CAPITAL_LETTER_SMALL_YUS_WITH_TAIL = 0x2C25;
const char32_t GLAGOLITIC_CAPITAL_LETTER_YO = 0x2C26;
const char32_t GLAGOLITIC_CAPITAL_LETTER_IOTATED_SMALL_YUS = 0x2C27;
const char32_t GLAGOLITIC_CAPITAL_LETTER_BIG_YUS = 0x2C28;
const char32_t GLAGOLITIC_CAPITAL_LETTER_IOTATED_BIG_YUS = 0x2C29;
const char32_t GLAGOLITIC_CAPITAL_LETTER_FITA = 0x2C2A;
const char32_t GLAGOLITIC_CAPITAL_LETTER_IZHITSA = 0x2C2B;
const char32_t GLAGOLITIC_CAPITAL_LETTER_SHTAPIC = 0x2C2C;
const char32_t GLAGOLITIC_CAPITAL_LETTER_TROKUTASTI_A = 0x2C2D;
const char32_t GLAGOLITIC_CAPITAL_LETTER_LATINATE_MYSLITE = 0x2C2E;
const char32_t GLAGOLITIC_CAPITAL_LETTER_CAUDATE_CHRIVI = 0x2C2F;
const char32_t GLAGOLITIC_SMALL_LETTER_AZU = 0x2C30;
const char32_t GLAGOLITIC_SMALL_LETTER_BUKY = 0x2C31;
const char32_t GLAGOLITIC_SMALL_LETTER_VEDE = 0x2C32;
const char32_t GLAGOLITIC_SMALL_LETTER_GLAGOLI = 0x2C33;
const char32_t GLAGOLITIC_SMALL_LETTER_DOBRO = 0x2C34;
const char32_t GLAGOLITIC_SMALL_LETTER_YESTU = 0x2C35;
const char32_t GLAGOLITIC_SMALL_LETTER_ZHIVETE = 0x2C36;
const char32_t GLAGOLITIC_SMALL_LETTER_DZELO = 0x2C37;
const char32_t GLAGOLITIC_SMALL_LETTER_ZEMLJA = 0x2C38;
const char32_t GLAGOLITIC_SMALL_LETTER_IZHE = 0x2C39;
const char32_t GLAGOLITIC_SMALL_LETTER_INITIAL_IZHE = 0x2C3A;
const char32_t GLAGOLITIC_SMALL_LETTER_I = 0x2C3B;
const char32_t GLAGOLITIC_SMALL_LETTER_DJERVI = 0x2C3C;
const char32_t GLAGOLITIC_SMALL_LETTER_KAKO = 0x2C3D;
const char32_t GLAGOLITIC_SMALL_LETTER_LJUDIJE = 0x2C3E;
const char32_t GLAGOLITIC_SMALL_LETTER_MYSLITE = 0x2C3F;
const char32_t GLAGOLITIC_SMALL_LETTER_NASHI = 0x2C40;
const char32_t GLAGOLITIC_SMALL_LETTER_ONU = 0x2C41;
const char32_t GLAGOLITIC_SMALL_LETTER_POKOJI = 0x2C42;
const char32_t GLAGOLITIC_SMALL_LETTER_RITSI = 0x2C43;
const char32_t GLAGOLITIC_SMALL_LETTER_SLOVO = 0x2C44;
const char32_t GLAGOLITIC_SMALL_LETTER_TVRIDO = 0x2C45;
const char32_t GLAGOLITIC_SMALL_LETTER_UKU = 0x2C46;
const char32_t GLAGOLITIC_SMALL_LETTER_FRITU = 0x2C47;
const char32_t GLAGOLITIC_SMALL_LETTER_HERU = 0x2C48;
const char32_t GLAGOLITIC_SMALL_LETTER_OTU = 0x2C49;
const char32_t GLAGOLITIC_SMALL_LETTER_PE = 0x2C4A;
const char32_t GLAGOLITIC_SMALL_LETTER_SHTA = 0x2C4B;
const char32_t GLAGOLITIC_SMALL_LETTER_TSI = 0x2C4C;
const char32_t GLAGOLITIC_SMALL_LETTER_CHRIVI = 0x2C4D;
const char32_t GLAGOLITIC_SMALL_LETTER_SHA = 0x2C4E;
const char32_t GLAGOLITIC_SMALL_LETTER_YERU = 0x2C4F;
const char32_t GLAGOLITIC_SMALL_LETTER_YERI = 0x2C50;
const char32_t GLAGOLITIC_SMALL_LETTER_YATI = 0x2C51;
const char32_t GLAGOLITIC_SMALL_LETTER_SPIDERY_HA = 0x2C52;
const char32_t GLAGOLITIC_SMALL_LETTER_YU = 0x2C53;
const char32_t GLAGOLITIC_SMALL_LETTER_SMALL_YUS = 0x2C54;
const char32_t GLAGOLITIC_SMALL_LETTER_SMALL_YUS_WITH_TAIL = 0x2C55;
const char32_t GLAGOLITIC_SMALL_LETTER_YO = 0x2C56;
const char32_t GLAGOLITIC_SMALL_LETTER_IOTATED_SMALL_YUS = 0x2C57;
const char32_t GLAGOLITIC_SMALL_LETTER_BIG_YUS = 0x2C58;
const char32_t GLAGOLITIC_SMALL_LETTER_IOTATED_BIG_YUS = 0x2C59;
const char32_t GLAGOLITIC_SMALL_LETTER_FITA = 0x2C5A;
const char32_t GLAGOLITIC_SMALL_LETTER_IZHITSA = 0x2C5B;
const char32_t GLAGOLITIC_SMALL_LETTER_SHTAPIC = 0x2C5C;
const char32_t GLAGOLITIC_SMALL_LETTER_TROKUTASTI_A = 0x2C5D;
const char32_t GLAGOLITIC_SMALL_LETTER_LATINATE_MYSLITE = 0x2C5E;
const char32_t GLAGOLITIC_SMALL_LETTER_CAUDATE_CHRIVI = 0x2C5F;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_DOUBLE_BAR = 0x2C60;
const char32_t LATIN_SMALL_LETTER_L_WITH_DOUBLE_BAR = 0x2C61;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_MIDDLE_TILDE = 0x2C62;
const char32_t LATIN_CAPITAL_LETTER_P_WITH_STROKE = 0x2C63;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_TAIL = 0x2C64;
const char32_t LATIN_SMALL_LETTER_A_WITH_STROKE = 0x2C65;
const char32_t LATIN_SMALL_LETTER_T_WITH_DIAGONAL_STROKE = 0x2C66;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_DESCENDER = 0x2C67;
const char32_t LATIN_SMALL_LETTER_H_WITH_DESCENDER = 0x2C68;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_DESCENDER = 0x2C69;
const char32_t LATIN_SMALL_LETTER_K_WITH_DESCENDER = 0x2C6A;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_DESCENDER = 0x2C6B;
const char32_t LATIN_SMALL_LETTER_Z_WITH_DESCENDER = 0x2C6C;
const char32_t LATIN_CAPITAL_LETTER_ALPHA = 0x2C6D;
const char32_t LATIN_CAPITAL_LETTER_M_WITH_HOOK = 0x2C6E;
const char32_t LATIN_CAPITAL_LETTER_TURNED_A = 0x2C6F;
const char32_t LATIN_CAPITAL_LETTER_TURNED_ALPHA = 0x2C70;
const char32_t LATIN_SMALL_LETTER_V_WITH_RIGHT_HOOK = 0x2C71;
const char32_t LATIN_CAPITAL_LETTER_W_WITH_HOOK = 0x2C72;
const char32_t LATIN_SMALL_LETTER_W_WITH_HOOK = 0x2C73;
const char32_t LATIN_SMALL_LETTER_V_WITH_CURL = 0x2C74;
const char32_t LATIN_CAPITAL_LETTER_HALF_H = 0x2C75;
const char32_t LATIN_SMALL_LETTER_HALF_H = 0x2C76;
const char32_t LATIN_SMALL_LETTER_TAILLESS_PHI = 0x2C77;
const char32_t LATIN_SMALL_LETTER_E_WITH_NOTCH = 0x2C78;
const char32_t LATIN_SMALL_LETTER_TURNED_R_WITH_TAIL = 0x2C79;
const char32_t LATIN_SMALL_LETTER_O_WITH_LOW_RING_INSIDE = 0x2C7A;
const char32_t LATIN_LETTER_SMALL_CAPITAL_TURNED_E = 0x2C7B;
const char32_t LATIN_SUBSCRIPT_SMALL_LETTER_J = 0x2C7C;
const char32_t MODIFIER_LETTER_CAPITAL_V = 0x2C7D;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_SWASH_TAIL = 0x2C7E;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_SWASH_TAIL = 0x2C7F;
const char32_t COPTIC_CAPITAL_LETTER_ALFA = 0x2C80;
const char32_t COPTIC_SMALL_LETTER_ALFA = 0x2C81;
const char32_t COPTIC_CAPITAL_LETTER_VIDA = 0x2C82;
const char32_t COPTIC_SMALL_LETTER_VIDA = 0x2C83;
const char32_t COPTIC_CAPITAL_LETTER_GAMMA = 0x2C84;
const char32_t COPTIC_SMALL_LETTER_GAMMA = 0x2C85;
const char32_t COPTIC_CAPITAL_LETTER_DALDA = 0x2C86;
const char32_t COPTIC_SMALL_LETTER_DALDA = 0x2C87;
const char32_t COPTIC_CAPITAL_LETTER_EIE = 0x2C88;
const char32_t COPTIC_SMALL_LETTER_EIE = 0x2C89;
const char32_t COPTIC_CAPITAL_LETTER_SOU = 0x2C8A;
const char32_t COPTIC_SMALL_LETTER_SOU = 0x2C8B;
const char32_t COPTIC_CAPITAL_LETTER_ZATA = 0x2C8C;
const char32_t COPTIC_SMALL_LETTER_ZATA = 0x2C8D;
const char32_t COPTIC_CAPITAL_LETTER_HATE = 0x2C8E;
const char32_t COPTIC_SMALL_LETTER_HATE = 0x2C8F;
const char32_t COPTIC_CAPITAL_LETTER_THETHE = 0x2C90;
const char32_t COPTIC_SMALL_LETTER_THETHE = 0x2C91;
const char32_t COPTIC_CAPITAL_LETTER_IAUDA = 0x2C92;
const char32_t COPTIC_SMALL_LETTER_IAUDA = 0x2C93;
const char32_t COPTIC_CAPITAL_LETTER_KAPA = 0x2C94;
const char32_t COPTIC_SMALL_LETTER_KAPA = 0x2C95;
const char32_t COPTIC_CAPITAL_LETTER_LAULA = 0x2C96;
const char32_t COPTIC_SMALL_LETTER_LAULA = 0x2C97;
const char32_t COPTIC_CAPITAL_LETTER_MI = 0x2C98;
const char32_t COPTIC_SMALL_LETTER_MI = 0x2C99;
const char32_t COPTIC_CAPITAL_LETTER_NI = 0x2C9A;
const char32_t COPTIC_SMALL_LETTER_NI = 0x2C9B;
const char32_t COPTIC_CAPITAL_LETTER_KSI = 0x2C9C;
const char32_t COPTIC_SMALL_LETTER_KSI = 0x2C9D;
const char32_t COPTIC_CAPITAL_LETTER_O = 0x2C9E;
const char32_t COPTIC_SMALL_LETTER_O = 0x2C9F;
const char32_t COPTIC_CAPITAL_LETTER_PI = 0x2CA0;
const char32_t COPTIC_SMALL_LETTER_PI = 0x2CA1;
const char32_t COPTIC_CAPITAL_LETTER_RO = 0x2CA2;
const char32_t COPTIC_SMALL_LETTER_RO = 0x2CA3;
const char32_t COPTIC_CAPITAL_LETTER_SIMA = 0x2CA4;
const char32_t COPTIC_SMALL_LETTER_SIMA = 0x2CA5;
const char32_t COPTIC_CAPITAL_LETTER_TAU = 0x2CA6;
const char32_t COPTIC_SMALL_LETTER_TAU = 0x2CA7;
const char32_t COPTIC_CAPITAL_LETTER_UA = 0x2CA8;
const char32_t COPTIC_SMALL_LETTER_UA = 0x2CA9;
const char32_t COPTIC_CAPITAL_LETTER_FI = 0x2CAA;
const char32_t COPTIC_SMALL_LETTER_FI = 0x2CAB;
const char32_t COPTIC_CAPITAL_LETTER_KHI = 0x2CAC;
const char32_t COPTIC_SMALL_LETTER_KHI = 0x2CAD;
const char32_t COPTIC_CAPITAL_LETTER_PSI = 0x2CAE;
const char32_t COPTIC_SMALL_LETTER_PSI = 0x2CAF;
const char32_t COPTIC_CAPITAL_LETTER_OOU = 0x2CB0;
const char32_t COPTIC_SMALL_LETTER_OOU = 0x2CB1;
const char32_t COPTIC_CAPITAL_LETTER_DIALECT_P_ALEF = 0x2CB2;
const char32_t COPTIC_SMALL_LETTER_DIALECT_P_ALEF = 0x2CB3;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_AIN = 0x2CB4;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_AIN = 0x2CB5;
const char32_t COPTIC_CAPITAL_LETTER_CRYPTOGRAMMIC_EIE = 0x2CB6;
const char32_t COPTIC_SMALL_LETTER_CRYPTOGRAMMIC_EIE = 0x2CB7;
const char32_t COPTIC_CAPITAL_LETTER_DIALECT_P_KAPA = 0x2CB8;
const char32_t COPTIC_SMALL_LETTER_DIALECT_P_KAPA = 0x2CB9;
const char32_t COPTIC_CAPITAL_LETTER_DIALECT_P_NI = 0x2CBA;
const char32_t COPTIC_SMALL_LETTER_DIALECT_P_NI = 0x2CBB;
const char32_t COPTIC_CAPITAL_LETTER_CRYPTOGRAMMIC_NI = 0x2CBC;
const char32_t COPTIC_SMALL_LETTER_CRYPTOGRAMMIC_NI = 0x2CBD;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_OOU = 0x2CBE;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_OOU = 0x2CBF;
const char32_t COPTIC_CAPITAL_LETTER_SAMPI = 0x2CC0;
const char32_t COPTIC_SMALL_LETTER_SAMPI = 0x2CC1;
const char32_t COPTIC_CAPITAL_LETTER_CROSSED_SHEI = 0x2CC2;
const char32_t COPTIC_SMALL_LETTER_CROSSED_SHEI = 0x2CC3;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_SHEI = 0x2CC4;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_SHEI = 0x2CC5;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_ESH = 0x2CC6;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_ESH = 0x2CC7;
const char32_t COPTIC_CAPITAL_LETTER_AKHMIMIC_KHEI = 0x2CC8;
const char32_t COPTIC_SMALL_LETTER_AKHMIMIC_KHEI = 0x2CC9;
const char32_t COPTIC_CAPITAL_LETTER_DIALECT_P_HORI = 0x2CCA;
const char32_t COPTIC_SMALL_LETTER_DIALECT_P_HORI = 0x2CCB;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_HORI = 0x2CCC;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_HORI = 0x2CCD;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_HA = 0x2CCE;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_HA = 0x2CCF;
const char32_t COPTIC_CAPITAL_LETTER_L_SHAPED_HA = 0x2CD0;
const char32_t COPTIC_SMALL_LETTER_L_SHAPED_HA = 0x2CD1;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_HEI = 0x2CD2;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_HEI = 0x2CD3;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_HAT = 0x2CD4;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_HAT = 0x2CD5;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_GANGIA = 0x2CD6;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_GANGIA = 0x2CD7;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_DJA = 0x2CD8;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_DJA = 0x2CD9;
const char32_t COPTIC_CAPITAL_LETTER_OLD_COPTIC_SHIMA = 0x2CDA;
const char32_t COPTIC_SMALL_LETTER_OLD_COPTIC_SHIMA = 0x2CDB;
const char32_t COPTIC_CAPITAL_LETTER_OLD_NUBIAN_SHIMA = 0x2CDC;
const char32_t COPTIC_SMALL_LETTER_OLD_NUBIAN_SHIMA = 0x2CDD;
const char32_t COPTIC_CAPITAL_LETTER_OLD_NUBIAN_NGI = 0x2CDE;
const char32_t COPTIC_SMALL_LETTER_OLD_NUBIAN_NGI = 0x2CDF;
const char32_t COPTIC_CAPITAL_LETTER_OLD_NUBIAN_NYI = 0x2CE0;
const char32_t COPTIC_SMALL_LETTER_OLD_NUBIAN_NYI = 0x2CE1;
const char32_t COPTIC_CAPITAL_LETTER_OLD_NUBIAN_WAU = 0x2CE2;
const char32_t COPTIC_SMALL_LETTER_OLD_NUBIAN_WAU = 0x2CE3;
const char32_t COPTIC_SYMBOL_KAI = 0x2CE4;
const char32_t COPTIC_SYMBOL_MI_RO = 0x2CE5;
const char32_t COPTIC_SYMBOL_PI_RO = 0x2CE6;
const char32_t COPTIC_SYMBOL_STAUROS = 0x2CE7;
const char32_t COPTIC_SYMBOL_TAU_RO = 0x2CE8;
const char32_t COPTIC_SYMBOL_KHI_RO = 0x2CE9;
const char32_t COPTIC_SYMBOL_SHIMA_SIMA = 0x2CEA;
const char32_t COPTIC_CAPITAL_LETTER_CRYPTOGRAMMIC_SHEI = 0x2CEB;
const char32_t COPTIC_SMALL_LETTER_CRYPTOGRAMMIC_SHEI = 0x2CEC;
const char32_t COPTIC_CAPITAL_LETTER_CRYPTOGRAMMIC_GANGIA = 0x2CED;
const char32_t COPTIC_SMALL_LETTER_CRYPTOGRAMMIC_GANGIA = 0x2CEE;
const char32_t COPTIC_COMBINING_NI_ABOVE = 0x2CEF;
const char32_t COPTIC_COMBINING_SPIRITUS_ASPER = 0x2CF0;
const char32_t COPTIC_COMBINING_SPIRITUS_LENIS = 0x2CF1;
const char32_t COPTIC_CAPITAL_LETTER_BOHAIRIC_KHEI = 0x2CF2;
const char32_t COPTIC_SMALL_LETTER_BOHAIRIC_KHEI = 0x2CF3;
const char32_t COPTIC_OLD_NUBIAN_FULL_STOP = 0x2CF9;
const char32_t COPTIC_OLD_NUBIAN_DIRECT_QUESTION_MARK = 0x2CFA;
const char32_t COPTIC_OLD_NUBIAN_INDIRECT_QUESTION_MARK = 0x2CFB;
const char32_t COPTIC_OLD_NUBIAN_VERSE_DIVIDER = 0x2CFC;
const char32_t COPTIC_FRACTION_ONE_HALF = 0x2CFD;
const char32_t COPTIC_FULL_STOP = 0x2CFE;
const char32_t COPTIC_MORPHOLOGICAL_DIVIDER = 0x2CFF;
const char32_t GEORGIAN_SMALL_LETTER_AN = 0x2D00;
const char32_t GEORGIAN_SMALL_LETTER_BAN = 0x2D01;
const char32_t GEORGIAN_SMALL_LETTER_GAN = 0x2D02;
const char32_t GEORGIAN_SMALL_LETTER_DON = 0x2D03;
const char32_t GEORGIAN_SMALL_LETTER_EN = 0x2D04;
const char32_t GEORGIAN_SMALL_LETTER_VIN = 0x2D05;
const char32_t GEORGIAN_SMALL_LETTER_ZEN = 0x2D06;
const char32_t GEORGIAN_SMALL_LETTER_TAN = 0x2D07;
const char32_t GEORGIAN_SMALL_LETTER_IN = 0x2D08;
const char32_t GEORGIAN_SMALL_LETTER_KAN = 0x2D09;
const char32_t GEORGIAN_SMALL_LETTER_LAS = 0x2D0A;
const char32_t GEORGIAN_SMALL_LETTER_MAN = 0x2D0B;
const char32_t GEORGIAN_SMALL_LETTER_NAR = 0x2D0C;
const char32_t GEORGIAN_SMALL_LETTER_ON = 0x2D0D;
const char32_t GEORGIAN_SMALL_LETTER_PAR = 0x2D0E;
const char32_t GEORGIAN_SMALL_LETTER_ZHAR = 0x2D0F;
const char32_t GEORGIAN_SMALL_LETTER_RAE = 0x2D10;
const char32_t GEORGIAN_SMALL_LETTER_SAN = 0x2D11;
const char32_t GEORGIAN_SMALL_LETTER_TAR = 0x2D12;
const char32_t GEORGIAN_SMALL_LETTER_UN = 0x2D13;
const char32_t GEORGIAN_SMALL_LETTER_PHAR = 0x2D14;
const char32_t GEORGIAN_SMALL_LETTER_KHAR = 0x2D15;
const char32_t GEORGIAN_SMALL_LETTER_GHAN = 0x2D16;
const char32_t GEORGIAN_SMALL_LETTER_QAR = 0x2D17;
const char32_t GEORGIAN_SMALL_LETTER_SHIN = 0x2D18;
const char32_t GEORGIAN_SMALL_LETTER_CHIN = 0x2D19;
const char32_t GEORGIAN_SMALL_LETTER_CAN = 0x2D1A;
const char32_t GEORGIAN_SMALL_LETTER_JIL = 0x2D1B;
const char32_t GEORGIAN_SMALL_LETTER_CIL = 0x2D1C;
const char32_t GEORGIAN_SMALL_LETTER_CHAR = 0x2D1D;
const char32_t GEORGIAN_SMALL_LETTER_XAN = 0x2D1E;
const char32_t GEORGIAN_SMALL_LETTER_JHAN = 0x2D1F;
const char32_t GEORGIAN_SMALL_LETTER_HAE = 0x2D20;
const char32_t GEORGIAN_SMALL_LETTER_HE = 0x2D21;
const char32_t GEORGIAN_SMALL_LETTER_HIE = 0x2D22;
const char32_t GEORGIAN_SMALL_LETTER_WE = 0x2D23;
const char32_t GEORGIAN_SMALL_LETTER_HAR = 0x2D24;
const char32_t GEORGIAN_SMALL_LETTER_HOE = 0x2D25;
const char32_t GEORGIAN_SMALL_LETTER_YN = 0x2D27;
const char32_t GEORGIAN_SMALL_LETTER_AEN = 0x2D2D;
const char32_t TIFINAGH_LETTER_YA = 0x2D30;
const char32_t TIFINAGH_LETTER_YAB = 0x2D31;
const char32_t TIFINAGH_LETTER_YABH = 0x2D32;
const char32_t TIFINAGH_LETTER_YAG = 0x2D33;
const char32_t TIFINAGH_LETTER_YAGHH = 0x2D34;
const char32_t TIFINAGH_LETTER_BERBER_ACADEMY_YAJ = 0x2D35;
const char32_t TIFINAGH_LETTER_YAJ = 0x2D36;
const char32_t TIFINAGH_LETTER_YAD = 0x2D37;
const char32_t TIFINAGH_LETTER_YADH = 0x2D38;
const char32_t TIFINAGH_LETTER_YADD = 0x2D39;
const char32_t TIFINAGH_LETTER_YADDH = 0x2D3A;
const char32_t TIFINAGH_LETTER_YEY = 0x2D3B;
const char32_t TIFINAGH_LETTER_YAF = 0x2D3C;
const char32_t TIFINAGH_LETTER_YAK = 0x2D3D;
const char32_t TIFINAGH_LETTER_TUAREG_YAK = 0x2D3E;
const char32_t TIFINAGH_LETTER_YAKHH = 0x2D3F;
const char32_t TIFINAGH_LETTER_YAH = 0x2D40;
const char32_t TIFINAGH_LETTER_BERBER_ACADEMY_YAH = 0x2D41;
const char32_t TIFINAGH_LETTER_TUAREG_YAH = 0x2D42;
const char32_t TIFINAGH_LETTER_YAHH = 0x2D43;
const char32_t TIFINAGH_LETTER_YAA = 0x2D44;
const char32_t TIFINAGH_LETTER_YAKH = 0x2D45;
const char32_t TIFINAGH_LETTER_TUAREG_YAKH = 0x2D46;
const char32_t TIFINAGH_LETTER_YAQ = 0x2D47;
const char32_t TIFINAGH_LETTER_TUAREG_YAQ = 0x2D48;
const char32_t TIFINAGH_LETTER_YI = 0x2D49;
const char32_t TIFINAGH_LETTER_YAZH = 0x2D4A;
const char32_t TIFINAGH_LETTER_AHAGGAR_YAZH = 0x2D4B;
const char32_t TIFINAGH_LETTER_TUAREG_YAZH = 0x2D4C;
const char32_t TIFINAGH_LETTER_YAL = 0x2D4D;
const char32_t TIFINAGH_LETTER_YAM = 0x2D4E;
const char32_t TIFINAGH_LETTER_YAN = 0x2D4F;
const char32_t TIFINAGH_LETTER_TUAREG_YAGN = 0x2D50;
const char32_t TIFINAGH_LETTER_TUAREG_YANG = 0x2D51;
const char32_t TIFINAGH_LETTER_YAP = 0x2D52;
const char32_t TIFINAGH_LETTER_YU = 0x2D53;
const char32_t TIFINAGH_LETTER_YAR = 0x2D54;
const char32_t TIFINAGH_LETTER_YARR = 0x2D55;
const char32_t TIFINAGH_LETTER_YAGH = 0x2D56;
const char32_t TIFINAGH_LETTER_TUAREG_YAGH = 0x2D57;
const char32_t TIFINAGH_LETTER_AYER_YAGH = 0x2D58;
const char32_t TIFINAGH_LETTER_YAS = 0x2D59;
const char32_t TIFINAGH_LETTER_YASS = 0x2D5A;
const char32_t TIFINAGH_LETTER_YASH = 0x2D5B;
const char32_t TIFINAGH_LETTER_YAT = 0x2D5C;
const char32_t TIFINAGH_LETTER_YATH = 0x2D5D;
const char32_t TIFINAGH_LETTER_YACH = 0x2D5E;
const char32_t TIFINAGH_LETTER_YATT = 0x2D5F;
const char32_t TIFINAGH_LETTER_YAV = 0x2D60;
const char32_t TIFINAGH_LETTER_YAW = 0x2D61;
const char32_t TIFINAGH_LETTER_YAY = 0x2D62;
const char32_t TIFINAGH_LETTER_YAZ = 0x2D63;
const char32_t TIFINAGH_LETTER_TAWELLEMET_YAZ = 0x2D64;
const char32_t TIFINAGH_LETTER_YAZZ = 0x2D65;
const char32_t TIFINAGH_LETTER_YE = 0x2D66;
const char32_t TIFINAGH_LETTER_YO = 0x2D67;
const char32_t TIFINAGH_MODIFIER_LETTER_LABIALIZATION_MARK = 0x2D6F;
const char32_t TIFINAGH_SEPARATOR_MARK = 0x2D70;
const char32_t TIFINAGH_CONSONANT_JOINER = 0x2D7F;
const char32_t ETHIOPIC_SYLLABLE_LOA = 0x2D80;
const char32_t ETHIOPIC_SYLLABLE_MOA = 0x2D81;
const char32_t ETHIOPIC_SYLLABLE_ROA = 0x2D82;
const char32_t ETHIOPIC_SYLLABLE_SOA = 0x2D83;
const char32_t ETHIOPIC_SYLLABLE_SHOA = 0x2D84;
const char32_t ETHIOPIC_SYLLABLE_BOA = 0x2D85;
const char32_t ETHIOPIC_SYLLABLE_TOA = 0x2D86;
const char32_t ETHIOPIC_SYLLABLE_COA = 0x2D87;
const char32_t ETHIOPIC_SYLLABLE_NOA = 0x2D88;
const char32_t ETHIOPIC_SYLLABLE_NYOA = 0x2D89;
const char32_t ETHIOPIC_SYLLABLE_GLOTTAL_OA = 0x2D8A;
const char32_t ETHIOPIC_SYLLABLE_ZOA = 0x2D8B;
const char32_t ETHIOPIC_SYLLABLE_DOA = 0x2D8C;
const char32_t ETHIOPIC_SYLLABLE_DDOA = 0x2D8D;
const char32_t ETHIOPIC_SYLLABLE_JOA = 0x2D8E;
const char32_t ETHIOPIC_SYLLABLE_THOA = 0x2D8F;
const char32_t ETHIOPIC_SYLLABLE_CHOA = 0x2D90;
const char32_t ETHIOPIC_SYLLABLE_PHOA = 0x2D91;
const char32_t ETHIOPIC_SYLLABLE_POA = 0x2D92;
const char32_t ETHIOPIC_SYLLABLE_GGWA = 0x2D93;
const char32_t ETHIOPIC_SYLLABLE_GGWI = 0x2D94;
const char32_t ETHIOPIC_SYLLABLE_GGWEE = 0x2D95;
const char32_t ETHIOPIC_SYLLABLE_GGWE = 0x2D96;
const char32_t ETHIOPIC_SYLLABLE_SSA = 0x2DA0;
const char32_t ETHIOPIC_SYLLABLE_SSU = 0x2DA1;
const char32_t ETHIOPIC_SYLLABLE_SSI = 0x2DA2;
const char32_t ETHIOPIC_SYLLABLE_SSAA = 0x2DA3;
const char32_t ETHIOPIC_SYLLABLE_SSEE = 0x2DA4;
const char32_t ETHIOPIC_SYLLABLE_SSE = 0x2DA5;
const char32_t ETHIOPIC_SYLLABLE_SSO = 0x2DA6;
const char32_t ETHIOPIC_SYLLABLE_CCA = 0x2DA8;
const char32_t ETHIOPIC_SYLLABLE_CCU = 0x2DA9;
const char32_t ETHIOPIC_SYLLABLE_CCI = 0x2DAA;
const char32_t ETHIOPIC_SYLLABLE_CCAA = 0x2DAB;
const char32_t ETHIOPIC_SYLLABLE_CCEE = 0x2DAC;
const char32_t ETHIOPIC_SYLLABLE_CCE = 0x2DAD;
const char32_t ETHIOPIC_SYLLABLE_CCO = 0x2DAE;
const char32_t ETHIOPIC_SYLLABLE_ZZA = 0x2DB0;
const char32_t ETHIOPIC_SYLLABLE_ZZU = 0x2DB1;
const char32_t ETHIOPIC_SYLLABLE_ZZI = 0x2DB2;
const char32_t ETHIOPIC_SYLLABLE_ZZAA = 0x2DB3;
const char32_t ETHIOPIC_SYLLABLE_ZZEE = 0x2DB4;
const char32_t ETHIOPIC_SYLLABLE_ZZE = 0x2DB5;
const char32_t ETHIOPIC_SYLLABLE_ZZO = 0x2DB6;
const char32_t ETHIOPIC_SYLLABLE_CCHA = 0x2DB8;
const char32_t ETHIOPIC_SYLLABLE_CCHU = 0x2DB9;
const char32_t ETHIOPIC_SYLLABLE_CCHI = 0x2DBA;
const char32_t ETHIOPIC_SYLLABLE_CCHAA = 0x2DBB;
const char32_t ETHIOPIC_SYLLABLE_CCHEE = 0x2DBC;
const char32_t ETHIOPIC_SYLLABLE_CCHE = 0x2DBD;
const char32_t ETHIOPIC_SYLLABLE_CCHO = 0x2DBE;
const char32_t ETHIOPIC_SYLLABLE_QYA = 0x2DC0;
const char32_t ETHIOPIC_SYLLABLE_QYU = 0x2DC1;
const char32_t ETHIOPIC_SYLLABLE_QYI = 0x2DC2;
const char32_t ETHIOPIC_SYLLABLE_QYAA = 0x2DC3;
const char32_t ETHIOPIC_SYLLABLE_QYEE = 0x2DC4;
const char32_t ETHIOPIC_SYLLABLE_QYE = 0x2DC5;
const char32_t ETHIOPIC_SYLLABLE_QYO = 0x2DC6;
const char32_t ETHIOPIC_SYLLABLE_KYA = 0x2DC8;
const char32_t ETHIOPIC_SYLLABLE_KYU = 0x2DC9;
const char32_t ETHIOPIC_SYLLABLE_KYI = 0x2DCA;
const char32_t ETHIOPIC_SYLLABLE_KYAA = 0x2DCB;
const char32_t ETHIOPIC_SYLLABLE_KYEE = 0x2DCC;
const char32_t ETHIOPIC_SYLLABLE_KYE = 0x2DCD;
const char32_t ETHIOPIC_SYLLABLE_KYO = 0x2DCE;
const char32_t ETHIOPIC_SYLLABLE_XYA = 0x2DD0;
const char32_t ETHIOPIC_SYLLABLE_XYU = 0x2DD1;
const char32_t ETHIOPIC_SYLLABLE_XYI = 0x2DD2;
const char32_t ETHIOPIC_SYLLABLE_XYAA = 0x2DD3;
const char32_t ETHIOPIC_SYLLABLE_XYEE = 0x2DD4;
const char32_t ETHIOPIC_SYLLABLE_XYE = 0x2DD5;
const char32_t ETHIOPIC_SYLLABLE_XYO = 0x2DD6;
const char32_t ETHIOPIC_SYLLABLE_GYA = 0x2DD8;
const char32_t ETHIOPIC_SYLLABLE_GYU = 0x2DD9;
const char32_t ETHIOPIC_SYLLABLE_GYI = 0x2DDA;
const char32_t ETHIOPIC_SYLLABLE_GYAA = 0x2DDB;
const char32_t ETHIOPIC_SYLLABLE_GYEE = 0x2DDC;
const char32_t ETHIOPIC_SYLLABLE_GYE = 0x2DDD;
const char32_t ETHIOPIC_SYLLABLE_GYO = 0x2DDE;
const char32_t COMBINING_CYRILLIC_LETTER_BE = 0x2DE0;
const char32_t COMBINING_CYRILLIC_LETTER_VE = 0x2DE1;
const char32_t COMBINING_CYRILLIC_LETTER_GHE = 0x2DE2;
const char32_t COMBINING_CYRILLIC_LETTER_DE = 0x2DE3;
const char32_t COMBINING_CYRILLIC_LETTER_ZHE = 0x2DE4;
const char32_t COMBINING_CYRILLIC_LETTER_ZE = 0x2DE5;
const char32_t COMBINING_CYRILLIC_LETTER_KA = 0x2DE6;
const char32_t COMBINING_CYRILLIC_LETTER_EL = 0x2DE7;
const char32_t COMBINING_CYRILLIC_LETTER_EM = 0x2DE8;
const char32_t COMBINING_CYRILLIC_LETTER_EN = 0x2DE9;
const char32_t COMBINING_CYRILLIC_LETTER_O = 0x2DEA;
const char32_t COMBINING_CYRILLIC_LETTER_PE = 0x2DEB;
const char32_t COMBINING_CYRILLIC_LETTER_ER = 0x2DEC;
const char32_t COMBINING_CYRILLIC_LETTER_ES = 0x2DED;
const char32_t COMBINING_CYRILLIC_LETTER_TE = 0x2DEE;
const char32_t COMBINING_CYRILLIC_LETTER_HA = 0x2DEF;
const char32_t COMBINING_CYRILLIC_LETTER_TSE = 0x2DF0;
const char32_t COMBINING_CYRILLIC_LETTER_CHE = 0x2DF1;
const char32_t COMBINING_CYRILLIC_LETTER_SHA = 0x2DF2;
const char32_t COMBINING_CYRILLIC_LETTER_SHCHA = 0x2DF3;
const char32_t COMBINING_CYRILLIC_LETTER_FITA = 0x2DF4;
const char32_t COMBINING_CYRILLIC_LETTER_ES_TE = 0x2DF5;
const char32_t COMBINING_CYRILLIC_LETTER_A = 0x2DF6;
const char32_t COMBINING_CYRILLIC_LETTER_IE = 0x2DF7;
const char32_t COMBINING_CYRILLIC_LETTER_DJERV = 0x2DF8;
const char32_t COMBINING_CYRILLIC_LETTER_MONOGRAPH_UK = 0x2DF9;
const char32_t COMBINING_CYRILLIC_LETTER_YAT = 0x2DFA;
const char32_t COMBINING_CYRILLIC_LETTER_YU = 0x2DFB;
const char32_t COMBINING_CYRILLIC_LETTER_IOTIFIED_A = 0x2DFC;
const char32_t COMBINING_CYRILLIC_LETTER_LITTLE_YUS = 0x2DFD;
const char32_t COMBINING_CYRILLIC_LETTER_BIG_YUS = 0x2DFE;
const char32_t COMBINING_CYRILLIC_LETTER_IOTIFIED_BIG_YUS = 0x2DFF;
const char32_t RIGHT_ANGLE_SUBSTITUTION_MARKER = 0x2E00;
const char32_t RIGHT_ANGLE_DOTTED_SUBSTITUTION_MARKER = 0x2E01;
const char32_t LEFT_SUBSTITUTION_BRACKET = 0x2E02;
const char32_t RIGHT_SUBSTITUTION_BRACKET = 0x2E03;
const char32_t LEFT_DOTTED_SUBSTITUTION_BRACKET = 0x2E04;
const char32_t RIGHT_DOTTED_SUBSTITUTION_BRACKET = 0x2E05;
const char32_t RAISED_INTERPOLATION_MARKER = 0x2E06;
const char32_t RAISED_DOTTED_INTERPOLATION_MARKER = 0x2E07;
const char32_t DOTTED_TRANSPOSITION_MARKER = 0x2E08;
const char32_t LEFT_TRANSPOSITION_BRACKET = 0x2E09;
const char32_t RIGHT_TRANSPOSITION_BRACKET = 0x2E0A;
const char32_t RAISED_SQUARE = 0x2E0B;
const char32_t LEFT_RAISED_OMISSION_BRACKET = 0x2E0C;
const char32_t RIGHT_RAISED_OMISSION_BRACKET = 0x2E0D;
const char32_t EDITORIAL_CORONIS = 0x2E0E;
const char32_t PARAGRAPHOS = 0x2E0F;
const char32_t FORKED_PARAGRAPHOS = 0x2E10;
const char32_t REVERSED_FORKED_PARAGRAPHOS = 0x2E11;
const char32_t HYPODIASTOLE = 0x2E12;
const char32_t DOTTED_OBELOS = 0x2E13;
const char32_t DOWNWARDS_ANCORA = 0x2E14;
const char32_t UPWARDS_ANCORA = 0x2E15;
const char32_t DOTTED_RIGHT_POINTING_ANGLE = 0x2E16;
const char32_t DOUBLE_OBLIQUE_HYPHEN = 0x2E17;
const char32_t INVERTED_INTERROBANG = 0x2E18;
const char32_t PALM_BRANCH = 0x2E19;
const char32_t HYPHEN_WITH_DIAERESIS = 0x2E1A;
const char32_t TILDE_WITH_RING_ABOVE = 0x2E1B;
const char32_t LEFT_LOW_PARAPHRASE_BRACKET = 0x2E1C;
const char32_t RIGHT_LOW_PARAPHRASE_BRACKET = 0x2E1D;
const char32_t TILDE_WITH_DOT_ABOVE = 0x2E1E;
const char32_t TILDE_WITH_DOT_BELOW = 0x2E1F;
const char32_t LEFT_VERTICAL_BAR_WITH_QUILL = 0x2E20;
const char32_t RIGHT_VERTICAL_BAR_WITH_QUILL = 0x2E21;
const char32_t TOP_LEFT_HALF_BRACKET = 0x2E22;
const char32_t TOP_RIGHT_HALF_BRACKET = 0x2E23;
const char32_t BOTTOM_LEFT_HALF_BRACKET = 0x2E24;
const char32_t BOTTOM_RIGHT_HALF_BRACKET = 0x2E25;
const char32_t LEFT_SIDEWAYS_U_BRACKET = 0x2E26;
const char32_t RIGHT_SIDEWAYS_U_BRACKET = 0x2E27;
const char32_t LEFT_DOUBLE_PARENTHESIS = 0x2E28;
const char32_t RIGHT_DOUBLE_PARENTHESIS = 0x2E29;
const char32_t TWO_DOTS_OVER_ONE_DOT_PUNCTUATION = 0x2E2A;
const char32_t ONE_DOT_OVER_TWO_DOTS_PUNCTUATION = 0x2E2B;
const char32_t SQUARED_FOUR_DOT_PUNCTUATION = 0x2E2C;
const char32_t FIVE_DOT_MARK = 0x2E2D;
const char32_t REVERSED_QUESTION_MARK = 0x2E2E;
const char32_t VERTICAL_TILDE = 0x2E2F;
const char32_t RING_POINT = 0x2E30;
const char32_t WORD_SEPARATOR_MIDDLE_DOT = 0x2E31;
const char32_t TURNED_COMMA = 0x2E32;
const char32_t RAISED_DOT = 0x2E33;
const char32_t RAISED_COMMA = 0x2E34;
const char32_t TURNED_SEMICOLON = 0x2E35;
const char32_t DAGGER_WITH_LEFT_GUARD = 0x2E36;
const char32_t DAGGER_WITH_RIGHT_GUARD = 0x2E37;
const char32_t TURNED_DAGGER = 0x2E38;
const char32_t TOP_HALF_SECTION_SIGN = 0x2E39;
const char32_t TWO_EM_DASH = 0x2E3A;
const char32_t THREE_EM_DASH = 0x2E3B;
const char32_t STENOGRAPHIC_FULL_STOP = 0x2E3C;
const char32_t VERTICAL_SIX_DOTS = 0x2E3D;
const char32_t WIGGLY_VERTICAL_LINE = 0x2E3E;
const char32_t CAPITULUM = 0x2E3F;
const char32_t DOUBLE_HYPHEN = 0x2E40;
const char32_t REVERSED_COMMA = 0x2E41;
const char32_t DOUBLE_LOW_REVERSED_9_QUOTATION_MARK = 0x2E42;
const char32_t DASH_WITH_LEFT_UPTURN = 0x2E43;
const char32_t DOUBLE_SUSPENSION_MARK = 0x2E44;
const char32_t INVERTED_LOW_KAVYKA = 0x2E45;
const char32_t INVERTED_LOW_KAVYKA_WITH_KAVYKA_ABOVE = 0x2E46;
const char32_t LOW_KAVYKA = 0x2E47;
const char32_t LOW_KAVYKA_WITH_DOT = 0x2E48;
const char32_t DOUBLE_STACKED_COMMA = 0x2E49;
const char32_t DOTTED_SOLIDUS = 0x2E4A;
const char32_t TRIPLE_DAGGER = 0x2E4B;
const char32_t MEDIEVAL_COMMA = 0x2E4C;
const char32_t PARAGRAPHUS_MARK = 0x2E4D;
const char32_t PUNCTUS_ELEVATUS_MARK = 0x2E4E;
const char32_t CORNISH_VERSE_DIVIDER = 0x2E4F;
const char32_t CROSS_PATTY_WITH_RIGHT_CROSSBAR = 0x2E50;
const char32_t CROSS_PATTY_WITH_LEFT_CROSSBAR = 0x2E51;
const char32_t TIRONIAN_SIGN_CAPITAL_ET = 0x2E52;
const char32_t MEDIEVAL_EXCLAMATION_MARK = 0x2E53;
const char32_t MEDIEVAL_QUESTION_MARK = 0x2E54;
const char32_t LEFT_SQUARE_BRACKET_WITH_STROKE = 0x2E55;
const char32_t RIGHT_SQUARE_BRACKET_WITH_STROKE = 0x2E56;
const char32_t LEFT_SQUARE_BRACKET_WITH_DOUBLE_STROKE = 0x2E57;
const char32_t RIGHT_SQUARE_BRACKET_WITH_DOUBLE_STROKE = 0x2E58;
const char32_t TOP_HALF_LEFT_PARENTHESIS = 0x2E59;
const char32_t TOP_HALF_RIGHT_PARENTHESIS = 0x2E5A;
const char32_t BOTTOM_HALF_LEFT_PARENTHESIS = 0x2E5B;
const char32_t BOTTOM_HALF_RIGHT_PARENTHESIS = 0x2E5C;
const char32_t OBLIQUE_HYPHEN = 0x2E5D;
const char32_t CJK_RADICAL_REPEAT = 0x2E80;
const char32_t CJK_RADICAL_CLIFF = 0x2E81;
const char32_t CJK_RADICAL_SECOND_ONE = 0x2E82;
const char32_t CJK_RADICAL_SECOND_TWO = 0x2E83;
const char32_t CJK_RADICAL_SECOND_THREE = 0x2E84;
const char32_t CJK_RADICAL_PERSON = 0x2E85;
const char32_t CJK_RADICAL_BOX = 0x2E86;
const char32_t CJK_RADICAL_TABLE = 0x2E87;
const char32_t CJK_RADICAL_KNIFE_ONE = 0x2E88;
const char32_t CJK_RADICAL_KNIFE_TWO = 0x2E89;
const char32_t CJK_RADICAL_DIVINATION = 0x2E8A;
const char32_t CJK_RADICAL_SEAL = 0x2E8B;
const char32_t CJK_RADICAL_SMALL_ONE = 0x2E8C;
const char32_t CJK_RADICAL_SMALL_TWO = 0x2E8D;
const char32_t CJK_RADICAL_LAME_ONE = 0x2E8E;
const char32_t CJK_RADICAL_LAME_TWO = 0x2E8F;
const char32_t CJK_RADICAL_LAME_THREE = 0x2E90;
const char32_t CJK_RADICAL_LAME_FOUR = 0x2E91;
const char32_t CJK_RADICAL_SNAKE = 0x2E92;
const char32_t CJK_RADICAL_THREAD = 0x2E93;
const char32_t CJK_RADICAL_SNOUT_ONE = 0x2E94;
const char32_t CJK_RADICAL_SNOUT_TWO = 0x2E95;
const char32_t CJK_RADICAL_HEART_ONE = 0x2E96;
const char32_t CJK_RADICAL_HEART_TWO = 0x2E97;
const char32_t CJK_RADICAL_HAND = 0x2E98;
const char32_t CJK_RADICAL_RAP = 0x2E99;
const char32_t CJK_RADICAL_CHOKE = 0x2E9B;
const char32_t CJK_RADICAL_SUN = 0x2E9C;
const char32_t CJK_RADICAL_MOON = 0x2E9D;
const char32_t CJK_RADICAL_DEATH = 0x2E9E;
const char32_t CJK_RADICAL_MOTHER = 0x2E9F;
const char32_t CJK_RADICAL_CIVILIAN = 0x2EA0;
const char32_t CJK_RADICAL_WATER_ONE = 0x2EA1;
const char32_t CJK_RADICAL_WATER_TWO = 0x2EA2;
const char32_t CJK_RADICAL_FIRE = 0x2EA3;
const char32_t CJK_RADICAL_PAW_ONE = 0x2EA4;
const char32_t CJK_RADICAL_PAW_TWO = 0x2EA5;
const char32_t CJK_RADICAL_SIMPLIFIED_HALF_TREE_TRUNK = 0x2EA6;
const char32_t CJK_RADICAL_COW = 0x2EA7;
const char32_t CJK_RADICAL_DOG = 0x2EA8;
const char32_t CJK_RADICAL_JADE = 0x2EA9;
const char32_t CJK_RADICAL_BOLT_OF_CLOTH = 0x2EAA;
const char32_t CJK_RADICAL_EYE = 0x2EAB;
const char32_t CJK_RADICAL_SPIRIT_ONE = 0x2EAC;
const char32_t CJK_RADICAL_SPIRIT_TWO = 0x2EAD;
const char32_t CJK_RADICAL_BAMBOO = 0x2EAE;
const char32_t CJK_RADICAL_SILK = 0x2EAF;
const char32_t CJK_RADICAL_C_SIMPLIFIED_SILK = 0x2EB0;
const char32_t CJK_RADICAL_NET_ONE = 0x2EB1;
const char32_t CJK_RADICAL_NET_TWO = 0x2EB2;
const char32_t CJK_RADICAL_NET_THREE = 0x2EB3;
const char32_t CJK_RADICAL_NET_FOUR = 0x2EB4;
const char32_t CJK_RADICAL_MESH = 0x2EB5;
const char32_t CJK_RADICAL_SHEEP = 0x2EB6;
const char32_t CJK_RADICAL_RAM = 0x2EB7;
const char32_t CJK_RADICAL_EWE = 0x2EB8;
const char32_t CJK_RADICAL_OLD = 0x2EB9;
const char32_t CJK_RADICAL_BRUSH_ONE = 0x2EBA;
const char32_t CJK_RADICAL_BRUSH_TWO = 0x2EBB;
const char32_t CJK_RADICAL_MEAT = 0x2EBC;
const char32_t CJK_RADICAL_MORTAR = 0x2EBD;
const char32_t CJK_RADICAL_GRASS_ONE = 0x2EBE;
const char32_t CJK_RADICAL_GRASS_TWO = 0x2EBF;
const char32_t CJK_RADICAL_GRASS_THREE = 0x2EC0;
const char32_t CJK_RADICAL_TIGER = 0x2EC1;
const char32_t CJK_RADICAL_CLOTHES = 0x2EC2;
const char32_t CJK_RADICAL_WEST_ONE = 0x2EC3;
const char32_t CJK_RADICAL_WEST_TWO = 0x2EC4;
const char32_t CJK_RADICAL_C_SIMPLIFIED_SEE = 0x2EC5;
const char32_t CJK_RADICAL_SIMPLIFIED_HORN = 0x2EC6;
const char32_t CJK_RADICAL_HORN = 0x2EC7;
const char32_t CJK_RADICAL_C_SIMPLIFIED_SPEECH = 0x2EC8;
const char32_t CJK_RADICAL_C_SIMPLIFIED_SHELL = 0x2EC9;
const char32_t CJK_RADICAL_FOOT = 0x2ECA;
const char32_t CJK_RADICAL_C_SIMPLIFIED_CART = 0x2ECB;
const char32_t CJK_RADICAL_SIMPLIFIED_WALK = 0x2ECC;
const char32_t CJK_RADICAL_WALK_ONE = 0x2ECD;
const char32_t CJK_RADICAL_WALK_TWO = 0x2ECE;
const char32_t CJK_RADICAL_CITY = 0x2ECF;
const char32_t CJK_RADICAL_C_SIMPLIFIED_GOLD = 0x2ED0;
const char32_t CJK_RADICAL_LONG_ONE = 0x2ED1;
const char32_t CJK_RADICAL_LONG_TWO = 0x2ED2;
const char32_t CJK_RADICAL_C_SIMPLIFIED_LONG = 0x2ED3;
const char32_t CJK_RADICAL_C_SIMPLIFIED_GATE = 0x2ED4;
const char32_t CJK_RADICAL_MOUND_ONE = 0x2ED5;
const char32_t CJK_RADICAL_MOUND_TWO = 0x2ED6;
const char32_t CJK_RADICAL_RAIN = 0x2ED7;
const char32_t CJK_RADICAL_BLUE = 0x2ED8;
const char32_t CJK_RADICAL_C_SIMPLIFIED_TANNED_LEATHER = 0x2ED9;
const char32_t CJK_RADICAL_C_SIMPLIFIED_LEAF = 0x2EDA;
const char32_t CJK_RADICAL_C_SIMPLIFIED_WIND = 0x2EDB;
const char32_t CJK_RADICAL_C_SIMPLIFIED_FLY = 0x2EDC;
const char32_t CJK_RADICAL_EAT_ONE = 0x2EDD;
const char32_t CJK_RADICAL_EAT_TWO = 0x2EDE;
const char32_t CJK_RADICAL_EAT_THREE = 0x2EDF;
const char32_t CJK_RADICAL_C_SIMPLIFIED_EAT = 0x2EE0;
const char32_t CJK_RADICAL_HEAD = 0x2EE1;
const char32_t CJK_RADICAL_C_SIMPLIFIED_HORSE = 0x2EE2;
const char32_t CJK_RADICAL_BONE = 0x2EE3;
const char32_t CJK_RADICAL_GHOST = 0x2EE4;
const char32_t CJK_RADICAL_C_SIMPLIFIED_FISH = 0x2EE5;
const char32_t CJK_RADICAL_C_SIMPLIFIED_BIRD = 0x2EE6;
const char32_t CJK_RADICAL_C_SIMPLIFIED_SALT = 0x2EE7;
const char32_t CJK_RADICAL_SIMPLIFIED_WHEAT = 0x2EE8;
const char32_t CJK_RADICAL_SIMPLIFIED_YELLOW = 0x2EE9;
const char32_t CJK_RADICAL_C_SIMPLIFIED_FROG = 0x2EEA;
const char32_t CJK_RADICAL_J_SIMPLIFIED_EVEN = 0x2EEB;
const char32_t CJK_RADICAL_C_SIMPLIFIED_EVEN = 0x2EEC;
const char32_t CJK_RADICAL_J_SIMPLIFIED_TOOTH = 0x2EED;
const char32_t CJK_RADICAL_C_SIMPLIFIED_TOOTH = 0x2EEE;
const char32_t CJK_RADICAL_J_SIMPLIFIED_DRAGON = 0x2EEF;
const char32_t CJK_RADICAL_C_SIMPLIFIED_DRAGON = 0x2EF0;
const char32_t CJK_RADICAL_TURTLE = 0x2EF1;
const char32_t CJK_RADICAL_J_SIMPLIFIED_TURTLE = 0x2EF2;
const char32_t CJK_RADICAL_C_SIMPLIFIED_TURTLE = 0x2EF3;
const char32_t KANGXI_RADICAL_ONE = 0x2F00;
const char32_t KANGXI_RADICAL_LINE = 0x2F01;
const char32_t KANGXI_RADICAL_DOT = 0x2F02;
const char32_t KANGXI_RADICAL_SLASH = 0x2F03;
const char32_t KANGXI_RADICAL_SECOND = 0x2F04;
const char32_t KANGXI_RADICAL_HOOK = 0x2F05;
const char32_t KANGXI_RADICAL_TWO = 0x2F06;
const char32_t KANGXI_RADICAL_LID = 0x2F07;
const char32_t KANGXI_RADICAL_MAN = 0x2F08;
const char32_t KANGXI_RADICAL_LEGS = 0x2F09;
const char32_t KANGXI_RADICAL_ENTER = 0x2F0A;
const char32_t KANGXI_RADICAL_EIGHT = 0x2F0B;
const char32_t KANGXI_RADICAL_DOWN_BOX = 0x2F0C;
const char32_t KANGXI_RADICAL_COVER = 0x2F0D;
const char32_t KANGXI_RADICAL_ICE = 0x2F0E;
const char32_t KANGXI_RADICAL_TABLE = 0x2F0F;
const char32_t KANGXI_RADICAL_OPEN_BOX = 0x2F10;
const char32_t KANGXI_RADICAL_KNIFE = 0x2F11;
const char32_t KANGXI_RADICAL_POWER = 0x2F12;
const char32_t KANGXI_RADICAL_WRAP = 0x2F13;
const char32_t KANGXI_RADICAL_SPOON = 0x2F14;
const char32_t KANGXI_RADICAL_RIGHT_OPEN_BOX = 0x2F15;
const char32_t KANGXI_RADICAL_HIDING_ENCLOSURE = 0x2F16;
const char32_t KANGXI_RADICAL_TEN = 0x2F17;
const char32_t KANGXI_RADICAL_DIVINATION = 0x2F18;
const char32_t KANGXI_RADICAL_SEAL = 0x2F19;
const char32_t KANGXI_RADICAL_CLIFF = 0x2F1A;
const char32_t KANGXI_RADICAL_PRIVATE = 0x2F1B;
const char32_t KANGXI_RADICAL_AGAIN = 0x2F1C;
const char32_t KANGXI_RADICAL_MOUTH = 0x2F1D;
const char32_t KANGXI_RADICAL_ENCLOSURE = 0x2F1E;
const char32_t KANGXI_RADICAL_EARTH = 0x2F1F;
const char32_t KANGXI_RADICAL_SCHOLAR = 0x2F20;
const char32_t KANGXI_RADICAL_GO = 0x2F21;
const char32_t KANGXI_RADICAL_GO_SLOWLY = 0x2F22;
const char32_t KANGXI_RADICAL_EVENING = 0x2F23;
const char32_t KANGXI_RADICAL_BIG = 0x2F24;
const char32_t KANGXI_RADICAL_WOMAN = 0x2F25;
const char32_t KANGXI_RADICAL_CHILD = 0x2F26;
const char32_t KANGXI_RADICAL_ROOF = 0x2F27;
const char32_t KANGXI_RADICAL_INCH = 0x2F28;
const char32_t KANGXI_RADICAL_SMALL = 0x2F29;
const char32_t KANGXI_RADICAL_LAME = 0x2F2A;
const char32_t KANGXI_RADICAL_CORPSE = 0x2F2B;
const char32_t KANGXI_RADICAL_SPROUT = 0x2F2C;
const char32_t KANGXI_RADICAL_MOUNTAIN = 0x2F2D;
const char32_t KANGXI_RADICAL_RIVER = 0x2F2E;
const char32_t KANGXI_RADICAL_WORK = 0x2F2F;
const char32_t KANGXI_RADICAL_ONESELF = 0x2F30;
const char32_t KANGXI_RADICAL_TURBAN = 0x2F31;
const char32_t KANGXI_RADICAL_DRY = 0x2F32;
const char32_t KANGXI_RADICAL_SHORT_THREAD = 0x2F33;
const char32_t KANGXI_RADICAL_DOTTED_CLIFF = 0x2F34;
const char32_t KANGXI_RADICAL_LONG_STRIDE = 0x2F35;
const char32_t KANGXI_RADICAL_TWO_HANDS = 0x2F36;
const char32_t KANGXI_RADICAL_SHOOT = 0x2F37;
const char32_t KANGXI_RADICAL_BOW = 0x2F38;
const char32_t KANGXI_RADICAL_SNOUT = 0x2F39;
const char32_t KANGXI_RADICAL_BRISTLE = 0x2F3A;
const char32_t KANGXI_RADICAL_STEP = 0x2F3B;
const char32_t KANGXI_RADICAL_HEART = 0x2F3C;
const char32_t KANGXI_RADICAL_HALBERD = 0x2F3D;
const char32_t KANGXI_RADICAL_DOOR = 0x2F3E;
const char32_t KANGXI_RADICAL_HAND = 0x2F3F;
const char32_t KANGXI_RADICAL_BRANCH = 0x2F40;
const char32_t KANGXI_RADICAL_RAP = 0x2F41;
const char32_t KANGXI_RADICAL_SCRIPT = 0x2F42;
const char32_t KANGXI_RADICAL_DIPPER = 0x2F43;
const char32_t KANGXI_RADICAL_AXE = 0x2F44;
const char32_t KANGXI_RADICAL_SQUARE = 0x2F45;
const char32_t KANGXI_RADICAL_NOT = 0x2F46;
const char32_t KANGXI_RADICAL_SUN = 0x2F47;
const char32_t KANGXI_RADICAL_SAY = 0x2F48;
const char32_t KANGXI_RADICAL_MOON = 0x2F49;
const char32_t KANGXI_RADICAL_TREE = 0x2F4A;
const char32_t KANGXI_RADICAL_LACK = 0x2F4B;
const char32_t KANGXI_RADICAL_STOP = 0x2F4C;
const char32_t KANGXI_RADICAL_DEATH = 0x2F4D;
const char32_t KANGXI_RADICAL_WEAPON = 0x2F4E;
const char32_t KANGXI_RADICAL_DO_NOT = 0x2F4F;
const char32_t KANGXI_RADICAL_COMPARE = 0x2F50;
const char32_t KANGXI_RADICAL_FUR = 0x2F51;
const char32_t KANGXI_RADICAL_CLAN = 0x2F52;
const char32_t KANGXI_RADICAL_STEAM = 0x2F53;
const char32_t KANGXI_RADICAL_WATER = 0x2F54;
const char32_t KANGXI_RADICAL_FIRE = 0x2F55;
const char32_t KANGXI_RADICAL_CLAW = 0x2F56;
const char32_t KANGXI_RADICAL_FATHER = 0x2F57;
const char32_t KANGXI_RADICAL_DOUBLE_X = 0x2F58;
const char32_t KANGXI_RADICAL_HALF_TREE_TRUNK = 0x2F59;
const char32_t KANGXI_RADICAL_SLICE = 0x2F5A;
const char32_t KANGXI_RADICAL_FANG = 0x2F5B;
const char32_t KANGXI_RADICAL_COW = 0x2F5C;
const char32_t KANGXI_RADICAL_DOG = 0x2F5D;
const char32_t KANGXI_RADICAL_PROFOUND = 0x2F5E;
const char32_t KANGXI_RADICAL_JADE = 0x2F5F;
const char32_t KANGXI_RADICAL_MELON = 0x2F60;
const char32_t KANGXI_RADICAL_TILE = 0x2F61;
const char32_t KANGXI_RADICAL_SWEET = 0x2F62;
const char32_t KANGXI_RADICAL_LIFE = 0x2F63;
const char32_t KANGXI_RADICAL_USE = 0x2F64;
const char32_t KANGXI_RADICAL_FIELD = 0x2F65;
const char32_t KANGXI_RADICAL_BOLT_OF_CLOTH = 0x2F66;
const char32_t KANGXI_RADICAL_SICKNESS = 0x2F67;
const char32_t KANGXI_RADICAL_DOTTED_TENT = 0x2F68;
const char32_t KANGXI_RADICAL_WHITE = 0x2F69;
const char32_t KANGXI_RADICAL_SKIN = 0x2F6A;
const char32_t KANGXI_RADICAL_DISH = 0x2F6B;
const char32_t KANGXI_RADICAL_EYE = 0x2F6C;
const char32_t KANGXI_RADICAL_SPEAR = 0x2F6D;
const char32_t KANGXI_RADICAL_ARROW = 0x2F6E;
const char32_t KANGXI_RADICAL_STONE = 0x2F6F;
const char32_t KANGXI_RADICAL_SPIRIT = 0x2F70;
const char32_t KANGXI_RADICAL_TRACK = 0x2F71;
const char32_t KANGXI_RADICAL_GRAIN = 0x2F72;
const char32_t KANGXI_RADICAL_CAVE = 0x2F73;
const char32_t KANGXI_RADICAL_STAND = 0x2F74;
const char32_t KANGXI_RADICAL_BAMBOO = 0x2F75;
const char32_t KANGXI_RADICAL_RICE = 0x2F76;
const char32_t KANGXI_RADICAL_SILK = 0x2F77;
const char32_t KANGXI_RADICAL_JAR = 0x2F78;
const char32_t KANGXI_RADICAL_NET = 0x2F79;
const char32_t KANGXI_RADICAL_SHEEP = 0x2F7A;
const char32_t KANGXI_RADICAL_FEATHER = 0x2F7B;
const char32_t KANGXI_RADICAL_OLD = 0x2F7C;
const char32_t KANGXI_RADICAL_AND = 0x2F7D;
const char32_t KANGXI_RADICAL_PLOW = 0x2F7E;
const char32_t KANGXI_RADICAL_EAR = 0x2F7F;
const char32_t KANGXI_RADICAL_BRUSH = 0x2F80;
const char32_t KANGXI_RADICAL_MEAT = 0x2F81;
const char32_t KANGXI_RADICAL_MINISTER = 0x2F82;
const char32_t KANGXI_RADICAL_SELF = 0x2F83;
const char32_t KANGXI_RADICAL_ARRIVE = 0x2F84;
const char32_t KANGXI_RADICAL_MORTAR = 0x2F85;
const char32_t KANGXI_RADICAL_TONGUE = 0x2F86;
const char32_t KANGXI_RADICAL_OPPOSE = 0x2F87;
const char32_t KANGXI_RADICAL_BOAT = 0x2F88;
const char32_t KANGXI_RADICAL_STOPPING = 0x2F89;
const char32_t KANGXI_RADICAL_COLOR = 0x2F8A;
const char32_t KANGXI_RADICAL_GRASS = 0x2F8B;
const char32_t KANGXI_RADICAL_TIGER = 0x2F8C;
const char32_t KANGXI_RADICAL_INSECT = 0x2F8D;
const char32_t KANGXI_RADICAL_BLOOD = 0x2F8E;
const char32_t KANGXI_RADICAL_WALK_ENCLOSURE = 0x2F8F;
const char32_t KANGXI_RADICAL_CLOTHES = 0x2F90;
const char32_t KANGXI_RADICAL_WEST = 0x2F91;
const char32_t KANGXI_RADICAL_SEE = 0x2F92;
const char32_t KANGXI_RADICAL_HORN = 0x2F93;
const char32_t KANGXI_RADICAL_SPEECH = 0x2F94;
const char32_t KANGXI_RADICAL_VALLEY = 0x2F95;
const char32_t KANGXI_RADICAL_BEAN = 0x2F96;
const char32_t KANGXI_RADICAL_PIG = 0x2F97;
const char32_t KANGXI_RADICAL_BADGER = 0x2F98;
const char32_t KANGXI_RADICAL_SHELL = 0x2F99;
const char32_t KANGXI_RADICAL_RED = 0x2F9A;
const char32_t KANGXI_RADICAL_RUN = 0x2F9B;
const char32_t KANGXI_RADICAL_FOOT = 0x2F9C;
const char32_t KANGXI_RADICAL_BODY = 0x2F9D;
const char32_t KANGXI_RADICAL_CART = 0x2F9E;
const char32_t KANGXI_RADICAL_BITTER = 0x2F9F;
const char32_t KANGXI_RADICAL_MORNING = 0x2FA0;
const char32_t KANGXI_RADICAL_WALK = 0x2FA1;
const char32_t KANGXI_RADICAL_CITY = 0x2FA2;
const char32_t KANGXI_RADICAL_WINE = 0x2FA3;
const char32_t KANGXI_RADICAL_DISTINGUISH = 0x2FA4;
const char32_t KANGXI_RADICAL_VILLAGE = 0x2FA5;
const char32_t KANGXI_RADICAL_GOLD = 0x2FA6;
const char32_t KANGXI_RADICAL_LONG = 0x2FA7;
const char32_t KANGXI_RADICAL_GATE = 0x2FA8;
const char32_t KANGXI_RADICAL_MOUND = 0x2FA9;
const char32_t KANGXI_RADICAL_SLAVE = 0x2FAA;
const char32_t KANGXI_RADICAL_SHORT_TAILED_BIRD = 0x2FAB;
const char32_t KANGXI_RADICAL_RAIN = 0x2FAC;
const char32_t KANGXI_RADICAL_BLUE = 0x2FAD;
const char32_t KANGXI_RADICAL_WRONG = 0x2FAE;
const char32_t KANGXI_RADICAL_FACE = 0x2FAF;
const char32_t KANGXI_RADICAL_LEATHER = 0x2FB0;
const char32_t KANGXI_RADICAL_TANNED_LEATHER = 0x2FB1;
const char32_t KANGXI_RADICAL_LEEK = 0x2FB2;
const char32_t KANGXI_RADICAL_SOUND = 0x2FB3;
const char32_t KANGXI_RADICAL_LEAF = 0x2FB4;
const char32_t KANGXI_RADICAL_WIND = 0x2FB5;
const char32_t KANGXI_RADICAL_FLY = 0x2FB6;
const char32_t KANGXI_RADICAL_EAT = 0x2FB7;
const char32_t KANGXI_RADICAL_HEAD = 0x2FB8;
const char32_t KANGXI_RADICAL_FRAGRANT = 0x2FB9;
const char32_t KANGXI_RADICAL_HORSE = 0x2FBA;
const char32_t KANGXI_RADICAL_BONE = 0x2FBB;
const char32_t KANGXI_RADICAL_TALL = 0x2FBC;
const char32_t KANGXI_RADICAL_HAIR = 0x2FBD;
const char32_t KANGXI_RADICAL_FIGHT = 0x2FBE;
const char32_t KANGXI_RADICAL_SACRIFICIAL_WINE = 0x2FBF;
const char32_t KANGXI_RADICAL_CAULDRON = 0x2FC0;
const char32_t KANGXI_RADICAL_GHOST = 0x2FC1;
const char32_t KANGXI_RADICAL_FISH = 0x2FC2;
const char32_t KANGXI_RADICAL_BIRD = 0x2FC3;
const char32_t KANGXI_RADICAL_SALT = 0x2FC4;
const char32_t KANGXI_RADICAL_DEER = 0x2FC5;
const char32_t KANGXI_RADICAL_WHEAT = 0x2FC6;
const char32_t KANGXI_RADICAL_HEMP = 0x2FC7;
const char32_t KANGXI_RADICAL_YELLOW = 0x2FC8;
const char32_t KANGXI_RADICAL_MILLET = 0x2FC9;
const char32_t KANGXI_RADICAL_BLACK = 0x2FCA;
const char32_t KANGXI_RADICAL_EMBROIDERY = 0x2FCB;
const char32_t KANGXI_RADICAL_FROG = 0x2FCC;
const char32_t KANGXI_RADICAL_TRIPOD = 0x2FCD;
const char32_t KANGXI_RADICAL_DRUM = 0x2FCE;
const char32_t KANGXI_RADICAL_RAT = 0x2FCF;
const char32_t KANGXI_RADICAL_NOSE = 0x2FD0;
const char32_t KANGXI_RADICAL_EVEN = 0x2FD1;
const char32_t KANGXI_RADICAL_TOOTH = 0x2FD2;
const char32_t KANGXI_RADICAL_DRAGON = 0x2FD3;
const char32_t KANGXI_RADICAL_TURTLE = 0x2FD4;
const char32_t KANGXI_RADICAL_FLUTE = 0x2FD5;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_LEFT_TO_RIGHT = 0x2FF0;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_ABOVE_TO_BELOW = 0x2FF1;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_LEFT_TO_MIDDLE_AND_RIGHT = 0x2FF2;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_ABOVE_TO_MIDDLE_AND_BELOW = 0x2FF3;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_FULL_SURROUND = 0x2FF4;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_ABOVE = 0x2FF5;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_BELOW = 0x2FF6;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_LEFT = 0x2FF7;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_UPPER_LEFT = 0x2FF8;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_UPPER_RIGHT = 0x2FF9;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_LOWER_LEFT = 0x2FFA;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_OVERLAID = 0x2FFB;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_RIGHT = 0x2FFC;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SURROUND_FROM_LOWER_RIGHT = 0x2FFD;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_HORIZONTAL_REFLECTION = 0x2FFE;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_ROTATION = 0x2FFF;
const char32_t IDEOGRAPHIC_SPACE = 0x3000;
const char32_t IDEOGRAPHIC_COMMA = 0x3001;
const char32_t IDEOGRAPHIC_FULL_STOP = 0x3002;
const char32_t DITTO_MARK = 0x3003;
const char32_t JAPANESE_INDUSTRIAL_STANDARD_SYMBOL = 0x3004;
const char32_t IDEOGRAPHIC_ITERATION_MARK = 0x3005;
const char32_t IDEOGRAPHIC_CLOSING_MARK = 0x3006;
const char32_t IDEOGRAPHIC_NUMBER_ZERO = 0x3007;
const char32_t LEFT_ANGLE_BRACKET = 0x3008;
const char32_t RIGHT_ANGLE_BRACKET = 0x3009;
const char32_t LEFT_DOUBLE_ANGLE_BRACKET = 0x300A;
const char32_t RIGHT_DOUBLE_ANGLE_BRACKET = 0x300B;
const char32_t LEFT_CORNER_BRACKET = 0x300C;
const char32_t RIGHT_CORNER_BRACKET = 0x300D;
const char32_t LEFT_WHITE_CORNER_BRACKET = 0x300E;
const char32_t RIGHT_WHITE_CORNER_BRACKET = 0x300F;
const char32_t LEFT_BLACK_LENTICULAR_BRACKET = 0x3010;
const char32_t RIGHT_BLACK_LENTICULAR_BRACKET = 0x3011;
const char32_t POSTAL_MARK = 0x3012;
const char32_t GETA_MARK = 0x3013;
const char32_t LEFT_TORTOISE_SHELL_BRACKET = 0x3014;
const char32_t RIGHT_TORTOISE_SHELL_BRACKET = 0x3015;
const char32_t LEFT_WHITE_LENTICULAR_BRACKET = 0x3016;
const char32_t RIGHT_WHITE_LENTICULAR_BRACKET = 0x3017;
const char32_t LEFT_WHITE_TORTOISE_SHELL_BRACKET = 0x3018;
const char32_t RIGHT_WHITE_TORTOISE_SHELL_BRACKET = 0x3019;
const char32_t LEFT_WHITE_SQUARE_BRACKET = 0x301A;
const char32_t RIGHT_WHITE_SQUARE_BRACKET = 0x301B;
const char32_t WAVE_DASH = 0x301C;
const char32_t REVERSED_DOUBLE_PRIME_QUOTATION_MARK = 0x301D;
const char32_t DOUBLE_PRIME_QUOTATION_MARK = 0x301E;
const char32_t LOW_DOUBLE_PRIME_QUOTATION_MARK = 0x301F;
const char32_t POSTAL_MARK_FACE = 0x3020;
const char32_t HANGZHOU_NUMERAL_ONE = 0x3021;
const char32_t HANGZHOU_NUMERAL_TWO = 0x3022;
const char32_t HANGZHOU_NUMERAL_THREE = 0x3023;
const char32_t HANGZHOU_NUMERAL_FOUR = 0x3024;
const char32_t HANGZHOU_NUMERAL_FIVE = 0x3025;
const char32_t HANGZHOU_NUMERAL_SIX = 0x3026;
const char32_t HANGZHOU_NUMERAL_SEVEN = 0x3027;
const char32_t HANGZHOU_NUMERAL_EIGHT = 0x3028;
const char32_t HANGZHOU_NUMERAL_NINE = 0x3029;
const char32_t IDEOGRAPHIC_LEVEL_TONE_MARK = 0x302A;
const char32_t IDEOGRAPHIC_RISING_TONE_MARK = 0x302B;
const char32_t IDEOGRAPHIC_DEPARTING_TONE_MARK = 0x302C;
const char32_t IDEOGRAPHIC_ENTERING_TONE_MARK = 0x302D;
const char32_t HANGUL_SINGLE_DOT_TONE_MARK = 0x302E;
const char32_t HANGUL_DOUBLE_DOT_TONE_MARK = 0x302F;
const char32_t WAVY_DASH = 0x3030;
const char32_t VERTICAL_KANA_REPEAT_MARK = 0x3031;
const char32_t VERTICAL_KANA_REPEAT_WITH_VOICED_SOUND_MARK = 0x3032;
const char32_t VERTICAL_KANA_REPEAT_MARK_UPPER_HALF = 0x3033;
const char32_t VERTICAL_KANA_REPEAT_WITH_VOICED_SOUND_MARK_UPPER_HALF = 0x3034;
const char32_t VERTICAL_KANA_REPEAT_MARK_LOWER_HALF = 0x3035;
const char32_t CIRCLED_POSTAL_MARK = 0x3036;
const char32_t IDEOGRAPHIC_TELEGRAPH_LINE_FEED_SEPARATOR_SYMBOL = 0x3037;
const char32_t HANGZHOU_NUMERAL_TEN = 0x3038;
const char32_t HANGZHOU_NUMERAL_TWENTY = 0x3039;
const char32_t HANGZHOU_NUMERAL_THIRTY = 0x303A;
const char32_t VERTICAL_IDEOGRAPHIC_ITERATION_MARK = 0x303B;
const char32_t MASU_MARK = 0x303C;
const char32_t PART_ALTERNATION_MARK = 0x303D;
const char32_t IDEOGRAPHIC_VARIATION_INDICATOR = 0x303E;
const char32_t IDEOGRAPHIC_HALF_FILL_SPACE = 0x303F;
const char32_t HIRAGANA_LETTER_SMALL_A = 0x3041;
const char32_t HIRAGANA_LETTER_A = 0x3042;
const char32_t HIRAGANA_LETTER_SMALL_I = 0x3043;
const char32_t HIRAGANA_LETTER_I = 0x3044;
const char32_t HIRAGANA_LETTER_SMALL_U = 0x3045;
const char32_t HIRAGANA_LETTER_U = 0x3046;
const char32_t HIRAGANA_LETTER_SMALL_E = 0x3047;
const char32_t HIRAGANA_LETTER_E = 0x3048;
const char32_t HIRAGANA_LETTER_SMALL_O = 0x3049;
const char32_t HIRAGANA_LETTER_O = 0x304A;
const char32_t HIRAGANA_LETTER_KA = 0x304B;
const char32_t HIRAGANA_LETTER_GA = 0x304C;
const char32_t HIRAGANA_LETTER_KI = 0x304D;
const char32_t HIRAGANA_LETTER_GI = 0x304E;
const char32_t HIRAGANA_LETTER_KU = 0x304F;
const char32_t HIRAGANA_LETTER_GU = 0x3050;
const char32_t HIRAGANA_LETTER_KE = 0x3051;
const char32_t HIRAGANA_LETTER_GE = 0x3052;
const char32_t HIRAGANA_LETTER_KO = 0x3053;
const char32_t HIRAGANA_LETTER_GO = 0x3054;
const char32_t HIRAGANA_LETTER_SA = 0x3055;
const char32_t HIRAGANA_LETTER_ZA = 0x3056;
const char32_t HIRAGANA_LETTER_SI = 0x3057;
const char32_t HIRAGANA_LETTER_ZI = 0x3058;
const char32_t HIRAGANA_LETTER_SU = 0x3059;
const char32_t HIRAGANA_LETTER_ZU = 0x305A;
const char32_t HIRAGANA_LETTER_SE = 0x305B;
const char32_t HIRAGANA_LETTER_ZE = 0x305C;
const char32_t HIRAGANA_LETTER_SO = 0x305D;
const char32_t HIRAGANA_LETTER_ZO = 0x305E;
const char32_t HIRAGANA_LETTER_TA = 0x305F;
const char32_t HIRAGANA_LETTER_DA = 0x3060;
const char32_t HIRAGANA_LETTER_TI = 0x3061;
const char32_t HIRAGANA_LETTER_DI = 0x3062;
const char32_t HIRAGANA_LETTER_SMALL_TU = 0x3063;
const char32_t HIRAGANA_LETTER_TU = 0x3064;
const char32_t HIRAGANA_LETTER_DU = 0x3065;
const char32_t HIRAGANA_LETTER_TE = 0x3066;
const char32_t HIRAGANA_LETTER_DE = 0x3067;
const char32_t HIRAGANA_LETTER_TO = 0x3068;
const char32_t HIRAGANA_LETTER_DO = 0x3069;
const char32_t HIRAGANA_LETTER_NA = 0x306A;
const char32_t HIRAGANA_LETTER_NI = 0x306B;
const char32_t HIRAGANA_LETTER_NU = 0x306C;
const char32_t HIRAGANA_LETTER_NE = 0x306D;
const char32_t HIRAGANA_LETTER_NO = 0x306E;
const char32_t HIRAGANA_LETTER_HA = 0x306F;
const char32_t HIRAGANA_LETTER_BA = 0x3070;
const char32_t HIRAGANA_LETTER_PA = 0x3071;
const char32_t HIRAGANA_LETTER_HI = 0x3072;
const char32_t HIRAGANA_LETTER_BI = 0x3073;
const char32_t HIRAGANA_LETTER_PI = 0x3074;
const char32_t HIRAGANA_LETTER_HU = 0x3075;
const char32_t HIRAGANA_LETTER_BU = 0x3076;
const char32_t HIRAGANA_LETTER_PU = 0x3077;
const char32_t HIRAGANA_LETTER_HE = 0x3078;
const char32_t HIRAGANA_LETTER_BE = 0x3079;
const char32_t HIRAGANA_LETTER_PE = 0x307A;
const char32_t HIRAGANA_LETTER_HO = 0x307B;
const char32_t HIRAGANA_LETTER_BO = 0x307C;
const char32_t HIRAGANA_LETTER_PO = 0x307D;
const char32_t HIRAGANA_LETTER_MA = 0x307E;
const char32_t HIRAGANA_LETTER_MI = 0x307F;
const char32_t HIRAGANA_LETTER_MU = 0x3080;
const char32_t HIRAGANA_LETTER_ME = 0x3081;
const char32_t HIRAGANA_LETTER_MO = 0x3082;
const char32_t HIRAGANA_LETTER_SMALL_YA = 0x3083;
const char32_t HIRAGANA_LETTER_YA = 0x3084;
const char32_t HIRAGANA_LETTER_SMALL_YU = 0x3085;
const char32_t HIRAGANA_LETTER_YU = 0x3086;
const char32_t HIRAGANA_LETTER_SMALL_YO = 0x3087;
const char32_t HIRAGANA_LETTER_YO = 0x3088;
const char32_t HIRAGANA_LETTER_RA = 0x3089;
const char32_t HIRAGANA_LETTER_RI = 0x308A;
const char32_t HIRAGANA_LETTER_RU = 0x308B;
const char32_t HIRAGANA_LETTER_RE = 0x308C;
const char32_t HIRAGANA_LETTER_RO = 0x308D;
const char32_t HIRAGANA_LETTER_SMALL_WA = 0x308E;
const char32_t HIRAGANA_LETTER_WA = 0x308F;
const char32_t HIRAGANA_LETTER_WI = 0x3090;
const char32_t HIRAGANA_LETTER_WE = 0x3091;
const char32_t HIRAGANA_LETTER_WO = 0x3092;
const char32_t HIRAGANA_LETTER_N = 0x3093;
const char32_t HIRAGANA_LETTER_VU = 0x3094;
const char32_t HIRAGANA_LETTER_SMALL_KA = 0x3095;
const char32_t HIRAGANA_LETTER_SMALL_KE = 0x3096;
const char32_t COMBINING_KATAKANA_HIRAGANA_VOICED_SOUND_MARK = 0x3099;
const char32_t COMBINING_KATAKANA_HIRAGANA_SEMI_VOICED_SOUND_MARK = 0x309A;
const char32_t KATAKANA_HIRAGANA_VOICED_SOUND_MARK = 0x309B;
const char32_t KATAKANA_HIRAGANA_SEMI_VOICED_SOUND_MARK = 0x309C;
const char32_t HIRAGANA_ITERATION_MARK = 0x309D;
const char32_t HIRAGANA_VOICED_ITERATION_MARK = 0x309E;
const char32_t HIRAGANA_DIGRAPH_YORI = 0x309F;
const char32_t KATAKANA_HIRAGANA_DOUBLE_HYPHEN = 0x30A0;
const char32_t KATAKANA_LETTER_SMALL_A = 0x30A1;
const char32_t KATAKANA_LETTER_A = 0x30A2;
const char32_t KATAKANA_LETTER_SMALL_I = 0x30A3;
const char32_t KATAKANA_LETTER_I = 0x30A4;
const char32_t KATAKANA_LETTER_SMALL_U = 0x30A5;
const char32_t KATAKANA_LETTER_U = 0x30A6;
const char32_t KATAKANA_LETTER_SMALL_E = 0x30A7;
const char32_t KATAKANA_LETTER_E = 0x30A8;
const char32_t KATAKANA_LETTER_SMALL_O = 0x30A9;
const char32_t KATAKANA_LETTER_O = 0x30AA;
const char32_t KATAKANA_LETTER_KA = 0x30AB;
const char32_t KATAKANA_LETTER_GA = 0x30AC;
const char32_t KATAKANA_LETTER_KI = 0x30AD;
const char32_t KATAKANA_LETTER_GI = 0x30AE;
const char32_t KATAKANA_LETTER_KU = 0x30AF;
const char32_t KATAKANA_LETTER_GU = 0x30B0;
const char32_t KATAKANA_LETTER_KE = 0x30B1;
const char32_t KATAKANA_LETTER_GE = 0x30B2;
const char32_t KATAKANA_LETTER_KO = 0x30B3;
const char32_t KATAKANA_LETTER_GO = 0x30B4;
const char32_t KATAKANA_LETTER_SA = 0x30B5;
const char32_t KATAKANA_LETTER_ZA = 0x30B6;
const char32_t KATAKANA_LETTER_SI = 0x30B7;
const char32_t KATAKANA_LETTER_ZI = 0x30B8;
const char32_t KATAKANA_LETTER_SU = 0x30B9;
const char32_t KATAKANA_LETTER_ZU = 0x30BA;
const char32_t KATAKANA_LETTER_SE = 0x30BB;
const char32_t KATAKANA_LETTER_ZE = 0x30BC;
const char32_t KATAKANA_LETTER_SO = 0x30BD;
const char32_t KATAKANA_LETTER_ZO = 0x30BE;
const char32_t KATAKANA_LETTER_TA = 0x30BF;
const char32_t KATAKANA_LETTER_DA = 0x30C0;
const char32_t KATAKANA_LETTER_TI = 0x30C1;
const char32_t KATAKANA_LETTER_DI = 0x30C2;
const char32_t KATAKANA_LETTER_SMALL_TU = 0x30C3;
const char32_t KATAKANA_LETTER_TU = 0x30C4;
const char32_t KATAKANA_LETTER_DU = 0x30C5;
const char32_t KATAKANA_LETTER_TE = 0x30C6;
const char32_t KATAKANA_LETTER_DE = 0x30C7;
const char32_t KATAKANA_LETTER_TO = 0x30C8;
const char32_t KATAKANA_LETTER_DO = 0x30C9;
const char32_t KATAKANA_LETTER_NA = 0x30CA;
const char32_t KATAKANA_LETTER_NI = 0x30CB;
const char32_t KATAKANA_LETTER_NU = 0x30CC;
const char32_t KATAKANA_LETTER_NE = 0x30CD;
const char32_t KATAKANA_LETTER_NO = 0x30CE;
const char32_t KATAKANA_LETTER_HA = 0x30CF;
const char32_t KATAKANA_LETTER_BA = 0x30D0;
const char32_t KATAKANA_LETTER_PA = 0x30D1;
const char32_t KATAKANA_LETTER_HI = 0x30D2;
const char32_t KATAKANA_LETTER_BI = 0x30D3;
const char32_t KATAKANA_LETTER_PI = 0x30D4;
const char32_t KATAKANA_LETTER_HU = 0x30D5;
const char32_t KATAKANA_LETTER_BU = 0x30D6;
const char32_t KATAKANA_LETTER_PU = 0x30D7;
const char32_t KATAKANA_LETTER_HE = 0x30D8;
const char32_t KATAKANA_LETTER_BE = 0x30D9;
const char32_t KATAKANA_LETTER_PE = 0x30DA;
const char32_t KATAKANA_LETTER_HO = 0x30DB;
const char32_t KATAKANA_LETTER_BO = 0x30DC;
const char32_t KATAKANA_LETTER_PO = 0x30DD;
const char32_t KATAKANA_LETTER_MA = 0x30DE;
const char32_t KATAKANA_LETTER_MI = 0x30DF;
const char32_t KATAKANA_LETTER_MU = 0x30E0;
const char32_t KATAKANA_LETTER_ME = 0x30E1;
const char32_t KATAKANA_LETTER_MO = 0x30E2;
const char32_t KATAKANA_LETTER_SMALL_YA = 0x30E3;
const char32_t KATAKANA_LETTER_YA = 0x30E4;
const char32_t KATAKANA_LETTER_SMALL_YU = 0x30E5;
const char32_t KATAKANA_LETTER_YU = 0x30E6;
const char32_t KATAKANA_LETTER_SMALL_YO = 0x30E7;
const char32_t KATAKANA_LETTER_YO = 0x30E8;
const char32_t KATAKANA_LETTER_RA = 0x30E9;
const char32_t KATAKANA_LETTER_RI = 0x30EA;
const char32_t KATAKANA_LETTER_RU = 0x30EB;
const char32_t KATAKANA_LETTER_RE = 0x30EC;
const char32_t KATAKANA_LETTER_RO = 0x30ED;
const char32_t KATAKANA_LETTER_SMALL_WA = 0x30EE;
const char32_t KATAKANA_LETTER_WA = 0x30EF;
const char32_t KATAKANA_LETTER_WI = 0x30F0;
const char32_t KATAKANA_LETTER_WE = 0x30F1;
const char32_t KATAKANA_LETTER_WO = 0x30F2;
const char32_t KATAKANA_LETTER_N = 0x30F3;
const char32_t KATAKANA_LETTER_VU = 0x30F4;
const char32_t KATAKANA_LETTER_SMALL_KA = 0x30F5;
const char32_t KATAKANA_LETTER_SMALL_KE = 0x30F6;
const char32_t KATAKANA_LETTER_VA = 0x30F7;
const char32_t KATAKANA_LETTER_VI = 0x30F8;
const char32_t KATAKANA_LETTER_VE = 0x30F9;
const char32_t KATAKANA_LETTER_VO = 0x30FA;
const char32_t KATAKANA_MIDDLE_DOT = 0x30FB;
const char32_t KATAKANA_HIRAGANA_PROLONGED_SOUND_MARK = 0x30FC;
const char32_t KATAKANA_ITERATION_MARK = 0x30FD;
const char32_t KATAKANA_VOICED_ITERATION_MARK = 0x30FE;
const char32_t KATAKANA_DIGRAPH_KOTO = 0x30FF;
const char32_t BOPOMOFO_LETTER_B = 0x3105;
const char32_t BOPOMOFO_LETTER_P = 0x3106;
const char32_t BOPOMOFO_LETTER_M = 0x3107;
const char32_t BOPOMOFO_LETTER_F = 0x3108;
const char32_t BOPOMOFO_LETTER_D = 0x3109;
const char32_t BOPOMOFO_LETTER_T = 0x310A;
const char32_t BOPOMOFO_LETTER_N = 0x310B;
const char32_t BOPOMOFO_LETTER_L = 0x310C;
const char32_t BOPOMOFO_LETTER_G = 0x310D;
const char32_t BOPOMOFO_LETTER_K = 0x310E;
const char32_t BOPOMOFO_LETTER_H = 0x310F;
const char32_t BOPOMOFO_LETTER_J = 0x3110;
const char32_t BOPOMOFO_LETTER_Q = 0x3111;
const char32_t BOPOMOFO_LETTER_X = 0x3112;
const char32_t BOPOMOFO_LETTER_ZH = 0x3113;
const char32_t BOPOMOFO_LETTER_CH = 0x3114;
const char32_t BOPOMOFO_LETTER_SH = 0x3115;
const char32_t BOPOMOFO_LETTER_R = 0x3116;
const char32_t BOPOMOFO_LETTER_Z = 0x3117;
const char32_t BOPOMOFO_LETTER_C = 0x3118;
const char32_t BOPOMOFO_LETTER_S = 0x3119;
const char32_t BOPOMOFO_LETTER_A = 0x311A;
const char32_t BOPOMOFO_LETTER_O = 0x311B;
const char32_t BOPOMOFO_LETTER_E = 0x311C;
const char32_t BOPOMOFO_LETTER_EH = 0x311D;
const char32_t BOPOMOFO_LETTER_AI = 0x311E;
const char32_t BOPOMOFO_LETTER_EI = 0x311F;
const char32_t BOPOMOFO_LETTER_AU = 0x3120;
const char32_t BOPOMOFO_LETTER_OU = 0x3121;
const char32_t BOPOMOFO_LETTER_AN = 0x3122;
const char32_t BOPOMOFO_LETTER_EN = 0x3123;
const char32_t BOPOMOFO_LETTER_ANG = 0x3124;
const char32_t BOPOMOFO_LETTER_ENG = 0x3125;
const char32_t BOPOMOFO_LETTER_ER = 0x3126;
const char32_t BOPOMOFO_LETTER_I = 0x3127;
const char32_t BOPOMOFO_LETTER_U = 0x3128;
const char32_t BOPOMOFO_LETTER_IU = 0x3129;
const char32_t BOPOMOFO_LETTER_V = 0x312A;
const char32_t BOPOMOFO_LETTER_NG = 0x312B;
const char32_t BOPOMOFO_LETTER_GN = 0x312C;
const char32_t BOPOMOFO_LETTER_IH = 0x312D;
const char32_t BOPOMOFO_LETTER_O_WITH_DOT_ABOVE = 0x312E;
const char32_t BOPOMOFO_LETTER_NN = 0x312F;
const char32_t HANGUL_LETTER_KIYEOK = 0x3131;
const char32_t HANGUL_LETTER_SSANGKIYEOK = 0x3132;
const char32_t HANGUL_LETTER_KIYEOK_SIOS = 0x3133;
const char32_t HANGUL_LETTER_NIEUN = 0x3134;
const char32_t HANGUL_LETTER_NIEUN_CIEUC = 0x3135;
const char32_t HANGUL_LETTER_NIEUN_HIEUH = 0x3136;
const char32_t HANGUL_LETTER_TIKEUT = 0x3137;
const char32_t HANGUL_LETTER_SSANGTIKEUT = 0x3138;
const char32_t HANGUL_LETTER_RIEUL = 0x3139;
const char32_t HANGUL_LETTER_RIEUL_KIYEOK = 0x313A;
const char32_t HANGUL_LETTER_RIEUL_MIEUM = 0x313B;
const char32_t HANGUL_LETTER_RIEUL_PIEUP = 0x313C;
const char32_t HANGUL_LETTER_RIEUL_SIOS = 0x313D;
const char32_t HANGUL_LETTER_RIEUL_THIEUTH = 0x313E;
const char32_t HANGUL_LETTER_RIEUL_PHIEUPH = 0x313F;
const char32_t HANGUL_LETTER_RIEUL_HIEUH = 0x3140;
const char32_t HANGUL_LETTER_MIEUM = 0x3141;
const char32_t HANGUL_LETTER_PIEUP = 0x3142;
const char32_t HANGUL_LETTER_SSANGPIEUP = 0x3143;
const char32_t HANGUL_LETTER_PIEUP_SIOS = 0x3144;
const char32_t HANGUL_LETTER_SIOS = 0x3145;
const char32_t HANGUL_LETTER_SSANGSIOS = 0x3146;
const char32_t HANGUL_LETTER_IEUNG = 0x3147;
const char32_t HANGUL_LETTER_CIEUC = 0x3148;
const char32_t HANGUL_LETTER_SSANGCIEUC = 0x3149;
const char32_t HANGUL_LETTER_CHIEUCH = 0x314A;
const char32_t HANGUL_LETTER_KHIEUKH = 0x314B;
const char32_t HANGUL_LETTER_THIEUTH = 0x314C;
const char32_t HANGUL_LETTER_PHIEUPH = 0x314D;
const char32_t HANGUL_LETTER_HIEUH = 0x314E;
const char32_t HANGUL_LETTER_A = 0x314F;
const char32_t HANGUL_LETTER_AE = 0x3150;
const char32_t HANGUL_LETTER_YA = 0x3151;
const char32_t HANGUL_LETTER_YAE = 0x3152;
const char32_t HANGUL_LETTER_EO = 0x3153;
const char32_t HANGUL_LETTER_E = 0x3154;
const char32_t HANGUL_LETTER_YEO = 0x3155;
const char32_t HANGUL_LETTER_YE = 0x3156;
const char32_t HANGUL_LETTER_O = 0x3157;
const char32_t HANGUL_LETTER_WA = 0x3158;
const char32_t HANGUL_LETTER_WAE = 0x3159;
const char32_t HANGUL_LETTER_OE = 0x315A;
const char32_t HANGUL_LETTER_YO = 0x315B;
const char32_t HANGUL_LETTER_U = 0x315C;
const char32_t HANGUL_LETTER_WEO = 0x315D;
const char32_t HANGUL_LETTER_WE = 0x315E;
const char32_t HANGUL_LETTER_WI = 0x315F;
const char32_t HANGUL_LETTER_YU = 0x3160;
const char32_t HANGUL_LETTER_EU = 0x3161;
const char32_t HANGUL_LETTER_YI = 0x3162;
const char32_t HANGUL_LETTER_I = 0x3163;
const char32_t HANGUL_FILLER = 0x3164;
const char32_t HANGUL_LETTER_SSANGNIEUN = 0x3165;
const char32_t HANGUL_LETTER_NIEUN_TIKEUT = 0x3166;
const char32_t HANGUL_LETTER_NIEUN_SIOS = 0x3167;
const char32_t HANGUL_LETTER_NIEUN_PANSIOS = 0x3168;
const char32_t HANGUL_LETTER_RIEUL_KIYEOK_SIOS = 0x3169;
const char32_t HANGUL_LETTER_RIEUL_TIKEUT = 0x316A;
const char32_t HANGUL_LETTER_RIEUL_PIEUP_SIOS = 0x316B;
const char32_t HANGUL_LETTER_RIEUL_PANSIOS = 0x316C;
const char32_t HANGUL_LETTER_RIEUL_YEORINHIEUH = 0x316D;
const char32_t HANGUL_LETTER_MIEUM_PIEUP = 0x316E;
const char32_t HANGUL_LETTER_MIEUM_SIOS = 0x316F;
const char32_t HANGUL_LETTER_MIEUM_PANSIOS = 0x3170;
const char32_t HANGUL_LETTER_KAPYEOUNMIEUM = 0x3171;
const char32_t HANGUL_LETTER_PIEUP_KIYEOK = 0x3172;
const char32_t HANGUL_LETTER_PIEUP_TIKEUT = 0x3173;
const char32_t HANGUL_LETTER_PIEUP_SIOS_KIYEOK = 0x3174;
const char32_t HANGUL_LETTER_PIEUP_SIOS_TIKEUT = 0x3175;
const char32_t HANGUL_LETTER_PIEUP_CIEUC = 0x3176;
const char32_t HANGUL_LETTER_PIEUP_THIEUTH = 0x3177;
const char32_t HANGUL_LETTER_KAPYEOUNPIEUP = 0x3178;
const char32_t HANGUL_LETTER_KAPYEOUNSSANGPIEUP = 0x3179;
const char32_t HANGUL_LETTER_SIOS_KIYEOK = 0x317A;
const char32_t HANGUL_LETTER_SIOS_NIEUN = 0x317B;
const char32_t HANGUL_LETTER_SIOS_TIKEUT = 0x317C;
const char32_t HANGUL_LETTER_SIOS_PIEUP = 0x317D;
const char32_t HANGUL_LETTER_SIOS_CIEUC = 0x317E;
const char32_t HANGUL_LETTER_PANSIOS = 0x317F;
const char32_t HANGUL_LETTER_SSANGIEUNG = 0x3180;
const char32_t HANGUL_LETTER_YESIEUNG = 0x3181;
const char32_t HANGUL_LETTER_YESIEUNG_SIOS = 0x3182;
const char32_t HANGUL_LETTER_YESIEUNG_PANSIOS = 0x3183;
const char32_t HANGUL_LETTER_KAPYEOUNPHIEUPH = 0x3184;
const char32_t HANGUL_LETTER_SSANGHIEUH = 0x3185;
const char32_t HANGUL_LETTER_YEORINHIEUH = 0x3186;
const char32_t HANGUL_LETTER_YO_YA = 0x3187;
const char32_t HANGUL_LETTER_YO_YAE = 0x3188;
const char32_t HANGUL_LETTER_YO_I = 0x3189;
const char32_t HANGUL_LETTER_YU_YEO = 0x318A;
const char32_t HANGUL_LETTER_YU_YE = 0x318B;
const char32_t HANGUL_LETTER_YU_I = 0x318C;
const char32_t HANGUL_LETTER_ARAEA = 0x318D;
const char32_t HANGUL_LETTER_ARAEAE = 0x318E;
const char32_t IDEOGRAPHIC_ANNOTATION_LINKING_MARK = 0x3190;
const char32_t IDEOGRAPHIC_ANNOTATION_REVERSE_MARK = 0x3191;
const char32_t IDEOGRAPHIC_ANNOTATION_ONE_MARK = 0x3192;
const char32_t IDEOGRAPHIC_ANNOTATION_TWO_MARK = 0x3193;
const char32_t IDEOGRAPHIC_ANNOTATION_THREE_MARK = 0x3194;
const char32_t IDEOGRAPHIC_ANNOTATION_FOUR_MARK = 0x3195;
const char32_t IDEOGRAPHIC_ANNOTATION_TOP_MARK = 0x3196;
const char32_t IDEOGRAPHIC_ANNOTATION_MIDDLE_MARK = 0x3197;
const char32_t IDEOGRAPHIC_ANNOTATION_BOTTOM_MARK = 0x3198;
const char32_t IDEOGRAPHIC_ANNOTATION_FIRST_MARK = 0x3199;
const char32_t IDEOGRAPHIC_ANNOTATION_SECOND_MARK = 0x319A;
const char32_t IDEOGRAPHIC_ANNOTATION_THIRD_MARK = 0x319B;
const char32_t IDEOGRAPHIC_ANNOTATION_FOURTH_MARK = 0x319C;
const char32_t IDEOGRAPHIC_ANNOTATION_HEAVEN_MARK = 0x319D;
const char32_t IDEOGRAPHIC_ANNOTATION_EARTH_MARK = 0x319E;
const char32_t IDEOGRAPHIC_ANNOTATION_MAN_MARK = 0x319F;
const char32_t BOPOMOFO_LETTER_BU = 0x31A0;
const char32_t BOPOMOFO_LETTER_ZI = 0x31A1;
const char32_t BOPOMOFO_LETTER_JI = 0x31A2;
const char32_t BOPOMOFO_LETTER_GU = 0x31A3;
const char32_t BOPOMOFO_LETTER_EE = 0x31A4;
const char32_t BOPOMOFO_LETTER_ENN = 0x31A5;
const char32_t BOPOMOFO_LETTER_OO = 0x31A6;
const char32_t BOPOMOFO_LETTER_ONN = 0x31A7;
const char32_t BOPOMOFO_LETTER_IR = 0x31A8;
const char32_t BOPOMOFO_LETTER_ANN = 0x31A9;
const char32_t BOPOMOFO_LETTER_INN = 0x31AA;
const char32_t BOPOMOFO_LETTER_UNN = 0x31AB;
const char32_t BOPOMOFO_LETTER_IM = 0x31AC;
const char32_t BOPOMOFO_LETTER_NGG = 0x31AD;
const char32_t BOPOMOFO_LETTER_AINN = 0x31AE;
const char32_t BOPOMOFO_LETTER_AUNN = 0x31AF;
const char32_t BOPOMOFO_LETTER_AM = 0x31B0;
const char32_t BOPOMOFO_LETTER_OM = 0x31B1;
const char32_t BOPOMOFO_LETTER_ONG = 0x31B2;
const char32_t BOPOMOFO_LETTER_INNN = 0x31B3;
const char32_t BOPOMOFO_FINAL_LETTER_P = 0x31B4;
const char32_t BOPOMOFO_FINAL_LETTER_T = 0x31B5;
const char32_t BOPOMOFO_FINAL_LETTER_K = 0x31B6;
const char32_t BOPOMOFO_FINAL_LETTER_H = 0x31B7;
const char32_t BOPOMOFO_LETTER_GH = 0x31B8;
const char32_t BOPOMOFO_LETTER_LH = 0x31B9;
const char32_t BOPOMOFO_LETTER_ZY = 0x31BA;
const char32_t BOPOMOFO_FINAL_LETTER_G = 0x31BB;
const char32_t BOPOMOFO_LETTER_GW = 0x31BC;
const char32_t BOPOMOFO_LETTER_KW = 0x31BD;
const char32_t BOPOMOFO_LETTER_OE = 0x31BE;
const char32_t BOPOMOFO_LETTER_AH = 0x31BF;
const char32_t CJK_STROKE_T = 0x31C0;
const char32_t CJK_STROKE_WG = 0x31C1;
const char32_t CJK_STROKE_XG = 0x31C2;
const char32_t CJK_STROKE_BXG = 0x31C3;
const char32_t CJK_STROKE_SW = 0x31C4;
const char32_t CJK_STROKE_HZZ = 0x31C5;
const char32_t CJK_STROKE_HZG = 0x31C6;
const char32_t CJK_STROKE_HP = 0x31C7;
const char32_t CJK_STROKE_HZWG = 0x31C8;
const char32_t CJK_STROKE_SZWG = 0x31C9;
const char32_t CJK_STROKE_HZT = 0x31CA;
const char32_t CJK_STROKE_HZZP = 0x31CB;
const char32_t CJK_STROKE_HPWG = 0x31CC;
const char32_t CJK_STROKE_HZW = 0x31CD;
const char32_t CJK_STROKE_HZZZ = 0x31CE;
const char32_t CJK_STROKE_N = 0x31CF;
const char32_t CJK_STROKE_H = 0x31D0;
const char32_t CJK_STROKE_S = 0x31D1;
const char32_t CJK_STROKE_P = 0x31D2;
const char32_t CJK_STROKE_SP = 0x31D3;
const char32_t CJK_STROKE_D = 0x31D4;
const char32_t CJK_STROKE_HZ = 0x31D5;
const char32_t CJK_STROKE_HG = 0x31D6;
const char32_t CJK_STROKE_SZ = 0x31D7;
const char32_t CJK_STROKE_SWZ = 0x31D8;
const char32_t CJK_STROKE_ST = 0x31D9;
const char32_t CJK_STROKE_SG = 0x31DA;
const char32_t CJK_STROKE_PD = 0x31DB;
const char32_t CJK_STROKE_PZ = 0x31DC;
const char32_t CJK_STROKE_TN = 0x31DD;
const char32_t CJK_STROKE_SZZ = 0x31DE;
const char32_t CJK_STROKE_SWG = 0x31DF;
const char32_t CJK_STROKE_HXWG = 0x31E0;
const char32_t CJK_STROKE_HZZZG = 0x31E1;
const char32_t CJK_STROKE_PG = 0x31E2;
const char32_t CJK_STROKE_Q = 0x31E3;
const char32_t CJK_STROKE_HXG = 0x31E4;
const char32_t CJK_STROKE_SZP = 0x31E5;
const char32_t IDEOGRAPHIC_DESCRIPTION_CHARACTER_SUBTRACTION = 0x31EF;
const char32_t KATAKANA_LETTER_SMALL_KU = 0x31F0;
const char32_t KATAKANA_LETTER_SMALL_SI = 0x31F1;
const char32_t KATAKANA_LETTER_SMALL_SU = 0x31F2;
const char32_t KATAKANA_LETTER_SMALL_TO = 0x31F3;
const char32_t KATAKANA_LETTER_SMALL_NU = 0x31F4;
const char32_t KATAKANA_LETTER_SMALL_HA = 0x31F5;
const char32_t KATAKANA_LETTER_SMALL_HI = 0x31F6;
const char32_t KATAKANA_LETTER_SMALL_HU = 0x31F7;
const char32_t KATAKANA_LETTER_SMALL_HE = 0x31F8;
const char32_t KATAKANA_LETTER_SMALL_HO = 0x31F9;
const char32_t KATAKANA_LETTER_SMALL_MU = 0x31FA;
const char32_t KATAKANA_LETTER_SMALL_RA = 0x31FB;
const char32_t KATAKANA_LETTER_SMALL_RI = 0x31FC;
const char32_t KATAKANA_LETTER_SMALL_RU = 0x31FD;
const char32_t KATAKANA_LETTER_SMALL_RE = 0x31FE;
const char32_t KATAKANA_LETTER_SMALL_RO = 0x31FF;
const char32_t PARENTHESIZED_HANGUL_KIYEOK = 0x3200;
const char32_t PARENTHESIZED_HANGUL_NIEUN = 0x3201;
const char32_t PARENTHESIZED_HANGUL_TIKEUT = 0x3202;
const char32_t PARENTHESIZED_HANGUL_RIEUL = 0x3203;
const char32_t PARENTHESIZED_HANGUL_MIEUM = 0x3204;
const char32_t PARENTHESIZED_HANGUL_PIEUP = 0x3205;
const char32_t PARENTHESIZED_HANGUL_SIOS = 0x3206;
const char32_t PARENTHESIZED_HANGUL_IEUNG = 0x3207;
const char32_t PARENTHESIZED_HANGUL_CIEUC = 0x3208;
const char32_t PARENTHESIZED_HANGUL_CHIEUCH = 0x3209;
const char32_t PARENTHESIZED_HANGUL_KHIEUKH = 0x320A;
const char32_t PARENTHESIZED_HANGUL_THIEUTH = 0x320B;
const char32_t PARENTHESIZED_HANGUL_PHIEUPH = 0x320C;
const char32_t PARENTHESIZED_HANGUL_HIEUH = 0x320D;
const char32_t PARENTHESIZED_HANGUL_KIYEOK_A = 0x320E;
const char32_t PARENTHESIZED_HANGUL_NIEUN_A = 0x320F;
const char32_t PARENTHESIZED_HANGUL_TIKEUT_A = 0x3210;
const char32_t PARENTHESIZED_HANGUL_RIEUL_A = 0x3211;
const char32_t PARENTHESIZED_HANGUL_MIEUM_A = 0x3212;
const char32_t PARENTHESIZED_HANGUL_PIEUP_A = 0x3213;
const char32_t PARENTHESIZED_HANGUL_SIOS_A = 0x3214;
const char32_t PARENTHESIZED_HANGUL_IEUNG_A = 0x3215;
const char32_t PARENTHESIZED_HANGUL_CIEUC_A = 0x3216;
const char32_t PARENTHESIZED_HANGUL_CHIEUCH_A = 0x3217;
const char32_t PARENTHESIZED_HANGUL_KHIEUKH_A = 0x3218;
const char32_t PARENTHESIZED_HANGUL_THIEUTH_A = 0x3219;
const char32_t PARENTHESIZED_HANGUL_PHIEUPH_A = 0x321A;
const char32_t PARENTHESIZED_HANGUL_HIEUH_A = 0x321B;
const char32_t PARENTHESIZED_HANGUL_CIEUC_U = 0x321C;
const char32_t PARENTHESIZED_KOREAN_CHARACTER_OJEON = 0x321D;
const char32_t PARENTHESIZED_KOREAN_CHARACTER_O_HU = 0x321E;
const char32_t PARENTHESIZED_IDEOGRAPH_ONE = 0x3220;
const char32_t PARENTHESIZED_IDEOGRAPH_TWO = 0x3221;
const char32_t PARENTHESIZED_IDEOGRAPH_THREE = 0x3222;
const char32_t PARENTHESIZED_IDEOGRAPH_FOUR = 0x3223;
const char32_t PARENTHESIZED_IDEOGRAPH_FIVE = 0x3224;
const char32_t PARENTHESIZED_IDEOGRAPH_SIX = 0x3225;
const char32_t PARENTHESIZED_IDEOGRAPH_SEVEN = 0x3226;
const char32_t PARENTHESIZED_IDEOGRAPH_EIGHT = 0x3227;
const char32_t PARENTHESIZED_IDEOGRAPH_NINE = 0x3228;
const char32_t PARENTHESIZED_IDEOGRAPH_TEN = 0x3229;
const char32_t PARENTHESIZED_IDEOGRAPH_MOON = 0x322A;
const char32_t PARENTHESIZED_IDEOGRAPH_FIRE = 0x322B;
const char32_t PARENTHESIZED_IDEOGRAPH_WATER = 0x322C;
const char32_t PARENTHESIZED_IDEOGRAPH_WOOD = 0x322D;
const char32_t PARENTHESIZED_IDEOGRAPH_METAL = 0x322E;
const char32_t PARENTHESIZED_IDEOGRAPH_EARTH = 0x322F;
const char32_t PARENTHESIZED_IDEOGRAPH_SUN = 0x3230;
const char32_t PARENTHESIZED_IDEOGRAPH_STOCK = 0x3231;
const char32_t PARENTHESIZED_IDEOGRAPH_HAVE = 0x3232;
const char32_t PARENTHESIZED_IDEOGRAPH_SOCIETY = 0x3233;
const char32_t PARENTHESIZED_IDEOGRAPH_NAME = 0x3234;
const char32_t PARENTHESIZED_IDEOGRAPH_SPECIAL = 0x3235;
const char32_t PARENTHESIZED_IDEOGRAPH_FINANCIAL = 0x3236;
const char32_t PARENTHESIZED_IDEOGRAPH_CONGRATULATION = 0x3237;
const char32_t PARENTHESIZED_IDEOGRAPH_LABOR = 0x3238;
const char32_t PARENTHESIZED_IDEOGRAPH_REPRESENT = 0x3239;
const char32_t PARENTHESIZED_IDEOGRAPH_CALL = 0x323A;
const char32_t PARENTHESIZED_IDEOGRAPH_STUDY = 0x323B;
const char32_t PARENTHESIZED_IDEOGRAPH_SUPERVISE = 0x323C;
const char32_t PARENTHESIZED_IDEOGRAPH_ENTERPRISE = 0x323D;
const char32_t PARENTHESIZED_IDEOGRAPH_RESOURCE = 0x323E;
const char32_t PARENTHESIZED_IDEOGRAPH_ALLIANCE = 0x323F;
const char32_t PARENTHESIZED_IDEOGRAPH_FESTIVAL = 0x3240;
const char32_t PARENTHESIZED_IDEOGRAPH_REST = 0x3241;
const char32_t PARENTHESIZED_IDEOGRAPH_SELF = 0x3242;
const char32_t PARENTHESIZED_IDEOGRAPH_REACH = 0x3243;
const char32_t CIRCLED_IDEOGRAPH_QUESTION = 0x3244;
const char32_t CIRCLED_IDEOGRAPH_KINDERGARTEN = 0x3245;
const char32_t CIRCLED_IDEOGRAPH_SCHOOL = 0x3246;
const char32_t CIRCLED_IDEOGRAPH_KOTO = 0x3247;
const char32_t CIRCLED_NUMBER_TEN_ON_BLACK_SQUARE = 0x3248;
const char32_t CIRCLED_NUMBER_TWENTY_ON_BLACK_SQUARE = 0x3249;
const char32_t CIRCLED_NUMBER_THIRTY_ON_BLACK_SQUARE = 0x324A;
const char32_t CIRCLED_NUMBER_FORTY_ON_BLACK_SQUARE = 0x324B;
const char32_t CIRCLED_NUMBER_FIFTY_ON_BLACK_SQUARE = 0x324C;
const char32_t CIRCLED_NUMBER_SIXTY_ON_BLACK_SQUARE = 0x324D;
const char32_t CIRCLED_NUMBER_SEVENTY_ON_BLACK_SQUARE = 0x324E;
const char32_t CIRCLED_NUMBER_EIGHTY_ON_BLACK_SQUARE = 0x324F;
const char32_t PARTNERSHIP_SIGN = 0x3250;
const char32_t CIRCLED_NUMBER_TWENTY_ONE = 0x3251;
const char32_t CIRCLED_NUMBER_TWENTY_TWO = 0x3252;
const char32_t CIRCLED_NUMBER_TWENTY_THREE = 0x3253;
const char32_t CIRCLED_NUMBER_TWENTY_FOUR = 0x3254;
const char32_t CIRCLED_NUMBER_TWENTY_FIVE = 0x3255;
const char32_t CIRCLED_NUMBER_TWENTY_SIX = 0x3256;
const char32_t CIRCLED_NUMBER_TWENTY_SEVEN = 0x3257;
const char32_t CIRCLED_NUMBER_TWENTY_EIGHT = 0x3258;
const char32_t CIRCLED_NUMBER_TWENTY_NINE = 0x3259;
const char32_t CIRCLED_NUMBER_THIRTY = 0x325A;
const char32_t CIRCLED_NUMBER_THIRTY_ONE = 0x325B;
const char32_t CIRCLED_NUMBER_THIRTY_TWO = 0x325C;
const char32_t CIRCLED_NUMBER_THIRTY_THREE = 0x325D;
const char32_t CIRCLED_NUMBER_THIRTY_FOUR = 0x325E;
const char32_t CIRCLED_NUMBER_THIRTY_FIVE = 0x325F;
const char32_t CIRCLED_HANGUL_KIYEOK = 0x3260;
const char32_t CIRCLED_HANGUL_NIEUN = 0x3261;
const char32_t CIRCLED_HANGUL_TIKEUT = 0x3262;
const char32_t CIRCLED_HANGUL_RIEUL = 0x3263;
const char32_t CIRCLED_HANGUL_MIEUM = 0x3264;
const char32_t CIRCLED_HANGUL_PIEUP = 0x3265;
const char32_t CIRCLED_HANGUL_SIOS = 0x3266;
const char32_t CIRCLED_HANGUL_IEUNG = 0x3267;
const char32_t CIRCLED_HANGUL_CIEUC = 0x3268;
const char32_t CIRCLED_HANGUL_CHIEUCH = 0x3269;
const char32_t CIRCLED_HANGUL_KHIEUKH = 0x326A;
const char32_t CIRCLED_HANGUL_THIEUTH = 0x326B;
const char32_t CIRCLED_HANGUL_PHIEUPH = 0x326C;
const char32_t CIRCLED_HANGUL_HIEUH = 0x326D;
const char32_t CIRCLED_HANGUL_KIYEOK_A = 0x326E;
const char32_t CIRCLED_HANGUL_NIEUN_A = 0x326F;
const char32_t CIRCLED_HANGUL_TIKEUT_A = 0x3270;
const char32_t CIRCLED_HANGUL_RIEUL_A = 0x3271;
const char32_t CIRCLED_HANGUL_MIEUM_A = 0x3272;
const char32_t CIRCLED_HANGUL_PIEUP_A = 0x3273;
const char32_t CIRCLED_HANGUL_SIOS_A = 0x3274;
const char32_t CIRCLED_HANGUL_IEUNG_A = 0x3275;
const char32_t CIRCLED_HANGUL_CIEUC_A = 0x3276;
const char32_t CIRCLED_HANGUL_CHIEUCH_A = 0x3277;
const char32_t CIRCLED_HANGUL_KHIEUKH_A = 0x3278;
const char32_t CIRCLED_HANGUL_THIEUTH_A = 0x3279;
const char32_t CIRCLED_HANGUL_PHIEUPH_A = 0x327A;
const char32_t CIRCLED_HANGUL_HIEUH_A = 0x327B;
const char32_t CIRCLED_KOREAN_CHARACTER_CHAMKO = 0x327C;
const char32_t CIRCLED_KOREAN_CHARACTER_JUEUI = 0x327D;
const char32_t CIRCLED_HANGUL_IEUNG_U = 0x327E;
const char32_t KOREAN_STANDARD_SYMBOL = 0x327F;
const char32_t CIRCLED_IDEOGRAPH_ONE = 0x3280;
const char32_t CIRCLED_IDEOGRAPH_TWO = 0x3281;
const char32_t CIRCLED_IDEOGRAPH_THREE = 0x3282;
const char32_t CIRCLED_IDEOGRAPH_FOUR = 0x3283;
const char32_t CIRCLED_IDEOGRAPH_FIVE = 0x3284;
const char32_t CIRCLED_IDEOGRAPH_SIX = 0x3285;
const char32_t CIRCLED_IDEOGRAPH_SEVEN = 0x3286;
const char32_t CIRCLED_IDEOGRAPH_EIGHT = 0x3287;
const char32_t CIRCLED_IDEOGRAPH_NINE = 0x3288;
const char32_t CIRCLED_IDEOGRAPH_TEN = 0x3289;
const char32_t CIRCLED_IDEOGRAPH_MOON = 0x328A;
const char32_t CIRCLED_IDEOGRAPH_FIRE = 0x328B;
const char32_t CIRCLED_IDEOGRAPH_WATER = 0x328C;
const char32_t CIRCLED_IDEOGRAPH_WOOD = 0x328D;
const char32_t CIRCLED_IDEOGRAPH_METAL = 0x328E;
const char32_t CIRCLED_IDEOGRAPH_EARTH = 0x328F;
const char32_t CIRCLED_IDEOGRAPH_SUN = 0x3290;
const char32_t CIRCLED_IDEOGRAPH_STOCK = 0x3291;
const char32_t CIRCLED_IDEOGRAPH_HAVE = 0x3292;
const char32_t CIRCLED_IDEOGRAPH_SOCIETY = 0x3293;
const char32_t CIRCLED_IDEOGRAPH_NAME = 0x3294;
const char32_t CIRCLED_IDEOGRAPH_SPECIAL = 0x3295;
const char32_t CIRCLED_IDEOGRAPH_FINANCIAL = 0x3296;
const char32_t CIRCLED_IDEOGRAPH_CONGRATULATION = 0x3297;
const char32_t CIRCLED_IDEOGRAPH_LABOR = 0x3298;
const char32_t CIRCLED_IDEOGRAPH_SECRET = 0x3299;
const char32_t CIRCLED_IDEOGRAPH_MALE = 0x329A;
const char32_t CIRCLED_IDEOGRAPH_FEMALE = 0x329B;
const char32_t CIRCLED_IDEOGRAPH_SUITABLE = 0x329C;
const char32_t CIRCLED_IDEOGRAPH_EXCELLENT = 0x329D;
const char32_t CIRCLED_IDEOGRAPH_PRINT = 0x329E;
const char32_t CIRCLED_IDEOGRAPH_ATTENTION = 0x329F;
const char32_t CIRCLED_IDEOGRAPH_ITEM = 0x32A0;
const char32_t CIRCLED_IDEOGRAPH_REST = 0x32A1;
const char32_t CIRCLED_IDEOGRAPH_COPY = 0x32A2;
const char32_t CIRCLED_IDEOGRAPH_CORRECT = 0x32A3;
const char32_t CIRCLED_IDEOGRAPH_HIGH = 0x32A4;
const char32_t CIRCLED_IDEOGRAPH_CENTRE = 0x32A5;
const char32_t CIRCLED_IDEOGRAPH_LOW = 0x32A6;
const char32_t CIRCLED_IDEOGRAPH_LEFT = 0x32A7;
const char32_t CIRCLED_IDEOGRAPH_RIGHT = 0x32A8;
const char32_t CIRCLED_IDEOGRAPH_MEDICINE = 0x32A9;
const char32_t CIRCLED_IDEOGRAPH_RELIGION = 0x32AA;
const char32_t CIRCLED_IDEOGRAPH_STUDY = 0x32AB;
const char32_t CIRCLED_IDEOGRAPH_SUPERVISE = 0x32AC;
const char32_t CIRCLED_IDEOGRAPH_ENTERPRISE = 0x32AD;
const char32_t CIRCLED_IDEOGRAPH_RESOURCE = 0x32AE;
const char32_t CIRCLED_IDEOGRAPH_ALLIANCE = 0x32AF;
const char32_t CIRCLED_IDEOGRAPH_NIGHT = 0x32B0;
const char32_t CIRCLED_NUMBER_THIRTY_SIX = 0x32B1;
const char32_t CIRCLED_NUMBER_THIRTY_SEVEN = 0x32B2;
const char32_t CIRCLED_NUMBER_THIRTY_EIGHT = 0x32B3;
const char32_t CIRCLED_NUMBER_THIRTY_NINE = 0x32B4;
const char32_t CIRCLED_NUMBER_FORTY = 0x32B5;
const char32_t CIRCLED_NUMBER_FORTY_ONE = 0x32B6;
const char32_t CIRCLED_NUMBER_FORTY_TWO = 0x32B7;
const char32_t CIRCLED_NUMBER_FORTY_THREE = 0x32B8;
const char32_t CIRCLED_NUMBER_FORTY_FOUR = 0x32B9;
const char32_t CIRCLED_NUMBER_FORTY_FIVE = 0x32BA;
const char32_t CIRCLED_NUMBER_FORTY_SIX = 0x32BB;
const char32_t CIRCLED_NUMBER_FORTY_SEVEN = 0x32BC;
const char32_t CIRCLED_NUMBER_FORTY_EIGHT = 0x32BD;
const char32_t CIRCLED_NUMBER_FORTY_NINE = 0x32BE;
const char32_t CIRCLED_NUMBER_FIFTY = 0x32BF;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_JANUARY = 0x32C0;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_FEBRUARY = 0x32C1;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_MARCH = 0x32C2;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_APRIL = 0x32C3;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_MAY = 0x32C4;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_JUNE = 0x32C5;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_JULY = 0x32C6;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_AUGUST = 0x32C7;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_SEPTEMBER = 0x32C8;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_OCTOBER = 0x32C9;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_NOVEMBER = 0x32CA;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DECEMBER = 0x32CB;
const char32_t SQUARE_HG = 0x32CC;
const char32_t SQUARE_ERG = 0x32CD;
const char32_t SQUARE_EV = 0x32CE;
const char32_t LIMITED_LIABILITY_SIGN = 0x32CF;
const char32_t CIRCLED_KATAKANA_A = 0x32D0;
const char32_t CIRCLED_KATAKANA_I = 0x32D1;
const char32_t CIRCLED_KATAKANA_U = 0x32D2;
const char32_t CIRCLED_KATAKANA_E = 0x32D3;
const char32_t CIRCLED_KATAKANA_O = 0x32D4;
const char32_t CIRCLED_KATAKANA_KA = 0x32D5;
const char32_t CIRCLED_KATAKANA_KI = 0x32D6;
const char32_t CIRCLED_KATAKANA_KU = 0x32D7;
const char32_t CIRCLED_KATAKANA_KE = 0x32D8;
const char32_t CIRCLED_KATAKANA_KO = 0x32D9;
const char32_t CIRCLED_KATAKANA_SA = 0x32DA;
const char32_t CIRCLED_KATAKANA_SI = 0x32DB;
const char32_t CIRCLED_KATAKANA_SU = 0x32DC;
const char32_t CIRCLED_KATAKANA_SE = 0x32DD;
const char32_t CIRCLED_KATAKANA_SO = 0x32DE;
const char32_t CIRCLED_KATAKANA_TA = 0x32DF;
const char32_t CIRCLED_KATAKANA_TI = 0x32E0;
const char32_t CIRCLED_KATAKANA_TU = 0x32E1;
const char32_t CIRCLED_KATAKANA_TE = 0x32E2;
const char32_t CIRCLED_KATAKANA_TO = 0x32E3;
const char32_t CIRCLED_KATAKANA_NA = 0x32E4;
const char32_t CIRCLED_KATAKANA_NI = 0x32E5;
const char32_t CIRCLED_KATAKANA_NU = 0x32E6;
const char32_t CIRCLED_KATAKANA_NE = 0x32E7;
const char32_t CIRCLED_KATAKANA_NO = 0x32E8;
const char32_t CIRCLED_KATAKANA_HA = 0x32E9;
const char32_t CIRCLED_KATAKANA_HI = 0x32EA;
const char32_t CIRCLED_KATAKANA_HU = 0x32EB;
const char32_t CIRCLED_KATAKANA_HE = 0x32EC;
const char32_t CIRCLED_KATAKANA_HO = 0x32ED;
const char32_t CIRCLED_KATAKANA_MA = 0x32EE;
const char32_t CIRCLED_KATAKANA_MI = 0x32EF;
const char32_t CIRCLED_KATAKANA_MU = 0x32F0;
const char32_t CIRCLED_KATAKANA_ME = 0x32F1;
const char32_t CIRCLED_KATAKANA_MO = 0x32F2;
const char32_t CIRCLED_KATAKANA_YA = 0x32F3;
const char32_t CIRCLED_KATAKANA_YU = 0x32F4;
const char32_t CIRCLED_KATAKANA_YO = 0x32F5;
const char32_t CIRCLED_KATAKANA_RA = 0x32F6;
const char32_t CIRCLED_KATAKANA_RI = 0x32F7;
const char32_t CIRCLED_KATAKANA_RU = 0x32F8;
const char32_t CIRCLED_KATAKANA_RE = 0x32F9;
const char32_t CIRCLED_KATAKANA_RO = 0x32FA;
const char32_t CIRCLED_KATAKANA_WA = 0x32FB;
const char32_t CIRCLED_KATAKANA_WI = 0x32FC;
const char32_t CIRCLED_KATAKANA_WE = 0x32FD;
const char32_t CIRCLED_KATAKANA_WO = 0x32FE;
const char32_t SQUARE_ERA_NAME_REIWA = 0x32FF;
const char32_t SQUARE_APAATO = 0x3300;
const char32_t SQUARE_ARUHUA = 0x3301;
const char32_t SQUARE_ANPEA = 0x3302;
const char32_t SQUARE_AARU = 0x3303;
const char32_t SQUARE_ININGU = 0x3304;
const char32_t SQUARE_INTI = 0x3305;
const char32_t SQUARE_UON = 0x3306;
const char32_t SQUARE_ESUKUUDO = 0x3307;
const char32_t SQUARE_EEKAA = 0x3308;
const char32_t SQUARE_ONSU = 0x3309;
const char32_t SQUARE_OOMU = 0x330A;
const char32_t SQUARE_KAIRI = 0x330B;
const char32_t SQUARE_KARATTO = 0x330C;
const char32_t SQUARE_KARORII = 0x330D;
const char32_t SQUARE_GARON = 0x330E;
const char32_t SQUARE_GANMA = 0x330F;
const char32_t SQUARE_GIGA = 0x3310;
const char32_t SQUARE_GINII = 0x3311;
const char32_t SQUARE_KYURII = 0x3312;
const char32_t SQUARE_GIRUDAA = 0x3313;
const char32_t SQUARE_KIRO = 0x3314;
const char32_t SQUARE_KIROGURAMU = 0x3315;
const char32_t SQUARE_KIROMEETORU = 0x3316;
const char32_t SQUARE_KIROWATTO = 0x3317;
const char32_t SQUARE_GURAMU = 0x3318;
const char32_t SQUARE_GURAMUTON = 0x3319;
const char32_t SQUARE_KURUZEIRO = 0x331A;
const char32_t SQUARE_KUROONE = 0x331B;
const char32_t SQUARE_KEESU = 0x331C;
const char32_t SQUARE_KORUNA = 0x331D;
const char32_t SQUARE_KOOPO = 0x331E;
const char32_t SQUARE_SAIKURU = 0x331F;
const char32_t SQUARE_SANTIIMU = 0x3320;
const char32_t SQUARE_SIRINGU = 0x3321;
const char32_t SQUARE_SENTI = 0x3322;
const char32_t SQUARE_SENTO = 0x3323;
const char32_t SQUARE_DAASU = 0x3324;
const char32_t SQUARE_DESI = 0x3325;
const char32_t SQUARE_DORU = 0x3326;
const char32_t SQUARE_TON = 0x3327;
const char32_t SQUARE_NANO = 0x3328;
const char32_t SQUARE_NOTTO = 0x3329;
const char32_t SQUARE_HAITU = 0x332A;
const char32_t SQUARE_PAASENTO = 0x332B;
const char32_t SQUARE_PAATU = 0x332C;
const char32_t SQUARE_BAARERU = 0x332D;
const char32_t SQUARE_PIASUTORU = 0x332E;
const char32_t SQUARE_PIKURU = 0x332F;
const char32_t SQUARE_PIKO = 0x3330;
const char32_t SQUARE_BIRU = 0x3331;
const char32_t SQUARE_HUARADDO = 0x3332;
const char32_t SQUARE_HUIITO = 0x3333;
const char32_t SQUARE_BUSSYERU = 0x3334;
const char32_t SQUARE_HURAN = 0x3335;
const char32_t SQUARE_HEKUTAARU = 0x3336;
const char32_t SQUARE_PESO = 0x3337;
const char32_t SQUARE_PENIHI = 0x3338;
const char32_t SQUARE_HERUTU = 0x3339;
const char32_t SQUARE_PENSU = 0x333A;
const char32_t SQUARE_PEEZI = 0x333B;
const char32_t SQUARE_BEETA = 0x333C;
const char32_t SQUARE_POINTO = 0x333D;
const char32_t SQUARE_BORUTO = 0x333E;
const char32_t SQUARE_HON = 0x333F;
const char32_t SQUARE_PONDO = 0x3340;
const char32_t SQUARE_HOORU = 0x3341;
const char32_t SQUARE_HOON = 0x3342;
const char32_t SQUARE_MAIKURO = 0x3343;
const char32_t SQUARE_MAIRU = 0x3344;
const char32_t SQUARE_MAHHA = 0x3345;
const char32_t SQUARE_MARUKU = 0x3346;
const char32_t SQUARE_MANSYON = 0x3347;
const char32_t SQUARE_MIKURON = 0x3348;
const char32_t SQUARE_MIRI = 0x3349;
const char32_t SQUARE_MIRIBAARU = 0x334A;
const char32_t SQUARE_MEGA = 0x334B;
const char32_t SQUARE_MEGATON = 0x334C;
const char32_t SQUARE_MEETORU = 0x334D;
const char32_t SQUARE_YAADO = 0x334E;
const char32_t SQUARE_YAARU = 0x334F;
const char32_t SQUARE_YUAN = 0x3350;
const char32_t SQUARE_RITTORU = 0x3351;
const char32_t SQUARE_RIRA = 0x3352;
const char32_t SQUARE_RUPII = 0x3353;
const char32_t SQUARE_RUUBURU = 0x3354;
const char32_t SQUARE_REMU = 0x3355;
const char32_t SQUARE_RENTOGEN = 0x3356;
const char32_t SQUARE_WATTO = 0x3357;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_ZERO = 0x3358;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_ONE = 0x3359;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TWO = 0x335A;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_THREE = 0x335B;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_FOUR = 0x335C;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_FIVE = 0x335D;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_SIX = 0x335E;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_SEVEN = 0x335F;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_EIGHT = 0x3360;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_NINE = 0x3361;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TEN = 0x3362;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_ELEVEN = 0x3363;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TWELVE = 0x3364;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_THIRTEEN = 0x3365;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_FOURTEEN = 0x3366;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_FIFTEEN = 0x3367;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_SIXTEEN = 0x3368;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_SEVENTEEN = 0x3369;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_EIGHTEEN = 0x336A;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_NINETEEN = 0x336B;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TWENTY = 0x336C;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TWENTY_ONE = 0x336D;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TWENTY_TWO = 0x336E;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TWENTY_THREE = 0x336F;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_HOUR_TWENTY_FOUR = 0x3370;
const char32_t SQUARE_HPA = 0x3371;
const char32_t SQUARE_DA = 0x3372;
const char32_t SQUARE_AU = 0x3373;
const char32_t SQUARE_BAR = 0x3374;
const char32_t SQUARE_OV = 0x3375;
const char32_t SQUARE_PC = 0x3376;
const char32_t SQUARE_DM = 0x3377;
const char32_t SQUARE_DM_SQUARED = 0x3378;
const char32_t SQUARE_DM_CUBED = 0x3379;
const char32_t SQUARE_IU = 0x337A;
const char32_t SQUARE_ERA_NAME_HEISEI = 0x337B;
const char32_t SQUARE_ERA_NAME_SYOUWA = 0x337C;
const char32_t SQUARE_ERA_NAME_TAISYOU = 0x337D;
const char32_t SQUARE_ERA_NAME_MEIZI = 0x337E;
const char32_t SQUARE_CORPORATION = 0x337F;
const char32_t SQUARE_PA_AMPS = 0x3380;
const char32_t SQUARE_NA = 0x3381;
const char32_t SQUARE_MU_A = 0x3382;
const char32_t SQUARE_MA = 0x3383;
const char32_t SQUARE_KA = 0x3384;
const char32_t SQUARE_KB = 0x3385;
const char32_t SQUARE_MB = 0x3386;
const char32_t SQUARE_GB = 0x3387;
const char32_t SQUARE_CAL = 0x3388;
const char32_t SQUARE_KCAL = 0x3389;
const char32_t SQUARE_PF = 0x338A;
const char32_t SQUARE_NF = 0x338B;
const char32_t SQUARE_MU_F = 0x338C;
const char32_t SQUARE_MU_G = 0x338D;
const char32_t SQUARE_MG = 0x338E;
const char32_t SQUARE_KG = 0x338F;
const char32_t SQUARE_HZ = 0x3390;
const char32_t SQUARE_KHZ = 0x3391;
const char32_t SQUARE_MHZ = 0x3392;
const char32_t SQUARE_GHZ = 0x3393;
const char32_t SQUARE_THZ = 0x3394;
const char32_t SQUARE_MU_L = 0x3395;
const char32_t SQUARE_ML = 0x3396;
const char32_t SQUARE_DL = 0x3397;
const char32_t SQUARE_KL = 0x3398;
const char32_t SQUARE_FM = 0x3399;
const char32_t SQUARE_NM = 0x339A;
const char32_t SQUARE_MU_M = 0x339B;
const char32_t SQUARE_MM = 0x339C;
const char32_t SQUARE_CM = 0x339D;
const char32_t SQUARE_KM = 0x339E;
const char32_t SQUARE_MM_SQUARED = 0x339F;
const char32_t SQUARE_CM_SQUARED = 0x33A0;
const char32_t SQUARE_M_SQUARED = 0x33A1;
const char32_t SQUARE_KM_SQUARED = 0x33A2;
const char32_t SQUARE_MM_CUBED = 0x33A3;
const char32_t SQUARE_CM_CUBED = 0x33A4;
const char32_t SQUARE_M_CUBED = 0x33A5;
const char32_t SQUARE_KM_CUBED = 0x33A6;
const char32_t SQUARE_M_OVER_S = 0x33A7;
const char32_t SQUARE_M_OVER_S_SQUARED = 0x33A8;
const char32_t SQUARE_PA = 0x33A9;
const char32_t SQUARE_KPA = 0x33AA;
const char32_t SQUARE_MPA = 0x33AB;
const char32_t SQUARE_GPA = 0x33AC;
const char32_t SQUARE_RAD = 0x33AD;
const char32_t SQUARE_RAD_OVER_S = 0x33AE;
const char32_t SQUARE_RAD_OVER_S_SQUARED = 0x33AF;
const char32_t SQUARE_PS = 0x33B0;
const char32_t SQUARE_NS = 0x33B1;
const char32_t SQUARE_MU_S = 0x33B2;
const char32_t SQUARE_MS = 0x33B3;
const char32_t SQUARE_PV = 0x33B4;
const char32_t SQUARE_NV = 0x33B5;
const char32_t SQUARE_MU_V = 0x33B6;
const char32_t SQUARE_MV = 0x33B7;
const char32_t SQUARE_KV = 0x33B8;
const char32_t SQUARE_MV_MEGA = 0x33B9;
const char32_t SQUARE_PW = 0x33BA;
const char32_t SQUARE_NW = 0x33BB;
const char32_t SQUARE_MU_W = 0x33BC;
const char32_t SQUARE_MW = 0x33BD;
const char32_t SQUARE_KW = 0x33BE;
const char32_t SQUARE_MW_MEGA = 0x33BF;
const char32_t SQUARE_K_OHM = 0x33C0;
const char32_t SQUARE_M_OHM = 0x33C1;
const char32_t SQUARE_AM = 0x33C2;
const char32_t SQUARE_BQ = 0x33C3;
const char32_t SQUARE_CC = 0x33C4;
const char32_t SQUARE_CD = 0x33C5;
const char32_t SQUARE_C_OVER_KG = 0x33C6;
const char32_t SQUARE_CO = 0x33C7;
const char32_t SQUARE_DB = 0x33C8;
const char32_t SQUARE_GY = 0x33C9;
const char32_t SQUARE_HA = 0x33CA;
const char32_t SQUARE_HP = 0x33CB;
const char32_t SQUARE_IN = 0x33CC;
const char32_t SQUARE_KK = 0x33CD;
const char32_t SQUARE_KM_CAPITAL = 0x33CE;
const char32_t SQUARE_KT = 0x33CF;
const char32_t SQUARE_LM = 0x33D0;
const char32_t SQUARE_LN = 0x33D1;
const char32_t SQUARE_LOG = 0x33D2;
const char32_t SQUARE_LX = 0x33D3;
const char32_t SQUARE_MB_SMALL = 0x33D4;
const char32_t SQUARE_MIL = 0x33D5;
const char32_t SQUARE_MOL = 0x33D6;
const char32_t SQUARE_PH = 0x33D7;
const char32_t SQUARE_PM = 0x33D8;
const char32_t SQUARE_PPM = 0x33D9;
const char32_t SQUARE_PR = 0x33DA;
const char32_t SQUARE_SR = 0x33DB;
const char32_t SQUARE_SV = 0x33DC;
const char32_t SQUARE_WB = 0x33DD;
const char32_t SQUARE_V_OVER_M = 0x33DE;
const char32_t SQUARE_A_OVER_M = 0x33DF;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_ONE = 0x33E0;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWO = 0x33E1;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_THREE = 0x33E2;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_FOUR = 0x33E3;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_FIVE = 0x33E4;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_SIX = 0x33E5;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_SEVEN = 0x33E6;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_EIGHT = 0x33E7;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_NINE = 0x33E8;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TEN = 0x33E9;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_ELEVEN = 0x33EA;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWELVE = 0x33EB;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_THIRTEEN = 0x33EC;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_FOURTEEN = 0x33ED;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_FIFTEEN = 0x33EE;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_SIXTEEN = 0x33EF;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_SEVENTEEN = 0x33F0;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_EIGHTEEN = 0x33F1;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_NINETEEN = 0x33F2;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY = 0x33F3;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_ONE = 0x33F4;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_TWO = 0x33F5;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_THREE = 0x33F6;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_FOUR = 0x33F7;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_FIVE = 0x33F8;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_SIX = 0x33F9;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_SEVEN = 0x33FA;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_EIGHT = 0x33FB;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_TWENTY_NINE = 0x33FC;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_THIRTY = 0x33FD;
const char32_t IDEOGRAPHIC_TELEGRAPH_SYMBOL_FOR_DAY_THIRTY_ONE = 0x33FE;
const char32_t SQUARE_GAL = 0x33FF;
const char32_t CJK_IDEOGRAPH_EXTENSION_A_FIRST = 0x3400;
const char32_t CJK_IDEOGRAPH_EXTENSION_A_LAST = 0x4DBF;
const char32_t HEXAGRAM_FOR_THE_CREATIVE_HEAVEN = 0x4DC0;
const char32_t HEXAGRAM_FOR_THE_RECEPTIVE_EARTH = 0x4DC1;
const char32_t HEXAGRAM_FOR_DIFFICULTY_AT_THE_BEGINNING = 0x4DC2;
const char32_t HEXAGRAM_FOR_YOUTHFUL_FOLLY = 0x4DC3;
const char32_t HEXAGRAM_FOR_WAITING = 0x4DC4;
const char32_t HEXAGRAM_FOR_CONFLICT = 0x4DC5;
const char32_t HEXAGRAM_FOR_THE_ARMY = 0x4DC6;
const char32_t HEXAGRAM_FOR_HOLDING_TOGETHER = 0x4DC7;
const char32_t HEXAGRAM_FOR_SMALL_TAMING = 0x4DC8;
const char32_t HEXAGRAM_FOR_TREADING = 0x4DC9;
const char32_t HEXAGRAM_FOR_PEACE = 0x4DCA;
const char32_t HEXAGRAM_FOR_STANDSTILL = 0x4DCB;
const char32_t HEXAGRAM_FOR_FELLOWSHIP = 0x4DCC;
const char32_t HEXAGRAM_FOR_GREAT_POSSESSION = 0x4DCD;
const char32_t HEXAGRAM_FOR_MODESTY = 0x4DCE;
const char32_t HEXAGRAM_FOR_ENTHUSIASM = 0x4DCF;
const char32_t HEXAGRAM_FOR_FOLLOWING = 0x4DD0;
const char32_t HEXAGRAM_FOR_WORK_ON_THE_DECAYED = 0x4DD1;
const char32_t HEXAGRAM_FOR_APPROACH = 0x4DD2;
const char32_t HEXAGRAM_FOR_CONTEMPLATION = 0x4DD3;
const char32_t HEXAGRAM_FOR_BITING_THROUGH = 0x4DD4;
const char32_t HEXAGRAM_FOR_GRACE = 0x4DD5;
const char32_t HEXAGRAM_FOR_SPLITTING_APART = 0x4DD6;
const char32_t HEXAGRAM_FOR_RETURN = 0x4DD7;
const char32_t HEXAGRAM_FOR_INNOCENCE = 0x4DD8;
const char32_t HEXAGRAM_FOR_GREAT_TAMING = 0x4DD9;
const char32_t HEXAGRAM_FOR_MOUTH_CORNERS = 0x4DDA;
const char32_t HEXAGRAM_FOR_GREAT_PREPONDERANCE = 0x4DDB;
const char32_t HEXAGRAM_FOR_THE_ABYSMAL_WATER = 0x4DDC;
const char32_t HEXAGRAM_FOR_THE_CLINGING_FIRE = 0x4DDD;
const char32_t HEXAGRAM_FOR_INFLUENCE = 0x4DDE;
const char32_t HEXAGRAM_FOR_DURATION = 0x4DDF;
const char32_t HEXAGRAM_FOR_RETREAT = 0x4DE0;
const char32_t HEXAGRAM_FOR_GREAT_POWER = 0x4DE1;
const char32_t HEXAGRAM_FOR_PROGRESS = 0x4DE2;
const char32_t HEXAGRAM_FOR_DARKENING_OF_THE_LIGHT = 0x4DE3;
const char32_t HEXAGRAM_FOR_THE_FAMILY = 0x4DE4;
const char32_t HEXAGRAM_FOR_OPPOSITION = 0x4DE5;
const char32_t HEXAGRAM_FOR_OBSTRUCTION = 0x4DE6;
const char32_t HEXAGRAM_FOR_DELIVERANCE = 0x4DE7;
const char32_t HEXAGRAM_FOR_DECREASE = 0x4DE8;
const char32_t HEXAGRAM_FOR_INCREASE = 0x4DE9;
const char32_t HEXAGRAM_FOR_BREAKTHROUGH = 0x4DEA;
const char32_t HEXAGRAM_FOR_COMING_TO_MEET = 0x4DEB;
const char32_t HEXAGRAM_FOR_GATHERING_TOGETHER = 0x4DEC;
const char32_t HEXAGRAM_FOR_PUSHING_UPWARD = 0x4DED;
const char32_t HEXAGRAM_FOR_OPPRESSION = 0x4DEE;
const char32_t HEXAGRAM_FOR_THE_WELL = 0x4DEF;
const char32_t HEXAGRAM_FOR_REVOLUTION = 0x4DF0;
const char32_t HEXAGRAM_FOR_THE_CAULDRON = 0x4DF1;
const char32_t HEXAGRAM_FOR_THE_AROUSING_THUNDER = 0x4DF2;
const char32_t HEXAGRAM_FOR_THE_KEEPING_STILL_MOUNTAIN = 0x4DF3;
const char32_t HEXAGRAM_FOR_DEVELOPMENT = 0x4DF4;
const char32_t HEXAGRAM_FOR_THE_MARRYING_MAIDEN = 0x4DF5;
const char32_t HEXAGRAM_FOR_ABUNDANCE = 0x4DF6;
const char32_t HEXAGRAM_FOR_THE_WANDERER = 0x4DF7;
const char32_t HEXAGRAM_FOR_THE_GENTLE_WIND = 0x4DF8;
const char32_t HEXAGRAM_FOR_THE_JOYOUS_LAKE = 0x4DF9;
const char32_t HEXAGRAM_FOR_DISPERSION = 0x4DFA;
const char32_t HEXAGRAM_FOR_LIMITATION = 0x4DFB;
const char32_t HEXAGRAM_FOR_INNER_TRUTH = 0x4DFC;
const char32_t HEXAGRAM_FOR_SMALL_PREPONDERANCE = 0x4DFD;
const char32_t HEXAGRAM_FOR_AFTER_COMPLETION = 0x4DFE;
const char32_t HEXAGRAM_FOR_BEFORE_COMPLETION = 0x4DFF;
const char32_t CJK_IDEOGRAPH_FIRST = 0x4E00;
const char32_t CJK_IDEOGRAPH_LAST = 0x9FFF;
const char32_t YI_SYLLABLE_IT = 0xA000;
const char32_t YI_SYLLABLE_IX = 0xA001;
const char32_t YI_SYLLABLE_I = 0xA002;
const char32_t YI_SYLLABLE_IP = 0xA003;
const char32_t YI_SYLLABLE_IET = 0xA004;
const char32_t YI_SYLLABLE_IEX = 0xA005;
const char32_t YI_SYLLABLE_IE = 0xA006;
const char32_t YI_SYLLABLE_IEP = 0xA007;
const char32_t YI_SYLLABLE_AT = 0xA008;
const char32_t YI_SYLLABLE_AX = 0xA009;
const char32_t YI_SYLLABLE_A = 0xA00A;
const char32_t YI_SYLLABLE_AP = 0xA00B;
const char32_t YI_SYLLABLE_UOX = 0xA00C;
const char32_t YI_SYLLABLE_UO = 0xA00D;
const char32_t YI_SYLLABLE_UOP = 0xA00E;
const char32_t YI_SYLLABLE_OT = 0xA00F;
const char32_t YI_SYLLABLE_OX = 0xA010;
const char32_t YI_SYLLABLE_O = 0xA011;
const char32_t YI_SYLLABLE_OP = 0xA012;
const char32_t YI_SYLLABLE_EX = 0xA013;
const char32_t YI_SYLLABLE_E = 0xA014;
const char32_t YI_SYLLABLE_WU = 0xA015;
const char32_t YI_SYLLABLE_BIT = 0xA016;
const char32_t YI_SYLLABLE_BIX = 0xA017;
const char32_t YI_SYLLABLE_BI = 0xA018;
const char32_t YI_SYLLABLE_BIP = 0xA019;
const char32_t YI_SYLLABLE_BIET = 0xA01A;
const char32_t YI_SYLLABLE_BIEX = 0xA01B;
const char32_t YI_SYLLABLE_BIE = 0xA01C;
const char32_t YI_SYLLABLE_BIEP = 0xA01D;
const char32_t YI_SYLLABLE_BAT = 0xA01E;
const char32_t YI_SYLLABLE_BAX = 0xA01F;
const char32_t YI_SYLLABLE_BA = 0xA020;
const char32_t YI_SYLLABLE_BAP = 0xA021;
const char32_t YI_SYLLABLE_BUOX = 0xA022;
const char32_t YI_SYLLABLE_BUO = 0xA023;
const char32_t YI_SYLLABLE_BUOP = 0xA024;
const char32_t YI_SYLLABLE_BOT = 0xA025;
const char32_t YI_SYLLABLE_BOX = 0xA026;
const char32_t YI_SYLLABLE_BO = 0xA027;
const char32_t YI_SYLLABLE_BOP = 0xA028;
const char32_t YI_SYLLABLE_BEX = 0xA029;
const char32_t YI_SYLLABLE_BE = 0xA02A;
const char32_t YI_SYLLABLE_BEP = 0xA02B;
const char32_t YI_SYLLABLE_BUT = 0xA02C;
const char32_t YI_SYLLABLE_BUX = 0xA02D;
const char32_t YI_SYLLABLE_BU = 0xA02E;
const char32_t YI_SYLLABLE_BUP = 0xA02F;
const char32_t YI_SYLLABLE_BURX = 0xA030;
const char32_t YI_SYLLABLE_BUR = 0xA031;
const char32_t YI_SYLLABLE_BYT = 0xA032;
const char32_t YI_SYLLABLE_BYX = 0xA033;
const char32_t YI_SYLLABLE_BY = 0xA034;
const char32_t YI_SYLLABLE_BYP = 0xA035;
const char32_t YI_SYLLABLE_BYRX = 0xA036;
const char32_t YI_SYLLABLE_BYR = 0xA037;
const char32_t YI_SYLLABLE_PIT = 0xA038;
const char32_t YI_SYLLABLE_PIX = 0xA039;
const char32_t YI_SYLLABLE_PI = 0xA03A;
const char32_t YI_SYLLABLE_PIP = 0xA03B;
const char32_t YI_SYLLABLE_PIEX = 0xA03C;
const char32_t YI_SYLLABLE_PIE = 0xA03D;
const char32_t YI_SYLLABLE_PIEP = 0xA03E;
const char32_t YI_SYLLABLE_PAT = 0xA03F;
const char32_t YI_SYLLABLE_PAX = 0xA040;
const char32_t YI_SYLLABLE_PA = 0xA041;
const char32_t YI_SYLLABLE_PAP = 0xA042;
const char32_t YI_SYLLABLE_PUOX = 0xA043;
const char32_t YI_SYLLABLE_PUO = 0xA044;
const char32_t YI_SYLLABLE_PUOP = 0xA045;
const char32_t YI_SYLLABLE_POT = 0xA046;
const char32_t YI_SYLLABLE_POX = 0xA047;
const char32_t YI_SYLLABLE_PO = 0xA048;
const char32_t YI_SYLLABLE_POP = 0xA049;
const char32_t YI_SYLLABLE_PUT = 0xA04A;
const char32_t YI_SYLLABLE_PUX = 0xA04B;
const char32_t YI_SYLLABLE_PU = 0xA04C;
const char32_t YI_SYLLABLE_PUP = 0xA04D;
const char32_t YI_SYLLABLE_PURX = 0xA04E;
const char32_t YI_SYLLABLE_PUR = 0xA04F;
const char32_t YI_SYLLABLE_PYT = 0xA050;
const char32_t YI_SYLLABLE_PYX = 0xA051;
const char32_t YI_SYLLABLE_PY = 0xA052;
const char32_t YI_SYLLABLE_PYP = 0xA053;
const char32_t YI_SYLLABLE_PYRX = 0xA054;
const char32_t YI_SYLLABLE_PYR = 0xA055;
const char32_t YI_SYLLABLE_BBIT = 0xA056;
const char32_t YI_SYLLABLE_BBIX = 0xA057;
const char32_t YI_SYLLABLE_BBI = 0xA058;
const char32_t YI_SYLLABLE_BBIP = 0xA059;
const char32_t YI_SYLLABLE_BBIET = 0xA05A;
const char32_t YI_SYLLABLE_BBIEX = 0xA05B;
const char32_t YI_SYLLABLE_BBIE = 0xA05C;
const char32_t YI_SYLLABLE_BBIEP = 0xA05D;
const char32_t YI_SYLLABLE_BBAT = 0xA05E;
const char32_t YI_SYLLABLE_BBAX = 0xA05F;
const char32_t YI_SYLLABLE_BBA = 0xA060;
const char32_t YI_SYLLABLE_BBAP = 0xA061;
const char32_t YI_SYLLABLE_BBUOX = 0xA062;
const char32_t YI_SYLLABLE_BBUO = 0xA063;
const char32_t YI_SYLLABLE_BBUOP = 0xA064;
const char32_t YI_SYLLABLE_BBOT = 0xA065;
const char32_t YI_SYLLABLE_BBOX = 0xA066;
const char32_t YI_SYLLABLE_BBO = 0xA067;
const char32_t YI_SYLLABLE_BBOP = 0xA068;
const char32_t YI_SYLLABLE_BBEX = 0xA069;
const char32_t YI_SYLLABLE_BBE = 0xA06A;
const char32_t YI_SYLLABLE_BBEP = 0xA06B;
const char32_t YI_SYLLABLE_BBUT = 0xA06C;
const char32_t YI_SYLLABLE_BBUX = 0xA06D;
const char32_t YI_SYLLABLE_BBU = 0xA06E;
const char32_t YI_SYLLABLE_BBUP = 0xA06F;
const char32_t YI_SYLLABLE_BBURX = 0xA070;
const char32_t YI_SYLLABLE_BBUR = 0xA071;
const char32_t YI_SYLLABLE_BBYT = 0xA072;
const char32_t YI_SYLLABLE_BBYX = 0xA073;
const char32_t YI_SYLLABLE_BBY = 0xA074;
const char32_t YI_SYLLABLE_BBYP = 0xA075;
const char32_t YI_SYLLABLE_NBIT = 0xA076;
const char32_t YI_SYLLABLE_NBIX = 0xA077;
const char32_t YI_SYLLABLE_NBI = 0xA078;
const char32_t YI_SYLLABLE_NBIP = 0xA079;
const char32_t YI_SYLLABLE_NBIEX = 0xA07A;
const char32_t YI_SYLLABLE_NBIE = 0xA07B;
const char32_t YI_SYLLABLE_NBIEP = 0xA07C;
const char32_t YI_SYLLABLE_NBAT = 0xA07D;
const char32_t YI_SYLLABLE_NBAX = 0xA07E;
const char32_t YI_SYLLABLE_NBA = 0xA07F;
const char32_t YI_SYLLABLE_NBAP = 0xA080;
const char32_t YI_SYLLABLE_NBOT = 0xA081;
const char32_t YI_SYLLABLE_NBOX = 0xA082;
const char32_t YI_SYLLABLE_NBO = 0xA083;
const char32_t YI_SYLLABLE_NBOP = 0xA084;
const char32_t YI_SYLLABLE_NBUT = 0xA085;
const char32_t YI_SYLLABLE_NBUX = 0xA086;
const char32_t YI_SYLLABLE_NBU = 0xA087;
const char32_t YI_SYLLABLE_NBUP = 0xA088;
const char32_t YI_SYLLABLE_NBURX = 0xA089;
const char32_t YI_SYLLABLE_NBUR = 0xA08A;
const char32_t YI_SYLLABLE_NBYT = 0xA08B;
const char32_t YI_SYLLABLE_NBYX = 0xA08C;
const char32_t YI_SYLLABLE_NBY = 0xA08D;
const char32_t YI_SYLLABLE_NBYP = 0xA08E;
const char32_t YI_SYLLABLE_NBYRX = 0xA08F;
const char32_t YI_SYLLABLE_NBYR = 0xA090;
const char32_t YI_SYLLABLE_HMIT = 0xA091;
const char32_t YI_SYLLABLE_HMIX = 0xA092;
const char32_t YI_SYLLABLE_HMI = 0xA093;
const char32_t YI_SYLLABLE_HMIP = 0xA094;
const char32_t YI_SYLLABLE_HMIEX = 0xA095;
const char32_t YI_SYLLABLE_HMIE = 0xA096;
const char32_t YI_SYLLABLE_HMIEP = 0xA097;
const char32_t YI_SYLLABLE_HMAT = 0xA098;
const char32_t YI_SYLLABLE_HMAX = 0xA099;
const char32_t YI_SYLLABLE_HMA = 0xA09A;
const char32_t YI_SYLLABLE_HMAP = 0xA09B;
const char32_t YI_SYLLABLE_HMUOX = 0xA09C;
const char32_t YI_SYLLABLE_HMUO = 0xA09D;
const char32_t YI_SYLLABLE_HMUOP = 0xA09E;
const char32_t YI_SYLLABLE_HMOT = 0xA09F;
const char32_t YI_SYLLABLE_HMOX = 0xA0A0;
const char32_t YI_SYLLABLE_HMO = 0xA0A1;
const char32_t YI_SYLLABLE_HMOP = 0xA0A2;
const char32_t YI_SYLLABLE_HMUT = 0xA0A3;
const char32_t YI_SYLLABLE_HMUX = 0xA0A4;
const char32_t YI_SYLLABLE_HMU = 0xA0A5;
const char32_t YI_SYLLABLE_HMUP = 0xA0A6;
const char32_t YI_SYLLABLE_HMURX = 0xA0A7;
const char32_t YI_SYLLABLE_HMUR = 0xA0A8;
const char32_t YI_SYLLABLE_HMYX = 0xA0A9;
const char32_t YI_SYLLABLE_HMY = 0xA0AA;
const char32_t YI_SYLLABLE_HMYP = 0xA0AB;
const char32_t YI_SYLLABLE_HMYRX = 0xA0AC;
const char32_t YI_SYLLABLE_HMYR = 0xA0AD;
const char32_t YI_SYLLABLE_MIT = 0xA0AE;
const char32_t YI_SYLLABLE_MIX = 0xA0AF;
const char32_t YI_SYLLABLE_MI = 0xA0B0;
const char32_t YI_SYLLABLE_MIP = 0xA0B1;
const char32_t YI_SYLLABLE_MIEX = 0xA0B2;
const char32_t YI_SYLLABLE_MIE = 0xA0B3;
const char32_t YI_SYLLABLE_MIEP = 0xA0B4;
const char32_t YI_SYLLABLE_MAT = 0xA0B5;
const char32_t YI_SYLLABLE_MAX = 0xA0B6;
const char32_t YI_SYLLABLE_MA = 0xA0B7;
const char32_t YI_SYLLABLE_MAP = 0xA0B8;
const char32_t YI_SYLLABLE_MUOT = 0xA0B9;
const char32_t YI_SYLLABLE_MUOX = 0xA0BA;
const char32_t YI_SYLLABLE_MUO = 0xA0BB;
const char32_t YI_SYLLABLE_MUOP = 0xA0BC;
const char32_t YI_SYLLABLE_MOT = 0xA0BD;
const char32_t YI_SYLLABLE_MOX = 0xA0BE;
const char32_t YI_SYLLABLE_MO = 0xA0BF;
const char32_t YI_SYLLABLE_MOP = 0xA0C0;
const char32_t YI_SYLLABLE_MEX = 0xA0C1;
const char32_t YI_SYLLABLE_ME = 0xA0C2;
const char32_t YI_SYLLABLE_MUT = 0xA0C3;
const char32_t YI_SYLLABLE_MUX = 0xA0C4;
const char32_t YI_SYLLABLE_MU = 0xA0C5;
const char32_t YI_SYLLABLE_MUP = 0xA0C6;
const char32_t YI_SYLLABLE_MURX = 0xA0C7;
const char32_t YI_SYLLABLE_MUR = 0xA0C8;
const char32_t YI_SYLLABLE_MYT = 0xA0C9;
const char32_t YI_SYLLABLE_MYX = 0xA0CA;
const char32_t YI_SYLLABLE_MY = 0xA0CB;
const char32_t YI_SYLLABLE_MYP = 0xA0CC;
const char32_t YI_SYLLABLE_FIT = 0xA0CD;
const char32_t YI_SYLLABLE_FIX = 0xA0CE;
const char32_t YI_SYLLABLE_FI = 0xA0CF;
const char32_t YI_SYLLABLE_FIP = 0xA0D0;
const char32_t YI_SYLLABLE_FAT = 0xA0D1;
const char32_t YI_SYLLABLE_FAX = 0xA0D2;
const char32_t YI_SYLLABLE_FA = 0xA0D3;
const char32_t YI_SYLLABLE_FAP = 0xA0D4;
const char32_t YI_SYLLABLE_FOX = 0xA0D5;
const char32_t YI_SYLLABLE_FO = 0xA0D6;
const char32_t YI_SYLLABLE_FOP = 0xA0D7;
const char32_t YI_SYLLABLE_FUT = 0xA0D8;
const char32_t YI_SYLLABLE_FUX = 0xA0D9;
const char32_t YI_SYLLABLE_FU = 0xA0DA;
const char32_t YI_SYLLABLE_FUP = 0xA0DB;
const char32_t YI_SYLLABLE_FURX = 0xA0DC;
const char32_t YI_SYLLABLE_FUR = 0xA0DD;
const char32_t YI_SYLLABLE_FYT = 0xA0DE;
const char32_t YI_SYLLABLE_FYX = 0xA0DF;
const char32_t YI_SYLLABLE_FY = 0xA0E0;
const char32_t YI_SYLLABLE_FYP = 0xA0E1;
const char32_t YI_SYLLABLE_VIT = 0xA0E2;
const char32_t YI_SYLLABLE_VIX = 0xA0E3;
const char32_t YI_SYLLABLE_VI = 0xA0E4;
const char32_t YI_SYLLABLE_VIP = 0xA0E5;
const char32_t YI_SYLLABLE_VIET = 0xA0E6;
const char32_t YI_SYLLABLE_VIEX = 0xA0E7;
const char32_t YI_SYLLABLE_VIE = 0xA0E8;
const char32_t YI_SYLLABLE_VIEP = 0xA0E9;
const char32_t YI_SYLLABLE_VAT = 0xA0EA;
const char32_t YI_SYLLABLE_VAX = 0xA0EB;
const char32_t YI_SYLLABLE_VA = 0xA0EC;
const char32_t YI_SYLLABLE_VAP = 0xA0ED;
const char32_t YI_SYLLABLE_VOT = 0xA0EE;
const char32_t YI_SYLLABLE_VOX = 0xA0EF;
const char32_t YI_SYLLABLE_VO = 0xA0F0;
const char32_t YI_SYLLABLE_VOP = 0xA0F1;
const char32_t YI_SYLLABLE_VEX = 0xA0F2;
const char32_t YI_SYLLABLE_VEP = 0xA0F3;
const char32_t YI_SYLLABLE_VUT = 0xA0F4;
const char32_t YI_SYLLABLE_VUX = 0xA0F5;
const char32_t YI_SYLLABLE_VU = 0xA0F6;
const char32_t YI_SYLLABLE_VUP = 0xA0F7;
const char32_t YI_SYLLABLE_VURX = 0xA0F8;
const char32_t YI_SYLLABLE_VUR = 0xA0F9;
const char32_t YI_SYLLABLE_VYT = 0xA0FA;
const char32_t YI_SYLLABLE_VYX = 0xA0FB;
const char32_t YI_SYLLABLE_VY = 0xA0FC;
const char32_t YI_SYLLABLE_VYP = 0xA0FD;
const char32_t YI_SYLLABLE_VYRX = 0xA0FE;
const char32_t YI_SYLLABLE_VYR = 0xA0FF;
const char32_t YI_SYLLABLE_DIT = 0xA100;
const char32_t YI_SYLLABLE_DIX = 0xA101;
const char32_t YI_SYLLABLE_DI = 0xA102;
const char32_t YI_SYLLABLE_DIP = 0xA103;
const char32_t YI_SYLLABLE_DIEX = 0xA104;
const char32_t YI_SYLLABLE_DIE = 0xA105;
const char32_t YI_SYLLABLE_DIEP = 0xA106;
const char32_t YI_SYLLABLE_DAT = 0xA107;
const char32_t YI_SYLLABLE_DAX = 0xA108;
const char32_t YI_SYLLABLE_DA = 0xA109;
const char32_t YI_SYLLABLE_DAP = 0xA10A;
const char32_t YI_SYLLABLE_DUOX = 0xA10B;
const char32_t YI_SYLLABLE_DUO = 0xA10C;
const char32_t YI_SYLLABLE_DOT = 0xA10D;
const char32_t YI_SYLLABLE_DOX = 0xA10E;
const char32_t YI_SYLLABLE_DO = 0xA10F;
const char32_t YI_SYLLABLE_DOP = 0xA110;
const char32_t YI_SYLLABLE_DEX = 0xA111;
const char32_t YI_SYLLABLE_DE = 0xA112;
const char32_t YI_SYLLABLE_DEP = 0xA113;
const char32_t YI_SYLLABLE_DUT = 0xA114;
const char32_t YI_SYLLABLE_DUX = 0xA115;
const char32_t YI_SYLLABLE_DU = 0xA116;
const char32_t YI_SYLLABLE_DUP = 0xA117;
const char32_t YI_SYLLABLE_DURX = 0xA118;
const char32_t YI_SYLLABLE_DUR = 0xA119;
const char32_t YI_SYLLABLE_TIT = 0xA11A;
const char32_t YI_SYLLABLE_TIX = 0xA11B;
const char32_t YI_SYLLABLE_TI = 0xA11C;
const char32_t YI_SYLLABLE_TIP = 0xA11D;
const char32_t YI_SYLLABLE_TIEX = 0xA11E;
const char32_t YI_SYLLABLE_TIE = 0xA11F;
const char32_t YI_SYLLABLE_TIEP = 0xA120;
const char32_t YI_SYLLABLE_TAT = 0xA121;
const char32_t YI_SYLLABLE_TAX = 0xA122;
const char32_t YI_SYLLABLE_TA = 0xA123;
const char32_t YI_SYLLABLE_TAP = 0xA124;
const char32_t YI_SYLLABLE_TUOT = 0xA125;
const char32_t YI_SYLLABLE_TUOX = 0xA126;
const char32_t YI_SYLLABLE_TUO = 0xA127;
const char32_t YI_SYLLABLE_TUOP = 0xA128;
const char32_t YI_SYLLABLE_TOT = 0xA129;
const char32_t YI_SYLLABLE_TOX = 0xA12A;
const char32_t YI_SYLLABLE_TO = 0xA12B;
const char32_t YI_SYLLABLE_TOP = 0xA12C;
const char32_t YI_SYLLABLE_TEX = 0xA12D;
const char32_t YI_SYLLABLE_TE = 0xA12E;
const char32_t YI_SYLLABLE_TEP = 0xA12F;
const char32_t YI_SYLLABLE_TUT = 0xA130;
const char32_t YI_SYLLABLE_TUX = 0xA131;
const char32_t YI_SYLLABLE_TU = 0xA132;
const char32_t YI_SYLLABLE_TUP = 0xA133;
const char32_t YI_SYLLABLE_TURX = 0xA134;
const char32_t YI_SYLLABLE_TUR = 0xA135;
const char32_t YI_SYLLABLE_DDIT = 0xA136;
const char32_t YI_SYLLABLE_DDIX = 0xA137;
const char32_t YI_SYLLABLE_DDI = 0xA138;
const char32_t YI_SYLLABLE_DDIP = 0xA139;
const char32_t YI_SYLLABLE_DDIEX = 0xA13A;
const char32_t YI_SYLLABLE_DDIE = 0xA13B;
const char32_t YI_SYLLABLE_DDIEP = 0xA13C;
const char32_t YI_SYLLABLE_DDAT = 0xA13D;
const char32_t YI_SYLLABLE_DDAX = 0xA13E;
const char32_t YI_SYLLABLE_DDA = 0xA13F;
const char32_t YI_SYLLABLE_DDAP = 0xA140;
const char32_t YI_SYLLABLE_DDUOX = 0xA141;
const char32_t YI_SYLLABLE_DDUO = 0xA142;
const char32_t YI_SYLLABLE_DDUOP = 0xA143;
const char32_t YI_SYLLABLE_DDOT = 0xA144;
const char32_t YI_SYLLABLE_DDOX = 0xA145;
const char32_t YI_SYLLABLE_DDO = 0xA146;
const char32_t YI_SYLLABLE_DDOP = 0xA147;
const char32_t YI_SYLLABLE_DDEX = 0xA148;
const char32_t YI_SYLLABLE_DDE = 0xA149;
const char32_t YI_SYLLABLE_DDEP = 0xA14A;
const char32_t YI_SYLLABLE_DDUT = 0xA14B;
const char32_t YI_SYLLABLE_DDUX = 0xA14C;
const char32_t YI_SYLLABLE_DDU = 0xA14D;
const char32_t YI_SYLLABLE_DDUP = 0xA14E;
const char32_t YI_SYLLABLE_DDURX = 0xA14F;
const char32_t YI_SYLLABLE_DDUR = 0xA150;
const char32_t YI_SYLLABLE_NDIT = 0xA151;
const char32_t YI_SYLLABLE_NDIX = 0xA152;
const char32_t YI_SYLLABLE_NDI = 0xA153;
const char32_t YI_SYLLABLE_NDIP = 0xA154;
const char32_t YI_SYLLABLE_NDIEX = 0xA155;
const char32_t YI_SYLLABLE_NDIE = 0xA156;
const char32_t YI_SYLLABLE_NDAT = 0xA157;
const char32_t YI_SYLLABLE_NDAX = 0xA158;
const char32_t YI_SYLLABLE_NDA = 0xA159;
const char32_t YI_SYLLABLE_NDAP = 0xA15A;
const char32_t YI_SYLLABLE_NDOT = 0xA15B;
const char32_t YI_SYLLABLE_NDOX = 0xA15C;
const char32_t YI_SYLLABLE_NDO = 0xA15D;
const char32_t YI_SYLLABLE_NDOP = 0xA15E;
const char32_t YI_SYLLABLE_NDEX = 0xA15F;
const char32_t YI_SYLLABLE_NDE = 0xA160;
const char32_t YI_SYLLABLE_NDEP = 0xA161;
const char32_t YI_SYLLABLE_NDUT = 0xA162;
const char32_t YI_SYLLABLE_NDUX = 0xA163;
const char32_t YI_SYLLABLE_NDU = 0xA164;
const char32_t YI_SYLLABLE_NDUP = 0xA165;
const char32_t YI_SYLLABLE_NDURX = 0xA166;
const char32_t YI_SYLLABLE_NDUR = 0xA167;
const char32_t YI_SYLLABLE_HNIT = 0xA168;
const char32_t YI_SYLLABLE_HNIX = 0xA169;
const char32_t YI_SYLLABLE_HNI = 0xA16A;
const char32_t YI_SYLLABLE_HNIP = 0xA16B;
const char32_t YI_SYLLABLE_HNIET = 0xA16C;
const char32_t YI_SYLLABLE_HNIEX = 0xA16D;
const char32_t YI_SYLLABLE_HNIE = 0xA16E;
const char32_t YI_SYLLABLE_HNIEP = 0xA16F;
const char32_t YI_SYLLABLE_HNAT = 0xA170;
const char32_t YI_SYLLABLE_HNAX = 0xA171;
const char32_t YI_SYLLABLE_HNA = 0xA172;
const char32_t YI_SYLLABLE_HNAP = 0xA173;
const char32_t YI_SYLLABLE_HNUOX = 0xA174;
const char32_t YI_SYLLABLE_HNUO = 0xA175;
const char32_t YI_SYLLABLE_HNOT = 0xA176;
const char32_t YI_SYLLABLE_HNOX = 0xA177;
const char32_t YI_SYLLABLE_HNOP = 0xA178;
const char32_t YI_SYLLABLE_HNEX = 0xA179;
const char32_t YI_SYLLABLE_HNE = 0xA17A;
const char32_t YI_SYLLABLE_HNEP = 0xA17B;
const char32_t YI_SYLLABLE_HNUT = 0xA17C;
const char32_t YI_SYLLABLE_NIT = 0xA17D;
const char32_t YI_SYLLABLE_NIX = 0xA17E;
const char32_t YI_SYLLABLE_NI = 0xA17F;
const char32_t YI_SYLLABLE_NIP = 0xA180;
const char32_t YI_SYLLABLE_NIEX = 0xA181;
const char32_t YI_SYLLABLE_NIE = 0xA182;
const char32_t YI_SYLLABLE_NIEP = 0xA183;
const char32_t YI_SYLLABLE_NAX = 0xA184;
const char32_t YI_SYLLABLE_NA = 0xA185;
const char32_t YI_SYLLABLE_NAP = 0xA186;
const char32_t YI_SYLLABLE_NUOX = 0xA187;
const char32_t YI_SYLLABLE_NUO = 0xA188;
const char32_t YI_SYLLABLE_NUOP = 0xA189;
const char32_t YI_SYLLABLE_NOT = 0xA18A;
const char32_t YI_SYLLABLE_NOX = 0xA18B;
const char32_t YI_SYLLABLE_NO = 0xA18C;
const char32_t YI_SYLLABLE_NOP = 0xA18D;
const char32_t YI_SYLLABLE_NEX = 0xA18E;
const char32_t YI_SYLLABLE_NE = 0xA18F;
const char32_t YI_SYLLABLE_NEP = 0xA190;
const char32_t YI_SYLLABLE_NUT = 0xA191;
const char32_t YI_SYLLABLE_NUX = 0xA192;
const char32_t YI_SYLLABLE_NU = 0xA193;
const char32_t YI_SYLLABLE_NUP = 0xA194;
const char32_t YI_SYLLABLE_NURX = 0xA195;
const char32_t YI_SYLLABLE_NUR = 0xA196;
const char32_t YI_SYLLABLE_HLIT = 0xA197;
const char32_t YI_SYLLABLE_HLIX = 0xA198;
const char32_t YI_SYLLABLE_HLI = 0xA199;
const char32_t YI_SYLLABLE_HLIP = 0xA19A;
const char32_t YI_SYLLABLE_HLIEX = 0xA19B;
const char32_t YI_SYLLABLE_HLIE = 0xA19C;
const char32_t YI_SYLLABLE_HLIEP = 0xA19D;
const char32_t YI_SYLLABLE_HLAT = 0xA19E;
const char32_t YI_SYLLABLE_HLAX = 0xA19F;
const char32_t YI_SYLLABLE_HLA = 0xA1A0;
const char32_t YI_SYLLABLE_HLAP = 0xA1A1;
const char32_t YI_SYLLABLE_HLUOX = 0xA1A2;
const char32_t YI_SYLLABLE_HLUO = 0xA1A3;
const char32_t YI_SYLLABLE_HLUOP = 0xA1A4;
const char32_t YI_SYLLABLE_HLOX = 0xA1A5;
const char32_t YI_SYLLABLE_HLO = 0xA1A6;
const char32_t YI_SYLLABLE_HLOP = 0xA1A7;
const char32_t YI_SYLLABLE_HLEX = 0xA1A8;
const char32_t YI_SYLLABLE_HLE = 0xA1A9;
const char32_t YI_SYLLABLE_HLEP = 0xA1AA;
const char32_t YI_SYLLABLE_HLUT = 0xA1AB;
const char32_t YI_SYLLABLE_HLUX = 0xA1AC;
const char32_t YI_SYLLABLE_HLU = 0xA1AD;
const char32_t YI_SYLLABLE_HLUP = 0xA1AE;
const char32_t YI_SYLLABLE_HLURX = 0xA1AF;
const char32_t YI_SYLLABLE_HLUR = 0xA1B0;
const char32_t YI_SYLLABLE_HLYT = 0xA1B1;
const char32_t YI_SYLLABLE_HLYX = 0xA1B2;
const char32_t YI_SYLLABLE_HLY = 0xA1B3;
const char32_t YI_SYLLABLE_HLYP = 0xA1B4;
const char32_t YI_SYLLABLE_HLYRX = 0xA1B5;
const char32_t YI_SYLLABLE_HLYR = 0xA1B6;
const char32_t YI_SYLLABLE_LIT = 0xA1B7;
const char32_t YI_SYLLABLE_LIX = 0xA1B8;
const char32_t YI_SYLLABLE_LI = 0xA1B9;
const char32_t YI_SYLLABLE_LIP = 0xA1BA;
const char32_t YI_SYLLABLE_LIET = 0xA1BB;
const char32_t YI_SYLLABLE_LIEX = 0xA1BC;
const char32_t YI_SYLLABLE_LIE = 0xA1BD;
const char32_t YI_SYLLABLE_LIEP = 0xA1BE;
const char32_t YI_SYLLABLE_LAT = 0xA1BF;
const char32_t YI_SYLLABLE_LAX = 0xA1C0;
const char32_t YI_SYLLABLE_LA = 0xA1C1;
const char32_t YI_SYLLABLE_LAP = 0xA1C2;
const char32_t YI_SYLLABLE_LUOT = 0xA1C3;
const char32_t YI_SYLLABLE_LUOX = 0xA1C4;
const char32_t YI_SYLLABLE_LUO = 0xA1C5;
const char32_t YI_SYLLABLE_LUOP = 0xA1C6;
const char32_t YI_SYLLABLE_LOT = 0xA1C7;
const char32_t YI_SYLLABLE_LOX = 0xA1C8;
const char32_t YI_SYLLABLE_LO = 0xA1C9;
const char32_t YI_SYLLABLE_LOP = 0xA1CA;
const char32_t YI_SYLLABLE_LEX = 0xA1CB;
const char32_t YI_SYLLABLE_LE = 0xA1CC;
const char32_t YI_SYLLABLE_LEP = 0xA1CD;
const char32_t YI_SYLLABLE_LUT = 0xA1CE;
const char32_t YI_SYLLABLE_LUX = 0xA1CF;
const char32_t YI_SYLLABLE_LU = 0xA1D0;
const char32_t YI_SYLLABLE_LUP = 0xA1D1;
const char32_t YI_SYLLABLE_LURX = 0xA1D2;
const char32_t YI_SYLLABLE_LUR = 0xA1D3;
const char32_t YI_SYLLABLE_LYT = 0xA1D4;
const char32_t YI_SYLLABLE_LYX = 0xA1D5;
const char32_t YI_SYLLABLE_LY = 0xA1D6;
const char32_t YI_SYLLABLE_LYP = 0xA1D7;
const char32_t YI_SYLLABLE_LYRX = 0xA1D8;
const char32_t YI_SYLLABLE_LYR = 0xA1D9;
const char32_t YI_SYLLABLE_GIT = 0xA1DA;
const char32_t YI_SYLLABLE_GIX = 0xA1DB;
const char32_t YI_SYLLABLE_GI = 0xA1DC;
const char32_t YI_SYLLABLE_GIP = 0xA1DD;
const char32_t YI_SYLLABLE_GIET = 0xA1DE;
const char32_t YI_SYLLABLE_GIEX = 0xA1DF;
const char32_t YI_SYLLABLE_GIE = 0xA1E0;
const char32_t YI_SYLLABLE_GIEP = 0xA1E1;
const char32_t YI_SYLLABLE_GAT = 0xA1E2;
const char32_t YI_SYLLABLE_GAX = 0xA1E3;
const char32_t YI_SYLLABLE_GA = 0xA1E4;
const char32_t YI_SYLLABLE_GAP = 0xA1E5;
const char32_t YI_SYLLABLE_GUOT = 0xA1E6;
const char32_t YI_SYLLABLE_GUOX = 0xA1E7;
const char32_t YI_SYLLABLE_GUO = 0xA1E8;
const char32_t YI_SYLLABLE_GUOP = 0xA1E9;
const char32_t YI_SYLLABLE_GOT = 0xA1EA;
const char32_t YI_SYLLABLE_GOX = 0xA1EB;
const char32_t YI_SYLLABLE_GO = 0xA1EC;
const char32_t YI_SYLLABLE_GOP = 0xA1ED;
const char32_t YI_SYLLABLE_GET = 0xA1EE;
const char32_t YI_SYLLABLE_GEX = 0xA1EF;
const char32_t YI_SYLLABLE_GE = 0xA1F0;
const char32_t YI_SYLLABLE_GEP = 0xA1F1;
const char32_t YI_SYLLABLE_GUT = 0xA1F2;
const char32_t YI_SYLLABLE_GUX = 0xA1F3;
const char32_t YI_SYLLABLE_GU = 0xA1F4;
const char32_t YI_SYLLABLE_GUP = 0xA1F5;
const char32_t YI_SYLLABLE_GURX = 0xA1F6;
const char32_t YI_SYLLABLE_GUR = 0xA1F7;
const char32_t YI_SYLLABLE_KIT = 0xA1F8;
const char32_t YI_SYLLABLE_KIX = 0xA1F9;
const char32_t YI_SYLLABLE_KI = 0xA1FA;
const char32_t YI_SYLLABLE_KIP = 0xA1FB;
const char32_t YI_SYLLABLE_KIEX = 0xA1FC;
const char32_t YI_SYLLABLE_KIE = 0xA1FD;
const char32_t YI_SYLLABLE_KIEP = 0xA1FE;
const char32_t YI_SYLLABLE_KAT = 0xA1FF;
const char32_t YI_SYLLABLE_KAX = 0xA200;
const char32_t YI_SYLLABLE_KA = 0xA201;
const char32_t YI_SYLLABLE_KAP = 0xA202;
const char32_t YI_SYLLABLE_KUOX = 0xA203;
const char32_t YI_SYLLABLE_KUO = 0xA204;
const char32_t YI_SYLLABLE_KUOP = 0xA205;
const char32_t YI_SYLLABLE_KOT = 0xA206;
const char32_t YI_SYLLABLE_KOX = 0xA207;
const char32_t YI_SYLLABLE_KO = 0xA208;
const char32_t YI_SYLLABLE_KOP = 0xA209;
const char32_t YI_SYLLABLE_KET = 0xA20A;
const char32_t YI_SYLLABLE_KEX = 0xA20B;
const char32_t YI_SYLLABLE_KE = 0xA20C;
const char32_t YI_SYLLABLE_KEP = 0xA20D;
const char32_t YI_SYLLABLE_KUT = 0xA20E;
const char32_t YI_SYLLABLE_KUX = 0xA20F;
const char32_t YI_SYLLABLE_KU = 0xA210;
const char32_t YI_SYLLABLE_KUP = 0xA211;
const char32_t YI_SYLLABLE_KURX = 0xA212;
const char32_t YI_SYLLABLE_KUR = 0xA213;
const char32_t YI_SYLLABLE_GGIT = 0xA214;
const char32_t YI_SYLLABLE_GGIX = 0xA215;
const char32_t YI_SYLLABLE_GGI = 0xA216;
const char32_t YI_SYLLABLE_GGIEX = 0xA217;
const char32_t YI_SYLLABLE_GGIE = 0xA218;
const char32_t YI_SYLLABLE_GGIEP = 0xA219;
const char32_t YI_SYLLABLE_GGAT = 0xA21A;
const char32_t YI_SYLLABLE_GGAX = 0xA21B;
const char32_t YI_SYLLABLE_GGA = 0xA21C;
const char32_t YI_SYLLABLE_GGAP = 0xA21D;
const char32_t YI_SYLLABLE_GGUOT = 0xA21E;
const char32_t YI_SYLLABLE_GGUOX = 0xA21F;
const char32_t YI_SYLLABLE_GGUO = 0xA220;
const char32_t YI_SYLLABLE_GGUOP = 0xA221;
const char32_t YI_SYLLABLE_GGOT = 0xA222;
const char32_t YI_SYLLABLE_GGOX = 0xA223;
const char32_t YI_SYLLABLE_GGO = 0xA224;
const char32_t YI_SYLLABLE_GGOP = 0xA225;
const char32_t YI_SYLLABLE_GGET = 0xA226;
const char32_t YI_SYLLABLE_GGEX = 0xA227;
const char32_t YI_SYLLABLE_GGE = 0xA228;
const char32_t YI_SYLLABLE_GGEP = 0xA229;
const char32_t YI_SYLLABLE_GGUT = 0xA22A;
const char32_t YI_SYLLABLE_GGUX = 0xA22B;
const char32_t YI_SYLLABLE_GGU = 0xA22C;
const char32_t YI_SYLLABLE_GGUP = 0xA22D;
const char32_t YI_SYLLABLE_GGURX = 0xA22E;
const char32_t YI_SYLLABLE_GGUR = 0xA22F;
const char32_t YI_SYLLABLE_MGIEX = 0xA230;
const char32_t YI_SYLLABLE_MGIE = 0xA231;
const char32_t YI_SYLLABLE_MGAT = 0xA232;
const char32_t YI_SYLLABLE_MGAX = 0xA233;
const char32_t YI_SYLLABLE_MGA = 0xA234;
const char32_t YI_SYLLABLE_MGAP = 0xA235;
const char32_t YI_SYLLABLE_MGUOX = 0xA236;
const char32_t YI_SYLLABLE_MGUO = 0xA237;
const char32_t YI_SYLLABLE_MGUOP = 0xA238;
const char32_t YI_SYLLABLE_MGOT = 0xA239;
const char32_t YI_SYLLABLE_MGOX = 0xA23A;
const char32_t YI_SYLLABLE_MGO = 0xA23B;
const char32_t YI_SYLLABLE_MGOP = 0xA23C;
const char32_t YI_SYLLABLE_MGEX = 0xA23D;
const char32_t YI_SYLLABLE_MGE = 0xA23E;
const char32_t YI_SYLLABLE_MGEP = 0xA23F;
const char32_t YI_SYLLABLE_MGUT = 0xA240;
const char32_t YI_SYLLABLE_MGUX = 0xA241;
const char32_t YI_SYLLABLE_MGU = 0xA242;
const char32_t YI_SYLLABLE_MGUP = 0xA243;
const char32_t YI_SYLLABLE_MGURX = 0xA244;
const char32_t YI_SYLLABLE_MGUR = 0xA245;
const char32_t YI_SYLLABLE_HXIT = 0xA246;
const char32_t YI_SYLLABLE_HXIX = 0xA247;
const char32_t YI_SYLLABLE_HXI = 0xA248;
const char32_t YI_SYLLABLE_HXIP = 0xA249;
const char32_t YI_SYLLABLE_HXIET = 0xA24A;
const char32_t YI_SYLLABLE_HXIEX = 0xA24B;
const char32_t YI_SYLLABLE_HXIE = 0xA24C;
const char32_t YI_SYLLABLE_HXIEP = 0xA24D;
const char32_t YI_SYLLABLE_HXAT = 0xA24E;
const char32_t YI_SYLLABLE_HXAX = 0xA24F;
const char32_t YI_SYLLABLE_HXA = 0xA250;
const char32_t YI_SYLLABLE_HXAP = 0xA251;
const char32_t YI_SYLLABLE_HXUOT = 0xA252;
const char32_t YI_SYLLABLE_HXUOX = 0xA253;
const char32_t YI_SYLLABLE_HXUO = 0xA254;
const char32_t YI_SYLLABLE_HXUOP = 0xA255;
const char32_t YI_SYLLABLE_HXOT = 0xA256;
const char32_t YI_SYLLABLE_HXOX = 0xA257;
const char32_t YI_SYLLABLE_HXO = 0xA258;
const char32_t YI_SYLLABLE_HXOP = 0xA259;
const char32_t YI_SYLLABLE_HXEX = 0xA25A;
const char32_t YI_SYLLABLE_HXE = 0xA25B;
const char32_t YI_SYLLABLE_HXEP = 0xA25C;
const char32_t YI_SYLLABLE_NGIEX = 0xA25D;
const char32_t YI_SYLLABLE_NGIE = 0xA25E;
const char32_t YI_SYLLABLE_NGIEP = 0xA25F;
const char32_t YI_SYLLABLE_NGAT = 0xA260;
const char32_t YI_SYLLABLE_NGAX = 0xA261;
const char32_t YI_SYLLABLE_NGA = 0xA262;
const char32_t YI_SYLLABLE_NGAP = 0xA263;
const char32_t YI_SYLLABLE_NGUOT = 0xA264;
const char32_t YI_SYLLABLE_NGUOX = 0xA265;
const char32_t YI_SYLLABLE_NGUO = 0xA266;
const char32_t YI_SYLLABLE_NGOT = 0xA267;
const char32_t YI_SYLLABLE_NGOX = 0xA268;
const char32_t YI_SYLLABLE_NGO = 0xA269;
const char32_t YI_SYLLABLE_NGOP = 0xA26A;
const char32_t YI_SYLLABLE_NGEX = 0xA26B;
const char32_t YI_SYLLABLE_NGE = 0xA26C;
const char32_t YI_SYLLABLE_NGEP = 0xA26D;
const char32_t YI_SYLLABLE_HIT = 0xA26E;
const char32_t YI_SYLLABLE_HIEX = 0xA26F;
const char32_t YI_SYLLABLE_HIE = 0xA270;
const char32_t YI_SYLLABLE_HAT = 0xA271;
const char32_t YI_SYLLABLE_HAX = 0xA272;
const char32_t YI_SYLLABLE_HA = 0xA273;
const char32_t YI_SYLLABLE_HAP = 0xA274;
const char32_t YI_SYLLABLE_HUOT = 0xA275;
const char32_t YI_SYLLABLE_HUOX = 0xA276;
const char32_t YI_SYLLABLE_HUO = 0xA277;
const char32_t YI_SYLLABLE_HUOP = 0xA278;
const char32_t YI_SYLLABLE_HOT = 0xA279;
const char32_t YI_SYLLABLE_HOX = 0xA27A;
const char32_t YI_SYLLABLE_HO = 0xA27B;
const char32_t YI_SYLLABLE_HOP = 0xA27C;
const char32_t YI_SYLLABLE_HEX = 0xA27D;
const char32_t YI_SYLLABLE_HE = 0xA27E;
const char32_t YI_SYLLABLE_HEP = 0xA27F;
const char32_t YI_SYLLABLE_WAT = 0xA280;
const char32_t YI_SYLLABLE_WAX = 0xA281;
const char32_t YI_SYLLABLE_WA = 0xA282;
const char32_t YI_SYLLABLE_WAP = 0xA283;
const char32_t YI_SYLLABLE_WUOX = 0xA284;
const char32_t YI_SYLLABLE_WUO = 0xA285;
const char32_t YI_SYLLABLE_WUOP = 0xA286;
const char32_t YI_SYLLABLE_WOX = 0xA287;
const char32_t YI_SYLLABLE_WO = 0xA288;
const char32_t YI_SYLLABLE_WOP = 0xA289;
const char32_t YI_SYLLABLE_WEX = 0xA28A;
const char32_t YI_SYLLABLE_WE = 0xA28B;
const char32_t YI_SYLLABLE_WEP = 0xA28C;
const char32_t YI_SYLLABLE_ZIT = 0xA28D;
const char32_t YI_SYLLABLE_ZIX = 0xA28E;
const char32_t YI_SYLLABLE_ZI = 0xA28F;
const char32_t YI_SYLLABLE_ZIP = 0xA290;
const char32_t YI_SYLLABLE_ZIEX = 0xA291;
const char32_t YI_SYLLABLE_ZIE = 0xA292;
const char32_t YI_SYLLABLE_ZIEP = 0xA293;
const char32_t YI_SYLLABLE_ZAT = 0xA294;
const char32_t YI_SYLLABLE_ZAX = 0xA295;
const char32_t YI_SYLLABLE_ZA = 0xA296;
const char32_t YI_SYLLABLE_ZAP = 0xA297;
const char32_t YI_SYLLABLE_ZUOX = 0xA298;
const char32_t YI_SYLLABLE_ZUO = 0xA299;
const char32_t YI_SYLLABLE_ZUOP = 0xA29A;
const char32_t YI_SYLLABLE_ZOT = 0xA29B;
const char32_t YI_SYLLABLE_ZOX = 0xA29C;
const char32_t YI_SYLLABLE_ZO = 0xA29D;
const char32_t YI_SYLLABLE_ZOP = 0xA29E;
const char32_t YI_SYLLABLE_ZEX = 0xA29F;
const char32_t YI_SYLLABLE_ZE = 0xA2A0;
const char32_t YI_SYLLABLE_ZEP = 0xA2A1;
const char32_t YI_SYLLABLE_ZUT = 0xA2A2;
const char32_t YI_SYLLABLE_ZUX = 0xA2A3;
const char32_t YI_SYLLABLE_ZU = 0xA2A4;
const char32_t YI_SYLLABLE_ZUP = 0xA2A5;
const char32_t YI_SYLLABLE_ZURX = 0xA2A6;
const char32_t YI_SYLLABLE_ZUR = 0xA2A7;
const char32_t YI_SYLLABLE_ZYT = 0xA2A8;
const char32_t YI_SYLLABLE_ZYX = 0xA2A9;
const char32_t YI_SYLLABLE_ZY = 0xA2AA;
const char32_t YI_SYLLABLE_ZYP = 0xA2AB;
const char32_t YI_SYLLABLE_ZYRX = 0xA2AC;
const char32_t YI_SYLLABLE_ZYR = 0xA2AD;
const char32_t YI_SYLLABLE_CIT = 0xA2AE;
const char32_t YI_SYLLABLE_CIX = 0xA2AF;
const char32_t YI_SYLLABLE_CI = 0xA2B0;
const char32_t YI_SYLLABLE_CIP = 0xA2B1;
const char32_t YI_SYLLABLE_CIET = 0xA2B2;
const char32_t YI_SYLLABLE_CIEX = 0xA2B3;
const char32_t YI_SYLLABLE_CIE = 0xA2B4;
const char32_t YI_SYLLABLE_CIEP = 0xA2B5;
const char32_t YI_SYLLABLE_CAT = 0xA2B6;
const char32_t YI_SYLLABLE_CAX = 0xA2B7;
const char32_t YI_SYLLABLE_CA = 0xA2B8;
const char32_t YI_SYLLABLE_CAP = 0xA2B9;
const char32_t YI_SYLLABLE_CUOX = 0xA2BA;
const char32_t YI_SYLLABLE_CUO = 0xA2BB;
const char32_t YI_SYLLABLE_CUOP = 0xA2BC;
const char32_t YI_SYLLABLE_COT = 0xA2BD;
const char32_t YI_SYLLABLE_COX = 0xA2BE;
const char32_t YI_SYLLABLE_CO = 0xA2BF;
const char32_t YI_SYLLABLE_COP = 0xA2C0;
const char32_t YI_SYLLABLE_CEX = 0xA2C1;
const char32_t YI_SYLLABLE_CE = 0xA2C2;
const char32_t YI_SYLLABLE_CEP = 0xA2C3;
const char32_t YI_SYLLABLE_CUT = 0xA2C4;
const char32_t YI_SYLLABLE_CUX = 0xA2C5;
const char32_t YI_SYLLABLE_CU = 0xA2C6;
const char32_t YI_SYLLABLE_CUP = 0xA2C7;
const char32_t YI_SYLLABLE_CURX = 0xA2C8;
const char32_t YI_SYLLABLE_CUR = 0xA2C9;
const char32_t YI_SYLLABLE_CYT = 0xA2CA;
const char32_t YI_SYLLABLE_CYX = 0xA2CB;
const char32_t YI_SYLLABLE_CY = 0xA2CC;
const char32_t YI_SYLLABLE_CYP = 0xA2CD;
const char32_t YI_SYLLABLE_CYRX = 0xA2CE;
const char32_t YI_SYLLABLE_CYR = 0xA2CF;
const char32_t YI_SYLLABLE_ZZIT = 0xA2D0;
const char32_t YI_SYLLABLE_ZZIX = 0xA2D1;
const char32_t YI_SYLLABLE_ZZI = 0xA2D2;
const char32_t YI_SYLLABLE_ZZIP = 0xA2D3;
const char32_t YI_SYLLABLE_ZZIET = 0xA2D4;
const char32_t YI_SYLLABLE_ZZIEX = 0xA2D5;
const char32_t YI_SYLLABLE_ZZIE = 0xA2D6;
const char32_t YI_SYLLABLE_ZZIEP = 0xA2D7;
const char32_t YI_SYLLABLE_ZZAT = 0xA2D8;
const char32_t YI_SYLLABLE_ZZAX = 0xA2D9;
const char32_t YI_SYLLABLE_ZZA = 0xA2DA;
const char32_t YI_SYLLABLE_ZZAP = 0xA2DB;
const char32_t YI_SYLLABLE_ZZOX = 0xA2DC;
const char32_t YI_SYLLABLE_ZZO = 0xA2DD;
const char32_t YI_SYLLABLE_ZZOP = 0xA2DE;
const char32_t YI_SYLLABLE_ZZEX = 0xA2DF;
const char32_t YI_SYLLABLE_ZZE = 0xA2E0;
const char32_t YI_SYLLABLE_ZZEP = 0xA2E1;
const char32_t YI_SYLLABLE_ZZUX = 0xA2E2;
const char32_t YI_SYLLABLE_ZZU = 0xA2E3;
const char32_t YI_SYLLABLE_ZZUP = 0xA2E4;
const char32_t YI_SYLLABLE_ZZURX = 0xA2E5;
const char32_t YI_SYLLABLE_ZZUR = 0xA2E6;
const char32_t YI_SYLLABLE_ZZYT = 0xA2E7;
const char32_t YI_SYLLABLE_ZZYX = 0xA2E8;
const char32_t YI_SYLLABLE_ZZY = 0xA2E9;
const char32_t YI_SYLLABLE_ZZYP = 0xA2EA;
const char32_t YI_SYLLABLE_ZZYRX = 0xA2EB;
const char32_t YI_SYLLABLE_ZZYR = 0xA2EC;
const char32_t YI_SYLLABLE_NZIT = 0xA2ED;
const char32_t YI_SYLLABLE_NZIX = 0xA2EE;
const char32_t YI_SYLLABLE_NZI = 0xA2EF;
const char32_t YI_SYLLABLE_NZIP = 0xA2F0;
const char32_t YI_SYLLABLE_NZIEX = 0xA2F1;
const char32_t YI_SYLLABLE_NZIE = 0xA2F2;
const char32_t YI_SYLLABLE_NZIEP = 0xA2F3;
const char32_t YI_SYLLABLE_NZAT = 0xA2F4;
const char32_t YI_SYLLABLE_NZAX = 0xA2F5;
const char32_t YI_SYLLABLE_NZA = 0xA2F6;
const char32_t YI_SYLLABLE_NZAP = 0xA2F7;
const char32_t YI_SYLLABLE_NZUOX = 0xA2F8;
const char32_t YI_SYLLABLE_NZUO = 0xA2F9;
const char32_t YI_SYLLABLE_NZOX = 0xA2FA;
const char32_t YI_SYLLABLE_NZOP = 0xA2FB;
const char32_t YI_SYLLABLE_NZEX = 0xA2FC;
const char32_t YI_SYLLABLE_NZE = 0xA2FD;
const char32_t YI_SYLLABLE_NZUX = 0xA2FE;
const char32_t YI_SYLLABLE_NZU = 0xA2FF;
const char32_t YI_SYLLABLE_NZUP = 0xA300;
const char32_t YI_SYLLABLE_NZURX = 0xA301;
const char32_t YI_SYLLABLE_NZUR = 0xA302;
const char32_t YI_SYLLABLE_NZYT = 0xA303;
const char32_t YI_SYLLABLE_NZYX = 0xA304;
const char32_t YI_SYLLABLE_NZY = 0xA305;
const char32_t YI_SYLLABLE_NZYP = 0xA306;
const char32_t YI_SYLLABLE_NZYRX = 0xA307;
const char32_t YI_SYLLABLE_NZYR = 0xA308;
const char32_t YI_SYLLABLE_SIT = 0xA309;
const char32_t YI_SYLLABLE_SIX = 0xA30A;
const char32_t YI_SYLLABLE_SI = 0xA30B;
const char32_t YI_SYLLABLE_SIP = 0xA30C;
const char32_t YI_SYLLABLE_SIEX = 0xA30D;
const char32_t YI_SYLLABLE_SIE = 0xA30E;
const char32_t YI_SYLLABLE_SIEP = 0xA30F;
const char32_t YI_SYLLABLE_SAT = 0xA310;
const char32_t YI_SYLLABLE_SAX = 0xA311;
const char32_t YI_SYLLABLE_SA = 0xA312;
const char32_t YI_SYLLABLE_SAP = 0xA313;
const char32_t YI_SYLLABLE_SUOX = 0xA314;
const char32_t YI_SYLLABLE_SUO = 0xA315;
const char32_t YI_SYLLABLE_SUOP = 0xA316;
const char32_t YI_SYLLABLE_SOT = 0xA317;
const char32_t YI_SYLLABLE_SOX = 0xA318;
const char32_t YI_SYLLABLE_SO = 0xA319;
const char32_t YI_SYLLABLE_SOP = 0xA31A;
const char32_t YI_SYLLABLE_SEX = 0xA31B;
const char32_t YI_SYLLABLE_SE = 0xA31C;
const char32_t YI_SYLLABLE_SEP = 0xA31D;
const char32_t YI_SYLLABLE_SUT = 0xA31E;
const char32_t YI_SYLLABLE_SUX = 0xA31F;
const char32_t YI_SYLLABLE_SU = 0xA320;
const char32_t YI_SYLLABLE_SUP = 0xA321;
const char32_t YI_SYLLABLE_SURX = 0xA322;
const char32_t YI_SYLLABLE_SUR = 0xA323;
const char32_t YI_SYLLABLE_SYT = 0xA324;
const char32_t YI_SYLLABLE_SYX = 0xA325;
const char32_t YI_SYLLABLE_SY = 0xA326;
const char32_t YI_SYLLABLE_SYP = 0xA327;
const char32_t YI_SYLLABLE_SYRX = 0xA328;
const char32_t YI_SYLLABLE_SYR = 0xA329;
const char32_t YI_SYLLABLE_SSIT = 0xA32A;
const char32_t YI_SYLLABLE_SSIX = 0xA32B;
const char32_t YI_SYLLABLE_SSI = 0xA32C;
const char32_t YI_SYLLABLE_SSIP = 0xA32D;
const char32_t YI_SYLLABLE_SSIEX = 0xA32E;
const char32_t YI_SYLLABLE_SSIE = 0xA32F;
const char32_t YI_SYLLABLE_SSIEP = 0xA330;
const char32_t YI_SYLLABLE_SSAT = 0xA331;
const char32_t YI_SYLLABLE_SSAX = 0xA332;
const char32_t YI_SYLLABLE_SSA = 0xA333;
const char32_t YI_SYLLABLE_SSAP = 0xA334;
const char32_t YI_SYLLABLE_SSOT = 0xA335;
const char32_t YI_SYLLABLE_SSOX = 0xA336;
const char32_t YI_SYLLABLE_SSO = 0xA337;
const char32_t YI_SYLLABLE_SSOP = 0xA338;
const char32_t YI_SYLLABLE_SSEX = 0xA339;
const char32_t YI_SYLLABLE_SSE = 0xA33A;
const char32_t YI_SYLLABLE_SSEP = 0xA33B;
const char32_t YI_SYLLABLE_SSUT = 0xA33C;
const char32_t YI_SYLLABLE_SSUX = 0xA33D;
const char32_t YI_SYLLABLE_SSU = 0xA33E;
const char32_t YI_SYLLABLE_SSUP = 0xA33F;
const char32_t YI_SYLLABLE_SSYT = 0xA340;
const char32_t YI_SYLLABLE_SSYX = 0xA341;
const char32_t YI_SYLLABLE_SSY = 0xA342;
const char32_t YI_SYLLABLE_SSYP = 0xA343;
const char32_t YI_SYLLABLE_SSYRX = 0xA344;
const char32_t YI_SYLLABLE_SSYR = 0xA345;
const char32_t YI_SYLLABLE_ZHAT = 0xA346;
const char32_t YI_SYLLABLE_ZHAX = 0xA347;
const char32_t YI_SYLLABLE_ZHA = 0xA348;
const char32_t YI_SYLLABLE_ZHAP = 0xA349;
const char32_t YI_SYLLABLE_ZHUOX = 0xA34A;
const char32_t YI_SYLLABLE_ZHUO = 0xA34B;
const char32_t YI_SYLLABLE_ZHUOP = 0xA34C;
const char32_t YI_SYLLABLE_ZHOT = 0xA34D;
const char32_t YI_SYLLABLE_ZHOX = 0xA34E;
const char32_t YI_SYLLABLE_ZHO = 0xA34F;
const char32_t YI_SYLLABLE_ZHOP = 0xA350;
const char32_t YI_SYLLABLE_ZHET = 0xA351;
const char32_t YI_SYLLABLE_ZHEX = 0xA352;
const char32_t YI_SYLLABLE_ZHE = 0xA353;
const char32_t YI_SYLLABLE_ZHEP = 0xA354;
const char32_t YI_SYLLABLE_ZHUT = 0xA355;
const char32_t YI_SYLLABLE_ZHUX = 0xA356;
const char32_t YI_SYLLABLE_ZHU = 0xA357;
const char32_t YI_SYLLABLE_ZHUP = 0xA358;
const char32_t YI_SYLLABLE_ZHURX = 0xA359;
const char32_t YI_SYLLABLE_ZHUR = 0xA35A;
const char32_t YI_SYLLABLE_ZHYT = 0xA35B;
const char32_t YI_SYLLABLE_ZHYX = 0xA35C;
const char32_t YI_SYLLABLE_ZHY = 0xA35D;
const char32_t YI_SYLLABLE_ZHYP = 0xA35E;
const char32_t YI_SYLLABLE_ZHYRX = 0xA35F;
const char32_t YI_SYLLABLE_ZHYR = 0xA360;
const char32_t YI_SYLLABLE_CHAT = 0xA361;
const char32_t YI_SYLLABLE_CHAX = 0xA362;
const char32_t YI_SYLLABLE_CHA = 0xA363;
const char32_t YI_SYLLABLE_CHAP = 0xA364;
const char32_t YI_SYLLABLE_CHUOT = 0xA365;
const char32_t YI_SYLLABLE_CHUOX = 0xA366;
const char32_t YI_SYLLABLE_CHUO = 0xA367;
const char32_t YI_SYLLABLE_CHUOP = 0xA368;
const char32_t YI_SYLLABLE_CHOT = 0xA369;
const char32_t YI_SYLLABLE_CHOX = 0xA36A;
const char32_t YI_SYLLABLE_CHO = 0xA36B;
const char32_t YI_SYLLABLE_CHOP = 0xA36C;
const char32_t YI_SYLLABLE_CHET = 0xA36D;
const char32_t YI_SYLLABLE_CHEX = 0xA36E;
const char32_t YI_SYLLABLE_CHE = 0xA36F;
const char32_t YI_SYLLABLE_CHEP = 0xA370;
const char32_t YI_SYLLABLE_CHUX = 0xA371;
const char32_t YI_SYLLABLE_CHU = 0xA372;
const char32_t YI_SYLLABLE_CHUP = 0xA373;
const char32_t YI_SYLLABLE_CHURX = 0xA374;
const char32_t YI_SYLLABLE_CHUR = 0xA375;
const char32_t YI_SYLLABLE_CHYT = 0xA376;
const char32_t YI_SYLLABLE_CHYX = 0xA377;
const char32_t YI_SYLLABLE_CHY = 0xA378;
const char32_t YI_SYLLABLE_CHYP = 0xA379;
const char32_t YI_SYLLABLE_CHYRX = 0xA37A;
const char32_t YI_SYLLABLE_CHYR = 0xA37B;
const char32_t YI_SYLLABLE_RRAX = 0xA37C;
const char32_t YI_SYLLABLE_RRA = 0xA37D;
const char32_t YI_SYLLABLE_RRUOX = 0xA37E;
const char32_t YI_SYLLABLE_RRUO = 0xA37F;
const char32_t YI_SYLLABLE_RROT = 0xA380;
const char32_t YI_SYLLABLE_RROX = 0xA381;
const char32_t YI_SYLLABLE_RRO = 0xA382;
const char32_t YI_SYLLABLE_RROP = 0xA383;
const char32_t YI_SYLLABLE_RRET = 0xA384;
const char32_t YI_SYLLABLE_RREX = 0xA385;
const char32_t YI_SYLLABLE_RRE = 0xA386;
const char32_t YI_SYLLABLE_RREP = 0xA387;
const char32_t YI_SYLLABLE_RRUT = 0xA388;
const char32_t YI_SYLLABLE_RRUX = 0xA389;
const char32_t YI_SYLLABLE_RRU = 0xA38A;
const char32_t YI_SYLLABLE_RRUP = 0xA38B;
const char32_t YI_SYLLABLE_RRURX = 0xA38C;
const char32_t YI_SYLLABLE_RRUR = 0xA38D;
const char32_t YI_SYLLABLE_RRYT = 0xA38E;
const char32_t YI_SYLLABLE_RRYX = 0xA38F;
const char32_t YI_SYLLABLE_RRY = 0xA390;
const char32_t YI_SYLLABLE_RRYP = 0xA391;
const char32_t YI_SYLLABLE_RRYRX = 0xA392;
const char32_t YI_SYLLABLE_RRYR = 0xA393;
const char32_t YI_SYLLABLE_NRAT = 0xA394;
const char32_t YI_SYLLABLE_NRAX = 0xA395;
const char32_t YI_SYLLABLE_NRA = 0xA396;
const char32_t YI_SYLLABLE_NRAP = 0xA397;
const char32_t YI_SYLLABLE_NROX = 0xA398;
const char32_t YI_SYLLABLE_NRO = 0xA399;
const char32_t YI_SYLLABLE_NROP = 0xA39A;
const char32_t YI_SYLLABLE_NRET = 0xA39B;
const char32_t YI_SYLLABLE_NREX = 0xA39C;
const char32_t YI_SYLLABLE_NRE = 0xA39D;
const char32_t YI_SYLLABLE_NREP = 0xA39E;
const char32_t YI_SYLLABLE_NRUT = 0xA39F;
const char32_t YI_SYLLABLE_NRUX = 0xA3A0;
const char32_t YI_SYLLABLE_NRU = 0xA3A1;
const char32_t YI_SYLLABLE_NRUP = 0xA3A2;
const char32_t YI_SYLLABLE_NRURX = 0xA3A3;
const char32_t YI_SYLLABLE_NRUR = 0xA3A4;
const char32_t YI_SYLLABLE_NRYT = 0xA3A5;
const char32_t YI_SYLLABLE_NRYX = 0xA3A6;
const char32_t YI_SYLLABLE_NRY = 0xA3A7;
const char32_t YI_SYLLABLE_NRYP = 0xA3A8;
const char32_t YI_SYLLABLE_NRYRX = 0xA3A9;
const char32_t YI_SYLLABLE_NRYR = 0xA3AA;
const char32_t YI_SYLLABLE_SHAT = 0xA3AB;
const char32_t YI_SYLLABLE_SHAX = 0xA3AC;
const char32_t YI_SYLLABLE_SHA = 0xA3AD;
const char32_t YI_SYLLABLE_SHAP = 0xA3AE;
const char32_t YI_SYLLABLE_SHUOX = 0xA3AF;
const char32_t YI_SYLLABLE_SHUO = 0xA3B0;
const char32_t YI_SYLLABLE_SHUOP = 0xA3B1;
const char32_t YI_SYLLABLE_SHOT = 0xA3B2;
const char32_t YI_SYLLABLE_SHOX = 0xA3B3;
const char32_t YI_SYLLABLE_SHO = 0xA3B4;
const char32_t YI_SYLLABLE_SHOP = 0xA3B5;
const char32_t YI_SYLLABLE_SHET = 0xA3B6;
const char32_t YI_SYLLABLE_SHEX = 0xA3B7;
const char32_t YI_SYLLABLE_SHE = 0xA3B8;
const char32_t YI_SYLLABLE_SHEP = 0xA3B9;
const char32_t YI_SYLLABLE_SHUT = 0xA3BA;
const char32_t YI_SYLLABLE_SHUX = 0xA3BB;
const char32_t YI_SYLLABLE_SHU = 0xA3BC;
const char32_t YI_SYLLABLE_SHUP = 0xA3BD;
const char32_t YI_SYLLABLE_SHURX = 0xA3BE;
const char32_t YI_SYLLABLE_SHUR = 0xA3BF;
const char32_t YI_SYLLABLE_SHYT = 0xA3C0;
const char32_t YI_SYLLABLE_SHYX = 0xA3C1;
const char32_t YI_SYLLABLE_SHY = 0xA3C2;
const char32_t YI_SYLLABLE_SHYP = 0xA3C3;
const char32_t YI_SYLLABLE_SHYRX = 0xA3C4;
const char32_t YI_SYLLABLE_SHYR = 0xA3C5;
const char32_t YI_SYLLABLE_RAT = 0xA3C6;
const char32_t YI_SYLLABLE_RAX = 0xA3C7;
const char32_t YI_SYLLABLE_RA = 0xA3C8;
const char32_t YI_SYLLABLE_RAP = 0xA3C9;
const char32_t YI_SYLLABLE_RUOX = 0xA3CA;
const char32_t YI_SYLLABLE_RUO = 0xA3CB;
const char32_t YI_SYLLABLE_RUOP = 0xA3CC;
const char32_t YI_SYLLABLE_ROT = 0xA3CD;
const char32_t YI_SYLLABLE_ROX = 0xA3CE;
const char32_t YI_SYLLABLE_RO = 0xA3CF;
const char32_t YI_SYLLABLE_ROP = 0xA3D0;
const char32_t YI_SYLLABLE_REX = 0xA3D1;
const char32_t YI_SYLLABLE_RE = 0xA3D2;
const char32_t YI_SYLLABLE_REP = 0xA3D3;
const char32_t YI_SYLLABLE_RUT = 0xA3D4;
const char32_t YI_SYLLABLE_RUX = 0xA3D5;
const char32_t YI_SYLLABLE_RU = 0xA3D6;
const char32_t YI_SYLLABLE_RUP = 0xA3D7;
const char32_t YI_SYLLABLE_RURX = 0xA3D8;
const char32_t YI_SYLLABLE_RUR = 0xA3D9;
const char32_t YI_SYLLABLE_RYT = 0xA3DA;
const char32_t YI_SYLLABLE_RYX = 0xA3DB;
const char32_t YI_SYLLABLE_RY = 0xA3DC;
const char32_t YI_SYLLABLE_RYP = 0xA3DD;
const char32_t YI_SYLLABLE_RYRX = 0xA3DE;
const char32_t YI_SYLLABLE_RYR = 0xA3DF;
const char32_t YI_SYLLABLE_JIT = 0xA3E0;
const char32_t YI_SYLLABLE_JIX = 0xA3E1;
const char32_t YI_SYLLABLE_JI = 0xA3E2;
const char32_t YI_SYLLABLE_JIP = 0xA3E3;
const char32_t YI_SYLLABLE_JIET = 0xA3E4;
const char32_t YI_SYLLABLE_JIEX = 0xA3E5;
const char32_t YI_SYLLABLE_JIE = 0xA3E6;
const char32_t YI_SYLLABLE_JIEP = 0xA3E7;
const char32_t YI_SYLLABLE_JUOT = 0xA3E8;
const char32_t YI_SYLLABLE_JUOX = 0xA3E9;
const char32_t YI_SYLLABLE_JUO = 0xA3EA;
const char32_t YI_SYLLABLE_JUOP = 0xA3EB;
const char32_t YI_SYLLABLE_JOT = 0xA3EC;
const char32_t YI_SYLLABLE_JOX = 0xA3ED;
const char32_t YI_SYLLABLE_JO = 0xA3EE;
const char32_t YI_SYLLABLE_JOP = 0xA3EF;
const char32_t YI_SYLLABLE_JUT = 0xA3F0;
const char32_t YI_SYLLABLE_JUX = 0xA3F1;
const char32_t YI_SYLLABLE_JU = 0xA3F2;
const char32_t YI_SYLLABLE_JUP = 0xA3F3;
const char32_t YI_SYLLABLE_JURX = 0xA3F4;
const char32_t YI_SYLLABLE_JUR = 0xA3F5;
const char32_t YI_SYLLABLE_JYT = 0xA3F6;
const char32_t YI_SYLLABLE_JYX = 0xA3F7;
const char32_t YI_SYLLABLE_JY = 0xA3F8;
const char32_t YI_SYLLABLE_JYP = 0xA3F9;
const char32_t YI_SYLLABLE_JYRX = 0xA3FA;
const char32_t YI_SYLLABLE_JYR = 0xA3FB;
const char32_t YI_SYLLABLE_QIT = 0xA3FC;
const char32_t YI_SYLLABLE_QIX = 0xA3FD;
const char32_t YI_SYLLABLE_QI = 0xA3FE;
const char32_t YI_SYLLABLE_QIP = 0xA3FF;
const char32_t YI_SYLLABLE_QIET = 0xA400;
const char32_t YI_SYLLABLE_QIEX = 0xA401;
const char32_t YI_SYLLABLE_QIE = 0xA402;
const char32_t YI_SYLLABLE_QIEP = 0xA403;
const char32_t YI_SYLLABLE_QUOT = 0xA404;
const char32_t YI_SYLLABLE_QUOX = 0xA405;
const char32_t YI_SYLLABLE_QUO = 0xA406;
const char32_t YI_SYLLABLE_QUOP = 0xA407;
const char32_t YI_SYLLABLE_QOT = 0xA408;
const char32_t YI_SYLLABLE_QOX = 0xA409;
const char32_t YI_SYLLABLE_QO = 0xA40A;
const char32_t YI_SYLLABLE_QOP = 0xA40B;
const char32_t YI_SYLLABLE_QUT = 0xA40C;
const char32_t YI_SYLLABLE_QUX = 0xA40D;
const char32_t YI_SYLLABLE_QU = 0xA40E;
const char32_t YI_SYLLABLE_QUP = 0xA40F;
const char32_t YI_SYLLABLE_QURX = 0xA410;
const char32_t YI_SYLLABLE_QUR = 0xA411;
const char32_t YI_SYLLABLE_QYT = 0xA412;
const char32_t YI_SYLLABLE_QYX = 0xA413;
const char32_t YI_SYLLABLE_QY = 0xA414;
const char32_t YI_SYLLABLE_QYP = 0xA415;
const char32_t YI_SYLLABLE_QYRX = 0xA416;
const char32_t YI_SYLLABLE_QYR = 0xA417;
const char32_t YI_SYLLABLE_JJIT = 0xA418;
const char32_t YI_SYLLABLE_JJIX = 0xA419;
const char32_t YI_SYLLABLE_JJI = 0xA41A;
const char32_t YI_SYLLABLE_JJIP = 0xA41B;
const char32_t YI_SYLLABLE_JJIET = 0xA41C;
const char32_t YI_SYLLABLE_JJIEX = 0xA41D;
const char32_t YI_SYLLABLE_JJIE = 0xA41E;
const char32_t YI_SYLLABLE_JJIEP = 0xA41F;
const char32_t YI_SYLLABLE_JJUOX = 0xA420;
const char32_t YI_SYLLABLE_JJUO = 0xA421;
const char32_t YI_SYLLABLE_JJUOP = 0xA422;
const char32_t YI_SYLLABLE_JJOT = 0xA423;
const char32_t YI_SYLLABLE_JJOX = 0xA424;
const char32_t YI_SYLLABLE_JJO = 0xA425;
const char32_t YI_SYLLABLE_JJOP = 0xA426;
const char32_t YI_SYLLABLE_JJUT = 0xA427;
const char32_t YI_SYLLABLE_JJUX = 0xA428;
const char32_t YI_SYLLABLE_JJU = 0xA429;
const char32_t YI_SYLLABLE_JJUP = 0xA42A;
const char32_t YI_SYLLABLE_JJURX = 0xA42B;
const char32_t YI_SYLLABLE_JJUR = 0xA42C;
const char32_t YI_SYLLABLE_JJYT = 0xA42D;
const char32_t YI_SYLLABLE_JJYX = 0xA42E;
const char32_t YI_SYLLABLE_JJY = 0xA42F;
const char32_t YI_SYLLABLE_JJYP = 0xA430;
const char32_t YI_SYLLABLE_NJIT = 0xA431;
const char32_t YI_SYLLABLE_NJIX = 0xA432;
const char32_t YI_SYLLABLE_NJI = 0xA433;
const char32_t YI_SYLLABLE_NJIP = 0xA434;
const char32_t YI_SYLLABLE_NJIET = 0xA435;
const char32_t YI_SYLLABLE_NJIEX = 0xA436;
const char32_t YI_SYLLABLE_NJIE = 0xA437;
const char32_t YI_SYLLABLE_NJIEP = 0xA438;
const char32_t YI_SYLLABLE_NJUOX = 0xA439;
const char32_t YI_SYLLABLE_NJUO = 0xA43A;
const char32_t YI_SYLLABLE_NJOT = 0xA43B;
const char32_t YI_SYLLABLE_NJOX = 0xA43C;
const char32_t YI_SYLLABLE_NJO = 0xA43D;
const char32_t YI_SYLLABLE_NJOP = 0xA43E;
const char32_t YI_SYLLABLE_NJUX = 0xA43F;
const char32_t YI_SYLLABLE_NJU = 0xA440;
const char32_t YI_SYLLABLE_NJUP = 0xA441;
const char32_t YI_SYLLABLE_NJURX = 0xA442;
const char32_t YI_SYLLABLE_NJUR = 0xA443;
const char32_t YI_SYLLABLE_NJYT = 0xA444;
const char32_t YI_SYLLABLE_NJYX = 0xA445;
const char32_t YI_SYLLABLE_NJY = 0xA446;
const char32_t YI_SYLLABLE_NJYP = 0xA447;
const char32_t YI_SYLLABLE_NJYRX = 0xA448;
const char32_t YI_SYLLABLE_NJYR = 0xA449;
const char32_t YI_SYLLABLE_NYIT = 0xA44A;
const char32_t YI_SYLLABLE_NYIX = 0xA44B;
const char32_t YI_SYLLABLE_NYI = 0xA44C;
const char32_t YI_SYLLABLE_NYIP = 0xA44D;
const char32_t YI_SYLLABLE_NYIET = 0xA44E;
const char32_t YI_SYLLABLE_NYIEX = 0xA44F;
const char32_t YI_SYLLABLE_NYIE = 0xA450;
const char32_t YI_SYLLABLE_NYIEP = 0xA451;
const char32_t YI_SYLLABLE_NYUOX = 0xA452;
const char32_t YI_SYLLABLE_NYUO = 0xA453;
const char32_t YI_SYLLABLE_NYUOP = 0xA454;
const char32_t YI_SYLLABLE_NYOT = 0xA455;
const char32_t YI_SYLLABLE_NYOX = 0xA456;
const char32_t YI_SYLLABLE_NYO = 0xA457;
const char32_t YI_SYLLABLE_NYOP = 0xA458;
const char32_t YI_SYLLABLE_NYUT = 0xA459;
const char32_t YI_SYLLABLE_NYUX = 0xA45A;
const char32_t YI_SYLLABLE_NYU = 0xA45B;
const char32_t YI_SYLLABLE_NYUP = 0xA45C;
const char32_t YI_SYLLABLE_XIT = 0xA45D;
const char32_t YI_SYLLABLE_XIX = 0xA45E;
const char32_t YI_SYLLABLE_XI = 0xA45F;
const char32_t YI_SYLLABLE_XIP = 0xA460;
const char32_t YI_SYLLABLE_XIET = 0xA461;
const char32_t YI_SYLLABLE_XIEX = 0xA462;
const char32_t YI_SYLLABLE_XIE = 0xA463;
const char32_t YI_SYLLABLE_XIEP = 0xA464;
const char32_t YI_SYLLABLE_XUOX = 0xA465;
const char32_t YI_SYLLABLE_XUO = 0xA466;
const char32_t YI_SYLLABLE_XOT = 0xA467;
const char32_t YI_SYLLABLE_XOX = 0xA468;
const char32_t YI_SYLLABLE_XO = 0xA469;
const char32_t YI_SYLLABLE_XOP = 0xA46A;
const char32_t YI_SYLLABLE_XYT = 0xA46B;
const char32_t YI_SYLLABLE_XYX = 0xA46C;
const char32_t YI_SYLLABLE_XY = 0xA46D;
const char32_t YI_SYLLABLE_XYP = 0xA46E;
const char32_t YI_SYLLABLE_XYRX = 0xA46F;
const char32_t YI_SYLLABLE_XYR = 0xA470;
const char32_t YI_SYLLABLE_YIT = 0xA471;
const char32_t YI_SYLLABLE_YIX = 0xA472;
const char32_t YI_SYLLABLE_YI = 0xA473;
const char32_t YI_SYLLABLE_YIP = 0xA474;
const char32_t YI_SYLLABLE_YIET = 0xA475;
const char32_t YI_SYLLABLE_YIEX = 0xA476;
const char32_t YI_SYLLABLE_YIE = 0xA477;
const char32_t YI_SYLLABLE_YIEP = 0xA478;
const char32_t YI_SYLLABLE_YUOT = 0xA479;
const char32_t YI_SYLLABLE_YUOX = 0xA47A;
const char32_t YI_SYLLABLE_YUO = 0xA47B;
const char32_t YI_SYLLABLE_YUOP = 0xA47C;
const char32_t YI_SYLLABLE_YOT = 0xA47D;
const char32_t YI_SYLLABLE_YOX = 0xA47E;
const char32_t YI_SYLLABLE_YO = 0xA47F;
const char32_t YI_SYLLABLE_YOP = 0xA480;
const char32_t YI_SYLLABLE_YUT = 0xA481;
const char32_t YI_SYLLABLE_YUX = 0xA482;
const char32_t YI_SYLLABLE_YU = 0xA483;
const char32_t YI_SYLLABLE_YUP = 0xA484;
const char32_t YI_SYLLABLE_YURX = 0xA485;
const char32_t YI_SYLLABLE_YUR = 0xA486;
const char32_t YI_SYLLABLE_YYT = 0xA487;
const char32_t YI_SYLLABLE_YYX = 0xA488;
const char32_t YI_SYLLABLE_YY = 0xA489;
const char32_t YI_SYLLABLE_YYP = 0xA48A;
const char32_t YI_SYLLABLE_YYRX = 0xA48B;
const char32_t YI_SYLLABLE_YYR = 0xA48C;
const char32_t YI_RADICAL_QOT = 0xA490;
const char32_t YI_RADICAL_LI = 0xA491;
const char32_t YI_RADICAL_KIT = 0xA492;
const char32_t YI_RADICAL_NYIP = 0xA493;
const char32_t YI_RADICAL_CYP = 0xA494;
const char32_t YI_RADICAL_SSI = 0xA495;
const char32_t YI_RADICAL_GGOP = 0xA496;
const char32_t YI_RADICAL_GEP = 0xA497;
const char32_t YI_RADICAL_MI = 0xA498;
const char32_t YI_RADICAL_HXIT = 0xA499;
const char32_t YI_RADICAL_LYR = 0xA49A;
const char32_t YI_RADICAL_BBUT = 0xA49B;
const char32_t YI_RADICAL_MOP = 0xA49C;
const char32_t YI_RADICAL_YO = 0xA49D;
const char32_t YI_RADICAL_PUT = 0xA49E;
const char32_t YI_RADICAL_HXUO = 0xA49F;
const char32_t YI_RADICAL_TAT = 0xA4A0;
const char32_t YI_RADICAL_GA = 0xA4A1;
const char32_t YI_RADICAL_ZUP = 0xA4A2;
const char32_t YI_RADICAL_CYT = 0xA4A3;
const char32_t YI_RADICAL_DDUR = 0xA4A4;
const char32_t YI_RADICAL_BUR = 0xA4A5;
const char32_t YI_RADICAL_GGUO = 0xA4A6;
const char32_t YI_RADICAL_NYOP = 0xA4A7;
const char32_t YI_RADICAL_TU = 0xA4A8;
const char32_t YI_RADICAL_OP = 0xA4A9;
const char32_t YI_RADICAL_JJUT = 0xA4AA;
const char32_t YI_RADICAL_ZOT = 0xA4AB;
const char32_t YI_RADICAL_PYT = 0xA4AC;
const char32_t YI_RADICAL_HMO = 0xA4AD;
const char32_t YI_RADICAL_YIT = 0xA4AE;
const char32_t YI_RADICAL_VUR = 0xA4AF;
const char32_t YI_RADICAL_SHY = 0xA4B0;
const char32_t YI_RADICAL_VEP = 0xA4B1;
const char32_t YI_RADICAL_ZA = 0xA4B2;
const char32_t YI_RADICAL_JO = 0xA4B3;
const char32_t YI_RADICAL_NZUP = 0xA4B4;
const char32_t YI_RADICAL_JJY = 0xA4B5;
const char32_t YI_RADICAL_GOT = 0xA4B6;
const char32_t YI_RADICAL_JJIE = 0xA4B7;
const char32_t YI_RADICAL_WO = 0xA4B8;
const char32_t YI_RADICAL_DU = 0xA4B9;
const char32_t YI_RADICAL_SHUR = 0xA4BA;
const char32_t YI_RADICAL_LIE = 0xA4BB;
const char32_t YI_RADICAL_CY = 0xA4BC;
const char32_t YI_RADICAL_CUOP = 0xA4BD;
const char32_t YI_RADICAL_CIP = 0xA4BE;
const char32_t YI_RADICAL_HXOP = 0xA4BF;
const char32_t YI_RADICAL_SHAT = 0xA4C0;
const char32_t YI_RADICAL_ZUR = 0xA4C1;
const char32_t YI_RADICAL_SHOP = 0xA4C2;
const char32_t YI_RADICAL_CHE = 0xA4C3;
const char32_t YI_RADICAL_ZZIET = 0xA4C4;
const char32_t YI_RADICAL_NBIE = 0xA4C5;
const char32_t YI_RADICAL_KE = 0xA4C6;
const char32_t LISU_LETTER_BA = 0xA4D0;
const char32_t LISU_LETTER_PA = 0xA4D1;
const char32_t LISU_LETTER_PHA = 0xA4D2;
const char32_t LISU_LETTER_DA = 0xA4D3;
const char32_t LISU_LETTER_TA = 0xA4D4;
const char32_t LISU_LETTER_THA = 0xA4D5;
const char32_t LISU_LETTER_GA = 0xA4D6;
const char32_t LISU_LETTER_KA = 0xA4D7;
const char32_t LISU_LETTER_KHA = 0xA4D8;
const char32_t LISU_LETTER_JA = 0xA4D9;
const char32_t LISU_LETTER_CA = 0xA4DA;
const char32_t LISU_LETTER_CHA = 0xA4DB;
const char32_t LISU_LETTER_DZA = 0xA4DC;
const char32_t LISU_LETTER_TSA = 0xA4DD;
const char32_t LISU_LETTER_TSHA = 0xA4DE;
const char32_t LISU_LETTER_MA = 0xA4DF;
const char32_t LISU_LETTER_NA = 0xA4E0;
const char32_t LISU_LETTER_LA = 0xA4E1;
const char32_t LISU_LETTER_SA = 0xA4E2;
const char32_t LISU_LETTER_ZHA = 0xA4E3;
const char32_t LISU_LETTER_ZA = 0xA4E4;
const char32_t LISU_LETTER_NGA = 0xA4E5;
const char32_t LISU_LETTER_HA = 0xA4E6;
const char32_t LISU_LETTER_XA = 0xA4E7;
const char32_t LISU_LETTER_HHA = 0xA4E8;
const char32_t LISU_LETTER_FA = 0xA4E9;
const char32_t LISU_LETTER_WA = 0xA4EA;
const char32_t LISU_LETTER_SHA = 0xA4EB;
const char32_t LISU_LETTER_YA = 0xA4EC;
const char32_t LISU_LETTER_GHA = 0xA4ED;
const char32_t LISU_LETTER_A = 0xA4EE;
const char32_t LISU_LETTER_AE = 0xA4EF;
const char32_t LISU_LETTER_E = 0xA4F0;
const char32_t LISU_LETTER_EU = 0xA4F1;
const char32_t LISU_LETTER_I = 0xA4F2;
const char32_t LISU_LETTER_O = 0xA4F3;
const char32_t LISU_LETTER_U = 0xA4F4;
const char32_t LISU_LETTER_UE = 0xA4F5;
const char32_t LISU_LETTER_UH = 0xA4F6;
const char32_t LISU_LETTER_OE = 0xA4F7;
const char32_t LISU_LETTER_TONE_MYA_TI = 0xA4F8;
const char32_t LISU_LETTER_TONE_NA_PO = 0xA4F9;
const char32_t LISU_LETTER_TONE_MYA_CYA = 0xA4FA;
const char32_t LISU_LETTER_TONE_MYA_BO = 0xA4FB;
const char32_t LISU_LETTER_TONE_MYA_NA = 0xA4FC;
const char32_t LISU_LETTER_TONE_MYA_JEU = 0xA4FD;
const char32_t LISU_PUNCTUATION_COMMA = 0xA4FE;
const char32_t LISU_PUNCTUATION_FULL_STOP = 0xA4FF;
const char32_t VAI_SYLLABLE_EE = 0xA500;
const char32_t VAI_SYLLABLE_EEN = 0xA501;
const char32_t VAI_SYLLABLE_HEE = 0xA502;
const char32_t VAI_SYLLABLE_WEE = 0xA503;
const char32_t VAI_SYLLABLE_WEEN = 0xA504;
const char32_t VAI_SYLLABLE_PEE = 0xA505;
const char32_t VAI_SYLLABLE_BHEE = 0xA506;
const char32_t VAI_SYLLABLE_BEE = 0xA507;
const char32_t VAI_SYLLABLE_MBEE = 0xA508;
const char32_t VAI_SYLLABLE_KPEE = 0xA509;
const char32_t VAI_SYLLABLE_MGBEE = 0xA50A;
const char32_t VAI_SYLLABLE_GBEE = 0xA50B;
const char32_t VAI_SYLLABLE_FEE = 0xA50C;
const char32_t VAI_SYLLABLE_VEE = 0xA50D;
const char32_t VAI_SYLLABLE_TEE = 0xA50E;
const char32_t VAI_SYLLABLE_THEE = 0xA50F;
const char32_t VAI_SYLLABLE_DHEE = 0xA510;
const char32_t VAI_SYLLABLE_DHHEE = 0xA511;
const char32_t VAI_SYLLABLE_LEE = 0xA512;
const char32_t VAI_SYLLABLE_REE = 0xA513;
const char32_t VAI_SYLLABLE_DEE = 0xA514;
const char32_t VAI_SYLLABLE_NDEE = 0xA515;
const char32_t VAI_SYLLABLE_SEE = 0xA516;
const char32_t VAI_SYLLABLE_SHEE = 0xA517;
const char32_t VAI_SYLLABLE_ZEE = 0xA518;
const char32_t VAI_SYLLABLE_ZHEE = 0xA519;
const char32_t VAI_SYLLABLE_CEE = 0xA51A;
const char32_t VAI_SYLLABLE_JEE = 0xA51B;
const char32_t VAI_SYLLABLE_NJEE = 0xA51C;
const char32_t VAI_SYLLABLE_YEE = 0xA51D;
const char32_t VAI_SYLLABLE_KEE = 0xA51E;
const char32_t VAI_SYLLABLE_NGGEE = 0xA51F;
const char32_t VAI_SYLLABLE_GEE = 0xA520;
const char32_t VAI_SYLLABLE_MEE = 0xA521;
const char32_t VAI_SYLLABLE_NEE = 0xA522;
const char32_t VAI_SYLLABLE_NYEE = 0xA523;
const char32_t VAI_SYLLABLE_I = 0xA524;
const char32_t VAI_SYLLABLE_IN = 0xA525;
const char32_t VAI_SYLLABLE_HI = 0xA526;
const char32_t VAI_SYLLABLE_HIN = 0xA527;
const char32_t VAI_SYLLABLE_WI = 0xA528;
const char32_t VAI_SYLLABLE_WIN = 0xA529;
const char32_t VAI_SYLLABLE_PI = 0xA52A;
const char32_t VAI_SYLLABLE_BHI = 0xA52B;
const char32_t VAI_SYLLABLE_BI = 0xA52C;
const char32_t VAI_SYLLABLE_MBI = 0xA52D;
const char32_t VAI_SYLLABLE_KPI = 0xA52E;
const char32_t VAI_SYLLABLE_MGBI = 0xA52F;
const char32_t VAI_SYLLABLE_GBI = 0xA530;
const char32_t VAI_SYLLABLE_FI = 0xA531;
const char32_t VAI_SYLLABLE_VI = 0xA532;
const char32_t VAI_SYLLABLE_TI = 0xA533;
const char32_t VAI_SYLLABLE_THI = 0xA534;
const char32_t VAI_SYLLABLE_DHI = 0xA535;
const char32_t VAI_SYLLABLE_DHHI = 0xA536;
const char32_t VAI_SYLLABLE_LI = 0xA537;
const char32_t VAI_SYLLABLE_RI = 0xA538;
const char32_t VAI_SYLLABLE_DI = 0xA539;
const char32_t VAI_SYLLABLE_NDI = 0xA53A;
const char32_t VAI_SYLLABLE_SI = 0xA53B;
const char32_t VAI_SYLLABLE_SHI = 0xA53C;
const char32_t VAI_SYLLABLE_ZI = 0xA53D;
const char32_t VAI_SYLLABLE_ZHI = 0xA53E;
const char32_t VAI_SYLLABLE_CI = 0xA53F;
const char32_t VAI_SYLLABLE_JI = 0xA540;
const char32_t VAI_SYLLABLE_NJI = 0xA541;
const char32_t VAI_SYLLABLE_YI = 0xA542;
const char32_t VAI_SYLLABLE_KI = 0xA543;
const char32_t VAI_SYLLABLE_NGGI = 0xA544;
const char32_t VAI_SYLLABLE_GI = 0xA545;
const char32_t VAI_SYLLABLE_MI = 0xA546;
const char32_t VAI_SYLLABLE_NI = 0xA547;
const char32_t VAI_SYLLABLE_NYI = 0xA548;
const char32_t VAI_SYLLABLE_A = 0xA549;
const char32_t VAI_SYLLABLE_AN = 0xA54A;
const char32_t VAI_SYLLABLE_NGAN = 0xA54B;
const char32_t VAI_SYLLABLE_HA = 0xA54C;
const char32_t VAI_SYLLABLE_HAN = 0xA54D;
const char32_t VAI_SYLLABLE_WA = 0xA54E;
const char32_t VAI_SYLLABLE_WAN = 0xA54F;
const char32_t VAI_SYLLABLE_PA = 0xA550;
const char32_t VAI_SYLLABLE_BHA = 0xA551;
const char32_t VAI_SYLLABLE_BA = 0xA552;
const char32_t VAI_SYLLABLE_MBA = 0xA553;
const char32_t VAI_SYLLABLE_KPA = 0xA554;
const char32_t VAI_SYLLABLE_KPAN = 0xA555;
const char32_t VAI_SYLLABLE_MGBA = 0xA556;
const char32_t VAI_SYLLABLE_GBA = 0xA557;
const char32_t VAI_SYLLABLE_FA = 0xA558;
const char32_t VAI_SYLLABLE_VA = 0xA559;
const char32_t VAI_SYLLABLE_TA = 0xA55A;
const char32_t VAI_SYLLABLE_THA = 0xA55B;
const char32_t VAI_SYLLABLE_DHA = 0xA55C;
const char32_t VAI_SYLLABLE_DHHA = 0xA55D;
const char32_t VAI_SYLLABLE_LA = 0xA55E;
const char32_t VAI_SYLLABLE_RA = 0xA55F;
const char32_t VAI_SYLLABLE_DA = 0xA560;
const char32_t VAI_SYLLABLE_NDA = 0xA561;
const char32_t VAI_SYLLABLE_SA = 0xA562;
const char32_t VAI_SYLLABLE_SHA = 0xA563;
const char32_t VAI_SYLLABLE_ZA = 0xA564;
const char32_t VAI_SYLLABLE_ZHA = 0xA565;
const char32_t VAI_SYLLABLE_CA = 0xA566;
const char32_t VAI_SYLLABLE_JA = 0xA567;
const char32_t VAI_SYLLABLE_NJA = 0xA568;
const char32_t VAI_SYLLABLE_YA = 0xA569;
const char32_t VAI_SYLLABLE_KA = 0xA56A;
const char32_t VAI_SYLLABLE_KAN = 0xA56B;
const char32_t VAI_SYLLABLE_NGGA = 0xA56C;
const char32_t VAI_SYLLABLE_GA = 0xA56D;
const char32_t VAI_SYLLABLE_MA = 0xA56E;
const char32_t VAI_SYLLABLE_NA = 0xA56F;
const char32_t VAI_SYLLABLE_NYA = 0xA570;
const char32_t VAI_SYLLABLE_OO = 0xA571;
const char32_t VAI_SYLLABLE_OON = 0xA572;
const char32_t VAI_SYLLABLE_HOO = 0xA573;
const char32_t VAI_SYLLABLE_WOO = 0xA574;
const char32_t VAI_SYLLABLE_WOON = 0xA575;
const char32_t VAI_SYLLABLE_POO = 0xA576;
const char32_t VAI_SYLLABLE_BHOO = 0xA577;
const char32_t VAI_SYLLABLE_BOO = 0xA578;
const char32_t VAI_SYLLABLE_MBOO = 0xA579;
const char32_t VAI_SYLLABLE_KPOO = 0xA57A;
const char32_t VAI_SYLLABLE_MGBOO = 0xA57B;
const char32_t VAI_SYLLABLE_GBOO = 0xA57C;
const char32_t VAI_SYLLABLE_FOO = 0xA57D;
const char32_t VAI_SYLLABLE_VOO = 0xA57E;
const char32_t VAI_SYLLABLE_TOO = 0xA57F;
const char32_t VAI_SYLLABLE_THOO = 0xA580;
const char32_t VAI_SYLLABLE_DHOO = 0xA581;
const char32_t VAI_SYLLABLE_DHHOO = 0xA582;
const char32_t VAI_SYLLABLE_LOO = 0xA583;
const char32_t VAI_SYLLABLE_ROO = 0xA584;
const char32_t VAI_SYLLABLE_DOO = 0xA585;
const char32_t VAI_SYLLABLE_NDOO = 0xA586;
const char32_t VAI_SYLLABLE_SOO = 0xA587;
const char32_t VAI_SYLLABLE_SHOO = 0xA588;
const char32_t VAI_SYLLABLE_ZOO = 0xA589;
const char32_t VAI_SYLLABLE_ZHOO = 0xA58A;
const char32_t VAI_SYLLABLE_COO = 0xA58B;
const char32_t VAI_SYLLABLE_JOO = 0xA58C;
const char32_t VAI_SYLLABLE_NJOO = 0xA58D;
const char32_t VAI_SYLLABLE_YOO = 0xA58E;
const char32_t VAI_SYLLABLE_KOO = 0xA58F;
const char32_t VAI_SYLLABLE_NGGOO = 0xA590;
const char32_t VAI_SYLLABLE_GOO = 0xA591;
const char32_t VAI_SYLLABLE_MOO = 0xA592;
const char32_t VAI_SYLLABLE_NOO = 0xA593;
const char32_t VAI_SYLLABLE_NYOO = 0xA594;
const char32_t VAI_SYLLABLE_U = 0xA595;
const char32_t VAI_SYLLABLE_UN = 0xA596;
const char32_t VAI_SYLLABLE_HU = 0xA597;
const char32_t VAI_SYLLABLE_HUN = 0xA598;
const char32_t VAI_SYLLABLE_WU = 0xA599;
const char32_t VAI_SYLLABLE_WUN = 0xA59A;
const char32_t VAI_SYLLABLE_PU = 0xA59B;
const char32_t VAI_SYLLABLE_BHU = 0xA59C;
const char32_t VAI_SYLLABLE_BU = 0xA59D;
const char32_t VAI_SYLLABLE_MBU = 0xA59E;
const char32_t VAI_SYLLABLE_KPU = 0xA59F;
const char32_t VAI_SYLLABLE_MGBU = 0xA5A0;
const char32_t VAI_SYLLABLE_GBU = 0xA5A1;
const char32_t VAI_SYLLABLE_FU = 0xA5A2;
const char32_t VAI_SYLLABLE_VU = 0xA5A3;
const char32_t VAI_SYLLABLE_TU = 0xA5A4;
const char32_t VAI_SYLLABLE_THU = 0xA5A5;
const char32_t VAI_SYLLABLE_DHU = 0xA5A6;
const char32_t VAI_SYLLABLE_DHHU = 0xA5A7;
const char32_t VAI_SYLLABLE_LU = 0xA5A8;
const char32_t VAI_SYLLABLE_RU = 0xA5A9;
const char32_t VAI_SYLLABLE_DU = 0xA5AA;
const char32_t VAI_SYLLABLE_NDU = 0xA5AB;
const char32_t VAI_SYLLABLE_SU = 0xA5AC;
const char32_t VAI_SYLLABLE_SHU = 0xA5AD;
const char32_t VAI_SYLLABLE_ZU = 0xA5AE;
const char32_t VAI_SYLLABLE_ZHU = 0xA5AF;
const char32_t VAI_SYLLABLE_CU = 0xA5B0;
const char32_t VAI_SYLLABLE_JU = 0xA5B1;
const char32_t VAI_SYLLABLE_NJU = 0xA5B2;
const char32_t VAI_SYLLABLE_YU = 0xA5B3;
const char32_t VAI_SYLLABLE_KU = 0xA5B4;
const char32_t VAI_SYLLABLE_NGGU = 0xA5B5;
const char32_t VAI_SYLLABLE_GU = 0xA5B6;
const char32_t VAI_SYLLABLE_MU = 0xA5B7;
const char32_t VAI_SYLLABLE_NU = 0xA5B8;
const char32_t VAI_SYLLABLE_NYU = 0xA5B9;
const char32_t VAI_SYLLABLE_O = 0xA5BA;
const char32_t VAI_SYLLABLE_ON = 0xA5BB;
const char32_t VAI_SYLLABLE_NGON = 0xA5BC;
const char32_t VAI_SYLLABLE_HO = 0xA5BD;
const char32_t VAI_SYLLABLE_HON = 0xA5BE;
const char32_t VAI_SYLLABLE_WO = 0xA5BF;
const char32_t VAI_SYLLABLE_WON = 0xA5C0;
const char32_t VAI_SYLLABLE_PO = 0xA5C1;
const char32_t VAI_SYLLABLE_BHO = 0xA5C2;
const char32_t VAI_SYLLABLE_BO = 0xA5C3;
const char32_t VAI_SYLLABLE_MBO = 0xA5C4;
const char32_t VAI_SYLLABLE_KPO = 0xA5C5;
const char32_t VAI_SYLLABLE_MGBO = 0xA5C6;
const char32_t VAI_SYLLABLE_GBO = 0xA5C7;
const char32_t VAI_SYLLABLE_GBON = 0xA5C8;
const char32_t VAI_SYLLABLE_FO = 0xA5C9;
const char32_t VAI_SYLLABLE_VO = 0xA5CA;
const char32_t VAI_SYLLABLE_TO = 0xA5CB;
const char32_t VAI_SYLLABLE_THO = 0xA5CC;
const char32_t VAI_SYLLABLE_DHO = 0xA5CD;
const char32_t VAI_SYLLABLE_DHHO = 0xA5CE;
const char32_t VAI_SYLLABLE_LO = 0xA5CF;
const char32_t VAI_SYLLABLE_RO = 0xA5D0;
const char32_t VAI_SYLLABLE_DO = 0xA5D1;
const char32_t VAI_SYLLABLE_NDO = 0xA5D2;
const char32_t VAI_SYLLABLE_SO = 0xA5D3;
const char32_t VAI_SYLLABLE_SHO = 0xA5D4;
const char32_t VAI_SYLLABLE_ZO = 0xA5D5;
const char32_t VAI_SYLLABLE_ZHO = 0xA5D6;
const char32_t VAI_SYLLABLE_CO = 0xA5D7;
const char32_t VAI_SYLLABLE_JO = 0xA5D8;
const char32_t VAI_SYLLABLE_NJO = 0xA5D9;
const char32_t VAI_SYLLABLE_YO = 0xA5DA;
const char32_t VAI_SYLLABLE_KO = 0xA5DB;
const char32_t VAI_SYLLABLE_NGGO = 0xA5DC;
const char32_t VAI_SYLLABLE_GO = 0xA5DD;
const char32_t VAI_SYLLABLE_MO = 0xA5DE;
const char32_t VAI_SYLLABLE_NO = 0xA5DF;
const char32_t VAI_SYLLABLE_NYO = 0xA5E0;
const char32_t VAI_SYLLABLE_E = 0xA5E1;
const char32_t VAI_SYLLABLE_EN = 0xA5E2;
const char32_t VAI_SYLLABLE_NGEN = 0xA5E3;
const char32_t VAI_SYLLABLE_HE = 0xA5E4;
const char32_t VAI_SYLLABLE_HEN = 0xA5E5;
const char32_t VAI_SYLLABLE_WE = 0xA5E6;
const char32_t VAI_SYLLABLE_WEN = 0xA5E7;
const char32_t VAI_SYLLABLE_PE = 0xA5E8;
const char32_t VAI_SYLLABLE_BHE = 0xA5E9;
const char32_t VAI_SYLLABLE_BE = 0xA5EA;
const char32_t VAI_SYLLABLE_MBE = 0xA5EB;
const char32_t VAI_SYLLABLE_KPE = 0xA5EC;
const char32_t VAI_SYLLABLE_KPEN = 0xA5ED;
const char32_t VAI_SYLLABLE_MGBE = 0xA5EE;
const char32_t VAI_SYLLABLE_GBE = 0xA5EF;
const char32_t VAI_SYLLABLE_GBEN = 0xA5F0;
const char32_t VAI_SYLLABLE_FE = 0xA5F1;
const char32_t VAI_SYLLABLE_VE = 0xA5F2;
const char32_t VAI_SYLLABLE_TE = 0xA5F3;
const char32_t VAI_SYLLABLE_THE = 0xA5F4;
const char32_t VAI_SYLLABLE_DHE = 0xA5F5;
const char32_t VAI_SYLLABLE_DHHE = 0xA5F6;
const char32_t VAI_SYLLABLE_LE = 0xA5F7;
const char32_t VAI_SYLLABLE_RE = 0xA5F8;
const char32_t VAI_SYLLABLE_DE = 0xA5F9;
const char32_t VAI_SYLLABLE_NDE = 0xA5FA;
const char32_t VAI_SYLLABLE_SE = 0xA5FB;
const char32_t VAI_SYLLABLE_SHE = 0xA5FC;
const char32_t VAI_SYLLABLE_ZE = 0xA5FD;
const char32_t VAI_SYLLABLE_ZHE = 0xA5FE;
const char32_t VAI_SYLLABLE_CE = 0xA5FF;
const char32_t VAI_SYLLABLE_JE = 0xA600;
const char32_t VAI_SYLLABLE_NJE = 0xA601;
const char32_t VAI_SYLLABLE_YE = 0xA602;
const char32_t VAI_SYLLABLE_KE = 0xA603;
const char32_t VAI_SYLLABLE_NGGE = 0xA604;
const char32_t VAI_SYLLABLE_NGGEN = 0xA605;
const char32_t VAI_SYLLABLE_GE = 0xA606;
const char32_t VAI_SYLLABLE_GEN = 0xA607;
const char32_t VAI_SYLLABLE_ME = 0xA608;
const char32_t VAI_SYLLABLE_NE = 0xA609;
const char32_t VAI_SYLLABLE_NYE = 0xA60A;
const char32_t VAI_SYLLABLE_NG = 0xA60B;
const char32_t VAI_SYLLABLE_LENGTHENER = 0xA60C;
const char32_t VAI_COMMA = 0xA60D;
const char32_t VAI_FULL_STOP = 0xA60E;
const char32_t VAI_QUESTION_MARK = 0xA60F;
const char32_t VAI_SYLLABLE_NDOLE_FA = 0xA610;
const char32_t VAI_SYLLABLE_NDOLE_KA = 0xA611;
const char32_t VAI_SYLLABLE_NDOLE_SOO = 0xA612;
const char32_t VAI_SYMBOL_FEENG = 0xA613;
const char32_t VAI_SYMBOL_KEENG = 0xA614;
const char32_t VAI_SYMBOL_TING = 0xA615;
const char32_t VAI_SYMBOL_NII = 0xA616;
const char32_t VAI_SYMBOL_BANG = 0xA617;
const char32_t VAI_SYMBOL_FAA = 0xA618;
const char32_t VAI_SYMBOL_TAA = 0xA619;
const char32_t VAI_SYMBOL_DANG = 0xA61A;
const char32_t VAI_SYMBOL_DOONG = 0xA61B;
const char32_t VAI_SYMBOL_KUNG = 0xA61C;
const char32_t VAI_SYMBOL_TONG = 0xA61D;
const char32_t VAI_SYMBOL_DO_O = 0xA61E;
const char32_t VAI_SYMBOL_JONG = 0xA61F;
const char32_t VAI_DIGIT_ZERO = 0xA620;
const char32_t VAI_DIGIT_ONE = 0xA621;
const char32_t VAI_DIGIT_TWO = 0xA622;
const char32_t VAI_DIGIT_THREE = 0xA623;
const char32_t VAI_DIGIT_FOUR = 0xA624;
const char32_t VAI_DIGIT_FIVE = 0xA625;
const char32_t VAI_DIGIT_SIX = 0xA626;
const char32_t VAI_DIGIT_SEVEN = 0xA627;
const char32_t VAI_DIGIT_EIGHT = 0xA628;
const char32_t VAI_DIGIT_NINE = 0xA629;
const char32_t VAI_SYLLABLE_NDOLE_MA = 0xA62A;
const char32_t VAI_SYLLABLE_NDOLE_DO = 0xA62B;
const char32_t CYRILLIC_CAPITAL_LETTER_ZEMLYA = 0xA640;
const char32_t CYRILLIC_SMALL_LETTER_ZEMLYA = 0xA641;
const char32_t CYRILLIC_CAPITAL_LETTER_DZELO = 0xA642;
const char32_t CYRILLIC_SMALL_LETTER_DZELO = 0xA643;
const char32_t CYRILLIC_CAPITAL_LETTER_REVERSED_DZE = 0xA644;
const char32_t CYRILLIC_SMALL_LETTER_REVERSED_DZE = 0xA645;
const char32_t CYRILLIC_CAPITAL_LETTER_IOTA = 0xA646;
const char32_t CYRILLIC_SMALL_LETTER_IOTA = 0xA647;
const char32_t CYRILLIC_CAPITAL_LETTER_DJERV = 0xA648;
const char32_t CYRILLIC_SMALL_LETTER_DJERV = 0xA649;
const char32_t CYRILLIC_CAPITAL_LETTER_MONOGRAPH_UK = 0xA64A;
const char32_t CYRILLIC_SMALL_LETTER_MONOGRAPH_UK = 0xA64B;
const char32_t CYRILLIC_CAPITAL_LETTER_BROAD_OMEGA = 0xA64C;
const char32_t CYRILLIC_SMALL_LETTER_BROAD_OMEGA = 0xA64D;
const char32_t CYRILLIC_CAPITAL_LETTER_NEUTRAL_YER = 0xA64E;
const char32_t CYRILLIC_SMALL_LETTER_NEUTRAL_YER = 0xA64F;
const char32_t CYRILLIC_CAPITAL_LETTER_YERU_WITH_BACK_YER = 0xA650;
const char32_t CYRILLIC_SMALL_LETTER_YERU_WITH_BACK_YER = 0xA651;
const char32_t CYRILLIC_CAPITAL_LETTER_IOTIFIED_YAT = 0xA652;
const char32_t CYRILLIC_SMALL_LETTER_IOTIFIED_YAT = 0xA653;
const char32_t CYRILLIC_CAPITAL_LETTER_REVERSED_YU = 0xA654;
const char32_t CYRILLIC_SMALL_LETTER_REVERSED_YU = 0xA655;
const char32_t CYRILLIC_CAPITAL_LETTER_IOTIFIED_A = 0xA656;
const char32_t CYRILLIC_SMALL_LETTER_IOTIFIED_A = 0xA657;
const char32_t CYRILLIC_CAPITAL_LETTER_CLOSED_LITTLE_YUS = 0xA658;
const char32_t CYRILLIC_SMALL_LETTER_CLOSED_LITTLE_YUS = 0xA659;
const char32_t CYRILLIC_CAPITAL_LETTER_BLENDED_YUS = 0xA65A;
const char32_t CYRILLIC_SMALL_LETTER_BLENDED_YUS = 0xA65B;
const char32_t CYRILLIC_CAPITAL_LETTER_IOTIFIED_CLOSED_LITTLE_YUS = 0xA65C;
const char32_t CYRILLIC_SMALL_LETTER_IOTIFIED_CLOSED_LITTLE_YUS = 0xA65D;
const char32_t CYRILLIC_CAPITAL_LETTER_YN = 0xA65E;
const char32_t CYRILLIC_SMALL_LETTER_YN = 0xA65F;
const char32_t CYRILLIC_CAPITAL_LETTER_REVERSED_TSE = 0xA660;
const char32_t CYRILLIC_SMALL_LETTER_REVERSED_TSE = 0xA661;
const char32_t CYRILLIC_CAPITAL_LETTER_SOFT_DE = 0xA662;
const char32_t CYRILLIC_SMALL_LETTER_SOFT_DE = 0xA663;
const char32_t CYRILLIC_CAPITAL_LETTER_SOFT_EL = 0xA664;
const char32_t CYRILLIC_SMALL_LETTER_SOFT_EL = 0xA665;
const char32_t CYRILLIC_CAPITAL_LETTER_SOFT_EM = 0xA666;
const char32_t CYRILLIC_SMALL_LETTER_SOFT_EM = 0xA667;
const char32_t CYRILLIC_CAPITAL_LETTER_MONOCULAR_O = 0xA668;
const char32_t CYRILLIC_SMALL_LETTER_MONOCULAR_O = 0xA669;
const char32_t CYRILLIC_CAPITAL_LETTER_BINOCULAR_O = 0xA66A;
const char32_t CYRILLIC_SMALL_LETTER_BINOCULAR_O = 0xA66B;
const char32_t CYRILLIC_CAPITAL_LETTER_DOUBLE_MONOCULAR_O = 0xA66C;
const char32_t CYRILLIC_SMALL_LETTER_DOUBLE_MONOCULAR_O = 0xA66D;
const char32_t CYRILLIC_LETTER_MULTIOCULAR_O = 0xA66E;
const char32_t COMBINING_CYRILLIC_VZMET = 0xA66F;
const char32_t COMBINING_CYRILLIC_TEN_MILLIONS_SIGN = 0xA670;
const char32_t COMBINING_CYRILLIC_HUNDRED_MILLIONS_SIGN = 0xA671;
const char32_t COMBINING_CYRILLIC_THOUSAND_MILLIONS_SIGN = 0xA672;
const char32_t SLAVONIC_ASTERISK = 0xA673;
const char32_t COMBINING_CYRILLIC_LETTER_UKRAINIAN_IE = 0xA674;
const char32_t COMBINING_CYRILLIC_LETTER_I = 0xA675;
const char32_t COMBINING_CYRILLIC_LETTER_YI = 0xA676;
const char32_t COMBINING_CYRILLIC_LETTER_U = 0xA677;
const char32_t COMBINING_CYRILLIC_LETTER_HARD_SIGN = 0xA678;
const char32_t COMBINING_CYRILLIC_LETTER_YERU = 0xA679;
const char32_t COMBINING_CYRILLIC_LETTER_SOFT_SIGN = 0xA67A;
const char32_t COMBINING_CYRILLIC_LETTER_OMEGA = 0xA67B;
const char32_t COMBINING_CYRILLIC_KAVYKA = 0xA67C;
const char32_t COMBINING_CYRILLIC_PAYEROK = 0xA67D;
const char32_t CYRILLIC_KAVYKA = 0xA67E;
const char32_t CYRILLIC_PAYEROK = 0xA67F;
const char32_t CYRILLIC_CAPITAL_LETTER_DWE = 0xA680;
const char32_t CYRILLIC_SMALL_LETTER_DWE = 0xA681;
const char32_t CYRILLIC_CAPITAL_LETTER_DZWE = 0xA682;
const char32_t CYRILLIC_SMALL_LETTER_DZWE = 0xA683;
const char32_t CYRILLIC_CAPITAL_LETTER_ZHWE = 0xA684;
const char32_t CYRILLIC_SMALL_LETTER_ZHWE = 0xA685;
const char32_t CYRILLIC_CAPITAL_LETTER_CCHE = 0xA686;
const char32_t CYRILLIC_SMALL_LETTER_CCHE = 0xA687;
const char32_t CYRILLIC_CAPITAL_LETTER_DZZE = 0xA688;
const char32_t CYRILLIC_SMALL_LETTER_DZZE = 0xA689;
const char32_t CYRILLIC_CAPITAL_LETTER_TE_WITH_MIDDLE_HOOK = 0xA68A;
const char32_t CYRILLIC_SMALL_LETTER_TE_WITH_MIDDLE_HOOK = 0xA68B;
const char32_t CYRILLIC_CAPITAL_LETTER_TWE = 0xA68C;
const char32_t CYRILLIC_SMALL_LETTER_TWE = 0xA68D;
const char32_t CYRILLIC_CAPITAL_LETTER_TSWE = 0xA68E;
const char32_t CYRILLIC_SMALL_LETTER_TSWE = 0xA68F;
const char32_t CYRILLIC_CAPITAL_LETTER_TSSE = 0xA690;
const char32_t CYRILLIC_SMALL_LETTER_TSSE = 0xA691;
const char32_t CYRILLIC_CAPITAL_LETTER_TCHE = 0xA692;
const char32_t CYRILLIC_SMALL_LETTER_TCHE = 0xA693;
const char32_t CYRILLIC_CAPITAL_LETTER_HWE = 0xA694;
const char32_t CYRILLIC_SMALL_LETTER_HWE = 0xA695;
const char32_t CYRILLIC_CAPITAL_LETTER_SHWE = 0xA696;
const char32_t CYRILLIC_SMALL_LETTER_SHWE = 0xA697;
const char32_t CYRILLIC_CAPITAL_LETTER_DOUBLE_O = 0xA698;
const char32_t CYRILLIC_SMALL_LETTER_DOUBLE_O = 0xA699;
const char32_t CYRILLIC_CAPITAL_LETTER_CROSSED_O = 0xA69A;
const char32_t CYRILLIC_SMALL_LETTER_CROSSED_O = 0xA69B;
const char32_t MODIFIER_LETTER_CYRILLIC_HARD_SIGN = 0xA69C;
const char32_t MODIFIER_LETTER_CYRILLIC_SOFT_SIGN = 0xA69D;
const char32_t COMBINING_CYRILLIC_LETTER_EF = 0xA69E;
const char32_t COMBINING_CYRILLIC_LETTER_IOTIFIED_E = 0xA69F;
const char32_t BAMUM_LETTER_A = 0xA6A0;
const char32_t BAMUM_LETTER_KA = 0xA6A1;
const char32_t BAMUM_LETTER_U = 0xA6A2;
const char32_t BAMUM_LETTER_KU = 0xA6A3;
const char32_t BAMUM_LETTER_EE = 0xA6A4;
const char32_t BAMUM_LETTER_REE = 0xA6A5;
const char32_t BAMUM_LETTER_TAE = 0xA6A6;
const char32_t BAMUM_LETTER_O = 0xA6A7;
const char32_t BAMUM_LETTER_NYI = 0xA6A8;
const char32_t BAMUM_LETTER_I = 0xA6A9;
const char32_t BAMUM_LETTER_LA = 0xA6AA;
const char32_t BAMUM_LETTER_PA = 0xA6AB;
const char32_t BAMUM_LETTER_RII = 0xA6AC;
const char32_t BAMUM_LETTER_RIEE = 0xA6AD;
const char32_t BAMUM_LETTER_LEEEE = 0xA6AE;
const char32_t BAMUM_LETTER_MEEEE = 0xA6AF;
const char32_t BAMUM_LETTER_TAA = 0xA6B0;
const char32_t BAMUM_LETTER_NDAA = 0xA6B1;
const char32_t BAMUM_LETTER_NJAEM = 0xA6B2;
const char32_t BAMUM_LETTER_M = 0xA6B3;
const char32_t BAMUM_LETTER_SUU = 0xA6B4;
const char32_t BAMUM_LETTER_MU = 0xA6B5;
const char32_t BAMUM_LETTER_SHII = 0xA6B6;
const char32_t BAMUM_LETTER_SI = 0xA6B7;
const char32_t BAMUM_LETTER_SHEUX = 0xA6B8;
const char32_t BAMUM_LETTER_SEUX = 0xA6B9;
const char32_t BAMUM_LETTER_KYEE = 0xA6BA;
const char32_t BAMUM_LETTER_KET = 0xA6BB;
const char32_t BAMUM_LETTER_NUAE = 0xA6BC;
const char32_t BAMUM_LETTER_NU = 0xA6BD;
const char32_t BAMUM_LETTER_NJUAE = 0xA6BE;
const char32_t BAMUM_LETTER_YOQ = 0xA6BF;
const char32_t BAMUM_LETTER_SHU = 0xA6C0;
const char32_t BAMUM_LETTER_YUQ = 0xA6C1;
const char32_t BAMUM_LETTER_YA = 0xA6C2;
const char32_t BAMUM_LETTER_NSHA = 0xA6C3;
const char32_t BAMUM_LETTER_KEUX = 0xA6C4;
const char32_t BAMUM_LETTER_PEUX = 0xA6C5;
const char32_t BAMUM_LETTER_NJEE = 0xA6C6;
const char32_t BAMUM_LETTER_NTEE = 0xA6C7;
const char32_t BAMUM_LETTER_PUE = 0xA6C8;
const char32_t BAMUM_LETTER_WUE = 0xA6C9;
const char32_t BAMUM_LETTER_PEE = 0xA6CA;
const char32_t BAMUM_LETTER_FEE = 0xA6CB;
const char32_t BAMUM_LETTER_RU = 0xA6CC;
const char32_t BAMUM_LETTER_LU = 0xA6CD;
const char32_t BAMUM_LETTER_MI = 0xA6CE;
const char32_t BAMUM_LETTER_NI = 0xA6CF;
const char32_t BAMUM_LETTER_REUX = 0xA6D0;
const char32_t BAMUM_LETTER_RAE = 0xA6D1;
const char32_t BAMUM_LETTER_KEN = 0xA6D2;
const char32_t BAMUM_LETTER_NGKWAEN = 0xA6D3;
const char32_t BAMUM_LETTER_NGGA = 0xA6D4;
const char32_t BAMUM_LETTER_NGA = 0xA6D5;
const char32_t BAMUM_LETTER_SHO = 0xA6D6;
const char32_t BAMUM_LETTER_PUAE = 0xA6D7;
const char32_t BAMUM_LETTER_FU = 0xA6D8;
const char32_t BAMUM_LETTER_FOM = 0xA6D9;
const char32_t BAMUM_LETTER_WA = 0xA6DA;
const char32_t BAMUM_LETTER_NA = 0xA6DB;
const char32_t BAMUM_LETTER_LI = 0xA6DC;
const char32_t BAMUM_LETTER_PI = 0xA6DD;
const char32_t BAMUM_LETTER_LOQ = 0xA6DE;
const char32_t BAMUM_LETTER_KO = 0xA6DF;
const char32_t BAMUM_LETTER_MBEN = 0xA6E0;
const char32_t BAMUM_LETTER_REN = 0xA6E1;
const char32_t BAMUM_LETTER_MEN = 0xA6E2;
const char32_t BAMUM_LETTER_MA = 0xA6E3;
const char32_t BAMUM_LETTER_TI = 0xA6E4;
const char32_t BAMUM_LETTER_KI = 0xA6E5;
const char32_t BAMUM_LETTER_MO = 0xA6E6;
const char32_t BAMUM_LETTER_MBAA = 0xA6E7;
const char32_t BAMUM_LETTER_TET = 0xA6E8;
const char32_t BAMUM_LETTER_KPA = 0xA6E9;
const char32_t BAMUM_LETTER_TEN = 0xA6EA;
const char32_t BAMUM_LETTER_NTUU = 0xA6EB;
const char32_t BAMUM_LETTER_SAMBA = 0xA6EC;
const char32_t BAMUM_LETTER_FAAMAE = 0xA6ED;
const char32_t BAMUM_LETTER_KOVUU = 0xA6EE;
const char32_t BAMUM_LETTER_KOGHOM = 0xA6EF;
const char32_t BAMUM_COMBINING_MARK_KOQNDON = 0xA6F0;
const char32_t BAMUM_COMBINING_MARK_TUKWENTIS = 0xA6F1;
const char32_t BAMUM_NJAEMLI = 0xA6F2;
const char32_t BAMUM_FULL_STOP = 0xA6F3;
const char32_t BAMUM_COLON = 0xA6F4;
const char32_t BAMUM_COMMA = 0xA6F5;
const char32_t BAMUM_SEMICOLON = 0xA6F6;
const char32_t BAMUM_QUESTION_MARK = 0xA6F7;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YIN_PING = 0xA700;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YANG_PING = 0xA701;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YIN_SHANG = 0xA702;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YANG_SHANG = 0xA703;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YIN_QU = 0xA704;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YANG_QU = 0xA705;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YIN_RU = 0xA706;
const char32_t MODIFIER_LETTER_CHINESE_TONE_YANG_RU = 0xA707;
const char32_t MODIFIER_LETTER_EXTRA_HIGH_DOTTED_TONE_BAR = 0xA708;
const char32_t MODIFIER_LETTER_HIGH_DOTTED_TONE_BAR = 0xA709;
const char32_t MODIFIER_LETTER_MID_DOTTED_TONE_BAR = 0xA70A;
const char32_t MODIFIER_LETTER_LOW_DOTTED_TONE_BAR = 0xA70B;
const char32_t MODIFIER_LETTER_EXTRA_LOW_DOTTED_TONE_BAR = 0xA70C;
const char32_t MODIFIER_LETTER_EXTRA_HIGH_DOTTED_LEFT_STEM_TONE_BAR = 0xA70D;
const char32_t MODIFIER_LETTER_HIGH_DOTTED_LEFT_STEM_TONE_BAR = 0xA70E;
const char32_t MODIFIER_LETTER_MID_DOTTED_LEFT_STEM_TONE_BAR = 0xA70F;
const char32_t MODIFIER_LETTER_LOW_DOTTED_LEFT_STEM_TONE_BAR = 0xA710;
const char32_t MODIFIER_LETTER_EXTRA_LOW_DOTTED_LEFT_STEM_TONE_BAR = 0xA711;
const char32_t MODIFIER_LETTER_EXTRA_HIGH_LEFT_STEM_TONE_BAR = 0xA712;
const char32_t MODIFIER_LETTER_HIGH_LEFT_STEM_TONE_BAR = 0xA713;
const char32_t MODIFIER_LETTER_MID_LEFT_STEM_TONE_BAR = 0xA714;
const char32_t MODIFIER_LETTER_LOW_LEFT_STEM_TONE_BAR = 0xA715;
const char32_t MODIFIER_LETTER_EXTRA_LOW_LEFT_STEM_TONE_BAR = 0xA716;
const char32_t MODIFIER_LETTER_DOT_VERTICAL_BAR = 0xA717;
const char32_t MODIFIER_LETTER_DOT_SLASH = 0xA718;
const char32_t MODIFIER_LETTER_DOT_HORIZONTAL_BAR = 0xA719;
const char32_t MODIFIER_LETTER_LOWER_RIGHT_CORNER_ANGLE = 0xA71A;
const char32_t MODIFIER_LETTER_RAISED_UP_ARROW = 0xA71B;
const char32_t MODIFIER_LETTER_RAISED_DOWN_ARROW = 0xA71C;
const char32_t MODIFIER_LETTER_RAISED_EXCLAMATION_MARK = 0xA71D;
const char32_t MODIFIER_LETTER_RAISED_INVERTED_EXCLAMATION_MARK = 0xA71E;
const char32_t MODIFIER_LETTER_LOW_INVERTED_EXCLAMATION_MARK = 0xA71F;
const char32_t MODIFIER_LETTER_STRESS_AND_HIGH_TONE = 0xA720;
const char32_t MODIFIER_LETTER_STRESS_AND_LOW_TONE = 0xA721;
const char32_t LATIN_CAPITAL_LETTER_EGYPTOLOGICAL_ALEF = 0xA722;
const char32_t LATIN_SMALL_LETTER_EGYPTOLOGICAL_ALEF = 0xA723;
const char32_t LATIN_CAPITAL_LETTER_EGYPTOLOGICAL_AIN = 0xA724;
const char32_t LATIN_SMALL_LETTER_EGYPTOLOGICAL_AIN = 0xA725;
const char32_t LATIN_CAPITAL_LETTER_HENG = 0xA726;
const char32_t LATIN_SMALL_LETTER_HENG = 0xA727;
const char32_t LATIN_CAPITAL_LETTER_TZ = 0xA728;
const char32_t LATIN_SMALL_LETTER_TZ = 0xA729;
const char32_t LATIN_CAPITAL_LETTER_TRESILLO = 0xA72A;
const char32_t LATIN_SMALL_LETTER_TRESILLO = 0xA72B;
const char32_t LATIN_CAPITAL_LETTER_CUATRILLO = 0xA72C;
const char32_t LATIN_SMALL_LETTER_CUATRILLO = 0xA72D;
const char32_t LATIN_CAPITAL_LETTER_CUATRILLO_WITH_COMMA = 0xA72E;
const char32_t LATIN_SMALL_LETTER_CUATRILLO_WITH_COMMA = 0xA72F;
const char32_t LATIN_LETTER_SMALL_CAPITAL_F = 0xA730;
const char32_t LATIN_LETTER_SMALL_CAPITAL_S = 0xA731;
const char32_t LATIN_CAPITAL_LETTER_AA = 0xA732;
const char32_t LATIN_SMALL_LETTER_AA = 0xA733;
const char32_t LATIN_CAPITAL_LETTER_AO = 0xA734;
const char32_t LATIN_SMALL_LETTER_AO = 0xA735;
const char32_t LATIN_CAPITAL_LETTER_AU = 0xA736;
const char32_t LATIN_SMALL_LETTER_AU = 0xA737;
const char32_t LATIN_CAPITAL_LETTER_AV = 0xA738;
const char32_t LATIN_SMALL_LETTER_AV = 0xA739;
const char32_t LATIN_CAPITAL_LETTER_AV_WITH_HORIZONTAL_BAR = 0xA73A;
const char32_t LATIN_SMALL_LETTER_AV_WITH_HORIZONTAL_BAR = 0xA73B;
const char32_t LATIN_CAPITAL_LETTER_AY = 0xA73C;
const char32_t LATIN_SMALL_LETTER_AY = 0xA73D;
const char32_t LATIN_CAPITAL_LETTER_REVERSED_C_WITH_DOT = 0xA73E;
const char32_t LATIN_SMALL_LETTER_REVERSED_C_WITH_DOT = 0xA73F;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_STROKE = 0xA740;
const char32_t LATIN_SMALL_LETTER_K_WITH_STROKE = 0xA741;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_DIAGONAL_STROKE = 0xA742;
const char32_t LATIN_SMALL_LETTER_K_WITH_DIAGONAL_STROKE = 0xA743;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_STROKE_AND_DIAGONAL_STROKE = 0xA744;
const char32_t LATIN_SMALL_LETTER_K_WITH_STROKE_AND_DIAGONAL_STROKE = 0xA745;
const char32_t LATIN_CAPITAL_LETTER_BROKEN_L = 0xA746;
const char32_t LATIN_SMALL_LETTER_BROKEN_L = 0xA747;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_HIGH_STROKE = 0xA748;
const char32_t LATIN_SMALL_LETTER_L_WITH_HIGH_STROKE = 0xA749;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_LONG_STROKE_OVERLAY = 0xA74A;
const char32_t LATIN_SMALL_LETTER_O_WITH_LONG_STROKE_OVERLAY = 0xA74B;
const char32_t LATIN_CAPITAL_LETTER_O_WITH_LOOP = 0xA74C;
const char32_t LATIN_SMALL_LETTER_O_WITH_LOOP = 0xA74D;
const char32_t LATIN_CAPITAL_LETTER_OO = 0xA74E;
const char32_t LATIN_SMALL_LETTER_OO = 0xA74F;
const char32_t LATIN_CAPITAL_LETTER_P_WITH_STROKE_THROUGH_DESCENDER = 0xA750;
const char32_t LATIN_SMALL_LETTER_P_WITH_STROKE_THROUGH_DESCENDER = 0xA751;
const char32_t LATIN_CAPITAL_LETTER_P_WITH_FLOURISH = 0xA752;
const char32_t LATIN_SMALL_LETTER_P_WITH_FLOURISH = 0xA753;
const char32_t LATIN_CAPITAL_LETTER_P_WITH_SQUIRREL_TAIL = 0xA754;
const char32_t LATIN_SMALL_LETTER_P_WITH_SQUIRREL_TAIL = 0xA755;
const char32_t LATIN_CAPITAL_LETTER_Q_WITH_STROKE_THROUGH_DESCENDER = 0xA756;
const char32_t LATIN_SMALL_LETTER_Q_WITH_STROKE_THROUGH_DESCENDER = 0xA757;
const char32_t LATIN_CAPITAL_LETTER_Q_WITH_DIAGONAL_STROKE = 0xA758;
const char32_t LATIN_SMALL_LETTER_Q_WITH_DIAGONAL_STROKE = 0xA759;
const char32_t LATIN_CAPITAL_LETTER_R_ROTUNDA = 0xA75A;
const char32_t LATIN_SMALL_LETTER_R_ROTUNDA = 0xA75B;
const char32_t LATIN_CAPITAL_LETTER_RUM_ROTUNDA = 0xA75C;
const char32_t LATIN_SMALL_LETTER_RUM_ROTUNDA = 0xA75D;
const char32_t LATIN_CAPITAL_LETTER_V_WITH_DIAGONAL_STROKE = 0xA75E;
const char32_t LATIN_SMALL_LETTER_V_WITH_DIAGONAL_STROKE = 0xA75F;
const char32_t LATIN_CAPITAL_LETTER_VY = 0xA760;
const char32_t LATIN_SMALL_LETTER_VY = 0xA761;
const char32_t LATIN_CAPITAL_LETTER_VISIGOTHIC_Z = 0xA762;
const char32_t LATIN_SMALL_LETTER_VISIGOTHIC_Z = 0xA763;
const char32_t LATIN_CAPITAL_LETTER_THORN_WITH_STROKE = 0xA764;
const char32_t LATIN_SMALL_LETTER_THORN_WITH_STROKE = 0xA765;
const char32_t LATIN_CAPITAL_LETTER_THORN_WITH_STROKE_THROUGH_DESCENDER = 0xA766;
const char32_t LATIN_SMALL_LETTER_THORN_WITH_STROKE_THROUGH_DESCENDER = 0xA767;
const char32_t LATIN_CAPITAL_LETTER_VEND = 0xA768;
const char32_t LATIN_SMALL_LETTER_VEND = 0xA769;
const char32_t LATIN_CAPITAL_LETTER_ET = 0xA76A;
const char32_t LATIN_SMALL_LETTER_ET = 0xA76B;
const char32_t LATIN_CAPITAL_LETTER_IS = 0xA76C;
const char32_t LATIN_SMALL_LETTER_IS = 0xA76D;
const char32_t LATIN_CAPITAL_LETTER_CON = 0xA76E;
const char32_t LATIN_SMALL_LETTER_CON = 0xA76F;
const char32_t MODIFIER_LETTER_US = 0xA770;
const char32_t LATIN_SMALL_LETTER_DUM = 0xA771;
const char32_t LATIN_SMALL_LETTER_LUM = 0xA772;
const char32_t LATIN_SMALL_LETTER_MUM = 0xA773;
const char32_t LATIN_SMALL_LETTER_NUM = 0xA774;
const char32_t LATIN_SMALL_LETTER_RUM = 0xA775;
const char32_t LATIN_LETTER_SMALL_CAPITAL_RUM = 0xA776;
const char32_t LATIN_SMALL_LETTER_TUM = 0xA777;
const char32_t LATIN_SMALL_LETTER_UM = 0xA778;
const char32_t LATIN_CAPITAL_LETTER_INSULAR_D = 0xA779;
const char32_t LATIN_SMALL_LETTER_INSULAR_D = 0xA77A;
const char32_t LATIN_CAPITAL_LETTER_INSULAR_F = 0xA77B;
const char32_t LATIN_SMALL_LETTER_INSULAR_F = 0xA77C;
const char32_t LATIN_CAPITAL_LETTER_INSULAR_G = 0xA77D;
const char32_t LATIN_CAPITAL_LETTER_TURNED_INSULAR_G = 0xA77E;
const char32_t LATIN_SMALL_LETTER_TURNED_INSULAR_G = 0xA77F;
const char32_t LATIN_CAPITAL_LETTER_TURNED_L = 0xA780;
const char32_t LATIN_SMALL_LETTER_TURNED_L = 0xA781;
const char32_t LATIN_CAPITAL_LETTER_INSULAR_R = 0xA782;
const char32_t LATIN_SMALL_LETTER_INSULAR_R = 0xA783;
const char32_t LATIN_CAPITAL_LETTER_INSULAR_S = 0xA784;
const char32_t LATIN_SMALL_LETTER_INSULAR_S = 0xA785;
const char32_t LATIN_CAPITAL_LETTER_INSULAR_T = 0xA786;
const char32_t LATIN_SMALL_LETTER_INSULAR_T = 0xA787;
const char32_t MODIFIER_LETTER_LOW_CIRCUMFLEX_ACCENT = 0xA788;
const char32_t MODIFIER_LETTER_COLON = 0xA789;
const char32_t MODIFIER_LETTER_SHORT_EQUALS_SIGN = 0xA78A;
const char32_t LATIN_CAPITAL_LETTER_SALTILLO = 0xA78B;
const char32_t LATIN_SMALL_LETTER_SALTILLO = 0xA78C;
const char32_t LATIN_CAPITAL_LETTER_TURNED_H = 0xA78D;
const char32_t LATIN_SMALL_LETTER_L_WITH_RETROFLEX_HOOK_AND_BELT = 0xA78E;
const char32_t LATIN_LETTER_SINOLOGICAL_DOT = 0xA78F;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_DESCENDER = 0xA790;
const char32_t LATIN_SMALL_LETTER_N_WITH_DESCENDER = 0xA791;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_BAR = 0xA792;
const char32_t LATIN_SMALL_LETTER_C_WITH_BAR = 0xA793;
const char32_t LATIN_SMALL_LETTER_C_WITH_PALATAL_HOOK = 0xA794;
const char32_t LATIN_SMALL_LETTER_H_WITH_PALATAL_HOOK = 0xA795;
const char32_t LATIN_CAPITAL_LETTER_B_WITH_FLOURISH = 0xA796;
const char32_t LATIN_SMALL_LETTER_B_WITH_FLOURISH = 0xA797;
const char32_t LATIN_CAPITAL_LETTER_F_WITH_STROKE = 0xA798;
const char32_t LATIN_SMALL_LETTER_F_WITH_STROKE = 0xA799;
const char32_t LATIN_CAPITAL_LETTER_VOLAPUK_AE = 0xA79A;
const char32_t LATIN_SMALL_LETTER_VOLAPUK_AE = 0xA79B;
const char32_t LATIN_CAPITAL_LETTER_VOLAPUK_OE = 0xA79C;
const char32_t LATIN_SMALL_LETTER_VOLAPUK_OE = 0xA79D;
const char32_t LATIN_CAPITAL_LETTER_VOLAPUK_UE = 0xA79E;
const char32_t LATIN_SMALL_LETTER_VOLAPUK_UE = 0xA79F;
const char32_t LATIN_CAPITAL_LETTER_G_WITH_OBLIQUE_STROKE = 0xA7A0;
const char32_t LATIN_SMALL_LETTER_G_WITH_OBLIQUE_STROKE = 0xA7A1;
const char32_t LATIN_CAPITAL_LETTER_K_WITH_OBLIQUE_STROKE = 0xA7A2;
const char32_t LATIN_SMALL_LETTER_K_WITH_OBLIQUE_STROKE = 0xA7A3;
const char32_t LATIN_CAPITAL_LETTER_N_WITH_OBLIQUE_STROKE = 0xA7A4;
const char32_t LATIN_SMALL_LETTER_N_WITH_OBLIQUE_STROKE = 0xA7A5;
const char32_t LATIN_CAPITAL_LETTER_R_WITH_OBLIQUE_STROKE = 0xA7A6;
const char32_t LATIN_SMALL_LETTER_R_WITH_OBLIQUE_STROKE = 0xA7A7;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_OBLIQUE_STROKE = 0xA7A8;
const char32_t LATIN_SMALL_LETTER_S_WITH_OBLIQUE_STROKE = 0xA7A9;
const char32_t LATIN_CAPITAL_LETTER_H_WITH_HOOK = 0xA7AA;
const char32_t LATIN_CAPITAL_LETTER_REVERSED_OPEN_E = 0xA7AB;
const char32_t LATIN_CAPITAL_LETTER_SCRIPT_G = 0xA7AC;
const char32_t LATIN_CAPITAL_LETTER_L_WITH_BELT = 0xA7AD;
const char32_t LATIN_CAPITAL_LETTER_SMALL_CAPITAL_I = 0xA7AE;
const char32_t LATIN_LETTER_SMALL_CAPITAL_Q = 0xA7AF;
const char32_t LATIN_CAPITAL_LETTER_TURNED_K = 0xA7B0;
const char32_t LATIN_CAPITAL_LETTER_TURNED_T = 0xA7B1;
const char32_t LATIN_CAPITAL_LETTER_J_WITH_CROSSED_TAIL = 0xA7B2;
const char32_t LATIN_CAPITAL_LETTER_CHI = 0xA7B3;
const char32_t LATIN_CAPITAL_LETTER_BETA = 0xA7B4;
const char32_t LATIN_SMALL_LETTER_BETA = 0xA7B5;
const char32_t LATIN_CAPITAL_LETTER_OMEGA = 0xA7B6;
const char32_t LATIN_SMALL_LETTER_OMEGA = 0xA7B7;
const char32_t LATIN_CAPITAL_LETTER_U_WITH_STROKE = 0xA7B8;
const char32_t LATIN_SMALL_LETTER_U_WITH_STROKE = 0xA7B9;
const char32_t LATIN_CAPITAL_LETTER_GLOTTAL_A = 0xA7BA;
const char32_t LATIN_SMALL_LETTER_GLOTTAL_A = 0xA7BB;
const char32_t LATIN_CAPITAL_LETTER_GLOTTAL_I = 0xA7BC;
const char32_t LATIN_SMALL_LETTER_GLOTTAL_I = 0xA7BD;
const char32_t LATIN_CAPITAL_LETTER_GLOTTAL_U = 0xA7BE;
const char32_t LATIN_SMALL_LETTER_GLOTTAL_U = 0xA7BF;
const char32_t LATIN_CAPITAL_LETTER_OLD_POLISH_O = 0xA7C0;
const char32_t LATIN_SMALL_LETTER_OLD_POLISH_O = 0xA7C1;
const char32_t LATIN_CAPITAL_LETTER_ANGLICANA_W = 0xA7C2;
const char32_t LATIN_SMALL_LETTER_ANGLICANA_W = 0xA7C3;
const char32_t LATIN_CAPITAL_LETTER_C_WITH_PALATAL_HOOK = 0xA7C4;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_HOOK = 0xA7C5;
const char32_t LATIN_CAPITAL_LETTER_Z_WITH_PALATAL_HOOK = 0xA7C6;
const char32_t LATIN_CAPITAL_LETTER_D_WITH_SHORT_STROKE_OVERLAY = 0xA7C7;
const char32_t LATIN_SMALL_LETTER_D_WITH_SHORT_STROKE_OVERLAY = 0xA7C8;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_SHORT_STROKE_OVERLAY = 0xA7C9;
const char32_t LATIN_SMALL_LETTER_S_WITH_SHORT_STROKE_OVERLAY = 0xA7CA;
const char32_t LATIN_CAPITAL_LETTER_RAMS_HORN = 0xA7CB;
const char32_t LATIN_CAPITAL_LETTER_S_WITH_DIAGONAL_STROKE = 0xA7CC;
const char32_t LATIN_SMALL_LETTER_S_WITH_DIAGONAL_STROKE = 0xA7CD;
const char32_t LATIN_CAPITAL_LETTER_PHARYNGEAL_VOICED_FRICATIVE = 0xA7CE;
const char32_t LATIN_SMALL_LETTER_PHARYNGEAL_VOICED_FRICATIVE = 0xA7CF;
const char32_t LATIN_CAPITAL_LETTER_CLOSED_INSULAR_G = 0xA7D0;
const char32_t LATIN_SMALL_LETTER_CLOSED_INSULAR_G = 0xA7D1;
const char32_t LATIN_CAPITAL_LETTER_DOUBLE_THORN = 0xA7D2;
const char32_t LATIN_SMALL_LETTER_DOUBLE_THORN = 0xA7D3;
const char32_t LATIN_CAPITAL_LETTER_DOUBLE_WYNN = 0xA7D4;
const char32_t LATIN_SMALL_LETTER_DOUBLE_WYNN = 0xA7D5;
const char32_t LATIN_CAPITAL_LETTER_MIDDLE_SCOTS_S = 0xA7D6;
const char32_t LATIN_SMALL_LETTER_MIDDLE_SCOTS_S = 0xA7D7;
const char32_t LATIN_CAPITAL_LETTER_SIGMOID_S = 0xA7D8;
const char32_t LATIN_SMALL_LETTER_SIGMOID_S = 0xA7D9;
const char32_t LATIN_CAPITAL_LETTER_LAMBDA = 0xA7DA;
const char32_t LATIN_SMALL_LETTER_LAMBDA = 0xA7DB;
const char32_t LATIN_CAPITAL_LETTER_LAMBDA_WITH_STROKE = 0xA7DC;
const char32_t MODIFIER_LETTER_CAPITAL_S = 0xA7F1;
const char32_t MODIFIER_LETTER_CAPITAL_C = 0xA7F2;
const char32_t MODIFIER_LETTER_CAPITAL_F = 0xA7F3;
const char32_t MODIFIER_LETTER_CAPITAL_Q = 0xA7F4;
const char32_t LATIN_CAPITAL_LETTER_REVERSED_HALF_H = 0xA7F5;
const char32_t LATIN_SMALL_LETTER_REVERSED_HALF_H = 0xA7F6;
const char32_t LATIN_EPIGRAPHIC_LETTER_SIDEWAYS_I = 0xA7F7;
const char32_t MODIFIER_LETTER_CAPITAL_H_WITH_STROKE = 0xA7F8;
const char32_t MODIFIER_LETTER_SMALL_LIGATURE_OE = 0xA7F9;
const char32_t LATIN_LETTER_SMALL_CAPITAL_TURNED_M = 0xA7FA;
const char32_t LATIN_EPIGRAPHIC_LETTER_REVERSED_F = 0xA7FB;
const char32_t LATIN_EPIGRAPHIC_LETTER_REVERSED_P = 0xA7FC;
const char32_t LATIN_EPIGRAPHIC_LETTER_INVERTED_M = 0xA7FD;
const char32_t LATIN_EPIGRAPHIC_LETTER_I_LONGA = 0xA7FE;
const char32_t LATIN_EPIGRAPHIC_LETTER_ARCHAIC_M = 0xA7FF;
const char32_t SYLOTI_NAGRI_LETTER_A = 0xA800;
const char32_t SYLOTI_NAGRI_LETTER_I = 0xA801;
const char32_t SYLOTI_NAGRI_SIGN_DVISVARA = 0xA802;
const char32_t SYLOTI_NAGRI_LETTER_U = 0xA803;
const char32_t SYLOTI_NAGRI_LETTER_E = 0xA804;
const char32_t SYLOTI_NAGRI_LETTER_O = 0xA805;
const char32_t SYLOTI_NAGRI_SIGN_HASANTA = 0xA806;
const char32_t SYLOTI_NAGRI_LETTER_KO = 0xA807;
const char32_t SYLOTI_NAGRI_LETTER_KHO = 0xA808;
const char32_t SYLOTI_NAGRI_LETTER_GO = 0xA809;
const char32_t SYLOTI_NAGRI_LETTER_GHO = 0xA80A;
const char32_t SYLOTI_NAGRI_SIGN_ANUSVARA = 0xA80B;
const char32_t SYLOTI_NAGRI_LETTER_CO = 0xA80C;
const char32_t SYLOTI_NAGRI_LETTER_CHO = 0xA80D;
const char32_t SYLOTI_NAGRI_LETTER_JO = 0xA80E;
const char32_t SYLOTI_NAGRI_LETTER_JHO = 0xA80F;
const char32_t SYLOTI_NAGRI_LETTER_TTO = 0xA810;
const char32_t SYLOTI_NAGRI_LETTER_TTHO = 0xA811;
const char32_t SYLOTI_NAGRI_LETTER_DDO = 0xA812;
const char32_t SYLOTI_NAGRI_LETTER_DDHO = 0xA813;
const char32_t SYLOTI_NAGRI_LETTER_TO = 0xA814;
const char32_t SYLOTI_NAGRI_LETTER_THO = 0xA815;
const char32_t SYLOTI_NAGRI_LETTER_DO = 0xA816;
const char32_t SYLOTI_NAGRI_LETTER_DHO = 0xA817;
const char32_t SYLOTI_NAGRI_LETTER_NO = 0xA818;
const char32_t SYLOTI_NAGRI_LETTER_PO = 0xA819;
const char32_t SYLOTI_NAGRI_LETTER_PHO = 0xA81A;
const char32_t SYLOTI_NAGRI_LETTER_BO = 0xA81B;
const char32_t SYLOTI_NAGRI_LETTER_BHO = 0xA81C;
const char32_t SYLOTI_NAGRI_LETTER_MO = 0xA81D;
const char32_t SYLOTI_NAGRI_LETTER_RO = 0xA81E;
const char32_t SYLOTI_NAGRI_LETTER_LO = 0xA81F;
const char32_t SYLOTI_NAGRI_LETTER_RRO = 0xA820;
const char32_t SYLOTI_NAGRI_LETTER_SO = 0xA821;
const char32_t SYLOTI_NAGRI_LETTER_HO = 0xA822;
const char32_t SYLOTI_NAGRI_VOWEL_SIGN_A = 0xA823;
const char32_t SYLOTI_NAGRI_VOWEL_SIGN_I = 0xA824;
const char32_t SYLOTI_NAGRI_VOWEL_SIGN_U = 0xA825;
const char32_t SYLOTI_NAGRI_VOWEL_SIGN_E = 0xA826;
const char32_t SYLOTI_NAGRI_VOWEL_SIGN_OO = 0xA827;
const char32_t SYLOTI_NAGRI_POETRY_MARK_1 = 0xA828;
const char32_t SYLOTI_NAGRI_POETRY_MARK_2 = 0xA829;
const char32_t SYLOTI_NAGRI_POETRY_MARK_3 = 0xA82A;
const char32_t SYLOTI_NAGRI_POETRY_MARK_4 = 0xA82B;
const char32_t SYLOTI_NAGRI_SIGN_ALTERNATE_HASANTA = 0xA82C;
const char32_t NORTH_INDIC_FRACTION_ONE_QUARTER = 0xA830;
const char32_t NORTH_INDIC_FRACTION_ONE_HALF = 0xA831;
const char32_t NORTH_INDIC_FRACTION_THREE_QUARTERS = 0xA832;
const char32_t NORTH_INDIC_FRACTION_ONE_SIXTEENTH = 0xA833;
const char32_t NORTH_INDIC_FRACTION_ONE_EIGHTH = 0xA834;
const char32_t NORTH_INDIC_FRACTION_THREE_SIXTEENTHS = 0xA835;
const char32_t NORTH_INDIC_QUARTER_MARK = 0xA836;
const char32_t NORTH_INDIC_PLACEHOLDER_MARK = 0xA837;
const char32_t NORTH_INDIC_RUPEE_MARK = 0xA838;
const char32_t NORTH_INDIC_QUANTITY_MARK = 0xA839;
const char32_t PHAGS_PA_LETTER_KA = 0xA840;
const char32_t PHAGS_PA_LETTER_KHA = 0xA841;
const char32_t PHAGS_PA_LETTER_GA = 0xA842;
const char32_t PHAGS_PA_LETTER_NGA = 0xA843;
const char32_t PHAGS_PA_LETTER_CA = 0xA844;
const char32_t PHAGS_PA_LETTER_CHA = 0xA845;
const char32_t PHAGS_PA_LETTER_JA = 0xA846;
const char32_t PHAGS_PA_LETTER_NYA = 0xA847;
const char32_t PHAGS_PA_LETTER_TA = 0xA848;
const char32_t PHAGS_PA_LETTER_THA = 0xA849;
const char32_t PHAGS_PA_LETTER_DA = 0xA84A;
const char32_t PHAGS_PA_LETTER_NA = 0xA84B;
const char32_t PHAGS_PA_LETTER_PA = 0xA84C;
const char32_t PHAGS_PA_LETTER_PHA = 0xA84D;
const char32_t PHAGS_PA_LETTER_BA = 0xA84E;
const char32_t PHAGS_PA_LETTER_MA = 0xA84F;
const char32_t PHAGS_PA_LETTER_TSA = 0xA850;
const char32_t PHAGS_PA_LETTER_TSHA = 0xA851;
const char32_t PHAGS_PA_LETTER_DZA = 0xA852;
const char32_t PHAGS_PA_LETTER_WA = 0xA853;
const char32_t PHAGS_PA_LETTER_ZHA = 0xA854;
const char32_t PHAGS_PA_LETTER_ZA = 0xA855;
const char32_t PHAGS_PA_LETTER_SMALL_A = 0xA856;
const char32_t PHAGS_PA_LETTER_YA = 0xA857;
const char32_t PHAGS_PA_LETTER_RA = 0xA858;
const char32_t PHAGS_PA_LETTER_LA = 0xA859;
const char32_t PHAGS_PA_LETTER_SHA = 0xA85A;
const char32_t PHAGS_PA_LETTER_SA = 0xA85B;
const char32_t PHAGS_PA_LETTER_HA = 0xA85C;
const char32_t PHAGS_PA_LETTER_A = 0xA85D;
const char32_t PHAGS_PA_LETTER_I = 0xA85E;
const char32_t PHAGS_PA_LETTER_U = 0xA85F;
const char32_t PHAGS_PA_LETTER_E = 0xA860;
const char32_t PHAGS_PA_LETTER_O = 0xA861;
const char32_t PHAGS_PA_LETTER_QA = 0xA862;
const char32_t PHAGS_PA_LETTER_XA = 0xA863;
const char32_t PHAGS_PA_LETTER_FA = 0xA864;
const char32_t PHAGS_PA_LETTER_GGA = 0xA865;
const char32_t PHAGS_PA_LETTER_EE = 0xA866;
const char32_t PHAGS_PA_SUBJOINED_LETTER_WA = 0xA867;
const char32_t PHAGS_PA_SUBJOINED_LETTER_YA = 0xA868;
const char32_t PHAGS_PA_LETTER_TTA = 0xA869;
const char32_t PHAGS_PA_LETTER_TTHA = 0xA86A;
const char32_t PHAGS_PA_LETTER_DDA = 0xA86B;
const char32_t PHAGS_PA_LETTER_NNA = 0xA86C;
const char32_t PHAGS_PA_LETTER_ALTERNATE_YA = 0xA86D;
const char32_t PHAGS_PA_LETTER_VOICELESS_SHA = 0xA86E;
const char32_t PHAGS_PA_LETTER_VOICED_HA = 0xA86F;
const char32_t PHAGS_PA_LETTER_ASPIRATED_FA = 0xA870;
const char32_t PHAGS_PA_SUBJOINED_LETTER_RA = 0xA871;
const char32_t PHAGS_PA_SUPERFIXED_LETTER_RA = 0xA872;
const char32_t PHAGS_PA_LETTER_CANDRABINDU = 0xA873;
const char32_t PHAGS_PA_SINGLE_HEAD_MARK = 0xA874;
const char32_t PHAGS_PA_DOUBLE_HEAD_MARK = 0xA875;
const char32_t PHAGS_PA_MARK_SHAD = 0xA876;
const char32_t PHAGS_PA_MARK_DOUBLE_SHAD = 0xA877;
const char32_t SAURASHTRA_SIGN_ANUSVARA = 0xA880;
const char32_t SAURASHTRA_SIGN_VISARGA = 0xA881;
const char32_t SAURASHTRA_LETTER_A = 0xA882;
const char32_t SAURASHTRA_LETTER_AA = 0xA883;
const char32_t SAURASHTRA_LETTER_I = 0xA884;
const char32_t SAURASHTRA_LETTER_II = 0xA885;
const char32_t SAURASHTRA_LETTER_U = 0xA886;
const char32_t SAURASHTRA_LETTER_UU = 0xA887;
const char32_t SAURASHTRA_LETTER_VOCALIC_R = 0xA888;
const char32_t SAURASHTRA_LETTER_VOCALIC_RR = 0xA889;
const char32_t SAURASHTRA_LETTER_VOCALIC_L = 0xA88A;
const char32_t SAURASHTRA_LETTER_VOCALIC_LL = 0xA88B;
const char32_t SAURASHTRA_LETTER_E = 0xA88C;
const char32_t SAURASHTRA_LETTER_EE = 0xA88D;
const char32_t SAURASHTRA_LETTER_AI = 0xA88E;
const char32_t SAURASHTRA_LETTER_O = 0xA88F;
const char32_t SAURASHTRA_LETTER_OO = 0xA890;
const char32_t SAURASHTRA_LETTER_AU = 0xA891;
const char32_t SAURASHTRA_LETTER_KA = 0xA892;
const char32_t SAURASHTRA_LETTER_KHA = 0xA893;
const char32_t SAURASHTRA_LETTER_GA = 0xA894;
const char32_t SAURASHTRA_LETTER_GHA = 0xA895;
const char32_t SAURASHTRA_LETTER_NGA = 0xA896;
const char32_t SAURASHTRA_LETTER_CA = 0xA897;
const char32_t SAURASHTRA_LETTER_CHA = 0xA898;
const char32_t SAURASHTRA_LETTER_JA = 0xA899;
const char32_t SAURASHTRA_LETTER_JHA = 0xA89A;
const char32_t SAURASHTRA_LETTER_NYA = 0xA89B;
const char32_t SAURASHTRA_LETTER_TTA = 0xA89C;
const char32_t SAURASHTRA_LETTER_TTHA = 0xA89D;
const char32_t SAURASHTRA_LETTER_DDA = 0xA89E;
const char32_t SAURASHTRA_LETTER_DDHA = 0xA89F;
const char32_t SAURASHTRA_LETTER_NNA = 0xA8A0;
const char32_t SAURASHTRA_LETTER_TA = 0xA8A1;
const char32_t SAURASHTRA_LETTER_THA = 0xA8A2;
const char32_t SAURASHTRA_LETTER_DA = 0xA8A3;
const char32_t SAURASHTRA_LETTER_DHA = 0xA8A4;
const char32_t SAURASHTRA_LETTER_NA = 0xA8A5;
const char32_t SAURASHTRA_LETTER_PA = 0xA8A6;
const char32_t SAURASHTRA_LETTER_PHA = 0xA8A7;
const char32_t SAURASHTRA_LETTER_BA = 0xA8A8;
const char32_t SAURASHTRA_LETTER_BHA = 0xA8A9;
const char32_t SAURASHTRA_LETTER_MA = 0xA8AA;
const char32_t SAURASHTRA_LETTER_YA = 0xA8AB;
const char32_t SAURASHTRA_LETTER_RA = 0xA8AC;
const char32_t SAURASHTRA_LETTER_LA = 0xA8AD;
const char32_t SAURASHTRA_LETTER_VA = 0xA8AE;
const char32_t SAURASHTRA_LETTER_SHA = 0xA8AF;
const char32_t SAURASHTRA_LETTER_SSA = 0xA8B0;
const char32_t SAURASHTRA_LETTER_SA = 0xA8B1;
const char32_t SAURASHTRA_LETTER_HA = 0xA8B2;
const char32_t SAURASHTRA_LETTER_LLA = 0xA8B3;
const char32_t SAURASHTRA_CONSONANT_SIGN_HAARU = 0xA8B4;
const char32_t SAURASHTRA_VOWEL_SIGN_AA = 0xA8B5;
const char32_t SAURASHTRA_VOWEL_SIGN_I = 0xA8B6;
const char32_t SAURASHTRA_VOWEL_SIGN_II = 0xA8B7;
const char32_t SAURASHTRA_VOWEL_SIGN_U = 0xA8B8;
const char32_t SAURASHTRA_VOWEL_SIGN_UU = 0xA8B9;
const char32_t SAURASHTRA_VOWEL_SIGN_VOCALIC_R = 0xA8BA;
const char32_t SAURASHTRA_VOWEL_SIGN_VOCALIC_RR = 0xA8BB;
const char32_t SAURASHTRA_VOWEL_SIGN_VOCALIC_L = 0xA8BC;
const char32_t SAURASHTRA_VOWEL_SIGN_VOCALIC_LL = 0xA8BD;
const char32_t SAURASHTRA_VOWEL_SIGN_E = 0xA8BE;
const char32_t SAURASHTRA_VOWEL_SIGN_EE = 0xA8BF;
const char32_t SAURASHTRA_VOWEL_SIGN_AI = 0xA8C0;
const char32_t SAURASHTRA_VOWEL_SIGN_O = 0xA8C1;
const char32_t SAURASHTRA_VOWEL_SIGN_OO = 0xA8C2;
const char32_t SAURASHTRA_VOWEL_SIGN_AU = 0xA8C3;
const char32_t SAURASHTRA_SIGN_VIRAMA = 0xA8C4;
const char32_t SAURASHTRA_SIGN_CANDRABINDU = 0xA8C5;
const char32_t SAURASHTRA_DANDA = 0xA8CE;
const char32_t SAURASHTRA_DOUBLE_DANDA = 0xA8CF;
const char32_t SAURASHTRA_DIGIT_ZERO = 0xA8D0;
const char32_t SAURASHTRA_DIGIT_ONE = 0xA8D1;
const char32_t SAURASHTRA_DIGIT_TWO = 0xA8D2;
const char32_t SAURASHTRA_DIGIT_THREE = 0xA8D3;
const char32_t SAURASHTRA_DIGIT_FOUR = 0xA8D4;
const char32_t SAURASHTRA_DIGIT_FIVE = 0xA8D5;
const char32_t SAURASHTRA_DIGIT_SIX = 0xA8D6;
const char32_t SAURASHTRA_DIGIT_SEVEN = 0xA8D7;
const char32_t SAURASHTRA_DIGIT_EIGHT = 0xA8D8;
const char32_t SAURASHTRA_DIGIT_NINE = 0xA8D9;
const char32_t COMBINING_DEVANAGARI_DIGIT_ZERO = 0xA8E0;
const char32_t COMBINING_DEVANAGARI_DIGIT_ONE = 0xA8E1;
const char32_t COMBINING_DEVANAGARI_DIGIT_TWO = 0xA8E2;
const char32_t COMBINING_DEVANAGARI_DIGIT_THREE = 0xA8E3;
const char32_t COMBINING_DEVANAGARI_DIGIT_FOUR = 0xA8E4;
const char32_t COMBINING_DEVANAGARI_DIGIT_FIVE = 0xA8E5;
const char32_t COMBINING_DEVANAGARI_DIGIT_SIX = 0xA8E6;
const char32_t COMBINING_DEVANAGARI_DIGIT_SEVEN = 0xA8E7;
const char32_t COMBINING_DEVANAGARI_DIGIT_EIGHT = 0xA8E8;
const char32_t COMBINING_DEVANAGARI_DIGIT_NINE = 0xA8E9;
const char32_t COMBINING_DEVANAGARI_LETTER_A = 0xA8EA;
const char32_t COMBINING_DEVANAGARI_LETTER_U = 0xA8EB;
const char32_t COMBINING_DEVANAGARI_LETTER_KA = 0xA8EC;
const char32_t COMBINING_DEVANAGARI_LETTER_NA = 0xA8ED;
const char32_t COMBINING_DEVANAGARI_LETTER_PA = 0xA8EE;
const char32_t COMBINING_DEVANAGARI_LETTER_RA = 0xA8EF;
const char32_t COMBINING_DEVANAGARI_LETTER_VI = 0xA8F0;
const char32_t COMBINING_DEVANAGARI_SIGN_AVAGRAHA = 0xA8F1;
const char32_t DEVANAGARI_SIGN_SPACING_CANDRABINDU = 0xA8F2;
const char32_t DEVANAGARI_SIGN_CANDRABINDU_VIRAMA = 0xA8F3;
const char32_t DEVANAGARI_SIGN_DOUBLE_CANDRABINDU_VIRAMA = 0xA8F4;
const char32_t DEVANAGARI_SIGN_CANDRABINDU_TWO = 0xA8F5;
const char32_t DEVANAGARI_SIGN_CANDRABINDU_THREE = 0xA8F6;
const char32_t DEVANAGARI_SIGN_CANDRABINDU_AVAGRAHA = 0xA8F7;
const char32_t DEVANAGARI_SIGN_PUSHPIKA = 0xA8F8;
const char32_t DEVANAGARI_GAP_FILLER = 0xA8F9;
const char32_t DEVANAGARI_CARET = 0xA8FA;
const char32_t DEVANAGARI_HEADSTROKE = 0xA8FB;
const char32_t DEVANAGARI_SIGN_SIDDHAM = 0xA8FC;
const char32_t DEVANAGARI_JAIN_OM = 0xA8FD;
const char32_t DEVANAGARI_LETTER_AY = 0xA8FE;
const char32_t DEVANAGARI_VOWEL_SIGN_AY = 0xA8FF;
const char32_t KAYAH_LI_DIGIT_ZERO = 0xA900;
const char32_t KAYAH_LI_DIGIT_ONE = 0xA901;
const char32_t KAYAH_LI_DIGIT_TWO = 0xA902;
const char32_t KAYAH_LI_DIGIT_THREE = 0xA903;
const char32_t KAYAH_LI_DIGIT_FOUR = 0xA904;
const char32_t KAYAH_LI_DIGIT_FIVE = 0xA905;
const char32_t KAYAH_LI_DIGIT_SIX = 0xA906;
const char32_t KAYAH_LI_DIGIT_SEVEN = 0xA907;
const char32_t KAYAH_LI_DIGIT_EIGHT = 0xA908;
const char32_t KAYAH_LI_DIGIT_NINE = 0xA909;
const char32_t KAYAH_LI_LETTER_KA = 0xA90A;
const char32_t KAYAH_LI_LETTER_KHA = 0xA90B;
const char32_t KAYAH_LI_LETTER_GA = 0xA90C;
const char32_t KAYAH_LI_LETTER_NGA = 0xA90D;
const char32_t KAYAH_LI_LETTER_SA = 0xA90E;
const char32_t KAYAH_LI_LETTER_SHA = 0xA90F;
const char32_t KAYAH_LI_LETTER_ZA = 0xA910;
const char32_t KAYAH_LI_LETTER_NYA = 0xA911;
const char32_t KAYAH_LI_LETTER_TA = 0xA912;
const char32_t KAYAH_LI_LETTER_HTA = 0xA913;
const char32_t KAYAH_LI_LETTER_NA = 0xA914;
const char32_t KAYAH_LI_LETTER_PA = 0xA915;
const char32_t KAYAH_LI_LETTER_PHA = 0xA916;
const char32_t KAYAH_LI_LETTER_MA = 0xA917;
const char32_t KAYAH_LI_LETTER_DA = 0xA918;
const char32_t KAYAH_LI_LETTER_BA = 0xA919;
const char32_t KAYAH_LI_LETTER_RA = 0xA91A;
const char32_t KAYAH_LI_LETTER_YA = 0xA91B;
const char32_t KAYAH_LI_LETTER_LA = 0xA91C;
const char32_t KAYAH_LI_LETTER_WA = 0xA91D;
const char32_t KAYAH_LI_LETTER_THA = 0xA91E;
const char32_t KAYAH_LI_LETTER_HA = 0xA91F;
const char32_t KAYAH_LI_LETTER_VA = 0xA920;
const char32_t KAYAH_LI_LETTER_CA = 0xA921;
const char32_t KAYAH_LI_LETTER_A = 0xA922;
const char32_t KAYAH_LI_LETTER_OE = 0xA923;
const char32_t KAYAH_LI_LETTER_I = 0xA924;
const char32_t KAYAH_LI_LETTER_OO = 0xA925;
const char32_t KAYAH_LI_VOWEL_UE = 0xA926;
const char32_t KAYAH_LI_VOWEL_E = 0xA927;
const char32_t KAYAH_LI_VOWEL_U = 0xA928;
const char32_t KAYAH_LI_VOWEL_EE = 0xA929;
const char32_t KAYAH_LI_VOWEL_O = 0xA92A;
const char32_t KAYAH_LI_TONE_PLOPHU = 0xA92B;
const char32_t KAYAH_LI_TONE_CALYA = 0xA92C;
const char32_t KAYAH_LI_TONE_CALYA_PLOPHU = 0xA92D;
const char32_t KAYAH_LI_SIGN_CWI = 0xA92E;
const char32_t KAYAH_LI_SIGN_SHYA = 0xA92F;
const char32_t REJANG_LETTER_KA = 0xA930;
const char32_t REJANG_LETTER_GA = 0xA931;
const char32_t REJANG_LETTER_NGA = 0xA932;
const char32_t REJANG_LETTER_TA = 0xA933;
const char32_t REJANG_LETTER_DA = 0xA934;
const char32_t REJANG_LETTER_NA = 0xA935;
const char32_t REJANG_LETTER_PA = 0xA936;
const char32_t REJANG_LETTER_BA = 0xA937;
const char32_t REJANG_LETTER_MA = 0xA938;
const char32_t REJANG_LETTER_CA = 0xA939;
const char32_t REJANG_LETTER_JA = 0xA93A;
const char32_t REJANG_LETTER_NYA = 0xA93B;
const char32_t REJANG_LETTER_SA = 0xA93C;
const char32_t REJANG_LETTER_RA = 0xA93D;
const char32_t REJANG_LETTER_LA = 0xA93E;
const char32_t REJANG_LETTER_YA = 0xA93F;
const char32_t REJANG_LETTER_WA = 0xA940;
const char32_t REJANG_LETTER_HA = 0xA941;
const char32_t REJANG_LETTER_MBA = 0xA942;
const char32_t REJANG_LETTER_NGGA = 0xA943;
const char32_t REJANG_LETTER_NDA = 0xA944;
const char32_t REJANG_LETTER_NYJA = 0xA945;
const char32_t REJANG_LETTER_A = 0xA946;
const char32_t REJANG_VOWEL_SIGN_I = 0xA947;
const char32_t REJANG_VOWEL_SIGN_U = 0xA948;
const char32_t REJANG_VOWEL_SIGN_E = 0xA949;
const char32_t REJANG_VOWEL_SIGN_AI = 0xA94A;
const char32_t REJANG_VOWEL_SIGN_O = 0xA94B;
const char32_t REJANG_VOWEL_SIGN_AU = 0xA94C;
const char32_t REJANG_VOWEL_SIGN_EU = 0xA94D;
const char32_t REJANG_VOWEL_SIGN_EA = 0xA94E;
const char32_t REJANG_CONSONANT_SIGN_NG = 0xA94F;
const char32_t REJANG_CONSONANT_SIGN_N = 0xA950;
const char32_t REJANG_CONSONANT_SIGN_R = 0xA951;
const char32_t REJANG_CONSONANT_SIGN_H = 0xA952;
const char32_t REJANG_VIRAMA = 0xA953;
const char32_t REJANG_SECTION_MARK = 0xA95F;
const char32_t HANGUL_CHOSEONG_TIKEUT_MIEUM = 0xA960;
const char32_t HANGUL_CHOSEONG_TIKEUT_PIEUP = 0xA961;
const char32_t HANGUL_CHOSEONG_TIKEUT_SIOS = 0xA962;
const char32_t HANGUL_CHOSEONG_TIKEUT_CIEUC = 0xA963;
const char32_t HANGUL_CHOSEONG_RIEUL_KIYEOK = 0xA964;
const char32_t HANGUL_CHOSEONG_RIEUL_SSANGKIYEOK = 0xA965;
const char32_t HANGUL_CHOSEONG_RIEUL_TIKEUT = 0xA966;
const char32_t HANGUL_CHOSEONG_RIEUL_SSANGTIKEUT = 0xA967;
const char32_t HANGUL_CHOSEONG_RIEUL_MIEUM = 0xA968;
const char32_t HANGUL_CHOSEONG_RIEUL_PIEUP = 0xA969;
const char32_t HANGUL_CHOSEONG_RIEUL_SSANGPIEUP = 0xA96A;
const char32_t HANGUL_CHOSEONG_RIEUL_KAPYEOUNPIEUP = 0xA96B;
const char32_t HANGUL_CHOSEONG_RIEUL_SIOS = 0xA96C;
const char32_t HANGUL_CHOSEONG_RIEUL_CIEUC = 0xA96D;
const char32_t HANGUL_CHOSEONG_RIEUL_KHIEUKH = 0xA96E;
const char32_t HANGUL_CHOSEONG_MIEUM_KIYEOK = 0xA96F;
const char32_t HANGUL_CHOSEONG_MIEUM_TIKEUT = 0xA970;
const char32_t HANGUL_CHOSEONG_MIEUM_SIOS = 0xA971;
const char32_t HANGUL_CHOSEONG_PIEUP_SIOS_THIEUTH = 0xA972;
const char32_t HANGUL_CHOSEONG_PIEUP_KHIEUKH = 0xA973;
const char32_t HANGUL_CHOSEONG_PIEUP_HIEUH = 0xA974;
const char32_t HANGUL_CHOSEONG_SSANGSIOS_PIEUP = 0xA975;
const char32_t HANGUL_CHOSEONG_IEUNG_RIEUL = 0xA976;
const char32_t HANGUL_CHOSEONG_IEUNG_HIEUH = 0xA977;
const char32_t HANGUL_CHOSEONG_SSANGCIEUC_HIEUH = 0xA978;
const char32_t HANGUL_CHOSEONG_SSANGTHIEUTH = 0xA979;
const char32_t HANGUL_CHOSEONG_PHIEUPH_HIEUH = 0xA97A;
const char32_t HANGUL_CHOSEONG_HIEUH_SIOS = 0xA97B;
const char32_t HANGUL_CHOSEONG_SSANGYEORINHIEUH = 0xA97C;
const char32_t JAVANESE_SIGN_PANYANGGA = 0xA980;
const char32_t JAVANESE_SIGN_CECAK = 0xA981;
const char32_t JAVANESE_SIGN_LAYAR = 0xA982;
const char32_t JAVANESE_SIGN_WIGNYAN = 0xA983;
const char32_t JAVANESE_LETTER_A = 0xA984;
const char32_t JAVANESE_LETTER_I_KAWI = 0xA985;
const char32_t JAVANESE_LETTER_I = 0xA986;
const char32_t JAVANESE_LETTER_II = 0xA987;
const char32_t JAVANESE_LETTER_U = 0xA988;
const char32_t JAVANESE_LETTER_PA_CEREK = 0xA989;
const char32_t JAVANESE_LETTER_NGA_LELET = 0xA98A;
const char32_t JAVANESE_LETTER_NGA_LELET_RASWADI = 0xA98B;
const char32_t JAVANESE_LETTER_E = 0xA98C;
const char32_t JAVANESE_LETTER_AI = 0xA98D;
const char32_t JAVANESE_LETTER_O = 0xA98E;
const char32_t JAVANESE_LETTER_KA = 0xA98F;
const char32_t JAVANESE_LETTER_KA_SASAK = 0xA990;
const char32_t JAVANESE_LETTER_KA_MURDA = 0xA991;
const char32_t JAVANESE_LETTER_GA = 0xA992;
const char32_t JAVANESE_LETTER_GA_MURDA = 0xA993;
const char32_t JAVANESE_LETTER_NGA = 0xA994;
const char32_t JAVANESE_LETTER_CA = 0xA995;
const char32_t JAVANESE_LETTER_CA_MURDA = 0xA996;
const char32_t JAVANESE_LETTER_JA = 0xA997;
const char32_t JAVANESE_LETTER_NYA_MURDA = 0xA998;
const char32_t JAVANESE_LETTER_JA_MAHAPRANA = 0xA999;
const char32_t JAVANESE_LETTER_NYA = 0xA99A;
const char32_t JAVANESE_LETTER_TTA = 0xA99B;
const char32_t JAVANESE_LETTER_TTA_MAHAPRANA = 0xA99C;
const char32_t JAVANESE_LETTER_DDA = 0xA99D;
const char32_t JAVANESE_LETTER_DDA_MAHAPRANA = 0xA99E;
const char32_t JAVANESE_LETTER_NA_MURDA = 0xA99F;
const char32_t JAVANESE_LETTER_TA = 0xA9A0;
const char32_t JAVANESE_LETTER_TA_MURDA = 0xA9A1;
const char32_t JAVANESE_LETTER_DA = 0xA9A2;
const char32_t JAVANESE_LETTER_DA_MAHAPRANA = 0xA9A3;
const char32_t JAVANESE_LETTER_NA = 0xA9A4;
const char32_t JAVANESE_LETTER_PA = 0xA9A5;
const char32_t JAVANESE_LETTER_PA_MURDA = 0xA9A6;
const char32_t JAVANESE_LETTER_BA = 0xA9A7;
const char32_t JAVANESE_LETTER_BA_MURDA = 0xA9A8;
const char32_t JAVANESE_LETTER_MA = 0xA9A9;
const char32_t JAVANESE_LETTER_YA = 0xA9AA;
const char32_t JAVANESE_LETTER_RA = 0xA9AB;
const char32_t JAVANESE_LETTER_RA_AGUNG = 0xA9AC;
const char32_t JAVANESE_LETTER_LA = 0xA9AD;
const char32_t JAVANESE_LETTER_WA = 0xA9AE;
const char32_t JAVANESE_LETTER_SA_MURDA = 0xA9AF;
const char32_t JAVANESE_LETTER_SA_MAHAPRANA = 0xA9B0;
const char32_t JAVANESE_LETTER_SA = 0xA9B1;
const char32_t JAVANESE_LETTER_HA = 0xA9B2;
const char32_t JAVANESE_SIGN_CECAK_TELU = 0xA9B3;
const char32_t JAVANESE_VOWEL_SIGN_TARUNG = 0xA9B4;
const char32_t JAVANESE_VOWEL_SIGN_TOLONG = 0xA9B5;
const char32_t JAVANESE_VOWEL_SIGN_WULU = 0xA9B6;
const char32_t JAVANESE_VOWEL_SIGN_WULU_MELIK = 0xA9B7;
const char32_t JAVANESE_VOWEL_SIGN_SUKU = 0xA9B8;
const char32_t JAVANESE_VOWEL_SIGN_SUKU_MENDUT = 0xA9B9;
const char32_t JAVANESE_VOWEL_SIGN_TALING = 0xA9BA;
const char32_t JAVANESE_VOWEL_SIGN_DIRGA_MURE = 0xA9BB;
const char32_t JAVANESE_VOWEL_SIGN_PEPET = 0xA9BC;
const char32_t JAVANESE_CONSONANT_SIGN_KERET = 0xA9BD;
const char32_t JAVANESE_CONSONANT_SIGN_PENGKAL = 0xA9BE;
const char32_t JAVANESE_CONSONANT_SIGN_CAKRA = 0xA9BF;
const char32_t JAVANESE_PANGKON = 0xA9C0;
const char32_t JAVANESE_LEFT_RERENGGAN = 0xA9C1;
const char32_t JAVANESE_RIGHT_RERENGGAN = 0xA9C2;
const char32_t JAVANESE_PADA_ANDAP = 0xA9C3;
const char32_t JAVANESE_PADA_MADYA = 0xA9C4;
const char32_t JAVANESE_PADA_LUHUR = 0xA9C5;
const char32_t JAVANESE_PADA_WINDU = 0xA9C6;
const char32_t JAVANESE_PADA_PANGKAT = 0xA9C7;
const char32_t JAVANESE_PADA_LINGSA = 0xA9C8;
const char32_t JAVANESE_PADA_LUNGSI = 0xA9C9;
const char32_t JAVANESE_PADA_ADEG = 0xA9CA;
const char32_t JAVANESE_PADA_ADEG_ADEG = 0xA9CB;
const char32_t JAVANESE_PADA_PISELEH = 0xA9CC;
const char32_t JAVANESE_TURNED_PADA_PISELEH = 0xA9CD;
const char32_t JAVANESE_PANGRANGKEP = 0xA9CF;
const char32_t JAVANESE_DIGIT_ZERO = 0xA9D0;
const char32_t JAVANESE_DIGIT_ONE = 0xA9D1;
const char32_t JAVANESE_DIGIT_TWO = 0xA9D2;
const char32_t JAVANESE_DIGIT_THREE = 0xA9D3;
const char32_t JAVANESE_DIGIT_FOUR = 0xA9D4;
const char32_t JAVANESE_DIGIT_FIVE = 0xA9D5;
const char32_t JAVANESE_DIGIT_SIX = 0xA9D6;
const char32_t JAVANESE_DIGIT_SEVEN = 0xA9D7;
const char32_t JAVANESE_DIGIT_EIGHT = 0xA9D8;
const char32_t JAVANESE_DIGIT_NINE = 0xA9D9;
const char32_t JAVANESE_PADA_TIRTA_TUMETES = 0xA9DE;
const char32_t JAVANESE_PADA_ISEN_ISEN = 0xA9DF;
const char32_t MYANMAR_LETTER_SHAN_GHA = 0xA9E0;
const char32_t MYANMAR_LETTER_SHAN_CHA = 0xA9E1;
const char32_t MYANMAR_LETTER_SHAN_JHA = 0xA9E2;
const char32_t MYANMAR_LETTER_SHAN_NNA = 0xA9E3;
const char32_t MYANMAR_LETTER_SHAN_BHA = 0xA9E4;
const char32_t MYANMAR_SIGN_SHAN_SAW = 0xA9E5;
const char32_t MYANMAR_MODIFIER_LETTER_SHAN_REDUPLICATION = 0xA9E6;
const char32_t MYANMAR_LETTER_TAI_LAING_NYA = 0xA9E7;
const char32_t MYANMAR_LETTER_TAI_LAING_FA = 0xA9E8;
const char32_t MYANMAR_LETTER_TAI_LAING_GA = 0xA9E9;
const char32_t MYANMAR_LETTER_TAI_LAING_GHA = 0xA9EA;
const char32_t MYANMAR_LETTER_TAI_LAING_JA = 0xA9EB;
const char32_t MYANMAR_LETTER_TAI_LAING_JHA = 0xA9EC;
const char32_t MYANMAR_LETTER_TAI_LAING_DDA = 0xA9ED;
const char32_t MYANMAR_LETTER_TAI_LAING_DDHA = 0xA9EE;
const char32_t MYANMAR_LETTER_TAI_LAING_NNA = 0xA9EF;
const char32_t MYANMAR_TAI_LAING_DIGIT_ZERO = 0xA9F0;
const char32_t MYANMAR_TAI_LAING_DIGIT_ONE = 0xA9F1;
const char32_t MYANMAR_TAI_LAING_DIGIT_TWO = 0xA9F2;
const char32_t MYANMAR_TAI_LAING_DIGIT_THREE = 0xA9F3;
const char32_t MYANMAR_TAI_LAING_DIGIT_FOUR = 0xA9F4;
const char32_t MYANMAR_TAI_LAING_DIGIT_FIVE = 0xA9F5;
const char32_t MYANMAR_TAI_LAING_DIGIT_SIX = 0xA9F6;
const char32_t MYANMAR_TAI_LAING_DIGIT_SEVEN = 0xA9F7;
const char32_t MYANMAR_TAI_LAING_DIGIT_EIGHT = 0xA9F8;
const char32_t MYANMAR_TAI_LAING_DIGIT_NINE = 0xA9F9;
const char32_t MYANMAR_LETTER_TAI_LAING_LLA = 0xA9FA;
const char32_t MYANMAR_LETTER_TAI_LAING_DA = 0xA9FB;
const char32_t MYANMAR_LETTER_TAI_LAING_DHA = 0xA9FC;
const char32_t MYANMAR_LETTER_TAI_LAING_BA = 0xA9FD;
const char32_t MYANMAR_LETTER_TAI_LAING_BHA = 0xA9FE;
const char32_t CHAM_LETTER_A = 0xAA00;
const char32_t CHAM_LETTER_I = 0xAA01;
const char32_t CHAM_LETTER_U = 0xAA02;
const char32_t CHAM_LETTER_E = 0xAA03;
const char32_t CHAM_LETTER_AI = 0xAA04;
const char32_t CHAM_LETTER_O = 0xAA05;
const char32_t CHAM_LETTER_KA = 0xAA06;
const char32_t CHAM_LETTER_KHA = 0xAA07;
const char32_t CHAM_LETTER_GA = 0xAA08;
const char32_t CHAM_LETTER_GHA = 0xAA09;
const char32_t CHAM_LETTER_NGUE = 0xAA0A;
const char32_t CHAM_LETTER_NGA = 0xAA0B;
const char32_t CHAM_LETTER_CHA = 0xAA0C;
const char32_t CHAM_LETTER_CHHA = 0xAA0D;
const char32_t CHAM_LETTER_JA = 0xAA0E;
const char32_t CHAM_LETTER_JHA = 0xAA0F;
const char32_t CHAM_LETTER_NHUE = 0xAA10;
const char32_t CHAM_LETTER_NHA = 0xAA11;
const char32_t CHAM_LETTER_NHJA = 0xAA12;
const char32_t CHAM_LETTER_TA = 0xAA13;
const char32_t CHAM_LETTER_THA = 0xAA14;
const char32_t CHAM_LETTER_DA = 0xAA15;
const char32_t CHAM_LETTER_DHA = 0xAA16;
const char32_t CHAM_LETTER_NUE = 0xAA17;
const char32_t CHAM_LETTER_NA = 0xAA18;
const char32_t CHAM_LETTER_DDA = 0xAA19;
const char32_t CHAM_LETTER_PA = 0xAA1A;
const char32_t CHAM_LETTER_PPA = 0xAA1B;
const char32_t CHAM_LETTER_PHA = 0xAA1C;
const char32_t CHAM_LETTER_BA = 0xAA1D;
const char32_t CHAM_LETTER_BHA = 0xAA1E;
const char32_t CHAM_LETTER_MUE = 0xAA1F;
const char32_t CHAM_LETTER_MA = 0xAA20;
const char32_t CHAM_LETTER_BBA = 0xAA21;
const char32_t CHAM_LETTER_YA = 0xAA22;
const char32_t CHAM_LETTER_RA = 0xAA23;
const char32_t CHAM_LETTER_LA = 0xAA24;
const char32_t CHAM_LETTER_VA = 0xAA25;
const char32_t CHAM_LETTER_SSA = 0xAA26;
const char32_t CHAM_LETTER_SA = 0xAA27;
const char32_t CHAM_LETTER_HA = 0xAA28;
const char32_t CHAM_VOWEL_SIGN_AA = 0xAA29;
const char32_t CHAM_VOWEL_SIGN_I = 0xAA2A;
const char32_t CHAM_VOWEL_SIGN_II = 0xAA2B;
const char32_t CHAM_VOWEL_SIGN_EI = 0xAA2C;
const char32_t CHAM_VOWEL_SIGN_U = 0xAA2D;
const char32_t CHAM_VOWEL_SIGN_OE = 0xAA2E;
const char32_t CHAM_VOWEL_SIGN_O = 0xAA2F;
const char32_t CHAM_VOWEL_SIGN_AI = 0xAA30;
const char32_t CHAM_VOWEL_SIGN_AU = 0xAA31;
const char32_t CHAM_VOWEL_SIGN_UE = 0xAA32;
const char32_t CHAM_CONSONANT_SIGN_YA = 0xAA33;
const char32_t CHAM_CONSONANT_SIGN_RA = 0xAA34;
const char32_t CHAM_CONSONANT_SIGN_LA = 0xAA35;
const char32_t CHAM_CONSONANT_SIGN_WA = 0xAA36;
const char32_t CHAM_LETTER_FINAL_K = 0xAA40;
const char32_t CHAM_LETTER_FINAL_G = 0xAA41;
const char32_t CHAM_LETTER_FINAL_NG = 0xAA42;
const char32_t CHAM_CONSONANT_SIGN_FINAL_NG = 0xAA43;
const char32_t CHAM_LETTER_FINAL_CH = 0xAA44;
const char32_t CHAM_LETTER_FINAL_T = 0xAA45;
const char32_t CHAM_LETTER_FINAL_N = 0xAA46;
const char32_t CHAM_LETTER_FINAL_P = 0xAA47;
const char32_t CHAM_LETTER_FINAL_Y = 0xAA48;
const char32_t CHAM_LETTER_FINAL_R = 0xAA49;
const char32_t CHAM_LETTER_FINAL_L = 0xAA4A;
const char32_t CHAM_LETTER_FINAL_SS = 0xAA4B;
const char32_t CHAM_CONSONANT_SIGN_FINAL_M = 0xAA4C;
const char32_t CHAM_CONSONANT_SIGN_FINAL_H = 0xAA4D;
const char32_t CHAM_DIGIT_ZERO = 0xAA50;
const char32_t CHAM_DIGIT_ONE = 0xAA51;
const char32_t CHAM_DIGIT_TWO = 0xAA52;
const char32_t CHAM_DIGIT_THREE = 0xAA53;
const char32_t CHAM_DIGIT_FOUR = 0xAA54;
const char32_t CHAM_DIGIT_FIVE = 0xAA55;
const char32_t CHAM_DIGIT_SIX = 0xAA56;
const char32_t CHAM_DIGIT_SEVEN = 0xAA57;
const char32_t CHAM_DIGIT_EIGHT = 0xAA58;
const char32_t CHAM_DIGIT_NINE = 0xAA59;
const char32_t CHAM_PUNCTUATION_SPIRAL = 0xAA5C;
const char32_t CHAM_PUNCTUATION_DANDA = 0xAA5D;
const char32_t CHAM_PUNCTUATION_DOUBLE_DANDA = 0xAA5E;
const char32_t CHAM_PUNCTUATION_TRIPLE_DANDA = 0xAA5F;
const char32_t MYANMAR_LETTER_KHAMTI_GA = 0xAA60;
const char32_t MYANMAR_LETTER_KHAMTI_CA = 0xAA61;
const char32_t MYANMAR_LETTER_KHAMTI_CHA = 0xAA62;
const char32_t MYANMAR_LETTER_KHAMTI_JA = 0xAA63;
const char32_t MYANMAR_LETTER_KHAMTI_JHA = 0xAA64;
const char32_t MYANMAR_LETTER_KHAMTI_NYA = 0xAA65;
const char32_t MYANMAR_LETTER_KHAMTI_TTA = 0xAA66;
const char32_t MYANMAR_LETTER_KHAMTI_TTHA = 0xAA67;
const char32_t MYANMAR_LETTER_KHAMTI_DDA = 0xAA68;
const char32_t MYANMAR_LETTER_KHAMTI_DDHA = 0xAA69;
const char32_t MYANMAR_LETTER_KHAMTI_DHA = 0xAA6A;
const char32_t MYANMAR_LETTER_KHAMTI_NA = 0xAA6B;
const char32_t MYANMAR_LETTER_KHAMTI_SA = 0xAA6C;
const char32_t MYANMAR_LETTER_KHAMTI_HA = 0xAA6D;
const char32_t MYANMAR_LETTER_KHAMTI_HHA = 0xAA6E;
const char32_t MYANMAR_LETTER_KHAMTI_FA = 0xAA6F;
const char32_t MYANMAR_MODIFIER_LETTER_KHAMTI_REDUPLICATION = 0xAA70;
const char32_t MYANMAR_LETTER_KHAMTI_XA = 0xAA71;
const char32_t MYANMAR_LETTER_KHAMTI_ZA = 0xAA72;
const char32_t MYANMAR_LETTER_KHAMTI_RA = 0xAA73;
const char32_t MYANMAR_LOGOGRAM_KHAMTI_OAY = 0xAA74;
const char32_t MYANMAR_LOGOGRAM_KHAMTI_QN = 0xAA75;
const char32_t MYANMAR_LOGOGRAM_KHAMTI_HM = 0xAA76;
const char32_t MYANMAR_SYMBOL_AITON_EXCLAMATION = 0xAA77;
const char32_t MYANMAR_SYMBOL_AITON_ONE = 0xAA78;
const char32_t MYANMAR_SYMBOL_AITON_TWO = 0xAA79;
const char32_t MYANMAR_LETTER_AITON_RA = 0xAA7A;
const char32_t MYANMAR_SIGN_PAO_KAREN_TONE = 0xAA7B;
const char32_t MYANMAR_SIGN_TAI_LAING_TONE_2 = 0xAA7C;
const char32_t MYANMAR_SIGN_TAI_LAING_TONE_5 = 0xAA7D;
const char32_t MYANMAR_LETTER_SHWE_PALAUNG_CHA = 0xAA7E;
const char32_t MYANMAR_LETTER_SHWE_PALAUNG_SHA = 0xAA7F;
const char32_t TAI_VIET_LETTER_LOW_KO = 0xAA80;
const char32_t TAI_VIET_LETTER_HIGH_KO = 0xAA81;
const char32_t TAI_VIET_LETTER_LOW_KHO = 0xAA82;
const char32_t TAI_VIET_LETTER_HIGH_KHO = 0xAA83;
const char32_t TAI_VIET_LETTER_LOW_KHHO = 0xAA84;
const char32_t TAI_VIET_LETTER_HIGH_KHHO = 0xAA85;
const char32_t TAI_VIET_LETTER_LOW_GO = 0xAA86;
const char32_t TAI_VIET_LETTER_HIGH_GO = 0xAA87;
const char32_t TAI_VIET_LETTER_LOW_NGO = 0xAA88;
const char32_t TAI_VIET_LETTER_HIGH_NGO = 0xAA89;
const char32_t TAI_VIET_LETTER_LOW_CO = 0xAA8A;
const char32_t TAI_VIET_LETTER_HIGH_CO = 0xAA8B;
const char32_t TAI_VIET_LETTER_LOW_CHO = 0xAA8C;
const char32_t TAI_VIET_LETTER_HIGH_CHO = 0xAA8D;
const char32_t TAI_VIET_LETTER_LOW_SO = 0xAA8E;
const char32_t TAI_VIET_LETTER_HIGH_SO = 0xAA8F;
const char32_t TAI_VIET_LETTER_LOW_NYO = 0xAA90;
const char32_t TAI_VIET_LETTER_HIGH_NYO = 0xAA91;
const char32_t TAI_VIET_LETTER_LOW_DO = 0xAA92;
const char32_t TAI_VIET_LETTER_HIGH_DO = 0xAA93;
const char32_t TAI_VIET_LETTER_LOW_TO = 0xAA94;
const char32_t TAI_VIET_LETTER_HIGH_TO = 0xAA95;
const char32_t TAI_VIET_LETTER_LOW_THO = 0xAA96;
const char32_t TAI_VIET_LETTER_HIGH_THO = 0xAA97;
const char32_t TAI_VIET_LETTER_LOW_NO = 0xAA98;
const char32_t TAI_VIET_LETTER_HIGH_NO = 0xAA99;
const char32_t TAI_VIET_LETTER_LOW_BO = 0xAA9A;
const char32_t TAI_VIET_LETTER_HIGH_BO = 0xAA9B;
const char32_t TAI_VIET_LETTER_LOW_PO = 0xAA9C;
const char32_t TAI_VIET_LETTER_HIGH_PO = 0xAA9D;
const char32_t TAI_VIET_LETTER_LOW_PHO = 0xAA9E;
const char32_t TAI_VIET_LETTER_HIGH_PHO = 0xAA9F;
const char32_t TAI_VIET_LETTER_LOW_FO = 0xAAA0;
const char32_t TAI_VIET_LETTER_HIGH_FO = 0xAAA1;
const char32_t TAI_VIET_LETTER_LOW_MO = 0xAAA2;
const char32_t TAI_VIET_LETTER_HIGH_MO = 0xAAA3;
const char32_t TAI_VIET_LETTER_LOW_YO = 0xAAA4;
const char32_t TAI_VIET_LETTER_HIGH_YO = 0xAAA5;
const char32_t TAI_VIET_LETTER_LOW_RO = 0xAAA6;
const char32_t TAI_VIET_LETTER_HIGH_RO = 0xAAA7;
const char32_t TAI_VIET_LETTER_LOW_LO = 0xAAA8;
const char32_t TAI_VIET_LETTER_HIGH_LO = 0xAAA9;
const char32_t TAI_VIET_LETTER_LOW_VO = 0xAAAA;
const char32_t TAI_VIET_LETTER_HIGH_VO = 0xAAAB;
const char32_t TAI_VIET_LETTER_LOW_HO = 0xAAAC;
const char32_t TAI_VIET_LETTER_HIGH_HO = 0xAAAD;
const char32_t TAI_VIET_LETTER_LOW_O = 0xAAAE;
const char32_t TAI_VIET_LETTER_HIGH_O = 0xAAAF;
const char32_t TAI_VIET_MAI_KANG = 0xAAB0;
const char32_t TAI_VIET_VOWEL_AA = 0xAAB1;
const char32_t TAI_VIET_VOWEL_I = 0xAAB2;
const char32_t TAI_VIET_VOWEL_UE = 0xAAB3;
const char32_t TAI_VIET_VOWEL_U = 0xAAB4;
const char32_t TAI_VIET_VOWEL_E = 0xAAB5;
const char32_t TAI_VIET_VOWEL_O = 0xAAB6;
const char32_t TAI_VIET_MAI_KHIT = 0xAAB7;
const char32_t TAI_VIET_VOWEL_IA = 0xAAB8;
const char32_t TAI_VIET_VOWEL_UEA = 0xAAB9;
const char32_t TAI_VIET_VOWEL_UA = 0xAABA;
const char32_t TAI_VIET_VOWEL_AUE = 0xAABB;
const char32_t TAI_VIET_VOWEL_AY = 0xAABC;
const char32_t TAI_VIET_VOWEL_AN = 0xAABD;
const char32_t TAI_VIET_VOWEL_AM = 0xAABE;
const char32_t TAI_VIET_TONE_MAI_EK = 0xAABF;
const char32_t TAI_VIET_TONE_MAI_NUENG = 0xAAC0;
const char32_t TAI_VIET_TONE_MAI_THO = 0xAAC1;
const char32_t TAI_VIET_TONE_MAI_SONG = 0xAAC2;
const char32_t TAI_VIET_SYMBOL_KON = 0xAADB;
const char32_t TAI_VIET_SYMBOL_NUENG = 0xAADC;
const char32_t TAI_VIET_SYMBOL_SAM = 0xAADD;
const char32_t TAI_VIET_SYMBOL_HO_HOI = 0xAADE;
const char32_t TAI_VIET_SYMBOL_KOI_KOI = 0xAADF;
const char32_t MEETEI_MAYEK_LETTER_E = 0xAAE0;
const char32_t MEETEI_MAYEK_LETTER_O = 0xAAE1;
const char32_t MEETEI_MAYEK_LETTER_CHA = 0xAAE2;
const char32_t MEETEI_MAYEK_LETTER_NYA = 0xAAE3;
const char32_t MEETEI_MAYEK_LETTER_TTA = 0xAAE4;
const char32_t MEETEI_MAYEK_LETTER_TTHA = 0xAAE5;
const char32_t MEETEI_MAYEK_LETTER_DDA = 0xAAE6;
const char32_t MEETEI_MAYEK_LETTER_DDHA = 0xAAE7;
const char32_t MEETEI_MAYEK_LETTER_NNA = 0xAAE8;
const char32_t MEETEI_MAYEK_LETTER_SHA = 0xAAE9;
const char32_t MEETEI_MAYEK_LETTER_SSA = 0xAAEA;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_II = 0xAAEB;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_UU = 0xAAEC;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_AAI = 0xAAED;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_AU = 0xAAEE;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_AAU = 0xAAEF;
const char32_t MEETEI_MAYEK_CHEIKHAN = 0xAAF0;
const char32_t MEETEI_MAYEK_AHANG_KHUDAM = 0xAAF1;
const char32_t MEETEI_MAYEK_ANJI = 0xAAF2;
const char32_t MEETEI_MAYEK_SYLLABLE_REPETITION_MARK = 0xAAF3;
const char32_t MEETEI_MAYEK_WORD_REPETITION_MARK = 0xAAF4;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_VISARGA = 0xAAF5;
const char32_t MEETEI_MAYEK_VIRAMA = 0xAAF6;
const char32_t ETHIOPIC_SYLLABLE_TTHU = 0xAB01;
const char32_t ETHIOPIC_SYLLABLE_TTHI = 0xAB02;
const char32_t ETHIOPIC_SYLLABLE_TTHAA = 0xAB03;
const char32_t ETHIOPIC_SYLLABLE_TTHEE = 0xAB04;
const char32_t ETHIOPIC_SYLLABLE_TTHE = 0xAB05;
const char32_t ETHIOPIC_SYLLABLE_TTHO = 0xAB06;
const char32_t ETHIOPIC_SYLLABLE_DDHU = 0xAB09;
const char32_t ETHIOPIC_SYLLABLE_DDHI = 0xAB0A;
const char32_t ETHIOPIC_SYLLABLE_DDHAA = 0xAB0B;
const char32_t ETHIOPIC_SYLLABLE_DDHEE = 0xAB0C;
const char32_t ETHIOPIC_SYLLABLE_DDHE = 0xAB0D;
const char32_t ETHIOPIC_SYLLABLE_DDHO = 0xAB0E;
const char32_t ETHIOPIC_SYLLABLE_DZU = 0xAB11;
const char32_t ETHIOPIC_SYLLABLE_DZI = 0xAB12;
const char32_t ETHIOPIC_SYLLABLE_DZAA = 0xAB13;
const char32_t ETHIOPIC_SYLLABLE_DZEE = 0xAB14;
const char32_t ETHIOPIC_SYLLABLE_DZE = 0xAB15;
const char32_t ETHIOPIC_SYLLABLE_DZO = 0xAB16;
const char32_t ETHIOPIC_SYLLABLE_CCHHA = 0xAB20;
const char32_t ETHIOPIC_SYLLABLE_CCHHU = 0xAB21;
const char32_t ETHIOPIC_SYLLABLE_CCHHI = 0xAB22;
const char32_t ETHIOPIC_SYLLABLE_CCHHAA = 0xAB23;
const char32_t ETHIOPIC_SYLLABLE_CCHHEE = 0xAB24;
const char32_t ETHIOPIC_SYLLABLE_CCHHE = 0xAB25;
const char32_t ETHIOPIC_SYLLABLE_CCHHO = 0xAB26;
const char32_t ETHIOPIC_SYLLABLE_BBA = 0xAB28;
const char32_t ETHIOPIC_SYLLABLE_BBU = 0xAB29;
const char32_t ETHIOPIC_SYLLABLE_BBI = 0xAB2A;
const char32_t ETHIOPIC_SYLLABLE_BBAA = 0xAB2B;
const char32_t ETHIOPIC_SYLLABLE_BBEE = 0xAB2C;
const char32_t ETHIOPIC_SYLLABLE_BBE = 0xAB2D;
const char32_t ETHIOPIC_SYLLABLE_BBO = 0xAB2E;
const char32_t LATIN_SMALL_LETTER_BARRED_ALPHA = 0xAB30;
const char32_t LATIN_SMALL_LETTER_A_REVERSED_SCHWA = 0xAB31;
const char32_t LATIN_SMALL_LETTER_BLACKLETTER_E = 0xAB32;
const char32_t LATIN_SMALL_LETTER_BARRED_E = 0xAB33;
const char32_t LATIN_SMALL_LETTER_E_WITH_FLOURISH = 0xAB34;
const char32_t LATIN_SMALL_LETTER_LENIS_F = 0xAB35;
const char32_t LATIN_SMALL_LETTER_SCRIPT_G_WITH_CROSSED_TAIL = 0xAB36;
const char32_t LATIN_SMALL_LETTER_L_WITH_INVERTED_LAZY_S = 0xAB37;
const char32_t LATIN_SMALL_LETTER_L_WITH_DOUBLE_MIDDLE_TILDE = 0xAB38;
const char32_t LATIN_SMALL_LETTER_L_WITH_MIDDLE_RING = 0xAB39;
const char32_t LATIN_SMALL_LETTER_M_WITH_CROSSED_TAIL = 0xAB3A;
const char32_t LATIN_SMALL_LETTER_N_WITH_CROSSED_TAIL = 0xAB3B;
const char32_t LATIN_SMALL_LETTER_ENG_WITH_CROSSED_TAIL = 0xAB3C;
const char32_t LATIN_SMALL_LETTER_BLACKLETTER_O = 0xAB3D;
const char32_t LATIN_SMALL_LETTER_BLACKLETTER_O_WITH_STROKE = 0xAB3E;
const char32_t LATIN_SMALL_LETTER_OPEN_O_WITH_STROKE = 0xAB3F;
const char32_t LATIN_SMALL_LETTER_INVERTED_OE = 0xAB40;
const char32_t LATIN_SMALL_LETTER_TURNED_OE_WITH_STROKE = 0xAB41;
const char32_t LATIN_SMALL_LETTER_TURNED_OE_WITH_HORIZONTAL_STROKE = 0xAB42;
const char32_t LATIN_SMALL_LETTER_TURNED_O_OPEN_O = 0xAB43;
const char32_t LATIN_SMALL_LETTER_TURNED_O_OPEN_O_WITH_STROKE = 0xAB44;
const char32_t LATIN_SMALL_LETTER_STIRRUP_R = 0xAB45;
const char32_t LATIN_LETTER_SMALL_CAPITAL_R_WITH_RIGHT_LEG = 0xAB46;
const char32_t LATIN_SMALL_LETTER_R_WITHOUT_HANDLE = 0xAB47;
const char32_t LATIN_SMALL_LETTER_DOUBLE_R = 0xAB48;
const char32_t LATIN_SMALL_LETTER_R_WITH_CROSSED_TAIL = 0xAB49;
const char32_t LATIN_SMALL_LETTER_DOUBLE_R_WITH_CROSSED_TAIL = 0xAB4A;
const char32_t LATIN_SMALL_LETTER_SCRIPT_R = 0xAB4B;
const char32_t LATIN_SMALL_LETTER_SCRIPT_R_WITH_RING = 0xAB4C;
const char32_t LATIN_SMALL_LETTER_BASELINE_ESH = 0xAB4D;
const char32_t LATIN_SMALL_LETTER_U_WITH_SHORT_RIGHT_LEG = 0xAB4E;
const char32_t LATIN_SMALL_LETTER_U_BAR_WITH_SHORT_RIGHT_LEG = 0xAB4F;
const char32_t LATIN_SMALL_LETTER_UI = 0xAB50;
const char32_t LATIN_SMALL_LETTER_TURNED_UI = 0xAB51;
const char32_t LATIN_SMALL_LETTER_U_WITH_LEFT_HOOK = 0xAB52;
const char32_t LATIN_SMALL_LETTER_CHI = 0xAB53;
const char32_t LATIN_SMALL_LETTER_CHI_WITH_LOW_RIGHT_RING = 0xAB54;
const char32_t LATIN_SMALL_LETTER_CHI_WITH_LOW_LEFT_SERIF = 0xAB55;
const char32_t LATIN_SMALL_LETTER_X_WITH_LOW_RIGHT_RING = 0xAB56;
const char32_t LATIN_SMALL_LETTER_X_WITH_LONG_LEFT_LEG = 0xAB57;
const char32_t LATIN_SMALL_LETTER_X_WITH_LONG_LEFT_LEG_AND_LOW_RIGHT_RING = 0xAB58;
const char32_t LATIN_SMALL_LETTER_X_WITH_LONG_LEFT_LEG_WITH_SERIF = 0xAB59;
const char32_t LATIN_SMALL_LETTER_Y_WITH_SHORT_RIGHT_LEG = 0xAB5A;
const char32_t MODIFIER_BREVE_WITH_INVERTED_BREVE = 0xAB5B;
const char32_t MODIFIER_LETTER_SMALL_HENG = 0xAB5C;
const char32_t MODIFIER_LETTER_SMALL_L_WITH_INVERTED_LAZY_S = 0xAB5D;
const char32_t MODIFIER_LETTER_SMALL_L_WITH_MIDDLE_TILDE = 0xAB5E;
const char32_t MODIFIER_LETTER_SMALL_U_WITH_LEFT_HOOK = 0xAB5F;
const char32_t LATIN_SMALL_LETTER_SAKHA_YAT = 0xAB60;
const char32_t LATIN_SMALL_LETTER_IOTIFIED_E = 0xAB61;
const char32_t LATIN_SMALL_LETTER_OPEN_OE = 0xAB62;
const char32_t LATIN_SMALL_LETTER_UO = 0xAB63;
const char32_t LATIN_SMALL_LETTER_INVERTED_ALPHA = 0xAB64;
const char32_t GREEK_LETTER_SMALL_CAPITAL_OMEGA = 0xAB65;
const char32_t LATIN_SMALL_LETTER_DZ_DIGRAPH_WITH_RETROFLEX_HOOK = 0xAB66;
const char32_t LATIN_SMALL_LETTER_TS_DIGRAPH_WITH_RETROFLEX_HOOK = 0xAB67;
const char32_t LATIN_SMALL_LETTER_TURNED_R_WITH_MIDDLE_TILDE = 0xAB68;
const char32_t MODIFIER_LETTER_SMALL_TURNED_W = 0xAB69;
const char32_t MODIFIER_LETTER_LEFT_TACK = 0xAB6A;
const char32_t MODIFIER_LETTER_RIGHT_TACK = 0xAB6B;
const char32_t CHEROKEE_SMALL_LETTER_A = 0xAB70;
const char32_t CHEROKEE_SMALL_LETTER_E = 0xAB71;
const char32_t CHEROKEE_SMALL_LETTER_I = 0xAB72;
const char32_t CHEROKEE_SMALL_LETTER_O = 0xAB73;
const char32_t CHEROKEE_SMALL_LETTER_U = 0xAB74;
const char32_t CHEROKEE_SMALL_LETTER_V = 0xAB75;
const char32_t CHEROKEE_SMALL_LETTER_GA = 0xAB76;
const char32_t CHEROKEE_SMALL_LETTER_KA = 0xAB77;
const char32_t CHEROKEE_SMALL_LETTER_GE = 0xAB78;
const char32_t CHEROKEE_SMALL_LETTER_GI = 0xAB79;
const char32_t CHEROKEE_SMALL_LETTER_GO = 0xAB7A;
const char32_t CHEROKEE_SMALL_LETTER_GU = 0xAB7B;
const char32_t CHEROKEE_SMALL_LETTER_GV = 0xAB7C;
const char32_t CHEROKEE_SMALL_LETTER_HA = 0xAB7D;
const char32_t CHEROKEE_SMALL_LETTER_HE = 0xAB7E;
const char32_t CHEROKEE_SMALL_LETTER_HI = 0xAB7F;
const char32_t CHEROKEE_SMALL_LETTER_HO = 0xAB80;
const char32_t CHEROKEE_SMALL_LETTER_HU = 0xAB81;
const char32_t CHEROKEE_SMALL_LETTER_HV = 0xAB82;
const char32_t CHEROKEE_SMALL_LETTER_LA = 0xAB83;
const char32_t CHEROKEE_SMALL_LETTER_LE = 0xAB84;
const char32_t CHEROKEE_SMALL_LETTER_LI = 0xAB85;
const char32_t CHEROKEE_SMALL_LETTER_LO = 0xAB86;
const char32_t CHEROKEE_SMALL_LETTER_LU = 0xAB87;
const char32_t CHEROKEE_SMALL_LETTER_LV = 0xAB88;
const char32_t CHEROKEE_SMALL_LETTER_MA = 0xAB89;
const char32_t CHEROKEE_SMALL_LETTER_ME = 0xAB8A;
const char32_t CHEROKEE_SMALL_LETTER_MI = 0xAB8B;
const char32_t CHEROKEE_SMALL_LETTER_MO = 0xAB8C;
const char32_t CHEROKEE_SMALL_LETTER_MU = 0xAB8D;
const char32_t CHEROKEE_SMALL_LETTER_NA = 0xAB8E;
const char32_t CHEROKEE_SMALL_LETTER_HNA = 0xAB8F;
const char32_t CHEROKEE_SMALL_LETTER_NAH = 0xAB90;
const char32_t CHEROKEE_SMALL_LETTER_NE = 0xAB91;
const char32_t CHEROKEE_SMALL_LETTER_NI = 0xAB92;
const char32_t CHEROKEE_SMALL_LETTER_NO = 0xAB93;
const char32_t CHEROKEE_SMALL_LETTER_NU = 0xAB94;
const char32_t CHEROKEE_SMALL_LETTER_NV = 0xAB95;
const char32_t CHEROKEE_SMALL_LETTER_QUA = 0xAB96;
const char32_t CHEROKEE_SMALL_LETTER_QUE = 0xAB97;
const char32_t CHEROKEE_SMALL_LETTER_QUI = 0xAB98;
const char32_t CHEROKEE_SMALL_LETTER_QUO = 0xAB99;
const char32_t CHEROKEE_SMALL_LETTER_QUU = 0xAB9A;
const char32_t CHEROKEE_SMALL_LETTER_QUV = 0xAB9B;
const char32_t CHEROKEE_SMALL_LETTER_SA = 0xAB9C;
const char32_t CHEROKEE_SMALL_LETTER_S = 0xAB9D;
const char32_t CHEROKEE_SMALL_LETTER_SE = 0xAB9E;
const char32_t CHEROKEE_SMALL_LETTER_SI = 0xAB9F;
const char32_t CHEROKEE_SMALL_LETTER_SO = 0xABA0;
const char32_t CHEROKEE_SMALL_LETTER_SU = 0xABA1;
const char32_t CHEROKEE_SMALL_LETTER_SV = 0xABA2;
const char32_t CHEROKEE_SMALL_LETTER_DA = 0xABA3;
const char32_t CHEROKEE_SMALL_LETTER_TA = 0xABA4;
const char32_t CHEROKEE_SMALL_LETTER_DE = 0xABA5;
const char32_t CHEROKEE_SMALL_LETTER_TE = 0xABA6;
const char32_t CHEROKEE_SMALL_LETTER_DI = 0xABA7;
const char32_t CHEROKEE_SMALL_LETTER_TI = 0xABA8;
const char32_t CHEROKEE_SMALL_LETTER_DO = 0xABA9;
const char32_t CHEROKEE_SMALL_LETTER_DU = 0xABAA;
const char32_t CHEROKEE_SMALL_LETTER_DV = 0xABAB;
const char32_t CHEROKEE_SMALL_LETTER_DLA = 0xABAC;
const char32_t CHEROKEE_SMALL_LETTER_TLA = 0xABAD;
const char32_t CHEROKEE_SMALL_LETTER_TLE = 0xABAE;
const char32_t CHEROKEE_SMALL_LETTER_TLI = 0xABAF;
const char32_t CHEROKEE_SMALL_LETTER_TLO = 0xABB0;
const char32_t CHEROKEE_SMALL_LETTER_TLU = 0xABB1;
const char32_t CHEROKEE_SMALL_LETTER_TLV = 0xABB2;
const char32_t CHEROKEE_SMALL_LETTER_TSA = 0xABB3;
const char32_t CHEROKEE_SMALL_LETTER_TSE = 0xABB4;
const char32_t CHEROKEE_SMALL_LETTER_TSI = 0xABB5;
const char32_t CHEROKEE_SMALL_LETTER_TSO = 0xABB6;
const char32_t CHEROKEE_SMALL_LETTER_TSU = 0xABB7;
const char32_t CHEROKEE_SMALL_LETTER_TSV = 0xABB8;
const char32_t CHEROKEE_SMALL_LETTER_WA = 0xABB9;
const char32_t CHEROKEE_SMALL_LETTER_WE = 0xABBA;
const char32_t CHEROKEE_SMALL_LETTER_WI = 0xABBB;
const char32_t CHEROKEE_SMALL_LETTER_WO = 0xABBC;
const char32_t CHEROKEE_SMALL_LETTER_WU = 0xABBD;
const char32_t CHEROKEE_SMALL_LETTER_WV = 0xABBE;
const char32_t CHEROKEE_SMALL_LETTER_YA = 0xABBF;
const char32_t MEETEI_MAYEK_LETTER_KOK = 0xABC0;
const char32_t MEETEI_MAYEK_LETTER_SAM = 0xABC1;
const char32_t MEETEI_MAYEK_LETTER_LAI = 0xABC2;
const char32_t MEETEI_MAYEK_LETTER_MIT = 0xABC3;
const char32_t MEETEI_MAYEK_LETTER_PA = 0xABC4;
const char32_t MEETEI_MAYEK_LETTER_NA = 0xABC5;
const char32_t MEETEI_MAYEK_LETTER_CHIL = 0xABC6;
const char32_t MEETEI_MAYEK_LETTER_TIL = 0xABC7;
const char32_t MEETEI_MAYEK_LETTER_KHOU = 0xABC8;
const char32_t MEETEI_MAYEK_LETTER_NGOU = 0xABC9;
const char32_t MEETEI_MAYEK_LETTER_THOU = 0xABCA;
const char32_t MEETEI_MAYEK_LETTER_WAI = 0xABCB;
const char32_t MEETEI_MAYEK_LETTER_YANG = 0xABCC;
const char32_t MEETEI_MAYEK_LETTER_HUK = 0xABCD;
const char32_t MEETEI_MAYEK_LETTER_UN = 0xABCE;
const char32_t MEETEI_MAYEK_LETTER_I = 0xABCF;
const char32_t MEETEI_MAYEK_LETTER_PHAM = 0xABD0;
const char32_t MEETEI_MAYEK_LETTER_ATIYA = 0xABD1;
const char32_t MEETEI_MAYEK_LETTER_GOK = 0xABD2;
const char32_t MEETEI_MAYEK_LETTER_JHAM = 0xABD3;
const char32_t MEETEI_MAYEK_LETTER_RAI = 0xABD4;
const char32_t MEETEI_MAYEK_LETTER_BA = 0xABD5;
const char32_t MEETEI_MAYEK_LETTER_JIL = 0xABD6;
const char32_t MEETEI_MAYEK_LETTER_DIL = 0xABD7;
const char32_t MEETEI_MAYEK_LETTER_GHOU = 0xABD8;
const char32_t MEETEI_MAYEK_LETTER_DHOU = 0xABD9;
const char32_t MEETEI_MAYEK_LETTER_BHAM = 0xABDA;
const char32_t MEETEI_MAYEK_LETTER_KOK_LONSUM = 0xABDB;
const char32_t MEETEI_MAYEK_LETTER_LAI_LONSUM = 0xABDC;
const char32_t MEETEI_MAYEK_LETTER_MIT_LONSUM = 0xABDD;
const char32_t MEETEI_MAYEK_LETTER_PA_LONSUM = 0xABDE;
const char32_t MEETEI_MAYEK_LETTER_NA_LONSUM = 0xABDF;
const char32_t MEETEI_MAYEK_LETTER_TIL_LONSUM = 0xABE0;
const char32_t MEETEI_MAYEK_LETTER_NGOU_LONSUM = 0xABE1;
const char32_t MEETEI_MAYEK_LETTER_I_LONSUM = 0xABE2;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_ONAP = 0xABE3;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_INAP = 0xABE4;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_ANAP = 0xABE5;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_YENAP = 0xABE6;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_SOUNAP = 0xABE7;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_UNAP = 0xABE8;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_CHEINAP = 0xABE9;
const char32_t MEETEI_MAYEK_VOWEL_SIGN_NUNG = 0xABEA;
const char32_t MEETEI_MAYEK_CHEIKHEI = 0xABEB;
const char32_t MEETEI_MAYEK_LUM_IYEK = 0xABEC;
const char32_t MEETEI_MAYEK_APUN_IYEK = 0xABED;
const char32_t MEETEI_MAYEK_DIGIT_ZERO = 0xABF0;
const char32_t MEETEI_MAYEK_DIGIT_ONE = 0xABF1;
const char32_t MEETEI_MAYEK_DIGIT_TWO = 0xABF2;
const char32_t MEETEI_MAYEK_DIGIT_THREE = 0xABF3;
const char32_t MEETEI_MAYEK_DIGIT_FOUR = 0xABF4;
const char32_t MEETEI_MAYEK_DIGIT_FIVE = 0xABF5;
const char32_t MEETEI_MAYEK_DIGIT_SIX = 0xABF6;
const char32_t MEETEI_MAYEK_DIGIT_SEVEN = 0xABF7;
const char32_t MEETEI_MAYEK_DIGIT_EIGHT = 0xABF8;
const char32_t MEETEI_MAYEK_DIGIT_NINE = 0xABF9;
const char32_t HANGUL_SYLLABLE_FIRST = 0xAC00;
const char32_t HANGUL_SYLLABLE_LAST = 0xD7A3;
const char32_t HANGUL_JUNGSEONG_O_YEO = 0xD7B0;
const char32_t HANGUL_JUNGSEONG_O_O_I = 0xD7B1;
const char32_t HANGUL_JUNGSEONG_YO_A = 0xD7B2;
const char32_t HANGUL_JUNGSEONG_YO_AE = 0xD7B3;
const char32_t HANGUL_JUNGSEONG_YO_EO = 0xD7B4;
const char32_t HANGUL_JUNGSEONG_U_YEO = 0xD7B5;
const char32_t HANGUL_JUNGSEONG_U_I_I = 0xD7B6;
const char32_t HANGUL_JUNGSEONG_YU_AE = 0xD7B7;
const char32_t HANGUL_JUNGSEONG_YU_O = 0xD7B8;
const char32_t HANGUL_JUNGSEONG_EU_A = 0xD7B9;
const char32_t HANGUL_JUNGSEONG_EU_EO = 0xD7BA;
const char32_t HANGUL_JUNGSEONG_EU_E = 0xD7BB;
const char32_t HANGUL_JUNGSEONG_EU_O = 0xD7BC;
const char32_t HANGUL_JUNGSEONG_I_YA_O = 0xD7BD;
const char32_t HANGUL_JUNGSEONG_I_YAE = 0xD7BE;
const char32_t HANGUL_JUNGSEONG_I_YEO = 0xD7BF;
const char32_t HANGUL_JUNGSEONG_I_YE = 0xD7C0;
const char32_t HANGUL_JUNGSEONG_I_O_I = 0xD7C1;
const char32_t HANGUL_JUNGSEONG_I_YO = 0xD7C2;
const char32_t HANGUL_JUNGSEONG_I_YU = 0xD7C3;
const char32_t HANGUL_JUNGSEONG_I_I = 0xD7C4;
const char32_t HANGUL_JUNGSEONG_ARAEA_A = 0xD7C5;
const char32_t HANGUL_JUNGSEONG_ARAEA_E = 0xD7C6;
const char32_t HANGUL_JONGSEONG_NIEUN_RIEUL = 0xD7CB;
const char32_t HANGUL_JONGSEONG_NIEUN_CHIEUCH = 0xD7CC;
const char32_t HANGUL_JONGSEONG_SSANGTIKEUT = 0xD7CD;
const char32_t HANGUL_JONGSEONG_SSANGTIKEUT_PIEUP = 0xD7CE;
const char32_t HANGUL_JONGSEONG_TIKEUT_PIEUP = 0xD7CF;
const char32_t HANGUL_JONGSEONG_TIKEUT_SIOS = 0xD7D0;
const char32_t HANGUL_JONGSEONG_TIKEUT_SIOS_KIYEOK = 0xD7D1;
const char32_t HANGUL_JONGSEONG_TIKEUT_CIEUC = 0xD7D2;
const char32_t HANGUL_JONGSEONG_TIKEUT_CHIEUCH = 0xD7D3;
const char32_t HANGUL_JONGSEONG_TIKEUT_THIEUTH = 0xD7D4;
const char32_t HANGUL_JONGSEONG_RIEUL_SSANGKIYEOK = 0xD7D5;
const char32_t HANGUL_JONGSEONG_RIEUL_KIYEOK_HIEUH = 0xD7D6;
const char32_t HANGUL_JONGSEONG_SSANGRIEUL_KHIEUKH = 0xD7D7;
const char32_t HANGUL_JONGSEONG_RIEUL_MIEUM_HIEUH = 0xD7D8;
const char32_t HANGUL_JONGSEONG_RIEUL_PIEUP_TIKEUT = 0xD7D9;
const char32_t HANGUL_JONGSEONG_RIEUL_PIEUP_PHIEUPH = 0xD7DA;
const char32_t HANGUL_JONGSEONG_RIEUL_YESIEUNG = 0xD7DB;
const char32_t HANGUL_JONGSEONG_RIEUL_YEORINHIEUH_HIEUH = 0xD7DC;
const char32_t HANGUL_JONGSEONG_KAPYEOUNRIEUL = 0xD7DD;
const char32_t HANGUL_JONGSEONG_MIEUM_NIEUN = 0xD7DE;
const char32_t HANGUL_JONGSEONG_MIEUM_SSANGNIEUN = 0xD7DF;
const char32_t HANGUL_JONGSEONG_SSANGMIEUM = 0xD7E0;
const char32_t HANGUL_JONGSEONG_MIEUM_PIEUP_SIOS = 0xD7E1;
const char32_t HANGUL_JONGSEONG_MIEUM_CIEUC = 0xD7E2;
const char32_t HANGUL_JONGSEONG_PIEUP_TIKEUT = 0xD7E3;
const char32_t HANGUL_JONGSEONG_PIEUP_RIEUL_PHIEUPH = 0xD7E4;
const char32_t HANGUL_JONGSEONG_PIEUP_MIEUM = 0xD7E5;
const char32_t HANGUL_JONGSEONG_SSANGPIEUP = 0xD7E6;
const char32_t HANGUL_JONGSEONG_PIEUP_SIOS_TIKEUT = 0xD7E7;
const char32_t HANGUL_JONGSEONG_PIEUP_CIEUC = 0xD7E8;
const char32_t HANGUL_JONGSEONG_PIEUP_CHIEUCH = 0xD7E9;
const char32_t HANGUL_JONGSEONG_SIOS_MIEUM = 0xD7EA;
const char32_t HANGUL_JONGSEONG_SIOS_KAPYEOUNPIEUP = 0xD7EB;
const char32_t HANGUL_JONGSEONG_SSANGSIOS_KIYEOK = 0xD7EC;
const char32_t HANGUL_JONGSEONG_SSANGSIOS_TIKEUT = 0xD7ED;
const char32_t HANGUL_JONGSEONG_SIOS_PANSIOS = 0xD7EE;
const char32_t HANGUL_JONGSEONG_SIOS_CIEUC = 0xD7EF;
const char32_t HANGUL_JONGSEONG_SIOS_CHIEUCH = 0xD7F0;
const char32_t HANGUL_JONGSEONG_SIOS_THIEUTH = 0xD7F1;
const char32_t HANGUL_JONGSEONG_SIOS_HIEUH = 0xD7F2;
const char32_t HANGUL_JONGSEONG_PANSIOS_PIEUP = 0xD7F3;
const char32_t HANGUL_JONGSEONG_PANSIOS_KAPYEOUNPIEUP = 0xD7F4;
const char32_t HANGUL_JONGSEONG_YESIEUNG_MIEUM = 0xD7F5;
const char32_t HANGUL_JONGSEONG_YESIEUNG_HIEUH = 0xD7F6;
const char32_t HANGUL_JONGSEONG_CIEUC_PIEUP = 0xD7F7;
const char32_t HANGUL_JONGSEONG_CIEUC_SSANGPIEUP = 0xD7F8;
const char32_t HANGUL_JONGSEONG_SSANGCIEUC = 0xD7F9;
const char32_t HANGUL_JONGSEONG_PHIEUPH_SIOS = 0xD7FA;
const char32_t HANGUL_JONGSEONG_PHIEUPH_THIEUTH = 0xD7FB;
const char32_t NON_PRIVATE_USE_HIGH_SURROGATE_FIRST = 0xD800;
const char32_t NON_PRIVATE_USE_HIGH_SURROGATE_LAST = 0xDB7F;
const char32_t PRIVATE_USE_HIGH_SURROGATE_FIRST = 0xDB80;
const char32_t PRIVATE_USE_HIGH_SURROGATE_LAST = 0xDBFF;
const char32_t LOW_SURROGATE_FIRST = 0xDC00;
const char32_t LOW_SURROGATE_LAST = 0xDFFF;
const char32_t PRIVATE_USE_FIRST = 0xE000;
const char32_t PRIVATE_USE_LAST = 0xF8FF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F900 = 0xF900;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F901 = 0xF901;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F902 = 0xF902;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F903 = 0xF903;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F904 = 0xF904;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F905 = 0xF905;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F906 = 0xF906;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F907 = 0xF907;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F908 = 0xF908;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F909 = 0xF909;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F90A = 0xF90A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F90B = 0xF90B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F90C = 0xF90C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F90D = 0xF90D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F90E = 0xF90E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F90F = 0xF90F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F910 = 0xF910;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F911 = 0xF911;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F912 = 0xF912;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F913 = 0xF913;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F914 = 0xF914;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F915 = 0xF915;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F916 = 0xF916;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F917 = 0xF917;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F918 = 0xF918;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F919 = 0xF919;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F91A = 0xF91A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F91B = 0xF91B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F91C = 0xF91C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F91D = 0xF91D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F91E = 0xF91E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F91F = 0xF91F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F920 = 0xF920;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F921 = 0xF921;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F922 = 0xF922;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F923 = 0xF923;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F924 = 0xF924;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F925 = 0xF925;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F926 = 0xF926;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F927 = 0xF927;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F928 = 0xF928;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F929 = 0xF929;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F92A = 0xF92A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F92B = 0xF92B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F92C = 0xF92C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F92D = 0xF92D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F92E = 0xF92E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F92F = 0xF92F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F930 = 0xF930;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F931 = 0xF931;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F932 = 0xF932;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F933 = 0xF933;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F934 = 0xF934;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F935 = 0xF935;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F936 = 0xF936;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F937 = 0xF937;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F938 = 0xF938;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F939 = 0xF939;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F93A = 0xF93A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F93B = 0xF93B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F93C = 0xF93C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F93D = 0xF93D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F93E = 0xF93E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F93F = 0xF93F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F940 = 0xF940;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F941 = 0xF941;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F942 = 0xF942;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F943 = 0xF943;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F944 = 0xF944;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F945 = 0xF945;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F946 = 0xF946;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F947 = 0xF947;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F948 = 0xF948;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F949 = 0xF949;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F94A = 0xF94A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F94B = 0xF94B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F94C = 0xF94C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F94D = 0xF94D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F94E = 0xF94E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F94F = 0xF94F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F950 = 0xF950;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F951 = 0xF951;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F952 = 0xF952;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F953 = 0xF953;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F954 = 0xF954;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F955 = 0xF955;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F956 = 0xF956;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F957 = 0xF957;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F958 = 0xF958;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F959 = 0xF959;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F95A = 0xF95A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F95B = 0xF95B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F95C = 0xF95C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F95D = 0xF95D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F95E = 0xF95E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F95F = 0xF95F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F960 = 0xF960;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F961 = 0xF961;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F962 = 0xF962;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F963 = 0xF963;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F964 = 0xF964;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F965 = 0xF965;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F966 = 0xF966;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F967 = 0xF967;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F968 = 0xF968;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F969 = 0xF969;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F96A = 0xF96A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F96B = 0xF96B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F96C = 0xF96C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F96D = 0xF96D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F96E = 0xF96E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F96F = 0xF96F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F970 = 0xF970;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F971 = 0xF971;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F972 = 0xF972;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F973 = 0xF973;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F974 = 0xF974;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F975 = 0xF975;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F976 = 0xF976;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F977 = 0xF977;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F978 = 0xF978;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F979 = 0xF979;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F97A = 0xF97A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F97B = 0xF97B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F97C = 0xF97C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F97D = 0xF97D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F97E = 0xF97E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F97F = 0xF97F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F980 = 0xF980;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F981 = 0xF981;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F982 = 0xF982;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F983 = 0xF983;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F984 = 0xF984;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F985 = 0xF985;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F986 = 0xF986;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F987 = 0xF987;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F988 = 0xF988;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F989 = 0xF989;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F98A = 0xF98A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F98B = 0xF98B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F98C = 0xF98C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F98D = 0xF98D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F98E = 0xF98E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F98F = 0xF98F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F990 = 0xF990;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F991 = 0xF991;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F992 = 0xF992;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F993 = 0xF993;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F994 = 0xF994;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F995 = 0xF995;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F996 = 0xF996;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F997 = 0xF997;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F998 = 0xF998;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F999 = 0xF999;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F99A = 0xF99A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F99B = 0xF99B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F99C = 0xF99C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F99D = 0xF99D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F99E = 0xF99E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F99F = 0xF99F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A0 = 0xF9A0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A1 = 0xF9A1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A2 = 0xF9A2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A3 = 0xF9A3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A4 = 0xF9A4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A5 = 0xF9A5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A6 = 0xF9A6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A7 = 0xF9A7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A8 = 0xF9A8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9A9 = 0xF9A9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9AA = 0xF9AA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9AB = 0xF9AB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9AC = 0xF9AC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9AD = 0xF9AD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9AE = 0xF9AE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9AF = 0xF9AF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B0 = 0xF9B0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B1 = 0xF9B1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B2 = 0xF9B2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B3 = 0xF9B3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B4 = 0xF9B4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B5 = 0xF9B5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B6 = 0xF9B6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B7 = 0xF9B7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B8 = 0xF9B8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9B9 = 0xF9B9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9BA = 0xF9BA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9BB = 0xF9BB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9BC = 0xF9BC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9BD = 0xF9BD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9BE = 0xF9BE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9BF = 0xF9BF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C0 = 0xF9C0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C1 = 0xF9C1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C2 = 0xF9C2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C3 = 0xF9C3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C4 = 0xF9C4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C5 = 0xF9C5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C6 = 0xF9C6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C7 = 0xF9C7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C8 = 0xF9C8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9C9 = 0xF9C9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9CA = 0xF9CA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9CB = 0xF9CB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9CC = 0xF9CC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9CD = 0xF9CD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9CE = 0xF9CE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9CF = 0xF9CF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D0 = 0xF9D0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D1 = 0xF9D1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D2 = 0xF9D2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D3 = 0xF9D3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D4 = 0xF9D4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D5 = 0xF9D5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D6 = 0xF9D6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D7 = 0xF9D7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D8 = 0xF9D8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9D9 = 0xF9D9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9DA = 0xF9DA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9DB = 0xF9DB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9DC = 0xF9DC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9DD = 0xF9DD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9DE = 0xF9DE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9DF = 0xF9DF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E0 = 0xF9E0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E1 = 0xF9E1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E2 = 0xF9E2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E3 = 0xF9E3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E4 = 0xF9E4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E5 = 0xF9E5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E6 = 0xF9E6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E7 = 0xF9E7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E8 = 0xF9E8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9E9 = 0xF9E9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9EA = 0xF9EA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9EB = 0xF9EB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9EC = 0xF9EC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9ED = 0xF9ED;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9EE = 0xF9EE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9EF = 0xF9EF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F0 = 0xF9F0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F1 = 0xF9F1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F2 = 0xF9F2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F3 = 0xF9F3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F4 = 0xF9F4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F5 = 0xF9F5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F6 = 0xF9F6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F7 = 0xF9F7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F8 = 0xF9F8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9F9 = 0xF9F9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9FA = 0xF9FA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9FB = 0xF9FB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9FC = 0xF9FC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9FD = 0xF9FD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9FE = 0xF9FE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_F9FF = 0xF9FF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA00 = 0xFA00;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA01 = 0xFA01;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA02 = 0xFA02;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA03 = 0xFA03;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA04 = 0xFA04;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA05 = 0xFA05;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA06 = 0xFA06;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA07 = 0xFA07;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA08 = 0xFA08;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA09 = 0xFA09;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA0A = 0xFA0A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA0B = 0xFA0B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA0C = 0xFA0C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA0D = 0xFA0D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA0E = 0xFA0E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA0F = 0xFA0F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA10 = 0xFA10;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA11 = 0xFA11;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA12 = 0xFA12;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA13 = 0xFA13;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA14 = 0xFA14;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA15 = 0xFA15;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA16 = 0xFA16;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA17 = 0xFA17;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA18 = 0xFA18;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA19 = 0xFA19;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA1A = 0xFA1A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA1B = 0xFA1B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA1C = 0xFA1C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA1D = 0xFA1D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA1E = 0xFA1E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA1F = 0xFA1F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA20 = 0xFA20;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA21 = 0xFA21;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA22 = 0xFA22;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA23 = 0xFA23;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA24 = 0xFA24;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA25 = 0xFA25;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA26 = 0xFA26;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA27 = 0xFA27;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA28 = 0xFA28;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA29 = 0xFA29;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA2A = 0xFA2A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA2B = 0xFA2B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA2C = 0xFA2C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA2D = 0xFA2D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA2E = 0xFA2E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA2F = 0xFA2F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA30 = 0xFA30;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA31 = 0xFA31;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA32 = 0xFA32;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA33 = 0xFA33;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA34 = 0xFA34;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA35 = 0xFA35;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA36 = 0xFA36;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA37 = 0xFA37;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA38 = 0xFA38;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA39 = 0xFA39;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA3A = 0xFA3A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA3B = 0xFA3B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA3C = 0xFA3C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA3D = 0xFA3D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA3E = 0xFA3E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA3F = 0xFA3F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA40 = 0xFA40;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA41 = 0xFA41;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA42 = 0xFA42;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA43 = 0xFA43;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA44 = 0xFA44;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA45 = 0xFA45;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA46 = 0xFA46;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA47 = 0xFA47;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA48 = 0xFA48;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA49 = 0xFA49;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA4A = 0xFA4A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA4B = 0xFA4B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA4C = 0xFA4C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA4D = 0xFA4D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA4E = 0xFA4E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA4F = 0xFA4F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA50 = 0xFA50;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA51 = 0xFA51;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA52 = 0xFA52;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA53 = 0xFA53;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA54 = 0xFA54;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA55 = 0xFA55;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA56 = 0xFA56;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA57 = 0xFA57;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA58 = 0xFA58;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA59 = 0xFA59;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA5A = 0xFA5A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA5B = 0xFA5B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA5C = 0xFA5C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA5D = 0xFA5D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA5E = 0xFA5E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA5F = 0xFA5F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA60 = 0xFA60;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA61 = 0xFA61;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA62 = 0xFA62;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA63 = 0xFA63;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA64 = 0xFA64;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA65 = 0xFA65;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA66 = 0xFA66;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA67 = 0xFA67;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA68 = 0xFA68;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA69 = 0xFA69;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA6A = 0xFA6A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA6B = 0xFA6B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA6C = 0xFA6C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA6D = 0xFA6D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA70 = 0xFA70;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA71 = 0xFA71;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA72 = 0xFA72;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA73 = 0xFA73;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA74 = 0xFA74;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA75 = 0xFA75;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA76 = 0xFA76;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA77 = 0xFA77;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA78 = 0xFA78;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA79 = 0xFA79;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA7A = 0xFA7A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA7B = 0xFA7B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA7C = 0xFA7C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA7D = 0xFA7D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA7E = 0xFA7E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA7F = 0xFA7F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA80 = 0xFA80;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA81 = 0xFA81;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA82 = 0xFA82;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA83 = 0xFA83;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA84 = 0xFA84;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA85 = 0xFA85;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA86 = 0xFA86;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA87 = 0xFA87;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA88 = 0xFA88;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA89 = 0xFA89;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA8A = 0xFA8A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA8B = 0xFA8B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA8C = 0xFA8C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA8D = 0xFA8D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA8E = 0xFA8E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA8F = 0xFA8F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA90 = 0xFA90;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA91 = 0xFA91;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA92 = 0xFA92;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA93 = 0xFA93;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA94 = 0xFA94;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA95 = 0xFA95;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA96 = 0xFA96;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA97 = 0xFA97;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA98 = 0xFA98;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA99 = 0xFA99;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA9A = 0xFA9A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA9B = 0xFA9B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA9C = 0xFA9C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA9D = 0xFA9D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA9E = 0xFA9E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FA9F = 0xFA9F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA0 = 0xFAA0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA1 = 0xFAA1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA2 = 0xFAA2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA3 = 0xFAA3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA4 = 0xFAA4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA5 = 0xFAA5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA6 = 0xFAA6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA7 = 0xFAA7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA8 = 0xFAA8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAA9 = 0xFAA9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAAA = 0xFAAA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAAB = 0xFAAB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAAC = 0xFAAC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAAD = 0xFAAD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAAE = 0xFAAE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAAF = 0xFAAF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB0 = 0xFAB0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB1 = 0xFAB1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB2 = 0xFAB2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB3 = 0xFAB3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB4 = 0xFAB4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB5 = 0xFAB5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB6 = 0xFAB6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB7 = 0xFAB7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB8 = 0xFAB8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAB9 = 0xFAB9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FABA = 0xFABA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FABB = 0xFABB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FABC = 0xFABC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FABD = 0xFABD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FABE = 0xFABE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FABF = 0xFABF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC0 = 0xFAC0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC1 = 0xFAC1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC2 = 0xFAC2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC3 = 0xFAC3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC4 = 0xFAC4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC5 = 0xFAC5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC6 = 0xFAC6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC7 = 0xFAC7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC8 = 0xFAC8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAC9 = 0xFAC9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FACA = 0xFACA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FACB = 0xFACB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FACC = 0xFACC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FACD = 0xFACD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FACE = 0xFACE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FACF = 0xFACF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD0 = 0xFAD0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD1 = 0xFAD1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD2 = 0xFAD2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD3 = 0xFAD3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD4 = 0xFAD4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD5 = 0xFAD5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD6 = 0xFAD6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD7 = 0xFAD7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD8 = 0xFAD8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_FAD9 = 0xFAD9;
const char32_t LATIN_SMALL_LIGATURE_FF = 0xFB00;
const char32_t LATIN_SMALL_LIGATURE_FI = 0xFB01;
const char32_t LATIN_SMALL_LIGATURE_FL = 0xFB02;
const char32_t LATIN_SMALL_LIGATURE_FFI = 0xFB03;
const char32_t LATIN_SMALL_LIGATURE_FFL = 0xFB04;
const char32_t LATIN_SMALL_LIGATURE_LONG_S_T = 0xFB05;
const char32_t LATIN_SMALL_LIGATURE_ST = 0xFB06;
const char32_t ARMENIAN_SMALL_LIGATURE_MEN_NOW = 0xFB13;
const char32_t ARMENIAN_SMALL_LIGATURE_MEN_ECH = 0xFB14;
const char32_t ARMENIAN_SMALL_LIGATURE_MEN_INI = 0xFB15;
const char32_t ARMENIAN_SMALL_LIGATURE_VEW_NOW = 0xFB16;
const char32_t ARMENIAN_SMALL_LIGATURE_MEN_XEH = 0xFB17;
const char32_t HEBREW_LETTER_YOD_WITH_HIRIQ = 0xFB1D;
const char32_t HEBREW_POINT_JUDEO_SPANISH_VARIKA = 0xFB1E;
const char32_t HEBREW_LIGATURE_YIDDISH_YOD_YOD_PATAH = 0xFB1F;
const char32_t HEBREW_LETTER_ALTERNATIVE_AYIN = 0xFB20;
const char32_t HEBREW_LETTER_WIDE_ALEF = 0xFB21;
const char32_t HEBREW_LETTER_WIDE_DALET = 0xFB22;
const char32_t HEBREW_LETTER_WIDE_HE = 0xFB23;
const char32_t HEBREW_LETTER_WIDE_KAF = 0xFB24;
const char32_t HEBREW_LETTER_WIDE_LAMED = 0xFB25;
const char32_t HEBREW_LETTER_WIDE_FINAL_MEM = 0xFB26;
const char32_t HEBREW_LETTER_WIDE_RESH = 0xFB27;
const char32_t HEBREW_LETTER_WIDE_TAV = 0xFB28;
const char32_t HEBREW_LETTER_ALTERNATIVE_PLUS_SIGN = 0xFB29;
const char32_t HEBREW_LETTER_SHIN_WITH_SHIN_DOT = 0xFB2A;
const char32_t HEBREW_LETTER_SHIN_WITH_SIN_DOT = 0xFB2B;
const char32_t HEBREW_LETTER_SHIN_WITH_DAGESH_AND_SHIN_DOT = 0xFB2C;
const char32_t HEBREW_LETTER_SHIN_WITH_DAGESH_AND_SIN_DOT = 0xFB2D;
const char32_t HEBREW_LETTER_ALEF_WITH_PATAH = 0xFB2E;
const char32_t HEBREW_LETTER_ALEF_WITH_QAMATS = 0xFB2F;
const char32_t HEBREW_LETTER_ALEF_WITH_MAPIQ = 0xFB30;
const char32_t HEBREW_LETTER_BET_WITH_DAGESH = 0xFB31;
const char32_t HEBREW_LETTER_GIMEL_WITH_DAGESH = 0xFB32;
const char32_t HEBREW_LETTER_DALET_WITH_DAGESH = 0xFB33;
const char32_t HEBREW_LETTER_HE_WITH_MAPIQ = 0xFB34;
const char32_t HEBREW_LETTER_VAV_WITH_DAGESH = 0xFB35;
const char32_t HEBREW_LETTER_ZAYIN_WITH_DAGESH = 0xFB36;
const char32_t HEBREW_LETTER_TET_WITH_DAGESH = 0xFB38;
const char32_t HEBREW_LETTER_YOD_WITH_DAGESH = 0xFB39;
const char32_t HEBREW_LETTER_FINAL_KAF_WITH_DAGESH = 0xFB3A;
const char32_t HEBREW_LETTER_KAF_WITH_DAGESH = 0xFB3B;
const char32_t HEBREW_LETTER_LAMED_WITH_DAGESH = 0xFB3C;
const char32_t HEBREW_LETTER_MEM_WITH_DAGESH = 0xFB3E;
const char32_t HEBREW_LETTER_NUN_WITH_DAGESH = 0xFB40;
const char32_t HEBREW_LETTER_SAMEKH_WITH_DAGESH = 0xFB41;
const char32_t HEBREW_LETTER_FINAL_PE_WITH_DAGESH = 0xFB43;
const char32_t HEBREW_LETTER_PE_WITH_DAGESH = 0xFB44;
const char32_t HEBREW_LETTER_TSADI_WITH_DAGESH = 0xFB46;
const char32_t HEBREW_LETTER_QOF_WITH_DAGESH = 0xFB47;
const char32_t HEBREW_LETTER_RESH_WITH_DAGESH = 0xFB48;
const char32_t HEBREW_LETTER_SHIN_WITH_DAGESH = 0xFB49;
const char32_t HEBREW_LETTER_TAV_WITH_DAGESH = 0xFB4A;
const char32_t HEBREW_LETTER_VAV_WITH_HOLAM = 0xFB4B;
const char32_t HEBREW_LETTER_BET_WITH_RAFE = 0xFB4C;
const char32_t HEBREW_LETTER_KAF_WITH_RAFE = 0xFB4D;
const char32_t HEBREW_LETTER_PE_WITH_RAFE = 0xFB4E;
const char32_t HEBREW_LIGATURE_ALEF_LAMED = 0xFB4F;
const char32_t ARABIC_LETTER_ALEF_WASLA_ISOLATED_FORM = 0xFB50;
const char32_t ARABIC_LETTER_ALEF_WASLA_FINAL_FORM = 0xFB51;
const char32_t ARABIC_LETTER_BEEH_ISOLATED_FORM = 0xFB52;
const char32_t ARABIC_LETTER_BEEH_FINAL_FORM = 0xFB53;
const char32_t ARABIC_LETTER_BEEH_INITIAL_FORM = 0xFB54;
const char32_t ARABIC_LETTER_BEEH_MEDIAL_FORM = 0xFB55;
const char32_t ARABIC_LETTER_PEH_ISOLATED_FORM = 0xFB56;
const char32_t ARABIC_LETTER_PEH_FINAL_FORM = 0xFB57;
const char32_t ARABIC_LETTER_PEH_INITIAL_FORM = 0xFB58;
const char32_t ARABIC_LETTER_PEH_MEDIAL_FORM = 0xFB59;
const char32_t ARABIC_LETTER_BEHEH_ISOLATED_FORM = 0xFB5A;
const char32_t ARABIC_LETTER_BEHEH_FINAL_FORM = 0xFB5B;
const char32_t ARABIC_LETTER_BEHEH_INITIAL_FORM = 0xFB5C;
const char32_t ARABIC_LETTER_BEHEH_MEDIAL_FORM = 0xFB5D;
const char32_t ARABIC_LETTER_TTEHEH_ISOLATED_FORM = 0xFB5E;
const char32_t ARABIC_LETTER_TTEHEH_FINAL_FORM = 0xFB5F;
const char32_t ARABIC_LETTER_TTEHEH_INITIAL_FORM = 0xFB60;
const char32_t ARABIC_LETTER_TTEHEH_MEDIAL_FORM = 0xFB61;
const char32_t ARABIC_LETTER_TEHEH_ISOLATED_FORM = 0xFB62;
const char32_t ARABIC_LETTER_TEHEH_FINAL_FORM = 0xFB63;
const char32_t ARABIC_LETTER_TEHEH_INITIAL_FORM = 0xFB64;
const char32_t ARABIC_LETTER_TEHEH_MEDIAL_FORM = 0xFB65;
const char32_t ARABIC_LETTER_TTEH_ISOLATED_FORM = 0xFB66;
const char32_t ARABIC_LETTER_TTEH_FINAL_FORM = 0xFB67;
const char32_t ARABIC_LETTER_TTEH_INITIAL_FORM = 0xFB68;
const char32_t ARABIC_LETTER_TTEH_MEDIAL_FORM = 0xFB69;
const char32_t ARABIC_LETTER_VEH_ISOLATED_FORM = 0xFB6A;
const char32_t ARABIC_LETTER_VEH_FINAL_FORM = 0xFB6B;
const char32_t ARABIC_LETTER_VEH_INITIAL_FORM = 0xFB6C;
const char32_t ARABIC_LETTER_VEH_MEDIAL_FORM = 0xFB6D;
const char32_t ARABIC_LETTER_PEHEH_ISOLATED_FORM = 0xFB6E;
const char32_t ARABIC_LETTER_PEHEH_FINAL_FORM = 0xFB6F;
const char32_t ARABIC_LETTER_PEHEH_INITIAL_FORM = 0xFB70;
const char32_t ARABIC_LETTER_PEHEH_MEDIAL_FORM = 0xFB71;
const char32_t ARABIC_LETTER_DYEH_ISOLATED_FORM = 0xFB72;
const char32_t ARABIC_LETTER_DYEH_FINAL_FORM = 0xFB73;
const char32_t ARABIC_LETTER_DYEH_INITIAL_FORM = 0xFB74;
const char32_t ARABIC_LETTER_DYEH_MEDIAL_FORM = 0xFB75;
const char32_t ARABIC_LETTER_NYEH_ISOLATED_FORM = 0xFB76;
const char32_t ARABIC_LETTER_NYEH_FINAL_FORM = 0xFB77;
const char32_t ARABIC_LETTER_NYEH_INITIAL_FORM = 0xFB78;
const char32_t ARABIC_LETTER_NYEH_MEDIAL_FORM = 0xFB79;
const char32_t ARABIC_LETTER_TCHEH_ISOLATED_FORM = 0xFB7A;
const char32_t ARABIC_LETTER_TCHEH_FINAL_FORM = 0xFB7B;
const char32_t ARABIC_LETTER_TCHEH_INITIAL_FORM = 0xFB7C;
const char32_t ARABIC_LETTER_TCHEH_MEDIAL_FORM = 0xFB7D;
const char32_t ARABIC_LETTER_TCHEHEH_ISOLATED_FORM = 0xFB7E;
const char32_t ARABIC_LETTER_TCHEHEH_FINAL_FORM = 0xFB7F;
const char32_t ARABIC_LETTER_TCHEHEH_INITIAL_FORM = 0xFB80;
const char32_t ARABIC_LETTER_TCHEHEH_MEDIAL_FORM = 0xFB81;
const char32_t ARABIC_LETTER_DDAHAL_ISOLATED_FORM = 0xFB82;
const char32_t ARABIC_LETTER_DDAHAL_FINAL_FORM = 0xFB83;
const char32_t ARABIC_LETTER_DAHAL_ISOLATED_FORM = 0xFB84;
const char32_t ARABIC_LETTER_DAHAL_FINAL_FORM = 0xFB85;
const char32_t ARABIC_LETTER_DUL_ISOLATED_FORM = 0xFB86;
const char32_t ARABIC_LETTER_DUL_FINAL_FORM = 0xFB87;
const char32_t ARABIC_LETTER_DDAL_ISOLATED_FORM = 0xFB88;
const char32_t ARABIC_LETTER_DDAL_FINAL_FORM = 0xFB89;
const char32_t ARABIC_LETTER_JEH_ISOLATED_FORM = 0xFB8A;
const char32_t ARABIC_LETTER_JEH_FINAL_FORM = 0xFB8B;
const char32_t ARABIC_LETTER_RREH_ISOLATED_FORM = 0xFB8C;
const char32_t ARABIC_LETTER_RREH_FINAL_FORM = 0xFB8D;
const char32_t ARABIC_LETTER_KEHEH_ISOLATED_FORM = 0xFB8E;
const char32_t ARABIC_LETTER_KEHEH_FINAL_FORM = 0xFB8F;
const char32_t ARABIC_LETTER_KEHEH_INITIAL_FORM = 0xFB90;
const char32_t ARABIC_LETTER_KEHEH_MEDIAL_FORM = 0xFB91;
const char32_t ARABIC_LETTER_GAF_ISOLATED_FORM = 0xFB92;
const char32_t ARABIC_LETTER_GAF_FINAL_FORM = 0xFB93;
const char32_t ARABIC_LETTER_GAF_INITIAL_FORM = 0xFB94;
const char32_t ARABIC_LETTER_GAF_MEDIAL_FORM = 0xFB95;
const char32_t ARABIC_LETTER_GUEH_ISOLATED_FORM = 0xFB96;
const char32_t ARABIC_LETTER_GUEH_FINAL_FORM = 0xFB97;
const char32_t ARABIC_LETTER_GUEH_INITIAL_FORM = 0xFB98;
const char32_t ARABIC_LETTER_GUEH_MEDIAL_FORM = 0xFB99;
const char32_t ARABIC_LETTER_NGOEH_ISOLATED_FORM = 0xFB9A;
const char32_t ARABIC_LETTER_NGOEH_FINAL_FORM = 0xFB9B;
const char32_t ARABIC_LETTER_NGOEH_INITIAL_FORM = 0xFB9C;
const char32_t ARABIC_LETTER_NGOEH_MEDIAL_FORM = 0xFB9D;
const char32_t ARABIC_LETTER_NOON_GHUNNA_ISOLATED_FORM = 0xFB9E;
const char32_t ARABIC_LETTER_NOON_GHUNNA_FINAL_FORM = 0xFB9F;
const char32_t ARABIC_LETTER_RNOON_ISOLATED_FORM = 0xFBA0;
const char32_t ARABIC_LETTER_RNOON_FINAL_FORM = 0xFBA1;
const char32_t ARABIC_LETTER_RNOON_INITIAL_FORM = 0xFBA2;
const char32_t ARABIC_LETTER_RNOON_MEDIAL_FORM = 0xFBA3;
const char32_t ARABIC_LETTER_HEH_WITH_YEH_ABOVE_ISOLATED_FORM = 0xFBA4;
const char32_t ARABIC_LETTER_HEH_WITH_YEH_ABOVE_FINAL_FORM = 0xFBA5;
const char32_t ARABIC_LETTER_HEH_GOAL_ISOLATED_FORM = 0xFBA6;
const char32_t ARABIC_LETTER_HEH_GOAL_FINAL_FORM = 0xFBA7;
const char32_t ARABIC_LETTER_HEH_GOAL_INITIAL_FORM = 0xFBA8;
const char32_t ARABIC_LETTER_HEH_GOAL_MEDIAL_FORM = 0xFBA9;
const char32_t ARABIC_LETTER_HEH_DOACHASHMEE_ISOLATED_FORM = 0xFBAA;
const char32_t ARABIC_LETTER_HEH_DOACHASHMEE_FINAL_FORM = 0xFBAB;
const char32_t ARABIC_LETTER_HEH_DOACHASHMEE_INITIAL_FORM = 0xFBAC;
const char32_t ARABIC_LETTER_HEH_DOACHASHMEE_MEDIAL_FORM = 0xFBAD;
const char32_t ARABIC_LETTER_YEH_BARREE_ISOLATED_FORM = 0xFBAE;
const char32_t ARABIC_LETTER_YEH_BARREE_FINAL_FORM = 0xFBAF;
const char32_t ARABIC_LETTER_YEH_BARREE_WITH_HAMZA_ABOVE_ISOLATED_FORM = 0xFBB0;
const char32_t ARABIC_LETTER_YEH_BARREE_WITH_HAMZA_ABOVE_FINAL_FORM = 0xFBB1;
const char32_t ARABIC_SYMBOL_DOT_ABOVE = 0xFBB2;
const char32_t ARABIC_SYMBOL_DOT_BELOW = 0xFBB3;
const char32_t ARABIC_SYMBOL_TWO_DOTS_ABOVE = 0xFBB4;
const char32_t ARABIC_SYMBOL_TWO_DOTS_BELOW = 0xFBB5;
const char32_t ARABIC_SYMBOL_THREE_DOTS_ABOVE = 0xFBB6;
const char32_t ARABIC_SYMBOL_THREE_DOTS_BELOW = 0xFBB7;
const char32_t ARABIC_SYMBOL_THREE_DOTS_POINTING_DOWNWARDS_ABOVE = 0xFBB8;
const char32_t ARABIC_SYMBOL_THREE_DOTS_POINTING_DOWNWARDS_BELOW = 0xFBB9;
const char32_t ARABIC_SYMBOL_FOUR_DOTS_ABOVE = 0xFBBA;
const char32_t ARABIC_SYMBOL_FOUR_DOTS_BELOW = 0xFBBB;
const char32_t ARABIC_SYMBOL_DOUBLE_VERTICAL_BAR_BELOW = 0xFBBC;
const char32_t ARABIC_SYMBOL_TWO_DOTS_VERTICALLY_ABOVE = 0xFBBD;
const char32_t ARABIC_SYMBOL_TWO_DOTS_VERTICALLY_BELOW = 0xFBBE;
const char32_t ARABIC_SYMBOL_RING = 0xFBBF;
const char32_t ARABIC_SYMBOL_SMALL_TAH_ABOVE = 0xFBC0;
const char32_t ARABIC_SYMBOL_SMALL_TAH_BELOW = 0xFBC1;
const char32_t ARABIC_SYMBOL_WASLA_ABOVE = 0xFBC2;
const char32_t ARABIC_LIGATURE_JALLA_WA_ALAA = 0xFBC3;
const char32_t ARABIC_LIGATURE_DAAMAT_BARAKAATUHUM = 0xFBC4;
const char32_t ARABIC_LIGATURE_RAHMATU_ALLAAHI_TAAALAA_ALAYH = 0xFBC5;
const char32_t ARABIC_LIGATURE_RAHMATU_ALLAAHI_ALAYHIM = 0xFBC6;
const char32_t ARABIC_LIGATURE_RAHMATU_ALLAAHI_ALAYHIMAA = 0xFBC7;
const char32_t ARABIC_LIGATURE_RAHIMAHUM_ALLAAHU_TAAALAA = 0xFBC8;
const char32_t ARABIC_LIGATURE_RAHIMAHUMAA_ALLAAH = 0xFBC9;
const char32_t ARABIC_LIGATURE_RAHIMAHUMAA_ALLAAHU_TAAALAA = 0xFBCA;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_TAAALAA_ANHUM = 0xFBCB;
const char32_t ARABIC_LIGATURE_HAFIZAHU_ALLAAH = 0xFBCC;
const char32_t ARABIC_LIGATURE_HAFIZAHU_ALLAAHU_TAAALAA = 0xFBCD;
const char32_t ARABIC_LIGATURE_HAFIZAHUM_ALLAAHU_TAAALAA = 0xFBCE;
const char32_t ARABIC_LIGATURE_HAFIZAHUMAA_ALLAAHU_TAAALAA = 0xFBCF;
const char32_t ARABIC_LIGATURE_SALLALLAAHU_TAAALAA_ALAYHI_WA_SALLAM = 0xFBD0;
const char32_t ARABIC_LIGATURE_AJJAL_ALLAAHU_FARAJAHU_ASH_SHAREEF = 0xFBD1;
const char32_t ARABIC_LIGATURE_ALAYHI_AR_RAHMAH = 0xFBD2;
const char32_t ARABIC_LETTER_NG_ISOLATED_FORM = 0xFBD3;
const char32_t ARABIC_LETTER_NG_FINAL_FORM = 0xFBD4;
const char32_t ARABIC_LETTER_NG_INITIAL_FORM = 0xFBD5;
const char32_t ARABIC_LETTER_NG_MEDIAL_FORM = 0xFBD6;
const char32_t ARABIC_LETTER_U_ISOLATED_FORM = 0xFBD7;
const char32_t ARABIC_LETTER_U_FINAL_FORM = 0xFBD8;
const char32_t ARABIC_LETTER_OE_ISOLATED_FORM = 0xFBD9;
const char32_t ARABIC_LETTER_OE_FINAL_FORM = 0xFBDA;
const char32_t ARABIC_LETTER_YU_ISOLATED_FORM = 0xFBDB;
const char32_t ARABIC_LETTER_YU_FINAL_FORM = 0xFBDC;
const char32_t ARABIC_LETTER_U_WITH_HAMZA_ABOVE_ISOLATED_FORM = 0xFBDD;
const char32_t ARABIC_LETTER_VE_ISOLATED_FORM = 0xFBDE;
const char32_t ARABIC_LETTER_VE_FINAL_FORM = 0xFBDF;
const char32_t ARABIC_LETTER_KIRGHIZ_OE_ISOLATED_FORM = 0xFBE0;
const char32_t ARABIC_LETTER_KIRGHIZ_OE_FINAL_FORM = 0xFBE1;
const char32_t ARABIC_LETTER_KIRGHIZ_YU_ISOLATED_FORM = 0xFBE2;
const char32_t ARABIC_LETTER_KIRGHIZ_YU_FINAL_FORM = 0xFBE3;
const char32_t ARABIC_LETTER_E_ISOLATED_FORM = 0xFBE4;
const char32_t ARABIC_LETTER_E_FINAL_FORM = 0xFBE5;
const char32_t ARABIC_LETTER_E_INITIAL_FORM = 0xFBE6;
const char32_t ARABIC_LETTER_E_MEDIAL_FORM = 0xFBE7;
const char32_t ARABIC_LETTER_UIGHUR_KAZAKH_KIRGHIZ_ALEF_MAKSURA_INITIAL_FORM = 0xFBE8;
const char32_t ARABIC_LETTER_UIGHUR_KAZAKH_KIRGHIZ_ALEF_MAKSURA_MEDIAL_FORM = 0xFBE9;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_ALEF_ISOLATED_FORM = 0xFBEA;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_ALEF_FINAL_FORM = 0xFBEB;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_AE_ISOLATED_FORM = 0xFBEC;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_AE_FINAL_FORM = 0xFBED;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_WAW_ISOLATED_FORM = 0xFBEE;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_WAW_FINAL_FORM = 0xFBEF;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_U_ISOLATED_FORM = 0xFBF0;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_U_FINAL_FORM = 0xFBF1;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_OE_ISOLATED_FORM = 0xFBF2;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_OE_FINAL_FORM = 0xFBF3;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_YU_ISOLATED_FORM = 0xFBF4;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_YU_FINAL_FORM = 0xFBF5;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_E_ISOLATED_FORM = 0xFBF6;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_E_FINAL_FORM = 0xFBF7;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_E_INITIAL_FORM = 0xFBF8;
const char32_t ARABIC_LIGATURE_UIGHUR_KIRGHIZ_YEH_WITH_HAMZA_ABOVE_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFBF9;
const char32_t ARABIC_LIGATURE_UIGHUR_KIRGHIZ_YEH_WITH_HAMZA_ABOVE_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFBFA;
const char32_t ARABIC_LIGATURE_UIGHUR_KIRGHIZ_YEH_WITH_HAMZA_ABOVE_WITH_ALEF_MAKSURA_INITIAL_FORM = 0xFBFB;
const char32_t ARABIC_LETTER_FARSI_YEH_ISOLATED_FORM = 0xFBFC;
const char32_t ARABIC_LETTER_FARSI_YEH_FINAL_FORM = 0xFBFD;
const char32_t ARABIC_LETTER_FARSI_YEH_INITIAL_FORM = 0xFBFE;
const char32_t ARABIC_LETTER_FARSI_YEH_MEDIAL_FORM = 0xFBFF;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_JEEM_ISOLATED_FORM = 0xFC00;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_HAH_ISOLATED_FORM = 0xFC01;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_MEEM_ISOLATED_FORM = 0xFC02;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC03;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_YEH_ISOLATED_FORM = 0xFC04;
const char32_t ARABIC_LIGATURE_BEH_WITH_JEEM_ISOLATED_FORM = 0xFC05;
const char32_t ARABIC_LIGATURE_BEH_WITH_HAH_ISOLATED_FORM = 0xFC06;
const char32_t ARABIC_LIGATURE_BEH_WITH_KHAH_ISOLATED_FORM = 0xFC07;
const char32_t ARABIC_LIGATURE_BEH_WITH_MEEM_ISOLATED_FORM = 0xFC08;
const char32_t ARABIC_LIGATURE_BEH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC09;
const char32_t ARABIC_LIGATURE_BEH_WITH_YEH_ISOLATED_FORM = 0xFC0A;
const char32_t ARABIC_LIGATURE_TEH_WITH_JEEM_ISOLATED_FORM = 0xFC0B;
const char32_t ARABIC_LIGATURE_TEH_WITH_HAH_ISOLATED_FORM = 0xFC0C;
const char32_t ARABIC_LIGATURE_TEH_WITH_KHAH_ISOLATED_FORM = 0xFC0D;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_ISOLATED_FORM = 0xFC0E;
const char32_t ARABIC_LIGATURE_TEH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC0F;
const char32_t ARABIC_LIGATURE_TEH_WITH_YEH_ISOLATED_FORM = 0xFC10;
const char32_t ARABIC_LIGATURE_THEH_WITH_JEEM_ISOLATED_FORM = 0xFC11;
const char32_t ARABIC_LIGATURE_THEH_WITH_MEEM_ISOLATED_FORM = 0xFC12;
const char32_t ARABIC_LIGATURE_THEH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC13;
const char32_t ARABIC_LIGATURE_THEH_WITH_YEH_ISOLATED_FORM = 0xFC14;
const char32_t ARABIC_LIGATURE_JEEM_WITH_HAH_ISOLATED_FORM = 0xFC15;
const char32_t ARABIC_LIGATURE_JEEM_WITH_MEEM_ISOLATED_FORM = 0xFC16;
const char32_t ARABIC_LIGATURE_HAH_WITH_JEEM_ISOLATED_FORM = 0xFC17;
const char32_t ARABIC_LIGATURE_HAH_WITH_MEEM_ISOLATED_FORM = 0xFC18;
const char32_t ARABIC_LIGATURE_KHAH_WITH_JEEM_ISOLATED_FORM = 0xFC19;
const char32_t ARABIC_LIGATURE_KHAH_WITH_HAH_ISOLATED_FORM = 0xFC1A;
const char32_t ARABIC_LIGATURE_KHAH_WITH_MEEM_ISOLATED_FORM = 0xFC1B;
const char32_t ARABIC_LIGATURE_SEEN_WITH_JEEM_ISOLATED_FORM = 0xFC1C;
const char32_t ARABIC_LIGATURE_SEEN_WITH_HAH_ISOLATED_FORM = 0xFC1D;
const char32_t ARABIC_LIGATURE_SEEN_WITH_KHAH_ISOLATED_FORM = 0xFC1E;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_ISOLATED_FORM = 0xFC1F;
const char32_t ARABIC_LIGATURE_SAD_WITH_HAH_ISOLATED_FORM = 0xFC20;
const char32_t ARABIC_LIGATURE_SAD_WITH_MEEM_ISOLATED_FORM = 0xFC21;
const char32_t ARABIC_LIGATURE_DAD_WITH_JEEM_ISOLATED_FORM = 0xFC22;
const char32_t ARABIC_LIGATURE_DAD_WITH_HAH_ISOLATED_FORM = 0xFC23;
const char32_t ARABIC_LIGATURE_DAD_WITH_KHAH_ISOLATED_FORM = 0xFC24;
const char32_t ARABIC_LIGATURE_DAD_WITH_MEEM_ISOLATED_FORM = 0xFC25;
const char32_t ARABIC_LIGATURE_TAH_WITH_HAH_ISOLATED_FORM = 0xFC26;
const char32_t ARABIC_LIGATURE_TAH_WITH_MEEM_ISOLATED_FORM = 0xFC27;
const char32_t ARABIC_LIGATURE_ZAH_WITH_MEEM_ISOLATED_FORM = 0xFC28;
const char32_t ARABIC_LIGATURE_AIN_WITH_JEEM_ISOLATED_FORM = 0xFC29;
const char32_t ARABIC_LIGATURE_AIN_WITH_MEEM_ISOLATED_FORM = 0xFC2A;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_JEEM_ISOLATED_FORM = 0xFC2B;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_MEEM_ISOLATED_FORM = 0xFC2C;
const char32_t ARABIC_LIGATURE_FEH_WITH_JEEM_ISOLATED_FORM = 0xFC2D;
const char32_t ARABIC_LIGATURE_FEH_WITH_HAH_ISOLATED_FORM = 0xFC2E;
const char32_t ARABIC_LIGATURE_FEH_WITH_KHAH_ISOLATED_FORM = 0xFC2F;
const char32_t ARABIC_LIGATURE_FEH_WITH_MEEM_ISOLATED_FORM = 0xFC30;
const char32_t ARABIC_LIGATURE_FEH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC31;
const char32_t ARABIC_LIGATURE_FEH_WITH_YEH_ISOLATED_FORM = 0xFC32;
const char32_t ARABIC_LIGATURE_QAF_WITH_HAH_ISOLATED_FORM = 0xFC33;
const char32_t ARABIC_LIGATURE_QAF_WITH_MEEM_ISOLATED_FORM = 0xFC34;
const char32_t ARABIC_LIGATURE_QAF_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC35;
const char32_t ARABIC_LIGATURE_QAF_WITH_YEH_ISOLATED_FORM = 0xFC36;
const char32_t ARABIC_LIGATURE_KAF_WITH_ALEF_ISOLATED_FORM = 0xFC37;
const char32_t ARABIC_LIGATURE_KAF_WITH_JEEM_ISOLATED_FORM = 0xFC38;
const char32_t ARABIC_LIGATURE_KAF_WITH_HAH_ISOLATED_FORM = 0xFC39;
const char32_t ARABIC_LIGATURE_KAF_WITH_KHAH_ISOLATED_FORM = 0xFC3A;
const char32_t ARABIC_LIGATURE_KAF_WITH_LAM_ISOLATED_FORM = 0xFC3B;
const char32_t ARABIC_LIGATURE_KAF_WITH_MEEM_ISOLATED_FORM = 0xFC3C;
const char32_t ARABIC_LIGATURE_KAF_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC3D;
const char32_t ARABIC_LIGATURE_KAF_WITH_YEH_ISOLATED_FORM = 0xFC3E;
const char32_t ARABIC_LIGATURE_LAM_WITH_JEEM_ISOLATED_FORM = 0xFC3F;
const char32_t ARABIC_LIGATURE_LAM_WITH_HAH_ISOLATED_FORM = 0xFC40;
const char32_t ARABIC_LIGATURE_LAM_WITH_KHAH_ISOLATED_FORM = 0xFC41;
const char32_t ARABIC_LIGATURE_LAM_WITH_MEEM_ISOLATED_FORM = 0xFC42;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC43;
const char32_t ARABIC_LIGATURE_LAM_WITH_YEH_ISOLATED_FORM = 0xFC44;
const char32_t ARABIC_LIGATURE_MEEM_WITH_JEEM_ISOLATED_FORM = 0xFC45;
const char32_t ARABIC_LIGATURE_MEEM_WITH_HAH_ISOLATED_FORM = 0xFC46;
const char32_t ARABIC_LIGATURE_MEEM_WITH_KHAH_ISOLATED_FORM = 0xFC47;
const char32_t ARABIC_LIGATURE_MEEM_WITH_MEEM_ISOLATED_FORM = 0xFC48;
const char32_t ARABIC_LIGATURE_MEEM_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC49;
const char32_t ARABIC_LIGATURE_MEEM_WITH_YEH_ISOLATED_FORM = 0xFC4A;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_ISOLATED_FORM = 0xFC4B;
const char32_t ARABIC_LIGATURE_NOON_WITH_HAH_ISOLATED_FORM = 0xFC4C;
const char32_t ARABIC_LIGATURE_NOON_WITH_KHAH_ISOLATED_FORM = 0xFC4D;
const char32_t ARABIC_LIGATURE_NOON_WITH_MEEM_ISOLATED_FORM = 0xFC4E;
const char32_t ARABIC_LIGATURE_NOON_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC4F;
const char32_t ARABIC_LIGATURE_NOON_WITH_YEH_ISOLATED_FORM = 0xFC50;
const char32_t ARABIC_LIGATURE_HEH_WITH_JEEM_ISOLATED_FORM = 0xFC51;
const char32_t ARABIC_LIGATURE_HEH_WITH_MEEM_ISOLATED_FORM = 0xFC52;
const char32_t ARABIC_LIGATURE_HEH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC53;
const char32_t ARABIC_LIGATURE_HEH_WITH_YEH_ISOLATED_FORM = 0xFC54;
const char32_t ARABIC_LIGATURE_YEH_WITH_JEEM_ISOLATED_FORM = 0xFC55;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAH_ISOLATED_FORM = 0xFC56;
const char32_t ARABIC_LIGATURE_YEH_WITH_KHAH_ISOLATED_FORM = 0xFC57;
const char32_t ARABIC_LIGATURE_YEH_WITH_MEEM_ISOLATED_FORM = 0xFC58;
const char32_t ARABIC_LIGATURE_YEH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFC59;
const char32_t ARABIC_LIGATURE_YEH_WITH_YEH_ISOLATED_FORM = 0xFC5A;
const char32_t ARABIC_LIGATURE_THAL_WITH_SUPERSCRIPT_ALEF_ISOLATED_FORM = 0xFC5B;
const char32_t ARABIC_LIGATURE_REH_WITH_SUPERSCRIPT_ALEF_ISOLATED_FORM = 0xFC5C;
const char32_t ARABIC_LIGATURE_ALEF_MAKSURA_WITH_SUPERSCRIPT_ALEF_ISOLATED_FORM = 0xFC5D;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_DAMMATAN_ISOLATED_FORM = 0xFC5E;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_KASRATAN_ISOLATED_FORM = 0xFC5F;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_FATHA_ISOLATED_FORM = 0xFC60;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_DAMMA_ISOLATED_FORM = 0xFC61;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_KASRA_ISOLATED_FORM = 0xFC62;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_SUPERSCRIPT_ALEF_ISOLATED_FORM = 0xFC63;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_REH_FINAL_FORM = 0xFC64;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_ZAIN_FINAL_FORM = 0xFC65;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_MEEM_FINAL_FORM = 0xFC66;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_NOON_FINAL_FORM = 0xFC67;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC68;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_YEH_FINAL_FORM = 0xFC69;
const char32_t ARABIC_LIGATURE_BEH_WITH_REH_FINAL_FORM = 0xFC6A;
const char32_t ARABIC_LIGATURE_BEH_WITH_ZAIN_FINAL_FORM = 0xFC6B;
const char32_t ARABIC_LIGATURE_BEH_WITH_MEEM_FINAL_FORM = 0xFC6C;
const char32_t ARABIC_LIGATURE_BEH_WITH_NOON_FINAL_FORM = 0xFC6D;
const char32_t ARABIC_LIGATURE_BEH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC6E;
const char32_t ARABIC_LIGATURE_BEH_WITH_YEH_FINAL_FORM = 0xFC6F;
const char32_t ARABIC_LIGATURE_TEH_WITH_REH_FINAL_FORM = 0xFC70;
const char32_t ARABIC_LIGATURE_TEH_WITH_ZAIN_FINAL_FORM = 0xFC71;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_FINAL_FORM = 0xFC72;
const char32_t ARABIC_LIGATURE_TEH_WITH_NOON_FINAL_FORM = 0xFC73;
const char32_t ARABIC_LIGATURE_TEH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC74;
const char32_t ARABIC_LIGATURE_TEH_WITH_YEH_FINAL_FORM = 0xFC75;
const char32_t ARABIC_LIGATURE_THEH_WITH_REH_FINAL_FORM = 0xFC76;
const char32_t ARABIC_LIGATURE_THEH_WITH_ZAIN_FINAL_FORM = 0xFC77;
const char32_t ARABIC_LIGATURE_THEH_WITH_MEEM_FINAL_FORM = 0xFC78;
const char32_t ARABIC_LIGATURE_THEH_WITH_NOON_FINAL_FORM = 0xFC79;
const char32_t ARABIC_LIGATURE_THEH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC7A;
const char32_t ARABIC_LIGATURE_THEH_WITH_YEH_FINAL_FORM = 0xFC7B;
const char32_t ARABIC_LIGATURE_FEH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC7C;
const char32_t ARABIC_LIGATURE_FEH_WITH_YEH_FINAL_FORM = 0xFC7D;
const char32_t ARABIC_LIGATURE_QAF_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC7E;
const char32_t ARABIC_LIGATURE_QAF_WITH_YEH_FINAL_FORM = 0xFC7F;
const char32_t ARABIC_LIGATURE_KAF_WITH_ALEF_FINAL_FORM = 0xFC80;
const char32_t ARABIC_LIGATURE_KAF_WITH_LAM_FINAL_FORM = 0xFC81;
const char32_t ARABIC_LIGATURE_KAF_WITH_MEEM_FINAL_FORM = 0xFC82;
const char32_t ARABIC_LIGATURE_KAF_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC83;
const char32_t ARABIC_LIGATURE_KAF_WITH_YEH_FINAL_FORM = 0xFC84;
const char32_t ARABIC_LIGATURE_LAM_WITH_MEEM_FINAL_FORM = 0xFC85;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC86;
const char32_t ARABIC_LIGATURE_LAM_WITH_YEH_FINAL_FORM = 0xFC87;
const char32_t ARABIC_LIGATURE_MEEM_WITH_ALEF_FINAL_FORM = 0xFC88;
const char32_t ARABIC_LIGATURE_MEEM_WITH_MEEM_FINAL_FORM = 0xFC89;
const char32_t ARABIC_LIGATURE_NOON_WITH_REH_FINAL_FORM = 0xFC8A;
const char32_t ARABIC_LIGATURE_NOON_WITH_ZAIN_FINAL_FORM = 0xFC8B;
const char32_t ARABIC_LIGATURE_NOON_WITH_MEEM_FINAL_FORM = 0xFC8C;
const char32_t ARABIC_LIGATURE_NOON_WITH_NOON_FINAL_FORM = 0xFC8D;
const char32_t ARABIC_LIGATURE_NOON_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC8E;
const char32_t ARABIC_LIGATURE_NOON_WITH_YEH_FINAL_FORM = 0xFC8F;
const char32_t ARABIC_LIGATURE_ALEF_MAKSURA_WITH_SUPERSCRIPT_ALEF_FINAL_FORM = 0xFC90;
const char32_t ARABIC_LIGATURE_YEH_WITH_REH_FINAL_FORM = 0xFC91;
const char32_t ARABIC_LIGATURE_YEH_WITH_ZAIN_FINAL_FORM = 0xFC92;
const char32_t ARABIC_LIGATURE_YEH_WITH_MEEM_FINAL_FORM = 0xFC93;
const char32_t ARABIC_LIGATURE_YEH_WITH_NOON_FINAL_FORM = 0xFC94;
const char32_t ARABIC_LIGATURE_YEH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFC95;
const char32_t ARABIC_LIGATURE_YEH_WITH_YEH_FINAL_FORM = 0xFC96;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_JEEM_INITIAL_FORM = 0xFC97;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_HAH_INITIAL_FORM = 0xFC98;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_KHAH_INITIAL_FORM = 0xFC99;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_MEEM_INITIAL_FORM = 0xFC9A;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_HEH_INITIAL_FORM = 0xFC9B;
const char32_t ARABIC_LIGATURE_BEH_WITH_JEEM_INITIAL_FORM = 0xFC9C;
const char32_t ARABIC_LIGATURE_BEH_WITH_HAH_INITIAL_FORM = 0xFC9D;
const char32_t ARABIC_LIGATURE_BEH_WITH_KHAH_INITIAL_FORM = 0xFC9E;
const char32_t ARABIC_LIGATURE_BEH_WITH_MEEM_INITIAL_FORM = 0xFC9F;
const char32_t ARABIC_LIGATURE_BEH_WITH_HEH_INITIAL_FORM = 0xFCA0;
const char32_t ARABIC_LIGATURE_TEH_WITH_JEEM_INITIAL_FORM = 0xFCA1;
const char32_t ARABIC_LIGATURE_TEH_WITH_HAH_INITIAL_FORM = 0xFCA2;
const char32_t ARABIC_LIGATURE_TEH_WITH_KHAH_INITIAL_FORM = 0xFCA3;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_INITIAL_FORM = 0xFCA4;
const char32_t ARABIC_LIGATURE_TEH_WITH_HEH_INITIAL_FORM = 0xFCA5;
const char32_t ARABIC_LIGATURE_THEH_WITH_MEEM_INITIAL_FORM = 0xFCA6;
const char32_t ARABIC_LIGATURE_JEEM_WITH_HAH_INITIAL_FORM = 0xFCA7;
const char32_t ARABIC_LIGATURE_JEEM_WITH_MEEM_INITIAL_FORM = 0xFCA8;
const char32_t ARABIC_LIGATURE_HAH_WITH_JEEM_INITIAL_FORM = 0xFCA9;
const char32_t ARABIC_LIGATURE_HAH_WITH_MEEM_INITIAL_FORM = 0xFCAA;
const char32_t ARABIC_LIGATURE_KHAH_WITH_JEEM_INITIAL_FORM = 0xFCAB;
const char32_t ARABIC_LIGATURE_KHAH_WITH_MEEM_INITIAL_FORM = 0xFCAC;
const char32_t ARABIC_LIGATURE_SEEN_WITH_JEEM_INITIAL_FORM = 0xFCAD;
const char32_t ARABIC_LIGATURE_SEEN_WITH_HAH_INITIAL_FORM = 0xFCAE;
const char32_t ARABIC_LIGATURE_SEEN_WITH_KHAH_INITIAL_FORM = 0xFCAF;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_INITIAL_FORM = 0xFCB0;
const char32_t ARABIC_LIGATURE_SAD_WITH_HAH_INITIAL_FORM = 0xFCB1;
const char32_t ARABIC_LIGATURE_SAD_WITH_KHAH_INITIAL_FORM = 0xFCB2;
const char32_t ARABIC_LIGATURE_SAD_WITH_MEEM_INITIAL_FORM = 0xFCB3;
const char32_t ARABIC_LIGATURE_DAD_WITH_JEEM_INITIAL_FORM = 0xFCB4;
const char32_t ARABIC_LIGATURE_DAD_WITH_HAH_INITIAL_FORM = 0xFCB5;
const char32_t ARABIC_LIGATURE_DAD_WITH_KHAH_INITIAL_FORM = 0xFCB6;
const char32_t ARABIC_LIGATURE_DAD_WITH_MEEM_INITIAL_FORM = 0xFCB7;
const char32_t ARABIC_LIGATURE_TAH_WITH_HAH_INITIAL_FORM = 0xFCB8;
const char32_t ARABIC_LIGATURE_ZAH_WITH_MEEM_INITIAL_FORM = 0xFCB9;
const char32_t ARABIC_LIGATURE_AIN_WITH_JEEM_INITIAL_FORM = 0xFCBA;
const char32_t ARABIC_LIGATURE_AIN_WITH_MEEM_INITIAL_FORM = 0xFCBB;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_JEEM_INITIAL_FORM = 0xFCBC;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_MEEM_INITIAL_FORM = 0xFCBD;
const char32_t ARABIC_LIGATURE_FEH_WITH_JEEM_INITIAL_FORM = 0xFCBE;
const char32_t ARABIC_LIGATURE_FEH_WITH_HAH_INITIAL_FORM = 0xFCBF;
const char32_t ARABIC_LIGATURE_FEH_WITH_KHAH_INITIAL_FORM = 0xFCC0;
const char32_t ARABIC_LIGATURE_FEH_WITH_MEEM_INITIAL_FORM = 0xFCC1;
const char32_t ARABIC_LIGATURE_QAF_WITH_HAH_INITIAL_FORM = 0xFCC2;
const char32_t ARABIC_LIGATURE_QAF_WITH_MEEM_INITIAL_FORM = 0xFCC3;
const char32_t ARABIC_LIGATURE_KAF_WITH_JEEM_INITIAL_FORM = 0xFCC4;
const char32_t ARABIC_LIGATURE_KAF_WITH_HAH_INITIAL_FORM = 0xFCC5;
const char32_t ARABIC_LIGATURE_KAF_WITH_KHAH_INITIAL_FORM = 0xFCC6;
const char32_t ARABIC_LIGATURE_KAF_WITH_LAM_INITIAL_FORM = 0xFCC7;
const char32_t ARABIC_LIGATURE_KAF_WITH_MEEM_INITIAL_FORM = 0xFCC8;
const char32_t ARABIC_LIGATURE_LAM_WITH_JEEM_INITIAL_FORM = 0xFCC9;
const char32_t ARABIC_LIGATURE_LAM_WITH_HAH_INITIAL_FORM = 0xFCCA;
const char32_t ARABIC_LIGATURE_LAM_WITH_KHAH_INITIAL_FORM = 0xFCCB;
const char32_t ARABIC_LIGATURE_LAM_WITH_MEEM_INITIAL_FORM = 0xFCCC;
const char32_t ARABIC_LIGATURE_LAM_WITH_HEH_INITIAL_FORM = 0xFCCD;
const char32_t ARABIC_LIGATURE_MEEM_WITH_JEEM_INITIAL_FORM = 0xFCCE;
const char32_t ARABIC_LIGATURE_MEEM_WITH_HAH_INITIAL_FORM = 0xFCCF;
const char32_t ARABIC_LIGATURE_MEEM_WITH_KHAH_INITIAL_FORM = 0xFCD0;
const char32_t ARABIC_LIGATURE_MEEM_WITH_MEEM_INITIAL_FORM = 0xFCD1;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_INITIAL_FORM = 0xFCD2;
const char32_t ARABIC_LIGATURE_NOON_WITH_HAH_INITIAL_FORM = 0xFCD3;
const char32_t ARABIC_LIGATURE_NOON_WITH_KHAH_INITIAL_FORM = 0xFCD4;
const char32_t ARABIC_LIGATURE_NOON_WITH_MEEM_INITIAL_FORM = 0xFCD5;
const char32_t ARABIC_LIGATURE_NOON_WITH_HEH_INITIAL_FORM = 0xFCD6;
const char32_t ARABIC_LIGATURE_HEH_WITH_JEEM_INITIAL_FORM = 0xFCD7;
const char32_t ARABIC_LIGATURE_HEH_WITH_MEEM_INITIAL_FORM = 0xFCD8;
const char32_t ARABIC_LIGATURE_HEH_WITH_SUPERSCRIPT_ALEF_INITIAL_FORM = 0xFCD9;
const char32_t ARABIC_LIGATURE_YEH_WITH_JEEM_INITIAL_FORM = 0xFCDA;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAH_INITIAL_FORM = 0xFCDB;
const char32_t ARABIC_LIGATURE_YEH_WITH_KHAH_INITIAL_FORM = 0xFCDC;
const char32_t ARABIC_LIGATURE_YEH_WITH_MEEM_INITIAL_FORM = 0xFCDD;
const char32_t ARABIC_LIGATURE_YEH_WITH_HEH_INITIAL_FORM = 0xFCDE;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_MEEM_MEDIAL_FORM = 0xFCDF;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAMZA_ABOVE_WITH_HEH_MEDIAL_FORM = 0xFCE0;
const char32_t ARABIC_LIGATURE_BEH_WITH_MEEM_MEDIAL_FORM = 0xFCE1;
const char32_t ARABIC_LIGATURE_BEH_WITH_HEH_MEDIAL_FORM = 0xFCE2;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_MEDIAL_FORM = 0xFCE3;
const char32_t ARABIC_LIGATURE_TEH_WITH_HEH_MEDIAL_FORM = 0xFCE4;
const char32_t ARABIC_LIGATURE_THEH_WITH_MEEM_MEDIAL_FORM = 0xFCE5;
const char32_t ARABIC_LIGATURE_THEH_WITH_HEH_MEDIAL_FORM = 0xFCE6;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_MEDIAL_FORM = 0xFCE7;
const char32_t ARABIC_LIGATURE_SEEN_WITH_HEH_MEDIAL_FORM = 0xFCE8;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_MEDIAL_FORM = 0xFCE9;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HEH_MEDIAL_FORM = 0xFCEA;
const char32_t ARABIC_LIGATURE_KAF_WITH_LAM_MEDIAL_FORM = 0xFCEB;
const char32_t ARABIC_LIGATURE_KAF_WITH_MEEM_MEDIAL_FORM = 0xFCEC;
const char32_t ARABIC_LIGATURE_LAM_WITH_MEEM_MEDIAL_FORM = 0xFCED;
const char32_t ARABIC_LIGATURE_NOON_WITH_MEEM_MEDIAL_FORM = 0xFCEE;
const char32_t ARABIC_LIGATURE_NOON_WITH_HEH_MEDIAL_FORM = 0xFCEF;
const char32_t ARABIC_LIGATURE_YEH_WITH_MEEM_MEDIAL_FORM = 0xFCF0;
const char32_t ARABIC_LIGATURE_YEH_WITH_HEH_MEDIAL_FORM = 0xFCF1;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_FATHA_MEDIAL_FORM = 0xFCF2;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_DAMMA_MEDIAL_FORM = 0xFCF3;
const char32_t ARABIC_LIGATURE_SHADDA_WITH_KASRA_MEDIAL_FORM = 0xFCF4;
const char32_t ARABIC_LIGATURE_TAH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFCF5;
const char32_t ARABIC_LIGATURE_TAH_WITH_YEH_ISOLATED_FORM = 0xFCF6;
const char32_t ARABIC_LIGATURE_AIN_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFCF7;
const char32_t ARABIC_LIGATURE_AIN_WITH_YEH_ISOLATED_FORM = 0xFCF8;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFCF9;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_YEH_ISOLATED_FORM = 0xFCFA;
const char32_t ARABIC_LIGATURE_SEEN_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFCFB;
const char32_t ARABIC_LIGATURE_SEEN_WITH_YEH_ISOLATED_FORM = 0xFCFC;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFCFD;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_YEH_ISOLATED_FORM = 0xFCFE;
const char32_t ARABIC_LIGATURE_HAH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFCFF;
const char32_t ARABIC_LIGATURE_HAH_WITH_YEH_ISOLATED_FORM = 0xFD00;
const char32_t ARABIC_LIGATURE_JEEM_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFD01;
const char32_t ARABIC_LIGATURE_JEEM_WITH_YEH_ISOLATED_FORM = 0xFD02;
const char32_t ARABIC_LIGATURE_KHAH_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFD03;
const char32_t ARABIC_LIGATURE_KHAH_WITH_YEH_ISOLATED_FORM = 0xFD04;
const char32_t ARABIC_LIGATURE_SAD_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFD05;
const char32_t ARABIC_LIGATURE_SAD_WITH_YEH_ISOLATED_FORM = 0xFD06;
const char32_t ARABIC_LIGATURE_DAD_WITH_ALEF_MAKSURA_ISOLATED_FORM = 0xFD07;
const char32_t ARABIC_LIGATURE_DAD_WITH_YEH_ISOLATED_FORM = 0xFD08;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_JEEM_ISOLATED_FORM = 0xFD09;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HAH_ISOLATED_FORM = 0xFD0A;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_KHAH_ISOLATED_FORM = 0xFD0B;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_ISOLATED_FORM = 0xFD0C;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_REH_ISOLATED_FORM = 0xFD0D;
const char32_t ARABIC_LIGATURE_SEEN_WITH_REH_ISOLATED_FORM = 0xFD0E;
const char32_t ARABIC_LIGATURE_SAD_WITH_REH_ISOLATED_FORM = 0xFD0F;
const char32_t ARABIC_LIGATURE_DAD_WITH_REH_ISOLATED_FORM = 0xFD10;
const char32_t ARABIC_LIGATURE_TAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD11;
const char32_t ARABIC_LIGATURE_TAH_WITH_YEH_FINAL_FORM = 0xFD12;
const char32_t ARABIC_LIGATURE_AIN_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD13;
const char32_t ARABIC_LIGATURE_AIN_WITH_YEH_FINAL_FORM = 0xFD14;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD15;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_YEH_FINAL_FORM = 0xFD16;
const char32_t ARABIC_LIGATURE_SEEN_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD17;
const char32_t ARABIC_LIGATURE_SEEN_WITH_YEH_FINAL_FORM = 0xFD18;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD19;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_YEH_FINAL_FORM = 0xFD1A;
const char32_t ARABIC_LIGATURE_HAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD1B;
const char32_t ARABIC_LIGATURE_HAH_WITH_YEH_FINAL_FORM = 0xFD1C;
const char32_t ARABIC_LIGATURE_JEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD1D;
const char32_t ARABIC_LIGATURE_JEEM_WITH_YEH_FINAL_FORM = 0xFD1E;
const char32_t ARABIC_LIGATURE_KHAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD1F;
const char32_t ARABIC_LIGATURE_KHAH_WITH_YEH_FINAL_FORM = 0xFD20;
const char32_t ARABIC_LIGATURE_SAD_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD21;
const char32_t ARABIC_LIGATURE_SAD_WITH_YEH_FINAL_FORM = 0xFD22;
const char32_t ARABIC_LIGATURE_DAD_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD23;
const char32_t ARABIC_LIGATURE_DAD_WITH_YEH_FINAL_FORM = 0xFD24;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_JEEM_FINAL_FORM = 0xFD25;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HAH_FINAL_FORM = 0xFD26;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_KHAH_FINAL_FORM = 0xFD27;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_FINAL_FORM = 0xFD28;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_REH_FINAL_FORM = 0xFD29;
const char32_t ARABIC_LIGATURE_SEEN_WITH_REH_FINAL_FORM = 0xFD2A;
const char32_t ARABIC_LIGATURE_SAD_WITH_REH_FINAL_FORM = 0xFD2B;
const char32_t ARABIC_LIGATURE_DAD_WITH_REH_FINAL_FORM = 0xFD2C;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_JEEM_INITIAL_FORM = 0xFD2D;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HAH_INITIAL_FORM = 0xFD2E;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_KHAH_INITIAL_FORM = 0xFD2F;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_INITIAL_FORM = 0xFD30;
const char32_t ARABIC_LIGATURE_SEEN_WITH_HEH_INITIAL_FORM = 0xFD31;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HEH_INITIAL_FORM = 0xFD32;
const char32_t ARABIC_LIGATURE_TAH_WITH_MEEM_INITIAL_FORM = 0xFD33;
const char32_t ARABIC_LIGATURE_SEEN_WITH_JEEM_MEDIAL_FORM = 0xFD34;
const char32_t ARABIC_LIGATURE_SEEN_WITH_HAH_MEDIAL_FORM = 0xFD35;
const char32_t ARABIC_LIGATURE_SEEN_WITH_KHAH_MEDIAL_FORM = 0xFD36;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_JEEM_MEDIAL_FORM = 0xFD37;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HAH_MEDIAL_FORM = 0xFD38;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_KHAH_MEDIAL_FORM = 0xFD39;
const char32_t ARABIC_LIGATURE_TAH_WITH_MEEM_MEDIAL_FORM = 0xFD3A;
const char32_t ARABIC_LIGATURE_ZAH_WITH_MEEM_MEDIAL_FORM = 0xFD3B;
const char32_t ARABIC_LIGATURE_ALEF_WITH_FATHATAN_FINAL_FORM = 0xFD3C;
const char32_t ARABIC_LIGATURE_ALEF_WITH_FATHATAN_ISOLATED_FORM = 0xFD3D;
const char32_t ORNATE_LEFT_PARENTHESIS = 0xFD3E;
const char32_t ORNATE_RIGHT_PARENTHESIS = 0xFD3F;
const char32_t ARABIC_LIGATURE_RAHIMAHU_ALLAAH = 0xFD40;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_ANH = 0xFD41;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_ANHAA = 0xFD42;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_ANHUM = 0xFD43;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_ANHUMAA = 0xFD44;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_ANHUNNA = 0xFD45;
const char32_t ARABIC_LIGATURE_SALLALLAAHU_ALAYHI_WA_AALIH = 0xFD46;
const char32_t ARABIC_LIGATURE_ALAYHI_AS_SALAAM = 0xFD47;
const char32_t ARABIC_LIGATURE_ALAYHIM_AS_SALAAM = 0xFD48;
const char32_t ARABIC_LIGATURE_ALAYHIMAA_AS_SALAAM = 0xFD49;
const char32_t ARABIC_LIGATURE_ALAYHI_AS_SALAATU_WAS_SALAAM = 0xFD4A;
const char32_t ARABIC_LIGATURE_QUDDISA_SIRRAH = 0xFD4B;
const char32_t ARABIC_LIGATURE_SALLALLAHU_ALAYHI_WAAALIHEE_WA_SALLAM = 0xFD4C;
const char32_t ARABIC_LIGATURE_ALAYHAA_AS_SALAAM = 0xFD4D;
const char32_t ARABIC_LIGATURE_TABAARAKA_WA_TAAALAA = 0xFD4E;
const char32_t ARABIC_LIGATURE_RAHIMAHUM_ALLAAH = 0xFD4F;
const char32_t ARABIC_LIGATURE_TEH_WITH_JEEM_WITH_MEEM_INITIAL_FORM = 0xFD50;
const char32_t ARABIC_LIGATURE_TEH_WITH_HAH_WITH_JEEM_FINAL_FORM = 0xFD51;
const char32_t ARABIC_LIGATURE_TEH_WITH_HAH_WITH_JEEM_INITIAL_FORM = 0xFD52;
const char32_t ARABIC_LIGATURE_TEH_WITH_HAH_WITH_MEEM_INITIAL_FORM = 0xFD53;
const char32_t ARABIC_LIGATURE_TEH_WITH_KHAH_WITH_MEEM_INITIAL_FORM = 0xFD54;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_WITH_JEEM_INITIAL_FORM = 0xFD55;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_WITH_HAH_INITIAL_FORM = 0xFD56;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_WITH_KHAH_INITIAL_FORM = 0xFD57;
const char32_t ARABIC_LIGATURE_JEEM_WITH_MEEM_WITH_HAH_FINAL_FORM = 0xFD58;
const char32_t ARABIC_LIGATURE_JEEM_WITH_MEEM_WITH_HAH_INITIAL_FORM = 0xFD59;
const char32_t ARABIC_LIGATURE_HAH_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFD5A;
const char32_t ARABIC_LIGATURE_HAH_WITH_MEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD5B;
const char32_t ARABIC_LIGATURE_SEEN_WITH_HAH_WITH_JEEM_INITIAL_FORM = 0xFD5C;
const char32_t ARABIC_LIGATURE_SEEN_WITH_JEEM_WITH_HAH_INITIAL_FORM = 0xFD5D;
const char32_t ARABIC_LIGATURE_SEEN_WITH_JEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD5E;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_WITH_HAH_FINAL_FORM = 0xFD5F;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_WITH_HAH_INITIAL_FORM = 0xFD60;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_WITH_JEEM_INITIAL_FORM = 0xFD61;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFD62;
const char32_t ARABIC_LIGATURE_SEEN_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFD63;
const char32_t ARABIC_LIGATURE_SAD_WITH_HAH_WITH_HAH_FINAL_FORM = 0xFD64;
const char32_t ARABIC_LIGATURE_SAD_WITH_HAH_WITH_HAH_INITIAL_FORM = 0xFD65;
const char32_t ARABIC_LIGATURE_SAD_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFD66;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HAH_WITH_MEEM_FINAL_FORM = 0xFD67;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HAH_WITH_MEEM_INITIAL_FORM = 0xFD68;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_JEEM_WITH_YEH_FINAL_FORM = 0xFD69;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_WITH_KHAH_FINAL_FORM = 0xFD6A;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_WITH_KHAH_INITIAL_FORM = 0xFD6B;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFD6C;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFD6D;
const char32_t ARABIC_LIGATURE_DAD_WITH_HAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD6E;
const char32_t ARABIC_LIGATURE_DAD_WITH_KHAH_WITH_MEEM_FINAL_FORM = 0xFD6F;
const char32_t ARABIC_LIGATURE_DAD_WITH_KHAH_WITH_MEEM_INITIAL_FORM = 0xFD70;
const char32_t ARABIC_LIGATURE_TAH_WITH_MEEM_WITH_HAH_FINAL_FORM = 0xFD71;
const char32_t ARABIC_LIGATURE_TAH_WITH_MEEM_WITH_HAH_INITIAL_FORM = 0xFD72;
const char32_t ARABIC_LIGATURE_TAH_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFD73;
const char32_t ARABIC_LIGATURE_TAH_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFD74;
const char32_t ARABIC_LIGATURE_AIN_WITH_JEEM_WITH_MEEM_FINAL_FORM = 0xFD75;
const char32_t ARABIC_LIGATURE_AIN_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFD76;
const char32_t ARABIC_LIGATURE_AIN_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFD77;
const char32_t ARABIC_LIGATURE_AIN_WITH_MEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD78;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFD79;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFD7A;
const char32_t ARABIC_LIGATURE_GHAIN_WITH_MEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD7B;
const char32_t ARABIC_LIGATURE_FEH_WITH_KHAH_WITH_MEEM_FINAL_FORM = 0xFD7C;
const char32_t ARABIC_LIGATURE_FEH_WITH_KHAH_WITH_MEEM_INITIAL_FORM = 0xFD7D;
const char32_t ARABIC_LIGATURE_QAF_WITH_MEEM_WITH_HAH_FINAL_FORM = 0xFD7E;
const char32_t ARABIC_LIGATURE_QAF_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFD7F;
const char32_t ARABIC_LIGATURE_LAM_WITH_HAH_WITH_MEEM_FINAL_FORM = 0xFD80;
const char32_t ARABIC_LIGATURE_LAM_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFD81;
const char32_t ARABIC_LIGATURE_LAM_WITH_HAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD82;
const char32_t ARABIC_LIGATURE_LAM_WITH_JEEM_WITH_JEEM_INITIAL_FORM = 0xFD83;
const char32_t ARABIC_LIGATURE_LAM_WITH_JEEM_WITH_JEEM_FINAL_FORM = 0xFD84;
const char32_t ARABIC_LIGATURE_LAM_WITH_KHAH_WITH_MEEM_FINAL_FORM = 0xFD85;
const char32_t ARABIC_LIGATURE_LAM_WITH_KHAH_WITH_MEEM_INITIAL_FORM = 0xFD86;
const char32_t ARABIC_LIGATURE_LAM_WITH_MEEM_WITH_HAH_FINAL_FORM = 0xFD87;
const char32_t ARABIC_LIGATURE_LAM_WITH_MEEM_WITH_HAH_INITIAL_FORM = 0xFD88;
const char32_t ARABIC_LIGATURE_MEEM_WITH_HAH_WITH_JEEM_INITIAL_FORM = 0xFD89;
const char32_t ARABIC_LIGATURE_MEEM_WITH_HAH_WITH_MEEM_INITIAL_FORM = 0xFD8A;
const char32_t ARABIC_LIGATURE_MEEM_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFD8B;
const char32_t ARABIC_LIGATURE_MEEM_WITH_JEEM_WITH_HAH_INITIAL_FORM = 0xFD8C;
const char32_t ARABIC_LIGATURE_MEEM_WITH_JEEM_WITH_MEEM_INITIAL_FORM = 0xFD8D;
const char32_t ARABIC_LIGATURE_MEEM_WITH_KHAH_WITH_JEEM_INITIAL_FORM = 0xFD8E;
const char32_t ARABIC_LIGATURE_MEEM_WITH_KHAH_WITH_MEEM_INITIAL_FORM = 0xFD8F;
const char32_t ARABIC_LIGATURE_RAHMATU_ALLAAHI_ALAYH = 0xFD90;
const char32_t ARABIC_LIGATURE_RAHMATU_ALLAAHI_ALAYHAA = 0xFD91;
const char32_t ARABIC_LIGATURE_MEEM_WITH_JEEM_WITH_KHAH_INITIAL_FORM = 0xFD92;
const char32_t ARABIC_LIGATURE_HEH_WITH_MEEM_WITH_JEEM_INITIAL_FORM = 0xFD93;
const char32_t ARABIC_LIGATURE_HEH_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFD94;
const char32_t ARABIC_LIGATURE_NOON_WITH_HAH_WITH_MEEM_INITIAL_FORM = 0xFD95;
const char32_t ARABIC_LIGATURE_NOON_WITH_HAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD96;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_WITH_MEEM_FINAL_FORM = 0xFD97;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_WITH_MEEM_INITIAL_FORM = 0xFD98;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD99;
const char32_t ARABIC_LIGATURE_NOON_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFD9A;
const char32_t ARABIC_LIGATURE_NOON_WITH_MEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFD9B;
const char32_t ARABIC_LIGATURE_YEH_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFD9C;
const char32_t ARABIC_LIGATURE_YEH_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFD9D;
const char32_t ARABIC_LIGATURE_BEH_WITH_KHAH_WITH_YEH_FINAL_FORM = 0xFD9E;
const char32_t ARABIC_LIGATURE_TEH_WITH_JEEM_WITH_YEH_FINAL_FORM = 0xFD9F;
const char32_t ARABIC_LIGATURE_TEH_WITH_JEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFDA0;
const char32_t ARABIC_LIGATURE_TEH_WITH_KHAH_WITH_YEH_FINAL_FORM = 0xFDA1;
const char32_t ARABIC_LIGATURE_TEH_WITH_KHAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFDA2;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDA3;
const char32_t ARABIC_LIGATURE_TEH_WITH_MEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFDA4;
const char32_t ARABIC_LIGATURE_JEEM_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDA5;
const char32_t ARABIC_LIGATURE_JEEM_WITH_HAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFDA6;
const char32_t ARABIC_LIGATURE_JEEM_WITH_MEEM_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFDA7;
const char32_t ARABIC_LIGATURE_SEEN_WITH_KHAH_WITH_ALEF_MAKSURA_FINAL_FORM = 0xFDA8;
const char32_t ARABIC_LIGATURE_SAD_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFDA9;
const char32_t ARABIC_LIGATURE_SHEEN_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFDAA;
const char32_t ARABIC_LIGATURE_DAD_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFDAB;
const char32_t ARABIC_LIGATURE_LAM_WITH_JEEM_WITH_YEH_FINAL_FORM = 0xFDAC;
const char32_t ARABIC_LIGATURE_LAM_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDAD;
const char32_t ARABIC_LIGATURE_YEH_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFDAE;
const char32_t ARABIC_LIGATURE_YEH_WITH_JEEM_WITH_YEH_FINAL_FORM = 0xFDAF;
const char32_t ARABIC_LIGATURE_YEH_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDB0;
const char32_t ARABIC_LIGATURE_MEEM_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDB1;
const char32_t ARABIC_LIGATURE_QAF_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDB2;
const char32_t ARABIC_LIGATURE_NOON_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFDB3;
const char32_t ARABIC_LIGATURE_QAF_WITH_MEEM_WITH_HAH_INITIAL_FORM = 0xFDB4;
const char32_t ARABIC_LIGATURE_LAM_WITH_HAH_WITH_MEEM_INITIAL_FORM = 0xFDB5;
const char32_t ARABIC_LIGATURE_AIN_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDB6;
const char32_t ARABIC_LIGATURE_KAF_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDB7;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_WITH_HAH_INITIAL_FORM = 0xFDB8;
const char32_t ARABIC_LIGATURE_MEEM_WITH_KHAH_WITH_YEH_FINAL_FORM = 0xFDB9;
const char32_t ARABIC_LIGATURE_LAM_WITH_JEEM_WITH_MEEM_INITIAL_FORM = 0xFDBA;
const char32_t ARABIC_LIGATURE_KAF_WITH_MEEM_WITH_MEEM_FINAL_FORM = 0xFDBB;
const char32_t ARABIC_LIGATURE_LAM_WITH_JEEM_WITH_MEEM_FINAL_FORM = 0xFDBC;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_WITH_HAH_FINAL_FORM = 0xFDBD;
const char32_t ARABIC_LIGATURE_JEEM_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFDBE;
const char32_t ARABIC_LIGATURE_HAH_WITH_JEEM_WITH_YEH_FINAL_FORM = 0xFDBF;
const char32_t ARABIC_LIGATURE_MEEM_WITH_JEEM_WITH_YEH_FINAL_FORM = 0xFDC0;
const char32_t ARABIC_LIGATURE_FEH_WITH_MEEM_WITH_YEH_FINAL_FORM = 0xFDC1;
const char32_t ARABIC_LIGATURE_BEH_WITH_HAH_WITH_YEH_FINAL_FORM = 0xFDC2;
const char32_t ARABIC_LIGATURE_KAF_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFDC3;
const char32_t ARABIC_LIGATURE_AIN_WITH_JEEM_WITH_MEEM_INITIAL_FORM = 0xFDC4;
const char32_t ARABIC_LIGATURE_SAD_WITH_MEEM_WITH_MEEM_INITIAL_FORM = 0xFDC5;
const char32_t ARABIC_LIGATURE_SEEN_WITH_KHAH_WITH_YEH_FINAL_FORM = 0xFDC6;
const char32_t ARABIC_LIGATURE_NOON_WITH_JEEM_WITH_YEH_FINAL_FORM = 0xFDC7;
const char32_t ARABIC_LIGATURE_RAHIMAHU_ALLAAH_TAAALAA = 0xFDC8;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_TAAALAA_ANH = 0xFDC9;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_TAAALAA_ANHAA = 0xFDCA;
const char32_t ARABIC_LIGATURE_RADI_ALLAAHU_TAAALAA_ANHUMAA = 0xFDCB;
const char32_t ARABIC_LIGATURE_SALLALLAHU_ALAYHI_WA_ALAA_AALIHEE_WA_SALLAM = 0xFDCC;
const char32_t ARABIC_LIGATURE_AJJAL_ALLAAHU_TAAALAA_FARAJAHU_ASH_SHAREEF = 0xFDCD;
const char32_t ARABIC_LIGATURE_KARRAMA_ALLAAHU_WAJHAH = 0xFDCE;
const char32_t ARABIC_LIGATURE_SALAAMUHU_ALAYNAA = 0xFDCF;
const char32_t ARABIC_LIGATURE_SALLA_USED_AS_KORANIC_STOP_SIGN_ISOLATED_FORM = 0xFDF0;
const char32_t ARABIC_LIGATURE_QALA_USED_AS_KORANIC_STOP_SIGN_ISOLATED_FORM = 0xFDF1;
const char32_t ARABIC_LIGATURE_ALLAH_ISOLATED_FORM = 0xFDF2;
const char32_t ARABIC_LIGATURE_AKBAR_ISOLATED_FORM = 0xFDF3;
const char32_t ARABIC_LIGATURE_MOHAMMAD_ISOLATED_FORM = 0xFDF4;
const char32_t ARABIC_LIGATURE_SALAM_ISOLATED_FORM = 0xFDF5;
const char32_t ARABIC_LIGATURE_RASOUL_ISOLATED_FORM = 0xFDF6;
const char32_t ARABIC_LIGATURE_ALAYHE_ISOLATED_FORM = 0xFDF7;
const char32_t ARABIC_LIGATURE_WASALLAM_ISOLATED_FORM = 0xFDF8;
const char32_t ARABIC_LIGATURE_SALLA_ISOLATED_FORM = 0xFDF9;
const char32_t ARABIC_LIGATURE_SALLALLAHOU_ALAYHE_WASALLAM = 0xFDFA;
const char32_t ARABIC_LIGATURE_JALLAJALALOUHOU = 0xFDFB;
const char32_t RIAL_SIGN = 0xFDFC;
const char32_t ARABIC_LIGATURE_BISMILLAH_AR_RAHMAN_AR_RAHEEM = 0xFDFD;
const char32_t ARABIC_LIGATURE_SUBHAANAHU_WA_TAAALAA = 0xFDFE;
const char32_t ARABIC_LIGATURE_AZZA_WA_JALL = 0xFDFF;
const char32_t VARIATION_SELECTOR_1 = 0xFE00;
const char32_t VARIATION_SELECTOR_2 = 0xFE01;
const char32_t VARIATION_SELECTOR_3 = 0xFE02;
const char32_t VARIATION_SELECTOR_4 = 0xFE03;
const char32_t VARIATION_SELECTOR_5 = 0xFE04;
const char32_t VARIATION_SELECTOR_6 = 0xFE05;
const char32_t VARIATION_SELECTOR_7 = 0xFE06;
const char32_t VARIATION_SELECTOR_8 = 0xFE07;
const char32_t VARIATION_SELECTOR_9 = 0xFE08;
const char32_t VARIATION_SELECTOR_10 = 0xFE09;
const char32_t VARIATION_SELECTOR_11 = 0xFE0A;
const char32_t VARIATION_SELECTOR_12 = 0xFE0B;
const char32_t VARIATION_SELECTOR_13 = 0xFE0C;
const char32_t VARIATION_SELECTOR_14 = 0xFE0D;
const char32_t VARIATION_SELECTOR_15 = 0xFE0E;
const char32_t VARIATION_SELECTOR_16 = 0xFE0F;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_COMMA = 0xFE10;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_IDEOGRAPHIC_COMMA = 0xFE11;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_IDEOGRAPHIC_FULL_STOP = 0xFE12;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_COLON = 0xFE13;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_SEMICOLON = 0xFE14;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_EXCLAMATION_MARK = 0xFE15;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_QUESTION_MARK = 0xFE16;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_WHITE_LENTICULAR_BRACKET = 0xFE17;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_WHITE_LENTICULAR_BRAKCET = 0xFE18;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_HORIZONTAL_ELLIPSIS = 0xFE19;
const char32_t COMBINING_LIGATURE_LEFT_HALF = 0xFE20;
const char32_t COMBINING_LIGATURE_RIGHT_HALF = 0xFE21;
const char32_t COMBINING_DOUBLE_TILDE_LEFT_HALF = 0xFE22;
const char32_t COMBINING_DOUBLE_TILDE_RIGHT_HALF = 0xFE23;
const char32_t COMBINING_MACRON_LEFT_HALF = 0xFE24;
const char32_t COMBINING_MACRON_RIGHT_HALF = 0xFE25;
const char32_t COMBINING_CONJOINING_MACRON = 0xFE26;
const char32_t COMBINING_LIGATURE_LEFT_HALF_BELOW = 0xFE27;
const char32_t COMBINING_LIGATURE_RIGHT_HALF_BELOW = 0xFE28;
const char32_t COMBINING_TILDE_LEFT_HALF_BELOW = 0xFE29;
const char32_t COMBINING_TILDE_RIGHT_HALF_BELOW = 0xFE2A;
const char32_t COMBINING_MACRON_LEFT_HALF_BELOW = 0xFE2B;
const char32_t COMBINING_MACRON_RIGHT_HALF_BELOW = 0xFE2C;
const char32_t COMBINING_CONJOINING_MACRON_BELOW = 0xFE2D;
const char32_t COMBINING_CYRILLIC_TITLO_LEFT_HALF = 0xFE2E;
const char32_t COMBINING_CYRILLIC_TITLO_RIGHT_HALF = 0xFE2F;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_TWO_DOT_LEADER = 0xFE30;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_EM_DASH = 0xFE31;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_EN_DASH = 0xFE32;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LOW_LINE = 0xFE33;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_WAVY_LOW_LINE = 0xFE34;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_PARENTHESIS = 0xFE35;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_PARENTHESIS = 0xFE36;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_CURLY_BRACKET = 0xFE37;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_CURLY_BRACKET = 0xFE38;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_TORTOISE_SHELL_BRACKET = 0xFE39;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_TORTOISE_SHELL_BRACKET = 0xFE3A;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_BLACK_LENTICULAR_BRACKET = 0xFE3B;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_BLACK_LENTICULAR_BRACKET = 0xFE3C;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_DOUBLE_ANGLE_BRACKET = 0xFE3D;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_DOUBLE_ANGLE_BRACKET = 0xFE3E;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_ANGLE_BRACKET = 0xFE3F;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_ANGLE_BRACKET = 0xFE40;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_CORNER_BRACKET = 0xFE41;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_CORNER_BRACKET = 0xFE42;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_WHITE_CORNER_BRACKET = 0xFE43;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_WHITE_CORNER_BRACKET = 0xFE44;
const char32_t SESAME_DOT = 0xFE45;
const char32_t WHITE_SESAME_DOT = 0xFE46;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_LEFT_SQUARE_BRACKET = 0xFE47;
const char32_t PRESENTATION_FORM_FOR_VERTICAL_RIGHT_SQUARE_BRACKET = 0xFE48;
const char32_t DASHED_OVERLINE = 0xFE49;
const char32_t CENTRELINE_OVERLINE = 0xFE4A;
const char32_t WAVY_OVERLINE = 0xFE4B;
const char32_t DOUBLE_WAVY_OVERLINE = 0xFE4C;
const char32_t DASHED_LOW_LINE = 0xFE4D;
const char32_t CENTRELINE_LOW_LINE = 0xFE4E;
const char32_t WAVY_LOW_LINE = 0xFE4F;
const char32_t SMALL_COMMA = 0xFE50;
const char32_t SMALL_IDEOGRAPHIC_COMMA = 0xFE51;
const char32_t SMALL_FULL_STOP = 0xFE52;
const char32_t SMALL_SEMICOLON = 0xFE54;
const char32_t SMALL_COLON = 0xFE55;
const char32_t SMALL_QUESTION_MARK = 0xFE56;
const char32_t SMALL_EXCLAMATION_MARK = 0xFE57;
const char32_t SMALL_EM_DASH = 0xFE58;
const char32_t SMALL_LEFT_PARENTHESIS = 0xFE59;
const char32_t SMALL_RIGHT_PARENTHESIS = 0xFE5A;
const char32_t SMALL_LEFT_CURLY_BRACKET = 0xFE5B;
const char32_t SMALL_RIGHT_CURLY_BRACKET = 0xFE5C;
const char32_t SMALL_LEFT_TORTOISE_SHELL_BRACKET = 0xFE5D;
const char32_t SMALL_RIGHT_TORTOISE_SHELL_BRACKET = 0xFE5E;
const char32_t SMALL_NUMBER_SIGN = 0xFE5F;
const char32_t SMALL_AMPERSAND = 0xFE60;
const char32_t SMALL_ASTERISK = 0xFE61;
const char32_t SMALL_PLUS_SIGN = 0xFE62;
const char32_t SMALL_HYPHEN_MINUS = 0xFE63;
const char32_t SMALL_LESS_THAN_SIGN = 0xFE64;
const char32_t SMALL_GREATER_THAN_SIGN = 0xFE65;
const char32_t SMALL_EQUALS_SIGN = 0xFE66;
const char32_t SMALL_REVERSE_SOLIDUS = 0xFE68;
const char32_t SMALL_DOLLAR_SIGN = 0xFE69;
const char32_t SMALL_PERCENT_SIGN = 0xFE6A;
const char32_t SMALL_COMMERCIAL_AT = 0xFE6B;
const char32_t ARABIC_FATHATAN_ISOLATED_FORM = 0xFE70;
const char32_t ARABIC_TATWEEL_WITH_FATHATAN_ABOVE = 0xFE71;
const char32_t ARABIC_DAMMATAN_ISOLATED_FORM = 0xFE72;
const char32_t ARABIC_TAIL_FRAGMENT = 0xFE73;
const char32_t ARABIC_KASRATAN_ISOLATED_FORM = 0xFE74;
const char32_t ARABIC_FATHA_ISOLATED_FORM = 0xFE76;
const char32_t ARABIC_FATHA_MEDIAL_FORM = 0xFE77;
const char32_t ARABIC_DAMMA_ISOLATED_FORM = 0xFE78;
const char32_t ARABIC_DAMMA_MEDIAL_FORM = 0xFE79;
const char32_t ARABIC_KASRA_ISOLATED_FORM = 0xFE7A;
const char32_t ARABIC_KASRA_MEDIAL_FORM = 0xFE7B;
const char32_t ARABIC_SHADDA_ISOLATED_FORM = 0xFE7C;
const char32_t ARABIC_SHADDA_MEDIAL_FORM = 0xFE7D;
const char32_t ARABIC_SUKUN_ISOLATED_FORM = 0xFE7E;
const char32_t ARABIC_SUKUN_MEDIAL_FORM = 0xFE7F;
const char32_t ARABIC_LETTER_HAMZA_ISOLATED_FORM = 0xFE80;
const char32_t ARABIC_LETTER_ALEF_WITH_MADDA_ABOVE_ISOLATED_FORM = 0xFE81;
const char32_t ARABIC_LETTER_ALEF_WITH_MADDA_ABOVE_FINAL_FORM = 0xFE82;
const char32_t ARABIC_LETTER_ALEF_WITH_HAMZA_ABOVE_ISOLATED_FORM = 0xFE83;
const char32_t ARABIC_LETTER_ALEF_WITH_HAMZA_ABOVE_FINAL_FORM = 0xFE84;
const char32_t ARABIC_LETTER_WAW_WITH_HAMZA_ABOVE_ISOLATED_FORM = 0xFE85;
const char32_t ARABIC_LETTER_WAW_WITH_HAMZA_ABOVE_FINAL_FORM = 0xFE86;
const char32_t ARABIC_LETTER_ALEF_WITH_HAMZA_BELOW_ISOLATED_FORM = 0xFE87;
const char32_t ARABIC_LETTER_ALEF_WITH_HAMZA_BELOW_FINAL_FORM = 0xFE88;
const char32_t ARABIC_LETTER_YEH_WITH_HAMZA_ABOVE_ISOLATED_FORM = 0xFE89;
const char32_t ARABIC_LETTER_YEH_WITH_HAMZA_ABOVE_FINAL_FORM = 0xFE8A;
const char32_t ARABIC_LETTER_YEH_WITH_HAMZA_ABOVE_INITIAL_FORM = 0xFE8B;
const char32_t ARABIC_LETTER_YEH_WITH_HAMZA_ABOVE_MEDIAL_FORM = 0xFE8C;
const char32_t ARABIC_LETTER_ALEF_ISOLATED_FORM = 0xFE8D;
const char32_t ARABIC_LETTER_ALEF_FINAL_FORM = 0xFE8E;
const char32_t ARABIC_LETTER_BEH_ISOLATED_FORM = 0xFE8F;
const char32_t ARABIC_LETTER_BEH_FINAL_FORM = 0xFE90;
const char32_t ARABIC_LETTER_BEH_INITIAL_FORM = 0xFE91;
const char32_t ARABIC_LETTER_BEH_MEDIAL_FORM = 0xFE92;
const char32_t ARABIC_LETTER_TEH_MARBUTA_ISOLATED_FORM = 0xFE93;
const char32_t ARABIC_LETTER_TEH_MARBUTA_FINAL_FORM = 0xFE94;
const char32_t ARABIC_LETTER_TEH_ISOLATED_FORM = 0xFE95;
const char32_t ARABIC_LETTER_TEH_FINAL_FORM = 0xFE96;
const char32_t ARABIC_LETTER_TEH_INITIAL_FORM = 0xFE97;
const char32_t ARABIC_LETTER_TEH_MEDIAL_FORM = 0xFE98;
const char32_t ARABIC_LETTER_THEH_ISOLATED_FORM = 0xFE99;
const char32_t ARABIC_LETTER_THEH_FINAL_FORM = 0xFE9A;
const char32_t ARABIC_LETTER_THEH_INITIAL_FORM = 0xFE9B;
const char32_t ARABIC_LETTER_THEH_MEDIAL_FORM = 0xFE9C;
const char32_t ARABIC_LETTER_JEEM_ISOLATED_FORM = 0xFE9D;
const char32_t ARABIC_LETTER_JEEM_FINAL_FORM = 0xFE9E;
const char32_t ARABIC_LETTER_JEEM_INITIAL_FORM = 0xFE9F;
const char32_t ARABIC_LETTER_JEEM_MEDIAL_FORM = 0xFEA0;
const char32_t ARABIC_LETTER_HAH_ISOLATED_FORM = 0xFEA1;
const char32_t ARABIC_LETTER_HAH_FINAL_FORM = 0xFEA2;
const char32_t ARABIC_LETTER_HAH_INITIAL_FORM = 0xFEA3;
const char32_t ARABIC_LETTER_HAH_MEDIAL_FORM = 0xFEA4;
const char32_t ARABIC_LETTER_KHAH_ISOLATED_FORM = 0xFEA5;
const char32_t ARABIC_LETTER_KHAH_FINAL_FORM = 0xFEA6;
const char32_t ARABIC_LETTER_KHAH_INITIAL_FORM = 0xFEA7;
const char32_t ARABIC_LETTER_KHAH_MEDIAL_FORM = 0xFEA8;
const char32_t ARABIC_LETTER_DAL_ISOLATED_FORM = 0xFEA9;
const char32_t ARABIC_LETTER_DAL_FINAL_FORM = 0xFEAA;
const char32_t ARABIC_LETTER_THAL_ISOLATED_FORM = 0xFEAB;
const char32_t ARABIC_LETTER_THAL_FINAL_FORM = 0xFEAC;
const char32_t ARABIC_LETTER_REH_ISOLATED_FORM = 0xFEAD;
const char32_t ARABIC_LETTER_REH_FINAL_FORM = 0xFEAE;
const char32_t ARABIC_LETTER_ZAIN_ISOLATED_FORM = 0xFEAF;
const char32_t ARABIC_LETTER_ZAIN_FINAL_FORM = 0xFEB0;
const char32_t ARABIC_LETTER_SEEN_ISOLATED_FORM = 0xFEB1;
const char32_t ARABIC_LETTER_SEEN_FINAL_FORM = 0xFEB2;
const char32_t ARABIC_LETTER_SEEN_INITIAL_FORM = 0xFEB3;
const char32_t ARABIC_LETTER_SEEN_MEDIAL_FORM = 0xFEB4;
const char32_t ARABIC_LETTER_SHEEN_ISOLATED_FORM = 0xFEB5;
const char32_t ARABIC_LETTER_SHEEN_FINAL_FORM = 0xFEB6;
const char32_t ARABIC_LETTER_SHEEN_INITIAL_FORM = 0xFEB7;
const char32_t ARABIC_LETTER_SHEEN_MEDIAL_FORM = 0xFEB8;
const char32_t ARABIC_LETTER_SAD_ISOLATED_FORM = 0xFEB9;
const char32_t ARABIC_LETTER_SAD_FINAL_FORM = 0xFEBA;
const char32_t ARABIC_LETTER_SAD_INITIAL_FORM = 0xFEBB;
const char32_t ARABIC_LETTER_SAD_MEDIAL_FORM = 0xFEBC;
const char32_t ARABIC_LETTER_DAD_ISOLATED_FORM = 0xFEBD;
const char32_t ARABIC_LETTER_DAD_FINAL_FORM = 0xFEBE;
const char32_t ARABIC_LETTER_DAD_INITIAL_FORM = 0xFEBF;
const char32_t ARABIC_LETTER_DAD_MEDIAL_FORM = 0xFEC0;
const char32_t ARABIC_LETTER_TAH_ISOLATED_FORM = 0xFEC1;
const char32_t ARABIC_LETTER_TAH_FINAL_FORM = 0xFEC2;
const char32_t ARABIC_LETTER_TAH_INITIAL_FORM = 0xFEC3;
const char32_t ARABIC_LETTER_TAH_MEDIAL_FORM = 0xFEC4;
const char32_t ARABIC_LETTER_ZAH_ISOLATED_FORM = 0xFEC5;
const char32_t ARABIC_LETTER_ZAH_FINAL_FORM = 0xFEC6;
const char32_t ARABIC_LETTER_ZAH_INITIAL_FORM = 0xFEC7;
const char32_t ARABIC_LETTER_ZAH_MEDIAL_FORM = 0xFEC8;
const char32_t ARABIC_LETTER_AIN_ISOLATED_FORM = 0xFEC9;
const char32_t ARABIC_LETTER_AIN_FINAL_FORM = 0xFECA;
const char32_t ARABIC_LETTER_AIN_INITIAL_FORM = 0xFECB;
const char32_t ARABIC_LETTER_AIN_MEDIAL_FORM = 0xFECC;
const char32_t ARABIC_LETTER_GHAIN_ISOLATED_FORM = 0xFECD;
const char32_t ARABIC_LETTER_GHAIN_FINAL_FORM = 0xFECE;
const char32_t ARABIC_LETTER_GHAIN_INITIAL_FORM = 0xFECF;
const char32_t ARABIC_LETTER_GHAIN_MEDIAL_FORM = 0xFED0;
const char32_t ARABIC_LETTER_FEH_ISOLATED_FORM = 0xFED1;
const char32_t ARABIC_LETTER_FEH_FINAL_FORM = 0xFED2;
const char32_t ARABIC_LETTER_FEH_INITIAL_FORM = 0xFED3;
const char32_t ARABIC_LETTER_FEH_MEDIAL_FORM = 0xFED4;
const char32_t ARABIC_LETTER_QAF_ISOLATED_FORM = 0xFED5;
const char32_t ARABIC_LETTER_QAF_FINAL_FORM = 0xFED6;
const char32_t ARABIC_LETTER_QAF_INITIAL_FORM = 0xFED7;
const char32_t ARABIC_LETTER_QAF_MEDIAL_FORM = 0xFED8;
const char32_t ARABIC_LETTER_KAF_ISOLATED_FORM = 0xFED9;
const char32_t ARABIC_LETTER_KAF_FINAL_FORM = 0xFEDA;
const char32_t ARABIC_LETTER_KAF_INITIAL_FORM = 0xFEDB;
const char32_t ARABIC_LETTER_KAF_MEDIAL_FORM = 0xFEDC;
const char32_t ARABIC_LETTER_LAM_ISOLATED_FORM = 0xFEDD;
const char32_t ARABIC_LETTER_LAM_FINAL_FORM = 0xFEDE;
const char32_t ARABIC_LETTER_LAM_INITIAL_FORM = 0xFEDF;
const char32_t ARABIC_LETTER_LAM_MEDIAL_FORM = 0xFEE0;
const char32_t ARABIC_LETTER_MEEM_ISOLATED_FORM = 0xFEE1;
const char32_t ARABIC_LETTER_MEEM_FINAL_FORM = 0xFEE2;
const char32_t ARABIC_LETTER_MEEM_INITIAL_FORM = 0xFEE3;
const char32_t ARABIC_LETTER_MEEM_MEDIAL_FORM = 0xFEE4;
const char32_t ARABIC_LETTER_NOON_ISOLATED_FORM = 0xFEE5;
const char32_t ARABIC_LETTER_NOON_FINAL_FORM = 0xFEE6;
const char32_t ARABIC_LETTER_NOON_INITIAL_FORM = 0xFEE7;
const char32_t ARABIC_LETTER_NOON_MEDIAL_FORM = 0xFEE8;
const char32_t ARABIC_LETTER_HEH_ISOLATED_FORM = 0xFEE9;
const char32_t ARABIC_LETTER_HEH_FINAL_FORM = 0xFEEA;
const char32_t ARABIC_LETTER_HEH_INITIAL_FORM = 0xFEEB;
const char32_t ARABIC_LETTER_HEH_MEDIAL_FORM = 0xFEEC;
const char32_t ARABIC_LETTER_WAW_ISOLATED_FORM = 0xFEED;
const char32_t ARABIC_LETTER_WAW_FINAL_FORM = 0xFEEE;
const char32_t ARABIC_LETTER_ALEF_MAKSURA_ISOLATED_FORM = 0xFEEF;
const char32_t ARABIC_LETTER_ALEF_MAKSURA_FINAL_FORM = 0xFEF0;
const char32_t ARABIC_LETTER_YEH_ISOLATED_FORM = 0xFEF1;
const char32_t ARABIC_LETTER_YEH_FINAL_FORM = 0xFEF2;
const char32_t ARABIC_LETTER_YEH_INITIAL_FORM = 0xFEF3;
const char32_t ARABIC_LETTER_YEH_MEDIAL_FORM = 0xFEF4;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_WITH_MADDA_ABOVE_ISOLATED_FORM = 0xFEF5;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_WITH_MADDA_ABOVE_FINAL_FORM = 0xFEF6;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_WITH_HAMZA_ABOVE_ISOLATED_FORM = 0xFEF7;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_WITH_HAMZA_ABOVE_FINAL_FORM = 0xFEF8;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_WITH_HAMZA_BELOW_ISOLATED_FORM = 0xFEF9;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_WITH_HAMZA_BELOW_FINAL_FORM = 0xFEFA;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_ISOLATED_FORM = 0xFEFB;
const char32_t ARABIC_LIGATURE_LAM_WITH_ALEF_FINAL_FORM = 0xFEFC;
const char32_t ZERO_WIDTH_NO_BREAK_SPACE = 0xFEFF;
const char32_t FULLWIDTH_EXCLAMATION_MARK = 0xFF01;
const char32_t FULLWIDTH_QUOTATION_MARK = 0xFF02;
const char32_t FULLWIDTH_NUMBER_SIGN = 0xFF03;
const char32_t FULLWIDTH_DOLLAR_SIGN = 0xFF04;
const char32_t FULLWIDTH_PERCENT_SIGN = 0xFF05;
const char32_t FULLWIDTH_AMPERSAND = 0xFF06;
const char32_t FULLWIDTH_APOSTROPHE = 0xFF07;
const char32_t FULLWIDTH_LEFT_PARENTHESIS = 0xFF08;
const char32_t FULLWIDTH_RIGHT_PARENTHESIS = 0xFF09;
const char32_t FULLWIDTH_ASTERISK = 0xFF0A;
const char32_t FULLWIDTH_PLUS_SIGN = 0xFF0B;
const char32_t FULLWIDTH_COMMA = 0xFF0C;
const char32_t FULLWIDTH_HYPHEN_MINUS = 0xFF0D;
const char32_t FULLWIDTH_FULL_STOP = 0xFF0E;
const char32_t FULLWIDTH_SOLIDUS = 0xFF0F;
const char32_t FULLWIDTH_DIGIT_ZERO = 0xFF10;
const char32_t FULLWIDTH_DIGIT_ONE = 0xFF11;
const char32_t FULLWIDTH_DIGIT_TWO = 0xFF12;
const char32_t FULLWIDTH_DIGIT_THREE = 0xFF13;
const char32_t FULLWIDTH_DIGIT_FOUR = 0xFF14;
const char32_t FULLWIDTH_DIGIT_FIVE = 0xFF15;
const char32_t FULLWIDTH_DIGIT_SIX = 0xFF16;
const char32_t FULLWIDTH_DIGIT_SEVEN = 0xFF17;
const char32_t FULLWIDTH_DIGIT_EIGHT = 0xFF18;
const char32_t FULLWIDTH_DIGIT_NINE = 0xFF19;
const char32_t FULLWIDTH_COLON = 0xFF1A;
const char32_t FULLWIDTH_SEMICOLON = 0xFF1B;
const char32_t FULLWIDTH_LESS_THAN_SIGN = 0xFF1C;
const char32_t FULLWIDTH_EQUALS_SIGN = 0xFF1D;
const char32_t FULLWIDTH_GREATER_THAN_SIGN = 0xFF1E;
const char32_t FULLWIDTH_QUESTION_MARK = 0xFF1F;
const char32_t FULLWIDTH_COMMERCIAL_AT = 0xFF20;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_A = 0xFF21;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_B = 0xFF22;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_C = 0xFF23;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_D = 0xFF24;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_E = 0xFF25;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_F = 0xFF26;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_G = 0xFF27;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_H = 0xFF28;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_I = 0xFF29;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_J = 0xFF2A;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_K = 0xFF2B;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_L = 0xFF2C;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_M = 0xFF2D;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_N = 0xFF2E;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_O = 0xFF2F;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_P = 0xFF30;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_Q = 0xFF31;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_R = 0xFF32;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_S = 0xFF33;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_T = 0xFF34;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_U = 0xFF35;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_V = 0xFF36;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_W = 0xFF37;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_X = 0xFF38;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_Y = 0xFF39;
const char32_t FULLWIDTH_LATIN_CAPITAL_LETTER_Z = 0xFF3A;
const char32_t FULLWIDTH_LEFT_SQUARE_BRACKET = 0xFF3B;
const char32_t FULLWIDTH_REVERSE_SOLIDUS = 0xFF3C;
const char32_t FULLWIDTH_RIGHT_SQUARE_BRACKET = 0xFF3D;
const char32_t FULLWIDTH_CIRCUMFLEX_ACCENT = 0xFF3E;
const char32_t FULLWIDTH_LOW_LINE = 0xFF3F;
const char32_t FULLWIDTH_GRAVE_ACCENT = 0xFF40;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_A = 0xFF41;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_B = 0xFF42;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_C = 0xFF43;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_D = 0xFF44;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_E = 0xFF45;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_F = 0xFF46;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_G = 0xFF47;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_H = 0xFF48;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_I = 0xFF49;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_J = 0xFF4A;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_K = 0xFF4B;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_L = 0xFF4C;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_M = 0xFF4D;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_N = 0xFF4E;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_O = 0xFF4F;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_P = 0xFF50;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_Q = 0xFF51;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_R = 0xFF52;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_S = 0xFF53;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_T = 0xFF54;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_U = 0xFF55;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_V = 0xFF56;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_W = 0xFF57;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_X = 0xFF58;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_Y = 0xFF59;
const char32_t FULLWIDTH_LATIN_SMALL_LETTER_Z = 0xFF5A;
const char32_t FULLWIDTH_LEFT_CURLY_BRACKET = 0xFF5B;
const char32_t FULLWIDTH_VERTICAL_LINE = 0xFF5C;
const char32_t FULLWIDTH_RIGHT_CURLY_BRACKET = 0xFF5D;
const char32_t FULLWIDTH_TILDE = 0xFF5E;
const char32_t FULLWIDTH_LEFT_WHITE_PARENTHESIS = 0xFF5F;
const char32_t FULLWIDTH_RIGHT_WHITE_PARENTHESIS = 0xFF60;
const char32_t HALFWIDTH_IDEOGRAPHIC_FULL_STOP = 0xFF61;
const char32_t HALFWIDTH_LEFT_CORNER_BRACKET = 0xFF62;
const char32_t HALFWIDTH_RIGHT_CORNER_BRACKET = 0xFF63;
const char32_t HALFWIDTH_IDEOGRAPHIC_COMMA = 0xFF64;
const char32_t HALFWIDTH_KATAKANA_MIDDLE_DOT = 0xFF65;
const char32_t HALFWIDTH_KATAKANA_LETTER_WO = 0xFF66;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_A = 0xFF67;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_I = 0xFF68;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_U = 0xFF69;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_E = 0xFF6A;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_O = 0xFF6B;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_YA = 0xFF6C;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_YU = 0xFF6D;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_YO = 0xFF6E;
const char32_t HALFWIDTH_KATAKANA_LETTER_SMALL_TU = 0xFF6F;
const char32_t HALFWIDTH_KATAKANA_HIRAGANA_PROLONGED_SOUND_MARK = 0xFF70;
const char32_t HALFWIDTH_KATAKANA_LETTER_A = 0xFF71;
const char32_t HALFWIDTH_KATAKANA_LETTER_I = 0xFF72;
const char32_t HALFWIDTH_KATAKANA_LETTER_U = 0xFF73;
const char32_t HALFWIDTH_KATAKANA_LETTER_E = 0xFF74;
const char32_t HALFWIDTH_KATAKANA_LETTER_O = 0xFF75;
const char32_t HALFWIDTH_KATAKANA_LETTER_KA = 0xFF76;
const char32_t HALFWIDTH_KATAKANA_LETTER_KI = 0xFF77;
const char32_t HALFWIDTH_KATAKANA_LETTER_KU = 0xFF78;
const char32_t HALFWIDTH_KATAKANA_LETTER_KE = 0xFF79;
const char32_t HALFWIDTH_KATAKANA_LETTER_KO = 0xFF7A;
const char32_t HALFWIDTH_KATAKANA_LETTER_SA = 0xFF7B;
const char32_t HALFWIDTH_KATAKANA_LETTER_SI = 0xFF7C;
const char32_t HALFWIDTH_KATAKANA_LETTER_SU = 0xFF7D;
const char32_t HALFWIDTH_KATAKANA_LETTER_SE = 0xFF7E;
const char32_t HALFWIDTH_KATAKANA_LETTER_SO = 0xFF7F;
const char32_t HALFWIDTH_KATAKANA_LETTER_TA = 0xFF80;
const char32_t HALFWIDTH_KATAKANA_LETTER_TI = 0xFF81;
const char32_t HALFWIDTH_KATAKANA_LETTER_TU = 0xFF82;
const char32_t HALFWIDTH_KATAKANA_LETTER_TE = 0xFF83;
const char32_t HALFWIDTH_KATAKANA_LETTER_TO = 0xFF84;
const char32_t HALFWIDTH_KATAKANA_LETTER_NA = 0xFF85;
const char32_t HALFWIDTH_KATAKANA_LETTER_NI = 0xFF86;
const char32_t HALFWIDTH_KATAKANA_LETTER_NU = 0xFF87;
const char32_t HALFWIDTH_KATAKANA_LETTER_NE = 0xFF88;
const char32_t HALFWIDTH_KATAKANA_LETTER_NO = 0xFF89;
const char32_t HALFWIDTH_KATAKANA_LETTER_HA = 0xFF8A;
const char32_t HALFWIDTH_KATAKANA_LETTER_HI = 0xFF8B;
const char32_t HALFWIDTH_KATAKANA_LETTER_HU = 0xFF8C;
const char32_t HALFWIDTH_KATAKANA_LETTER_HE = 0xFF8D;
const char32_t HALFWIDTH_KATAKANA_LETTER_HO = 0xFF8E;
const char32_t HALFWIDTH_KATAKANA_LETTER_MA = 0xFF8F;
const char32_t HALFWIDTH_KATAKANA_LETTER_MI = 0xFF90;
const char32_t HALFWIDTH_KATAKANA_LETTER_MU = 0xFF91;
const char32_t HALFWIDTH_KATAKANA_LETTER_ME = 0xFF92;
const char32_t HALFWIDTH_KATAKANA_LETTER_MO = 0xFF93;
const char32_t HALFWIDTH_KATAKANA_LETTER_YA = 0xFF94;
const char32_t HALFWIDTH_KATAKANA_LETTER_YU = 0xFF95;
const char32_t HALFWIDTH_KATAKANA_LETTER_YO = 0xFF96;
const char32_t HALFWIDTH_KATAKANA_LETTER_RA = 0xFF97;
const char32_t HALFWIDTH_KATAKANA_LETTER_RI = 0xFF98;
const char32_t HALFWIDTH_KATAKANA_LETTER_RU = 0xFF99;
const char32_t HALFWIDTH_KATAKANA_LETTER_RE = 0xFF9A;
const char32_t HALFWIDTH_KATAKANA_LETTER_RO = 0xFF9B;
const char32_t HALFWIDTH_KATAKANA_LETTER_WA = 0xFF9C;
const char32_t HALFWIDTH_KATAKANA_LETTER_N = 0xFF9D;
const char32_t HALFWIDTH_KATAKANA_VOICED_SOUND_MARK = 0xFF9E;
const char32_t HALFWIDTH_KATAKANA_SEMI_VOICED_SOUND_MARK = 0xFF9F;
const char32_t HALFWIDTH_HANGUL_FILLER = 0xFFA0;
const char32_t HALFWIDTH_HANGUL_LETTER_KIYEOK = 0xFFA1;
const char32_t HALFWIDTH_HANGUL_LETTER_SSANGKIYEOK = 0xFFA2;
const char32_t HALFWIDTH_HANGUL_LETTER_KIYEOK_SIOS = 0xFFA3;
const char32_t HALFWIDTH_HANGUL_LETTER_NIEUN = 0xFFA4;
const char32_t HALFWIDTH_HANGUL_LETTER_NIEUN_CIEUC = 0xFFA5;
const char32_t HALFWIDTH_HANGUL_LETTER_NIEUN_HIEUH = 0xFFA6;
const char32_t HALFWIDTH_HANGUL_LETTER_TIKEUT = 0xFFA7;
const char32_t HALFWIDTH_HANGUL_LETTER_SSANGTIKEUT = 0xFFA8;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL = 0xFFA9;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL_KIYEOK = 0xFFAA;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL_MIEUM = 0xFFAB;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL_PIEUP = 0xFFAC;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL_SIOS = 0xFFAD;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL_THIEUTH = 0xFFAE;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL_PHIEUPH = 0xFFAF;
const char32_t HALFWIDTH_HANGUL_LETTER_RIEUL_HIEUH = 0xFFB0;
const char32_t HALFWIDTH_HANGUL_LETTER_MIEUM = 0xFFB1;
const char32_t HALFWIDTH_HANGUL_LETTER_PIEUP = 0xFFB2;
const char32_t HALFWIDTH_HANGUL_LETTER_SSANGPIEUP = 0xFFB3;
const char32_t HALFWIDTH_HANGUL_LETTER_PIEUP_SIOS = 0xFFB4;
const char32_t HALFWIDTH_HANGUL_LETTER_SIOS = 0xFFB5;
const char32_t HALFWIDTH_HANGUL_LETTER_SSANGSIOS = 0xFFB6;
const char32_t HALFWIDTH_HANGUL_LETTER_IEUNG = 0xFFB7;
const char32_t HALFWIDTH_HANGUL_LETTER_CIEUC = 0xFFB8;
const char32_t HALFWIDTH_HANGUL_LETTER_SSANGCIEUC = 0xFFB9;
const char32_t HALFWIDTH_HANGUL_LETTER_CHIEUCH = 0xFFBA;
const char32_t HALFWIDTH_HANGUL_LETTER_KHIEUKH = 0xFFBB;
const char32_t HALFWIDTH_HANGUL_LETTER_THIEUTH = 0xFFBC;
const char32_t HALFWIDTH_HANGUL_LETTER_PHIEUPH = 0xFFBD;
const char32_t HALFWIDTH_HANGUL_LETTER_HIEUH = 0xFFBE;
const char32_t HALFWIDTH_HANGUL_LETTER_A = 0xFFC2;
const char32_t HALFWIDTH_HANGUL_LETTER_AE = 0xFFC3;
const char32_t HALFWIDTH_HANGUL_LETTER_YA = 0xFFC4;
const char32_t HALFWIDTH_HANGUL_LETTER_YAE = 0xFFC5;
const char32_t HALFWIDTH_HANGUL_LETTER_EO = 0xFFC6;
const char32_t HALFWIDTH_HANGUL_LETTER_E = 0xFFC7;
const char32_t HALFWIDTH_HANGUL_LETTER_YEO = 0xFFCA;
const char32_t HALFWIDTH_HANGUL_LETTER_YE = 0xFFCB;
const char32_t HALFWIDTH_HANGUL_LETTER_O = 0xFFCC;
const char32_t HALFWIDTH_HANGUL_LETTER_WA = 0xFFCD;
const char32_t HALFWIDTH_HANGUL_LETTER_WAE = 0xFFCE;
const char32_t HALFWIDTH_HANGUL_LETTER_OE = 0xFFCF;
const char32_t HALFWIDTH_HANGUL_LETTER_YO = 0xFFD2;
const char32_t HALFWIDTH_HANGUL_LETTER_U = 0xFFD3;
const char32_t HALFWIDTH_HANGUL_LETTER_WEO = 0xFFD4;
const char32_t HALFWIDTH_HANGUL_LETTER_WE = 0xFFD5;
const char32_t HALFWIDTH_HANGUL_LETTER_WI = 0xFFD6;
const char32_t HALFWIDTH_HANGUL_LETTER_YU = 0xFFD7;
const char32_t HALFWIDTH_HANGUL_LETTER_EU = 0xFFDA;
const char32_t HALFWIDTH_HANGUL_LETTER_YI = 0xFFDB;
const char32_t HALFWIDTH_HANGUL_LETTER_I = 0xFFDC;
const char32_t FULLWIDTH_CENT_SIGN = 0xFFE0;
const char32_t FULLWIDTH_POUND_SIGN = 0xFFE1;
const char32_t FULLWIDTH_NOT_SIGN = 0xFFE2;
const char32_t FULLWIDTH_MACRON = 0xFFE3;
const char32_t FULLWIDTH_BROKEN_BAR = 0xFFE4;
const char32_t FULLWIDTH_YEN_SIGN = 0xFFE5;
const char32_t FULLWIDTH_WON_SIGN = 0xFFE6;
const char32_t HALFWIDTH_FORMS_LIGHT_VERTICAL = 0xFFE8;
const char32_t HALFWIDTH_LEFTWARDS_ARROW = 0xFFE9;
const char32_t HALFWIDTH_UPWARDS_ARROW = 0xFFEA;
const char32_t HALFWIDTH_RIGHTWARDS_ARROW = 0xFFEB;
const char32_t HALFWIDTH_DOWNWARDS_ARROW = 0xFFEC;
const char32_t HALFWIDTH_BLACK_SQUARE = 0xFFED;
const char32_t HALFWIDTH_WHITE_CIRCLE = 0xFFEE;
const char32_t INTERLINEAR_ANNOTATION_ANCHOR = 0xFFF9;
const char32_t INTERLINEAR_ANNOTATION_SEPARATOR = 0xFFFA;
const char32_t INTERLINEAR_ANNOTATION_TERMINATOR = 0xFFFB;
const char32_t OBJECT_REPLACEMENT_CHARACTER = 0xFFFC;
const char32_t REPLACEMENT_CHARACTER = 0xFFFD;
const char32_t LINEAR_B_SYLLABLE_B008_A = 0x00010000;
const char32_t LINEAR_B_SYLLABLE_B038_E = 0x00010001;
const char32_t LINEAR_B_SYLLABLE_B028_I = 0x00010002;
const char32_t LINEAR_B_SYLLABLE_B061_O = 0x00010003;
const char32_t LINEAR_B_SYLLABLE_B010_U = 0x00010004;
const char32_t LINEAR_B_SYLLABLE_B001_DA = 0x00010005;
const char32_t LINEAR_B_SYLLABLE_B045_DE = 0x00010006;
const char32_t LINEAR_B_SYLLABLE_B007_DI = 0x00010007;
const char32_t LINEAR_B_SYLLABLE_B014_DO = 0x00010008;
const char32_t LINEAR_B_SYLLABLE_B051_DU = 0x00010009;
const char32_t LINEAR_B_SYLLABLE_B057_JA = 0x0001000A;
const char32_t LINEAR_B_SYLLABLE_B046_JE = 0x0001000B;
const char32_t LINEAR_B_SYLLABLE_B036_JO = 0x0001000D;
const char32_t LINEAR_B_SYLLABLE_B065_JU = 0x0001000E;
const char32_t LINEAR_B_SYLLABLE_B077_KA = 0x0001000F;
const char32_t LINEAR_B_SYLLABLE_B044_KE = 0x00010010;
const char32_t LINEAR_B_SYLLABLE_B067_KI = 0x00010011;
const char32_t LINEAR_B_SYLLABLE_B070_KO = 0x00010012;
const char32_t LINEAR_B_SYLLABLE_B081_KU = 0x00010013;
const char32_t LINEAR_B_SYLLABLE_B080_MA = 0x00010014;
const char32_t LINEAR_B_SYLLABLE_B013_ME = 0x00010015;
const char32_t LINEAR_B_SYLLABLE_B073_MI = 0x00010016;
const char32_t LINEAR_B_SYLLABLE_B015_MO = 0x00010017;
const char32_t LINEAR_B_SYLLABLE_B023_MU = 0x00010018;
const char32_t LINEAR_B_SYLLABLE_B006_NA = 0x00010019;
const char32_t LINEAR_B_SYLLABLE_B024_NE = 0x0001001A;
const char32_t LINEAR_B_SYLLABLE_B030_NI = 0x0001001B;
const char32_t LINEAR_B_SYLLABLE_B052_NO = 0x0001001C;
const char32_t LINEAR_B_SYLLABLE_B055_NU = 0x0001001D;
const char32_t LINEAR_B_SYLLABLE_B003_PA = 0x0001001E;
const char32_t LINEAR_B_SYLLABLE_B072_PE = 0x0001001F;
const char32_t LINEAR_B_SYLLABLE_B039_PI = 0x00010020;
const char32_t LINEAR_B_SYLLABLE_B011_PO = 0x00010021;
const char32_t LINEAR_B_SYLLABLE_B050_PU = 0x00010022;
const char32_t LINEAR_B_SYLLABLE_B016_QA = 0x00010023;
const char32_t LINEAR_B_SYLLABLE_B078_QE = 0x00010024;
const char32_t LINEAR_B_SYLLABLE_B021_QI = 0x00010025;
const char32_t LINEAR_B_SYLLABLE_B032_QO = 0x00010026;
const char32_t LINEAR_B_SYLLABLE_B060_RA = 0x00010028;
const char32_t LINEAR_B_SYLLABLE_B027_RE = 0x00010029;
const char32_t LINEAR_B_SYLLABLE_B053_RI = 0x0001002A;
const char32_t LINEAR_B_SYLLABLE_B002_RO = 0x0001002B;
const char32_t LINEAR_B_SYLLABLE_B026_RU = 0x0001002C;
const char32_t LINEAR_B_SYLLABLE_B031_SA = 0x0001002D;
const char32_t LINEAR_B_SYLLABLE_B009_SE = 0x0001002E;
const char32_t LINEAR_B_SYLLABLE_B041_SI = 0x0001002F;
const char32_t LINEAR_B_SYLLABLE_B012_SO = 0x00010030;
const char32_t LINEAR_B_SYLLABLE_B058_SU = 0x00010031;
const char32_t LINEAR_B_SYLLABLE_B059_TA = 0x00010032;
const char32_t LINEAR_B_SYLLABLE_B004_TE = 0x00010033;
const char32_t LINEAR_B_SYLLABLE_B037_TI = 0x00010034;
const char32_t LINEAR_B_SYLLABLE_B005_TO = 0x00010035;
const char32_t LINEAR_B_SYLLABLE_B069_TU = 0x00010036;
const char32_t LINEAR_B_SYLLABLE_B054_WA = 0x00010037;
const char32_t LINEAR_B_SYLLABLE_B075_WE = 0x00010038;
const char32_t LINEAR_B_SYLLABLE_B040_WI = 0x00010039;
const char32_t LINEAR_B_SYLLABLE_B042_WO = 0x0001003A;
const char32_t LINEAR_B_SYLLABLE_B017_ZA = 0x0001003C;
const char32_t LINEAR_B_SYLLABLE_B074_ZE = 0x0001003D;
const char32_t LINEAR_B_SYLLABLE_B020_ZO = 0x0001003F;
const char32_t LINEAR_B_SYLLABLE_B025_A2 = 0x00010040;
const char32_t LINEAR_B_SYLLABLE_B043_A3 = 0x00010041;
const char32_t LINEAR_B_SYLLABLE_B085_AU = 0x00010042;
const char32_t LINEAR_B_SYLLABLE_B071_DWE = 0x00010043;
const char32_t LINEAR_B_SYLLABLE_B090_DWO = 0x00010044;
const char32_t LINEAR_B_SYLLABLE_B048_NWA = 0x00010045;
const char32_t LINEAR_B_SYLLABLE_B029_PU2 = 0x00010046;
const char32_t LINEAR_B_SYLLABLE_B062_PTE = 0x00010047;
const char32_t LINEAR_B_SYLLABLE_B076_RA2 = 0x00010048;
const char32_t LINEAR_B_SYLLABLE_B033_RA3 = 0x00010049;
const char32_t LINEAR_B_SYLLABLE_B068_RO2 = 0x0001004A;
const char32_t LINEAR_B_SYLLABLE_B066_TA2 = 0x0001004B;
const char32_t LINEAR_B_SYLLABLE_B087_TWE = 0x0001004C;
const char32_t LINEAR_B_SYLLABLE_B091_TWO = 0x0001004D;
const char32_t LINEAR_B_SYMBOL_B018 = 0x00010050;
const char32_t LINEAR_B_SYMBOL_B019 = 0x00010051;
const char32_t LINEAR_B_SYMBOL_B022 = 0x00010052;
const char32_t LINEAR_B_SYMBOL_B034 = 0x00010053;
const char32_t LINEAR_B_SYMBOL_B047 = 0x00010054;
const char32_t LINEAR_B_SYMBOL_B049 = 0x00010055;
const char32_t LINEAR_B_SYMBOL_B056 = 0x00010056;
const char32_t LINEAR_B_SYMBOL_B063 = 0x00010057;
const char32_t LINEAR_B_SYMBOL_B064 = 0x00010058;
const char32_t LINEAR_B_SYMBOL_B079 = 0x00010059;
const char32_t LINEAR_B_SYMBOL_B082 = 0x0001005A;
const char32_t LINEAR_B_SYMBOL_B083 = 0x0001005B;
const char32_t LINEAR_B_SYMBOL_B086 = 0x0001005C;
const char32_t LINEAR_B_SYMBOL_B089 = 0x0001005D;
const char32_t LINEAR_B_IDEOGRAM_B100_MAN = 0x00010080;
const char32_t LINEAR_B_IDEOGRAM_B102_WOMAN = 0x00010081;
const char32_t LINEAR_B_IDEOGRAM_B104_DEER = 0x00010082;
const char32_t LINEAR_B_IDEOGRAM_B105_EQUID = 0x00010083;
const char32_t LINEAR_B_IDEOGRAM_B105F_MARE = 0x00010084;
const char32_t LINEAR_B_IDEOGRAM_B105M_STALLION = 0x00010085;
const char32_t LINEAR_B_IDEOGRAM_B106F_EWE = 0x00010086;
const char32_t LINEAR_B_IDEOGRAM_B106M_RAM = 0x00010087;
const char32_t LINEAR_B_IDEOGRAM_B107F_SHE_GOAT = 0x00010088;
const char32_t LINEAR_B_IDEOGRAM_B107M_HE_GOAT = 0x00010089;
const char32_t LINEAR_B_IDEOGRAM_B108F_SOW = 0x0001008A;
const char32_t LINEAR_B_IDEOGRAM_B108M_BOAR = 0x0001008B;
const char32_t LINEAR_B_IDEOGRAM_B109F_COW = 0x0001008C;
const char32_t LINEAR_B_IDEOGRAM_B109M_BULL = 0x0001008D;
const char32_t LINEAR_B_IDEOGRAM_B120_WHEAT = 0x0001008E;
const char32_t LINEAR_B_IDEOGRAM_B121_BARLEY = 0x0001008F;
const char32_t LINEAR_B_IDEOGRAM_B122_OLIVE = 0x00010090;
const char32_t LINEAR_B_IDEOGRAM_B123_SPICE = 0x00010091;
const char32_t LINEAR_B_IDEOGRAM_B125_CYPERUS = 0x00010092;
const char32_t LINEAR_B_MONOGRAM_B127_KAPO = 0x00010093;
const char32_t LINEAR_B_MONOGRAM_B128_KANAKO = 0x00010094;
const char32_t LINEAR_B_IDEOGRAM_B130_OIL = 0x00010095;
const char32_t LINEAR_B_IDEOGRAM_B131_WINE = 0x00010096;
const char32_t LINEAR_B_IDEOGRAM_B132 = 0x00010097;
const char32_t LINEAR_B_MONOGRAM_B133_AREPA = 0x00010098;
const char32_t LINEAR_B_MONOGRAM_B135_MERI = 0x00010099;
const char32_t LINEAR_B_IDEOGRAM_B140_BRONZE = 0x0001009A;
const char32_t LINEAR_B_IDEOGRAM_B141_GOLD = 0x0001009B;
const char32_t LINEAR_B_IDEOGRAM_B142 = 0x0001009C;
const char32_t LINEAR_B_IDEOGRAM_B145_WOOL = 0x0001009D;
const char32_t LINEAR_B_IDEOGRAM_B146 = 0x0001009E;
const char32_t LINEAR_B_IDEOGRAM_B150 = 0x0001009F;
const char32_t LINEAR_B_IDEOGRAM_B151_HORN = 0x000100A0;
const char32_t LINEAR_B_IDEOGRAM_B152 = 0x000100A1;
const char32_t LINEAR_B_IDEOGRAM_B153 = 0x000100A2;
const char32_t LINEAR_B_IDEOGRAM_B154 = 0x000100A3;
const char32_t LINEAR_B_MONOGRAM_B156_TURO2 = 0x000100A4;
const char32_t LINEAR_B_IDEOGRAM_B157 = 0x000100A5;
const char32_t LINEAR_B_IDEOGRAM_B158 = 0x000100A6;
const char32_t LINEAR_B_IDEOGRAM_B159_CLOTH = 0x000100A7;
const char32_t LINEAR_B_IDEOGRAM_B160 = 0x000100A8;
const char32_t LINEAR_B_IDEOGRAM_B161 = 0x000100A9;
const char32_t LINEAR_B_IDEOGRAM_B162_GARMENT = 0x000100AA;
const char32_t LINEAR_B_IDEOGRAM_B163_ARMOUR = 0x000100AB;
const char32_t LINEAR_B_IDEOGRAM_B164 = 0x000100AC;
const char32_t LINEAR_B_IDEOGRAM_B165 = 0x000100AD;
const char32_t LINEAR_B_IDEOGRAM_B166 = 0x000100AE;
const char32_t LINEAR_B_IDEOGRAM_B167 = 0x000100AF;
const char32_t LINEAR_B_IDEOGRAM_B168 = 0x000100B0;
const char32_t LINEAR_B_IDEOGRAM_B169 = 0x000100B1;
const char32_t LINEAR_B_IDEOGRAM_B170 = 0x000100B2;
const char32_t LINEAR_B_IDEOGRAM_B171 = 0x000100B3;
const char32_t LINEAR_B_IDEOGRAM_B172 = 0x000100B4;
const char32_t LINEAR_B_IDEOGRAM_B173_MONTH = 0x000100B5;
const char32_t LINEAR_B_IDEOGRAM_B174 = 0x000100B6;
const char32_t LINEAR_B_IDEOGRAM_B176_TREE = 0x000100B7;
const char32_t LINEAR_B_IDEOGRAM_B177 = 0x000100B8;
const char32_t LINEAR_B_IDEOGRAM_B178 = 0x000100B9;
const char32_t LINEAR_B_IDEOGRAM_B179 = 0x000100BA;
const char32_t LINEAR_B_IDEOGRAM_B180 = 0x000100BB;
const char32_t LINEAR_B_IDEOGRAM_B181 = 0x000100BC;
const char32_t LINEAR_B_IDEOGRAM_B182 = 0x000100BD;
const char32_t LINEAR_B_IDEOGRAM_B183 = 0x000100BE;
const char32_t LINEAR_B_IDEOGRAM_B184 = 0x000100BF;
const char32_t LINEAR_B_IDEOGRAM_B185 = 0x000100C0;
const char32_t LINEAR_B_IDEOGRAM_B189 = 0x000100C1;
const char32_t LINEAR_B_IDEOGRAM_B190 = 0x000100C2;
const char32_t LINEAR_B_IDEOGRAM_B191_HELMET = 0x000100C3;
const char32_t LINEAR_B_IDEOGRAM_B220_FOOTSTOOL = 0x000100C4;
const char32_t LINEAR_B_IDEOGRAM_B225_BATHTUB = 0x000100C5;
const char32_t LINEAR_B_IDEOGRAM_B230_SPEAR = 0x000100C6;
const char32_t LINEAR_B_IDEOGRAM_B231_ARROW = 0x000100C7;
const char32_t LINEAR_B_IDEOGRAM_B232 = 0x000100C8;
const char32_t LINEAR_B_IDEOGRAM_B233_SWORD = 0x000100C9;
const char32_t LINEAR_B_IDEOGRAM_B234 = 0x000100CA;
const char32_t LINEAR_B_IDEOGRAM_B236 = 0x000100CB;
const char32_t LINEAR_B_IDEOGRAM_B240_WHEELED_CHARIOT = 0x000100CC;
const char32_t LINEAR_B_IDEOGRAM_B241_CHARIOT = 0x000100CD;
const char32_t LINEAR_B_IDEOGRAM_B242_CHARIOT_FRAME = 0x000100CE;
const char32_t LINEAR_B_IDEOGRAM_B243_WHEEL = 0x000100CF;
const char32_t LINEAR_B_IDEOGRAM_B245 = 0x000100D0;
const char32_t LINEAR_B_IDEOGRAM_B246 = 0x000100D1;
const char32_t LINEAR_B_MONOGRAM_B247_DIPTE = 0x000100D2;
const char32_t LINEAR_B_IDEOGRAM_B248 = 0x000100D3;
const char32_t LINEAR_B_IDEOGRAM_B249 = 0x000100D4;
const char32_t LINEAR_B_IDEOGRAM_B251 = 0x000100D5;
const char32_t LINEAR_B_IDEOGRAM_B252 = 0x000100D6;
const char32_t LINEAR_B_IDEOGRAM_B253 = 0x000100D7;
const char32_t LINEAR_B_IDEOGRAM_B254_DART = 0x000100D8;
const char32_t LINEAR_B_IDEOGRAM_B255 = 0x000100D9;
const char32_t LINEAR_B_IDEOGRAM_B256 = 0x000100DA;
const char32_t LINEAR_B_IDEOGRAM_B257 = 0x000100DB;
const char32_t LINEAR_B_IDEOGRAM_B258 = 0x000100DC;
const char32_t LINEAR_B_IDEOGRAM_B259 = 0x000100DD;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B155 = 0x000100DE;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B200 = 0x000100DF;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B201 = 0x000100E0;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B202 = 0x000100E1;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B203 = 0x000100E2;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B204 = 0x000100E3;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B205 = 0x000100E4;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B206 = 0x000100E5;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B207 = 0x000100E6;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B208 = 0x000100E7;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B209 = 0x000100E8;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B210 = 0x000100E9;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B211 = 0x000100EA;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B212 = 0x000100EB;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B213 = 0x000100EC;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B214 = 0x000100ED;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B215 = 0x000100EE;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B216 = 0x000100EF;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B217 = 0x000100F0;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B218 = 0x000100F1;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B219 = 0x000100F2;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B221 = 0x000100F3;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B222 = 0x000100F4;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B226 = 0x000100F5;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B227 = 0x000100F6;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B228 = 0x000100F7;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B229 = 0x000100F8;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B250 = 0x000100F9;
const char32_t LINEAR_B_IDEOGRAM_VESSEL_B305 = 0x000100FA;
const char32_t AEGEAN_WORD_SEPARATOR_LINE = 0x00010100;
const char32_t AEGEAN_WORD_SEPARATOR_DOT = 0x00010101;
const char32_t AEGEAN_CHECK_MARK = 0x00010102;
const char32_t AEGEAN_NUMBER_ONE = 0x00010107;
const char32_t AEGEAN_NUMBER_TWO = 0x00010108;
const char32_t AEGEAN_NUMBER_THREE = 0x00010109;
const char32_t AEGEAN_NUMBER_FOUR = 0x0001010A;
const char32_t AEGEAN_NUMBER_FIVE = 0x0001010B;
const char32_t AEGEAN_NUMBER_SIX = 0x0001010C;
const char32_t AEGEAN_NUMBER_SEVEN = 0x0001010D;
const char32_t AEGEAN_NUMBER_EIGHT = 0x0001010E;
const char32_t AEGEAN_NUMBER_NINE = 0x0001010F;
const char32_t AEGEAN_NUMBER_TEN = 0x00010110;
const char32_t AEGEAN_NUMBER_TWENTY = 0x00010111;
const char32_t AEGEAN_NUMBER_THIRTY = 0x00010112;
const char32_t AEGEAN_NUMBER_FORTY = 0x00010113;
const char32_t AEGEAN_NUMBER_FIFTY = 0x00010114;
const char32_t AEGEAN_NUMBER_SIXTY = 0x00010115;
const char32_t AEGEAN_NUMBER_SEVENTY = 0x00010116;
const char32_t AEGEAN_NUMBER_EIGHTY = 0x00010117;
const char32_t AEGEAN_NUMBER_NINETY = 0x00010118;
const char32_t AEGEAN_NUMBER_ONE_HUNDRED = 0x00010119;
const char32_t AEGEAN_NUMBER_TWO_HUNDRED = 0x0001011A;
const char32_t AEGEAN_NUMBER_THREE_HUNDRED = 0x0001011B;
const char32_t AEGEAN_NUMBER_FOUR_HUNDRED = 0x0001011C;
const char32_t AEGEAN_NUMBER_FIVE_HUNDRED = 0x0001011D;
const char32_t AEGEAN_NUMBER_SIX_HUNDRED = 0x0001011E;
const char32_t AEGEAN_NUMBER_SEVEN_HUNDRED = 0x0001011F;
const char32_t AEGEAN_NUMBER_EIGHT_HUNDRED = 0x00010120;
const char32_t AEGEAN_NUMBER_NINE_HUNDRED = 0x00010121;
const char32_t AEGEAN_NUMBER_ONE_THOUSAND = 0x00010122;
const char32_t AEGEAN_NUMBER_TWO_THOUSAND = 0x00010123;
const char32_t AEGEAN_NUMBER_THREE_THOUSAND = 0x00010124;
const char32_t AEGEAN_NUMBER_FOUR_THOUSAND = 0x00010125;
const char32_t AEGEAN_NUMBER_FIVE_THOUSAND = 0x00010126;
const char32_t AEGEAN_NUMBER_SIX_THOUSAND = 0x00010127;
const char32_t AEGEAN_NUMBER_SEVEN_THOUSAND = 0x00010128;
const char32_t AEGEAN_NUMBER_EIGHT_THOUSAND = 0x00010129;
const char32_t AEGEAN_NUMBER_NINE_THOUSAND = 0x0001012A;
const char32_t AEGEAN_NUMBER_TEN_THOUSAND = 0x0001012B;
const char32_t AEGEAN_NUMBER_TWENTY_THOUSAND = 0x0001012C;
const char32_t AEGEAN_NUMBER_THIRTY_THOUSAND = 0x0001012D;
const char32_t AEGEAN_NUMBER_FORTY_THOUSAND = 0x0001012E;
const char32_t AEGEAN_NUMBER_FIFTY_THOUSAND = 0x0001012F;
const char32_t AEGEAN_NUMBER_SIXTY_THOUSAND = 0x00010130;
const char32_t AEGEAN_NUMBER_SEVENTY_THOUSAND = 0x00010131;
const char32_t AEGEAN_NUMBER_EIGHTY_THOUSAND = 0x00010132;
const char32_t AEGEAN_NUMBER_NINETY_THOUSAND = 0x00010133;
const char32_t AEGEAN_WEIGHT_BASE_UNIT = 0x00010137;
const char32_t AEGEAN_WEIGHT_FIRST_SUBUNIT = 0x00010138;
const char32_t AEGEAN_WEIGHT_SECOND_SUBUNIT = 0x00010139;
const char32_t AEGEAN_WEIGHT_THIRD_SUBUNIT = 0x0001013A;
const char32_t AEGEAN_WEIGHT_FOURTH_SUBUNIT = 0x0001013B;
const char32_t AEGEAN_DRY_MEASURE_FIRST_SUBUNIT = 0x0001013C;
const char32_t AEGEAN_LIQUID_MEASURE_FIRST_SUBUNIT = 0x0001013D;
const char32_t AEGEAN_MEASURE_SECOND_SUBUNIT = 0x0001013E;
const char32_t AEGEAN_MEASURE_THIRD_SUBUNIT = 0x0001013F;
const char32_t GREEK_ACROPHONIC_ATTIC_ONE_QUARTER = 0x00010140;
const char32_t GREEK_ACROPHONIC_ATTIC_ONE_HALF = 0x00010141;
const char32_t GREEK_ACROPHONIC_ATTIC_ONE_DRACHMA = 0x00010142;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE = 0x00010143;
const char32_t GREEK_ACROPHONIC_ATTIC_FIFTY = 0x00010144;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE_HUNDRED = 0x00010145;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE_THOUSAND = 0x00010146;
const char32_t GREEK_ACROPHONIC_ATTIC_FIFTY_THOUSAND = 0x00010147;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE_TALENTS = 0x00010148;
const char32_t GREEK_ACROPHONIC_ATTIC_TEN_TALENTS = 0x00010149;
const char32_t GREEK_ACROPHONIC_ATTIC_FIFTY_TALENTS = 0x0001014A;
const char32_t GREEK_ACROPHONIC_ATTIC_ONE_HUNDRED_TALENTS = 0x0001014B;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE_HUNDRED_TALENTS = 0x0001014C;
const char32_t GREEK_ACROPHONIC_ATTIC_ONE_THOUSAND_TALENTS = 0x0001014D;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE_THOUSAND_TALENTS = 0x0001014E;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE_STATERS = 0x0001014F;
const char32_t GREEK_ACROPHONIC_ATTIC_TEN_STATERS = 0x00010150;
const char32_t GREEK_ACROPHONIC_ATTIC_FIFTY_STATERS = 0x00010151;
const char32_t GREEK_ACROPHONIC_ATTIC_ONE_HUNDRED_STATERS = 0x00010152;
const char32_t GREEK_ACROPHONIC_ATTIC_FIVE_HUNDRED_STATERS = 0x00010153;
const char32_t GREEK_ACROPHONIC_ATTIC_ONE_THOUSAND_STATERS = 0x00010154;
const char32_t GREEK_ACROPHONIC_ATTIC_TEN_THOUSAND_STATERS = 0x00010155;
const char32_t GREEK_ACROPHONIC_ATTIC_FIFTY_THOUSAND_STATERS = 0x00010156;
const char32_t GREEK_ACROPHONIC_ATTIC_TEN_MNAS = 0x00010157;
const char32_t GREEK_ACROPHONIC_HERAEUM_ONE_PLETHRON = 0x00010158;
const char32_t GREEK_ACROPHONIC_THESPIAN_ONE = 0x00010159;
const char32_t GREEK_ACROPHONIC_HERMIONIAN_ONE = 0x0001015A;
const char32_t GREEK_ACROPHONIC_EPIDAUREAN_TWO = 0x0001015B;
const char32_t GREEK_ACROPHONIC_THESPIAN_TWO = 0x0001015C;
const char32_t GREEK_ACROPHONIC_CYRENAIC_TWO_DRACHMAS = 0x0001015D;
const char32_t GREEK_ACROPHONIC_EPIDAUREAN_TWO_DRACHMAS = 0x0001015E;
const char32_t GREEK_ACROPHONIC_TROEZENIAN_FIVE = 0x0001015F;
const char32_t GREEK_ACROPHONIC_TROEZENIAN_TEN = 0x00010160;
const char32_t GREEK_ACROPHONIC_TROEZENIAN_TEN_ALTERNATE_FORM = 0x00010161;
const char32_t GREEK_ACROPHONIC_HERMIONIAN_TEN = 0x00010162;
const char32_t GREEK_ACROPHONIC_MESSENIAN_TEN = 0x00010163;
const char32_t GREEK_ACROPHONIC_THESPIAN_TEN = 0x00010164;
const char32_t GREEK_ACROPHONIC_THESPIAN_THIRTY = 0x00010165;
const char32_t GREEK_ACROPHONIC_TROEZENIAN_FIFTY = 0x00010166;
const char32_t GREEK_ACROPHONIC_TROEZENIAN_FIFTY_ALTERNATE_FORM = 0x00010167;
const char32_t GREEK_ACROPHONIC_HERMIONIAN_FIFTY = 0x00010168;
const char32_t GREEK_ACROPHONIC_THESPIAN_FIFTY = 0x00010169;
const char32_t GREEK_ACROPHONIC_THESPIAN_ONE_HUNDRED = 0x0001016A;
const char32_t GREEK_ACROPHONIC_THESPIAN_THREE_HUNDRED = 0x0001016B;
const char32_t GREEK_ACROPHONIC_EPIDAUREAN_FIVE_HUNDRED = 0x0001016C;
const char32_t GREEK_ACROPHONIC_TROEZENIAN_FIVE_HUNDRED = 0x0001016D;
const char32_t GREEK_ACROPHONIC_THESPIAN_FIVE_HUNDRED = 0x0001016E;
const char32_t GREEK_ACROPHONIC_CARYSTIAN_FIVE_HUNDRED = 0x0001016F;
const char32_t GREEK_ACROPHONIC_NAXIAN_FIVE_HUNDRED = 0x00010170;
const char32_t GREEK_ACROPHONIC_THESPIAN_ONE_THOUSAND = 0x00010171;
const char32_t GREEK_ACROPHONIC_THESPIAN_FIVE_THOUSAND = 0x00010172;
const char32_t GREEK_ACROPHONIC_DELPHIC_FIVE_MNAS = 0x00010173;
const char32_t GREEK_ACROPHONIC_STRATIAN_FIFTY_MNAS = 0x00010174;
const char32_t GREEK_ONE_HALF_SIGN = 0x00010175;
const char32_t GREEK_ONE_HALF_SIGN_ALTERNATE_FORM = 0x00010176;
const char32_t GREEK_TWO_THIRDS_SIGN = 0x00010177;
const char32_t GREEK_THREE_QUARTERS_SIGN = 0x00010178;
const char32_t GREEK_YEAR_SIGN = 0x00010179;
const char32_t GREEK_TALENT_SIGN = 0x0001017A;
const char32_t GREEK_DRACHMA_SIGN = 0x0001017B;
const char32_t GREEK_OBOL_SIGN = 0x0001017C;
const char32_t GREEK_TWO_OBOLS_SIGN = 0x0001017D;
const char32_t GREEK_THREE_OBOLS_SIGN = 0x0001017E;
const char32_t GREEK_FOUR_OBOLS_SIGN = 0x0001017F;
const char32_t GREEK_FIVE_OBOLS_SIGN = 0x00010180;
const char32_t GREEK_METRETES_SIGN = 0x00010181;
const char32_t GREEK_KYATHOS_BASE_SIGN = 0x00010182;
const char32_t GREEK_LITRA_SIGN = 0x00010183;
const char32_t GREEK_OUNKIA_SIGN = 0x00010184;
const char32_t GREEK_XESTES_SIGN = 0x00010185;
const char32_t GREEK_ARTABE_SIGN = 0x00010186;
const char32_t GREEK_AROURA_SIGN = 0x00010187;
const char32_t GREEK_GRAMMA_SIGN = 0x00010188;
const char32_t GREEK_TRYBLION_BASE_SIGN = 0x00010189;
const char32_t GREEK_ZERO_SIGN = 0x0001018A;
const char32_t GREEK_ONE_QUARTER_SIGN = 0x0001018B;
const char32_t GREEK_SINUSOID_SIGN = 0x0001018C;
const char32_t GREEK_INDICTION_SIGN = 0x0001018D;
const char32_t NOMISMA_SIGN = 0x0001018E;
const char32_t ROMAN_SEXTANS_SIGN = 0x00010190;
const char32_t ROMAN_UNCIA_SIGN = 0x00010191;
const char32_t ROMAN_SEMUNCIA_SIGN = 0x00010192;
const char32_t ROMAN_SEXTULA_SIGN = 0x00010193;
const char32_t ROMAN_DIMIDIA_SEXTULA_SIGN = 0x00010194;
const char32_t ROMAN_SILIQUA_SIGN = 0x00010195;
const char32_t ROMAN_DENARIUS_SIGN = 0x00010196;
const char32_t ROMAN_QUINARIUS_SIGN = 0x00010197;
const char32_t ROMAN_SESTERTIUS_SIGN = 0x00010198;
const char32_t ROMAN_DUPONDIUS_SIGN = 0x00010199;
const char32_t ROMAN_AS_SIGN = 0x0001019A;
const char32_t ROMAN_CENTURIAL_SIGN = 0x0001019B;
const char32_t ASCIA_SYMBOL = 0x0001019C;
const char32_t GREEK_SYMBOL_TAU_RHO = 0x000101A0;
const char32_t PHAISTOS_DISC_SIGN_PEDESTRIAN = 0x000101D0;
const char32_t PHAISTOS_DISC_SIGN_PLUMED_HEAD = 0x000101D1;
const char32_t PHAISTOS_DISC_SIGN_TATTOOED_HEAD = 0x000101D2;
const char32_t PHAISTOS_DISC_SIGN_CAPTIVE = 0x000101D3;
const char32_t PHAISTOS_DISC_SIGN_CHILD = 0x000101D4;
const char32_t PHAISTOS_DISC_SIGN_WOMAN = 0x000101D5;
const char32_t PHAISTOS_DISC_SIGN_HELMET = 0x000101D6;
const char32_t PHAISTOS_DISC_SIGN_GAUNTLET = 0x000101D7;
const char32_t PHAISTOS_DISC_SIGN_TIARA = 0x000101D8;
const char32_t PHAISTOS_DISC_SIGN_ARROW = 0x000101D9;
const char32_t PHAISTOS_DISC_SIGN_BOW = 0x000101DA;
const char32_t PHAISTOS_DISC_SIGN_SHIELD = 0x000101DB;
const char32_t PHAISTOS_DISC_SIGN_CLUB = 0x000101DC;
const char32_t PHAISTOS_DISC_SIGN_MANACLES = 0x000101DD;
const char32_t PHAISTOS_DISC_SIGN_MATTOCK = 0x000101DE;
const char32_t PHAISTOS_DISC_SIGN_SAW = 0x000101DF;
const char32_t PHAISTOS_DISC_SIGN_LID = 0x000101E0;
const char32_t PHAISTOS_DISC_SIGN_BOOMERANG = 0x000101E1;
const char32_t PHAISTOS_DISC_SIGN_CARPENTRY_PLANE = 0x000101E2;
const char32_t PHAISTOS_DISC_SIGN_DOLIUM = 0x000101E3;
const char32_t PHAISTOS_DISC_SIGN_COMB = 0x000101E4;
const char32_t PHAISTOS_DISC_SIGN_SLING = 0x000101E5;
const char32_t PHAISTOS_DISC_SIGN_COLUMN = 0x000101E6;
const char32_t PHAISTOS_DISC_SIGN_BEEHIVE = 0x000101E7;
const char32_t PHAISTOS_DISC_SIGN_SHIP = 0x000101E8;
const char32_t PHAISTOS_DISC_SIGN_HORN = 0x000101E9;
const char32_t PHAISTOS_DISC_SIGN_HIDE = 0x000101EA;
const char32_t PHAISTOS_DISC_SIGN_BULLS_LEG = 0x000101EB;
const char32_t PHAISTOS_DISC_SIGN_CAT = 0x000101EC;
const char32_t PHAISTOS_DISC_SIGN_RAM = 0x000101ED;
const char32_t PHAISTOS_DISC_SIGN_EAGLE = 0x000101EE;
const char32_t PHAISTOS_DISC_SIGN_DOVE = 0x000101EF;
const char32_t PHAISTOS_DISC_SIGN_TUNNY = 0x000101F0;
const char32_t PHAISTOS_DISC_SIGN_BEE = 0x000101F1;
const char32_t PHAISTOS_DISC_SIGN_PLANE_TREE = 0x000101F2;
const char32_t PHAISTOS_DISC_SIGN_VINE = 0x000101F3;
const char32_t PHAISTOS_DISC_SIGN_PAPYRUS = 0x000101F4;
const char32_t PHAISTOS_DISC_SIGN_ROSETTE = 0x000101F5;
const char32_t PHAISTOS_DISC_SIGN_LILY = 0x000101F6;
const char32_t PHAISTOS_DISC_SIGN_OX_BACK = 0x000101F7;
const char32_t PHAISTOS_DISC_SIGN_FLUTE = 0x000101F8;
const char32_t PHAISTOS_DISC_SIGN_GRATER = 0x000101F9;
const char32_t PHAISTOS_DISC_SIGN_STRAINER = 0x000101FA;
const char32_t PHAISTOS_DISC_SIGN_SMALL_AXE = 0x000101FB;
const char32_t PHAISTOS_DISC_SIGN_WAVY_BAND = 0x000101FC;
const char32_t PHAISTOS_DISC_SIGN_COMBINING_OBLIQUE_STROKE = 0x000101FD;
const char32_t LYCIAN_LETTER_A = 0x00010280;
const char32_t LYCIAN_LETTER_E = 0x00010281;
const char32_t LYCIAN_LETTER_B = 0x00010282;
const char32_t LYCIAN_LETTER_BH = 0x00010283;
const char32_t LYCIAN_LETTER_G = 0x00010284;
const char32_t LYCIAN_LETTER_D = 0x00010285;
const char32_t LYCIAN_LETTER_I = 0x00010286;
const char32_t LYCIAN_LETTER_W = 0x00010287;
const char32_t LYCIAN_LETTER_Z = 0x00010288;
const char32_t LYCIAN_LETTER_TH = 0x00010289;
const char32_t LYCIAN_LETTER_J = 0x0001028A;
const char32_t LYCIAN_LETTER_K = 0x0001028B;
const char32_t LYCIAN_LETTER_Q = 0x0001028C;
const char32_t LYCIAN_LETTER_L = 0x0001028D;
const char32_t LYCIAN_LETTER_M = 0x0001028E;
const char32_t LYCIAN_LETTER_N = 0x0001028F;
const char32_t LYCIAN_LETTER_MM = 0x00010290;
const char32_t LYCIAN_LETTER_NN = 0x00010291;
const char32_t LYCIAN_LETTER_U = 0x00010292;
const char32_t LYCIAN_LETTER_P = 0x00010293;
const char32_t LYCIAN_LETTER_KK = 0x00010294;
const char32_t LYCIAN_LETTER_R = 0x00010295;
const char32_t LYCIAN_LETTER_S = 0x00010296;
const char32_t LYCIAN_LETTER_T = 0x00010297;
const char32_t LYCIAN_LETTER_TT = 0x00010298;
const char32_t LYCIAN_LETTER_AN = 0x00010299;
const char32_t LYCIAN_LETTER_EN = 0x0001029A;
const char32_t LYCIAN_LETTER_H = 0x0001029B;
const char32_t LYCIAN_LETTER_X = 0x0001029C;
const char32_t CARIAN_LETTER_A = 0x000102A0;
const char32_t CARIAN_LETTER_P2 = 0x000102A1;
const char32_t CARIAN_LETTER_D = 0x000102A2;
const char32_t CARIAN_LETTER_L = 0x000102A3;
const char32_t CARIAN_LETTER_UUU = 0x000102A4;
const char32_t CARIAN_LETTER_R = 0x000102A5;
const char32_t CARIAN_LETTER_LD = 0x000102A6;
const char32_t CARIAN_LETTER_A2 = 0x000102A7;
const char32_t CARIAN_LETTER_Q = 0x000102A8;
const char32_t CARIAN_LETTER_B = 0x000102A9;
const char32_t CARIAN_LETTER_M = 0x000102AA;
const char32_t CARIAN_LETTER_O = 0x000102AB;
const char32_t CARIAN_LETTER_D2 = 0x000102AC;
const char32_t CARIAN_LETTER_T = 0x000102AD;
const char32_t CARIAN_LETTER_SH = 0x000102AE;
const char32_t CARIAN_LETTER_SH2 = 0x000102AF;
const char32_t CARIAN_LETTER_S = 0x000102B0;
const char32_t CARIAN_LETTER_C_18 = 0x000102B1;
const char32_t CARIAN_LETTER_U = 0x000102B2;
const char32_t CARIAN_LETTER_NN = 0x000102B3;
const char32_t CARIAN_LETTER_X = 0x000102B4;
const char32_t CARIAN_LETTER_N = 0x000102B5;
const char32_t CARIAN_LETTER_TT2 = 0x000102B6;
const char32_t CARIAN_LETTER_P = 0x000102B7;
const char32_t CARIAN_LETTER_SS = 0x000102B8;
const char32_t CARIAN_LETTER_I = 0x000102B9;
const char32_t CARIAN_LETTER_E = 0x000102BA;
const char32_t CARIAN_LETTER_UUUU = 0x000102BB;
const char32_t CARIAN_LETTER_K = 0x000102BC;
const char32_t CARIAN_LETTER_K2 = 0x000102BD;
const char32_t CARIAN_LETTER_ND = 0x000102BE;
const char32_t CARIAN_LETTER_UU = 0x000102BF;
const char32_t CARIAN_LETTER_G = 0x000102C0;
const char32_t CARIAN_LETTER_G2 = 0x000102C1;
const char32_t CARIAN_LETTER_ST = 0x000102C2;
const char32_t CARIAN_LETTER_ST2 = 0x000102C3;
const char32_t CARIAN_LETTER_NG = 0x000102C4;
const char32_t CARIAN_LETTER_II = 0x000102C5;
const char32_t CARIAN_LETTER_C_39 = 0x000102C6;
const char32_t CARIAN_LETTER_TT = 0x000102C7;
const char32_t CARIAN_LETTER_UUU2 = 0x000102C8;
const char32_t CARIAN_LETTER_RR = 0x000102C9;
const char32_t CARIAN_LETTER_MB = 0x000102CA;
const char32_t CARIAN_LETTER_MB2 = 0x000102CB;
const char32_t CARIAN_LETTER_MB3 = 0x000102CC;
const char32_t CARIAN_LETTER_MB4 = 0x000102CD;
const char32_t CARIAN_LETTER_LD2 = 0x000102CE;
const char32_t CARIAN_LETTER_E2 = 0x000102CF;
const char32_t CARIAN_LETTER_UUU3 = 0x000102D0;
const char32_t COPTIC_EPACT_THOUSANDS_MARK = 0x000102E0;
const char32_t COPTIC_EPACT_DIGIT_ONE = 0x000102E1;
const char32_t COPTIC_EPACT_DIGIT_TWO = 0x000102E2;
const char32_t COPTIC_EPACT_DIGIT_THREE = 0x000102E3;
const char32_t COPTIC_EPACT_DIGIT_FOUR = 0x000102E4;
const char32_t COPTIC_EPACT_DIGIT_FIVE = 0x000102E5;
const char32_t COPTIC_EPACT_DIGIT_SIX = 0x000102E6;
const char32_t COPTIC_EPACT_DIGIT_SEVEN = 0x000102E7;
const char32_t COPTIC_EPACT_DIGIT_EIGHT = 0x000102E8;
const char32_t COPTIC_EPACT_DIGIT_NINE = 0x000102E9;
const char32_t COPTIC_EPACT_NUMBER_TEN = 0x000102EA;
const char32_t COPTIC_EPACT_NUMBER_TWENTY = 0x000102EB;
const char32_t COPTIC_EPACT_NUMBER_THIRTY = 0x000102EC;
const char32_t COPTIC_EPACT_NUMBER_FORTY = 0x000102ED;
const char32_t COPTIC_EPACT_NUMBER_FIFTY = 0x000102EE;
const char32_t COPTIC_EPACT_NUMBER_SIXTY = 0x000102EF;
const char32_t COPTIC_EPACT_NUMBER_SEVENTY = 0x000102F0;
const char32_t COPTIC_EPACT_NUMBER_EIGHTY = 0x000102F1;
const char32_t COPTIC_EPACT_NUMBER_NINETY = 0x000102F2;
const char32_t COPTIC_EPACT_NUMBER_ONE_HUNDRED = 0x000102F3;
const char32_t COPTIC_EPACT_NUMBER_TWO_HUNDRED = 0x000102F4;
const char32_t COPTIC_EPACT_NUMBER_THREE_HUNDRED = 0x000102F5;
const char32_t COPTIC_EPACT_NUMBER_FOUR_HUNDRED = 0x000102F6;
const char32_t COPTIC_EPACT_NUMBER_FIVE_HUNDRED = 0x000102F7;
const char32_t COPTIC_EPACT_NUMBER_SIX_HUNDRED = 0x000102F8;
const char32_t COPTIC_EPACT_NUMBER_SEVEN_HUNDRED = 0x000102F9;
const char32_t COPTIC_EPACT_NUMBER_EIGHT_HUNDRED = 0x000102FA;
const char32_t COPTIC_EPACT_NUMBER_NINE_HUNDRED = 0x000102FB;
const char32_t OLD_ITALIC_LETTER_A = 0x00010300;
const char32_t OLD_ITALIC_LETTER_BE = 0x00010301;
const char32_t OLD_ITALIC_LETTER_KE = 0x00010302;
const char32_t OLD_ITALIC_LETTER_DE = 0x00010303;
const char32_t OLD_ITALIC_LETTER_E = 0x00010304;
const char32_t OLD_ITALIC_LETTER_VE = 0x00010305;
const char32_t OLD_ITALIC_LETTER_ZE = 0x00010306;
const char32_t OLD_ITALIC_LETTER_HE = 0x00010307;
const char32_t OLD_ITALIC_LETTER_THE = 0x00010308;
const char32_t OLD_ITALIC_LETTER_I = 0x00010309;
const char32_t OLD_ITALIC_LETTER_KA = 0x0001030A;
const char32_t OLD_ITALIC_LETTER_EL = 0x0001030B;
const char32_t OLD_ITALIC_LETTER_EM = 0x0001030C;
const char32_t OLD_ITALIC_LETTER_EN = 0x0001030D;
const char32_t OLD_ITALIC_LETTER_ESH = 0x0001030E;
const char32_t OLD_ITALIC_LETTER_O = 0x0001030F;
const char32_t OLD_ITALIC_LETTER_PE = 0x00010310;
const char32_t OLD_ITALIC_LETTER_SHE = 0x00010311;
const char32_t OLD_ITALIC_LETTER_KU = 0x00010312;
const char32_t OLD_ITALIC_LETTER_ER = 0x00010313;
const char32_t OLD_ITALIC_LETTER_ES = 0x00010314;
const char32_t OLD_ITALIC_LETTER_TE = 0x00010315;
const char32_t OLD_ITALIC_LETTER_U = 0x00010316;
const char32_t OLD_ITALIC_LETTER_EKS = 0x00010317;
const char32_t OLD_ITALIC_LETTER_PHE = 0x00010318;
const char32_t OLD_ITALIC_LETTER_KHE = 0x00010319;
const char32_t OLD_ITALIC_LETTER_EF = 0x0001031A;
const char32_t OLD_ITALIC_LETTER_ERS = 0x0001031B;
const char32_t OLD_ITALIC_LETTER_CHE = 0x0001031C;
const char32_t OLD_ITALIC_LETTER_II = 0x0001031D;
const char32_t OLD_ITALIC_LETTER_UU = 0x0001031E;
const char32_t OLD_ITALIC_LETTER_ESS = 0x0001031F;
const char32_t OLD_ITALIC_NUMERAL_ONE = 0x00010320;
const char32_t OLD_ITALIC_NUMERAL_FIVE = 0x00010321;
const char32_t OLD_ITALIC_NUMERAL_TEN = 0x00010322;
const char32_t OLD_ITALIC_NUMERAL_FIFTY = 0x00010323;
const char32_t OLD_ITALIC_LETTER_YE = 0x0001032D;
const char32_t OLD_ITALIC_LETTER_NORTHERN_TSE = 0x0001032E;
const char32_t OLD_ITALIC_LETTER_SOUTHERN_TSE = 0x0001032F;
const char32_t GOTHIC_LETTER_AHSA = 0x00010330;
const char32_t GOTHIC_LETTER_BAIRKAN = 0x00010331;
const char32_t GOTHIC_LETTER_GIBA = 0x00010332;
const char32_t GOTHIC_LETTER_DAGS = 0x00010333;
const char32_t GOTHIC_LETTER_AIHVUS = 0x00010334;
const char32_t GOTHIC_LETTER_QAIRTHRA = 0x00010335;
const char32_t GOTHIC_LETTER_IUJA = 0x00010336;
const char32_t GOTHIC_LETTER_HAGL = 0x00010337;
const char32_t GOTHIC_LETTER_THIUTH = 0x00010338;
const char32_t GOTHIC_LETTER_EIS = 0x00010339;
const char32_t GOTHIC_LETTER_KUSMA = 0x0001033A;
const char32_t GOTHIC_LETTER_LAGUS = 0x0001033B;
const char32_t GOTHIC_LETTER_MANNA = 0x0001033C;
const char32_t GOTHIC_LETTER_NAUTHS = 0x0001033D;
const char32_t GOTHIC_LETTER_JER = 0x0001033E;
const char32_t GOTHIC_LETTER_URUS = 0x0001033F;
const char32_t GOTHIC_LETTER_PAIRTHRA = 0x00010340;
const char32_t GOTHIC_LETTER_NINETY = 0x00010341;
const char32_t GOTHIC_LETTER_RAIDA = 0x00010342;
const char32_t GOTHIC_LETTER_SAUIL = 0x00010343;
const char32_t GOTHIC_LETTER_TEIWS = 0x00010344;
const char32_t GOTHIC_LETTER_WINJA = 0x00010345;
const char32_t GOTHIC_LETTER_FAIHU = 0x00010346;
const char32_t GOTHIC_LETTER_IGGWS = 0x00010347;
const char32_t GOTHIC_LETTER_HWAIR = 0x00010348;
const char32_t GOTHIC_LETTER_OTHAL = 0x00010349;
const char32_t GOTHIC_LETTER_NINE_HUNDRED = 0x0001034A;
const char32_t OLD_PERMIC_LETTER_AN = 0x00010350;
const char32_t OLD_PERMIC_LETTER_BUR = 0x00010351;
const char32_t OLD_PERMIC_LETTER_GAI = 0x00010352;
const char32_t OLD_PERMIC_LETTER_DOI = 0x00010353;
const char32_t OLD_PERMIC_LETTER_E = 0x00010354;
const char32_t OLD_PERMIC_LETTER_ZHOI = 0x00010355;
const char32_t OLD_PERMIC_LETTER_DZHOI = 0x00010356;
const char32_t OLD_PERMIC_LETTER_ZATA = 0x00010357;
const char32_t OLD_PERMIC_LETTER_DZITA = 0x00010358;
const char32_t OLD_PERMIC_LETTER_I = 0x00010359;
const char32_t OLD_PERMIC_LETTER_KOKE = 0x0001035A;
const char32_t OLD_PERMIC_LETTER_LEI = 0x0001035B;
const char32_t OLD_PERMIC_LETTER_MENOE = 0x0001035C;
const char32_t OLD_PERMIC_LETTER_NENOE = 0x0001035D;
const char32_t OLD_PERMIC_LETTER_VOOI = 0x0001035E;
const char32_t OLD_PERMIC_LETTER_PEEI = 0x0001035F;
const char32_t OLD_PERMIC_LETTER_REI = 0x00010360;
const char32_t OLD_PERMIC_LETTER_SII = 0x00010361;
const char32_t OLD_PERMIC_LETTER_TAI = 0x00010362;
const char32_t OLD_PERMIC_LETTER_U = 0x00010363;
const char32_t OLD_PERMIC_LETTER_CHERY = 0x00010364;
const char32_t OLD_PERMIC_LETTER_SHOOI = 0x00010365;
const char32_t OLD_PERMIC_LETTER_SHCHOOI = 0x00010366;
const char32_t OLD_PERMIC_LETTER_YRY = 0x00010367;
const char32_t OLD_PERMIC_LETTER_YERU = 0x00010368;
const char32_t OLD_PERMIC_LETTER_O = 0x00010369;
const char32_t OLD_PERMIC_LETTER_OO = 0x0001036A;
const char32_t OLD_PERMIC_LETTER_EF = 0x0001036B;
const char32_t OLD_PERMIC_LETTER_HA = 0x0001036C;
const char32_t OLD_PERMIC_LETTER_TSIU = 0x0001036D;
const char32_t OLD_PERMIC_LETTER_VER = 0x0001036E;
const char32_t OLD_PERMIC_LETTER_YER = 0x0001036F;
const char32_t OLD_PERMIC_LETTER_YERI = 0x00010370;
const char32_t OLD_PERMIC_LETTER_YAT = 0x00010371;
const char32_t OLD_PERMIC_LETTER_IE = 0x00010372;
const char32_t OLD_PERMIC_LETTER_YU = 0x00010373;
const char32_t OLD_PERMIC_LETTER_YA = 0x00010374;
const char32_t OLD_PERMIC_LETTER_IA = 0x00010375;
const char32_t COMBINING_OLD_PERMIC_LETTER_AN = 0x00010376;
const char32_t COMBINING_OLD_PERMIC_LETTER_DOI = 0x00010377;
const char32_t COMBINING_OLD_PERMIC_LETTER_ZATA = 0x00010378;
const char32_t COMBINING_OLD_PERMIC_LETTER_NENOE = 0x00010379;
const char32_t COMBINING_OLD_PERMIC_LETTER_SII = 0x0001037A;
const char32_t UGARITIC_LETTER_ALPA = 0x00010380;
const char32_t UGARITIC_LETTER_BETA = 0x00010381;
const char32_t UGARITIC_LETTER_GAMLA = 0x00010382;
const char32_t UGARITIC_LETTER_KHA = 0x00010383;
const char32_t UGARITIC_LETTER_DELTA = 0x00010384;
const char32_t UGARITIC_LETTER_HO = 0x00010385;
const char32_t UGARITIC_LETTER_WO = 0x00010386;
const char32_t UGARITIC_LETTER_ZETA = 0x00010387;
const char32_t UGARITIC_LETTER_HOTA = 0x00010388;
const char32_t UGARITIC_LETTER_TET = 0x00010389;
const char32_t UGARITIC_LETTER_YOD = 0x0001038A;
const char32_t UGARITIC_LETTER_KAF = 0x0001038B;
const char32_t UGARITIC_LETTER_SHIN = 0x0001038C;
const char32_t UGARITIC_LETTER_LAMDA = 0x0001038D;
const char32_t UGARITIC_LETTER_MEM = 0x0001038E;
const char32_t UGARITIC_LETTER_DHAL = 0x0001038F;
const char32_t UGARITIC_LETTER_NUN = 0x00010390;
const char32_t UGARITIC_LETTER_ZU = 0x00010391;
const char32_t UGARITIC_LETTER_SAMKA = 0x00010392;
const char32_t UGARITIC_LETTER_AIN = 0x00010393;
const char32_t UGARITIC_LETTER_PU = 0x00010394;
const char32_t UGARITIC_LETTER_SADE = 0x00010395;
const char32_t UGARITIC_LETTER_QOPA = 0x00010396;
const char32_t UGARITIC_LETTER_RASHA = 0x00010397;
const char32_t UGARITIC_LETTER_THANNA = 0x00010398;
const char32_t UGARITIC_LETTER_GHAIN = 0x00010399;
const char32_t UGARITIC_LETTER_TO = 0x0001039A;
const char32_t UGARITIC_LETTER_I = 0x0001039B;
const char32_t UGARITIC_LETTER_U = 0x0001039C;
const char32_t UGARITIC_LETTER_SSU = 0x0001039D;
const char32_t UGARITIC_WORD_DIVIDER = 0x0001039F;
const char32_t OLD_PERSIAN_SIGN_A = 0x000103A0;
const char32_t OLD_PERSIAN_SIGN_I = 0x000103A1;
const char32_t OLD_PERSIAN_SIGN_U = 0x000103A2;
const char32_t OLD_PERSIAN_SIGN_KA = 0x000103A3;
const char32_t OLD_PERSIAN_SIGN_KU = 0x000103A4;
const char32_t OLD_PERSIAN_SIGN_GA = 0x000103A5;
const char32_t OLD_PERSIAN_SIGN_GU = 0x000103A6;
const char32_t OLD_PERSIAN_SIGN_XA = 0x000103A7;
const char32_t OLD_PERSIAN_SIGN_CA = 0x000103A8;
const char32_t OLD_PERSIAN_SIGN_JA = 0x000103A9;
const char32_t OLD_PERSIAN_SIGN_JI = 0x000103AA;
const char32_t OLD_PERSIAN_SIGN_TA = 0x000103AB;
const char32_t OLD_PERSIAN_SIGN_TU = 0x000103AC;
const char32_t OLD_PERSIAN_SIGN_DA = 0x000103AD;
const char32_t OLD_PERSIAN_SIGN_DI = 0x000103AE;
const char32_t OLD_PERSIAN_SIGN_DU = 0x000103AF;
const char32_t OLD_PERSIAN_SIGN_THA = 0x000103B0;
const char32_t OLD_PERSIAN_SIGN_PA = 0x000103B1;
const char32_t OLD_PERSIAN_SIGN_BA = 0x000103B2;
const char32_t OLD_PERSIAN_SIGN_FA = 0x000103B3;
const char32_t OLD_PERSIAN_SIGN_NA = 0x000103B4;
const char32_t OLD_PERSIAN_SIGN_NU = 0x000103B5;
const char32_t OLD_PERSIAN_SIGN_MA = 0x000103B6;
const char32_t OLD_PERSIAN_SIGN_MI = 0x000103B7;
const char32_t OLD_PERSIAN_SIGN_MU = 0x000103B8;
const char32_t OLD_PERSIAN_SIGN_YA = 0x000103B9;
const char32_t OLD_PERSIAN_SIGN_VA = 0x000103BA;
const char32_t OLD_PERSIAN_SIGN_VI = 0x000103BB;
const char32_t OLD_PERSIAN_SIGN_RA = 0x000103BC;
const char32_t OLD_PERSIAN_SIGN_RU = 0x000103BD;
const char32_t OLD_PERSIAN_SIGN_LA = 0x000103BE;
const char32_t OLD_PERSIAN_SIGN_SA = 0x000103BF;
const char32_t OLD_PERSIAN_SIGN_ZA = 0x000103C0;
const char32_t OLD_PERSIAN_SIGN_SHA = 0x000103C1;
const char32_t OLD_PERSIAN_SIGN_SSA = 0x000103C2;
const char32_t OLD_PERSIAN_SIGN_HA = 0x000103C3;
const char32_t OLD_PERSIAN_SIGN_AURAMAZDAA = 0x000103C8;
const char32_t OLD_PERSIAN_SIGN_AURAMAZDAA_2 = 0x000103C9;
const char32_t OLD_PERSIAN_SIGN_AURAMAZDAAHA = 0x000103CA;
const char32_t OLD_PERSIAN_SIGN_XSHAAYATHIYA = 0x000103CB;
const char32_t OLD_PERSIAN_SIGN_DAHYAAUSH = 0x000103CC;
const char32_t OLD_PERSIAN_SIGN_DAHYAAUSH_2 = 0x000103CD;
const char32_t OLD_PERSIAN_SIGN_BAGA = 0x000103CE;
const char32_t OLD_PERSIAN_SIGN_BUUMISH = 0x000103CF;
const char32_t OLD_PERSIAN_WORD_DIVIDER = 0x000103D0;
const char32_t OLD_PERSIAN_NUMBER_ONE = 0x000103D1;
const char32_t OLD_PERSIAN_NUMBER_TWO = 0x000103D2;
const char32_t OLD_PERSIAN_NUMBER_TEN = 0x000103D3;
const char32_t OLD_PERSIAN_NUMBER_TWENTY = 0x000103D4;
const char32_t OLD_PERSIAN_NUMBER_HUNDRED = 0x000103D5;
const char32_t DESERET_CAPITAL_LETTER_LONG_I = 0x00010400;
const char32_t DESERET_CAPITAL_LETTER_LONG_E = 0x00010401;
const char32_t DESERET_CAPITAL_LETTER_LONG_A = 0x00010402;
const char32_t DESERET_CAPITAL_LETTER_LONG_AH = 0x00010403;
const char32_t DESERET_CAPITAL_LETTER_LONG_O = 0x00010404;
const char32_t DESERET_CAPITAL_LETTER_LONG_OO = 0x00010405;
const char32_t DESERET_CAPITAL_LETTER_SHORT_I = 0x00010406;
const char32_t DESERET_CAPITAL_LETTER_SHORT_E = 0x00010407;
const char32_t DESERET_CAPITAL_LETTER_SHORT_A = 0x00010408;
const char32_t DESERET_CAPITAL_LETTER_SHORT_AH = 0x00010409;
const char32_t DESERET_CAPITAL_LETTER_SHORT_O = 0x0001040A;
const char32_t DESERET_CAPITAL_LETTER_SHORT_OO = 0x0001040B;
const char32_t DESERET_CAPITAL_LETTER_AY = 0x0001040C;
const char32_t DESERET_CAPITAL_LETTER_OW = 0x0001040D;
const char32_t DESERET_CAPITAL_LETTER_WU = 0x0001040E;
const char32_t DESERET_CAPITAL_LETTER_YEE = 0x0001040F;
const char32_t DESERET_CAPITAL_LETTER_H = 0x00010410;
const char32_t DESERET_CAPITAL_LETTER_PEE = 0x00010411;
const char32_t DESERET_CAPITAL_LETTER_BEE = 0x00010412;
const char32_t DESERET_CAPITAL_LETTER_TEE = 0x00010413;
const char32_t DESERET_CAPITAL_LETTER_DEE = 0x00010414;
const char32_t DESERET_CAPITAL_LETTER_CHEE = 0x00010415;
const char32_t DESERET_CAPITAL_LETTER_JEE = 0x00010416;
const char32_t DESERET_CAPITAL_LETTER_KAY = 0x00010417;
const char32_t DESERET_CAPITAL_LETTER_GAY = 0x00010418;
const char32_t DESERET_CAPITAL_LETTER_EF = 0x00010419;
const char32_t DESERET_CAPITAL_LETTER_VEE = 0x0001041A;
const char32_t DESERET_CAPITAL_LETTER_ETH = 0x0001041B;
const char32_t DESERET_CAPITAL_LETTER_THEE = 0x0001041C;
const char32_t DESERET_CAPITAL_LETTER_ES = 0x0001041D;
const char32_t DESERET_CAPITAL_LETTER_ZEE = 0x0001041E;
const char32_t DESERET_CAPITAL_LETTER_ESH = 0x0001041F;
const char32_t DESERET_CAPITAL_LETTER_ZHEE = 0x00010420;
const char32_t DESERET_CAPITAL_LETTER_ER = 0x00010421;
const char32_t DESERET_CAPITAL_LETTER_EL = 0x00010422;
const char32_t DESERET_CAPITAL_LETTER_EM = 0x00010423;
const char32_t DESERET_CAPITAL_LETTER_EN = 0x00010424;
const char32_t DESERET_CAPITAL_LETTER_ENG = 0x00010425;
const char32_t DESERET_CAPITAL_LETTER_OI = 0x00010426;
const char32_t DESERET_CAPITAL_LETTER_EW = 0x00010427;
const char32_t DESERET_SMALL_LETTER_LONG_I = 0x00010428;
const char32_t DESERET_SMALL_LETTER_LONG_E = 0x00010429;
const char32_t DESERET_SMALL_LETTER_LONG_A = 0x0001042A;
const char32_t DESERET_SMALL_LETTER_LONG_AH = 0x0001042B;
const char32_t DESERET_SMALL_LETTER_LONG_O = 0x0001042C;
const char32_t DESERET_SMALL_LETTER_LONG_OO = 0x0001042D;
const char32_t DESERET_SMALL_LETTER_SHORT_I = 0x0001042E;
const char32_t DESERET_SMALL_LETTER_SHORT_E = 0x0001042F;
const char32_t DESERET_SMALL_LETTER_SHORT_A = 0x00010430;
const char32_t DESERET_SMALL_LETTER_SHORT_AH = 0x00010431;
const char32_t DESERET_SMALL_LETTER_SHORT_O = 0x00010432;
const char32_t DESERET_SMALL_LETTER_SHORT_OO = 0x00010433;
const char32_t DESERET_SMALL_LETTER_AY = 0x00010434;
const char32_t DESERET_SMALL_LETTER_OW = 0x00010435;
const char32_t DESERET_SMALL_LETTER_WU = 0x00010436;
const char32_t DESERET_SMALL_LETTER_YEE = 0x00010437;
const char32_t DESERET_SMALL_LETTER_H = 0x00010438;
const char32_t DESERET_SMALL_LETTER_PEE = 0x00010439;
const char32_t DESERET_SMALL_LETTER_BEE = 0x0001043A;
const char32_t DESERET_SMALL_LETTER_TEE = 0x0001043B;
const char32_t DESERET_SMALL_LETTER_DEE = 0x0001043C;
const char32_t DESERET_SMALL_LETTER_CHEE = 0x0001043D;
const char32_t DESERET_SMALL_LETTER_JEE = 0x0001043E;
const char32_t DESERET_SMALL_LETTER_KAY = 0x0001043F;
const char32_t DESERET_SMALL_LETTER_GAY = 0x00010440;
const char32_t DESERET_SMALL_LETTER_EF = 0x00010441;
const char32_t DESERET_SMALL_LETTER_VEE = 0x00010442;
const char32_t DESERET_SMALL_LETTER_ETH = 0x00010443;
const char32_t DESERET_SMALL_LETTER_THEE = 0x00010444;
const char32_t DESERET_SMALL_LETTER_ES = 0x00010445;
const char32_t DESERET_SMALL_LETTER_ZEE = 0x00010446;
const char32_t DESERET_SMALL_LETTER_ESH = 0x00010447;
const char32_t DESERET_SMALL_LETTER_ZHEE = 0x00010448;
const char32_t DESERET_SMALL_LETTER_ER = 0x00010449;
const char32_t DESERET_SMALL_LETTER_EL = 0x0001044A;
const char32_t DESERET_SMALL_LETTER_EM = 0x0001044B;
const char32_t DESERET_SMALL_LETTER_EN = 0x0001044C;
const char32_t DESERET_SMALL_LETTER_ENG = 0x0001044D;
const char32_t DESERET_SMALL_LETTER_OI = 0x0001044E;
const char32_t DESERET_SMALL_LETTER_EW = 0x0001044F;
const char32_t SHAVIAN_LETTER_PEEP = 0x00010450;
const char32_t SHAVIAN_LETTER_TOT = 0x00010451;
const char32_t SHAVIAN_LETTER_KICK = 0x00010452;
const char32_t SHAVIAN_LETTER_FEE = 0x00010453;
const char32_t SHAVIAN_LETTER_THIGH = 0x00010454;
const char32_t SHAVIAN_LETTER_SO = 0x00010455;
const char32_t SHAVIAN_LETTER_SURE = 0x00010456;
const char32_t SHAVIAN_LETTER_CHURCH = 0x00010457;
const char32_t SHAVIAN_LETTER_YEA = 0x00010458;
const char32_t SHAVIAN_LETTER_HUNG = 0x00010459;
const char32_t SHAVIAN_LETTER_BIB = 0x0001045A;
const char32_t SHAVIAN_LETTER_DEAD = 0x0001045B;
const char32_t SHAVIAN_LETTER_GAG = 0x0001045C;
const char32_t SHAVIAN_LETTER_VOW = 0x0001045D;
const char32_t SHAVIAN_LETTER_THEY = 0x0001045E;
const char32_t SHAVIAN_LETTER_ZOO = 0x0001045F;
const char32_t SHAVIAN_LETTER_MEASURE = 0x00010460;
const char32_t SHAVIAN_LETTER_JUDGE = 0x00010461;
const char32_t SHAVIAN_LETTER_WOE = 0x00010462;
const char32_t SHAVIAN_LETTER_HA_HA = 0x00010463;
const char32_t SHAVIAN_LETTER_LOLL = 0x00010464;
const char32_t SHAVIAN_LETTER_MIME = 0x00010465;
const char32_t SHAVIAN_LETTER_IF = 0x00010466;
const char32_t SHAVIAN_LETTER_EGG = 0x00010467;
const char32_t SHAVIAN_LETTER_ASH = 0x00010468;
const char32_t SHAVIAN_LETTER_ADO = 0x00010469;
const char32_t SHAVIAN_LETTER_ON = 0x0001046A;
const char32_t SHAVIAN_LETTER_WOOL = 0x0001046B;
const char32_t SHAVIAN_LETTER_OUT = 0x0001046C;
const char32_t SHAVIAN_LETTER_AH = 0x0001046D;
const char32_t SHAVIAN_LETTER_ROAR = 0x0001046E;
const char32_t SHAVIAN_LETTER_NUN = 0x0001046F;
const char32_t SHAVIAN_LETTER_EAT = 0x00010470;
const char32_t SHAVIAN_LETTER_AGE = 0x00010471;
const char32_t SHAVIAN_LETTER_ICE = 0x00010472;
const char32_t SHAVIAN_LETTER_UP = 0x00010473;
const char32_t SHAVIAN_LETTER_OAK = 0x00010474;
const char32_t SHAVIAN_LETTER_OOZE = 0x00010475;
const char32_t SHAVIAN_LETTER_OIL = 0x00010476;
const char32_t SHAVIAN_LETTER_AWE = 0x00010477;
const char32_t SHAVIAN_LETTER_ARE = 0x00010478;
const char32_t SHAVIAN_LETTER_OR = 0x00010479;
const char32_t SHAVIAN_LETTER_AIR = 0x0001047A;
const char32_t SHAVIAN_LETTER_ERR = 0x0001047B;
const char32_t SHAVIAN_LETTER_ARRAY = 0x0001047C;
const char32_t SHAVIAN_LETTER_EAR = 0x0001047D;
const char32_t SHAVIAN_LETTER_IAN = 0x0001047E;
const char32_t SHAVIAN_LETTER_YEW = 0x0001047F;
const char32_t OSMANYA_LETTER_ALEF = 0x00010480;
const char32_t OSMANYA_LETTER_BA = 0x00010481;
const char32_t OSMANYA_LETTER_TA = 0x00010482;
const char32_t OSMANYA_LETTER_JA = 0x00010483;
const char32_t OSMANYA_LETTER_XA = 0x00010484;
const char32_t OSMANYA_LETTER_KHA = 0x00010485;
const char32_t OSMANYA_LETTER_DEEL = 0x00010486;
const char32_t OSMANYA_LETTER_RA = 0x00010487;
const char32_t OSMANYA_LETTER_SA = 0x00010488;
const char32_t OSMANYA_LETTER_SHIIN = 0x00010489;
const char32_t OSMANYA_LETTER_DHA = 0x0001048A;
const char32_t OSMANYA_LETTER_CAYN = 0x0001048B;
const char32_t OSMANYA_LETTER_GA = 0x0001048C;
const char32_t OSMANYA_LETTER_FA = 0x0001048D;
const char32_t OSMANYA_LETTER_QAAF = 0x0001048E;
const char32_t OSMANYA_LETTER_KAAF = 0x0001048F;
const char32_t OSMANYA_LETTER_LAAN = 0x00010490;
const char32_t OSMANYA_LETTER_MIIN = 0x00010491;
const char32_t OSMANYA_LETTER_NUUN = 0x00010492;
const char32_t OSMANYA_LETTER_WAW = 0x00010493;
const char32_t OSMANYA_LETTER_HA = 0x00010494;
const char32_t OSMANYA_LETTER_YA = 0x00010495;
const char32_t OSMANYA_LETTER_A = 0x00010496;
const char32_t OSMANYA_LETTER_E = 0x00010497;
const char32_t OSMANYA_LETTER_I = 0x00010498;
const char32_t OSMANYA_LETTER_O = 0x00010499;
const char32_t OSMANYA_LETTER_U = 0x0001049A;
const char32_t OSMANYA_LETTER_AA = 0x0001049B;
const char32_t OSMANYA_LETTER_EE = 0x0001049C;
const char32_t OSMANYA_LETTER_OO = 0x0001049D;
const char32_t OSMANYA_DIGIT_ZERO = 0x000104A0;
const char32_t OSMANYA_DIGIT_ONE = 0x000104A1;
const char32_t OSMANYA_DIGIT_TWO = 0x000104A2;
const char32_t OSMANYA_DIGIT_THREE = 0x000104A3;
const char32_t OSMANYA_DIGIT_FOUR = 0x000104A4;
const char32_t OSMANYA_DIGIT_FIVE = 0x000104A5;
const char32_t OSMANYA_DIGIT_SIX = 0x000104A6;
const char32_t OSMANYA_DIGIT_SEVEN = 0x000104A7;
const char32_t OSMANYA_DIGIT_EIGHT = 0x000104A8;
const char32_t OSMANYA_DIGIT_NINE = 0x000104A9;
const char32_t OSAGE_CAPITAL_LETTER_A = 0x000104B0;
const char32_t OSAGE_CAPITAL_LETTER_AI = 0x000104B1;
const char32_t OSAGE_CAPITAL_LETTER_AIN = 0x000104B2;
const char32_t OSAGE_CAPITAL_LETTER_AH = 0x000104B3;
const char32_t OSAGE_CAPITAL_LETTER_BRA = 0x000104B4;
const char32_t OSAGE_CAPITAL_LETTER_CHA = 0x000104B5;
const char32_t OSAGE_CAPITAL_LETTER_EHCHA = 0x000104B6;
const char32_t OSAGE_CAPITAL_LETTER_E = 0x000104B7;
const char32_t OSAGE_CAPITAL_LETTER_EIN = 0x000104B8;
const char32_t OSAGE_CAPITAL_LETTER_HA = 0x000104B9;
const char32_t OSAGE_CAPITAL_LETTER_HYA = 0x000104BA;
const char32_t OSAGE_CAPITAL_LETTER_I = 0x000104BB;
const char32_t OSAGE_CAPITAL_LETTER_KA = 0x000104BC;
const char32_t OSAGE_CAPITAL_LETTER_EHKA = 0x000104BD;
const char32_t OSAGE_CAPITAL_LETTER_KYA = 0x000104BE;
const char32_t OSAGE_CAPITAL_LETTER_LA = 0x000104BF;
const char32_t OSAGE_CAPITAL_LETTER_MA = 0x000104C0;
const char32_t OSAGE_CAPITAL_LETTER_NA = 0x000104C1;
const char32_t OSAGE_CAPITAL_LETTER_O = 0x000104C2;
const char32_t OSAGE_CAPITAL_LETTER_OIN = 0x000104C3;
const char32_t OSAGE_CAPITAL_LETTER_PA = 0x000104C4;
const char32_t OSAGE_CAPITAL_LETTER_EHPA = 0x000104C5;
const char32_t OSAGE_CAPITAL_LETTER_SA = 0x000104C6;
const char32_t OSAGE_CAPITAL_LETTER_SHA = 0x000104C7;
const char32_t OSAGE_CAPITAL_LETTER_TA = 0x000104C8;
const char32_t OSAGE_CAPITAL_LETTER_EHTA = 0x000104C9;
const char32_t OSAGE_CAPITAL_LETTER_TSA = 0x000104CA;
const char32_t OSAGE_CAPITAL_LETTER_EHTSA = 0x000104CB;
const char32_t OSAGE_CAPITAL_LETTER_TSHA = 0x000104CC;
const char32_t OSAGE_CAPITAL_LETTER_DHA = 0x000104CD;
const char32_t OSAGE_CAPITAL_LETTER_U = 0x000104CE;
const char32_t OSAGE_CAPITAL_LETTER_WA = 0x000104CF;
const char32_t OSAGE_CAPITAL_LETTER_KHA = 0x000104D0;
const char32_t OSAGE_CAPITAL_LETTER_GHA = 0x000104D1;
const char32_t OSAGE_CAPITAL_LETTER_ZA = 0x000104D2;
const char32_t OSAGE_CAPITAL_LETTER_ZHA = 0x000104D3;
const char32_t OSAGE_SMALL_LETTER_A = 0x000104D8;
const char32_t OSAGE_SMALL_LETTER_AI = 0x000104D9;
const char32_t OSAGE_SMALL_LETTER_AIN = 0x000104DA;
const char32_t OSAGE_SMALL_LETTER_AH = 0x000104DB;
const char32_t OSAGE_SMALL_LETTER_BRA = 0x000104DC;
const char32_t OSAGE_SMALL_LETTER_CHA = 0x000104DD;
const char32_t OSAGE_SMALL_LETTER_EHCHA = 0x000104DE;
const char32_t OSAGE_SMALL_LETTER_E = 0x000104DF;
const char32_t OSAGE_SMALL_LETTER_EIN = 0x000104E0;
const char32_t OSAGE_SMALL_LETTER_HA = 0x000104E1;
const char32_t OSAGE_SMALL_LETTER_HYA = 0x000104E2;
const char32_t OSAGE_SMALL_LETTER_I = 0x000104E3;
const char32_t OSAGE_SMALL_LETTER_KA = 0x000104E4;
const char32_t OSAGE_SMALL_LETTER_EHKA = 0x000104E5;
const char32_t OSAGE_SMALL_LETTER_KYA = 0x000104E6;
const char32_t OSAGE_SMALL_LETTER_LA = 0x000104E7;
const char32_t OSAGE_SMALL_LETTER_MA = 0x000104E8;
const char32_t OSAGE_SMALL_LETTER_NA = 0x000104E9;
const char32_t OSAGE_SMALL_LETTER_O = 0x000104EA;
const char32_t OSAGE_SMALL_LETTER_OIN = 0x000104EB;
const char32_t OSAGE_SMALL_LETTER_PA = 0x000104EC;
const char32_t OSAGE_SMALL_LETTER_EHPA = 0x000104ED;
const char32_t OSAGE_SMALL_LETTER_SA = 0x000104EE;
const char32_t OSAGE_SMALL_LETTER_SHA = 0x000104EF;
const char32_t OSAGE_SMALL_LETTER_TA = 0x000104F0;
const char32_t OSAGE_SMALL_LETTER_EHTA = 0x000104F1;
const char32_t OSAGE_SMALL_LETTER_TSA = 0x000104F2;
const char32_t OSAGE_SMALL_LETTER_EHTSA = 0x000104F3;
const char32_t OSAGE_SMALL_LETTER_TSHA = 0x000104F4;
const char32_t OSAGE_SMALL_LETTER_DHA = 0x000104F5;
const char32_t OSAGE_SMALL_LETTER_U = 0x000104F6;
const char32_t OSAGE_SMALL_LETTER_WA = 0x000104F7;
const char32_t OSAGE_SMALL_LETTER_KHA = 0x000104F8;
const char32_t OSAGE_SMALL_LETTER_GHA = 0x000104F9;
const char32_t OSAGE_SMALL_LETTER_ZA = 0x000104FA;
const char32_t OSAGE_SMALL_LETTER_ZHA = 0x000104FB;
const char32_t ELBASAN_LETTER_A = 0x00010500;
const char32_t ELBASAN_LETTER_BE = 0x00010501;
const char32_t ELBASAN_LETTER_CE = 0x00010502;
const char32_t ELBASAN_LETTER_CHE = 0x00010503;
const char32_t ELBASAN_LETTER_DE = 0x00010504;
const char32_t ELBASAN_LETTER_NDE = 0x00010505;
const char32_t ELBASAN_LETTER_DHE = 0x00010506;
const char32_t ELBASAN_LETTER_EI = 0x00010507;
const char32_t ELBASAN_LETTER_E = 0x00010508;
const char32_t ELBASAN_LETTER_FE = 0x00010509;
const char32_t ELBASAN_LETTER_GE = 0x0001050A;
const char32_t ELBASAN_LETTER_GJE = 0x0001050B;
const char32_t ELBASAN_LETTER_HE = 0x0001050C;
const char32_t ELBASAN_LETTER_I = 0x0001050D;
const char32_t ELBASAN_LETTER_JE = 0x0001050E;
const char32_t ELBASAN_LETTER_KE = 0x0001050F;
const char32_t ELBASAN_LETTER_LE = 0x00010510;
const char32_t ELBASAN_LETTER_LLE = 0x00010511;
const char32_t ELBASAN_LETTER_ME = 0x00010512;
const char32_t ELBASAN_LETTER_NE = 0x00010513;
const char32_t ELBASAN_LETTER_NA = 0x00010514;
const char32_t ELBASAN_LETTER_NJE = 0x00010515;
const char32_t ELBASAN_LETTER_O = 0x00010516;
const char32_t ELBASAN_LETTER_PE = 0x00010517;
const char32_t ELBASAN_LETTER_QE = 0x00010518;
const char32_t ELBASAN_LETTER_RE = 0x00010519;
const char32_t ELBASAN_LETTER_RRE = 0x0001051A;
const char32_t ELBASAN_LETTER_SE = 0x0001051B;
const char32_t ELBASAN_LETTER_SHE = 0x0001051C;
const char32_t ELBASAN_LETTER_TE = 0x0001051D;
const char32_t ELBASAN_LETTER_THE = 0x0001051E;
const char32_t ELBASAN_LETTER_U = 0x0001051F;
const char32_t ELBASAN_LETTER_VE = 0x00010520;
const char32_t ELBASAN_LETTER_XE = 0x00010521;
const char32_t ELBASAN_LETTER_Y = 0x00010522;
const char32_t ELBASAN_LETTER_ZE = 0x00010523;
const char32_t ELBASAN_LETTER_ZHE = 0x00010524;
const char32_t ELBASAN_LETTER_GHE = 0x00010525;
const char32_t ELBASAN_LETTER_GHAMMA = 0x00010526;
const char32_t ELBASAN_LETTER_KHE = 0x00010527;
const char32_t CAUCASIAN_ALBANIAN_LETTER_ALT = 0x00010530;
const char32_t CAUCASIAN_ALBANIAN_LETTER_BET = 0x00010531;
const char32_t CAUCASIAN_ALBANIAN_LETTER_GIM = 0x00010532;
const char32_t CAUCASIAN_ALBANIAN_LETTER_DAT = 0x00010533;
const char32_t CAUCASIAN_ALBANIAN_LETTER_EB = 0x00010534;
const char32_t CAUCASIAN_ALBANIAN_LETTER_ZARL = 0x00010535;
const char32_t CAUCASIAN_ALBANIAN_LETTER_EYN = 0x00010536;
const char32_t CAUCASIAN_ALBANIAN_LETTER_ZHIL = 0x00010537;
const char32_t CAUCASIAN_ALBANIAN_LETTER_TAS = 0x00010538;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CHA = 0x00010539;
const char32_t CAUCASIAN_ALBANIAN_LETTER_YOWD = 0x0001053A;
const char32_t CAUCASIAN_ALBANIAN_LETTER_ZHA = 0x0001053B;
const char32_t CAUCASIAN_ALBANIAN_LETTER_IRB = 0x0001053C;
const char32_t CAUCASIAN_ALBANIAN_LETTER_SHA = 0x0001053D;
const char32_t CAUCASIAN_ALBANIAN_LETTER_LAN = 0x0001053E;
const char32_t CAUCASIAN_ALBANIAN_LETTER_INYA = 0x0001053F;
const char32_t CAUCASIAN_ALBANIAN_LETTER_XEYN = 0x00010540;
const char32_t CAUCASIAN_ALBANIAN_LETTER_DYAN = 0x00010541;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CAR = 0x00010542;
const char32_t CAUCASIAN_ALBANIAN_LETTER_JHOX = 0x00010543;
const char32_t CAUCASIAN_ALBANIAN_LETTER_KAR = 0x00010544;
const char32_t CAUCASIAN_ALBANIAN_LETTER_LYIT = 0x00010545;
const char32_t CAUCASIAN_ALBANIAN_LETTER_HEYT = 0x00010546;
const char32_t CAUCASIAN_ALBANIAN_LETTER_QAY = 0x00010547;
const char32_t CAUCASIAN_ALBANIAN_LETTER_AOR = 0x00010548;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CHOY = 0x00010549;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CHI = 0x0001054A;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CYAY = 0x0001054B;
const char32_t CAUCASIAN_ALBANIAN_LETTER_MAQ = 0x0001054C;
const char32_t CAUCASIAN_ALBANIAN_LETTER_QAR = 0x0001054D;
const char32_t CAUCASIAN_ALBANIAN_LETTER_NOWC = 0x0001054E;
const char32_t CAUCASIAN_ALBANIAN_LETTER_DZYAY = 0x0001054F;
const char32_t CAUCASIAN_ALBANIAN_LETTER_SHAK = 0x00010550;
const char32_t CAUCASIAN_ALBANIAN_LETTER_JAYN = 0x00010551;
const char32_t CAUCASIAN_ALBANIAN_LETTER_ON = 0x00010552;
const char32_t CAUCASIAN_ALBANIAN_LETTER_TYAY = 0x00010553;
const char32_t CAUCASIAN_ALBANIAN_LETTER_FAM = 0x00010554;
const char32_t CAUCASIAN_ALBANIAN_LETTER_DZAY = 0x00010555;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CHAT = 0x00010556;
const char32_t CAUCASIAN_ALBANIAN_LETTER_PEN = 0x00010557;
const char32_t CAUCASIAN_ALBANIAN_LETTER_GHEYS = 0x00010558;
const char32_t CAUCASIAN_ALBANIAN_LETTER_RAT = 0x00010559;
const char32_t CAUCASIAN_ALBANIAN_LETTER_SEYK = 0x0001055A;
const char32_t CAUCASIAN_ALBANIAN_LETTER_VEYZ = 0x0001055B;
const char32_t CAUCASIAN_ALBANIAN_LETTER_TIWR = 0x0001055C;
const char32_t CAUCASIAN_ALBANIAN_LETTER_SHOY = 0x0001055D;
const char32_t CAUCASIAN_ALBANIAN_LETTER_IWN = 0x0001055E;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CYAW = 0x0001055F;
const char32_t CAUCASIAN_ALBANIAN_LETTER_CAYN = 0x00010560;
const char32_t CAUCASIAN_ALBANIAN_LETTER_YAYD = 0x00010561;
const char32_t CAUCASIAN_ALBANIAN_LETTER_PIWR = 0x00010562;
const char32_t CAUCASIAN_ALBANIAN_LETTER_KIW = 0x00010563;
const char32_t CAUCASIAN_ALBANIAN_CITATION_MARK = 0x0001056F;
const char32_t VITHKUQI_CAPITAL_LETTER_A = 0x00010570;
const char32_t VITHKUQI_CAPITAL_LETTER_BBE = 0x00010571;
const char32_t VITHKUQI_CAPITAL_LETTER_BE = 0x00010572;
const char32_t VITHKUQI_CAPITAL_LETTER_CE = 0x00010573;
const char32_t VITHKUQI_CAPITAL_LETTER_CHE = 0x00010574;
const char32_t VITHKUQI_CAPITAL_LETTER_DE = 0x00010575;
const char32_t VITHKUQI_CAPITAL_LETTER_DHE = 0x00010576;
const char32_t VITHKUQI_CAPITAL_LETTER_EI = 0x00010577;
const char32_t VITHKUQI_CAPITAL_LETTER_E = 0x00010578;
const char32_t VITHKUQI_CAPITAL_LETTER_FE = 0x00010579;
const char32_t VITHKUQI_CAPITAL_LETTER_GA = 0x0001057A;
const char32_t VITHKUQI_CAPITAL_LETTER_HA = 0x0001057C;
const char32_t VITHKUQI_CAPITAL_LETTER_HHA = 0x0001057D;
const char32_t VITHKUQI_CAPITAL_LETTER_I = 0x0001057E;
const char32_t VITHKUQI_CAPITAL_LETTER_IJE = 0x0001057F;
const char32_t VITHKUQI_CAPITAL_LETTER_JE = 0x00010580;
const char32_t VITHKUQI_CAPITAL_LETTER_KA = 0x00010581;
const char32_t VITHKUQI_CAPITAL_LETTER_LA = 0x00010582;
const char32_t VITHKUQI_CAPITAL_LETTER_LLA = 0x00010583;
const char32_t VITHKUQI_CAPITAL_LETTER_ME = 0x00010584;
const char32_t VITHKUQI_CAPITAL_LETTER_NE = 0x00010585;
const char32_t VITHKUQI_CAPITAL_LETTER_NJE = 0x00010586;
const char32_t VITHKUQI_CAPITAL_LETTER_O = 0x00010587;
const char32_t VITHKUQI_CAPITAL_LETTER_PE = 0x00010588;
const char32_t VITHKUQI_CAPITAL_LETTER_QA = 0x00010589;
const char32_t VITHKUQI_CAPITAL_LETTER_RE = 0x0001058A;
const char32_t VITHKUQI_CAPITAL_LETTER_SE = 0x0001058C;
const char32_t VITHKUQI_CAPITAL_LETTER_SHE = 0x0001058D;
const char32_t VITHKUQI_CAPITAL_LETTER_TE = 0x0001058E;
const char32_t VITHKUQI_CAPITAL_LETTER_THE = 0x0001058F;
const char32_t VITHKUQI_CAPITAL_LETTER_U = 0x00010590;
const char32_t VITHKUQI_CAPITAL_LETTER_VE = 0x00010591;
const char32_t VITHKUQI_CAPITAL_LETTER_XE = 0x00010592;
const char32_t VITHKUQI_CAPITAL_LETTER_Y = 0x00010594;
const char32_t VITHKUQI_CAPITAL_LETTER_ZE = 0x00010595;
const char32_t VITHKUQI_SMALL_LETTER_A = 0x00010597;
const char32_t VITHKUQI_SMALL_LETTER_BBE = 0x00010598;
const char32_t VITHKUQI_SMALL_LETTER_BE = 0x00010599;
const char32_t VITHKUQI_SMALL_LETTER_CE = 0x0001059A;
const char32_t VITHKUQI_SMALL_LETTER_CHE = 0x0001059B;
const char32_t VITHKUQI_SMALL_LETTER_DE = 0x0001059C;
const char32_t VITHKUQI_SMALL_LETTER_DHE = 0x0001059D;
const char32_t VITHKUQI_SMALL_LETTER_EI = 0x0001059E;
const char32_t VITHKUQI_SMALL_LETTER_E = 0x0001059F;
const char32_t VITHKUQI_SMALL_LETTER_FE = 0x000105A0;
const char32_t VITHKUQI_SMALL_LETTER_GA = 0x000105A1;
const char32_t VITHKUQI_SMALL_LETTER_HA = 0x000105A3;
const char32_t VITHKUQI_SMALL_LETTER_HHA = 0x000105A4;
const char32_t VITHKUQI_SMALL_LETTER_I = 0x000105A5;
const char32_t VITHKUQI_SMALL_LETTER_IJE = 0x000105A6;
const char32_t VITHKUQI_SMALL_LETTER_JE = 0x000105A7;
const char32_t VITHKUQI_SMALL_LETTER_KA = 0x000105A8;
const char32_t VITHKUQI_SMALL_LETTER_LA = 0x000105A9;
const char32_t VITHKUQI_SMALL_LETTER_LLA = 0x000105AA;
const char32_t VITHKUQI_SMALL_LETTER_ME = 0x000105AB;
const char32_t VITHKUQI_SMALL_LETTER_NE = 0x000105AC;
const char32_t VITHKUQI_SMALL_LETTER_NJE = 0x000105AD;
const char32_t VITHKUQI_SMALL_LETTER_O = 0x000105AE;
const char32_t VITHKUQI_SMALL_LETTER_PE = 0x000105AF;
const char32_t VITHKUQI_SMALL_LETTER_QA = 0x000105B0;
const char32_t VITHKUQI_SMALL_LETTER_RE = 0x000105B1;
const char32_t VITHKUQI_SMALL_LETTER_SE = 0x000105B3;
const char32_t VITHKUQI_SMALL_LETTER_SHE = 0x000105B4;
const char32_t VITHKUQI_SMALL_LETTER_TE = 0x000105B5;
const char32_t VITHKUQI_SMALL_LETTER_THE = 0x000105B6;
const char32_t VITHKUQI_SMALL_LETTER_U = 0x000105B7;
const char32_t VITHKUQI_SMALL_LETTER_VE = 0x000105B8;
const char32_t VITHKUQI_SMALL_LETTER_XE = 0x000105B9;
const char32_t VITHKUQI_SMALL_LETTER_Y = 0x000105BB;
const char32_t VITHKUQI_SMALL_LETTER_ZE = 0x000105BC;
const char32_t TODHRI_LETTER_A = 0x000105C0;
const char32_t TODHRI_LETTER_AS = 0x000105C1;
const char32_t TODHRI_LETTER_BA = 0x000105C2;
const char32_t TODHRI_LETTER_MBA = 0x000105C3;
const char32_t TODHRI_LETTER_CA = 0x000105C4;
const char32_t TODHRI_LETTER_CHA = 0x000105C5;
const char32_t TODHRI_LETTER_DA = 0x000105C6;
const char32_t TODHRI_LETTER_NDA = 0x000105C7;
const char32_t TODHRI_LETTER_DHA = 0x000105C8;
const char32_t TODHRI_LETTER_EI = 0x000105C9;
const char32_t TODHRI_LETTER_E = 0x000105CA;
const char32_t TODHRI_LETTER_FA = 0x000105CB;
const char32_t TODHRI_LETTER_GA = 0x000105CC;
const char32_t TODHRI_LETTER_NGA = 0x000105CD;
const char32_t TODHRI_LETTER_GJA = 0x000105CE;
const char32_t TODHRI_LETTER_NGJA = 0x000105CF;
const char32_t TODHRI_LETTER_HA = 0x000105D0;
const char32_t TODHRI_LETTER_HJA = 0x000105D1;
const char32_t TODHRI_LETTER_I = 0x000105D2;
const char32_t TODHRI_LETTER_JA = 0x000105D3;
const char32_t TODHRI_LETTER_KA = 0x000105D4;
const char32_t TODHRI_LETTER_LA = 0x000105D5;
const char32_t TODHRI_LETTER_LLA = 0x000105D6;
const char32_t TODHRI_LETTER_MA = 0x000105D7;
const char32_t TODHRI_LETTER_NA = 0x000105D8;
const char32_t TODHRI_LETTER_NJAN = 0x000105D9;
const char32_t TODHRI_LETTER_O = 0x000105DA;
const char32_t TODHRI_LETTER_PA = 0x000105DB;
const char32_t TODHRI_LETTER_QA = 0x000105DC;
const char32_t TODHRI_LETTER_RA = 0x000105DD;
const char32_t TODHRI_LETTER_RRA = 0x000105DE;
const char32_t TODHRI_LETTER_SA = 0x000105DF;
const char32_t TODHRI_LETTER_SHA = 0x000105E0;
const char32_t TODHRI_LETTER_SHTA = 0x000105E1;
const char32_t TODHRI_LETTER_TA = 0x000105E2;
const char32_t TODHRI_LETTER_THA = 0x000105E3;
const char32_t TODHRI_LETTER_U = 0x000105E4;
const char32_t TODHRI_LETTER_VA = 0x000105E5;
const char32_t TODHRI_LETTER_XA = 0x000105E6;
const char32_t TODHRI_LETTER_NXA = 0x000105E7;
const char32_t TODHRI_LETTER_XHA = 0x000105E8;
const char32_t TODHRI_LETTER_NXHA = 0x000105E9;
const char32_t TODHRI_LETTER_Y = 0x000105EA;
const char32_t TODHRI_LETTER_JY = 0x000105EB;
const char32_t TODHRI_LETTER_ZA = 0x000105EC;
const char32_t TODHRI_LETTER_ZHA = 0x000105ED;
const char32_t TODHRI_LETTER_GHA = 0x000105EE;
const char32_t TODHRI_LETTER_STA = 0x000105EF;
const char32_t TODHRI_LETTER_SKAN = 0x000105F0;
const char32_t TODHRI_LETTER_KHA = 0x000105F1;
const char32_t TODHRI_LETTER_PSA = 0x000105F2;
const char32_t TODHRI_LETTER_OO = 0x000105F3;
const char32_t LINEAR_A_SIGN_AB001 = 0x00010600;
const char32_t LINEAR_A_SIGN_AB002 = 0x00010601;
const char32_t LINEAR_A_SIGN_AB003 = 0x00010602;
const char32_t LINEAR_A_SIGN_AB004 = 0x00010603;
const char32_t LINEAR_A_SIGN_AB005 = 0x00010604;
const char32_t LINEAR_A_SIGN_AB006 = 0x00010605;
const char32_t LINEAR_A_SIGN_AB007 = 0x00010606;
const char32_t LINEAR_A_SIGN_AB008 = 0x00010607;
const char32_t LINEAR_A_SIGN_AB009 = 0x00010608;
const char32_t LINEAR_A_SIGN_AB010 = 0x00010609;
const char32_t LINEAR_A_SIGN_AB011 = 0x0001060A;
const char32_t LINEAR_A_SIGN_AB013 = 0x0001060B;
const char32_t LINEAR_A_SIGN_AB016 = 0x0001060C;
const char32_t LINEAR_A_SIGN_AB017 = 0x0001060D;
const char32_t LINEAR_A_SIGN_AB020 = 0x0001060E;
const char32_t LINEAR_A_SIGN_AB021 = 0x0001060F;
const char32_t LINEAR_A_SIGN_AB021F = 0x00010610;
const char32_t LINEAR_A_SIGN_AB021M = 0x00010611;
const char32_t LINEAR_A_SIGN_AB022 = 0x00010612;
const char32_t LINEAR_A_SIGN_AB022F = 0x00010613;
const char32_t LINEAR_A_SIGN_AB022M = 0x00010614;
const char32_t LINEAR_A_SIGN_AB023 = 0x00010615;
const char32_t LINEAR_A_SIGN_AB023M = 0x00010616;
const char32_t LINEAR_A_SIGN_AB024 = 0x00010617;
const char32_t LINEAR_A_SIGN_AB026 = 0x00010618;
const char32_t LINEAR_A_SIGN_AB027 = 0x00010619;
const char32_t LINEAR_A_SIGN_AB028 = 0x0001061A;
const char32_t LINEAR_A_SIGN_A028B = 0x0001061B;
const char32_t LINEAR_A_SIGN_AB029 = 0x0001061C;
const char32_t LINEAR_A_SIGN_AB030 = 0x0001061D;
const char32_t LINEAR_A_SIGN_AB031 = 0x0001061E;
const char32_t LINEAR_A_SIGN_AB034 = 0x0001061F;
const char32_t LINEAR_A_SIGN_AB037 = 0x00010620;
const char32_t LINEAR_A_SIGN_AB038 = 0x00010621;
const char32_t LINEAR_A_SIGN_AB039 = 0x00010622;
const char32_t LINEAR_A_SIGN_AB040 = 0x00010623;
const char32_t LINEAR_A_SIGN_AB041 = 0x00010624;
const char32_t LINEAR_A_SIGN_AB044 = 0x00010625;
const char32_t LINEAR_A_SIGN_AB045 = 0x00010626;
const char32_t LINEAR_A_SIGN_AB046 = 0x00010627;
const char32_t LINEAR_A_SIGN_AB047 = 0x00010628;
const char32_t LINEAR_A_SIGN_AB048 = 0x00010629;
const char32_t LINEAR_A_SIGN_AB049 = 0x0001062A;
const char32_t LINEAR_A_SIGN_AB050 = 0x0001062B;
const char32_t LINEAR_A_SIGN_AB051 = 0x0001062C;
const char32_t LINEAR_A_SIGN_AB053 = 0x0001062D;
const char32_t LINEAR_A_SIGN_AB054 = 0x0001062E;
const char32_t LINEAR_A_SIGN_AB055 = 0x0001062F;
const char32_t LINEAR_A_SIGN_AB056 = 0x00010630;
const char32_t LINEAR_A_SIGN_AB057 = 0x00010631;
const char32_t LINEAR_A_SIGN_AB058 = 0x00010632;
const char32_t LINEAR_A_SIGN_AB059 = 0x00010633;
const char32_t LINEAR_A_SIGN_AB060 = 0x00010634;
const char32_t LINEAR_A_SIGN_AB061 = 0x00010635;
const char32_t LINEAR_A_SIGN_AB065 = 0x00010636;
const char32_t LINEAR_A_SIGN_AB066 = 0x00010637;
const char32_t LINEAR_A_SIGN_AB067 = 0x00010638;
const char32_t LINEAR_A_SIGN_AB069 = 0x00010639;
const char32_t LINEAR_A_SIGN_AB070 = 0x0001063A;
const char32_t LINEAR_A_SIGN_AB073 = 0x0001063B;
const char32_t LINEAR_A_SIGN_AB074 = 0x0001063C;
const char32_t LINEAR_A_SIGN_AB076 = 0x0001063D;
const char32_t LINEAR_A_SIGN_AB077 = 0x0001063E;
const char32_t LINEAR_A_SIGN_AB078 = 0x0001063F;
const char32_t LINEAR_A_SIGN_AB079 = 0x00010640;
const char32_t LINEAR_A_SIGN_AB080 = 0x00010641;
const char32_t LINEAR_A_SIGN_AB081 = 0x00010642;
const char32_t LINEAR_A_SIGN_AB082 = 0x00010643;
const char32_t LINEAR_A_SIGN_AB085 = 0x00010644;
const char32_t LINEAR_A_SIGN_AB086 = 0x00010645;
const char32_t LINEAR_A_SIGN_AB087 = 0x00010646;
const char32_t LINEAR_A_SIGN_A100_102 = 0x00010647;
const char32_t LINEAR_A_SIGN_AB118 = 0x00010648;
const char32_t LINEAR_A_SIGN_AB120 = 0x00010649;
const char32_t LINEAR_A_SIGN_A120B = 0x0001064A;
const char32_t LINEAR_A_SIGN_AB122 = 0x0001064B;
const char32_t LINEAR_A_SIGN_AB123 = 0x0001064C;
const char32_t LINEAR_A_SIGN_AB131A = 0x0001064D;
const char32_t LINEAR_A_SIGN_AB131B = 0x0001064E;
const char32_t LINEAR_A_SIGN_A131C = 0x0001064F;
const char32_t LINEAR_A_SIGN_AB164 = 0x00010650;
const char32_t LINEAR_A_SIGN_AB171 = 0x00010651;
const char32_t LINEAR_A_SIGN_AB180 = 0x00010652;
const char32_t LINEAR_A_SIGN_AB188 = 0x00010653;
const char32_t LINEAR_A_SIGN_AB191 = 0x00010654;
const char32_t LINEAR_A_SIGN_A301 = 0x00010655;
const char32_t LINEAR_A_SIGN_A302 = 0x00010656;
const char32_t LINEAR_A_SIGN_A303 = 0x00010657;
const char32_t LINEAR_A_SIGN_A304 = 0x00010658;
const char32_t LINEAR_A_SIGN_A305 = 0x00010659;
const char32_t LINEAR_A_SIGN_A306 = 0x0001065A;
const char32_t LINEAR_A_SIGN_A307 = 0x0001065B;
const char32_t LINEAR_A_SIGN_A308 = 0x0001065C;
const char32_t LINEAR_A_SIGN_A309A = 0x0001065D;
const char32_t LINEAR_A_SIGN_A309B = 0x0001065E;
const char32_t LINEAR_A_SIGN_A309C = 0x0001065F;
const char32_t LINEAR_A_SIGN_A310 = 0x00010660;
const char32_t LINEAR_A_SIGN_A311 = 0x00010661;
const char32_t LINEAR_A_SIGN_A312 = 0x00010662;
const char32_t LINEAR_A_SIGN_A313A = 0x00010663;
const char32_t LINEAR_A_SIGN_A313B = 0x00010664;
const char32_t LINEAR_A_SIGN_A313C = 0x00010665;
const char32_t LINEAR_A_SIGN_A314 = 0x00010666;
const char32_t LINEAR_A_SIGN_A315 = 0x00010667;
const char32_t LINEAR_A_SIGN_A316 = 0x00010668;
const char32_t LINEAR_A_SIGN_A317 = 0x00010669;
const char32_t LINEAR_A_SIGN_A318 = 0x0001066A;
const char32_t LINEAR_A_SIGN_A319 = 0x0001066B;
const char32_t LINEAR_A_SIGN_A320 = 0x0001066C;
const char32_t LINEAR_A_SIGN_A321 = 0x0001066D;
const char32_t LINEAR_A_SIGN_A322 = 0x0001066E;
const char32_t LINEAR_A_SIGN_A323 = 0x0001066F;
const char32_t LINEAR_A_SIGN_A324 = 0x00010670;
const char32_t LINEAR_A_SIGN_A325 = 0x00010671;
const char32_t LINEAR_A_SIGN_A326 = 0x00010672;
const char32_t LINEAR_A_SIGN_A327 = 0x00010673;
const char32_t LINEAR_A_SIGN_A328 = 0x00010674;
const char32_t LINEAR_A_SIGN_A329 = 0x00010675;
const char32_t LINEAR_A_SIGN_A330 = 0x00010676;
const char32_t LINEAR_A_SIGN_A331 = 0x00010677;
const char32_t LINEAR_A_SIGN_A332 = 0x00010678;
const char32_t LINEAR_A_SIGN_A333 = 0x00010679;
const char32_t LINEAR_A_SIGN_A334 = 0x0001067A;
const char32_t LINEAR_A_SIGN_A335 = 0x0001067B;
const char32_t LINEAR_A_SIGN_A336 = 0x0001067C;
const char32_t LINEAR_A_SIGN_A337 = 0x0001067D;
const char32_t LINEAR_A_SIGN_A338 = 0x0001067E;
const char32_t LINEAR_A_SIGN_A339 = 0x0001067F;
const char32_t LINEAR_A_SIGN_A340 = 0x00010680;
const char32_t LINEAR_A_SIGN_A341 = 0x00010681;
const char32_t LINEAR_A_SIGN_A342 = 0x00010682;
const char32_t LINEAR_A_SIGN_A343 = 0x00010683;
const char32_t LINEAR_A_SIGN_A344 = 0x00010684;
const char32_t LINEAR_A_SIGN_A345 = 0x00010685;
const char32_t LINEAR_A_SIGN_A346 = 0x00010686;
const char32_t LINEAR_A_SIGN_A347 = 0x00010687;
const char32_t LINEAR_A_SIGN_A348 = 0x00010688;
const char32_t LINEAR_A_SIGN_A349 = 0x00010689;
const char32_t LINEAR_A_SIGN_A350 = 0x0001068A;
const char32_t LINEAR_A_SIGN_A351 = 0x0001068B;
const char32_t LINEAR_A_SIGN_A352 = 0x0001068C;
const char32_t LINEAR_A_SIGN_A353 = 0x0001068D;
const char32_t LINEAR_A_SIGN_A354 = 0x0001068E;
const char32_t LINEAR_A_SIGN_A355 = 0x0001068F;
const char32_t LINEAR_A_SIGN_A356 = 0x00010690;
const char32_t LINEAR_A_SIGN_A357 = 0x00010691;
const char32_t LINEAR_A_SIGN_A358 = 0x00010692;
const char32_t LINEAR_A_SIGN_A359 = 0x00010693;
const char32_t LINEAR_A_SIGN_A360 = 0x00010694;
const char32_t LINEAR_A_SIGN_A361 = 0x00010695;
const char32_t LINEAR_A_SIGN_A362 = 0x00010696;
const char32_t LINEAR_A_SIGN_A363 = 0x00010697;
const char32_t LINEAR_A_SIGN_A364 = 0x00010698;
const char32_t LINEAR_A_SIGN_A365 = 0x00010699;
const char32_t LINEAR_A_SIGN_A366 = 0x0001069A;
const char32_t LINEAR_A_SIGN_A367 = 0x0001069B;
const char32_t LINEAR_A_SIGN_A368 = 0x0001069C;
const char32_t LINEAR_A_SIGN_A369 = 0x0001069D;
const char32_t LINEAR_A_SIGN_A370 = 0x0001069E;
const char32_t LINEAR_A_SIGN_A371 = 0x0001069F;
const char32_t LINEAR_A_SIGN_A400_VAS = 0x000106A0;
const char32_t LINEAR_A_SIGN_A401_VAS = 0x000106A1;
const char32_t LINEAR_A_SIGN_A402_VAS = 0x000106A2;
const char32_t LINEAR_A_SIGN_A403_VAS = 0x000106A3;
const char32_t LINEAR_A_SIGN_A404_VAS = 0x000106A4;
const char32_t LINEAR_A_SIGN_A405_VAS = 0x000106A5;
const char32_t LINEAR_A_SIGN_A406_VAS = 0x000106A6;
const char32_t LINEAR_A_SIGN_A407_VAS = 0x000106A7;
const char32_t LINEAR_A_SIGN_A408_VAS = 0x000106A8;
const char32_t LINEAR_A_SIGN_A409_VAS = 0x000106A9;
const char32_t LINEAR_A_SIGN_A410_VAS = 0x000106AA;
const char32_t LINEAR_A_SIGN_A411_VAS = 0x000106AB;
const char32_t LINEAR_A_SIGN_A412_VAS = 0x000106AC;
const char32_t LINEAR_A_SIGN_A413_VAS = 0x000106AD;
const char32_t LINEAR_A_SIGN_A414_VAS = 0x000106AE;
const char32_t LINEAR_A_SIGN_A415_VAS = 0x000106AF;
const char32_t LINEAR_A_SIGN_A416_VAS = 0x000106B0;
const char32_t LINEAR_A_SIGN_A417_VAS = 0x000106B1;
const char32_t LINEAR_A_SIGN_A418_VAS = 0x000106B2;
const char32_t LINEAR_A_SIGN_A501 = 0x000106B3;
const char32_t LINEAR_A_SIGN_A502 = 0x000106B4;
const char32_t LINEAR_A_SIGN_A503 = 0x000106B5;
const char32_t LINEAR_A_SIGN_A504 = 0x000106B6;
const char32_t LINEAR_A_SIGN_A505 = 0x000106B7;
const char32_t LINEAR_A_SIGN_A506 = 0x000106B8;
const char32_t LINEAR_A_SIGN_A508 = 0x000106B9;
const char32_t LINEAR_A_SIGN_A509 = 0x000106BA;
const char32_t LINEAR_A_SIGN_A510 = 0x000106BB;
const char32_t LINEAR_A_SIGN_A511 = 0x000106BC;
const char32_t LINEAR_A_SIGN_A512 = 0x000106BD;
const char32_t LINEAR_A_SIGN_A513 = 0x000106BE;
const char32_t LINEAR_A_SIGN_A515 = 0x000106BF;
const char32_t LINEAR_A_SIGN_A516 = 0x000106C0;
const char32_t LINEAR_A_SIGN_A520 = 0x000106C1;
const char32_t LINEAR_A_SIGN_A521 = 0x000106C2;
const char32_t LINEAR_A_SIGN_A523 = 0x000106C3;
const char32_t LINEAR_A_SIGN_A524 = 0x000106C4;
const char32_t LINEAR_A_SIGN_A525 = 0x000106C5;
const char32_t LINEAR_A_SIGN_A526 = 0x000106C6;
const char32_t LINEAR_A_SIGN_A527 = 0x000106C7;
const char32_t LINEAR_A_SIGN_A528 = 0x000106C8;
const char32_t LINEAR_A_SIGN_A529 = 0x000106C9;
const char32_t LINEAR_A_SIGN_A530 = 0x000106CA;
const char32_t LINEAR_A_SIGN_A531 = 0x000106CB;
const char32_t LINEAR_A_SIGN_A532 = 0x000106CC;
const char32_t LINEAR_A_SIGN_A534 = 0x000106CD;
const char32_t LINEAR_A_SIGN_A535 = 0x000106CE;
const char32_t LINEAR_A_SIGN_A536 = 0x000106CF;
const char32_t LINEAR_A_SIGN_A537 = 0x000106D0;
const char32_t LINEAR_A_SIGN_A538 = 0x000106D1;
const char32_t LINEAR_A_SIGN_A539 = 0x000106D2;
const char32_t LINEAR_A_SIGN_A540 = 0x000106D3;
const char32_t LINEAR_A_SIGN_A541 = 0x000106D4;
const char32_t LINEAR_A_SIGN_A542 = 0x000106D5;
const char32_t LINEAR_A_SIGN_A545 = 0x000106D6;
const char32_t LINEAR_A_SIGN_A547 = 0x000106D7;
const char32_t LINEAR_A_SIGN_A548 = 0x000106D8;
const char32_t LINEAR_A_SIGN_A549 = 0x000106D9;
const char32_t LINEAR_A_SIGN_A550 = 0x000106DA;
const char32_t LINEAR_A_SIGN_A551 = 0x000106DB;
const char32_t LINEAR_A_SIGN_A552 = 0x000106DC;
const char32_t LINEAR_A_SIGN_A553 = 0x000106DD;
const char32_t LINEAR_A_SIGN_A554 = 0x000106DE;
const char32_t LINEAR_A_SIGN_A555 = 0x000106DF;
const char32_t LINEAR_A_SIGN_A556 = 0x000106E0;
const char32_t LINEAR_A_SIGN_A557 = 0x000106E1;
const char32_t LINEAR_A_SIGN_A559 = 0x000106E2;
const char32_t LINEAR_A_SIGN_A563 = 0x000106E3;
const char32_t LINEAR_A_SIGN_A564 = 0x000106E4;
const char32_t LINEAR_A_SIGN_A565 = 0x000106E5;
const char32_t LINEAR_A_SIGN_A566 = 0x000106E6;
const char32_t LINEAR_A_SIGN_A568 = 0x000106E7;
const char32_t LINEAR_A_SIGN_A569 = 0x000106E8;
const char32_t LINEAR_A_SIGN_A570 = 0x000106E9;
const char32_t LINEAR_A_SIGN_A571 = 0x000106EA;
const char32_t LINEAR_A_SIGN_A572 = 0x000106EB;
const char32_t LINEAR_A_SIGN_A573 = 0x000106EC;
const char32_t LINEAR_A_SIGN_A574 = 0x000106ED;
const char32_t LINEAR_A_SIGN_A575 = 0x000106EE;
const char32_t LINEAR_A_SIGN_A576 = 0x000106EF;
const char32_t LINEAR_A_SIGN_A577 = 0x000106F0;
const char32_t LINEAR_A_SIGN_A578 = 0x000106F1;
const char32_t LINEAR_A_SIGN_A579 = 0x000106F2;
const char32_t LINEAR_A_SIGN_A580 = 0x000106F3;
const char32_t LINEAR_A_SIGN_A581 = 0x000106F4;
const char32_t LINEAR_A_SIGN_A582 = 0x000106F5;
const char32_t LINEAR_A_SIGN_A583 = 0x000106F6;
const char32_t LINEAR_A_SIGN_A584 = 0x000106F7;
const char32_t LINEAR_A_SIGN_A585 = 0x000106F8;
const char32_t LINEAR_A_SIGN_A586 = 0x000106F9;
const char32_t LINEAR_A_SIGN_A587 = 0x000106FA;
const char32_t LINEAR_A_SIGN_A588 = 0x000106FB;
const char32_t LINEAR_A_SIGN_A589 = 0x000106FC;
const char32_t LINEAR_A_SIGN_A591 = 0x000106FD;
const char32_t LINEAR_A_SIGN_A592 = 0x000106FE;
const char32_t LINEAR_A_SIGN_A594 = 0x000106FF;
const char32_t LINEAR_A_SIGN_A595 = 0x00010700;
const char32_t LINEAR_A_SIGN_A596 = 0x00010701;
const char32_t LINEAR_A_SIGN_A598 = 0x00010702;
const char32_t LINEAR_A_SIGN_A600 = 0x00010703;
const char32_t LINEAR_A_SIGN_A601 = 0x00010704;
const char32_t LINEAR_A_SIGN_A602 = 0x00010705;
const char32_t LINEAR_A_SIGN_A603 = 0x00010706;
const char32_t LINEAR_A_SIGN_A604 = 0x00010707;
const char32_t LINEAR_A_SIGN_A606 = 0x00010708;
const char32_t LINEAR_A_SIGN_A608 = 0x00010709;
const char32_t LINEAR_A_SIGN_A609 = 0x0001070A;
const char32_t LINEAR_A_SIGN_A610 = 0x0001070B;
const char32_t LINEAR_A_SIGN_A611 = 0x0001070C;
const char32_t LINEAR_A_SIGN_A612 = 0x0001070D;
const char32_t LINEAR_A_SIGN_A613 = 0x0001070E;
const char32_t LINEAR_A_SIGN_A614 = 0x0001070F;
const char32_t LINEAR_A_SIGN_A615 = 0x00010710;
const char32_t LINEAR_A_SIGN_A616 = 0x00010711;
const char32_t LINEAR_A_SIGN_A617 = 0x00010712;
const char32_t LINEAR_A_SIGN_A618 = 0x00010713;
const char32_t LINEAR_A_SIGN_A619 = 0x00010714;
const char32_t LINEAR_A_SIGN_A620 = 0x00010715;
const char32_t LINEAR_A_SIGN_A621 = 0x00010716;
const char32_t LINEAR_A_SIGN_A622 = 0x00010717;
const char32_t LINEAR_A_SIGN_A623 = 0x00010718;
const char32_t LINEAR_A_SIGN_A624 = 0x00010719;
const char32_t LINEAR_A_SIGN_A626 = 0x0001071A;
const char32_t LINEAR_A_SIGN_A627 = 0x0001071B;
const char32_t LINEAR_A_SIGN_A628 = 0x0001071C;
const char32_t LINEAR_A_SIGN_A629 = 0x0001071D;
const char32_t LINEAR_A_SIGN_A634 = 0x0001071E;
const char32_t LINEAR_A_SIGN_A637 = 0x0001071F;
const char32_t LINEAR_A_SIGN_A638 = 0x00010720;
const char32_t LINEAR_A_SIGN_A640 = 0x00010721;
const char32_t LINEAR_A_SIGN_A642 = 0x00010722;
const char32_t LINEAR_A_SIGN_A643 = 0x00010723;
const char32_t LINEAR_A_SIGN_A644 = 0x00010724;
const char32_t LINEAR_A_SIGN_A645 = 0x00010725;
const char32_t LINEAR_A_SIGN_A646 = 0x00010726;
const char32_t LINEAR_A_SIGN_A648 = 0x00010727;
const char32_t LINEAR_A_SIGN_A649 = 0x00010728;
const char32_t LINEAR_A_SIGN_A651 = 0x00010729;
const char32_t LINEAR_A_SIGN_A652 = 0x0001072A;
const char32_t LINEAR_A_SIGN_A653 = 0x0001072B;
const char32_t LINEAR_A_SIGN_A654 = 0x0001072C;
const char32_t LINEAR_A_SIGN_A655 = 0x0001072D;
const char32_t LINEAR_A_SIGN_A656 = 0x0001072E;
const char32_t LINEAR_A_SIGN_A657 = 0x0001072F;
const char32_t LINEAR_A_SIGN_A658 = 0x00010730;
const char32_t LINEAR_A_SIGN_A659 = 0x00010731;
const char32_t LINEAR_A_SIGN_A660 = 0x00010732;
const char32_t LINEAR_A_SIGN_A661 = 0x00010733;
const char32_t LINEAR_A_SIGN_A662 = 0x00010734;
const char32_t LINEAR_A_SIGN_A663 = 0x00010735;
const char32_t LINEAR_A_SIGN_A664 = 0x00010736;
const char32_t LINEAR_A_SIGN_A701_A = 0x00010740;
const char32_t LINEAR_A_SIGN_A702_B = 0x00010741;
const char32_t LINEAR_A_SIGN_A703_D = 0x00010742;
const char32_t LINEAR_A_SIGN_A704_E = 0x00010743;
const char32_t LINEAR_A_SIGN_A705_F = 0x00010744;
const char32_t LINEAR_A_SIGN_A706_H = 0x00010745;
const char32_t LINEAR_A_SIGN_A707_J = 0x00010746;
const char32_t LINEAR_A_SIGN_A708_K = 0x00010747;
const char32_t LINEAR_A_SIGN_A709_L = 0x00010748;
const char32_t LINEAR_A_SIGN_A709_2_L2 = 0x00010749;
const char32_t LINEAR_A_SIGN_A709_3_L3 = 0x0001074A;
const char32_t LINEAR_A_SIGN_A709_4_L4 = 0x0001074B;
const char32_t LINEAR_A_SIGN_A709_6_L6 = 0x0001074C;
const char32_t LINEAR_A_SIGN_A710_W = 0x0001074D;
const char32_t LINEAR_A_SIGN_A711_X = 0x0001074E;
const char32_t LINEAR_A_SIGN_A712_Y = 0x0001074F;
const char32_t LINEAR_A_SIGN_A713_OMEGA = 0x00010750;
const char32_t LINEAR_A_SIGN_A714_ABB = 0x00010751;
const char32_t LINEAR_A_SIGN_A715_BB = 0x00010752;
const char32_t LINEAR_A_SIGN_A717_DD = 0x00010753;
const char32_t LINEAR_A_SIGN_A726_EYYY = 0x00010754;
const char32_t LINEAR_A_SIGN_A732_JE = 0x00010755;
const char32_t LINEAR_A_SIGN_A800 = 0x00010760;
const char32_t LINEAR_A_SIGN_A801 = 0x00010761;
const char32_t LINEAR_A_SIGN_A802 = 0x00010762;
const char32_t LINEAR_A_SIGN_A803 = 0x00010763;
const char32_t LINEAR_A_SIGN_A804 = 0x00010764;
const char32_t LINEAR_A_SIGN_A805 = 0x00010765;
const char32_t LINEAR_A_SIGN_A806 = 0x00010766;
const char32_t LINEAR_A_SIGN_A807 = 0x00010767;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_AA = 0x00010780;
const char32_t MODIFIER_LETTER_SUPERSCRIPT_TRIANGULAR_COLON = 0x00010781;
const char32_t MODIFIER_LETTER_SUPERSCRIPT_HALF_TRIANGULAR_COLON = 0x00010782;
const char32_t MODIFIER_LETTER_SMALL_AE = 0x00010783;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_B = 0x00010784;
const char32_t MODIFIER_LETTER_SMALL_B_WITH_HOOK = 0x00010785;
const char32_t MODIFIER_LETTER_SMALL_DZ_DIGRAPH = 0x00010787;
const char32_t MODIFIER_LETTER_SMALL_DZ_DIGRAPH_WITH_RETROFLEX_HOOK = 0x00010788;
const char32_t MODIFIER_LETTER_SMALL_DZ_DIGRAPH_WITH_CURL = 0x00010789;
const char32_t MODIFIER_LETTER_SMALL_DEZH_DIGRAPH = 0x0001078A;
const char32_t MODIFIER_LETTER_SMALL_D_WITH_TAIL = 0x0001078B;
const char32_t MODIFIER_LETTER_SMALL_D_WITH_HOOK = 0x0001078C;
const char32_t MODIFIER_LETTER_SMALL_D_WITH_HOOK_AND_TAIL = 0x0001078D;
const char32_t MODIFIER_LETTER_SMALL_REVERSED_E = 0x0001078E;
const char32_t MODIFIER_LETTER_SMALL_CLOSED_REVERSED_OPEN_E = 0x0001078F;
const char32_t MODIFIER_LETTER_SMALL_FENG_DIGRAPH = 0x00010790;
const char32_t MODIFIER_LETTER_SMALL_RAMS_HORN = 0x00010791;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_G = 0x00010792;
const char32_t MODIFIER_LETTER_SMALL_G_WITH_HOOK = 0x00010793;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_G_WITH_HOOK = 0x00010794;
const char32_t MODIFIER_LETTER_SMALL_H_WITH_STROKE = 0x00010795;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_H = 0x00010796;
const char32_t MODIFIER_LETTER_SMALL_HENG_WITH_HOOK = 0x00010797;
const char32_t MODIFIER_LETTER_SMALL_DOTLESS_J_WITH_STROKE_AND_HOOK = 0x00010798;
const char32_t MODIFIER_LETTER_SMALL_LS_DIGRAPH = 0x00010799;
const char32_t MODIFIER_LETTER_SMALL_LZ_DIGRAPH = 0x0001079A;
const char32_t MODIFIER_LETTER_SMALL_L_WITH_BELT = 0x0001079B;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_L_WITH_BELT = 0x0001079C;
const char32_t MODIFIER_LETTER_SMALL_L_WITH_RETROFLEX_HOOK_AND_BELT = 0x0001079D;
const char32_t MODIFIER_LETTER_SMALL_LEZH = 0x0001079E;
const char32_t MODIFIER_LETTER_SMALL_LEZH_WITH_RETROFLEX_HOOK = 0x0001079F;
const char32_t MODIFIER_LETTER_SMALL_TURNED_Y = 0x000107A0;
const char32_t MODIFIER_LETTER_SMALL_TURNED_Y_WITH_BELT = 0x000107A1;
const char32_t MODIFIER_LETTER_SMALL_O_WITH_STROKE = 0x000107A2;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_OE = 0x000107A3;
const char32_t MODIFIER_LETTER_SMALL_CLOSED_OMEGA = 0x000107A4;
const char32_t MODIFIER_LETTER_SMALL_Q = 0x000107A5;
const char32_t MODIFIER_LETTER_SMALL_TURNED_R_WITH_LONG_LEG = 0x000107A6;
const char32_t MODIFIER_LETTER_SMALL_TURNED_R_WITH_LONG_LEG_AND_RETROFLEX_HOOK = 0x000107A7;
const char32_t MODIFIER_LETTER_SMALL_R_WITH_TAIL = 0x000107A8;
const char32_t MODIFIER_LETTER_SMALL_R_WITH_FISHHOOK = 0x000107A9;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_R = 0x000107AA;
const char32_t MODIFIER_LETTER_SMALL_TC_DIGRAPH_WITH_CURL = 0x000107AB;
const char32_t MODIFIER_LETTER_SMALL_TS_DIGRAPH = 0x000107AC;
const char32_t MODIFIER_LETTER_SMALL_TS_DIGRAPH_WITH_RETROFLEX_HOOK = 0x000107AD;
const char32_t MODIFIER_LETTER_SMALL_TESH_DIGRAPH = 0x000107AE;
const char32_t MODIFIER_LETTER_SMALL_T_WITH_RETROFLEX_HOOK = 0x000107AF;
const char32_t MODIFIER_LETTER_SMALL_V_WITH_RIGHT_HOOK = 0x000107B0;
const char32_t MODIFIER_LETTER_SMALL_CAPITAL_Y = 0x000107B2;
const char32_t MODIFIER_LETTER_GLOTTAL_STOP_WITH_STROKE = 0x000107B3;
const char32_t MODIFIER_LETTER_REVERSED_GLOTTAL_STOP_WITH_STROKE = 0x000107B4;
const char32_t MODIFIER_LETTER_BILABIAL_CLICK = 0x000107B5;
const char32_t MODIFIER_LETTER_DENTAL_CLICK = 0x000107B6;
const char32_t MODIFIER_LETTER_LATERAL_CLICK = 0x000107B7;
const char32_t MODIFIER_LETTER_ALVEOLAR_CLICK = 0x000107B8;
const char32_t MODIFIER_LETTER_RETROFLEX_CLICK_WITH_RETROFLEX_HOOK = 0x000107B9;
const char32_t MODIFIER_LETTER_SMALL_S_WITH_CURL = 0x000107BA;
const char32_t CYPRIOT_SYLLABLE_A = 0x00010800;
const char32_t CYPRIOT_SYLLABLE_E = 0x00010801;
const char32_t CYPRIOT_SYLLABLE_I = 0x00010802;
const char32_t CYPRIOT_SYLLABLE_O = 0x00010803;
const char32_t CYPRIOT_SYLLABLE_U = 0x00010804;
const char32_t CYPRIOT_SYLLABLE_JA = 0x00010805;
const char32_t CYPRIOT_SYLLABLE_JO = 0x00010808;
const char32_t CYPRIOT_SYLLABLE_KA = 0x0001080A;
const char32_t CYPRIOT_SYLLABLE_KE = 0x0001080B;
const char32_t CYPRIOT_SYLLABLE_KI = 0x0001080C;
const char32_t CYPRIOT_SYLLABLE_KO = 0x0001080D;
const char32_t CYPRIOT_SYLLABLE_KU = 0x0001080E;
const char32_t CYPRIOT_SYLLABLE_LA = 0x0001080F;
const char32_t CYPRIOT_SYLLABLE_LE = 0x00010810;
const char32_t CYPRIOT_SYLLABLE_LI = 0x00010811;
const char32_t CYPRIOT_SYLLABLE_LO = 0x00010812;
const char32_t CYPRIOT_SYLLABLE_LU = 0x00010813;
const char32_t CYPRIOT_SYLLABLE_MA = 0x00010814;
const char32_t CYPRIOT_SYLLABLE_ME = 0x00010815;
const char32_t CYPRIOT_SYLLABLE_MI = 0x00010816;
const char32_t CYPRIOT_SYLLABLE_MO = 0x00010817;
const char32_t CYPRIOT_SYLLABLE_MU = 0x00010818;
const char32_t CYPRIOT_SYLLABLE_NA = 0x00010819;
const char32_t CYPRIOT_SYLLABLE_NE = 0x0001081A;
const char32_t CYPRIOT_SYLLABLE_NI = 0x0001081B;
const char32_t CYPRIOT_SYLLABLE_NO = 0x0001081C;
const char32_t CYPRIOT_SYLLABLE_NU = 0x0001081D;
const char32_t CYPRIOT_SYLLABLE_PA = 0x0001081E;
const char32_t CYPRIOT_SYLLABLE_PE = 0x0001081F;
const char32_t CYPRIOT_SYLLABLE_PI = 0x00010820;
const char32_t CYPRIOT_SYLLABLE_PO = 0x00010821;
const char32_t CYPRIOT_SYLLABLE_PU = 0x00010822;
const char32_t CYPRIOT_SYLLABLE_RA = 0x00010823;
const char32_t CYPRIOT_SYLLABLE_RE = 0x00010824;
const char32_t CYPRIOT_SYLLABLE_RI = 0x00010825;
const char32_t CYPRIOT_SYLLABLE_RO = 0x00010826;
const char32_t CYPRIOT_SYLLABLE_RU = 0x00010827;
const char32_t CYPRIOT_SYLLABLE_SA = 0x00010828;
const char32_t CYPRIOT_SYLLABLE_SE = 0x00010829;
const char32_t CYPRIOT_SYLLABLE_SI = 0x0001082A;
const char32_t CYPRIOT_SYLLABLE_SO = 0x0001082B;
const char32_t CYPRIOT_SYLLABLE_SU = 0x0001082C;
const char32_t CYPRIOT_SYLLABLE_TA = 0x0001082D;
const char32_t CYPRIOT_SYLLABLE_TE = 0x0001082E;
const char32_t CYPRIOT_SYLLABLE_TI = 0x0001082F;
const char32_t CYPRIOT_SYLLABLE_TO = 0x00010830;
const char32_t CYPRIOT_SYLLABLE_TU = 0x00010831;
const char32_t CYPRIOT_SYLLABLE_WA = 0x00010832;
const char32_t CYPRIOT_SYLLABLE_WE = 0x00010833;
const char32_t CYPRIOT_SYLLABLE_WI = 0x00010834;
const char32_t CYPRIOT_SYLLABLE_WO = 0x00010835;
const char32_t CYPRIOT_SYLLABLE_XA = 0x00010837;
const char32_t CYPRIOT_SYLLABLE_XE = 0x00010838;
const char32_t CYPRIOT_SYLLABLE_ZA = 0x0001083C;
const char32_t CYPRIOT_SYLLABLE_ZO = 0x0001083F;
const char32_t IMPERIAL_ARAMAIC_LETTER_ALEPH = 0x00010840;
const char32_t IMPERIAL_ARAMAIC_LETTER_BETH = 0x00010841;
const char32_t IMPERIAL_ARAMAIC_LETTER_GIMEL = 0x00010842;
const char32_t IMPERIAL_ARAMAIC_LETTER_DALETH = 0x00010843;
const char32_t IMPERIAL_ARAMAIC_LETTER_HE = 0x00010844;
const char32_t IMPERIAL_ARAMAIC_LETTER_WAW = 0x00010845;
const char32_t IMPERIAL_ARAMAIC_LETTER_ZAYIN = 0x00010846;
const char32_t IMPERIAL_ARAMAIC_LETTER_HETH = 0x00010847;
const char32_t IMPERIAL_ARAMAIC_LETTER_TETH = 0x00010848;
const char32_t IMPERIAL_ARAMAIC_LETTER_YODH = 0x00010849;
const char32_t IMPERIAL_ARAMAIC_LETTER_KAPH = 0x0001084A;
const char32_t IMPERIAL_ARAMAIC_LETTER_LAMEDH = 0x0001084B;
const char32_t IMPERIAL_ARAMAIC_LETTER_MEM = 0x0001084C;
const char32_t IMPERIAL_ARAMAIC_LETTER_NUN = 0x0001084D;
const char32_t IMPERIAL_ARAMAIC_LETTER_SAMEKH = 0x0001084E;
const char32_t IMPERIAL_ARAMAIC_LETTER_AYIN = 0x0001084F;
const char32_t IMPERIAL_ARAMAIC_LETTER_PE = 0x00010850;
const char32_t IMPERIAL_ARAMAIC_LETTER_SADHE = 0x00010851;
const char32_t IMPERIAL_ARAMAIC_LETTER_QOPH = 0x00010852;
const char32_t IMPERIAL_ARAMAIC_LETTER_RESH = 0x00010853;
const char32_t IMPERIAL_ARAMAIC_LETTER_SHIN = 0x00010854;
const char32_t IMPERIAL_ARAMAIC_LETTER_TAW = 0x00010855;
const char32_t IMPERIAL_ARAMAIC_SECTION_SIGN = 0x00010857;
const char32_t IMPERIAL_ARAMAIC_NUMBER_ONE = 0x00010858;
const char32_t IMPERIAL_ARAMAIC_NUMBER_TWO = 0x00010859;
const char32_t IMPERIAL_ARAMAIC_NUMBER_THREE = 0x0001085A;
const char32_t IMPERIAL_ARAMAIC_NUMBER_TEN = 0x0001085B;
const char32_t IMPERIAL_ARAMAIC_NUMBER_TWENTY = 0x0001085C;
const char32_t IMPERIAL_ARAMAIC_NUMBER_ONE_HUNDRED = 0x0001085D;
const char32_t IMPERIAL_ARAMAIC_NUMBER_ONE_THOUSAND = 0x0001085E;
const char32_t IMPERIAL_ARAMAIC_NUMBER_TEN_THOUSAND = 0x0001085F;
const char32_t PALMYRENE_LETTER_ALEPH = 0x00010860;
const char32_t PALMYRENE_LETTER_BETH = 0x00010861;
const char32_t PALMYRENE_LETTER_GIMEL = 0x00010862;
const char32_t PALMYRENE_LETTER_DALETH = 0x00010863;
const char32_t PALMYRENE_LETTER_HE = 0x00010864;
const char32_t PALMYRENE_LETTER_WAW = 0x00010865;
const char32_t PALMYRENE_LETTER_ZAYIN = 0x00010866;
const char32_t PALMYRENE_LETTER_HETH = 0x00010867;
const char32_t PALMYRENE_LETTER_TETH = 0x00010868;
const char32_t PALMYRENE_LETTER_YODH = 0x00010869;
const char32_t PALMYRENE_LETTER_KAPH = 0x0001086A;
const char32_t PALMYRENE_LETTER_LAMEDH = 0x0001086B;
const char32_t PALMYRENE_LETTER_MEM = 0x0001086C;
const char32_t PALMYRENE_LETTER_FINAL_NUN = 0x0001086D;
const char32_t PALMYRENE_LETTER_NUN = 0x0001086E;
const char32_t PALMYRENE_LETTER_SAMEKH = 0x0001086F;
const char32_t PALMYRENE_LETTER_AYIN = 0x00010870;
const char32_t PALMYRENE_LETTER_PE = 0x00010871;
const char32_t PALMYRENE_LETTER_SADHE = 0x00010872;
const char32_t PALMYRENE_LETTER_QOPH = 0x00010873;
const char32_t PALMYRENE_LETTER_RESH = 0x00010874;
const char32_t PALMYRENE_LETTER_SHIN = 0x00010875;
const char32_t PALMYRENE_LETTER_TAW = 0x00010876;
const char32_t PALMYRENE_LEFT_POINTING_FLEURON = 0x00010877;
const char32_t PALMYRENE_RIGHT_POINTING_FLEURON = 0x00010878;
const char32_t PALMYRENE_NUMBER_ONE = 0x00010879;
const char32_t PALMYRENE_NUMBER_TWO = 0x0001087A;
const char32_t PALMYRENE_NUMBER_THREE = 0x0001087B;
const char32_t PALMYRENE_NUMBER_FOUR = 0x0001087C;
const char32_t PALMYRENE_NUMBER_FIVE = 0x0001087D;
const char32_t PALMYRENE_NUMBER_TEN = 0x0001087E;
const char32_t PALMYRENE_NUMBER_TWENTY = 0x0001087F;
const char32_t NABATAEAN_LETTER_FINAL_ALEPH = 0x00010880;
const char32_t NABATAEAN_LETTER_ALEPH = 0x00010881;
const char32_t NABATAEAN_LETTER_FINAL_BETH = 0x00010882;
const char32_t NABATAEAN_LETTER_BETH = 0x00010883;
const char32_t NABATAEAN_LETTER_GIMEL = 0x00010884;
const char32_t NABATAEAN_LETTER_DALETH = 0x00010885;
const char32_t NABATAEAN_LETTER_FINAL_HE = 0x00010886;
const char32_t NABATAEAN_LETTER_HE = 0x00010887;
const char32_t NABATAEAN_LETTER_WAW = 0x00010888;
const char32_t NABATAEAN_LETTER_ZAYIN = 0x00010889;
const char32_t NABATAEAN_LETTER_HETH = 0x0001088A;
const char32_t NABATAEAN_LETTER_TETH = 0x0001088B;
const char32_t NABATAEAN_LETTER_FINAL_YODH = 0x0001088C;
const char32_t NABATAEAN_LETTER_YODH = 0x0001088D;
const char32_t NABATAEAN_LETTER_FINAL_KAPH = 0x0001088E;
const char32_t NABATAEAN_LETTER_KAPH = 0x0001088F;
const char32_t NABATAEAN_LETTER_FINAL_LAMEDH = 0x00010890;
const char32_t NABATAEAN_LETTER_LAMEDH = 0x00010891;
const char32_t NABATAEAN_LETTER_FINAL_MEM = 0x00010892;
const char32_t NABATAEAN_LETTER_MEM = 0x00010893;
const char32_t NABATAEAN_LETTER_FINAL_NUN = 0x00010894;
const char32_t NABATAEAN_LETTER_NUN = 0x00010895;
const char32_t NABATAEAN_LETTER_SAMEKH = 0x00010896;
const char32_t NABATAEAN_LETTER_AYIN = 0x00010897;
const char32_t NABATAEAN_LETTER_PE = 0x00010898;
const char32_t NABATAEAN_LETTER_SADHE = 0x00010899;
const char32_t NABATAEAN_LETTER_QOPH = 0x0001089A;
const char32_t NABATAEAN_LETTER_RESH = 0x0001089B;
const char32_t NABATAEAN_LETTER_FINAL_SHIN = 0x0001089C;
const char32_t NABATAEAN_LETTER_SHIN = 0x0001089D;
const char32_t NABATAEAN_LETTER_TAW = 0x0001089E;
const char32_t NABATAEAN_NUMBER_ONE = 0x000108A7;
const char32_t NABATAEAN_NUMBER_TWO = 0x000108A8;
const char32_t NABATAEAN_NUMBER_THREE = 0x000108A9;
const char32_t NABATAEAN_NUMBER_FOUR = 0x000108AA;
const char32_t NABATAEAN_CRUCIFORM_NUMBER_FOUR = 0x000108AB;
const char32_t NABATAEAN_NUMBER_FIVE = 0x000108AC;
const char32_t NABATAEAN_NUMBER_TEN = 0x000108AD;
const char32_t NABATAEAN_NUMBER_TWENTY = 0x000108AE;
const char32_t NABATAEAN_NUMBER_ONE_HUNDRED = 0x000108AF;
const char32_t HATRAN_LETTER_ALEPH = 0x000108E0;
const char32_t HATRAN_LETTER_BETH = 0x000108E1;
const char32_t HATRAN_LETTER_GIMEL = 0x000108E2;
const char32_t HATRAN_LETTER_DALETH_RESH = 0x000108E3;
const char32_t HATRAN_LETTER_HE = 0x000108E4;
const char32_t HATRAN_LETTER_WAW = 0x000108E5;
const char32_t HATRAN_LETTER_ZAYN = 0x000108E6;
const char32_t HATRAN_LETTER_HETH = 0x000108E7;
const char32_t HATRAN_LETTER_TETH = 0x000108E8;
const char32_t HATRAN_LETTER_YODH = 0x000108E9;
const char32_t HATRAN_LETTER_KAPH = 0x000108EA;
const char32_t HATRAN_LETTER_LAMEDH = 0x000108EB;
const char32_t HATRAN_LETTER_MEM = 0x000108EC;
const char32_t HATRAN_LETTER_NUN = 0x000108ED;
const char32_t HATRAN_LETTER_SAMEKH = 0x000108EE;
const char32_t HATRAN_LETTER_AYN = 0x000108EF;
const char32_t HATRAN_LETTER_PE = 0x000108F0;
const char32_t HATRAN_LETTER_SADHE = 0x000108F1;
const char32_t HATRAN_LETTER_QOPH = 0x000108F2;
const char32_t HATRAN_LETTER_SHIN = 0x000108F4;
const char32_t HATRAN_LETTER_TAW = 0x000108F5;
const char32_t HATRAN_NUMBER_ONE = 0x000108FB;
const char32_t HATRAN_NUMBER_FIVE = 0x000108FC;
const char32_t HATRAN_NUMBER_TEN = 0x000108FD;
const char32_t HATRAN_NUMBER_TWENTY = 0x000108FE;
const char32_t HATRAN_NUMBER_ONE_HUNDRED = 0x000108FF;
const char32_t PHOENICIAN_LETTER_ALF = 0x00010900;
const char32_t PHOENICIAN_LETTER_BET = 0x00010901;
const char32_t PHOENICIAN_LETTER_GAML = 0x00010902;
const char32_t PHOENICIAN_LETTER_DELT = 0x00010903;
const char32_t PHOENICIAN_LETTER_HE = 0x00010904;
const char32_t PHOENICIAN_LETTER_WAU = 0x00010905;
const char32_t PHOENICIAN_LETTER_ZAI = 0x00010906;
const char32_t PHOENICIAN_LETTER_HET = 0x00010907;
const char32_t PHOENICIAN_LETTER_TET = 0x00010908;
const char32_t PHOENICIAN_LETTER_YOD = 0x00010909;
const char32_t PHOENICIAN_LETTER_KAF = 0x0001090A;
const char32_t PHOENICIAN_LETTER_LAMD = 0x0001090B;
const char32_t PHOENICIAN_LETTER_MEM = 0x0001090C;
const char32_t PHOENICIAN_LETTER_NUN = 0x0001090D;
const char32_t PHOENICIAN_LETTER_SEMK = 0x0001090E;
const char32_t PHOENICIAN_LETTER_AIN = 0x0001090F;
const char32_t PHOENICIAN_LETTER_PE = 0x00010910;
const char32_t PHOENICIAN_LETTER_SADE = 0x00010911;
const char32_t PHOENICIAN_LETTER_QOF = 0x00010912;
const char32_t PHOENICIAN_LETTER_ROSH = 0x00010913;
const char32_t PHOENICIAN_LETTER_SHIN = 0x00010914;
const char32_t PHOENICIAN_LETTER_TAU = 0x00010915;
const char32_t PHOENICIAN_NUMBER_ONE = 0x00010916;
const char32_t PHOENICIAN_NUMBER_TEN = 0x00010917;
const char32_t PHOENICIAN_NUMBER_TWENTY = 0x00010918;
const char32_t PHOENICIAN_NUMBER_ONE_HUNDRED = 0x00010919;
const char32_t PHOENICIAN_NUMBER_TWO = 0x0001091A;
const char32_t PHOENICIAN_NUMBER_THREE = 0x0001091B;
const char32_t PHOENICIAN_WORD_SEPARATOR = 0x0001091F;
const char32_t LYDIAN_LETTER_A = 0x00010920;
const char32_t LYDIAN_LETTER_B = 0x00010921;
const char32_t LYDIAN_LETTER_G = 0x00010922;
const char32_t LYDIAN_LETTER_D = 0x00010923;
const char32_t LYDIAN_LETTER_E = 0x00010924;
const char32_t LYDIAN_LETTER_V = 0x00010925;
const char32_t LYDIAN_LETTER_I = 0x00010926;
const char32_t LYDIAN_LETTER_Y = 0x00010927;
const char32_t LYDIAN_LETTER_K = 0x00010928;
const char32_t LYDIAN_LETTER_L = 0x00010929;
const char32_t LYDIAN_LETTER_M = 0x0001092A;
const char32_t LYDIAN_LETTER_N = 0x0001092B;
const char32_t LYDIAN_LETTER_O = 0x0001092C;
const char32_t LYDIAN_LETTER_R = 0x0001092D;
const char32_t LYDIAN_LETTER_SS = 0x0001092E;
const char32_t LYDIAN_LETTER_T = 0x0001092F;
const char32_t LYDIAN_LETTER_U = 0x00010930;
const char32_t LYDIAN_LETTER_F = 0x00010931;
const char32_t LYDIAN_LETTER_Q = 0x00010932;
const char32_t LYDIAN_LETTER_S = 0x00010933;
const char32_t LYDIAN_LETTER_TT = 0x00010934;
const char32_t LYDIAN_LETTER_AN = 0x00010935;
const char32_t LYDIAN_LETTER_EN = 0x00010936;
const char32_t LYDIAN_LETTER_LY = 0x00010937;
const char32_t LYDIAN_LETTER_NN = 0x00010938;
const char32_t LYDIAN_LETTER_C = 0x00010939;
const char32_t LYDIAN_TRIANGULAR_MARK = 0x0001093F;
const char32_t SIDETIC_LETTER_N01 = 0x00010940;
const char32_t SIDETIC_LETTER_N02 = 0x00010941;
const char32_t SIDETIC_LETTER_N03 = 0x00010942;
const char32_t SIDETIC_LETTER_N04 = 0x00010943;
const char32_t SIDETIC_LETTER_N05 = 0x00010944;
const char32_t SIDETIC_LETTER_N06 = 0x00010945;
const char32_t SIDETIC_LETTER_N07 = 0x00010946;
const char32_t SIDETIC_LETTER_N08 = 0x00010947;
const char32_t SIDETIC_LETTER_N09 = 0x00010948;
const char32_t SIDETIC_LETTER_N10 = 0x00010949;
const char32_t SIDETIC_LETTER_N11 = 0x0001094A;
const char32_t SIDETIC_LETTER_N12 = 0x0001094B;
const char32_t SIDETIC_LETTER_N13 = 0x0001094C;
const char32_t SIDETIC_LETTER_N14 = 0x0001094D;
const char32_t SIDETIC_LETTER_N15 = 0x0001094E;
const char32_t SIDETIC_LETTER_N16 = 0x0001094F;
const char32_t SIDETIC_LETTER_N17 = 0x00010950;
const char32_t SIDETIC_LETTER_N18 = 0x00010951;
const char32_t SIDETIC_LETTER_N19 = 0x00010952;
const char32_t SIDETIC_LETTER_N20 = 0x00010953;
const char32_t SIDETIC_LETTER_N21 = 0x00010954;
const char32_t SIDETIC_LETTER_N22 = 0x00010955;
const char32_t SIDETIC_LETTER_N23 = 0x00010956;
const char32_t SIDETIC_LETTER_N24 = 0x00010957;
const char32_t SIDETIC_LETTER_N25 = 0x00010958;
const char32_t SIDETIC_LETTER_N26 = 0x00010959;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_A = 0x00010980;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_E = 0x00010981;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_I = 0x00010982;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_O = 0x00010983;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_YA = 0x00010984;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_WA = 0x00010985;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_BA = 0x00010986;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_BA_2 = 0x00010987;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_PA = 0x00010988;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_MA = 0x00010989;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_NA = 0x0001098A;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_NA_2 = 0x0001098B;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_NE = 0x0001098C;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_NE_2 = 0x0001098D;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_RA = 0x0001098E;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_RA_2 = 0x0001098F;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_LA = 0x00010990;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_KHA = 0x00010991;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_HHA = 0x00010992;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_SA = 0x00010993;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_SA_2 = 0x00010994;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_SE = 0x00010995;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_KA = 0x00010996;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_QA = 0x00010997;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_TA = 0x00010998;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_TA_2 = 0x00010999;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_TE = 0x0001099A;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_TE_2 = 0x0001099B;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_TO = 0x0001099C;
const char32_t MEROITIC_HIEROGLYPHIC_LETTER_DA = 0x0001099D;
const char32_t MEROITIC_HIEROGLYPHIC_SYMBOL_VIDJ = 0x0001099E;
const char32_t MEROITIC_HIEROGLYPHIC_SYMBOL_VIDJ_2 = 0x0001099F;
const char32_t MEROITIC_CURSIVE_LETTER_A = 0x000109A0;
const char32_t MEROITIC_CURSIVE_LETTER_E = 0x000109A1;
const char32_t MEROITIC_CURSIVE_LETTER_I = 0x000109A2;
const char32_t MEROITIC_CURSIVE_LETTER_O = 0x000109A3;
const char32_t MEROITIC_CURSIVE_LETTER_YA = 0x000109A4;
const char32_t MEROITIC_CURSIVE_LETTER_WA = 0x000109A5;
const char32_t MEROITIC_CURSIVE_LETTER_BA = 0x000109A6;
const char32_t MEROITIC_CURSIVE_LETTER_PA = 0x000109A7;
const char32_t MEROITIC_CURSIVE_LETTER_MA = 0x000109A8;
const char32_t MEROITIC_CURSIVE_LETTER_NA = 0x000109A9;
const char32_t MEROITIC_CURSIVE_LETTER_NE = 0x000109AA;
const char32_t MEROITIC_CURSIVE_LETTER_RA = 0x000109AB;
const char32_t MEROITIC_CURSIVE_LETTER_LA = 0x000109AC;
const char32_t MEROITIC_CURSIVE_LETTER_KHA = 0x000109AD;
const char32_t MEROITIC_CURSIVE_LETTER_HHA = 0x000109AE;
const char32_t MEROITIC_CURSIVE_LETTER_SA = 0x000109AF;
const char32_t MEROITIC_CURSIVE_LETTER_ARCHAIC_SA = 0x000109B0;
const char32_t MEROITIC_CURSIVE_LETTER_SE = 0x000109B1;
const char32_t MEROITIC_CURSIVE_LETTER_KA = 0x000109B2;
const char32_t MEROITIC_CURSIVE_LETTER_QA = 0x000109B3;
const char32_t MEROITIC_CURSIVE_LETTER_TA = 0x000109B4;
const char32_t MEROITIC_CURSIVE_LETTER_TE = 0x000109B5;
const char32_t MEROITIC_CURSIVE_LETTER_TO = 0x000109B6;
const char32_t MEROITIC_CURSIVE_LETTER_DA = 0x000109B7;
const char32_t MEROITIC_CURSIVE_FRACTION_ELEVEN_TWELFTHS = 0x000109BC;
const char32_t MEROITIC_CURSIVE_FRACTION_ONE_HALF = 0x000109BD;
const char32_t MEROITIC_CURSIVE_LOGOGRAM_RMT = 0x000109BE;
const char32_t MEROITIC_CURSIVE_LOGOGRAM_IMN = 0x000109BF;
const char32_t MEROITIC_CURSIVE_NUMBER_ONE = 0x000109C0;
const char32_t MEROITIC_CURSIVE_NUMBER_TWO = 0x000109C1;
const char32_t MEROITIC_CURSIVE_NUMBER_THREE = 0x000109C2;
const char32_t MEROITIC_CURSIVE_NUMBER_FOUR = 0x000109C3;
const char32_t MEROITIC_CURSIVE_NUMBER_FIVE = 0x000109C4;
const char32_t MEROITIC_CURSIVE_NUMBER_SIX = 0x000109C5;
const char32_t MEROITIC_CURSIVE_NUMBER_SEVEN = 0x000109C6;
const char32_t MEROITIC_CURSIVE_NUMBER_EIGHT = 0x000109C7;
const char32_t MEROITIC_CURSIVE_NUMBER_NINE = 0x000109C8;
const char32_t MEROITIC_CURSIVE_NUMBER_TEN = 0x000109C9;
const char32_t MEROITIC_CURSIVE_NUMBER_TWENTY = 0x000109CA;
const char32_t MEROITIC_CURSIVE_NUMBER_THIRTY = 0x000109CB;
const char32_t MEROITIC_CURSIVE_NUMBER_FORTY = 0x000109CC;
const char32_t MEROITIC_CURSIVE_NUMBER_FIFTY = 0x000109CD;
const char32_t MEROITIC_CURSIVE_NUMBER_SIXTY = 0x000109CE;
const char32_t MEROITIC_CURSIVE_NUMBER_SEVENTY = 0x000109CF;
const char32_t MEROITIC_CURSIVE_NUMBER_ONE_HUNDRED = 0x000109D2;
const char32_t MEROITIC_CURSIVE_NUMBER_TWO_HUNDRED = 0x000109D3;
const char32_t MEROITIC_CURSIVE_NUMBER_THREE_HUNDRED = 0x000109D4;
const char32_t MEROITIC_CURSIVE_NUMBER_FOUR_HUNDRED = 0x000109D5;
const char32_t MEROITIC_CURSIVE_NUMBER_FIVE_HUNDRED = 0x000109D6;
const char32_t MEROITIC_CURSIVE_NUMBER_SIX_HUNDRED = 0x000109D7;
const char32_t MEROITIC_CURSIVE_NUMBER_SEVEN_HUNDRED = 0x000109D8;
const char32_t MEROITIC_CURSIVE_NUMBER_EIGHT_HUNDRED = 0x000109D9;
const char32_t MEROITIC_CURSIVE_NUMBER_NINE_HUNDRED = 0x000109DA;
const char32_t MEROITIC_CURSIVE_NUMBER_ONE_THOUSAND = 0x000109DB;
const char32_t MEROITIC_CURSIVE_NUMBER_TWO_THOUSAND = 0x000109DC;
const char32_t MEROITIC_CURSIVE_NUMBER_THREE_THOUSAND = 0x000109DD;
const char32_t MEROITIC_CURSIVE_NUMBER_FOUR_THOUSAND = 0x000109DE;
const char32_t MEROITIC_CURSIVE_NUMBER_FIVE_THOUSAND = 0x000109DF;
const char32_t MEROITIC_CURSIVE_NUMBER_SIX_THOUSAND = 0x000109E0;
const char32_t MEROITIC_CURSIVE_NUMBER_SEVEN_THOUSAND = 0x000109E1;
const char32_t MEROITIC_CURSIVE_NUMBER_EIGHT_THOUSAND = 0x000109E2;
const char32_t MEROITIC_CURSIVE_NUMBER_NINE_THOUSAND = 0x000109E3;
const char32_t MEROITIC_CURSIVE_NUMBER_TEN_THOUSAND = 0x000109E4;
const char32_t MEROITIC_CURSIVE_NUMBER_TWENTY_THOUSAND = 0x000109E5;
const char32_t MEROITIC_CURSIVE_NUMBER_THIRTY_THOUSAND = 0x000109E6;
const char32_t MEROITIC_CURSIVE_NUMBER_FORTY_THOUSAND = 0x000109E7;
const char32_t MEROITIC_CURSIVE_NUMBER_FIFTY_THOUSAND = 0x000109E8;
const char32_t MEROITIC_CURSIVE_NUMBER_SIXTY_THOUSAND = 0x000109E9;
const char32_t MEROITIC_CURSIVE_NUMBER_SEVENTY_THOUSAND = 0x000109EA;
const char32_t MEROITIC_CURSIVE_NUMBER_EIGHTY_THOUSAND = 0x000109EB;
const char32_t MEROITIC_CURSIVE_NUMBER_NINETY_THOUSAND = 0x000109EC;
const char32_t MEROITIC_CURSIVE_NUMBER_ONE_HUNDRED_THOUSAND = 0x000109ED;
const char32_t MEROITIC_CURSIVE_NUMBER_TWO_HUNDRED_THOUSAND = 0x000109EE;
const char32_t MEROITIC_CURSIVE_NUMBER_THREE_HUNDRED_THOUSAND = 0x000109EF;
const char32_t MEROITIC_CURSIVE_NUMBER_FOUR_HUNDRED_THOUSAND = 0x000109F0;
const char32_t MEROITIC_CURSIVE_NUMBER_FIVE_HUNDRED_THOUSAND = 0x000109F1;
const char32_t MEROITIC_CURSIVE_NUMBER_SIX_HUNDRED_THOUSAND = 0x000109F2;
const char32_t MEROITIC_CURSIVE_NUMBER_SEVEN_HUNDRED_THOUSAND = 0x000109F3;
const char32_t MEROITIC_CURSIVE_NUMBER_EIGHT_HUNDRED_THOUSAND = 0x000109F4;
const char32_t MEROITIC_CURSIVE_NUMBER_NINE_HUNDRED_THOUSAND = 0x000109F5;
const char32_t MEROITIC_CURSIVE_FRACTION_ONE_TWELFTH = 0x000109F6;
const char32_t MEROITIC_CURSIVE_FRACTION_TWO_TWELFTHS = 0x000109F7;
const char32_t MEROITIC_CURSIVE_FRACTION_THREE_TWELFTHS = 0x000109F8;
const char32_t MEROITIC_CURSIVE_FRACTION_FOUR_TWELFTHS = 0x000109F9;
const char32_t MEROITIC_CURSIVE_FRACTION_FIVE_TWELFTHS = 0x000109FA;
const char32_t MEROITIC_CURSIVE_FRACTION_SIX_TWELFTHS = 0x000109FB;
const char32_t MEROITIC_CURSIVE_FRACTION_SEVEN_TWELFTHS = 0x000109FC;
const char32_t MEROITIC_CURSIVE_FRACTION_EIGHT_TWELFTHS = 0x000109FD;
const char32_t MEROITIC_CURSIVE_FRACTION_NINE_TWELFTHS = 0x000109FE;
const char32_t MEROITIC_CURSIVE_FRACTION_TEN_TWELFTHS = 0x000109FF;
const char32_t KHAROSHTHI_LETTER_A = 0x00010A00;
const char32_t KHAROSHTHI_VOWEL_SIGN_I = 0x00010A01;
const char32_t KHAROSHTHI_VOWEL_SIGN_U = 0x00010A02;
const char32_t KHAROSHTHI_VOWEL_SIGN_VOCALIC_R = 0x00010A03;
const char32_t KHAROSHTHI_VOWEL_SIGN_E = 0x00010A05;
const char32_t KHAROSHTHI_VOWEL_SIGN_O = 0x00010A06;
const char32_t KHAROSHTHI_VOWEL_LENGTH_MARK = 0x00010A0C;
const char32_t KHAROSHTHI_SIGN_DOUBLE_RING_BELOW = 0x00010A0D;
const char32_t KHAROSHTHI_SIGN_ANUSVARA = 0x00010A0E;
const char32_t KHAROSHTHI_SIGN_VISARGA = 0x00010A0F;
const char32_t KHAROSHTHI_LETTER_KA = 0x00010A10;
const char32_t KHAROSHTHI_LETTER_KHA = 0x00010A11;
const char32_t KHAROSHTHI_LETTER_GA = 0x00010A12;
const char32_t KHAROSHTHI_LETTER_GHA = 0x00010A13;
const char32_t KHAROSHTHI_LETTER_CA = 0x00010A15;
const char32_t KHAROSHTHI_LETTER_CHA = 0x00010A16;
const char32_t KHAROSHTHI_LETTER_JA = 0x00010A17;
const char32_t KHAROSHTHI_LETTER_NYA = 0x00010A19;
const char32_t KHAROSHTHI_LETTER_TTA = 0x00010A1A;
const char32_t KHAROSHTHI_LETTER_TTHA = 0x00010A1B;
const char32_t KHAROSHTHI_LETTER_DDA = 0x00010A1C;
const char32_t KHAROSHTHI_LETTER_DDHA = 0x00010A1D;
const char32_t KHAROSHTHI_LETTER_NNA = 0x00010A1E;
const char32_t KHAROSHTHI_LETTER_TA = 0x00010A1F;
const char32_t KHAROSHTHI_LETTER_THA = 0x00010A20;
const char32_t KHAROSHTHI_LETTER_DA = 0x00010A21;
const char32_t KHAROSHTHI_LETTER_DHA = 0x00010A22;
const char32_t KHAROSHTHI_LETTER_NA = 0x00010A23;
const char32_t KHAROSHTHI_LETTER_PA = 0x00010A24;
const char32_t KHAROSHTHI_LETTER_PHA = 0x00010A25;
const char32_t KHAROSHTHI_LETTER_BA = 0x00010A26;
const char32_t KHAROSHTHI_LETTER_BHA = 0x00010A27;
const char32_t KHAROSHTHI_LETTER_MA = 0x00010A28;
const char32_t KHAROSHTHI_LETTER_YA = 0x00010A29;
const char32_t KHAROSHTHI_LETTER_RA = 0x00010A2A;
const char32_t KHAROSHTHI_LETTER_LA = 0x00010A2B;
const char32_t KHAROSHTHI_LETTER_VA = 0x00010A2C;
const char32_t KHAROSHTHI_LETTER_SHA = 0x00010A2D;
const char32_t KHAROSHTHI_LETTER_SSA = 0x00010A2E;
const char32_t KHAROSHTHI_LETTER_SA = 0x00010A2F;
const char32_t KHAROSHTHI_LETTER_ZA = 0x00010A30;
const char32_t KHAROSHTHI_LETTER_HA = 0x00010A31;
const char32_t KHAROSHTHI_LETTER_KKA = 0x00010A32;
const char32_t KHAROSHTHI_LETTER_TTTHA = 0x00010A33;
const char32_t KHAROSHTHI_LETTER_TTTA = 0x00010A34;
const char32_t KHAROSHTHI_LETTER_VHA = 0x00010A35;
const char32_t KHAROSHTHI_SIGN_BAR_ABOVE = 0x00010A38;
const char32_t KHAROSHTHI_SIGN_CAUDA = 0x00010A39;
const char32_t KHAROSHTHI_SIGN_DOT_BELOW = 0x00010A3A;
const char32_t KHAROSHTHI_VIRAMA = 0x00010A3F;
const char32_t KHAROSHTHI_DIGIT_ONE = 0x00010A40;
const char32_t KHAROSHTHI_DIGIT_TWO = 0x00010A41;
const char32_t KHAROSHTHI_DIGIT_THREE = 0x00010A42;
const char32_t KHAROSHTHI_DIGIT_FOUR = 0x00010A43;
const char32_t KHAROSHTHI_NUMBER_TEN = 0x00010A44;
const char32_t KHAROSHTHI_NUMBER_TWENTY = 0x00010A45;
const char32_t KHAROSHTHI_NUMBER_ONE_HUNDRED = 0x00010A46;
const char32_t KHAROSHTHI_NUMBER_ONE_THOUSAND = 0x00010A47;
const char32_t KHAROSHTHI_FRACTION_ONE_HALF = 0x00010A48;
const char32_t KHAROSHTHI_PUNCTUATION_DOT = 0x00010A50;
const char32_t KHAROSHTHI_PUNCTUATION_SMALL_CIRCLE = 0x00010A51;
const char32_t KHAROSHTHI_PUNCTUATION_CIRCLE = 0x00010A52;
const char32_t KHAROSHTHI_PUNCTUATION_CRESCENT_BAR = 0x00010A53;
const char32_t KHAROSHTHI_PUNCTUATION_MANGALAM = 0x00010A54;
const char32_t KHAROSHTHI_PUNCTUATION_LOTUS = 0x00010A55;
const char32_t KHAROSHTHI_PUNCTUATION_DANDA = 0x00010A56;
const char32_t KHAROSHTHI_PUNCTUATION_DOUBLE_DANDA = 0x00010A57;
const char32_t KHAROSHTHI_PUNCTUATION_LINES = 0x00010A58;
const char32_t OLD_SOUTH_ARABIAN_LETTER_HE = 0x00010A60;
const char32_t OLD_SOUTH_ARABIAN_LETTER_LAMEDH = 0x00010A61;
const char32_t OLD_SOUTH_ARABIAN_LETTER_HETH = 0x00010A62;
const char32_t OLD_SOUTH_ARABIAN_LETTER_MEM = 0x00010A63;
const char32_t OLD_SOUTH_ARABIAN_LETTER_QOPH = 0x00010A64;
const char32_t OLD_SOUTH_ARABIAN_LETTER_WAW = 0x00010A65;
const char32_t OLD_SOUTH_ARABIAN_LETTER_SHIN = 0x00010A66;
const char32_t OLD_SOUTH_ARABIAN_LETTER_RESH = 0x00010A67;
const char32_t OLD_SOUTH_ARABIAN_LETTER_BETH = 0x00010A68;
const char32_t OLD_SOUTH_ARABIAN_LETTER_TAW = 0x00010A69;
const char32_t OLD_SOUTH_ARABIAN_LETTER_SAT = 0x00010A6A;
const char32_t OLD_SOUTH_ARABIAN_LETTER_KAPH = 0x00010A6B;
const char32_t OLD_SOUTH_ARABIAN_LETTER_NUN = 0x00010A6C;
const char32_t OLD_SOUTH_ARABIAN_LETTER_KHETH = 0x00010A6D;
const char32_t OLD_SOUTH_ARABIAN_LETTER_SADHE = 0x00010A6E;
const char32_t OLD_SOUTH_ARABIAN_LETTER_SAMEKH = 0x00010A6F;
const char32_t OLD_SOUTH_ARABIAN_LETTER_FE = 0x00010A70;
const char32_t OLD_SOUTH_ARABIAN_LETTER_ALEF = 0x00010A71;
const char32_t OLD_SOUTH_ARABIAN_LETTER_AYN = 0x00010A72;
const char32_t OLD_SOUTH_ARABIAN_LETTER_DHADHE = 0x00010A73;
const char32_t OLD_SOUTH_ARABIAN_LETTER_GIMEL = 0x00010A74;
const char32_t OLD_SOUTH_ARABIAN_LETTER_DALETH = 0x00010A75;
const char32_t OLD_SOUTH_ARABIAN_LETTER_GHAYN = 0x00010A76;
const char32_t OLD_SOUTH_ARABIAN_LETTER_TETH = 0x00010A77;
const char32_t OLD_SOUTH_ARABIAN_LETTER_ZAYN = 0x00010A78;
const char32_t OLD_SOUTH_ARABIAN_LETTER_DHALETH = 0x00010A79;
const char32_t OLD_SOUTH_ARABIAN_LETTER_YODH = 0x00010A7A;
const char32_t OLD_SOUTH_ARABIAN_LETTER_THAW = 0x00010A7B;
const char32_t OLD_SOUTH_ARABIAN_LETTER_THETH = 0x00010A7C;
const char32_t OLD_SOUTH_ARABIAN_NUMBER_ONE = 0x00010A7D;
const char32_t OLD_SOUTH_ARABIAN_NUMBER_FIFTY = 0x00010A7E;
const char32_t OLD_SOUTH_ARABIAN_NUMERIC_INDICATOR = 0x00010A7F;
const char32_t OLD_NORTH_ARABIAN_LETTER_HEH = 0x00010A80;
const char32_t OLD_NORTH_ARABIAN_LETTER_LAM = 0x00010A81;
const char32_t OLD_NORTH_ARABIAN_LETTER_HAH = 0x00010A82;
const char32_t OLD_NORTH_ARABIAN_LETTER_MEEM = 0x00010A83;
const char32_t OLD_NORTH_ARABIAN_LETTER_QAF = 0x00010A84;
const char32_t OLD_NORTH_ARABIAN_LETTER_WAW = 0x00010A85;
const char32_t OLD_NORTH_ARABIAN_LETTER_ES_2 = 0x00010A86;
const char32_t OLD_NORTH_ARABIAN_LETTER_REH = 0x00010A87;
const char32_t OLD_NORTH_ARABIAN_LETTER_BEH = 0x00010A88;
const char32_t OLD_NORTH_ARABIAN_LETTER_TEH = 0x00010A89;
const char32_t OLD_NORTH_ARABIAN_LETTER_ES_1 = 0x00010A8A;
const char32_t OLD_NORTH_ARABIAN_LETTER_KAF = 0x00010A8B;
const char32_t OLD_NORTH_ARABIAN_LETTER_NOON = 0x00010A8C;
const char32_t OLD_NORTH_ARABIAN_LETTER_KHAH = 0x00010A8D;
const char32_t OLD_NORTH_ARABIAN_LETTER_SAD = 0x00010A8E;
const char32_t OLD_NORTH_ARABIAN_LETTER_ES_3 = 0x00010A8F;
const char32_t OLD_NORTH_ARABIAN_LETTER_FEH = 0x00010A90;
const char32_t OLD_NORTH_ARABIAN_LETTER_ALEF = 0x00010A91;
const char32_t OLD_NORTH_ARABIAN_LETTER_AIN = 0x00010A92;
const char32_t OLD_NORTH_ARABIAN_LETTER_DAD = 0x00010A93;
const char32_t OLD_NORTH_ARABIAN_LETTER_GEEM = 0x00010A94;
const char32_t OLD_NORTH_ARABIAN_LETTER_DAL = 0x00010A95;
const char32_t OLD_NORTH_ARABIAN_LETTER_GHAIN = 0x00010A96;
const char32_t OLD_NORTH_ARABIAN_LETTER_TAH = 0x00010A97;
const char32_t OLD_NORTH_ARABIAN_LETTER_ZAIN = 0x00010A98;
const char32_t OLD_NORTH_ARABIAN_LETTER_THAL = 0x00010A99;
const char32_t OLD_NORTH_ARABIAN_LETTER_YEH = 0x00010A9A;
const char32_t OLD_NORTH_ARABIAN_LETTER_THEH = 0x00010A9B;
const char32_t OLD_NORTH_ARABIAN_LETTER_ZAH = 0x00010A9C;
const char32_t OLD_NORTH_ARABIAN_NUMBER_ONE = 0x00010A9D;
const char32_t OLD_NORTH_ARABIAN_NUMBER_TEN = 0x00010A9E;
const char32_t OLD_NORTH_ARABIAN_NUMBER_TWENTY = 0x00010A9F;
const char32_t MANICHAEAN_LETTER_ALEPH = 0x00010AC0;
const char32_t MANICHAEAN_LETTER_BETH = 0x00010AC1;
const char32_t MANICHAEAN_LETTER_BHETH = 0x00010AC2;
const char32_t MANICHAEAN_LETTER_GIMEL = 0x00010AC3;
const char32_t MANICHAEAN_LETTER_GHIMEL = 0x00010AC4;
const char32_t MANICHAEAN_LETTER_DALETH = 0x00010AC5;
const char32_t MANICHAEAN_LETTER_HE = 0x00010AC6;
const char32_t MANICHAEAN_LETTER_WAW = 0x00010AC7;
const char32_t MANICHAEAN_SIGN_UD = 0x00010AC8;
const char32_t MANICHAEAN_LETTER_ZAYIN = 0x00010AC9;
const char32_t MANICHAEAN_LETTER_ZHAYIN = 0x00010ACA;
const char32_t MANICHAEAN_LETTER_JAYIN = 0x00010ACB;
const char32_t MANICHAEAN_LETTER_JHAYIN = 0x00010ACC;
const char32_t MANICHAEAN_LETTER_HETH = 0x00010ACD;
const char32_t MANICHAEAN_LETTER_TETH = 0x00010ACE;
const char32_t MANICHAEAN_LETTER_YODH = 0x00010ACF;
const char32_t MANICHAEAN_LETTER_KAPH = 0x00010AD0;
const char32_t MANICHAEAN_LETTER_XAPH = 0x00010AD1;
const char32_t MANICHAEAN_LETTER_KHAPH = 0x00010AD2;
const char32_t MANICHAEAN_LETTER_LAMEDH = 0x00010AD3;
const char32_t MANICHAEAN_LETTER_DHAMEDH = 0x00010AD4;
const char32_t MANICHAEAN_LETTER_THAMEDH = 0x00010AD5;
const char32_t MANICHAEAN_LETTER_MEM = 0x00010AD6;
const char32_t MANICHAEAN_LETTER_NUN = 0x00010AD7;
const char32_t MANICHAEAN_LETTER_SAMEKH = 0x00010AD8;
const char32_t MANICHAEAN_LETTER_AYIN = 0x00010AD9;
const char32_t MANICHAEAN_LETTER_AAYIN = 0x00010ADA;
const char32_t MANICHAEAN_LETTER_PE = 0x00010ADB;
const char32_t MANICHAEAN_LETTER_FE = 0x00010ADC;
const char32_t MANICHAEAN_LETTER_SADHE = 0x00010ADD;
const char32_t MANICHAEAN_LETTER_QOPH = 0x00010ADE;
const char32_t MANICHAEAN_LETTER_XOPH = 0x00010ADF;
const char32_t MANICHAEAN_LETTER_QHOPH = 0x00010AE0;
const char32_t MANICHAEAN_LETTER_RESH = 0x00010AE1;
const char32_t MANICHAEAN_LETTER_SHIN = 0x00010AE2;
const char32_t MANICHAEAN_LETTER_SSHIN = 0x00010AE3;
const char32_t MANICHAEAN_LETTER_TAW = 0x00010AE4;
const char32_t MANICHAEAN_ABBREVIATION_MARK_ABOVE = 0x00010AE5;
const char32_t MANICHAEAN_ABBREVIATION_MARK_BELOW = 0x00010AE6;
const char32_t MANICHAEAN_NUMBER_ONE = 0x00010AEB;
const char32_t MANICHAEAN_NUMBER_FIVE = 0x00010AEC;
const char32_t MANICHAEAN_NUMBER_TEN = 0x00010AED;
const char32_t MANICHAEAN_NUMBER_TWENTY = 0x00010AEE;
const char32_t MANICHAEAN_NUMBER_ONE_HUNDRED = 0x00010AEF;
const char32_t MANICHAEAN_PUNCTUATION_STAR = 0x00010AF0;
const char32_t MANICHAEAN_PUNCTUATION_FLEURON = 0x00010AF1;
const char32_t MANICHAEAN_PUNCTUATION_DOUBLE_DOT_WITHIN_DOT = 0x00010AF2;
const char32_t MANICHAEAN_PUNCTUATION_DOT_WITHIN_DOT = 0x00010AF3;
const char32_t MANICHAEAN_PUNCTUATION_DOT = 0x00010AF4;
const char32_t MANICHAEAN_PUNCTUATION_TWO_DOTS = 0x00010AF5;
const char32_t MANICHAEAN_PUNCTUATION_LINE_FILLER = 0x00010AF6;
const char32_t AVESTAN_LETTER_A = 0x00010B00;
const char32_t AVESTAN_LETTER_AA = 0x00010B01;
const char32_t AVESTAN_LETTER_AO = 0x00010B02;
const char32_t AVESTAN_LETTER_AAO = 0x00010B03;
const char32_t AVESTAN_LETTER_AN = 0x00010B04;
const char32_t AVESTAN_LETTER_AAN = 0x00010B05;
const char32_t AVESTAN_LETTER_AE = 0x00010B06;
const char32_t AVESTAN_LETTER_AEE = 0x00010B07;
const char32_t AVESTAN_LETTER_E = 0x00010B08;
const char32_t AVESTAN_LETTER_EE = 0x00010B09;
const char32_t AVESTAN_LETTER_O = 0x00010B0A;
const char32_t AVESTAN_LETTER_OO = 0x00010B0B;
const char32_t AVESTAN_LETTER_I = 0x00010B0C;
const char32_t AVESTAN_LETTER_II = 0x00010B0D;
const char32_t AVESTAN_LETTER_U = 0x00010B0E;
const char32_t AVESTAN_LETTER_UU = 0x00010B0F;
const char32_t AVESTAN_LETTER_KE = 0x00010B10;
const char32_t AVESTAN_LETTER_XE = 0x00010B11;
const char32_t AVESTAN_LETTER_XYE = 0x00010B12;
const char32_t AVESTAN_LETTER_XVE = 0x00010B13;
const char32_t AVESTAN_LETTER_GE = 0x00010B14;
const char32_t AVESTAN_LETTER_GGE = 0x00010B15;
const char32_t AVESTAN_LETTER_GHE = 0x00010B16;
const char32_t AVESTAN_LETTER_CE = 0x00010B17;
const char32_t AVESTAN_LETTER_JE = 0x00010B18;
const char32_t AVESTAN_LETTER_TE = 0x00010B19;
const char32_t AVESTAN_LETTER_THE = 0x00010B1A;
const char32_t AVESTAN_LETTER_DE = 0x00010B1B;
const char32_t AVESTAN_LETTER_DHE = 0x00010B1C;
const char32_t AVESTAN_LETTER_TTE = 0x00010B1D;
const char32_t AVESTAN_LETTER_PE = 0x00010B1E;
const char32_t AVESTAN_LETTER_FE = 0x00010B1F;
const char32_t AVESTAN_LETTER_BE = 0x00010B20;
const char32_t AVESTAN_LETTER_BHE = 0x00010B21;
const char32_t AVESTAN_LETTER_NGE = 0x00010B22;
const char32_t AVESTAN_LETTER_NGYE = 0x00010B23;
const char32_t AVESTAN_LETTER_NGVE = 0x00010B24;
const char32_t AVESTAN_LETTER_NE = 0x00010B25;
const char32_t AVESTAN_LETTER_NYE = 0x00010B26;
const char32_t AVESTAN_LETTER_NNE = 0x00010B27;
const char32_t AVESTAN_LETTER_ME = 0x00010B28;
const char32_t AVESTAN_LETTER_HME = 0x00010B29;
const char32_t AVESTAN_LETTER_YYE = 0x00010B2A;
const char32_t AVESTAN_LETTER_YE = 0x00010B2B;
const char32_t AVESTAN_LETTER_VE = 0x00010B2C;
const char32_t AVESTAN_LETTER_RE = 0x00010B2D;
const char32_t AVESTAN_LETTER_LE = 0x00010B2E;
const char32_t AVESTAN_LETTER_SE = 0x00010B2F;
const char32_t AVESTAN_LETTER_ZE = 0x00010B30;
const char32_t AVESTAN_LETTER_SHE = 0x00010B31;
const char32_t AVESTAN_LETTER_ZHE = 0x00010B32;
const char32_t AVESTAN_LETTER_SHYE = 0x00010B33;
const char32_t AVESTAN_LETTER_SSHE = 0x00010B34;
const char32_t AVESTAN_LETTER_HE = 0x00010B35;
const char32_t AVESTAN_ABBREVIATION_MARK = 0x00010B39;
const char32_t TINY_TWO_DOTS_OVER_ONE_DOT_PUNCTUATION = 0x00010B3A;
const char32_t SMALL_TWO_DOTS_OVER_ONE_DOT_PUNCTUATION = 0x00010B3B;
const char32_t LARGE_TWO_DOTS_OVER_ONE_DOT_PUNCTUATION = 0x00010B3C;
const char32_t LARGE_ONE_DOT_OVER_TWO_DOTS_PUNCTUATION = 0x00010B3D;
const char32_t LARGE_TWO_RINGS_OVER_ONE_RING_PUNCTUATION = 0x00010B3E;
const char32_t LARGE_ONE_RING_OVER_TWO_RINGS_PUNCTUATION = 0x00010B3F;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_ALEPH = 0x00010B40;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_BETH = 0x00010B41;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_GIMEL = 0x00010B42;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_DALETH = 0x00010B43;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_HE = 0x00010B44;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_WAW = 0x00010B45;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_ZAYIN = 0x00010B46;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_HETH = 0x00010B47;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_TETH = 0x00010B48;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_YODH = 0x00010B49;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_KAPH = 0x00010B4A;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_LAMEDH = 0x00010B4B;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_MEM = 0x00010B4C;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_NUN = 0x00010B4D;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_SAMEKH = 0x00010B4E;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_AYIN = 0x00010B4F;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_PE = 0x00010B50;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_SADHE = 0x00010B51;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_QOPH = 0x00010B52;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_RESH = 0x00010B53;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_SHIN = 0x00010B54;
const char32_t INSCRIPTIONAL_PARTHIAN_LETTER_TAW = 0x00010B55;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_ONE = 0x00010B58;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_TWO = 0x00010B59;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_THREE = 0x00010B5A;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_FOUR = 0x00010B5B;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_TEN = 0x00010B5C;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_TWENTY = 0x00010B5D;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_ONE_HUNDRED = 0x00010B5E;
const char32_t INSCRIPTIONAL_PARTHIAN_NUMBER_ONE_THOUSAND = 0x00010B5F;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_ALEPH = 0x00010B60;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_BETH = 0x00010B61;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_GIMEL = 0x00010B62;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_DALETH = 0x00010B63;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_HE = 0x00010B64;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_WAW_AYIN_RESH = 0x00010B65;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_ZAYIN = 0x00010B66;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_HETH = 0x00010B67;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_TETH = 0x00010B68;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_YODH = 0x00010B69;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_KAPH = 0x00010B6A;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_LAMEDH = 0x00010B6B;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_MEM_QOPH = 0x00010B6C;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_NUN = 0x00010B6D;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_SAMEKH = 0x00010B6E;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_PE = 0x00010B6F;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_SADHE = 0x00010B70;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_SHIN = 0x00010B71;
const char32_t INSCRIPTIONAL_PAHLAVI_LETTER_TAW = 0x00010B72;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_ONE = 0x00010B78;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_TWO = 0x00010B79;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_THREE = 0x00010B7A;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_FOUR = 0x00010B7B;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_TEN = 0x00010B7C;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_TWENTY = 0x00010B7D;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_ONE_HUNDRED = 0x00010B7E;
const char32_t INSCRIPTIONAL_PAHLAVI_NUMBER_ONE_THOUSAND = 0x00010B7F;
const char32_t PSALTER_PAHLAVI_LETTER_ALEPH = 0x00010B80;
const char32_t PSALTER_PAHLAVI_LETTER_BETH = 0x00010B81;
const char32_t PSALTER_PAHLAVI_LETTER_GIMEL = 0x00010B82;
const char32_t PSALTER_PAHLAVI_LETTER_DALETH = 0x00010B83;
const char32_t PSALTER_PAHLAVI_LETTER_HE = 0x00010B84;
const char32_t PSALTER_PAHLAVI_LETTER_WAW_AYIN_RESH = 0x00010B85;
const char32_t PSALTER_PAHLAVI_LETTER_ZAYIN = 0x00010B86;
const char32_t PSALTER_PAHLAVI_LETTER_HETH = 0x00010B87;
const char32_t PSALTER_PAHLAVI_LETTER_YODH = 0x00010B88;
const char32_t PSALTER_PAHLAVI_LETTER_KAPH = 0x00010B89;
const char32_t PSALTER_PAHLAVI_LETTER_LAMEDH = 0x00010B8A;
const char32_t PSALTER_PAHLAVI_LETTER_MEM_QOPH = 0x00010B8B;
const char32_t PSALTER_PAHLAVI_LETTER_NUN = 0x00010B8C;
const char32_t PSALTER_PAHLAVI_LETTER_SAMEKH = 0x00010B8D;
const char32_t PSALTER_PAHLAVI_LETTER_PE = 0x00010B8E;
const char32_t PSALTER_PAHLAVI_LETTER_SADHE = 0x00010B8F;
const char32_t PSALTER_PAHLAVI_LETTER_SHIN = 0x00010B90;
const char32_t PSALTER_PAHLAVI_LETTER_TAW = 0x00010B91;
const char32_t PSALTER_PAHLAVI_SECTION_MARK = 0x00010B99;
const char32_t PSALTER_PAHLAVI_TURNED_SECTION_MARK = 0x00010B9A;
const char32_t PSALTER_PAHLAVI_FOUR_DOTS_WITH_CROSS = 0x00010B9B;
const char32_t PSALTER_PAHLAVI_FOUR_DOTS_WITH_DOT = 0x00010B9C;
const char32_t PSALTER_PAHLAVI_NUMBER_ONE = 0x00010BA9;
const char32_t PSALTER_PAHLAVI_NUMBER_TWO = 0x00010BAA;
const char32_t PSALTER_PAHLAVI_NUMBER_THREE = 0x00010BAB;
const char32_t PSALTER_PAHLAVI_NUMBER_FOUR = 0x00010BAC;
const char32_t PSALTER_PAHLAVI_NUMBER_TEN = 0x00010BAD;
const char32_t PSALTER_PAHLAVI_NUMBER_TWENTY = 0x00010BAE;
const char32_t PSALTER_PAHLAVI_NUMBER_ONE_HUNDRED = 0x00010BAF;
const char32_t OLD_TURKIC_LETTER_ORKHON_A = 0x00010C00;
const char32_t OLD_TURKIC_LETTER_YENISEI_A = 0x00010C01;
const char32_t OLD_TURKIC_LETTER_YENISEI_AE = 0x00010C02;
const char32_t OLD_TURKIC_LETTER_ORKHON_I = 0x00010C03;
const char32_t OLD_TURKIC_LETTER_YENISEI_I = 0x00010C04;
const char32_t OLD_TURKIC_LETTER_YENISEI_E = 0x00010C05;
const char32_t OLD_TURKIC_LETTER_ORKHON_O = 0x00010C06;
const char32_t OLD_TURKIC_LETTER_ORKHON_OE = 0x00010C07;
const char32_t OLD_TURKIC_LETTER_YENISEI_OE = 0x00010C08;
const char32_t OLD_TURKIC_LETTER_ORKHON_AB = 0x00010C09;
const char32_t OLD_TURKIC_LETTER_YENISEI_AB = 0x00010C0A;
const char32_t OLD_TURKIC_LETTER_ORKHON_AEB = 0x00010C0B;
const char32_t OLD_TURKIC_LETTER_YENISEI_AEB = 0x00010C0C;
const char32_t OLD_TURKIC_LETTER_ORKHON_AG = 0x00010C0D;
const char32_t OLD_TURKIC_LETTER_YENISEI_AG = 0x00010C0E;
const char32_t OLD_TURKIC_LETTER_ORKHON_AEG = 0x00010C0F;
const char32_t OLD_TURKIC_LETTER_YENISEI_AEG = 0x00010C10;
const char32_t OLD_TURKIC_LETTER_ORKHON_AD = 0x00010C11;
const char32_t OLD_TURKIC_LETTER_YENISEI_AD = 0x00010C12;
const char32_t OLD_TURKIC_LETTER_ORKHON_AED = 0x00010C13;
const char32_t OLD_TURKIC_LETTER_ORKHON_EZ = 0x00010C14;
const char32_t OLD_TURKIC_LETTER_YENISEI_EZ = 0x00010C15;
const char32_t OLD_TURKIC_LETTER_ORKHON_AY = 0x00010C16;
const char32_t OLD_TURKIC_LETTER_YENISEI_AY = 0x00010C17;
const char32_t OLD_TURKIC_LETTER_ORKHON_AEY = 0x00010C18;
const char32_t OLD_TURKIC_LETTER_YENISEI_AEY = 0x00010C19;
const char32_t OLD_TURKIC_LETTER_ORKHON_AEK = 0x00010C1A;
const char32_t OLD_TURKIC_LETTER_YENISEI_AEK = 0x00010C1B;
const char32_t OLD_TURKIC_LETTER_ORKHON_OEK = 0x00010C1C;
const char32_t OLD_TURKIC_LETTER_YENISEI_OEK = 0x00010C1D;
const char32_t OLD_TURKIC_LETTER_ORKHON_AL = 0x00010C1E;
const char32_t OLD_TURKIC_LETTER_YENISEI_AL = 0x00010C1F;
const char32_t OLD_TURKIC_LETTER_ORKHON_AEL = 0x00010C20;
const char32_t OLD_TURKIC_LETTER_ORKHON_ELT = 0x00010C21;
const char32_t OLD_TURKIC_LETTER_ORKHON_EM = 0x00010C22;
const char32_t OLD_TURKIC_LETTER_ORKHON_AN = 0x00010C23;
const char32_t OLD_TURKIC_LETTER_ORKHON_AEN = 0x00010C24;
const char32_t OLD_TURKIC_LETTER_YENISEI_AEN = 0x00010C25;
const char32_t OLD_TURKIC_LETTER_ORKHON_ENT = 0x00010C26;
const char32_t OLD_TURKIC_LETTER_YENISEI_ENT = 0x00010C27;
const char32_t OLD_TURKIC_LETTER_ORKHON_ENC = 0x00010C28;
const char32_t OLD_TURKIC_LETTER_YENISEI_ENC = 0x00010C29;
const char32_t OLD_TURKIC_LETTER_ORKHON_ENY = 0x00010C2A;
const char32_t OLD_TURKIC_LETTER_YENISEI_ENY = 0x00010C2B;
const char32_t OLD_TURKIC_LETTER_YENISEI_ANG = 0x00010C2C;
const char32_t OLD_TURKIC_LETTER_ORKHON_ENG = 0x00010C2D;
const char32_t OLD_TURKIC_LETTER_YENISEI_AENG = 0x00010C2E;
const char32_t OLD_TURKIC_LETTER_ORKHON_EP = 0x00010C2F;
const char32_t OLD_TURKIC_LETTER_ORKHON_OP = 0x00010C30;
const char32_t OLD_TURKIC_LETTER_ORKHON_IC = 0x00010C31;
const char32_t OLD_TURKIC_LETTER_ORKHON_EC = 0x00010C32;
const char32_t OLD_TURKIC_LETTER_YENISEI_EC = 0x00010C33;
const char32_t OLD_TURKIC_LETTER_ORKHON_AQ = 0x00010C34;
const char32_t OLD_TURKIC_LETTER_YENISEI_AQ = 0x00010C35;
const char32_t OLD_TURKIC_LETTER_ORKHON_IQ = 0x00010C36;
const char32_t OLD_TURKIC_LETTER_YENISEI_IQ = 0x00010C37;
const char32_t OLD_TURKIC_LETTER_ORKHON_OQ = 0x00010C38;
const char32_t OLD_TURKIC_LETTER_YENISEI_OQ = 0x00010C39;
const char32_t OLD_TURKIC_LETTER_ORKHON_AR = 0x00010C3A;
const char32_t OLD_TURKIC_LETTER_YENISEI_AR = 0x00010C3B;
const char32_t OLD_TURKIC_LETTER_ORKHON_AER = 0x00010C3C;
const char32_t OLD_TURKIC_LETTER_ORKHON_AS = 0x00010C3D;
const char32_t OLD_TURKIC_LETTER_ORKHON_AES = 0x00010C3E;
const char32_t OLD_TURKIC_LETTER_ORKHON_ASH = 0x00010C3F;
const char32_t OLD_TURKIC_LETTER_YENISEI_ASH = 0x00010C40;
const char32_t OLD_TURKIC_LETTER_ORKHON_ESH = 0x00010C41;
const char32_t OLD_TURKIC_LETTER_YENISEI_ESH = 0x00010C42;
const char32_t OLD_TURKIC_LETTER_ORKHON_AT = 0x00010C43;
const char32_t OLD_TURKIC_LETTER_YENISEI_AT = 0x00010C44;
const char32_t OLD_TURKIC_LETTER_ORKHON_AET = 0x00010C45;
const char32_t OLD_TURKIC_LETTER_YENISEI_AET = 0x00010C46;
const char32_t OLD_TURKIC_LETTER_ORKHON_OT = 0x00010C47;
const char32_t OLD_TURKIC_LETTER_ORKHON_BASH = 0x00010C48;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_A = 0x00010C80;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_AA = 0x00010C81;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EB = 0x00010C82;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_AMB = 0x00010C83;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EC = 0x00010C84;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ENC = 0x00010C85;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ECS = 0x00010C86;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ED = 0x00010C87;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_AND = 0x00010C88;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_E = 0x00010C89;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_CLOSE_E = 0x00010C8A;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EE = 0x00010C8B;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EF = 0x00010C8C;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EG = 0x00010C8D;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EGY = 0x00010C8E;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EH = 0x00010C8F;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_I = 0x00010C90;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_II = 0x00010C91;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EJ = 0x00010C92;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EK = 0x00010C93;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_AK = 0x00010C94;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_UNK = 0x00010C95;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EL = 0x00010C96;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ELY = 0x00010C97;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EM = 0x00010C98;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EN = 0x00010C99;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ENY = 0x00010C9A;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_O = 0x00010C9B;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_OO = 0x00010C9C;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_NIKOLSBURG_OE = 0x00010C9D;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_RUDIMENTA_OE = 0x00010C9E;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_OEE = 0x00010C9F;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EP = 0x00010CA0;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EMP = 0x00010CA1;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ER = 0x00010CA2;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_SHORT_ER = 0x00010CA3;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ES = 0x00010CA4;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ESZ = 0x00010CA5;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ET = 0x00010CA6;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ENT = 0x00010CA7;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ETY = 0x00010CA8;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ECH = 0x00010CA9;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_U = 0x00010CAA;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_UU = 0x00010CAB;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_NIKOLSBURG_UE = 0x00010CAC;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_RUDIMENTA_UE = 0x00010CAD;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EV = 0x00010CAE;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EZ = 0x00010CAF;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_EZS = 0x00010CB0;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_ENT_SHAPED_SIGN = 0x00010CB1;
const char32_t OLD_HUNGARIAN_CAPITAL_LETTER_US = 0x00010CB2;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_A = 0x00010CC0;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_AA = 0x00010CC1;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EB = 0x00010CC2;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_AMB = 0x00010CC3;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EC = 0x00010CC4;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ENC = 0x00010CC5;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ECS = 0x00010CC6;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ED = 0x00010CC7;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_AND = 0x00010CC8;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_E = 0x00010CC9;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_CLOSE_E = 0x00010CCA;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EE = 0x00010CCB;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EF = 0x00010CCC;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EG = 0x00010CCD;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EGY = 0x00010CCE;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EH = 0x00010CCF;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_I = 0x00010CD0;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_II = 0x00010CD1;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EJ = 0x00010CD2;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EK = 0x00010CD3;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_AK = 0x00010CD4;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_UNK = 0x00010CD5;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EL = 0x00010CD6;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ELY = 0x00010CD7;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EM = 0x00010CD8;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EN = 0x00010CD9;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ENY = 0x00010CDA;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_O = 0x00010CDB;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_OO = 0x00010CDC;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_NIKOLSBURG_OE = 0x00010CDD;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_RUDIMENTA_OE = 0x00010CDE;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_OEE = 0x00010CDF;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EP = 0x00010CE0;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EMP = 0x00010CE1;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ER = 0x00010CE2;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_SHORT_ER = 0x00010CE3;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ES = 0x00010CE4;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ESZ = 0x00010CE5;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ET = 0x00010CE6;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ENT = 0x00010CE7;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ETY = 0x00010CE8;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ECH = 0x00010CE9;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_U = 0x00010CEA;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_UU = 0x00010CEB;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_NIKOLSBURG_UE = 0x00010CEC;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_RUDIMENTA_UE = 0x00010CED;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EV = 0x00010CEE;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EZ = 0x00010CEF;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_EZS = 0x00010CF0;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_ENT_SHAPED_SIGN = 0x00010CF1;
const char32_t OLD_HUNGARIAN_SMALL_LETTER_US = 0x00010CF2;
const char32_t OLD_HUNGARIAN_NUMBER_ONE = 0x00010CFA;
const char32_t OLD_HUNGARIAN_NUMBER_FIVE = 0x00010CFB;
const char32_t OLD_HUNGARIAN_NUMBER_TEN = 0x00010CFC;
const char32_t OLD_HUNGARIAN_NUMBER_FIFTY = 0x00010CFD;
const char32_t OLD_HUNGARIAN_NUMBER_ONE_HUNDRED = 0x00010CFE;
const char32_t OLD_HUNGARIAN_NUMBER_ONE_THOUSAND = 0x00010CFF;
const char32_t HANIFI_ROHINGYA_LETTER_A = 0x00010D00;
const char32_t HANIFI_ROHINGYA_LETTER_BA = 0x00010D01;
const char32_t HANIFI_ROHINGYA_LETTER_PA = 0x00010D02;
const char32_t HANIFI_ROHINGYA_LETTER_TA = 0x00010D03;
const char32_t HANIFI_ROHINGYA_LETTER_TTA = 0x00010D04;
const char32_t HANIFI_ROHINGYA_LETTER_JA = 0x00010D05;
const char32_t HANIFI_ROHINGYA_LETTER_CA = 0x00010D06;
const char32_t HANIFI_ROHINGYA_LETTER_HA = 0x00010D07;
const char32_t HANIFI_ROHINGYA_LETTER_KHA = 0x00010D08;
const char32_t HANIFI_ROHINGYA_LETTER_FA = 0x00010D09;
const char32_t HANIFI_ROHINGYA_LETTER_DA = 0x00010D0A;
const char32_t HANIFI_ROHINGYA_LETTER_DDA = 0x00010D0B;
const char32_t HANIFI_ROHINGYA_LETTER_RA = 0x00010D0C;
const char32_t HANIFI_ROHINGYA_LETTER_RRA = 0x00010D0D;
const char32_t HANIFI_ROHINGYA_LETTER_ZA = 0x00010D0E;
const char32_t HANIFI_ROHINGYA_LETTER_SA = 0x00010D0F;
const char32_t HANIFI_ROHINGYA_LETTER_SHA = 0x00010D10;
const char32_t HANIFI_ROHINGYA_LETTER_KA = 0x00010D11;
const char32_t HANIFI_ROHINGYA_LETTER_GA = 0x00010D12;
const char32_t HANIFI_ROHINGYA_LETTER_LA = 0x00010D13;
const char32_t HANIFI_ROHINGYA_LETTER_MA = 0x00010D14;
const char32_t HANIFI_ROHINGYA_LETTER_NA = 0x00010D15;
const char32_t HANIFI_ROHINGYA_LETTER_WA = 0x00010D16;
const char32_t HANIFI_ROHINGYA_LETTER_KINNA_WA = 0x00010D17;
const char32_t HANIFI_ROHINGYA_LETTER_YA = 0x00010D18;
const char32_t HANIFI_ROHINGYA_LETTER_KINNA_YA = 0x00010D19;
const char32_t HANIFI_ROHINGYA_LETTER_NGA = 0x00010D1A;
const char32_t HANIFI_ROHINGYA_LETTER_NYA = 0x00010D1B;
const char32_t HANIFI_ROHINGYA_LETTER_VA = 0x00010D1C;
const char32_t HANIFI_ROHINGYA_VOWEL_A = 0x00010D1D;
const char32_t HANIFI_ROHINGYA_VOWEL_I = 0x00010D1E;
const char32_t HANIFI_ROHINGYA_VOWEL_U = 0x00010D1F;
const char32_t HANIFI_ROHINGYA_VOWEL_E = 0x00010D20;
const char32_t HANIFI_ROHINGYA_VOWEL_O = 0x00010D21;
const char32_t HANIFI_ROHINGYA_MARK_SAKIN = 0x00010D22;
const char32_t HANIFI_ROHINGYA_MARK_NA_KHONNA = 0x00010D23;
const char32_t HANIFI_ROHINGYA_SIGN_HARBAHAY = 0x00010D24;
const char32_t HANIFI_ROHINGYA_SIGN_TAHALA = 0x00010D25;
const char32_t HANIFI_ROHINGYA_SIGN_TANA = 0x00010D26;
const char32_t HANIFI_ROHINGYA_SIGN_TASSI = 0x00010D27;
const char32_t HANIFI_ROHINGYA_DIGIT_ZERO = 0x00010D30;
const char32_t HANIFI_ROHINGYA_DIGIT_ONE = 0x00010D31;
const char32_t HANIFI_ROHINGYA_DIGIT_TWO = 0x00010D32;
const char32_t HANIFI_ROHINGYA_DIGIT_THREE = 0x00010D33;
const char32_t HANIFI_ROHINGYA_DIGIT_FOUR = 0x00010D34;
const char32_t HANIFI_ROHINGYA_DIGIT_FIVE = 0x00010D35;
const char32_t HANIFI_ROHINGYA_DIGIT_SIX = 0x00010D36;
const char32_t HANIFI_ROHINGYA_DIGIT_SEVEN = 0x00010D37;
const char32_t HANIFI_ROHINGYA_DIGIT_EIGHT = 0x00010D38;
const char32_t HANIFI_ROHINGYA_DIGIT_NINE = 0x00010D39;
const char32_t GARAY_DIGIT_ZERO = 0x00010D40;
const char32_t GARAY_DIGIT_ONE = 0x00010D41;
const char32_t GARAY_DIGIT_TWO = 0x00010D42;
const char32_t GARAY_DIGIT_THREE = 0x00010D43;
const char32_t GARAY_DIGIT_FOUR = 0x00010D44;
const char32_t GARAY_DIGIT_FIVE = 0x00010D45;
const char32_t GARAY_DIGIT_SIX = 0x00010D46;
const char32_t GARAY_DIGIT_SEVEN = 0x00010D47;
const char32_t GARAY_DIGIT_EIGHT = 0x00010D48;
const char32_t GARAY_DIGIT_NINE = 0x00010D49;
const char32_t GARAY_VOWEL_SIGN_A = 0x00010D4A;
const char32_t GARAY_VOWEL_SIGN_I = 0x00010D4B;
const char32_t GARAY_VOWEL_SIGN_O = 0x00010D4C;
const char32_t GARAY_VOWEL_SIGN_EE = 0x00010D4D;
const char32_t GARAY_VOWEL_LENGTH_MARK = 0x00010D4E;
const char32_t GARAY_SUKUN = 0x00010D4F;
const char32_t GARAY_CAPITAL_LETTER_A = 0x00010D50;
const char32_t GARAY_CAPITAL_LETTER_CA = 0x00010D51;
const char32_t GARAY_CAPITAL_LETTER_MA = 0x00010D52;
const char32_t GARAY_CAPITAL_LETTER_KA = 0x00010D53;
const char32_t GARAY_CAPITAL_LETTER_BA = 0x00010D54;
const char32_t GARAY_CAPITAL_LETTER_JA = 0x00010D55;
const char32_t GARAY_CAPITAL_LETTER_SA = 0x00010D56;
const char32_t GARAY_CAPITAL_LETTER_WA = 0x00010D57;
const char32_t GARAY_CAPITAL_LETTER_LA = 0x00010D58;
const char32_t GARAY_CAPITAL_LETTER_GA = 0x00010D59;
const char32_t GARAY_CAPITAL_LETTER_DA = 0x00010D5A;
const char32_t GARAY_CAPITAL_LETTER_XA = 0x00010D5B;
const char32_t GARAY_CAPITAL_LETTER_YA = 0x00010D5C;
const char32_t GARAY_CAPITAL_LETTER_TA = 0x00010D5D;
const char32_t GARAY_CAPITAL_LETTER_RA = 0x00010D5E;
const char32_t GARAY_CAPITAL_LETTER_NYA = 0x00010D5F;
const char32_t GARAY_CAPITAL_LETTER_FA = 0x00010D60;
const char32_t GARAY_CAPITAL_LETTER_NA = 0x00010D61;
const char32_t GARAY_CAPITAL_LETTER_PA = 0x00010D62;
const char32_t GARAY_CAPITAL_LETTER_HA = 0x00010D63;
const char32_t GARAY_CAPITAL_LETTER_OLD_KA = 0x00010D64;
const char32_t GARAY_CAPITAL_LETTER_OLD_NA = 0x00010D65;
const char32_t GARAY_VOWEL_SIGN_E = 0x00010D69;
const char32_t GARAY_CONSONANT_GEMINATION_MARK = 0x00010D6A;
const char32_t GARAY_COMBINING_DOT_ABOVE = 0x00010D6B;
const char32_t GARAY_COMBINING_DOUBLE_DOT_ABOVE = 0x00010D6C;
const char32_t GARAY_CONSONANT_NASALIZATION_MARK = 0x00010D6D;
const char32_t GARAY_HYPHEN = 0x00010D6E;
const char32_t GARAY_REDUPLICATION_MARK = 0x00010D6F;
const char32_t GARAY_SMALL_LETTER_A = 0x00010D70;
const char32_t GARAY_SMALL_LETTER_CA = 0x00010D71;
const char32_t GARAY_SMALL_LETTER_MA = 0x00010D72;
const char32_t GARAY_SMALL_LETTER_KA = 0x00010D73;
const char32_t GARAY_SMALL_LETTER_BA = 0x00010D74;
const char32_t GARAY_SMALL_LETTER_JA = 0x00010D75;
const char32_t GARAY_SMALL_LETTER_SA = 0x00010D76;
const char32_t GARAY_SMALL_LETTER_WA = 0x00010D77;
const char32_t GARAY_SMALL_LETTER_LA = 0x00010D78;
const char32_t GARAY_SMALL_LETTER_GA = 0x00010D79;
const char32_t GARAY_SMALL_LETTER_DA = 0x00010D7A;
const char32_t GARAY_SMALL_LETTER_XA = 0x00010D7B;
const char32_t GARAY_SMALL_LETTER_YA = 0x00010D7C;
const char32_t GARAY_SMALL_LETTER_TA = 0x00010D7D;
const char32_t GARAY_SMALL_LETTER_RA = 0x00010D7E;
const char32_t GARAY_SMALL_LETTER_NYA = 0x00010D7F;
const char32_t GARAY_SMALL_LETTER_FA = 0x00010D80;
const char32_t GARAY_SMALL_LETTER_NA = 0x00010D81;
const char32_t GARAY_SMALL_LETTER_PA = 0x00010D82;
const char32_t GARAY_SMALL_LETTER_HA = 0x00010D83;
const char32_t GARAY_SMALL_LETTER_OLD_KA = 0x00010D84;
const char32_t GARAY_SMALL_LETTER_OLD_NA = 0x00010D85;
const char32_t GARAY_PLUS_SIGN = 0x00010D8E;
const char32_t GARAY_MINUS_SIGN = 0x00010D8F;
const char32_t RUMI_DIGIT_ONE = 0x00010E60;
const char32_t RUMI_DIGIT_TWO = 0x00010E61;
const char32_t RUMI_DIGIT_THREE = 0x00010E62;
const char32_t RUMI_DIGIT_FOUR = 0x00010E63;
const char32_t RUMI_DIGIT_FIVE = 0x00010E64;
const char32_t RUMI_DIGIT_SIX = 0x00010E65;
const char32_t RUMI_DIGIT_SEVEN = 0x00010E66;
const char32_t RUMI_DIGIT_EIGHT = 0x00010E67;
const char32_t RUMI_DIGIT_NINE = 0x00010E68;
const char32_t RUMI_NUMBER_TEN = 0x00010E69;
const char32_t RUMI_NUMBER_TWENTY = 0x00010E6A;
const char32_t RUMI_NUMBER_THIRTY = 0x00010E6B;
const char32_t RUMI_NUMBER_FORTY = 0x00010E6C;
const char32_t RUMI_NUMBER_FIFTY = 0x00010E6D;
const char32_t RUMI_NUMBER_SIXTY = 0x00010E6E;
const char32_t RUMI_NUMBER_SEVENTY = 0x00010E6F;
const char32_t RUMI_NUMBER_EIGHTY = 0x00010E70;
const char32_t RUMI_NUMBER_NINETY = 0x00010E71;
const char32_t RUMI_NUMBER_ONE_HUNDRED = 0x00010E72;
const char32_t RUMI_NUMBER_TWO_HUNDRED = 0x00010E73;
const char32_t RUMI_NUMBER_THREE_HUNDRED = 0x00010E74;
const char32_t RUMI_NUMBER_FOUR_HUNDRED = 0x00010E75;
const char32_t RUMI_NUMBER_FIVE_HUNDRED = 0x00010E76;
const char32_t RUMI_NUMBER_SIX_HUNDRED = 0x00010E77;
const char32_t RUMI_NUMBER_SEVEN_HUNDRED = 0x00010E78;
const char32_t RUMI_NUMBER_EIGHT_HUNDRED = 0x00010E79;
const char32_t RUMI_NUMBER_NINE_HUNDRED = 0x00010E7A;
const char32_t RUMI_FRACTION_ONE_HALF = 0x00010E7B;
const char32_t RUMI_FRACTION_ONE_QUARTER = 0x00010E7C;
const char32_t RUMI_FRACTION_ONE_THIRD = 0x00010E7D;
const char32_t RUMI_FRACTION_TWO_THIRDS = 0x00010E7E;
const char32_t YEZIDI_LETTER_ELIF = 0x00010E80;
const char32_t YEZIDI_LETTER_BE = 0x00010E81;
const char32_t YEZIDI_LETTER_PE = 0x00010E82;
const char32_t YEZIDI_LETTER_PHE = 0x00010E83;
const char32_t YEZIDI_LETTER_THE = 0x00010E84;
const char32_t YEZIDI_LETTER_SE = 0x00010E85;
const char32_t YEZIDI_LETTER_CIM = 0x00010E86;
const char32_t YEZIDI_LETTER_CHIM = 0x00010E87;
const char32_t YEZIDI_LETTER_CHHIM = 0x00010E88;
const char32_t YEZIDI_LETTER_HHA = 0x00010E89;
const char32_t YEZIDI_LETTER_XA = 0x00010E8A;
const char32_t YEZIDI_LETTER_DAL = 0x00010E8B;
const char32_t YEZIDI_LETTER_ZAL = 0x00010E8C;
const char32_t YEZIDI_LETTER_RA = 0x00010E8D;
const char32_t YEZIDI_LETTER_RHA = 0x00010E8E;
const char32_t YEZIDI_LETTER_ZA = 0x00010E8F;
const char32_t YEZIDI_LETTER_JA = 0x00010E90;
const char32_t YEZIDI_LETTER_SIN = 0x00010E91;
const char32_t YEZIDI_LETTER_SHIN = 0x00010E92;
const char32_t YEZIDI_LETTER_SAD = 0x00010E93;
const char32_t YEZIDI_LETTER_DAD = 0x00010E94;
const char32_t YEZIDI_LETTER_TA = 0x00010E95;
const char32_t YEZIDI_LETTER_ZE = 0x00010E96;
const char32_t YEZIDI_LETTER_EYN = 0x00010E97;
const char32_t YEZIDI_LETTER_XHEYN = 0x00010E98;
const char32_t YEZIDI_LETTER_FA = 0x00010E99;
const char32_t YEZIDI_LETTER_VA = 0x00010E9A;
const char32_t YEZIDI_LETTER_VA_ALTERNATE_FORM = 0x00010E9B;
const char32_t YEZIDI_LETTER_QAF = 0x00010E9C;
const char32_t YEZIDI_LETTER_KAF = 0x00010E9D;
const char32_t YEZIDI_LETTER_KHAF = 0x00010E9E;
const char32_t YEZIDI_LETTER_GAF = 0x00010E9F;
const char32_t YEZIDI_LETTER_LAM = 0x00010EA0;
const char32_t YEZIDI_LETTER_MIM = 0x00010EA1;
const char32_t YEZIDI_LETTER_NUN = 0x00010EA2;
const char32_t YEZIDI_LETTER_UM = 0x00010EA3;
const char32_t YEZIDI_LETTER_WAW = 0x00010EA4;
const char32_t YEZIDI_LETTER_OW = 0x00010EA5;
const char32_t YEZIDI_LETTER_EW = 0x00010EA6;
const char32_t YEZIDI_LETTER_HAY = 0x00010EA7;
const char32_t YEZIDI_LETTER_YOT = 0x00010EA8;
const char32_t YEZIDI_LETTER_ET = 0x00010EA9;
const char32_t YEZIDI_COMBINING_HAMZA_MARK = 0x00010EAB;
const char32_t YEZIDI_COMBINING_MADDA_MARK = 0x00010EAC;
const char32_t YEZIDI_HYPHENATION_MARK = 0x00010EAD;
const char32_t YEZIDI_LETTER_LAM_WITH_DOT_ABOVE = 0x00010EB0;
const char32_t YEZIDI_LETTER_YOT_WITH_CIRCUMFLEX_ABOVE = 0x00010EB1;
const char32_t ARABIC_LETTER_DAL_WITH_TWO_DOTS_VERTICALLY_BELOW = 0x00010EC2;
const char32_t ARABIC_LETTER_TAH_WITH_TWO_DOTS_VERTICALLY_BELOW = 0x00010EC3;
const char32_t ARABIC_LETTER_KAF_WITH_TWO_DOTS_VERTICALLY_BELOW = 0x00010EC4;
const char32_t ARABIC_SMALL_YEH_BARREE_WITH_TWO_DOTS_BELOW = 0x00010EC5;
const char32_t ARABIC_LETTER_THIN_NOON = 0x00010EC6;
const char32_t ARABIC_LETTER_YEH_WITH_FOUR_DOTS_BELOW = 0x00010EC7;
const char32_t ARABIC_BIBLICAL_END_OF_VERSE = 0x00010ED0;
const char32_t ARABIC_LIGATURE_ALAYHAA_AS_SALAATU_WAS_SALAAM = 0x00010ED1;
const char32_t ARABIC_LIGATURE_ALAYHIM_AS_SALAATU_WAS_SALAAM = 0x00010ED2;
const char32_t ARABIC_LIGATURE_ALAYHIMAA_AS_SALAATU_WAS_SALAAM = 0x00010ED3;
const char32_t ARABIC_LIGATURE_QADDASA_ALLAAHU_SIRRAH = 0x00010ED4;
const char32_t ARABIC_LIGATURE_QUDDISA_SIRRUHUM = 0x00010ED5;
const char32_t ARABIC_LIGATURE_QUDDISA_SIRRUHUMAA = 0x00010ED6;
const char32_t ARABIC_LIGATURE_QUDDISAT_ASRAARUHUM = 0x00010ED7;
const char32_t ARABIC_LIGATURE_NAWWARA_ALLAAHU_MARQADAH = 0x00010ED8;
const char32_t ARABIC_DOUBLE_VERTICAL_BAR_BELOW = 0x00010EFA;
const char32_t ARABIC_SMALL_LOW_NOON = 0x00010EFB;
const char32_t ARABIC_COMBINING_ALEF_OVERLAY = 0x00010EFC;
const char32_t ARABIC_SMALL_LOW_WORD_SAKTA = 0x00010EFD;
const char32_t ARABIC_SMALL_LOW_WORD_QASR = 0x00010EFE;
const char32_t ARABIC_SMALL_LOW_WORD_MADDA = 0x00010EFF;
const char32_t OLD_SOGDIAN_LETTER_ALEPH = 0x00010F00;
const char32_t OLD_SOGDIAN_LETTER_FINAL_ALEPH = 0x00010F01;
const char32_t OLD_SOGDIAN_LETTER_BETH = 0x00010F02;
const char32_t OLD_SOGDIAN_LETTER_FINAL_BETH = 0x00010F03;
const char32_t OLD_SOGDIAN_LETTER_GIMEL = 0x00010F04;
const char32_t OLD_SOGDIAN_LETTER_HE = 0x00010F05;
const char32_t OLD_SOGDIAN_LETTER_FINAL_HE = 0x00010F06;
const char32_t OLD_SOGDIAN_LETTER_WAW = 0x00010F07;
const char32_t OLD_SOGDIAN_LETTER_ZAYIN = 0x00010F08;
const char32_t OLD_SOGDIAN_LETTER_HETH = 0x00010F09;
const char32_t OLD_SOGDIAN_LETTER_YODH = 0x00010F0A;
const char32_t OLD_SOGDIAN_LETTER_KAPH = 0x00010F0B;
const char32_t OLD_SOGDIAN_LETTER_LAMEDH = 0x00010F0C;
const char32_t OLD_SOGDIAN_LETTER_MEM = 0x00010F0D;
const char32_t OLD_SOGDIAN_LETTER_NUN = 0x00010F0E;
const char32_t OLD_SOGDIAN_LETTER_FINAL_NUN = 0x00010F0F;
const char32_t OLD_SOGDIAN_LETTER_FINAL_NUN_WITH_VERTICAL_TAIL = 0x00010F10;
const char32_t OLD_SOGDIAN_LETTER_SAMEKH = 0x00010F11;
const char32_t OLD_SOGDIAN_LETTER_AYIN = 0x00010F12;
const char32_t OLD_SOGDIAN_LETTER_ALTERNATE_AYIN = 0x00010F13;
const char32_t OLD_SOGDIAN_LETTER_PE = 0x00010F14;
const char32_t OLD_SOGDIAN_LETTER_SADHE = 0x00010F15;
const char32_t OLD_SOGDIAN_LETTER_FINAL_SADHE = 0x00010F16;
const char32_t OLD_SOGDIAN_LETTER_FINAL_SADHE_WITH_VERTICAL_TAIL = 0x00010F17;
const char32_t OLD_SOGDIAN_LETTER_RESH_AYIN_DALETH = 0x00010F18;
const char32_t OLD_SOGDIAN_LETTER_SHIN = 0x00010F19;
const char32_t OLD_SOGDIAN_LETTER_TAW = 0x00010F1A;
const char32_t OLD_SOGDIAN_LETTER_FINAL_TAW = 0x00010F1B;
const char32_t OLD_SOGDIAN_LETTER_FINAL_TAW_WITH_VERTICAL_TAIL = 0x00010F1C;
const char32_t OLD_SOGDIAN_NUMBER_ONE = 0x00010F1D;
const char32_t OLD_SOGDIAN_NUMBER_TWO = 0x00010F1E;
const char32_t OLD_SOGDIAN_NUMBER_THREE = 0x00010F1F;
const char32_t OLD_SOGDIAN_NUMBER_FOUR = 0x00010F20;
const char32_t OLD_SOGDIAN_NUMBER_FIVE = 0x00010F21;
const char32_t OLD_SOGDIAN_NUMBER_TEN = 0x00010F22;
const char32_t OLD_SOGDIAN_NUMBER_TWENTY = 0x00010F23;
const char32_t OLD_SOGDIAN_NUMBER_THIRTY = 0x00010F24;
const char32_t OLD_SOGDIAN_NUMBER_ONE_HUNDRED = 0x00010F25;
const char32_t OLD_SOGDIAN_FRACTION_ONE_HALF = 0x00010F26;
const char32_t OLD_SOGDIAN_LIGATURE_AYIN_DALETH = 0x00010F27;
const char32_t SOGDIAN_LETTER_ALEPH = 0x00010F30;
const char32_t SOGDIAN_LETTER_BETH = 0x00010F31;
const char32_t SOGDIAN_LETTER_GIMEL = 0x00010F32;
const char32_t SOGDIAN_LETTER_HE = 0x00010F33;
const char32_t SOGDIAN_LETTER_WAW = 0x00010F34;
const char32_t SOGDIAN_LETTER_ZAYIN = 0x00010F35;
const char32_t SOGDIAN_LETTER_HETH = 0x00010F36;
const char32_t SOGDIAN_LETTER_YODH = 0x00010F37;
const char32_t SOGDIAN_LETTER_KAPH = 0x00010F38;
const char32_t SOGDIAN_LETTER_LAMEDH = 0x00010F39;
const char32_t SOGDIAN_LETTER_MEM = 0x00010F3A;
const char32_t SOGDIAN_LETTER_NUN = 0x00010F3B;
const char32_t SOGDIAN_LETTER_SAMEKH = 0x00010F3C;
const char32_t SOGDIAN_LETTER_AYIN = 0x00010F3D;
const char32_t SOGDIAN_LETTER_PE = 0x00010F3E;
const char32_t SOGDIAN_LETTER_SADHE = 0x00010F3F;
const char32_t SOGDIAN_LETTER_RESH_AYIN = 0x00010F40;
const char32_t SOGDIAN_LETTER_SHIN = 0x00010F41;
const char32_t SOGDIAN_LETTER_TAW = 0x00010F42;
const char32_t SOGDIAN_LETTER_FETH = 0x00010F43;
const char32_t SOGDIAN_LETTER_LESH = 0x00010F44;
const char32_t SOGDIAN_INDEPENDENT_SHIN = 0x00010F45;
const char32_t SOGDIAN_COMBINING_DOT_BELOW = 0x00010F46;
const char32_t SOGDIAN_COMBINING_TWO_DOTS_BELOW = 0x00010F47;
const char32_t SOGDIAN_COMBINING_DOT_ABOVE = 0x00010F48;
const char32_t SOGDIAN_COMBINING_TWO_DOTS_ABOVE = 0x00010F49;
const char32_t SOGDIAN_COMBINING_CURVE_ABOVE = 0x00010F4A;
const char32_t SOGDIAN_COMBINING_CURVE_BELOW = 0x00010F4B;
const char32_t SOGDIAN_COMBINING_HOOK_ABOVE = 0x00010F4C;
const char32_t SOGDIAN_COMBINING_HOOK_BELOW = 0x00010F4D;
const char32_t SOGDIAN_COMBINING_LONG_HOOK_BELOW = 0x00010F4E;
const char32_t SOGDIAN_COMBINING_RESH_BELOW = 0x00010F4F;
const char32_t SOGDIAN_COMBINING_STROKE_BELOW = 0x00010F50;
const char32_t SOGDIAN_NUMBER_ONE = 0x00010F51;
const char32_t SOGDIAN_NUMBER_TEN = 0x00010F52;
const char32_t SOGDIAN_NUMBER_TWENTY = 0x00010F53;
const char32_t SOGDIAN_NUMBER_ONE_HUNDRED = 0x00010F54;
const char32_t SOGDIAN_PUNCTUATION_TWO_VERTICAL_BARS = 0x00010F55;
const char32_t SOGDIAN_PUNCTUATION_TWO_VERTICAL_BARS_WITH_DOTS = 0x00010F56;
const char32_t SOGDIAN_PUNCTUATION_CIRCLE_WITH_DOT = 0x00010F57;
const char32_t SOGDIAN_PUNCTUATION_TWO_CIRCLES_WITH_DOTS = 0x00010F58;
const char32_t SOGDIAN_PUNCTUATION_HALF_CIRCLE_WITH_DOT = 0x00010F59;
const char32_t OLD_UYGHUR_LETTER_ALEPH = 0x00010F70;
const char32_t OLD_UYGHUR_LETTER_BETH = 0x00010F71;
const char32_t OLD_UYGHUR_LETTER_GIMEL_HETH = 0x00010F72;
const char32_t OLD_UYGHUR_LETTER_WAW = 0x00010F73;
const char32_t OLD_UYGHUR_LETTER_ZAYIN = 0x00010F74;
const char32_t OLD_UYGHUR_LETTER_FINAL_HETH = 0x00010F75;
const char32_t OLD_UYGHUR_LETTER_YODH = 0x00010F76;
const char32_t OLD_UYGHUR_LETTER_KAPH = 0x00010F77;
const char32_t OLD_UYGHUR_LETTER_LAMEDH = 0x00010F78;
const char32_t OLD_UYGHUR_LETTER_MEM = 0x00010F79;
const char32_t OLD_UYGHUR_LETTER_NUN = 0x00010F7A;
const char32_t OLD_UYGHUR_LETTER_SAMEKH = 0x00010F7B;
const char32_t OLD_UYGHUR_LETTER_PE = 0x00010F7C;
const char32_t OLD_UYGHUR_LETTER_SADHE = 0x00010F7D;
const char32_t OLD_UYGHUR_LETTER_RESH = 0x00010F7E;
const char32_t OLD_UYGHUR_LETTER_SHIN = 0x00010F7F;
const char32_t OLD_UYGHUR_LETTER_TAW = 0x00010F80;
const char32_t OLD_UYGHUR_LETTER_LESH = 0x00010F81;
const char32_t OLD_UYGHUR_COMBINING_DOT_ABOVE = 0x00010F82;
const char32_t OLD_UYGHUR_COMBINING_DOT_BELOW = 0x00010F83;
const char32_t OLD_UYGHUR_COMBINING_TWO_DOTS_ABOVE = 0x00010F84;
const char32_t OLD_UYGHUR_COMBINING_TWO_DOTS_BELOW = 0x00010F85;
const char32_t OLD_UYGHUR_PUNCTUATION_BAR = 0x00010F86;
const char32_t OLD_UYGHUR_PUNCTUATION_TWO_BARS = 0x00010F87;
const char32_t OLD_UYGHUR_PUNCTUATION_TWO_DOTS = 0x00010F88;
const char32_t OLD_UYGHUR_PUNCTUATION_FOUR_DOTS = 0x00010F89;
const char32_t CHORASMIAN_LETTER_ALEPH = 0x00010FB0;
const char32_t CHORASMIAN_LETTER_SMALL_ALEPH = 0x00010FB1;
const char32_t CHORASMIAN_LETTER_BETH = 0x00010FB2;
const char32_t CHORASMIAN_LETTER_GIMEL = 0x00010FB3;
const char32_t CHORASMIAN_LETTER_DALETH = 0x00010FB4;
const char32_t CHORASMIAN_LETTER_HE = 0x00010FB5;
const char32_t CHORASMIAN_LETTER_WAW = 0x00010FB6;
const char32_t CHORASMIAN_LETTER_CURLED_WAW = 0x00010FB7;
const char32_t CHORASMIAN_LETTER_ZAYIN = 0x00010FB8;
const char32_t CHORASMIAN_LETTER_HETH = 0x00010FB9;
const char32_t CHORASMIAN_LETTER_YODH = 0x00010FBA;
const char32_t CHORASMIAN_LETTER_KAPH = 0x00010FBB;
const char32_t CHORASMIAN_LETTER_LAMEDH = 0x00010FBC;
const char32_t CHORASMIAN_LETTER_MEM = 0x00010FBD;
const char32_t CHORASMIAN_LETTER_NUN = 0x00010FBE;
const char32_t CHORASMIAN_LETTER_SAMEKH = 0x00010FBF;
const char32_t CHORASMIAN_LETTER_AYIN = 0x00010FC0;
const char32_t CHORASMIAN_LETTER_PE = 0x00010FC1;
const char32_t CHORASMIAN_LETTER_RESH = 0x00010FC2;
const char32_t CHORASMIAN_LETTER_SHIN = 0x00010FC3;
const char32_t CHORASMIAN_LETTER_TAW = 0x00010FC4;
const char32_t CHORASMIAN_NUMBER_ONE = 0x00010FC5;
const char32_t CHORASMIAN_NUMBER_TWO = 0x00010FC6;
const char32_t CHORASMIAN_NUMBER_THREE = 0x00010FC7;
const char32_t CHORASMIAN_NUMBER_FOUR = 0x00010FC8;
const char32_t CHORASMIAN_NUMBER_TEN = 0x00010FC9;
const char32_t CHORASMIAN_NUMBER_TWENTY = 0x00010FCA;
const char32_t CHORASMIAN_NUMBER_ONE_HUNDRED = 0x00010FCB;
const char32_t ELYMAIC_LETTER_ALEPH = 0x00010FE0;
const char32_t ELYMAIC_LETTER_BETH = 0x00010FE1;
const char32_t ELYMAIC_LETTER_GIMEL = 0x00010FE2;
const char32_t ELYMAIC_LETTER_DALETH = 0x00010FE3;
const char32_t ELYMAIC_LETTER_HE = 0x00010FE4;
const char32_t ELYMAIC_LETTER_WAW = 0x00010FE5;
const char32_t ELYMAIC_LETTER_ZAYIN = 0x00010FE6;
const char32_t ELYMAIC_LETTER_HETH = 0x00010FE7;
const char32_t ELYMAIC_LETTER_TETH = 0x00010FE8;
const char32_t ELYMAIC_LETTER_YODH = 0x00010FE9;
const char32_t ELYMAIC_LETTER_KAPH = 0x00010FEA;
const char32_t ELYMAIC_LETTER_LAMEDH = 0x00010FEB;
const char32_t ELYMAIC_LETTER_MEM = 0x00010FEC;
const char32_t ELYMAIC_LETTER_NUN = 0x00010FED;
const char32_t ELYMAIC_LETTER_SAMEKH = 0x00010FEE;
const char32_t ELYMAIC_LETTER_AYIN = 0x00010FEF;
const char32_t ELYMAIC_LETTER_PE = 0x00010FF0;
const char32_t ELYMAIC_LETTER_SADHE = 0x00010FF1;
const char32_t ELYMAIC_LETTER_QOPH = 0x00010FF2;
const char32_t ELYMAIC_LETTER_RESH = 0x00010FF3;
const char32_t ELYMAIC_LETTER_SHIN = 0x00010FF4;
const char32_t ELYMAIC_LETTER_TAW = 0x00010FF5;
const char32_t ELYMAIC_LIGATURE_ZAYIN_YODH = 0x00010FF6;
const char32_t BRAHMI_SIGN_CANDRABINDU = 0x00011000;
const char32_t BRAHMI_SIGN_ANUSVARA = 0x00011001;
const char32_t BRAHMI_SIGN_VISARGA = 0x00011002;
const char32_t BRAHMI_SIGN_JIHVAMULIYA = 0x00011003;
const char32_t BRAHMI_SIGN_UPADHMANIYA = 0x00011004;
const char32_t BRAHMI_LETTER_A = 0x00011005;
const char32_t BRAHMI_LETTER_AA = 0x00011006;
const char32_t BRAHMI_LETTER_I = 0x00011007;
const char32_t BRAHMI_LETTER_II = 0x00011008;
const char32_t BRAHMI_LETTER_U = 0x00011009;
const char32_t BRAHMI_LETTER_UU = 0x0001100A;
const char32_t BRAHMI_LETTER_VOCALIC_R = 0x0001100B;
const char32_t BRAHMI_LETTER_VOCALIC_RR = 0x0001100C;
const char32_t BRAHMI_LETTER_VOCALIC_L = 0x0001100D;
const char32_t BRAHMI_LETTER_VOCALIC_LL = 0x0001100E;
const char32_t BRAHMI_LETTER_E = 0x0001100F;
const char32_t BRAHMI_LETTER_AI = 0x00011010;
const char32_t BRAHMI_LETTER_O = 0x00011011;
const char32_t BRAHMI_LETTER_AU = 0x00011012;
const char32_t BRAHMI_LETTER_KA = 0x00011013;
const char32_t BRAHMI_LETTER_KHA = 0x00011014;
const char32_t BRAHMI_LETTER_GA = 0x00011015;
const char32_t BRAHMI_LETTER_GHA = 0x00011016;
const char32_t BRAHMI_LETTER_NGA = 0x00011017;
const char32_t BRAHMI_LETTER_CA = 0x00011018;
const char32_t BRAHMI_LETTER_CHA = 0x00011019;
const char32_t BRAHMI_LETTER_JA = 0x0001101A;
const char32_t BRAHMI_LETTER_JHA = 0x0001101B;
const char32_t BRAHMI_LETTER_NYA = 0x0001101C;
const char32_t BRAHMI_LETTER_TTA = 0x0001101D;
const char32_t BRAHMI_LETTER_TTHA = 0x0001101E;
const char32_t BRAHMI_LETTER_DDA = 0x0001101F;
const char32_t BRAHMI_LETTER_DDHA = 0x00011020;
const char32_t BRAHMI_LETTER_NNA = 0x00011021;
const char32_t BRAHMI_LETTER_TA = 0x00011022;
const char32_t BRAHMI_LETTER_THA = 0x00011023;
const char32_t BRAHMI_LETTER_DA = 0x00011024;
const char32_t BRAHMI_LETTER_DHA = 0x00011025;
const char32_t BRAHMI_LETTER_NA = 0x00011026;
const char32_t BRAHMI_LETTER_PA = 0x00011027;
const char32_t BRAHMI_LETTER_PHA = 0x00011028;
const char32_t BRAHMI_LETTER_BA = 0x00011029;
const char32_t BRAHMI_LETTER_BHA = 0x0001102A;
const char32_t BRAHMI_LETTER_MA = 0x0001102B;
const char32_t BRAHMI_LETTER_YA = 0x0001102C;
const char32_t BRAHMI_LETTER_RA = 0x0001102D;
const char32_t BRAHMI_LETTER_LA = 0x0001102E;
const char32_t BRAHMI_LETTER_VA = 0x0001102F;
const char32_t BRAHMI_LETTER_SHA = 0x00011030;
const char32_t BRAHMI_LETTER_SSA = 0x00011031;
const char32_t BRAHMI_LETTER_SA = 0x00011032;
const char32_t BRAHMI_LETTER_HA = 0x00011033;
const char32_t BRAHMI_LETTER_LLA = 0x00011034;
const char32_t BRAHMI_LETTER_OLD_TAMIL_LLLA = 0x00011035;
const char32_t BRAHMI_LETTER_OLD_TAMIL_RRA = 0x00011036;
const char32_t BRAHMI_LETTER_OLD_TAMIL_NNNA = 0x00011037;
const char32_t BRAHMI_VOWEL_SIGN_AA = 0x00011038;
const char32_t BRAHMI_VOWEL_SIGN_BHATTIPROLU_AA = 0x00011039;
const char32_t BRAHMI_VOWEL_SIGN_I = 0x0001103A;
const char32_t BRAHMI_VOWEL_SIGN_II = 0x0001103B;
const char32_t BRAHMI_VOWEL_SIGN_U = 0x0001103C;
const char32_t BRAHMI_VOWEL_SIGN_UU = 0x0001103D;
const char32_t BRAHMI_VOWEL_SIGN_VOCALIC_R = 0x0001103E;
const char32_t BRAHMI_VOWEL_SIGN_VOCALIC_RR = 0x0001103F;
const char32_t BRAHMI_VOWEL_SIGN_VOCALIC_L = 0x00011040;
const char32_t BRAHMI_VOWEL_SIGN_VOCALIC_LL = 0x00011041;
const char32_t BRAHMI_VOWEL_SIGN_E = 0x00011042;
const char32_t BRAHMI_VOWEL_SIGN_AI = 0x00011043;
const char32_t BRAHMI_VOWEL_SIGN_O = 0x00011044;
const char32_t BRAHMI_VOWEL_SIGN_AU = 0x00011045;
const char32_t BRAHMI_VIRAMA = 0x00011046;
const char32_t BRAHMI_DANDA = 0x00011047;
const char32_t BRAHMI_DOUBLE_DANDA = 0x00011048;
const char32_t BRAHMI_PUNCTUATION_DOT = 0x00011049;
const char32_t BRAHMI_PUNCTUATION_DOUBLE_DOT = 0x0001104A;
const char32_t BRAHMI_PUNCTUATION_LINE = 0x0001104B;
const char32_t BRAHMI_PUNCTUATION_CRESCENT_BAR = 0x0001104C;
const char32_t BRAHMI_PUNCTUATION_LOTUS = 0x0001104D;
const char32_t BRAHMI_NUMBER_ONE = 0x00011052;
const char32_t BRAHMI_NUMBER_TWO = 0x00011053;
const char32_t BRAHMI_NUMBER_THREE = 0x00011054;
const char32_t BRAHMI_NUMBER_FOUR = 0x00011055;
const char32_t BRAHMI_NUMBER_FIVE = 0x00011056;
const char32_t BRAHMI_NUMBER_SIX = 0x00011057;
const char32_t BRAHMI_NUMBER_SEVEN = 0x00011058;
const char32_t BRAHMI_NUMBER_EIGHT = 0x00011059;
const char32_t BRAHMI_NUMBER_NINE = 0x0001105A;
const char32_t BRAHMI_NUMBER_TEN = 0x0001105B;
const char32_t BRAHMI_NUMBER_TWENTY = 0x0001105C;
const char32_t BRAHMI_NUMBER_THIRTY = 0x0001105D;
const char32_t BRAHMI_NUMBER_FORTY = 0x0001105E;
const char32_t BRAHMI_NUMBER_FIFTY = 0x0001105F;
const char32_t BRAHMI_NUMBER_SIXTY = 0x00011060;
const char32_t BRAHMI_NUMBER_SEVENTY = 0x00011061;
const char32_t BRAHMI_NUMBER_EIGHTY = 0x00011062;
const char32_t BRAHMI_NUMBER_NINETY = 0x00011063;
const char32_t BRAHMI_NUMBER_ONE_HUNDRED = 0x00011064;
const char32_t BRAHMI_NUMBER_ONE_THOUSAND = 0x00011065;
const char32_t BRAHMI_DIGIT_ZERO = 0x00011066;
const char32_t BRAHMI_DIGIT_ONE = 0x00011067;
const char32_t BRAHMI_DIGIT_TWO = 0x00011068;
const char32_t BRAHMI_DIGIT_THREE = 0x00011069;
const char32_t BRAHMI_DIGIT_FOUR = 0x0001106A;
const char32_t BRAHMI_DIGIT_FIVE = 0x0001106B;
const char32_t BRAHMI_DIGIT_SIX = 0x0001106C;
const char32_t BRAHMI_DIGIT_SEVEN = 0x0001106D;
const char32_t BRAHMI_DIGIT_EIGHT = 0x0001106E;
const char32_t BRAHMI_DIGIT_NINE = 0x0001106F;
const char32_t BRAHMI_SIGN_OLD_TAMIL_VIRAMA = 0x00011070;
const char32_t BRAHMI_LETTER_OLD_TAMIL_SHORT_E = 0x00011071;
const char32_t BRAHMI_LETTER_OLD_TAMIL_SHORT_O = 0x00011072;
const char32_t BRAHMI_VOWEL_SIGN_OLD_TAMIL_SHORT_E = 0x00011073;
const char32_t BRAHMI_VOWEL_SIGN_OLD_TAMIL_SHORT_O = 0x00011074;
const char32_t BRAHMI_LETTER_OLD_TAMIL_LLA = 0x00011075;
const char32_t BRAHMI_NUMBER_JOINER = 0x0001107F;
const char32_t KAITHI_SIGN_CANDRABINDU = 0x00011080;
const char32_t KAITHI_SIGN_ANUSVARA = 0x00011081;
const char32_t KAITHI_SIGN_VISARGA = 0x00011082;
const char32_t KAITHI_LETTER_A = 0x00011083;
const char32_t KAITHI_LETTER_AA = 0x00011084;
const char32_t KAITHI_LETTER_I = 0x00011085;
const char32_t KAITHI_LETTER_II = 0x00011086;
const char32_t KAITHI_LETTER_U = 0x00011087;
const char32_t KAITHI_LETTER_UU = 0x00011088;
const char32_t KAITHI_LETTER_E = 0x00011089;
const char32_t KAITHI_LETTER_AI = 0x0001108A;
const char32_t KAITHI_LETTER_O = 0x0001108B;
const char32_t KAITHI_LETTER_AU = 0x0001108C;
const char32_t KAITHI_LETTER_KA = 0x0001108D;
const char32_t KAITHI_LETTER_KHA = 0x0001108E;
const char32_t KAITHI_LETTER_GA = 0x0001108F;
const char32_t KAITHI_LETTER_GHA = 0x00011090;
const char32_t KAITHI_LETTER_NGA = 0x00011091;
const char32_t KAITHI_LETTER_CA = 0x00011092;
const char32_t KAITHI_LETTER_CHA = 0x00011093;
const char32_t KAITHI_LETTER_JA = 0x00011094;
const char32_t KAITHI_LETTER_JHA = 0x00011095;
const char32_t KAITHI_LETTER_NYA = 0x00011096;
const char32_t KAITHI_LETTER_TTA = 0x00011097;
const char32_t KAITHI_LETTER_TTHA = 0x00011098;
const char32_t KAITHI_LETTER_DDA = 0x00011099;
const char32_t KAITHI_LETTER_DDDHA = 0x0001109A;
const char32_t KAITHI_LETTER_DDHA = 0x0001109B;
const char32_t KAITHI_LETTER_RHA = 0x0001109C;
const char32_t KAITHI_LETTER_NNA = 0x0001109D;
const char32_t KAITHI_LETTER_TA = 0x0001109E;
const char32_t KAITHI_LETTER_THA = 0x0001109F;
const char32_t KAITHI_LETTER_DA = 0x000110A0;
const char32_t KAITHI_LETTER_DHA = 0x000110A1;
const char32_t KAITHI_LETTER_NA = 0x000110A2;
const char32_t KAITHI_LETTER_PA = 0x000110A3;
const char32_t KAITHI_LETTER_PHA = 0x000110A4;
const char32_t KAITHI_LETTER_BA = 0x000110A5;
const char32_t KAITHI_LETTER_BHA = 0x000110A6;
const char32_t KAITHI_LETTER_MA = 0x000110A7;
const char32_t KAITHI_LETTER_YA = 0x000110A8;
const char32_t KAITHI_LETTER_RA = 0x000110A9;
const char32_t KAITHI_LETTER_LA = 0x000110AA;
const char32_t KAITHI_LETTER_VA = 0x000110AB;
const char32_t KAITHI_LETTER_SHA = 0x000110AC;
const char32_t KAITHI_LETTER_SSA = 0x000110AD;
const char32_t KAITHI_LETTER_SA = 0x000110AE;
const char32_t KAITHI_LETTER_HA = 0x000110AF;
const char32_t KAITHI_VOWEL_SIGN_AA = 0x000110B0;
const char32_t KAITHI_VOWEL_SIGN_I = 0x000110B1;
const char32_t KAITHI_VOWEL_SIGN_II = 0x000110B2;
const char32_t KAITHI_VOWEL_SIGN_U = 0x000110B3;
const char32_t KAITHI_VOWEL_SIGN_UU = 0x000110B4;
const char32_t KAITHI_VOWEL_SIGN_E = 0x000110B5;
const char32_t KAITHI_VOWEL_SIGN_AI = 0x000110B6;
const char32_t KAITHI_VOWEL_SIGN_O = 0x000110B7;
const char32_t KAITHI_VOWEL_SIGN_AU = 0x000110B8;
const char32_t KAITHI_SIGN_VIRAMA = 0x000110B9;
const char32_t KAITHI_SIGN_NUKTA = 0x000110BA;
const char32_t KAITHI_ABBREVIATION_SIGN = 0x000110BB;
const char32_t KAITHI_ENUMERATION_SIGN = 0x000110BC;
const char32_t KAITHI_NUMBER_SIGN = 0x000110BD;
const char32_t KAITHI_SECTION_MARK = 0x000110BE;
const char32_t KAITHI_DOUBLE_SECTION_MARK = 0x000110BF;
const char32_t KAITHI_DANDA = 0x000110C0;
const char32_t KAITHI_DOUBLE_DANDA = 0x000110C1;
const char32_t KAITHI_VOWEL_SIGN_VOCALIC_R = 0x000110C2;
const char32_t KAITHI_NUMBER_SIGN_ABOVE = 0x000110CD;
const char32_t SORA_SOMPENG_LETTER_SAH = 0x000110D0;
const char32_t SORA_SOMPENG_LETTER_TAH = 0x000110D1;
const char32_t SORA_SOMPENG_LETTER_BAH = 0x000110D2;
const char32_t SORA_SOMPENG_LETTER_CAH = 0x000110D3;
const char32_t SORA_SOMPENG_LETTER_DAH = 0x000110D4;
const char32_t SORA_SOMPENG_LETTER_GAH = 0x000110D5;
const char32_t SORA_SOMPENG_LETTER_MAH = 0x000110D6;
const char32_t SORA_SOMPENG_LETTER_NGAH = 0x000110D7;
const char32_t SORA_SOMPENG_LETTER_LAH = 0x000110D8;
const char32_t SORA_SOMPENG_LETTER_NAH = 0x000110D9;
const char32_t SORA_SOMPENG_LETTER_VAH = 0x000110DA;
const char32_t SORA_SOMPENG_LETTER_PAH = 0x000110DB;
const char32_t SORA_SOMPENG_LETTER_YAH = 0x000110DC;
const char32_t SORA_SOMPENG_LETTER_RAH = 0x000110DD;
const char32_t SORA_SOMPENG_LETTER_HAH = 0x000110DE;
const char32_t SORA_SOMPENG_LETTER_KAH = 0x000110DF;
const char32_t SORA_SOMPENG_LETTER_JAH = 0x000110E0;
const char32_t SORA_SOMPENG_LETTER_NYAH = 0x000110E1;
const char32_t SORA_SOMPENG_LETTER_AH = 0x000110E2;
const char32_t SORA_SOMPENG_LETTER_EEH = 0x000110E3;
const char32_t SORA_SOMPENG_LETTER_IH = 0x000110E4;
const char32_t SORA_SOMPENG_LETTER_UH = 0x000110E5;
const char32_t SORA_SOMPENG_LETTER_OH = 0x000110E6;
const char32_t SORA_SOMPENG_LETTER_EH = 0x000110E7;
const char32_t SORA_SOMPENG_LETTER_MAE = 0x000110E8;
const char32_t SORA_SOMPENG_DIGIT_ZERO = 0x000110F0;
const char32_t SORA_SOMPENG_DIGIT_ONE = 0x000110F1;
const char32_t SORA_SOMPENG_DIGIT_TWO = 0x000110F2;
const char32_t SORA_SOMPENG_DIGIT_THREE = 0x000110F3;
const char32_t SORA_SOMPENG_DIGIT_FOUR = 0x000110F4;
const char32_t SORA_SOMPENG_DIGIT_FIVE = 0x000110F5;
const char32_t SORA_SOMPENG_DIGIT_SIX = 0x000110F6;
const char32_t SORA_SOMPENG_DIGIT_SEVEN = 0x000110F7;
const char32_t SORA_SOMPENG_DIGIT_EIGHT = 0x000110F8;
const char32_t SORA_SOMPENG_DIGIT_NINE = 0x000110F9;
const char32_t CHAKMA_SIGN_CANDRABINDU = 0x00011100;
const char32_t CHAKMA_SIGN_ANUSVARA = 0x00011101;
const char32_t CHAKMA_SIGN_VISARGA = 0x00011102;
const char32_t CHAKMA_LETTER_AA = 0x00011103;
const char32_t CHAKMA_LETTER_I = 0x00011104;
const char32_t CHAKMA_LETTER_U = 0x00011105;
const char32_t CHAKMA_LETTER_E = 0x00011106;
const char32_t CHAKMA_LETTER_KAA = 0x00011107;
const char32_t CHAKMA_LETTER_KHAA = 0x00011108;
const char32_t CHAKMA_LETTER_GAA = 0x00011109;
const char32_t CHAKMA_LETTER_GHAA = 0x0001110A;
const char32_t CHAKMA_LETTER_NGAA = 0x0001110B;
const char32_t CHAKMA_LETTER_CAA = 0x0001110C;
const char32_t CHAKMA_LETTER_CHAA = 0x0001110D;
const char32_t CHAKMA_LETTER_JAA = 0x0001110E;
const char32_t CHAKMA_LETTER_JHAA = 0x0001110F;
const char32_t CHAKMA_LETTER_NYAA = 0x00011110;
const char32_t CHAKMA_LETTER_TTAA = 0x00011111;
const char32_t CHAKMA_LETTER_TTHAA = 0x00011112;
const char32_t CHAKMA_LETTER_DDAA = 0x00011113;
const char32_t CHAKMA_LETTER_DDHAA = 0x00011114;
const char32_t CHAKMA_LETTER_NNAA = 0x00011115;
const char32_t CHAKMA_LETTER_TAA = 0x00011116;
const char32_t CHAKMA_LETTER_THAA = 0x00011117;
const char32_t CHAKMA_LETTER_DAA = 0x00011118;
const char32_t CHAKMA_LETTER_DHAA = 0x00011119;
const char32_t CHAKMA_LETTER_NAA = 0x0001111A;
const char32_t CHAKMA_LETTER_PAA = 0x0001111B;
const char32_t CHAKMA_LETTER_PHAA = 0x0001111C;
const char32_t CHAKMA_LETTER_BAA = 0x0001111D;
const char32_t CHAKMA_LETTER_BHAA = 0x0001111E;
const char32_t CHAKMA_LETTER_MAA = 0x0001111F;
const char32_t CHAKMA_LETTER_YYAA = 0x00011120;
const char32_t CHAKMA_LETTER_YAA = 0x00011121;
const char32_t CHAKMA_LETTER_RAA = 0x00011122;
const char32_t CHAKMA_LETTER_LAA = 0x00011123;
const char32_t CHAKMA_LETTER_WAA = 0x00011124;
const char32_t CHAKMA_LETTER_SAA = 0x00011125;
const char32_t CHAKMA_LETTER_HAA = 0x00011126;
const char32_t CHAKMA_VOWEL_SIGN_A = 0x00011127;
const char32_t CHAKMA_VOWEL_SIGN_I = 0x00011128;
const char32_t CHAKMA_VOWEL_SIGN_II = 0x00011129;
const char32_t CHAKMA_VOWEL_SIGN_U = 0x0001112A;
const char32_t CHAKMA_VOWEL_SIGN_UU = 0x0001112B;
const char32_t CHAKMA_VOWEL_SIGN_E = 0x0001112C;
const char32_t CHAKMA_VOWEL_SIGN_AI = 0x0001112D;
const char32_t CHAKMA_VOWEL_SIGN_O = 0x0001112E;
const char32_t CHAKMA_VOWEL_SIGN_AU = 0x0001112F;
const char32_t CHAKMA_VOWEL_SIGN_OI = 0x00011130;
const char32_t CHAKMA_O_MARK = 0x00011131;
const char32_t CHAKMA_AU_MARK = 0x00011132;
const char32_t CHAKMA_VIRAMA = 0x00011133;
const char32_t CHAKMA_MAAYYAA = 0x00011134;
const char32_t CHAKMA_DIGIT_ZERO = 0x00011136;
const char32_t CHAKMA_DIGIT_ONE = 0x00011137;
const char32_t CHAKMA_DIGIT_TWO = 0x00011138;
const char32_t CHAKMA_DIGIT_THREE = 0x00011139;
const char32_t CHAKMA_DIGIT_FOUR = 0x0001113A;
const char32_t CHAKMA_DIGIT_FIVE = 0x0001113B;
const char32_t CHAKMA_DIGIT_SIX = 0x0001113C;
const char32_t CHAKMA_DIGIT_SEVEN = 0x0001113D;
const char32_t CHAKMA_DIGIT_EIGHT = 0x0001113E;
const char32_t CHAKMA_DIGIT_NINE = 0x0001113F;
const char32_t CHAKMA_SECTION_MARK = 0x00011140;
const char32_t CHAKMA_DANDA = 0x00011141;
const char32_t CHAKMA_DOUBLE_DANDA = 0x00011142;
const char32_t CHAKMA_QUESTION_MARK = 0x00011143;
const char32_t CHAKMA_LETTER_LHAA = 0x00011144;
const char32_t CHAKMA_VOWEL_SIGN_AA = 0x00011145;
const char32_t CHAKMA_VOWEL_SIGN_EI = 0x00011146;
const char32_t CHAKMA_LETTER_VAA = 0x00011147;
const char32_t MAHAJANI_LETTER_A = 0x00011150;
const char32_t MAHAJANI_LETTER_I = 0x00011151;
const char32_t MAHAJANI_LETTER_U = 0x00011152;
const char32_t MAHAJANI_LETTER_E = 0x00011153;
const char32_t MAHAJANI_LETTER_O = 0x00011154;
const char32_t MAHAJANI_LETTER_KA = 0x00011155;
const char32_t MAHAJANI_LETTER_KHA = 0x00011156;
const char32_t MAHAJANI_LETTER_GA = 0x00011157;
const char32_t MAHAJANI_LETTER_GHA = 0x00011158;
const char32_t MAHAJANI_LETTER_CA = 0x00011159;
const char32_t MAHAJANI_LETTER_CHA = 0x0001115A;
const char32_t MAHAJANI_LETTER_JA = 0x0001115B;
const char32_t MAHAJANI_LETTER_JHA = 0x0001115C;
const char32_t MAHAJANI_LETTER_NYA = 0x0001115D;
const char32_t MAHAJANI_LETTER_TTA = 0x0001115E;
const char32_t MAHAJANI_LETTER_TTHA = 0x0001115F;
const char32_t MAHAJANI_LETTER_DDA = 0x00011160;
const char32_t MAHAJANI_LETTER_DDHA = 0x00011161;
const char32_t MAHAJANI_LETTER_NNA = 0x00011162;
const char32_t MAHAJANI_LETTER_TA = 0x00011163;
const char32_t MAHAJANI_LETTER_THA = 0x00011164;
const char32_t MAHAJANI_LETTER_DA = 0x00011165;
const char32_t MAHAJANI_LETTER_DHA = 0x00011166;
const char32_t MAHAJANI_LETTER_NA = 0x00011167;
const char32_t MAHAJANI_LETTER_PA = 0x00011168;
const char32_t MAHAJANI_LETTER_PHA = 0x00011169;
const char32_t MAHAJANI_LETTER_BA = 0x0001116A;
const char32_t MAHAJANI_LETTER_BHA = 0x0001116B;
const char32_t MAHAJANI_LETTER_MA = 0x0001116C;
const char32_t MAHAJANI_LETTER_RA = 0x0001116D;
const char32_t MAHAJANI_LETTER_LA = 0x0001116E;
const char32_t MAHAJANI_LETTER_VA = 0x0001116F;
const char32_t MAHAJANI_LETTER_SA = 0x00011170;
const char32_t MAHAJANI_LETTER_HA = 0x00011171;
const char32_t MAHAJANI_LETTER_RRA = 0x00011172;
const char32_t MAHAJANI_SIGN_NUKTA = 0x00011173;
const char32_t MAHAJANI_ABBREVIATION_SIGN = 0x00011174;
const char32_t MAHAJANI_SECTION_MARK = 0x00011175;
const char32_t MAHAJANI_LIGATURE_SHRI = 0x00011176;
const char32_t SHARADA_SIGN_CANDRABINDU = 0x00011180;
const char32_t SHARADA_SIGN_ANUSVARA = 0x00011181;
const char32_t SHARADA_SIGN_VISARGA = 0x00011182;
const char32_t SHARADA_LETTER_A = 0x00011183;
const char32_t SHARADA_LETTER_AA = 0x00011184;
const char32_t SHARADA_LETTER_I = 0x00011185;
const char32_t SHARADA_LETTER_II = 0x00011186;
const char32_t SHARADA_LETTER_U = 0x00011187;
const char32_t SHARADA_LETTER_UU = 0x00011188;
const char32_t SHARADA_LETTER_VOCALIC_R = 0x00011189;
const char32_t SHARADA_LETTER_VOCALIC_RR = 0x0001118A;
const char32_t SHARADA_LETTER_VOCALIC_L = 0x0001118B;
const char32_t SHARADA_LETTER_VOCALIC_LL = 0x0001118C;
const char32_t SHARADA_LETTER_E = 0x0001118D;
const char32_t SHARADA_LETTER_AI = 0x0001118E;
const char32_t SHARADA_LETTER_O = 0x0001118F;
const char32_t SHARADA_LETTER_AU = 0x00011190;
const char32_t SHARADA_LETTER_KA = 0x00011191;
const char32_t SHARADA_LETTER_KHA = 0x00011192;
const char32_t SHARADA_LETTER_GA = 0x00011193;
const char32_t SHARADA_LETTER_GHA = 0x00011194;
const char32_t SHARADA_LETTER_NGA = 0x00011195;
const char32_t SHARADA_LETTER_CA = 0x00011196;
const char32_t SHARADA_LETTER_CHA = 0x00011197;
const char32_t SHARADA_LETTER_JA = 0x00011198;
const char32_t SHARADA_LETTER_JHA = 0x00011199;
const char32_t SHARADA_LETTER_NYA = 0x0001119A;
const char32_t SHARADA_LETTER_TTA = 0x0001119B;
const char32_t SHARADA_LETTER_TTHA = 0x0001119C;
const char32_t SHARADA_LETTER_DDA = 0x0001119D;
const char32_t SHARADA_LETTER_DDHA = 0x0001119E;
const char32_t SHARADA_LETTER_NNA = 0x0001119F;
const char32_t SHARADA_LETTER_TA = 0x000111A0;
const char32_t SHARADA_LETTER_THA = 0x000111A1;
const char32_t SHARADA_LETTER_DA = 0x000111A2;
const char32_t SHARADA_LETTER_DHA = 0x000111A3;
const char32_t SHARADA_LETTER_NA = 0x000111A4;
const char32_t SHARADA_LETTER_PA = 0x000111A5;
const char32_t SHARADA_LETTER_PHA = 0x000111A6;
const char32_t SHARADA_LETTER_BA = 0x000111A7;
const char32_t SHARADA_LETTER_BHA = 0x000111A8;
const char32_t SHARADA_LETTER_MA = 0x000111A9;
const char32_t SHARADA_LETTER_YA = 0x000111AA;
const char32_t SHARADA_LETTER_RA = 0x000111AB;
const char32_t SHARADA_LETTER_LA = 0x000111AC;
const char32_t SHARADA_LETTER_LLA = 0x000111AD;
const char32_t SHARADA_LETTER_VA = 0x000111AE;
const char32_t SHARADA_LETTER_SHA = 0x000111AF;
const char32_t SHARADA_LETTER_SSA = 0x000111B0;
const char32_t SHARADA_LETTER_SA = 0x000111B1;
const char32_t SHARADA_LETTER_HA = 0x000111B2;
const char32_t SHARADA_VOWEL_SIGN_AA = 0x000111B3;
const char32_t SHARADA_VOWEL_SIGN_I = 0x000111B4;
const char32_t SHARADA_VOWEL_SIGN_II = 0x000111B5;
const char32_t SHARADA_VOWEL_SIGN_U = 0x000111B6;
const char32_t SHARADA_VOWEL_SIGN_UU = 0x000111B7;
const char32_t SHARADA_VOWEL_SIGN_VOCALIC_R = 0x000111B8;
const char32_t SHARADA_VOWEL_SIGN_VOCALIC_RR = 0x000111B9;
const char32_t SHARADA_VOWEL_SIGN_VOCALIC_L = 0x000111BA;
const char32_t SHARADA_VOWEL_SIGN_VOCALIC_LL = 0x000111BB;
const char32_t SHARADA_VOWEL_SIGN_E = 0x000111BC;
const char32_t SHARADA_VOWEL_SIGN_AI = 0x000111BD;
const char32_t SHARADA_VOWEL_SIGN_O = 0x000111BE;
const char32_t SHARADA_VOWEL_SIGN_AU = 0x000111BF;
const char32_t SHARADA_SIGN_VIRAMA = 0x000111C0;
const char32_t SHARADA_SIGN_AVAGRAHA = 0x000111C1;
const char32_t SHARADA_SIGN_JIHVAMULIYA = 0x000111C2;
const char32_t SHARADA_SIGN_UPADHMANIYA = 0x000111C3;
const char32_t SHARADA_OM = 0x000111C4;
const char32_t SHARADA_DANDA = 0x000111C5;
const char32_t SHARADA_DOUBLE_DANDA = 0x000111C6;
const char32_t SHARADA_ABBREVIATION_SIGN = 0x000111C7;
const char32_t SHARADA_SEPARATOR = 0x000111C8;
const char32_t SHARADA_SANDHI_MARK = 0x000111C9;
const char32_t SHARADA_SIGN_NUKTA = 0x000111CA;
const char32_t SHARADA_VOWEL_MODIFIER_MARK = 0x000111CB;
const char32_t SHARADA_EXTRA_SHORT_VOWEL_MARK = 0x000111CC;
const char32_t SHARADA_SUTRA_MARK = 0x000111CD;
const char32_t SHARADA_VOWEL_SIGN_PRISHTHAMATRA_E = 0x000111CE;
const char32_t SHARADA_SIGN_INVERTED_CANDRABINDU = 0x000111CF;
const char32_t SHARADA_DIGIT_ZERO = 0x000111D0;
const char32_t SHARADA_DIGIT_ONE = 0x000111D1;
const char32_t SHARADA_DIGIT_TWO = 0x000111D2;
const char32_t SHARADA_DIGIT_THREE = 0x000111D3;
const char32_t SHARADA_DIGIT_FOUR = 0x000111D4;
const char32_t SHARADA_DIGIT_FIVE = 0x000111D5;
const char32_t SHARADA_DIGIT_SIX = 0x000111D6;
const char32_t SHARADA_DIGIT_SEVEN = 0x000111D7;
const char32_t SHARADA_DIGIT_EIGHT = 0x000111D8;
const char32_t SHARADA_DIGIT_NINE = 0x000111D9;
const char32_t SHARADA_EKAM = 0x000111DA;
const char32_t SHARADA_SIGN_SIDDHAM = 0x000111DB;
const char32_t SHARADA_HEADSTROKE = 0x000111DC;
const char32_t SHARADA_CONTINUATION_SIGN = 0x000111DD;
const char32_t SHARADA_SECTION_MARK_1 = 0x000111DE;
const char32_t SHARADA_SECTION_MARK_2 = 0x000111DF;
const char32_t SINHALA_ARCHAIC_DIGIT_ONE = 0x000111E1;
const char32_t SINHALA_ARCHAIC_DIGIT_TWO = 0x000111E2;
const char32_t SINHALA_ARCHAIC_DIGIT_THREE = 0x000111E3;
const char32_t SINHALA_ARCHAIC_DIGIT_FOUR = 0x000111E4;
const char32_t SINHALA_ARCHAIC_DIGIT_FIVE = 0x000111E5;
const char32_t SINHALA_ARCHAIC_DIGIT_SIX = 0x000111E6;
const char32_t SINHALA_ARCHAIC_DIGIT_SEVEN = 0x000111E7;
const char32_t SINHALA_ARCHAIC_DIGIT_EIGHT = 0x000111E8;
const char32_t SINHALA_ARCHAIC_DIGIT_NINE = 0x000111E9;
const char32_t SINHALA_ARCHAIC_NUMBER_TEN = 0x000111EA;
const char32_t SINHALA_ARCHAIC_NUMBER_TWENTY = 0x000111EB;
const char32_t SINHALA_ARCHAIC_NUMBER_THIRTY = 0x000111EC;
const char32_t SINHALA_ARCHAIC_NUMBER_FORTY = 0x000111ED;
const char32_t SINHALA_ARCHAIC_NUMBER_FIFTY = 0x000111EE;
const char32_t SINHALA_ARCHAIC_NUMBER_SIXTY = 0x000111EF;
const char32_t SINHALA_ARCHAIC_NUMBER_SEVENTY = 0x000111F0;
const char32_t SINHALA_ARCHAIC_NUMBER_EIGHTY = 0x000111F1;
const char32_t SINHALA_ARCHAIC_NUMBER_NINETY = 0x000111F2;
const char32_t SINHALA_ARCHAIC_NUMBER_ONE_HUNDRED = 0x000111F3;
const char32_t SINHALA_ARCHAIC_NUMBER_ONE_THOUSAND = 0x000111F4;
const char32_t KHOJKI_LETTER_A = 0x00011200;
const char32_t KHOJKI_LETTER_AA = 0x00011201;
const char32_t KHOJKI_LETTER_I = 0x00011202;
const char32_t KHOJKI_LETTER_U = 0x00011203;
const char32_t KHOJKI_LETTER_E = 0x00011204;
const char32_t KHOJKI_LETTER_AI = 0x00011205;
const char32_t KHOJKI_LETTER_O = 0x00011206;
const char32_t KHOJKI_LETTER_AU = 0x00011207;
const char32_t KHOJKI_LETTER_KA = 0x00011208;
const char32_t KHOJKI_LETTER_KHA = 0x00011209;
const char32_t KHOJKI_LETTER_GA = 0x0001120A;
const char32_t KHOJKI_LETTER_GGA = 0x0001120B;
const char32_t KHOJKI_LETTER_GHA = 0x0001120C;
const char32_t KHOJKI_LETTER_NGA = 0x0001120D;
const char32_t KHOJKI_LETTER_CA = 0x0001120E;
const char32_t KHOJKI_LETTER_CHA = 0x0001120F;
const char32_t KHOJKI_LETTER_JA = 0x00011210;
const char32_t KHOJKI_LETTER_JJA = 0x00011211;
const char32_t KHOJKI_LETTER_NYA = 0x00011213;
const char32_t KHOJKI_LETTER_TTA = 0x00011214;
const char32_t KHOJKI_LETTER_TTHA = 0x00011215;
const char32_t KHOJKI_LETTER_DDA = 0x00011216;
const char32_t KHOJKI_LETTER_DDHA = 0x00011217;
const char32_t KHOJKI_LETTER_NNA = 0x00011218;
const char32_t KHOJKI_LETTER_TA = 0x00011219;
const char32_t KHOJKI_LETTER_THA = 0x0001121A;
const char32_t KHOJKI_LETTER_DA = 0x0001121B;
const char32_t KHOJKI_LETTER_DDDA = 0x0001121C;
const char32_t KHOJKI_LETTER_DHA = 0x0001121D;
const char32_t KHOJKI_LETTER_NA = 0x0001121E;
const char32_t KHOJKI_LETTER_PA = 0x0001121F;
const char32_t KHOJKI_LETTER_PHA = 0x00011220;
const char32_t KHOJKI_LETTER_BA = 0x00011221;
const char32_t KHOJKI_LETTER_BBA = 0x00011222;
const char32_t KHOJKI_LETTER_BHA = 0x00011223;
const char32_t KHOJKI_LETTER_MA = 0x00011224;
const char32_t KHOJKI_LETTER_YA = 0x00011225;
const char32_t KHOJKI_LETTER_RA = 0x00011226;
const char32_t KHOJKI_LETTER_LA = 0x00011227;
const char32_t KHOJKI_LETTER_VA = 0x00011228;
const char32_t KHOJKI_LETTER_SA = 0x00011229;
const char32_t KHOJKI_LETTER_HA = 0x0001122A;
const char32_t KHOJKI_LETTER_LLA = 0x0001122B;
const char32_t KHOJKI_VOWEL_SIGN_AA = 0x0001122C;
const char32_t KHOJKI_VOWEL_SIGN_I = 0x0001122D;
const char32_t KHOJKI_VOWEL_SIGN_II = 0x0001122E;
const char32_t KHOJKI_VOWEL_SIGN_U = 0x0001122F;
const char32_t KHOJKI_VOWEL_SIGN_E = 0x00011230;
const char32_t KHOJKI_VOWEL_SIGN_AI = 0x00011231;
const char32_t KHOJKI_VOWEL_SIGN_O = 0x00011232;
const char32_t KHOJKI_VOWEL_SIGN_AU = 0x00011233;
const char32_t KHOJKI_SIGN_ANUSVARA = 0x00011234;
const char32_t KHOJKI_SIGN_VIRAMA = 0x00011235;
const char32_t KHOJKI_SIGN_NUKTA = 0x00011236;
const char32_t KHOJKI_SIGN_SHADDA = 0x00011237;
const char32_t KHOJKI_DANDA = 0x00011238;
const char32_t KHOJKI_DOUBLE_DANDA = 0x00011239;
const char32_t KHOJKI_WORD_SEPARATOR = 0x0001123A;
const char32_t KHOJKI_SECTION_MARK = 0x0001123B;
const char32_t KHOJKI_DOUBLE_SECTION_MARK = 0x0001123C;
const char32_t KHOJKI_ABBREVIATION_SIGN = 0x0001123D;
const char32_t KHOJKI_SIGN_SUKUN = 0x0001123E;
const char32_t KHOJKI_LETTER_QA = 0x0001123F;
const char32_t KHOJKI_LETTER_SHORT_I = 0x00011240;
const char32_t KHOJKI_VOWEL_SIGN_VOCALIC_R = 0x00011241;
const char32_t MULTANI_LETTER_A = 0x00011280;
const char32_t MULTANI_LETTER_I = 0x00011281;
const char32_t MULTANI_LETTER_U = 0x00011282;
const char32_t MULTANI_LETTER_E = 0x00011283;
const char32_t MULTANI_LETTER_KA = 0x00011284;
const char32_t MULTANI_LETTER_KHA = 0x00011285;
const char32_t MULTANI_LETTER_GA = 0x00011286;
const char32_t MULTANI_LETTER_GHA = 0x00011288;
const char32_t MULTANI_LETTER_CA = 0x0001128A;
const char32_t MULTANI_LETTER_CHA = 0x0001128B;
const char32_t MULTANI_LETTER_JA = 0x0001128C;
const char32_t MULTANI_LETTER_JJA = 0x0001128D;
const char32_t MULTANI_LETTER_NYA = 0x0001128F;
const char32_t MULTANI_LETTER_TTA = 0x00011290;
const char32_t MULTANI_LETTER_TTHA = 0x00011291;
const char32_t MULTANI_LETTER_DDA = 0x00011292;
const char32_t MULTANI_LETTER_DDDA = 0x00011293;
const char32_t MULTANI_LETTER_DDHA = 0x00011294;
const char32_t MULTANI_LETTER_NNA = 0x00011295;
const char32_t MULTANI_LETTER_TA = 0x00011296;
const char32_t MULTANI_LETTER_THA = 0x00011297;
const char32_t MULTANI_LETTER_DA = 0x00011298;
const char32_t MULTANI_LETTER_DHA = 0x00011299;
const char32_t MULTANI_LETTER_NA = 0x0001129A;
const char32_t MULTANI_LETTER_PA = 0x0001129B;
const char32_t MULTANI_LETTER_PHA = 0x0001129C;
const char32_t MULTANI_LETTER_BA = 0x0001129D;
const char32_t MULTANI_LETTER_BHA = 0x0001129F;
const char32_t MULTANI_LETTER_MA = 0x000112A0;
const char32_t MULTANI_LETTER_YA = 0x000112A1;
const char32_t MULTANI_LETTER_RA = 0x000112A2;
const char32_t MULTANI_LETTER_LA = 0x000112A3;
const char32_t MULTANI_LETTER_VA = 0x000112A4;
const char32_t MULTANI_LETTER_SA = 0x000112A5;
const char32_t MULTANI_LETTER_HA = 0x000112A6;
const char32_t MULTANI_LETTER_RRA = 0x000112A7;
const char32_t MULTANI_LETTER_RHA = 0x000112A8;
const char32_t MULTANI_SECTION_MARK = 0x000112A9;
const char32_t KHUDAWADI_LETTER_A = 0x000112B0;
const char32_t KHUDAWADI_LETTER_AA = 0x000112B1;
const char32_t KHUDAWADI_LETTER_I = 0x000112B2;
const char32_t KHUDAWADI_LETTER_II = 0x000112B3;
const char32_t KHUDAWADI_LETTER_U = 0x000112B4;
const char32_t KHUDAWADI_LETTER_UU = 0x000112B5;
const char32_t KHUDAWADI_LETTER_E = 0x000112B6;
const char32_t KHUDAWADI_LETTER_AI = 0x000112B7;
const char32_t KHUDAWADI_LETTER_O = 0x000112B8;
const char32_t KHUDAWADI_LETTER_AU = 0x000112B9;
const char32_t KHUDAWADI_LETTER_KA = 0x000112BA;
const char32_t KHUDAWADI_LETTER_KHA = 0x000112BB;
const char32_t KHUDAWADI_LETTER_GA = 0x000112BC;
const char32_t KHUDAWADI_LETTER_GGA = 0x000112BD;
const char32_t KHUDAWADI_LETTER_GHA = 0x000112BE;
const char32_t KHUDAWADI_LETTER_NGA = 0x000112BF;
const char32_t KHUDAWADI_LETTER_CA = 0x000112C0;
const char32_t KHUDAWADI_LETTER_CHA = 0x000112C1;
const char32_t KHUDAWADI_LETTER_JA = 0x000112C2;
const char32_t KHUDAWADI_LETTER_JJA = 0x000112C3;
const char32_t KHUDAWADI_LETTER_JHA = 0x000112C4;
const char32_t KHUDAWADI_LETTER_NYA = 0x000112C5;
const char32_t KHUDAWADI_LETTER_TTA = 0x000112C6;
const char32_t KHUDAWADI_LETTER_TTHA = 0x000112C7;
const char32_t KHUDAWADI_LETTER_DDA = 0x000112C8;
const char32_t KHUDAWADI_LETTER_DDDA = 0x000112C9;
const char32_t KHUDAWADI_LETTER_RRA = 0x000112CA;
const char32_t KHUDAWADI_LETTER_DDHA = 0x000112CB;
const char32_t KHUDAWADI_LETTER_NNA = 0x000112CC;
const char32_t KHUDAWADI_LETTER_TA = 0x000112CD;
const char32_t KHUDAWADI_LETTER_THA = 0x000112CE;
const char32_t KHUDAWADI_LETTER_DA = 0x000112CF;
const char32_t KHUDAWADI_LETTER_DHA = 0x000112D0;
const char32_t KHUDAWADI_LETTER_NA = 0x000112D1;
const char32_t KHUDAWADI_LETTER_PA = 0x000112D2;
const char32_t KHUDAWADI_LETTER_PHA = 0x000112D3;
const char32_t KHUDAWADI_LETTER_BA = 0x000112D4;
const char32_t KHUDAWADI_LETTER_BBA = 0x000112D5;
const char32_t KHUDAWADI_LETTER_BHA = 0x000112D6;
const char32_t KHUDAWADI_LETTER_MA = 0x000112D7;
const char32_t KHUDAWADI_LETTER_YA = 0x000112D8;
const char32_t KHUDAWADI_LETTER_RA = 0x000112D9;
const char32_t KHUDAWADI_LETTER_LA = 0x000112DA;
const char32_t KHUDAWADI_LETTER_VA = 0x000112DB;
const char32_t KHUDAWADI_LETTER_SHA = 0x000112DC;
const char32_t KHUDAWADI_LETTER_SA = 0x000112DD;
const char32_t KHUDAWADI_LETTER_HA = 0x000112DE;
const char32_t KHUDAWADI_SIGN_ANUSVARA = 0x000112DF;
const char32_t KHUDAWADI_VOWEL_SIGN_AA = 0x000112E0;
const char32_t KHUDAWADI_VOWEL_SIGN_I = 0x000112E1;
const char32_t KHUDAWADI_VOWEL_SIGN_II = 0x000112E2;
const char32_t KHUDAWADI_VOWEL_SIGN_U = 0x000112E3;
const char32_t KHUDAWADI_VOWEL_SIGN_UU = 0x000112E4;
const char32_t KHUDAWADI_VOWEL_SIGN_E = 0x000112E5;
const char32_t KHUDAWADI_VOWEL_SIGN_AI = 0x000112E6;
const char32_t KHUDAWADI_VOWEL_SIGN_O = 0x000112E7;
const char32_t KHUDAWADI_VOWEL_SIGN_AU = 0x000112E8;
const char32_t KHUDAWADI_SIGN_NUKTA = 0x000112E9;
const char32_t KHUDAWADI_SIGN_VIRAMA = 0x000112EA;
const char32_t KHUDAWADI_DIGIT_ZERO = 0x000112F0;
const char32_t KHUDAWADI_DIGIT_ONE = 0x000112F1;
const char32_t KHUDAWADI_DIGIT_TWO = 0x000112F2;
const char32_t KHUDAWADI_DIGIT_THREE = 0x000112F3;
const char32_t KHUDAWADI_DIGIT_FOUR = 0x000112F4;
const char32_t KHUDAWADI_DIGIT_FIVE = 0x000112F5;
const char32_t KHUDAWADI_DIGIT_SIX = 0x000112F6;
const char32_t KHUDAWADI_DIGIT_SEVEN = 0x000112F7;
const char32_t KHUDAWADI_DIGIT_EIGHT = 0x000112F8;
const char32_t KHUDAWADI_DIGIT_NINE = 0x000112F9;
const char32_t GRANTHA_SIGN_COMBINING_ANUSVARA_ABOVE = 0x00011300;
const char32_t GRANTHA_SIGN_CANDRABINDU = 0x00011301;
const char32_t GRANTHA_SIGN_ANUSVARA = 0x00011302;
const char32_t GRANTHA_SIGN_VISARGA = 0x00011303;
const char32_t GRANTHA_LETTER_A = 0x00011305;
const char32_t GRANTHA_LETTER_AA = 0x00011306;
const char32_t GRANTHA_LETTER_I = 0x00011307;
const char32_t GRANTHA_LETTER_II = 0x00011308;
const char32_t GRANTHA_LETTER_U = 0x00011309;
const char32_t GRANTHA_LETTER_UU = 0x0001130A;
const char32_t GRANTHA_LETTER_VOCALIC_R = 0x0001130B;
const char32_t GRANTHA_LETTER_VOCALIC_L = 0x0001130C;
const char32_t GRANTHA_LETTER_EE = 0x0001130F;
const char32_t GRANTHA_LETTER_AI = 0x00011310;
const char32_t GRANTHA_LETTER_OO = 0x00011313;
const char32_t GRANTHA_LETTER_AU = 0x00011314;
const char32_t GRANTHA_LETTER_KA = 0x00011315;
const char32_t GRANTHA_LETTER_KHA = 0x00011316;
const char32_t GRANTHA_LETTER_GA = 0x00011317;
const char32_t GRANTHA_LETTER_GHA = 0x00011318;
const char32_t GRANTHA_LETTER_NGA = 0x00011319;
const char32_t GRANTHA_LETTER_CA = 0x0001131A;
const char32_t GRANTHA_LETTER_CHA = 0x0001131B;
const char32_t GRANTHA_LETTER_JA = 0x0001131C;
const char32_t GRANTHA_LETTER_JHA = 0x0001131D;
const char32_t GRANTHA_LETTER_NYA = 0x0001131E;
const char32_t GRANTHA_LETTER_TTA = 0x0001131F;
const char32_t GRANTHA_LETTER_TTHA = 0x00011320;
const char32_t GRANTHA_LETTER_DDA = 0x00011321;
const char32_t GRANTHA_LETTER_DDHA = 0x00011322;
const char32_t GRANTHA_LETTER_NNA = 0x00011323;
const char32_t GRANTHA_LETTER_TA = 0x00011324;
const char32_t GRANTHA_LETTER_THA = 0x00011325;
const char32_t GRANTHA_LETTER_DA = 0x00011326;
const char32_t GRANTHA_LETTER_DHA = 0x00011327;
const char32_t GRANTHA_LETTER_NA = 0x00011328;
const char32_t GRANTHA_LETTER_PA = 0x0001132A;
const char32_t GRANTHA_LETTER_PHA = 0x0001132B;
const char32_t GRANTHA_LETTER_BA = 0x0001132C;
const char32_t GRANTHA_LETTER_BHA = 0x0001132D;
const char32_t GRANTHA_LETTER_MA = 0x0001132E;
const char32_t GRANTHA_LETTER_YA = 0x0001132F;
const char32_t GRANTHA_LETTER_RA = 0x00011330;
const char32_t GRANTHA_LETTER_LA = 0x00011332;
const char32_t GRANTHA_LETTER_LLA = 0x00011333;
const char32_t GRANTHA_LETTER_VA = 0x00011335;
const char32_t GRANTHA_LETTER_SHA = 0x00011336;
const char32_t GRANTHA_LETTER_SSA = 0x00011337;
const char32_t GRANTHA_LETTER_SA = 0x00011338;
const char32_t GRANTHA_LETTER_HA = 0x00011339;
const char32_t COMBINING_BINDU_BELOW = 0x0001133B;
const char32_t GRANTHA_SIGN_NUKTA = 0x0001133C;
const char32_t GRANTHA_SIGN_AVAGRAHA = 0x0001133D;
const char32_t GRANTHA_VOWEL_SIGN_AA = 0x0001133E;
const char32_t GRANTHA_VOWEL_SIGN_I = 0x0001133F;
const char32_t GRANTHA_VOWEL_SIGN_II = 0x00011340;
const char32_t GRANTHA_VOWEL_SIGN_U = 0x00011341;
const char32_t GRANTHA_VOWEL_SIGN_UU = 0x00011342;
const char32_t GRANTHA_VOWEL_SIGN_VOCALIC_R = 0x00011343;
const char32_t GRANTHA_VOWEL_SIGN_VOCALIC_RR = 0x00011344;
const char32_t GRANTHA_VOWEL_SIGN_EE = 0x00011347;
const char32_t GRANTHA_VOWEL_SIGN_AI = 0x00011348;
const char32_t GRANTHA_VOWEL_SIGN_OO = 0x0001134B;
const char32_t GRANTHA_VOWEL_SIGN_AU = 0x0001134C;
const char32_t GRANTHA_SIGN_VIRAMA = 0x0001134D;
const char32_t GRANTHA_OM = 0x00011350;
const char32_t GRANTHA_AU_LENGTH_MARK = 0x00011357;
const char32_t GRANTHA_SIGN_PLUTA = 0x0001135D;
const char32_t GRANTHA_LETTER_VEDIC_ANUSVARA = 0x0001135E;
const char32_t GRANTHA_LETTER_VEDIC_DOUBLE_ANUSVARA = 0x0001135F;
const char32_t GRANTHA_LETTER_VOCALIC_RR = 0x00011360;
const char32_t GRANTHA_LETTER_VOCALIC_LL = 0x00011361;
const char32_t GRANTHA_VOWEL_SIGN_VOCALIC_L = 0x00011362;
const char32_t GRANTHA_VOWEL_SIGN_VOCALIC_LL = 0x00011363;
const char32_t COMBINING_GRANTHA_DIGIT_ZERO = 0x00011366;
const char32_t COMBINING_GRANTHA_DIGIT_ONE = 0x00011367;
const char32_t COMBINING_GRANTHA_DIGIT_TWO = 0x00011368;
const char32_t COMBINING_GRANTHA_DIGIT_THREE = 0x00011369;
const char32_t COMBINING_GRANTHA_DIGIT_FOUR = 0x0001136A;
const char32_t COMBINING_GRANTHA_DIGIT_FIVE = 0x0001136B;
const char32_t COMBINING_GRANTHA_DIGIT_SIX = 0x0001136C;
const char32_t COMBINING_GRANTHA_LETTER_A = 0x00011370;
const char32_t COMBINING_GRANTHA_LETTER_KA = 0x00011371;
const char32_t COMBINING_GRANTHA_LETTER_NA = 0x00011372;
const char32_t COMBINING_GRANTHA_LETTER_VI = 0x00011373;
const char32_t COMBINING_GRANTHA_LETTER_PA = 0x00011374;
const char32_t TULU_TIGALARI_LETTER_A = 0x00011380;
const char32_t TULU_TIGALARI_LETTER_AA = 0x00011381;
const char32_t TULU_TIGALARI_LETTER_I = 0x00011382;
const char32_t TULU_TIGALARI_LETTER_II = 0x00011383;
const char32_t TULU_TIGALARI_LETTER_U = 0x00011384;
const char32_t TULU_TIGALARI_LETTER_UU = 0x00011385;
const char32_t TULU_TIGALARI_LETTER_VOCALIC_R = 0x00011386;
const char32_t TULU_TIGALARI_LETTER_VOCALIC_RR = 0x00011387;
const char32_t TULU_TIGALARI_LETTER_VOCALIC_L = 0x00011388;
const char32_t TULU_TIGALARI_LETTER_VOCALIC_LL = 0x00011389;
const char32_t TULU_TIGALARI_LETTER_EE = 0x0001138B;
const char32_t TULU_TIGALARI_LETTER_AI = 0x0001138E;
const char32_t TULU_TIGALARI_LETTER_OO = 0x00011390;
const char32_t TULU_TIGALARI_LETTER_AU = 0x00011391;
const char32_t TULU_TIGALARI_LETTER_KA = 0x00011392;
const char32_t TULU_TIGALARI_LETTER_KHA = 0x00011393;
const char32_t TULU_TIGALARI_LETTER_GA = 0x00011394;
const char32_t TULU_TIGALARI_LETTER_GHA = 0x00011395;
const char32_t TULU_TIGALARI_LETTER_NGA = 0x00011396;
const char32_t TULU_TIGALARI_LETTER_CA = 0x00011397;
const char32_t TULU_TIGALARI_LETTER_CHA = 0x00011398;
const char32_t TULU_TIGALARI_LETTER_JA = 0x00011399;
const char32_t TULU_TIGALARI_LETTER_JHA = 0x0001139A;
const char32_t TULU_TIGALARI_LETTER_NYA = 0x0001139B;
const char32_t TULU_TIGALARI_LETTER_TTA = 0x0001139C;
const char32_t TULU_TIGALARI_LETTER_TTHA = 0x0001139D;
const char32_t TULU_TIGALARI_LETTER_DDA = 0x0001139E;
const char32_t TULU_TIGALARI_LETTER_DDHA = 0x0001139F;
const char32_t TULU_TIGALARI_LETTER_NNA = 0x000113A0;
const char32_t TULU_TIGALARI_LETTER_TA = 0x000113A1;
const char32_t TULU_TIGALARI_LETTER_THA = 0x000113A2;
const char32_t TULU_TIGALARI_LETTER_DA = 0x000113A3;
const char32_t TULU_TIGALARI_LETTER_DHA = 0x000113A4;
const char32_t TULU_TIGALARI_LETTER_NA = 0x000113A5;
const char32_t TULU_TIGALARI_LETTER_PA = 0x000113A6;
const char32_t TULU_TIGALARI_LETTER_PHA = 0x000113A7;
const char32_t TULU_TIGALARI_LETTER_BA = 0x000113A8;
const char32_t TULU_TIGALARI_LETTER_BHA = 0x000113A9;
const char32_t TULU_TIGALARI_LETTER_MA = 0x000113AA;
const char32_t TULU_TIGALARI_LETTER_YA = 0x000113AB;
const char32_t TULU_TIGALARI_LETTER_RA = 0x000113AC;
const char32_t TULU_TIGALARI_LETTER_LA = 0x000113AD;
const char32_t TULU_TIGALARI_LETTER_VA = 0x000113AE;
const char32_t TULU_TIGALARI_LETTER_SHA = 0x000113AF;
const char32_t TULU_TIGALARI_LETTER_SSA = 0x000113B0;
const char32_t TULU_TIGALARI_LETTER_SA = 0x000113B1;
const char32_t TULU_TIGALARI_LETTER_HA = 0x000113B2;
const char32_t TULU_TIGALARI_LETTER_LLA = 0x000113B3;
const char32_t TULU_TIGALARI_LETTER_RRA = 0x000113B4;
const char32_t TULU_TIGALARI_LETTER_LLLA = 0x000113B5;
const char32_t TULU_TIGALARI_SIGN_AVAGRAHA = 0x000113B7;
const char32_t TULU_TIGALARI_VOWEL_SIGN_AA = 0x000113B8;
const char32_t TULU_TIGALARI_VOWEL_SIGN_I = 0x000113B9;
const char32_t TULU_TIGALARI_VOWEL_SIGN_II = 0x000113BA;
const char32_t TULU_TIGALARI_VOWEL_SIGN_U = 0x000113BB;
const char32_t TULU_TIGALARI_VOWEL_SIGN_UU = 0x000113BC;
const char32_t TULU_TIGALARI_VOWEL_SIGN_VOCALIC_R = 0x000113BD;
const char32_t TULU_TIGALARI_VOWEL_SIGN_VOCALIC_RR = 0x000113BE;
const char32_t TULU_TIGALARI_VOWEL_SIGN_VOCALIC_L = 0x000113BF;
const char32_t TULU_TIGALARI_VOWEL_SIGN_VOCALIC_LL = 0x000113C0;
const char32_t TULU_TIGALARI_VOWEL_SIGN_EE = 0x000113C2;
const char32_t TULU_TIGALARI_VOWEL_SIGN_AI = 0x000113C5;
const char32_t TULU_TIGALARI_VOWEL_SIGN_OO = 0x000113C7;
const char32_t TULU_TIGALARI_VOWEL_SIGN_AU = 0x000113C8;
const char32_t TULU_TIGALARI_AU_LENGTH_MARK = 0x000113C9;
const char32_t TULU_TIGALARI_SIGN_CANDRA_ANUNASIKA = 0x000113CA;
const char32_t TULU_TIGALARI_SIGN_ANUSVARA = 0x000113CC;
const char32_t TULU_TIGALARI_SIGN_VISARGA = 0x000113CD;
const char32_t TULU_TIGALARI_SIGN_VIRAMA = 0x000113CE;
const char32_t TULU_TIGALARI_SIGN_LOOPED_VIRAMA = 0x000113CF;
const char32_t TULU_TIGALARI_CONJOINER = 0x000113D0;
const char32_t TULU_TIGALARI_REPHA = 0x000113D1;
const char32_t TULU_TIGALARI_GEMINATION_MARK = 0x000113D2;
const char32_t TULU_TIGALARI_SIGN_PLUTA = 0x000113D3;
const char32_t TULU_TIGALARI_DANDA = 0x000113D4;
const char32_t TULU_TIGALARI_DOUBLE_DANDA = 0x000113D5;
const char32_t TULU_TIGALARI_SIGN_OM_PUSHPIKA = 0x000113D7;
const char32_t TULU_TIGALARI_SIGN_SHRII_PUSHPIKA = 0x000113D8;
const char32_t TULU_TIGALARI_VEDIC_TONE_SVARITA = 0x000113E1;
const char32_t TULU_TIGALARI_VEDIC_TONE_ANUDATTA = 0x000113E2;
const char32_t NEWA_LETTER_A = 0x00011400;
const char32_t NEWA_LETTER_AA = 0x00011401;
const char32_t NEWA_LETTER_I = 0x00011402;
const char32_t NEWA_LETTER_II = 0x00011403;
const char32_t NEWA_LETTER_U = 0x00011404;
const char32_t NEWA_LETTER_UU = 0x00011405;
const char32_t NEWA_LETTER_VOCALIC_R = 0x00011406;
const char32_t NEWA_LETTER_VOCALIC_RR = 0x00011407;
const char32_t NEWA_LETTER_VOCALIC_L = 0x00011408;
const char32_t NEWA_LETTER_VOCALIC_LL = 0x00011409;
const char32_t NEWA_LETTER_E = 0x0001140A;
const char32_t NEWA_LETTER_AI = 0x0001140B;
const char32_t NEWA_LETTER_O = 0x0001140C;
const char32_t NEWA_LETTER_AU = 0x0001140D;
const char32_t NEWA_LETTER_KA = 0x0001140E;
const char32_t NEWA_LETTER_KHA = 0x0001140F;
const char32_t NEWA_LETTER_GA = 0x00011410;
const char32_t NEWA_LETTER_GHA = 0x00011411;
const char32_t NEWA_LETTER_NGA = 0x00011412;
const char32_t NEWA_LETTER_NGHA = 0x00011413;
const char32_t NEWA_LETTER_CA = 0x00011414;
const char32_t NEWA_LETTER_CHA = 0x00011415;
const char32_t NEWA_LETTER_JA = 0x00011416;
const char32_t NEWA_LETTER_JHA = 0x00011417;
const char32_t NEWA_LETTER_NYA = 0x00011418;
const char32_t NEWA_LETTER_NYHA = 0x00011419;
const char32_t NEWA_LETTER_TTA = 0x0001141A;
const char32_t NEWA_LETTER_TTHA = 0x0001141B;
const char32_t NEWA_LETTER_DDA = 0x0001141C;
const char32_t NEWA_LETTER_DDHA = 0x0001141D;
const char32_t NEWA_LETTER_NNA = 0x0001141E;
const char32_t NEWA_LETTER_TA = 0x0001141F;
const char32_t NEWA_LETTER_THA = 0x00011420;
const char32_t NEWA_LETTER_DA = 0x00011421;
const char32_t NEWA_LETTER_DHA = 0x00011422;
const char32_t NEWA_LETTER_NA = 0x00011423;
const char32_t NEWA_LETTER_NHA = 0x00011424;
const char32_t NEWA_LETTER_PA = 0x00011425;
const char32_t NEWA_LETTER_PHA = 0x00011426;
const char32_t NEWA_LETTER_BA = 0x00011427;
const char32_t NEWA_LETTER_BHA = 0x00011428;
const char32_t NEWA_LETTER_MA = 0x00011429;
const char32_t NEWA_LETTER_MHA = 0x0001142A;
const char32_t NEWA_LETTER_YA = 0x0001142B;
const char32_t NEWA_LETTER_RA = 0x0001142C;
const char32_t NEWA_LETTER_RHA = 0x0001142D;
const char32_t NEWA_LETTER_LA = 0x0001142E;
const char32_t NEWA_LETTER_LHA = 0x0001142F;
const char32_t NEWA_LETTER_WA = 0x00011430;
const char32_t NEWA_LETTER_SHA = 0x00011431;
const char32_t NEWA_LETTER_SSA = 0x00011432;
const char32_t NEWA_LETTER_SA = 0x00011433;
const char32_t NEWA_LETTER_HA = 0x00011434;
const char32_t NEWA_VOWEL_SIGN_AA = 0x00011435;
const char32_t NEWA_VOWEL_SIGN_I = 0x00011436;
const char32_t NEWA_VOWEL_SIGN_II = 0x00011437;
const char32_t NEWA_VOWEL_SIGN_U = 0x00011438;
const char32_t NEWA_VOWEL_SIGN_UU = 0x00011439;
const char32_t NEWA_VOWEL_SIGN_VOCALIC_R = 0x0001143A;
const char32_t NEWA_VOWEL_SIGN_VOCALIC_RR = 0x0001143B;
const char32_t NEWA_VOWEL_SIGN_VOCALIC_L = 0x0001143C;
const char32_t NEWA_VOWEL_SIGN_VOCALIC_LL = 0x0001143D;
const char32_t NEWA_VOWEL_SIGN_E = 0x0001143E;
const char32_t NEWA_VOWEL_SIGN_AI = 0x0001143F;
const char32_t NEWA_VOWEL_SIGN_O = 0x00011440;
const char32_t NEWA_VOWEL_SIGN_AU = 0x00011441;
const char32_t NEWA_SIGN_VIRAMA = 0x00011442;
const char32_t NEWA_SIGN_CANDRABINDU = 0x00011443;
const char32_t NEWA_SIGN_ANUSVARA = 0x00011444;
const char32_t NEWA_SIGN_VISARGA = 0x00011445;
const char32_t NEWA_SIGN_NUKTA = 0x00011446;
const char32_t NEWA_SIGN_AVAGRAHA = 0x00011447;
const char32_t NEWA_SIGN_FINAL_ANUSVARA = 0x00011448;
const char32_t NEWA_OM = 0x00011449;
const char32_t NEWA_SIDDHI = 0x0001144A;
const char32_t NEWA_DANDA = 0x0001144B;
const char32_t NEWA_DOUBLE_DANDA = 0x0001144C;
const char32_t NEWA_COMMA = 0x0001144D;
const char32_t NEWA_GAP_FILLER = 0x0001144E;
const char32_t NEWA_ABBREVIATION_SIGN = 0x0001144F;
const char32_t NEWA_DIGIT_ZERO = 0x00011450;
const char32_t NEWA_DIGIT_ONE = 0x00011451;
const char32_t NEWA_DIGIT_TWO = 0x00011452;
const char32_t NEWA_DIGIT_THREE = 0x00011453;
const char32_t NEWA_DIGIT_FOUR = 0x00011454;
const char32_t NEWA_DIGIT_FIVE = 0x00011455;
const char32_t NEWA_DIGIT_SIX = 0x00011456;
const char32_t NEWA_DIGIT_SEVEN = 0x00011457;
const char32_t NEWA_DIGIT_EIGHT = 0x00011458;
const char32_t NEWA_DIGIT_NINE = 0x00011459;
const char32_t NEWA_DOUBLE_COMMA = 0x0001145A;
const char32_t NEWA_PLACEHOLDER_MARK = 0x0001145B;
const char32_t NEWA_INSERTION_SIGN = 0x0001145D;
const char32_t NEWA_SANDHI_MARK = 0x0001145E;
const char32_t NEWA_LETTER_VEDIC_ANUSVARA = 0x0001145F;
const char32_t NEWA_SIGN_JIHVAMULIYA = 0x00011460;
const char32_t NEWA_SIGN_UPADHMANIYA = 0x00011461;
const char32_t TIRHUTA_ANJI = 0x00011480;
const char32_t TIRHUTA_LETTER_A = 0x00011481;
const char32_t TIRHUTA_LETTER_AA = 0x00011482;
const char32_t TIRHUTA_LETTER_I = 0x00011483;
const char32_t TIRHUTA_LETTER_II = 0x00011484;
const char32_t TIRHUTA_LETTER_U = 0x00011485;
const char32_t TIRHUTA_LETTER_UU = 0x00011486;
const char32_t TIRHUTA_LETTER_VOCALIC_R = 0x00011487;
const char32_t TIRHUTA_LETTER_VOCALIC_RR = 0x00011488;
const char32_t TIRHUTA_LETTER_VOCALIC_L = 0x00011489;
const char32_t TIRHUTA_LETTER_VOCALIC_LL = 0x0001148A;
const char32_t TIRHUTA_LETTER_E = 0x0001148B;
const char32_t TIRHUTA_LETTER_AI = 0x0001148C;
const char32_t TIRHUTA_LETTER_O = 0x0001148D;
const char32_t TIRHUTA_LETTER_AU = 0x0001148E;
const char32_t TIRHUTA_LETTER_KA = 0x0001148F;
const char32_t TIRHUTA_LETTER_KHA = 0x00011490;
const char32_t TIRHUTA_LETTER_GA = 0x00011491;
const char32_t TIRHUTA_LETTER_GHA = 0x00011492;
const char32_t TIRHUTA_LETTER_NGA = 0x00011493;
const char32_t TIRHUTA_LETTER_CA = 0x00011494;
const char32_t TIRHUTA_LETTER_CHA = 0x00011495;
const char32_t TIRHUTA_LETTER_JA = 0x00011496;
const char32_t TIRHUTA_LETTER_JHA = 0x00011497;
const char32_t TIRHUTA_LETTER_NYA = 0x00011498;
const char32_t TIRHUTA_LETTER_TTA = 0x00011499;
const char32_t TIRHUTA_LETTER_TTHA = 0x0001149A;
const char32_t TIRHUTA_LETTER_DDA = 0x0001149B;
const char32_t TIRHUTA_LETTER_DDHA = 0x0001149C;
const char32_t TIRHUTA_LETTER_NNA = 0x0001149D;
const char32_t TIRHUTA_LETTER_TA = 0x0001149E;
const char32_t TIRHUTA_LETTER_THA = 0x0001149F;
const char32_t TIRHUTA_LETTER_DA = 0x000114A0;
const char32_t TIRHUTA_LETTER_DHA = 0x000114A1;
const char32_t TIRHUTA_LETTER_NA = 0x000114A2;
const char32_t TIRHUTA_LETTER_PA = 0x000114A3;
const char32_t TIRHUTA_LETTER_PHA = 0x000114A4;
const char32_t TIRHUTA_LETTER_BA = 0x000114A5;
const char32_t TIRHUTA_LETTER_BHA = 0x000114A6;
const char32_t TIRHUTA_LETTER_MA = 0x000114A7;
const char32_t TIRHUTA_LETTER_YA = 0x000114A8;
const char32_t TIRHUTA_LETTER_RA = 0x000114A9;
const char32_t TIRHUTA_LETTER_LA = 0x000114AA;
const char32_t TIRHUTA_LETTER_VA = 0x000114AB;
const char32_t TIRHUTA_LETTER_SHA = 0x000114AC;
const char32_t TIRHUTA_LETTER_SSA = 0x000114AD;
const char32_t TIRHUTA_LETTER_SA = 0x000114AE;
const char32_t TIRHUTA_LETTER_HA = 0x000114AF;
const char32_t TIRHUTA_VOWEL_SIGN_AA = 0x000114B0;
const char32_t TIRHUTA_VOWEL_SIGN_I = 0x000114B1;
const char32_t TIRHUTA_VOWEL_SIGN_II = 0x000114B2;
const char32_t TIRHUTA_VOWEL_SIGN_U = 0x000114B3;
const char32_t TIRHUTA_VOWEL_SIGN_UU = 0x000114B4;
const char32_t TIRHUTA_VOWEL_SIGN_VOCALIC_R = 0x000114B5;
const char32_t TIRHUTA_VOWEL_SIGN_VOCALIC_RR = 0x000114B6;
const char32_t TIRHUTA_VOWEL_SIGN_VOCALIC_L = 0x000114B7;
const char32_t TIRHUTA_VOWEL_SIGN_VOCALIC_LL = 0x000114B8;
const char32_t TIRHUTA_VOWEL_SIGN_E = 0x000114B9;
const char32_t TIRHUTA_VOWEL_SIGN_SHORT_E = 0x000114BA;
const char32_t TIRHUTA_VOWEL_SIGN_AI = 0x000114BB;
const char32_t TIRHUTA_VOWEL_SIGN_O = 0x000114BC;
const char32_t TIRHUTA_VOWEL_SIGN_SHORT_O = 0x000114BD;
const char32_t TIRHUTA_VOWEL_SIGN_AU = 0x000114BE;
const char32_t TIRHUTA_SIGN_CANDRABINDU = 0x000114BF;
const char32_t TIRHUTA_SIGN_ANUSVARA = 0x000114C0;
const char32_t TIRHUTA_SIGN_VISARGA = 0x000114C1;
const char32_t TIRHUTA_SIGN_VIRAMA = 0x000114C2;
const char32_t TIRHUTA_SIGN_NUKTA = 0x000114C3;
const char32_t TIRHUTA_SIGN_AVAGRAHA = 0x000114C4;
const char32_t TIRHUTA_GVANG = 0x000114C5;
const char32_t TIRHUTA_ABBREVIATION_SIGN = 0x000114C6;
const char32_t TIRHUTA_OM = 0x000114C7;
const char32_t TIRHUTA_DIGIT_ZERO = 0x000114D0;
const char32_t TIRHUTA_DIGIT_ONE = 0x000114D1;
const char32_t TIRHUTA_DIGIT_TWO = 0x000114D2;
const char32_t TIRHUTA_DIGIT_THREE = 0x000114D3;
const char32_t TIRHUTA_DIGIT_FOUR = 0x000114D4;
const char32_t TIRHUTA_DIGIT_FIVE = 0x000114D5;
const char32_t TIRHUTA_DIGIT_SIX = 0x000114D6;
const char32_t TIRHUTA_DIGIT_SEVEN = 0x000114D7;
const char32_t TIRHUTA_DIGIT_EIGHT = 0x000114D8;
const char32_t TIRHUTA_DIGIT_NINE = 0x000114D9;
const char32_t SIDDHAM_LETTER_A = 0x00011580;
const char32_t SIDDHAM_LETTER_AA = 0x00011581;
const char32_t SIDDHAM_LETTER_I = 0x00011582;
const char32_t SIDDHAM_LETTER_II = 0x00011583;
const char32_t SIDDHAM_LETTER_U = 0x00011584;
const char32_t SIDDHAM_LETTER_UU = 0x00011585;
const char32_t SIDDHAM_LETTER_VOCALIC_R = 0x00011586;
const char32_t SIDDHAM_LETTER_VOCALIC_RR = 0x00011587;
const char32_t SIDDHAM_LETTER_VOCALIC_L = 0x00011588;
const char32_t SIDDHAM_LETTER_VOCALIC_LL = 0x00011589;
const char32_t SIDDHAM_LETTER_E = 0x0001158A;
const char32_t SIDDHAM_LETTER_AI = 0x0001158B;
const char32_t SIDDHAM_LETTER_O = 0x0001158C;
const char32_t SIDDHAM_LETTER_AU = 0x0001158D;
const char32_t SIDDHAM_LETTER_KA = 0x0001158E;
const char32_t SIDDHAM_LETTER_KHA = 0x0001158F;
const char32_t SIDDHAM_LETTER_GA = 0x00011590;
const char32_t SIDDHAM_LETTER_GHA = 0x00011591;
const char32_t SIDDHAM_LETTER_NGA = 0x00011592;
const char32_t SIDDHAM_LETTER_CA = 0x00011593;
const char32_t SIDDHAM_LETTER_CHA = 0x00011594;
const char32_t SIDDHAM_LETTER_JA = 0x00011595;
const char32_t SIDDHAM_LETTER_JHA = 0x00011596;
const char32_t SIDDHAM_LETTER_NYA = 0x00011597;
const char32_t SIDDHAM_LETTER_TTA = 0x00011598;
const char32_t SIDDHAM_LETTER_TTHA = 0x00011599;
const char32_t SIDDHAM_LETTER_DDA = 0x0001159A;
const char32_t SIDDHAM_LETTER_DDHA = 0x0001159B;
const char32_t SIDDHAM_LETTER_NNA = 0x0001159C;
const char32_t SIDDHAM_LETTER_TA = 0x0001159D;
const char32_t SIDDHAM_LETTER_THA = 0x0001159E;
const char32_t SIDDHAM_LETTER_DA = 0x0001159F;
const char32_t SIDDHAM_LETTER_DHA = 0x000115A0;
const char32_t SIDDHAM_LETTER_NA = 0x000115A1;
const char32_t SIDDHAM_LETTER_PA = 0x000115A2;
const char32_t SIDDHAM_LETTER_PHA = 0x000115A3;
const char32_t SIDDHAM_LETTER_BA = 0x000115A4;
const char32_t SIDDHAM_LETTER_BHA = 0x000115A5;
const char32_t SIDDHAM_LETTER_MA = 0x000115A6;
const char32_t SIDDHAM_LETTER_YA = 0x000115A7;
const char32_t SIDDHAM_LETTER_RA = 0x000115A8;
const char32_t SIDDHAM_LETTER_LA = 0x000115A9;
const char32_t SIDDHAM_LETTER_VA = 0x000115AA;
const char32_t SIDDHAM_LETTER_SHA = 0x000115AB;
const char32_t SIDDHAM_LETTER_SSA = 0x000115AC;
const char32_t SIDDHAM_LETTER_SA = 0x000115AD;
const char32_t SIDDHAM_LETTER_HA = 0x000115AE;
const char32_t SIDDHAM_VOWEL_SIGN_AA = 0x000115AF;
const char32_t SIDDHAM_VOWEL_SIGN_I = 0x000115B0;
const char32_t SIDDHAM_VOWEL_SIGN_II = 0x000115B1;
const char32_t SIDDHAM_VOWEL_SIGN_U = 0x000115B2;
const char32_t SIDDHAM_VOWEL_SIGN_UU = 0x000115B3;
const char32_t SIDDHAM_VOWEL_SIGN_VOCALIC_R = 0x000115B4;
const char32_t SIDDHAM_VOWEL_SIGN_VOCALIC_RR = 0x000115B5;
const char32_t SIDDHAM_VOWEL_SIGN_E = 0x000115B8;
const char32_t SIDDHAM_VOWEL_SIGN_AI = 0x000115B9;
const char32_t SIDDHAM_VOWEL_SIGN_O = 0x000115BA;
const char32_t SIDDHAM_VOWEL_SIGN_AU = 0x000115BB;
const char32_t SIDDHAM_SIGN_CANDRABINDU = 0x000115BC;
const char32_t SIDDHAM_SIGN_ANUSVARA = 0x000115BD;
const char32_t SIDDHAM_SIGN_VISARGA = 0x000115BE;
const char32_t SIDDHAM_SIGN_VIRAMA = 0x000115BF;
const char32_t SIDDHAM_SIGN_NUKTA = 0x000115C0;
const char32_t SIDDHAM_SIGN_SIDDHAM = 0x000115C1;
const char32_t SIDDHAM_DANDA = 0x000115C2;
const char32_t SIDDHAM_DOUBLE_DANDA = 0x000115C3;
const char32_t SIDDHAM_SEPARATOR_DOT = 0x000115C4;
const char32_t SIDDHAM_SEPARATOR_BAR = 0x000115C5;
const char32_t SIDDHAM_REPETITION_MARK_1 = 0x000115C6;
const char32_t SIDDHAM_REPETITION_MARK_2 = 0x000115C7;
const char32_t SIDDHAM_REPETITION_MARK_3 = 0x000115C8;
const char32_t SIDDHAM_END_OF_TEXT_MARK = 0x000115C9;
const char32_t SIDDHAM_SECTION_MARK_WITH_TRIDENT_AND_U_SHAPED_ORNAMENTS = 0x000115CA;
const char32_t SIDDHAM_SECTION_MARK_WITH_TRIDENT_AND_DOTTED_CRESCENTS = 0x000115CB;
const char32_t SIDDHAM_SECTION_MARK_WITH_RAYS_AND_DOTTED_CRESCENTS = 0x000115CC;
const char32_t SIDDHAM_SECTION_MARK_WITH_RAYS_AND_DOTTED_DOUBLE_CRESCENTS = 0x000115CD;
const char32_t SIDDHAM_SECTION_MARK_WITH_RAYS_AND_DOTTED_TRIPLE_CRESCENTS = 0x000115CE;
const char32_t SIDDHAM_SECTION_MARK_DOUBLE_RING = 0x000115CF;
const char32_t SIDDHAM_SECTION_MARK_DOUBLE_RING_WITH_RAYS = 0x000115D0;
const char32_t SIDDHAM_SECTION_MARK_WITH_DOUBLE_CRESCENTS = 0x000115D1;
const char32_t SIDDHAM_SECTION_MARK_WITH_TRIPLE_CRESCENTS = 0x000115D2;
const char32_t SIDDHAM_SECTION_MARK_WITH_QUADRUPLE_CRESCENTS = 0x000115D3;
const char32_t SIDDHAM_SECTION_MARK_WITH_SEPTUPLE_CRESCENTS = 0x000115D4;
const char32_t SIDDHAM_SECTION_MARK_WITH_CIRCLES_AND_RAYS = 0x000115D5;
const char32_t SIDDHAM_SECTION_MARK_WITH_CIRCLES_AND_TWO_ENCLOSURES = 0x000115D6;
const char32_t SIDDHAM_SECTION_MARK_WITH_CIRCLES_AND_FOUR_ENCLOSURES = 0x000115D7;
const char32_t SIDDHAM_LETTER_THREE_CIRCLE_ALTERNATE_I = 0x000115D8;
const char32_t SIDDHAM_LETTER_TWO_CIRCLE_ALTERNATE_I = 0x000115D9;
const char32_t SIDDHAM_LETTER_TWO_CIRCLE_ALTERNATE_II = 0x000115DA;
const char32_t SIDDHAM_LETTER_ALTERNATE_U = 0x000115DB;
const char32_t SIDDHAM_VOWEL_SIGN_ALTERNATE_U = 0x000115DC;
const char32_t SIDDHAM_VOWEL_SIGN_ALTERNATE_UU = 0x000115DD;
const char32_t MODI_LETTER_A = 0x00011600;
const char32_t MODI_LETTER_AA = 0x00011601;
const char32_t MODI_LETTER_I = 0x00011602;
const char32_t MODI_LETTER_II = 0x00011603;
const char32_t MODI_LETTER_U = 0x00011604;
const char32_t MODI_LETTER_UU = 0x00011605;
const char32_t MODI_LETTER_VOCALIC_R = 0x00011606;
const char32_t MODI_LETTER_VOCALIC_RR = 0x00011607;
const char32_t MODI_LETTER_VOCALIC_L = 0x00011608;
const char32_t MODI_LETTER_VOCALIC_LL = 0x00011609;
const char32_t MODI_LETTER_E = 0x0001160A;
const char32_t MODI_LETTER_AI = 0x0001160B;
const char32_t MODI_LETTER_O = 0x0001160C;
const char32_t MODI_LETTER_AU = 0x0001160D;
const char32_t MODI_LETTER_KA = 0x0001160E;
const char32_t MODI_LETTER_KHA = 0x0001160F;
const char32_t MODI_LETTER_GA = 0x00011610;
const char32_t MODI_LETTER_GHA = 0x00011611;
const char32_t MODI_LETTER_NGA = 0x00011612;
const char32_t MODI_LETTER_CA = 0x00011613;
const char32_t MODI_LETTER_CHA = 0x00011614;
const char32_t MODI_LETTER_JA = 0x00011615;
const char32_t MODI_LETTER_JHA = 0x00011616;
const char32_t MODI_LETTER_NYA = 0x00011617;
const char32_t MODI_LETTER_TTA = 0x00011618;
const char32_t MODI_LETTER_TTHA = 0x00011619;
const char32_t MODI_LETTER_DDA = 0x0001161A;
const char32_t MODI_LETTER_DDHA = 0x0001161B;
const char32_t MODI_LETTER_NNA = 0x0001161C;
const char32_t MODI_LETTER_TA = 0x0001161D;
const char32_t MODI_LETTER_THA = 0x0001161E;
const char32_t MODI_LETTER_DA = 0x0001161F;
const char32_t MODI_LETTER_DHA = 0x00011620;
const char32_t MODI_LETTER_NA = 0x00011621;
const char32_t MODI_LETTER_PA = 0x00011622;
const char32_t MODI_LETTER_PHA = 0x00011623;
const char32_t MODI_LETTER_BA = 0x00011624;
const char32_t MODI_LETTER_BHA = 0x00011625;
const char32_t MODI_LETTER_MA = 0x00011626;
const char32_t MODI_LETTER_YA = 0x00011627;
const char32_t MODI_LETTER_RA = 0x00011628;
const char32_t MODI_LETTER_LA = 0x00011629;
const char32_t MODI_LETTER_VA = 0x0001162A;
const char32_t MODI_LETTER_SHA = 0x0001162B;
const char32_t MODI_LETTER_SSA = 0x0001162C;
const char32_t MODI_LETTER_SA = 0x0001162D;
const char32_t MODI_LETTER_HA = 0x0001162E;
const char32_t MODI_LETTER_LLA = 0x0001162F;
const char32_t MODI_VOWEL_SIGN_AA = 0x00011630;
const char32_t MODI_VOWEL_SIGN_I = 0x00011631;
const char32_t MODI_VOWEL_SIGN_II = 0x00011632;
const char32_t MODI_VOWEL_SIGN_U = 0x00011633;
const char32_t MODI_VOWEL_SIGN_UU = 0x00011634;
const char32_t MODI_VOWEL_SIGN_VOCALIC_R = 0x00011635;
const char32_t MODI_VOWEL_SIGN_VOCALIC_RR = 0x00011636;
const char32_t MODI_VOWEL_SIGN_VOCALIC_L = 0x00011637;
const char32_t MODI_VOWEL_SIGN_VOCALIC_LL = 0x00011638;
const char32_t MODI_VOWEL_SIGN_E = 0x00011639;
const char32_t MODI_VOWEL_SIGN_AI = 0x0001163A;
const char32_t MODI_VOWEL_SIGN_O = 0x0001163B;
const char32_t MODI_VOWEL_SIGN_AU = 0x0001163C;
const char32_t MODI_SIGN_ANUSVARA = 0x0001163D;
const char32_t MODI_SIGN_VISARGA = 0x0001163E;
const char32_t MODI_SIGN_VIRAMA = 0x0001163F;
const char32_t MODI_SIGN_ARDHACANDRA = 0x00011640;
const char32_t MODI_DANDA = 0x00011641;
const char32_t MODI_DOUBLE_DANDA = 0x00011642;
const char32_t MODI_ABBREVIATION_SIGN = 0x00011643;
const char32_t MODI_SIGN_HUVA = 0x00011644;
const char32_t MODI_DIGIT_ZERO = 0x00011650;
const char32_t MODI_DIGIT_ONE = 0x00011651;
const char32_t MODI_DIGIT_TWO = 0x00011652;
const char32_t MODI_DIGIT_THREE = 0x00011653;
const char32_t MODI_DIGIT_FOUR = 0x00011654;
const char32_t MODI_DIGIT_FIVE = 0x00011655;
const char32_t MODI_DIGIT_SIX = 0x00011656;
const char32_t MODI_DIGIT_SEVEN = 0x00011657;
const char32_t MODI_DIGIT_EIGHT = 0x00011658;
const char32_t MODI_DIGIT_NINE = 0x00011659;
const char32_t MONGOLIAN_BIRGA_WITH_ORNAMENT = 0x00011660;
const char32_t MONGOLIAN_ROTATED_BIRGA = 0x00011661;
const char32_t MONGOLIAN_DOUBLE_BIRGA_WITH_ORNAMENT = 0x00011662;
const char32_t MONGOLIAN_TRIPLE_BIRGA_WITH_ORNAMENT = 0x00011663;
const char32_t MONGOLIAN_BIRGA_WITH_DOUBLE_ORNAMENT = 0x00011664;
const char32_t MONGOLIAN_ROTATED_BIRGA_WITH_ORNAMENT = 0x00011665;
const char32_t MONGOLIAN_ROTATED_BIRGA_WITH_DOUBLE_ORNAMENT = 0x00011666;
const char32_t MONGOLIAN_INVERTED_BIRGA = 0x00011667;
const char32_t MONGOLIAN_INVERTED_BIRGA_WITH_DOUBLE_ORNAMENT = 0x00011668;
const char32_t MONGOLIAN_SWIRL_BIRGA = 0x00011669;
const char32_t MONGOLIAN_SWIRL_BIRGA_WITH_ORNAMENT = 0x0001166A;
const char32_t MONGOLIAN_SWIRL_BIRGA_WITH_DOUBLE_ORNAMENT = 0x0001166B;
const char32_t MONGOLIAN_TURNED_SWIRL_BIRGA_WITH_DOUBLE_ORNAMENT = 0x0001166C;
const char32_t TAKRI_LETTER_A = 0x00011680;
const char32_t TAKRI_LETTER_AA = 0x00011681;
const char32_t TAKRI_LETTER_I = 0x00011682;
const char32_t TAKRI_LETTER_II = 0x00011683;
const char32_t TAKRI_LETTER_U = 0x00011684;
const char32_t TAKRI_LETTER_UU = 0x00011685;
const char32_t TAKRI_LETTER_E = 0x00011686;
const char32_t TAKRI_LETTER_AI = 0x00011687;
const char32_t TAKRI_LETTER_O = 0x00011688;
const char32_t TAKRI_LETTER_AU = 0x00011689;
const char32_t TAKRI_LETTER_KA = 0x0001168A;
const char32_t TAKRI_LETTER_KHA = 0x0001168B;
const char32_t TAKRI_LETTER_GA = 0x0001168C;
const char32_t TAKRI_LETTER_GHA = 0x0001168D;
const char32_t TAKRI_LETTER_NGA = 0x0001168E;
const char32_t TAKRI_LETTER_CA = 0x0001168F;
const char32_t TAKRI_LETTER_CHA = 0x00011690;
const char32_t TAKRI_LETTER_JA = 0x00011691;
const char32_t TAKRI_LETTER_JHA = 0x00011692;
const char32_t TAKRI_LETTER_NYA = 0x00011693;
const char32_t TAKRI_LETTER_TTA = 0x00011694;
const char32_t TAKRI_LETTER_TTHA = 0x00011695;
const char32_t TAKRI_LETTER_DDA = 0x00011696;
const char32_t TAKRI_LETTER_DDHA = 0x00011697;
const char32_t TAKRI_LETTER_NNA = 0x00011698;
const char32_t TAKRI_LETTER_TA = 0x00011699;
const char32_t TAKRI_LETTER_THA = 0x0001169A;
const char32_t TAKRI_LETTER_DA = 0x0001169B;
const char32_t TAKRI_LETTER_DHA = 0x0001169C;
const char32_t TAKRI_LETTER_NA = 0x0001169D;
const char32_t TAKRI_LETTER_PA = 0x0001169E;
const char32_t TAKRI_LETTER_PHA = 0x0001169F;
const char32_t TAKRI_LETTER_BA = 0x000116A0;
const char32_t TAKRI_LETTER_BHA = 0x000116A1;
const char32_t TAKRI_LETTER_MA = 0x000116A2;
const char32_t TAKRI_LETTER_YA = 0x000116A3;
const char32_t TAKRI_LETTER_RA = 0x000116A4;
const char32_t TAKRI_LETTER_LA = 0x000116A5;
const char32_t TAKRI_LETTER_VA = 0x000116A6;
const char32_t TAKRI_LETTER_SHA = 0x000116A7;
const char32_t TAKRI_LETTER_SA = 0x000116A8;
const char32_t TAKRI_LETTER_HA = 0x000116A9;
const char32_t TAKRI_LETTER_RRA = 0x000116AA;
const char32_t TAKRI_SIGN_ANUSVARA = 0x000116AB;
const char32_t TAKRI_SIGN_VISARGA = 0x000116AC;
const char32_t TAKRI_VOWEL_SIGN_AA = 0x000116AD;
const char32_t TAKRI_VOWEL_SIGN_I = 0x000116AE;
const char32_t TAKRI_VOWEL_SIGN_II = 0x000116AF;
const char32_t TAKRI_VOWEL_SIGN_U = 0x000116B0;
const char32_t TAKRI_VOWEL_SIGN_UU = 0x000116B1;
const char32_t TAKRI_VOWEL_SIGN_E = 0x000116B2;
const char32_t TAKRI_VOWEL_SIGN_AI = 0x000116B3;
const char32_t TAKRI_VOWEL_SIGN_O = 0x000116B4;
const char32_t TAKRI_VOWEL_SIGN_AU = 0x000116B5;
const char32_t TAKRI_SIGN_VIRAMA = 0x000116B6;
const char32_t TAKRI_SIGN_NUKTA = 0x000116B7;
const char32_t TAKRI_LETTER_ARCHAIC_KHA = 0x000116B8;
const char32_t TAKRI_ABBREVIATION_SIGN = 0x000116B9;
const char32_t TAKRI_DIGIT_ZERO = 0x000116C0;
const char32_t TAKRI_DIGIT_ONE = 0x000116C1;
const char32_t TAKRI_DIGIT_TWO = 0x000116C2;
const char32_t TAKRI_DIGIT_THREE = 0x000116C3;
const char32_t TAKRI_DIGIT_FOUR = 0x000116C4;
const char32_t TAKRI_DIGIT_FIVE = 0x000116C5;
const char32_t TAKRI_DIGIT_SIX = 0x000116C6;
const char32_t TAKRI_DIGIT_SEVEN = 0x000116C7;
const char32_t TAKRI_DIGIT_EIGHT = 0x000116C8;
const char32_t TAKRI_DIGIT_NINE = 0x000116C9;
const char32_t MYANMAR_PAO_DIGIT_ZERO = 0x000116D0;
const char32_t MYANMAR_PAO_DIGIT_ONE = 0x000116D1;
const char32_t MYANMAR_PAO_DIGIT_TWO = 0x000116D2;
const char32_t MYANMAR_PAO_DIGIT_THREE = 0x000116D3;
const char32_t MYANMAR_PAO_DIGIT_FOUR = 0x000116D4;
const char32_t MYANMAR_PAO_DIGIT_FIVE = 0x000116D5;
const char32_t MYANMAR_PAO_DIGIT_SIX = 0x000116D6;
const char32_t MYANMAR_PAO_DIGIT_SEVEN = 0x000116D7;
const char32_t MYANMAR_PAO_DIGIT_EIGHT = 0x000116D8;
const char32_t MYANMAR_PAO_DIGIT_NINE = 0x000116D9;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_ZERO = 0x000116DA;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_ONE = 0x000116DB;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_TWO = 0x000116DC;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_THREE = 0x000116DD;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_FOUR = 0x000116DE;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_FIVE = 0x000116DF;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_SIX = 0x000116E0;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_SEVEN = 0x000116E1;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_EIGHT = 0x000116E2;
const char32_t MYANMAR_EASTERN_PWO_KAREN_DIGIT_NINE = 0x000116E3;
const char32_t AHOM_LETTER_KA = 0x00011700;
const char32_t AHOM_LETTER_KHA = 0x00011701;
const char32_t AHOM_LETTER_NGA = 0x00011702;
const char32_t AHOM_LETTER_NA = 0x00011703;
const char32_t AHOM_LETTER_TA = 0x00011704;
const char32_t AHOM_LETTER_ALTERNATE_TA = 0x00011705;
const char32_t AHOM_LETTER_PA = 0x00011706;
const char32_t AHOM_LETTER_PHA = 0x00011707;
const char32_t AHOM_LETTER_BA = 0x00011708;
const char32_t AHOM_LETTER_MA = 0x00011709;
const char32_t AHOM_LETTER_JA = 0x0001170A;
const char32_t AHOM_LETTER_CHA = 0x0001170B;
const char32_t AHOM_LETTER_THA = 0x0001170C;
const char32_t AHOM_LETTER_RA = 0x0001170D;
const char32_t AHOM_LETTER_LA = 0x0001170E;
const char32_t AHOM_LETTER_SA = 0x0001170F;
const char32_t AHOM_LETTER_NYA = 0x00011710;
const char32_t AHOM_LETTER_HA = 0x00011711;
const char32_t AHOM_LETTER_A = 0x00011712;
const char32_t AHOM_LETTER_DA = 0x00011713;
const char32_t AHOM_LETTER_DHA = 0x00011714;
const char32_t AHOM_LETTER_GA = 0x00011715;
const char32_t AHOM_LETTER_ALTERNATE_GA = 0x00011716;
const char32_t AHOM_LETTER_GHA = 0x00011717;
const char32_t AHOM_LETTER_BHA = 0x00011718;
const char32_t AHOM_LETTER_JHA = 0x00011719;
const char32_t AHOM_LETTER_ALTERNATE_BA = 0x0001171A;
const char32_t AHOM_CONSONANT_SIGN_MEDIAL_LA = 0x0001171D;
const char32_t AHOM_CONSONANT_SIGN_MEDIAL_RA = 0x0001171E;
const char32_t AHOM_CONSONANT_SIGN_MEDIAL_LIGATING_RA = 0x0001171F;
const char32_t AHOM_VOWEL_SIGN_A = 0x00011720;
const char32_t AHOM_VOWEL_SIGN_AA = 0x00011721;
const char32_t AHOM_VOWEL_SIGN_I = 0x00011722;
const char32_t AHOM_VOWEL_SIGN_II = 0x00011723;
const char32_t AHOM_VOWEL_SIGN_U = 0x00011724;
const char32_t AHOM_VOWEL_SIGN_UU = 0x00011725;
const char32_t AHOM_VOWEL_SIGN_E = 0x00011726;
const char32_t AHOM_VOWEL_SIGN_AW = 0x00011727;
const char32_t AHOM_VOWEL_SIGN_O = 0x00011728;
const char32_t AHOM_VOWEL_SIGN_AI = 0x00011729;
const char32_t AHOM_VOWEL_SIGN_AM = 0x0001172A;
const char32_t AHOM_SIGN_KILLER = 0x0001172B;
const char32_t AHOM_DIGIT_ZERO = 0x00011730;
const char32_t AHOM_DIGIT_ONE = 0x00011731;
const char32_t AHOM_DIGIT_TWO = 0x00011732;
const char32_t AHOM_DIGIT_THREE = 0x00011733;
const char32_t AHOM_DIGIT_FOUR = 0x00011734;
const char32_t AHOM_DIGIT_FIVE = 0x00011735;
const char32_t AHOM_DIGIT_SIX = 0x00011736;
const char32_t AHOM_DIGIT_SEVEN = 0x00011737;
const char32_t AHOM_DIGIT_EIGHT = 0x00011738;
const char32_t AHOM_DIGIT_NINE = 0x00011739;
const char32_t AHOM_NUMBER_TEN = 0x0001173A;
const char32_t AHOM_NUMBER_TWENTY = 0x0001173B;
const char32_t AHOM_SIGN_SMALL_SECTION = 0x0001173C;
const char32_t AHOM_SIGN_SECTION = 0x0001173D;
const char32_t AHOM_SIGN_RULAI = 0x0001173E;
const char32_t AHOM_SYMBOL_VI = 0x0001173F;
const char32_t AHOM_LETTER_CA = 0x00011740;
const char32_t AHOM_LETTER_TTA = 0x00011741;
const char32_t AHOM_LETTER_TTHA = 0x00011742;
const char32_t AHOM_LETTER_DDA = 0x00011743;
const char32_t AHOM_LETTER_DDHA = 0x00011744;
const char32_t AHOM_LETTER_NNA = 0x00011745;
const char32_t AHOM_LETTER_LLA = 0x00011746;
const char32_t DOGRA_LETTER_A = 0x00011800;
const char32_t DOGRA_LETTER_AA = 0x00011801;
const char32_t DOGRA_LETTER_I = 0x00011802;
const char32_t DOGRA_LETTER_II = 0x00011803;
const char32_t DOGRA_LETTER_U = 0x00011804;
const char32_t DOGRA_LETTER_UU = 0x00011805;
const char32_t DOGRA_LETTER_E = 0x00011806;
const char32_t DOGRA_LETTER_AI = 0x00011807;
const char32_t DOGRA_LETTER_O = 0x00011808;
const char32_t DOGRA_LETTER_AU = 0x00011809;
const char32_t DOGRA_LETTER_KA = 0x0001180A;
const char32_t DOGRA_LETTER_KHA = 0x0001180B;
const char32_t DOGRA_LETTER_GA = 0x0001180C;
const char32_t DOGRA_LETTER_GHA = 0x0001180D;
const char32_t DOGRA_LETTER_NGA = 0x0001180E;
const char32_t DOGRA_LETTER_CA = 0x0001180F;
const char32_t DOGRA_LETTER_CHA = 0x00011810;
const char32_t DOGRA_LETTER_JA = 0x00011811;
const char32_t DOGRA_LETTER_JHA = 0x00011812;
const char32_t DOGRA_LETTER_NYA = 0x00011813;
const char32_t DOGRA_LETTER_TTA = 0x00011814;
const char32_t DOGRA_LETTER_TTHA = 0x00011815;
const char32_t DOGRA_LETTER_DDA = 0x00011816;
const char32_t DOGRA_LETTER_DDHA = 0x00011817;
const char32_t DOGRA_LETTER_NNA = 0x00011818;
const char32_t DOGRA_LETTER_TA = 0x00011819;
const char32_t DOGRA_LETTER_THA = 0x0001181A;
const char32_t DOGRA_LETTER_DA = 0x0001181B;
const char32_t DOGRA_LETTER_DHA = 0x0001181C;
const char32_t DOGRA_LETTER_NA = 0x0001181D;
const char32_t DOGRA_LETTER_PA = 0x0001181E;
const char32_t DOGRA_LETTER_PHA = 0x0001181F;
const char32_t DOGRA_LETTER_BA = 0x00011820;
const char32_t DOGRA_LETTER_BHA = 0x00011821;
const char32_t DOGRA_LETTER_MA = 0x00011822;
const char32_t DOGRA_LETTER_YA = 0x00011823;
const char32_t DOGRA_LETTER_RA = 0x00011824;
const char32_t DOGRA_LETTER_LA = 0x00011825;
const char32_t DOGRA_LETTER_VA = 0x00011826;
const char32_t DOGRA_LETTER_SHA = 0x00011827;
const char32_t DOGRA_LETTER_SSA = 0x00011828;
const char32_t DOGRA_LETTER_SA = 0x00011829;
const char32_t DOGRA_LETTER_HA = 0x0001182A;
const char32_t DOGRA_LETTER_RRA = 0x0001182B;
const char32_t DOGRA_VOWEL_SIGN_AA = 0x0001182C;
const char32_t DOGRA_VOWEL_SIGN_I = 0x0001182D;
const char32_t DOGRA_VOWEL_SIGN_II = 0x0001182E;
const char32_t DOGRA_VOWEL_SIGN_U = 0x0001182F;
const char32_t DOGRA_VOWEL_SIGN_UU = 0x00011830;
const char32_t DOGRA_VOWEL_SIGN_VOCALIC_R = 0x00011831;
const char32_t DOGRA_VOWEL_SIGN_VOCALIC_RR = 0x00011832;
const char32_t DOGRA_VOWEL_SIGN_E = 0x00011833;
const char32_t DOGRA_VOWEL_SIGN_AI = 0x00011834;
const char32_t DOGRA_VOWEL_SIGN_O = 0x00011835;
const char32_t DOGRA_VOWEL_SIGN_AU = 0x00011836;
const char32_t DOGRA_SIGN_ANUSVARA = 0x00011837;
const char32_t DOGRA_SIGN_VISARGA = 0x00011838;
const char32_t DOGRA_SIGN_VIRAMA = 0x00011839;
const char32_t DOGRA_SIGN_NUKTA = 0x0001183A;
const char32_t DOGRA_ABBREVIATION_SIGN = 0x0001183B;
const char32_t WARANG_CITI_CAPITAL_LETTER_NGAA = 0x000118A0;
const char32_t WARANG_CITI_CAPITAL_LETTER_A = 0x000118A1;
const char32_t WARANG_CITI_CAPITAL_LETTER_WI = 0x000118A2;
const char32_t WARANG_CITI_CAPITAL_LETTER_YU = 0x000118A3;
const char32_t WARANG_CITI_CAPITAL_LETTER_YA = 0x000118A4;
const char32_t WARANG_CITI_CAPITAL_LETTER_YO = 0x000118A5;
const char32_t WARANG_CITI_CAPITAL_LETTER_II = 0x000118A6;
const char32_t WARANG_CITI_CAPITAL_LETTER_UU = 0x000118A7;
const char32_t WARANG_CITI_CAPITAL_LETTER_E = 0x000118A8;
const char32_t WARANG_CITI_CAPITAL_LETTER_O = 0x000118A9;
const char32_t WARANG_CITI_CAPITAL_LETTER_ANG = 0x000118AA;
const char32_t WARANG_CITI_CAPITAL_LETTER_GA = 0x000118AB;
const char32_t WARANG_CITI_CAPITAL_LETTER_KO = 0x000118AC;
const char32_t WARANG_CITI_CAPITAL_LETTER_ENY = 0x000118AD;
const char32_t WARANG_CITI_CAPITAL_LETTER_YUJ = 0x000118AE;
const char32_t WARANG_CITI_CAPITAL_LETTER_UC = 0x000118AF;
const char32_t WARANG_CITI_CAPITAL_LETTER_ENN = 0x000118B0;
const char32_t WARANG_CITI_CAPITAL_LETTER_ODD = 0x000118B1;
const char32_t WARANG_CITI_CAPITAL_LETTER_TTE = 0x000118B2;
const char32_t WARANG_CITI_CAPITAL_LETTER_NUNG = 0x000118B3;
const char32_t WARANG_CITI_CAPITAL_LETTER_DA = 0x000118B4;
const char32_t WARANG_CITI_CAPITAL_LETTER_AT = 0x000118B5;
const char32_t WARANG_CITI_CAPITAL_LETTER_AM = 0x000118B6;
const char32_t WARANG_CITI_CAPITAL_LETTER_BU = 0x000118B7;
const char32_t WARANG_CITI_CAPITAL_LETTER_PU = 0x000118B8;
const char32_t WARANG_CITI_CAPITAL_LETTER_HIYO = 0x000118B9;
const char32_t WARANG_CITI_CAPITAL_LETTER_HOLO = 0x000118BA;
const char32_t WARANG_CITI_CAPITAL_LETTER_HORR = 0x000118BB;
const char32_t WARANG_CITI_CAPITAL_LETTER_HAR = 0x000118BC;
const char32_t WARANG_CITI_CAPITAL_LETTER_SSUU = 0x000118BD;
const char32_t WARANG_CITI_CAPITAL_LETTER_SII = 0x000118BE;
const char32_t WARANG_CITI_CAPITAL_LETTER_VIYO = 0x000118BF;
const char32_t WARANG_CITI_SMALL_LETTER_NGAA = 0x000118C0;
const char32_t WARANG_CITI_SMALL_LETTER_A = 0x000118C1;
const char32_t WARANG_CITI_SMALL_LETTER_WI = 0x000118C2;
const char32_t WARANG_CITI_SMALL_LETTER_YU = 0x000118C3;
const char32_t WARANG_CITI_SMALL_LETTER_YA = 0x000118C4;
const char32_t WARANG_CITI_SMALL_LETTER_YO = 0x000118C5;
const char32_t WARANG_CITI_SMALL_LETTER_II = 0x000118C6;
const char32_t WARANG_CITI_SMALL_LETTER_UU = 0x000118C7;
const char32_t WARANG_CITI_SMALL_LETTER_E = 0x000118C8;
const char32_t WARANG_CITI_SMALL_LETTER_O = 0x000118C9;
const char32_t WARANG_CITI_SMALL_LETTER_ANG = 0x000118CA;
const char32_t WARANG_CITI_SMALL_LETTER_GA = 0x000118CB;
const char32_t WARANG_CITI_SMALL_LETTER_KO = 0x000118CC;
const char32_t WARANG_CITI_SMALL_LETTER_ENY = 0x000118CD;
const char32_t WARANG_CITI_SMALL_LETTER_YUJ = 0x000118CE;
const char32_t WARANG_CITI_SMALL_LETTER_UC = 0x000118CF;
const char32_t WARANG_CITI_SMALL_LETTER_ENN = 0x000118D0;
const char32_t WARANG_CITI_SMALL_LETTER_ODD = 0x000118D1;
const char32_t WARANG_CITI_SMALL_LETTER_TTE = 0x000118D2;
const char32_t WARANG_CITI_SMALL_LETTER_NUNG = 0x000118D3;
const char32_t WARANG_CITI_SMALL_LETTER_DA = 0x000118D4;
const char32_t WARANG_CITI_SMALL_LETTER_AT = 0x000118D5;
const char32_t WARANG_CITI_SMALL_LETTER_AM = 0x000118D6;
const char32_t WARANG_CITI_SMALL_LETTER_BU = 0x000118D7;
const char32_t WARANG_CITI_SMALL_LETTER_PU = 0x000118D8;
const char32_t WARANG_CITI_SMALL_LETTER_HIYO = 0x000118D9;
const char32_t WARANG_CITI_SMALL_LETTER_HOLO = 0x000118DA;
const char32_t WARANG_CITI_SMALL_LETTER_HORR = 0x000118DB;
const char32_t WARANG_CITI_SMALL_LETTER_HAR = 0x000118DC;
const char32_t WARANG_CITI_SMALL_LETTER_SSUU = 0x000118DD;
const char32_t WARANG_CITI_SMALL_LETTER_SII = 0x000118DE;
const char32_t WARANG_CITI_SMALL_LETTER_VIYO = 0x000118DF;
const char32_t WARANG_CITI_DIGIT_ZERO = 0x000118E0;
const char32_t WARANG_CITI_DIGIT_ONE = 0x000118E1;
const char32_t WARANG_CITI_DIGIT_TWO = 0x000118E2;
const char32_t WARANG_CITI_DIGIT_THREE = 0x000118E3;
const char32_t WARANG_CITI_DIGIT_FOUR = 0x000118E4;
const char32_t WARANG_CITI_DIGIT_FIVE = 0x000118E5;
const char32_t WARANG_CITI_DIGIT_SIX = 0x000118E6;
const char32_t WARANG_CITI_DIGIT_SEVEN = 0x000118E7;
const char32_t WARANG_CITI_DIGIT_EIGHT = 0x000118E8;
const char32_t WARANG_CITI_DIGIT_NINE = 0x000118E9;
const char32_t WARANG_CITI_NUMBER_TEN = 0x000118EA;
const char32_t WARANG_CITI_NUMBER_TWENTY = 0x000118EB;
const char32_t WARANG_CITI_NUMBER_THIRTY = 0x000118EC;
const char32_t WARANG_CITI_NUMBER_FORTY = 0x000118ED;
const char32_t WARANG_CITI_NUMBER_FIFTY = 0x000118EE;
const char32_t WARANG_CITI_NUMBER_SIXTY = 0x000118EF;
const char32_t WARANG_CITI_NUMBER_SEVENTY = 0x000118F0;
const char32_t WARANG_CITI_NUMBER_EIGHTY = 0x000118F1;
const char32_t WARANG_CITI_NUMBER_NINETY = 0x000118F2;
const char32_t WARANG_CITI_OM = 0x000118FF;
const char32_t DIVES_AKURU_LETTER_A = 0x00011900;
const char32_t DIVES_AKURU_LETTER_AA = 0x00011901;
const char32_t DIVES_AKURU_LETTER_I = 0x00011902;
const char32_t DIVES_AKURU_LETTER_II = 0x00011903;
const char32_t DIVES_AKURU_LETTER_U = 0x00011904;
const char32_t DIVES_AKURU_LETTER_UU = 0x00011905;
const char32_t DIVES_AKURU_LETTER_E = 0x00011906;
const char32_t DIVES_AKURU_LETTER_O = 0x00011909;
const char32_t DIVES_AKURU_LETTER_KA = 0x0001190C;
const char32_t DIVES_AKURU_LETTER_KHA = 0x0001190D;
const char32_t DIVES_AKURU_LETTER_GA = 0x0001190E;
const char32_t DIVES_AKURU_LETTER_GHA = 0x0001190F;
const char32_t DIVES_AKURU_LETTER_NGA = 0x00011910;
const char32_t DIVES_AKURU_LETTER_CA = 0x00011911;
const char32_t DIVES_AKURU_LETTER_CHA = 0x00011912;
const char32_t DIVES_AKURU_LETTER_JA = 0x00011913;
const char32_t DIVES_AKURU_LETTER_NYA = 0x00011915;
const char32_t DIVES_AKURU_LETTER_TTA = 0x00011916;
const char32_t DIVES_AKURU_LETTER_DDA = 0x00011918;
const char32_t DIVES_AKURU_LETTER_DDHA = 0x00011919;
const char32_t DIVES_AKURU_LETTER_NNA = 0x0001191A;
const char32_t DIVES_AKURU_LETTER_TA = 0x0001191B;
const char32_t DIVES_AKURU_LETTER_THA = 0x0001191C;
const char32_t DIVES_AKURU_LETTER_DA = 0x0001191D;
const char32_t DIVES_AKURU_LETTER_DHA = 0x0001191E;
const char32_t DIVES_AKURU_LETTER_NA = 0x0001191F;
const char32_t DIVES_AKURU_LETTER_PA = 0x00011920;
const char32_t DIVES_AKURU_LETTER_PHA = 0x00011921;
const char32_t DIVES_AKURU_LETTER_BA = 0x00011922;
const char32_t DIVES_AKURU_LETTER_BHA = 0x00011923;
const char32_t DIVES_AKURU_LETTER_MA = 0x00011924;
const char32_t DIVES_AKURU_LETTER_YA = 0x00011925;
const char32_t DIVES_AKURU_LETTER_YYA = 0x00011926;
const char32_t DIVES_AKURU_LETTER_RA = 0x00011927;
const char32_t DIVES_AKURU_LETTER_LA = 0x00011928;
const char32_t DIVES_AKURU_LETTER_VA = 0x00011929;
const char32_t DIVES_AKURU_LETTER_SHA = 0x0001192A;
const char32_t DIVES_AKURU_LETTER_SSA = 0x0001192B;
const char32_t DIVES_AKURU_LETTER_SA = 0x0001192C;
const char32_t DIVES_AKURU_LETTER_HA = 0x0001192D;
const char32_t DIVES_AKURU_LETTER_LLA = 0x0001192E;
const char32_t DIVES_AKURU_LETTER_ZA = 0x0001192F;
const char32_t DIVES_AKURU_VOWEL_SIGN_AA = 0x00011930;
const char32_t DIVES_AKURU_VOWEL_SIGN_I = 0x00011931;
const char32_t DIVES_AKURU_VOWEL_SIGN_II = 0x00011932;
const char32_t DIVES_AKURU_VOWEL_SIGN_U = 0x00011933;
const char32_t DIVES_AKURU_VOWEL_SIGN_UU = 0x00011934;
const char32_t DIVES_AKURU_VOWEL_SIGN_E = 0x00011935;
const char32_t DIVES_AKURU_VOWEL_SIGN_AI = 0x00011937;
const char32_t DIVES_AKURU_VOWEL_SIGN_O = 0x00011938;
const char32_t DIVES_AKURU_SIGN_ANUSVARA = 0x0001193B;
const char32_t DIVES_AKURU_SIGN_CANDRABINDU = 0x0001193C;
const char32_t DIVES_AKURU_SIGN_HALANTA = 0x0001193D;
const char32_t DIVES_AKURU_VIRAMA = 0x0001193E;
const char32_t DIVES_AKURU_PREFIXED_NASAL_SIGN = 0x0001193F;
const char32_t DIVES_AKURU_MEDIAL_YA = 0x00011940;
const char32_t DIVES_AKURU_INITIAL_RA = 0x00011941;
const char32_t DIVES_AKURU_MEDIAL_RA = 0x00011942;
const char32_t DIVES_AKURU_SIGN_NUKTA = 0x00011943;
const char32_t DIVES_AKURU_DOUBLE_DANDA = 0x00011944;
const char32_t DIVES_AKURU_GAP_FILLER = 0x00011945;
const char32_t DIVES_AKURU_END_OF_TEXT_MARK = 0x00011946;
const char32_t DIVES_AKURU_DIGIT_ZERO = 0x00011950;
const char32_t DIVES_AKURU_DIGIT_ONE = 0x00011951;
const char32_t DIVES_AKURU_DIGIT_TWO = 0x00011952;
const char32_t DIVES_AKURU_DIGIT_THREE = 0x00011953;
const char32_t DIVES_AKURU_DIGIT_FOUR = 0x00011954;
const char32_t DIVES_AKURU_DIGIT_FIVE = 0x00011955;
const char32_t DIVES_AKURU_DIGIT_SIX = 0x00011956;
const char32_t DIVES_AKURU_DIGIT_SEVEN = 0x00011957;
const char32_t DIVES_AKURU_DIGIT_EIGHT = 0x00011958;
const char32_t DIVES_AKURU_DIGIT_NINE = 0x00011959;
const char32_t NANDINAGARI_LETTER_A = 0x000119A0;
const char32_t NANDINAGARI_LETTER_AA = 0x000119A1;
const char32_t NANDINAGARI_LETTER_I = 0x000119A2;
const char32_t NANDINAGARI_LETTER_II = 0x000119A3;
const char32_t NANDINAGARI_LETTER_U = 0x000119A4;
const char32_t NANDINAGARI_LETTER_UU = 0x000119A5;
const char32_t NANDINAGARI_LETTER_VOCALIC_R = 0x000119A6;
const char32_t NANDINAGARI_LETTER_VOCALIC_RR = 0x000119A7;
const char32_t NANDINAGARI_LETTER_E = 0x000119AA;
const char32_t NANDINAGARI_LETTER_AI = 0x000119AB;
const char32_t NANDINAGARI_LETTER_O = 0x000119AC;
const char32_t NANDINAGARI_LETTER_AU = 0x000119AD;
const char32_t NANDINAGARI_LETTER_KA = 0x000119AE;
const char32_t NANDINAGARI_LETTER_KHA = 0x000119AF;
const char32_t NANDINAGARI_LETTER_GA = 0x000119B0;
const char32_t NANDINAGARI_LETTER_GHA = 0x000119B1;
const char32_t NANDINAGARI_LETTER_NGA = 0x000119B2;
const char32_t NANDINAGARI_LETTER_CA = 0x000119B3;
const char32_t NANDINAGARI_LETTER_CHA = 0x000119B4;
const char32_t NANDINAGARI_LETTER_JA = 0x000119B5;
const char32_t NANDINAGARI_LETTER_JHA = 0x000119B6;
const char32_t NANDINAGARI_LETTER_NYA = 0x000119B7;
const char32_t NANDINAGARI_LETTER_TTA = 0x000119B8;
const char32_t NANDINAGARI_LETTER_TTHA = 0x000119B9;
const char32_t NANDINAGARI_LETTER_DDA = 0x000119BA;
const char32_t NANDINAGARI_LETTER_DDHA = 0x000119BB;
const char32_t NANDINAGARI_LETTER_NNA = 0x000119BC;
const char32_t NANDINAGARI_LETTER_TA = 0x000119BD;
const char32_t NANDINAGARI_LETTER_THA = 0x000119BE;
const char32_t NANDINAGARI_LETTER_DA = 0x000119BF;
const char32_t NANDINAGARI_LETTER_DHA = 0x000119C0;
const char32_t NANDINAGARI_LETTER_NA = 0x000119C1;
const char32_t NANDINAGARI_LETTER_PA = 0x000119C2;
const char32_t NANDINAGARI_LETTER_PHA = 0x000119C3;
const char32_t NANDINAGARI_LETTER_BA = 0x000119C4;
const char32_t NANDINAGARI_LETTER_BHA = 0x000119C5;
const char32_t NANDINAGARI_LETTER_MA = 0x000119C6;
const char32_t NANDINAGARI_LETTER_YA = 0x000119C7;
const char32_t NANDINAGARI_LETTER_RA = 0x000119C8;
const char32_t NANDINAGARI_LETTER_LA = 0x000119C9;
const char32_t NANDINAGARI_LETTER_VA = 0x000119CA;
const char32_t NANDINAGARI_LETTER_SHA = 0x000119CB;
const char32_t NANDINAGARI_LETTER_SSA = 0x000119CC;
const char32_t NANDINAGARI_LETTER_SA = 0x000119CD;
const char32_t NANDINAGARI_LETTER_HA = 0x000119CE;
const char32_t NANDINAGARI_LETTER_LLA = 0x000119CF;
const char32_t NANDINAGARI_LETTER_RRA = 0x000119D0;
const char32_t NANDINAGARI_VOWEL_SIGN_AA = 0x000119D1;
const char32_t NANDINAGARI_VOWEL_SIGN_I = 0x000119D2;
const char32_t NANDINAGARI_VOWEL_SIGN_II = 0x000119D3;
const char32_t NANDINAGARI_VOWEL_SIGN_U = 0x000119D4;
const char32_t NANDINAGARI_VOWEL_SIGN_UU = 0x000119D5;
const char32_t NANDINAGARI_VOWEL_SIGN_VOCALIC_R = 0x000119D6;
const char32_t NANDINAGARI_VOWEL_SIGN_VOCALIC_RR = 0x000119D7;
const char32_t NANDINAGARI_VOWEL_SIGN_E = 0x000119DA;
const char32_t NANDINAGARI_VOWEL_SIGN_AI = 0x000119DB;
const char32_t NANDINAGARI_VOWEL_SIGN_O = 0x000119DC;
const char32_t NANDINAGARI_VOWEL_SIGN_AU = 0x000119DD;
const char32_t NANDINAGARI_SIGN_ANUSVARA = 0x000119DE;
const char32_t NANDINAGARI_SIGN_VISARGA = 0x000119DF;
const char32_t NANDINAGARI_SIGN_VIRAMA = 0x000119E0;
const char32_t NANDINAGARI_SIGN_AVAGRAHA = 0x000119E1;
const char32_t NANDINAGARI_SIGN_SIDDHAM = 0x000119E2;
const char32_t NANDINAGARI_HEADSTROKE = 0x000119E3;
const char32_t NANDINAGARI_VOWEL_SIGN_PRISHTHAMATRA_E = 0x000119E4;
const char32_t ZANABAZAR_SQUARE_LETTER_A = 0x00011A00;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_I = 0x00011A01;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_UE = 0x00011A02;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_U = 0x00011A03;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_E = 0x00011A04;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_OE = 0x00011A05;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_O = 0x00011A06;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_AI = 0x00011A07;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_AU = 0x00011A08;
const char32_t ZANABAZAR_SQUARE_VOWEL_SIGN_REVERSED_I = 0x00011A09;
const char32_t ZANABAZAR_SQUARE_VOWEL_LENGTH_MARK = 0x00011A0A;
const char32_t ZANABAZAR_SQUARE_LETTER_KA = 0x00011A0B;
const char32_t ZANABAZAR_SQUARE_LETTER_KHA = 0x00011A0C;
const char32_t ZANABAZAR_SQUARE_LETTER_GA = 0x00011A0D;
const char32_t ZANABAZAR_SQUARE_LETTER_GHA = 0x00011A0E;
const char32_t ZANABAZAR_SQUARE_LETTER_NGA = 0x00011A0F;
const char32_t ZANABAZAR_SQUARE_LETTER_CA = 0x00011A10;
const char32_t ZANABAZAR_SQUARE_LETTER_CHA = 0x00011A11;
const char32_t ZANABAZAR_SQUARE_LETTER_JA = 0x00011A12;
const char32_t ZANABAZAR_SQUARE_LETTER_NYA = 0x00011A13;
const char32_t ZANABAZAR_SQUARE_LETTER_TTA = 0x00011A14;
const char32_t ZANABAZAR_SQUARE_LETTER_TTHA = 0x00011A15;
const char32_t ZANABAZAR_SQUARE_LETTER_DDA = 0x00011A16;
const char32_t ZANABAZAR_SQUARE_LETTER_DDHA = 0x00011A17;
const char32_t ZANABAZAR_SQUARE_LETTER_NNA = 0x00011A18;
const char32_t ZANABAZAR_SQUARE_LETTER_TA = 0x00011A19;
const char32_t ZANABAZAR_SQUARE_LETTER_THA = 0x00011A1A;
const char32_t ZANABAZAR_SQUARE_LETTER_DA = 0x00011A1B;
const char32_t ZANABAZAR_SQUARE_LETTER_DHA = 0x00011A1C;
const char32_t ZANABAZAR_SQUARE_LETTER_NA = 0x00011A1D;
const char32_t ZANABAZAR_SQUARE_LETTER_PA = 0x00011A1E;
const char32_t ZANABAZAR_SQUARE_LETTER_PHA = 0x00011A1F;
const char32_t ZANABAZAR_SQUARE_LETTER_BA = 0x00011A20;
const char32_t ZANABAZAR_SQUARE_LETTER_BHA = 0x00011A21;
const char32_t ZANABAZAR_SQUARE_LETTER_MA = 0x00011A22;
const char32_t ZANABAZAR_SQUARE_LETTER_TSA = 0x00011A23;
const char32_t ZANABAZAR_SQUARE_LETTER_TSHA = 0x00011A24;
const char32_t ZANABAZAR_SQUARE_LETTER_DZA = 0x00011A25;
const char32_t ZANABAZAR_SQUARE_LETTER_DZHA = 0x00011A26;
const char32_t ZANABAZAR_SQUARE_LETTER_ZHA = 0x00011A27;
const char32_t ZANABAZAR_SQUARE_LETTER_ZA = 0x00011A28;
const char32_t ZANABAZAR_SQUARE_LETTER__A = 0x00011A29;
const char32_t ZANABAZAR_SQUARE_LETTER_YA = 0x00011A2A;
const char32_t ZANABAZAR_SQUARE_LETTER_RA = 0x00011A2B;
const char32_t ZANABAZAR_SQUARE_LETTER_LA = 0x00011A2C;
const char32_t ZANABAZAR_SQUARE_LETTER_VA = 0x00011A2D;
const char32_t ZANABAZAR_SQUARE_LETTER_SHA = 0x00011A2E;
const char32_t ZANABAZAR_SQUARE_LETTER_SSA = 0x00011A2F;
const char32_t ZANABAZAR_SQUARE_LETTER_SA = 0x00011A30;
const char32_t ZANABAZAR_SQUARE_LETTER_HA = 0x00011A31;
const char32_t ZANABAZAR_SQUARE_LETTER_KSSA = 0x00011A32;
const char32_t ZANABAZAR_SQUARE_FINAL_CONSONANT_MARK = 0x00011A33;
const char32_t ZANABAZAR_SQUARE_SIGN_VIRAMA = 0x00011A34;
const char32_t ZANABAZAR_SQUARE_SIGN_CANDRABINDU = 0x00011A35;
const char32_t ZANABAZAR_SQUARE_SIGN_CANDRABINDU_WITH_ORNAMENT = 0x00011A36;
const char32_t ZANABAZAR_SQUARE_SIGN_CANDRA_WITH_ORNAMENT = 0x00011A37;
const char32_t ZANABAZAR_SQUARE_SIGN_ANUSVARA = 0x00011A38;
const char32_t ZANABAZAR_SQUARE_SIGN_VISARGA = 0x00011A39;
const char32_t ZANABAZAR_SQUARE_CLUSTER_INITIAL_LETTER_RA = 0x00011A3A;
const char32_t ZANABAZAR_SQUARE_CLUSTER_FINAL_LETTER_YA = 0x00011A3B;
const char32_t ZANABAZAR_SQUARE_CLUSTER_FINAL_LETTER_RA = 0x00011A3C;
const char32_t ZANABAZAR_SQUARE_CLUSTER_FINAL_LETTER_LA = 0x00011A3D;
const char32_t ZANABAZAR_SQUARE_CLUSTER_FINAL_LETTER_VA = 0x00011A3E;
const char32_t ZANABAZAR_SQUARE_INITIAL_HEAD_MARK = 0x00011A3F;
const char32_t ZANABAZAR_SQUARE_CLOSING_HEAD_MARK = 0x00011A40;
const char32_t ZANABAZAR_SQUARE_MARK_TSHEG = 0x00011A41;
const char32_t ZANABAZAR_SQUARE_MARK_SHAD = 0x00011A42;
const char32_t ZANABAZAR_SQUARE_MARK_DOUBLE_SHAD = 0x00011A43;
const char32_t ZANABAZAR_SQUARE_MARK_LONG_TSHEG = 0x00011A44;
const char32_t ZANABAZAR_SQUARE_INITIAL_DOUBLE_LINED_HEAD_MARK = 0x00011A45;
const char32_t ZANABAZAR_SQUARE_CLOSING_DOUBLE_LINED_HEAD_MARK = 0x00011A46;
const char32_t ZANABAZAR_SQUARE_SUBJOINER = 0x00011A47;
const char32_t SOYOMBO_LETTER_A = 0x00011A50;
const char32_t SOYOMBO_VOWEL_SIGN_I = 0x00011A51;
const char32_t SOYOMBO_VOWEL_SIGN_UE = 0x00011A52;
const char32_t SOYOMBO_VOWEL_SIGN_U = 0x00011A53;
const char32_t SOYOMBO_VOWEL_SIGN_E = 0x00011A54;
const char32_t SOYOMBO_VOWEL_SIGN_O = 0x00011A55;
const char32_t SOYOMBO_VOWEL_SIGN_OE = 0x00011A56;
const char32_t SOYOMBO_VOWEL_SIGN_AI = 0x00011A57;
const char32_t SOYOMBO_VOWEL_SIGN_AU = 0x00011A58;
const char32_t SOYOMBO_VOWEL_SIGN_VOCALIC_R = 0x00011A59;
const char32_t SOYOMBO_VOWEL_SIGN_VOCALIC_L = 0x00011A5A;
const char32_t SOYOMBO_VOWEL_LENGTH_MARK = 0x00011A5B;
const char32_t SOYOMBO_LETTER_KA = 0x00011A5C;
const char32_t SOYOMBO_LETTER_KHA = 0x00011A5D;
const char32_t SOYOMBO_LETTER_GA = 0x00011A5E;
const char32_t SOYOMBO_LETTER_GHA = 0x00011A5F;
const char32_t SOYOMBO_LETTER_NGA = 0x00011A60;
const char32_t SOYOMBO_LETTER_CA = 0x00011A61;
const char32_t SOYOMBO_LETTER_CHA = 0x00011A62;
const char32_t SOYOMBO_LETTER_JA = 0x00011A63;
const char32_t SOYOMBO_LETTER_JHA = 0x00011A64;
const char32_t SOYOMBO_LETTER_NYA = 0x00011A65;
const char32_t SOYOMBO_LETTER_TTA = 0x00011A66;
const char32_t SOYOMBO_LETTER_TTHA = 0x00011A67;
const char32_t SOYOMBO_LETTER_DDA = 0x00011A68;
const char32_t SOYOMBO_LETTER_DDHA = 0x00011A69;
const char32_t SOYOMBO_LETTER_NNA = 0x00011A6A;
const char32_t SOYOMBO_LETTER_TA = 0x00011A6B;
const char32_t SOYOMBO_LETTER_THA = 0x00011A6C;
const char32_t SOYOMBO_LETTER_DA = 0x00011A6D;
const char32_t SOYOMBO_LETTER_DHA = 0x00011A6E;
const char32_t SOYOMBO_LETTER_NA = 0x00011A6F;
const char32_t SOYOMBO_LETTER_PA = 0x00011A70;
const char32_t SOYOMBO_LETTER_PHA = 0x00011A71;
const char32_t SOYOMBO_LETTER_BA = 0x00011A72;
const char32_t SOYOMBO_LETTER_BHA = 0x00011A73;
const char32_t SOYOMBO_LETTER_MA = 0x00011A74;
const char32_t SOYOMBO_LETTER_TSA = 0x00011A75;
const char32_t SOYOMBO_LETTER_TSHA = 0x00011A76;
const char32_t SOYOMBO_LETTER_DZA = 0x00011A77;
const char32_t SOYOMBO_LETTER_ZHA = 0x00011A78;
const char32_t SOYOMBO_LETTER_ZA = 0x00011A79;
const char32_t SOYOMBO_LETTER__A = 0x00011A7A;
const char32_t SOYOMBO_LETTER_YA = 0x00011A7B;
const char32_t SOYOMBO_LETTER_RA = 0x00011A7C;
const char32_t SOYOMBO_LETTER_LA = 0x00011A7D;
const char32_t SOYOMBO_LETTER_VA = 0x00011A7E;
const char32_t SOYOMBO_LETTER_SHA = 0x00011A7F;
const char32_t SOYOMBO_LETTER_SSA = 0x00011A80;
const char32_t SOYOMBO_LETTER_SA = 0x00011A81;
const char32_t SOYOMBO_LETTER_HA = 0x00011A82;
const char32_t SOYOMBO_LETTER_KSSA = 0x00011A83;
const char32_t SOYOMBO_SIGN_JIHVAMULIYA = 0x00011A84;
const char32_t SOYOMBO_SIGN_UPADHMANIYA = 0x00011A85;
const char32_t SOYOMBO_CLUSTER_INITIAL_LETTER_RA = 0x00011A86;
const char32_t SOYOMBO_CLUSTER_INITIAL_LETTER_LA = 0x00011A87;
const char32_t SOYOMBO_CLUSTER_INITIAL_LETTER_SHA = 0x00011A88;
const char32_t SOYOMBO_CLUSTER_INITIAL_LETTER_SA = 0x00011A89;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_G = 0x00011A8A;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_K = 0x00011A8B;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_NG = 0x00011A8C;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_D = 0x00011A8D;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_N = 0x00011A8E;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_B = 0x00011A8F;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_M = 0x00011A90;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_R = 0x00011A91;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_L = 0x00011A92;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_SH = 0x00011A93;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN_S = 0x00011A94;
const char32_t SOYOMBO_FINAL_CONSONANT_SIGN__A = 0x00011A95;
const char32_t SOYOMBO_SIGN_ANUSVARA = 0x00011A96;
const char32_t SOYOMBO_SIGN_VISARGA = 0x00011A97;
const char32_t SOYOMBO_GEMINATION_MARK = 0x00011A98;
const char32_t SOYOMBO_SUBJOINER = 0x00011A99;
const char32_t SOYOMBO_MARK_TSHEG = 0x00011A9A;
const char32_t SOYOMBO_MARK_SHAD = 0x00011A9B;
const char32_t SOYOMBO_MARK_DOUBLE_SHAD = 0x00011A9C;
const char32_t SOYOMBO_MARK_PLUTA = 0x00011A9D;
const char32_t SOYOMBO_HEAD_MARK_WITH_MOON_AND_SUN_AND_TRIPLE_FLAME = 0x00011A9E;
const char32_t SOYOMBO_HEAD_MARK_WITH_MOON_AND_SUN_AND_FLAME = 0x00011A9F;
const char32_t SOYOMBO_HEAD_MARK_WITH_MOON_AND_SUN = 0x00011AA0;
const char32_t SOYOMBO_TERMINAL_MARK_1 = 0x00011AA1;
const char32_t SOYOMBO_TERMINAL_MARK_2 = 0x00011AA2;
const char32_t CANADIAN_SYLLABICS_NATTILIK_HI = 0x00011AB0;
const char32_t CANADIAN_SYLLABICS_NATTILIK_HII = 0x00011AB1;
const char32_t CANADIAN_SYLLABICS_NATTILIK_HO = 0x00011AB2;
const char32_t CANADIAN_SYLLABICS_NATTILIK_HOO = 0x00011AB3;
const char32_t CANADIAN_SYLLABICS_NATTILIK_HA = 0x00011AB4;
const char32_t CANADIAN_SYLLABICS_NATTILIK_HAA = 0x00011AB5;
const char32_t CANADIAN_SYLLABICS_NATTILIK_SHRI = 0x00011AB6;
const char32_t CANADIAN_SYLLABICS_NATTILIK_SHRII = 0x00011AB7;
const char32_t CANADIAN_SYLLABICS_NATTILIK_SHRO = 0x00011AB8;
const char32_t CANADIAN_SYLLABICS_NATTILIK_SHROO = 0x00011AB9;
const char32_t CANADIAN_SYLLABICS_NATTILIK_SHRA = 0x00011ABA;
const char32_t CANADIAN_SYLLABICS_NATTILIK_SHRAA = 0x00011ABB;
const char32_t CANADIAN_SYLLABICS_SPE = 0x00011ABC;
const char32_t CANADIAN_SYLLABICS_SPI = 0x00011ABD;
const char32_t CANADIAN_SYLLABICS_SPO = 0x00011ABE;
const char32_t CANADIAN_SYLLABICS_SPA = 0x00011ABF;
const char32_t PAU_CIN_HAU_LETTER_PA = 0x00011AC0;
const char32_t PAU_CIN_HAU_LETTER_KA = 0x00011AC1;
const char32_t PAU_CIN_HAU_LETTER_LA = 0x00011AC2;
const char32_t PAU_CIN_HAU_LETTER_MA = 0x00011AC3;
const char32_t PAU_CIN_HAU_LETTER_DA = 0x00011AC4;
const char32_t PAU_CIN_HAU_LETTER_ZA = 0x00011AC5;
const char32_t PAU_CIN_HAU_LETTER_VA = 0x00011AC6;
const char32_t PAU_CIN_HAU_LETTER_NGA = 0x00011AC7;
const char32_t PAU_CIN_HAU_LETTER_HA = 0x00011AC8;
const char32_t PAU_CIN_HAU_LETTER_GA = 0x00011AC9;
const char32_t PAU_CIN_HAU_LETTER_KHA = 0x00011ACA;
const char32_t PAU_CIN_HAU_LETTER_SA = 0x00011ACB;
const char32_t PAU_CIN_HAU_LETTER_BA = 0x00011ACC;
const char32_t PAU_CIN_HAU_LETTER_CA = 0x00011ACD;
const char32_t PAU_CIN_HAU_LETTER_TA = 0x00011ACE;
const char32_t PAU_CIN_HAU_LETTER_THA = 0x00011ACF;
const char32_t PAU_CIN_HAU_LETTER_NA = 0x00011AD0;
const char32_t PAU_CIN_HAU_LETTER_PHA = 0x00011AD1;
const char32_t PAU_CIN_HAU_LETTER_RA = 0x00011AD2;
const char32_t PAU_CIN_HAU_LETTER_FA = 0x00011AD3;
const char32_t PAU_CIN_HAU_LETTER_CHA = 0x00011AD4;
const char32_t PAU_CIN_HAU_LETTER_A = 0x00011AD5;
const char32_t PAU_CIN_HAU_LETTER_E = 0x00011AD6;
const char32_t PAU_CIN_HAU_LETTER_I = 0x00011AD7;
const char32_t PAU_CIN_HAU_LETTER_O = 0x00011AD8;
const char32_t PAU_CIN_HAU_LETTER_U = 0x00011AD9;
const char32_t PAU_CIN_HAU_LETTER_UA = 0x00011ADA;
const char32_t PAU_CIN_HAU_LETTER_IA = 0x00011ADB;
const char32_t PAU_CIN_HAU_LETTER_FINAL_P = 0x00011ADC;
const char32_t PAU_CIN_HAU_LETTER_FINAL_K = 0x00011ADD;
const char32_t PAU_CIN_HAU_LETTER_FINAL_T = 0x00011ADE;
const char32_t PAU_CIN_HAU_LETTER_FINAL_M = 0x00011ADF;
const char32_t PAU_CIN_HAU_LETTER_FINAL_N = 0x00011AE0;
const char32_t PAU_CIN_HAU_LETTER_FINAL_L = 0x00011AE1;
const char32_t PAU_CIN_HAU_LETTER_FINAL_W = 0x00011AE2;
const char32_t PAU_CIN_HAU_LETTER_FINAL_NG = 0x00011AE3;
const char32_t PAU_CIN_HAU_LETTER_FINAL_Y = 0x00011AE4;
const char32_t PAU_CIN_HAU_RISING_TONE_LONG = 0x00011AE5;
const char32_t PAU_CIN_HAU_RISING_TONE = 0x00011AE6;
const char32_t PAU_CIN_HAU_SANDHI_GLOTTAL_STOP = 0x00011AE7;
const char32_t PAU_CIN_HAU_RISING_TONE_LONG_FINAL = 0x00011AE8;
const char32_t PAU_CIN_HAU_RISING_TONE_FINAL = 0x00011AE9;
const char32_t PAU_CIN_HAU_SANDHI_GLOTTAL_STOP_FINAL = 0x00011AEA;
const char32_t PAU_CIN_HAU_SANDHI_TONE_LONG = 0x00011AEB;
const char32_t PAU_CIN_HAU_SANDHI_TONE = 0x00011AEC;
const char32_t PAU_CIN_HAU_SANDHI_TONE_LONG_FINAL = 0x00011AED;
const char32_t PAU_CIN_HAU_SANDHI_TONE_FINAL = 0x00011AEE;
const char32_t PAU_CIN_HAU_MID_LEVEL_TONE = 0x00011AEF;
const char32_t PAU_CIN_HAU_GLOTTAL_STOP_VARIANT = 0x00011AF0;
const char32_t PAU_CIN_HAU_MID_LEVEL_TONE_LONG_FINAL = 0x00011AF1;
const char32_t PAU_CIN_HAU_MID_LEVEL_TONE_FINAL = 0x00011AF2;
const char32_t PAU_CIN_HAU_LOW_FALLING_TONE_LONG = 0x00011AF3;
const char32_t PAU_CIN_HAU_LOW_FALLING_TONE = 0x00011AF4;
const char32_t PAU_CIN_HAU_GLOTTAL_STOP = 0x00011AF5;
const char32_t PAU_CIN_HAU_LOW_FALLING_TONE_LONG_FINAL = 0x00011AF6;
const char32_t PAU_CIN_HAU_LOW_FALLING_TONE_FINAL = 0x00011AF7;
const char32_t PAU_CIN_HAU_GLOTTAL_STOP_FINAL = 0x00011AF8;
const char32_t DEVANAGARI_HEAD_MARK = 0x00011B00;
const char32_t DEVANAGARI_HEAD_MARK_WITH_HEADSTROKE = 0x00011B01;
const char32_t DEVANAGARI_SIGN_BHALE = 0x00011B02;
const char32_t DEVANAGARI_SIGN_BHALE_WITH_HOOK = 0x00011B03;
const char32_t DEVANAGARI_SIGN_EXTENDED_BHALE = 0x00011B04;
const char32_t DEVANAGARI_SIGN_EXTENDED_BHALE_WITH_HOOK = 0x00011B05;
const char32_t DEVANAGARI_SIGN_WESTERN_FIVE_LIKE_BHALE = 0x00011B06;
const char32_t DEVANAGARI_SIGN_WESTERN_NINE_LIKE_BHALE = 0x00011B07;
const char32_t DEVANAGARI_SIGN_REVERSED_NINE_LIKE_BHALE = 0x00011B08;
const char32_t DEVANAGARI_SIGN_MINDU = 0x00011B09;
const char32_t SHARADA_VOWEL_SIGN_OE = 0x00011B60;
const char32_t SHARADA_VOWEL_SIGN_OOE = 0x00011B61;
const char32_t SHARADA_VOWEL_SIGN_UE = 0x00011B62;
const char32_t SHARADA_VOWEL_SIGN_UUE = 0x00011B63;
const char32_t SHARADA_VOWEL_SIGN_SHORT_E = 0x00011B64;
const char32_t SHARADA_VOWEL_SIGN_SHORT_O = 0x00011B65;
const char32_t SHARADA_VOWEL_SIGN_CANDRA_E = 0x00011B66;
const char32_t SHARADA_VOWEL_SIGN_CANDRA_O = 0x00011B67;
const char32_t SUNUWAR_LETTER_DEVI = 0x00011BC0;
const char32_t SUNUWAR_LETTER_TASLA = 0x00011BC1;
const char32_t SUNUWAR_LETTER_EKO = 0x00011BC2;
const char32_t SUNUWAR_LETTER_IMAR = 0x00011BC3;
const char32_t SUNUWAR_LETTER_REU = 0x00011BC4;
const char32_t SUNUWAR_LETTER_UTTHI = 0x00011BC5;
const char32_t SUNUWAR_LETTER_KIK = 0x00011BC6;
const char32_t SUNUWAR_LETTER_MA = 0x00011BC7;
const char32_t SUNUWAR_LETTER_APPHO = 0x00011BC8;
const char32_t SUNUWAR_LETTER_PIP = 0x00011BC9;
const char32_t SUNUWAR_LETTER_GIL = 0x00011BCA;
const char32_t SUNUWAR_LETTER_HAMSO = 0x00011BCB;
const char32_t SUNUWAR_LETTER_CARMI = 0x00011BCC;
const char32_t SUNUWAR_LETTER_NAH = 0x00011BCD;
const char32_t SUNUWAR_LETTER_BUR = 0x00011BCE;
const char32_t SUNUWAR_LETTER_JYAH = 0x00011BCF;
const char32_t SUNUWAR_LETTER_LOACHA = 0x00011BD0;
const char32_t SUNUWAR_LETTER_OTTHI = 0x00011BD1;
const char32_t SUNUWAR_LETTER_SHYELE = 0x00011BD2;
const char32_t SUNUWAR_LETTER_VARCA = 0x00011BD3;
const char32_t SUNUWAR_LETTER_YAT = 0x00011BD4;
const char32_t SUNUWAR_LETTER_AVA = 0x00011BD5;
const char32_t SUNUWAR_LETTER_AAL = 0x00011BD6;
const char32_t SUNUWAR_LETTER_DONGA = 0x00011BD7;
const char32_t SUNUWAR_LETTER_THARI = 0x00011BD8;
const char32_t SUNUWAR_LETTER_PHAR = 0x00011BD9;
const char32_t SUNUWAR_LETTER_NGAR = 0x00011BDA;
const char32_t SUNUWAR_LETTER_KHA = 0x00011BDB;
const char32_t SUNUWAR_LETTER_SHYER = 0x00011BDC;
const char32_t SUNUWAR_LETTER_CHELAP = 0x00011BDD;
const char32_t SUNUWAR_LETTER_TENTU = 0x00011BDE;
const char32_t SUNUWAR_LETTER_THELE = 0x00011BDF;
const char32_t SUNUWAR_LETTER_KLOKO = 0x00011BE0;
const char32_t SUNUWAR_SIGN_PVO = 0x00011BE1;
const char32_t SUNUWAR_DIGIT_ZERO = 0x00011BF0;
const char32_t SUNUWAR_DIGIT_ONE = 0x00011BF1;
const char32_t SUNUWAR_DIGIT_TWO = 0x00011BF2;
const char32_t SUNUWAR_DIGIT_THREE = 0x00011BF3;
const char32_t SUNUWAR_DIGIT_FOUR = 0x00011BF4;
const char32_t SUNUWAR_DIGIT_FIVE = 0x00011BF5;
const char32_t SUNUWAR_DIGIT_SIX = 0x00011BF6;
const char32_t SUNUWAR_DIGIT_SEVEN = 0x00011BF7;
const char32_t SUNUWAR_DIGIT_EIGHT = 0x00011BF8;
const char32_t SUNUWAR_DIGIT_NINE = 0x00011BF9;
const char32_t BHAIKSUKI_LETTER_A = 0x00011C00;
const char32_t BHAIKSUKI_LETTER_AA = 0x00011C01;
const char32_t BHAIKSUKI_LETTER_I = 0x00011C02;
const char32_t BHAIKSUKI_LETTER_II = 0x00011C03;
const char32_t BHAIKSUKI_LETTER_U = 0x00011C04;
const char32_t BHAIKSUKI_LETTER_UU = 0x00011C05;
const char32_t BHAIKSUKI_LETTER_VOCALIC_R = 0x00011C06;
const char32_t BHAIKSUKI_LETTER_VOCALIC_RR = 0x00011C07;
const char32_t BHAIKSUKI_LETTER_VOCALIC_L = 0x00011C08;
const char32_t BHAIKSUKI_LETTER_E = 0x00011C0A;
const char32_t BHAIKSUKI_LETTER_AI = 0x00011C0B;
const char32_t BHAIKSUKI_LETTER_O = 0x00011C0C;
const char32_t BHAIKSUKI_LETTER_AU = 0x00011C0D;
const char32_t BHAIKSUKI_LETTER_KA = 0x00011C0E;
const char32_t BHAIKSUKI_LETTER_KHA = 0x00011C0F;
const char32_t BHAIKSUKI_LETTER_GA = 0x00011C10;
const char32_t BHAIKSUKI_LETTER_GHA = 0x00011C11;
const char32_t BHAIKSUKI_LETTER_NGA = 0x00011C12;
const char32_t BHAIKSUKI_LETTER_CA = 0x00011C13;
const char32_t BHAIKSUKI_LETTER_CHA = 0x00011C14;
const char32_t BHAIKSUKI_LETTER_JA = 0x00011C15;
const char32_t BHAIKSUKI_LETTER_JHA = 0x00011C16;
const char32_t BHAIKSUKI_LETTER_NYA = 0x00011C17;
const char32_t BHAIKSUKI_LETTER_TTA = 0x00011C18;
const char32_t BHAIKSUKI_LETTER_TTHA = 0x00011C19;
const char32_t BHAIKSUKI_LETTER_DDA = 0x00011C1A;
const char32_t BHAIKSUKI_LETTER_DDHA = 0x00011C1B;
const char32_t BHAIKSUKI_LETTER_NNA = 0x00011C1C;
const char32_t BHAIKSUKI_LETTER_TA = 0x00011C1D;
const char32_t BHAIKSUKI_LETTER_THA = 0x00011C1E;
const char32_t BHAIKSUKI_LETTER_DA = 0x00011C1F;
const char32_t BHAIKSUKI_LETTER_DHA = 0x00011C20;
const char32_t BHAIKSUKI_LETTER_NA = 0x00011C21;
const char32_t BHAIKSUKI_LETTER_PA = 0x00011C22;
const char32_t BHAIKSUKI_LETTER_PHA = 0x00011C23;
const char32_t BHAIKSUKI_LETTER_BA = 0x00011C24;
const char32_t BHAIKSUKI_LETTER_BHA = 0x00011C25;
const char32_t BHAIKSUKI_LETTER_MA = 0x00011C26;
const char32_t BHAIKSUKI_LETTER_YA = 0x00011C27;
const char32_t BHAIKSUKI_LETTER_RA = 0x00011C28;
const char32_t BHAIKSUKI_LETTER_LA = 0x00011C29;
const char32_t BHAIKSUKI_LETTER_VA = 0x00011C2A;
const char32_t BHAIKSUKI_LETTER_SHA = 0x00011C2B;
const char32_t BHAIKSUKI_LETTER_SSA = 0x00011C2C;
const char32_t BHAIKSUKI_LETTER_SA = 0x00011C2D;
const char32_t BHAIKSUKI_LETTER_HA = 0x00011C2E;
const char32_t BHAIKSUKI_VOWEL_SIGN_AA = 0x00011C2F;
const char32_t BHAIKSUKI_VOWEL_SIGN_I = 0x00011C30;
const char32_t BHAIKSUKI_VOWEL_SIGN_II = 0x00011C31;
const char32_t BHAIKSUKI_VOWEL_SIGN_U = 0x00011C32;
const char32_t BHAIKSUKI_VOWEL_SIGN_UU = 0x00011C33;
const char32_t BHAIKSUKI_VOWEL_SIGN_VOCALIC_R = 0x00011C34;
const char32_t BHAIKSUKI_VOWEL_SIGN_VOCALIC_RR = 0x00011C35;
const char32_t BHAIKSUKI_VOWEL_SIGN_VOCALIC_L = 0x00011C36;
const char32_t BHAIKSUKI_VOWEL_SIGN_E = 0x00011C38;
const char32_t BHAIKSUKI_VOWEL_SIGN_AI = 0x00011C39;
const char32_t BHAIKSUKI_VOWEL_SIGN_O = 0x00011C3A;
const char32_t BHAIKSUKI_VOWEL_SIGN_AU = 0x00011C3B;
const char32_t BHAIKSUKI_SIGN_CANDRABINDU = 0x00011C3C;
const char32_t BHAIKSUKI_SIGN_ANUSVARA = 0x00011C3D;
const char32_t BHAIKSUKI_SIGN_VISARGA = 0x00011C3E;
const char32_t BHAIKSUKI_SIGN_VIRAMA = 0x00011C3F;
const char32_t BHAIKSUKI_SIGN_AVAGRAHA = 0x00011C40;
const char32_t BHAIKSUKI_DANDA = 0x00011C41;
const char32_t BHAIKSUKI_DOUBLE_DANDA = 0x00011C42;
const char32_t BHAIKSUKI_WORD_SEPARATOR = 0x00011C43;
const char32_t BHAIKSUKI_GAP_FILLER_1 = 0x00011C44;
const char32_t BHAIKSUKI_GAP_FILLER_2 = 0x00011C45;
const char32_t BHAIKSUKI_DIGIT_ZERO = 0x00011C50;
const char32_t BHAIKSUKI_DIGIT_ONE = 0x00011C51;
const char32_t BHAIKSUKI_DIGIT_TWO = 0x00011C52;
const char32_t BHAIKSUKI_DIGIT_THREE = 0x00011C53;
const char32_t BHAIKSUKI_DIGIT_FOUR = 0x00011C54;
const char32_t BHAIKSUKI_DIGIT_FIVE = 0x00011C55;
const char32_t BHAIKSUKI_DIGIT_SIX = 0x00011C56;
const char32_t BHAIKSUKI_DIGIT_SEVEN = 0x00011C57;
const char32_t BHAIKSUKI_DIGIT_EIGHT = 0x00011C58;
const char32_t BHAIKSUKI_DIGIT_NINE = 0x00011C59;
const char32_t BHAIKSUKI_NUMBER_ONE = 0x00011C5A;
const char32_t BHAIKSUKI_NUMBER_TWO = 0x00011C5B;
const char32_t BHAIKSUKI_NUMBER_THREE = 0x00011C5C;
const char32_t BHAIKSUKI_NUMBER_FOUR = 0x00011C5D;
const char32_t BHAIKSUKI_NUMBER_FIVE = 0x00011C5E;
const char32_t BHAIKSUKI_NUMBER_SIX = 0x00011C5F;
const char32_t BHAIKSUKI_NUMBER_SEVEN = 0x00011C60;
const char32_t BHAIKSUKI_NUMBER_EIGHT = 0x00011C61;
const char32_t BHAIKSUKI_NUMBER_NINE = 0x00011C62;
const char32_t BHAIKSUKI_NUMBER_TEN = 0x00011C63;
const char32_t BHAIKSUKI_NUMBER_TWENTY = 0x00011C64;
const char32_t BHAIKSUKI_NUMBER_THIRTY = 0x00011C65;
const char32_t BHAIKSUKI_NUMBER_FORTY = 0x00011C66;
const char32_t BHAIKSUKI_NUMBER_FIFTY = 0x00011C67;
const char32_t BHAIKSUKI_NUMBER_SIXTY = 0x00011C68;
const char32_t BHAIKSUKI_NUMBER_SEVENTY = 0x00011C69;
const char32_t BHAIKSUKI_NUMBER_EIGHTY = 0x00011C6A;
const char32_t BHAIKSUKI_NUMBER_NINETY = 0x00011C6B;
const char32_t BHAIKSUKI_HUNDREDS_UNIT_MARK = 0x00011C6C;
const char32_t MARCHEN_HEAD_MARK = 0x00011C70;
const char32_t MARCHEN_MARK_SHAD = 0x00011C71;
const char32_t MARCHEN_LETTER_KA = 0x00011C72;
const char32_t MARCHEN_LETTER_KHA = 0x00011C73;
const char32_t MARCHEN_LETTER_GA = 0x00011C74;
const char32_t MARCHEN_LETTER_NGA = 0x00011C75;
const char32_t MARCHEN_LETTER_CA = 0x00011C76;
const char32_t MARCHEN_LETTER_CHA = 0x00011C77;
const char32_t MARCHEN_LETTER_JA = 0x00011C78;
const char32_t MARCHEN_LETTER_NYA = 0x00011C79;
const char32_t MARCHEN_LETTER_TA = 0x00011C7A;
const char32_t MARCHEN_LETTER_THA = 0x00011C7B;
const char32_t MARCHEN_LETTER_DA = 0x00011C7C;
const char32_t MARCHEN_LETTER_NA = 0x00011C7D;
const char32_t MARCHEN_LETTER_PA = 0x00011C7E;
const char32_t MARCHEN_LETTER_PHA = 0x00011C7F;
const char32_t MARCHEN_LETTER_BA = 0x00011C80;
const char32_t MARCHEN_LETTER_MA = 0x00011C81;
const char32_t MARCHEN_LETTER_TSA = 0x00011C82;
const char32_t MARCHEN_LETTER_TSHA = 0x00011C83;
const char32_t MARCHEN_LETTER_DZA = 0x00011C84;
const char32_t MARCHEN_LETTER_WA = 0x00011C85;
const char32_t MARCHEN_LETTER_ZHA = 0x00011C86;
const char32_t MARCHEN_LETTER_ZA = 0x00011C87;
const char32_t MARCHEN_LETTER__A = 0x00011C88;
const char32_t MARCHEN_LETTER_YA = 0x00011C89;
const char32_t MARCHEN_LETTER_RA = 0x00011C8A;
const char32_t MARCHEN_LETTER_LA = 0x00011C8B;
const char32_t MARCHEN_LETTER_SHA = 0x00011C8C;
const char32_t MARCHEN_LETTER_SA = 0x00011C8D;
const char32_t MARCHEN_LETTER_HA = 0x00011C8E;
const char32_t MARCHEN_LETTER_A = 0x00011C8F;
const char32_t MARCHEN_SUBJOINED_LETTER_KA = 0x00011C92;
const char32_t MARCHEN_SUBJOINED_LETTER_KHA = 0x00011C93;
const char32_t MARCHEN_SUBJOINED_LETTER_GA = 0x00011C94;
const char32_t MARCHEN_SUBJOINED_LETTER_NGA = 0x00011C95;
const char32_t MARCHEN_SUBJOINED_LETTER_CA = 0x00011C96;
const char32_t MARCHEN_SUBJOINED_LETTER_CHA = 0x00011C97;
const char32_t MARCHEN_SUBJOINED_LETTER_JA = 0x00011C98;
const char32_t MARCHEN_SUBJOINED_LETTER_NYA = 0x00011C99;
const char32_t MARCHEN_SUBJOINED_LETTER_TA = 0x00011C9A;
const char32_t MARCHEN_SUBJOINED_LETTER_THA = 0x00011C9B;
const char32_t MARCHEN_SUBJOINED_LETTER_DA = 0x00011C9C;
const char32_t MARCHEN_SUBJOINED_LETTER_NA = 0x00011C9D;
const char32_t MARCHEN_SUBJOINED_LETTER_PA = 0x00011C9E;
const char32_t MARCHEN_SUBJOINED_LETTER_PHA = 0x00011C9F;
const char32_t MARCHEN_SUBJOINED_LETTER_BA = 0x00011CA0;
const char32_t MARCHEN_SUBJOINED_LETTER_MA = 0x00011CA1;
const char32_t MARCHEN_SUBJOINED_LETTER_TSA = 0x00011CA2;
const char32_t MARCHEN_SUBJOINED_LETTER_TSHA = 0x00011CA3;
const char32_t MARCHEN_SUBJOINED_LETTER_DZA = 0x00011CA4;
const char32_t MARCHEN_SUBJOINED_LETTER_WA = 0x00011CA5;
const char32_t MARCHEN_SUBJOINED_LETTER_ZHA = 0x00011CA6;
const char32_t MARCHEN_SUBJOINED_LETTER_ZA = 0x00011CA7;
const char32_t MARCHEN_SUBJOINED_LETTER_YA = 0x00011CA9;
const char32_t MARCHEN_SUBJOINED_LETTER_RA = 0x00011CAA;
const char32_t MARCHEN_SUBJOINED_LETTER_LA = 0x00011CAB;
const char32_t MARCHEN_SUBJOINED_LETTER_SHA = 0x00011CAC;
const char32_t MARCHEN_SUBJOINED_LETTER_SA = 0x00011CAD;
const char32_t MARCHEN_SUBJOINED_LETTER_HA = 0x00011CAE;
const char32_t MARCHEN_SUBJOINED_LETTER_A = 0x00011CAF;
const char32_t MARCHEN_VOWEL_SIGN_AA = 0x00011CB0;
const char32_t MARCHEN_VOWEL_SIGN_I = 0x00011CB1;
const char32_t MARCHEN_VOWEL_SIGN_U = 0x00011CB2;
const char32_t MARCHEN_VOWEL_SIGN_E = 0x00011CB3;
const char32_t MARCHEN_VOWEL_SIGN_O = 0x00011CB4;
const char32_t MARCHEN_SIGN_ANUSVARA = 0x00011CB5;
const char32_t MARCHEN_SIGN_CANDRABINDU = 0x00011CB6;
const char32_t MASARAM_GONDI_LETTER_A = 0x00011D00;
const char32_t MASARAM_GONDI_LETTER_AA = 0x00011D01;
const char32_t MASARAM_GONDI_LETTER_I = 0x00011D02;
const char32_t MASARAM_GONDI_LETTER_II = 0x00011D03;
const char32_t MASARAM_GONDI_LETTER_U = 0x00011D04;
const char32_t MASARAM_GONDI_LETTER_UU = 0x00011D05;
const char32_t MASARAM_GONDI_LETTER_E = 0x00011D06;
const char32_t MASARAM_GONDI_LETTER_AI = 0x00011D08;
const char32_t MASARAM_GONDI_LETTER_O = 0x00011D09;
const char32_t MASARAM_GONDI_LETTER_AU = 0x00011D0B;
const char32_t MASARAM_GONDI_LETTER_KA = 0x00011D0C;
const char32_t MASARAM_GONDI_LETTER_KHA = 0x00011D0D;
const char32_t MASARAM_GONDI_LETTER_GA = 0x00011D0E;
const char32_t MASARAM_GONDI_LETTER_GHA = 0x00011D0F;
const char32_t MASARAM_GONDI_LETTER_NGA = 0x00011D10;
const char32_t MASARAM_GONDI_LETTER_CA = 0x00011D11;
const char32_t MASARAM_GONDI_LETTER_CHA = 0x00011D12;
const char32_t MASARAM_GONDI_LETTER_JA = 0x00011D13;
const char32_t MASARAM_GONDI_LETTER_JHA = 0x00011D14;
const char32_t MASARAM_GONDI_LETTER_NYA = 0x00011D15;
const char32_t MASARAM_GONDI_LETTER_TTA = 0x00011D16;
const char32_t MASARAM_GONDI_LETTER_TTHA = 0x00011D17;
const char32_t MASARAM_GONDI_LETTER_DDA = 0x00011D18;
const char32_t MASARAM_GONDI_LETTER_DDHA = 0x00011D19;
const char32_t MASARAM_GONDI_LETTER_NNA = 0x00011D1A;
const char32_t MASARAM_GONDI_LETTER_TA = 0x00011D1B;
const char32_t MASARAM_GONDI_LETTER_THA = 0x00011D1C;
const char32_t MASARAM_GONDI_LETTER_DA = 0x00011D1D;
const char32_t MASARAM_GONDI_LETTER_DHA = 0x00011D1E;
const char32_t MASARAM_GONDI_LETTER_NA = 0x00011D1F;
const char32_t MASARAM_GONDI_LETTER_PA = 0x00011D20;
const char32_t MASARAM_GONDI_LETTER_PHA = 0x00011D21;
const char32_t MASARAM_GONDI_LETTER_BA = 0x00011D22;
const char32_t MASARAM_GONDI_LETTER_BHA = 0x00011D23;
const char32_t MASARAM_GONDI_LETTER_MA = 0x00011D24;
const char32_t MASARAM_GONDI_LETTER_YA = 0x00011D25;
const char32_t MASARAM_GONDI_LETTER_RA = 0x00011D26;
const char32_t MASARAM_GONDI_LETTER_LA = 0x00011D27;
const char32_t MASARAM_GONDI_LETTER_VA = 0x00011D28;
const char32_t MASARAM_GONDI_LETTER_SHA = 0x00011D29;
const char32_t MASARAM_GONDI_LETTER_SSA = 0x00011D2A;
const char32_t MASARAM_GONDI_LETTER_SA = 0x00011D2B;
const char32_t MASARAM_GONDI_LETTER_HA = 0x00011D2C;
const char32_t MASARAM_GONDI_LETTER_LLA = 0x00011D2D;
const char32_t MASARAM_GONDI_LETTER_KSSA = 0x00011D2E;
const char32_t MASARAM_GONDI_LETTER_JNYA = 0x00011D2F;
const char32_t MASARAM_GONDI_LETTER_TRA = 0x00011D30;
const char32_t MASARAM_GONDI_VOWEL_SIGN_AA = 0x00011D31;
const char32_t MASARAM_GONDI_VOWEL_SIGN_I = 0x00011D32;
const char32_t MASARAM_GONDI_VOWEL_SIGN_II = 0x00011D33;
const char32_t MASARAM_GONDI_VOWEL_SIGN_U = 0x00011D34;
const char32_t MASARAM_GONDI_VOWEL_SIGN_UU = 0x00011D35;
const char32_t MASARAM_GONDI_VOWEL_SIGN_VOCALIC_R = 0x00011D36;
const char32_t MASARAM_GONDI_VOWEL_SIGN_E = 0x00011D3A;
const char32_t MASARAM_GONDI_VOWEL_SIGN_AI = 0x00011D3C;
const char32_t MASARAM_GONDI_VOWEL_SIGN_O = 0x00011D3D;
const char32_t MASARAM_GONDI_VOWEL_SIGN_AU = 0x00011D3F;
const char32_t MASARAM_GONDI_SIGN_ANUSVARA = 0x00011D40;
const char32_t MASARAM_GONDI_SIGN_VISARGA = 0x00011D41;
const char32_t MASARAM_GONDI_SIGN_NUKTA = 0x00011D42;
const char32_t MASARAM_GONDI_SIGN_CANDRA = 0x00011D43;
const char32_t MASARAM_GONDI_SIGN_HALANTA = 0x00011D44;
const char32_t MASARAM_GONDI_VIRAMA = 0x00011D45;
const char32_t MASARAM_GONDI_REPHA = 0x00011D46;
const char32_t MASARAM_GONDI_RA_KARA = 0x00011D47;
const char32_t MASARAM_GONDI_DIGIT_ZERO = 0x00011D50;
const char32_t MASARAM_GONDI_DIGIT_ONE = 0x00011D51;
const char32_t MASARAM_GONDI_DIGIT_TWO = 0x00011D52;
const char32_t MASARAM_GONDI_DIGIT_THREE = 0x00011D53;
const char32_t MASARAM_GONDI_DIGIT_FOUR = 0x00011D54;
const char32_t MASARAM_GONDI_DIGIT_FIVE = 0x00011D55;
const char32_t MASARAM_GONDI_DIGIT_SIX = 0x00011D56;
const char32_t MASARAM_GONDI_DIGIT_SEVEN = 0x00011D57;
const char32_t MASARAM_GONDI_DIGIT_EIGHT = 0x00011D58;
const char32_t MASARAM_GONDI_DIGIT_NINE = 0x00011D59;
const char32_t GUNJALA_GONDI_LETTER_A = 0x00011D60;
const char32_t GUNJALA_GONDI_LETTER_AA = 0x00011D61;
const char32_t GUNJALA_GONDI_LETTER_I = 0x00011D62;
const char32_t GUNJALA_GONDI_LETTER_II = 0x00011D63;
const char32_t GUNJALA_GONDI_LETTER_U = 0x00011D64;
const char32_t GUNJALA_GONDI_LETTER_UU = 0x00011D65;
const char32_t GUNJALA_GONDI_LETTER_EE = 0x00011D67;
const char32_t GUNJALA_GONDI_LETTER_AI = 0x00011D68;
const char32_t GUNJALA_GONDI_LETTER_OO = 0x00011D6A;
const char32_t GUNJALA_GONDI_LETTER_AU = 0x00011D6B;
const char32_t GUNJALA_GONDI_LETTER_YA = 0x00011D6C;
const char32_t GUNJALA_GONDI_LETTER_VA = 0x00011D6D;
const char32_t GUNJALA_GONDI_LETTER_BA = 0x00011D6E;
const char32_t GUNJALA_GONDI_LETTER_BHA = 0x00011D6F;
const char32_t GUNJALA_GONDI_LETTER_MA = 0x00011D70;
const char32_t GUNJALA_GONDI_LETTER_KA = 0x00011D71;
const char32_t GUNJALA_GONDI_LETTER_KHA = 0x00011D72;
const char32_t GUNJALA_GONDI_LETTER_TA = 0x00011D73;
const char32_t GUNJALA_GONDI_LETTER_THA = 0x00011D74;
const char32_t GUNJALA_GONDI_LETTER_LA = 0x00011D75;
const char32_t GUNJALA_GONDI_LETTER_GA = 0x00011D76;
const char32_t GUNJALA_GONDI_LETTER_GHA = 0x00011D77;
const char32_t GUNJALA_GONDI_LETTER_DA = 0x00011D78;
const char32_t GUNJALA_GONDI_LETTER_DHA = 0x00011D79;
const char32_t GUNJALA_GONDI_LETTER_NA = 0x00011D7A;
const char32_t GUNJALA_GONDI_LETTER_CA = 0x00011D7B;
const char32_t GUNJALA_GONDI_LETTER_CHA = 0x00011D7C;
const char32_t GUNJALA_GONDI_LETTER_TTA = 0x00011D7D;
const char32_t GUNJALA_GONDI_LETTER_TTHA = 0x00011D7E;
const char32_t GUNJALA_GONDI_LETTER_LLA = 0x00011D7F;
const char32_t GUNJALA_GONDI_LETTER_JA = 0x00011D80;
const char32_t GUNJALA_GONDI_LETTER_JHA = 0x00011D81;
const char32_t GUNJALA_GONDI_LETTER_DDA = 0x00011D82;
const char32_t GUNJALA_GONDI_LETTER_DDHA = 0x00011D83;
const char32_t GUNJALA_GONDI_LETTER_NGA = 0x00011D84;
const char32_t GUNJALA_GONDI_LETTER_PA = 0x00011D85;
const char32_t GUNJALA_GONDI_LETTER_PHA = 0x00011D86;
const char32_t GUNJALA_GONDI_LETTER_HA = 0x00011D87;
const char32_t GUNJALA_GONDI_LETTER_RA = 0x00011D88;
const char32_t GUNJALA_GONDI_LETTER_SA = 0x00011D89;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_AA = 0x00011D8A;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_I = 0x00011D8B;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_II = 0x00011D8C;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_U = 0x00011D8D;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_UU = 0x00011D8E;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_EE = 0x00011D90;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_AI = 0x00011D91;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_OO = 0x00011D93;
const char32_t GUNJALA_GONDI_VOWEL_SIGN_AU = 0x00011D94;
const char32_t GUNJALA_GONDI_SIGN_ANUSVARA = 0x00011D95;
const char32_t GUNJALA_GONDI_SIGN_VISARGA = 0x00011D96;
const char32_t GUNJALA_GONDI_VIRAMA = 0x00011D97;
const char32_t GUNJALA_GONDI_OM = 0x00011D98;
const char32_t GUNJALA_GONDI_DIGIT_ZERO = 0x00011DA0;
const char32_t GUNJALA_GONDI_DIGIT_ONE = 0x00011DA1;
const char32_t GUNJALA_GONDI_DIGIT_TWO = 0x00011DA2;
const char32_t GUNJALA_GONDI_DIGIT_THREE = 0x00011DA3;
const char32_t GUNJALA_GONDI_DIGIT_FOUR = 0x00011DA4;
const char32_t GUNJALA_GONDI_DIGIT_FIVE = 0x00011DA5;
const char32_t GUNJALA_GONDI_DIGIT_SIX = 0x00011DA6;
const char32_t GUNJALA_GONDI_DIGIT_SEVEN = 0x00011DA7;
const char32_t GUNJALA_GONDI_DIGIT_EIGHT = 0x00011DA8;
const char32_t GUNJALA_GONDI_DIGIT_NINE = 0x00011DA9;
const char32_t TOLONG_SIKI_LETTER_I = 0x00011DB0;
const char32_t TOLONG_SIKI_LETTER_E = 0x00011DB1;
const char32_t TOLONG_SIKI_LETTER_U = 0x00011DB2;
const char32_t TOLONG_SIKI_LETTER_O = 0x00011DB3;
const char32_t TOLONG_SIKI_LETTER_A = 0x00011DB4;
const char32_t TOLONG_SIKI_LETTER_AA = 0x00011DB5;
const char32_t TOLONG_SIKI_LETTER_P = 0x00011DB6;
const char32_t TOLONG_SIKI_LETTER_PH = 0x00011DB7;
const char32_t TOLONG_SIKI_LETTER_B = 0x00011DB8;
const char32_t TOLONG_SIKI_LETTER_BH = 0x00011DB9;
const char32_t TOLONG_SIKI_LETTER_M = 0x00011DBA;
const char32_t TOLONG_SIKI_LETTER_T = 0x00011DBB;
const char32_t TOLONG_SIKI_LETTER_TH = 0x00011DBC;
const char32_t TOLONG_SIKI_LETTER_D = 0x00011DBD;
const char32_t TOLONG_SIKI_LETTER_DH = 0x00011DBE;
const char32_t TOLONG_SIKI_LETTER_N = 0x00011DBF;
const char32_t TOLONG_SIKI_LETTER_TT = 0x00011DC0;
const char32_t TOLONG_SIKI_LETTER_TTH = 0x00011DC1;
const char32_t TOLONG_SIKI_LETTER_DD = 0x00011DC2;
const char32_t TOLONG_SIKI_LETTER_DDH = 0x00011DC3;
const char32_t TOLONG_SIKI_LETTER_NN = 0x00011DC4;
const char32_t TOLONG_SIKI_LETTER_C = 0x00011DC5;
const char32_t TOLONG_SIKI_LETTER_CH = 0x00011DC6;
const char32_t TOLONG_SIKI_LETTER_J = 0x00011DC7;
const char32_t TOLONG_SIKI_LETTER_JH = 0x00011DC8;
const char32_t TOLONG_SIKI_LETTER_NY = 0x00011DC9;
const char32_t TOLONG_SIKI_LETTER_K = 0x00011DCA;
const char32_t TOLONG_SIKI_LETTER_KH = 0x00011DCB;
const char32_t TOLONG_SIKI_LETTER_G = 0x00011DCC;
const char32_t TOLONG_SIKI_LETTER_GH = 0x00011DCD;
const char32_t TOLONG_SIKI_LETTER_NG = 0x00011DCE;
const char32_t TOLONG_SIKI_LETTER_Y = 0x00011DCF;
const char32_t TOLONG_SIKI_LETTER_R = 0x00011DD0;
const char32_t TOLONG_SIKI_LETTER_L = 0x00011DD1;
const char32_t TOLONG_SIKI_LETTER_V = 0x00011DD2;
const char32_t TOLONG_SIKI_LETTER_NNY = 0x00011DD3;
const char32_t TOLONG_SIKI_LETTER_S = 0x00011DD4;
const char32_t TOLONG_SIKI_LETTER_H = 0x00011DD5;
const char32_t TOLONG_SIKI_LETTER_X = 0x00011DD6;
const char32_t TOLONG_SIKI_LETTER_RR = 0x00011DD7;
const char32_t TOLONG_SIKI_LETTER_RRH = 0x00011DD8;
const char32_t TOLONG_SIKI_SIGN_SELA = 0x00011DD9;
const char32_t TOLONG_SIKI_SIGN_HECAKA = 0x00011DDA;
const char32_t TOLONG_SIKI_UNGGA = 0x00011DDB;
const char32_t TOLONG_SIKI_DIGIT_ZERO = 0x00011DE0;
const char32_t TOLONG_SIKI_DIGIT_ONE = 0x00011DE1;
const char32_t TOLONG_SIKI_DIGIT_TWO = 0x00011DE2;
const char32_t TOLONG_SIKI_DIGIT_THREE = 0x00011DE3;
const char32_t TOLONG_SIKI_DIGIT_FOUR = 0x00011DE4;
const char32_t TOLONG_SIKI_DIGIT_FIVE = 0x00011DE5;
const char32_t TOLONG_SIKI_DIGIT_SIX = 0x00011DE6;
const char32_t TOLONG_SIKI_DIGIT_SEVEN = 0x00011DE7;
const char32_t TOLONG_SIKI_DIGIT_EIGHT = 0x00011DE8;
const char32_t TOLONG_SIKI_DIGIT_NINE = 0x00011DE9;
const char32_t MAKASAR_LETTER_KA = 0x00011EE0;
const char32_t MAKASAR_LETTER_GA = 0x00011EE1;
const char32_t MAKASAR_LETTER_NGA = 0x00011EE2;
const char32_t MAKASAR_LETTER_PA = 0x00011EE3;
const char32_t MAKASAR_LETTER_BA = 0x00011EE4;
const char32_t MAKASAR_LETTER_MA = 0x00011EE5;
const char32_t MAKASAR_LETTER_TA = 0x00011EE6;
const char32_t MAKASAR_LETTER_DA = 0x00011EE7;
const char32_t MAKASAR_LETTER_NA = 0x00011EE8;
const char32_t MAKASAR_LETTER_CA = 0x00011EE9;
const char32_t MAKASAR_LETTER_JA = 0x00011EEA;
const char32_t MAKASAR_LETTER_NYA = 0x00011EEB;
const char32_t MAKASAR_LETTER_YA = 0x00011EEC;
const char32_t MAKASAR_LETTER_RA = 0x00011EED;
const char32_t MAKASAR_LETTER_LA = 0x00011EEE;
const char32_t MAKASAR_LETTER_VA = 0x00011EEF;
const char32_t MAKASAR_LETTER_SA = 0x00011EF0;
const char32_t MAKASAR_LETTER_A = 0x00011EF1;
const char32_t MAKASAR_ANGKA = 0x00011EF2;
const char32_t MAKASAR_VOWEL_SIGN_I = 0x00011EF3;
const char32_t MAKASAR_VOWEL_SIGN_U = 0x00011EF4;
const char32_t MAKASAR_VOWEL_SIGN_E = 0x00011EF5;
const char32_t MAKASAR_VOWEL_SIGN_O = 0x00011EF6;
const char32_t MAKASAR_PASSIMBANG = 0x00011EF7;
const char32_t MAKASAR_END_OF_SECTION = 0x00011EF8;
const char32_t KAWI_SIGN_CANDRABINDU = 0x00011F00;
const char32_t KAWI_SIGN_ANUSVARA = 0x00011F01;
const char32_t KAWI_SIGN_REPHA = 0x00011F02;
const char32_t KAWI_SIGN_VISARGA = 0x00011F03;
const char32_t KAWI_LETTER_A = 0x00011F04;
const char32_t KAWI_LETTER_AA = 0x00011F05;
const char32_t KAWI_LETTER_I = 0x00011F06;
const char32_t KAWI_LETTER_II = 0x00011F07;
const char32_t KAWI_LETTER_U = 0x00011F08;
const char32_t KAWI_LETTER_UU = 0x00011F09;
const char32_t KAWI_LETTER_VOCALIC_R = 0x00011F0A;
const char32_t KAWI_LETTER_VOCALIC_RR = 0x00011F0B;
const char32_t KAWI_LETTER_VOCALIC_L = 0x00011F0C;
const char32_t KAWI_LETTER_VOCALIC_LL = 0x00011F0D;
const char32_t KAWI_LETTER_E = 0x00011F0E;
const char32_t KAWI_LETTER_AI = 0x00011F0F;
const char32_t KAWI_LETTER_O = 0x00011F10;
const char32_t KAWI_LETTER_KA = 0x00011F12;
const char32_t KAWI_LETTER_KHA = 0x00011F13;
const char32_t KAWI_LETTER_GA = 0x00011F14;
const char32_t KAWI_LETTER_GHA = 0x00011F15;
const char32_t KAWI_LETTER_NGA = 0x00011F16;
const char32_t KAWI_LETTER_CA = 0x00011F17;
const char32_t KAWI_LETTER_CHA = 0x00011F18;
const char32_t KAWI_LETTER_JA = 0x00011F19;
const char32_t KAWI_LETTER_JHA = 0x00011F1A;
const char32_t KAWI_LETTER_NYA = 0x00011F1B;
const char32_t KAWI_LETTER_TTA = 0x00011F1C;
const char32_t KAWI_LETTER_TTHA = 0x00011F1D;
const char32_t KAWI_LETTER_DDA = 0x00011F1E;
const char32_t KAWI_LETTER_DDHA = 0x00011F1F;
const char32_t KAWI_LETTER_NNA = 0x00011F20;
const char32_t KAWI_LETTER_TA = 0x00011F21;
const char32_t KAWI_LETTER_THA = 0x00011F22;
const char32_t KAWI_LETTER_DA = 0x00011F23;
const char32_t KAWI_LETTER_DHA = 0x00011F24;
const char32_t KAWI_LETTER_NA = 0x00011F25;
const char32_t KAWI_LETTER_PA = 0x00011F26;
const char32_t KAWI_LETTER_PHA = 0x00011F27;
const char32_t KAWI_LETTER_BA = 0x00011F28;
const char32_t KAWI_LETTER_BHA = 0x00011F29;
const char32_t KAWI_LETTER_MA = 0x00011F2A;
const char32_t KAWI_LETTER_YA = 0x00011F2B;
const char32_t KAWI_LETTER_RA = 0x00011F2C;
const char32_t KAWI_LETTER_LA = 0x00011F2D;
const char32_t KAWI_LETTER_WA = 0x00011F2E;
const char32_t KAWI_LETTER_SHA = 0x00011F2F;
const char32_t KAWI_LETTER_SSA = 0x00011F30;
const char32_t KAWI_LETTER_SA = 0x00011F31;
const char32_t KAWI_LETTER_HA = 0x00011F32;
const char32_t KAWI_LETTER_JNYA = 0x00011F33;
const char32_t KAWI_VOWEL_SIGN_AA = 0x00011F34;
const char32_t KAWI_VOWEL_SIGN_ALTERNATE_AA = 0x00011F35;
const char32_t KAWI_VOWEL_SIGN_I = 0x00011F36;
const char32_t KAWI_VOWEL_SIGN_II = 0x00011F37;
const char32_t KAWI_VOWEL_SIGN_U = 0x00011F38;
const char32_t KAWI_VOWEL_SIGN_UU = 0x00011F39;
const char32_t KAWI_VOWEL_SIGN_VOCALIC_R = 0x00011F3A;
const char32_t KAWI_VOWEL_SIGN_E = 0x00011F3E;
const char32_t KAWI_VOWEL_SIGN_AI = 0x00011F3F;
const char32_t KAWI_VOWEL_SIGN_EU = 0x00011F40;
const char32_t KAWI_SIGN_KILLER = 0x00011F41;
const char32_t KAWI_CONJOINER = 0x00011F42;
const char32_t KAWI_DANDA = 0x00011F43;
const char32_t KAWI_DOUBLE_DANDA = 0x00011F44;
const char32_t KAWI_PUNCTUATION_SECTION_MARKER = 0x00011F45;
const char32_t KAWI_PUNCTUATION_ALTERNATE_SECTION_MARKER = 0x00011F46;
const char32_t KAWI_PUNCTUATION_FLOWER = 0x00011F47;
const char32_t KAWI_PUNCTUATION_SPACE_FILLER = 0x00011F48;
const char32_t KAWI_PUNCTUATION_DOT = 0x00011F49;
const char32_t KAWI_PUNCTUATION_DOUBLE_DOT = 0x00011F4A;
const char32_t KAWI_PUNCTUATION_TRIPLE_DOT = 0x00011F4B;
const char32_t KAWI_PUNCTUATION_CIRCLE = 0x00011F4C;
const char32_t KAWI_PUNCTUATION_FILLED_CIRCLE = 0x00011F4D;
const char32_t KAWI_PUNCTUATION_SPIRAL = 0x00011F4E;
const char32_t KAWI_PUNCTUATION_CLOSING_SPIRAL = 0x00011F4F;
const char32_t KAWI_DIGIT_ZERO = 0x00011F50;
const char32_t KAWI_DIGIT_ONE = 0x00011F51;
const char32_t KAWI_DIGIT_TWO = 0x00011F52;
const char32_t KAWI_DIGIT_THREE = 0x00011F53;
const char32_t KAWI_DIGIT_FOUR = 0x00011F54;
const char32_t KAWI_DIGIT_FIVE = 0x00011F55;
const char32_t KAWI_DIGIT_SIX = 0x00011F56;
const char32_t KAWI_DIGIT_SEVEN = 0x00011F57;
const char32_t KAWI_DIGIT_EIGHT = 0x00011F58;
const char32_t KAWI_DIGIT_NINE = 0x00011F59;
const char32_t KAWI_SIGN_NUKTA = 0x00011F5A;
const char32_t LISU_LETTER_YHA = 0x00011FB0;
const char32_t TAMIL_FRACTION_ONE_THREE_HUNDRED_AND_TWENTIETH = 0x00011FC0;
const char32_t TAMIL_FRACTION_ONE_ONE_HUNDRED_AND_SIXTIETH = 0x00011FC1;
const char32_t TAMIL_FRACTION_ONE_EIGHTIETH = 0x00011FC2;
const char32_t TAMIL_FRACTION_ONE_SIXTY_FOURTH = 0x00011FC3;
const char32_t TAMIL_FRACTION_ONE_FORTIETH = 0x00011FC4;
const char32_t TAMIL_FRACTION_ONE_THIRTY_SECOND = 0x00011FC5;
const char32_t TAMIL_FRACTION_THREE_EIGHTIETHS = 0x00011FC6;
const char32_t TAMIL_FRACTION_THREE_SIXTY_FOURTHS = 0x00011FC7;
const char32_t TAMIL_FRACTION_ONE_TWENTIETH = 0x00011FC8;
const char32_t TAMIL_FRACTION_ONE_SIXTEENTH_1 = 0x00011FC9;
const char32_t TAMIL_FRACTION_ONE_SIXTEENTH_2 = 0x00011FCA;
const char32_t TAMIL_FRACTION_ONE_TENTH = 0x00011FCB;
const char32_t TAMIL_FRACTION_ONE_EIGHTH = 0x00011FCC;
const char32_t TAMIL_FRACTION_THREE_TWENTIETHS = 0x00011FCD;
const char32_t TAMIL_FRACTION_THREE_SIXTEENTHS = 0x00011FCE;
const char32_t TAMIL_FRACTION_ONE_FIFTH = 0x00011FCF;
const char32_t TAMIL_FRACTION_ONE_QUARTER = 0x00011FD0;
const char32_t TAMIL_FRACTION_ONE_HALF_1 = 0x00011FD1;
const char32_t TAMIL_FRACTION_ONE_HALF_2 = 0x00011FD2;
const char32_t TAMIL_FRACTION_THREE_QUARTERS = 0x00011FD3;
const char32_t TAMIL_FRACTION_DOWNSCALING_FACTOR_KIIZH = 0x00011FD4;
const char32_t TAMIL_SIGN_NEL = 0x00011FD5;
const char32_t TAMIL_SIGN_CEVITU = 0x00011FD6;
const char32_t TAMIL_SIGN_AAZHAAKKU = 0x00011FD7;
const char32_t TAMIL_SIGN_UZHAKKU = 0x00011FD8;
const char32_t TAMIL_SIGN_MUUVUZHAKKU = 0x00011FD9;
const char32_t TAMIL_SIGN_KURUNI = 0x00011FDA;
const char32_t TAMIL_SIGN_PATHAKKU = 0x00011FDB;
const char32_t TAMIL_SIGN_MUKKURUNI = 0x00011FDC;
const char32_t TAMIL_SIGN_KAACU = 0x00011FDD;
const char32_t TAMIL_SIGN_PANAM = 0x00011FDE;
const char32_t TAMIL_SIGN_PON = 0x00011FDF;
const char32_t TAMIL_SIGN_VARAAKAN = 0x00011FE0;
const char32_t TAMIL_SIGN_PAARAM = 0x00011FE1;
const char32_t TAMIL_SIGN_KUZHI = 0x00011FE2;
const char32_t TAMIL_SIGN_VELI = 0x00011FE3;
const char32_t TAMIL_WET_CULTIVATION_SIGN = 0x00011FE4;
const char32_t TAMIL_DRY_CULTIVATION_SIGN = 0x00011FE5;
const char32_t TAMIL_LAND_SIGN = 0x00011FE6;
const char32_t TAMIL_SALT_PAN_SIGN = 0x00011FE7;
const char32_t TAMIL_TRADITIONAL_CREDIT_SIGN = 0x00011FE8;
const char32_t TAMIL_TRADITIONAL_NUMBER_SIGN = 0x00011FE9;
const char32_t TAMIL_CURRENT_SIGN = 0x00011FEA;
const char32_t TAMIL_AND_ODD_SIGN = 0x00011FEB;
const char32_t TAMIL_SPENT_SIGN = 0x00011FEC;
const char32_t TAMIL_TOTAL_SIGN = 0x00011FED;
const char32_t TAMIL_IN_POSSESSION_SIGN = 0x00011FEE;
const char32_t TAMIL_STARTING_FROM_SIGN = 0x00011FEF;
const char32_t TAMIL_SIGN_MUTHALIYA = 0x00011FF0;
const char32_t TAMIL_SIGN_VAKAIYARAA = 0x00011FF1;
const char32_t TAMIL_PUNCTUATION_END_OF_TEXT = 0x00011FFF;
const char32_t CUNEIFORM_SIGN_A = 0x00012000;
const char32_t CUNEIFORM_SIGN_A_TIMES_A = 0x00012001;
const char32_t CUNEIFORM_SIGN_A_TIMES_BAD = 0x00012002;
const char32_t CUNEIFORM_SIGN_A_TIMES_GAN2_TENU = 0x00012003;
const char32_t CUNEIFORM_SIGN_A_TIMES_HA = 0x00012004;
const char32_t CUNEIFORM_SIGN_A_TIMES_IGI = 0x00012005;
const char32_t CUNEIFORM_SIGN_A_TIMES_LAGAR_GUNU = 0x00012006;
const char32_t CUNEIFORM_SIGN_A_TIMES_MUSH = 0x00012007;
const char32_t CUNEIFORM_SIGN_A_TIMES_SAG = 0x00012008;
const char32_t CUNEIFORM_SIGN_A2 = 0x00012009;
const char32_t CUNEIFORM_SIGN_AB = 0x0001200A;
const char32_t CUNEIFORM_SIGN_AB_TIMES_ASH2 = 0x0001200B;
const char32_t CUNEIFORM_SIGN_AB_TIMES_DUN3_GUNU = 0x0001200C;
const char32_t CUNEIFORM_SIGN_AB_TIMES_GAL = 0x0001200D;
const char32_t CUNEIFORM_SIGN_AB_TIMES_GAN2_TENU = 0x0001200E;
const char32_t CUNEIFORM_SIGN_AB_TIMES_HA = 0x0001200F;
const char32_t CUNEIFORM_SIGN_AB_TIMES_IGI_GUNU = 0x00012010;
const char32_t CUNEIFORM_SIGN_AB_TIMES_IMIN = 0x00012011;
const char32_t CUNEIFORM_SIGN_AB_TIMES_LAGAB = 0x00012012;
const char32_t CUNEIFORM_SIGN_AB_TIMES_SHESH = 0x00012013;
const char32_t CUNEIFORM_SIGN_AB_TIMES_U_PLUS_U_PLUS_U = 0x00012014;
const char32_t CUNEIFORM_SIGN_AB_GUNU = 0x00012015;
const char32_t CUNEIFORM_SIGN_AB2 = 0x00012016;
const char32_t CUNEIFORM_SIGN_AB2_TIMES_BALAG = 0x00012017;
const char32_t CUNEIFORM_SIGN_AB2_TIMES_GAN2_TENU = 0x00012018;
const char32_t CUNEIFORM_SIGN_AB2_TIMES_ME_PLUS_EN = 0x00012019;
const char32_t CUNEIFORM_SIGN_AB2_TIMES_SHA3 = 0x0001201A;
const char32_t CUNEIFORM_SIGN_AB2_TIMES_TAK4 = 0x0001201B;
const char32_t CUNEIFORM_SIGN_AD = 0x0001201C;
const char32_t CUNEIFORM_SIGN_AK = 0x0001201D;
const char32_t CUNEIFORM_SIGN_AK_TIMES_ERIN2 = 0x0001201E;
const char32_t CUNEIFORM_SIGN_AK_TIMES_SHITA_PLUS_GISH = 0x0001201F;
const char32_t CUNEIFORM_SIGN_AL = 0x00012020;
const char32_t CUNEIFORM_SIGN_AL_TIMES_AL = 0x00012021;
const char32_t CUNEIFORM_SIGN_AL_TIMES_DIM2 = 0x00012022;
const char32_t CUNEIFORM_SIGN_AL_TIMES_GISH = 0x00012023;
const char32_t CUNEIFORM_SIGN_AL_TIMES_HA = 0x00012024;
const char32_t CUNEIFORM_SIGN_AL_TIMES_KAD3 = 0x00012025;
const char32_t CUNEIFORM_SIGN_AL_TIMES_KI = 0x00012026;
const char32_t CUNEIFORM_SIGN_AL_TIMES_SHE = 0x00012027;
const char32_t CUNEIFORM_SIGN_AL_TIMES_USH = 0x00012028;
const char32_t CUNEIFORM_SIGN_ALAN = 0x00012029;
const char32_t CUNEIFORM_SIGN_ALEPH = 0x0001202A;
const char32_t CUNEIFORM_SIGN_AMAR = 0x0001202B;
const char32_t CUNEIFORM_SIGN_AMAR_TIMES_SHE = 0x0001202C;
const char32_t CUNEIFORM_SIGN_AN = 0x0001202D;
const char32_t CUNEIFORM_SIGN_AN_OVER_AN = 0x0001202E;
const char32_t CUNEIFORM_SIGN_AN_THREE_TIMES = 0x0001202F;
const char32_t CUNEIFORM_SIGN_AN_PLUS_NAGA_OPPOSING_AN_PLUS_NAGA = 0x00012030;
const char32_t CUNEIFORM_SIGN_AN_PLUS_NAGA_SQUARED = 0x00012031;
const char32_t CUNEIFORM_SIGN_ANSHE = 0x00012032;
const char32_t CUNEIFORM_SIGN_APIN = 0x00012033;
const char32_t CUNEIFORM_SIGN_ARAD = 0x00012034;
const char32_t CUNEIFORM_SIGN_ARAD_TIMES_KUR = 0x00012035;
const char32_t CUNEIFORM_SIGN_ARKAB = 0x00012036;
const char32_t CUNEIFORM_SIGN_ASAL2 = 0x00012037;
const char32_t CUNEIFORM_SIGN_ASH = 0x00012038;
const char32_t CUNEIFORM_SIGN_ASH_ZIDA_TENU = 0x00012039;
const char32_t CUNEIFORM_SIGN_ASH_KABA_TENU = 0x0001203A;
const char32_t CUNEIFORM_SIGN_ASH_OVER_ASH_TUG2_OVER_TUG2_TUG2_OVER_TUG2_PAP = 0x0001203B;
const char32_t CUNEIFORM_SIGN_ASH_OVER_ASH_OVER_ASH = 0x0001203C;
const char32_t CUNEIFORM_SIGN_ASH_OVER_ASH_OVER_ASH_CROSSING_ASH_OVER_ASH_OVER_ASH = 0x0001203D;
const char32_t CUNEIFORM_SIGN_ASH2 = 0x0001203E;
const char32_t CUNEIFORM_SIGN_ASHGAB = 0x0001203F;
const char32_t CUNEIFORM_SIGN_BA = 0x00012040;
const char32_t CUNEIFORM_SIGN_BAD = 0x00012041;
const char32_t CUNEIFORM_SIGN_BAG3 = 0x00012042;
const char32_t CUNEIFORM_SIGN_BAHAR2 = 0x00012043;
const char32_t CUNEIFORM_SIGN_BAL = 0x00012044;
const char32_t CUNEIFORM_SIGN_BAL_OVER_BAL = 0x00012045;
const char32_t CUNEIFORM_SIGN_BALAG = 0x00012046;
const char32_t CUNEIFORM_SIGN_BAR = 0x00012047;
const char32_t CUNEIFORM_SIGN_BARA2 = 0x00012048;
const char32_t CUNEIFORM_SIGN_BI = 0x00012049;
const char32_t CUNEIFORM_SIGN_BI_TIMES_A = 0x0001204A;
const char32_t CUNEIFORM_SIGN_BI_TIMES_GAR = 0x0001204B;
const char32_t CUNEIFORM_SIGN_BI_TIMES_IGI_GUNU = 0x0001204C;
const char32_t CUNEIFORM_SIGN_BU = 0x0001204D;
const char32_t CUNEIFORM_SIGN_BU_OVER_BU_AB = 0x0001204E;
const char32_t CUNEIFORM_SIGN_BU_OVER_BU_UN = 0x0001204F;
const char32_t CUNEIFORM_SIGN_BU_CROSSING_BU = 0x00012050;
const char32_t CUNEIFORM_SIGN_BULUG = 0x00012051;
const char32_t CUNEIFORM_SIGN_BULUG_OVER_BULUG = 0x00012052;
const char32_t CUNEIFORM_SIGN_BUR = 0x00012053;
const char32_t CUNEIFORM_SIGN_BUR2 = 0x00012054;
const char32_t CUNEIFORM_SIGN_DA = 0x00012055;
const char32_t CUNEIFORM_SIGN_DAG = 0x00012056;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_A_PLUS_MASH = 0x00012057;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_AMAR = 0x00012058;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_BALAG = 0x00012059;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_BI = 0x0001205A;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_GA = 0x0001205B;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_GA_PLUS_MASH = 0x0001205C;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_GI = 0x0001205D;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_GIR2 = 0x0001205E;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_GUD = 0x0001205F;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_HA = 0x00012060;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_IR = 0x00012061;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_IR_PLUS_LU = 0x00012062;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_KAK = 0x00012063;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_LA = 0x00012064;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_LU = 0x00012065;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_LU_PLUS_MASH2 = 0x00012066;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_LUM = 0x00012067;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_NE = 0x00012068;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_PAP_PLUS_PAP = 0x00012069;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_SI = 0x0001206A;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_TAK4 = 0x0001206B;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_U2_PLUS_GIR2 = 0x0001206C;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_USH = 0x0001206D;
const char32_t CUNEIFORM_SIGN_DAM = 0x0001206E;
const char32_t CUNEIFORM_SIGN_DAR = 0x0001206F;
const char32_t CUNEIFORM_SIGN_DARA3 = 0x00012070;
const char32_t CUNEIFORM_SIGN_DARA4 = 0x00012071;
const char32_t CUNEIFORM_SIGN_DI = 0x00012072;
const char32_t CUNEIFORM_SIGN_DIB = 0x00012073;
const char32_t CUNEIFORM_SIGN_DIM = 0x00012074;
const char32_t CUNEIFORM_SIGN_DIM_TIMES_SHE = 0x00012075;
const char32_t CUNEIFORM_SIGN_DIM2 = 0x00012076;
const char32_t CUNEIFORM_SIGN_DIN = 0x00012077;
const char32_t CUNEIFORM_SIGN_DIN_KASKAL_U_GUNU_DISH = 0x00012078;
const char32_t CUNEIFORM_SIGN_DISH = 0x00012079;
const char32_t CUNEIFORM_SIGN_DU = 0x0001207A;
const char32_t CUNEIFORM_SIGN_DU_OVER_DU = 0x0001207B;
const char32_t CUNEIFORM_SIGN_DU_GUNU = 0x0001207C;
const char32_t CUNEIFORM_SIGN_DU_SHESHIG = 0x0001207D;
const char32_t CUNEIFORM_SIGN_DUB = 0x0001207E;
const char32_t CUNEIFORM_SIGN_DUB_TIMES_ESH2 = 0x0001207F;
const char32_t CUNEIFORM_SIGN_DUB2 = 0x00012080;
const char32_t CUNEIFORM_SIGN_DUG = 0x00012081;
const char32_t CUNEIFORM_SIGN_DUGUD = 0x00012082;
const char32_t CUNEIFORM_SIGN_DUH = 0x00012083;
const char32_t CUNEIFORM_SIGN_DUN = 0x00012084;
const char32_t CUNEIFORM_SIGN_DUN3 = 0x00012085;
const char32_t CUNEIFORM_SIGN_DUN3_GUNU = 0x00012086;
const char32_t CUNEIFORM_SIGN_DUN3_GUNU_GUNU = 0x00012087;
const char32_t CUNEIFORM_SIGN_DUN4 = 0x00012088;
const char32_t CUNEIFORM_SIGN_DUR2 = 0x00012089;
const char32_t CUNEIFORM_SIGN_E = 0x0001208A;
const char32_t CUNEIFORM_SIGN_E_TIMES_PAP = 0x0001208B;
const char32_t CUNEIFORM_SIGN_E_OVER_E_NUN_OVER_NUN = 0x0001208C;
const char32_t CUNEIFORM_SIGN_E2 = 0x0001208D;
const char32_t CUNEIFORM_SIGN_E2_TIMES_A_PLUS_HA_PLUS_DA = 0x0001208E;
const char32_t CUNEIFORM_SIGN_E2_TIMES_GAR = 0x0001208F;
const char32_t CUNEIFORM_SIGN_E2_TIMES_MI = 0x00012090;
const char32_t CUNEIFORM_SIGN_E2_TIMES_SAL = 0x00012091;
const char32_t CUNEIFORM_SIGN_E2_TIMES_SHE = 0x00012092;
const char32_t CUNEIFORM_SIGN_E2_TIMES_U = 0x00012093;
const char32_t CUNEIFORM_SIGN_EDIN = 0x00012094;
const char32_t CUNEIFORM_SIGN_EGIR = 0x00012095;
const char32_t CUNEIFORM_SIGN_EL = 0x00012096;
const char32_t CUNEIFORM_SIGN_EN = 0x00012097;
const char32_t CUNEIFORM_SIGN_EN_TIMES_GAN2 = 0x00012098;
const char32_t CUNEIFORM_SIGN_EN_TIMES_GAN2_TENU = 0x00012099;
const char32_t CUNEIFORM_SIGN_EN_TIMES_ME = 0x0001209A;
const char32_t CUNEIFORM_SIGN_EN_CROSSING_EN = 0x0001209B;
const char32_t CUNEIFORM_SIGN_EN_OPPOSING_EN = 0x0001209C;
const char32_t CUNEIFORM_SIGN_EN_SQUARED = 0x0001209D;
const char32_t CUNEIFORM_SIGN_EREN = 0x0001209E;
const char32_t CUNEIFORM_SIGN_ERIN2 = 0x0001209F;
const char32_t CUNEIFORM_SIGN_ESH2 = 0x000120A0;
const char32_t CUNEIFORM_SIGN_EZEN = 0x000120A1;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_A = 0x000120A2;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_A_PLUS_LAL = 0x000120A3;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_A_PLUS_LAL_TIMES_LAL = 0x000120A4;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_AN = 0x000120A5;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_BAD = 0x000120A6;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_DUN3_GUNU = 0x000120A7;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_DUN3_GUNU_GUNU = 0x000120A8;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_HA = 0x000120A9;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_HA_GUNU = 0x000120AA;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_IGI_GUNU = 0x000120AB;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_KASKAL = 0x000120AC;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_KASKAL_SQUARED = 0x000120AD;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_KU3 = 0x000120AE;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_LA = 0x000120AF;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_LAL_TIMES_LAL = 0x000120B0;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_LI = 0x000120B1;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_LU = 0x000120B2;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_U2 = 0x000120B3;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_UD = 0x000120B4;
const char32_t CUNEIFORM_SIGN_GA = 0x000120B5;
const char32_t CUNEIFORM_SIGN_GA_GUNU = 0x000120B6;
const char32_t CUNEIFORM_SIGN_GA2 = 0x000120B7;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_A_PLUS_DA_PLUS_HA = 0x000120B8;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_A_PLUS_HA = 0x000120B9;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_A_PLUS_IGI = 0x000120BA;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_AB2_TENU_PLUS_TAB = 0x000120BB;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_AN = 0x000120BC;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_ASH = 0x000120BD;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_ASH2_PLUS_GAL = 0x000120BE;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_BAD = 0x000120BF;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_BAR_PLUS_RA = 0x000120C0;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_BUR = 0x000120C1;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_BUR_PLUS_RA = 0x000120C2;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_DA = 0x000120C3;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_DI = 0x000120C4;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_DIM_TIMES_SHE = 0x000120C5;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_DUB = 0x000120C6;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_EL = 0x000120C7;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_EL_PLUS_LA = 0x000120C8;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_EN = 0x000120C9;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_EN_TIMES_GAN2_TENU = 0x000120CA;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GAN2_TENU = 0x000120CB;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GAR = 0x000120CC;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GI = 0x000120CD;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GI4 = 0x000120CE;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GI4_PLUS_A = 0x000120CF;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GIR2_PLUS_SU = 0x000120D0;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_HA_PLUS_LU_PLUS_ESH2 = 0x000120D1;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_HAL = 0x000120D2;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_HAL_PLUS_LA = 0x000120D3;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_HI_PLUS_LI = 0x000120D4;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_HUB2 = 0x000120D5;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_IGI_GUNU = 0x000120D6;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_ISH_PLUS_HU_PLUS_ASH = 0x000120D7;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_KAK = 0x000120D8;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_KASKAL = 0x000120D9;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_KID = 0x000120DA;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_KID_PLUS_LAL = 0x000120DB;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_KU3_PLUS_AN = 0x000120DC;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_LA = 0x000120DD;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_ME_PLUS_EN = 0x000120DE;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_MI = 0x000120DF;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_NUN = 0x000120E0;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_NUN_OVER_NUN = 0x000120E1;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_PA = 0x000120E2;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_SAL = 0x000120E3;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_SAR = 0x000120E4;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_SHE = 0x000120E5;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_SHE_PLUS_TUR = 0x000120E6;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_SHID = 0x000120E7;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_SUM = 0x000120E8;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_TAK4 = 0x000120E9;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_U = 0x000120EA;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_UD = 0x000120EB;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_UD_PLUS_DU = 0x000120EC;
const char32_t CUNEIFORM_SIGN_GA2_OVER_GA2 = 0x000120ED;
const char32_t CUNEIFORM_SIGN_GABA = 0x000120EE;
const char32_t CUNEIFORM_SIGN_GABA_CROSSING_GABA = 0x000120EF;
const char32_t CUNEIFORM_SIGN_GAD = 0x000120F0;
const char32_t CUNEIFORM_SIGN_GAD_OVER_GAD_GAR_OVER_GAR = 0x000120F1;
const char32_t CUNEIFORM_SIGN_GAL = 0x000120F2;
const char32_t CUNEIFORM_SIGN_GAL_GAD_OVER_GAD_GAR_OVER_GAR = 0x000120F3;
const char32_t CUNEIFORM_SIGN_GALAM = 0x000120F4;
const char32_t CUNEIFORM_SIGN_GAM = 0x000120F5;
const char32_t CUNEIFORM_SIGN_GAN = 0x000120F6;
const char32_t CUNEIFORM_SIGN_GAN2 = 0x000120F7;
const char32_t CUNEIFORM_SIGN_GAN2_TENU = 0x000120F8;
const char32_t CUNEIFORM_SIGN_GAN2_OVER_GAN2 = 0x000120F9;
const char32_t CUNEIFORM_SIGN_GAN2_CROSSING_GAN2 = 0x000120FA;
const char32_t CUNEIFORM_SIGN_GAR = 0x000120FB;
const char32_t CUNEIFORM_SIGN_GAR3 = 0x000120FC;
const char32_t CUNEIFORM_SIGN_GASHAN = 0x000120FD;
const char32_t CUNEIFORM_SIGN_GESHTIN = 0x000120FE;
const char32_t CUNEIFORM_SIGN_GESHTIN_TIMES_KUR = 0x000120FF;
const char32_t CUNEIFORM_SIGN_GI = 0x00012100;
const char32_t CUNEIFORM_SIGN_GI_TIMES_E = 0x00012101;
const char32_t CUNEIFORM_SIGN_GI_TIMES_U = 0x00012102;
const char32_t CUNEIFORM_SIGN_GI_CROSSING_GI = 0x00012103;
const char32_t CUNEIFORM_SIGN_GI4 = 0x00012104;
const char32_t CUNEIFORM_SIGN_GI4_OVER_GI4 = 0x00012105;
const char32_t CUNEIFORM_SIGN_GI4_CROSSING_GI4 = 0x00012106;
const char32_t CUNEIFORM_SIGN_GIDIM = 0x00012107;
const char32_t CUNEIFORM_SIGN_GIR2 = 0x00012108;
const char32_t CUNEIFORM_SIGN_GIR2_GUNU = 0x00012109;
const char32_t CUNEIFORM_SIGN_GIR3 = 0x0001210A;
const char32_t CUNEIFORM_SIGN_GIR3_TIMES_A_PLUS_IGI = 0x0001210B;
const char32_t CUNEIFORM_SIGN_GIR3_TIMES_GAN2_TENU = 0x0001210C;
const char32_t CUNEIFORM_SIGN_GIR3_TIMES_IGI = 0x0001210D;
const char32_t CUNEIFORM_SIGN_GIR3_TIMES_LU_PLUS_IGI = 0x0001210E;
const char32_t CUNEIFORM_SIGN_GIR3_TIMES_PA = 0x0001210F;
const char32_t CUNEIFORM_SIGN_GISAL = 0x00012110;
const char32_t CUNEIFORM_SIGN_GISH = 0x00012111;
const char32_t CUNEIFORM_SIGN_GISH_CROSSING_GISH = 0x00012112;
const char32_t CUNEIFORM_SIGN_GISH_TIMES_BAD = 0x00012113;
const char32_t CUNEIFORM_SIGN_GISH_TIMES_TAK4 = 0x00012114;
const char32_t CUNEIFORM_SIGN_GISH_TENU = 0x00012115;
const char32_t CUNEIFORM_SIGN_GU = 0x00012116;
const char32_t CUNEIFORM_SIGN_GU_CROSSING_GU = 0x00012117;
const char32_t CUNEIFORM_SIGN_GU2 = 0x00012118;
const char32_t CUNEIFORM_SIGN_GU2_TIMES_KAK = 0x00012119;
const char32_t CUNEIFORM_SIGN_GU2_TIMES_KAK_TIMES_IGI_GUNU = 0x0001211A;
const char32_t CUNEIFORM_SIGN_GU2_TIMES_NUN = 0x0001211B;
const char32_t CUNEIFORM_SIGN_GU2_TIMES_SAL_PLUS_TUG2 = 0x0001211C;
const char32_t CUNEIFORM_SIGN_GU2_GUNU = 0x0001211D;
const char32_t CUNEIFORM_SIGN_GUD = 0x0001211E;
const char32_t CUNEIFORM_SIGN_GUD_TIMES_A_PLUS_KUR = 0x0001211F;
const char32_t CUNEIFORM_SIGN_GUD_TIMES_KUR = 0x00012120;
const char32_t CUNEIFORM_SIGN_GUD_OVER_GUD_LUGAL = 0x00012121;
const char32_t CUNEIFORM_SIGN_GUL = 0x00012122;
const char32_t CUNEIFORM_SIGN_GUM = 0x00012123;
const char32_t CUNEIFORM_SIGN_GUM_TIMES_SHE = 0x00012124;
const char32_t CUNEIFORM_SIGN_GUR = 0x00012125;
const char32_t CUNEIFORM_SIGN_GUR7 = 0x00012126;
const char32_t CUNEIFORM_SIGN_GURUN = 0x00012127;
const char32_t CUNEIFORM_SIGN_GURUSH = 0x00012128;
const char32_t CUNEIFORM_SIGN_HA = 0x00012129;
const char32_t CUNEIFORM_SIGN_HA_TENU = 0x0001212A;
const char32_t CUNEIFORM_SIGN_HA_GUNU = 0x0001212B;
const char32_t CUNEIFORM_SIGN_HAL = 0x0001212C;
const char32_t CUNEIFORM_SIGN_HI = 0x0001212D;
const char32_t CUNEIFORM_SIGN_HI_TIMES_ASH = 0x0001212E;
const char32_t CUNEIFORM_SIGN_HI_TIMES_ASH2 = 0x0001212F;
const char32_t CUNEIFORM_SIGN_HI_TIMES_BAD = 0x00012130;
const char32_t CUNEIFORM_SIGN_HI_TIMES_DISH = 0x00012131;
const char32_t CUNEIFORM_SIGN_HI_TIMES_GAD = 0x00012132;
const char32_t CUNEIFORM_SIGN_HI_TIMES_KIN = 0x00012133;
const char32_t CUNEIFORM_SIGN_HI_TIMES_NUN = 0x00012134;
const char32_t CUNEIFORM_SIGN_HI_TIMES_SHE = 0x00012135;
const char32_t CUNEIFORM_SIGN_HI_TIMES_U = 0x00012136;
const char32_t CUNEIFORM_SIGN_HU = 0x00012137;
const char32_t CUNEIFORM_SIGN_HUB2 = 0x00012138;
const char32_t CUNEIFORM_SIGN_HUB2_TIMES_AN = 0x00012139;
const char32_t CUNEIFORM_SIGN_HUB2_TIMES_HAL = 0x0001213A;
const char32_t CUNEIFORM_SIGN_HUB2_TIMES_KASKAL = 0x0001213B;
const char32_t CUNEIFORM_SIGN_HUB2_TIMES_LISH = 0x0001213C;
const char32_t CUNEIFORM_SIGN_HUB2_TIMES_UD = 0x0001213D;
const char32_t CUNEIFORM_SIGN_HUL2 = 0x0001213E;
const char32_t CUNEIFORM_SIGN_I = 0x0001213F;
const char32_t CUNEIFORM_SIGN_I_A = 0x00012140;
const char32_t CUNEIFORM_SIGN_IB = 0x00012141;
const char32_t CUNEIFORM_SIGN_IDIM = 0x00012142;
const char32_t CUNEIFORM_SIGN_IDIM_OVER_IDIM_BUR = 0x00012143;
const char32_t CUNEIFORM_SIGN_IDIM_OVER_IDIM_SQUARED = 0x00012144;
const char32_t CUNEIFORM_SIGN_IG = 0x00012145;
const char32_t CUNEIFORM_SIGN_IGI = 0x00012146;
const char32_t CUNEIFORM_SIGN_IGI_DIB = 0x00012147;
const char32_t CUNEIFORM_SIGN_IGI_RI = 0x00012148;
const char32_t CUNEIFORM_SIGN_IGI_OVER_IGI_SHIR_OVER_SHIR_UD_OVER_UD = 0x00012149;
const char32_t CUNEIFORM_SIGN_IGI_GUNU = 0x0001214A;
const char32_t CUNEIFORM_SIGN_IL = 0x0001214B;
const char32_t CUNEIFORM_SIGN_IL_TIMES_GAN2_TENU = 0x0001214C;
const char32_t CUNEIFORM_SIGN_IL2 = 0x0001214D;
const char32_t CUNEIFORM_SIGN_IM = 0x0001214E;
const char32_t CUNEIFORM_SIGN_IM_TIMES_TAK4 = 0x0001214F;
const char32_t CUNEIFORM_SIGN_IM_CROSSING_IM = 0x00012150;
const char32_t CUNEIFORM_SIGN_IM_OPPOSING_IM = 0x00012151;
const char32_t CUNEIFORM_SIGN_IM_SQUARED = 0x00012152;
const char32_t CUNEIFORM_SIGN_IMIN = 0x00012153;
const char32_t CUNEIFORM_SIGN_IN = 0x00012154;
const char32_t CUNEIFORM_SIGN_IR = 0x00012155;
const char32_t CUNEIFORM_SIGN_ISH = 0x00012156;
const char32_t CUNEIFORM_SIGN_KA = 0x00012157;
const char32_t CUNEIFORM_SIGN_KA_TIMES_A = 0x00012158;
const char32_t CUNEIFORM_SIGN_KA_TIMES_AD = 0x00012159;
const char32_t CUNEIFORM_SIGN_KA_TIMES_AD_PLUS_KU3 = 0x0001215A;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ASH2 = 0x0001215B;
const char32_t CUNEIFORM_SIGN_KA_TIMES_BAD = 0x0001215C;
const char32_t CUNEIFORM_SIGN_KA_TIMES_BALAG = 0x0001215D;
const char32_t CUNEIFORM_SIGN_KA_TIMES_BAR = 0x0001215E;
const char32_t CUNEIFORM_SIGN_KA_TIMES_BI = 0x0001215F;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ERIN2 = 0x00012160;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ESH2 = 0x00012161;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GA = 0x00012162;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GAL = 0x00012163;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GAN2_TENU = 0x00012164;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GAR = 0x00012165;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GAR_PLUS_SHA3_PLUS_A = 0x00012166;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GI = 0x00012167;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GIR2 = 0x00012168;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GISH_PLUS_SAR = 0x00012169;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GISH_CROSSING_GISH = 0x0001216A;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GU = 0x0001216B;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GUR7 = 0x0001216C;
const char32_t CUNEIFORM_SIGN_KA_TIMES_IGI = 0x0001216D;
const char32_t CUNEIFORM_SIGN_KA_TIMES_IM = 0x0001216E;
const char32_t CUNEIFORM_SIGN_KA_TIMES_KAK = 0x0001216F;
const char32_t CUNEIFORM_SIGN_KA_TIMES_KI = 0x00012170;
const char32_t CUNEIFORM_SIGN_KA_TIMES_KID = 0x00012171;
const char32_t CUNEIFORM_SIGN_KA_TIMES_LI = 0x00012172;
const char32_t CUNEIFORM_SIGN_KA_TIMES_LU = 0x00012173;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ME = 0x00012174;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ME_PLUS_DU = 0x00012175;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ME_PLUS_GI = 0x00012176;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ME_PLUS_TE = 0x00012177;
const char32_t CUNEIFORM_SIGN_KA_TIMES_MI = 0x00012178;
const char32_t CUNEIFORM_SIGN_KA_TIMES_MI_PLUS_NUNUZ = 0x00012179;
const char32_t CUNEIFORM_SIGN_KA_TIMES_NE = 0x0001217A;
const char32_t CUNEIFORM_SIGN_KA_TIMES_NUN = 0x0001217B;
const char32_t CUNEIFORM_SIGN_KA_TIMES_PI = 0x0001217C;
const char32_t CUNEIFORM_SIGN_KA_TIMES_RU = 0x0001217D;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SA = 0x0001217E;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SAR = 0x0001217F;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SHA = 0x00012180;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SHE = 0x00012181;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SHID = 0x00012182;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SHU = 0x00012183;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SIG = 0x00012184;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SUHUR = 0x00012185;
const char32_t CUNEIFORM_SIGN_KA_TIMES_TAR = 0x00012186;
const char32_t CUNEIFORM_SIGN_KA_TIMES_U = 0x00012187;
const char32_t CUNEIFORM_SIGN_KA_TIMES_U2 = 0x00012188;
const char32_t CUNEIFORM_SIGN_KA_TIMES_UD = 0x00012189;
const char32_t CUNEIFORM_SIGN_KA_TIMES_UMUM_TIMES_PA = 0x0001218A;
const char32_t CUNEIFORM_SIGN_KA_TIMES_USH = 0x0001218B;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ZI = 0x0001218C;
const char32_t CUNEIFORM_SIGN_KA2 = 0x0001218D;
const char32_t CUNEIFORM_SIGN_KA2_CROSSING_KA2 = 0x0001218E;
const char32_t CUNEIFORM_SIGN_KAB = 0x0001218F;
const char32_t CUNEIFORM_SIGN_KAD2 = 0x00012190;
const char32_t CUNEIFORM_SIGN_KAD3 = 0x00012191;
const char32_t CUNEIFORM_SIGN_KAD4 = 0x00012192;
const char32_t CUNEIFORM_SIGN_KAD5 = 0x00012193;
const char32_t CUNEIFORM_SIGN_KAD5_OVER_KAD5 = 0x00012194;
const char32_t CUNEIFORM_SIGN_KAK = 0x00012195;
const char32_t CUNEIFORM_SIGN_KAK_TIMES_IGI_GUNU = 0x00012196;
const char32_t CUNEIFORM_SIGN_KAL = 0x00012197;
const char32_t CUNEIFORM_SIGN_KAL_TIMES_BAD = 0x00012198;
const char32_t CUNEIFORM_SIGN_KAL_CROSSING_KAL = 0x00012199;
const char32_t CUNEIFORM_SIGN_KAM2 = 0x0001219A;
const char32_t CUNEIFORM_SIGN_KAM4 = 0x0001219B;
const char32_t CUNEIFORM_SIGN_KASKAL = 0x0001219C;
const char32_t CUNEIFORM_SIGN_KASKAL_LAGAB_TIMES_U_OVER_LAGAB_TIMES_U = 0x0001219D;
const char32_t CUNEIFORM_SIGN_KASKAL_OVER_KASKAL_LAGAB_TIMES_U_OVER_LAGAB_TIMES_U = 0x0001219E;
const char32_t CUNEIFORM_SIGN_KESH2 = 0x0001219F;
const char32_t CUNEIFORM_SIGN_KI = 0x000121A0;
const char32_t CUNEIFORM_SIGN_KI_TIMES_BAD = 0x000121A1;
const char32_t CUNEIFORM_SIGN_KI_TIMES_U = 0x000121A2;
const char32_t CUNEIFORM_SIGN_KI_TIMES_UD = 0x000121A3;
const char32_t CUNEIFORM_SIGN_KID = 0x000121A4;
const char32_t CUNEIFORM_SIGN_KIN = 0x000121A5;
const char32_t CUNEIFORM_SIGN_KISAL = 0x000121A6;
const char32_t CUNEIFORM_SIGN_KISH = 0x000121A7;
const char32_t CUNEIFORM_SIGN_KISIM5 = 0x000121A8;
const char32_t CUNEIFORM_SIGN_KISIM5_OVER_KISIM5 = 0x000121A9;
const char32_t CUNEIFORM_SIGN_KU = 0x000121AA;
const char32_t CUNEIFORM_SIGN_KU_OVER_HI_TIMES_ASH2_KU_OVER_HI_TIMES_ASH2 = 0x000121AB;
const char32_t CUNEIFORM_SIGN_KU3 = 0x000121AC;
const char32_t CUNEIFORM_SIGN_KU4 = 0x000121AD;
const char32_t CUNEIFORM_SIGN_KU4_VARIANT_FORM = 0x000121AE;
const char32_t CUNEIFORM_SIGN_KU7 = 0x000121AF;
const char32_t CUNEIFORM_SIGN_KUL = 0x000121B0;
const char32_t CUNEIFORM_SIGN_KUL_GUNU = 0x000121B1;
const char32_t CUNEIFORM_SIGN_KUN = 0x000121B2;
const char32_t CUNEIFORM_SIGN_KUR = 0x000121B3;
const char32_t CUNEIFORM_SIGN_KUR_OPPOSING_KUR = 0x000121B4;
const char32_t CUNEIFORM_SIGN_KUSHU2 = 0x000121B5;
const char32_t CUNEIFORM_SIGN_KWU318 = 0x000121B6;
const char32_t CUNEIFORM_SIGN_LA = 0x000121B7;
const char32_t CUNEIFORM_SIGN_LAGAB = 0x000121B8;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_A = 0x000121B9;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_A_PLUS_DA_PLUS_HA = 0x000121BA;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_A_PLUS_GAR = 0x000121BB;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_A_PLUS_LAL = 0x000121BC;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_AL = 0x000121BD;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_AN = 0x000121BE;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_ASH_ZIDA_TENU = 0x000121BF;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_BAD = 0x000121C0;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_BI = 0x000121C1;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_DAR = 0x000121C2;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_EN = 0x000121C3;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_GA = 0x000121C4;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_GAR = 0x000121C5;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_GUD = 0x000121C6;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_GUD_PLUS_GUD = 0x000121C7;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_HA = 0x000121C8;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_HAL = 0x000121C9;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_HI_TIMES_NUN = 0x000121CA;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_IGI_GUNU = 0x000121CB;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_IM = 0x000121CC;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_IM_PLUS_HA = 0x000121CD;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_IM_PLUS_LU = 0x000121CE;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_KI = 0x000121CF;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_KIN = 0x000121D0;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_KU3 = 0x000121D1;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_KUL = 0x000121D2;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_KUL_PLUS_HI_PLUS_A = 0x000121D3;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_LAGAB = 0x000121D4;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_LISH = 0x000121D5;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_LU = 0x000121D6;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_LUL = 0x000121D7;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_ME = 0x000121D8;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_ME_PLUS_EN = 0x000121D9;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_MUSH = 0x000121DA;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_NE = 0x000121DB;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_SHE_PLUS_SUM = 0x000121DC;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_SHITA_PLUS_GISH_PLUS_ERIN2 = 0x000121DD;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_SHITA_PLUS_GISH_TENU = 0x000121DE;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_SHU2 = 0x000121DF;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_SHU2_PLUS_SHU2 = 0x000121E0;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_SUM = 0x000121E1;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_TAG = 0x000121E2;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_TAK4 = 0x000121E3;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_TE_PLUS_A_PLUS_SU_PLUS_NA = 0x000121E4;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_U = 0x000121E5;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_U_PLUS_A = 0x000121E6;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_U_PLUS_U_PLUS_U = 0x000121E7;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_U2_PLUS_ASH = 0x000121E8;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_UD = 0x000121E9;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_USH = 0x000121EA;
const char32_t CUNEIFORM_SIGN_LAGAB_SQUARED = 0x000121EB;
const char32_t CUNEIFORM_SIGN_LAGAR = 0x000121EC;
const char32_t CUNEIFORM_SIGN_LAGAR_TIMES_SHE = 0x000121ED;
const char32_t CUNEIFORM_SIGN_LAGAR_TIMES_SHE_PLUS_SUM = 0x000121EE;
const char32_t CUNEIFORM_SIGN_LAGAR_GUNU = 0x000121EF;
const char32_t CUNEIFORM_SIGN_LAGAR_GUNU_OVER_LAGAR_GUNU_SHE = 0x000121F0;
const char32_t CUNEIFORM_SIGN_LAHSHU = 0x000121F1;
const char32_t CUNEIFORM_SIGN_LAL = 0x000121F2;
const char32_t CUNEIFORM_SIGN_LAL_TIMES_LAL = 0x000121F3;
const char32_t CUNEIFORM_SIGN_LAM = 0x000121F4;
const char32_t CUNEIFORM_SIGN_LAM_TIMES_KUR = 0x000121F5;
const char32_t CUNEIFORM_SIGN_LAM_TIMES_KUR_PLUS_RU = 0x000121F6;
const char32_t CUNEIFORM_SIGN_LI = 0x000121F7;
const char32_t CUNEIFORM_SIGN_LIL = 0x000121F8;
const char32_t CUNEIFORM_SIGN_LIMMU2 = 0x000121F9;
const char32_t CUNEIFORM_SIGN_LISH = 0x000121FA;
const char32_t CUNEIFORM_SIGN_LU = 0x000121FB;
const char32_t CUNEIFORM_SIGN_LU_TIMES_BAD = 0x000121FC;
const char32_t CUNEIFORM_SIGN_LU2 = 0x000121FD;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_AL = 0x000121FE;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_BAD = 0x000121FF;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_ESH2 = 0x00012200;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_ESH2_TENU = 0x00012201;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_GAN2_TENU = 0x00012202;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_HI_TIMES_BAD = 0x00012203;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_IM = 0x00012204;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_KAD2 = 0x00012205;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_KAD3 = 0x00012206;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_KAD3_PLUS_ASH = 0x00012207;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_KI = 0x00012208;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_LA_PLUS_ASH = 0x00012209;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_LAGAB = 0x0001220A;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_ME_PLUS_EN = 0x0001220B;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_NE = 0x0001220C;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_NU = 0x0001220D;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_SI_PLUS_ASH = 0x0001220E;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_SIK2_PLUS_BU = 0x0001220F;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_TUG2 = 0x00012210;
const char32_t CUNEIFORM_SIGN_LU2_TENU = 0x00012211;
const char32_t CUNEIFORM_SIGN_LU2_CROSSING_LU2 = 0x00012212;
const char32_t CUNEIFORM_SIGN_LU2_OPPOSING_LU2 = 0x00012213;
const char32_t CUNEIFORM_SIGN_LU2_SQUARED = 0x00012214;
const char32_t CUNEIFORM_SIGN_LU2_SHESHIG = 0x00012215;
const char32_t CUNEIFORM_SIGN_LU3 = 0x00012216;
const char32_t CUNEIFORM_SIGN_LUGAL = 0x00012217;
const char32_t CUNEIFORM_SIGN_LUGAL_OVER_LUGAL = 0x00012218;
const char32_t CUNEIFORM_SIGN_LUGAL_OPPOSING_LUGAL = 0x00012219;
const char32_t CUNEIFORM_SIGN_LUGAL_SHESHIG = 0x0001221A;
const char32_t CUNEIFORM_SIGN_LUH = 0x0001221B;
const char32_t CUNEIFORM_SIGN_LUL = 0x0001221C;
const char32_t CUNEIFORM_SIGN_LUM = 0x0001221D;
const char32_t CUNEIFORM_SIGN_LUM_OVER_LUM = 0x0001221E;
const char32_t CUNEIFORM_SIGN_LUM_OVER_LUM_GAR_OVER_GAR = 0x0001221F;
const char32_t CUNEIFORM_SIGN_MA = 0x00012220;
const char32_t CUNEIFORM_SIGN_MA_TIMES_TAK4 = 0x00012221;
const char32_t CUNEIFORM_SIGN_MA_GUNU = 0x00012222;
const char32_t CUNEIFORM_SIGN_MA2 = 0x00012223;
const char32_t CUNEIFORM_SIGN_MAH = 0x00012224;
const char32_t CUNEIFORM_SIGN_MAR = 0x00012225;
const char32_t CUNEIFORM_SIGN_MASH = 0x00012226;
const char32_t CUNEIFORM_SIGN_MASH2 = 0x00012227;
const char32_t CUNEIFORM_SIGN_ME = 0x00012228;
const char32_t CUNEIFORM_SIGN_MES = 0x00012229;
const char32_t CUNEIFORM_SIGN_MI = 0x0001222A;
const char32_t CUNEIFORM_SIGN_MIN = 0x0001222B;
const char32_t CUNEIFORM_SIGN_MU = 0x0001222C;
const char32_t CUNEIFORM_SIGN_MU_OVER_MU = 0x0001222D;
const char32_t CUNEIFORM_SIGN_MUG = 0x0001222E;
const char32_t CUNEIFORM_SIGN_MUG_GUNU = 0x0001222F;
const char32_t CUNEIFORM_SIGN_MUNSUB = 0x00012230;
const char32_t CUNEIFORM_SIGN_MURGU2 = 0x00012231;
const char32_t CUNEIFORM_SIGN_MUSH = 0x00012232;
const char32_t CUNEIFORM_SIGN_MUSH_TIMES_A = 0x00012233;
const char32_t CUNEIFORM_SIGN_MUSH_TIMES_KUR = 0x00012234;
const char32_t CUNEIFORM_SIGN_MUSH_TIMES_ZA = 0x00012235;
const char32_t CUNEIFORM_SIGN_MUSH_OVER_MUSH = 0x00012236;
const char32_t CUNEIFORM_SIGN_MUSH_OVER_MUSH_TIMES_A_PLUS_NA = 0x00012237;
const char32_t CUNEIFORM_SIGN_MUSH_CROSSING_MUSH = 0x00012238;
const char32_t CUNEIFORM_SIGN_MUSH3 = 0x00012239;
const char32_t CUNEIFORM_SIGN_MUSH3_TIMES_A = 0x0001223A;
const char32_t CUNEIFORM_SIGN_MUSH3_TIMES_A_PLUS_DI = 0x0001223B;
const char32_t CUNEIFORM_SIGN_MUSH3_TIMES_DI = 0x0001223C;
const char32_t CUNEIFORM_SIGN_MUSH3_GUNU = 0x0001223D;
const char32_t CUNEIFORM_SIGN_NA = 0x0001223E;
const char32_t CUNEIFORM_SIGN_NA2 = 0x0001223F;
const char32_t CUNEIFORM_SIGN_NAGA = 0x00012240;
const char32_t CUNEIFORM_SIGN_NAGA_INVERTED = 0x00012241;
const char32_t CUNEIFORM_SIGN_NAGA_TIMES_SHU_TENU = 0x00012242;
const char32_t CUNEIFORM_SIGN_NAGA_OPPOSING_NAGA = 0x00012243;
const char32_t CUNEIFORM_SIGN_NAGAR = 0x00012244;
const char32_t CUNEIFORM_SIGN_NAM_NUTILLU = 0x00012245;
const char32_t CUNEIFORM_SIGN_NAM = 0x00012246;
const char32_t CUNEIFORM_SIGN_NAM2 = 0x00012247;
const char32_t CUNEIFORM_SIGN_NE = 0x00012248;
const char32_t CUNEIFORM_SIGN_NE_TIMES_A = 0x00012249;
const char32_t CUNEIFORM_SIGN_NE_TIMES_UD = 0x0001224A;
const char32_t CUNEIFORM_SIGN_NE_SHESHIG = 0x0001224B;
const char32_t CUNEIFORM_SIGN_NI = 0x0001224C;
const char32_t CUNEIFORM_SIGN_NI_TIMES_E = 0x0001224D;
const char32_t CUNEIFORM_SIGN_NI2 = 0x0001224E;
const char32_t CUNEIFORM_SIGN_NIM = 0x0001224F;
const char32_t CUNEIFORM_SIGN_NIM_TIMES_GAN2_TENU = 0x00012250;
const char32_t CUNEIFORM_SIGN_NIM_TIMES_GAR_PLUS_GAN2_TENU = 0x00012251;
const char32_t CUNEIFORM_SIGN_NINDA2 = 0x00012252;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_AN = 0x00012253;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_ASH = 0x00012254;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_ASH_PLUS_ASH = 0x00012255;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_GUD = 0x00012256;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_ME_PLUS_GAN2_TENU = 0x00012257;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_NE = 0x00012258;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_NUN = 0x00012259;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_SHE = 0x0001225A;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_SHE_PLUS_A_AN = 0x0001225B;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_SHE_PLUS_ASH = 0x0001225C;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_SHE_PLUS_ASH_PLUS_ASH = 0x0001225D;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_U2_PLUS_ASH = 0x0001225E;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_USH = 0x0001225F;
const char32_t CUNEIFORM_SIGN_NISAG = 0x00012260;
const char32_t CUNEIFORM_SIGN_NU = 0x00012261;
const char32_t CUNEIFORM_SIGN_NU11 = 0x00012262;
const char32_t CUNEIFORM_SIGN_NUN = 0x00012263;
const char32_t CUNEIFORM_SIGN_NUN_LAGAR_TIMES_GAR = 0x00012264;
const char32_t CUNEIFORM_SIGN_NUN_LAGAR_TIMES_MASH = 0x00012265;
const char32_t CUNEIFORM_SIGN_NUN_LAGAR_TIMES_SAL = 0x00012266;
const char32_t CUNEIFORM_SIGN_NUN_LAGAR_TIMES_SAL_OVER_NUN_LAGAR_TIMES_SAL = 0x00012267;
const char32_t CUNEIFORM_SIGN_NUN_LAGAR_TIMES_USH = 0x00012268;
const char32_t CUNEIFORM_SIGN_NUN_TENU = 0x00012269;
const char32_t CUNEIFORM_SIGN_NUN_OVER_NUN = 0x0001226A;
const char32_t CUNEIFORM_SIGN_NUN_CROSSING_NUN = 0x0001226B;
const char32_t CUNEIFORM_SIGN_NUN_CROSSING_NUN_LAGAR_OVER_LAGAR = 0x0001226C;
const char32_t CUNEIFORM_SIGN_NUNUZ = 0x0001226D;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_ASHGAB = 0x0001226E;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_BI = 0x0001226F;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_DUG = 0x00012270;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_GUD = 0x00012271;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_IGI_GUNU = 0x00012272;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_KAD3 = 0x00012273;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_LA = 0x00012274;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_NE = 0x00012275;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_SILA3 = 0x00012276;
const char32_t CUNEIFORM_SIGN_NUNUZ_AB2_TIMES_U2 = 0x00012277;
const char32_t CUNEIFORM_SIGN_NUNUZ_KISIM5_TIMES_BI = 0x00012278;
const char32_t CUNEIFORM_SIGN_NUNUZ_KISIM5_TIMES_BI_U = 0x00012279;
const char32_t CUNEIFORM_SIGN_PA = 0x0001227A;
const char32_t CUNEIFORM_SIGN_PAD = 0x0001227B;
const char32_t CUNEIFORM_SIGN_PAN = 0x0001227C;
const char32_t CUNEIFORM_SIGN_PAP = 0x0001227D;
const char32_t CUNEIFORM_SIGN_PESH2 = 0x0001227E;
const char32_t CUNEIFORM_SIGN_PI = 0x0001227F;
const char32_t CUNEIFORM_SIGN_PI_TIMES_A = 0x00012280;
const char32_t CUNEIFORM_SIGN_PI_TIMES_AB = 0x00012281;
const char32_t CUNEIFORM_SIGN_PI_TIMES_BI = 0x00012282;
const char32_t CUNEIFORM_SIGN_PI_TIMES_BU = 0x00012283;
const char32_t CUNEIFORM_SIGN_PI_TIMES_E = 0x00012284;
const char32_t CUNEIFORM_SIGN_PI_TIMES_I = 0x00012285;
const char32_t CUNEIFORM_SIGN_PI_TIMES_IB = 0x00012286;
const char32_t CUNEIFORM_SIGN_PI_TIMES_U = 0x00012287;
const char32_t CUNEIFORM_SIGN_PI_TIMES_U2 = 0x00012288;
const char32_t CUNEIFORM_SIGN_PI_CROSSING_PI = 0x00012289;
const char32_t CUNEIFORM_SIGN_PIRIG = 0x0001228A;
const char32_t CUNEIFORM_SIGN_PIRIG_TIMES_KAL = 0x0001228B;
const char32_t CUNEIFORM_SIGN_PIRIG_TIMES_UD = 0x0001228C;
const char32_t CUNEIFORM_SIGN_PIRIG_TIMES_ZA = 0x0001228D;
const char32_t CUNEIFORM_SIGN_PIRIG_OPPOSING_PIRIG = 0x0001228E;
const char32_t CUNEIFORM_SIGN_RA = 0x0001228F;
const char32_t CUNEIFORM_SIGN_RAB = 0x00012290;
const char32_t CUNEIFORM_SIGN_RI = 0x00012291;
const char32_t CUNEIFORM_SIGN_RU = 0x00012292;
const char32_t CUNEIFORM_SIGN_SA = 0x00012293;
const char32_t CUNEIFORM_SIGN_SAG_NUTILLU = 0x00012294;
const char32_t CUNEIFORM_SIGN_SAG = 0x00012295;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_A = 0x00012296;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_DU = 0x00012297;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_DUB = 0x00012298;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_HA = 0x00012299;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_KAK = 0x0001229A;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_KUR = 0x0001229B;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_LUM = 0x0001229C;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_MI = 0x0001229D;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_NUN = 0x0001229E;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_SAL = 0x0001229F;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_SHID = 0x000122A0;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_TAB = 0x000122A1;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_U2 = 0x000122A2;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_UB = 0x000122A3;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_UM = 0x000122A4;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_UR = 0x000122A5;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_USH = 0x000122A6;
const char32_t CUNEIFORM_SIGN_SAG_OVER_SAG = 0x000122A7;
const char32_t CUNEIFORM_SIGN_SAG_GUNU = 0x000122A8;
const char32_t CUNEIFORM_SIGN_SAL = 0x000122A9;
const char32_t CUNEIFORM_SIGN_SAL_LAGAB_TIMES_ASH2 = 0x000122AA;
const char32_t CUNEIFORM_SIGN_SANGA2 = 0x000122AB;
const char32_t CUNEIFORM_SIGN_SAR = 0x000122AC;
const char32_t CUNEIFORM_SIGN_SHA = 0x000122AD;
const char32_t CUNEIFORM_SIGN_SHA3 = 0x000122AE;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_A = 0x000122AF;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_BAD = 0x000122B0;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_GISH = 0x000122B1;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_NE = 0x000122B2;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_SHU2 = 0x000122B3;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_TUR = 0x000122B4;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_U = 0x000122B5;
const char32_t CUNEIFORM_SIGN_SHA3_TIMES_U_PLUS_A = 0x000122B6;
const char32_t CUNEIFORM_SIGN_SHA6 = 0x000122B7;
const char32_t CUNEIFORM_SIGN_SHAB6 = 0x000122B8;
const char32_t CUNEIFORM_SIGN_SHAR2 = 0x000122B9;
const char32_t CUNEIFORM_SIGN_SHE = 0x000122BA;
const char32_t CUNEIFORM_SIGN_SHE_HU = 0x000122BB;
const char32_t CUNEIFORM_SIGN_SHE_OVER_SHE_GAD_OVER_GAD_GAR_OVER_GAR = 0x000122BC;
const char32_t CUNEIFORM_SIGN_SHE_OVER_SHE_TAB_OVER_TAB_GAR_OVER_GAR = 0x000122BD;
const char32_t CUNEIFORM_SIGN_SHEG9 = 0x000122BE;
const char32_t CUNEIFORM_SIGN_SHEN = 0x000122BF;
const char32_t CUNEIFORM_SIGN_SHESH = 0x000122C0;
const char32_t CUNEIFORM_SIGN_SHESH2 = 0x000122C1;
const char32_t CUNEIFORM_SIGN_SHESHLAM = 0x000122C2;
const char32_t CUNEIFORM_SIGN_SHID = 0x000122C3;
const char32_t CUNEIFORM_SIGN_SHID_TIMES_A = 0x000122C4;
const char32_t CUNEIFORM_SIGN_SHID_TIMES_IM = 0x000122C5;
const char32_t CUNEIFORM_SIGN_SHIM = 0x000122C6;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_A = 0x000122C7;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_BAL = 0x000122C8;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_BULUG = 0x000122C9;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_DIN = 0x000122CA;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_GAR = 0x000122CB;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_IGI = 0x000122CC;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_IGI_GUNU = 0x000122CD;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_KUSHU2 = 0x000122CE;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_LUL = 0x000122CF;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_MUG = 0x000122D0;
const char32_t CUNEIFORM_SIGN_SHIM_TIMES_SAL = 0x000122D1;
const char32_t CUNEIFORM_SIGN_SHINIG = 0x000122D2;
const char32_t CUNEIFORM_SIGN_SHIR = 0x000122D3;
const char32_t CUNEIFORM_SIGN_SHIR_TENU = 0x000122D4;
const char32_t CUNEIFORM_SIGN_SHIR_OVER_SHIR_BUR_OVER_BUR = 0x000122D5;
const char32_t CUNEIFORM_SIGN_SHITA = 0x000122D6;
const char32_t CUNEIFORM_SIGN_SHU = 0x000122D7;
const char32_t CUNEIFORM_SIGN_SHU_OVER_INVERTED_SHU = 0x000122D8;
const char32_t CUNEIFORM_SIGN_SHU2 = 0x000122D9;
const char32_t CUNEIFORM_SIGN_SHUBUR = 0x000122DA;
const char32_t CUNEIFORM_SIGN_SI = 0x000122DB;
const char32_t CUNEIFORM_SIGN_SI_GUNU = 0x000122DC;
const char32_t CUNEIFORM_SIGN_SIG = 0x000122DD;
const char32_t CUNEIFORM_SIGN_SIG4 = 0x000122DE;
const char32_t CUNEIFORM_SIGN_SIG4_OVER_SIG4_SHU2 = 0x000122DF;
const char32_t CUNEIFORM_SIGN_SIK2 = 0x000122E0;
const char32_t CUNEIFORM_SIGN_SILA3 = 0x000122E1;
const char32_t CUNEIFORM_SIGN_SU = 0x000122E2;
const char32_t CUNEIFORM_SIGN_SU_OVER_SU = 0x000122E3;
const char32_t CUNEIFORM_SIGN_SUD = 0x000122E4;
const char32_t CUNEIFORM_SIGN_SUD2 = 0x000122E5;
const char32_t CUNEIFORM_SIGN_SUHUR = 0x000122E6;
const char32_t CUNEIFORM_SIGN_SUM = 0x000122E7;
const char32_t CUNEIFORM_SIGN_SUMASH = 0x000122E8;
const char32_t CUNEIFORM_SIGN_SUR = 0x000122E9;
const char32_t CUNEIFORM_SIGN_SUR9 = 0x000122EA;
const char32_t CUNEIFORM_SIGN_TA = 0x000122EB;
const char32_t CUNEIFORM_SIGN_TA_ASTERISK = 0x000122EC;
const char32_t CUNEIFORM_SIGN_TA_TIMES_HI = 0x000122ED;
const char32_t CUNEIFORM_SIGN_TA_TIMES_MI = 0x000122EE;
const char32_t CUNEIFORM_SIGN_TA_GUNU = 0x000122EF;
const char32_t CUNEIFORM_SIGN_TAB = 0x000122F0;
const char32_t CUNEIFORM_SIGN_TAB_OVER_TAB_NI_OVER_NI_DISH_OVER_DISH = 0x000122F1;
const char32_t CUNEIFORM_SIGN_TAB_SQUARED = 0x000122F2;
const char32_t CUNEIFORM_SIGN_TAG = 0x000122F3;
const char32_t CUNEIFORM_SIGN_TAG_TIMES_BI = 0x000122F4;
const char32_t CUNEIFORM_SIGN_TAG_TIMES_GUD = 0x000122F5;
const char32_t CUNEIFORM_SIGN_TAG_TIMES_SHE = 0x000122F6;
const char32_t CUNEIFORM_SIGN_TAG_TIMES_SHU = 0x000122F7;
const char32_t CUNEIFORM_SIGN_TAG_TIMES_TUG2 = 0x000122F8;
const char32_t CUNEIFORM_SIGN_TAG_TIMES_UD = 0x000122F9;
const char32_t CUNEIFORM_SIGN_TAK4 = 0x000122FA;
const char32_t CUNEIFORM_SIGN_TAR = 0x000122FB;
const char32_t CUNEIFORM_SIGN_TE = 0x000122FC;
const char32_t CUNEIFORM_SIGN_TE_GUNU = 0x000122FD;
const char32_t CUNEIFORM_SIGN_TI = 0x000122FE;
const char32_t CUNEIFORM_SIGN_TI_TENU = 0x000122FF;
const char32_t CUNEIFORM_SIGN_TIL = 0x00012300;
const char32_t CUNEIFORM_SIGN_TIR = 0x00012301;
const char32_t CUNEIFORM_SIGN_TIR_TIMES_TAK4 = 0x00012302;
const char32_t CUNEIFORM_SIGN_TIR_OVER_TIR = 0x00012303;
const char32_t CUNEIFORM_SIGN_TIR_OVER_TIR_GAD_OVER_GAD_GAR_OVER_GAR = 0x00012304;
const char32_t CUNEIFORM_SIGN_TU = 0x00012305;
const char32_t CUNEIFORM_SIGN_TUG2 = 0x00012306;
const char32_t CUNEIFORM_SIGN_TUK = 0x00012307;
const char32_t CUNEIFORM_SIGN_TUM = 0x00012308;
const char32_t CUNEIFORM_SIGN_TUR = 0x00012309;
const char32_t CUNEIFORM_SIGN_TUR_OVER_TUR_ZA_OVER_ZA = 0x0001230A;
const char32_t CUNEIFORM_SIGN_U = 0x0001230B;
const char32_t CUNEIFORM_SIGN_U_GUD = 0x0001230C;
const char32_t CUNEIFORM_SIGN_U_U_U = 0x0001230D;
const char32_t CUNEIFORM_SIGN_U_OVER_U_PA_OVER_PA_GAR_OVER_GAR = 0x0001230E;
const char32_t CUNEIFORM_SIGN_U_OVER_U_SUR_OVER_SUR = 0x0001230F;
const char32_t CUNEIFORM_SIGN_U_OVER_U_U_REVERSED_OVER_U_REVERSED = 0x00012310;
const char32_t CUNEIFORM_SIGN_U2 = 0x00012311;
const char32_t CUNEIFORM_SIGN_UB = 0x00012312;
const char32_t CUNEIFORM_SIGN_UD = 0x00012313;
const char32_t CUNEIFORM_SIGN_UD_KUSHU2 = 0x00012314;
const char32_t CUNEIFORM_SIGN_UD_TIMES_BAD = 0x00012315;
const char32_t CUNEIFORM_SIGN_UD_TIMES_MI = 0x00012316;
const char32_t CUNEIFORM_SIGN_UD_TIMES_U_PLUS_U_PLUS_U = 0x00012317;
const char32_t CUNEIFORM_SIGN_UD_TIMES_U_PLUS_U_PLUS_U_GUNU = 0x00012318;
const char32_t CUNEIFORM_SIGN_UD_GUNU = 0x00012319;
const char32_t CUNEIFORM_SIGN_UD_SHESHIG = 0x0001231A;
const char32_t CUNEIFORM_SIGN_UD_SHESHIG_TIMES_BAD = 0x0001231B;
const char32_t CUNEIFORM_SIGN_UDUG = 0x0001231C;
const char32_t CUNEIFORM_SIGN_UM = 0x0001231D;
const char32_t CUNEIFORM_SIGN_UM_TIMES_LAGAB = 0x0001231E;
const char32_t CUNEIFORM_SIGN_UM_TIMES_ME_PLUS_DA = 0x0001231F;
const char32_t CUNEIFORM_SIGN_UM_TIMES_SHA3 = 0x00012320;
const char32_t CUNEIFORM_SIGN_UM_TIMES_U = 0x00012321;
const char32_t CUNEIFORM_SIGN_UMBIN = 0x00012322;
const char32_t CUNEIFORM_SIGN_UMUM = 0x00012323;
const char32_t CUNEIFORM_SIGN_UMUM_TIMES_KASKAL = 0x00012324;
const char32_t CUNEIFORM_SIGN_UMUM_TIMES_PA = 0x00012325;
const char32_t CUNEIFORM_SIGN_UN = 0x00012326;
const char32_t CUNEIFORM_SIGN_UN_GUNU = 0x00012327;
const char32_t CUNEIFORM_SIGN_UR = 0x00012328;
const char32_t CUNEIFORM_SIGN_UR_CROSSING_UR = 0x00012329;
const char32_t CUNEIFORM_SIGN_UR_SHESHIG = 0x0001232A;
const char32_t CUNEIFORM_SIGN_UR2 = 0x0001232B;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_A_PLUS_HA = 0x0001232C;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_A_PLUS_NA = 0x0001232D;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_AL = 0x0001232E;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_HA = 0x0001232F;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_NUN = 0x00012330;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_U2 = 0x00012331;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_U2_PLUS_ASH = 0x00012332;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_U2_PLUS_BI = 0x00012333;
const char32_t CUNEIFORM_SIGN_UR4 = 0x00012334;
const char32_t CUNEIFORM_SIGN_URI = 0x00012335;
const char32_t CUNEIFORM_SIGN_URI3 = 0x00012336;
const char32_t CUNEIFORM_SIGN_URU = 0x00012337;
const char32_t CUNEIFORM_SIGN_URU_TIMES_A = 0x00012338;
const char32_t CUNEIFORM_SIGN_URU_TIMES_ASHGAB = 0x00012339;
const char32_t CUNEIFORM_SIGN_URU_TIMES_BAR = 0x0001233A;
const char32_t CUNEIFORM_SIGN_URU_TIMES_DUN = 0x0001233B;
const char32_t CUNEIFORM_SIGN_URU_TIMES_GA = 0x0001233C;
const char32_t CUNEIFORM_SIGN_URU_TIMES_GAL = 0x0001233D;
const char32_t CUNEIFORM_SIGN_URU_TIMES_GAN2_TENU = 0x0001233E;
const char32_t CUNEIFORM_SIGN_URU_TIMES_GAR = 0x0001233F;
const char32_t CUNEIFORM_SIGN_URU_TIMES_GU = 0x00012340;
const char32_t CUNEIFORM_SIGN_URU_TIMES_HA = 0x00012341;
const char32_t CUNEIFORM_SIGN_URU_TIMES_IGI = 0x00012342;
const char32_t CUNEIFORM_SIGN_URU_TIMES_IM = 0x00012343;
const char32_t CUNEIFORM_SIGN_URU_TIMES_ISH = 0x00012344;
const char32_t CUNEIFORM_SIGN_URU_TIMES_KI = 0x00012345;
const char32_t CUNEIFORM_SIGN_URU_TIMES_LUM = 0x00012346;
const char32_t CUNEIFORM_SIGN_URU_TIMES_MIN = 0x00012347;
const char32_t CUNEIFORM_SIGN_URU_TIMES_PA = 0x00012348;
const char32_t CUNEIFORM_SIGN_URU_TIMES_SHE = 0x00012349;
const char32_t CUNEIFORM_SIGN_URU_TIMES_SIG4 = 0x0001234A;
const char32_t CUNEIFORM_SIGN_URU_TIMES_TU = 0x0001234B;
const char32_t CUNEIFORM_SIGN_URU_TIMES_U_PLUS_GUD = 0x0001234C;
const char32_t CUNEIFORM_SIGN_URU_TIMES_UD = 0x0001234D;
const char32_t CUNEIFORM_SIGN_URU_TIMES_URUDA = 0x0001234E;
const char32_t CUNEIFORM_SIGN_URUDA = 0x0001234F;
const char32_t CUNEIFORM_SIGN_URUDA_TIMES_U = 0x00012350;
const char32_t CUNEIFORM_SIGN_USH = 0x00012351;
const char32_t CUNEIFORM_SIGN_USH_TIMES_A = 0x00012352;
const char32_t CUNEIFORM_SIGN_USH_TIMES_KU = 0x00012353;
const char32_t CUNEIFORM_SIGN_USH_TIMES_KUR = 0x00012354;
const char32_t CUNEIFORM_SIGN_USH_TIMES_TAK4 = 0x00012355;
const char32_t CUNEIFORM_SIGN_USHX = 0x00012356;
const char32_t CUNEIFORM_SIGN_USH2 = 0x00012357;
const char32_t CUNEIFORM_SIGN_USHUMX = 0x00012358;
const char32_t CUNEIFORM_SIGN_UTUKI = 0x00012359;
const char32_t CUNEIFORM_SIGN_UZ3 = 0x0001235A;
const char32_t CUNEIFORM_SIGN_UZ3_TIMES_KASKAL = 0x0001235B;
const char32_t CUNEIFORM_SIGN_UZU = 0x0001235C;
const char32_t CUNEIFORM_SIGN_ZA = 0x0001235D;
const char32_t CUNEIFORM_SIGN_ZA_TENU = 0x0001235E;
const char32_t CUNEIFORM_SIGN_ZA_SQUARED_TIMES_KUR = 0x0001235F;
const char32_t CUNEIFORM_SIGN_ZAG = 0x00012360;
const char32_t CUNEIFORM_SIGN_ZAMX = 0x00012361;
const char32_t CUNEIFORM_SIGN_ZE2 = 0x00012362;
const char32_t CUNEIFORM_SIGN_ZI = 0x00012363;
const char32_t CUNEIFORM_SIGN_ZI_OVER_ZI = 0x00012364;
const char32_t CUNEIFORM_SIGN_ZI3 = 0x00012365;
const char32_t CUNEIFORM_SIGN_ZIB = 0x00012366;
const char32_t CUNEIFORM_SIGN_ZIB_KABA_TENU = 0x00012367;
const char32_t CUNEIFORM_SIGN_ZIG = 0x00012368;
const char32_t CUNEIFORM_SIGN_ZIZ2 = 0x00012369;
const char32_t CUNEIFORM_SIGN_ZU = 0x0001236A;
const char32_t CUNEIFORM_SIGN_ZU5 = 0x0001236B;
const char32_t CUNEIFORM_SIGN_ZU5_TIMES_A = 0x0001236C;
const char32_t CUNEIFORM_SIGN_ZUBUR = 0x0001236D;
const char32_t CUNEIFORM_SIGN_ZUM = 0x0001236E;
const char32_t CUNEIFORM_SIGN_KAP_ELAMITE = 0x0001236F;
const char32_t CUNEIFORM_SIGN_AB_TIMES_NUN = 0x00012370;
const char32_t CUNEIFORM_SIGN_AB2_TIMES_A = 0x00012371;
const char32_t CUNEIFORM_SIGN_AMAR_TIMES_KUG = 0x00012372;
const char32_t CUNEIFORM_SIGN_DAG_KISIM5_TIMES_U2_PLUS_MASH = 0x00012373;
const char32_t CUNEIFORM_SIGN_DAG3 = 0x00012374;
const char32_t CUNEIFORM_SIGN_DISH_PLUS_SHU = 0x00012375;
const char32_t CUNEIFORM_SIGN_DUB_TIMES_SHE = 0x00012376;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_GUD = 0x00012377;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_SHE = 0x00012378;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_AN_PLUS_KAK_PLUS_A = 0x00012379;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_ASH2 = 0x0001237A;
const char32_t CUNEIFORM_SIGN_GE22 = 0x0001237B;
const char32_t CUNEIFORM_SIGN_GIG = 0x0001237C;
const char32_t CUNEIFORM_SIGN_HUSH = 0x0001237D;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ANSHE = 0x0001237E;
const char32_t CUNEIFORM_SIGN_KA_TIMES_ASH3 = 0x0001237F;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GISH = 0x00012380;
const char32_t CUNEIFORM_SIGN_KA_TIMES_GUD = 0x00012381;
const char32_t CUNEIFORM_SIGN_KA_TIMES_HI_TIMES_ASH2 = 0x00012382;
const char32_t CUNEIFORM_SIGN_KA_TIMES_LUM = 0x00012383;
const char32_t CUNEIFORM_SIGN_KA_TIMES_PA = 0x00012384;
const char32_t CUNEIFORM_SIGN_KA_TIMES_SHUL = 0x00012385;
const char32_t CUNEIFORM_SIGN_KA_TIMES_TU = 0x00012386;
const char32_t CUNEIFORM_SIGN_KA_TIMES_UR2 = 0x00012387;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_GI = 0x00012388;
const char32_t CUNEIFORM_SIGN_LU2_SHESHIG_TIMES_BAD = 0x00012389;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_ESH2_PLUS_LAL = 0x0001238A;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_SHU = 0x0001238B;
const char32_t CUNEIFORM_SIGN_MESH = 0x0001238C;
const char32_t CUNEIFORM_SIGN_MUSH3_TIMES_ZA = 0x0001238D;
const char32_t CUNEIFORM_SIGN_NA4 = 0x0001238E;
const char32_t CUNEIFORM_SIGN_NIN = 0x0001238F;
const char32_t CUNEIFORM_SIGN_NIN9 = 0x00012390;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_BAL = 0x00012391;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_GI = 0x00012392;
const char32_t CUNEIFORM_SIGN_NU11_ROTATED_NINETY_DEGREES = 0x00012393;
const char32_t CUNEIFORM_SIGN_PESH2_ASTERISK = 0x00012394;
const char32_t CUNEIFORM_SIGN_PIR2 = 0x00012395;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_IGI_GUNU = 0x00012396;
const char32_t CUNEIFORM_SIGN_TI2 = 0x00012397;
const char32_t CUNEIFORM_SIGN_UM_TIMES_ME = 0x00012398;
const char32_t CUNEIFORM_SIGN_U_U = 0x00012399;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_ASH = 0x00012400;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_ASH = 0x00012401;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_ASH = 0x00012402;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_ASH = 0x00012403;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_ASH = 0x00012404;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_ASH = 0x00012405;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_ASH = 0x00012406;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_ASH = 0x00012407;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_DISH = 0x00012408;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_DISH = 0x00012409;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_DISH = 0x0001240A;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_DISH = 0x0001240B;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_DISH = 0x0001240C;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_DISH = 0x0001240D;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_DISH = 0x0001240E;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_U = 0x0001240F;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_U = 0x00012410;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_U = 0x00012411;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_U = 0x00012412;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_U = 0x00012413;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_U = 0x00012414;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_GESH2 = 0x00012415;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_GESH2 = 0x00012416;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_GESH2 = 0x00012417;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_GESH2 = 0x00012418;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_GESH2 = 0x00012419;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_GESH2 = 0x0001241A;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_GESH2 = 0x0001241B;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_GESH2 = 0x0001241C;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_GESH2 = 0x0001241D;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_GESHU = 0x0001241E;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_GESHU = 0x0001241F;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_GESHU = 0x00012420;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_GESHU = 0x00012421;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_GESHU = 0x00012422;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_SHAR2 = 0x00012423;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_SHAR2 = 0x00012424;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_SHAR2_VARIANT_FORM = 0x00012425;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_SHAR2 = 0x00012426;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_SHAR2 = 0x00012427;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_SHAR2 = 0x00012428;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_SHAR2 = 0x00012429;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_SHAR2 = 0x0001242A;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_SHAR2 = 0x0001242B;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_SHARU = 0x0001242C;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_SHARU = 0x0001242D;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_SHARU = 0x0001242E;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_SHARU_VARIANT_FORM = 0x0001242F;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_SHARU = 0x00012430;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_SHARU = 0x00012431;
const char32_t CUNEIFORM_NUMERIC_SIGN_SHAR2_TIMES_GAL_PLUS_DISH = 0x00012432;
const char32_t CUNEIFORM_NUMERIC_SIGN_SHAR2_TIMES_GAL_PLUS_MIN = 0x00012433;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_BURU = 0x00012434;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_BURU = 0x00012435;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_BURU = 0x00012436;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_BURU_VARIANT_FORM = 0x00012437;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_BURU = 0x00012438;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_BURU = 0x00012439;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_VARIANT_FORM_ESH16 = 0x0001243A;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_VARIANT_FORM_ESH21 = 0x0001243B;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_VARIANT_FORM_LIMMU = 0x0001243C;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_VARIANT_FORM_LIMMU4 = 0x0001243D;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_VARIANT_FORM_LIMMU_A = 0x0001243E;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_VARIANT_FORM_LIMMU_B = 0x0001243F;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_VARIANT_FORM_ASH9 = 0x00012440;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_VARIANT_FORM_IMIN3 = 0x00012441;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_VARIANT_FORM_IMIN_A = 0x00012442;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_VARIANT_FORM_IMIN_B = 0x00012443;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_VARIANT_FORM_USSU = 0x00012444;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_VARIANT_FORM_USSU3 = 0x00012445;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_VARIANT_FORM_ILIMMU = 0x00012446;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_VARIANT_FORM_ILIMMU3 = 0x00012447;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_VARIANT_FORM_ILIMMU4 = 0x00012448;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_VARIANT_FORM_ILIMMU_A = 0x00012449;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_ASH_TENU = 0x0001244A;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_ASH_TENU = 0x0001244B;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_ASH_TENU = 0x0001244C;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_ASH_TENU = 0x0001244D;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_ASH_TENU = 0x0001244E;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_BAN2 = 0x0001244F;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_BAN2 = 0x00012450;
const char32_t CUNEIFORM_NUMERIC_SIGN_THREE_BAN2 = 0x00012451;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_BAN2 = 0x00012452;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_BAN2_VARIANT_FORM = 0x00012453;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_BAN2 = 0x00012454;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_BAN2_VARIANT_FORM = 0x00012455;
const char32_t CUNEIFORM_NUMERIC_SIGN_NIGIDAMIN = 0x00012456;
const char32_t CUNEIFORM_NUMERIC_SIGN_NIGIDAESH = 0x00012457;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_ESHE3 = 0x00012458;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_ESHE3 = 0x00012459;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_THIRD_DISH = 0x0001245A;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_THIRDS_DISH = 0x0001245B;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_SIXTHS_DISH = 0x0001245C;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_THIRD_VARIANT_FORM_A = 0x0001245D;
const char32_t CUNEIFORM_NUMERIC_SIGN_TWO_THIRDS_VARIANT_FORM_A = 0x0001245E;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_EIGHTH_ASH = 0x0001245F;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_QUARTER_ASH = 0x00012460;
const char32_t CUNEIFORM_NUMERIC_SIGN_OLD_ASSYRIAN_ONE_SIXTH = 0x00012461;
const char32_t CUNEIFORM_NUMERIC_SIGN_OLD_ASSYRIAN_ONE_QUARTER = 0x00012462;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_QUARTER_GUR = 0x00012463;
const char32_t CUNEIFORM_NUMERIC_SIGN_ONE_HALF_GUR = 0x00012464;
const char32_t CUNEIFORM_NUMERIC_SIGN_ELAMITE_ONE_THIRD = 0x00012465;
const char32_t CUNEIFORM_NUMERIC_SIGN_ELAMITE_TWO_THIRDS = 0x00012466;
const char32_t CUNEIFORM_NUMERIC_SIGN_ELAMITE_FORTY = 0x00012467;
const char32_t CUNEIFORM_NUMERIC_SIGN_ELAMITE_FIFTY = 0x00012468;
const char32_t CUNEIFORM_NUMERIC_SIGN_FOUR_U_VARIANT_FORM = 0x00012469;
const char32_t CUNEIFORM_NUMERIC_SIGN_FIVE_U_VARIANT_FORM = 0x0001246A;
const char32_t CUNEIFORM_NUMERIC_SIGN_SIX_U_VARIANT_FORM = 0x0001246B;
const char32_t CUNEIFORM_NUMERIC_SIGN_SEVEN_U_VARIANT_FORM = 0x0001246C;
const char32_t CUNEIFORM_NUMERIC_SIGN_EIGHT_U_VARIANT_FORM = 0x0001246D;
const char32_t CUNEIFORM_NUMERIC_SIGN_NINE_U_VARIANT_FORM = 0x0001246E;
const char32_t CUNEIFORM_PUNCTUATION_SIGN_OLD_ASSYRIAN_WORD_DIVIDER = 0x00012470;
const char32_t CUNEIFORM_PUNCTUATION_SIGN_VERTICAL_COLON = 0x00012471;
const char32_t CUNEIFORM_PUNCTUATION_SIGN_DIAGONAL_COLON = 0x00012472;
const char32_t CUNEIFORM_PUNCTUATION_SIGN_DIAGONAL_TRICOLON = 0x00012473;
const char32_t CUNEIFORM_PUNCTUATION_SIGN_DIAGONAL_QUADCOLON = 0x00012474;
const char32_t CUNEIFORM_SIGN_AB_TIMES_NUN_TENU = 0x00012480;
const char32_t CUNEIFORM_SIGN_AB_TIMES_SHU2 = 0x00012481;
const char32_t CUNEIFORM_SIGN_AD_TIMES_ESH2 = 0x00012482;
const char32_t CUNEIFORM_SIGN_BAD_TIMES_DISH_TENU = 0x00012483;
const char32_t CUNEIFORM_SIGN_BAHAR2_TIMES_AB2 = 0x00012484;
const char32_t CUNEIFORM_SIGN_BAHAR2_TIMES_NI = 0x00012485;
const char32_t CUNEIFORM_SIGN_BAHAR2_TIMES_ZA = 0x00012486;
const char32_t CUNEIFORM_SIGN_BU_OVER_BU_TIMES_NA2 = 0x00012487;
const char32_t CUNEIFORM_SIGN_DA_TIMES_TAK4 = 0x00012488;
const char32_t CUNEIFORM_SIGN_DAG_TIMES_KUR = 0x00012489;
const char32_t CUNEIFORM_SIGN_DIM_TIMES_IGI = 0x0001248A;
const char32_t CUNEIFORM_SIGN_DIM_TIMES_U_U_U = 0x0001248B;
const char32_t CUNEIFORM_SIGN_DIM2_TIMES_UD = 0x0001248C;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_ANSHE = 0x0001248D;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_ASH = 0x0001248E;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_ASH_AT_LEFT = 0x0001248F;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_DIN = 0x00012490;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_DUN = 0x00012491;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_ERIN2 = 0x00012492;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_GA = 0x00012493;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_GI = 0x00012494;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_GIR2_GUNU = 0x00012495;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_GISH = 0x00012496;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_HA = 0x00012497;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_HI = 0x00012498;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_IGI_GUNU = 0x00012499;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_KASKAL = 0x0001249A;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_KUR = 0x0001249B;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_KUSHU2 = 0x0001249C;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_KUSHU2_PLUS_KASKAL = 0x0001249D;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_LAK_020 = 0x0001249E;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_LAM = 0x0001249F;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_LAM_TIMES_KUR = 0x000124A0;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_LUH_PLUS_GISH = 0x000124A1;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_MASH = 0x000124A2;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_MES = 0x000124A3;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_MI = 0x000124A4;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_NI = 0x000124A5;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_PI = 0x000124A6;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_SHE = 0x000124A7;
const char32_t CUNEIFORM_SIGN_DUG_TIMES_SI_GUNU = 0x000124A8;
const char32_t CUNEIFORM_SIGN_E2_TIMES_KUR = 0x000124A9;
const char32_t CUNEIFORM_SIGN_E2_TIMES_PAP = 0x000124AA;
const char32_t CUNEIFORM_SIGN_ERIN2_X = 0x000124AB;
const char32_t CUNEIFORM_SIGN_ESH2_CROSSING_ESH2 = 0x000124AC;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_ASH = 0x000124AD;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_HI = 0x000124AE;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_IGI_GUNU = 0x000124AF;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_LA = 0x000124B0;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_LAL = 0x000124B1;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_ME = 0x000124B2;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_MES = 0x000124B3;
const char32_t CUNEIFORM_SIGN_EZEN_SHESHIG_TIMES_SU = 0x000124B4;
const char32_t CUNEIFORM_SIGN_EZEN_TIMES_SU = 0x000124B5;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_BAHAR2 = 0x000124B6;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_DIM_GUNU = 0x000124B7;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_DUG_TIMES_IGI_GUNU = 0x000124B8;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_DUG_TIMES_KASKAL = 0x000124B9;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_EREN = 0x000124BA;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GA = 0x000124BB;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GAR_PLUS_DI = 0x000124BC;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_GAR_PLUS_NE = 0x000124BD;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_HA_PLUS_A = 0x000124BE;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_KUSHU2_PLUS_KASKAL = 0x000124BF;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_LAM = 0x000124C0;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_LAM_TIMES_KUR = 0x000124C1;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_LUH = 0x000124C2;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_MUSH = 0x000124C3;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_NE = 0x000124C4;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_NE_PLUS_E2 = 0x000124C5;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_NE_PLUS_GI = 0x000124C6;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_SHIM = 0x000124C7;
const char32_t CUNEIFORM_SIGN_GA2_TIMES_ZIZ2 = 0x000124C8;
const char32_t CUNEIFORM_SIGN_GABA_ROTATED_NINETY_DEGREES = 0x000124C9;
const char32_t CUNEIFORM_SIGN_GESHTIN_TIMES_U = 0x000124CA;
const char32_t CUNEIFORM_SIGN_GISH_TIMES_GISH_CROSSING_GISH = 0x000124CB;
const char32_t CUNEIFORM_SIGN_GU2_TIMES_IGI_GUNU = 0x000124CC;
const char32_t CUNEIFORM_SIGN_GUD_PLUS_GISH_TIMES_TAK4 = 0x000124CD;
const char32_t CUNEIFORM_SIGN_HA_TENU_GUNU = 0x000124CE;
const char32_t CUNEIFORM_SIGN_HI_TIMES_ASH_OVER_HI_TIMES_ASH = 0x000124CF;
const char32_t CUNEIFORM_SIGN_KA_TIMES_BU = 0x000124D0;
const char32_t CUNEIFORM_SIGN_KA_TIMES_KA = 0x000124D1;
const char32_t CUNEIFORM_SIGN_KA_TIMES_U_U_U = 0x000124D2;
const char32_t CUNEIFORM_SIGN_KA_TIMES_UR = 0x000124D3;
const char32_t CUNEIFORM_SIGN_LAGAB_TIMES_ZU_OVER_ZU = 0x000124D4;
const char32_t CUNEIFORM_SIGN_LAK_003 = 0x000124D5;
const char32_t CUNEIFORM_SIGN_LAK_021 = 0x000124D6;
const char32_t CUNEIFORM_SIGN_LAK_025 = 0x000124D7;
const char32_t CUNEIFORM_SIGN_LAK_030 = 0x000124D8;
const char32_t CUNEIFORM_SIGN_LAK_050 = 0x000124D9;
const char32_t CUNEIFORM_SIGN_LAK_051 = 0x000124DA;
const char32_t CUNEIFORM_SIGN_LAK_062 = 0x000124DB;
const char32_t CUNEIFORM_SIGN_LAK_079_OVER_LAK_079_GUNU = 0x000124DC;
const char32_t CUNEIFORM_SIGN_LAK_080 = 0x000124DD;
const char32_t CUNEIFORM_SIGN_LAK_081_OVER_LAK_081 = 0x000124DE;
const char32_t CUNEIFORM_SIGN_LAK_092 = 0x000124DF;
const char32_t CUNEIFORM_SIGN_LAK_130 = 0x000124E0;
const char32_t CUNEIFORM_SIGN_LAK_142 = 0x000124E1;
const char32_t CUNEIFORM_SIGN_LAK_210 = 0x000124E2;
const char32_t CUNEIFORM_SIGN_LAK_219 = 0x000124E3;
const char32_t CUNEIFORM_SIGN_LAK_220 = 0x000124E4;
const char32_t CUNEIFORM_SIGN_LAK_225 = 0x000124E5;
const char32_t CUNEIFORM_SIGN_LAK_228 = 0x000124E6;
const char32_t CUNEIFORM_SIGN_LAK_238 = 0x000124E7;
const char32_t CUNEIFORM_SIGN_LAK_265 = 0x000124E8;
const char32_t CUNEIFORM_SIGN_LAK_266 = 0x000124E9;
const char32_t CUNEIFORM_SIGN_LAK_343 = 0x000124EA;
const char32_t CUNEIFORM_SIGN_LAK_347 = 0x000124EB;
const char32_t CUNEIFORM_SIGN_LAK_348 = 0x000124EC;
const char32_t CUNEIFORM_SIGN_LAK_383 = 0x000124ED;
const char32_t CUNEIFORM_SIGN_LAK_384 = 0x000124EE;
const char32_t CUNEIFORM_SIGN_LAK_390 = 0x000124EF;
const char32_t CUNEIFORM_SIGN_LAK_441 = 0x000124F0;
const char32_t CUNEIFORM_SIGN_LAK_449 = 0x000124F1;
const char32_t CUNEIFORM_SIGN_LAK_449_TIMES_GU = 0x000124F2;
const char32_t CUNEIFORM_SIGN_LAK_449_TIMES_IGI = 0x000124F3;
const char32_t CUNEIFORM_SIGN_LAK_449_TIMES_PAP_PLUS_LU3 = 0x000124F4;
const char32_t CUNEIFORM_SIGN_LAK_449_TIMES_PAP_PLUS_PAP_PLUS_LU3 = 0x000124F5;
const char32_t CUNEIFORM_SIGN_LAK_449_TIMES_U2_PLUS_BA = 0x000124F6;
const char32_t CUNEIFORM_SIGN_LAK_450 = 0x000124F7;
const char32_t CUNEIFORM_SIGN_LAK_457 = 0x000124F8;
const char32_t CUNEIFORM_SIGN_LAK_470 = 0x000124F9;
const char32_t CUNEIFORM_SIGN_LAK_483 = 0x000124FA;
const char32_t CUNEIFORM_SIGN_LAK_490 = 0x000124FB;
const char32_t CUNEIFORM_SIGN_LAK_492 = 0x000124FC;
const char32_t CUNEIFORM_SIGN_LAK_493 = 0x000124FD;
const char32_t CUNEIFORM_SIGN_LAK_495 = 0x000124FE;
const char32_t CUNEIFORM_SIGN_LAK_550 = 0x000124FF;
const char32_t CUNEIFORM_SIGN_LAK_608 = 0x00012500;
const char32_t CUNEIFORM_SIGN_LAK_617 = 0x00012501;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_ASH = 0x00012502;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_BAD = 0x00012503;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_DUN3_GUNU_GUNU = 0x00012504;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_KU3 = 0x00012505;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_LA = 0x00012506;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_TAR = 0x00012507;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_TE = 0x00012508;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_U2 = 0x00012509;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_UD = 0x0001250A;
const char32_t CUNEIFORM_SIGN_LAK_617_TIMES_URUDA = 0x0001250B;
const char32_t CUNEIFORM_SIGN_LAK_636 = 0x0001250C;
const char32_t CUNEIFORM_SIGN_LAK_648 = 0x0001250D;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_DUB = 0x0001250E;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_GA = 0x0001250F;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_IGI = 0x00012510;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_IGI_GUNU = 0x00012511;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_NI = 0x00012512;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_PAP_PLUS_PAP_PLUS_LU3 = 0x00012513;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_SHESH_PLUS_KI = 0x00012514;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_UD = 0x00012515;
const char32_t CUNEIFORM_SIGN_LAK_648_TIMES_URUDA = 0x00012516;
const char32_t CUNEIFORM_SIGN_LAK_724 = 0x00012517;
const char32_t CUNEIFORM_SIGN_LAK_749 = 0x00012518;
const char32_t CUNEIFORM_SIGN_LU2_GUNU_TIMES_ASH = 0x00012519;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_DISH = 0x0001251A;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_HAL = 0x0001251B;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_PAP = 0x0001251C;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_PAP_PLUS_PAP_PLUS_LU3 = 0x0001251D;
const char32_t CUNEIFORM_SIGN_LU2_TIMES_TAK4 = 0x0001251E;
const char32_t CUNEIFORM_SIGN_MI_PLUS_ZA7 = 0x0001251F;
const char32_t CUNEIFORM_SIGN_MUSH_OVER_MUSH_TIMES_GA = 0x00012520;
const char32_t CUNEIFORM_SIGN_MUSH_OVER_MUSH_TIMES_KAK = 0x00012521;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_DIM_GUNU = 0x00012522;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_GISH = 0x00012523;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_GUL = 0x00012524;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_HI = 0x00012525;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_KESH2 = 0x00012526;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_LAK_050 = 0x00012527;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_MASH = 0x00012528;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_PAP_PLUS_PAP = 0x00012529;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_U = 0x0001252A;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_U_PLUS_U = 0x0001252B;
const char32_t CUNEIFORM_SIGN_NINDA2_TIMES_URUDA = 0x0001252C;
const char32_t CUNEIFORM_SIGN_SAG_GUNU_TIMES_HA = 0x0001252D;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_EN = 0x0001252E;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_SHE_AT_LEFT = 0x0001252F;
const char32_t CUNEIFORM_SIGN_SAG_TIMES_TAK4 = 0x00012530;
const char32_t CUNEIFORM_SIGN_SHA6_TENU = 0x00012531;
const char32_t CUNEIFORM_SIGN_SHE_OVER_SHE = 0x00012532;
const char32_t CUNEIFORM_SIGN_SHE_PLUS_HUB2 = 0x00012533;
const char32_t CUNEIFORM_SIGN_SHE_PLUS_NAM2 = 0x00012534;
const char32_t CUNEIFORM_SIGN_SHE_PLUS_SAR = 0x00012535;
const char32_t CUNEIFORM_SIGN_SHU2_PLUS_DUG_TIMES_NI = 0x00012536;
const char32_t CUNEIFORM_SIGN_SHU2_PLUS_E2_TIMES_AN = 0x00012537;
const char32_t CUNEIFORM_SIGN_SI_TIMES_TAK4 = 0x00012538;
const char32_t CUNEIFORM_SIGN_TAK4_PLUS_SAG = 0x00012539;
const char32_t CUNEIFORM_SIGN_TUM_TIMES_GAN2_TENU = 0x0001253A;
const char32_t CUNEIFORM_SIGN_TUM_TIMES_THREE_DISH = 0x0001253B;
const char32_t CUNEIFORM_SIGN_UR2_INVERTED = 0x0001253C;
const char32_t CUNEIFORM_SIGN_UR2_TIMES_UD = 0x0001253D;
const char32_t CUNEIFORM_SIGN_URU_TIMES_DARA3 = 0x0001253E;
const char32_t CUNEIFORM_SIGN_URU_TIMES_LAK_668 = 0x0001253F;
const char32_t CUNEIFORM_SIGN_URU_TIMES_LU3 = 0x00012540;
const char32_t CUNEIFORM_SIGN_ZA7 = 0x00012541;
const char32_t CUNEIFORM_SIGN_ZU_OVER_ZU_PLUS_SAR = 0x00012542;
const char32_t CUNEIFORM_SIGN_ZU5_TIMES_THREE_DISH_TENU = 0x00012543;
const char32_t CYPRO_MINOAN_SIGN_CM001 = 0x00012F90;
const char32_t CYPRO_MINOAN_SIGN_CM002 = 0x00012F91;
const char32_t CYPRO_MINOAN_SIGN_CM004 = 0x00012F92;
const char32_t CYPRO_MINOAN_SIGN_CM005 = 0x00012F93;
const char32_t CYPRO_MINOAN_SIGN_CM006 = 0x00012F94;
const char32_t CYPRO_MINOAN_SIGN_CM007 = 0x00012F95;
const char32_t CYPRO_MINOAN_SIGN_CM008 = 0x00012F96;
const char32_t CYPRO_MINOAN_SIGN_CM009 = 0x00012F97;
const char32_t CYPRO_MINOAN_SIGN_CM010 = 0x00012F98;
const char32_t CYPRO_MINOAN_SIGN_CM011 = 0x00012F99;
const char32_t CYPRO_MINOAN_SIGN_CM012 = 0x00012F9A;
const char32_t CYPRO_MINOAN_SIGN_CM012B = 0x00012F9B;
const char32_t CYPRO_MINOAN_SIGN_CM013 = 0x00012F9C;
const char32_t CYPRO_MINOAN_SIGN_CM015 = 0x00012F9D;
const char32_t CYPRO_MINOAN_SIGN_CM017 = 0x00012F9E;
const char32_t CYPRO_MINOAN_SIGN_CM019 = 0x00012F9F;
const char32_t CYPRO_MINOAN_SIGN_CM021 = 0x00012FA0;
const char32_t CYPRO_MINOAN_SIGN_CM023 = 0x00012FA1;
const char32_t CYPRO_MINOAN_SIGN_CM024 = 0x00012FA2;
const char32_t CYPRO_MINOAN_SIGN_CM025 = 0x00012FA3;
const char32_t CYPRO_MINOAN_SIGN_CM026 = 0x00012FA4;
const char32_t CYPRO_MINOAN_SIGN_CM027 = 0x00012FA5;
const char32_t CYPRO_MINOAN_SIGN_CM028 = 0x00012FA6;
const char32_t CYPRO_MINOAN_SIGN_CM029 = 0x00012FA7;
const char32_t CYPRO_MINOAN_SIGN_CM030 = 0x00012FA8;
const char32_t CYPRO_MINOAN_SIGN_CM033 = 0x00012FA9;
const char32_t CYPRO_MINOAN_SIGN_CM034 = 0x00012FAA;
const char32_t CYPRO_MINOAN_SIGN_CM035 = 0x00012FAB;
const char32_t CYPRO_MINOAN_SIGN_CM036 = 0x00012FAC;
const char32_t CYPRO_MINOAN_SIGN_CM037 = 0x00012FAD;
const char32_t CYPRO_MINOAN_SIGN_CM038 = 0x00012FAE;
const char32_t CYPRO_MINOAN_SIGN_CM039 = 0x00012FAF;
const char32_t CYPRO_MINOAN_SIGN_CM040 = 0x00012FB0;
const char32_t CYPRO_MINOAN_SIGN_CM041 = 0x00012FB1;
const char32_t CYPRO_MINOAN_SIGN_CM044 = 0x00012FB2;
const char32_t CYPRO_MINOAN_SIGN_CM046 = 0x00012FB3;
const char32_t CYPRO_MINOAN_SIGN_CM047 = 0x00012FB4;
const char32_t CYPRO_MINOAN_SIGN_CM049 = 0x00012FB5;
const char32_t CYPRO_MINOAN_SIGN_CM050 = 0x00012FB6;
const char32_t CYPRO_MINOAN_SIGN_CM051 = 0x00012FB7;
const char32_t CYPRO_MINOAN_SIGN_CM052 = 0x00012FB8;
const char32_t CYPRO_MINOAN_SIGN_CM053 = 0x00012FB9;
const char32_t CYPRO_MINOAN_SIGN_CM054 = 0x00012FBA;
const char32_t CYPRO_MINOAN_SIGN_CM055 = 0x00012FBB;
const char32_t CYPRO_MINOAN_SIGN_CM056 = 0x00012FBC;
const char32_t CYPRO_MINOAN_SIGN_CM058 = 0x00012FBD;
const char32_t CYPRO_MINOAN_SIGN_CM059 = 0x00012FBE;
const char32_t CYPRO_MINOAN_SIGN_CM060 = 0x00012FBF;
const char32_t CYPRO_MINOAN_SIGN_CM061 = 0x00012FC0;
const char32_t CYPRO_MINOAN_SIGN_CM062 = 0x00012FC1;
const char32_t CYPRO_MINOAN_SIGN_CM063 = 0x00012FC2;
const char32_t CYPRO_MINOAN_SIGN_CM064 = 0x00012FC3;
const char32_t CYPRO_MINOAN_SIGN_CM066 = 0x00012FC4;
const char32_t CYPRO_MINOAN_SIGN_CM067 = 0x00012FC5;
const char32_t CYPRO_MINOAN_SIGN_CM068 = 0x00012FC6;
const char32_t CYPRO_MINOAN_SIGN_CM069 = 0x00012FC7;
const char32_t CYPRO_MINOAN_SIGN_CM070 = 0x00012FC8;
const char32_t CYPRO_MINOAN_SIGN_CM071 = 0x00012FC9;
const char32_t CYPRO_MINOAN_SIGN_CM072 = 0x00012FCA;
const char32_t CYPRO_MINOAN_SIGN_CM073 = 0x00012FCB;
const char32_t CYPRO_MINOAN_SIGN_CM074 = 0x00012FCC;
const char32_t CYPRO_MINOAN_SIGN_CM075 = 0x00012FCD;
const char32_t CYPRO_MINOAN_SIGN_CM075B = 0x00012FCE;
const char32_t CYPRO_MINOAN_SIGN_CM076 = 0x00012FCF;
const char32_t CYPRO_MINOAN_SIGN_CM078 = 0x00012FD0;
const char32_t CYPRO_MINOAN_SIGN_CM079 = 0x00012FD1;
const char32_t CYPRO_MINOAN_SIGN_CM080 = 0x00012FD2;
const char32_t CYPRO_MINOAN_SIGN_CM081 = 0x00012FD3;
const char32_t CYPRO_MINOAN_SIGN_CM082 = 0x00012FD4;
const char32_t CYPRO_MINOAN_SIGN_CM083 = 0x00012FD5;
const char32_t CYPRO_MINOAN_SIGN_CM084 = 0x00012FD6;
const char32_t CYPRO_MINOAN_SIGN_CM085 = 0x00012FD7;
const char32_t CYPRO_MINOAN_SIGN_CM086 = 0x00012FD8;
const char32_t CYPRO_MINOAN_SIGN_CM087 = 0x00012FD9;
const char32_t CYPRO_MINOAN_SIGN_CM088 = 0x00012FDA;
const char32_t CYPRO_MINOAN_SIGN_CM089 = 0x00012FDB;
const char32_t CYPRO_MINOAN_SIGN_CM090 = 0x00012FDC;
const char32_t CYPRO_MINOAN_SIGN_CM091 = 0x00012FDD;
const char32_t CYPRO_MINOAN_SIGN_CM092 = 0x00012FDE;
const char32_t CYPRO_MINOAN_SIGN_CM094 = 0x00012FDF;
const char32_t CYPRO_MINOAN_SIGN_CM095 = 0x00012FE0;
const char32_t CYPRO_MINOAN_SIGN_CM096 = 0x00012FE1;
const char32_t CYPRO_MINOAN_SIGN_CM097 = 0x00012FE2;
const char32_t CYPRO_MINOAN_SIGN_CM098 = 0x00012FE3;
const char32_t CYPRO_MINOAN_SIGN_CM099 = 0x00012FE4;
const char32_t CYPRO_MINOAN_SIGN_CM100 = 0x00012FE5;
const char32_t CYPRO_MINOAN_SIGN_CM101 = 0x00012FE6;
const char32_t CYPRO_MINOAN_SIGN_CM102 = 0x00012FE7;
const char32_t CYPRO_MINOAN_SIGN_CM103 = 0x00012FE8;
const char32_t CYPRO_MINOAN_SIGN_CM104 = 0x00012FE9;
const char32_t CYPRO_MINOAN_SIGN_CM105 = 0x00012FEA;
const char32_t CYPRO_MINOAN_SIGN_CM107 = 0x00012FEB;
const char32_t CYPRO_MINOAN_SIGN_CM108 = 0x00012FEC;
const char32_t CYPRO_MINOAN_SIGN_CM109 = 0x00012FED;
const char32_t CYPRO_MINOAN_SIGN_CM110 = 0x00012FEE;
const char32_t CYPRO_MINOAN_SIGN_CM112 = 0x00012FEF;
const char32_t CYPRO_MINOAN_SIGN_CM114 = 0x00012FF0;
const char32_t CYPRO_MINOAN_SIGN_CM301 = 0x00012FF1;
const char32_t CYPRO_MINOAN_SIGN_CM302 = 0x00012FF2;
const char32_t EGYPTIAN_HIEROGLYPH_A001 = 0x00013000;
const char32_t EGYPTIAN_HIEROGLYPH_A002 = 0x00013001;
const char32_t EGYPTIAN_HIEROGLYPH_A003 = 0x00013002;
const char32_t EGYPTIAN_HIEROGLYPH_A004 = 0x00013003;
const char32_t EGYPTIAN_HIEROGLYPH_A005 = 0x00013004;
const char32_t EGYPTIAN_HIEROGLYPH_A005A = 0x00013005;
const char32_t EGYPTIAN_HIEROGLYPH_A006 = 0x00013006;
const char32_t EGYPTIAN_HIEROGLYPH_A006A = 0x00013007;
const char32_t EGYPTIAN_HIEROGLYPH_A006B = 0x00013008;
const char32_t EGYPTIAN_HIEROGLYPH_A007 = 0x00013009;
const char32_t EGYPTIAN_HIEROGLYPH_A008 = 0x0001300A;
const char32_t EGYPTIAN_HIEROGLYPH_A009 = 0x0001300B;
const char32_t EGYPTIAN_HIEROGLYPH_A010 = 0x0001300C;
const char32_t EGYPTIAN_HIEROGLYPH_A011 = 0x0001300D;
const char32_t EGYPTIAN_HIEROGLYPH_A012 = 0x0001300E;
const char32_t EGYPTIAN_HIEROGLYPH_A013 = 0x0001300F;
const char32_t EGYPTIAN_HIEROGLYPH_A014 = 0x00013010;
const char32_t EGYPTIAN_HIEROGLYPH_A014A = 0x00013011;
const char32_t EGYPTIAN_HIEROGLYPH_A015 = 0x00013012;
const char32_t EGYPTIAN_HIEROGLYPH_A016 = 0x00013013;
const char32_t EGYPTIAN_HIEROGLYPH_A017 = 0x00013014;
const char32_t EGYPTIAN_HIEROGLYPH_A017A = 0x00013015;
const char32_t EGYPTIAN_HIEROGLYPH_A018 = 0x00013016;
const char32_t EGYPTIAN_HIEROGLYPH_A019 = 0x00013017;
const char32_t EGYPTIAN_HIEROGLYPH_A020 = 0x00013018;
const char32_t EGYPTIAN_HIEROGLYPH_A021 = 0x00013019;
const char32_t EGYPTIAN_HIEROGLYPH_A022 = 0x0001301A;
const char32_t EGYPTIAN_HIEROGLYPH_A023 = 0x0001301B;
const char32_t EGYPTIAN_HIEROGLYPH_A024 = 0x0001301C;
const char32_t EGYPTIAN_HIEROGLYPH_A025 = 0x0001301D;
const char32_t EGYPTIAN_HIEROGLYPH_A026 = 0x0001301E;
const char32_t EGYPTIAN_HIEROGLYPH_A027 = 0x0001301F;
const char32_t EGYPTIAN_HIEROGLYPH_A028 = 0x00013020;
const char32_t EGYPTIAN_HIEROGLYPH_A029 = 0x00013021;
const char32_t EGYPTIAN_HIEROGLYPH_A030 = 0x00013022;
const char32_t EGYPTIAN_HIEROGLYPH_A031 = 0x00013023;
const char32_t EGYPTIAN_HIEROGLYPH_A032 = 0x00013024;
const char32_t EGYPTIAN_HIEROGLYPH_A032A = 0x00013025;
const char32_t EGYPTIAN_HIEROGLYPH_A033 = 0x00013026;
const char32_t EGYPTIAN_HIEROGLYPH_A034 = 0x00013027;
const char32_t EGYPTIAN_HIEROGLYPH_A035 = 0x00013028;
const char32_t EGYPTIAN_HIEROGLYPH_A036 = 0x00013029;
const char32_t EGYPTIAN_HIEROGLYPH_A037 = 0x0001302A;
const char32_t EGYPTIAN_HIEROGLYPH_A038 = 0x0001302B;
const char32_t EGYPTIAN_HIEROGLYPH_A039 = 0x0001302C;
const char32_t EGYPTIAN_HIEROGLYPH_A040 = 0x0001302D;
const char32_t EGYPTIAN_HIEROGLYPH_A040A = 0x0001302E;
const char32_t EGYPTIAN_HIEROGLYPH_A041 = 0x0001302F;
const char32_t EGYPTIAN_HIEROGLYPH_A042 = 0x00013030;
const char32_t EGYPTIAN_HIEROGLYPH_A042A = 0x00013031;
const char32_t EGYPTIAN_HIEROGLYPH_A043 = 0x00013032;
const char32_t EGYPTIAN_HIEROGLYPH_A043A = 0x00013033;
const char32_t EGYPTIAN_HIEROGLYPH_A044 = 0x00013034;
const char32_t EGYPTIAN_HIEROGLYPH_A045 = 0x00013035;
const char32_t EGYPTIAN_HIEROGLYPH_A045A = 0x00013036;
const char32_t EGYPTIAN_HIEROGLYPH_A046 = 0x00013037;
const char32_t EGYPTIAN_HIEROGLYPH_A047 = 0x00013038;
const char32_t EGYPTIAN_HIEROGLYPH_A048 = 0x00013039;
const char32_t EGYPTIAN_HIEROGLYPH_A049 = 0x0001303A;
const char32_t EGYPTIAN_HIEROGLYPH_A050 = 0x0001303B;
const char32_t EGYPTIAN_HIEROGLYPH_A051 = 0x0001303C;
const char32_t EGYPTIAN_HIEROGLYPH_A052 = 0x0001303D;
const char32_t EGYPTIAN_HIEROGLYPH_A053 = 0x0001303E;
const char32_t EGYPTIAN_HIEROGLYPH_A054 = 0x0001303F;
const char32_t EGYPTIAN_HIEROGLYPH_A055 = 0x00013040;
const char32_t EGYPTIAN_HIEROGLYPH_A056 = 0x00013041;
const char32_t EGYPTIAN_HIEROGLYPH_A057 = 0x00013042;
const char32_t EGYPTIAN_HIEROGLYPH_A058 = 0x00013043;
const char32_t EGYPTIAN_HIEROGLYPH_A059 = 0x00013044;
const char32_t EGYPTIAN_HIEROGLYPH_A060 = 0x00013045;
const char32_t EGYPTIAN_HIEROGLYPH_A061 = 0x00013046;
const char32_t EGYPTIAN_HIEROGLYPH_A062 = 0x00013047;
const char32_t EGYPTIAN_HIEROGLYPH_A063 = 0x00013048;
const char32_t EGYPTIAN_HIEROGLYPH_A064 = 0x00013049;
const char32_t EGYPTIAN_HIEROGLYPH_A065 = 0x0001304A;
const char32_t EGYPTIAN_HIEROGLYPH_A066 = 0x0001304B;
const char32_t EGYPTIAN_HIEROGLYPH_A067 = 0x0001304C;
const char32_t EGYPTIAN_HIEROGLYPH_A068 = 0x0001304D;
const char32_t EGYPTIAN_HIEROGLYPH_A069 = 0x0001304E;
const char32_t EGYPTIAN_HIEROGLYPH_A070 = 0x0001304F;
const char32_t EGYPTIAN_HIEROGLYPH_B001 = 0x00013050;
const char32_t EGYPTIAN_HIEROGLYPH_B002 = 0x00013051;
const char32_t EGYPTIAN_HIEROGLYPH_B003 = 0x00013052;
const char32_t EGYPTIAN_HIEROGLYPH_B004 = 0x00013053;
const char32_t EGYPTIAN_HIEROGLYPH_B005 = 0x00013054;
const char32_t EGYPTIAN_HIEROGLYPH_B005A = 0x00013055;
const char32_t EGYPTIAN_HIEROGLYPH_B006 = 0x00013056;
const char32_t EGYPTIAN_HIEROGLYPH_B007 = 0x00013057;
const char32_t EGYPTIAN_HIEROGLYPH_B008 = 0x00013058;
const char32_t EGYPTIAN_HIEROGLYPH_B009 = 0x00013059;
const char32_t EGYPTIAN_HIEROGLYPH_C001 = 0x0001305A;
const char32_t EGYPTIAN_HIEROGLYPH_C002 = 0x0001305B;
const char32_t EGYPTIAN_HIEROGLYPH_C002A = 0x0001305C;
const char32_t EGYPTIAN_HIEROGLYPH_C002B = 0x0001305D;
const char32_t EGYPTIAN_HIEROGLYPH_C002C = 0x0001305E;
const char32_t EGYPTIAN_HIEROGLYPH_C003 = 0x0001305F;
const char32_t EGYPTIAN_HIEROGLYPH_C004 = 0x00013060;
const char32_t EGYPTIAN_HIEROGLYPH_C005 = 0x00013061;
const char32_t EGYPTIAN_HIEROGLYPH_C006 = 0x00013062;
const char32_t EGYPTIAN_HIEROGLYPH_C007 = 0x00013063;
const char32_t EGYPTIAN_HIEROGLYPH_C008 = 0x00013064;
const char32_t EGYPTIAN_HIEROGLYPH_C009 = 0x00013065;
const char32_t EGYPTIAN_HIEROGLYPH_C010 = 0x00013066;
const char32_t EGYPTIAN_HIEROGLYPH_C010A = 0x00013067;
const char32_t EGYPTIAN_HIEROGLYPH_C011 = 0x00013068;
const char32_t EGYPTIAN_HIEROGLYPH_C012 = 0x00013069;
const char32_t EGYPTIAN_HIEROGLYPH_C013 = 0x0001306A;
const char32_t EGYPTIAN_HIEROGLYPH_C014 = 0x0001306B;
const char32_t EGYPTIAN_HIEROGLYPH_C015 = 0x0001306C;
const char32_t EGYPTIAN_HIEROGLYPH_C016 = 0x0001306D;
const char32_t EGYPTIAN_HIEROGLYPH_C017 = 0x0001306E;
const char32_t EGYPTIAN_HIEROGLYPH_C018 = 0x0001306F;
const char32_t EGYPTIAN_HIEROGLYPH_C019 = 0x00013070;
const char32_t EGYPTIAN_HIEROGLYPH_C020 = 0x00013071;
const char32_t EGYPTIAN_HIEROGLYPH_C021 = 0x00013072;
const char32_t EGYPTIAN_HIEROGLYPH_C022 = 0x00013073;
const char32_t EGYPTIAN_HIEROGLYPH_C023 = 0x00013074;
const char32_t EGYPTIAN_HIEROGLYPH_C024 = 0x00013075;
const char32_t EGYPTIAN_HIEROGLYPH_D001 = 0x00013076;
const char32_t EGYPTIAN_HIEROGLYPH_D002 = 0x00013077;
const char32_t EGYPTIAN_HIEROGLYPH_D003 = 0x00013078;
const char32_t EGYPTIAN_HIEROGLYPH_D004 = 0x00013079;
const char32_t EGYPTIAN_HIEROGLYPH_D005 = 0x0001307A;
const char32_t EGYPTIAN_HIEROGLYPH_D006 = 0x0001307B;
const char32_t EGYPTIAN_HIEROGLYPH_D007 = 0x0001307C;
const char32_t EGYPTIAN_HIEROGLYPH_D008 = 0x0001307D;
const char32_t EGYPTIAN_HIEROGLYPH_D008A = 0x0001307E;
const char32_t EGYPTIAN_HIEROGLYPH_D009 = 0x0001307F;
const char32_t EGYPTIAN_HIEROGLYPH_D010 = 0x00013080;
const char32_t EGYPTIAN_HIEROGLYPH_D011 = 0x00013081;
const char32_t EGYPTIAN_HIEROGLYPH_D012 = 0x00013082;
const char32_t EGYPTIAN_HIEROGLYPH_D013 = 0x00013083;
const char32_t EGYPTIAN_HIEROGLYPH_D014 = 0x00013084;
const char32_t EGYPTIAN_HIEROGLYPH_D015 = 0x00013085;
const char32_t EGYPTIAN_HIEROGLYPH_D016 = 0x00013086;
const char32_t EGYPTIAN_HIEROGLYPH_D017 = 0x00013087;
const char32_t EGYPTIAN_HIEROGLYPH_D018 = 0x00013088;
const char32_t EGYPTIAN_HIEROGLYPH_D019 = 0x00013089;
const char32_t EGYPTIAN_HIEROGLYPH_D020 = 0x0001308A;
const char32_t EGYPTIAN_HIEROGLYPH_D021 = 0x0001308B;
const char32_t EGYPTIAN_HIEROGLYPH_D022 = 0x0001308C;
const char32_t EGYPTIAN_HIEROGLYPH_D023 = 0x0001308D;
const char32_t EGYPTIAN_HIEROGLYPH_D024 = 0x0001308E;
const char32_t EGYPTIAN_HIEROGLYPH_D025 = 0x0001308F;
const char32_t EGYPTIAN_HIEROGLYPH_D026 = 0x00013090;
const char32_t EGYPTIAN_HIEROGLYPH_D027 = 0x00013091;
const char32_t EGYPTIAN_HIEROGLYPH_D027A = 0x00013092;
const char32_t EGYPTIAN_HIEROGLYPH_D028 = 0x00013093;
const char32_t EGYPTIAN_HIEROGLYPH_D029 = 0x00013094;
const char32_t EGYPTIAN_HIEROGLYPH_D030 = 0x00013095;
const char32_t EGYPTIAN_HIEROGLYPH_D031 = 0x00013096;
const char32_t EGYPTIAN_HIEROGLYPH_D031A = 0x00013097;
const char32_t EGYPTIAN_HIEROGLYPH_D032 = 0x00013098;
const char32_t EGYPTIAN_HIEROGLYPH_D033 = 0x00013099;
const char32_t EGYPTIAN_HIEROGLYPH_D034 = 0x0001309A;
const char32_t EGYPTIAN_HIEROGLYPH_D034A = 0x0001309B;
const char32_t EGYPTIAN_HIEROGLYPH_D035 = 0x0001309C;
const char32_t EGYPTIAN_HIEROGLYPH_D036 = 0x0001309D;
const char32_t EGYPTIAN_HIEROGLYPH_D037 = 0x0001309E;
const char32_t EGYPTIAN_HIEROGLYPH_D038 = 0x0001309F;
const char32_t EGYPTIAN_HIEROGLYPH_D039 = 0x000130A0;
const char32_t EGYPTIAN_HIEROGLYPH_D040 = 0x000130A1;
const char32_t EGYPTIAN_HIEROGLYPH_D041 = 0x000130A2;
const char32_t EGYPTIAN_HIEROGLYPH_D042 = 0x000130A3;
const char32_t EGYPTIAN_HIEROGLYPH_D043 = 0x000130A4;
const char32_t EGYPTIAN_HIEROGLYPH_D044 = 0x000130A5;
const char32_t EGYPTIAN_HIEROGLYPH_D045 = 0x000130A6;
const char32_t EGYPTIAN_HIEROGLYPH_D046 = 0x000130A7;
const char32_t EGYPTIAN_HIEROGLYPH_D046A = 0x000130A8;
const char32_t EGYPTIAN_HIEROGLYPH_D047 = 0x000130A9;
const char32_t EGYPTIAN_HIEROGLYPH_D048 = 0x000130AA;
const char32_t EGYPTIAN_HIEROGLYPH_D048A = 0x000130AB;
const char32_t EGYPTIAN_HIEROGLYPH_D049 = 0x000130AC;
const char32_t EGYPTIAN_HIEROGLYPH_D050 = 0x000130AD;
const char32_t EGYPTIAN_HIEROGLYPH_D050A = 0x000130AE;
const char32_t EGYPTIAN_HIEROGLYPH_D050B = 0x000130AF;
const char32_t EGYPTIAN_HIEROGLYPH_D050C = 0x000130B0;
const char32_t EGYPTIAN_HIEROGLYPH_D050D = 0x000130B1;
const char32_t EGYPTIAN_HIEROGLYPH_D050E = 0x000130B2;
const char32_t EGYPTIAN_HIEROGLYPH_D050F = 0x000130B3;
const char32_t EGYPTIAN_HIEROGLYPH_D050G = 0x000130B4;
const char32_t EGYPTIAN_HIEROGLYPH_D050H = 0x000130B5;
const char32_t EGYPTIAN_HIEROGLYPH_D050I = 0x000130B6;
const char32_t EGYPTIAN_HIEROGLYPH_D051 = 0x000130B7;
const char32_t EGYPTIAN_HIEROGLYPH_D052 = 0x000130B8;
const char32_t EGYPTIAN_HIEROGLYPH_D052A = 0x000130B9;
const char32_t EGYPTIAN_HIEROGLYPH_D053 = 0x000130BA;
const char32_t EGYPTIAN_HIEROGLYPH_D054 = 0x000130BB;
const char32_t EGYPTIAN_HIEROGLYPH_D054A = 0x000130BC;
const char32_t EGYPTIAN_HIEROGLYPH_D055 = 0x000130BD;
const char32_t EGYPTIAN_HIEROGLYPH_D056 = 0x000130BE;
const char32_t EGYPTIAN_HIEROGLYPH_D057 = 0x000130BF;
const char32_t EGYPTIAN_HIEROGLYPH_D058 = 0x000130C0;
const char32_t EGYPTIAN_HIEROGLYPH_D059 = 0x000130C1;
const char32_t EGYPTIAN_HIEROGLYPH_D060 = 0x000130C2;
const char32_t EGYPTIAN_HIEROGLYPH_D061 = 0x000130C3;
const char32_t EGYPTIAN_HIEROGLYPH_D062 = 0x000130C4;
const char32_t EGYPTIAN_HIEROGLYPH_D063 = 0x000130C5;
const char32_t EGYPTIAN_HIEROGLYPH_D064 = 0x000130C6;
const char32_t EGYPTIAN_HIEROGLYPH_D065 = 0x000130C7;
const char32_t EGYPTIAN_HIEROGLYPH_D066 = 0x000130C8;
const char32_t EGYPTIAN_HIEROGLYPH_D067 = 0x000130C9;
const char32_t EGYPTIAN_HIEROGLYPH_D067A = 0x000130CA;
const char32_t EGYPTIAN_HIEROGLYPH_D067B = 0x000130CB;
const char32_t EGYPTIAN_HIEROGLYPH_D067C = 0x000130CC;
const char32_t EGYPTIAN_HIEROGLYPH_D067D = 0x000130CD;
const char32_t EGYPTIAN_HIEROGLYPH_D067E = 0x000130CE;
const char32_t EGYPTIAN_HIEROGLYPH_D067F = 0x000130CF;
const char32_t EGYPTIAN_HIEROGLYPH_D067G = 0x000130D0;
const char32_t EGYPTIAN_HIEROGLYPH_D067H = 0x000130D1;
const char32_t EGYPTIAN_HIEROGLYPH_E001 = 0x000130D2;
const char32_t EGYPTIAN_HIEROGLYPH_E002 = 0x000130D3;
const char32_t EGYPTIAN_HIEROGLYPH_E003 = 0x000130D4;
const char32_t EGYPTIAN_HIEROGLYPH_E004 = 0x000130D5;
const char32_t EGYPTIAN_HIEROGLYPH_E005 = 0x000130D6;
const char32_t EGYPTIAN_HIEROGLYPH_E006 = 0x000130D7;
const char32_t EGYPTIAN_HIEROGLYPH_E007 = 0x000130D8;
const char32_t EGYPTIAN_HIEROGLYPH_E008 = 0x000130D9;
const char32_t EGYPTIAN_HIEROGLYPH_E008A = 0x000130DA;
const char32_t EGYPTIAN_HIEROGLYPH_E009 = 0x000130DB;
const char32_t EGYPTIAN_HIEROGLYPH_E009A = 0x000130DC;
const char32_t EGYPTIAN_HIEROGLYPH_E010 = 0x000130DD;
const char32_t EGYPTIAN_HIEROGLYPH_E011 = 0x000130DE;
const char32_t EGYPTIAN_HIEROGLYPH_E012 = 0x000130DF;
const char32_t EGYPTIAN_HIEROGLYPH_E013 = 0x000130E0;
const char32_t EGYPTIAN_HIEROGLYPH_E014 = 0x000130E1;
const char32_t EGYPTIAN_HIEROGLYPH_E015 = 0x000130E2;
const char32_t EGYPTIAN_HIEROGLYPH_E016 = 0x000130E3;
const char32_t EGYPTIAN_HIEROGLYPH_E016A = 0x000130E4;
const char32_t EGYPTIAN_HIEROGLYPH_E017 = 0x000130E5;
const char32_t EGYPTIAN_HIEROGLYPH_E017A = 0x000130E6;
const char32_t EGYPTIAN_HIEROGLYPH_E018 = 0x000130E7;
const char32_t EGYPTIAN_HIEROGLYPH_E019 = 0x000130E8;
const char32_t EGYPTIAN_HIEROGLYPH_E020 = 0x000130E9;
const char32_t EGYPTIAN_HIEROGLYPH_E020A = 0x000130EA;
const char32_t EGYPTIAN_HIEROGLYPH_E021 = 0x000130EB;
const char32_t EGYPTIAN_HIEROGLYPH_E022 = 0x000130EC;
const char32_t EGYPTIAN_HIEROGLYPH_E023 = 0x000130ED;
const char32_t EGYPTIAN_HIEROGLYPH_E024 = 0x000130EE;
const char32_t EGYPTIAN_HIEROGLYPH_E025 = 0x000130EF;
const char32_t EGYPTIAN_HIEROGLYPH_E026 = 0x000130F0;
const char32_t EGYPTIAN_HIEROGLYPH_E027 = 0x000130F1;
const char32_t EGYPTIAN_HIEROGLYPH_E028 = 0x000130F2;
const char32_t EGYPTIAN_HIEROGLYPH_E028A = 0x000130F3;
const char32_t EGYPTIAN_HIEROGLYPH_E029 = 0x000130F4;
const char32_t EGYPTIAN_HIEROGLYPH_E030 = 0x000130F5;
const char32_t EGYPTIAN_HIEROGLYPH_E031 = 0x000130F6;
const char32_t EGYPTIAN_HIEROGLYPH_E032 = 0x000130F7;
const char32_t EGYPTIAN_HIEROGLYPH_E033 = 0x000130F8;
const char32_t EGYPTIAN_HIEROGLYPH_E034 = 0x000130F9;
const char32_t EGYPTIAN_HIEROGLYPH_E034A = 0x000130FA;
const char32_t EGYPTIAN_HIEROGLYPH_E036 = 0x000130FB;
const char32_t EGYPTIAN_HIEROGLYPH_E037 = 0x000130FC;
const char32_t EGYPTIAN_HIEROGLYPH_E038 = 0x000130FD;
const char32_t EGYPTIAN_HIEROGLYPH_F001 = 0x000130FE;
const char32_t EGYPTIAN_HIEROGLYPH_F001A = 0x000130FF;
const char32_t EGYPTIAN_HIEROGLYPH_F002 = 0x00013100;
const char32_t EGYPTIAN_HIEROGLYPH_F003 = 0x00013101;
const char32_t EGYPTIAN_HIEROGLYPH_F004 = 0x00013102;
const char32_t EGYPTIAN_HIEROGLYPH_F005 = 0x00013103;
const char32_t EGYPTIAN_HIEROGLYPH_F006 = 0x00013104;
const char32_t EGYPTIAN_HIEROGLYPH_F007 = 0x00013105;
const char32_t EGYPTIAN_HIEROGLYPH_F008 = 0x00013106;
const char32_t EGYPTIAN_HIEROGLYPH_F009 = 0x00013107;
const char32_t EGYPTIAN_HIEROGLYPH_F010 = 0x00013108;
const char32_t EGYPTIAN_HIEROGLYPH_F011 = 0x00013109;
const char32_t EGYPTIAN_HIEROGLYPH_F012 = 0x0001310A;
const char32_t EGYPTIAN_HIEROGLYPH_F013 = 0x0001310B;
const char32_t EGYPTIAN_HIEROGLYPH_F013A = 0x0001310C;
const char32_t EGYPTIAN_HIEROGLYPH_F014 = 0x0001310D;
const char32_t EGYPTIAN_HIEROGLYPH_F015 = 0x0001310E;
const char32_t EGYPTIAN_HIEROGLYPH_F016 = 0x0001310F;
const char32_t EGYPTIAN_HIEROGLYPH_F017 = 0x00013110;
const char32_t EGYPTIAN_HIEROGLYPH_F018 = 0x00013111;
const char32_t EGYPTIAN_HIEROGLYPH_F019 = 0x00013112;
const char32_t EGYPTIAN_HIEROGLYPH_F020 = 0x00013113;
const char32_t EGYPTIAN_HIEROGLYPH_F021 = 0x00013114;
const char32_t EGYPTIAN_HIEROGLYPH_F021A = 0x00013115;
const char32_t EGYPTIAN_HIEROGLYPH_F022 = 0x00013116;
const char32_t EGYPTIAN_HIEROGLYPH_F023 = 0x00013117;
const char32_t EGYPTIAN_HIEROGLYPH_F024 = 0x00013118;
const char32_t EGYPTIAN_HIEROGLYPH_F025 = 0x00013119;
const char32_t EGYPTIAN_HIEROGLYPH_F026 = 0x0001311A;
const char32_t EGYPTIAN_HIEROGLYPH_F027 = 0x0001311B;
const char32_t EGYPTIAN_HIEROGLYPH_F028 = 0x0001311C;
const char32_t EGYPTIAN_HIEROGLYPH_F029 = 0x0001311D;
const char32_t EGYPTIAN_HIEROGLYPH_F030 = 0x0001311E;
const char32_t EGYPTIAN_HIEROGLYPH_F031 = 0x0001311F;
const char32_t EGYPTIAN_HIEROGLYPH_F031A = 0x00013120;
const char32_t EGYPTIAN_HIEROGLYPH_F032 = 0x00013121;
const char32_t EGYPTIAN_HIEROGLYPH_F033 = 0x00013122;
const char32_t EGYPTIAN_HIEROGLYPH_F034 = 0x00013123;
const char32_t EGYPTIAN_HIEROGLYPH_F035 = 0x00013124;
const char32_t EGYPTIAN_HIEROGLYPH_F036 = 0x00013125;
const char32_t EGYPTIAN_HIEROGLYPH_F037 = 0x00013126;
const char32_t EGYPTIAN_HIEROGLYPH_F037A = 0x00013127;
const char32_t EGYPTIAN_HIEROGLYPH_F038 = 0x00013128;
const char32_t EGYPTIAN_HIEROGLYPH_F038A = 0x00013129;
const char32_t EGYPTIAN_HIEROGLYPH_F039 = 0x0001312A;
const char32_t EGYPTIAN_HIEROGLYPH_F040 = 0x0001312B;
const char32_t EGYPTIAN_HIEROGLYPH_F041 = 0x0001312C;
const char32_t EGYPTIAN_HIEROGLYPH_F042 = 0x0001312D;
const char32_t EGYPTIAN_HIEROGLYPH_F043 = 0x0001312E;
const char32_t EGYPTIAN_HIEROGLYPH_F044 = 0x0001312F;
const char32_t EGYPTIAN_HIEROGLYPH_F045 = 0x00013130;
const char32_t EGYPTIAN_HIEROGLYPH_F045A = 0x00013131;
const char32_t EGYPTIAN_HIEROGLYPH_F046 = 0x00013132;
const char32_t EGYPTIAN_HIEROGLYPH_F046A = 0x00013133;
const char32_t EGYPTIAN_HIEROGLYPH_F047 = 0x00013134;
const char32_t EGYPTIAN_HIEROGLYPH_F047A = 0x00013135;
const char32_t EGYPTIAN_HIEROGLYPH_F048 = 0x00013136;
const char32_t EGYPTIAN_HIEROGLYPH_F049 = 0x00013137;
const char32_t EGYPTIAN_HIEROGLYPH_F050 = 0x00013138;
const char32_t EGYPTIAN_HIEROGLYPH_F051 = 0x00013139;
const char32_t EGYPTIAN_HIEROGLYPH_F051A = 0x0001313A;
const char32_t EGYPTIAN_HIEROGLYPH_F051B = 0x0001313B;
const char32_t EGYPTIAN_HIEROGLYPH_F051C = 0x0001313C;
const char32_t EGYPTIAN_HIEROGLYPH_F052 = 0x0001313D;
const char32_t EGYPTIAN_HIEROGLYPH_F053 = 0x0001313E;
const char32_t EGYPTIAN_HIEROGLYPH_G001 = 0x0001313F;
const char32_t EGYPTIAN_HIEROGLYPH_G002 = 0x00013140;
const char32_t EGYPTIAN_HIEROGLYPH_G003 = 0x00013141;
const char32_t EGYPTIAN_HIEROGLYPH_G004 = 0x00013142;
const char32_t EGYPTIAN_HIEROGLYPH_G005 = 0x00013143;
const char32_t EGYPTIAN_HIEROGLYPH_G006 = 0x00013144;
const char32_t EGYPTIAN_HIEROGLYPH_G006A = 0x00013145;
const char32_t EGYPTIAN_HIEROGLYPH_G007 = 0x00013146;
const char32_t EGYPTIAN_HIEROGLYPH_G007A = 0x00013147;
const char32_t EGYPTIAN_HIEROGLYPH_G007B = 0x00013148;
const char32_t EGYPTIAN_HIEROGLYPH_G008 = 0x00013149;
const char32_t EGYPTIAN_HIEROGLYPH_G009 = 0x0001314A;
const char32_t EGYPTIAN_HIEROGLYPH_G010 = 0x0001314B;
const char32_t EGYPTIAN_HIEROGLYPH_G011 = 0x0001314C;
const char32_t EGYPTIAN_HIEROGLYPH_G011A = 0x0001314D;
const char32_t EGYPTIAN_HIEROGLYPH_G012 = 0x0001314E;
const char32_t EGYPTIAN_HIEROGLYPH_G013 = 0x0001314F;
const char32_t EGYPTIAN_HIEROGLYPH_G014 = 0x00013150;
const char32_t EGYPTIAN_HIEROGLYPH_G015 = 0x00013151;
const char32_t EGYPTIAN_HIEROGLYPH_G016 = 0x00013152;
const char32_t EGYPTIAN_HIEROGLYPH_G017 = 0x00013153;
const char32_t EGYPTIAN_HIEROGLYPH_G018 = 0x00013154;
const char32_t EGYPTIAN_HIEROGLYPH_G019 = 0x00013155;
const char32_t EGYPTIAN_HIEROGLYPH_G020 = 0x00013156;
const char32_t EGYPTIAN_HIEROGLYPH_G020A = 0x00013157;
const char32_t EGYPTIAN_HIEROGLYPH_G021 = 0x00013158;
const char32_t EGYPTIAN_HIEROGLYPH_G022 = 0x00013159;
const char32_t EGYPTIAN_HIEROGLYPH_G023 = 0x0001315A;
const char32_t EGYPTIAN_HIEROGLYPH_G024 = 0x0001315B;
const char32_t EGYPTIAN_HIEROGLYPH_G025 = 0x0001315C;
const char32_t EGYPTIAN_HIEROGLYPH_G026 = 0x0001315D;
const char32_t EGYPTIAN_HIEROGLYPH_G026A = 0x0001315E;
const char32_t EGYPTIAN_HIEROGLYPH_G027 = 0x0001315F;
const char32_t EGYPTIAN_HIEROGLYPH_G028 = 0x00013160;
const char32_t EGYPTIAN_HIEROGLYPH_G029 = 0x00013161;
const char32_t EGYPTIAN_HIEROGLYPH_G030 = 0x00013162;
const char32_t EGYPTIAN_HIEROGLYPH_G031 = 0x00013163;
const char32_t EGYPTIAN_HIEROGLYPH_G032 = 0x00013164;
const char32_t EGYPTIAN_HIEROGLYPH_G033 = 0x00013165;
const char32_t EGYPTIAN_HIEROGLYPH_G034 = 0x00013166;
const char32_t EGYPTIAN_HIEROGLYPH_G035 = 0x00013167;
const char32_t EGYPTIAN_HIEROGLYPH_G036 = 0x00013168;
const char32_t EGYPTIAN_HIEROGLYPH_G036A = 0x00013169;
const char32_t EGYPTIAN_HIEROGLYPH_G037 = 0x0001316A;
const char32_t EGYPTIAN_HIEROGLYPH_G037A = 0x0001316B;
const char32_t EGYPTIAN_HIEROGLYPH_G038 = 0x0001316C;
const char32_t EGYPTIAN_HIEROGLYPH_G039 = 0x0001316D;
const char32_t EGYPTIAN_HIEROGLYPH_G040 = 0x0001316E;
const char32_t EGYPTIAN_HIEROGLYPH_G041 = 0x0001316F;
const char32_t EGYPTIAN_HIEROGLYPH_G042 = 0x00013170;
const char32_t EGYPTIAN_HIEROGLYPH_G043 = 0x00013171;
const char32_t EGYPTIAN_HIEROGLYPH_G043A = 0x00013172;
const char32_t EGYPTIAN_HIEROGLYPH_G044 = 0x00013173;
const char32_t EGYPTIAN_HIEROGLYPH_G045 = 0x00013174;
const char32_t EGYPTIAN_HIEROGLYPH_G045A = 0x00013175;
const char32_t EGYPTIAN_HIEROGLYPH_G046 = 0x00013176;
const char32_t EGYPTIAN_HIEROGLYPH_G047 = 0x00013177;
const char32_t EGYPTIAN_HIEROGLYPH_G048 = 0x00013178;
const char32_t EGYPTIAN_HIEROGLYPH_G049 = 0x00013179;
const char32_t EGYPTIAN_HIEROGLYPH_G050 = 0x0001317A;
const char32_t EGYPTIAN_HIEROGLYPH_G051 = 0x0001317B;
const char32_t EGYPTIAN_HIEROGLYPH_G052 = 0x0001317C;
const char32_t EGYPTIAN_HIEROGLYPH_G053 = 0x0001317D;
const char32_t EGYPTIAN_HIEROGLYPH_G054 = 0x0001317E;
const char32_t EGYPTIAN_HIEROGLYPH_H001 = 0x0001317F;
const char32_t EGYPTIAN_HIEROGLYPH_H002 = 0x00013180;
const char32_t EGYPTIAN_HIEROGLYPH_H003 = 0x00013181;
const char32_t EGYPTIAN_HIEROGLYPH_H004 = 0x00013182;
const char32_t EGYPTIAN_HIEROGLYPH_H005 = 0x00013183;
const char32_t EGYPTIAN_HIEROGLYPH_H006 = 0x00013184;
const char32_t EGYPTIAN_HIEROGLYPH_H006A = 0x00013185;
const char32_t EGYPTIAN_HIEROGLYPH_H007 = 0x00013186;
const char32_t EGYPTIAN_HIEROGLYPH_H008 = 0x00013187;
const char32_t EGYPTIAN_HIEROGLYPH_I001 = 0x00013188;
const char32_t EGYPTIAN_HIEROGLYPH_I002 = 0x00013189;
const char32_t EGYPTIAN_HIEROGLYPH_I003 = 0x0001318A;
const char32_t EGYPTIAN_HIEROGLYPH_I004 = 0x0001318B;
const char32_t EGYPTIAN_HIEROGLYPH_I005 = 0x0001318C;
const char32_t EGYPTIAN_HIEROGLYPH_I005A = 0x0001318D;
const char32_t EGYPTIAN_HIEROGLYPH_I006 = 0x0001318E;
const char32_t EGYPTIAN_HIEROGLYPH_I007 = 0x0001318F;
const char32_t EGYPTIAN_HIEROGLYPH_I008 = 0x00013190;
const char32_t EGYPTIAN_HIEROGLYPH_I009 = 0x00013191;
const char32_t EGYPTIAN_HIEROGLYPH_I009A = 0x00013192;
const char32_t EGYPTIAN_HIEROGLYPH_I010 = 0x00013193;
const char32_t EGYPTIAN_HIEROGLYPH_I010A = 0x00013194;
const char32_t EGYPTIAN_HIEROGLYPH_I011 = 0x00013195;
const char32_t EGYPTIAN_HIEROGLYPH_I011A = 0x00013196;
const char32_t EGYPTIAN_HIEROGLYPH_I012 = 0x00013197;
const char32_t EGYPTIAN_HIEROGLYPH_I013 = 0x00013198;
const char32_t EGYPTIAN_HIEROGLYPH_I014 = 0x00013199;
const char32_t EGYPTIAN_HIEROGLYPH_I015 = 0x0001319A;
const char32_t EGYPTIAN_HIEROGLYPH_K001 = 0x0001319B;
const char32_t EGYPTIAN_HIEROGLYPH_K002 = 0x0001319C;
const char32_t EGYPTIAN_HIEROGLYPH_K003 = 0x0001319D;
const char32_t EGYPTIAN_HIEROGLYPH_K004 = 0x0001319E;
const char32_t EGYPTIAN_HIEROGLYPH_K005 = 0x0001319F;
const char32_t EGYPTIAN_HIEROGLYPH_K006 = 0x000131A0;
const char32_t EGYPTIAN_HIEROGLYPH_K007 = 0x000131A1;
const char32_t EGYPTIAN_HIEROGLYPH_K008 = 0x000131A2;
const char32_t EGYPTIAN_HIEROGLYPH_L001 = 0x000131A3;
const char32_t EGYPTIAN_HIEROGLYPH_L002 = 0x000131A4;
const char32_t EGYPTIAN_HIEROGLYPH_L002A = 0x000131A5;
const char32_t EGYPTIAN_HIEROGLYPH_L003 = 0x000131A6;
const char32_t EGYPTIAN_HIEROGLYPH_L004 = 0x000131A7;
const char32_t EGYPTIAN_HIEROGLYPH_L005 = 0x000131A8;
const char32_t EGYPTIAN_HIEROGLYPH_L006 = 0x000131A9;
const char32_t EGYPTIAN_HIEROGLYPH_L006A = 0x000131AA;
const char32_t EGYPTIAN_HIEROGLYPH_L007 = 0x000131AB;
const char32_t EGYPTIAN_HIEROGLYPH_L008 = 0x000131AC;
const char32_t EGYPTIAN_HIEROGLYPH_M001 = 0x000131AD;
const char32_t EGYPTIAN_HIEROGLYPH_M001A = 0x000131AE;
const char32_t EGYPTIAN_HIEROGLYPH_M001B = 0x000131AF;
const char32_t EGYPTIAN_HIEROGLYPH_M002 = 0x000131B0;
const char32_t EGYPTIAN_HIEROGLYPH_M003 = 0x000131B1;
const char32_t EGYPTIAN_HIEROGLYPH_M003A = 0x000131B2;
const char32_t EGYPTIAN_HIEROGLYPH_M004 = 0x000131B3;
const char32_t EGYPTIAN_HIEROGLYPH_M005 = 0x000131B4;
const char32_t EGYPTIAN_HIEROGLYPH_M006 = 0x000131B5;
const char32_t EGYPTIAN_HIEROGLYPH_M007 = 0x000131B6;
const char32_t EGYPTIAN_HIEROGLYPH_M008 = 0x000131B7;
const char32_t EGYPTIAN_HIEROGLYPH_M009 = 0x000131B8;
const char32_t EGYPTIAN_HIEROGLYPH_M010 = 0x000131B9;
const char32_t EGYPTIAN_HIEROGLYPH_M010A = 0x000131BA;
const char32_t EGYPTIAN_HIEROGLYPH_M011 = 0x000131BB;
const char32_t EGYPTIAN_HIEROGLYPH_M012 = 0x000131BC;
const char32_t EGYPTIAN_HIEROGLYPH_M012A = 0x000131BD;
const char32_t EGYPTIAN_HIEROGLYPH_M012B = 0x000131BE;
const char32_t EGYPTIAN_HIEROGLYPH_M012C = 0x000131BF;
const char32_t EGYPTIAN_HIEROGLYPH_M012D = 0x000131C0;
const char32_t EGYPTIAN_HIEROGLYPH_M012E = 0x000131C1;
const char32_t EGYPTIAN_HIEROGLYPH_M012F = 0x000131C2;
const char32_t EGYPTIAN_HIEROGLYPH_M012G = 0x000131C3;
const char32_t EGYPTIAN_HIEROGLYPH_M012H = 0x000131C4;
const char32_t EGYPTIAN_HIEROGLYPH_M013 = 0x000131C5;
const char32_t EGYPTIAN_HIEROGLYPH_M014 = 0x000131C6;
const char32_t EGYPTIAN_HIEROGLYPH_M015 = 0x000131C7;
const char32_t EGYPTIAN_HIEROGLYPH_M015A = 0x000131C8;
const char32_t EGYPTIAN_HIEROGLYPH_M016 = 0x000131C9;
const char32_t EGYPTIAN_HIEROGLYPH_M016A = 0x000131CA;
const char32_t EGYPTIAN_HIEROGLYPH_M017 = 0x000131CB;
const char32_t EGYPTIAN_HIEROGLYPH_M017A = 0x000131CC;
const char32_t EGYPTIAN_HIEROGLYPH_M018 = 0x000131CD;
const char32_t EGYPTIAN_HIEROGLYPH_M019 = 0x000131CE;
const char32_t EGYPTIAN_HIEROGLYPH_M020 = 0x000131CF;
const char32_t EGYPTIAN_HIEROGLYPH_M021 = 0x000131D0;
const char32_t EGYPTIAN_HIEROGLYPH_M022 = 0x000131D1;
const char32_t EGYPTIAN_HIEROGLYPH_M022A = 0x000131D2;
const char32_t EGYPTIAN_HIEROGLYPH_M023 = 0x000131D3;
const char32_t EGYPTIAN_HIEROGLYPH_M024 = 0x000131D4;
const char32_t EGYPTIAN_HIEROGLYPH_M024A = 0x000131D5;
const char32_t EGYPTIAN_HIEROGLYPH_M025 = 0x000131D6;
const char32_t EGYPTIAN_HIEROGLYPH_M026 = 0x000131D7;
const char32_t EGYPTIAN_HIEROGLYPH_M027 = 0x000131D8;
const char32_t EGYPTIAN_HIEROGLYPH_M028 = 0x000131D9;
const char32_t EGYPTIAN_HIEROGLYPH_M028A = 0x000131DA;
const char32_t EGYPTIAN_HIEROGLYPH_M029 = 0x000131DB;
const char32_t EGYPTIAN_HIEROGLYPH_M030 = 0x000131DC;
const char32_t EGYPTIAN_HIEROGLYPH_M031 = 0x000131DD;
const char32_t EGYPTIAN_HIEROGLYPH_M031A = 0x000131DE;
const char32_t EGYPTIAN_HIEROGLYPH_M032 = 0x000131DF;
const char32_t EGYPTIAN_HIEROGLYPH_M033 = 0x000131E0;
const char32_t EGYPTIAN_HIEROGLYPH_M033A = 0x000131E1;
const char32_t EGYPTIAN_HIEROGLYPH_M033B = 0x000131E2;
const char32_t EGYPTIAN_HIEROGLYPH_M034 = 0x000131E3;
const char32_t EGYPTIAN_HIEROGLYPH_M035 = 0x000131E4;
const char32_t EGYPTIAN_HIEROGLYPH_M036 = 0x000131E5;
const char32_t EGYPTIAN_HIEROGLYPH_M037 = 0x000131E6;
const char32_t EGYPTIAN_HIEROGLYPH_M038 = 0x000131E7;
const char32_t EGYPTIAN_HIEROGLYPH_M039 = 0x000131E8;
const char32_t EGYPTIAN_HIEROGLYPH_M040 = 0x000131E9;
const char32_t EGYPTIAN_HIEROGLYPH_M040A = 0x000131EA;
const char32_t EGYPTIAN_HIEROGLYPH_M041 = 0x000131EB;
const char32_t EGYPTIAN_HIEROGLYPH_M042 = 0x000131EC;
const char32_t EGYPTIAN_HIEROGLYPH_M043 = 0x000131ED;
const char32_t EGYPTIAN_HIEROGLYPH_M044 = 0x000131EE;
const char32_t EGYPTIAN_HIEROGLYPH_N001 = 0x000131EF;
const char32_t EGYPTIAN_HIEROGLYPH_N002 = 0x000131F0;
const char32_t EGYPTIAN_HIEROGLYPH_N003 = 0x000131F1;
const char32_t EGYPTIAN_HIEROGLYPH_N004 = 0x000131F2;
const char32_t EGYPTIAN_HIEROGLYPH_N005 = 0x000131F3;
const char32_t EGYPTIAN_HIEROGLYPH_N006 = 0x000131F4;
const char32_t EGYPTIAN_HIEROGLYPH_N007 = 0x000131F5;
const char32_t EGYPTIAN_HIEROGLYPH_N008 = 0x000131F6;
const char32_t EGYPTIAN_HIEROGLYPH_N009 = 0x000131F7;
const char32_t EGYPTIAN_HIEROGLYPH_N010 = 0x000131F8;
const char32_t EGYPTIAN_HIEROGLYPH_N011 = 0x000131F9;
const char32_t EGYPTIAN_HIEROGLYPH_N012 = 0x000131FA;
const char32_t EGYPTIAN_HIEROGLYPH_N013 = 0x000131FB;
const char32_t EGYPTIAN_HIEROGLYPH_N014 = 0x000131FC;
const char32_t EGYPTIAN_HIEROGLYPH_N015 = 0x000131FD;
const char32_t EGYPTIAN_HIEROGLYPH_N016 = 0x000131FE;
const char32_t EGYPTIAN_HIEROGLYPH_N017 = 0x000131FF;
const char32_t EGYPTIAN_HIEROGLYPH_N018 = 0x00013200;
const char32_t EGYPTIAN_HIEROGLYPH_N018A = 0x00013201;
const char32_t EGYPTIAN_HIEROGLYPH_N018B = 0x00013202;
const char32_t EGYPTIAN_HIEROGLYPH_N019 = 0x00013203;
const char32_t EGYPTIAN_HIEROGLYPH_N020 = 0x00013204;
const char32_t EGYPTIAN_HIEROGLYPH_N021 = 0x00013205;
const char32_t EGYPTIAN_HIEROGLYPH_N022 = 0x00013206;
const char32_t EGYPTIAN_HIEROGLYPH_N023 = 0x00013207;
const char32_t EGYPTIAN_HIEROGLYPH_N024 = 0x00013208;
const char32_t EGYPTIAN_HIEROGLYPH_N025 = 0x00013209;
const char32_t EGYPTIAN_HIEROGLYPH_N025A = 0x0001320A;
const char32_t EGYPTIAN_HIEROGLYPH_N026 = 0x0001320B;
const char32_t EGYPTIAN_HIEROGLYPH_N027 = 0x0001320C;
const char32_t EGYPTIAN_HIEROGLYPH_N028 = 0x0001320D;
const char32_t EGYPTIAN_HIEROGLYPH_N029 = 0x0001320E;
const char32_t EGYPTIAN_HIEROGLYPH_N030 = 0x0001320F;
const char32_t EGYPTIAN_HIEROGLYPH_N031 = 0x00013210;
const char32_t EGYPTIAN_HIEROGLYPH_N032 = 0x00013211;
const char32_t EGYPTIAN_HIEROGLYPH_N033 = 0x00013212;
const char32_t EGYPTIAN_HIEROGLYPH_N033A = 0x00013213;
const char32_t EGYPTIAN_HIEROGLYPH_N034 = 0x00013214;
const char32_t EGYPTIAN_HIEROGLYPH_N034A = 0x00013215;
const char32_t EGYPTIAN_HIEROGLYPH_N035 = 0x00013216;
const char32_t EGYPTIAN_HIEROGLYPH_N035A = 0x00013217;
const char32_t EGYPTIAN_HIEROGLYPH_N036 = 0x00013218;
const char32_t EGYPTIAN_HIEROGLYPH_N037 = 0x00013219;
const char32_t EGYPTIAN_HIEROGLYPH_N037A = 0x0001321A;
const char32_t EGYPTIAN_HIEROGLYPH_N038 = 0x0001321B;
const char32_t EGYPTIAN_HIEROGLYPH_N039 = 0x0001321C;
const char32_t EGYPTIAN_HIEROGLYPH_N040 = 0x0001321D;
const char32_t EGYPTIAN_HIEROGLYPH_N041 = 0x0001321E;
const char32_t EGYPTIAN_HIEROGLYPH_N042 = 0x0001321F;
const char32_t EGYPTIAN_HIEROGLYPH_NL001 = 0x00013220;
const char32_t EGYPTIAN_HIEROGLYPH_NL002 = 0x00013221;
const char32_t EGYPTIAN_HIEROGLYPH_NL003 = 0x00013222;
const char32_t EGYPTIAN_HIEROGLYPH_NL004 = 0x00013223;
const char32_t EGYPTIAN_HIEROGLYPH_NL005 = 0x00013224;
const char32_t EGYPTIAN_HIEROGLYPH_NL005A = 0x00013225;
const char32_t EGYPTIAN_HIEROGLYPH_NL006 = 0x00013226;
const char32_t EGYPTIAN_HIEROGLYPH_NL007 = 0x00013227;
const char32_t EGYPTIAN_HIEROGLYPH_NL008 = 0x00013228;
const char32_t EGYPTIAN_HIEROGLYPH_NL009 = 0x00013229;
const char32_t EGYPTIAN_HIEROGLYPH_NL010 = 0x0001322A;
const char32_t EGYPTIAN_HIEROGLYPH_NL011 = 0x0001322B;
const char32_t EGYPTIAN_HIEROGLYPH_NL012 = 0x0001322C;
const char32_t EGYPTIAN_HIEROGLYPH_NL013 = 0x0001322D;
const char32_t EGYPTIAN_HIEROGLYPH_NL014 = 0x0001322E;
const char32_t EGYPTIAN_HIEROGLYPH_NL015 = 0x0001322F;
const char32_t EGYPTIAN_HIEROGLYPH_NL016 = 0x00013230;
const char32_t EGYPTIAN_HIEROGLYPH_NL017 = 0x00013231;
const char32_t EGYPTIAN_HIEROGLYPH_NL017A = 0x00013232;
const char32_t EGYPTIAN_HIEROGLYPH_NL018 = 0x00013233;
const char32_t EGYPTIAN_HIEROGLYPH_NL019 = 0x00013234;
const char32_t EGYPTIAN_HIEROGLYPH_NL020 = 0x00013235;
const char32_t EGYPTIAN_HIEROGLYPH_NU001 = 0x00013236;
const char32_t EGYPTIAN_HIEROGLYPH_NU002 = 0x00013237;
const char32_t EGYPTIAN_HIEROGLYPH_NU003 = 0x00013238;
const char32_t EGYPTIAN_HIEROGLYPH_NU004 = 0x00013239;
const char32_t EGYPTIAN_HIEROGLYPH_NU005 = 0x0001323A;
const char32_t EGYPTIAN_HIEROGLYPH_NU006 = 0x0001323B;
const char32_t EGYPTIAN_HIEROGLYPH_NU007 = 0x0001323C;
const char32_t EGYPTIAN_HIEROGLYPH_NU008 = 0x0001323D;
const char32_t EGYPTIAN_HIEROGLYPH_NU009 = 0x0001323E;
const char32_t EGYPTIAN_HIEROGLYPH_NU010 = 0x0001323F;
const char32_t EGYPTIAN_HIEROGLYPH_NU010A = 0x00013240;
const char32_t EGYPTIAN_HIEROGLYPH_NU011 = 0x00013241;
const char32_t EGYPTIAN_HIEROGLYPH_NU011A = 0x00013242;
const char32_t EGYPTIAN_HIEROGLYPH_NU012 = 0x00013243;
const char32_t EGYPTIAN_HIEROGLYPH_NU013 = 0x00013244;
const char32_t EGYPTIAN_HIEROGLYPH_NU014 = 0x00013245;
const char32_t EGYPTIAN_HIEROGLYPH_NU015 = 0x00013246;
const char32_t EGYPTIAN_HIEROGLYPH_NU016 = 0x00013247;
const char32_t EGYPTIAN_HIEROGLYPH_NU017 = 0x00013248;
const char32_t EGYPTIAN_HIEROGLYPH_NU018 = 0x00013249;
const char32_t EGYPTIAN_HIEROGLYPH_NU018A = 0x0001324A;
const char32_t EGYPTIAN_HIEROGLYPH_NU019 = 0x0001324B;
const char32_t EGYPTIAN_HIEROGLYPH_NU020 = 0x0001324C;
const char32_t EGYPTIAN_HIEROGLYPH_NU021 = 0x0001324D;
const char32_t EGYPTIAN_HIEROGLYPH_NU022 = 0x0001324E;
const char32_t EGYPTIAN_HIEROGLYPH_NU022A = 0x0001324F;
const char32_t EGYPTIAN_HIEROGLYPH_O001 = 0x00013250;
const char32_t EGYPTIAN_HIEROGLYPH_O001A = 0x00013251;
const char32_t EGYPTIAN_HIEROGLYPH_O002 = 0x00013252;
const char32_t EGYPTIAN_HIEROGLYPH_O003 = 0x00013253;
const char32_t EGYPTIAN_HIEROGLYPH_O004 = 0x00013254;
const char32_t EGYPTIAN_HIEROGLYPH_O005 = 0x00013255;
const char32_t EGYPTIAN_HIEROGLYPH_O005A = 0x00013256;
const char32_t EGYPTIAN_HIEROGLYPH_O006 = 0x00013257;
const char32_t EGYPTIAN_HIEROGLYPH_O006A = 0x00013258;
const char32_t EGYPTIAN_HIEROGLYPH_O006B = 0x00013259;
const char32_t EGYPTIAN_HIEROGLYPH_O006C = 0x0001325A;
const char32_t EGYPTIAN_HIEROGLYPH_O006D = 0x0001325B;
const char32_t EGYPTIAN_HIEROGLYPH_O006E = 0x0001325C;
const char32_t EGYPTIAN_HIEROGLYPH_O006F = 0x0001325D;
const char32_t EGYPTIAN_HIEROGLYPH_O007 = 0x0001325E;
const char32_t EGYPTIAN_HIEROGLYPH_O008 = 0x0001325F;
const char32_t EGYPTIAN_HIEROGLYPH_O009 = 0x00013260;
const char32_t EGYPTIAN_HIEROGLYPH_O010 = 0x00013261;
const char32_t EGYPTIAN_HIEROGLYPH_O010A = 0x00013262;
const char32_t EGYPTIAN_HIEROGLYPH_O010B = 0x00013263;
const char32_t EGYPTIAN_HIEROGLYPH_O010C = 0x00013264;
const char32_t EGYPTIAN_HIEROGLYPH_O011 = 0x00013265;
const char32_t EGYPTIAN_HIEROGLYPH_O012 = 0x00013266;
const char32_t EGYPTIAN_HIEROGLYPH_O013 = 0x00013267;
const char32_t EGYPTIAN_HIEROGLYPH_O014 = 0x00013268;
const char32_t EGYPTIAN_HIEROGLYPH_O015 = 0x00013269;
const char32_t EGYPTIAN_HIEROGLYPH_O016 = 0x0001326A;
const char32_t EGYPTIAN_HIEROGLYPH_O017 = 0x0001326B;
const char32_t EGYPTIAN_HIEROGLYPH_O018 = 0x0001326C;
const char32_t EGYPTIAN_HIEROGLYPH_O019 = 0x0001326D;
const char32_t EGYPTIAN_HIEROGLYPH_O019A = 0x0001326E;
const char32_t EGYPTIAN_HIEROGLYPH_O020 = 0x0001326F;
const char32_t EGYPTIAN_HIEROGLYPH_O020A = 0x00013270;
const char32_t EGYPTIAN_HIEROGLYPH_O021 = 0x00013271;
const char32_t EGYPTIAN_HIEROGLYPH_O022 = 0x00013272;
const char32_t EGYPTIAN_HIEROGLYPH_O023 = 0x00013273;
const char32_t EGYPTIAN_HIEROGLYPH_O024 = 0x00013274;
const char32_t EGYPTIAN_HIEROGLYPH_O024A = 0x00013275;
const char32_t EGYPTIAN_HIEROGLYPH_O025 = 0x00013276;
const char32_t EGYPTIAN_HIEROGLYPH_O025A = 0x00013277;
const char32_t EGYPTIAN_HIEROGLYPH_O026 = 0x00013278;
const char32_t EGYPTIAN_HIEROGLYPH_O027 = 0x00013279;
const char32_t EGYPTIAN_HIEROGLYPH_O028 = 0x0001327A;
const char32_t EGYPTIAN_HIEROGLYPH_O029 = 0x0001327B;
const char32_t EGYPTIAN_HIEROGLYPH_O029A = 0x0001327C;
const char32_t EGYPTIAN_HIEROGLYPH_O030 = 0x0001327D;
const char32_t EGYPTIAN_HIEROGLYPH_O030A = 0x0001327E;
const char32_t EGYPTIAN_HIEROGLYPH_O031 = 0x0001327F;
const char32_t EGYPTIAN_HIEROGLYPH_O032 = 0x00013280;
const char32_t EGYPTIAN_HIEROGLYPH_O033 = 0x00013281;
const char32_t EGYPTIAN_HIEROGLYPH_O033A = 0x00013282;
const char32_t EGYPTIAN_HIEROGLYPH_O034 = 0x00013283;
const char32_t EGYPTIAN_HIEROGLYPH_O035 = 0x00013284;
const char32_t EGYPTIAN_HIEROGLYPH_O036 = 0x00013285;
const char32_t EGYPTIAN_HIEROGLYPH_O036A = 0x00013286;
const char32_t EGYPTIAN_HIEROGLYPH_O036B = 0x00013287;
const char32_t EGYPTIAN_HIEROGLYPH_O036C = 0x00013288;
const char32_t EGYPTIAN_HIEROGLYPH_O036D = 0x00013289;
const char32_t EGYPTIAN_HIEROGLYPH_O037 = 0x0001328A;
const char32_t EGYPTIAN_HIEROGLYPH_O038 = 0x0001328B;
const char32_t EGYPTIAN_HIEROGLYPH_O039 = 0x0001328C;
const char32_t EGYPTIAN_HIEROGLYPH_O040 = 0x0001328D;
const char32_t EGYPTIAN_HIEROGLYPH_O041 = 0x0001328E;
const char32_t EGYPTIAN_HIEROGLYPH_O042 = 0x0001328F;
const char32_t EGYPTIAN_HIEROGLYPH_O043 = 0x00013290;
const char32_t EGYPTIAN_HIEROGLYPH_O044 = 0x00013291;
const char32_t EGYPTIAN_HIEROGLYPH_O045 = 0x00013292;
const char32_t EGYPTIAN_HIEROGLYPH_O046 = 0x00013293;
const char32_t EGYPTIAN_HIEROGLYPH_O047 = 0x00013294;
const char32_t EGYPTIAN_HIEROGLYPH_O048 = 0x00013295;
const char32_t EGYPTIAN_HIEROGLYPH_O049 = 0x00013296;
const char32_t EGYPTIAN_HIEROGLYPH_O050 = 0x00013297;
const char32_t EGYPTIAN_HIEROGLYPH_O050A = 0x00013298;
const char32_t EGYPTIAN_HIEROGLYPH_O050B = 0x00013299;
const char32_t EGYPTIAN_HIEROGLYPH_O051 = 0x0001329A;
const char32_t EGYPTIAN_HIEROGLYPH_P001 = 0x0001329B;
const char32_t EGYPTIAN_HIEROGLYPH_P001A = 0x0001329C;
const char32_t EGYPTIAN_HIEROGLYPH_P002 = 0x0001329D;
const char32_t EGYPTIAN_HIEROGLYPH_P003 = 0x0001329E;
const char32_t EGYPTIAN_HIEROGLYPH_P003A = 0x0001329F;
const char32_t EGYPTIAN_HIEROGLYPH_P004 = 0x000132A0;
const char32_t EGYPTIAN_HIEROGLYPH_P005 = 0x000132A1;
const char32_t EGYPTIAN_HIEROGLYPH_P006 = 0x000132A2;
const char32_t EGYPTIAN_HIEROGLYPH_P007 = 0x000132A3;
const char32_t EGYPTIAN_HIEROGLYPH_P008 = 0x000132A4;
const char32_t EGYPTIAN_HIEROGLYPH_P009 = 0x000132A5;
const char32_t EGYPTIAN_HIEROGLYPH_P010 = 0x000132A6;
const char32_t EGYPTIAN_HIEROGLYPH_P011 = 0x000132A7;
const char32_t EGYPTIAN_HIEROGLYPH_Q001 = 0x000132A8;
const char32_t EGYPTIAN_HIEROGLYPH_Q002 = 0x000132A9;
const char32_t EGYPTIAN_HIEROGLYPH_Q003 = 0x000132AA;
const char32_t EGYPTIAN_HIEROGLYPH_Q004 = 0x000132AB;
const char32_t EGYPTIAN_HIEROGLYPH_Q005 = 0x000132AC;
const char32_t EGYPTIAN_HIEROGLYPH_Q006 = 0x000132AD;
const char32_t EGYPTIAN_HIEROGLYPH_Q007 = 0x000132AE;
const char32_t EGYPTIAN_HIEROGLYPH_R001 = 0x000132AF;
const char32_t EGYPTIAN_HIEROGLYPH_R002 = 0x000132B0;
const char32_t EGYPTIAN_HIEROGLYPH_R002A = 0x000132B1;
const char32_t EGYPTIAN_HIEROGLYPH_R003 = 0x000132B2;
const char32_t EGYPTIAN_HIEROGLYPH_R003A = 0x000132B3;
const char32_t EGYPTIAN_HIEROGLYPH_R003B = 0x000132B4;
const char32_t EGYPTIAN_HIEROGLYPH_R004 = 0x000132B5;
const char32_t EGYPTIAN_HIEROGLYPH_R005 = 0x000132B6;
const char32_t EGYPTIAN_HIEROGLYPH_R006 = 0x000132B7;
const char32_t EGYPTIAN_HIEROGLYPH_R007 = 0x000132B8;
const char32_t EGYPTIAN_HIEROGLYPH_R008 = 0x000132B9;
const char32_t EGYPTIAN_HIEROGLYPH_R009 = 0x000132BA;
const char32_t EGYPTIAN_HIEROGLYPH_R010 = 0x000132BB;
const char32_t EGYPTIAN_HIEROGLYPH_R010A = 0x000132BC;
const char32_t EGYPTIAN_HIEROGLYPH_R011 = 0x000132BD;
const char32_t EGYPTIAN_HIEROGLYPH_R012 = 0x000132BE;
const char32_t EGYPTIAN_HIEROGLYPH_R013 = 0x000132BF;
const char32_t EGYPTIAN_HIEROGLYPH_R014 = 0x000132C0;
const char32_t EGYPTIAN_HIEROGLYPH_R015 = 0x000132C1;
const char32_t EGYPTIAN_HIEROGLYPH_R016 = 0x000132C2;
const char32_t EGYPTIAN_HIEROGLYPH_R016A = 0x000132C3;
const char32_t EGYPTIAN_HIEROGLYPH_R017 = 0x000132C4;
const char32_t EGYPTIAN_HIEROGLYPH_R018 = 0x000132C5;
const char32_t EGYPTIAN_HIEROGLYPH_R019 = 0x000132C6;
const char32_t EGYPTIAN_HIEROGLYPH_R020 = 0x000132C7;
const char32_t EGYPTIAN_HIEROGLYPH_R021 = 0x000132C8;
const char32_t EGYPTIAN_HIEROGLYPH_R022 = 0x000132C9;
const char32_t EGYPTIAN_HIEROGLYPH_R023 = 0x000132CA;
const char32_t EGYPTIAN_HIEROGLYPH_R024 = 0x000132CB;
const char32_t EGYPTIAN_HIEROGLYPH_R025 = 0x000132CC;
const char32_t EGYPTIAN_HIEROGLYPH_R026 = 0x000132CD;
const char32_t EGYPTIAN_HIEROGLYPH_R027 = 0x000132CE;
const char32_t EGYPTIAN_HIEROGLYPH_R028 = 0x000132CF;
const char32_t EGYPTIAN_HIEROGLYPH_R029 = 0x000132D0;
const char32_t EGYPTIAN_HIEROGLYPH_S001 = 0x000132D1;
const char32_t EGYPTIAN_HIEROGLYPH_S002 = 0x000132D2;
const char32_t EGYPTIAN_HIEROGLYPH_S002A = 0x000132D3;
const char32_t EGYPTIAN_HIEROGLYPH_S003 = 0x000132D4;
const char32_t EGYPTIAN_HIEROGLYPH_S004 = 0x000132D5;
const char32_t EGYPTIAN_HIEROGLYPH_S005 = 0x000132D6;
const char32_t EGYPTIAN_HIEROGLYPH_S006 = 0x000132D7;
const char32_t EGYPTIAN_HIEROGLYPH_S006A = 0x000132D8;
const char32_t EGYPTIAN_HIEROGLYPH_S007 = 0x000132D9;
const char32_t EGYPTIAN_HIEROGLYPH_S008 = 0x000132DA;
const char32_t EGYPTIAN_HIEROGLYPH_S009 = 0x000132DB;
const char32_t EGYPTIAN_HIEROGLYPH_S010 = 0x000132DC;
const char32_t EGYPTIAN_HIEROGLYPH_S011 = 0x000132DD;
const char32_t EGYPTIAN_HIEROGLYPH_S012 = 0x000132DE;
const char32_t EGYPTIAN_HIEROGLYPH_S013 = 0x000132DF;
const char32_t EGYPTIAN_HIEROGLYPH_S014 = 0x000132E0;
const char32_t EGYPTIAN_HIEROGLYPH_S014A = 0x000132E1;
const char32_t EGYPTIAN_HIEROGLYPH_S014B = 0x000132E2;
const char32_t EGYPTIAN_HIEROGLYPH_S015 = 0x000132E3;
const char32_t EGYPTIAN_HIEROGLYPH_S016 = 0x000132E4;
const char32_t EGYPTIAN_HIEROGLYPH_S017 = 0x000132E5;
const char32_t EGYPTIAN_HIEROGLYPH_S017A = 0x000132E6;
const char32_t EGYPTIAN_HIEROGLYPH_S018 = 0x000132E7;
const char32_t EGYPTIAN_HIEROGLYPH_S019 = 0x000132E8;
const char32_t EGYPTIAN_HIEROGLYPH_S020 = 0x000132E9;
const char32_t EGYPTIAN_HIEROGLYPH_S021 = 0x000132EA;
const char32_t EGYPTIAN_HIEROGLYPH_S022 = 0x000132EB;
const char32_t EGYPTIAN_HIEROGLYPH_S023 = 0x000132EC;
const char32_t EGYPTIAN_HIEROGLYPH_S024 = 0x000132ED;
const char32_t EGYPTIAN_HIEROGLYPH_S025 = 0x000132EE;
const char32_t EGYPTIAN_HIEROGLYPH_S026 = 0x000132EF;
const char32_t EGYPTIAN_HIEROGLYPH_S026A = 0x000132F0;
const char32_t EGYPTIAN_HIEROGLYPH_S026B = 0x000132F1;
const char32_t EGYPTIAN_HIEROGLYPH_S027 = 0x000132F2;
const char32_t EGYPTIAN_HIEROGLYPH_S028 = 0x000132F3;
const char32_t EGYPTIAN_HIEROGLYPH_S029 = 0x000132F4;
const char32_t EGYPTIAN_HIEROGLYPH_S030 = 0x000132F5;
const char32_t EGYPTIAN_HIEROGLYPH_S031 = 0x000132F6;
const char32_t EGYPTIAN_HIEROGLYPH_S032 = 0x000132F7;
const char32_t EGYPTIAN_HIEROGLYPH_S033 = 0x000132F8;
const char32_t EGYPTIAN_HIEROGLYPH_S034 = 0x000132F9;
const char32_t EGYPTIAN_HIEROGLYPH_S035 = 0x000132FA;
const char32_t EGYPTIAN_HIEROGLYPH_S035A = 0x000132FB;
const char32_t EGYPTIAN_HIEROGLYPH_S036 = 0x000132FC;
const char32_t EGYPTIAN_HIEROGLYPH_S037 = 0x000132FD;
const char32_t EGYPTIAN_HIEROGLYPH_S038 = 0x000132FE;
const char32_t EGYPTIAN_HIEROGLYPH_S039 = 0x000132FF;
const char32_t EGYPTIAN_HIEROGLYPH_S040 = 0x00013300;
const char32_t EGYPTIAN_HIEROGLYPH_S041 = 0x00013301;
const char32_t EGYPTIAN_HIEROGLYPH_S042 = 0x00013302;
const char32_t EGYPTIAN_HIEROGLYPH_S043 = 0x00013303;
const char32_t EGYPTIAN_HIEROGLYPH_S044 = 0x00013304;
const char32_t EGYPTIAN_HIEROGLYPH_S045 = 0x00013305;
const char32_t EGYPTIAN_HIEROGLYPH_S046 = 0x00013306;
const char32_t EGYPTIAN_HIEROGLYPH_T001 = 0x00013307;
const char32_t EGYPTIAN_HIEROGLYPH_T002 = 0x00013308;
const char32_t EGYPTIAN_HIEROGLYPH_T003 = 0x00013309;
const char32_t EGYPTIAN_HIEROGLYPH_T003A = 0x0001330A;
const char32_t EGYPTIAN_HIEROGLYPH_T004 = 0x0001330B;
const char32_t EGYPTIAN_HIEROGLYPH_T005 = 0x0001330C;
const char32_t EGYPTIAN_HIEROGLYPH_T006 = 0x0001330D;
const char32_t EGYPTIAN_HIEROGLYPH_T007 = 0x0001330E;
const char32_t EGYPTIAN_HIEROGLYPH_T007A = 0x0001330F;
const char32_t EGYPTIAN_HIEROGLYPH_T008 = 0x00013310;
const char32_t EGYPTIAN_HIEROGLYPH_T008A = 0x00013311;
const char32_t EGYPTIAN_HIEROGLYPH_T009 = 0x00013312;
const char32_t EGYPTIAN_HIEROGLYPH_T009A = 0x00013313;
const char32_t EGYPTIAN_HIEROGLYPH_T010 = 0x00013314;
const char32_t EGYPTIAN_HIEROGLYPH_T011 = 0x00013315;
const char32_t EGYPTIAN_HIEROGLYPH_T011A = 0x00013316;
const char32_t EGYPTIAN_HIEROGLYPH_T012 = 0x00013317;
const char32_t EGYPTIAN_HIEROGLYPH_T013 = 0x00013318;
const char32_t EGYPTIAN_HIEROGLYPH_T014 = 0x00013319;
const char32_t EGYPTIAN_HIEROGLYPH_T015 = 0x0001331A;
const char32_t EGYPTIAN_HIEROGLYPH_T016 = 0x0001331B;
const char32_t EGYPTIAN_HIEROGLYPH_T016A = 0x0001331C;
const char32_t EGYPTIAN_HIEROGLYPH_T017 = 0x0001331D;
const char32_t EGYPTIAN_HIEROGLYPH_T018 = 0x0001331E;
const char32_t EGYPTIAN_HIEROGLYPH_T019 = 0x0001331F;
const char32_t EGYPTIAN_HIEROGLYPH_T020 = 0x00013320;
const char32_t EGYPTIAN_HIEROGLYPH_T021 = 0x00013321;
const char32_t EGYPTIAN_HIEROGLYPH_T022 = 0x00013322;
const char32_t EGYPTIAN_HIEROGLYPH_T023 = 0x00013323;
const char32_t EGYPTIAN_HIEROGLYPH_T024 = 0x00013324;
const char32_t EGYPTIAN_HIEROGLYPH_T025 = 0x00013325;
const char32_t EGYPTIAN_HIEROGLYPH_T026 = 0x00013326;
const char32_t EGYPTIAN_HIEROGLYPH_T027 = 0x00013327;
const char32_t EGYPTIAN_HIEROGLYPH_T028 = 0x00013328;
const char32_t EGYPTIAN_HIEROGLYPH_T029 = 0x00013329;
const char32_t EGYPTIAN_HIEROGLYPH_T030 = 0x0001332A;
const char32_t EGYPTIAN_HIEROGLYPH_T031 = 0x0001332B;
const char32_t EGYPTIAN_HIEROGLYPH_T032 = 0x0001332C;
const char32_t EGYPTIAN_HIEROGLYPH_T032A = 0x0001332D;
const char32_t EGYPTIAN_HIEROGLYPH_T033 = 0x0001332E;
const char32_t EGYPTIAN_HIEROGLYPH_T033A = 0x0001332F;
const char32_t EGYPTIAN_HIEROGLYPH_T034 = 0x00013330;
const char32_t EGYPTIAN_HIEROGLYPH_T035 = 0x00013331;
const char32_t EGYPTIAN_HIEROGLYPH_T036 = 0x00013332;
const char32_t EGYPTIAN_HIEROGLYPH_U001 = 0x00013333;
const char32_t EGYPTIAN_HIEROGLYPH_U002 = 0x00013334;
const char32_t EGYPTIAN_HIEROGLYPH_U003 = 0x00013335;
const char32_t EGYPTIAN_HIEROGLYPH_U004 = 0x00013336;
const char32_t EGYPTIAN_HIEROGLYPH_U005 = 0x00013337;
const char32_t EGYPTIAN_HIEROGLYPH_U006 = 0x00013338;
const char32_t EGYPTIAN_HIEROGLYPH_U006A = 0x00013339;
const char32_t EGYPTIAN_HIEROGLYPH_U006B = 0x0001333A;
const char32_t EGYPTIAN_HIEROGLYPH_U007 = 0x0001333B;
const char32_t EGYPTIAN_HIEROGLYPH_U008 = 0x0001333C;
const char32_t EGYPTIAN_HIEROGLYPH_U009 = 0x0001333D;
const char32_t EGYPTIAN_HIEROGLYPH_U010 = 0x0001333E;
const char32_t EGYPTIAN_HIEROGLYPH_U011 = 0x0001333F;
const char32_t EGYPTIAN_HIEROGLYPH_U012 = 0x00013340;
const char32_t EGYPTIAN_HIEROGLYPH_U013 = 0x00013341;
const char32_t EGYPTIAN_HIEROGLYPH_U014 = 0x00013342;
const char32_t EGYPTIAN_HIEROGLYPH_U015 = 0x00013343;
const char32_t EGYPTIAN_HIEROGLYPH_U016 = 0x00013344;
const char32_t EGYPTIAN_HIEROGLYPH_U017 = 0x00013345;
const char32_t EGYPTIAN_HIEROGLYPH_U018 = 0x00013346;
const char32_t EGYPTIAN_HIEROGLYPH_U019 = 0x00013347;
const char32_t EGYPTIAN_HIEROGLYPH_U020 = 0x00013348;
const char32_t EGYPTIAN_HIEROGLYPH_U021 = 0x00013349;
const char32_t EGYPTIAN_HIEROGLYPH_U022 = 0x0001334A;
const char32_t EGYPTIAN_HIEROGLYPH_U023 = 0x0001334B;
const char32_t EGYPTIAN_HIEROGLYPH_U023A = 0x0001334C;
const char32_t EGYPTIAN_HIEROGLYPH_U024 = 0x0001334D;
const char32_t EGYPTIAN_HIEROGLYPH_U025 = 0x0001334E;
const char32_t EGYPTIAN_HIEROGLYPH_U026 = 0x0001334F;
const char32_t EGYPTIAN_HIEROGLYPH_U027 = 0x00013350;
const char32_t EGYPTIAN_HIEROGLYPH_U028 = 0x00013351;
const char32_t EGYPTIAN_HIEROGLYPH_U029 = 0x00013352;
const char32_t EGYPTIAN_HIEROGLYPH_U029A = 0x00013353;
const char32_t EGYPTIAN_HIEROGLYPH_U030 = 0x00013354;
const char32_t EGYPTIAN_HIEROGLYPH_U031 = 0x00013355;
const char32_t EGYPTIAN_HIEROGLYPH_U032 = 0x00013356;
const char32_t EGYPTIAN_HIEROGLYPH_U032A = 0x00013357;
const char32_t EGYPTIAN_HIEROGLYPH_U033 = 0x00013358;
const char32_t EGYPTIAN_HIEROGLYPH_U034 = 0x00013359;
const char32_t EGYPTIAN_HIEROGLYPH_U035 = 0x0001335A;
const char32_t EGYPTIAN_HIEROGLYPH_U036 = 0x0001335B;
const char32_t EGYPTIAN_HIEROGLYPH_U037 = 0x0001335C;
const char32_t EGYPTIAN_HIEROGLYPH_U038 = 0x0001335D;
const char32_t EGYPTIAN_HIEROGLYPH_U039 = 0x0001335E;
const char32_t EGYPTIAN_HIEROGLYPH_U040 = 0x0001335F;
const char32_t EGYPTIAN_HIEROGLYPH_U041 = 0x00013360;
const char32_t EGYPTIAN_HIEROGLYPH_U042 = 0x00013361;
const char32_t EGYPTIAN_HIEROGLYPH_V001 = 0x00013362;
const char32_t EGYPTIAN_HIEROGLYPH_V001A = 0x00013363;
const char32_t EGYPTIAN_HIEROGLYPH_V001B = 0x00013364;
const char32_t EGYPTIAN_HIEROGLYPH_V001C = 0x00013365;
const char32_t EGYPTIAN_HIEROGLYPH_V001D = 0x00013366;
const char32_t EGYPTIAN_HIEROGLYPH_V001E = 0x00013367;
const char32_t EGYPTIAN_HIEROGLYPH_V001F = 0x00013368;
const char32_t EGYPTIAN_HIEROGLYPH_V001G = 0x00013369;
const char32_t EGYPTIAN_HIEROGLYPH_V001H = 0x0001336A;
const char32_t EGYPTIAN_HIEROGLYPH_V001I = 0x0001336B;
const char32_t EGYPTIAN_HIEROGLYPH_V002 = 0x0001336C;
const char32_t EGYPTIAN_HIEROGLYPH_V002A = 0x0001336D;
const char32_t EGYPTIAN_HIEROGLYPH_V003 = 0x0001336E;
const char32_t EGYPTIAN_HIEROGLYPH_V004 = 0x0001336F;
const char32_t EGYPTIAN_HIEROGLYPH_V005 = 0x00013370;
const char32_t EGYPTIAN_HIEROGLYPH_V006 = 0x00013371;
const char32_t EGYPTIAN_HIEROGLYPH_V007 = 0x00013372;
const char32_t EGYPTIAN_HIEROGLYPH_V007A = 0x00013373;
const char32_t EGYPTIAN_HIEROGLYPH_V007B = 0x00013374;
const char32_t EGYPTIAN_HIEROGLYPH_V008 = 0x00013375;
const char32_t EGYPTIAN_HIEROGLYPH_V009 = 0x00013376;
const char32_t EGYPTIAN_HIEROGLYPH_V010 = 0x00013377;
const char32_t EGYPTIAN_HIEROGLYPH_V011 = 0x00013378;
const char32_t EGYPTIAN_HIEROGLYPH_V011A = 0x00013379;
const char32_t EGYPTIAN_HIEROGLYPH_V011B = 0x0001337A;
const char32_t EGYPTIAN_HIEROGLYPH_V011C = 0x0001337B;
const char32_t EGYPTIAN_HIEROGLYPH_V012 = 0x0001337C;
const char32_t EGYPTIAN_HIEROGLYPH_V012A = 0x0001337D;
const char32_t EGYPTIAN_HIEROGLYPH_V012B = 0x0001337E;
const char32_t EGYPTIAN_HIEROGLYPH_V013 = 0x0001337F;
const char32_t EGYPTIAN_HIEROGLYPH_V014 = 0x00013380;
const char32_t EGYPTIAN_HIEROGLYPH_V015 = 0x00013381;
const char32_t EGYPTIAN_HIEROGLYPH_V016 = 0x00013382;
const char32_t EGYPTIAN_HIEROGLYPH_V017 = 0x00013383;
const char32_t EGYPTIAN_HIEROGLYPH_V018 = 0x00013384;
const char32_t EGYPTIAN_HIEROGLYPH_V019 = 0x00013385;
const char32_t EGYPTIAN_HIEROGLYPH_V020 = 0x00013386;
const char32_t EGYPTIAN_HIEROGLYPH_V020A = 0x00013387;
const char32_t EGYPTIAN_HIEROGLYPH_V020B = 0x00013388;
const char32_t EGYPTIAN_HIEROGLYPH_V020C = 0x00013389;
const char32_t EGYPTIAN_HIEROGLYPH_V020D = 0x0001338A;
const char32_t EGYPTIAN_HIEROGLYPH_V020E = 0x0001338B;
const char32_t EGYPTIAN_HIEROGLYPH_V020F = 0x0001338C;
const char32_t EGYPTIAN_HIEROGLYPH_V020G = 0x0001338D;
const char32_t EGYPTIAN_HIEROGLYPH_V020H = 0x0001338E;
const char32_t EGYPTIAN_HIEROGLYPH_V020I = 0x0001338F;
const char32_t EGYPTIAN_HIEROGLYPH_V020J = 0x00013390;
const char32_t EGYPTIAN_HIEROGLYPH_V020K = 0x00013391;
const char32_t EGYPTIAN_HIEROGLYPH_V020L = 0x00013392;
const char32_t EGYPTIAN_HIEROGLYPH_V021 = 0x00013393;
const char32_t EGYPTIAN_HIEROGLYPH_V022 = 0x00013394;
const char32_t EGYPTIAN_HIEROGLYPH_V023 = 0x00013395;
const char32_t EGYPTIAN_HIEROGLYPH_V023A = 0x00013396;
const char32_t EGYPTIAN_HIEROGLYPH_V024 = 0x00013397;
const char32_t EGYPTIAN_HIEROGLYPH_V025 = 0x00013398;
const char32_t EGYPTIAN_HIEROGLYPH_V026 = 0x00013399;
const char32_t EGYPTIAN_HIEROGLYPH_V027 = 0x0001339A;
const char32_t EGYPTIAN_HIEROGLYPH_V028 = 0x0001339B;
const char32_t EGYPTIAN_HIEROGLYPH_V028A = 0x0001339C;
const char32_t EGYPTIAN_HIEROGLYPH_V029 = 0x0001339D;
const char32_t EGYPTIAN_HIEROGLYPH_V029A = 0x0001339E;
const char32_t EGYPTIAN_HIEROGLYPH_V030 = 0x0001339F;
const char32_t EGYPTIAN_HIEROGLYPH_V030A = 0x000133A0;
const char32_t EGYPTIAN_HIEROGLYPH_V031 = 0x000133A1;
const char32_t EGYPTIAN_HIEROGLYPH_V031A = 0x000133A2;
const char32_t EGYPTIAN_HIEROGLYPH_V032 = 0x000133A3;
const char32_t EGYPTIAN_HIEROGLYPH_V033 = 0x000133A4;
const char32_t EGYPTIAN_HIEROGLYPH_V033A = 0x000133A5;
const char32_t EGYPTIAN_HIEROGLYPH_V034 = 0x000133A6;
const char32_t EGYPTIAN_HIEROGLYPH_V035 = 0x000133A7;
const char32_t EGYPTIAN_HIEROGLYPH_V036 = 0x000133A8;
const char32_t EGYPTIAN_HIEROGLYPH_V037 = 0x000133A9;
const char32_t EGYPTIAN_HIEROGLYPH_V037A = 0x000133AA;
const char32_t EGYPTIAN_HIEROGLYPH_V038 = 0x000133AB;
const char32_t EGYPTIAN_HIEROGLYPH_V039 = 0x000133AC;
const char32_t EGYPTIAN_HIEROGLYPH_V040 = 0x000133AD;
const char32_t EGYPTIAN_HIEROGLYPH_V040A = 0x000133AE;
const char32_t EGYPTIAN_HIEROGLYPH_W001 = 0x000133AF;
const char32_t EGYPTIAN_HIEROGLYPH_W002 = 0x000133B0;
const char32_t EGYPTIAN_HIEROGLYPH_W003 = 0x000133B1;
const char32_t EGYPTIAN_HIEROGLYPH_W003A = 0x000133B2;
const char32_t EGYPTIAN_HIEROGLYPH_W004 = 0x000133B3;
const char32_t EGYPTIAN_HIEROGLYPH_W005 = 0x000133B4;
const char32_t EGYPTIAN_HIEROGLYPH_W006 = 0x000133B5;
const char32_t EGYPTIAN_HIEROGLYPH_W007 = 0x000133B6;
const char32_t EGYPTIAN_HIEROGLYPH_W008 = 0x000133B7;
const char32_t EGYPTIAN_HIEROGLYPH_W009 = 0x000133B8;
const char32_t EGYPTIAN_HIEROGLYPH_W009A = 0x000133B9;
const char32_t EGYPTIAN_HIEROGLYPH_W010 = 0x000133BA;
const char32_t EGYPTIAN_HIEROGLYPH_W010A = 0x000133BB;
const char32_t EGYPTIAN_HIEROGLYPH_W011 = 0x000133BC;
const char32_t EGYPTIAN_HIEROGLYPH_W012 = 0x000133BD;
const char32_t EGYPTIAN_HIEROGLYPH_W013 = 0x000133BE;
const char32_t EGYPTIAN_HIEROGLYPH_W014 = 0x000133BF;
const char32_t EGYPTIAN_HIEROGLYPH_W014A = 0x000133C0;
const char32_t EGYPTIAN_HIEROGLYPH_W015 = 0x000133C1;
const char32_t EGYPTIAN_HIEROGLYPH_W016 = 0x000133C2;
const char32_t EGYPTIAN_HIEROGLYPH_W017 = 0x000133C3;
const char32_t EGYPTIAN_HIEROGLYPH_W017A = 0x000133C4;
const char32_t EGYPTIAN_HIEROGLYPH_W018 = 0x000133C5;
const char32_t EGYPTIAN_HIEROGLYPH_W018A = 0x000133C6;
const char32_t EGYPTIAN_HIEROGLYPH_W019 = 0x000133C7;
const char32_t EGYPTIAN_HIEROGLYPH_W020 = 0x000133C8;
const char32_t EGYPTIAN_HIEROGLYPH_W021 = 0x000133C9;
const char32_t EGYPTIAN_HIEROGLYPH_W022 = 0x000133CA;
const char32_t EGYPTIAN_HIEROGLYPH_W023 = 0x000133CB;
const char32_t EGYPTIAN_HIEROGLYPH_W024 = 0x000133CC;
const char32_t EGYPTIAN_HIEROGLYPH_W024A = 0x000133CD;
const char32_t EGYPTIAN_HIEROGLYPH_W025 = 0x000133CE;
const char32_t EGYPTIAN_HIEROGLYPH_X001 = 0x000133CF;
const char32_t EGYPTIAN_HIEROGLYPH_X002 = 0x000133D0;
const char32_t EGYPTIAN_HIEROGLYPH_X003 = 0x000133D1;
const char32_t EGYPTIAN_HIEROGLYPH_X004 = 0x000133D2;
const char32_t EGYPTIAN_HIEROGLYPH_X004A = 0x000133D3;
const char32_t EGYPTIAN_HIEROGLYPH_X004B = 0x000133D4;
const char32_t EGYPTIAN_HIEROGLYPH_X005 = 0x000133D5;
const char32_t EGYPTIAN_HIEROGLYPH_X006 = 0x000133D6;
const char32_t EGYPTIAN_HIEROGLYPH_X006A = 0x000133D7;
const char32_t EGYPTIAN_HIEROGLYPH_X007 = 0x000133D8;
const char32_t EGYPTIAN_HIEROGLYPH_X008 = 0x000133D9;
const char32_t EGYPTIAN_HIEROGLYPH_X008A = 0x000133DA;
const char32_t EGYPTIAN_HIEROGLYPH_Y001 = 0x000133DB;
const char32_t EGYPTIAN_HIEROGLYPH_Y001A = 0x000133DC;
const char32_t EGYPTIAN_HIEROGLYPH_Y002 = 0x000133DD;
const char32_t EGYPTIAN_HIEROGLYPH_Y003 = 0x000133DE;
const char32_t EGYPTIAN_HIEROGLYPH_Y004 = 0x000133DF;
const char32_t EGYPTIAN_HIEROGLYPH_Y005 = 0x000133E0;
const char32_t EGYPTIAN_HIEROGLYPH_Y006 = 0x000133E1;
const char32_t EGYPTIAN_HIEROGLYPH_Y007 = 0x000133E2;
const char32_t EGYPTIAN_HIEROGLYPH_Y008 = 0x000133E3;
const char32_t EGYPTIAN_HIEROGLYPH_Z001 = 0x000133E4;
const char32_t EGYPTIAN_HIEROGLYPH_Z002 = 0x000133E5;
const char32_t EGYPTIAN_HIEROGLYPH_Z002A = 0x000133E6;
const char32_t EGYPTIAN_HIEROGLYPH_Z002B = 0x000133E7;
const char32_t EGYPTIAN_HIEROGLYPH_Z002C = 0x000133E8;
const char32_t EGYPTIAN_HIEROGLYPH_Z002D = 0x000133E9;
const char32_t EGYPTIAN_HIEROGLYPH_Z003 = 0x000133EA;
const char32_t EGYPTIAN_HIEROGLYPH_Z003A = 0x000133EB;
const char32_t EGYPTIAN_HIEROGLYPH_Z003B = 0x000133EC;
const char32_t EGYPTIAN_HIEROGLYPH_Z004 = 0x000133ED;
const char32_t EGYPTIAN_HIEROGLYPH_Z004A = 0x000133EE;
const char32_t EGYPTIAN_HIEROGLYPH_Z005 = 0x000133EF;
const char32_t EGYPTIAN_HIEROGLYPH_Z005A = 0x000133F0;
const char32_t EGYPTIAN_HIEROGLYPH_Z006 = 0x000133F1;
const char32_t EGYPTIAN_HIEROGLYPH_Z007 = 0x000133F2;
const char32_t EGYPTIAN_HIEROGLYPH_Z008 = 0x000133F3;
const char32_t EGYPTIAN_HIEROGLYPH_Z009 = 0x000133F4;
const char32_t EGYPTIAN_HIEROGLYPH_Z010 = 0x000133F5;
const char32_t EGYPTIAN_HIEROGLYPH_Z011 = 0x000133F6;
const char32_t EGYPTIAN_HIEROGLYPH_Z012 = 0x000133F7;
const char32_t EGYPTIAN_HIEROGLYPH_Z013 = 0x000133F8;
const char32_t EGYPTIAN_HIEROGLYPH_Z014 = 0x000133F9;
const char32_t EGYPTIAN_HIEROGLYPH_Z015 = 0x000133FA;
const char32_t EGYPTIAN_HIEROGLYPH_Z015A = 0x000133FB;
const char32_t EGYPTIAN_HIEROGLYPH_Z015B = 0x000133FC;
const char32_t EGYPTIAN_HIEROGLYPH_Z015C = 0x000133FD;
const char32_t EGYPTIAN_HIEROGLYPH_Z015D = 0x000133FE;
const char32_t EGYPTIAN_HIEROGLYPH_Z015E = 0x000133FF;
const char32_t EGYPTIAN_HIEROGLYPH_Z015F = 0x00013400;
const char32_t EGYPTIAN_HIEROGLYPH_Z015G = 0x00013401;
const char32_t EGYPTIAN_HIEROGLYPH_Z015H = 0x00013402;
const char32_t EGYPTIAN_HIEROGLYPH_Z015I = 0x00013403;
const char32_t EGYPTIAN_HIEROGLYPH_Z016 = 0x00013404;
const char32_t EGYPTIAN_HIEROGLYPH_Z016A = 0x00013405;
const char32_t EGYPTIAN_HIEROGLYPH_Z016B = 0x00013406;
const char32_t EGYPTIAN_HIEROGLYPH_Z016C = 0x00013407;
const char32_t EGYPTIAN_HIEROGLYPH_Z016D = 0x00013408;
const char32_t EGYPTIAN_HIEROGLYPH_Z016E = 0x00013409;
const char32_t EGYPTIAN_HIEROGLYPH_Z016F = 0x0001340A;
const char32_t EGYPTIAN_HIEROGLYPH_Z016G = 0x0001340B;
const char32_t EGYPTIAN_HIEROGLYPH_Z016H = 0x0001340C;
const char32_t EGYPTIAN_HIEROGLYPH_AA001 = 0x0001340D;
const char32_t EGYPTIAN_HIEROGLYPH_AA002 = 0x0001340E;
const char32_t EGYPTIAN_HIEROGLYPH_AA003 = 0x0001340F;
const char32_t EGYPTIAN_HIEROGLYPH_AA004 = 0x00013410;
const char32_t EGYPTIAN_HIEROGLYPH_AA005 = 0x00013411;
const char32_t EGYPTIAN_HIEROGLYPH_AA006 = 0x00013412;
const char32_t EGYPTIAN_HIEROGLYPH_AA007 = 0x00013413;
const char32_t EGYPTIAN_HIEROGLYPH_AA007A = 0x00013414;
const char32_t EGYPTIAN_HIEROGLYPH_AA007B = 0x00013415;
const char32_t EGYPTIAN_HIEROGLYPH_AA008 = 0x00013416;
const char32_t EGYPTIAN_HIEROGLYPH_AA009 = 0x00013417;
const char32_t EGYPTIAN_HIEROGLYPH_AA010 = 0x00013418;
const char32_t EGYPTIAN_HIEROGLYPH_AA011 = 0x00013419;
const char32_t EGYPTIAN_HIEROGLYPH_AA012 = 0x0001341A;
const char32_t EGYPTIAN_HIEROGLYPH_AA013 = 0x0001341B;
const char32_t EGYPTIAN_HIEROGLYPH_AA014 = 0x0001341C;
const char32_t EGYPTIAN_HIEROGLYPH_AA015 = 0x0001341D;
const char32_t EGYPTIAN_HIEROGLYPH_AA016 = 0x0001341E;
const char32_t EGYPTIAN_HIEROGLYPH_AA017 = 0x0001341F;
const char32_t EGYPTIAN_HIEROGLYPH_AA018 = 0x00013420;
const char32_t EGYPTIAN_HIEROGLYPH_AA019 = 0x00013421;
const char32_t EGYPTIAN_HIEROGLYPH_AA020 = 0x00013422;
const char32_t EGYPTIAN_HIEROGLYPH_AA021 = 0x00013423;
const char32_t EGYPTIAN_HIEROGLYPH_AA022 = 0x00013424;
const char32_t EGYPTIAN_HIEROGLYPH_AA023 = 0x00013425;
const char32_t EGYPTIAN_HIEROGLYPH_AA024 = 0x00013426;
const char32_t EGYPTIAN_HIEROGLYPH_AA025 = 0x00013427;
const char32_t EGYPTIAN_HIEROGLYPH_AA026 = 0x00013428;
const char32_t EGYPTIAN_HIEROGLYPH_AA027 = 0x00013429;
const char32_t EGYPTIAN_HIEROGLYPH_AA028 = 0x0001342A;
const char32_t EGYPTIAN_HIEROGLYPH_AA029 = 0x0001342B;
const char32_t EGYPTIAN_HIEROGLYPH_AA030 = 0x0001342C;
const char32_t EGYPTIAN_HIEROGLYPH_AA031 = 0x0001342D;
const char32_t EGYPTIAN_HIEROGLYPH_AA032 = 0x0001342E;
const char32_t EGYPTIAN_HIEROGLYPH_V011D = 0x0001342F;
const char32_t EGYPTIAN_HIEROGLYPH_VERTICAL_JOINER = 0x00013430;
const char32_t EGYPTIAN_HIEROGLYPH_HORIZONTAL_JOINER = 0x00013431;
const char32_t EGYPTIAN_HIEROGLYPH_INSERT_AT_TOP_START = 0x00013432;
const char32_t EGYPTIAN_HIEROGLYPH_INSERT_AT_BOTTOM_START = 0x00013433;
const char32_t EGYPTIAN_HIEROGLYPH_INSERT_AT_TOP_END = 0x00013434;
const char32_t EGYPTIAN_HIEROGLYPH_INSERT_AT_BOTTOM_END = 0x00013435;
const char32_t EGYPTIAN_HIEROGLYPH_OVERLAY_MIDDLE = 0x00013436;
const char32_t EGYPTIAN_HIEROGLYPH_BEGIN_SEGMENT = 0x00013437;
const char32_t EGYPTIAN_HIEROGLYPH_END_SEGMENT = 0x00013438;
const char32_t EGYPTIAN_HIEROGLYPH_INSERT_AT_MIDDLE = 0x00013439;
const char32_t EGYPTIAN_HIEROGLYPH_INSERT_AT_TOP = 0x0001343A;
const char32_t EGYPTIAN_HIEROGLYPH_INSERT_AT_BOTTOM = 0x0001343B;
const char32_t EGYPTIAN_HIEROGLYPH_BEGIN_ENCLOSURE = 0x0001343C;
const char32_t EGYPTIAN_HIEROGLYPH_END_ENCLOSURE = 0x0001343D;
const char32_t EGYPTIAN_HIEROGLYPH_BEGIN_WALLED_ENCLOSURE = 0x0001343E;
const char32_t EGYPTIAN_HIEROGLYPH_END_WALLED_ENCLOSURE = 0x0001343F;
const char32_t EGYPTIAN_HIEROGLYPH_MIRROR_HORIZONTALLY = 0x00013440;
const char32_t EGYPTIAN_HIEROGLYPH_FULL_BLANK = 0x00013441;
const char32_t EGYPTIAN_HIEROGLYPH_HALF_BLANK = 0x00013442;
const char32_t EGYPTIAN_HIEROGLYPH_LOST_SIGN = 0x00013443;
const char32_t EGYPTIAN_HIEROGLYPH_HALF_LOST_SIGN = 0x00013444;
const char32_t EGYPTIAN_HIEROGLYPH_TALL_LOST_SIGN = 0x00013445;
const char32_t EGYPTIAN_HIEROGLYPH_WIDE_LOST_SIGN = 0x00013446;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_TOP_START = 0x00013447;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_BOTTOM_START = 0x00013448;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_START = 0x00013449;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_TOP_END = 0x0001344A;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_TOP = 0x0001344B;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_BOTTOM_START_AND_TOP_END = 0x0001344C;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_START_AND_TOP = 0x0001344D;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_BOTTOM_END = 0x0001344E;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_TOP_START_AND_BOTTOM_END = 0x0001344F;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_BOTTOM = 0x00013450;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_START_AND_BOTTOM = 0x00013451;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_END = 0x00013452;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_TOP_AND_END = 0x00013453;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED_AT_BOTTOM_AND_END = 0x00013454;
const char32_t EGYPTIAN_HIEROGLYPH_MODIFIER_DAMAGED = 0x00013455;
const char32_t EGYPTIAN_HIEROGLYPH_13460 = 0x00013460;
const char32_t EGYPTIAN_HIEROGLYPH_13461 = 0x00013461;
const char32_t EGYPTIAN_HIEROGLYPH_13462 = 0x00013462;
const char32_t EGYPTIAN_HIEROGLYPH_13463 = 0x00013463;
const char32_t EGYPTIAN_HIEROGLYPH_13464 = 0x00013464;
const char32_t EGYPTIAN_HIEROGLYPH_13465 = 0x00013465;
const char32_t EGYPTIAN_HIEROGLYPH_13466 = 0x00013466;
const char32_t EGYPTIAN_HIEROGLYPH_13467 = 0x00013467;
const char32_t EGYPTIAN_HIEROGLYPH_13468 = 0x00013468;
const char32_t EGYPTIAN_HIEROGLYPH_13469 = 0x00013469;
const char32_t EGYPTIAN_HIEROGLYPH_1346A = 0x0001346A;
const char32_t EGYPTIAN_HIEROGLYPH_1346B = 0x0001346B;
const char32_t EGYPTIAN_HIEROGLYPH_1346C = 0x0001346C;
const char32_t EGYPTIAN_HIEROGLYPH_1346D = 0x0001346D;
const char32_t EGYPTIAN_HIEROGLYPH_1346E = 0x0001346E;
const char32_t EGYPTIAN_HIEROGLYPH_1346F = 0x0001346F;
const char32_t EGYPTIAN_HIEROGLYPH_13470 = 0x00013470;
const char32_t EGYPTIAN_HIEROGLYPH_13471 = 0x00013471;
const char32_t EGYPTIAN_HIEROGLYPH_13472 = 0x00013472;
const char32_t EGYPTIAN_HIEROGLYPH_13473 = 0x00013473;
const char32_t EGYPTIAN_HIEROGLYPH_13474 = 0x00013474;
const char32_t EGYPTIAN_HIEROGLYPH_13475 = 0x00013475;
const char32_t EGYPTIAN_HIEROGLYPH_13476 = 0x00013476;
const char32_t EGYPTIAN_HIEROGLYPH_13477 = 0x00013477;
const char32_t EGYPTIAN_HIEROGLYPH_13478 = 0x00013478;
const char32_t EGYPTIAN_HIEROGLYPH_13479 = 0x00013479;
const char32_t EGYPTIAN_HIEROGLYPH_1347A = 0x0001347A;
const char32_t EGYPTIAN_HIEROGLYPH_1347B = 0x0001347B;
const char32_t EGYPTIAN_HIEROGLYPH_1347C = 0x0001347C;
const char32_t EGYPTIAN_HIEROGLYPH_1347D = 0x0001347D;
const char32_t EGYPTIAN_HIEROGLYPH_1347E = 0x0001347E;
const char32_t EGYPTIAN_HIEROGLYPH_1347F = 0x0001347F;
const char32_t EGYPTIAN_HIEROGLYPH_13480 = 0x00013480;
const char32_t EGYPTIAN_HIEROGLYPH_13481 = 0x00013481;
const char32_t EGYPTIAN_HIEROGLYPH_13482 = 0x00013482;
const char32_t EGYPTIAN_HIEROGLYPH_13483 = 0x00013483;
const char32_t EGYPTIAN_HIEROGLYPH_13484 = 0x00013484;
const char32_t EGYPTIAN_HIEROGLYPH_13485 = 0x00013485;
const char32_t EGYPTIAN_HIEROGLYPH_13486 = 0x00013486;
const char32_t EGYPTIAN_HIEROGLYPH_13487 = 0x00013487;
const char32_t EGYPTIAN_HIEROGLYPH_13488 = 0x00013488;
const char32_t EGYPTIAN_HIEROGLYPH_13489 = 0x00013489;
const char32_t EGYPTIAN_HIEROGLYPH_1348A = 0x0001348A;
const char32_t EGYPTIAN_HIEROGLYPH_1348B = 0x0001348B;
const char32_t EGYPTIAN_HIEROGLYPH_1348C = 0x0001348C;
const char32_t EGYPTIAN_HIEROGLYPH_1348D = 0x0001348D;
const char32_t EGYPTIAN_HIEROGLYPH_1348E = 0x0001348E;
const char32_t EGYPTIAN_HIEROGLYPH_1348F = 0x0001348F;
const char32_t EGYPTIAN_HIEROGLYPH_13490 = 0x00013490;
const char32_t EGYPTIAN_HIEROGLYPH_13491 = 0x00013491;
const char32_t EGYPTIAN_HIEROGLYPH_13492 = 0x00013492;
const char32_t EGYPTIAN_HIEROGLYPH_13493 = 0x00013493;
const char32_t EGYPTIAN_HIEROGLYPH_13494 = 0x00013494;
const char32_t EGYPTIAN_HIEROGLYPH_13495 = 0x00013495;
const char32_t EGYPTIAN_HIEROGLYPH_13496 = 0x00013496;
const char32_t EGYPTIAN_HIEROGLYPH_13497 = 0x00013497;
const char32_t EGYPTIAN_HIEROGLYPH_13498 = 0x00013498;
const char32_t EGYPTIAN_HIEROGLYPH_13499 = 0x00013499;
const char32_t EGYPTIAN_HIEROGLYPH_1349A = 0x0001349A;
const char32_t EGYPTIAN_HIEROGLYPH_1349B = 0x0001349B;
const char32_t EGYPTIAN_HIEROGLYPH_1349C = 0x0001349C;
const char32_t EGYPTIAN_HIEROGLYPH_1349D = 0x0001349D;
const char32_t EGYPTIAN_HIEROGLYPH_1349E = 0x0001349E;
const char32_t EGYPTIAN_HIEROGLYPH_1349F = 0x0001349F;
const char32_t EGYPTIAN_HIEROGLYPH_134A0 = 0x000134A0;
const char32_t EGYPTIAN_HIEROGLYPH_134A1 = 0x000134A1;
const char32_t EGYPTIAN_HIEROGLYPH_134A2 = 0x000134A2;
const char32_t EGYPTIAN_HIEROGLYPH_134A3 = 0x000134A3;
const char32_t EGYPTIAN_HIEROGLYPH_134A4 = 0x000134A4;
const char32_t EGYPTIAN_HIEROGLYPH_134A5 = 0x000134A5;
const char32_t EGYPTIAN_HIEROGLYPH_134A6 = 0x000134A6;
const char32_t EGYPTIAN_HIEROGLYPH_134A7 = 0x000134A7;
const char32_t EGYPTIAN_HIEROGLYPH_134A8 = 0x000134A8;
const char32_t EGYPTIAN_HIEROGLYPH_134A9 = 0x000134A9;
const char32_t EGYPTIAN_HIEROGLYPH_134AA = 0x000134AA;
const char32_t EGYPTIAN_HIEROGLYPH_134AB = 0x000134AB;
const char32_t EGYPTIAN_HIEROGLYPH_134AC = 0x000134AC;
const char32_t EGYPTIAN_HIEROGLYPH_134AD = 0x000134AD;
const char32_t EGYPTIAN_HIEROGLYPH_134AE = 0x000134AE;
const char32_t EGYPTIAN_HIEROGLYPH_134AF = 0x000134AF;
const char32_t EGYPTIAN_HIEROGLYPH_134B0 = 0x000134B0;
const char32_t EGYPTIAN_HIEROGLYPH_134B1 = 0x000134B1;
const char32_t EGYPTIAN_HIEROGLYPH_134B2 = 0x000134B2;
const char32_t EGYPTIAN_HIEROGLYPH_134B3 = 0x000134B3;
const char32_t EGYPTIAN_HIEROGLYPH_134B4 = 0x000134B4;
const char32_t EGYPTIAN_HIEROGLYPH_134B5 = 0x000134B5;
const char32_t EGYPTIAN_HIEROGLYPH_134B6 = 0x000134B6;
const char32_t EGYPTIAN_HIEROGLYPH_134B7 = 0x000134B7;
const char32_t EGYPTIAN_HIEROGLYPH_134B8 = 0x000134B8;
const char32_t EGYPTIAN_HIEROGLYPH_134B9 = 0x000134B9;
const char32_t EGYPTIAN_HIEROGLYPH_134BA = 0x000134BA;
const char32_t EGYPTIAN_HIEROGLYPH_134BB = 0x000134BB;
const char32_t EGYPTIAN_HIEROGLYPH_134BC = 0x000134BC;
const char32_t EGYPTIAN_HIEROGLYPH_134BD = 0x000134BD;
const char32_t EGYPTIAN_HIEROGLYPH_134BE = 0x000134BE;
const char32_t EGYPTIAN_HIEROGLYPH_134BF = 0x000134BF;
const char32_t EGYPTIAN_HIEROGLYPH_134C0 = 0x000134C0;
const char32_t EGYPTIAN_HIEROGLYPH_134C1 = 0x000134C1;
const char32_t EGYPTIAN_HIEROGLYPH_134C2 = 0x000134C2;
const char32_t EGYPTIAN_HIEROGLYPH_134C3 = 0x000134C3;
const char32_t EGYPTIAN_HIEROGLYPH_134C4 = 0x000134C4;
const char32_t EGYPTIAN_HIEROGLYPH_134C5 = 0x000134C5;
const char32_t EGYPTIAN_HIEROGLYPH_134C6 = 0x000134C6;
const char32_t EGYPTIAN_HIEROGLYPH_134C7 = 0x000134C7;
const char32_t EGYPTIAN_HIEROGLYPH_134C8 = 0x000134C8;
const char32_t EGYPTIAN_HIEROGLYPH_134C9 = 0x000134C9;
const char32_t EGYPTIAN_HIEROGLYPH_134CA = 0x000134CA;
const char32_t EGYPTIAN_HIEROGLYPH_134CB = 0x000134CB;
const char32_t EGYPTIAN_HIEROGLYPH_134CC = 0x000134CC;
const char32_t EGYPTIAN_HIEROGLYPH_134CD = 0x000134CD;
const char32_t EGYPTIAN_HIEROGLYPH_134CE = 0x000134CE;
const char32_t EGYPTIAN_HIEROGLYPH_134CF = 0x000134CF;
const char32_t EGYPTIAN_HIEROGLYPH_134D0 = 0x000134D0;
const char32_t EGYPTIAN_HIEROGLYPH_134D1 = 0x000134D1;
const char32_t EGYPTIAN_HIEROGLYPH_134D2 = 0x000134D2;
const char32_t EGYPTIAN_HIEROGLYPH_134D3 = 0x000134D3;
const char32_t EGYPTIAN_HIEROGLYPH_134D4 = 0x000134D4;
const char32_t EGYPTIAN_HIEROGLYPH_134D5 = 0x000134D5;
const char32_t EGYPTIAN_HIEROGLYPH_134D6 = 0x000134D6;
const char32_t EGYPTIAN_HIEROGLYPH_134D7 = 0x000134D7;
const char32_t EGYPTIAN_HIEROGLYPH_134D8 = 0x000134D8;
const char32_t EGYPTIAN_HIEROGLYPH_134D9 = 0x000134D9;
const char32_t EGYPTIAN_HIEROGLYPH_134DA = 0x000134DA;
const char32_t EGYPTIAN_HIEROGLYPH_134DB = 0x000134DB;
const char32_t EGYPTIAN_HIEROGLYPH_134DC = 0x000134DC;
const char32_t EGYPTIAN_HIEROGLYPH_134DD = 0x000134DD;
const char32_t EGYPTIAN_HIEROGLYPH_134DE = 0x000134DE;
const char32_t EGYPTIAN_HIEROGLYPH_134DF = 0x000134DF;
const char32_t EGYPTIAN_HIEROGLYPH_134E0 = 0x000134E0;
const char32_t EGYPTIAN_HIEROGLYPH_134E1 = 0x000134E1;
const char32_t EGYPTIAN_HIEROGLYPH_134E2 = 0x000134E2;
const char32_t EGYPTIAN_HIEROGLYPH_134E3 = 0x000134E3;
const char32_t EGYPTIAN_HIEROGLYPH_134E4 = 0x000134E4;
const char32_t EGYPTIAN_HIEROGLYPH_134E5 = 0x000134E5;
const char32_t EGYPTIAN_HIEROGLYPH_134E6 = 0x000134E6;
const char32_t EGYPTIAN_HIEROGLYPH_134E7 = 0x000134E7;
const char32_t EGYPTIAN_HIEROGLYPH_134E8 = 0x000134E8;
const char32_t EGYPTIAN_HIEROGLYPH_134E9 = 0x000134E9;
const char32_t EGYPTIAN_HIEROGLYPH_134EA = 0x000134EA;
const char32_t EGYPTIAN_HIEROGLYPH_134EB = 0x000134EB;
const char32_t EGYPTIAN_HIEROGLYPH_134EC = 0x000134EC;
const char32_t EGYPTIAN_HIEROGLYPH_134ED = 0x000134ED;
const char32_t EGYPTIAN_HIEROGLYPH_134EE = 0x000134EE;
const char32_t EGYPTIAN_HIEROGLYPH_134EF = 0x000134EF;
const char32_t EGYPTIAN_HIEROGLYPH_134F0 = 0x000134F0;
const char32_t EGYPTIAN_HIEROGLYPH_134F1 = 0x000134F1;
const char32_t EGYPTIAN_HIEROGLYPH_134F2 = 0x000134F2;
const char32_t EGYPTIAN_HIEROGLYPH_134F3 = 0x000134F3;
const char32_t EGYPTIAN_HIEROGLYPH_134F4 = 0x000134F4;
const char32_t EGYPTIAN_HIEROGLYPH_134F5 = 0x000134F5;
const char32_t EGYPTIAN_HIEROGLYPH_134F6 = 0x000134F6;
const char32_t EGYPTIAN_HIEROGLYPH_134F7 = 0x000134F7;
const char32_t EGYPTIAN_HIEROGLYPH_134F8 = 0x000134F8;
const char32_t EGYPTIAN_HIEROGLYPH_134F9 = 0x000134F9;
const char32_t EGYPTIAN_HIEROGLYPH_134FA = 0x000134FA;
const char32_t EGYPTIAN_HIEROGLYPH_134FB = 0x000134FB;
const char32_t EGYPTIAN_HIEROGLYPH_134FC = 0x000134FC;
const char32_t EGYPTIAN_HIEROGLYPH_134FD = 0x000134FD;
const char32_t EGYPTIAN_HIEROGLYPH_134FE = 0x000134FE;
const char32_t EGYPTIAN_HIEROGLYPH_134FF = 0x000134FF;
const char32_t EGYPTIAN_HIEROGLYPH_13500 = 0x00013500;
const char32_t EGYPTIAN_HIEROGLYPH_13501 = 0x00013501;
const char32_t EGYPTIAN_HIEROGLYPH_13502 = 0x00013502;
const char32_t EGYPTIAN_HIEROGLYPH_13503 = 0x00013503;
const char32_t EGYPTIAN_HIEROGLYPH_13504 = 0x00013504;
const char32_t EGYPTIAN_HIEROGLYPH_13505 = 0x00013505;
const char32_t EGYPTIAN_HIEROGLYPH_13506 = 0x00013506;
const char32_t EGYPTIAN_HIEROGLYPH_13507 = 0x00013507;
const char32_t EGYPTIAN_HIEROGLYPH_13508 = 0x00013508;
const char32_t EGYPTIAN_HIEROGLYPH_13509 = 0x00013509;
const char32_t EGYPTIAN_HIEROGLYPH_1350A = 0x0001350A;
const char32_t EGYPTIAN_HIEROGLYPH_1350B = 0x0001350B;
const char32_t EGYPTIAN_HIEROGLYPH_1350C = 0x0001350C;
const char32_t EGYPTIAN_HIEROGLYPH_1350D = 0x0001350D;
const char32_t EGYPTIAN_HIEROGLYPH_1350E = 0x0001350E;
const char32_t EGYPTIAN_HIEROGLYPH_1350F = 0x0001350F;
const char32_t EGYPTIAN_HIEROGLYPH_13510 = 0x00013510;
const char32_t EGYPTIAN_HIEROGLYPH_13511 = 0x00013511;
const char32_t EGYPTIAN_HIEROGLYPH_13512 = 0x00013512;
const char32_t EGYPTIAN_HIEROGLYPH_13513 = 0x00013513;
const char32_t EGYPTIAN_HIEROGLYPH_13514 = 0x00013514;
const char32_t EGYPTIAN_HIEROGLYPH_13515 = 0x00013515;
const char32_t EGYPTIAN_HIEROGLYPH_13516 = 0x00013516;
const char32_t EGYPTIAN_HIEROGLYPH_13517 = 0x00013517;
const char32_t EGYPTIAN_HIEROGLYPH_13518 = 0x00013518;
const char32_t EGYPTIAN_HIEROGLYPH_13519 = 0x00013519;
const char32_t EGYPTIAN_HIEROGLYPH_1351A = 0x0001351A;
const char32_t EGYPTIAN_HIEROGLYPH_1351B = 0x0001351B;
const char32_t EGYPTIAN_HIEROGLYPH_1351C = 0x0001351C;
const char32_t EGYPTIAN_HIEROGLYPH_1351D = 0x0001351D;
const char32_t EGYPTIAN_HIEROGLYPH_1351E = 0x0001351E;
const char32_t EGYPTIAN_HIEROGLYPH_1351F = 0x0001351F;
const char32_t EGYPTIAN_HIEROGLYPH_13520 = 0x00013520;
const char32_t EGYPTIAN_HIEROGLYPH_13521 = 0x00013521;
const char32_t EGYPTIAN_HIEROGLYPH_13522 = 0x00013522;
const char32_t EGYPTIAN_HIEROGLYPH_13523 = 0x00013523;
const char32_t EGYPTIAN_HIEROGLYPH_13524 = 0x00013524;
const char32_t EGYPTIAN_HIEROGLYPH_13525 = 0x00013525;
const char32_t EGYPTIAN_HIEROGLYPH_13526 = 0x00013526;
const char32_t EGYPTIAN_HIEROGLYPH_13527 = 0x00013527;
const char32_t EGYPTIAN_HIEROGLYPH_13528 = 0x00013528;
const char32_t EGYPTIAN_HIEROGLYPH_13529 = 0x00013529;
const char32_t EGYPTIAN_HIEROGLYPH_1352A = 0x0001352A;
const char32_t EGYPTIAN_HIEROGLYPH_1352B = 0x0001352B;
const char32_t EGYPTIAN_HIEROGLYPH_1352C = 0x0001352C;
const char32_t EGYPTIAN_HIEROGLYPH_1352D = 0x0001352D;
const char32_t EGYPTIAN_HIEROGLYPH_1352E = 0x0001352E;
const char32_t EGYPTIAN_HIEROGLYPH_1352F = 0x0001352F;
const char32_t EGYPTIAN_HIEROGLYPH_13530 = 0x00013530;
const char32_t EGYPTIAN_HIEROGLYPH_13531 = 0x00013531;
const char32_t EGYPTIAN_HIEROGLYPH_13532 = 0x00013532;
const char32_t EGYPTIAN_HIEROGLYPH_13533 = 0x00013533;
const char32_t EGYPTIAN_HIEROGLYPH_13534 = 0x00013534;
const char32_t EGYPTIAN_HIEROGLYPH_13535 = 0x00013535;
const char32_t EGYPTIAN_HIEROGLYPH_13536 = 0x00013536;
const char32_t EGYPTIAN_HIEROGLYPH_13537 = 0x00013537;
const char32_t EGYPTIAN_HIEROGLYPH_13538 = 0x00013538;
const char32_t EGYPTIAN_HIEROGLYPH_13539 = 0x00013539;
const char32_t EGYPTIAN_HIEROGLYPH_1353A = 0x0001353A;
const char32_t EGYPTIAN_HIEROGLYPH_1353B = 0x0001353B;
const char32_t EGYPTIAN_HIEROGLYPH_1353C = 0x0001353C;
const char32_t EGYPTIAN_HIEROGLYPH_1353D = 0x0001353D;
const char32_t EGYPTIAN_HIEROGLYPH_1353E = 0x0001353E;
const char32_t EGYPTIAN_HIEROGLYPH_1353F = 0x0001353F;
const char32_t EGYPTIAN_HIEROGLYPH_13540 = 0x00013540;
const char32_t EGYPTIAN_HIEROGLYPH_13541 = 0x00013541;
const char32_t EGYPTIAN_HIEROGLYPH_13542 = 0x00013542;
const char32_t EGYPTIAN_HIEROGLYPH_13543 = 0x00013543;
const char32_t EGYPTIAN_HIEROGLYPH_13544 = 0x00013544;
const char32_t EGYPTIAN_HIEROGLYPH_13545 = 0x00013545;
const char32_t EGYPTIAN_HIEROGLYPH_13546 = 0x00013546;
const char32_t EGYPTIAN_HIEROGLYPH_13547 = 0x00013547;
const char32_t EGYPTIAN_HIEROGLYPH_13548 = 0x00013548;
const char32_t EGYPTIAN_HIEROGLYPH_13549 = 0x00013549;
const char32_t EGYPTIAN_HIEROGLYPH_1354A = 0x0001354A;
const char32_t EGYPTIAN_HIEROGLYPH_1354B = 0x0001354B;
const char32_t EGYPTIAN_HIEROGLYPH_1354C = 0x0001354C;
const char32_t EGYPTIAN_HIEROGLYPH_1354D = 0x0001354D;
const char32_t EGYPTIAN_HIEROGLYPH_1354E = 0x0001354E;
const char32_t EGYPTIAN_HIEROGLYPH_1354F = 0x0001354F;
const char32_t EGYPTIAN_HIEROGLYPH_13550 = 0x00013550;
const char32_t EGYPTIAN_HIEROGLYPH_13551 = 0x00013551;
const char32_t EGYPTIAN_HIEROGLYPH_13552 = 0x00013552;
const char32_t EGYPTIAN_HIEROGLYPH_13553 = 0x00013553;
const char32_t EGYPTIAN_HIEROGLYPH_13554 = 0x00013554;
const char32_t EGYPTIAN_HIEROGLYPH_13555 = 0x00013555;
const char32_t EGYPTIAN_HIEROGLYPH_13556 = 0x00013556;
const char32_t EGYPTIAN_HIEROGLYPH_13557 = 0x00013557;
const char32_t EGYPTIAN_HIEROGLYPH_13558 = 0x00013558;
const char32_t EGYPTIAN_HIEROGLYPH_13559 = 0x00013559;
const char32_t EGYPTIAN_HIEROGLYPH_1355A = 0x0001355A;
const char32_t EGYPTIAN_HIEROGLYPH_1355B = 0x0001355B;
const char32_t EGYPTIAN_HIEROGLYPH_1355C = 0x0001355C;
const char32_t EGYPTIAN_HIEROGLYPH_1355D = 0x0001355D;
const char32_t EGYPTIAN_HIEROGLYPH_1355E = 0x0001355E;
const char32_t EGYPTIAN_HIEROGLYPH_1355F = 0x0001355F;
const char32_t EGYPTIAN_HIEROGLYPH_13560 = 0x00013560;
const char32_t EGYPTIAN_HIEROGLYPH_13561 = 0x00013561;
const char32_t EGYPTIAN_HIEROGLYPH_13562 = 0x00013562;
const char32_t EGYPTIAN_HIEROGLYPH_13563 = 0x00013563;
const char32_t EGYPTIAN_HIEROGLYPH_13564 = 0x00013564;
const char32_t EGYPTIAN_HIEROGLYPH_13565 = 0x00013565;
const char32_t EGYPTIAN_HIEROGLYPH_13566 = 0x00013566;
const char32_t EGYPTIAN_HIEROGLYPH_13567 = 0x00013567;
const char32_t EGYPTIAN_HIEROGLYPH_13568 = 0x00013568;
const char32_t EGYPTIAN_HIEROGLYPH_13569 = 0x00013569;
const char32_t EGYPTIAN_HIEROGLYPH_1356A = 0x0001356A;
const char32_t EGYPTIAN_HIEROGLYPH_1356B = 0x0001356B;
const char32_t EGYPTIAN_HIEROGLYPH_1356C = 0x0001356C;
const char32_t EGYPTIAN_HIEROGLYPH_1356D = 0x0001356D;
const char32_t EGYPTIAN_HIEROGLYPH_1356E = 0x0001356E;
const char32_t EGYPTIAN_HIEROGLYPH_1356F = 0x0001356F;
const char32_t EGYPTIAN_HIEROGLYPH_13570 = 0x00013570;
const char32_t EGYPTIAN_HIEROGLYPH_13571 = 0x00013571;
const char32_t EGYPTIAN_HIEROGLYPH_13572 = 0x00013572;
const char32_t EGYPTIAN_HIEROGLYPH_13573 = 0x00013573;
const char32_t EGYPTIAN_HIEROGLYPH_13574 = 0x00013574;
const char32_t EGYPTIAN_HIEROGLYPH_13575 = 0x00013575;
const char32_t EGYPTIAN_HIEROGLYPH_13576 = 0x00013576;
const char32_t EGYPTIAN_HIEROGLYPH_13577 = 0x00013577;
const char32_t EGYPTIAN_HIEROGLYPH_13578 = 0x00013578;
const char32_t EGYPTIAN_HIEROGLYPH_13579 = 0x00013579;
const char32_t EGYPTIAN_HIEROGLYPH_1357A = 0x0001357A;
const char32_t EGYPTIAN_HIEROGLYPH_1357B = 0x0001357B;
const char32_t EGYPTIAN_HIEROGLYPH_1357C = 0x0001357C;
const char32_t EGYPTIAN_HIEROGLYPH_1357D = 0x0001357D;
const char32_t EGYPTIAN_HIEROGLYPH_1357E = 0x0001357E;
const char32_t EGYPTIAN_HIEROGLYPH_1357F = 0x0001357F;
const char32_t EGYPTIAN_HIEROGLYPH_13580 = 0x00013580;
const char32_t EGYPTIAN_HIEROGLYPH_13581 = 0x00013581;
const char32_t EGYPTIAN_HIEROGLYPH_13582 = 0x00013582;
const char32_t EGYPTIAN_HIEROGLYPH_13583 = 0x00013583;
const char32_t EGYPTIAN_HIEROGLYPH_13584 = 0x00013584;
const char32_t EGYPTIAN_HIEROGLYPH_13585 = 0x00013585;
const char32_t EGYPTIAN_HIEROGLYPH_13586 = 0x00013586;
const char32_t EGYPTIAN_HIEROGLYPH_13587 = 0x00013587;
const char32_t EGYPTIAN_HIEROGLYPH_13588 = 0x00013588;
const char32_t EGYPTIAN_HIEROGLYPH_13589 = 0x00013589;
const char32_t EGYPTIAN_HIEROGLYPH_1358A = 0x0001358A;
const char32_t EGYPTIAN_HIEROGLYPH_1358B = 0x0001358B;
const char32_t EGYPTIAN_HIEROGLYPH_1358C = 0x0001358C;
const char32_t EGYPTIAN_HIEROGLYPH_1358D = 0x0001358D;
const char32_t EGYPTIAN_HIEROGLYPH_1358E = 0x0001358E;
const char32_t EGYPTIAN_HIEROGLYPH_1358F = 0x0001358F;
const char32_t EGYPTIAN_HIEROGLYPH_13590 = 0x00013590;
const char32_t EGYPTIAN_HIEROGLYPH_13591 = 0x00013591;
const char32_t EGYPTIAN_HIEROGLYPH_13592 = 0x00013592;
const char32_t EGYPTIAN_HIEROGLYPH_13593 = 0x00013593;
const char32_t EGYPTIAN_HIEROGLYPH_13594 = 0x00013594;
const char32_t EGYPTIAN_HIEROGLYPH_13595 = 0x00013595;
const char32_t EGYPTIAN_HIEROGLYPH_13596 = 0x00013596;
const char32_t EGYPTIAN_HIEROGLYPH_13597 = 0x00013597;
const char32_t EGYPTIAN_HIEROGLYPH_13598 = 0x00013598;
const char32_t EGYPTIAN_HIEROGLYPH_13599 = 0x00013599;
const char32_t EGYPTIAN_HIEROGLYPH_1359A = 0x0001359A;
const char32_t EGYPTIAN_HIEROGLYPH_1359B = 0x0001359B;
const char32_t EGYPTIAN_HIEROGLYPH_1359C = 0x0001359C;
const char32_t EGYPTIAN_HIEROGLYPH_1359D = 0x0001359D;
const char32_t EGYPTIAN_HIEROGLYPH_1359E = 0x0001359E;
const char32_t EGYPTIAN_HIEROGLYPH_1359F = 0x0001359F;
const char32_t EGYPTIAN_HIEROGLYPH_135A0 = 0x000135A0;
const char32_t EGYPTIAN_HIEROGLYPH_135A1 = 0x000135A1;
const char32_t EGYPTIAN_HIEROGLYPH_135A2 = 0x000135A2;
const char32_t EGYPTIAN_HIEROGLYPH_135A3 = 0x000135A3;
const char32_t EGYPTIAN_HIEROGLYPH_135A4 = 0x000135A4;
const char32_t EGYPTIAN_HIEROGLYPH_135A5 = 0x000135A5;
const char32_t EGYPTIAN_HIEROGLYPH_135A6 = 0x000135A6;
const char32_t EGYPTIAN_HIEROGLYPH_135A7 = 0x000135A7;
const char32_t EGYPTIAN_HIEROGLYPH_135A8 = 0x000135A8;
const char32_t EGYPTIAN_HIEROGLYPH_135A9 = 0x000135A9;
const char32_t EGYPTIAN_HIEROGLYPH_135AA = 0x000135AA;
const char32_t EGYPTIAN_HIEROGLYPH_135AB = 0x000135AB;
const char32_t EGYPTIAN_HIEROGLYPH_135AC = 0x000135AC;
const char32_t EGYPTIAN_HIEROGLYPH_135AD = 0x000135AD;
const char32_t EGYPTIAN_HIEROGLYPH_135AE = 0x000135AE;
const char32_t EGYPTIAN_HIEROGLYPH_135AF = 0x000135AF;
const char32_t EGYPTIAN_HIEROGLYPH_135B0 = 0x000135B0;
const char32_t EGYPTIAN_HIEROGLYPH_135B1 = 0x000135B1;
const char32_t EGYPTIAN_HIEROGLYPH_135B2 = 0x000135B2;
const char32_t EGYPTIAN_HIEROGLYPH_135B3 = 0x000135B3;
const char32_t EGYPTIAN_HIEROGLYPH_135B4 = 0x000135B4;
const char32_t EGYPTIAN_HIEROGLYPH_135B5 = 0x000135B5;
const char32_t EGYPTIAN_HIEROGLYPH_135B6 = 0x000135B6;
const char32_t EGYPTIAN_HIEROGLYPH_135B7 = 0x000135B7;
const char32_t EGYPTIAN_HIEROGLYPH_135B8 = 0x000135B8;
const char32_t EGYPTIAN_HIEROGLYPH_135B9 = 0x000135B9;
const char32_t EGYPTIAN_HIEROGLYPH_135BA = 0x000135BA;
const char32_t EGYPTIAN_HIEROGLYPH_135BB = 0x000135BB;
const char32_t EGYPTIAN_HIEROGLYPH_135BC = 0x000135BC;
const char32_t EGYPTIAN_HIEROGLYPH_135BD = 0x000135BD;
const char32_t EGYPTIAN_HIEROGLYPH_135BE = 0x000135BE;
const char32_t EGYPTIAN_HIEROGLYPH_135BF = 0x000135BF;
const char32_t EGYPTIAN_HIEROGLYPH_135C0 = 0x000135C0;
const char32_t EGYPTIAN_HIEROGLYPH_135C1 = 0x000135C1;
const char32_t EGYPTIAN_HIEROGLYPH_135C2 = 0x000135C2;
const char32_t EGYPTIAN_HIEROGLYPH_135C3 = 0x000135C3;
const char32_t EGYPTIAN_HIEROGLYPH_135C4 = 0x000135C4;
const char32_t EGYPTIAN_HIEROGLYPH_135C5 = 0x000135C5;
const char32_t EGYPTIAN_HIEROGLYPH_135C6 = 0x000135C6;
const char32_t EGYPTIAN_HIEROGLYPH_135C7 = 0x000135C7;
const char32_t EGYPTIAN_HIEROGLYPH_135C8 = 0x000135C8;
const char32_t EGYPTIAN_HIEROGLYPH_135C9 = 0x000135C9;
const char32_t EGYPTIAN_HIEROGLYPH_135CA = 0x000135CA;
const char32_t EGYPTIAN_HIEROGLYPH_135CB = 0x000135CB;
const char32_t EGYPTIAN_HIEROGLYPH_135CC = 0x000135CC;
const char32_t EGYPTIAN_HIEROGLYPH_135CD = 0x000135CD;
const char32_t EGYPTIAN_HIEROGLYPH_135CE = 0x000135CE;
const char32_t EGYPTIAN_HIEROGLYPH_135CF = 0x000135CF;
const char32_t EGYPTIAN_HIEROGLYPH_135D0 = 0x000135D0;
const char32_t EGYPTIAN_HIEROGLYPH_135D1 = 0x000135D1;
const char32_t EGYPTIAN_HIEROGLYPH_135D2 = 0x000135D2;
const char32_t EGYPTIAN_HIEROGLYPH_135D3 = 0x000135D3;
const char32_t EGYPTIAN_HIEROGLYPH_135D4 = 0x000135D4;
const char32_t EGYPTIAN_HIEROGLYPH_135D5 = 0x000135D5;
const char32_t EGYPTIAN_HIEROGLYPH_135D6 = 0x000135D6;
const char32_t EGYPTIAN_HIEROGLYPH_135D7 = 0x000135D7;
const char32_t EGYPTIAN_HIEROGLYPH_135D8 = 0x000135D8;
const char32_t EGYPTIAN_HIEROGLYPH_135D9 = 0x000135D9;
const char32_t EGYPTIAN_HIEROGLYPH_135DA = 0x000135DA;
const char32_t EGYPTIAN_HIEROGLYPH_135DB = 0x000135DB;
const char32_t EGYPTIAN_HIEROGLYPH_135DC = 0x000135DC;
const char32_t EGYPTIAN_HIEROGLYPH_135DD = 0x000135DD;
const char32_t EGYPTIAN_HIEROGLYPH_135DE = 0x000135DE;
const char32_t EGYPTIAN_HIEROGLYPH_135DF = 0x000135DF;
const char32_t EGYPTIAN_HIEROGLYPH_135E0 = 0x000135E0;
const char32_t EGYPTIAN_HIEROGLYPH_135E1 = 0x000135E1;
const char32_t EGYPTIAN_HIEROGLYPH_135E2 = 0x000135E2;
const char32_t EGYPTIAN_HIEROGLYPH_135E3 = 0x000135E3;
const char32_t EGYPTIAN_HIEROGLYPH_135E4 = 0x000135E4;
const char32_t EGYPTIAN_HIEROGLYPH_135E5 = 0x000135E5;
const char32_t EGYPTIAN_HIEROGLYPH_135E6 = 0x000135E6;
const char32_t EGYPTIAN_HIEROGLYPH_135E7 = 0x000135E7;
const char32_t EGYPTIAN_HIEROGLYPH_135E8 = 0x000135E8;
const char32_t EGYPTIAN_HIEROGLYPH_135E9 = 0x000135E9;
const char32_t EGYPTIAN_HIEROGLYPH_135EA = 0x000135EA;
const char32_t EGYPTIAN_HIEROGLYPH_135EB = 0x000135EB;
const char32_t EGYPTIAN_HIEROGLYPH_135EC = 0x000135EC;
const char32_t EGYPTIAN_HIEROGLYPH_135ED = 0x000135ED;
const char32_t EGYPTIAN_HIEROGLYPH_135EE = 0x000135EE;
const char32_t EGYPTIAN_HIEROGLYPH_135EF = 0x000135EF;
const char32_t EGYPTIAN_HIEROGLYPH_135F0 = 0x000135F0;
const char32_t EGYPTIAN_HIEROGLYPH_135F1 = 0x000135F1;
const char32_t EGYPTIAN_HIEROGLYPH_135F2 = 0x000135F2;
const char32_t EGYPTIAN_HIEROGLYPH_135F3 = 0x000135F3;
const char32_t EGYPTIAN_HIEROGLYPH_135F4 = 0x000135F4;
const char32_t EGYPTIAN_HIEROGLYPH_135F5 = 0x000135F5;
const char32_t EGYPTIAN_HIEROGLYPH_135F6 = 0x000135F6;
const char32_t EGYPTIAN_HIEROGLYPH_135F7 = 0x000135F7;
const char32_t EGYPTIAN_HIEROGLYPH_135F8 = 0x000135F8;
const char32_t EGYPTIAN_HIEROGLYPH_135F9 = 0x000135F9;
const char32_t EGYPTIAN_HIEROGLYPH_135FA = 0x000135FA;
const char32_t EGYPTIAN_HIEROGLYPH_135FB = 0x000135FB;
const char32_t EGYPTIAN_HIEROGLYPH_135FC = 0x000135FC;
const char32_t EGYPTIAN_HIEROGLYPH_135FD = 0x000135FD;
const char32_t EGYPTIAN_HIEROGLYPH_135FE = 0x000135FE;
const char32_t EGYPTIAN_HIEROGLYPH_135FF = 0x000135FF;
const char32_t EGYPTIAN_HIEROGLYPH_13600 = 0x00013600;
const char32_t EGYPTIAN_HIEROGLYPH_13601 = 0x00013601;
const char32_t EGYPTIAN_HIEROGLYPH_13602 = 0x00013602;
const char32_t EGYPTIAN_HIEROGLYPH_13603 = 0x00013603;
const char32_t EGYPTIAN_HIEROGLYPH_13604 = 0x00013604;
const char32_t EGYPTIAN_HIEROGLYPH_13605 = 0x00013605;
const char32_t EGYPTIAN_HIEROGLYPH_13606 = 0x00013606;
const char32_t EGYPTIAN_HIEROGLYPH_13607 = 0x00013607;
const char32_t EGYPTIAN_HIEROGLYPH_13608 = 0x00013608;
const char32_t EGYPTIAN_HIEROGLYPH_13609 = 0x00013609;
const char32_t EGYPTIAN_HIEROGLYPH_1360A = 0x0001360A;
const char32_t EGYPTIAN_HIEROGLYPH_1360B = 0x0001360B;
const char32_t EGYPTIAN_HIEROGLYPH_1360C = 0x0001360C;
const char32_t EGYPTIAN_HIEROGLYPH_1360D = 0x0001360D;
const char32_t EGYPTIAN_HIEROGLYPH_1360E = 0x0001360E;
const char32_t EGYPTIAN_HIEROGLYPH_1360F = 0x0001360F;
const char32_t EGYPTIAN_HIEROGLYPH_13610 = 0x00013610;
const char32_t EGYPTIAN_HIEROGLYPH_13611 = 0x00013611;
const char32_t EGYPTIAN_HIEROGLYPH_13612 = 0x00013612;
const char32_t EGYPTIAN_HIEROGLYPH_13613 = 0x00013613;
const char32_t EGYPTIAN_HIEROGLYPH_13614 = 0x00013614;
const char32_t EGYPTIAN_HIEROGLYPH_13615 = 0x00013615;
const char32_t EGYPTIAN_HIEROGLYPH_13616 = 0x00013616;
const char32_t EGYPTIAN_HIEROGLYPH_13617 = 0x00013617;
const char32_t EGYPTIAN_HIEROGLYPH_13618 = 0x00013618;
const char32_t EGYPTIAN_HIEROGLYPH_13619 = 0x00013619;
const char32_t EGYPTIAN_HIEROGLYPH_1361A = 0x0001361A;
const char32_t EGYPTIAN_HIEROGLYPH_1361B = 0x0001361B;
const char32_t EGYPTIAN_HIEROGLYPH_1361C = 0x0001361C;
const char32_t EGYPTIAN_HIEROGLYPH_1361D = 0x0001361D;
const char32_t EGYPTIAN_HIEROGLYPH_1361E = 0x0001361E;
const char32_t EGYPTIAN_HIEROGLYPH_1361F = 0x0001361F;
const char32_t EGYPTIAN_HIEROGLYPH_13620 = 0x00013620;
const char32_t EGYPTIAN_HIEROGLYPH_13621 = 0x00013621;
const char32_t EGYPTIAN_HIEROGLYPH_13622 = 0x00013622;
const char32_t EGYPTIAN_HIEROGLYPH_13623 = 0x00013623;
const char32_t EGYPTIAN_HIEROGLYPH_13624 = 0x00013624;
const char32_t EGYPTIAN_HIEROGLYPH_13625 = 0x00013625;
const char32_t EGYPTIAN_HIEROGLYPH_13626 = 0x00013626;
const char32_t EGYPTIAN_HIEROGLYPH_13627 = 0x00013627;
const char32_t EGYPTIAN_HIEROGLYPH_13628 = 0x00013628;
const char32_t EGYPTIAN_HIEROGLYPH_13629 = 0x00013629;
const char32_t EGYPTIAN_HIEROGLYPH_1362A = 0x0001362A;
const char32_t EGYPTIAN_HIEROGLYPH_1362B = 0x0001362B;
const char32_t EGYPTIAN_HIEROGLYPH_1362C = 0x0001362C;
const char32_t EGYPTIAN_HIEROGLYPH_1362D = 0x0001362D;
const char32_t EGYPTIAN_HIEROGLYPH_1362E = 0x0001362E;
const char32_t EGYPTIAN_HIEROGLYPH_1362F = 0x0001362F;
const char32_t EGYPTIAN_HIEROGLYPH_13630 = 0x00013630;
const char32_t EGYPTIAN_HIEROGLYPH_13631 = 0x00013631;
const char32_t EGYPTIAN_HIEROGLYPH_13632 = 0x00013632;
const char32_t EGYPTIAN_HIEROGLYPH_13633 = 0x00013633;
const char32_t EGYPTIAN_HIEROGLYPH_13634 = 0x00013634;
const char32_t EGYPTIAN_HIEROGLYPH_13635 = 0x00013635;
const char32_t EGYPTIAN_HIEROGLYPH_13636 = 0x00013636;
const char32_t EGYPTIAN_HIEROGLYPH_13637 = 0x00013637;
const char32_t EGYPTIAN_HIEROGLYPH_13638 = 0x00013638;
const char32_t EGYPTIAN_HIEROGLYPH_13639 = 0x00013639;
const char32_t EGYPTIAN_HIEROGLYPH_1363A = 0x0001363A;
const char32_t EGYPTIAN_HIEROGLYPH_1363B = 0x0001363B;
const char32_t EGYPTIAN_HIEROGLYPH_1363C = 0x0001363C;
const char32_t EGYPTIAN_HIEROGLYPH_1363D = 0x0001363D;
const char32_t EGYPTIAN_HIEROGLYPH_1363E = 0x0001363E;
const char32_t EGYPTIAN_HIEROGLYPH_1363F = 0x0001363F;
const char32_t EGYPTIAN_HIEROGLYPH_13640 = 0x00013640;
const char32_t EGYPTIAN_HIEROGLYPH_13641 = 0x00013641;
const char32_t EGYPTIAN_HIEROGLYPH_13642 = 0x00013642;
const char32_t EGYPTIAN_HIEROGLYPH_13643 = 0x00013643;
const char32_t EGYPTIAN_HIEROGLYPH_13644 = 0x00013644;
const char32_t EGYPTIAN_HIEROGLYPH_13645 = 0x00013645;
const char32_t EGYPTIAN_HIEROGLYPH_13646 = 0x00013646;
const char32_t EGYPTIAN_HIEROGLYPH_13647 = 0x00013647;
const char32_t EGYPTIAN_HIEROGLYPH_13648 = 0x00013648;
const char32_t EGYPTIAN_HIEROGLYPH_13649 = 0x00013649;
const char32_t EGYPTIAN_HIEROGLYPH_1364A = 0x0001364A;
const char32_t EGYPTIAN_HIEROGLYPH_1364B = 0x0001364B;
const char32_t EGYPTIAN_HIEROGLYPH_1364C = 0x0001364C;
const char32_t EGYPTIAN_HIEROGLYPH_1364D = 0x0001364D;
const char32_t EGYPTIAN_HIEROGLYPH_1364E = 0x0001364E;
const char32_t EGYPTIAN_HIEROGLYPH_1364F = 0x0001364F;
const char32_t EGYPTIAN_HIEROGLYPH_13650 = 0x00013650;
const char32_t EGYPTIAN_HIEROGLYPH_13651 = 0x00013651;
const char32_t EGYPTIAN_HIEROGLYPH_13652 = 0x00013652;
const char32_t EGYPTIAN_HIEROGLYPH_13653 = 0x00013653;
const char32_t EGYPTIAN_HIEROGLYPH_13654 = 0x00013654;
const char32_t EGYPTIAN_HIEROGLYPH_13655 = 0x00013655;
const char32_t EGYPTIAN_HIEROGLYPH_13656 = 0x00013656;
const char32_t EGYPTIAN_HIEROGLYPH_13657 = 0x00013657;
const char32_t EGYPTIAN_HIEROGLYPH_13658 = 0x00013658;
const char32_t EGYPTIAN_HIEROGLYPH_13659 = 0x00013659;
const char32_t EGYPTIAN_HIEROGLYPH_1365A = 0x0001365A;
const char32_t EGYPTIAN_HIEROGLYPH_1365B = 0x0001365B;
const char32_t EGYPTIAN_HIEROGLYPH_1365C = 0x0001365C;
const char32_t EGYPTIAN_HIEROGLYPH_1365D = 0x0001365D;
const char32_t EGYPTIAN_HIEROGLYPH_1365E = 0x0001365E;
const char32_t EGYPTIAN_HIEROGLYPH_1365F = 0x0001365F;
const char32_t EGYPTIAN_HIEROGLYPH_13660 = 0x00013660;
const char32_t EGYPTIAN_HIEROGLYPH_13661 = 0x00013661;
const char32_t EGYPTIAN_HIEROGLYPH_13662 = 0x00013662;
const char32_t EGYPTIAN_HIEROGLYPH_13663 = 0x00013663;
const char32_t EGYPTIAN_HIEROGLYPH_13664 = 0x00013664;
const char32_t EGYPTIAN_HIEROGLYPH_13665 = 0x00013665;
const char32_t EGYPTIAN_HIEROGLYPH_13666 = 0x00013666;
const char32_t EGYPTIAN_HIEROGLYPH_13667 = 0x00013667;
const char32_t EGYPTIAN_HIEROGLYPH_13668 = 0x00013668;
const char32_t EGYPTIAN_HIEROGLYPH_13669 = 0x00013669;
const char32_t EGYPTIAN_HIEROGLYPH_1366A = 0x0001366A;
const char32_t EGYPTIAN_HIEROGLYPH_1366B = 0x0001366B;
const char32_t EGYPTIAN_HIEROGLYPH_1366C = 0x0001366C;
const char32_t EGYPTIAN_HIEROGLYPH_1366D = 0x0001366D;
const char32_t EGYPTIAN_HIEROGLYPH_1366E = 0x0001366E;
const char32_t EGYPTIAN_HIEROGLYPH_1366F = 0x0001366F;
const char32_t EGYPTIAN_HIEROGLYPH_13670 = 0x00013670;
const char32_t EGYPTIAN_HIEROGLYPH_13671 = 0x00013671;
const char32_t EGYPTIAN_HIEROGLYPH_13672 = 0x00013672;
const char32_t EGYPTIAN_HIEROGLYPH_13673 = 0x00013673;
const char32_t EGYPTIAN_HIEROGLYPH_13674 = 0x00013674;
const char32_t EGYPTIAN_HIEROGLYPH_13675 = 0x00013675;
const char32_t EGYPTIAN_HIEROGLYPH_13676 = 0x00013676;
const char32_t EGYPTIAN_HIEROGLYPH_13677 = 0x00013677;
const char32_t EGYPTIAN_HIEROGLYPH_13678 = 0x00013678;
const char32_t EGYPTIAN_HIEROGLYPH_13679 = 0x00013679;
const char32_t EGYPTIAN_HIEROGLYPH_1367A = 0x0001367A;
const char32_t EGYPTIAN_HIEROGLYPH_1367B = 0x0001367B;
const char32_t EGYPTIAN_HIEROGLYPH_1367C = 0x0001367C;
const char32_t EGYPTIAN_HIEROGLYPH_1367D = 0x0001367D;
const char32_t EGYPTIAN_HIEROGLYPH_1367E = 0x0001367E;
const char32_t EGYPTIAN_HIEROGLYPH_1367F = 0x0001367F;
const char32_t EGYPTIAN_HIEROGLYPH_13680 = 0x00013680;
const char32_t EGYPTIAN_HIEROGLYPH_13681 = 0x00013681;
const char32_t EGYPTIAN_HIEROGLYPH_13682 = 0x00013682;
const char32_t EGYPTIAN_HIEROGLYPH_13683 = 0x00013683;
const char32_t EGYPTIAN_HIEROGLYPH_13684 = 0x00013684;
const char32_t EGYPTIAN_HIEROGLYPH_13685 = 0x00013685;
const char32_t EGYPTIAN_HIEROGLYPH_13686 = 0x00013686;
const char32_t EGYPTIAN_HIEROGLYPH_13687 = 0x00013687;
const char32_t EGYPTIAN_HIEROGLYPH_13688 = 0x00013688;
const char32_t EGYPTIAN_HIEROGLYPH_13689 = 0x00013689;
const char32_t EGYPTIAN_HIEROGLYPH_1368A = 0x0001368A;
const char32_t EGYPTIAN_HIEROGLYPH_1368B = 0x0001368B;
const char32_t EGYPTIAN_HIEROGLYPH_1368C = 0x0001368C;
const char32_t EGYPTIAN_HIEROGLYPH_1368D = 0x0001368D;
const char32_t EGYPTIAN_HIEROGLYPH_1368E = 0x0001368E;
const char32_t EGYPTIAN_HIEROGLYPH_1368F = 0x0001368F;
const char32_t EGYPTIAN_HIEROGLYPH_13690 = 0x00013690;
const char32_t EGYPTIAN_HIEROGLYPH_13691 = 0x00013691;
const char32_t EGYPTIAN_HIEROGLYPH_13692 = 0x00013692;
const char32_t EGYPTIAN_HIEROGLYPH_13693 = 0x00013693;
const char32_t EGYPTIAN_HIEROGLYPH_13694 = 0x00013694;
const char32_t EGYPTIAN_HIEROGLYPH_13695 = 0x00013695;
const char32_t EGYPTIAN_HIEROGLYPH_13696 = 0x00013696;
const char32_t EGYPTIAN_HIEROGLYPH_13697 = 0x00013697;
const char32_t EGYPTIAN_HIEROGLYPH_13698 = 0x00013698;
const char32_t EGYPTIAN_HIEROGLYPH_13699 = 0x00013699;
const char32_t EGYPTIAN_HIEROGLYPH_1369A = 0x0001369A;
const char32_t EGYPTIAN_HIEROGLYPH_1369B = 0x0001369B;
const char32_t EGYPTIAN_HIEROGLYPH_1369C = 0x0001369C;
const char32_t EGYPTIAN_HIEROGLYPH_1369D = 0x0001369D;
const char32_t EGYPTIAN_HIEROGLYPH_1369E = 0x0001369E;
const char32_t EGYPTIAN_HIEROGLYPH_1369F = 0x0001369F;
const char32_t EGYPTIAN_HIEROGLYPH_136A0 = 0x000136A0;
const char32_t EGYPTIAN_HIEROGLYPH_136A1 = 0x000136A1;
const char32_t EGYPTIAN_HIEROGLYPH_136A2 = 0x000136A2;
const char32_t EGYPTIAN_HIEROGLYPH_136A3 = 0x000136A3;
const char32_t EGYPTIAN_HIEROGLYPH_136A4 = 0x000136A4;
const char32_t EGYPTIAN_HIEROGLYPH_136A5 = 0x000136A5;
const char32_t EGYPTIAN_HIEROGLYPH_136A6 = 0x000136A6;
const char32_t EGYPTIAN_HIEROGLYPH_136A7 = 0x000136A7;
const char32_t EGYPTIAN_HIEROGLYPH_136A8 = 0x000136A8;
const char32_t EGYPTIAN_HIEROGLYPH_136A9 = 0x000136A9;
const char32_t EGYPTIAN_HIEROGLYPH_136AA = 0x000136AA;
const char32_t EGYPTIAN_HIEROGLYPH_136AB = 0x000136AB;
const char32_t EGYPTIAN_HIEROGLYPH_136AC = 0x000136AC;
const char32_t EGYPTIAN_HIEROGLYPH_136AD = 0x000136AD;
const char32_t EGYPTIAN_HIEROGLYPH_136AE = 0x000136AE;
const char32_t EGYPTIAN_HIEROGLYPH_136AF = 0x000136AF;
const char32_t EGYPTIAN_HIEROGLYPH_136B0 = 0x000136B0;
const char32_t EGYPTIAN_HIEROGLYPH_136B1 = 0x000136B1;
const char32_t EGYPTIAN_HIEROGLYPH_136B2 = 0x000136B2;
const char32_t EGYPTIAN_HIEROGLYPH_136B3 = 0x000136B3;
const char32_t EGYPTIAN_HIEROGLYPH_136B4 = 0x000136B4;
const char32_t EGYPTIAN_HIEROGLYPH_136B5 = 0x000136B5;
const char32_t EGYPTIAN_HIEROGLYPH_136B6 = 0x000136B6;
const char32_t EGYPTIAN_HIEROGLYPH_136B7 = 0x000136B7;
const char32_t EGYPTIAN_HIEROGLYPH_136B8 = 0x000136B8;
const char32_t EGYPTIAN_HIEROGLYPH_136B9 = 0x000136B9;
const char32_t EGYPTIAN_HIEROGLYPH_136BA = 0x000136BA;
const char32_t EGYPTIAN_HIEROGLYPH_136BB = 0x000136BB;
const char32_t EGYPTIAN_HIEROGLYPH_136BC = 0x000136BC;
const char32_t EGYPTIAN_HIEROGLYPH_136BD = 0x000136BD;
const char32_t EGYPTIAN_HIEROGLYPH_136BE = 0x000136BE;
const char32_t EGYPTIAN_HIEROGLYPH_136BF = 0x000136BF;
const char32_t EGYPTIAN_HIEROGLYPH_136C0 = 0x000136C0;
const char32_t EGYPTIAN_HIEROGLYPH_136C1 = 0x000136C1;
const char32_t EGYPTIAN_HIEROGLYPH_136C2 = 0x000136C2;
const char32_t EGYPTIAN_HIEROGLYPH_136C3 = 0x000136C3;
const char32_t EGYPTIAN_HIEROGLYPH_136C4 = 0x000136C4;
const char32_t EGYPTIAN_HIEROGLYPH_136C5 = 0x000136C5;
const char32_t EGYPTIAN_HIEROGLYPH_136C6 = 0x000136C6;
const char32_t EGYPTIAN_HIEROGLYPH_136C7 = 0x000136C7;
const char32_t EGYPTIAN_HIEROGLYPH_136C8 = 0x000136C8;
const char32_t EGYPTIAN_HIEROGLYPH_136C9 = 0x000136C9;
const char32_t EGYPTIAN_HIEROGLYPH_136CA = 0x000136CA;
const char32_t EGYPTIAN_HIEROGLYPH_136CB = 0x000136CB;
const char32_t EGYPTIAN_HIEROGLYPH_136CC = 0x000136CC;
const char32_t EGYPTIAN_HIEROGLYPH_136CD = 0x000136CD;
const char32_t EGYPTIAN_HIEROGLYPH_136CE = 0x000136CE;
const char32_t EGYPTIAN_HIEROGLYPH_136CF = 0x000136CF;
const char32_t EGYPTIAN_HIEROGLYPH_136D0 = 0x000136D0;
const char32_t EGYPTIAN_HIEROGLYPH_136D1 = 0x000136D1;
const char32_t EGYPTIAN_HIEROGLYPH_136D2 = 0x000136D2;
const char32_t EGYPTIAN_HIEROGLYPH_136D3 = 0x000136D3;
const char32_t EGYPTIAN_HIEROGLYPH_136D4 = 0x000136D4;
const char32_t EGYPTIAN_HIEROGLYPH_136D5 = 0x000136D5;
const char32_t EGYPTIAN_HIEROGLYPH_136D6 = 0x000136D6;
const char32_t EGYPTIAN_HIEROGLYPH_136D7 = 0x000136D7;
const char32_t EGYPTIAN_HIEROGLYPH_136D8 = 0x000136D8;
const char32_t EGYPTIAN_HIEROGLYPH_136D9 = 0x000136D9;
const char32_t EGYPTIAN_HIEROGLYPH_136DA = 0x000136DA;
const char32_t EGYPTIAN_HIEROGLYPH_136DB = 0x000136DB;
const char32_t EGYPTIAN_HIEROGLYPH_136DC = 0x000136DC;
const char32_t EGYPTIAN_HIEROGLYPH_136DD = 0x000136DD;
const char32_t EGYPTIAN_HIEROGLYPH_136DE = 0x000136DE;
const char32_t EGYPTIAN_HIEROGLYPH_136DF = 0x000136DF;
const char32_t EGYPTIAN_HIEROGLYPH_136E0 = 0x000136E0;
const char32_t EGYPTIAN_HIEROGLYPH_136E1 = 0x000136E1;
const char32_t EGYPTIAN_HIEROGLYPH_136E2 = 0x000136E2;
const char32_t EGYPTIAN_HIEROGLYPH_136E3 = 0x000136E3;
const char32_t EGYPTIAN_HIEROGLYPH_136E4 = 0x000136E4;
const char32_t EGYPTIAN_HIEROGLYPH_136E5 = 0x000136E5;
const char32_t EGYPTIAN_HIEROGLYPH_136E6 = 0x000136E6;
const char32_t EGYPTIAN_HIEROGLYPH_136E7 = 0x000136E7;
const char32_t EGYPTIAN_HIEROGLYPH_136E8 = 0x000136E8;
const char32_t EGYPTIAN_HIEROGLYPH_136E9 = 0x000136E9;
const char32_t EGYPTIAN_HIEROGLYPH_136EA = 0x000136EA;
const char32_t EGYPTIAN_HIEROGLYPH_136EB = 0x000136EB;
const char32_t EGYPTIAN_HIEROGLYPH_136EC = 0x000136EC;
const char32_t EGYPTIAN_HIEROGLYPH_136ED = 0x000136ED;
const char32_t EGYPTIAN_HIEROGLYPH_136EE = 0x000136EE;
const char32_t EGYPTIAN_HIEROGLYPH_136EF = 0x000136EF;
const char32_t EGYPTIAN_HIEROGLYPH_136F0 = 0x000136F0;
const char32_t EGYPTIAN_HIEROGLYPH_136F1 = 0x000136F1;
const char32_t EGYPTIAN_HIEROGLYPH_136F2 = 0x000136F2;
const char32_t EGYPTIAN_HIEROGLYPH_136F3 = 0x000136F3;
const char32_t EGYPTIAN_HIEROGLYPH_136F4 = 0x000136F4;
const char32_t EGYPTIAN_HIEROGLYPH_136F5 = 0x000136F5;
const char32_t EGYPTIAN_HIEROGLYPH_136F6 = 0x000136F6;
const char32_t EGYPTIAN_HIEROGLYPH_136F7 = 0x000136F7;
const char32_t EGYPTIAN_HIEROGLYPH_136F8 = 0x000136F8;
const char32_t EGYPTIAN_HIEROGLYPH_136F9 = 0x000136F9;
const char32_t EGYPTIAN_HIEROGLYPH_136FA = 0x000136FA;
const char32_t EGYPTIAN_HIEROGLYPH_136FB = 0x000136FB;
const char32_t EGYPTIAN_HIEROGLYPH_136FC = 0x000136FC;
const char32_t EGYPTIAN_HIEROGLYPH_136FD = 0x000136FD;
const char32_t EGYPTIAN_HIEROGLYPH_136FE = 0x000136FE;
const char32_t EGYPTIAN_HIEROGLYPH_136FF = 0x000136FF;
const char32_t EGYPTIAN_HIEROGLYPH_13700 = 0x00013700;
const char32_t EGYPTIAN_HIEROGLYPH_13701 = 0x00013701;
const char32_t EGYPTIAN_HIEROGLYPH_13702 = 0x00013702;
const char32_t EGYPTIAN_HIEROGLYPH_13703 = 0x00013703;
const char32_t EGYPTIAN_HIEROGLYPH_13704 = 0x00013704;
const char32_t EGYPTIAN_HIEROGLYPH_13705 = 0x00013705;
const char32_t EGYPTIAN_HIEROGLYPH_13706 = 0x00013706;
const char32_t EGYPTIAN_HIEROGLYPH_13707 = 0x00013707;
const char32_t EGYPTIAN_HIEROGLYPH_13708 = 0x00013708;
const char32_t EGYPTIAN_HIEROGLYPH_13709 = 0x00013709;
const char32_t EGYPTIAN_HIEROGLYPH_1370A = 0x0001370A;
const char32_t EGYPTIAN_HIEROGLYPH_1370B = 0x0001370B;
const char32_t EGYPTIAN_HIEROGLYPH_1370C = 0x0001370C;
const char32_t EGYPTIAN_HIEROGLYPH_1370D = 0x0001370D;
const char32_t EGYPTIAN_HIEROGLYPH_1370E = 0x0001370E;
const char32_t EGYPTIAN_HIEROGLYPH_1370F = 0x0001370F;
const char32_t EGYPTIAN_HIEROGLYPH_13710 = 0x00013710;
const char32_t EGYPTIAN_HIEROGLYPH_13711 = 0x00013711;
const char32_t EGYPTIAN_HIEROGLYPH_13712 = 0x00013712;
const char32_t EGYPTIAN_HIEROGLYPH_13713 = 0x00013713;
const char32_t EGYPTIAN_HIEROGLYPH_13714 = 0x00013714;
const char32_t EGYPTIAN_HIEROGLYPH_13715 = 0x00013715;
const char32_t EGYPTIAN_HIEROGLYPH_13716 = 0x00013716;
const char32_t EGYPTIAN_HIEROGLYPH_13717 = 0x00013717;
const char32_t EGYPTIAN_HIEROGLYPH_13718 = 0x00013718;
const char32_t EGYPTIAN_HIEROGLYPH_13719 = 0x00013719;
const char32_t EGYPTIAN_HIEROGLYPH_1371A = 0x0001371A;
const char32_t EGYPTIAN_HIEROGLYPH_1371B = 0x0001371B;
const char32_t EGYPTIAN_HIEROGLYPH_1371C = 0x0001371C;
const char32_t EGYPTIAN_HIEROGLYPH_1371D = 0x0001371D;
const char32_t EGYPTIAN_HIEROGLYPH_1371E = 0x0001371E;
const char32_t EGYPTIAN_HIEROGLYPH_1371F = 0x0001371F;
const char32_t EGYPTIAN_HIEROGLYPH_13720 = 0x00013720;
const char32_t EGYPTIAN_HIEROGLYPH_13721 = 0x00013721;
const char32_t EGYPTIAN_HIEROGLYPH_13722 = 0x00013722;
const char32_t EGYPTIAN_HIEROGLYPH_13723 = 0x00013723;
const char32_t EGYPTIAN_HIEROGLYPH_13724 = 0x00013724;
const char32_t EGYPTIAN_HIEROGLYPH_13725 = 0x00013725;
const char32_t EGYPTIAN_HIEROGLYPH_13726 = 0x00013726;
const char32_t EGYPTIAN_HIEROGLYPH_13727 = 0x00013727;
const char32_t EGYPTIAN_HIEROGLYPH_13728 = 0x00013728;
const char32_t EGYPTIAN_HIEROGLYPH_13729 = 0x00013729;
const char32_t EGYPTIAN_HIEROGLYPH_1372A = 0x0001372A;
const char32_t EGYPTIAN_HIEROGLYPH_1372B = 0x0001372B;
const char32_t EGYPTIAN_HIEROGLYPH_1372C = 0x0001372C;
const char32_t EGYPTIAN_HIEROGLYPH_1372D = 0x0001372D;
const char32_t EGYPTIAN_HIEROGLYPH_1372E = 0x0001372E;
const char32_t EGYPTIAN_HIEROGLYPH_1372F = 0x0001372F;
const char32_t EGYPTIAN_HIEROGLYPH_13730 = 0x00013730;
const char32_t EGYPTIAN_HIEROGLYPH_13731 = 0x00013731;
const char32_t EGYPTIAN_HIEROGLYPH_13732 = 0x00013732;
const char32_t EGYPTIAN_HIEROGLYPH_13733 = 0x00013733;
const char32_t EGYPTIAN_HIEROGLYPH_13734 = 0x00013734;
const char32_t EGYPTIAN_HIEROGLYPH_13735 = 0x00013735;
const char32_t EGYPTIAN_HIEROGLYPH_13736 = 0x00013736;
const char32_t EGYPTIAN_HIEROGLYPH_13737 = 0x00013737;
const char32_t EGYPTIAN_HIEROGLYPH_13738 = 0x00013738;
const char32_t EGYPTIAN_HIEROGLYPH_13739 = 0x00013739;
const char32_t EGYPTIAN_HIEROGLYPH_1373A = 0x0001373A;
const char32_t EGYPTIAN_HIEROGLYPH_1373B = 0x0001373B;
const char32_t EGYPTIAN_HIEROGLYPH_1373C = 0x0001373C;
const char32_t EGYPTIAN_HIEROGLYPH_1373D = 0x0001373D;
const char32_t EGYPTIAN_HIEROGLYPH_1373E = 0x0001373E;
const char32_t EGYPTIAN_HIEROGLYPH_1373F = 0x0001373F;
const char32_t EGYPTIAN_HIEROGLYPH_13740 = 0x00013740;
const char32_t EGYPTIAN_HIEROGLYPH_13741 = 0x00013741;
const char32_t EGYPTIAN_HIEROGLYPH_13742 = 0x00013742;
const char32_t EGYPTIAN_HIEROGLYPH_13743 = 0x00013743;
const char32_t EGYPTIAN_HIEROGLYPH_13744 = 0x00013744;
const char32_t EGYPTIAN_HIEROGLYPH_13745 = 0x00013745;
const char32_t EGYPTIAN_HIEROGLYPH_13746 = 0x00013746;
const char32_t EGYPTIAN_HIEROGLYPH_13747 = 0x00013747;
const char32_t EGYPTIAN_HIEROGLYPH_13748 = 0x00013748;
const char32_t EGYPTIAN_HIEROGLYPH_13749 = 0x00013749;
const char32_t EGYPTIAN_HIEROGLYPH_1374A = 0x0001374A;
const char32_t EGYPTIAN_HIEROGLYPH_1374B = 0x0001374B;
const char32_t EGYPTIAN_HIEROGLYPH_1374C = 0x0001374C;
const char32_t EGYPTIAN_HIEROGLYPH_1374D = 0x0001374D;
const char32_t EGYPTIAN_HIEROGLYPH_1374E = 0x0001374E;
const char32_t EGYPTIAN_HIEROGLYPH_1374F = 0x0001374F;
const char32_t EGYPTIAN_HIEROGLYPH_13750 = 0x00013750;
const char32_t EGYPTIAN_HIEROGLYPH_13751 = 0x00013751;
const char32_t EGYPTIAN_HIEROGLYPH_13752 = 0x00013752;
const char32_t EGYPTIAN_HIEROGLYPH_13753 = 0x00013753;
const char32_t EGYPTIAN_HIEROGLYPH_13754 = 0x00013754;
const char32_t EGYPTIAN_HIEROGLYPH_13755 = 0x00013755;
const char32_t EGYPTIAN_HIEROGLYPH_13756 = 0x00013756;
const char32_t EGYPTIAN_HIEROGLYPH_13757 = 0x00013757;
const char32_t EGYPTIAN_HIEROGLYPH_13758 = 0x00013758;
const char32_t EGYPTIAN_HIEROGLYPH_13759 = 0x00013759;
const char32_t EGYPTIAN_HIEROGLYPH_1375A = 0x0001375A;
const char32_t EGYPTIAN_HIEROGLYPH_1375B = 0x0001375B;
const char32_t EGYPTIAN_HIEROGLYPH_1375C = 0x0001375C;
const char32_t EGYPTIAN_HIEROGLYPH_1375D = 0x0001375D;
const char32_t EGYPTIAN_HIEROGLYPH_1375E = 0x0001375E;
const char32_t EGYPTIAN_HIEROGLYPH_1375F = 0x0001375F;
const char32_t EGYPTIAN_HIEROGLYPH_13760 = 0x00013760;
const char32_t EGYPTIAN_HIEROGLYPH_13761 = 0x00013761;
const char32_t EGYPTIAN_HIEROGLYPH_13762 = 0x00013762;
const char32_t EGYPTIAN_HIEROGLYPH_13763 = 0x00013763;
const char32_t EGYPTIAN_HIEROGLYPH_13764 = 0x00013764;
const char32_t EGYPTIAN_HIEROGLYPH_13765 = 0x00013765;
const char32_t EGYPTIAN_HIEROGLYPH_13766 = 0x00013766;
const char32_t EGYPTIAN_HIEROGLYPH_13767 = 0x00013767;
const char32_t EGYPTIAN_HIEROGLYPH_13768 = 0x00013768;
const char32_t EGYPTIAN_HIEROGLYPH_13769 = 0x00013769;
const char32_t EGYPTIAN_HIEROGLYPH_1376A = 0x0001376A;
const char32_t EGYPTIAN_HIEROGLYPH_1376B = 0x0001376B;
const char32_t EGYPTIAN_HIEROGLYPH_1376C = 0x0001376C;
const char32_t EGYPTIAN_HIEROGLYPH_1376D = 0x0001376D;
const char32_t EGYPTIAN_HIEROGLYPH_1376E = 0x0001376E;
const char32_t EGYPTIAN_HIEROGLYPH_1376F = 0x0001376F;
const char32_t EGYPTIAN_HIEROGLYPH_13770 = 0x00013770;
const char32_t EGYPTIAN_HIEROGLYPH_13771 = 0x00013771;
const char32_t EGYPTIAN_HIEROGLYPH_13772 = 0x00013772;
const char32_t EGYPTIAN_HIEROGLYPH_13773 = 0x00013773;
const char32_t EGYPTIAN_HIEROGLYPH_13774 = 0x00013774;
const char32_t EGYPTIAN_HIEROGLYPH_13775 = 0x00013775;
const char32_t EGYPTIAN_HIEROGLYPH_13776 = 0x00013776;
const char32_t EGYPTIAN_HIEROGLYPH_13777 = 0x00013777;
const char32_t EGYPTIAN_HIEROGLYPH_13778 = 0x00013778;
const char32_t EGYPTIAN_HIEROGLYPH_13779 = 0x00013779;
const char32_t EGYPTIAN_HIEROGLYPH_1377A = 0x0001377A;
const char32_t EGYPTIAN_HIEROGLYPH_1377B = 0x0001377B;
const char32_t EGYPTIAN_HIEROGLYPH_1377C = 0x0001377C;
const char32_t EGYPTIAN_HIEROGLYPH_1377D = 0x0001377D;
const char32_t EGYPTIAN_HIEROGLYPH_1377E = 0x0001377E;
const char32_t EGYPTIAN_HIEROGLYPH_1377F = 0x0001377F;
const char32_t EGYPTIAN_HIEROGLYPH_13780 = 0x00013780;
const char32_t EGYPTIAN_HIEROGLYPH_13781 = 0x00013781;
const char32_t EGYPTIAN_HIEROGLYPH_13782 = 0x00013782;
const char32_t EGYPTIAN_HIEROGLYPH_13783 = 0x00013783;
const char32_t EGYPTIAN_HIEROGLYPH_13784 = 0x00013784;
const char32_t EGYPTIAN_HIEROGLYPH_13785 = 0x00013785;
const char32_t EGYPTIAN_HIEROGLYPH_13786 = 0x00013786;
const char32_t EGYPTIAN_HIEROGLYPH_13787 = 0x00013787;
const char32_t EGYPTIAN_HIEROGLYPH_13788 = 0x00013788;
const char32_t EGYPTIAN_HIEROGLYPH_13789 = 0x00013789;
const char32_t EGYPTIAN_HIEROGLYPH_1378A = 0x0001378A;
const char32_t EGYPTIAN_HIEROGLYPH_1378B = 0x0001378B;
const char32_t EGYPTIAN_HIEROGLYPH_1378C = 0x0001378C;
const char32_t EGYPTIAN_HIEROGLYPH_1378D = 0x0001378D;
const char32_t EGYPTIAN_HIEROGLYPH_1378E = 0x0001378E;
const char32_t EGYPTIAN_HIEROGLYPH_1378F = 0x0001378F;
const char32_t EGYPTIAN_HIEROGLYPH_13790 = 0x00013790;
const char32_t EGYPTIAN_HIEROGLYPH_13791 = 0x00013791;
const char32_t EGYPTIAN_HIEROGLYPH_13792 = 0x00013792;
const char32_t EGYPTIAN_HIEROGLYPH_13793 = 0x00013793;
const char32_t EGYPTIAN_HIEROGLYPH_13794 = 0x00013794;
const char32_t EGYPTIAN_HIEROGLYPH_13795 = 0x00013795;
const char32_t EGYPTIAN_HIEROGLYPH_13796 = 0x00013796;
const char32_t EGYPTIAN_HIEROGLYPH_13797 = 0x00013797;
const char32_t EGYPTIAN_HIEROGLYPH_13798 = 0x00013798;
const char32_t EGYPTIAN_HIEROGLYPH_13799 = 0x00013799;
const char32_t EGYPTIAN_HIEROGLYPH_1379A = 0x0001379A;
const char32_t EGYPTIAN_HIEROGLYPH_1379B = 0x0001379B;
const char32_t EGYPTIAN_HIEROGLYPH_1379C = 0x0001379C;
const char32_t EGYPTIAN_HIEROGLYPH_1379D = 0x0001379D;
const char32_t EGYPTIAN_HIEROGLYPH_1379E = 0x0001379E;
const char32_t EGYPTIAN_HIEROGLYPH_1379F = 0x0001379F;
const char32_t EGYPTIAN_HIEROGLYPH_137A0 = 0x000137A0;
const char32_t EGYPTIAN_HIEROGLYPH_137A1 = 0x000137A1;
const char32_t EGYPTIAN_HIEROGLYPH_137A2 = 0x000137A2;
const char32_t EGYPTIAN_HIEROGLYPH_137A3 = 0x000137A3;
const char32_t EGYPTIAN_HIEROGLYPH_137A4 = 0x000137A4;
const char32_t EGYPTIAN_HIEROGLYPH_137A5 = 0x000137A5;
const char32_t EGYPTIAN_HIEROGLYPH_137A6 = 0x000137A6;
const char32_t EGYPTIAN_HIEROGLYPH_137A7 = 0x000137A7;
const char32_t EGYPTIAN_HIEROGLYPH_137A8 = 0x000137A8;
const char32_t EGYPTIAN_HIEROGLYPH_137A9 = 0x000137A9;
const char32_t EGYPTIAN_HIEROGLYPH_137AA = 0x000137AA;
const char32_t EGYPTIAN_HIEROGLYPH_137AB = 0x000137AB;
const char32_t EGYPTIAN_HIEROGLYPH_137AC = 0x000137AC;
const char32_t EGYPTIAN_HIEROGLYPH_137AD = 0x000137AD;
const char32_t EGYPTIAN_HIEROGLYPH_137AE = 0x000137AE;
const char32_t EGYPTIAN_HIEROGLYPH_137AF = 0x000137AF;
const char32_t EGYPTIAN_HIEROGLYPH_137B0 = 0x000137B0;
const char32_t EGYPTIAN_HIEROGLYPH_137B1 = 0x000137B1;
const char32_t EGYPTIAN_HIEROGLYPH_137B2 = 0x000137B2;
const char32_t EGYPTIAN_HIEROGLYPH_137B3 = 0x000137B3;
const char32_t EGYPTIAN_HIEROGLYPH_137B4 = 0x000137B4;
const char32_t EGYPTIAN_HIEROGLYPH_137B5 = 0x000137B5;
const char32_t EGYPTIAN_HIEROGLYPH_137B6 = 0x000137B6;
const char32_t EGYPTIAN_HIEROGLYPH_137B7 = 0x000137B7;
const char32_t EGYPTIAN_HIEROGLYPH_137B8 = 0x000137B8;
const char32_t EGYPTIAN_HIEROGLYPH_137B9 = 0x000137B9;
const char32_t EGYPTIAN_HIEROGLYPH_137BA = 0x000137BA;
const char32_t EGYPTIAN_HIEROGLYPH_137BB = 0x000137BB;
const char32_t EGYPTIAN_HIEROGLYPH_137BC = 0x000137BC;
const char32_t EGYPTIAN_HIEROGLYPH_137BD = 0x000137BD;
const char32_t EGYPTIAN_HIEROGLYPH_137BE = 0x000137BE;
const char32_t EGYPTIAN_HIEROGLYPH_137BF = 0x000137BF;
const char32_t EGYPTIAN_HIEROGLYPH_137C0 = 0x000137C0;
const char32_t EGYPTIAN_HIEROGLYPH_137C1 = 0x000137C1;
const char32_t EGYPTIAN_HIEROGLYPH_137C2 = 0x000137C2;
const char32_t EGYPTIAN_HIEROGLYPH_137C3 = 0x000137C3;
const char32_t EGYPTIAN_HIEROGLYPH_137C4 = 0x000137C4;
const char32_t EGYPTIAN_HIEROGLYPH_137C5 = 0x000137C5;
const char32_t EGYPTIAN_HIEROGLYPH_137C6 = 0x000137C6;
const char32_t EGYPTIAN_HIEROGLYPH_137C7 = 0x000137C7;
const char32_t EGYPTIAN_HIEROGLYPH_137C8 = 0x000137C8;
const char32_t EGYPTIAN_HIEROGLYPH_137C9 = 0x000137C9;
const char32_t EGYPTIAN_HIEROGLYPH_137CA = 0x000137CA;
const char32_t EGYPTIAN_HIEROGLYPH_137CB = 0x000137CB;
const char32_t EGYPTIAN_HIEROGLYPH_137CC = 0x000137CC;
const char32_t EGYPTIAN_HIEROGLYPH_137CD = 0x000137CD;
const char32_t EGYPTIAN_HIEROGLYPH_137CE = 0x000137CE;
const char32_t EGYPTIAN_HIEROGLYPH_137CF = 0x000137CF;
const char32_t EGYPTIAN_HIEROGLYPH_137D0 = 0x000137D0;
const char32_t EGYPTIAN_HIEROGLYPH_137D1 = 0x000137D1;
const char32_t EGYPTIAN_HIEROGLYPH_137D2 = 0x000137D2;
const char32_t EGYPTIAN_HIEROGLYPH_137D3 = 0x000137D3;
const char32_t EGYPTIAN_HIEROGLYPH_137D4 = 0x000137D4;
const char32_t EGYPTIAN_HIEROGLYPH_137D5 = 0x000137D5;
const char32_t EGYPTIAN_HIEROGLYPH_137D6 = 0x000137D6;
const char32_t EGYPTIAN_HIEROGLYPH_137D7 = 0x000137D7;
const char32_t EGYPTIAN_HIEROGLYPH_137D8 = 0x000137D8;
const char32_t EGYPTIAN_HIEROGLYPH_137D9 = 0x000137D9;
const char32_t EGYPTIAN_HIEROGLYPH_137DA = 0x000137DA;
const char32_t EGYPTIAN_HIEROGLYPH_137DB = 0x000137DB;
const char32_t EGYPTIAN_HIEROGLYPH_137DC = 0x000137DC;
const char32_t EGYPTIAN_HIEROGLYPH_137DD = 0x000137DD;
const char32_t EGYPTIAN_HIEROGLYPH_137DE = 0x000137DE;
const char32_t EGYPTIAN_HIEROGLYPH_137DF = 0x000137DF;
const char32_t EGYPTIAN_HIEROGLYPH_137E0 = 0x000137E0;
const char32_t EGYPTIAN_HIEROGLYPH_137E1 = 0x000137E1;
const char32_t EGYPTIAN_HIEROGLYPH_137E2 = 0x000137E2;
const char32_t EGYPTIAN_HIEROGLYPH_137E3 = 0x000137E3;
const char32_t EGYPTIAN_HIEROGLYPH_137E4 = 0x000137E4;
const char32_t EGYPTIAN_HIEROGLYPH_137E5 = 0x000137E5;
const char32_t EGYPTIAN_HIEROGLYPH_137E6 = 0x000137E6;
const char32_t EGYPTIAN_HIEROGLYPH_137E7 = 0x000137E7;
const char32_t EGYPTIAN_HIEROGLYPH_137E8 = 0x000137E8;
const char32_t EGYPTIAN_HIEROGLYPH_137E9 = 0x000137E9;
const char32_t EGYPTIAN_HIEROGLYPH_137EA = 0x000137EA;
const char32_t EGYPTIAN_HIEROGLYPH_137EB = 0x000137EB;
const char32_t EGYPTIAN_HIEROGLYPH_137EC = 0x000137EC;
const char32_t EGYPTIAN_HIEROGLYPH_137ED = 0x000137ED;
const char32_t EGYPTIAN_HIEROGLYPH_137EE = 0x000137EE;
const char32_t EGYPTIAN_HIEROGLYPH_137EF = 0x000137EF;
const char32_t EGYPTIAN_HIEROGLYPH_137F0 = 0x000137F0;
const char32_t EGYPTIAN_HIEROGLYPH_137F1 = 0x000137F1;
const char32_t EGYPTIAN_HIEROGLYPH_137F2 = 0x000137F2;
const char32_t EGYPTIAN_HIEROGLYPH_137F3 = 0x000137F3;
const char32_t EGYPTIAN_HIEROGLYPH_137F4 = 0x000137F4;
const char32_t EGYPTIAN_HIEROGLYPH_137F5 = 0x000137F5;
const char32_t EGYPTIAN_HIEROGLYPH_137F6 = 0x000137F6;
const char32_t EGYPTIAN_HIEROGLYPH_137F7 = 0x000137F7;
const char32_t EGYPTIAN_HIEROGLYPH_137F8 = 0x000137F8;
const char32_t EGYPTIAN_HIEROGLYPH_137F9 = 0x000137F9;
const char32_t EGYPTIAN_HIEROGLYPH_137FA = 0x000137FA;
const char32_t EGYPTIAN_HIEROGLYPH_137FB = 0x000137FB;
const char32_t EGYPTIAN_HIEROGLYPH_137FC = 0x000137FC;
const char32_t EGYPTIAN_HIEROGLYPH_137FD = 0x000137FD;
const char32_t EGYPTIAN_HIEROGLYPH_137FE = 0x000137FE;
const char32_t EGYPTIAN_HIEROGLYPH_137FF = 0x000137FF;
const char32_t EGYPTIAN_HIEROGLYPH_13800 = 0x00013800;
const char32_t EGYPTIAN_HIEROGLYPH_13801 = 0x00013801;
const char32_t EGYPTIAN_HIEROGLYPH_13802 = 0x00013802;
const char32_t EGYPTIAN_HIEROGLYPH_13803 = 0x00013803;
const char32_t EGYPTIAN_HIEROGLYPH_13804 = 0x00013804;
const char32_t EGYPTIAN_HIEROGLYPH_13805 = 0x00013805;
const char32_t EGYPTIAN_HIEROGLYPH_13806 = 0x00013806;
const char32_t EGYPTIAN_HIEROGLYPH_13807 = 0x00013807;
const char32_t EGYPTIAN_HIEROGLYPH_13808 = 0x00013808;
const char32_t EGYPTIAN_HIEROGLYPH_13809 = 0x00013809;
const char32_t EGYPTIAN_HIEROGLYPH_1380A = 0x0001380A;
const char32_t EGYPTIAN_HIEROGLYPH_1380B = 0x0001380B;
const char32_t EGYPTIAN_HIEROGLYPH_1380C = 0x0001380C;
const char32_t EGYPTIAN_HIEROGLYPH_1380D = 0x0001380D;
const char32_t EGYPTIAN_HIEROGLYPH_1380E = 0x0001380E;
const char32_t EGYPTIAN_HIEROGLYPH_1380F = 0x0001380F;
const char32_t EGYPTIAN_HIEROGLYPH_13810 = 0x00013810;
const char32_t EGYPTIAN_HIEROGLYPH_13811 = 0x00013811;
const char32_t EGYPTIAN_HIEROGLYPH_13812 = 0x00013812;
const char32_t EGYPTIAN_HIEROGLYPH_13813 = 0x00013813;
const char32_t EGYPTIAN_HIEROGLYPH_13814 = 0x00013814;
const char32_t EGYPTIAN_HIEROGLYPH_13815 = 0x00013815;
const char32_t EGYPTIAN_HIEROGLYPH_13816 = 0x00013816;
const char32_t EGYPTIAN_HIEROGLYPH_13817 = 0x00013817;
const char32_t EGYPTIAN_HIEROGLYPH_13818 = 0x00013818;
const char32_t EGYPTIAN_HIEROGLYPH_13819 = 0x00013819;
const char32_t EGYPTIAN_HIEROGLYPH_1381A = 0x0001381A;
const char32_t EGYPTIAN_HIEROGLYPH_1381B = 0x0001381B;
const char32_t EGYPTIAN_HIEROGLYPH_1381C = 0x0001381C;
const char32_t EGYPTIAN_HIEROGLYPH_1381D = 0x0001381D;
const char32_t EGYPTIAN_HIEROGLYPH_1381E = 0x0001381E;
const char32_t EGYPTIAN_HIEROGLYPH_1381F = 0x0001381F;
const char32_t EGYPTIAN_HIEROGLYPH_13820 = 0x00013820;
const char32_t EGYPTIAN_HIEROGLYPH_13821 = 0x00013821;
const char32_t EGYPTIAN_HIEROGLYPH_13822 = 0x00013822;
const char32_t EGYPTIAN_HIEROGLYPH_13823 = 0x00013823;
const char32_t EGYPTIAN_HIEROGLYPH_13824 = 0x00013824;
const char32_t EGYPTIAN_HIEROGLYPH_13825 = 0x00013825;
const char32_t EGYPTIAN_HIEROGLYPH_13826 = 0x00013826;
const char32_t EGYPTIAN_HIEROGLYPH_13827 = 0x00013827;
const char32_t EGYPTIAN_HIEROGLYPH_13828 = 0x00013828;
const char32_t EGYPTIAN_HIEROGLYPH_13829 = 0x00013829;
const char32_t EGYPTIAN_HIEROGLYPH_1382A = 0x0001382A;
const char32_t EGYPTIAN_HIEROGLYPH_1382B = 0x0001382B;
const char32_t EGYPTIAN_HIEROGLYPH_1382C = 0x0001382C;
const char32_t EGYPTIAN_HIEROGLYPH_1382D = 0x0001382D;
const char32_t EGYPTIAN_HIEROGLYPH_1382E = 0x0001382E;
const char32_t EGYPTIAN_HIEROGLYPH_1382F = 0x0001382F;
const char32_t EGYPTIAN_HIEROGLYPH_13830 = 0x00013830;
const char32_t EGYPTIAN_HIEROGLYPH_13831 = 0x00013831;
const char32_t EGYPTIAN_HIEROGLYPH_13832 = 0x00013832;
const char32_t EGYPTIAN_HIEROGLYPH_13833 = 0x00013833;
const char32_t EGYPTIAN_HIEROGLYPH_13834 = 0x00013834;
const char32_t EGYPTIAN_HIEROGLYPH_13835 = 0x00013835;
const char32_t EGYPTIAN_HIEROGLYPH_13836 = 0x00013836;
const char32_t EGYPTIAN_HIEROGLYPH_13837 = 0x00013837;
const char32_t EGYPTIAN_HIEROGLYPH_13838 = 0x00013838;
const char32_t EGYPTIAN_HIEROGLYPH_13839 = 0x00013839;
const char32_t EGYPTIAN_HIEROGLYPH_1383A = 0x0001383A;
const char32_t EGYPTIAN_HIEROGLYPH_1383B = 0x0001383B;
const char32_t EGYPTIAN_HIEROGLYPH_1383C = 0x0001383C;
const char32_t EGYPTIAN_HIEROGLYPH_1383D = 0x0001383D;
const char32_t EGYPTIAN_HIEROGLYPH_1383E = 0x0001383E;
const char32_t EGYPTIAN_HIEROGLYPH_1383F = 0x0001383F;
const char32_t EGYPTIAN_HIEROGLYPH_13840 = 0x00013840;
const char32_t EGYPTIAN_HIEROGLYPH_13841 = 0x00013841;
const char32_t EGYPTIAN_HIEROGLYPH_13842 = 0x00013842;
const char32_t EGYPTIAN_HIEROGLYPH_13843 = 0x00013843;
const char32_t EGYPTIAN_HIEROGLYPH_13844 = 0x00013844;
const char32_t EGYPTIAN_HIEROGLYPH_13845 = 0x00013845;
const char32_t EGYPTIAN_HIEROGLYPH_13846 = 0x00013846;
const char32_t EGYPTIAN_HIEROGLYPH_13847 = 0x00013847;
const char32_t EGYPTIAN_HIEROGLYPH_13848 = 0x00013848;
const char32_t EGYPTIAN_HIEROGLYPH_13849 = 0x00013849;
const char32_t EGYPTIAN_HIEROGLYPH_1384A = 0x0001384A;
const char32_t EGYPTIAN_HIEROGLYPH_1384B = 0x0001384B;
const char32_t EGYPTIAN_HIEROGLYPH_1384C = 0x0001384C;
const char32_t EGYPTIAN_HIEROGLYPH_1384D = 0x0001384D;
const char32_t EGYPTIAN_HIEROGLYPH_1384E = 0x0001384E;
const char32_t EGYPTIAN_HIEROGLYPH_1384F = 0x0001384F;
const char32_t EGYPTIAN_HIEROGLYPH_13850 = 0x00013850;
const char32_t EGYPTIAN_HIEROGLYPH_13851 = 0x00013851;
const char32_t EGYPTIAN_HIEROGLYPH_13852 = 0x00013852;
const char32_t EGYPTIAN_HIEROGLYPH_13853 = 0x00013853;
const char32_t EGYPTIAN_HIEROGLYPH_13854 = 0x00013854;
const char32_t EGYPTIAN_HIEROGLYPH_13855 = 0x00013855;
const char32_t EGYPTIAN_HIEROGLYPH_13856 = 0x00013856;
const char32_t EGYPTIAN_HIEROGLYPH_13857 = 0x00013857;
const char32_t EGYPTIAN_HIEROGLYPH_13858 = 0x00013858;
const char32_t EGYPTIAN_HIEROGLYPH_13859 = 0x00013859;
const char32_t EGYPTIAN_HIEROGLYPH_1385A = 0x0001385A;
const char32_t EGYPTIAN_HIEROGLYPH_1385B = 0x0001385B;
const char32_t EGYPTIAN_HIEROGLYPH_1385C = 0x0001385C;
const char32_t EGYPTIAN_HIEROGLYPH_1385D = 0x0001385D;
const char32_t EGYPTIAN_HIEROGLYPH_1385E = 0x0001385E;
const char32_t EGYPTIAN_HIEROGLYPH_1385F = 0x0001385F;
const char32_t EGYPTIAN_HIEROGLYPH_13860 = 0x00013860;
const char32_t EGYPTIAN_HIEROGLYPH_13861 = 0x00013861;
const char32_t EGYPTIAN_HIEROGLYPH_13862 = 0x00013862;
const char32_t EGYPTIAN_HIEROGLYPH_13863 = 0x00013863;
const char32_t EGYPTIAN_HIEROGLYPH_13864 = 0x00013864;
const char32_t EGYPTIAN_HIEROGLYPH_13865 = 0x00013865;
const char32_t EGYPTIAN_HIEROGLYPH_13866 = 0x00013866;
const char32_t EGYPTIAN_HIEROGLYPH_13867 = 0x00013867;
const char32_t EGYPTIAN_HIEROGLYPH_13868 = 0x00013868;
const char32_t EGYPTIAN_HIEROGLYPH_13869 = 0x00013869;
const char32_t EGYPTIAN_HIEROGLYPH_1386A = 0x0001386A;
const char32_t EGYPTIAN_HIEROGLYPH_1386B = 0x0001386B;
const char32_t EGYPTIAN_HIEROGLYPH_1386C = 0x0001386C;
const char32_t EGYPTIAN_HIEROGLYPH_1386D = 0x0001386D;
const char32_t EGYPTIAN_HIEROGLYPH_1386E = 0x0001386E;
const char32_t EGYPTIAN_HIEROGLYPH_1386F = 0x0001386F;
const char32_t EGYPTIAN_HIEROGLYPH_13870 = 0x00013870;
const char32_t EGYPTIAN_HIEROGLYPH_13871 = 0x00013871;
const char32_t EGYPTIAN_HIEROGLYPH_13872 = 0x00013872;
const char32_t EGYPTIAN_HIEROGLYPH_13873 = 0x00013873;
const char32_t EGYPTIAN_HIEROGLYPH_13874 = 0x00013874;
const char32_t EGYPTIAN_HIEROGLYPH_13875 = 0x00013875;
const char32_t EGYPTIAN_HIEROGLYPH_13876 = 0x00013876;
const char32_t EGYPTIAN_HIEROGLYPH_13877 = 0x00013877;
const char32_t EGYPTIAN_HIEROGLYPH_13878 = 0x00013878;
const char32_t EGYPTIAN_HIEROGLYPH_13879 = 0x00013879;
const char32_t EGYPTIAN_HIEROGLYPH_1387A = 0x0001387A;
const char32_t EGYPTIAN_HIEROGLYPH_1387B = 0x0001387B;
const char32_t EGYPTIAN_HIEROGLYPH_1387C = 0x0001387C;
const char32_t EGYPTIAN_HIEROGLYPH_1387D = 0x0001387D;
const char32_t EGYPTIAN_HIEROGLYPH_1387E = 0x0001387E;
const char32_t EGYPTIAN_HIEROGLYPH_1387F = 0x0001387F;
const char32_t EGYPTIAN_HIEROGLYPH_13880 = 0x00013880;
const char32_t EGYPTIAN_HIEROGLYPH_13881 = 0x00013881;
const char32_t EGYPTIAN_HIEROGLYPH_13882 = 0x00013882;
const char32_t EGYPTIAN_HIEROGLYPH_13883 = 0x00013883;
const char32_t EGYPTIAN_HIEROGLYPH_13884 = 0x00013884;
const char32_t EGYPTIAN_HIEROGLYPH_13885 = 0x00013885;
const char32_t EGYPTIAN_HIEROGLYPH_13886 = 0x00013886;
const char32_t EGYPTIAN_HIEROGLYPH_13887 = 0x00013887;
const char32_t EGYPTIAN_HIEROGLYPH_13888 = 0x00013888;
const char32_t EGYPTIAN_HIEROGLYPH_13889 = 0x00013889;
const char32_t EGYPTIAN_HIEROGLYPH_1388A = 0x0001388A;
const char32_t EGYPTIAN_HIEROGLYPH_1388B = 0x0001388B;
const char32_t EGYPTIAN_HIEROGLYPH_1388C = 0x0001388C;
const char32_t EGYPTIAN_HIEROGLYPH_1388D = 0x0001388D;
const char32_t EGYPTIAN_HIEROGLYPH_1388E = 0x0001388E;
const char32_t EGYPTIAN_HIEROGLYPH_1388F = 0x0001388F;
const char32_t EGYPTIAN_HIEROGLYPH_13890 = 0x00013890;
const char32_t EGYPTIAN_HIEROGLYPH_13891 = 0x00013891;
const char32_t EGYPTIAN_HIEROGLYPH_13892 = 0x00013892;
const char32_t EGYPTIAN_HIEROGLYPH_13893 = 0x00013893;
const char32_t EGYPTIAN_HIEROGLYPH_13894 = 0x00013894;
const char32_t EGYPTIAN_HIEROGLYPH_13895 = 0x00013895;
const char32_t EGYPTIAN_HIEROGLYPH_13896 = 0x00013896;
const char32_t EGYPTIAN_HIEROGLYPH_13897 = 0x00013897;
const char32_t EGYPTIAN_HIEROGLYPH_13898 = 0x00013898;
const char32_t EGYPTIAN_HIEROGLYPH_13899 = 0x00013899;
const char32_t EGYPTIAN_HIEROGLYPH_1389A = 0x0001389A;
const char32_t EGYPTIAN_HIEROGLYPH_1389B = 0x0001389B;
const char32_t EGYPTIAN_HIEROGLYPH_1389C = 0x0001389C;
const char32_t EGYPTIAN_HIEROGLYPH_1389D = 0x0001389D;
const char32_t EGYPTIAN_HIEROGLYPH_1389E = 0x0001389E;
const char32_t EGYPTIAN_HIEROGLYPH_1389F = 0x0001389F;
const char32_t EGYPTIAN_HIEROGLYPH_138A0 = 0x000138A0;
const char32_t EGYPTIAN_HIEROGLYPH_138A1 = 0x000138A1;
const char32_t EGYPTIAN_HIEROGLYPH_138A2 = 0x000138A2;
const char32_t EGYPTIAN_HIEROGLYPH_138A3 = 0x000138A3;
const char32_t EGYPTIAN_HIEROGLYPH_138A4 = 0x000138A4;
const char32_t EGYPTIAN_HIEROGLYPH_138A5 = 0x000138A5;
const char32_t EGYPTIAN_HIEROGLYPH_138A6 = 0x000138A6;
const char32_t EGYPTIAN_HIEROGLYPH_138A7 = 0x000138A7;
const char32_t EGYPTIAN_HIEROGLYPH_138A8 = 0x000138A8;
const char32_t EGYPTIAN_HIEROGLYPH_138A9 = 0x000138A9;
const char32_t EGYPTIAN_HIEROGLYPH_138AA = 0x000138AA;
const char32_t EGYPTIAN_HIEROGLYPH_138AB = 0x000138AB;
const char32_t EGYPTIAN_HIEROGLYPH_138AC = 0x000138AC;
const char32_t EGYPTIAN_HIEROGLYPH_138AD = 0x000138AD;
const char32_t EGYPTIAN_HIEROGLYPH_138AE = 0x000138AE;
const char32_t EGYPTIAN_HIEROGLYPH_138AF = 0x000138AF;
const char32_t EGYPTIAN_HIEROGLYPH_138B0 = 0x000138B0;
const char32_t EGYPTIAN_HIEROGLYPH_138B1 = 0x000138B1;
const char32_t EGYPTIAN_HIEROGLYPH_138B2 = 0x000138B2;
const char32_t EGYPTIAN_HIEROGLYPH_138B3 = 0x000138B3;
const char32_t EGYPTIAN_HIEROGLYPH_138B4 = 0x000138B4;
const char32_t EGYPTIAN_HIEROGLYPH_138B5 = 0x000138B5;
const char32_t EGYPTIAN_HIEROGLYPH_138B6 = 0x000138B6;
const char32_t EGYPTIAN_HIEROGLYPH_138B7 = 0x000138B7;
const char32_t EGYPTIAN_HIEROGLYPH_138B8 = 0x000138B8;
const char32_t EGYPTIAN_HIEROGLYPH_138B9 = 0x000138B9;
const char32_t EGYPTIAN_HIEROGLYPH_138BA = 0x000138BA;
const char32_t EGYPTIAN_HIEROGLYPH_138BB = 0x000138BB;
const char32_t EGYPTIAN_HIEROGLYPH_138BC = 0x000138BC;
const char32_t EGYPTIAN_HIEROGLYPH_138BD = 0x000138BD;
const char32_t EGYPTIAN_HIEROGLYPH_138BE = 0x000138BE;
const char32_t EGYPTIAN_HIEROGLYPH_138BF = 0x000138BF;
const char32_t EGYPTIAN_HIEROGLYPH_138C0 = 0x000138C0;
const char32_t EGYPTIAN_HIEROGLYPH_138C1 = 0x000138C1;
const char32_t EGYPTIAN_HIEROGLYPH_138C2 = 0x000138C2;
const char32_t EGYPTIAN_HIEROGLYPH_138C3 = 0x000138C3;
const char32_t EGYPTIAN_HIEROGLYPH_138C4 = 0x000138C4;
const char32_t EGYPTIAN_HIEROGLYPH_138C5 = 0x000138C5;
const char32_t EGYPTIAN_HIEROGLYPH_138C6 = 0x000138C6;
const char32_t EGYPTIAN_HIEROGLYPH_138C7 = 0x000138C7;
const char32_t EGYPTIAN_HIEROGLYPH_138C8 = 0x000138C8;
const char32_t EGYPTIAN_HIEROGLYPH_138C9 = 0x000138C9;
const char32_t EGYPTIAN_HIEROGLYPH_138CA = 0x000138CA;
const char32_t EGYPTIAN_HIEROGLYPH_138CB = 0x000138CB;
const char32_t EGYPTIAN_HIEROGLYPH_138CC = 0x000138CC;
const char32_t EGYPTIAN_HIEROGLYPH_138CD = 0x000138CD;
const char32_t EGYPTIAN_HIEROGLYPH_138CE = 0x000138CE;
const char32_t EGYPTIAN_HIEROGLYPH_138CF = 0x000138CF;
const char32_t EGYPTIAN_HIEROGLYPH_138D0 = 0x000138D0;
const char32_t EGYPTIAN_HIEROGLYPH_138D1 = 0x000138D1;
const char32_t EGYPTIAN_HIEROGLYPH_138D2 = 0x000138D2;
const char32_t EGYPTIAN_HIEROGLYPH_138D3 = 0x000138D3;
const char32_t EGYPTIAN_HIEROGLYPH_138D4 = 0x000138D4;
const char32_t EGYPTIAN_HIEROGLYPH_138D5 = 0x000138D5;
const char32_t EGYPTIAN_HIEROGLYPH_138D6 = 0x000138D6;
const char32_t EGYPTIAN_HIEROGLYPH_138D7 = 0x000138D7;
const char32_t EGYPTIAN_HIEROGLYPH_138D8 = 0x000138D8;
const char32_t EGYPTIAN_HIEROGLYPH_138D9 = 0x000138D9;
const char32_t EGYPTIAN_HIEROGLYPH_138DA = 0x000138DA;
const char32_t EGYPTIAN_HIEROGLYPH_138DB = 0x000138DB;
const char32_t EGYPTIAN_HIEROGLYPH_138DC = 0x000138DC;
const char32_t EGYPTIAN_HIEROGLYPH_138DD = 0x000138DD;
const char32_t EGYPTIAN_HIEROGLYPH_138DE = 0x000138DE;
const char32_t EGYPTIAN_HIEROGLYPH_138DF = 0x000138DF;
const char32_t EGYPTIAN_HIEROGLYPH_138E0 = 0x000138E0;
const char32_t EGYPTIAN_HIEROGLYPH_138E1 = 0x000138E1;
const char32_t EGYPTIAN_HIEROGLYPH_138E2 = 0x000138E2;
const char32_t EGYPTIAN_HIEROGLYPH_138E3 = 0x000138E3;
const char32_t EGYPTIAN_HIEROGLYPH_138E4 = 0x000138E4;
const char32_t EGYPTIAN_HIEROGLYPH_138E5 = 0x000138E5;
const char32_t EGYPTIAN_HIEROGLYPH_138E6 = 0x000138E6;
const char32_t EGYPTIAN_HIEROGLYPH_138E7 = 0x000138E7;
const char32_t EGYPTIAN_HIEROGLYPH_138E8 = 0x000138E8;
const char32_t EGYPTIAN_HIEROGLYPH_138E9 = 0x000138E9;
const char32_t EGYPTIAN_HIEROGLYPH_138EA = 0x000138EA;
const char32_t EGYPTIAN_HIEROGLYPH_138EB = 0x000138EB;
const char32_t EGYPTIAN_HIEROGLYPH_138EC = 0x000138EC;
const char32_t EGYPTIAN_HIEROGLYPH_138ED = 0x000138ED;
const char32_t EGYPTIAN_HIEROGLYPH_138EE = 0x000138EE;
const char32_t EGYPTIAN_HIEROGLYPH_138EF = 0x000138EF;
const char32_t EGYPTIAN_HIEROGLYPH_138F0 = 0x000138F0;
const char32_t EGYPTIAN_HIEROGLYPH_138F1 = 0x000138F1;
const char32_t EGYPTIAN_HIEROGLYPH_138F2 = 0x000138F2;
const char32_t EGYPTIAN_HIEROGLYPH_138F3 = 0x000138F3;
const char32_t EGYPTIAN_HIEROGLYPH_138F4 = 0x000138F4;
const char32_t EGYPTIAN_HIEROGLYPH_138F5 = 0x000138F5;
const char32_t EGYPTIAN_HIEROGLYPH_138F6 = 0x000138F6;
const char32_t EGYPTIAN_HIEROGLYPH_138F7 = 0x000138F7;
const char32_t EGYPTIAN_HIEROGLYPH_138F8 = 0x000138F8;
const char32_t EGYPTIAN_HIEROGLYPH_138F9 = 0x000138F9;
const char32_t EGYPTIAN_HIEROGLYPH_138FA = 0x000138FA;
const char32_t EGYPTIAN_HIEROGLYPH_138FB = 0x000138FB;
const char32_t EGYPTIAN_HIEROGLYPH_138FC = 0x000138FC;
const char32_t EGYPTIAN_HIEROGLYPH_138FD = 0x000138FD;
const char32_t EGYPTIAN_HIEROGLYPH_138FE = 0x000138FE;
const char32_t EGYPTIAN_HIEROGLYPH_138FF = 0x000138FF;
const char32_t EGYPTIAN_HIEROGLYPH_13900 = 0x00013900;
const char32_t EGYPTIAN_HIEROGLYPH_13901 = 0x00013901;
const char32_t EGYPTIAN_HIEROGLYPH_13902 = 0x00013902;
const char32_t EGYPTIAN_HIEROGLYPH_13903 = 0x00013903;
const char32_t EGYPTIAN_HIEROGLYPH_13904 = 0x00013904;
const char32_t EGYPTIAN_HIEROGLYPH_13905 = 0x00013905;
const char32_t EGYPTIAN_HIEROGLYPH_13906 = 0x00013906;
const char32_t EGYPTIAN_HIEROGLYPH_13907 = 0x00013907;
const char32_t EGYPTIAN_HIEROGLYPH_13908 = 0x00013908;
const char32_t EGYPTIAN_HIEROGLYPH_13909 = 0x00013909;
const char32_t EGYPTIAN_HIEROGLYPH_1390A = 0x0001390A;
const char32_t EGYPTIAN_HIEROGLYPH_1390B = 0x0001390B;
const char32_t EGYPTIAN_HIEROGLYPH_1390C = 0x0001390C;
const char32_t EGYPTIAN_HIEROGLYPH_1390D = 0x0001390D;
const char32_t EGYPTIAN_HIEROGLYPH_1390E = 0x0001390E;
const char32_t EGYPTIAN_HIEROGLYPH_1390F = 0x0001390F;
const char32_t EGYPTIAN_HIEROGLYPH_13910 = 0x00013910;
const char32_t EGYPTIAN_HIEROGLYPH_13911 = 0x00013911;
const char32_t EGYPTIAN_HIEROGLYPH_13912 = 0x00013912;
const char32_t EGYPTIAN_HIEROGLYPH_13913 = 0x00013913;
const char32_t EGYPTIAN_HIEROGLYPH_13914 = 0x00013914;
const char32_t EGYPTIAN_HIEROGLYPH_13915 = 0x00013915;
const char32_t EGYPTIAN_HIEROGLYPH_13916 = 0x00013916;
const char32_t EGYPTIAN_HIEROGLYPH_13917 = 0x00013917;
const char32_t EGYPTIAN_HIEROGLYPH_13918 = 0x00013918;
const char32_t EGYPTIAN_HIEROGLYPH_13919 = 0x00013919;
const char32_t EGYPTIAN_HIEROGLYPH_1391A = 0x0001391A;
const char32_t EGYPTIAN_HIEROGLYPH_1391B = 0x0001391B;
const char32_t EGYPTIAN_HIEROGLYPH_1391C = 0x0001391C;
const char32_t EGYPTIAN_HIEROGLYPH_1391D = 0x0001391D;
const char32_t EGYPTIAN_HIEROGLYPH_1391E = 0x0001391E;
const char32_t EGYPTIAN_HIEROGLYPH_1391F = 0x0001391F;
const char32_t EGYPTIAN_HIEROGLYPH_13920 = 0x00013920;
const char32_t EGYPTIAN_HIEROGLYPH_13921 = 0x00013921;
const char32_t EGYPTIAN_HIEROGLYPH_13922 = 0x00013922;
const char32_t EGYPTIAN_HIEROGLYPH_13923 = 0x00013923;
const char32_t EGYPTIAN_HIEROGLYPH_13924 = 0x00013924;
const char32_t EGYPTIAN_HIEROGLYPH_13925 = 0x00013925;
const char32_t EGYPTIAN_HIEROGLYPH_13926 = 0x00013926;
const char32_t EGYPTIAN_HIEROGLYPH_13927 = 0x00013927;
const char32_t EGYPTIAN_HIEROGLYPH_13928 = 0x00013928;
const char32_t EGYPTIAN_HIEROGLYPH_13929 = 0x00013929;
const char32_t EGYPTIAN_HIEROGLYPH_1392A = 0x0001392A;
const char32_t EGYPTIAN_HIEROGLYPH_1392B = 0x0001392B;
const char32_t EGYPTIAN_HIEROGLYPH_1392C = 0x0001392C;
const char32_t EGYPTIAN_HIEROGLYPH_1392D = 0x0001392D;
const char32_t EGYPTIAN_HIEROGLYPH_1392E = 0x0001392E;
const char32_t EGYPTIAN_HIEROGLYPH_1392F = 0x0001392F;
const char32_t EGYPTIAN_HIEROGLYPH_13930 = 0x00013930;
const char32_t EGYPTIAN_HIEROGLYPH_13931 = 0x00013931;
const char32_t EGYPTIAN_HIEROGLYPH_13932 = 0x00013932;
const char32_t EGYPTIAN_HIEROGLYPH_13933 = 0x00013933;
const char32_t EGYPTIAN_HIEROGLYPH_13934 = 0x00013934;
const char32_t EGYPTIAN_HIEROGLYPH_13935 = 0x00013935;
const char32_t EGYPTIAN_HIEROGLYPH_13936 = 0x00013936;
const char32_t EGYPTIAN_HIEROGLYPH_13937 = 0x00013937;
const char32_t EGYPTIAN_HIEROGLYPH_13938 = 0x00013938;
const char32_t EGYPTIAN_HIEROGLYPH_13939 = 0x00013939;
const char32_t EGYPTIAN_HIEROGLYPH_1393A = 0x0001393A;
const char32_t EGYPTIAN_HIEROGLYPH_1393B = 0x0001393B;
const char32_t EGYPTIAN_HIEROGLYPH_1393C = 0x0001393C;
const char32_t EGYPTIAN_HIEROGLYPH_1393D = 0x0001393D;
const char32_t EGYPTIAN_HIEROGLYPH_1393E = 0x0001393E;
const char32_t EGYPTIAN_HIEROGLYPH_1393F = 0x0001393F;
const char32_t EGYPTIAN_HIEROGLYPH_13940 = 0x00013940;
const char32_t EGYPTIAN_HIEROGLYPH_13941 = 0x00013941;
const char32_t EGYPTIAN_HIEROGLYPH_13942 = 0x00013942;
const char32_t EGYPTIAN_HIEROGLYPH_13943 = 0x00013943;
const char32_t EGYPTIAN_HIEROGLYPH_13944 = 0x00013944;
const char32_t EGYPTIAN_HIEROGLYPH_13945 = 0x00013945;
const char32_t EGYPTIAN_HIEROGLYPH_13946 = 0x00013946;
const char32_t EGYPTIAN_HIEROGLYPH_13947 = 0x00013947;
const char32_t EGYPTIAN_HIEROGLYPH_13948 = 0x00013948;
const char32_t EGYPTIAN_HIEROGLYPH_13949 = 0x00013949;
const char32_t EGYPTIAN_HIEROGLYPH_1394A = 0x0001394A;
const char32_t EGYPTIAN_HIEROGLYPH_1394B = 0x0001394B;
const char32_t EGYPTIAN_HIEROGLYPH_1394C = 0x0001394C;
const char32_t EGYPTIAN_HIEROGLYPH_1394D = 0x0001394D;
const char32_t EGYPTIAN_HIEROGLYPH_1394E = 0x0001394E;
const char32_t EGYPTIAN_HIEROGLYPH_1394F = 0x0001394F;
const char32_t EGYPTIAN_HIEROGLYPH_13950 = 0x00013950;
const char32_t EGYPTIAN_HIEROGLYPH_13951 = 0x00013951;
const char32_t EGYPTIAN_HIEROGLYPH_13952 = 0x00013952;
const char32_t EGYPTIAN_HIEROGLYPH_13953 = 0x00013953;
const char32_t EGYPTIAN_HIEROGLYPH_13954 = 0x00013954;
const char32_t EGYPTIAN_HIEROGLYPH_13955 = 0x00013955;
const char32_t EGYPTIAN_HIEROGLYPH_13956 = 0x00013956;
const char32_t EGYPTIAN_HIEROGLYPH_13957 = 0x00013957;
const char32_t EGYPTIAN_HIEROGLYPH_13958 = 0x00013958;
const char32_t EGYPTIAN_HIEROGLYPH_13959 = 0x00013959;
const char32_t EGYPTIAN_HIEROGLYPH_1395A = 0x0001395A;
const char32_t EGYPTIAN_HIEROGLYPH_1395B = 0x0001395B;
const char32_t EGYPTIAN_HIEROGLYPH_1395C = 0x0001395C;
const char32_t EGYPTIAN_HIEROGLYPH_1395D = 0x0001395D;
const char32_t EGYPTIAN_HIEROGLYPH_1395E = 0x0001395E;
const char32_t EGYPTIAN_HIEROGLYPH_1395F = 0x0001395F;
const char32_t EGYPTIAN_HIEROGLYPH_13960 = 0x00013960;
const char32_t EGYPTIAN_HIEROGLYPH_13961 = 0x00013961;
const char32_t EGYPTIAN_HIEROGLYPH_13962 = 0x00013962;
const char32_t EGYPTIAN_HIEROGLYPH_13963 = 0x00013963;
const char32_t EGYPTIAN_HIEROGLYPH_13964 = 0x00013964;
const char32_t EGYPTIAN_HIEROGLYPH_13965 = 0x00013965;
const char32_t EGYPTIAN_HIEROGLYPH_13966 = 0x00013966;
const char32_t EGYPTIAN_HIEROGLYPH_13967 = 0x00013967;
const char32_t EGYPTIAN_HIEROGLYPH_13968 = 0x00013968;
const char32_t EGYPTIAN_HIEROGLYPH_13969 = 0x00013969;
const char32_t EGYPTIAN_HIEROGLYPH_1396A = 0x0001396A;
const char32_t EGYPTIAN_HIEROGLYPH_1396B = 0x0001396B;
const char32_t EGYPTIAN_HIEROGLYPH_1396C = 0x0001396C;
const char32_t EGYPTIAN_HIEROGLYPH_1396D = 0x0001396D;
const char32_t EGYPTIAN_HIEROGLYPH_1396E = 0x0001396E;
const char32_t EGYPTIAN_HIEROGLYPH_1396F = 0x0001396F;
const char32_t EGYPTIAN_HIEROGLYPH_13970 = 0x00013970;
const char32_t EGYPTIAN_HIEROGLYPH_13971 = 0x00013971;
const char32_t EGYPTIAN_HIEROGLYPH_13972 = 0x00013972;
const char32_t EGYPTIAN_HIEROGLYPH_13973 = 0x00013973;
const char32_t EGYPTIAN_HIEROGLYPH_13974 = 0x00013974;
const char32_t EGYPTIAN_HIEROGLYPH_13975 = 0x00013975;
const char32_t EGYPTIAN_HIEROGLYPH_13976 = 0x00013976;
const char32_t EGYPTIAN_HIEROGLYPH_13977 = 0x00013977;
const char32_t EGYPTIAN_HIEROGLYPH_13978 = 0x00013978;
const char32_t EGYPTIAN_HIEROGLYPH_13979 = 0x00013979;
const char32_t EGYPTIAN_HIEROGLYPH_1397A = 0x0001397A;
const char32_t EGYPTIAN_HIEROGLYPH_1397B = 0x0001397B;
const char32_t EGYPTIAN_HIEROGLYPH_1397C = 0x0001397C;
const char32_t EGYPTIAN_HIEROGLYPH_1397D = 0x0001397D;
const char32_t EGYPTIAN_HIEROGLYPH_1397E = 0x0001397E;
const char32_t EGYPTIAN_HIEROGLYPH_1397F = 0x0001397F;
const char32_t EGYPTIAN_HIEROGLYPH_13980 = 0x00013980;
const char32_t EGYPTIAN_HIEROGLYPH_13981 = 0x00013981;
const char32_t EGYPTIAN_HIEROGLYPH_13982 = 0x00013982;
const char32_t EGYPTIAN_HIEROGLYPH_13983 = 0x00013983;
const char32_t EGYPTIAN_HIEROGLYPH_13984 = 0x00013984;
const char32_t EGYPTIAN_HIEROGLYPH_13985 = 0x00013985;
const char32_t EGYPTIAN_HIEROGLYPH_13986 = 0x00013986;
const char32_t EGYPTIAN_HIEROGLYPH_13987 = 0x00013987;
const char32_t EGYPTIAN_HIEROGLYPH_13988 = 0x00013988;
const char32_t EGYPTIAN_HIEROGLYPH_13989 = 0x00013989;
const char32_t EGYPTIAN_HIEROGLYPH_1398A = 0x0001398A;
const char32_t EGYPTIAN_HIEROGLYPH_1398B = 0x0001398B;
const char32_t EGYPTIAN_HIEROGLYPH_1398C = 0x0001398C;
const char32_t EGYPTIAN_HIEROGLYPH_1398D = 0x0001398D;
const char32_t EGYPTIAN_HIEROGLYPH_1398E = 0x0001398E;
const char32_t EGYPTIAN_HIEROGLYPH_1398F = 0x0001398F;
const char32_t EGYPTIAN_HIEROGLYPH_13990 = 0x00013990;
const char32_t EGYPTIAN_HIEROGLYPH_13991 = 0x00013991;
const char32_t EGYPTIAN_HIEROGLYPH_13992 = 0x00013992;
const char32_t EGYPTIAN_HIEROGLYPH_13993 = 0x00013993;
const char32_t EGYPTIAN_HIEROGLYPH_13994 = 0x00013994;
const char32_t EGYPTIAN_HIEROGLYPH_13995 = 0x00013995;
const char32_t EGYPTIAN_HIEROGLYPH_13996 = 0x00013996;
const char32_t EGYPTIAN_HIEROGLYPH_13997 = 0x00013997;
const char32_t EGYPTIAN_HIEROGLYPH_13998 = 0x00013998;
const char32_t EGYPTIAN_HIEROGLYPH_13999 = 0x00013999;
const char32_t EGYPTIAN_HIEROGLYPH_1399A = 0x0001399A;
const char32_t EGYPTIAN_HIEROGLYPH_1399B = 0x0001399B;
const char32_t EGYPTIAN_HIEROGLYPH_1399C = 0x0001399C;
const char32_t EGYPTIAN_HIEROGLYPH_1399D = 0x0001399D;
const char32_t EGYPTIAN_HIEROGLYPH_1399E = 0x0001399E;
const char32_t EGYPTIAN_HIEROGLYPH_1399F = 0x0001399F;
const char32_t EGYPTIAN_HIEROGLYPH_139A0 = 0x000139A0;
const char32_t EGYPTIAN_HIEROGLYPH_139A1 = 0x000139A1;
const char32_t EGYPTIAN_HIEROGLYPH_139A2 = 0x000139A2;
const char32_t EGYPTIAN_HIEROGLYPH_139A3 = 0x000139A3;
const char32_t EGYPTIAN_HIEROGLYPH_139A4 = 0x000139A4;
const char32_t EGYPTIAN_HIEROGLYPH_139A5 = 0x000139A5;
const char32_t EGYPTIAN_HIEROGLYPH_139A6 = 0x000139A6;
const char32_t EGYPTIAN_HIEROGLYPH_139A7 = 0x000139A7;
const char32_t EGYPTIAN_HIEROGLYPH_139A8 = 0x000139A8;
const char32_t EGYPTIAN_HIEROGLYPH_139A9 = 0x000139A9;
const char32_t EGYPTIAN_HIEROGLYPH_139AA = 0x000139AA;
const char32_t EGYPTIAN_HIEROGLYPH_139AB = 0x000139AB;
const char32_t EGYPTIAN_HIEROGLYPH_139AC = 0x000139AC;
const char32_t EGYPTIAN_HIEROGLYPH_139AD = 0x000139AD;
const char32_t EGYPTIAN_HIEROGLYPH_139AE = 0x000139AE;
const char32_t EGYPTIAN_HIEROGLYPH_139AF = 0x000139AF;
const char32_t EGYPTIAN_HIEROGLYPH_139B0 = 0x000139B0;
const char32_t EGYPTIAN_HIEROGLYPH_139B1 = 0x000139B1;
const char32_t EGYPTIAN_HIEROGLYPH_139B2 = 0x000139B2;
const char32_t EGYPTIAN_HIEROGLYPH_139B3 = 0x000139B3;
const char32_t EGYPTIAN_HIEROGLYPH_139B4 = 0x000139B4;
const char32_t EGYPTIAN_HIEROGLYPH_139B5 = 0x000139B5;
const char32_t EGYPTIAN_HIEROGLYPH_139B6 = 0x000139B6;
const char32_t EGYPTIAN_HIEROGLYPH_139B7 = 0x000139B7;
const char32_t EGYPTIAN_HIEROGLYPH_139B8 = 0x000139B8;
const char32_t EGYPTIAN_HIEROGLYPH_139B9 = 0x000139B9;
const char32_t EGYPTIAN_HIEROGLYPH_139BA = 0x000139BA;
const char32_t EGYPTIAN_HIEROGLYPH_139BB = 0x000139BB;
const char32_t EGYPTIAN_HIEROGLYPH_139BC = 0x000139BC;
const char32_t EGYPTIAN_HIEROGLYPH_139BD = 0x000139BD;
const char32_t EGYPTIAN_HIEROGLYPH_139BE = 0x000139BE;
const char32_t EGYPTIAN_HIEROGLYPH_139BF = 0x000139BF;
const char32_t EGYPTIAN_HIEROGLYPH_139C0 = 0x000139C0;
const char32_t EGYPTIAN_HIEROGLYPH_139C1 = 0x000139C1;
const char32_t EGYPTIAN_HIEROGLYPH_139C2 = 0x000139C2;
const char32_t EGYPTIAN_HIEROGLYPH_139C3 = 0x000139C3;
const char32_t EGYPTIAN_HIEROGLYPH_139C4 = 0x000139C4;
const char32_t EGYPTIAN_HIEROGLYPH_139C5 = 0x000139C5;
const char32_t EGYPTIAN_HIEROGLYPH_139C6 = 0x000139C6;
const char32_t EGYPTIAN_HIEROGLYPH_139C7 = 0x000139C7;
const char32_t EGYPTIAN_HIEROGLYPH_139C8 = 0x000139C8;
const char32_t EGYPTIAN_HIEROGLYPH_139C9 = 0x000139C9;
const char32_t EGYPTIAN_HIEROGLYPH_139CA = 0x000139CA;
const char32_t EGYPTIAN_HIEROGLYPH_139CB = 0x000139CB;
const char32_t EGYPTIAN_HIEROGLYPH_139CC = 0x000139CC;
const char32_t EGYPTIAN_HIEROGLYPH_139CD = 0x000139CD;
const char32_t EGYPTIAN_HIEROGLYPH_139CE = 0x000139CE;
const char32_t EGYPTIAN_HIEROGLYPH_139CF = 0x000139CF;
const char32_t EGYPTIAN_HIEROGLYPH_139D0 = 0x000139D0;
const char32_t EGYPTIAN_HIEROGLYPH_139D1 = 0x000139D1;
const char32_t EGYPTIAN_HIEROGLYPH_139D2 = 0x000139D2;
const char32_t EGYPTIAN_HIEROGLYPH_139D3 = 0x000139D3;
const char32_t EGYPTIAN_HIEROGLYPH_139D4 = 0x000139D4;
const char32_t EGYPTIAN_HIEROGLYPH_139D5 = 0x000139D5;
const char32_t EGYPTIAN_HIEROGLYPH_139D6 = 0x000139D6;
const char32_t EGYPTIAN_HIEROGLYPH_139D7 = 0x000139D7;
const char32_t EGYPTIAN_HIEROGLYPH_139D8 = 0x000139D8;
const char32_t EGYPTIAN_HIEROGLYPH_139D9 = 0x000139D9;
const char32_t EGYPTIAN_HIEROGLYPH_139DA = 0x000139DA;
const char32_t EGYPTIAN_HIEROGLYPH_139DB = 0x000139DB;
const char32_t EGYPTIAN_HIEROGLYPH_139DC = 0x000139DC;
const char32_t EGYPTIAN_HIEROGLYPH_139DD = 0x000139DD;
const char32_t EGYPTIAN_HIEROGLYPH_139DE = 0x000139DE;
const char32_t EGYPTIAN_HIEROGLYPH_139DF = 0x000139DF;
const char32_t EGYPTIAN_HIEROGLYPH_139E0 = 0x000139E0;
const char32_t EGYPTIAN_HIEROGLYPH_139E1 = 0x000139E1;
const char32_t EGYPTIAN_HIEROGLYPH_139E2 = 0x000139E2;
const char32_t EGYPTIAN_HIEROGLYPH_139E3 = 0x000139E3;
const char32_t EGYPTIAN_HIEROGLYPH_139E4 = 0x000139E4;
const char32_t EGYPTIAN_HIEROGLYPH_139E5 = 0x000139E5;
const char32_t EGYPTIAN_HIEROGLYPH_139E6 = 0x000139E6;
const char32_t EGYPTIAN_HIEROGLYPH_139E7 = 0x000139E7;
const char32_t EGYPTIAN_HIEROGLYPH_139E8 = 0x000139E8;
const char32_t EGYPTIAN_HIEROGLYPH_139E9 = 0x000139E9;
const char32_t EGYPTIAN_HIEROGLYPH_139EA = 0x000139EA;
const char32_t EGYPTIAN_HIEROGLYPH_139EB = 0x000139EB;
const char32_t EGYPTIAN_HIEROGLYPH_139EC = 0x000139EC;
const char32_t EGYPTIAN_HIEROGLYPH_139ED = 0x000139ED;
const char32_t EGYPTIAN_HIEROGLYPH_139EE = 0x000139EE;
const char32_t EGYPTIAN_HIEROGLYPH_139EF = 0x000139EF;
const char32_t EGYPTIAN_HIEROGLYPH_139F0 = 0x000139F0;
const char32_t EGYPTIAN_HIEROGLYPH_139F1 = 0x000139F1;
const char32_t EGYPTIAN_HIEROGLYPH_139F2 = 0x000139F2;
const char32_t EGYPTIAN_HIEROGLYPH_139F3 = 0x000139F3;
const char32_t EGYPTIAN_HIEROGLYPH_139F4 = 0x000139F4;
const char32_t EGYPTIAN_HIEROGLYPH_139F5 = 0x000139F5;
const char32_t EGYPTIAN_HIEROGLYPH_139F6 = 0x000139F6;
const char32_t EGYPTIAN_HIEROGLYPH_139F7 = 0x000139F7;
const char32_t EGYPTIAN_HIEROGLYPH_139F8 = 0x000139F8;
const char32_t EGYPTIAN_HIEROGLYPH_139F9 = 0x000139F9;
const char32_t EGYPTIAN_HIEROGLYPH_139FA = 0x000139FA;
const char32_t EGYPTIAN_HIEROGLYPH_139FB = 0x000139FB;
const char32_t EGYPTIAN_HIEROGLYPH_139FC = 0x000139FC;
const char32_t EGYPTIAN_HIEROGLYPH_139FD = 0x000139FD;
const char32_t EGYPTIAN_HIEROGLYPH_139FE = 0x000139FE;
const char32_t EGYPTIAN_HIEROGLYPH_139FF = 0x000139FF;
const char32_t EGYPTIAN_HIEROGLYPH_13A00 = 0x00013A00;
const char32_t EGYPTIAN_HIEROGLYPH_13A01 = 0x00013A01;
const char32_t EGYPTIAN_HIEROGLYPH_13A02 = 0x00013A02;
const char32_t EGYPTIAN_HIEROGLYPH_13A03 = 0x00013A03;
const char32_t EGYPTIAN_HIEROGLYPH_13A04 = 0x00013A04;
const char32_t EGYPTIAN_HIEROGLYPH_13A05 = 0x00013A05;
const char32_t EGYPTIAN_HIEROGLYPH_13A06 = 0x00013A06;
const char32_t EGYPTIAN_HIEROGLYPH_13A07 = 0x00013A07;
const char32_t EGYPTIAN_HIEROGLYPH_13A08 = 0x00013A08;
const char32_t EGYPTIAN_HIEROGLYPH_13A09 = 0x00013A09;
const char32_t EGYPTIAN_HIEROGLYPH_13A0A = 0x00013A0A;
const char32_t EGYPTIAN_HIEROGLYPH_13A0B = 0x00013A0B;
const char32_t EGYPTIAN_HIEROGLYPH_13A0C = 0x00013A0C;
const char32_t EGYPTIAN_HIEROGLYPH_13A0D = 0x00013A0D;
const char32_t EGYPTIAN_HIEROGLYPH_13A0E = 0x00013A0E;
const char32_t EGYPTIAN_HIEROGLYPH_13A0F = 0x00013A0F;
const char32_t EGYPTIAN_HIEROGLYPH_13A10 = 0x00013A10;
const char32_t EGYPTIAN_HIEROGLYPH_13A11 = 0x00013A11;
const char32_t EGYPTIAN_HIEROGLYPH_13A12 = 0x00013A12;
const char32_t EGYPTIAN_HIEROGLYPH_13A13 = 0x00013A13;
const char32_t EGYPTIAN_HIEROGLYPH_13A14 = 0x00013A14;
const char32_t EGYPTIAN_HIEROGLYPH_13A15 = 0x00013A15;
const char32_t EGYPTIAN_HIEROGLYPH_13A16 = 0x00013A16;
const char32_t EGYPTIAN_HIEROGLYPH_13A17 = 0x00013A17;
const char32_t EGYPTIAN_HIEROGLYPH_13A18 = 0x00013A18;
const char32_t EGYPTIAN_HIEROGLYPH_13A19 = 0x00013A19;
const char32_t EGYPTIAN_HIEROGLYPH_13A1A = 0x00013A1A;
const char32_t EGYPTIAN_HIEROGLYPH_13A1B = 0x00013A1B;
const char32_t EGYPTIAN_HIEROGLYPH_13A1C = 0x00013A1C;
const char32_t EGYPTIAN_HIEROGLYPH_13A1D = 0x00013A1D;
const char32_t EGYPTIAN_HIEROGLYPH_13A1E = 0x00013A1E;
const char32_t EGYPTIAN_HIEROGLYPH_13A1F = 0x00013A1F;
const char32_t EGYPTIAN_HIEROGLYPH_13A20 = 0x00013A20;
const char32_t EGYPTIAN_HIEROGLYPH_13A21 = 0x00013A21;
const char32_t EGYPTIAN_HIEROGLYPH_13A22 = 0x00013A22;
const char32_t EGYPTIAN_HIEROGLYPH_13A23 = 0x00013A23;
const char32_t EGYPTIAN_HIEROGLYPH_13A24 = 0x00013A24;
const char32_t EGYPTIAN_HIEROGLYPH_13A25 = 0x00013A25;
const char32_t EGYPTIAN_HIEROGLYPH_13A26 = 0x00013A26;
const char32_t EGYPTIAN_HIEROGLYPH_13A27 = 0x00013A27;
const char32_t EGYPTIAN_HIEROGLYPH_13A28 = 0x00013A28;
const char32_t EGYPTIAN_HIEROGLYPH_13A29 = 0x00013A29;
const char32_t EGYPTIAN_HIEROGLYPH_13A2A = 0x00013A2A;
const char32_t EGYPTIAN_HIEROGLYPH_13A2B = 0x00013A2B;
const char32_t EGYPTIAN_HIEROGLYPH_13A2C = 0x00013A2C;
const char32_t EGYPTIAN_HIEROGLYPH_13A2D = 0x00013A2D;
const char32_t EGYPTIAN_HIEROGLYPH_13A2E = 0x00013A2E;
const char32_t EGYPTIAN_HIEROGLYPH_13A2F = 0x00013A2F;
const char32_t EGYPTIAN_HIEROGLYPH_13A30 = 0x00013A30;
const char32_t EGYPTIAN_HIEROGLYPH_13A31 = 0x00013A31;
const char32_t EGYPTIAN_HIEROGLYPH_13A32 = 0x00013A32;
const char32_t EGYPTIAN_HIEROGLYPH_13A33 = 0x00013A33;
const char32_t EGYPTIAN_HIEROGLYPH_13A34 = 0x00013A34;
const char32_t EGYPTIAN_HIEROGLYPH_13A35 = 0x00013A35;
const char32_t EGYPTIAN_HIEROGLYPH_13A36 = 0x00013A36;
const char32_t EGYPTIAN_HIEROGLYPH_13A37 = 0x00013A37;
const char32_t EGYPTIAN_HIEROGLYPH_13A38 = 0x00013A38;
const char32_t EGYPTIAN_HIEROGLYPH_13A39 = 0x00013A39;
const char32_t EGYPTIAN_HIEROGLYPH_13A3A = 0x00013A3A;
const char32_t EGYPTIAN_HIEROGLYPH_13A3B = 0x00013A3B;
const char32_t EGYPTIAN_HIEROGLYPH_13A3C = 0x00013A3C;
const char32_t EGYPTIAN_HIEROGLYPH_13A3D = 0x00013A3D;
const char32_t EGYPTIAN_HIEROGLYPH_13A3E = 0x00013A3E;
const char32_t EGYPTIAN_HIEROGLYPH_13A3F = 0x00013A3F;
const char32_t EGYPTIAN_HIEROGLYPH_13A40 = 0x00013A40;
const char32_t EGYPTIAN_HIEROGLYPH_13A41 = 0x00013A41;
const char32_t EGYPTIAN_HIEROGLYPH_13A42 = 0x00013A42;
const char32_t EGYPTIAN_HIEROGLYPH_13A43 = 0x00013A43;
const char32_t EGYPTIAN_HIEROGLYPH_13A44 = 0x00013A44;
const char32_t EGYPTIAN_HIEROGLYPH_13A45 = 0x00013A45;
const char32_t EGYPTIAN_HIEROGLYPH_13A46 = 0x00013A46;
const char32_t EGYPTIAN_HIEROGLYPH_13A47 = 0x00013A47;
const char32_t EGYPTIAN_HIEROGLYPH_13A48 = 0x00013A48;
const char32_t EGYPTIAN_HIEROGLYPH_13A49 = 0x00013A49;
const char32_t EGYPTIAN_HIEROGLYPH_13A4A = 0x00013A4A;
const char32_t EGYPTIAN_HIEROGLYPH_13A4B = 0x00013A4B;
const char32_t EGYPTIAN_HIEROGLYPH_13A4C = 0x00013A4C;
const char32_t EGYPTIAN_HIEROGLYPH_13A4D = 0x00013A4D;
const char32_t EGYPTIAN_HIEROGLYPH_13A4E = 0x00013A4E;
const char32_t EGYPTIAN_HIEROGLYPH_13A4F = 0x00013A4F;
const char32_t EGYPTIAN_HIEROGLYPH_13A50 = 0x00013A50;
const char32_t EGYPTIAN_HIEROGLYPH_13A51 = 0x00013A51;
const char32_t EGYPTIAN_HIEROGLYPH_13A52 = 0x00013A52;
const char32_t EGYPTIAN_HIEROGLYPH_13A53 = 0x00013A53;
const char32_t EGYPTIAN_HIEROGLYPH_13A54 = 0x00013A54;
const char32_t EGYPTIAN_HIEROGLYPH_13A55 = 0x00013A55;
const char32_t EGYPTIAN_HIEROGLYPH_13A56 = 0x00013A56;
const char32_t EGYPTIAN_HIEROGLYPH_13A57 = 0x00013A57;
const char32_t EGYPTIAN_HIEROGLYPH_13A58 = 0x00013A58;
const char32_t EGYPTIAN_HIEROGLYPH_13A59 = 0x00013A59;
const char32_t EGYPTIAN_HIEROGLYPH_13A5A = 0x00013A5A;
const char32_t EGYPTIAN_HIEROGLYPH_13A5B = 0x00013A5B;
const char32_t EGYPTIAN_HIEROGLYPH_13A5C = 0x00013A5C;
const char32_t EGYPTIAN_HIEROGLYPH_13A5D = 0x00013A5D;
const char32_t EGYPTIAN_HIEROGLYPH_13A5E = 0x00013A5E;
const char32_t EGYPTIAN_HIEROGLYPH_13A5F = 0x00013A5F;
const char32_t EGYPTIAN_HIEROGLYPH_13A60 = 0x00013A60;
const char32_t EGYPTIAN_HIEROGLYPH_13A61 = 0x00013A61;
const char32_t EGYPTIAN_HIEROGLYPH_13A62 = 0x00013A62;
const char32_t EGYPTIAN_HIEROGLYPH_13A63 = 0x00013A63;
const char32_t EGYPTIAN_HIEROGLYPH_13A64 = 0x00013A64;
const char32_t EGYPTIAN_HIEROGLYPH_13A65 = 0x00013A65;
const char32_t EGYPTIAN_HIEROGLYPH_13A66 = 0x00013A66;
const char32_t EGYPTIAN_HIEROGLYPH_13A67 = 0x00013A67;
const char32_t EGYPTIAN_HIEROGLYPH_13A68 = 0x00013A68;
const char32_t EGYPTIAN_HIEROGLYPH_13A69 = 0x00013A69;
const char32_t EGYPTIAN_HIEROGLYPH_13A6A = 0x00013A6A;
const char32_t EGYPTIAN_HIEROGLYPH_13A6B = 0x00013A6B;
const char32_t EGYPTIAN_HIEROGLYPH_13A6C = 0x00013A6C;
const char32_t EGYPTIAN_HIEROGLYPH_13A6D = 0x00013A6D;
const char32_t EGYPTIAN_HIEROGLYPH_13A6E = 0x00013A6E;
const char32_t EGYPTIAN_HIEROGLYPH_13A6F = 0x00013A6F;
const char32_t EGYPTIAN_HIEROGLYPH_13A70 = 0x00013A70;
const char32_t EGYPTIAN_HIEROGLYPH_13A71 = 0x00013A71;
const char32_t EGYPTIAN_HIEROGLYPH_13A72 = 0x00013A72;
const char32_t EGYPTIAN_HIEROGLYPH_13A73 = 0x00013A73;
const char32_t EGYPTIAN_HIEROGLYPH_13A74 = 0x00013A74;
const char32_t EGYPTIAN_HIEROGLYPH_13A75 = 0x00013A75;
const char32_t EGYPTIAN_HIEROGLYPH_13A76 = 0x00013A76;
const char32_t EGYPTIAN_HIEROGLYPH_13A77 = 0x00013A77;
const char32_t EGYPTIAN_HIEROGLYPH_13A78 = 0x00013A78;
const char32_t EGYPTIAN_HIEROGLYPH_13A79 = 0x00013A79;
const char32_t EGYPTIAN_HIEROGLYPH_13A7A = 0x00013A7A;
const char32_t EGYPTIAN_HIEROGLYPH_13A7B = 0x00013A7B;
const char32_t EGYPTIAN_HIEROGLYPH_13A7C = 0x00013A7C;
const char32_t EGYPTIAN_HIEROGLYPH_13A7D = 0x00013A7D;
const char32_t EGYPTIAN_HIEROGLYPH_13A7E = 0x00013A7E;
const char32_t EGYPTIAN_HIEROGLYPH_13A7F = 0x00013A7F;
const char32_t EGYPTIAN_HIEROGLYPH_13A80 = 0x00013A80;
const char32_t EGYPTIAN_HIEROGLYPH_13A81 = 0x00013A81;
const char32_t EGYPTIAN_HIEROGLYPH_13A82 = 0x00013A82;
const char32_t EGYPTIAN_HIEROGLYPH_13A83 = 0x00013A83;
const char32_t EGYPTIAN_HIEROGLYPH_13A84 = 0x00013A84;
const char32_t EGYPTIAN_HIEROGLYPH_13A85 = 0x00013A85;
const char32_t EGYPTIAN_HIEROGLYPH_13A86 = 0x00013A86;
const char32_t EGYPTIAN_HIEROGLYPH_13A87 = 0x00013A87;
const char32_t EGYPTIAN_HIEROGLYPH_13A88 = 0x00013A88;
const char32_t EGYPTIAN_HIEROGLYPH_13A89 = 0x00013A89;
const char32_t EGYPTIAN_HIEROGLYPH_13A8A = 0x00013A8A;
const char32_t EGYPTIAN_HIEROGLYPH_13A8B = 0x00013A8B;
const char32_t EGYPTIAN_HIEROGLYPH_13A8C = 0x00013A8C;
const char32_t EGYPTIAN_HIEROGLYPH_13A8D = 0x00013A8D;
const char32_t EGYPTIAN_HIEROGLYPH_13A8E = 0x00013A8E;
const char32_t EGYPTIAN_HIEROGLYPH_13A8F = 0x00013A8F;
const char32_t EGYPTIAN_HIEROGLYPH_13A90 = 0x00013A90;
const char32_t EGYPTIAN_HIEROGLYPH_13A91 = 0x00013A91;
const char32_t EGYPTIAN_HIEROGLYPH_13A92 = 0x00013A92;
const char32_t EGYPTIAN_HIEROGLYPH_13A93 = 0x00013A93;
const char32_t EGYPTIAN_HIEROGLYPH_13A94 = 0x00013A94;
const char32_t EGYPTIAN_HIEROGLYPH_13A95 = 0x00013A95;
const char32_t EGYPTIAN_HIEROGLYPH_13A96 = 0x00013A96;
const char32_t EGYPTIAN_HIEROGLYPH_13A97 = 0x00013A97;
const char32_t EGYPTIAN_HIEROGLYPH_13A98 = 0x00013A98;
const char32_t EGYPTIAN_HIEROGLYPH_13A99 = 0x00013A99;
const char32_t EGYPTIAN_HIEROGLYPH_13A9A = 0x00013A9A;
const char32_t EGYPTIAN_HIEROGLYPH_13A9B = 0x00013A9B;
const char32_t EGYPTIAN_HIEROGLYPH_13A9C = 0x00013A9C;
const char32_t EGYPTIAN_HIEROGLYPH_13A9D = 0x00013A9D;
const char32_t EGYPTIAN_HIEROGLYPH_13A9E = 0x00013A9E;
const char32_t EGYPTIAN_HIEROGLYPH_13A9F = 0x00013A9F;
const char32_t EGYPTIAN_HIEROGLYPH_13AA0 = 0x00013AA0;
const char32_t EGYPTIAN_HIEROGLYPH_13AA1 = 0x00013AA1;
const char32_t EGYPTIAN_HIEROGLYPH_13AA2 = 0x00013AA2;
const char32_t EGYPTIAN_HIEROGLYPH_13AA3 = 0x00013AA3;
const char32_t EGYPTIAN_HIEROGLYPH_13AA4 = 0x00013AA4;
const char32_t EGYPTIAN_HIEROGLYPH_13AA5 = 0x00013AA5;
const char32_t EGYPTIAN_HIEROGLYPH_13AA6 = 0x00013AA6;
const char32_t EGYPTIAN_HIEROGLYPH_13AA7 = 0x00013AA7;
const char32_t EGYPTIAN_HIEROGLYPH_13AA8 = 0x00013AA8;
const char32_t EGYPTIAN_HIEROGLYPH_13AA9 = 0x00013AA9;
const char32_t EGYPTIAN_HIEROGLYPH_13AAA = 0x00013AAA;
const char32_t EGYPTIAN_HIEROGLYPH_13AAB = 0x00013AAB;
const char32_t EGYPTIAN_HIEROGLYPH_13AAC = 0x00013AAC;
const char32_t EGYPTIAN_HIEROGLYPH_13AAD = 0x00013AAD;
const char32_t EGYPTIAN_HIEROGLYPH_13AAE = 0x00013AAE;
const char32_t EGYPTIAN_HIEROGLYPH_13AAF = 0x00013AAF;
const char32_t EGYPTIAN_HIEROGLYPH_13AB0 = 0x00013AB0;
const char32_t EGYPTIAN_HIEROGLYPH_13AB1 = 0x00013AB1;
const char32_t EGYPTIAN_HIEROGLYPH_13AB2 = 0x00013AB2;
const char32_t EGYPTIAN_HIEROGLYPH_13AB3 = 0x00013AB3;
const char32_t EGYPTIAN_HIEROGLYPH_13AB4 = 0x00013AB4;
const char32_t EGYPTIAN_HIEROGLYPH_13AB5 = 0x00013AB5;
const char32_t EGYPTIAN_HIEROGLYPH_13AB6 = 0x00013AB6;
const char32_t EGYPTIAN_HIEROGLYPH_13AB7 = 0x00013AB7;
const char32_t EGYPTIAN_HIEROGLYPH_13AB8 = 0x00013AB8;
const char32_t EGYPTIAN_HIEROGLYPH_13AB9 = 0x00013AB9;
const char32_t EGYPTIAN_HIEROGLYPH_13ABA = 0x00013ABA;
const char32_t EGYPTIAN_HIEROGLYPH_13ABB = 0x00013ABB;
const char32_t EGYPTIAN_HIEROGLYPH_13ABC = 0x00013ABC;
const char32_t EGYPTIAN_HIEROGLYPH_13ABD = 0x00013ABD;
const char32_t EGYPTIAN_HIEROGLYPH_13ABE = 0x00013ABE;
const char32_t EGYPTIAN_HIEROGLYPH_13ABF = 0x00013ABF;
const char32_t EGYPTIAN_HIEROGLYPH_13AC0 = 0x00013AC0;
const char32_t EGYPTIAN_HIEROGLYPH_13AC1 = 0x00013AC1;
const char32_t EGYPTIAN_HIEROGLYPH_13AC2 = 0x00013AC2;
const char32_t EGYPTIAN_HIEROGLYPH_13AC3 = 0x00013AC3;
const char32_t EGYPTIAN_HIEROGLYPH_13AC4 = 0x00013AC4;
const char32_t EGYPTIAN_HIEROGLYPH_13AC5 = 0x00013AC5;
const char32_t EGYPTIAN_HIEROGLYPH_13AC6 = 0x00013AC6;
const char32_t EGYPTIAN_HIEROGLYPH_13AC7 = 0x00013AC7;
const char32_t EGYPTIAN_HIEROGLYPH_13AC8 = 0x00013AC8;
const char32_t EGYPTIAN_HIEROGLYPH_13AC9 = 0x00013AC9;
const char32_t EGYPTIAN_HIEROGLYPH_13ACA = 0x00013ACA;
const char32_t EGYPTIAN_HIEROGLYPH_13ACB = 0x00013ACB;
const char32_t EGYPTIAN_HIEROGLYPH_13ACC = 0x00013ACC;
const char32_t EGYPTIAN_HIEROGLYPH_13ACD = 0x00013ACD;
const char32_t EGYPTIAN_HIEROGLYPH_13ACE = 0x00013ACE;
const char32_t EGYPTIAN_HIEROGLYPH_13ACF = 0x00013ACF;
const char32_t EGYPTIAN_HIEROGLYPH_13AD0 = 0x00013AD0;
const char32_t EGYPTIAN_HIEROGLYPH_13AD1 = 0x00013AD1;
const char32_t EGYPTIAN_HIEROGLYPH_13AD2 = 0x00013AD2;
const char32_t EGYPTIAN_HIEROGLYPH_13AD3 = 0x00013AD3;
const char32_t EGYPTIAN_HIEROGLYPH_13AD4 = 0x00013AD4;
const char32_t EGYPTIAN_HIEROGLYPH_13AD5 = 0x00013AD5;
const char32_t EGYPTIAN_HIEROGLYPH_13AD6 = 0x00013AD6;
const char32_t EGYPTIAN_HIEROGLYPH_13AD7 = 0x00013AD7;
const char32_t EGYPTIAN_HIEROGLYPH_13AD8 = 0x00013AD8;
const char32_t EGYPTIAN_HIEROGLYPH_13AD9 = 0x00013AD9;
const char32_t EGYPTIAN_HIEROGLYPH_13ADA = 0x00013ADA;
const char32_t EGYPTIAN_HIEROGLYPH_13ADB = 0x00013ADB;
const char32_t EGYPTIAN_HIEROGLYPH_13ADC = 0x00013ADC;
const char32_t EGYPTIAN_HIEROGLYPH_13ADD = 0x00013ADD;
const char32_t EGYPTIAN_HIEROGLYPH_13ADE = 0x00013ADE;
const char32_t EGYPTIAN_HIEROGLYPH_13ADF = 0x00013ADF;
const char32_t EGYPTIAN_HIEROGLYPH_13AE0 = 0x00013AE0;
const char32_t EGYPTIAN_HIEROGLYPH_13AE1 = 0x00013AE1;
const char32_t EGYPTIAN_HIEROGLYPH_13AE2 = 0x00013AE2;
const char32_t EGYPTIAN_HIEROGLYPH_13AE3 = 0x00013AE3;
const char32_t EGYPTIAN_HIEROGLYPH_13AE4 = 0x00013AE4;
const char32_t EGYPTIAN_HIEROGLYPH_13AE5 = 0x00013AE5;
const char32_t EGYPTIAN_HIEROGLYPH_13AE6 = 0x00013AE6;
const char32_t EGYPTIAN_HIEROGLYPH_13AE7 = 0x00013AE7;
const char32_t EGYPTIAN_HIEROGLYPH_13AE8 = 0x00013AE8;
const char32_t EGYPTIAN_HIEROGLYPH_13AE9 = 0x00013AE9;
const char32_t EGYPTIAN_HIEROGLYPH_13AEA = 0x00013AEA;
const char32_t EGYPTIAN_HIEROGLYPH_13AEB = 0x00013AEB;
const char32_t EGYPTIAN_HIEROGLYPH_13AEC = 0x00013AEC;
const char32_t EGYPTIAN_HIEROGLYPH_13AED = 0x00013AED;
const char32_t EGYPTIAN_HIEROGLYPH_13AEE = 0x00013AEE;
const char32_t EGYPTIAN_HIEROGLYPH_13AEF = 0x00013AEF;
const char32_t EGYPTIAN_HIEROGLYPH_13AF0 = 0x00013AF0;
const char32_t EGYPTIAN_HIEROGLYPH_13AF1 = 0x00013AF1;
const char32_t EGYPTIAN_HIEROGLYPH_13AF2 = 0x00013AF2;
const char32_t EGYPTIAN_HIEROGLYPH_13AF3 = 0x00013AF3;
const char32_t EGYPTIAN_HIEROGLYPH_13AF4 = 0x00013AF4;
const char32_t EGYPTIAN_HIEROGLYPH_13AF5 = 0x00013AF5;
const char32_t EGYPTIAN_HIEROGLYPH_13AF6 = 0x00013AF6;
const char32_t EGYPTIAN_HIEROGLYPH_13AF7 = 0x00013AF7;
const char32_t EGYPTIAN_HIEROGLYPH_13AF8 = 0x00013AF8;
const char32_t EGYPTIAN_HIEROGLYPH_13AF9 = 0x00013AF9;
const char32_t EGYPTIAN_HIEROGLYPH_13AFA = 0x00013AFA;
const char32_t EGYPTIAN_HIEROGLYPH_13AFB = 0x00013AFB;
const char32_t EGYPTIAN_HIEROGLYPH_13AFC = 0x00013AFC;
const char32_t EGYPTIAN_HIEROGLYPH_13AFD = 0x00013AFD;
const char32_t EGYPTIAN_HIEROGLYPH_13AFE = 0x00013AFE;
const char32_t EGYPTIAN_HIEROGLYPH_13AFF = 0x00013AFF;
const char32_t EGYPTIAN_HIEROGLYPH_13B00 = 0x00013B00;
const char32_t EGYPTIAN_HIEROGLYPH_13B01 = 0x00013B01;
const char32_t EGYPTIAN_HIEROGLYPH_13B02 = 0x00013B02;
const char32_t EGYPTIAN_HIEROGLYPH_13B03 = 0x00013B03;
const char32_t EGYPTIAN_HIEROGLYPH_13B04 = 0x00013B04;
const char32_t EGYPTIAN_HIEROGLYPH_13B05 = 0x00013B05;
const char32_t EGYPTIAN_HIEROGLYPH_13B06 = 0x00013B06;
const char32_t EGYPTIAN_HIEROGLYPH_13B07 = 0x00013B07;
const char32_t EGYPTIAN_HIEROGLYPH_13B08 = 0x00013B08;
const char32_t EGYPTIAN_HIEROGLYPH_13B09 = 0x00013B09;
const char32_t EGYPTIAN_HIEROGLYPH_13B0A = 0x00013B0A;
const char32_t EGYPTIAN_HIEROGLYPH_13B0B = 0x00013B0B;
const char32_t EGYPTIAN_HIEROGLYPH_13B0C = 0x00013B0C;
const char32_t EGYPTIAN_HIEROGLYPH_13B0D = 0x00013B0D;
const char32_t EGYPTIAN_HIEROGLYPH_13B0E = 0x00013B0E;
const char32_t EGYPTIAN_HIEROGLYPH_13B0F = 0x00013B0F;
const char32_t EGYPTIAN_HIEROGLYPH_13B10 = 0x00013B10;
const char32_t EGYPTIAN_HIEROGLYPH_13B11 = 0x00013B11;
const char32_t EGYPTIAN_HIEROGLYPH_13B12 = 0x00013B12;
const char32_t EGYPTIAN_HIEROGLYPH_13B13 = 0x00013B13;
const char32_t EGYPTIAN_HIEROGLYPH_13B14 = 0x00013B14;
const char32_t EGYPTIAN_HIEROGLYPH_13B15 = 0x00013B15;
const char32_t EGYPTIAN_HIEROGLYPH_13B16 = 0x00013B16;
const char32_t EGYPTIAN_HIEROGLYPH_13B17 = 0x00013B17;
const char32_t EGYPTIAN_HIEROGLYPH_13B18 = 0x00013B18;
const char32_t EGYPTIAN_HIEROGLYPH_13B19 = 0x00013B19;
const char32_t EGYPTIAN_HIEROGLYPH_13B1A = 0x00013B1A;
const char32_t EGYPTIAN_HIEROGLYPH_13B1B = 0x00013B1B;
const char32_t EGYPTIAN_HIEROGLYPH_13B1C = 0x00013B1C;
const char32_t EGYPTIAN_HIEROGLYPH_13B1D = 0x00013B1D;
const char32_t EGYPTIAN_HIEROGLYPH_13B1E = 0x00013B1E;
const char32_t EGYPTIAN_HIEROGLYPH_13B1F = 0x00013B1F;
const char32_t EGYPTIAN_HIEROGLYPH_13B20 = 0x00013B20;
const char32_t EGYPTIAN_HIEROGLYPH_13B21 = 0x00013B21;
const char32_t EGYPTIAN_HIEROGLYPH_13B22 = 0x00013B22;
const char32_t EGYPTIAN_HIEROGLYPH_13B23 = 0x00013B23;
const char32_t EGYPTIAN_HIEROGLYPH_13B24 = 0x00013B24;
const char32_t EGYPTIAN_HIEROGLYPH_13B25 = 0x00013B25;
const char32_t EGYPTIAN_HIEROGLYPH_13B26 = 0x00013B26;
const char32_t EGYPTIAN_HIEROGLYPH_13B27 = 0x00013B27;
const char32_t EGYPTIAN_HIEROGLYPH_13B28 = 0x00013B28;
const char32_t EGYPTIAN_HIEROGLYPH_13B29 = 0x00013B29;
const char32_t EGYPTIAN_HIEROGLYPH_13B2A = 0x00013B2A;
const char32_t EGYPTIAN_HIEROGLYPH_13B2B = 0x00013B2B;
const char32_t EGYPTIAN_HIEROGLYPH_13B2C = 0x00013B2C;
const char32_t EGYPTIAN_HIEROGLYPH_13B2D = 0x00013B2D;
const char32_t EGYPTIAN_HIEROGLYPH_13B2E = 0x00013B2E;
const char32_t EGYPTIAN_HIEROGLYPH_13B2F = 0x00013B2F;
const char32_t EGYPTIAN_HIEROGLYPH_13B30 = 0x00013B30;
const char32_t EGYPTIAN_HIEROGLYPH_13B31 = 0x00013B31;
const char32_t EGYPTIAN_HIEROGLYPH_13B32 = 0x00013B32;
const char32_t EGYPTIAN_HIEROGLYPH_13B33 = 0x00013B33;
const char32_t EGYPTIAN_HIEROGLYPH_13B34 = 0x00013B34;
const char32_t EGYPTIAN_HIEROGLYPH_13B35 = 0x00013B35;
const char32_t EGYPTIAN_HIEROGLYPH_13B36 = 0x00013B36;
const char32_t EGYPTIAN_HIEROGLYPH_13B37 = 0x00013B37;
const char32_t EGYPTIAN_HIEROGLYPH_13B38 = 0x00013B38;
const char32_t EGYPTIAN_HIEROGLYPH_13B39 = 0x00013B39;
const char32_t EGYPTIAN_HIEROGLYPH_13B3A = 0x00013B3A;
const char32_t EGYPTIAN_HIEROGLYPH_13B3B = 0x00013B3B;
const char32_t EGYPTIAN_HIEROGLYPH_13B3C = 0x00013B3C;
const char32_t EGYPTIAN_HIEROGLYPH_13B3D = 0x00013B3D;
const char32_t EGYPTIAN_HIEROGLYPH_13B3E = 0x00013B3E;
const char32_t EGYPTIAN_HIEROGLYPH_13B3F = 0x00013B3F;
const char32_t EGYPTIAN_HIEROGLYPH_13B40 = 0x00013B40;
const char32_t EGYPTIAN_HIEROGLYPH_13B41 = 0x00013B41;
const char32_t EGYPTIAN_HIEROGLYPH_13B42 = 0x00013B42;
const char32_t EGYPTIAN_HIEROGLYPH_13B43 = 0x00013B43;
const char32_t EGYPTIAN_HIEROGLYPH_13B44 = 0x00013B44;
const char32_t EGYPTIAN_HIEROGLYPH_13B45 = 0x00013B45;
const char32_t EGYPTIAN_HIEROGLYPH_13B46 = 0x00013B46;
const char32_t EGYPTIAN_HIEROGLYPH_13B47 = 0x00013B47;
const char32_t EGYPTIAN_HIEROGLYPH_13B48 = 0x00013B48;
const char32_t EGYPTIAN_HIEROGLYPH_13B49 = 0x00013B49;
const char32_t EGYPTIAN_HIEROGLYPH_13B4A = 0x00013B4A;
const char32_t EGYPTIAN_HIEROGLYPH_13B4B = 0x00013B4B;
const char32_t EGYPTIAN_HIEROGLYPH_13B4C = 0x00013B4C;
const char32_t EGYPTIAN_HIEROGLYPH_13B4D = 0x00013B4D;
const char32_t EGYPTIAN_HIEROGLYPH_13B4E = 0x00013B4E;
const char32_t EGYPTIAN_HIEROGLYPH_13B4F = 0x00013B4F;
const char32_t EGYPTIAN_HIEROGLYPH_13B50 = 0x00013B50;
const char32_t EGYPTIAN_HIEROGLYPH_13B51 = 0x00013B51;
const char32_t EGYPTIAN_HIEROGLYPH_13B52 = 0x00013B52;
const char32_t EGYPTIAN_HIEROGLYPH_13B53 = 0x00013B53;
const char32_t EGYPTIAN_HIEROGLYPH_13B54 = 0x00013B54;
const char32_t EGYPTIAN_HIEROGLYPH_13B55 = 0x00013B55;
const char32_t EGYPTIAN_HIEROGLYPH_13B56 = 0x00013B56;
const char32_t EGYPTIAN_HIEROGLYPH_13B57 = 0x00013B57;
const char32_t EGYPTIAN_HIEROGLYPH_13B58 = 0x00013B58;
const char32_t EGYPTIAN_HIEROGLYPH_13B59 = 0x00013B59;
const char32_t EGYPTIAN_HIEROGLYPH_13B5A = 0x00013B5A;
const char32_t EGYPTIAN_HIEROGLYPH_13B5B = 0x00013B5B;
const char32_t EGYPTIAN_HIEROGLYPH_13B5C = 0x00013B5C;
const char32_t EGYPTIAN_HIEROGLYPH_13B5D = 0x00013B5D;
const char32_t EGYPTIAN_HIEROGLYPH_13B5E = 0x00013B5E;
const char32_t EGYPTIAN_HIEROGLYPH_13B5F = 0x00013B5F;
const char32_t EGYPTIAN_HIEROGLYPH_13B60 = 0x00013B60;
const char32_t EGYPTIAN_HIEROGLYPH_13B61 = 0x00013B61;
const char32_t EGYPTIAN_HIEROGLYPH_13B62 = 0x00013B62;
const char32_t EGYPTIAN_HIEROGLYPH_13B63 = 0x00013B63;
const char32_t EGYPTIAN_HIEROGLYPH_13B64 = 0x00013B64;
const char32_t EGYPTIAN_HIEROGLYPH_13B65 = 0x00013B65;
const char32_t EGYPTIAN_HIEROGLYPH_13B66 = 0x00013B66;
const char32_t EGYPTIAN_HIEROGLYPH_13B67 = 0x00013B67;
const char32_t EGYPTIAN_HIEROGLYPH_13B68 = 0x00013B68;
const char32_t EGYPTIAN_HIEROGLYPH_13B69 = 0x00013B69;
const char32_t EGYPTIAN_HIEROGLYPH_13B6A = 0x00013B6A;
const char32_t EGYPTIAN_HIEROGLYPH_13B6B = 0x00013B6B;
const char32_t EGYPTIAN_HIEROGLYPH_13B6C = 0x00013B6C;
const char32_t EGYPTIAN_HIEROGLYPH_13B6D = 0x00013B6D;
const char32_t EGYPTIAN_HIEROGLYPH_13B6E = 0x00013B6E;
const char32_t EGYPTIAN_HIEROGLYPH_13B6F = 0x00013B6F;
const char32_t EGYPTIAN_HIEROGLYPH_13B70 = 0x00013B70;
const char32_t EGYPTIAN_HIEROGLYPH_13B71 = 0x00013B71;
const char32_t EGYPTIAN_HIEROGLYPH_13B72 = 0x00013B72;
const char32_t EGYPTIAN_HIEROGLYPH_13B73 = 0x00013B73;
const char32_t EGYPTIAN_HIEROGLYPH_13B74 = 0x00013B74;
const char32_t EGYPTIAN_HIEROGLYPH_13B75 = 0x00013B75;
const char32_t EGYPTIAN_HIEROGLYPH_13B76 = 0x00013B76;
const char32_t EGYPTIAN_HIEROGLYPH_13B77 = 0x00013B77;
const char32_t EGYPTIAN_HIEROGLYPH_13B78 = 0x00013B78;
const char32_t EGYPTIAN_HIEROGLYPH_13B79 = 0x00013B79;
const char32_t EGYPTIAN_HIEROGLYPH_13B7A = 0x00013B7A;
const char32_t EGYPTIAN_HIEROGLYPH_13B7B = 0x00013B7B;
const char32_t EGYPTIAN_HIEROGLYPH_13B7C = 0x00013B7C;
const char32_t EGYPTIAN_HIEROGLYPH_13B7D = 0x00013B7D;
const char32_t EGYPTIAN_HIEROGLYPH_13B7E = 0x00013B7E;
const char32_t EGYPTIAN_HIEROGLYPH_13B7F = 0x00013B7F;
const char32_t EGYPTIAN_HIEROGLYPH_13B80 = 0x00013B80;
const char32_t EGYPTIAN_HIEROGLYPH_13B81 = 0x00013B81;
const char32_t EGYPTIAN_HIEROGLYPH_13B82 = 0x00013B82;
const char32_t EGYPTIAN_HIEROGLYPH_13B83 = 0x00013B83;
const char32_t EGYPTIAN_HIEROGLYPH_13B84 = 0x00013B84;
const char32_t EGYPTIAN_HIEROGLYPH_13B85 = 0x00013B85;
const char32_t EGYPTIAN_HIEROGLYPH_13B86 = 0x00013B86;
const char32_t EGYPTIAN_HIEROGLYPH_13B87 = 0x00013B87;
const char32_t EGYPTIAN_HIEROGLYPH_13B88 = 0x00013B88;
const char32_t EGYPTIAN_HIEROGLYPH_13B89 = 0x00013B89;
const char32_t EGYPTIAN_HIEROGLYPH_13B8A = 0x00013B8A;
const char32_t EGYPTIAN_HIEROGLYPH_13B8B = 0x00013B8B;
const char32_t EGYPTIAN_HIEROGLYPH_13B8C = 0x00013B8C;
const char32_t EGYPTIAN_HIEROGLYPH_13B8D = 0x00013B8D;
const char32_t EGYPTIAN_HIEROGLYPH_13B8E = 0x00013B8E;
const char32_t EGYPTIAN_HIEROGLYPH_13B8F = 0x00013B8F;
const char32_t EGYPTIAN_HIEROGLYPH_13B90 = 0x00013B90;
const char32_t EGYPTIAN_HIEROGLYPH_13B91 = 0x00013B91;
const char32_t EGYPTIAN_HIEROGLYPH_13B92 = 0x00013B92;
const char32_t EGYPTIAN_HIEROGLYPH_13B93 = 0x00013B93;
const char32_t EGYPTIAN_HIEROGLYPH_13B94 = 0x00013B94;
const char32_t EGYPTIAN_HIEROGLYPH_13B95 = 0x00013B95;
const char32_t EGYPTIAN_HIEROGLYPH_13B96 = 0x00013B96;
const char32_t EGYPTIAN_HIEROGLYPH_13B97 = 0x00013B97;
const char32_t EGYPTIAN_HIEROGLYPH_13B98 = 0x00013B98;
const char32_t EGYPTIAN_HIEROGLYPH_13B99 = 0x00013B99;
const char32_t EGYPTIAN_HIEROGLYPH_13B9A = 0x00013B9A;
const char32_t EGYPTIAN_HIEROGLYPH_13B9B = 0x00013B9B;
const char32_t EGYPTIAN_HIEROGLYPH_13B9C = 0x00013B9C;
const char32_t EGYPTIAN_HIEROGLYPH_13B9D = 0x00013B9D;
const char32_t EGYPTIAN_HIEROGLYPH_13B9E = 0x00013B9E;
const char32_t EGYPTIAN_HIEROGLYPH_13B9F = 0x00013B9F;
const char32_t EGYPTIAN_HIEROGLYPH_13BA0 = 0x00013BA0;
const char32_t EGYPTIAN_HIEROGLYPH_13BA1 = 0x00013BA1;
const char32_t EGYPTIAN_HIEROGLYPH_13BA2 = 0x00013BA2;
const char32_t EGYPTIAN_HIEROGLYPH_13BA3 = 0x00013BA3;
const char32_t EGYPTIAN_HIEROGLYPH_13BA4 = 0x00013BA4;
const char32_t EGYPTIAN_HIEROGLYPH_13BA5 = 0x00013BA5;
const char32_t EGYPTIAN_HIEROGLYPH_13BA6 = 0x00013BA6;
const char32_t EGYPTIAN_HIEROGLYPH_13BA7 = 0x00013BA7;
const char32_t EGYPTIAN_HIEROGLYPH_13BA8 = 0x00013BA8;
const char32_t EGYPTIAN_HIEROGLYPH_13BA9 = 0x00013BA9;
const char32_t EGYPTIAN_HIEROGLYPH_13BAA = 0x00013BAA;
const char32_t EGYPTIAN_HIEROGLYPH_13BAB = 0x00013BAB;
const char32_t EGYPTIAN_HIEROGLYPH_13BAC = 0x00013BAC;
const char32_t EGYPTIAN_HIEROGLYPH_13BAD = 0x00013BAD;
const char32_t EGYPTIAN_HIEROGLYPH_13BAE = 0x00013BAE;
const char32_t EGYPTIAN_HIEROGLYPH_13BAF = 0x00013BAF;
const char32_t EGYPTIAN_HIEROGLYPH_13BB0 = 0x00013BB0;
const char32_t EGYPTIAN_HIEROGLYPH_13BB1 = 0x00013BB1;
const char32_t EGYPTIAN_HIEROGLYPH_13BB2 = 0x00013BB2;
const char32_t EGYPTIAN_HIEROGLYPH_13BB3 = 0x00013BB3;
const char32_t EGYPTIAN_HIEROGLYPH_13BB4 = 0x00013BB4;
const char32_t EGYPTIAN_HIEROGLYPH_13BB5 = 0x00013BB5;
const char32_t EGYPTIAN_HIEROGLYPH_13BB6 = 0x00013BB6;
const char32_t EGYPTIAN_HIEROGLYPH_13BB7 = 0x00013BB7;
const char32_t EGYPTIAN_HIEROGLYPH_13BB8 = 0x00013BB8;
const char32_t EGYPTIAN_HIEROGLYPH_13BB9 = 0x00013BB9;
const char32_t EGYPTIAN_HIEROGLYPH_13BBA = 0x00013BBA;
const char32_t EGYPTIAN_HIEROGLYPH_13BBB = 0x00013BBB;
const char32_t EGYPTIAN_HIEROGLYPH_13BBC = 0x00013BBC;
const char32_t EGYPTIAN_HIEROGLYPH_13BBD = 0x00013BBD;
const char32_t EGYPTIAN_HIEROGLYPH_13BBE = 0x00013BBE;
const char32_t EGYPTIAN_HIEROGLYPH_13BBF = 0x00013BBF;
const char32_t EGYPTIAN_HIEROGLYPH_13BC0 = 0x00013BC0;
const char32_t EGYPTIAN_HIEROGLYPH_13BC1 = 0x00013BC1;
const char32_t EGYPTIAN_HIEROGLYPH_13BC2 = 0x00013BC2;
const char32_t EGYPTIAN_HIEROGLYPH_13BC3 = 0x00013BC3;
const char32_t EGYPTIAN_HIEROGLYPH_13BC4 = 0x00013BC4;
const char32_t EGYPTIAN_HIEROGLYPH_13BC5 = 0x00013BC5;
const char32_t EGYPTIAN_HIEROGLYPH_13BC6 = 0x00013BC6;
const char32_t EGYPTIAN_HIEROGLYPH_13BC7 = 0x00013BC7;
const char32_t EGYPTIAN_HIEROGLYPH_13BC8 = 0x00013BC8;
const char32_t EGYPTIAN_HIEROGLYPH_13BC9 = 0x00013BC9;
const char32_t EGYPTIAN_HIEROGLYPH_13BCA = 0x00013BCA;
const char32_t EGYPTIAN_HIEROGLYPH_13BCB = 0x00013BCB;
const char32_t EGYPTIAN_HIEROGLYPH_13BCC = 0x00013BCC;
const char32_t EGYPTIAN_HIEROGLYPH_13BCD = 0x00013BCD;
const char32_t EGYPTIAN_HIEROGLYPH_13BCE = 0x00013BCE;
const char32_t EGYPTIAN_HIEROGLYPH_13BCF = 0x00013BCF;
const char32_t EGYPTIAN_HIEROGLYPH_13BD0 = 0x00013BD0;
const char32_t EGYPTIAN_HIEROGLYPH_13BD1 = 0x00013BD1;
const char32_t EGYPTIAN_HIEROGLYPH_13BD2 = 0x00013BD2;
const char32_t EGYPTIAN_HIEROGLYPH_13BD3 = 0x00013BD3;
const char32_t EGYPTIAN_HIEROGLYPH_13BD4 = 0x00013BD4;
const char32_t EGYPTIAN_HIEROGLYPH_13BD5 = 0x00013BD5;
const char32_t EGYPTIAN_HIEROGLYPH_13BD6 = 0x00013BD6;
const char32_t EGYPTIAN_HIEROGLYPH_13BD7 = 0x00013BD7;
const char32_t EGYPTIAN_HIEROGLYPH_13BD8 = 0x00013BD8;
const char32_t EGYPTIAN_HIEROGLYPH_13BD9 = 0x00013BD9;
const char32_t EGYPTIAN_HIEROGLYPH_13BDA = 0x00013BDA;
const char32_t EGYPTIAN_HIEROGLYPH_13BDB = 0x00013BDB;
const char32_t EGYPTIAN_HIEROGLYPH_13BDC = 0x00013BDC;
const char32_t EGYPTIAN_HIEROGLYPH_13BDD = 0x00013BDD;
const char32_t EGYPTIAN_HIEROGLYPH_13BDE = 0x00013BDE;
const char32_t EGYPTIAN_HIEROGLYPH_13BDF = 0x00013BDF;
const char32_t EGYPTIAN_HIEROGLYPH_13BE0 = 0x00013BE0;
const char32_t EGYPTIAN_HIEROGLYPH_13BE1 = 0x00013BE1;
const char32_t EGYPTIAN_HIEROGLYPH_13BE2 = 0x00013BE2;
const char32_t EGYPTIAN_HIEROGLYPH_13BE3 = 0x00013BE3;
const char32_t EGYPTIAN_HIEROGLYPH_13BE4 = 0x00013BE4;
const char32_t EGYPTIAN_HIEROGLYPH_13BE5 = 0x00013BE5;
const char32_t EGYPTIAN_HIEROGLYPH_13BE6 = 0x00013BE6;
const char32_t EGYPTIAN_HIEROGLYPH_13BE7 = 0x00013BE7;
const char32_t EGYPTIAN_HIEROGLYPH_13BE8 = 0x00013BE8;
const char32_t EGYPTIAN_HIEROGLYPH_13BE9 = 0x00013BE9;
const char32_t EGYPTIAN_HIEROGLYPH_13BEA = 0x00013BEA;
const char32_t EGYPTIAN_HIEROGLYPH_13BEB = 0x00013BEB;
const char32_t EGYPTIAN_HIEROGLYPH_13BEC = 0x00013BEC;
const char32_t EGYPTIAN_HIEROGLYPH_13BED = 0x00013BED;
const char32_t EGYPTIAN_HIEROGLYPH_13BEE = 0x00013BEE;
const char32_t EGYPTIAN_HIEROGLYPH_13BEF = 0x00013BEF;
const char32_t EGYPTIAN_HIEROGLYPH_13BF0 = 0x00013BF0;
const char32_t EGYPTIAN_HIEROGLYPH_13BF1 = 0x00013BF1;
const char32_t EGYPTIAN_HIEROGLYPH_13BF2 = 0x00013BF2;
const char32_t EGYPTIAN_HIEROGLYPH_13BF3 = 0x00013BF3;
const char32_t EGYPTIAN_HIEROGLYPH_13BF4 = 0x00013BF4;
const char32_t EGYPTIAN_HIEROGLYPH_13BF5 = 0x00013BF5;
const char32_t EGYPTIAN_HIEROGLYPH_13BF6 = 0x00013BF6;
const char32_t EGYPTIAN_HIEROGLYPH_13BF7 = 0x00013BF7;
const char32_t EGYPTIAN_HIEROGLYPH_13BF8 = 0x00013BF8;
const char32_t EGYPTIAN_HIEROGLYPH_13BF9 = 0x00013BF9;
const char32_t EGYPTIAN_HIEROGLYPH_13BFA = 0x00013BFA;
const char32_t EGYPTIAN_HIEROGLYPH_13BFB = 0x00013BFB;
const char32_t EGYPTIAN_HIEROGLYPH_13BFC = 0x00013BFC;
const char32_t EGYPTIAN_HIEROGLYPH_13BFD = 0x00013BFD;
const char32_t EGYPTIAN_HIEROGLYPH_13BFE = 0x00013BFE;
const char32_t EGYPTIAN_HIEROGLYPH_13BFF = 0x00013BFF;
const char32_t EGYPTIAN_HIEROGLYPH_13C00 = 0x00013C00;
const char32_t EGYPTIAN_HIEROGLYPH_13C01 = 0x00013C01;
const char32_t EGYPTIAN_HIEROGLYPH_13C02 = 0x00013C02;
const char32_t EGYPTIAN_HIEROGLYPH_13C03 = 0x00013C03;
const char32_t EGYPTIAN_HIEROGLYPH_13C04 = 0x00013C04;
const char32_t EGYPTIAN_HIEROGLYPH_13C05 = 0x00013C05;
const char32_t EGYPTIAN_HIEROGLYPH_13C06 = 0x00013C06;
const char32_t EGYPTIAN_HIEROGLYPH_13C07 = 0x00013C07;
const char32_t EGYPTIAN_HIEROGLYPH_13C08 = 0x00013C08;
const char32_t EGYPTIAN_HIEROGLYPH_13C09 = 0x00013C09;
const char32_t EGYPTIAN_HIEROGLYPH_13C0A = 0x00013C0A;
const char32_t EGYPTIAN_HIEROGLYPH_13C0B = 0x00013C0B;
const char32_t EGYPTIAN_HIEROGLYPH_13C0C = 0x00013C0C;
const char32_t EGYPTIAN_HIEROGLYPH_13C0D = 0x00013C0D;
const char32_t EGYPTIAN_HIEROGLYPH_13C0E = 0x00013C0E;
const char32_t EGYPTIAN_HIEROGLYPH_13C0F = 0x00013C0F;
const char32_t EGYPTIAN_HIEROGLYPH_13C10 = 0x00013C10;
const char32_t EGYPTIAN_HIEROGLYPH_13C11 = 0x00013C11;
const char32_t EGYPTIAN_HIEROGLYPH_13C12 = 0x00013C12;
const char32_t EGYPTIAN_HIEROGLYPH_13C13 = 0x00013C13;
const char32_t EGYPTIAN_HIEROGLYPH_13C14 = 0x00013C14;
const char32_t EGYPTIAN_HIEROGLYPH_13C15 = 0x00013C15;
const char32_t EGYPTIAN_HIEROGLYPH_13C16 = 0x00013C16;
const char32_t EGYPTIAN_HIEROGLYPH_13C17 = 0x00013C17;
const char32_t EGYPTIAN_HIEROGLYPH_13C18 = 0x00013C18;
const char32_t EGYPTIAN_HIEROGLYPH_13C19 = 0x00013C19;
const char32_t EGYPTIAN_HIEROGLYPH_13C1A = 0x00013C1A;
const char32_t EGYPTIAN_HIEROGLYPH_13C1B = 0x00013C1B;
const char32_t EGYPTIAN_HIEROGLYPH_13C1C = 0x00013C1C;
const char32_t EGYPTIAN_HIEROGLYPH_13C1D = 0x00013C1D;
const char32_t EGYPTIAN_HIEROGLYPH_13C1E = 0x00013C1E;
const char32_t EGYPTIAN_HIEROGLYPH_13C1F = 0x00013C1F;
const char32_t EGYPTIAN_HIEROGLYPH_13C20 = 0x00013C20;
const char32_t EGYPTIAN_HIEROGLYPH_13C21 = 0x00013C21;
const char32_t EGYPTIAN_HIEROGLYPH_13C22 = 0x00013C22;
const char32_t EGYPTIAN_HIEROGLYPH_13C23 = 0x00013C23;
const char32_t EGYPTIAN_HIEROGLYPH_13C24 = 0x00013C24;
const char32_t EGYPTIAN_HIEROGLYPH_13C25 = 0x00013C25;
const char32_t EGYPTIAN_HIEROGLYPH_13C26 = 0x00013C26;
const char32_t EGYPTIAN_HIEROGLYPH_13C27 = 0x00013C27;
const char32_t EGYPTIAN_HIEROGLYPH_13C28 = 0x00013C28;
const char32_t EGYPTIAN_HIEROGLYPH_13C29 = 0x00013C29;
const char32_t EGYPTIAN_HIEROGLYPH_13C2A = 0x00013C2A;
const char32_t EGYPTIAN_HIEROGLYPH_13C2B = 0x00013C2B;
const char32_t EGYPTIAN_HIEROGLYPH_13C2C = 0x00013C2C;
const char32_t EGYPTIAN_HIEROGLYPH_13C2D = 0x00013C2D;
const char32_t EGYPTIAN_HIEROGLYPH_13C2E = 0x00013C2E;
const char32_t EGYPTIAN_HIEROGLYPH_13C2F = 0x00013C2F;
const char32_t EGYPTIAN_HIEROGLYPH_13C30 = 0x00013C30;
const char32_t EGYPTIAN_HIEROGLYPH_13C31 = 0x00013C31;
const char32_t EGYPTIAN_HIEROGLYPH_13C32 = 0x00013C32;
const char32_t EGYPTIAN_HIEROGLYPH_13C33 = 0x00013C33;
const char32_t EGYPTIAN_HIEROGLYPH_13C34 = 0x00013C34;
const char32_t EGYPTIAN_HIEROGLYPH_13C35 = 0x00013C35;
const char32_t EGYPTIAN_HIEROGLYPH_13C36 = 0x00013C36;
const char32_t EGYPTIAN_HIEROGLYPH_13C37 = 0x00013C37;
const char32_t EGYPTIAN_HIEROGLYPH_13C38 = 0x00013C38;
const char32_t EGYPTIAN_HIEROGLYPH_13C39 = 0x00013C39;
const char32_t EGYPTIAN_HIEROGLYPH_13C3A = 0x00013C3A;
const char32_t EGYPTIAN_HIEROGLYPH_13C3B = 0x00013C3B;
const char32_t EGYPTIAN_HIEROGLYPH_13C3C = 0x00013C3C;
const char32_t EGYPTIAN_HIEROGLYPH_13C3D = 0x00013C3D;
const char32_t EGYPTIAN_HIEROGLYPH_13C3E = 0x00013C3E;
const char32_t EGYPTIAN_HIEROGLYPH_13C3F = 0x00013C3F;
const char32_t EGYPTIAN_HIEROGLYPH_13C40 = 0x00013C40;
const char32_t EGYPTIAN_HIEROGLYPH_13C41 = 0x00013C41;
const char32_t EGYPTIAN_HIEROGLYPH_13C42 = 0x00013C42;
const char32_t EGYPTIAN_HIEROGLYPH_13C43 = 0x00013C43;
const char32_t EGYPTIAN_HIEROGLYPH_13C44 = 0x00013C44;
const char32_t EGYPTIAN_HIEROGLYPH_13C45 = 0x00013C45;
const char32_t EGYPTIAN_HIEROGLYPH_13C46 = 0x00013C46;
const char32_t EGYPTIAN_HIEROGLYPH_13C47 = 0x00013C47;
const char32_t EGYPTIAN_HIEROGLYPH_13C48 = 0x00013C48;
const char32_t EGYPTIAN_HIEROGLYPH_13C49 = 0x00013C49;
const char32_t EGYPTIAN_HIEROGLYPH_13C4A = 0x00013C4A;
const char32_t EGYPTIAN_HIEROGLYPH_13C4B = 0x00013C4B;
const char32_t EGYPTIAN_HIEROGLYPH_13C4C = 0x00013C4C;
const char32_t EGYPTIAN_HIEROGLYPH_13C4D = 0x00013C4D;
const char32_t EGYPTIAN_HIEROGLYPH_13C4E = 0x00013C4E;
const char32_t EGYPTIAN_HIEROGLYPH_13C4F = 0x00013C4F;
const char32_t EGYPTIAN_HIEROGLYPH_13C50 = 0x00013C50;
const char32_t EGYPTIAN_HIEROGLYPH_13C51 = 0x00013C51;
const char32_t EGYPTIAN_HIEROGLYPH_13C52 = 0x00013C52;
const char32_t EGYPTIAN_HIEROGLYPH_13C53 = 0x00013C53;
const char32_t EGYPTIAN_HIEROGLYPH_13C54 = 0x00013C54;
const char32_t EGYPTIAN_HIEROGLYPH_13C55 = 0x00013C55;
const char32_t EGYPTIAN_HIEROGLYPH_13C56 = 0x00013C56;
const char32_t EGYPTIAN_HIEROGLYPH_13C57 = 0x00013C57;
const char32_t EGYPTIAN_HIEROGLYPH_13C58 = 0x00013C58;
const char32_t EGYPTIAN_HIEROGLYPH_13C59 = 0x00013C59;
const char32_t EGYPTIAN_HIEROGLYPH_13C5A = 0x00013C5A;
const char32_t EGYPTIAN_HIEROGLYPH_13C5B = 0x00013C5B;
const char32_t EGYPTIAN_HIEROGLYPH_13C5C = 0x00013C5C;
const char32_t EGYPTIAN_HIEROGLYPH_13C5D = 0x00013C5D;
const char32_t EGYPTIAN_HIEROGLYPH_13C5E = 0x00013C5E;
const char32_t EGYPTIAN_HIEROGLYPH_13C5F = 0x00013C5F;
const char32_t EGYPTIAN_HIEROGLYPH_13C60 = 0x00013C60;
const char32_t EGYPTIAN_HIEROGLYPH_13C61 = 0x00013C61;
const char32_t EGYPTIAN_HIEROGLYPH_13C62 = 0x00013C62;
const char32_t EGYPTIAN_HIEROGLYPH_13C63 = 0x00013C63;
const char32_t EGYPTIAN_HIEROGLYPH_13C64 = 0x00013C64;
const char32_t EGYPTIAN_HIEROGLYPH_13C65 = 0x00013C65;
const char32_t EGYPTIAN_HIEROGLYPH_13C66 = 0x00013C66;
const char32_t EGYPTIAN_HIEROGLYPH_13C67 = 0x00013C67;
const char32_t EGYPTIAN_HIEROGLYPH_13C68 = 0x00013C68;
const char32_t EGYPTIAN_HIEROGLYPH_13C69 = 0x00013C69;
const char32_t EGYPTIAN_HIEROGLYPH_13C6A = 0x00013C6A;
const char32_t EGYPTIAN_HIEROGLYPH_13C6B = 0x00013C6B;
const char32_t EGYPTIAN_HIEROGLYPH_13C6C = 0x00013C6C;
const char32_t EGYPTIAN_HIEROGLYPH_13C6D = 0x00013C6D;
const char32_t EGYPTIAN_HIEROGLYPH_13C6E = 0x00013C6E;
const char32_t EGYPTIAN_HIEROGLYPH_13C6F = 0x00013C6F;
const char32_t EGYPTIAN_HIEROGLYPH_13C70 = 0x00013C70;
const char32_t EGYPTIAN_HIEROGLYPH_13C71 = 0x00013C71;
const char32_t EGYPTIAN_HIEROGLYPH_13C72 = 0x00013C72;
const char32_t EGYPTIAN_HIEROGLYPH_13C73 = 0x00013C73;
const char32_t EGYPTIAN_HIEROGLYPH_13C74 = 0x00013C74;
const char32_t EGYPTIAN_HIEROGLYPH_13C75 = 0x00013C75;
const char32_t EGYPTIAN_HIEROGLYPH_13C76 = 0x00013C76;
const char32_t EGYPTIAN_HIEROGLYPH_13C77 = 0x00013C77;
const char32_t EGYPTIAN_HIEROGLYPH_13C78 = 0x00013C78;
const char32_t EGYPTIAN_HIEROGLYPH_13C79 = 0x00013C79;
const char32_t EGYPTIAN_HIEROGLYPH_13C7A = 0x00013C7A;
const char32_t EGYPTIAN_HIEROGLYPH_13C7B = 0x00013C7B;
const char32_t EGYPTIAN_HIEROGLYPH_13C7C = 0x00013C7C;
const char32_t EGYPTIAN_HIEROGLYPH_13C7D = 0x00013C7D;
const char32_t EGYPTIAN_HIEROGLYPH_13C7E = 0x00013C7E;
const char32_t EGYPTIAN_HIEROGLYPH_13C7F = 0x00013C7F;
const char32_t EGYPTIAN_HIEROGLYPH_13C80 = 0x00013C80;
const char32_t EGYPTIAN_HIEROGLYPH_13C81 = 0x00013C81;
const char32_t EGYPTIAN_HIEROGLYPH_13C82 = 0x00013C82;
const char32_t EGYPTIAN_HIEROGLYPH_13C83 = 0x00013C83;
const char32_t EGYPTIAN_HIEROGLYPH_13C84 = 0x00013C84;
const char32_t EGYPTIAN_HIEROGLYPH_13C85 = 0x00013C85;
const char32_t EGYPTIAN_HIEROGLYPH_13C86 = 0x00013C86;
const char32_t EGYPTIAN_HIEROGLYPH_13C87 = 0x00013C87;
const char32_t EGYPTIAN_HIEROGLYPH_13C88 = 0x00013C88;
const char32_t EGYPTIAN_HIEROGLYPH_13C89 = 0x00013C89;
const char32_t EGYPTIAN_HIEROGLYPH_13C8A = 0x00013C8A;
const char32_t EGYPTIAN_HIEROGLYPH_13C8B = 0x00013C8B;
const char32_t EGYPTIAN_HIEROGLYPH_13C8C = 0x00013C8C;
const char32_t EGYPTIAN_HIEROGLYPH_13C8D = 0x00013C8D;
const char32_t EGYPTIAN_HIEROGLYPH_13C8E = 0x00013C8E;
const char32_t EGYPTIAN_HIEROGLYPH_13C8F = 0x00013C8F;
const char32_t EGYPTIAN_HIEROGLYPH_13C90 = 0x00013C90;
const char32_t EGYPTIAN_HIEROGLYPH_13C91 = 0x00013C91;
const char32_t EGYPTIAN_HIEROGLYPH_13C92 = 0x00013C92;
const char32_t EGYPTIAN_HIEROGLYPH_13C93 = 0x00013C93;
const char32_t EGYPTIAN_HIEROGLYPH_13C94 = 0x00013C94;
const char32_t EGYPTIAN_HIEROGLYPH_13C95 = 0x00013C95;
const char32_t EGYPTIAN_HIEROGLYPH_13C96 = 0x00013C96;
const char32_t EGYPTIAN_HIEROGLYPH_13C97 = 0x00013C97;
const char32_t EGYPTIAN_HIEROGLYPH_13C98 = 0x00013C98;
const char32_t EGYPTIAN_HIEROGLYPH_13C99 = 0x00013C99;
const char32_t EGYPTIAN_HIEROGLYPH_13C9A = 0x00013C9A;
const char32_t EGYPTIAN_HIEROGLYPH_13C9B = 0x00013C9B;
const char32_t EGYPTIAN_HIEROGLYPH_13C9C = 0x00013C9C;
const char32_t EGYPTIAN_HIEROGLYPH_13C9D = 0x00013C9D;
const char32_t EGYPTIAN_HIEROGLYPH_13C9E = 0x00013C9E;
const char32_t EGYPTIAN_HIEROGLYPH_13C9F = 0x00013C9F;
const char32_t EGYPTIAN_HIEROGLYPH_13CA0 = 0x00013CA0;
const char32_t EGYPTIAN_HIEROGLYPH_13CA1 = 0x00013CA1;
const char32_t EGYPTIAN_HIEROGLYPH_13CA2 = 0x00013CA2;
const char32_t EGYPTIAN_HIEROGLYPH_13CA3 = 0x00013CA3;
const char32_t EGYPTIAN_HIEROGLYPH_13CA4 = 0x00013CA4;
const char32_t EGYPTIAN_HIEROGLYPH_13CA5 = 0x00013CA5;
const char32_t EGYPTIAN_HIEROGLYPH_13CA6 = 0x00013CA6;
const char32_t EGYPTIAN_HIEROGLYPH_13CA7 = 0x00013CA7;
const char32_t EGYPTIAN_HIEROGLYPH_13CA8 = 0x00013CA8;
const char32_t EGYPTIAN_HIEROGLYPH_13CA9 = 0x00013CA9;
const char32_t EGYPTIAN_HIEROGLYPH_13CAA = 0x00013CAA;
const char32_t EGYPTIAN_HIEROGLYPH_13CAB = 0x00013CAB;
const char32_t EGYPTIAN_HIEROGLYPH_13CAC = 0x00013CAC;
const char32_t EGYPTIAN_HIEROGLYPH_13CAD = 0x00013CAD;
const char32_t EGYPTIAN_HIEROGLYPH_13CAE = 0x00013CAE;
const char32_t EGYPTIAN_HIEROGLYPH_13CAF = 0x00013CAF;
const char32_t EGYPTIAN_HIEROGLYPH_13CB0 = 0x00013CB0;
const char32_t EGYPTIAN_HIEROGLYPH_13CB1 = 0x00013CB1;
const char32_t EGYPTIAN_HIEROGLYPH_13CB2 = 0x00013CB2;
const char32_t EGYPTIAN_HIEROGLYPH_13CB3 = 0x00013CB3;
const char32_t EGYPTIAN_HIEROGLYPH_13CB4 = 0x00013CB4;
const char32_t EGYPTIAN_HIEROGLYPH_13CB5 = 0x00013CB5;
const char32_t EGYPTIAN_HIEROGLYPH_13CB6 = 0x00013CB6;
const char32_t EGYPTIAN_HIEROGLYPH_13CB7 = 0x00013CB7;
const char32_t EGYPTIAN_HIEROGLYPH_13CB8 = 0x00013CB8;
const char32_t EGYPTIAN_HIEROGLYPH_13CB9 = 0x00013CB9;
const char32_t EGYPTIAN_HIEROGLYPH_13CBA = 0x00013CBA;
const char32_t EGYPTIAN_HIEROGLYPH_13CBB = 0x00013CBB;
const char32_t EGYPTIAN_HIEROGLYPH_13CBC = 0x00013CBC;
const char32_t EGYPTIAN_HIEROGLYPH_13CBD = 0x00013CBD;
const char32_t EGYPTIAN_HIEROGLYPH_13CBE = 0x00013CBE;
const char32_t EGYPTIAN_HIEROGLYPH_13CBF = 0x00013CBF;
const char32_t EGYPTIAN_HIEROGLYPH_13CC0 = 0x00013CC0;
const char32_t EGYPTIAN_HIEROGLYPH_13CC1 = 0x00013CC1;
const char32_t EGYPTIAN_HIEROGLYPH_13CC2 = 0x00013CC2;
const char32_t EGYPTIAN_HIEROGLYPH_13CC3 = 0x00013CC3;
const char32_t EGYPTIAN_HIEROGLYPH_13CC4 = 0x00013CC4;
const char32_t EGYPTIAN_HIEROGLYPH_13CC5 = 0x00013CC5;
const char32_t EGYPTIAN_HIEROGLYPH_13CC6 = 0x00013CC6;
const char32_t EGYPTIAN_HIEROGLYPH_13CC7 = 0x00013CC7;
const char32_t EGYPTIAN_HIEROGLYPH_13CC8 = 0x00013CC8;
const char32_t EGYPTIAN_HIEROGLYPH_13CC9 = 0x00013CC9;
const char32_t EGYPTIAN_HIEROGLYPH_13CCA = 0x00013CCA;
const char32_t EGYPTIAN_HIEROGLYPH_13CCB = 0x00013CCB;
const char32_t EGYPTIAN_HIEROGLYPH_13CCC = 0x00013CCC;
const char32_t EGYPTIAN_HIEROGLYPH_13CCD = 0x00013CCD;
const char32_t EGYPTIAN_HIEROGLYPH_13CCE = 0x00013CCE;
const char32_t EGYPTIAN_HIEROGLYPH_13CCF = 0x00013CCF;
const char32_t EGYPTIAN_HIEROGLYPH_13CD0 = 0x00013CD0;
const char32_t EGYPTIAN_HIEROGLYPH_13CD1 = 0x00013CD1;
const char32_t EGYPTIAN_HIEROGLYPH_13CD2 = 0x00013CD2;
const char32_t EGYPTIAN_HIEROGLYPH_13CD3 = 0x00013CD3;
const char32_t EGYPTIAN_HIEROGLYPH_13CD4 = 0x00013CD4;
const char32_t EGYPTIAN_HIEROGLYPH_13CD5 = 0x00013CD5;
const char32_t EGYPTIAN_HIEROGLYPH_13CD6 = 0x00013CD6;
const char32_t EGYPTIAN_HIEROGLYPH_13CD7 = 0x00013CD7;
const char32_t EGYPTIAN_HIEROGLYPH_13CD8 = 0x00013CD8;
const char32_t EGYPTIAN_HIEROGLYPH_13CD9 = 0x00013CD9;
const char32_t EGYPTIAN_HIEROGLYPH_13CDA = 0x00013CDA;
const char32_t EGYPTIAN_HIEROGLYPH_13CDB = 0x00013CDB;
const char32_t EGYPTIAN_HIEROGLYPH_13CDC = 0x00013CDC;
const char32_t EGYPTIAN_HIEROGLYPH_13CDD = 0x00013CDD;
const char32_t EGYPTIAN_HIEROGLYPH_13CDE = 0x00013CDE;
const char32_t EGYPTIAN_HIEROGLYPH_13CDF = 0x00013CDF;
const char32_t EGYPTIAN_HIEROGLYPH_13CE0 = 0x00013CE0;
const char32_t EGYPTIAN_HIEROGLYPH_13CE1 = 0x00013CE1;
const char32_t EGYPTIAN_HIEROGLYPH_13CE2 = 0x00013CE2;
const char32_t EGYPTIAN_HIEROGLYPH_13CE3 = 0x00013CE3;
const char32_t EGYPTIAN_HIEROGLYPH_13CE4 = 0x00013CE4;
const char32_t EGYPTIAN_HIEROGLYPH_13CE5 = 0x00013CE5;
const char32_t EGYPTIAN_HIEROGLYPH_13CE6 = 0x00013CE6;
const char32_t EGYPTIAN_HIEROGLYPH_13CE7 = 0x00013CE7;
const char32_t EGYPTIAN_HIEROGLYPH_13CE8 = 0x00013CE8;
const char32_t EGYPTIAN_HIEROGLYPH_13CE9 = 0x00013CE9;
const char32_t EGYPTIAN_HIEROGLYPH_13CEA = 0x00013CEA;
const char32_t EGYPTIAN_HIEROGLYPH_13CEB = 0x00013CEB;
const char32_t EGYPTIAN_HIEROGLYPH_13CEC = 0x00013CEC;
const char32_t EGYPTIAN_HIEROGLYPH_13CED = 0x00013CED;
const char32_t EGYPTIAN_HIEROGLYPH_13CEE = 0x00013CEE;
const char32_t EGYPTIAN_HIEROGLYPH_13CEF = 0x00013CEF;
const char32_t EGYPTIAN_HIEROGLYPH_13CF0 = 0x00013CF0;
const char32_t EGYPTIAN_HIEROGLYPH_13CF1 = 0x00013CF1;
const char32_t EGYPTIAN_HIEROGLYPH_13CF2 = 0x00013CF2;
const char32_t EGYPTIAN_HIEROGLYPH_13CF3 = 0x00013CF3;
const char32_t EGYPTIAN_HIEROGLYPH_13CF4 = 0x00013CF4;
const char32_t EGYPTIAN_HIEROGLYPH_13CF5 = 0x00013CF5;
const char32_t EGYPTIAN_HIEROGLYPH_13CF6 = 0x00013CF6;
const char32_t EGYPTIAN_HIEROGLYPH_13CF7 = 0x00013CF7;
const char32_t EGYPTIAN_HIEROGLYPH_13CF8 = 0x00013CF8;
const char32_t EGYPTIAN_HIEROGLYPH_13CF9 = 0x00013CF9;
const char32_t EGYPTIAN_HIEROGLYPH_13CFA = 0x00013CFA;
const char32_t EGYPTIAN_HIEROGLYPH_13CFB = 0x00013CFB;
const char32_t EGYPTIAN_HIEROGLYPH_13CFC = 0x00013CFC;
const char32_t EGYPTIAN_HIEROGLYPH_13CFD = 0x00013CFD;
const char32_t EGYPTIAN_HIEROGLYPH_13CFE = 0x00013CFE;
const char32_t EGYPTIAN_HIEROGLYPH_13CFF = 0x00013CFF;
const char32_t EGYPTIAN_HIEROGLYPH_13D00 = 0x00013D00;
const char32_t EGYPTIAN_HIEROGLYPH_13D01 = 0x00013D01;
const char32_t EGYPTIAN_HIEROGLYPH_13D02 = 0x00013D02;
const char32_t EGYPTIAN_HIEROGLYPH_13D03 = 0x00013D03;
const char32_t EGYPTIAN_HIEROGLYPH_13D04 = 0x00013D04;
const char32_t EGYPTIAN_HIEROGLYPH_13D05 = 0x00013D05;
const char32_t EGYPTIAN_HIEROGLYPH_13D06 = 0x00013D06;
const char32_t EGYPTIAN_HIEROGLYPH_13D07 = 0x00013D07;
const char32_t EGYPTIAN_HIEROGLYPH_13D08 = 0x00013D08;
const char32_t EGYPTIAN_HIEROGLYPH_13D09 = 0x00013D09;
const char32_t EGYPTIAN_HIEROGLYPH_13D0A = 0x00013D0A;
const char32_t EGYPTIAN_HIEROGLYPH_13D0B = 0x00013D0B;
const char32_t EGYPTIAN_HIEROGLYPH_13D0C = 0x00013D0C;
const char32_t EGYPTIAN_HIEROGLYPH_13D0D = 0x00013D0D;
const char32_t EGYPTIAN_HIEROGLYPH_13D0E = 0x00013D0E;
const char32_t EGYPTIAN_HIEROGLYPH_13D0F = 0x00013D0F;
const char32_t EGYPTIAN_HIEROGLYPH_13D10 = 0x00013D10;
const char32_t EGYPTIAN_HIEROGLYPH_13D11 = 0x00013D11;
const char32_t EGYPTIAN_HIEROGLYPH_13D12 = 0x00013D12;
const char32_t EGYPTIAN_HIEROGLYPH_13D13 = 0x00013D13;
const char32_t EGYPTIAN_HIEROGLYPH_13D14 = 0x00013D14;
const char32_t EGYPTIAN_HIEROGLYPH_13D15 = 0x00013D15;
const char32_t EGYPTIAN_HIEROGLYPH_13D16 = 0x00013D16;
const char32_t EGYPTIAN_HIEROGLYPH_13D17 = 0x00013D17;
const char32_t EGYPTIAN_HIEROGLYPH_13D18 = 0x00013D18;
const char32_t EGYPTIAN_HIEROGLYPH_13D19 = 0x00013D19;
const char32_t EGYPTIAN_HIEROGLYPH_13D1A = 0x00013D1A;
const char32_t EGYPTIAN_HIEROGLYPH_13D1B = 0x00013D1B;
const char32_t EGYPTIAN_HIEROGLYPH_13D1C = 0x00013D1C;
const char32_t EGYPTIAN_HIEROGLYPH_13D1D = 0x00013D1D;
const char32_t EGYPTIAN_HIEROGLYPH_13D1E = 0x00013D1E;
const char32_t EGYPTIAN_HIEROGLYPH_13D1F = 0x00013D1F;
const char32_t EGYPTIAN_HIEROGLYPH_13D20 = 0x00013D20;
const char32_t EGYPTIAN_HIEROGLYPH_13D21 = 0x00013D21;
const char32_t EGYPTIAN_HIEROGLYPH_13D22 = 0x00013D22;
const char32_t EGYPTIAN_HIEROGLYPH_13D23 = 0x00013D23;
const char32_t EGYPTIAN_HIEROGLYPH_13D24 = 0x00013D24;
const char32_t EGYPTIAN_HIEROGLYPH_13D25 = 0x00013D25;
const char32_t EGYPTIAN_HIEROGLYPH_13D26 = 0x00013D26;
const char32_t EGYPTIAN_HIEROGLYPH_13D27 = 0x00013D27;
const char32_t EGYPTIAN_HIEROGLYPH_13D28 = 0x00013D28;
const char32_t EGYPTIAN_HIEROGLYPH_13D29 = 0x00013D29;
const char32_t EGYPTIAN_HIEROGLYPH_13D2A = 0x00013D2A;
const char32_t EGYPTIAN_HIEROGLYPH_13D2B = 0x00013D2B;
const char32_t EGYPTIAN_HIEROGLYPH_13D2C = 0x00013D2C;
const char32_t EGYPTIAN_HIEROGLYPH_13D2D = 0x00013D2D;
const char32_t EGYPTIAN_HIEROGLYPH_13D2E = 0x00013D2E;
const char32_t EGYPTIAN_HIEROGLYPH_13D2F = 0x00013D2F;
const char32_t EGYPTIAN_HIEROGLYPH_13D30 = 0x00013D30;
const char32_t EGYPTIAN_HIEROGLYPH_13D31 = 0x00013D31;
const char32_t EGYPTIAN_HIEROGLYPH_13D32 = 0x00013D32;
const char32_t EGYPTIAN_HIEROGLYPH_13D33 = 0x00013D33;
const char32_t EGYPTIAN_HIEROGLYPH_13D34 = 0x00013D34;
const char32_t EGYPTIAN_HIEROGLYPH_13D35 = 0x00013D35;
const char32_t EGYPTIAN_HIEROGLYPH_13D36 = 0x00013D36;
const char32_t EGYPTIAN_HIEROGLYPH_13D37 = 0x00013D37;
const char32_t EGYPTIAN_HIEROGLYPH_13D38 = 0x00013D38;
const char32_t EGYPTIAN_HIEROGLYPH_13D39 = 0x00013D39;
const char32_t EGYPTIAN_HIEROGLYPH_13D3A = 0x00013D3A;
const char32_t EGYPTIAN_HIEROGLYPH_13D3B = 0x00013D3B;
const char32_t EGYPTIAN_HIEROGLYPH_13D3C = 0x00013D3C;
const char32_t EGYPTIAN_HIEROGLYPH_13D3D = 0x00013D3D;
const char32_t EGYPTIAN_HIEROGLYPH_13D3E = 0x00013D3E;
const char32_t EGYPTIAN_HIEROGLYPH_13D3F = 0x00013D3F;
const char32_t EGYPTIAN_HIEROGLYPH_13D40 = 0x00013D40;
const char32_t EGYPTIAN_HIEROGLYPH_13D41 = 0x00013D41;
const char32_t EGYPTIAN_HIEROGLYPH_13D42 = 0x00013D42;
const char32_t EGYPTIAN_HIEROGLYPH_13D43 = 0x00013D43;
const char32_t EGYPTIAN_HIEROGLYPH_13D44 = 0x00013D44;
const char32_t EGYPTIAN_HIEROGLYPH_13D45 = 0x00013D45;
const char32_t EGYPTIAN_HIEROGLYPH_13D46 = 0x00013D46;
const char32_t EGYPTIAN_HIEROGLYPH_13D47 = 0x00013D47;
const char32_t EGYPTIAN_HIEROGLYPH_13D48 = 0x00013D48;
const char32_t EGYPTIAN_HIEROGLYPH_13D49 = 0x00013D49;
const char32_t EGYPTIAN_HIEROGLYPH_13D4A = 0x00013D4A;
const char32_t EGYPTIAN_HIEROGLYPH_13D4B = 0x00013D4B;
const char32_t EGYPTIAN_HIEROGLYPH_13D4C = 0x00013D4C;
const char32_t EGYPTIAN_HIEROGLYPH_13D4D = 0x00013D4D;
const char32_t EGYPTIAN_HIEROGLYPH_13D4E = 0x00013D4E;
const char32_t EGYPTIAN_HIEROGLYPH_13D4F = 0x00013D4F;
const char32_t EGYPTIAN_HIEROGLYPH_13D50 = 0x00013D50;
const char32_t EGYPTIAN_HIEROGLYPH_13D51 = 0x00013D51;
const char32_t EGYPTIAN_HIEROGLYPH_13D52 = 0x00013D52;
const char32_t EGYPTIAN_HIEROGLYPH_13D53 = 0x00013D53;
const char32_t EGYPTIAN_HIEROGLYPH_13D54 = 0x00013D54;
const char32_t EGYPTIAN_HIEROGLYPH_13D55 = 0x00013D55;
const char32_t EGYPTIAN_HIEROGLYPH_13D56 = 0x00013D56;
const char32_t EGYPTIAN_HIEROGLYPH_13D57 = 0x00013D57;
const char32_t EGYPTIAN_HIEROGLYPH_13D58 = 0x00013D58;
const char32_t EGYPTIAN_HIEROGLYPH_13D59 = 0x00013D59;
const char32_t EGYPTIAN_HIEROGLYPH_13D5A = 0x00013D5A;
const char32_t EGYPTIAN_HIEROGLYPH_13D5B = 0x00013D5B;
const char32_t EGYPTIAN_HIEROGLYPH_13D5C = 0x00013D5C;
const char32_t EGYPTIAN_HIEROGLYPH_13D5D = 0x00013D5D;
const char32_t EGYPTIAN_HIEROGLYPH_13D5E = 0x00013D5E;
const char32_t EGYPTIAN_HIEROGLYPH_13D5F = 0x00013D5F;
const char32_t EGYPTIAN_HIEROGLYPH_13D60 = 0x00013D60;
const char32_t EGYPTIAN_HIEROGLYPH_13D61 = 0x00013D61;
const char32_t EGYPTIAN_HIEROGLYPH_13D62 = 0x00013D62;
const char32_t EGYPTIAN_HIEROGLYPH_13D63 = 0x00013D63;
const char32_t EGYPTIAN_HIEROGLYPH_13D64 = 0x00013D64;
const char32_t EGYPTIAN_HIEROGLYPH_13D65 = 0x00013D65;
const char32_t EGYPTIAN_HIEROGLYPH_13D66 = 0x00013D66;
const char32_t EGYPTIAN_HIEROGLYPH_13D67 = 0x00013D67;
const char32_t EGYPTIAN_HIEROGLYPH_13D68 = 0x00013D68;
const char32_t EGYPTIAN_HIEROGLYPH_13D69 = 0x00013D69;
const char32_t EGYPTIAN_HIEROGLYPH_13D6A = 0x00013D6A;
const char32_t EGYPTIAN_HIEROGLYPH_13D6B = 0x00013D6B;
const char32_t EGYPTIAN_HIEROGLYPH_13D6C = 0x00013D6C;
const char32_t EGYPTIAN_HIEROGLYPH_13D6D = 0x00013D6D;
const char32_t EGYPTIAN_HIEROGLYPH_13D6E = 0x00013D6E;
const char32_t EGYPTIAN_HIEROGLYPH_13D6F = 0x00013D6F;
const char32_t EGYPTIAN_HIEROGLYPH_13D70 = 0x00013D70;
const char32_t EGYPTIAN_HIEROGLYPH_13D71 = 0x00013D71;
const char32_t EGYPTIAN_HIEROGLYPH_13D72 = 0x00013D72;
const char32_t EGYPTIAN_HIEROGLYPH_13D73 = 0x00013D73;
const char32_t EGYPTIAN_HIEROGLYPH_13D74 = 0x00013D74;
const char32_t EGYPTIAN_HIEROGLYPH_13D75 = 0x00013D75;
const char32_t EGYPTIAN_HIEROGLYPH_13D76 = 0x00013D76;
const char32_t EGYPTIAN_HIEROGLYPH_13D77 = 0x00013D77;
const char32_t EGYPTIAN_HIEROGLYPH_13D78 = 0x00013D78;
const char32_t EGYPTIAN_HIEROGLYPH_13D79 = 0x00013D79;
const char32_t EGYPTIAN_HIEROGLYPH_13D7A = 0x00013D7A;
const char32_t EGYPTIAN_HIEROGLYPH_13D7B = 0x00013D7B;
const char32_t EGYPTIAN_HIEROGLYPH_13D7C = 0x00013D7C;
const char32_t EGYPTIAN_HIEROGLYPH_13D7D = 0x00013D7D;
const char32_t EGYPTIAN_HIEROGLYPH_13D7E = 0x00013D7E;
const char32_t EGYPTIAN_HIEROGLYPH_13D7F = 0x00013D7F;
const char32_t EGYPTIAN_HIEROGLYPH_13D80 = 0x00013D80;
const char32_t EGYPTIAN_HIEROGLYPH_13D81 = 0x00013D81;
const char32_t EGYPTIAN_HIEROGLYPH_13D82 = 0x00013D82;
const char32_t EGYPTIAN_HIEROGLYPH_13D83 = 0x00013D83;
const char32_t EGYPTIAN_HIEROGLYPH_13D84 = 0x00013D84;
const char32_t EGYPTIAN_HIEROGLYPH_13D85 = 0x00013D85;
const char32_t EGYPTIAN_HIEROGLYPH_13D86 = 0x00013D86;
const char32_t EGYPTIAN_HIEROGLYPH_13D87 = 0x00013D87;
const char32_t EGYPTIAN_HIEROGLYPH_13D88 = 0x00013D88;
const char32_t EGYPTIAN_HIEROGLYPH_13D89 = 0x00013D89;
const char32_t EGYPTIAN_HIEROGLYPH_13D8A = 0x00013D8A;
const char32_t EGYPTIAN_HIEROGLYPH_13D8B = 0x00013D8B;
const char32_t EGYPTIAN_HIEROGLYPH_13D8C = 0x00013D8C;
const char32_t EGYPTIAN_HIEROGLYPH_13D8D = 0x00013D8D;
const char32_t EGYPTIAN_HIEROGLYPH_13D8E = 0x00013D8E;
const char32_t EGYPTIAN_HIEROGLYPH_13D8F = 0x00013D8F;
const char32_t EGYPTIAN_HIEROGLYPH_13D90 = 0x00013D90;
const char32_t EGYPTIAN_HIEROGLYPH_13D91 = 0x00013D91;
const char32_t EGYPTIAN_HIEROGLYPH_13D92 = 0x00013D92;
const char32_t EGYPTIAN_HIEROGLYPH_13D93 = 0x00013D93;
const char32_t EGYPTIAN_HIEROGLYPH_13D94 = 0x00013D94;
const char32_t EGYPTIAN_HIEROGLYPH_13D95 = 0x00013D95;
const char32_t EGYPTIAN_HIEROGLYPH_13D96 = 0x00013D96;
const char32_t EGYPTIAN_HIEROGLYPH_13D97 = 0x00013D97;
const char32_t EGYPTIAN_HIEROGLYPH_13D98 = 0x00013D98;
const char32_t EGYPTIAN_HIEROGLYPH_13D99 = 0x00013D99;
const char32_t EGYPTIAN_HIEROGLYPH_13D9A = 0x00013D9A;
const char32_t EGYPTIAN_HIEROGLYPH_13D9B = 0x00013D9B;
const char32_t EGYPTIAN_HIEROGLYPH_13D9C = 0x00013D9C;
const char32_t EGYPTIAN_HIEROGLYPH_13D9D = 0x00013D9D;
const char32_t EGYPTIAN_HIEROGLYPH_13D9E = 0x00013D9E;
const char32_t EGYPTIAN_HIEROGLYPH_13D9F = 0x00013D9F;
const char32_t EGYPTIAN_HIEROGLYPH_13DA0 = 0x00013DA0;
const char32_t EGYPTIAN_HIEROGLYPH_13DA1 = 0x00013DA1;
const char32_t EGYPTIAN_HIEROGLYPH_13DA2 = 0x00013DA2;
const char32_t EGYPTIAN_HIEROGLYPH_13DA3 = 0x00013DA3;
const char32_t EGYPTIAN_HIEROGLYPH_13DA4 = 0x00013DA4;
const char32_t EGYPTIAN_HIEROGLYPH_13DA5 = 0x00013DA5;
const char32_t EGYPTIAN_HIEROGLYPH_13DA6 = 0x00013DA6;
const char32_t EGYPTIAN_HIEROGLYPH_13DA7 = 0x00013DA7;
const char32_t EGYPTIAN_HIEROGLYPH_13DA8 = 0x00013DA8;
const char32_t EGYPTIAN_HIEROGLYPH_13DA9 = 0x00013DA9;
const char32_t EGYPTIAN_HIEROGLYPH_13DAA = 0x00013DAA;
const char32_t EGYPTIAN_HIEROGLYPH_13DAB = 0x00013DAB;
const char32_t EGYPTIAN_HIEROGLYPH_13DAC = 0x00013DAC;
const char32_t EGYPTIAN_HIEROGLYPH_13DAD = 0x00013DAD;
const char32_t EGYPTIAN_HIEROGLYPH_13DAE = 0x00013DAE;
const char32_t EGYPTIAN_HIEROGLYPH_13DAF = 0x00013DAF;
const char32_t EGYPTIAN_HIEROGLYPH_13DB0 = 0x00013DB0;
const char32_t EGYPTIAN_HIEROGLYPH_13DB1 = 0x00013DB1;
const char32_t EGYPTIAN_HIEROGLYPH_13DB2 = 0x00013DB2;
const char32_t EGYPTIAN_HIEROGLYPH_13DB3 = 0x00013DB3;
const char32_t EGYPTIAN_HIEROGLYPH_13DB4 = 0x00013DB4;
const char32_t EGYPTIAN_HIEROGLYPH_13DB5 = 0x00013DB5;
const char32_t EGYPTIAN_HIEROGLYPH_13DB6 = 0x00013DB6;
const char32_t EGYPTIAN_HIEROGLYPH_13DB7 = 0x00013DB7;
const char32_t EGYPTIAN_HIEROGLYPH_13DB8 = 0x00013DB8;
const char32_t EGYPTIAN_HIEROGLYPH_13DB9 = 0x00013DB9;
const char32_t EGYPTIAN_HIEROGLYPH_13DBA = 0x00013DBA;
const char32_t EGYPTIAN_HIEROGLYPH_13DBB = 0x00013DBB;
const char32_t EGYPTIAN_HIEROGLYPH_13DBC = 0x00013DBC;
const char32_t EGYPTIAN_HIEROGLYPH_13DBD = 0x00013DBD;
const char32_t EGYPTIAN_HIEROGLYPH_13DBE = 0x00013DBE;
const char32_t EGYPTIAN_HIEROGLYPH_13DBF = 0x00013DBF;
const char32_t EGYPTIAN_HIEROGLYPH_13DC0 = 0x00013DC0;
const char32_t EGYPTIAN_HIEROGLYPH_13DC1 = 0x00013DC1;
const char32_t EGYPTIAN_HIEROGLYPH_13DC2 = 0x00013DC2;
const char32_t EGYPTIAN_HIEROGLYPH_13DC3 = 0x00013DC3;
const char32_t EGYPTIAN_HIEROGLYPH_13DC4 = 0x00013DC4;
const char32_t EGYPTIAN_HIEROGLYPH_13DC5 = 0x00013DC5;
const char32_t EGYPTIAN_HIEROGLYPH_13DC6 = 0x00013DC6;
const char32_t EGYPTIAN_HIEROGLYPH_13DC7 = 0x00013DC7;
const char32_t EGYPTIAN_HIEROGLYPH_13DC8 = 0x00013DC8;
const char32_t EGYPTIAN_HIEROGLYPH_13DC9 = 0x00013DC9;
const char32_t EGYPTIAN_HIEROGLYPH_13DCA = 0x00013DCA;
const char32_t EGYPTIAN_HIEROGLYPH_13DCB = 0x00013DCB;
const char32_t EGYPTIAN_HIEROGLYPH_13DCC = 0x00013DCC;
const char32_t EGYPTIAN_HIEROGLYPH_13DCD = 0x00013DCD;
const char32_t EGYPTIAN_HIEROGLYPH_13DCE = 0x00013DCE;
const char32_t EGYPTIAN_HIEROGLYPH_13DCF = 0x00013DCF;
const char32_t EGYPTIAN_HIEROGLYPH_13DD0 = 0x00013DD0;
const char32_t EGYPTIAN_HIEROGLYPH_13DD1 = 0x00013DD1;
const char32_t EGYPTIAN_HIEROGLYPH_13DD2 = 0x00013DD2;
const char32_t EGYPTIAN_HIEROGLYPH_13DD3 = 0x00013DD3;
const char32_t EGYPTIAN_HIEROGLYPH_13DD4 = 0x00013DD4;
const char32_t EGYPTIAN_HIEROGLYPH_13DD5 = 0x00013DD5;
const char32_t EGYPTIAN_HIEROGLYPH_13DD6 = 0x00013DD6;
const char32_t EGYPTIAN_HIEROGLYPH_13DD7 = 0x00013DD7;
const char32_t EGYPTIAN_HIEROGLYPH_13DD8 = 0x00013DD8;
const char32_t EGYPTIAN_HIEROGLYPH_13DD9 = 0x00013DD9;
const char32_t EGYPTIAN_HIEROGLYPH_13DDA = 0x00013DDA;
const char32_t EGYPTIAN_HIEROGLYPH_13DDB = 0x00013DDB;
const char32_t EGYPTIAN_HIEROGLYPH_13DDC = 0x00013DDC;
const char32_t EGYPTIAN_HIEROGLYPH_13DDD = 0x00013DDD;
const char32_t EGYPTIAN_HIEROGLYPH_13DDE = 0x00013DDE;
const char32_t EGYPTIAN_HIEROGLYPH_13DDF = 0x00013DDF;
const char32_t EGYPTIAN_HIEROGLYPH_13DE0 = 0x00013DE0;
const char32_t EGYPTIAN_HIEROGLYPH_13DE1 = 0x00013DE1;
const char32_t EGYPTIAN_HIEROGLYPH_13DE2 = 0x00013DE2;
const char32_t EGYPTIAN_HIEROGLYPH_13DE3 = 0x00013DE3;
const char32_t EGYPTIAN_HIEROGLYPH_13DE4 = 0x00013DE4;
const char32_t EGYPTIAN_HIEROGLYPH_13DE5 = 0x00013DE5;
const char32_t EGYPTIAN_HIEROGLYPH_13DE6 = 0x00013DE6;
const char32_t EGYPTIAN_HIEROGLYPH_13DE7 = 0x00013DE7;
const char32_t EGYPTIAN_HIEROGLYPH_13DE8 = 0x00013DE8;
const char32_t EGYPTIAN_HIEROGLYPH_13DE9 = 0x00013DE9;
const char32_t EGYPTIAN_HIEROGLYPH_13DEA = 0x00013DEA;
const char32_t EGYPTIAN_HIEROGLYPH_13DEB = 0x00013DEB;
const char32_t EGYPTIAN_HIEROGLYPH_13DEC = 0x00013DEC;
const char32_t EGYPTIAN_HIEROGLYPH_13DED = 0x00013DED;
const char32_t EGYPTIAN_HIEROGLYPH_13DEE = 0x00013DEE;
const char32_t EGYPTIAN_HIEROGLYPH_13DEF = 0x00013DEF;
const char32_t EGYPTIAN_HIEROGLYPH_13DF0 = 0x00013DF0;
const char32_t EGYPTIAN_HIEROGLYPH_13DF1 = 0x00013DF1;
const char32_t EGYPTIAN_HIEROGLYPH_13DF2 = 0x00013DF2;
const char32_t EGYPTIAN_HIEROGLYPH_13DF3 = 0x00013DF3;
const char32_t EGYPTIAN_HIEROGLYPH_13DF4 = 0x00013DF4;
const char32_t EGYPTIAN_HIEROGLYPH_13DF5 = 0x00013DF5;
const char32_t EGYPTIAN_HIEROGLYPH_13DF6 = 0x00013DF6;
const char32_t EGYPTIAN_HIEROGLYPH_13DF7 = 0x00013DF7;
const char32_t EGYPTIAN_HIEROGLYPH_13DF8 = 0x00013DF8;
const char32_t EGYPTIAN_HIEROGLYPH_13DF9 = 0x00013DF9;
const char32_t EGYPTIAN_HIEROGLYPH_13DFA = 0x00013DFA;
const char32_t EGYPTIAN_HIEROGLYPH_13DFB = 0x00013DFB;
const char32_t EGYPTIAN_HIEROGLYPH_13DFC = 0x00013DFC;
const char32_t EGYPTIAN_HIEROGLYPH_13DFD = 0x00013DFD;
const char32_t EGYPTIAN_HIEROGLYPH_13DFE = 0x00013DFE;
const char32_t EGYPTIAN_HIEROGLYPH_13DFF = 0x00013DFF;
const char32_t EGYPTIAN_HIEROGLYPH_13E00 = 0x00013E00;
const char32_t EGYPTIAN_HIEROGLYPH_13E01 = 0x00013E01;
const char32_t EGYPTIAN_HIEROGLYPH_13E02 = 0x00013E02;
const char32_t EGYPTIAN_HIEROGLYPH_13E03 = 0x00013E03;
const char32_t EGYPTIAN_HIEROGLYPH_13E04 = 0x00013E04;
const char32_t EGYPTIAN_HIEROGLYPH_13E05 = 0x00013E05;
const char32_t EGYPTIAN_HIEROGLYPH_13E06 = 0x00013E06;
const char32_t EGYPTIAN_HIEROGLYPH_13E07 = 0x00013E07;
const char32_t EGYPTIAN_HIEROGLYPH_13E08 = 0x00013E08;
const char32_t EGYPTIAN_HIEROGLYPH_13E09 = 0x00013E09;
const char32_t EGYPTIAN_HIEROGLYPH_13E0A = 0x00013E0A;
const char32_t EGYPTIAN_HIEROGLYPH_13E0B = 0x00013E0B;
const char32_t EGYPTIAN_HIEROGLYPH_13E0C = 0x00013E0C;
const char32_t EGYPTIAN_HIEROGLYPH_13E0D = 0x00013E0D;
const char32_t EGYPTIAN_HIEROGLYPH_13E0E = 0x00013E0E;
const char32_t EGYPTIAN_HIEROGLYPH_13E0F = 0x00013E0F;
const char32_t EGYPTIAN_HIEROGLYPH_13E10 = 0x00013E10;
const char32_t EGYPTIAN_HIEROGLYPH_13E11 = 0x00013E11;
const char32_t EGYPTIAN_HIEROGLYPH_13E12 = 0x00013E12;
const char32_t EGYPTIAN_HIEROGLYPH_13E13 = 0x00013E13;
const char32_t EGYPTIAN_HIEROGLYPH_13E14 = 0x00013E14;
const char32_t EGYPTIAN_HIEROGLYPH_13E15 = 0x00013E15;
const char32_t EGYPTIAN_HIEROGLYPH_13E16 = 0x00013E16;
const char32_t EGYPTIAN_HIEROGLYPH_13E17 = 0x00013E17;
const char32_t EGYPTIAN_HIEROGLYPH_13E18 = 0x00013E18;
const char32_t EGYPTIAN_HIEROGLYPH_13E19 = 0x00013E19;
const char32_t EGYPTIAN_HIEROGLYPH_13E1A = 0x00013E1A;
const char32_t EGYPTIAN_HIEROGLYPH_13E1B = 0x00013E1B;
const char32_t EGYPTIAN_HIEROGLYPH_13E1C = 0x00013E1C;
const char32_t EGYPTIAN_HIEROGLYPH_13E1D = 0x00013E1D;
const char32_t EGYPTIAN_HIEROGLYPH_13E1E = 0x00013E1E;
const char32_t EGYPTIAN_HIEROGLYPH_13E1F = 0x00013E1F;
const char32_t EGYPTIAN_HIEROGLYPH_13E20 = 0x00013E20;
const char32_t EGYPTIAN_HIEROGLYPH_13E21 = 0x00013E21;
const char32_t EGYPTIAN_HIEROGLYPH_13E22 = 0x00013E22;
const char32_t EGYPTIAN_HIEROGLYPH_13E23 = 0x00013E23;
const char32_t EGYPTIAN_HIEROGLYPH_13E24 = 0x00013E24;
const char32_t EGYPTIAN_HIEROGLYPH_13E25 = 0x00013E25;
const char32_t EGYPTIAN_HIEROGLYPH_13E26 = 0x00013E26;
const char32_t EGYPTIAN_HIEROGLYPH_13E27 = 0x00013E27;
const char32_t EGYPTIAN_HIEROGLYPH_13E28 = 0x00013E28;
const char32_t EGYPTIAN_HIEROGLYPH_13E29 = 0x00013E29;
const char32_t EGYPTIAN_HIEROGLYPH_13E2A = 0x00013E2A;
const char32_t EGYPTIAN_HIEROGLYPH_13E2B = 0x00013E2B;
const char32_t EGYPTIAN_HIEROGLYPH_13E2C = 0x00013E2C;
const char32_t EGYPTIAN_HIEROGLYPH_13E2D = 0x00013E2D;
const char32_t EGYPTIAN_HIEROGLYPH_13E2E = 0x00013E2E;
const char32_t EGYPTIAN_HIEROGLYPH_13E2F = 0x00013E2F;
const char32_t EGYPTIAN_HIEROGLYPH_13E30 = 0x00013E30;
const char32_t EGYPTIAN_HIEROGLYPH_13E31 = 0x00013E31;
const char32_t EGYPTIAN_HIEROGLYPH_13E32 = 0x00013E32;
const char32_t EGYPTIAN_HIEROGLYPH_13E33 = 0x00013E33;
const char32_t EGYPTIAN_HIEROGLYPH_13E34 = 0x00013E34;
const char32_t EGYPTIAN_HIEROGLYPH_13E35 = 0x00013E35;
const char32_t EGYPTIAN_HIEROGLYPH_13E36 = 0x00013E36;
const char32_t EGYPTIAN_HIEROGLYPH_13E37 = 0x00013E37;
const char32_t EGYPTIAN_HIEROGLYPH_13E38 = 0x00013E38;
const char32_t EGYPTIAN_HIEROGLYPH_13E39 = 0x00013E39;
const char32_t EGYPTIAN_HIEROGLYPH_13E3A = 0x00013E3A;
const char32_t EGYPTIAN_HIEROGLYPH_13E3B = 0x00013E3B;
const char32_t EGYPTIAN_HIEROGLYPH_13E3C = 0x00013E3C;
const char32_t EGYPTIAN_HIEROGLYPH_13E3D = 0x00013E3D;
const char32_t EGYPTIAN_HIEROGLYPH_13E3E = 0x00013E3E;
const char32_t EGYPTIAN_HIEROGLYPH_13E3F = 0x00013E3F;
const char32_t EGYPTIAN_HIEROGLYPH_13E40 = 0x00013E40;
const char32_t EGYPTIAN_HIEROGLYPH_13E41 = 0x00013E41;
const char32_t EGYPTIAN_HIEROGLYPH_13E42 = 0x00013E42;
const char32_t EGYPTIAN_HIEROGLYPH_13E43 = 0x00013E43;
const char32_t EGYPTIAN_HIEROGLYPH_13E44 = 0x00013E44;
const char32_t EGYPTIAN_HIEROGLYPH_13E45 = 0x00013E45;
const char32_t EGYPTIAN_HIEROGLYPH_13E46 = 0x00013E46;
const char32_t EGYPTIAN_HIEROGLYPH_13E47 = 0x00013E47;
const char32_t EGYPTIAN_HIEROGLYPH_13E48 = 0x00013E48;
const char32_t EGYPTIAN_HIEROGLYPH_13E49 = 0x00013E49;
const char32_t EGYPTIAN_HIEROGLYPH_13E4A = 0x00013E4A;
const char32_t EGYPTIAN_HIEROGLYPH_13E4B = 0x00013E4B;
const char32_t EGYPTIAN_HIEROGLYPH_13E4C = 0x00013E4C;
const char32_t EGYPTIAN_HIEROGLYPH_13E4D = 0x00013E4D;
const char32_t EGYPTIAN_HIEROGLYPH_13E4E = 0x00013E4E;
const char32_t EGYPTIAN_HIEROGLYPH_13E4F = 0x00013E4F;
const char32_t EGYPTIAN_HIEROGLYPH_13E50 = 0x00013E50;
const char32_t EGYPTIAN_HIEROGLYPH_13E51 = 0x00013E51;
const char32_t EGYPTIAN_HIEROGLYPH_13E52 = 0x00013E52;
const char32_t EGYPTIAN_HIEROGLYPH_13E53 = 0x00013E53;
const char32_t EGYPTIAN_HIEROGLYPH_13E54 = 0x00013E54;
const char32_t EGYPTIAN_HIEROGLYPH_13E55 = 0x00013E55;
const char32_t EGYPTIAN_HIEROGLYPH_13E56 = 0x00013E56;
const char32_t EGYPTIAN_HIEROGLYPH_13E57 = 0x00013E57;
const char32_t EGYPTIAN_HIEROGLYPH_13E58 = 0x00013E58;
const char32_t EGYPTIAN_HIEROGLYPH_13E59 = 0x00013E59;
const char32_t EGYPTIAN_HIEROGLYPH_13E5A = 0x00013E5A;
const char32_t EGYPTIAN_HIEROGLYPH_13E5B = 0x00013E5B;
const char32_t EGYPTIAN_HIEROGLYPH_13E5C = 0x00013E5C;
const char32_t EGYPTIAN_HIEROGLYPH_13E5D = 0x00013E5D;
const char32_t EGYPTIAN_HIEROGLYPH_13E5E = 0x00013E5E;
const char32_t EGYPTIAN_HIEROGLYPH_13E5F = 0x00013E5F;
const char32_t EGYPTIAN_HIEROGLYPH_13E60 = 0x00013E60;
const char32_t EGYPTIAN_HIEROGLYPH_13E61 = 0x00013E61;
const char32_t EGYPTIAN_HIEROGLYPH_13E62 = 0x00013E62;
const char32_t EGYPTIAN_HIEROGLYPH_13E63 = 0x00013E63;
const char32_t EGYPTIAN_HIEROGLYPH_13E64 = 0x00013E64;
const char32_t EGYPTIAN_HIEROGLYPH_13E65 = 0x00013E65;
const char32_t EGYPTIAN_HIEROGLYPH_13E66 = 0x00013E66;
const char32_t EGYPTIAN_HIEROGLYPH_13E67 = 0x00013E67;
const char32_t EGYPTIAN_HIEROGLYPH_13E68 = 0x00013E68;
const char32_t EGYPTIAN_HIEROGLYPH_13E69 = 0x00013E69;
const char32_t EGYPTIAN_HIEROGLYPH_13E6A = 0x00013E6A;
const char32_t EGYPTIAN_HIEROGLYPH_13E6B = 0x00013E6B;
const char32_t EGYPTIAN_HIEROGLYPH_13E6C = 0x00013E6C;
const char32_t EGYPTIAN_HIEROGLYPH_13E6D = 0x00013E6D;
const char32_t EGYPTIAN_HIEROGLYPH_13E6E = 0x00013E6E;
const char32_t EGYPTIAN_HIEROGLYPH_13E6F = 0x00013E6F;
const char32_t EGYPTIAN_HIEROGLYPH_13E70 = 0x00013E70;
const char32_t EGYPTIAN_HIEROGLYPH_13E71 = 0x00013E71;
const char32_t EGYPTIAN_HIEROGLYPH_13E72 = 0x00013E72;
const char32_t EGYPTIAN_HIEROGLYPH_13E73 = 0x00013E73;
const char32_t EGYPTIAN_HIEROGLYPH_13E74 = 0x00013E74;
const char32_t EGYPTIAN_HIEROGLYPH_13E75 = 0x00013E75;
const char32_t EGYPTIAN_HIEROGLYPH_13E76 = 0x00013E76;
const char32_t EGYPTIAN_HIEROGLYPH_13E77 = 0x00013E77;
const char32_t EGYPTIAN_HIEROGLYPH_13E78 = 0x00013E78;
const char32_t EGYPTIAN_HIEROGLYPH_13E79 = 0x00013E79;
const char32_t EGYPTIAN_HIEROGLYPH_13E7A = 0x00013E7A;
const char32_t EGYPTIAN_HIEROGLYPH_13E7B = 0x00013E7B;
const char32_t EGYPTIAN_HIEROGLYPH_13E7C = 0x00013E7C;
const char32_t EGYPTIAN_HIEROGLYPH_13E7D = 0x00013E7D;
const char32_t EGYPTIAN_HIEROGLYPH_13E7E = 0x00013E7E;
const char32_t EGYPTIAN_HIEROGLYPH_13E7F = 0x00013E7F;
const char32_t EGYPTIAN_HIEROGLYPH_13E80 = 0x00013E80;
const char32_t EGYPTIAN_HIEROGLYPH_13E81 = 0x00013E81;
const char32_t EGYPTIAN_HIEROGLYPH_13E82 = 0x00013E82;
const char32_t EGYPTIAN_HIEROGLYPH_13E83 = 0x00013E83;
const char32_t EGYPTIAN_HIEROGLYPH_13E84 = 0x00013E84;
const char32_t EGYPTIAN_HIEROGLYPH_13E85 = 0x00013E85;
const char32_t EGYPTIAN_HIEROGLYPH_13E86 = 0x00013E86;
const char32_t EGYPTIAN_HIEROGLYPH_13E87 = 0x00013E87;
const char32_t EGYPTIAN_HIEROGLYPH_13E88 = 0x00013E88;
const char32_t EGYPTIAN_HIEROGLYPH_13E89 = 0x00013E89;
const char32_t EGYPTIAN_HIEROGLYPH_13E8A = 0x00013E8A;
const char32_t EGYPTIAN_HIEROGLYPH_13E8B = 0x00013E8B;
const char32_t EGYPTIAN_HIEROGLYPH_13E8C = 0x00013E8C;
const char32_t EGYPTIAN_HIEROGLYPH_13E8D = 0x00013E8D;
const char32_t EGYPTIAN_HIEROGLYPH_13E8E = 0x00013E8E;
const char32_t EGYPTIAN_HIEROGLYPH_13E8F = 0x00013E8F;
const char32_t EGYPTIAN_HIEROGLYPH_13E90 = 0x00013E90;
const char32_t EGYPTIAN_HIEROGLYPH_13E91 = 0x00013E91;
const char32_t EGYPTIAN_HIEROGLYPH_13E92 = 0x00013E92;
const char32_t EGYPTIAN_HIEROGLYPH_13E93 = 0x00013E93;
const char32_t EGYPTIAN_HIEROGLYPH_13E94 = 0x00013E94;
const char32_t EGYPTIAN_HIEROGLYPH_13E95 = 0x00013E95;
const char32_t EGYPTIAN_HIEROGLYPH_13E96 = 0x00013E96;
const char32_t EGYPTIAN_HIEROGLYPH_13E97 = 0x00013E97;
const char32_t EGYPTIAN_HIEROGLYPH_13E98 = 0x00013E98;
const char32_t EGYPTIAN_HIEROGLYPH_13E99 = 0x00013E99;
const char32_t EGYPTIAN_HIEROGLYPH_13E9A = 0x00013E9A;
const char32_t EGYPTIAN_HIEROGLYPH_13E9B = 0x00013E9B;
const char32_t EGYPTIAN_HIEROGLYPH_13E9C = 0x00013E9C;
const char32_t EGYPTIAN_HIEROGLYPH_13E9D = 0x00013E9D;
const char32_t EGYPTIAN_HIEROGLYPH_13E9E = 0x00013E9E;
const char32_t EGYPTIAN_HIEROGLYPH_13E9F = 0x00013E9F;
const char32_t EGYPTIAN_HIEROGLYPH_13EA0 = 0x00013EA0;
const char32_t EGYPTIAN_HIEROGLYPH_13EA1 = 0x00013EA1;
const char32_t EGYPTIAN_HIEROGLYPH_13EA2 = 0x00013EA2;
const char32_t EGYPTIAN_HIEROGLYPH_13EA3 = 0x00013EA3;
const char32_t EGYPTIAN_HIEROGLYPH_13EA4 = 0x00013EA4;
const char32_t EGYPTIAN_HIEROGLYPH_13EA5 = 0x00013EA5;
const char32_t EGYPTIAN_HIEROGLYPH_13EA6 = 0x00013EA6;
const char32_t EGYPTIAN_HIEROGLYPH_13EA7 = 0x00013EA7;
const char32_t EGYPTIAN_HIEROGLYPH_13EA8 = 0x00013EA8;
const char32_t EGYPTIAN_HIEROGLYPH_13EA9 = 0x00013EA9;
const char32_t EGYPTIAN_HIEROGLYPH_13EAA = 0x00013EAA;
const char32_t EGYPTIAN_HIEROGLYPH_13EAB = 0x00013EAB;
const char32_t EGYPTIAN_HIEROGLYPH_13EAC = 0x00013EAC;
const char32_t EGYPTIAN_HIEROGLYPH_13EAD = 0x00013EAD;
const char32_t EGYPTIAN_HIEROGLYPH_13EAE = 0x00013EAE;
const char32_t EGYPTIAN_HIEROGLYPH_13EAF = 0x00013EAF;
const char32_t EGYPTIAN_HIEROGLYPH_13EB0 = 0x00013EB0;
const char32_t EGYPTIAN_HIEROGLYPH_13EB1 = 0x00013EB1;
const char32_t EGYPTIAN_HIEROGLYPH_13EB2 = 0x00013EB2;
const char32_t EGYPTIAN_HIEROGLYPH_13EB3 = 0x00013EB3;
const char32_t EGYPTIAN_HIEROGLYPH_13EB4 = 0x00013EB4;
const char32_t EGYPTIAN_HIEROGLYPH_13EB5 = 0x00013EB5;
const char32_t EGYPTIAN_HIEROGLYPH_13EB6 = 0x00013EB6;
const char32_t EGYPTIAN_HIEROGLYPH_13EB7 = 0x00013EB7;
const char32_t EGYPTIAN_HIEROGLYPH_13EB8 = 0x00013EB8;
const char32_t EGYPTIAN_HIEROGLYPH_13EB9 = 0x00013EB9;
const char32_t EGYPTIAN_HIEROGLYPH_13EBA = 0x00013EBA;
const char32_t EGYPTIAN_HIEROGLYPH_13EBB = 0x00013EBB;
const char32_t EGYPTIAN_HIEROGLYPH_13EBC = 0x00013EBC;
const char32_t EGYPTIAN_HIEROGLYPH_13EBD = 0x00013EBD;
const char32_t EGYPTIAN_HIEROGLYPH_13EBE = 0x00013EBE;
const char32_t EGYPTIAN_HIEROGLYPH_13EBF = 0x00013EBF;
const char32_t EGYPTIAN_HIEROGLYPH_13EC0 = 0x00013EC0;
const char32_t EGYPTIAN_HIEROGLYPH_13EC1 = 0x00013EC1;
const char32_t EGYPTIAN_HIEROGLYPH_13EC2 = 0x00013EC2;
const char32_t EGYPTIAN_HIEROGLYPH_13EC3 = 0x00013EC3;
const char32_t EGYPTIAN_HIEROGLYPH_13EC4 = 0x00013EC4;
const char32_t EGYPTIAN_HIEROGLYPH_13EC5 = 0x00013EC5;
const char32_t EGYPTIAN_HIEROGLYPH_13EC6 = 0x00013EC6;
const char32_t EGYPTIAN_HIEROGLYPH_13EC7 = 0x00013EC7;
const char32_t EGYPTIAN_HIEROGLYPH_13EC8 = 0x00013EC8;
const char32_t EGYPTIAN_HIEROGLYPH_13EC9 = 0x00013EC9;
const char32_t EGYPTIAN_HIEROGLYPH_13ECA = 0x00013ECA;
const char32_t EGYPTIAN_HIEROGLYPH_13ECB = 0x00013ECB;
const char32_t EGYPTIAN_HIEROGLYPH_13ECC = 0x00013ECC;
const char32_t EGYPTIAN_HIEROGLYPH_13ECD = 0x00013ECD;
const char32_t EGYPTIAN_HIEROGLYPH_13ECE = 0x00013ECE;
const char32_t EGYPTIAN_HIEROGLYPH_13ECF = 0x00013ECF;
const char32_t EGYPTIAN_HIEROGLYPH_13ED0 = 0x00013ED0;
const char32_t EGYPTIAN_HIEROGLYPH_13ED1 = 0x00013ED1;
const char32_t EGYPTIAN_HIEROGLYPH_13ED2 = 0x00013ED2;
const char32_t EGYPTIAN_HIEROGLYPH_13ED3 = 0x00013ED3;
const char32_t EGYPTIAN_HIEROGLYPH_13ED4 = 0x00013ED4;
const char32_t EGYPTIAN_HIEROGLYPH_13ED5 = 0x00013ED5;
const char32_t EGYPTIAN_HIEROGLYPH_13ED6 = 0x00013ED6;
const char32_t EGYPTIAN_HIEROGLYPH_13ED7 = 0x00013ED7;
const char32_t EGYPTIAN_HIEROGLYPH_13ED8 = 0x00013ED8;
const char32_t EGYPTIAN_HIEROGLYPH_13ED9 = 0x00013ED9;
const char32_t EGYPTIAN_HIEROGLYPH_13EDA = 0x00013EDA;
const char32_t EGYPTIAN_HIEROGLYPH_13EDB = 0x00013EDB;
const char32_t EGYPTIAN_HIEROGLYPH_13EDC = 0x00013EDC;
const char32_t EGYPTIAN_HIEROGLYPH_13EDD = 0x00013EDD;
const char32_t EGYPTIAN_HIEROGLYPH_13EDE = 0x00013EDE;
const char32_t EGYPTIAN_HIEROGLYPH_13EDF = 0x00013EDF;
const char32_t EGYPTIAN_HIEROGLYPH_13EE0 = 0x00013EE0;
const char32_t EGYPTIAN_HIEROGLYPH_13EE1 = 0x00013EE1;
const char32_t EGYPTIAN_HIEROGLYPH_13EE2 = 0x00013EE2;
const char32_t EGYPTIAN_HIEROGLYPH_13EE3 = 0x00013EE3;
const char32_t EGYPTIAN_HIEROGLYPH_13EE4 = 0x00013EE4;
const char32_t EGYPTIAN_HIEROGLYPH_13EE5 = 0x00013EE5;
const char32_t EGYPTIAN_HIEROGLYPH_13EE6 = 0x00013EE6;
const char32_t EGYPTIAN_HIEROGLYPH_13EE7 = 0x00013EE7;
const char32_t EGYPTIAN_HIEROGLYPH_13EE8 = 0x00013EE8;
const char32_t EGYPTIAN_HIEROGLYPH_13EE9 = 0x00013EE9;
const char32_t EGYPTIAN_HIEROGLYPH_13EEA = 0x00013EEA;
const char32_t EGYPTIAN_HIEROGLYPH_13EEB = 0x00013EEB;
const char32_t EGYPTIAN_HIEROGLYPH_13EEC = 0x00013EEC;
const char32_t EGYPTIAN_HIEROGLYPH_13EED = 0x00013EED;
const char32_t EGYPTIAN_HIEROGLYPH_13EEE = 0x00013EEE;
const char32_t EGYPTIAN_HIEROGLYPH_13EEF = 0x00013EEF;
const char32_t EGYPTIAN_HIEROGLYPH_13EF0 = 0x00013EF0;
const char32_t EGYPTIAN_HIEROGLYPH_13EF1 = 0x00013EF1;
const char32_t EGYPTIAN_HIEROGLYPH_13EF2 = 0x00013EF2;
const char32_t EGYPTIAN_HIEROGLYPH_13EF3 = 0x00013EF3;
const char32_t EGYPTIAN_HIEROGLYPH_13EF4 = 0x00013EF4;
const char32_t EGYPTIAN_HIEROGLYPH_13EF5 = 0x00013EF5;
const char32_t EGYPTIAN_HIEROGLYPH_13EF6 = 0x00013EF6;
const char32_t EGYPTIAN_HIEROGLYPH_13EF7 = 0x00013EF7;
const char32_t EGYPTIAN_HIEROGLYPH_13EF8 = 0x00013EF8;
const char32_t EGYPTIAN_HIEROGLYPH_13EF9 = 0x00013EF9;
const char32_t EGYPTIAN_HIEROGLYPH_13EFA = 0x00013EFA;
const char32_t EGYPTIAN_HIEROGLYPH_13EFB = 0x00013EFB;
const char32_t EGYPTIAN_HIEROGLYPH_13EFC = 0x00013EFC;
const char32_t EGYPTIAN_HIEROGLYPH_13EFD = 0x00013EFD;
const char32_t EGYPTIAN_HIEROGLYPH_13EFE = 0x00013EFE;
const char32_t EGYPTIAN_HIEROGLYPH_13EFF = 0x00013EFF;
const char32_t EGYPTIAN_HIEROGLYPH_13F00 = 0x00013F00;
const char32_t EGYPTIAN_HIEROGLYPH_13F01 = 0x00013F01;
const char32_t EGYPTIAN_HIEROGLYPH_13F02 = 0x00013F02;
const char32_t EGYPTIAN_HIEROGLYPH_13F03 = 0x00013F03;
const char32_t EGYPTIAN_HIEROGLYPH_13F04 = 0x00013F04;
const char32_t EGYPTIAN_HIEROGLYPH_13F05 = 0x00013F05;
const char32_t EGYPTIAN_HIEROGLYPH_13F06 = 0x00013F06;
const char32_t EGYPTIAN_HIEROGLYPH_13F07 = 0x00013F07;
const char32_t EGYPTIAN_HIEROGLYPH_13F08 = 0x00013F08;
const char32_t EGYPTIAN_HIEROGLYPH_13F09 = 0x00013F09;
const char32_t EGYPTIAN_HIEROGLYPH_13F0A = 0x00013F0A;
const char32_t EGYPTIAN_HIEROGLYPH_13F0B = 0x00013F0B;
const char32_t EGYPTIAN_HIEROGLYPH_13F0C = 0x00013F0C;
const char32_t EGYPTIAN_HIEROGLYPH_13F0D = 0x00013F0D;
const char32_t EGYPTIAN_HIEROGLYPH_13F0E = 0x00013F0E;
const char32_t EGYPTIAN_HIEROGLYPH_13F0F = 0x00013F0F;
const char32_t EGYPTIAN_HIEROGLYPH_13F10 = 0x00013F10;
const char32_t EGYPTIAN_HIEROGLYPH_13F11 = 0x00013F11;
const char32_t EGYPTIAN_HIEROGLYPH_13F12 = 0x00013F12;
const char32_t EGYPTIAN_HIEROGLYPH_13F13 = 0x00013F13;
const char32_t EGYPTIAN_HIEROGLYPH_13F14 = 0x00013F14;
const char32_t EGYPTIAN_HIEROGLYPH_13F15 = 0x00013F15;
const char32_t EGYPTIAN_HIEROGLYPH_13F16 = 0x00013F16;
const char32_t EGYPTIAN_HIEROGLYPH_13F17 = 0x00013F17;
const char32_t EGYPTIAN_HIEROGLYPH_13F18 = 0x00013F18;
const char32_t EGYPTIAN_HIEROGLYPH_13F19 = 0x00013F19;
const char32_t EGYPTIAN_HIEROGLYPH_13F1A = 0x00013F1A;
const char32_t EGYPTIAN_HIEROGLYPH_13F1B = 0x00013F1B;
const char32_t EGYPTIAN_HIEROGLYPH_13F1C = 0x00013F1C;
const char32_t EGYPTIAN_HIEROGLYPH_13F1D = 0x00013F1D;
const char32_t EGYPTIAN_HIEROGLYPH_13F1E = 0x00013F1E;
const char32_t EGYPTIAN_HIEROGLYPH_13F1F = 0x00013F1F;
const char32_t EGYPTIAN_HIEROGLYPH_13F20 = 0x00013F20;
const char32_t EGYPTIAN_HIEROGLYPH_13F21 = 0x00013F21;
const char32_t EGYPTIAN_HIEROGLYPH_13F22 = 0x00013F22;
const char32_t EGYPTIAN_HIEROGLYPH_13F23 = 0x00013F23;
const char32_t EGYPTIAN_HIEROGLYPH_13F24 = 0x00013F24;
const char32_t EGYPTIAN_HIEROGLYPH_13F25 = 0x00013F25;
const char32_t EGYPTIAN_HIEROGLYPH_13F26 = 0x00013F26;
const char32_t EGYPTIAN_HIEROGLYPH_13F27 = 0x00013F27;
const char32_t EGYPTIAN_HIEROGLYPH_13F28 = 0x00013F28;
const char32_t EGYPTIAN_HIEROGLYPH_13F29 = 0x00013F29;
const char32_t EGYPTIAN_HIEROGLYPH_13F2A = 0x00013F2A;
const char32_t EGYPTIAN_HIEROGLYPH_13F2B = 0x00013F2B;
const char32_t EGYPTIAN_HIEROGLYPH_13F2C = 0x00013F2C;
const char32_t EGYPTIAN_HIEROGLYPH_13F2D = 0x00013F2D;
const char32_t EGYPTIAN_HIEROGLYPH_13F2E = 0x00013F2E;
const char32_t EGYPTIAN_HIEROGLYPH_13F2F = 0x00013F2F;
const char32_t EGYPTIAN_HIEROGLYPH_13F30 = 0x00013F30;
const char32_t EGYPTIAN_HIEROGLYPH_13F31 = 0x00013F31;
const char32_t EGYPTIAN_HIEROGLYPH_13F32 = 0x00013F32;
const char32_t EGYPTIAN_HIEROGLYPH_13F33 = 0x00013F33;
const char32_t EGYPTIAN_HIEROGLYPH_13F34 = 0x00013F34;
const char32_t EGYPTIAN_HIEROGLYPH_13F35 = 0x00013F35;
const char32_t EGYPTIAN_HIEROGLYPH_13F36 = 0x00013F36;
const char32_t EGYPTIAN_HIEROGLYPH_13F37 = 0x00013F37;
const char32_t EGYPTIAN_HIEROGLYPH_13F38 = 0x00013F38;
const char32_t EGYPTIAN_HIEROGLYPH_13F39 = 0x00013F39;
const char32_t EGYPTIAN_HIEROGLYPH_13F3A = 0x00013F3A;
const char32_t EGYPTIAN_HIEROGLYPH_13F3B = 0x00013F3B;
const char32_t EGYPTIAN_HIEROGLYPH_13F3C = 0x00013F3C;
const char32_t EGYPTIAN_HIEROGLYPH_13F3D = 0x00013F3D;
const char32_t EGYPTIAN_HIEROGLYPH_13F3E = 0x00013F3E;
const char32_t EGYPTIAN_HIEROGLYPH_13F3F = 0x00013F3F;
const char32_t EGYPTIAN_HIEROGLYPH_13F40 = 0x00013F40;
const char32_t EGYPTIAN_HIEROGLYPH_13F41 = 0x00013F41;
const char32_t EGYPTIAN_HIEROGLYPH_13F42 = 0x00013F42;
const char32_t EGYPTIAN_HIEROGLYPH_13F43 = 0x00013F43;
const char32_t EGYPTIAN_HIEROGLYPH_13F44 = 0x00013F44;
const char32_t EGYPTIAN_HIEROGLYPH_13F45 = 0x00013F45;
const char32_t EGYPTIAN_HIEROGLYPH_13F46 = 0x00013F46;
const char32_t EGYPTIAN_HIEROGLYPH_13F47 = 0x00013F47;
const char32_t EGYPTIAN_HIEROGLYPH_13F48 = 0x00013F48;
const char32_t EGYPTIAN_HIEROGLYPH_13F49 = 0x00013F49;
const char32_t EGYPTIAN_HIEROGLYPH_13F4A = 0x00013F4A;
const char32_t EGYPTIAN_HIEROGLYPH_13F4B = 0x00013F4B;
const char32_t EGYPTIAN_HIEROGLYPH_13F4C = 0x00013F4C;
const char32_t EGYPTIAN_HIEROGLYPH_13F4D = 0x00013F4D;
const char32_t EGYPTIAN_HIEROGLYPH_13F4E = 0x00013F4E;
const char32_t EGYPTIAN_HIEROGLYPH_13F4F = 0x00013F4F;
const char32_t EGYPTIAN_HIEROGLYPH_13F50 = 0x00013F50;
const char32_t EGYPTIAN_HIEROGLYPH_13F51 = 0x00013F51;
const char32_t EGYPTIAN_HIEROGLYPH_13F52 = 0x00013F52;
const char32_t EGYPTIAN_HIEROGLYPH_13F53 = 0x00013F53;
const char32_t EGYPTIAN_HIEROGLYPH_13F54 = 0x00013F54;
const char32_t EGYPTIAN_HIEROGLYPH_13F55 = 0x00013F55;
const char32_t EGYPTIAN_HIEROGLYPH_13F56 = 0x00013F56;
const char32_t EGYPTIAN_HIEROGLYPH_13F57 = 0x00013F57;
const char32_t EGYPTIAN_HIEROGLYPH_13F58 = 0x00013F58;
const char32_t EGYPTIAN_HIEROGLYPH_13F59 = 0x00013F59;
const char32_t EGYPTIAN_HIEROGLYPH_13F5A = 0x00013F5A;
const char32_t EGYPTIAN_HIEROGLYPH_13F5B = 0x00013F5B;
const char32_t EGYPTIAN_HIEROGLYPH_13F5C = 0x00013F5C;
const char32_t EGYPTIAN_HIEROGLYPH_13F5D = 0x00013F5D;
const char32_t EGYPTIAN_HIEROGLYPH_13F5E = 0x00013F5E;
const char32_t EGYPTIAN_HIEROGLYPH_13F5F = 0x00013F5F;
const char32_t EGYPTIAN_HIEROGLYPH_13F60 = 0x00013F60;
const char32_t EGYPTIAN_HIEROGLYPH_13F61 = 0x00013F61;
const char32_t EGYPTIAN_HIEROGLYPH_13F62 = 0x00013F62;
const char32_t EGYPTIAN_HIEROGLYPH_13F63 = 0x00013F63;
const char32_t EGYPTIAN_HIEROGLYPH_13F64 = 0x00013F64;
const char32_t EGYPTIAN_HIEROGLYPH_13F65 = 0x00013F65;
const char32_t EGYPTIAN_HIEROGLYPH_13F66 = 0x00013F66;
const char32_t EGYPTIAN_HIEROGLYPH_13F67 = 0x00013F67;
const char32_t EGYPTIAN_HIEROGLYPH_13F68 = 0x00013F68;
const char32_t EGYPTIAN_HIEROGLYPH_13F69 = 0x00013F69;
const char32_t EGYPTIAN_HIEROGLYPH_13F6A = 0x00013F6A;
const char32_t EGYPTIAN_HIEROGLYPH_13F6B = 0x00013F6B;
const char32_t EGYPTIAN_HIEROGLYPH_13F6C = 0x00013F6C;
const char32_t EGYPTIAN_HIEROGLYPH_13F6D = 0x00013F6D;
const char32_t EGYPTIAN_HIEROGLYPH_13F6E = 0x00013F6E;
const char32_t EGYPTIAN_HIEROGLYPH_13F6F = 0x00013F6F;
const char32_t EGYPTIAN_HIEROGLYPH_13F70 = 0x00013F70;
const char32_t EGYPTIAN_HIEROGLYPH_13F71 = 0x00013F71;
const char32_t EGYPTIAN_HIEROGLYPH_13F72 = 0x00013F72;
const char32_t EGYPTIAN_HIEROGLYPH_13F73 = 0x00013F73;
const char32_t EGYPTIAN_HIEROGLYPH_13F74 = 0x00013F74;
const char32_t EGYPTIAN_HIEROGLYPH_13F75 = 0x00013F75;
const char32_t EGYPTIAN_HIEROGLYPH_13F76 = 0x00013F76;
const char32_t EGYPTIAN_HIEROGLYPH_13F77 = 0x00013F77;
const char32_t EGYPTIAN_HIEROGLYPH_13F78 = 0x00013F78;
const char32_t EGYPTIAN_HIEROGLYPH_13F79 = 0x00013F79;
const char32_t EGYPTIAN_HIEROGLYPH_13F7A = 0x00013F7A;
const char32_t EGYPTIAN_HIEROGLYPH_13F7B = 0x00013F7B;
const char32_t EGYPTIAN_HIEROGLYPH_13F7C = 0x00013F7C;
const char32_t EGYPTIAN_HIEROGLYPH_13F7D = 0x00013F7D;
const char32_t EGYPTIAN_HIEROGLYPH_13F7E = 0x00013F7E;
const char32_t EGYPTIAN_HIEROGLYPH_13F7F = 0x00013F7F;
const char32_t EGYPTIAN_HIEROGLYPH_13F80 = 0x00013F80;
const char32_t EGYPTIAN_HIEROGLYPH_13F81 = 0x00013F81;
const char32_t EGYPTIAN_HIEROGLYPH_13F82 = 0x00013F82;
const char32_t EGYPTIAN_HIEROGLYPH_13F83 = 0x00013F83;
const char32_t EGYPTIAN_HIEROGLYPH_13F84 = 0x00013F84;
const char32_t EGYPTIAN_HIEROGLYPH_13F85 = 0x00013F85;
const char32_t EGYPTIAN_HIEROGLYPH_13F86 = 0x00013F86;
const char32_t EGYPTIAN_HIEROGLYPH_13F87 = 0x00013F87;
const char32_t EGYPTIAN_HIEROGLYPH_13F88 = 0x00013F88;
const char32_t EGYPTIAN_HIEROGLYPH_13F89 = 0x00013F89;
const char32_t EGYPTIAN_HIEROGLYPH_13F8A = 0x00013F8A;
const char32_t EGYPTIAN_HIEROGLYPH_13F8B = 0x00013F8B;
const char32_t EGYPTIAN_HIEROGLYPH_13F8C = 0x00013F8C;
const char32_t EGYPTIAN_HIEROGLYPH_13F8D = 0x00013F8D;
const char32_t EGYPTIAN_HIEROGLYPH_13F8E = 0x00013F8E;
const char32_t EGYPTIAN_HIEROGLYPH_13F8F = 0x00013F8F;
const char32_t EGYPTIAN_HIEROGLYPH_13F90 = 0x00013F90;
const char32_t EGYPTIAN_HIEROGLYPH_13F91 = 0x00013F91;
const char32_t EGYPTIAN_HIEROGLYPH_13F92 = 0x00013F92;
const char32_t EGYPTIAN_HIEROGLYPH_13F93 = 0x00013F93;
const char32_t EGYPTIAN_HIEROGLYPH_13F94 = 0x00013F94;
const char32_t EGYPTIAN_HIEROGLYPH_13F95 = 0x00013F95;
const char32_t EGYPTIAN_HIEROGLYPH_13F96 = 0x00013F96;
const char32_t EGYPTIAN_HIEROGLYPH_13F97 = 0x00013F97;
const char32_t EGYPTIAN_HIEROGLYPH_13F98 = 0x00013F98;
const char32_t EGYPTIAN_HIEROGLYPH_13F99 = 0x00013F99;
const char32_t EGYPTIAN_HIEROGLYPH_13F9A = 0x00013F9A;
const char32_t EGYPTIAN_HIEROGLYPH_13F9B = 0x00013F9B;
const char32_t EGYPTIAN_HIEROGLYPH_13F9C = 0x00013F9C;
const char32_t EGYPTIAN_HIEROGLYPH_13F9D = 0x00013F9D;
const char32_t EGYPTIAN_HIEROGLYPH_13F9E = 0x00013F9E;
const char32_t EGYPTIAN_HIEROGLYPH_13F9F = 0x00013F9F;
const char32_t EGYPTIAN_HIEROGLYPH_13FA0 = 0x00013FA0;
const char32_t EGYPTIAN_HIEROGLYPH_13FA1 = 0x00013FA1;
const char32_t EGYPTIAN_HIEROGLYPH_13FA2 = 0x00013FA2;
const char32_t EGYPTIAN_HIEROGLYPH_13FA3 = 0x00013FA3;
const char32_t EGYPTIAN_HIEROGLYPH_13FA4 = 0x00013FA4;
const char32_t EGYPTIAN_HIEROGLYPH_13FA5 = 0x00013FA5;
const char32_t EGYPTIAN_HIEROGLYPH_13FA6 = 0x00013FA6;
const char32_t EGYPTIAN_HIEROGLYPH_13FA7 = 0x00013FA7;
const char32_t EGYPTIAN_HIEROGLYPH_13FA8 = 0x00013FA8;
const char32_t EGYPTIAN_HIEROGLYPH_13FA9 = 0x00013FA9;
const char32_t EGYPTIAN_HIEROGLYPH_13FAA = 0x00013FAA;
const char32_t EGYPTIAN_HIEROGLYPH_13FAB = 0x00013FAB;
const char32_t EGYPTIAN_HIEROGLYPH_13FAC = 0x00013FAC;
const char32_t EGYPTIAN_HIEROGLYPH_13FAD = 0x00013FAD;
const char32_t EGYPTIAN_HIEROGLYPH_13FAE = 0x00013FAE;
const char32_t EGYPTIAN_HIEROGLYPH_13FAF = 0x00013FAF;
const char32_t EGYPTIAN_HIEROGLYPH_13FB0 = 0x00013FB0;
const char32_t EGYPTIAN_HIEROGLYPH_13FB1 = 0x00013FB1;
const char32_t EGYPTIAN_HIEROGLYPH_13FB2 = 0x00013FB2;
const char32_t EGYPTIAN_HIEROGLYPH_13FB3 = 0x00013FB3;
const char32_t EGYPTIAN_HIEROGLYPH_13FB4 = 0x00013FB4;
const char32_t EGYPTIAN_HIEROGLYPH_13FB5 = 0x00013FB5;
const char32_t EGYPTIAN_HIEROGLYPH_13FB6 = 0x00013FB6;
const char32_t EGYPTIAN_HIEROGLYPH_13FB7 = 0x00013FB7;
const char32_t EGYPTIAN_HIEROGLYPH_13FB8 = 0x00013FB8;
const char32_t EGYPTIAN_HIEROGLYPH_13FB9 = 0x00013FB9;
const char32_t EGYPTIAN_HIEROGLYPH_13FBA = 0x00013FBA;
const char32_t EGYPTIAN_HIEROGLYPH_13FBB = 0x00013FBB;
const char32_t EGYPTIAN_HIEROGLYPH_13FBC = 0x00013FBC;
const char32_t EGYPTIAN_HIEROGLYPH_13FBD = 0x00013FBD;
const char32_t EGYPTIAN_HIEROGLYPH_13FBE = 0x00013FBE;
const char32_t EGYPTIAN_HIEROGLYPH_13FBF = 0x00013FBF;
const char32_t EGYPTIAN_HIEROGLYPH_13FC0 = 0x00013FC0;
const char32_t EGYPTIAN_HIEROGLYPH_13FC1 = 0x00013FC1;
const char32_t EGYPTIAN_HIEROGLYPH_13FC2 = 0x00013FC2;
const char32_t EGYPTIAN_HIEROGLYPH_13FC3 = 0x00013FC3;
const char32_t EGYPTIAN_HIEROGLYPH_13FC4 = 0x00013FC4;
const char32_t EGYPTIAN_HIEROGLYPH_13FC5 = 0x00013FC5;
const char32_t EGYPTIAN_HIEROGLYPH_13FC6 = 0x00013FC6;
const char32_t EGYPTIAN_HIEROGLYPH_13FC7 = 0x00013FC7;
const char32_t EGYPTIAN_HIEROGLYPH_13FC8 = 0x00013FC8;
const char32_t EGYPTIAN_HIEROGLYPH_13FC9 = 0x00013FC9;
const char32_t EGYPTIAN_HIEROGLYPH_13FCA = 0x00013FCA;
const char32_t EGYPTIAN_HIEROGLYPH_13FCB = 0x00013FCB;
const char32_t EGYPTIAN_HIEROGLYPH_13FCC = 0x00013FCC;
const char32_t EGYPTIAN_HIEROGLYPH_13FCD = 0x00013FCD;
const char32_t EGYPTIAN_HIEROGLYPH_13FCE = 0x00013FCE;
const char32_t EGYPTIAN_HIEROGLYPH_13FCF = 0x00013FCF;
const char32_t EGYPTIAN_HIEROGLYPH_13FD0 = 0x00013FD0;
const char32_t EGYPTIAN_HIEROGLYPH_13FD1 = 0x00013FD1;
const char32_t EGYPTIAN_HIEROGLYPH_13FD2 = 0x00013FD2;
const char32_t EGYPTIAN_HIEROGLYPH_13FD3 = 0x00013FD3;
const char32_t EGYPTIAN_HIEROGLYPH_13FD4 = 0x00013FD4;
const char32_t EGYPTIAN_HIEROGLYPH_13FD5 = 0x00013FD5;
const char32_t EGYPTIAN_HIEROGLYPH_13FD6 = 0x00013FD6;
const char32_t EGYPTIAN_HIEROGLYPH_13FD7 = 0x00013FD7;
const char32_t EGYPTIAN_HIEROGLYPH_13FD8 = 0x00013FD8;
const char32_t EGYPTIAN_HIEROGLYPH_13FD9 = 0x00013FD9;
const char32_t EGYPTIAN_HIEROGLYPH_13FDA = 0x00013FDA;
const char32_t EGYPTIAN_HIEROGLYPH_13FDB = 0x00013FDB;
const char32_t EGYPTIAN_HIEROGLYPH_13FDC = 0x00013FDC;
const char32_t EGYPTIAN_HIEROGLYPH_13FDD = 0x00013FDD;
const char32_t EGYPTIAN_HIEROGLYPH_13FDE = 0x00013FDE;
const char32_t EGYPTIAN_HIEROGLYPH_13FDF = 0x00013FDF;
const char32_t EGYPTIAN_HIEROGLYPH_13FE0 = 0x00013FE0;
const char32_t EGYPTIAN_HIEROGLYPH_13FE1 = 0x00013FE1;
const char32_t EGYPTIAN_HIEROGLYPH_13FE2 = 0x00013FE2;
const char32_t EGYPTIAN_HIEROGLYPH_13FE3 = 0x00013FE3;
const char32_t EGYPTIAN_HIEROGLYPH_13FE4 = 0x00013FE4;
const char32_t EGYPTIAN_HIEROGLYPH_13FE5 = 0x00013FE5;
const char32_t EGYPTIAN_HIEROGLYPH_13FE6 = 0x00013FE6;
const char32_t EGYPTIAN_HIEROGLYPH_13FE7 = 0x00013FE7;
const char32_t EGYPTIAN_HIEROGLYPH_13FE8 = 0x00013FE8;
const char32_t EGYPTIAN_HIEROGLYPH_13FE9 = 0x00013FE9;
const char32_t EGYPTIAN_HIEROGLYPH_13FEA = 0x00013FEA;
const char32_t EGYPTIAN_HIEROGLYPH_13FEB = 0x00013FEB;
const char32_t EGYPTIAN_HIEROGLYPH_13FEC = 0x00013FEC;
const char32_t EGYPTIAN_HIEROGLYPH_13FED = 0x00013FED;
const char32_t EGYPTIAN_HIEROGLYPH_13FEE = 0x00013FEE;
const char32_t EGYPTIAN_HIEROGLYPH_13FEF = 0x00013FEF;
const char32_t EGYPTIAN_HIEROGLYPH_13FF0 = 0x00013FF0;
const char32_t EGYPTIAN_HIEROGLYPH_13FF1 = 0x00013FF1;
const char32_t EGYPTIAN_HIEROGLYPH_13FF2 = 0x00013FF2;
const char32_t EGYPTIAN_HIEROGLYPH_13FF3 = 0x00013FF3;
const char32_t EGYPTIAN_HIEROGLYPH_13FF4 = 0x00013FF4;
const char32_t EGYPTIAN_HIEROGLYPH_13FF5 = 0x00013FF5;
const char32_t EGYPTIAN_HIEROGLYPH_13FF6 = 0x00013FF6;
const char32_t EGYPTIAN_HIEROGLYPH_13FF7 = 0x00013FF7;
const char32_t EGYPTIAN_HIEROGLYPH_13FF8 = 0x00013FF8;
const char32_t EGYPTIAN_HIEROGLYPH_13FF9 = 0x00013FF9;
const char32_t EGYPTIAN_HIEROGLYPH_13FFA = 0x00013FFA;
const char32_t EGYPTIAN_HIEROGLYPH_13FFB = 0x00013FFB;
const char32_t EGYPTIAN_HIEROGLYPH_13FFC = 0x00013FFC;
const char32_t EGYPTIAN_HIEROGLYPH_13FFD = 0x00013FFD;
const char32_t EGYPTIAN_HIEROGLYPH_13FFE = 0x00013FFE;
const char32_t EGYPTIAN_HIEROGLYPH_13FFF = 0x00013FFF;
const char32_t EGYPTIAN_HIEROGLYPH_14000 = 0x00014000;
const char32_t EGYPTIAN_HIEROGLYPH_14001 = 0x00014001;
const char32_t EGYPTIAN_HIEROGLYPH_14002 = 0x00014002;
const char32_t EGYPTIAN_HIEROGLYPH_14003 = 0x00014003;
const char32_t EGYPTIAN_HIEROGLYPH_14004 = 0x00014004;
const char32_t EGYPTIAN_HIEROGLYPH_14005 = 0x00014005;
const char32_t EGYPTIAN_HIEROGLYPH_14006 = 0x00014006;
const char32_t EGYPTIAN_HIEROGLYPH_14007 = 0x00014007;
const char32_t EGYPTIAN_HIEROGLYPH_14008 = 0x00014008;
const char32_t EGYPTIAN_HIEROGLYPH_14009 = 0x00014009;
const char32_t EGYPTIAN_HIEROGLYPH_1400A = 0x0001400A;
const char32_t EGYPTIAN_HIEROGLYPH_1400B = 0x0001400B;
const char32_t EGYPTIAN_HIEROGLYPH_1400C = 0x0001400C;
const char32_t EGYPTIAN_HIEROGLYPH_1400D = 0x0001400D;
const char32_t EGYPTIAN_HIEROGLYPH_1400E = 0x0001400E;
const char32_t EGYPTIAN_HIEROGLYPH_1400F = 0x0001400F;
const char32_t EGYPTIAN_HIEROGLYPH_14010 = 0x00014010;
const char32_t EGYPTIAN_HIEROGLYPH_14011 = 0x00014011;
const char32_t EGYPTIAN_HIEROGLYPH_14012 = 0x00014012;
const char32_t EGYPTIAN_HIEROGLYPH_14013 = 0x00014013;
const char32_t EGYPTIAN_HIEROGLYPH_14014 = 0x00014014;
const char32_t EGYPTIAN_HIEROGLYPH_14015 = 0x00014015;
const char32_t EGYPTIAN_HIEROGLYPH_14016 = 0x00014016;
const char32_t EGYPTIAN_HIEROGLYPH_14017 = 0x00014017;
const char32_t EGYPTIAN_HIEROGLYPH_14018 = 0x00014018;
const char32_t EGYPTIAN_HIEROGLYPH_14019 = 0x00014019;
const char32_t EGYPTIAN_HIEROGLYPH_1401A = 0x0001401A;
const char32_t EGYPTIAN_HIEROGLYPH_1401B = 0x0001401B;
const char32_t EGYPTIAN_HIEROGLYPH_1401C = 0x0001401C;
const char32_t EGYPTIAN_HIEROGLYPH_1401D = 0x0001401D;
const char32_t EGYPTIAN_HIEROGLYPH_1401E = 0x0001401E;
const char32_t EGYPTIAN_HIEROGLYPH_1401F = 0x0001401F;
const char32_t EGYPTIAN_HIEROGLYPH_14020 = 0x00014020;
const char32_t EGYPTIAN_HIEROGLYPH_14021 = 0x00014021;
const char32_t EGYPTIAN_HIEROGLYPH_14022 = 0x00014022;
const char32_t EGYPTIAN_HIEROGLYPH_14023 = 0x00014023;
const char32_t EGYPTIAN_HIEROGLYPH_14024 = 0x00014024;
const char32_t EGYPTIAN_HIEROGLYPH_14025 = 0x00014025;
const char32_t EGYPTIAN_HIEROGLYPH_14026 = 0x00014026;
const char32_t EGYPTIAN_HIEROGLYPH_14027 = 0x00014027;
const char32_t EGYPTIAN_HIEROGLYPH_14028 = 0x00014028;
const char32_t EGYPTIAN_HIEROGLYPH_14029 = 0x00014029;
const char32_t EGYPTIAN_HIEROGLYPH_1402A = 0x0001402A;
const char32_t EGYPTIAN_HIEROGLYPH_1402B = 0x0001402B;
const char32_t EGYPTIAN_HIEROGLYPH_1402C = 0x0001402C;
const char32_t EGYPTIAN_HIEROGLYPH_1402D = 0x0001402D;
const char32_t EGYPTIAN_HIEROGLYPH_1402E = 0x0001402E;
const char32_t EGYPTIAN_HIEROGLYPH_1402F = 0x0001402F;
const char32_t EGYPTIAN_HIEROGLYPH_14030 = 0x00014030;
const char32_t EGYPTIAN_HIEROGLYPH_14031 = 0x00014031;
const char32_t EGYPTIAN_HIEROGLYPH_14032 = 0x00014032;
const char32_t EGYPTIAN_HIEROGLYPH_14033 = 0x00014033;
const char32_t EGYPTIAN_HIEROGLYPH_14034 = 0x00014034;
const char32_t EGYPTIAN_HIEROGLYPH_14035 = 0x00014035;
const char32_t EGYPTIAN_HIEROGLYPH_14036 = 0x00014036;
const char32_t EGYPTIAN_HIEROGLYPH_14037 = 0x00014037;
const char32_t EGYPTIAN_HIEROGLYPH_14038 = 0x00014038;
const char32_t EGYPTIAN_HIEROGLYPH_14039 = 0x00014039;
const char32_t EGYPTIAN_HIEROGLYPH_1403A = 0x0001403A;
const char32_t EGYPTIAN_HIEROGLYPH_1403B = 0x0001403B;
const char32_t EGYPTIAN_HIEROGLYPH_1403C = 0x0001403C;
const char32_t EGYPTIAN_HIEROGLYPH_1403D = 0x0001403D;
const char32_t EGYPTIAN_HIEROGLYPH_1403E = 0x0001403E;
const char32_t EGYPTIAN_HIEROGLYPH_1403F = 0x0001403F;
const char32_t EGYPTIAN_HIEROGLYPH_14040 = 0x00014040;
const char32_t EGYPTIAN_HIEROGLYPH_14041 = 0x00014041;
const char32_t EGYPTIAN_HIEROGLYPH_14042 = 0x00014042;
const char32_t EGYPTIAN_HIEROGLYPH_14043 = 0x00014043;
const char32_t EGYPTIAN_HIEROGLYPH_14044 = 0x00014044;
const char32_t EGYPTIAN_HIEROGLYPH_14045 = 0x00014045;
const char32_t EGYPTIAN_HIEROGLYPH_14046 = 0x00014046;
const char32_t EGYPTIAN_HIEROGLYPH_14047 = 0x00014047;
const char32_t EGYPTIAN_HIEROGLYPH_14048 = 0x00014048;
const char32_t EGYPTIAN_HIEROGLYPH_14049 = 0x00014049;
const char32_t EGYPTIAN_HIEROGLYPH_1404A = 0x0001404A;
const char32_t EGYPTIAN_HIEROGLYPH_1404B = 0x0001404B;
const char32_t EGYPTIAN_HIEROGLYPH_1404C = 0x0001404C;
const char32_t EGYPTIAN_HIEROGLYPH_1404D = 0x0001404D;
const char32_t EGYPTIAN_HIEROGLYPH_1404E = 0x0001404E;
const char32_t EGYPTIAN_HIEROGLYPH_1404F = 0x0001404F;
const char32_t EGYPTIAN_HIEROGLYPH_14050 = 0x00014050;
const char32_t EGYPTIAN_HIEROGLYPH_14051 = 0x00014051;
const char32_t EGYPTIAN_HIEROGLYPH_14052 = 0x00014052;
const char32_t EGYPTIAN_HIEROGLYPH_14053 = 0x00014053;
const char32_t EGYPTIAN_HIEROGLYPH_14054 = 0x00014054;
const char32_t EGYPTIAN_HIEROGLYPH_14055 = 0x00014055;
const char32_t EGYPTIAN_HIEROGLYPH_14056 = 0x00014056;
const char32_t EGYPTIAN_HIEROGLYPH_14057 = 0x00014057;
const char32_t EGYPTIAN_HIEROGLYPH_14058 = 0x00014058;
const char32_t EGYPTIAN_HIEROGLYPH_14059 = 0x00014059;
const char32_t EGYPTIAN_HIEROGLYPH_1405A = 0x0001405A;
const char32_t EGYPTIAN_HIEROGLYPH_1405B = 0x0001405B;
const char32_t EGYPTIAN_HIEROGLYPH_1405C = 0x0001405C;
const char32_t EGYPTIAN_HIEROGLYPH_1405D = 0x0001405D;
const char32_t EGYPTIAN_HIEROGLYPH_1405E = 0x0001405E;
const char32_t EGYPTIAN_HIEROGLYPH_1405F = 0x0001405F;
const char32_t EGYPTIAN_HIEROGLYPH_14060 = 0x00014060;
const char32_t EGYPTIAN_HIEROGLYPH_14061 = 0x00014061;
const char32_t EGYPTIAN_HIEROGLYPH_14062 = 0x00014062;
const char32_t EGYPTIAN_HIEROGLYPH_14063 = 0x00014063;
const char32_t EGYPTIAN_HIEROGLYPH_14064 = 0x00014064;
const char32_t EGYPTIAN_HIEROGLYPH_14065 = 0x00014065;
const char32_t EGYPTIAN_HIEROGLYPH_14066 = 0x00014066;
const char32_t EGYPTIAN_HIEROGLYPH_14067 = 0x00014067;
const char32_t EGYPTIAN_HIEROGLYPH_14068 = 0x00014068;
const char32_t EGYPTIAN_HIEROGLYPH_14069 = 0x00014069;
const char32_t EGYPTIAN_HIEROGLYPH_1406A = 0x0001406A;
const char32_t EGYPTIAN_HIEROGLYPH_1406B = 0x0001406B;
const char32_t EGYPTIAN_HIEROGLYPH_1406C = 0x0001406C;
const char32_t EGYPTIAN_HIEROGLYPH_1406D = 0x0001406D;
const char32_t EGYPTIAN_HIEROGLYPH_1406E = 0x0001406E;
const char32_t EGYPTIAN_HIEROGLYPH_1406F = 0x0001406F;
const char32_t EGYPTIAN_HIEROGLYPH_14070 = 0x00014070;
const char32_t EGYPTIAN_HIEROGLYPH_14071 = 0x00014071;
const char32_t EGYPTIAN_HIEROGLYPH_14072 = 0x00014072;
const char32_t EGYPTIAN_HIEROGLYPH_14073 = 0x00014073;
const char32_t EGYPTIAN_HIEROGLYPH_14074 = 0x00014074;
const char32_t EGYPTIAN_HIEROGLYPH_14075 = 0x00014075;
const char32_t EGYPTIAN_HIEROGLYPH_14076 = 0x00014076;
const char32_t EGYPTIAN_HIEROGLYPH_14077 = 0x00014077;
const char32_t EGYPTIAN_HIEROGLYPH_14078 = 0x00014078;
const char32_t EGYPTIAN_HIEROGLYPH_14079 = 0x00014079;
const char32_t EGYPTIAN_HIEROGLYPH_1407A = 0x0001407A;
const char32_t EGYPTIAN_HIEROGLYPH_1407B = 0x0001407B;
const char32_t EGYPTIAN_HIEROGLYPH_1407C = 0x0001407C;
const char32_t EGYPTIAN_HIEROGLYPH_1407D = 0x0001407D;
const char32_t EGYPTIAN_HIEROGLYPH_1407E = 0x0001407E;
const char32_t EGYPTIAN_HIEROGLYPH_1407F = 0x0001407F;
const char32_t EGYPTIAN_HIEROGLYPH_14080 = 0x00014080;
const char32_t EGYPTIAN_HIEROGLYPH_14081 = 0x00014081;
const char32_t EGYPTIAN_HIEROGLYPH_14082 = 0x00014082;
const char32_t EGYPTIAN_HIEROGLYPH_14083 = 0x00014083;
const char32_t EGYPTIAN_HIEROGLYPH_14084 = 0x00014084;
const char32_t EGYPTIAN_HIEROGLYPH_14085 = 0x00014085;
const char32_t EGYPTIAN_HIEROGLYPH_14086 = 0x00014086;
const char32_t EGYPTIAN_HIEROGLYPH_14087 = 0x00014087;
const char32_t EGYPTIAN_HIEROGLYPH_14088 = 0x00014088;
const char32_t EGYPTIAN_HIEROGLYPH_14089 = 0x00014089;
const char32_t EGYPTIAN_HIEROGLYPH_1408A = 0x0001408A;
const char32_t EGYPTIAN_HIEROGLYPH_1408B = 0x0001408B;
const char32_t EGYPTIAN_HIEROGLYPH_1408C = 0x0001408C;
const char32_t EGYPTIAN_HIEROGLYPH_1408D = 0x0001408D;
const char32_t EGYPTIAN_HIEROGLYPH_1408E = 0x0001408E;
const char32_t EGYPTIAN_HIEROGLYPH_1408F = 0x0001408F;
const char32_t EGYPTIAN_HIEROGLYPH_14090 = 0x00014090;
const char32_t EGYPTIAN_HIEROGLYPH_14091 = 0x00014091;
const char32_t EGYPTIAN_HIEROGLYPH_14092 = 0x00014092;
const char32_t EGYPTIAN_HIEROGLYPH_14093 = 0x00014093;
const char32_t EGYPTIAN_HIEROGLYPH_14094 = 0x00014094;
const char32_t EGYPTIAN_HIEROGLYPH_14095 = 0x00014095;
const char32_t EGYPTIAN_HIEROGLYPH_14096 = 0x00014096;
const char32_t EGYPTIAN_HIEROGLYPH_14097 = 0x00014097;
const char32_t EGYPTIAN_HIEROGLYPH_14098 = 0x00014098;
const char32_t EGYPTIAN_HIEROGLYPH_14099 = 0x00014099;
const char32_t EGYPTIAN_HIEROGLYPH_1409A = 0x0001409A;
const char32_t EGYPTIAN_HIEROGLYPH_1409B = 0x0001409B;
const char32_t EGYPTIAN_HIEROGLYPH_1409C = 0x0001409C;
const char32_t EGYPTIAN_HIEROGLYPH_1409D = 0x0001409D;
const char32_t EGYPTIAN_HIEROGLYPH_1409E = 0x0001409E;
const char32_t EGYPTIAN_HIEROGLYPH_1409F = 0x0001409F;
const char32_t EGYPTIAN_HIEROGLYPH_140A0 = 0x000140A0;
const char32_t EGYPTIAN_HIEROGLYPH_140A1 = 0x000140A1;
const char32_t EGYPTIAN_HIEROGLYPH_140A2 = 0x000140A2;
const char32_t EGYPTIAN_HIEROGLYPH_140A3 = 0x000140A3;
const char32_t EGYPTIAN_HIEROGLYPH_140A4 = 0x000140A4;
const char32_t EGYPTIAN_HIEROGLYPH_140A5 = 0x000140A5;
const char32_t EGYPTIAN_HIEROGLYPH_140A6 = 0x000140A6;
const char32_t EGYPTIAN_HIEROGLYPH_140A7 = 0x000140A7;
const char32_t EGYPTIAN_HIEROGLYPH_140A8 = 0x000140A8;
const char32_t EGYPTIAN_HIEROGLYPH_140A9 = 0x000140A9;
const char32_t EGYPTIAN_HIEROGLYPH_140AA = 0x000140AA;
const char32_t EGYPTIAN_HIEROGLYPH_140AB = 0x000140AB;
const char32_t EGYPTIAN_HIEROGLYPH_140AC = 0x000140AC;
const char32_t EGYPTIAN_HIEROGLYPH_140AD = 0x000140AD;
const char32_t EGYPTIAN_HIEROGLYPH_140AE = 0x000140AE;
const char32_t EGYPTIAN_HIEROGLYPH_140AF = 0x000140AF;
const char32_t EGYPTIAN_HIEROGLYPH_140B0 = 0x000140B0;
const char32_t EGYPTIAN_HIEROGLYPH_140B1 = 0x000140B1;
const char32_t EGYPTIAN_HIEROGLYPH_140B2 = 0x000140B2;
const char32_t EGYPTIAN_HIEROGLYPH_140B3 = 0x000140B3;
const char32_t EGYPTIAN_HIEROGLYPH_140B4 = 0x000140B4;
const char32_t EGYPTIAN_HIEROGLYPH_140B5 = 0x000140B5;
const char32_t EGYPTIAN_HIEROGLYPH_140B6 = 0x000140B6;
const char32_t EGYPTIAN_HIEROGLYPH_140B7 = 0x000140B7;
const char32_t EGYPTIAN_HIEROGLYPH_140B8 = 0x000140B8;
const char32_t EGYPTIAN_HIEROGLYPH_140B9 = 0x000140B9;
const char32_t EGYPTIAN_HIEROGLYPH_140BA = 0x000140BA;
const char32_t EGYPTIAN_HIEROGLYPH_140BB = 0x000140BB;
const char32_t EGYPTIAN_HIEROGLYPH_140BC = 0x000140BC;
const char32_t EGYPTIAN_HIEROGLYPH_140BD = 0x000140BD;
const char32_t EGYPTIAN_HIEROGLYPH_140BE = 0x000140BE;
const char32_t EGYPTIAN_HIEROGLYPH_140BF = 0x000140BF;
const char32_t EGYPTIAN_HIEROGLYPH_140C0 = 0x000140C0;
const char32_t EGYPTIAN_HIEROGLYPH_140C1 = 0x000140C1;
const char32_t EGYPTIAN_HIEROGLYPH_140C2 = 0x000140C2;
const char32_t EGYPTIAN_HIEROGLYPH_140C3 = 0x000140C3;
const char32_t EGYPTIAN_HIEROGLYPH_140C4 = 0x000140C4;
const char32_t EGYPTIAN_HIEROGLYPH_140C5 = 0x000140C5;
const char32_t EGYPTIAN_HIEROGLYPH_140C6 = 0x000140C6;
const char32_t EGYPTIAN_HIEROGLYPH_140C7 = 0x000140C7;
const char32_t EGYPTIAN_HIEROGLYPH_140C8 = 0x000140C8;
const char32_t EGYPTIAN_HIEROGLYPH_140C9 = 0x000140C9;
const char32_t EGYPTIAN_HIEROGLYPH_140CA = 0x000140CA;
const char32_t EGYPTIAN_HIEROGLYPH_140CB = 0x000140CB;
const char32_t EGYPTIAN_HIEROGLYPH_140CC = 0x000140CC;
const char32_t EGYPTIAN_HIEROGLYPH_140CD = 0x000140CD;
const char32_t EGYPTIAN_HIEROGLYPH_140CE = 0x000140CE;
const char32_t EGYPTIAN_HIEROGLYPH_140CF = 0x000140CF;
const char32_t EGYPTIAN_HIEROGLYPH_140D0 = 0x000140D0;
const char32_t EGYPTIAN_HIEROGLYPH_140D1 = 0x000140D1;
const char32_t EGYPTIAN_HIEROGLYPH_140D2 = 0x000140D2;
const char32_t EGYPTIAN_HIEROGLYPH_140D3 = 0x000140D3;
const char32_t EGYPTIAN_HIEROGLYPH_140D4 = 0x000140D4;
const char32_t EGYPTIAN_HIEROGLYPH_140D5 = 0x000140D5;
const char32_t EGYPTIAN_HIEROGLYPH_140D6 = 0x000140D6;
const char32_t EGYPTIAN_HIEROGLYPH_140D7 = 0x000140D7;
const char32_t EGYPTIAN_HIEROGLYPH_140D8 = 0x000140D8;
const char32_t EGYPTIAN_HIEROGLYPH_140D9 = 0x000140D9;
const char32_t EGYPTIAN_HIEROGLYPH_140DA = 0x000140DA;
const char32_t EGYPTIAN_HIEROGLYPH_140DB = 0x000140DB;
const char32_t EGYPTIAN_HIEROGLYPH_140DC = 0x000140DC;
const char32_t EGYPTIAN_HIEROGLYPH_140DD = 0x000140DD;
const char32_t EGYPTIAN_HIEROGLYPH_140DE = 0x000140DE;
const char32_t EGYPTIAN_HIEROGLYPH_140DF = 0x000140DF;
const char32_t EGYPTIAN_HIEROGLYPH_140E0 = 0x000140E0;
const char32_t EGYPTIAN_HIEROGLYPH_140E1 = 0x000140E1;
const char32_t EGYPTIAN_HIEROGLYPH_140E2 = 0x000140E2;
const char32_t EGYPTIAN_HIEROGLYPH_140E3 = 0x000140E3;
const char32_t EGYPTIAN_HIEROGLYPH_140E4 = 0x000140E4;
const char32_t EGYPTIAN_HIEROGLYPH_140E5 = 0x000140E5;
const char32_t EGYPTIAN_HIEROGLYPH_140E6 = 0x000140E6;
const char32_t EGYPTIAN_HIEROGLYPH_140E7 = 0x000140E7;
const char32_t EGYPTIAN_HIEROGLYPH_140E8 = 0x000140E8;
const char32_t EGYPTIAN_HIEROGLYPH_140E9 = 0x000140E9;
const char32_t EGYPTIAN_HIEROGLYPH_140EA = 0x000140EA;
const char32_t EGYPTIAN_HIEROGLYPH_140EB = 0x000140EB;
const char32_t EGYPTIAN_HIEROGLYPH_140EC = 0x000140EC;
const char32_t EGYPTIAN_HIEROGLYPH_140ED = 0x000140ED;
const char32_t EGYPTIAN_HIEROGLYPH_140EE = 0x000140EE;
const char32_t EGYPTIAN_HIEROGLYPH_140EF = 0x000140EF;
const char32_t EGYPTIAN_HIEROGLYPH_140F0 = 0x000140F0;
const char32_t EGYPTIAN_HIEROGLYPH_140F1 = 0x000140F1;
const char32_t EGYPTIAN_HIEROGLYPH_140F2 = 0x000140F2;
const char32_t EGYPTIAN_HIEROGLYPH_140F3 = 0x000140F3;
const char32_t EGYPTIAN_HIEROGLYPH_140F4 = 0x000140F4;
const char32_t EGYPTIAN_HIEROGLYPH_140F5 = 0x000140F5;
const char32_t EGYPTIAN_HIEROGLYPH_140F6 = 0x000140F6;
const char32_t EGYPTIAN_HIEROGLYPH_140F7 = 0x000140F7;
const char32_t EGYPTIAN_HIEROGLYPH_140F8 = 0x000140F8;
const char32_t EGYPTIAN_HIEROGLYPH_140F9 = 0x000140F9;
const char32_t EGYPTIAN_HIEROGLYPH_140FA = 0x000140FA;
const char32_t EGYPTIAN_HIEROGLYPH_140FB = 0x000140FB;
const char32_t EGYPTIAN_HIEROGLYPH_140FC = 0x000140FC;
const char32_t EGYPTIAN_HIEROGLYPH_140FD = 0x000140FD;
const char32_t EGYPTIAN_HIEROGLYPH_140FE = 0x000140FE;
const char32_t EGYPTIAN_HIEROGLYPH_140FF = 0x000140FF;
const char32_t EGYPTIAN_HIEROGLYPH_14100 = 0x00014100;
const char32_t EGYPTIAN_HIEROGLYPH_14101 = 0x00014101;
const char32_t EGYPTIAN_HIEROGLYPH_14102 = 0x00014102;
const char32_t EGYPTIAN_HIEROGLYPH_14103 = 0x00014103;
const char32_t EGYPTIAN_HIEROGLYPH_14104 = 0x00014104;
const char32_t EGYPTIAN_HIEROGLYPH_14105 = 0x00014105;
const char32_t EGYPTIAN_HIEROGLYPH_14106 = 0x00014106;
const char32_t EGYPTIAN_HIEROGLYPH_14107 = 0x00014107;
const char32_t EGYPTIAN_HIEROGLYPH_14108 = 0x00014108;
const char32_t EGYPTIAN_HIEROGLYPH_14109 = 0x00014109;
const char32_t EGYPTIAN_HIEROGLYPH_1410A = 0x0001410A;
const char32_t EGYPTIAN_HIEROGLYPH_1410B = 0x0001410B;
const char32_t EGYPTIAN_HIEROGLYPH_1410C = 0x0001410C;
const char32_t EGYPTIAN_HIEROGLYPH_1410D = 0x0001410D;
const char32_t EGYPTIAN_HIEROGLYPH_1410E = 0x0001410E;
const char32_t EGYPTIAN_HIEROGLYPH_1410F = 0x0001410F;
const char32_t EGYPTIAN_HIEROGLYPH_14110 = 0x00014110;
const char32_t EGYPTIAN_HIEROGLYPH_14111 = 0x00014111;
const char32_t EGYPTIAN_HIEROGLYPH_14112 = 0x00014112;
const char32_t EGYPTIAN_HIEROGLYPH_14113 = 0x00014113;
const char32_t EGYPTIAN_HIEROGLYPH_14114 = 0x00014114;
const char32_t EGYPTIAN_HIEROGLYPH_14115 = 0x00014115;
const char32_t EGYPTIAN_HIEROGLYPH_14116 = 0x00014116;
const char32_t EGYPTIAN_HIEROGLYPH_14117 = 0x00014117;
const char32_t EGYPTIAN_HIEROGLYPH_14118 = 0x00014118;
const char32_t EGYPTIAN_HIEROGLYPH_14119 = 0x00014119;
const char32_t EGYPTIAN_HIEROGLYPH_1411A = 0x0001411A;
const char32_t EGYPTIAN_HIEROGLYPH_1411B = 0x0001411B;
const char32_t EGYPTIAN_HIEROGLYPH_1411C = 0x0001411C;
const char32_t EGYPTIAN_HIEROGLYPH_1411D = 0x0001411D;
const char32_t EGYPTIAN_HIEROGLYPH_1411E = 0x0001411E;
const char32_t EGYPTIAN_HIEROGLYPH_1411F = 0x0001411F;
const char32_t EGYPTIAN_HIEROGLYPH_14120 = 0x00014120;
const char32_t EGYPTIAN_HIEROGLYPH_14121 = 0x00014121;
const char32_t EGYPTIAN_HIEROGLYPH_14122 = 0x00014122;
const char32_t EGYPTIAN_HIEROGLYPH_14123 = 0x00014123;
const char32_t EGYPTIAN_HIEROGLYPH_14124 = 0x00014124;
const char32_t EGYPTIAN_HIEROGLYPH_14125 = 0x00014125;
const char32_t EGYPTIAN_HIEROGLYPH_14126 = 0x00014126;
const char32_t EGYPTIAN_HIEROGLYPH_14127 = 0x00014127;
const char32_t EGYPTIAN_HIEROGLYPH_14128 = 0x00014128;
const char32_t EGYPTIAN_HIEROGLYPH_14129 = 0x00014129;
const char32_t EGYPTIAN_HIEROGLYPH_1412A = 0x0001412A;
const char32_t EGYPTIAN_HIEROGLYPH_1412B = 0x0001412B;
const char32_t EGYPTIAN_HIEROGLYPH_1412C = 0x0001412C;
const char32_t EGYPTIAN_HIEROGLYPH_1412D = 0x0001412D;
const char32_t EGYPTIAN_HIEROGLYPH_1412E = 0x0001412E;
const char32_t EGYPTIAN_HIEROGLYPH_1412F = 0x0001412F;
const char32_t EGYPTIAN_HIEROGLYPH_14130 = 0x00014130;
const char32_t EGYPTIAN_HIEROGLYPH_14131 = 0x00014131;
const char32_t EGYPTIAN_HIEROGLYPH_14132 = 0x00014132;
const char32_t EGYPTIAN_HIEROGLYPH_14133 = 0x00014133;
const char32_t EGYPTIAN_HIEROGLYPH_14134 = 0x00014134;
const char32_t EGYPTIAN_HIEROGLYPH_14135 = 0x00014135;
const char32_t EGYPTIAN_HIEROGLYPH_14136 = 0x00014136;
const char32_t EGYPTIAN_HIEROGLYPH_14137 = 0x00014137;
const char32_t EGYPTIAN_HIEROGLYPH_14138 = 0x00014138;
const char32_t EGYPTIAN_HIEROGLYPH_14139 = 0x00014139;
const char32_t EGYPTIAN_HIEROGLYPH_1413A = 0x0001413A;
const char32_t EGYPTIAN_HIEROGLYPH_1413B = 0x0001413B;
const char32_t EGYPTIAN_HIEROGLYPH_1413C = 0x0001413C;
const char32_t EGYPTIAN_HIEROGLYPH_1413D = 0x0001413D;
const char32_t EGYPTIAN_HIEROGLYPH_1413E = 0x0001413E;
const char32_t EGYPTIAN_HIEROGLYPH_1413F = 0x0001413F;
const char32_t EGYPTIAN_HIEROGLYPH_14140 = 0x00014140;
const char32_t EGYPTIAN_HIEROGLYPH_14141 = 0x00014141;
const char32_t EGYPTIAN_HIEROGLYPH_14142 = 0x00014142;
const char32_t EGYPTIAN_HIEROGLYPH_14143 = 0x00014143;
const char32_t EGYPTIAN_HIEROGLYPH_14144 = 0x00014144;
const char32_t EGYPTIAN_HIEROGLYPH_14145 = 0x00014145;
const char32_t EGYPTIAN_HIEROGLYPH_14146 = 0x00014146;
const char32_t EGYPTIAN_HIEROGLYPH_14147 = 0x00014147;
const char32_t EGYPTIAN_HIEROGLYPH_14148 = 0x00014148;
const char32_t EGYPTIAN_HIEROGLYPH_14149 = 0x00014149;
const char32_t EGYPTIAN_HIEROGLYPH_1414A = 0x0001414A;
const char32_t EGYPTIAN_HIEROGLYPH_1414B = 0x0001414B;
const char32_t EGYPTIAN_HIEROGLYPH_1414C = 0x0001414C;
const char32_t EGYPTIAN_HIEROGLYPH_1414D = 0x0001414D;
const char32_t EGYPTIAN_HIEROGLYPH_1414E = 0x0001414E;
const char32_t EGYPTIAN_HIEROGLYPH_1414F = 0x0001414F;
const char32_t EGYPTIAN_HIEROGLYPH_14150 = 0x00014150;
const char32_t EGYPTIAN_HIEROGLYPH_14151 = 0x00014151;
const char32_t EGYPTIAN_HIEROGLYPH_14152 = 0x00014152;
const char32_t EGYPTIAN_HIEROGLYPH_14153 = 0x00014153;
const char32_t EGYPTIAN_HIEROGLYPH_14154 = 0x00014154;
const char32_t EGYPTIAN_HIEROGLYPH_14155 = 0x00014155;
const char32_t EGYPTIAN_HIEROGLYPH_14156 = 0x00014156;
const char32_t EGYPTIAN_HIEROGLYPH_14157 = 0x00014157;
const char32_t EGYPTIAN_HIEROGLYPH_14158 = 0x00014158;
const char32_t EGYPTIAN_HIEROGLYPH_14159 = 0x00014159;
const char32_t EGYPTIAN_HIEROGLYPH_1415A = 0x0001415A;
const char32_t EGYPTIAN_HIEROGLYPH_1415B = 0x0001415B;
const char32_t EGYPTIAN_HIEROGLYPH_1415C = 0x0001415C;
const char32_t EGYPTIAN_HIEROGLYPH_1415D = 0x0001415D;
const char32_t EGYPTIAN_HIEROGLYPH_1415E = 0x0001415E;
const char32_t EGYPTIAN_HIEROGLYPH_1415F = 0x0001415F;
const char32_t EGYPTIAN_HIEROGLYPH_14160 = 0x00014160;
const char32_t EGYPTIAN_HIEROGLYPH_14161 = 0x00014161;
const char32_t EGYPTIAN_HIEROGLYPH_14162 = 0x00014162;
const char32_t EGYPTIAN_HIEROGLYPH_14163 = 0x00014163;
const char32_t EGYPTIAN_HIEROGLYPH_14164 = 0x00014164;
const char32_t EGYPTIAN_HIEROGLYPH_14165 = 0x00014165;
const char32_t EGYPTIAN_HIEROGLYPH_14166 = 0x00014166;
const char32_t EGYPTIAN_HIEROGLYPH_14167 = 0x00014167;
const char32_t EGYPTIAN_HIEROGLYPH_14168 = 0x00014168;
const char32_t EGYPTIAN_HIEROGLYPH_14169 = 0x00014169;
const char32_t EGYPTIAN_HIEROGLYPH_1416A = 0x0001416A;
const char32_t EGYPTIAN_HIEROGLYPH_1416B = 0x0001416B;
const char32_t EGYPTIAN_HIEROGLYPH_1416C = 0x0001416C;
const char32_t EGYPTIAN_HIEROGLYPH_1416D = 0x0001416D;
const char32_t EGYPTIAN_HIEROGLYPH_1416E = 0x0001416E;
const char32_t EGYPTIAN_HIEROGLYPH_1416F = 0x0001416F;
const char32_t EGYPTIAN_HIEROGLYPH_14170 = 0x00014170;
const char32_t EGYPTIAN_HIEROGLYPH_14171 = 0x00014171;
const char32_t EGYPTIAN_HIEROGLYPH_14172 = 0x00014172;
const char32_t EGYPTIAN_HIEROGLYPH_14173 = 0x00014173;
const char32_t EGYPTIAN_HIEROGLYPH_14174 = 0x00014174;
const char32_t EGYPTIAN_HIEROGLYPH_14175 = 0x00014175;
const char32_t EGYPTIAN_HIEROGLYPH_14176 = 0x00014176;
const char32_t EGYPTIAN_HIEROGLYPH_14177 = 0x00014177;
const char32_t EGYPTIAN_HIEROGLYPH_14178 = 0x00014178;
const char32_t EGYPTIAN_HIEROGLYPH_14179 = 0x00014179;
const char32_t EGYPTIAN_HIEROGLYPH_1417A = 0x0001417A;
const char32_t EGYPTIAN_HIEROGLYPH_1417B = 0x0001417B;
const char32_t EGYPTIAN_HIEROGLYPH_1417C = 0x0001417C;
const char32_t EGYPTIAN_HIEROGLYPH_1417D = 0x0001417D;
const char32_t EGYPTIAN_HIEROGLYPH_1417E = 0x0001417E;
const char32_t EGYPTIAN_HIEROGLYPH_1417F = 0x0001417F;
const char32_t EGYPTIAN_HIEROGLYPH_14180 = 0x00014180;
const char32_t EGYPTIAN_HIEROGLYPH_14181 = 0x00014181;
const char32_t EGYPTIAN_HIEROGLYPH_14182 = 0x00014182;
const char32_t EGYPTIAN_HIEROGLYPH_14183 = 0x00014183;
const char32_t EGYPTIAN_HIEROGLYPH_14184 = 0x00014184;
const char32_t EGYPTIAN_HIEROGLYPH_14185 = 0x00014185;
const char32_t EGYPTIAN_HIEROGLYPH_14186 = 0x00014186;
const char32_t EGYPTIAN_HIEROGLYPH_14187 = 0x00014187;
const char32_t EGYPTIAN_HIEROGLYPH_14188 = 0x00014188;
const char32_t EGYPTIAN_HIEROGLYPH_14189 = 0x00014189;
const char32_t EGYPTIAN_HIEROGLYPH_1418A = 0x0001418A;
const char32_t EGYPTIAN_HIEROGLYPH_1418B = 0x0001418B;
const char32_t EGYPTIAN_HIEROGLYPH_1418C = 0x0001418C;
const char32_t EGYPTIAN_HIEROGLYPH_1418D = 0x0001418D;
const char32_t EGYPTIAN_HIEROGLYPH_1418E = 0x0001418E;
const char32_t EGYPTIAN_HIEROGLYPH_1418F = 0x0001418F;
const char32_t EGYPTIAN_HIEROGLYPH_14190 = 0x00014190;
const char32_t EGYPTIAN_HIEROGLYPH_14191 = 0x00014191;
const char32_t EGYPTIAN_HIEROGLYPH_14192 = 0x00014192;
const char32_t EGYPTIAN_HIEROGLYPH_14193 = 0x00014193;
const char32_t EGYPTIAN_HIEROGLYPH_14194 = 0x00014194;
const char32_t EGYPTIAN_HIEROGLYPH_14195 = 0x00014195;
const char32_t EGYPTIAN_HIEROGLYPH_14196 = 0x00014196;
const char32_t EGYPTIAN_HIEROGLYPH_14197 = 0x00014197;
const char32_t EGYPTIAN_HIEROGLYPH_14198 = 0x00014198;
const char32_t EGYPTIAN_HIEROGLYPH_14199 = 0x00014199;
const char32_t EGYPTIAN_HIEROGLYPH_1419A = 0x0001419A;
const char32_t EGYPTIAN_HIEROGLYPH_1419B = 0x0001419B;
const char32_t EGYPTIAN_HIEROGLYPH_1419C = 0x0001419C;
const char32_t EGYPTIAN_HIEROGLYPH_1419D = 0x0001419D;
const char32_t EGYPTIAN_HIEROGLYPH_1419E = 0x0001419E;
const char32_t EGYPTIAN_HIEROGLYPH_1419F = 0x0001419F;
const char32_t EGYPTIAN_HIEROGLYPH_141A0 = 0x000141A0;
const char32_t EGYPTIAN_HIEROGLYPH_141A1 = 0x000141A1;
const char32_t EGYPTIAN_HIEROGLYPH_141A2 = 0x000141A2;
const char32_t EGYPTIAN_HIEROGLYPH_141A3 = 0x000141A3;
const char32_t EGYPTIAN_HIEROGLYPH_141A4 = 0x000141A4;
const char32_t EGYPTIAN_HIEROGLYPH_141A5 = 0x000141A5;
const char32_t EGYPTIAN_HIEROGLYPH_141A6 = 0x000141A6;
const char32_t EGYPTIAN_HIEROGLYPH_141A7 = 0x000141A7;
const char32_t EGYPTIAN_HIEROGLYPH_141A8 = 0x000141A8;
const char32_t EGYPTIAN_HIEROGLYPH_141A9 = 0x000141A9;
const char32_t EGYPTIAN_HIEROGLYPH_141AA = 0x000141AA;
const char32_t EGYPTIAN_HIEROGLYPH_141AB = 0x000141AB;
const char32_t EGYPTIAN_HIEROGLYPH_141AC = 0x000141AC;
const char32_t EGYPTIAN_HIEROGLYPH_141AD = 0x000141AD;
const char32_t EGYPTIAN_HIEROGLYPH_141AE = 0x000141AE;
const char32_t EGYPTIAN_HIEROGLYPH_141AF = 0x000141AF;
const char32_t EGYPTIAN_HIEROGLYPH_141B0 = 0x000141B0;
const char32_t EGYPTIAN_HIEROGLYPH_141B1 = 0x000141B1;
const char32_t EGYPTIAN_HIEROGLYPH_141B2 = 0x000141B2;
const char32_t EGYPTIAN_HIEROGLYPH_141B3 = 0x000141B3;
const char32_t EGYPTIAN_HIEROGLYPH_141B4 = 0x000141B4;
const char32_t EGYPTIAN_HIEROGLYPH_141B5 = 0x000141B5;
const char32_t EGYPTIAN_HIEROGLYPH_141B6 = 0x000141B6;
const char32_t EGYPTIAN_HIEROGLYPH_141B7 = 0x000141B7;
const char32_t EGYPTIAN_HIEROGLYPH_141B8 = 0x000141B8;
const char32_t EGYPTIAN_HIEROGLYPH_141B9 = 0x000141B9;
const char32_t EGYPTIAN_HIEROGLYPH_141BA = 0x000141BA;
const char32_t EGYPTIAN_HIEROGLYPH_141BB = 0x000141BB;
const char32_t EGYPTIAN_HIEROGLYPH_141BC = 0x000141BC;
const char32_t EGYPTIAN_HIEROGLYPH_141BD = 0x000141BD;
const char32_t EGYPTIAN_HIEROGLYPH_141BE = 0x000141BE;
const char32_t EGYPTIAN_HIEROGLYPH_141BF = 0x000141BF;
const char32_t EGYPTIAN_HIEROGLYPH_141C0 = 0x000141C0;
const char32_t EGYPTIAN_HIEROGLYPH_141C1 = 0x000141C1;
const char32_t EGYPTIAN_HIEROGLYPH_141C2 = 0x000141C2;
const char32_t EGYPTIAN_HIEROGLYPH_141C3 = 0x000141C3;
const char32_t EGYPTIAN_HIEROGLYPH_141C4 = 0x000141C4;
const char32_t EGYPTIAN_HIEROGLYPH_141C5 = 0x000141C5;
const char32_t EGYPTIAN_HIEROGLYPH_141C6 = 0x000141C6;
const char32_t EGYPTIAN_HIEROGLYPH_141C7 = 0x000141C7;
const char32_t EGYPTIAN_HIEROGLYPH_141C8 = 0x000141C8;
const char32_t EGYPTIAN_HIEROGLYPH_141C9 = 0x000141C9;
const char32_t EGYPTIAN_HIEROGLYPH_141CA = 0x000141CA;
const char32_t EGYPTIAN_HIEROGLYPH_141CB = 0x000141CB;
const char32_t EGYPTIAN_HIEROGLYPH_141CC = 0x000141CC;
const char32_t EGYPTIAN_HIEROGLYPH_141CD = 0x000141CD;
const char32_t EGYPTIAN_HIEROGLYPH_141CE = 0x000141CE;
const char32_t EGYPTIAN_HIEROGLYPH_141CF = 0x000141CF;
const char32_t EGYPTIAN_HIEROGLYPH_141D0 = 0x000141D0;
const char32_t EGYPTIAN_HIEROGLYPH_141D1 = 0x000141D1;
const char32_t EGYPTIAN_HIEROGLYPH_141D2 = 0x000141D2;
const char32_t EGYPTIAN_HIEROGLYPH_141D3 = 0x000141D3;
const char32_t EGYPTIAN_HIEROGLYPH_141D4 = 0x000141D4;
const char32_t EGYPTIAN_HIEROGLYPH_141D5 = 0x000141D5;
const char32_t EGYPTIAN_HIEROGLYPH_141D6 = 0x000141D6;
const char32_t EGYPTIAN_HIEROGLYPH_141D7 = 0x000141D7;
const char32_t EGYPTIAN_HIEROGLYPH_141D8 = 0x000141D8;
const char32_t EGYPTIAN_HIEROGLYPH_141D9 = 0x000141D9;
const char32_t EGYPTIAN_HIEROGLYPH_141DA = 0x000141DA;
const char32_t EGYPTIAN_HIEROGLYPH_141DB = 0x000141DB;
const char32_t EGYPTIAN_HIEROGLYPH_141DC = 0x000141DC;
const char32_t EGYPTIAN_HIEROGLYPH_141DD = 0x000141DD;
const char32_t EGYPTIAN_HIEROGLYPH_141DE = 0x000141DE;
const char32_t EGYPTIAN_HIEROGLYPH_141DF = 0x000141DF;
const char32_t EGYPTIAN_HIEROGLYPH_141E0 = 0x000141E0;
const char32_t EGYPTIAN_HIEROGLYPH_141E1 = 0x000141E1;
const char32_t EGYPTIAN_HIEROGLYPH_141E2 = 0x000141E2;
const char32_t EGYPTIAN_HIEROGLYPH_141E3 = 0x000141E3;
const char32_t EGYPTIAN_HIEROGLYPH_141E4 = 0x000141E4;
const char32_t EGYPTIAN_HIEROGLYPH_141E5 = 0x000141E5;
const char32_t EGYPTIAN_HIEROGLYPH_141E6 = 0x000141E6;
const char32_t EGYPTIAN_HIEROGLYPH_141E7 = 0x000141E7;
const char32_t EGYPTIAN_HIEROGLYPH_141E8 = 0x000141E8;
const char32_t EGYPTIAN_HIEROGLYPH_141E9 = 0x000141E9;
const char32_t EGYPTIAN_HIEROGLYPH_141EA = 0x000141EA;
const char32_t EGYPTIAN_HIEROGLYPH_141EB = 0x000141EB;
const char32_t EGYPTIAN_HIEROGLYPH_141EC = 0x000141EC;
const char32_t EGYPTIAN_HIEROGLYPH_141ED = 0x000141ED;
const char32_t EGYPTIAN_HIEROGLYPH_141EE = 0x000141EE;
const char32_t EGYPTIAN_HIEROGLYPH_141EF = 0x000141EF;
const char32_t EGYPTIAN_HIEROGLYPH_141F0 = 0x000141F0;
const char32_t EGYPTIAN_HIEROGLYPH_141F1 = 0x000141F1;
const char32_t EGYPTIAN_HIEROGLYPH_141F2 = 0x000141F2;
const char32_t EGYPTIAN_HIEROGLYPH_141F3 = 0x000141F3;
const char32_t EGYPTIAN_HIEROGLYPH_141F4 = 0x000141F4;
const char32_t EGYPTIAN_HIEROGLYPH_141F5 = 0x000141F5;
const char32_t EGYPTIAN_HIEROGLYPH_141F6 = 0x000141F6;
const char32_t EGYPTIAN_HIEROGLYPH_141F7 = 0x000141F7;
const char32_t EGYPTIAN_HIEROGLYPH_141F8 = 0x000141F8;
const char32_t EGYPTIAN_HIEROGLYPH_141F9 = 0x000141F9;
const char32_t EGYPTIAN_HIEROGLYPH_141FA = 0x000141FA;
const char32_t EGYPTIAN_HIEROGLYPH_141FB = 0x000141FB;
const char32_t EGYPTIAN_HIEROGLYPH_141FC = 0x000141FC;
const char32_t EGYPTIAN_HIEROGLYPH_141FD = 0x000141FD;
const char32_t EGYPTIAN_HIEROGLYPH_141FE = 0x000141FE;
const char32_t EGYPTIAN_HIEROGLYPH_141FF = 0x000141FF;
const char32_t EGYPTIAN_HIEROGLYPH_14200 = 0x00014200;
const char32_t EGYPTIAN_HIEROGLYPH_14201 = 0x00014201;
const char32_t EGYPTIAN_HIEROGLYPH_14202 = 0x00014202;
const char32_t EGYPTIAN_HIEROGLYPH_14203 = 0x00014203;
const char32_t EGYPTIAN_HIEROGLYPH_14204 = 0x00014204;
const char32_t EGYPTIAN_HIEROGLYPH_14205 = 0x00014205;
const char32_t EGYPTIAN_HIEROGLYPH_14206 = 0x00014206;
const char32_t EGYPTIAN_HIEROGLYPH_14207 = 0x00014207;
const char32_t EGYPTIAN_HIEROGLYPH_14208 = 0x00014208;
const char32_t EGYPTIAN_HIEROGLYPH_14209 = 0x00014209;
const char32_t EGYPTIAN_HIEROGLYPH_1420A = 0x0001420A;
const char32_t EGYPTIAN_HIEROGLYPH_1420B = 0x0001420B;
const char32_t EGYPTIAN_HIEROGLYPH_1420C = 0x0001420C;
const char32_t EGYPTIAN_HIEROGLYPH_1420D = 0x0001420D;
const char32_t EGYPTIAN_HIEROGLYPH_1420E = 0x0001420E;
const char32_t EGYPTIAN_HIEROGLYPH_1420F = 0x0001420F;
const char32_t EGYPTIAN_HIEROGLYPH_14210 = 0x00014210;
const char32_t EGYPTIAN_HIEROGLYPH_14211 = 0x00014211;
const char32_t EGYPTIAN_HIEROGLYPH_14212 = 0x00014212;
const char32_t EGYPTIAN_HIEROGLYPH_14213 = 0x00014213;
const char32_t EGYPTIAN_HIEROGLYPH_14214 = 0x00014214;
const char32_t EGYPTIAN_HIEROGLYPH_14215 = 0x00014215;
const char32_t EGYPTIAN_HIEROGLYPH_14216 = 0x00014216;
const char32_t EGYPTIAN_HIEROGLYPH_14217 = 0x00014217;
const char32_t EGYPTIAN_HIEROGLYPH_14218 = 0x00014218;
const char32_t EGYPTIAN_HIEROGLYPH_14219 = 0x00014219;
const char32_t EGYPTIAN_HIEROGLYPH_1421A = 0x0001421A;
const char32_t EGYPTIAN_HIEROGLYPH_1421B = 0x0001421B;
const char32_t EGYPTIAN_HIEROGLYPH_1421C = 0x0001421C;
const char32_t EGYPTIAN_HIEROGLYPH_1421D = 0x0001421D;
const char32_t EGYPTIAN_HIEROGLYPH_1421E = 0x0001421E;
const char32_t EGYPTIAN_HIEROGLYPH_1421F = 0x0001421F;
const char32_t EGYPTIAN_HIEROGLYPH_14220 = 0x00014220;
const char32_t EGYPTIAN_HIEROGLYPH_14221 = 0x00014221;
const char32_t EGYPTIAN_HIEROGLYPH_14222 = 0x00014222;
const char32_t EGYPTIAN_HIEROGLYPH_14223 = 0x00014223;
const char32_t EGYPTIAN_HIEROGLYPH_14224 = 0x00014224;
const char32_t EGYPTIAN_HIEROGLYPH_14225 = 0x00014225;
const char32_t EGYPTIAN_HIEROGLYPH_14226 = 0x00014226;
const char32_t EGYPTIAN_HIEROGLYPH_14227 = 0x00014227;
const char32_t EGYPTIAN_HIEROGLYPH_14228 = 0x00014228;
const char32_t EGYPTIAN_HIEROGLYPH_14229 = 0x00014229;
const char32_t EGYPTIAN_HIEROGLYPH_1422A = 0x0001422A;
const char32_t EGYPTIAN_HIEROGLYPH_1422B = 0x0001422B;
const char32_t EGYPTIAN_HIEROGLYPH_1422C = 0x0001422C;
const char32_t EGYPTIAN_HIEROGLYPH_1422D = 0x0001422D;
const char32_t EGYPTIAN_HIEROGLYPH_1422E = 0x0001422E;
const char32_t EGYPTIAN_HIEROGLYPH_1422F = 0x0001422F;
const char32_t EGYPTIAN_HIEROGLYPH_14230 = 0x00014230;
const char32_t EGYPTIAN_HIEROGLYPH_14231 = 0x00014231;
const char32_t EGYPTIAN_HIEROGLYPH_14232 = 0x00014232;
const char32_t EGYPTIAN_HIEROGLYPH_14233 = 0x00014233;
const char32_t EGYPTIAN_HIEROGLYPH_14234 = 0x00014234;
const char32_t EGYPTIAN_HIEROGLYPH_14235 = 0x00014235;
const char32_t EGYPTIAN_HIEROGLYPH_14236 = 0x00014236;
const char32_t EGYPTIAN_HIEROGLYPH_14237 = 0x00014237;
const char32_t EGYPTIAN_HIEROGLYPH_14238 = 0x00014238;
const char32_t EGYPTIAN_HIEROGLYPH_14239 = 0x00014239;
const char32_t EGYPTIAN_HIEROGLYPH_1423A = 0x0001423A;
const char32_t EGYPTIAN_HIEROGLYPH_1423B = 0x0001423B;
const char32_t EGYPTIAN_HIEROGLYPH_1423C = 0x0001423C;
const char32_t EGYPTIAN_HIEROGLYPH_1423D = 0x0001423D;
const char32_t EGYPTIAN_HIEROGLYPH_1423E = 0x0001423E;
const char32_t EGYPTIAN_HIEROGLYPH_1423F = 0x0001423F;
const char32_t EGYPTIAN_HIEROGLYPH_14240 = 0x00014240;
const char32_t EGYPTIAN_HIEROGLYPH_14241 = 0x00014241;
const char32_t EGYPTIAN_HIEROGLYPH_14242 = 0x00014242;
const char32_t EGYPTIAN_HIEROGLYPH_14243 = 0x00014243;
const char32_t EGYPTIAN_HIEROGLYPH_14244 = 0x00014244;
const char32_t EGYPTIAN_HIEROGLYPH_14245 = 0x00014245;
const char32_t EGYPTIAN_HIEROGLYPH_14246 = 0x00014246;
const char32_t EGYPTIAN_HIEROGLYPH_14247 = 0x00014247;
const char32_t EGYPTIAN_HIEROGLYPH_14248 = 0x00014248;
const char32_t EGYPTIAN_HIEROGLYPH_14249 = 0x00014249;
const char32_t EGYPTIAN_HIEROGLYPH_1424A = 0x0001424A;
const char32_t EGYPTIAN_HIEROGLYPH_1424B = 0x0001424B;
const char32_t EGYPTIAN_HIEROGLYPH_1424C = 0x0001424C;
const char32_t EGYPTIAN_HIEROGLYPH_1424D = 0x0001424D;
const char32_t EGYPTIAN_HIEROGLYPH_1424E = 0x0001424E;
const char32_t EGYPTIAN_HIEROGLYPH_1424F = 0x0001424F;
const char32_t EGYPTIAN_HIEROGLYPH_14250 = 0x00014250;
const char32_t EGYPTIAN_HIEROGLYPH_14251 = 0x00014251;
const char32_t EGYPTIAN_HIEROGLYPH_14252 = 0x00014252;
const char32_t EGYPTIAN_HIEROGLYPH_14253 = 0x00014253;
const char32_t EGYPTIAN_HIEROGLYPH_14254 = 0x00014254;
const char32_t EGYPTIAN_HIEROGLYPH_14255 = 0x00014255;
const char32_t EGYPTIAN_HIEROGLYPH_14256 = 0x00014256;
const char32_t EGYPTIAN_HIEROGLYPH_14257 = 0x00014257;
const char32_t EGYPTIAN_HIEROGLYPH_14258 = 0x00014258;
const char32_t EGYPTIAN_HIEROGLYPH_14259 = 0x00014259;
const char32_t EGYPTIAN_HIEROGLYPH_1425A = 0x0001425A;
const char32_t EGYPTIAN_HIEROGLYPH_1425B = 0x0001425B;
const char32_t EGYPTIAN_HIEROGLYPH_1425C = 0x0001425C;
const char32_t EGYPTIAN_HIEROGLYPH_1425D = 0x0001425D;
const char32_t EGYPTIAN_HIEROGLYPH_1425E = 0x0001425E;
const char32_t EGYPTIAN_HIEROGLYPH_1425F = 0x0001425F;
const char32_t EGYPTIAN_HIEROGLYPH_14260 = 0x00014260;
const char32_t EGYPTIAN_HIEROGLYPH_14261 = 0x00014261;
const char32_t EGYPTIAN_HIEROGLYPH_14262 = 0x00014262;
const char32_t EGYPTIAN_HIEROGLYPH_14263 = 0x00014263;
const char32_t EGYPTIAN_HIEROGLYPH_14264 = 0x00014264;
const char32_t EGYPTIAN_HIEROGLYPH_14265 = 0x00014265;
const char32_t EGYPTIAN_HIEROGLYPH_14266 = 0x00014266;
const char32_t EGYPTIAN_HIEROGLYPH_14267 = 0x00014267;
const char32_t EGYPTIAN_HIEROGLYPH_14268 = 0x00014268;
const char32_t EGYPTIAN_HIEROGLYPH_14269 = 0x00014269;
const char32_t EGYPTIAN_HIEROGLYPH_1426A = 0x0001426A;
const char32_t EGYPTIAN_HIEROGLYPH_1426B = 0x0001426B;
const char32_t EGYPTIAN_HIEROGLYPH_1426C = 0x0001426C;
const char32_t EGYPTIAN_HIEROGLYPH_1426D = 0x0001426D;
const char32_t EGYPTIAN_HIEROGLYPH_1426E = 0x0001426E;
const char32_t EGYPTIAN_HIEROGLYPH_1426F = 0x0001426F;
const char32_t EGYPTIAN_HIEROGLYPH_14270 = 0x00014270;
const char32_t EGYPTIAN_HIEROGLYPH_14271 = 0x00014271;
const char32_t EGYPTIAN_HIEROGLYPH_14272 = 0x00014272;
const char32_t EGYPTIAN_HIEROGLYPH_14273 = 0x00014273;
const char32_t EGYPTIAN_HIEROGLYPH_14274 = 0x00014274;
const char32_t EGYPTIAN_HIEROGLYPH_14275 = 0x00014275;
const char32_t EGYPTIAN_HIEROGLYPH_14276 = 0x00014276;
const char32_t EGYPTIAN_HIEROGLYPH_14277 = 0x00014277;
const char32_t EGYPTIAN_HIEROGLYPH_14278 = 0x00014278;
const char32_t EGYPTIAN_HIEROGLYPH_14279 = 0x00014279;
const char32_t EGYPTIAN_HIEROGLYPH_1427A = 0x0001427A;
const char32_t EGYPTIAN_HIEROGLYPH_1427B = 0x0001427B;
const char32_t EGYPTIAN_HIEROGLYPH_1427C = 0x0001427C;
const char32_t EGYPTIAN_HIEROGLYPH_1427D = 0x0001427D;
const char32_t EGYPTIAN_HIEROGLYPH_1427E = 0x0001427E;
const char32_t EGYPTIAN_HIEROGLYPH_1427F = 0x0001427F;
const char32_t EGYPTIAN_HIEROGLYPH_14280 = 0x00014280;
const char32_t EGYPTIAN_HIEROGLYPH_14281 = 0x00014281;
const char32_t EGYPTIAN_HIEROGLYPH_14282 = 0x00014282;
const char32_t EGYPTIAN_HIEROGLYPH_14283 = 0x00014283;
const char32_t EGYPTIAN_HIEROGLYPH_14284 = 0x00014284;
const char32_t EGYPTIAN_HIEROGLYPH_14285 = 0x00014285;
const char32_t EGYPTIAN_HIEROGLYPH_14286 = 0x00014286;
const char32_t EGYPTIAN_HIEROGLYPH_14287 = 0x00014287;
const char32_t EGYPTIAN_HIEROGLYPH_14288 = 0x00014288;
const char32_t EGYPTIAN_HIEROGLYPH_14289 = 0x00014289;
const char32_t EGYPTIAN_HIEROGLYPH_1428A = 0x0001428A;
const char32_t EGYPTIAN_HIEROGLYPH_1428B = 0x0001428B;
const char32_t EGYPTIAN_HIEROGLYPH_1428C = 0x0001428C;
const char32_t EGYPTIAN_HIEROGLYPH_1428D = 0x0001428D;
const char32_t EGYPTIAN_HIEROGLYPH_1428E = 0x0001428E;
const char32_t EGYPTIAN_HIEROGLYPH_1428F = 0x0001428F;
const char32_t EGYPTIAN_HIEROGLYPH_14290 = 0x00014290;
const char32_t EGYPTIAN_HIEROGLYPH_14291 = 0x00014291;
const char32_t EGYPTIAN_HIEROGLYPH_14292 = 0x00014292;
const char32_t EGYPTIAN_HIEROGLYPH_14293 = 0x00014293;
const char32_t EGYPTIAN_HIEROGLYPH_14294 = 0x00014294;
const char32_t EGYPTIAN_HIEROGLYPH_14295 = 0x00014295;
const char32_t EGYPTIAN_HIEROGLYPH_14296 = 0x00014296;
const char32_t EGYPTIAN_HIEROGLYPH_14297 = 0x00014297;
const char32_t EGYPTIAN_HIEROGLYPH_14298 = 0x00014298;
const char32_t EGYPTIAN_HIEROGLYPH_14299 = 0x00014299;
const char32_t EGYPTIAN_HIEROGLYPH_1429A = 0x0001429A;
const char32_t EGYPTIAN_HIEROGLYPH_1429B = 0x0001429B;
const char32_t EGYPTIAN_HIEROGLYPH_1429C = 0x0001429C;
const char32_t EGYPTIAN_HIEROGLYPH_1429D = 0x0001429D;
const char32_t EGYPTIAN_HIEROGLYPH_1429E = 0x0001429E;
const char32_t EGYPTIAN_HIEROGLYPH_1429F = 0x0001429F;
const char32_t EGYPTIAN_HIEROGLYPH_142A0 = 0x000142A0;
const char32_t EGYPTIAN_HIEROGLYPH_142A1 = 0x000142A1;
const char32_t EGYPTIAN_HIEROGLYPH_142A2 = 0x000142A2;
const char32_t EGYPTIAN_HIEROGLYPH_142A3 = 0x000142A3;
const char32_t EGYPTIAN_HIEROGLYPH_142A4 = 0x000142A4;
const char32_t EGYPTIAN_HIEROGLYPH_142A5 = 0x000142A5;
const char32_t EGYPTIAN_HIEROGLYPH_142A6 = 0x000142A6;
const char32_t EGYPTIAN_HIEROGLYPH_142A7 = 0x000142A7;
const char32_t EGYPTIAN_HIEROGLYPH_142A8 = 0x000142A8;
const char32_t EGYPTIAN_HIEROGLYPH_142A9 = 0x000142A9;
const char32_t EGYPTIAN_HIEROGLYPH_142AA = 0x000142AA;
const char32_t EGYPTIAN_HIEROGLYPH_142AB = 0x000142AB;
const char32_t EGYPTIAN_HIEROGLYPH_142AC = 0x000142AC;
const char32_t EGYPTIAN_HIEROGLYPH_142AD = 0x000142AD;
const char32_t EGYPTIAN_HIEROGLYPH_142AE = 0x000142AE;
const char32_t EGYPTIAN_HIEROGLYPH_142AF = 0x000142AF;
const char32_t EGYPTIAN_HIEROGLYPH_142B0 = 0x000142B0;
const char32_t EGYPTIAN_HIEROGLYPH_142B1 = 0x000142B1;
const char32_t EGYPTIAN_HIEROGLYPH_142B2 = 0x000142B2;
const char32_t EGYPTIAN_HIEROGLYPH_142B3 = 0x000142B3;
const char32_t EGYPTIAN_HIEROGLYPH_142B4 = 0x000142B4;
const char32_t EGYPTIAN_HIEROGLYPH_142B5 = 0x000142B5;
const char32_t EGYPTIAN_HIEROGLYPH_142B6 = 0x000142B6;
const char32_t EGYPTIAN_HIEROGLYPH_142B7 = 0x000142B7;
const char32_t EGYPTIAN_HIEROGLYPH_142B8 = 0x000142B8;
const char32_t EGYPTIAN_HIEROGLYPH_142B9 = 0x000142B9;
const char32_t EGYPTIAN_HIEROGLYPH_142BA = 0x000142BA;
const char32_t EGYPTIAN_HIEROGLYPH_142BB = 0x000142BB;
const char32_t EGYPTIAN_HIEROGLYPH_142BC = 0x000142BC;
const char32_t EGYPTIAN_HIEROGLYPH_142BD = 0x000142BD;
const char32_t EGYPTIAN_HIEROGLYPH_142BE = 0x000142BE;
const char32_t EGYPTIAN_HIEROGLYPH_142BF = 0x000142BF;
const char32_t EGYPTIAN_HIEROGLYPH_142C0 = 0x000142C0;
const char32_t EGYPTIAN_HIEROGLYPH_142C1 = 0x000142C1;
const char32_t EGYPTIAN_HIEROGLYPH_142C2 = 0x000142C2;
const char32_t EGYPTIAN_HIEROGLYPH_142C3 = 0x000142C3;
const char32_t EGYPTIAN_HIEROGLYPH_142C4 = 0x000142C4;
const char32_t EGYPTIAN_HIEROGLYPH_142C5 = 0x000142C5;
const char32_t EGYPTIAN_HIEROGLYPH_142C6 = 0x000142C6;
const char32_t EGYPTIAN_HIEROGLYPH_142C7 = 0x000142C7;
const char32_t EGYPTIAN_HIEROGLYPH_142C8 = 0x000142C8;
const char32_t EGYPTIAN_HIEROGLYPH_142C9 = 0x000142C9;
const char32_t EGYPTIAN_HIEROGLYPH_142CA = 0x000142CA;
const char32_t EGYPTIAN_HIEROGLYPH_142CB = 0x000142CB;
const char32_t EGYPTIAN_HIEROGLYPH_142CC = 0x000142CC;
const char32_t EGYPTIAN_HIEROGLYPH_142CD = 0x000142CD;
const char32_t EGYPTIAN_HIEROGLYPH_142CE = 0x000142CE;
const char32_t EGYPTIAN_HIEROGLYPH_142CF = 0x000142CF;
const char32_t EGYPTIAN_HIEROGLYPH_142D0 = 0x000142D0;
const char32_t EGYPTIAN_HIEROGLYPH_142D1 = 0x000142D1;
const char32_t EGYPTIAN_HIEROGLYPH_142D2 = 0x000142D2;
const char32_t EGYPTIAN_HIEROGLYPH_142D3 = 0x000142D3;
const char32_t EGYPTIAN_HIEROGLYPH_142D4 = 0x000142D4;
const char32_t EGYPTIAN_HIEROGLYPH_142D5 = 0x000142D5;
const char32_t EGYPTIAN_HIEROGLYPH_142D6 = 0x000142D6;
const char32_t EGYPTIAN_HIEROGLYPH_142D7 = 0x000142D7;
const char32_t EGYPTIAN_HIEROGLYPH_142D8 = 0x000142D8;
const char32_t EGYPTIAN_HIEROGLYPH_142D9 = 0x000142D9;
const char32_t EGYPTIAN_HIEROGLYPH_142DA = 0x000142DA;
const char32_t EGYPTIAN_HIEROGLYPH_142DB = 0x000142DB;
const char32_t EGYPTIAN_HIEROGLYPH_142DC = 0x000142DC;
const char32_t EGYPTIAN_HIEROGLYPH_142DD = 0x000142DD;
const char32_t EGYPTIAN_HIEROGLYPH_142DE = 0x000142DE;
const char32_t EGYPTIAN_HIEROGLYPH_142DF = 0x000142DF;
const char32_t EGYPTIAN_HIEROGLYPH_142E0 = 0x000142E0;
const char32_t EGYPTIAN_HIEROGLYPH_142E1 = 0x000142E1;
const char32_t EGYPTIAN_HIEROGLYPH_142E2 = 0x000142E2;
const char32_t EGYPTIAN_HIEROGLYPH_142E3 = 0x000142E3;
const char32_t EGYPTIAN_HIEROGLYPH_142E4 = 0x000142E4;
const char32_t EGYPTIAN_HIEROGLYPH_142E5 = 0x000142E5;
const char32_t EGYPTIAN_HIEROGLYPH_142E6 = 0x000142E6;
const char32_t EGYPTIAN_HIEROGLYPH_142E7 = 0x000142E7;
const char32_t EGYPTIAN_HIEROGLYPH_142E8 = 0x000142E8;
const char32_t EGYPTIAN_HIEROGLYPH_142E9 = 0x000142E9;
const char32_t EGYPTIAN_HIEROGLYPH_142EA = 0x000142EA;
const char32_t EGYPTIAN_HIEROGLYPH_142EB = 0x000142EB;
const char32_t EGYPTIAN_HIEROGLYPH_142EC = 0x000142EC;
const char32_t EGYPTIAN_HIEROGLYPH_142ED = 0x000142ED;
const char32_t EGYPTIAN_HIEROGLYPH_142EE = 0x000142EE;
const char32_t EGYPTIAN_HIEROGLYPH_142EF = 0x000142EF;
const char32_t EGYPTIAN_HIEROGLYPH_142F0 = 0x000142F0;
const char32_t EGYPTIAN_HIEROGLYPH_142F1 = 0x000142F1;
const char32_t EGYPTIAN_HIEROGLYPH_142F2 = 0x000142F2;
const char32_t EGYPTIAN_HIEROGLYPH_142F3 = 0x000142F3;
const char32_t EGYPTIAN_HIEROGLYPH_142F4 = 0x000142F4;
const char32_t EGYPTIAN_HIEROGLYPH_142F5 = 0x000142F5;
const char32_t EGYPTIAN_HIEROGLYPH_142F6 = 0x000142F6;
const char32_t EGYPTIAN_HIEROGLYPH_142F7 = 0x000142F7;
const char32_t EGYPTIAN_HIEROGLYPH_142F8 = 0x000142F8;
const char32_t EGYPTIAN_HIEROGLYPH_142F9 = 0x000142F9;
const char32_t EGYPTIAN_HIEROGLYPH_142FA = 0x000142FA;
const char32_t EGYPTIAN_HIEROGLYPH_142FB = 0x000142FB;
const char32_t EGYPTIAN_HIEROGLYPH_142FC = 0x000142FC;
const char32_t EGYPTIAN_HIEROGLYPH_142FD = 0x000142FD;
const char32_t EGYPTIAN_HIEROGLYPH_142FE = 0x000142FE;
const char32_t EGYPTIAN_HIEROGLYPH_142FF = 0x000142FF;
const char32_t EGYPTIAN_HIEROGLYPH_14300 = 0x00014300;
const char32_t EGYPTIAN_HIEROGLYPH_14301 = 0x00014301;
const char32_t EGYPTIAN_HIEROGLYPH_14302 = 0x00014302;
const char32_t EGYPTIAN_HIEROGLYPH_14303 = 0x00014303;
const char32_t EGYPTIAN_HIEROGLYPH_14304 = 0x00014304;
const char32_t EGYPTIAN_HIEROGLYPH_14305 = 0x00014305;
const char32_t EGYPTIAN_HIEROGLYPH_14306 = 0x00014306;
const char32_t EGYPTIAN_HIEROGLYPH_14307 = 0x00014307;
const char32_t EGYPTIAN_HIEROGLYPH_14308 = 0x00014308;
const char32_t EGYPTIAN_HIEROGLYPH_14309 = 0x00014309;
const char32_t EGYPTIAN_HIEROGLYPH_1430A = 0x0001430A;
const char32_t EGYPTIAN_HIEROGLYPH_1430B = 0x0001430B;
const char32_t EGYPTIAN_HIEROGLYPH_1430C = 0x0001430C;
const char32_t EGYPTIAN_HIEROGLYPH_1430D = 0x0001430D;
const char32_t EGYPTIAN_HIEROGLYPH_1430E = 0x0001430E;
const char32_t EGYPTIAN_HIEROGLYPH_1430F = 0x0001430F;
const char32_t EGYPTIAN_HIEROGLYPH_14310 = 0x00014310;
const char32_t EGYPTIAN_HIEROGLYPH_14311 = 0x00014311;
const char32_t EGYPTIAN_HIEROGLYPH_14312 = 0x00014312;
const char32_t EGYPTIAN_HIEROGLYPH_14313 = 0x00014313;
const char32_t EGYPTIAN_HIEROGLYPH_14314 = 0x00014314;
const char32_t EGYPTIAN_HIEROGLYPH_14315 = 0x00014315;
const char32_t EGYPTIAN_HIEROGLYPH_14316 = 0x00014316;
const char32_t EGYPTIAN_HIEROGLYPH_14317 = 0x00014317;
const char32_t EGYPTIAN_HIEROGLYPH_14318 = 0x00014318;
const char32_t EGYPTIAN_HIEROGLYPH_14319 = 0x00014319;
const char32_t EGYPTIAN_HIEROGLYPH_1431A = 0x0001431A;
const char32_t EGYPTIAN_HIEROGLYPH_1431B = 0x0001431B;
const char32_t EGYPTIAN_HIEROGLYPH_1431C = 0x0001431C;
const char32_t EGYPTIAN_HIEROGLYPH_1431D = 0x0001431D;
const char32_t EGYPTIAN_HIEROGLYPH_1431E = 0x0001431E;
const char32_t EGYPTIAN_HIEROGLYPH_1431F = 0x0001431F;
const char32_t EGYPTIAN_HIEROGLYPH_14320 = 0x00014320;
const char32_t EGYPTIAN_HIEROGLYPH_14321 = 0x00014321;
const char32_t EGYPTIAN_HIEROGLYPH_14322 = 0x00014322;
const char32_t EGYPTIAN_HIEROGLYPH_14323 = 0x00014323;
const char32_t EGYPTIAN_HIEROGLYPH_14324 = 0x00014324;
const char32_t EGYPTIAN_HIEROGLYPH_14325 = 0x00014325;
const char32_t EGYPTIAN_HIEROGLYPH_14326 = 0x00014326;
const char32_t EGYPTIAN_HIEROGLYPH_14327 = 0x00014327;
const char32_t EGYPTIAN_HIEROGLYPH_14328 = 0x00014328;
const char32_t EGYPTIAN_HIEROGLYPH_14329 = 0x00014329;
const char32_t EGYPTIAN_HIEROGLYPH_1432A = 0x0001432A;
const char32_t EGYPTIAN_HIEROGLYPH_1432B = 0x0001432B;
const char32_t EGYPTIAN_HIEROGLYPH_1432C = 0x0001432C;
const char32_t EGYPTIAN_HIEROGLYPH_1432D = 0x0001432D;
const char32_t EGYPTIAN_HIEROGLYPH_1432E = 0x0001432E;
const char32_t EGYPTIAN_HIEROGLYPH_1432F = 0x0001432F;
const char32_t EGYPTIAN_HIEROGLYPH_14330 = 0x00014330;
const char32_t EGYPTIAN_HIEROGLYPH_14331 = 0x00014331;
const char32_t EGYPTIAN_HIEROGLYPH_14332 = 0x00014332;
const char32_t EGYPTIAN_HIEROGLYPH_14333 = 0x00014333;
const char32_t EGYPTIAN_HIEROGLYPH_14334 = 0x00014334;
const char32_t EGYPTIAN_HIEROGLYPH_14335 = 0x00014335;
const char32_t EGYPTIAN_HIEROGLYPH_14336 = 0x00014336;
const char32_t EGYPTIAN_HIEROGLYPH_14337 = 0x00014337;
const char32_t EGYPTIAN_HIEROGLYPH_14338 = 0x00014338;
const char32_t EGYPTIAN_HIEROGLYPH_14339 = 0x00014339;
const char32_t EGYPTIAN_HIEROGLYPH_1433A = 0x0001433A;
const char32_t EGYPTIAN_HIEROGLYPH_1433B = 0x0001433B;
const char32_t EGYPTIAN_HIEROGLYPH_1433C = 0x0001433C;
const char32_t EGYPTIAN_HIEROGLYPH_1433D = 0x0001433D;
const char32_t EGYPTIAN_HIEROGLYPH_1433E = 0x0001433E;
const char32_t EGYPTIAN_HIEROGLYPH_1433F = 0x0001433F;
const char32_t EGYPTIAN_HIEROGLYPH_14340 = 0x00014340;
const char32_t EGYPTIAN_HIEROGLYPH_14341 = 0x00014341;
const char32_t EGYPTIAN_HIEROGLYPH_14342 = 0x00014342;
const char32_t EGYPTIAN_HIEROGLYPH_14343 = 0x00014343;
const char32_t EGYPTIAN_HIEROGLYPH_14344 = 0x00014344;
const char32_t EGYPTIAN_HIEROGLYPH_14345 = 0x00014345;
const char32_t EGYPTIAN_HIEROGLYPH_14346 = 0x00014346;
const char32_t EGYPTIAN_HIEROGLYPH_14347 = 0x00014347;
const char32_t EGYPTIAN_HIEROGLYPH_14348 = 0x00014348;
const char32_t EGYPTIAN_HIEROGLYPH_14349 = 0x00014349;
const char32_t EGYPTIAN_HIEROGLYPH_1434A = 0x0001434A;
const char32_t EGYPTIAN_HIEROGLYPH_1434B = 0x0001434B;
const char32_t EGYPTIAN_HIEROGLYPH_1434C = 0x0001434C;
const char32_t EGYPTIAN_HIEROGLYPH_1434D = 0x0001434D;
const char32_t EGYPTIAN_HIEROGLYPH_1434E = 0x0001434E;
const char32_t EGYPTIAN_HIEROGLYPH_1434F = 0x0001434F;
const char32_t EGYPTIAN_HIEROGLYPH_14350 = 0x00014350;
const char32_t EGYPTIAN_HIEROGLYPH_14351 = 0x00014351;
const char32_t EGYPTIAN_HIEROGLYPH_14352 = 0x00014352;
const char32_t EGYPTIAN_HIEROGLYPH_14353 = 0x00014353;
const char32_t EGYPTIAN_HIEROGLYPH_14354 = 0x00014354;
const char32_t EGYPTIAN_HIEROGLYPH_14355 = 0x00014355;
const char32_t EGYPTIAN_HIEROGLYPH_14356 = 0x00014356;
const char32_t EGYPTIAN_HIEROGLYPH_14357 = 0x00014357;
const char32_t EGYPTIAN_HIEROGLYPH_14358 = 0x00014358;
const char32_t EGYPTIAN_HIEROGLYPH_14359 = 0x00014359;
const char32_t EGYPTIAN_HIEROGLYPH_1435A = 0x0001435A;
const char32_t EGYPTIAN_HIEROGLYPH_1435B = 0x0001435B;
const char32_t EGYPTIAN_HIEROGLYPH_1435C = 0x0001435C;
const char32_t EGYPTIAN_HIEROGLYPH_1435D = 0x0001435D;
const char32_t EGYPTIAN_HIEROGLYPH_1435E = 0x0001435E;
const char32_t EGYPTIAN_HIEROGLYPH_1435F = 0x0001435F;
const char32_t EGYPTIAN_HIEROGLYPH_14360 = 0x00014360;
const char32_t EGYPTIAN_HIEROGLYPH_14361 = 0x00014361;
const char32_t EGYPTIAN_HIEROGLYPH_14362 = 0x00014362;
const char32_t EGYPTIAN_HIEROGLYPH_14363 = 0x00014363;
const char32_t EGYPTIAN_HIEROGLYPH_14364 = 0x00014364;
const char32_t EGYPTIAN_HIEROGLYPH_14365 = 0x00014365;
const char32_t EGYPTIAN_HIEROGLYPH_14366 = 0x00014366;
const char32_t EGYPTIAN_HIEROGLYPH_14367 = 0x00014367;
const char32_t EGYPTIAN_HIEROGLYPH_14368 = 0x00014368;
const char32_t EGYPTIAN_HIEROGLYPH_14369 = 0x00014369;
const char32_t EGYPTIAN_HIEROGLYPH_1436A = 0x0001436A;
const char32_t EGYPTIAN_HIEROGLYPH_1436B = 0x0001436B;
const char32_t EGYPTIAN_HIEROGLYPH_1436C = 0x0001436C;
const char32_t EGYPTIAN_HIEROGLYPH_1436D = 0x0001436D;
const char32_t EGYPTIAN_HIEROGLYPH_1436E = 0x0001436E;
const char32_t EGYPTIAN_HIEROGLYPH_1436F = 0x0001436F;
const char32_t EGYPTIAN_HIEROGLYPH_14370 = 0x00014370;
const char32_t EGYPTIAN_HIEROGLYPH_14371 = 0x00014371;
const char32_t EGYPTIAN_HIEROGLYPH_14372 = 0x00014372;
const char32_t EGYPTIAN_HIEROGLYPH_14373 = 0x00014373;
const char32_t EGYPTIAN_HIEROGLYPH_14374 = 0x00014374;
const char32_t EGYPTIAN_HIEROGLYPH_14375 = 0x00014375;
const char32_t EGYPTIAN_HIEROGLYPH_14376 = 0x00014376;
const char32_t EGYPTIAN_HIEROGLYPH_14377 = 0x00014377;
const char32_t EGYPTIAN_HIEROGLYPH_14378 = 0x00014378;
const char32_t EGYPTIAN_HIEROGLYPH_14379 = 0x00014379;
const char32_t EGYPTIAN_HIEROGLYPH_1437A = 0x0001437A;
const char32_t EGYPTIAN_HIEROGLYPH_1437B = 0x0001437B;
const char32_t EGYPTIAN_HIEROGLYPH_1437C = 0x0001437C;
const char32_t EGYPTIAN_HIEROGLYPH_1437D = 0x0001437D;
const char32_t EGYPTIAN_HIEROGLYPH_1437E = 0x0001437E;
const char32_t EGYPTIAN_HIEROGLYPH_1437F = 0x0001437F;
const char32_t EGYPTIAN_HIEROGLYPH_14380 = 0x00014380;
const char32_t EGYPTIAN_HIEROGLYPH_14381 = 0x00014381;
const char32_t EGYPTIAN_HIEROGLYPH_14382 = 0x00014382;
const char32_t EGYPTIAN_HIEROGLYPH_14383 = 0x00014383;
const char32_t EGYPTIAN_HIEROGLYPH_14384 = 0x00014384;
const char32_t EGYPTIAN_HIEROGLYPH_14385 = 0x00014385;
const char32_t EGYPTIAN_HIEROGLYPH_14386 = 0x00014386;
const char32_t EGYPTIAN_HIEROGLYPH_14387 = 0x00014387;
const char32_t EGYPTIAN_HIEROGLYPH_14388 = 0x00014388;
const char32_t EGYPTIAN_HIEROGLYPH_14389 = 0x00014389;
const char32_t EGYPTIAN_HIEROGLYPH_1438A = 0x0001438A;
const char32_t EGYPTIAN_HIEROGLYPH_1438B = 0x0001438B;
const char32_t EGYPTIAN_HIEROGLYPH_1438C = 0x0001438C;
const char32_t EGYPTIAN_HIEROGLYPH_1438D = 0x0001438D;
const char32_t EGYPTIAN_HIEROGLYPH_1438E = 0x0001438E;
const char32_t EGYPTIAN_HIEROGLYPH_1438F = 0x0001438F;
const char32_t EGYPTIAN_HIEROGLYPH_14390 = 0x00014390;
const char32_t EGYPTIAN_HIEROGLYPH_14391 = 0x00014391;
const char32_t EGYPTIAN_HIEROGLYPH_14392 = 0x00014392;
const char32_t EGYPTIAN_HIEROGLYPH_14393 = 0x00014393;
const char32_t EGYPTIAN_HIEROGLYPH_14394 = 0x00014394;
const char32_t EGYPTIAN_HIEROGLYPH_14395 = 0x00014395;
const char32_t EGYPTIAN_HIEROGLYPH_14396 = 0x00014396;
const char32_t EGYPTIAN_HIEROGLYPH_14397 = 0x00014397;
const char32_t EGYPTIAN_HIEROGLYPH_14398 = 0x00014398;
const char32_t EGYPTIAN_HIEROGLYPH_14399 = 0x00014399;
const char32_t EGYPTIAN_HIEROGLYPH_1439A = 0x0001439A;
const char32_t EGYPTIAN_HIEROGLYPH_1439B = 0x0001439B;
const char32_t EGYPTIAN_HIEROGLYPH_1439C = 0x0001439C;
const char32_t EGYPTIAN_HIEROGLYPH_1439D = 0x0001439D;
const char32_t EGYPTIAN_HIEROGLYPH_1439E = 0x0001439E;
const char32_t EGYPTIAN_HIEROGLYPH_1439F = 0x0001439F;
const char32_t EGYPTIAN_HIEROGLYPH_143A0 = 0x000143A0;
const char32_t EGYPTIAN_HIEROGLYPH_143A1 = 0x000143A1;
const char32_t EGYPTIAN_HIEROGLYPH_143A2 = 0x000143A2;
const char32_t EGYPTIAN_HIEROGLYPH_143A3 = 0x000143A3;
const char32_t EGYPTIAN_HIEROGLYPH_143A4 = 0x000143A4;
const char32_t EGYPTIAN_HIEROGLYPH_143A5 = 0x000143A5;
const char32_t EGYPTIAN_HIEROGLYPH_143A6 = 0x000143A6;
const char32_t EGYPTIAN_HIEROGLYPH_143A7 = 0x000143A7;
const char32_t EGYPTIAN_HIEROGLYPH_143A8 = 0x000143A8;
const char32_t EGYPTIAN_HIEROGLYPH_143A9 = 0x000143A9;
const char32_t EGYPTIAN_HIEROGLYPH_143AA = 0x000143AA;
const char32_t EGYPTIAN_HIEROGLYPH_143AB = 0x000143AB;
const char32_t EGYPTIAN_HIEROGLYPH_143AC = 0x000143AC;
const char32_t EGYPTIAN_HIEROGLYPH_143AD = 0x000143AD;
const char32_t EGYPTIAN_HIEROGLYPH_143AE = 0x000143AE;
const char32_t EGYPTIAN_HIEROGLYPH_143AF = 0x000143AF;
const char32_t EGYPTIAN_HIEROGLYPH_143B0 = 0x000143B0;
const char32_t EGYPTIAN_HIEROGLYPH_143B1 = 0x000143B1;
const char32_t EGYPTIAN_HIEROGLYPH_143B2 = 0x000143B2;
const char32_t EGYPTIAN_HIEROGLYPH_143B3 = 0x000143B3;
const char32_t EGYPTIAN_HIEROGLYPH_143B4 = 0x000143B4;
const char32_t EGYPTIAN_HIEROGLYPH_143B5 = 0x000143B5;
const char32_t EGYPTIAN_HIEROGLYPH_143B6 = 0x000143B6;
const char32_t EGYPTIAN_HIEROGLYPH_143B7 = 0x000143B7;
const char32_t EGYPTIAN_HIEROGLYPH_143B8 = 0x000143B8;
const char32_t EGYPTIAN_HIEROGLYPH_143B9 = 0x000143B9;
const char32_t EGYPTIAN_HIEROGLYPH_143BA = 0x000143BA;
const char32_t EGYPTIAN_HIEROGLYPH_143BB = 0x000143BB;
const char32_t EGYPTIAN_HIEROGLYPH_143BC = 0x000143BC;
const char32_t EGYPTIAN_HIEROGLYPH_143BD = 0x000143BD;
const char32_t EGYPTIAN_HIEROGLYPH_143BE = 0x000143BE;
const char32_t EGYPTIAN_HIEROGLYPH_143BF = 0x000143BF;
const char32_t EGYPTIAN_HIEROGLYPH_143C0 = 0x000143C0;
const char32_t EGYPTIAN_HIEROGLYPH_143C1 = 0x000143C1;
const char32_t EGYPTIAN_HIEROGLYPH_143C2 = 0x000143C2;
const char32_t EGYPTIAN_HIEROGLYPH_143C3 = 0x000143C3;
const char32_t EGYPTIAN_HIEROGLYPH_143C4 = 0x000143C4;
const char32_t EGYPTIAN_HIEROGLYPH_143C5 = 0x000143C5;
const char32_t EGYPTIAN_HIEROGLYPH_143C6 = 0x000143C6;
const char32_t EGYPTIAN_HIEROGLYPH_143C7 = 0x000143C7;
const char32_t EGYPTIAN_HIEROGLYPH_143C8 = 0x000143C8;
const char32_t EGYPTIAN_HIEROGLYPH_143C9 = 0x000143C9;
const char32_t EGYPTIAN_HIEROGLYPH_143CA = 0x000143CA;
const char32_t EGYPTIAN_HIEROGLYPH_143CB = 0x000143CB;
const char32_t EGYPTIAN_HIEROGLYPH_143CC = 0x000143CC;
const char32_t EGYPTIAN_HIEROGLYPH_143CD = 0x000143CD;
const char32_t EGYPTIAN_HIEROGLYPH_143CE = 0x000143CE;
const char32_t EGYPTIAN_HIEROGLYPH_143CF = 0x000143CF;
const char32_t EGYPTIAN_HIEROGLYPH_143D0 = 0x000143D0;
const char32_t EGYPTIAN_HIEROGLYPH_143D1 = 0x000143D1;
const char32_t EGYPTIAN_HIEROGLYPH_143D2 = 0x000143D2;
const char32_t EGYPTIAN_HIEROGLYPH_143D3 = 0x000143D3;
const char32_t EGYPTIAN_HIEROGLYPH_143D4 = 0x000143D4;
const char32_t EGYPTIAN_HIEROGLYPH_143D5 = 0x000143D5;
const char32_t EGYPTIAN_HIEROGLYPH_143D6 = 0x000143D6;
const char32_t EGYPTIAN_HIEROGLYPH_143D7 = 0x000143D7;
const char32_t EGYPTIAN_HIEROGLYPH_143D8 = 0x000143D8;
const char32_t EGYPTIAN_HIEROGLYPH_143D9 = 0x000143D9;
const char32_t EGYPTIAN_HIEROGLYPH_143DA = 0x000143DA;
const char32_t EGYPTIAN_HIEROGLYPH_143DB = 0x000143DB;
const char32_t EGYPTIAN_HIEROGLYPH_143DC = 0x000143DC;
const char32_t EGYPTIAN_HIEROGLYPH_143DD = 0x000143DD;
const char32_t EGYPTIAN_HIEROGLYPH_143DE = 0x000143DE;
const char32_t EGYPTIAN_HIEROGLYPH_143DF = 0x000143DF;
const char32_t EGYPTIAN_HIEROGLYPH_143E0 = 0x000143E0;
const char32_t EGYPTIAN_HIEROGLYPH_143E1 = 0x000143E1;
const char32_t EGYPTIAN_HIEROGLYPH_143E2 = 0x000143E2;
const char32_t EGYPTIAN_HIEROGLYPH_143E3 = 0x000143E3;
const char32_t EGYPTIAN_HIEROGLYPH_143E4 = 0x000143E4;
const char32_t EGYPTIAN_HIEROGLYPH_143E5 = 0x000143E5;
const char32_t EGYPTIAN_HIEROGLYPH_143E6 = 0x000143E6;
const char32_t EGYPTIAN_HIEROGLYPH_143E7 = 0x000143E7;
const char32_t EGYPTIAN_HIEROGLYPH_143E8 = 0x000143E8;
const char32_t EGYPTIAN_HIEROGLYPH_143E9 = 0x000143E9;
const char32_t EGYPTIAN_HIEROGLYPH_143EA = 0x000143EA;
const char32_t EGYPTIAN_HIEROGLYPH_143EB = 0x000143EB;
const char32_t EGYPTIAN_HIEROGLYPH_143EC = 0x000143EC;
const char32_t EGYPTIAN_HIEROGLYPH_143ED = 0x000143ED;
const char32_t EGYPTIAN_HIEROGLYPH_143EE = 0x000143EE;
const char32_t EGYPTIAN_HIEROGLYPH_143EF = 0x000143EF;
const char32_t EGYPTIAN_HIEROGLYPH_143F0 = 0x000143F0;
const char32_t EGYPTIAN_HIEROGLYPH_143F1 = 0x000143F1;
const char32_t EGYPTIAN_HIEROGLYPH_143F2 = 0x000143F2;
const char32_t EGYPTIAN_HIEROGLYPH_143F3 = 0x000143F3;
const char32_t EGYPTIAN_HIEROGLYPH_143F4 = 0x000143F4;
const char32_t EGYPTIAN_HIEROGLYPH_143F5 = 0x000143F5;
const char32_t EGYPTIAN_HIEROGLYPH_143F6 = 0x000143F6;
const char32_t EGYPTIAN_HIEROGLYPH_143F7 = 0x000143F7;
const char32_t EGYPTIAN_HIEROGLYPH_143F8 = 0x000143F8;
const char32_t EGYPTIAN_HIEROGLYPH_143F9 = 0x000143F9;
const char32_t EGYPTIAN_HIEROGLYPH_143FA = 0x000143FA;
const char32_t ANATOLIAN_HIEROGLYPH_A001 = 0x00014400;
const char32_t ANATOLIAN_HIEROGLYPH_A002 = 0x00014401;
const char32_t ANATOLIAN_HIEROGLYPH_A003 = 0x00014402;
const char32_t ANATOLIAN_HIEROGLYPH_A004 = 0x00014403;
const char32_t ANATOLIAN_HIEROGLYPH_A005 = 0x00014404;
const char32_t ANATOLIAN_HIEROGLYPH_A006 = 0x00014405;
const char32_t ANATOLIAN_HIEROGLYPH_A007 = 0x00014406;
const char32_t ANATOLIAN_HIEROGLYPH_A008 = 0x00014407;
const char32_t ANATOLIAN_HIEROGLYPH_A009 = 0x00014408;
const char32_t ANATOLIAN_HIEROGLYPH_A010 = 0x00014409;
const char32_t ANATOLIAN_HIEROGLYPH_A010A = 0x0001440A;
const char32_t ANATOLIAN_HIEROGLYPH_A011 = 0x0001440B;
const char32_t ANATOLIAN_HIEROGLYPH_A012 = 0x0001440C;
const char32_t ANATOLIAN_HIEROGLYPH_A013 = 0x0001440D;
const char32_t ANATOLIAN_HIEROGLYPH_A014 = 0x0001440E;
const char32_t ANATOLIAN_HIEROGLYPH_A015 = 0x0001440F;
const char32_t ANATOLIAN_HIEROGLYPH_A016 = 0x00014410;
const char32_t ANATOLIAN_HIEROGLYPH_A017 = 0x00014411;
const char32_t ANATOLIAN_HIEROGLYPH_A018 = 0x00014412;
const char32_t ANATOLIAN_HIEROGLYPH_A019 = 0x00014413;
const char32_t ANATOLIAN_HIEROGLYPH_A020 = 0x00014414;
const char32_t ANATOLIAN_HIEROGLYPH_A021 = 0x00014415;
const char32_t ANATOLIAN_HIEROGLYPH_A022 = 0x00014416;
const char32_t ANATOLIAN_HIEROGLYPH_A023 = 0x00014417;
const char32_t ANATOLIAN_HIEROGLYPH_A024 = 0x00014418;
const char32_t ANATOLIAN_HIEROGLYPH_A025 = 0x00014419;
const char32_t ANATOLIAN_HIEROGLYPH_A026 = 0x0001441A;
const char32_t ANATOLIAN_HIEROGLYPH_A026A = 0x0001441B;
const char32_t ANATOLIAN_HIEROGLYPH_A027 = 0x0001441C;
const char32_t ANATOLIAN_HIEROGLYPH_A028 = 0x0001441D;
const char32_t ANATOLIAN_HIEROGLYPH_A029 = 0x0001441E;
const char32_t ANATOLIAN_HIEROGLYPH_A030 = 0x0001441F;
const char32_t ANATOLIAN_HIEROGLYPH_A031 = 0x00014420;
const char32_t ANATOLIAN_HIEROGLYPH_A032 = 0x00014421;
const char32_t ANATOLIAN_HIEROGLYPH_A033 = 0x00014422;
const char32_t ANATOLIAN_HIEROGLYPH_A034 = 0x00014423;
const char32_t ANATOLIAN_HIEROGLYPH_A035 = 0x00014424;
const char32_t ANATOLIAN_HIEROGLYPH_A036 = 0x00014425;
const char32_t ANATOLIAN_HIEROGLYPH_A037 = 0x00014426;
const char32_t ANATOLIAN_HIEROGLYPH_A038 = 0x00014427;
const char32_t ANATOLIAN_HIEROGLYPH_A039 = 0x00014428;
const char32_t ANATOLIAN_HIEROGLYPH_A039A = 0x00014429;
const char32_t ANATOLIAN_HIEROGLYPH_A040 = 0x0001442A;
const char32_t ANATOLIAN_HIEROGLYPH_A041 = 0x0001442B;
const char32_t ANATOLIAN_HIEROGLYPH_A041A = 0x0001442C;
const char32_t ANATOLIAN_HIEROGLYPH_A042 = 0x0001442D;
const char32_t ANATOLIAN_HIEROGLYPH_A043 = 0x0001442E;
const char32_t ANATOLIAN_HIEROGLYPH_A044 = 0x0001442F;
const char32_t ANATOLIAN_HIEROGLYPH_A045 = 0x00014430;
const char32_t ANATOLIAN_HIEROGLYPH_A045A = 0x00014431;
const char32_t ANATOLIAN_HIEROGLYPH_A046 = 0x00014432;
const char32_t ANATOLIAN_HIEROGLYPH_A046A = 0x00014433;
const char32_t ANATOLIAN_HIEROGLYPH_A046B = 0x00014434;
const char32_t ANATOLIAN_HIEROGLYPH_A047 = 0x00014435;
const char32_t ANATOLIAN_HIEROGLYPH_A048 = 0x00014436;
const char32_t ANATOLIAN_HIEROGLYPH_A049 = 0x00014437;
const char32_t ANATOLIAN_HIEROGLYPH_A050 = 0x00014438;
const char32_t ANATOLIAN_HIEROGLYPH_A051 = 0x00014439;
const char32_t ANATOLIAN_HIEROGLYPH_A052 = 0x0001443A;
const char32_t ANATOLIAN_HIEROGLYPH_A053 = 0x0001443B;
const char32_t ANATOLIAN_HIEROGLYPH_A054 = 0x0001443C;
const char32_t ANATOLIAN_HIEROGLYPH_A055 = 0x0001443D;
const char32_t ANATOLIAN_HIEROGLYPH_A056 = 0x0001443E;
const char32_t ANATOLIAN_HIEROGLYPH_A057 = 0x0001443F;
const char32_t ANATOLIAN_HIEROGLYPH_A058 = 0x00014440;
const char32_t ANATOLIAN_HIEROGLYPH_A059 = 0x00014441;
const char32_t ANATOLIAN_HIEROGLYPH_A060 = 0x00014442;
const char32_t ANATOLIAN_HIEROGLYPH_A061 = 0x00014443;
const char32_t ANATOLIAN_HIEROGLYPH_A062 = 0x00014444;
const char32_t ANATOLIAN_HIEROGLYPH_A063 = 0x00014445;
const char32_t ANATOLIAN_HIEROGLYPH_A064 = 0x00014446;
const char32_t ANATOLIAN_HIEROGLYPH_A065 = 0x00014447;
const char32_t ANATOLIAN_HIEROGLYPH_A066 = 0x00014448;
const char32_t ANATOLIAN_HIEROGLYPH_A066A = 0x00014449;
const char32_t ANATOLIAN_HIEROGLYPH_A066B = 0x0001444A;
const char32_t ANATOLIAN_HIEROGLYPH_A066C = 0x0001444B;
const char32_t ANATOLIAN_HIEROGLYPH_A067 = 0x0001444C;
const char32_t ANATOLIAN_HIEROGLYPH_A068 = 0x0001444D;
const char32_t ANATOLIAN_HIEROGLYPH_A069 = 0x0001444E;
const char32_t ANATOLIAN_HIEROGLYPH_A070 = 0x0001444F;
const char32_t ANATOLIAN_HIEROGLYPH_A071 = 0x00014450;
const char32_t ANATOLIAN_HIEROGLYPH_A072 = 0x00014451;
const char32_t ANATOLIAN_HIEROGLYPH_A073 = 0x00014452;
const char32_t ANATOLIAN_HIEROGLYPH_A074 = 0x00014453;
const char32_t ANATOLIAN_HIEROGLYPH_A075 = 0x00014454;
const char32_t ANATOLIAN_HIEROGLYPH_A076 = 0x00014455;
const char32_t ANATOLIAN_HIEROGLYPH_A077 = 0x00014456;
const char32_t ANATOLIAN_HIEROGLYPH_A078 = 0x00014457;
const char32_t ANATOLIAN_HIEROGLYPH_A079 = 0x00014458;
const char32_t ANATOLIAN_HIEROGLYPH_A080 = 0x00014459;
const char32_t ANATOLIAN_HIEROGLYPH_A081 = 0x0001445A;
const char32_t ANATOLIAN_HIEROGLYPH_A082 = 0x0001445B;
const char32_t ANATOLIAN_HIEROGLYPH_A083 = 0x0001445C;
const char32_t ANATOLIAN_HIEROGLYPH_A084 = 0x0001445D;
const char32_t ANATOLIAN_HIEROGLYPH_A085 = 0x0001445E;
const char32_t ANATOLIAN_HIEROGLYPH_A086 = 0x0001445F;
const char32_t ANATOLIAN_HIEROGLYPH_A087 = 0x00014460;
const char32_t ANATOLIAN_HIEROGLYPH_A088 = 0x00014461;
const char32_t ANATOLIAN_HIEROGLYPH_A089 = 0x00014462;
const char32_t ANATOLIAN_HIEROGLYPH_A090 = 0x00014463;
const char32_t ANATOLIAN_HIEROGLYPH_A091 = 0x00014464;
const char32_t ANATOLIAN_HIEROGLYPH_A092 = 0x00014465;
const char32_t ANATOLIAN_HIEROGLYPH_A093 = 0x00014466;
const char32_t ANATOLIAN_HIEROGLYPH_A094 = 0x00014467;
const char32_t ANATOLIAN_HIEROGLYPH_A095 = 0x00014468;
const char32_t ANATOLIAN_HIEROGLYPH_A096 = 0x00014469;
const char32_t ANATOLIAN_HIEROGLYPH_A097 = 0x0001446A;
const char32_t ANATOLIAN_HIEROGLYPH_A097A = 0x0001446B;
const char32_t ANATOLIAN_HIEROGLYPH_A098 = 0x0001446C;
const char32_t ANATOLIAN_HIEROGLYPH_A098A = 0x0001446D;
const char32_t ANATOLIAN_HIEROGLYPH_A099 = 0x0001446E;
const char32_t ANATOLIAN_HIEROGLYPH_A100 = 0x0001446F;
const char32_t ANATOLIAN_HIEROGLYPH_A100A = 0x00014470;
const char32_t ANATOLIAN_HIEROGLYPH_A101 = 0x00014471;
const char32_t ANATOLIAN_HIEROGLYPH_A101A = 0x00014472;
const char32_t ANATOLIAN_HIEROGLYPH_A102 = 0x00014473;
const char32_t ANATOLIAN_HIEROGLYPH_A102A = 0x00014474;
const char32_t ANATOLIAN_HIEROGLYPH_A103 = 0x00014475;
const char32_t ANATOLIAN_HIEROGLYPH_A104 = 0x00014476;
const char32_t ANATOLIAN_HIEROGLYPH_A104A = 0x00014477;
const char32_t ANATOLIAN_HIEROGLYPH_A104B = 0x00014478;
const char32_t ANATOLIAN_HIEROGLYPH_A104C = 0x00014479;
const char32_t ANATOLIAN_HIEROGLYPH_A105 = 0x0001447A;
const char32_t ANATOLIAN_HIEROGLYPH_A105A = 0x0001447B;
const char32_t ANATOLIAN_HIEROGLYPH_A105B = 0x0001447C;
const char32_t ANATOLIAN_HIEROGLYPH_A106 = 0x0001447D;
const char32_t ANATOLIAN_HIEROGLYPH_A107 = 0x0001447E;
const char32_t ANATOLIAN_HIEROGLYPH_A107A = 0x0001447F;
const char32_t ANATOLIAN_HIEROGLYPH_A107B = 0x00014480;
const char32_t ANATOLIAN_HIEROGLYPH_A107C = 0x00014481;
const char32_t ANATOLIAN_HIEROGLYPH_A108 = 0x00014482;
const char32_t ANATOLIAN_HIEROGLYPH_A109 = 0x00014483;
const char32_t ANATOLIAN_HIEROGLYPH_A110 = 0x00014484;
const char32_t ANATOLIAN_HIEROGLYPH_A110A = 0x00014485;
const char32_t ANATOLIAN_HIEROGLYPH_A110B = 0x00014486;
const char32_t ANATOLIAN_HIEROGLYPH_A111 = 0x00014487;
const char32_t ANATOLIAN_HIEROGLYPH_A112 = 0x00014488;
const char32_t ANATOLIAN_HIEROGLYPH_A113 = 0x00014489;
const char32_t ANATOLIAN_HIEROGLYPH_A114 = 0x0001448A;
const char32_t ANATOLIAN_HIEROGLYPH_A115 = 0x0001448B;
const char32_t ANATOLIAN_HIEROGLYPH_A115A = 0x0001448C;
const char32_t ANATOLIAN_HIEROGLYPH_A116 = 0x0001448D;
const char32_t ANATOLIAN_HIEROGLYPH_A117 = 0x0001448E;
const char32_t ANATOLIAN_HIEROGLYPH_A118 = 0x0001448F;
const char32_t ANATOLIAN_HIEROGLYPH_A119 = 0x00014490;
const char32_t ANATOLIAN_HIEROGLYPH_A120 = 0x00014491;
const char32_t ANATOLIAN_HIEROGLYPH_A121 = 0x00014492;
const char32_t ANATOLIAN_HIEROGLYPH_A122 = 0x00014493;
const char32_t ANATOLIAN_HIEROGLYPH_A123 = 0x00014494;
const char32_t ANATOLIAN_HIEROGLYPH_A124 = 0x00014495;
const char32_t ANATOLIAN_HIEROGLYPH_A125 = 0x00014496;
const char32_t ANATOLIAN_HIEROGLYPH_A125A = 0x00014497;
const char32_t ANATOLIAN_HIEROGLYPH_A126 = 0x00014498;
const char32_t ANATOLIAN_HIEROGLYPH_A127 = 0x00014499;
const char32_t ANATOLIAN_HIEROGLYPH_A128 = 0x0001449A;
const char32_t ANATOLIAN_HIEROGLYPH_A129 = 0x0001449B;
const char32_t ANATOLIAN_HIEROGLYPH_A130 = 0x0001449C;
const char32_t ANATOLIAN_HIEROGLYPH_A131 = 0x0001449D;
const char32_t ANATOLIAN_HIEROGLYPH_A132 = 0x0001449E;
const char32_t ANATOLIAN_HIEROGLYPH_A133 = 0x0001449F;
const char32_t ANATOLIAN_HIEROGLYPH_A134 = 0x000144A0;
const char32_t ANATOLIAN_HIEROGLYPH_A135 = 0x000144A1;
const char32_t ANATOLIAN_HIEROGLYPH_A135A = 0x000144A2;
const char32_t ANATOLIAN_HIEROGLYPH_A136 = 0x000144A3;
const char32_t ANATOLIAN_HIEROGLYPH_A137 = 0x000144A4;
const char32_t ANATOLIAN_HIEROGLYPH_A138 = 0x000144A5;
const char32_t ANATOLIAN_HIEROGLYPH_A139 = 0x000144A6;
const char32_t ANATOLIAN_HIEROGLYPH_A140 = 0x000144A7;
const char32_t ANATOLIAN_HIEROGLYPH_A141 = 0x000144A8;
const char32_t ANATOLIAN_HIEROGLYPH_A142 = 0x000144A9;
const char32_t ANATOLIAN_HIEROGLYPH_A143 = 0x000144AA;
const char32_t ANATOLIAN_HIEROGLYPH_A144 = 0x000144AB;
const char32_t ANATOLIAN_HIEROGLYPH_A145 = 0x000144AC;
const char32_t ANATOLIAN_HIEROGLYPH_A146 = 0x000144AD;
const char32_t ANATOLIAN_HIEROGLYPH_A147 = 0x000144AE;
const char32_t ANATOLIAN_HIEROGLYPH_A148 = 0x000144AF;
const char32_t ANATOLIAN_HIEROGLYPH_A149 = 0x000144B0;
const char32_t ANATOLIAN_HIEROGLYPH_A150 = 0x000144B1;
const char32_t ANATOLIAN_HIEROGLYPH_A151 = 0x000144B2;
const char32_t ANATOLIAN_HIEROGLYPH_A152 = 0x000144B3;
const char32_t ANATOLIAN_HIEROGLYPH_A153 = 0x000144B4;
const char32_t ANATOLIAN_HIEROGLYPH_A154 = 0x000144B5;
const char32_t ANATOLIAN_HIEROGLYPH_A155 = 0x000144B6;
const char32_t ANATOLIAN_HIEROGLYPH_A156 = 0x000144B7;
const char32_t ANATOLIAN_HIEROGLYPH_A157 = 0x000144B8;
const char32_t ANATOLIAN_HIEROGLYPH_A158 = 0x000144B9;
const char32_t ANATOLIAN_HIEROGLYPH_A159 = 0x000144BA;
const char32_t ANATOLIAN_HIEROGLYPH_A160 = 0x000144BB;
const char32_t ANATOLIAN_HIEROGLYPH_A161 = 0x000144BC;
const char32_t ANATOLIAN_HIEROGLYPH_A162 = 0x000144BD;
const char32_t ANATOLIAN_HIEROGLYPH_A163 = 0x000144BE;
const char32_t ANATOLIAN_HIEROGLYPH_A164 = 0x000144BF;
const char32_t ANATOLIAN_HIEROGLYPH_A165 = 0x000144C0;
const char32_t ANATOLIAN_HIEROGLYPH_A166 = 0x000144C1;
const char32_t ANATOLIAN_HIEROGLYPH_A167 = 0x000144C2;
const char32_t ANATOLIAN_HIEROGLYPH_A168 = 0x000144C3;
const char32_t ANATOLIAN_HIEROGLYPH_A169 = 0x000144C4;
const char32_t ANATOLIAN_HIEROGLYPH_A170 = 0x000144C5;
const char32_t ANATOLIAN_HIEROGLYPH_A171 = 0x000144C6;
const char32_t ANATOLIAN_HIEROGLYPH_A172 = 0x000144C7;
const char32_t ANATOLIAN_HIEROGLYPH_A173 = 0x000144C8;
const char32_t ANATOLIAN_HIEROGLYPH_A174 = 0x000144C9;
const char32_t ANATOLIAN_HIEROGLYPH_A175 = 0x000144CA;
const char32_t ANATOLIAN_HIEROGLYPH_A176 = 0x000144CB;
const char32_t ANATOLIAN_HIEROGLYPH_A177 = 0x000144CC;
const char32_t ANATOLIAN_HIEROGLYPH_A178 = 0x000144CD;
const char32_t ANATOLIAN_HIEROGLYPH_A179 = 0x000144CE;
const char32_t ANATOLIAN_HIEROGLYPH_A180 = 0x000144CF;
const char32_t ANATOLIAN_HIEROGLYPH_A181 = 0x000144D0;
const char32_t ANATOLIAN_HIEROGLYPH_A182 = 0x000144D1;
const char32_t ANATOLIAN_HIEROGLYPH_A183 = 0x000144D2;
const char32_t ANATOLIAN_HIEROGLYPH_A184 = 0x000144D3;
const char32_t ANATOLIAN_HIEROGLYPH_A185 = 0x000144D4;
const char32_t ANATOLIAN_HIEROGLYPH_A186 = 0x000144D5;
const char32_t ANATOLIAN_HIEROGLYPH_A187 = 0x000144D6;
const char32_t ANATOLIAN_HIEROGLYPH_A188 = 0x000144D7;
const char32_t ANATOLIAN_HIEROGLYPH_A189 = 0x000144D8;
const char32_t ANATOLIAN_HIEROGLYPH_A190 = 0x000144D9;
const char32_t ANATOLIAN_HIEROGLYPH_A191 = 0x000144DA;
const char32_t ANATOLIAN_HIEROGLYPH_A192 = 0x000144DB;
const char32_t ANATOLIAN_HIEROGLYPH_A193 = 0x000144DC;
const char32_t ANATOLIAN_HIEROGLYPH_A194 = 0x000144DD;
const char32_t ANATOLIAN_HIEROGLYPH_A195 = 0x000144DE;
const char32_t ANATOLIAN_HIEROGLYPH_A196 = 0x000144DF;
const char32_t ANATOLIAN_HIEROGLYPH_A197 = 0x000144E0;
const char32_t ANATOLIAN_HIEROGLYPH_A198 = 0x000144E1;
const char32_t ANATOLIAN_HIEROGLYPH_A199 = 0x000144E2;
const char32_t ANATOLIAN_HIEROGLYPH_A200 = 0x000144E3;
const char32_t ANATOLIAN_HIEROGLYPH_A201 = 0x000144E4;
const char32_t ANATOLIAN_HIEROGLYPH_A202 = 0x000144E5;
const char32_t ANATOLIAN_HIEROGLYPH_A202A = 0x000144E6;
const char32_t ANATOLIAN_HIEROGLYPH_A202B = 0x000144E7;
const char32_t ANATOLIAN_HIEROGLYPH_A203 = 0x000144E8;
const char32_t ANATOLIAN_HIEROGLYPH_A204 = 0x000144E9;
const char32_t ANATOLIAN_HIEROGLYPH_A205 = 0x000144EA;
const char32_t ANATOLIAN_HIEROGLYPH_A206 = 0x000144EB;
const char32_t ANATOLIAN_HIEROGLYPH_A207 = 0x000144EC;
const char32_t ANATOLIAN_HIEROGLYPH_A207A = 0x000144ED;
const char32_t ANATOLIAN_HIEROGLYPH_A208 = 0x000144EE;
const char32_t ANATOLIAN_HIEROGLYPH_A209 = 0x000144EF;
const char32_t ANATOLIAN_HIEROGLYPH_A209A = 0x000144F0;
const char32_t ANATOLIAN_HIEROGLYPH_A210 = 0x000144F1;
const char32_t ANATOLIAN_HIEROGLYPH_A211 = 0x000144F2;
const char32_t ANATOLIAN_HIEROGLYPH_A212 = 0x000144F3;
const char32_t ANATOLIAN_HIEROGLYPH_A213 = 0x000144F4;
const char32_t ANATOLIAN_HIEROGLYPH_A214 = 0x000144F5;
const char32_t ANATOLIAN_HIEROGLYPH_A215 = 0x000144F6;
const char32_t ANATOLIAN_HIEROGLYPH_A215A = 0x000144F7;
const char32_t ANATOLIAN_HIEROGLYPH_A216 = 0x000144F8;
const char32_t ANATOLIAN_HIEROGLYPH_A216A = 0x000144F9;
const char32_t ANATOLIAN_HIEROGLYPH_A217 = 0x000144FA;
const char32_t ANATOLIAN_HIEROGLYPH_A218 = 0x000144FB;
const char32_t ANATOLIAN_HIEROGLYPH_A219 = 0x000144FC;
const char32_t ANATOLIAN_HIEROGLYPH_A220 = 0x000144FD;
const char32_t ANATOLIAN_HIEROGLYPH_A221 = 0x000144FE;
const char32_t ANATOLIAN_HIEROGLYPH_A222 = 0x000144FF;
const char32_t ANATOLIAN_HIEROGLYPH_A223 = 0x00014500;
const char32_t ANATOLIAN_HIEROGLYPH_A224 = 0x00014501;
const char32_t ANATOLIAN_HIEROGLYPH_A225 = 0x00014502;
const char32_t ANATOLIAN_HIEROGLYPH_A226 = 0x00014503;
const char32_t ANATOLIAN_HIEROGLYPH_A227 = 0x00014504;
const char32_t ANATOLIAN_HIEROGLYPH_A227A = 0x00014505;
const char32_t ANATOLIAN_HIEROGLYPH_A228 = 0x00014506;
const char32_t ANATOLIAN_HIEROGLYPH_A229 = 0x00014507;
const char32_t ANATOLIAN_HIEROGLYPH_A230 = 0x00014508;
const char32_t ANATOLIAN_HIEROGLYPH_A231 = 0x00014509;
const char32_t ANATOLIAN_HIEROGLYPH_A232 = 0x0001450A;
const char32_t ANATOLIAN_HIEROGLYPH_A233 = 0x0001450B;
const char32_t ANATOLIAN_HIEROGLYPH_A234 = 0x0001450C;
const char32_t ANATOLIAN_HIEROGLYPH_A235 = 0x0001450D;
const char32_t ANATOLIAN_HIEROGLYPH_A236 = 0x0001450E;
const char32_t ANATOLIAN_HIEROGLYPH_A237 = 0x0001450F;
const char32_t ANATOLIAN_HIEROGLYPH_A238 = 0x00014510;
const char32_t ANATOLIAN_HIEROGLYPH_A239 = 0x00014511;
const char32_t ANATOLIAN_HIEROGLYPH_A240 = 0x00014512;
const char32_t ANATOLIAN_HIEROGLYPH_A241 = 0x00014513;
const char32_t ANATOLIAN_HIEROGLYPH_A242 = 0x00014514;
const char32_t ANATOLIAN_HIEROGLYPH_A243 = 0x00014515;
const char32_t ANATOLIAN_HIEROGLYPH_A244 = 0x00014516;
const char32_t ANATOLIAN_HIEROGLYPH_A245 = 0x00014517;
const char32_t ANATOLIAN_HIEROGLYPH_A246 = 0x00014518;
const char32_t ANATOLIAN_HIEROGLYPH_A247 = 0x00014519;
const char32_t ANATOLIAN_HIEROGLYPH_A248 = 0x0001451A;
const char32_t ANATOLIAN_HIEROGLYPH_A249 = 0x0001451B;
const char32_t ANATOLIAN_HIEROGLYPH_A250 = 0x0001451C;
const char32_t ANATOLIAN_HIEROGLYPH_A251 = 0x0001451D;
const char32_t ANATOLIAN_HIEROGLYPH_A252 = 0x0001451E;
const char32_t ANATOLIAN_HIEROGLYPH_A253 = 0x0001451F;
const char32_t ANATOLIAN_HIEROGLYPH_A254 = 0x00014520;
const char32_t ANATOLIAN_HIEROGLYPH_A255 = 0x00014521;
const char32_t ANATOLIAN_HIEROGLYPH_A256 = 0x00014522;
const char32_t ANATOLIAN_HIEROGLYPH_A257 = 0x00014523;
const char32_t ANATOLIAN_HIEROGLYPH_A258 = 0x00014524;
const char32_t ANATOLIAN_HIEROGLYPH_A259 = 0x00014525;
const char32_t ANATOLIAN_HIEROGLYPH_A260 = 0x00014526;
const char32_t ANATOLIAN_HIEROGLYPH_A261 = 0x00014527;
const char32_t ANATOLIAN_HIEROGLYPH_A262 = 0x00014528;
const char32_t ANATOLIAN_HIEROGLYPH_A263 = 0x00014529;
const char32_t ANATOLIAN_HIEROGLYPH_A264 = 0x0001452A;
const char32_t ANATOLIAN_HIEROGLYPH_A265 = 0x0001452B;
const char32_t ANATOLIAN_HIEROGLYPH_A266 = 0x0001452C;
const char32_t ANATOLIAN_HIEROGLYPH_A267 = 0x0001452D;
const char32_t ANATOLIAN_HIEROGLYPH_A267A = 0x0001452E;
const char32_t ANATOLIAN_HIEROGLYPH_A268 = 0x0001452F;
const char32_t ANATOLIAN_HIEROGLYPH_A269 = 0x00014530;
const char32_t ANATOLIAN_HIEROGLYPH_A270 = 0x00014531;
const char32_t ANATOLIAN_HIEROGLYPH_A271 = 0x00014532;
const char32_t ANATOLIAN_HIEROGLYPH_A272 = 0x00014533;
const char32_t ANATOLIAN_HIEROGLYPH_A273 = 0x00014534;
const char32_t ANATOLIAN_HIEROGLYPH_A274 = 0x00014535;
const char32_t ANATOLIAN_HIEROGLYPH_A275 = 0x00014536;
const char32_t ANATOLIAN_HIEROGLYPH_A276 = 0x00014537;
const char32_t ANATOLIAN_HIEROGLYPH_A277 = 0x00014538;
const char32_t ANATOLIAN_HIEROGLYPH_A278 = 0x00014539;
const char32_t ANATOLIAN_HIEROGLYPH_A279 = 0x0001453A;
const char32_t ANATOLIAN_HIEROGLYPH_A280 = 0x0001453B;
const char32_t ANATOLIAN_HIEROGLYPH_A281 = 0x0001453C;
const char32_t ANATOLIAN_HIEROGLYPH_A282 = 0x0001453D;
const char32_t ANATOLIAN_HIEROGLYPH_A283 = 0x0001453E;
const char32_t ANATOLIAN_HIEROGLYPH_A284 = 0x0001453F;
const char32_t ANATOLIAN_HIEROGLYPH_A285 = 0x00014540;
const char32_t ANATOLIAN_HIEROGLYPH_A286 = 0x00014541;
const char32_t ANATOLIAN_HIEROGLYPH_A287 = 0x00014542;
const char32_t ANATOLIAN_HIEROGLYPH_A288 = 0x00014543;
const char32_t ANATOLIAN_HIEROGLYPH_A289 = 0x00014544;
const char32_t ANATOLIAN_HIEROGLYPH_A289A = 0x00014545;
const char32_t ANATOLIAN_HIEROGLYPH_A290 = 0x00014546;
const char32_t ANATOLIAN_HIEROGLYPH_A291 = 0x00014547;
const char32_t ANATOLIAN_HIEROGLYPH_A292 = 0x00014548;
const char32_t ANATOLIAN_HIEROGLYPH_A293 = 0x00014549;
const char32_t ANATOLIAN_HIEROGLYPH_A294 = 0x0001454A;
const char32_t ANATOLIAN_HIEROGLYPH_A294A = 0x0001454B;
const char32_t ANATOLIAN_HIEROGLYPH_A295 = 0x0001454C;
const char32_t ANATOLIAN_HIEROGLYPH_A296 = 0x0001454D;
const char32_t ANATOLIAN_HIEROGLYPH_A297 = 0x0001454E;
const char32_t ANATOLIAN_HIEROGLYPH_A298 = 0x0001454F;
const char32_t ANATOLIAN_HIEROGLYPH_A299 = 0x00014550;
const char32_t ANATOLIAN_HIEROGLYPH_A299A = 0x00014551;
const char32_t ANATOLIAN_HIEROGLYPH_A300 = 0x00014552;
const char32_t ANATOLIAN_HIEROGLYPH_A301 = 0x00014553;
const char32_t ANATOLIAN_HIEROGLYPH_A302 = 0x00014554;
const char32_t ANATOLIAN_HIEROGLYPH_A303 = 0x00014555;
const char32_t ANATOLIAN_HIEROGLYPH_A304 = 0x00014556;
const char32_t ANATOLIAN_HIEROGLYPH_A305 = 0x00014557;
const char32_t ANATOLIAN_HIEROGLYPH_A306 = 0x00014558;
const char32_t ANATOLIAN_HIEROGLYPH_A307 = 0x00014559;
const char32_t ANATOLIAN_HIEROGLYPH_A308 = 0x0001455A;
const char32_t ANATOLIAN_HIEROGLYPH_A309 = 0x0001455B;
const char32_t ANATOLIAN_HIEROGLYPH_A309A = 0x0001455C;
const char32_t ANATOLIAN_HIEROGLYPH_A310 = 0x0001455D;
const char32_t ANATOLIAN_HIEROGLYPH_A311 = 0x0001455E;
const char32_t ANATOLIAN_HIEROGLYPH_A312 = 0x0001455F;
const char32_t ANATOLIAN_HIEROGLYPH_A313 = 0x00014560;
const char32_t ANATOLIAN_HIEROGLYPH_A314 = 0x00014561;
const char32_t ANATOLIAN_HIEROGLYPH_A315 = 0x00014562;
const char32_t ANATOLIAN_HIEROGLYPH_A316 = 0x00014563;
const char32_t ANATOLIAN_HIEROGLYPH_A317 = 0x00014564;
const char32_t ANATOLIAN_HIEROGLYPH_A318 = 0x00014565;
const char32_t ANATOLIAN_HIEROGLYPH_A319 = 0x00014566;
const char32_t ANATOLIAN_HIEROGLYPH_A320 = 0x00014567;
const char32_t ANATOLIAN_HIEROGLYPH_A321 = 0x00014568;
const char32_t ANATOLIAN_HIEROGLYPH_A322 = 0x00014569;
const char32_t ANATOLIAN_HIEROGLYPH_A323 = 0x0001456A;
const char32_t ANATOLIAN_HIEROGLYPH_A324 = 0x0001456B;
const char32_t ANATOLIAN_HIEROGLYPH_A325 = 0x0001456C;
const char32_t ANATOLIAN_HIEROGLYPH_A326 = 0x0001456D;
const char32_t ANATOLIAN_HIEROGLYPH_A327 = 0x0001456E;
const char32_t ANATOLIAN_HIEROGLYPH_A328 = 0x0001456F;
const char32_t ANATOLIAN_HIEROGLYPH_A329 = 0x00014570;
const char32_t ANATOLIAN_HIEROGLYPH_A329A = 0x00014571;
const char32_t ANATOLIAN_HIEROGLYPH_A330 = 0x00014572;
const char32_t ANATOLIAN_HIEROGLYPH_A331 = 0x00014573;
const char32_t ANATOLIAN_HIEROGLYPH_A332A = 0x00014574;
const char32_t ANATOLIAN_HIEROGLYPH_A332B = 0x00014575;
const char32_t ANATOLIAN_HIEROGLYPH_A332C = 0x00014576;
const char32_t ANATOLIAN_HIEROGLYPH_A333 = 0x00014577;
const char32_t ANATOLIAN_HIEROGLYPH_A334 = 0x00014578;
const char32_t ANATOLIAN_HIEROGLYPH_A335 = 0x00014579;
const char32_t ANATOLIAN_HIEROGLYPH_A336 = 0x0001457A;
const char32_t ANATOLIAN_HIEROGLYPH_A336A = 0x0001457B;
const char32_t ANATOLIAN_HIEROGLYPH_A336B = 0x0001457C;
const char32_t ANATOLIAN_HIEROGLYPH_A336C = 0x0001457D;
const char32_t ANATOLIAN_HIEROGLYPH_A337 = 0x0001457E;
const char32_t ANATOLIAN_HIEROGLYPH_A338 = 0x0001457F;
const char32_t ANATOLIAN_HIEROGLYPH_A339 = 0x00014580;
const char32_t ANATOLIAN_HIEROGLYPH_A340 = 0x00014581;
const char32_t ANATOLIAN_HIEROGLYPH_A341 = 0x00014582;
const char32_t ANATOLIAN_HIEROGLYPH_A342 = 0x00014583;
const char32_t ANATOLIAN_HIEROGLYPH_A343 = 0x00014584;
const char32_t ANATOLIAN_HIEROGLYPH_A344 = 0x00014585;
const char32_t ANATOLIAN_HIEROGLYPH_A345 = 0x00014586;
const char32_t ANATOLIAN_HIEROGLYPH_A346 = 0x00014587;
const char32_t ANATOLIAN_HIEROGLYPH_A347 = 0x00014588;
const char32_t ANATOLIAN_HIEROGLYPH_A348 = 0x00014589;
const char32_t ANATOLIAN_HIEROGLYPH_A349 = 0x0001458A;
const char32_t ANATOLIAN_HIEROGLYPH_A350 = 0x0001458B;
const char32_t ANATOLIAN_HIEROGLYPH_A351 = 0x0001458C;
const char32_t ANATOLIAN_HIEROGLYPH_A352 = 0x0001458D;
const char32_t ANATOLIAN_HIEROGLYPH_A353 = 0x0001458E;
const char32_t ANATOLIAN_HIEROGLYPH_A354 = 0x0001458F;
const char32_t ANATOLIAN_HIEROGLYPH_A355 = 0x00014590;
const char32_t ANATOLIAN_HIEROGLYPH_A356 = 0x00014591;
const char32_t ANATOLIAN_HIEROGLYPH_A357 = 0x00014592;
const char32_t ANATOLIAN_HIEROGLYPH_A358 = 0x00014593;
const char32_t ANATOLIAN_HIEROGLYPH_A359 = 0x00014594;
const char32_t ANATOLIAN_HIEROGLYPH_A359A = 0x00014595;
const char32_t ANATOLIAN_HIEROGLYPH_A360 = 0x00014596;
const char32_t ANATOLIAN_HIEROGLYPH_A361 = 0x00014597;
const char32_t ANATOLIAN_HIEROGLYPH_A362 = 0x00014598;
const char32_t ANATOLIAN_HIEROGLYPH_A363 = 0x00014599;
const char32_t ANATOLIAN_HIEROGLYPH_A364 = 0x0001459A;
const char32_t ANATOLIAN_HIEROGLYPH_A364A = 0x0001459B;
const char32_t ANATOLIAN_HIEROGLYPH_A365 = 0x0001459C;
const char32_t ANATOLIAN_HIEROGLYPH_A366 = 0x0001459D;
const char32_t ANATOLIAN_HIEROGLYPH_A367 = 0x0001459E;
const char32_t ANATOLIAN_HIEROGLYPH_A368 = 0x0001459F;
const char32_t ANATOLIAN_HIEROGLYPH_A368A = 0x000145A0;
const char32_t ANATOLIAN_HIEROGLYPH_A369 = 0x000145A1;
const char32_t ANATOLIAN_HIEROGLYPH_A370 = 0x000145A2;
const char32_t ANATOLIAN_HIEROGLYPH_A371 = 0x000145A3;
const char32_t ANATOLIAN_HIEROGLYPH_A371A = 0x000145A4;
const char32_t ANATOLIAN_HIEROGLYPH_A372 = 0x000145A5;
const char32_t ANATOLIAN_HIEROGLYPH_A373 = 0x000145A6;
const char32_t ANATOLIAN_HIEROGLYPH_A374 = 0x000145A7;
const char32_t ANATOLIAN_HIEROGLYPH_A375 = 0x000145A8;
const char32_t ANATOLIAN_HIEROGLYPH_A376 = 0x000145A9;
const char32_t ANATOLIAN_HIEROGLYPH_A377 = 0x000145AA;
const char32_t ANATOLIAN_HIEROGLYPH_A378 = 0x000145AB;
const char32_t ANATOLIAN_HIEROGLYPH_A379 = 0x000145AC;
const char32_t ANATOLIAN_HIEROGLYPH_A380 = 0x000145AD;
const char32_t ANATOLIAN_HIEROGLYPH_A381 = 0x000145AE;
const char32_t ANATOLIAN_HIEROGLYPH_A381A = 0x000145AF;
const char32_t ANATOLIAN_HIEROGLYPH_A382 = 0x000145B0;
const char32_t ANATOLIAN_HIEROGLYPH_A383_RA_OR_RI = 0x000145B1;
const char32_t ANATOLIAN_HIEROGLYPH_A383A = 0x000145B2;
const char32_t ANATOLIAN_HIEROGLYPH_A384 = 0x000145B3;
const char32_t ANATOLIAN_HIEROGLYPH_A385 = 0x000145B4;
const char32_t ANATOLIAN_HIEROGLYPH_A386 = 0x000145B5;
const char32_t ANATOLIAN_HIEROGLYPH_A386A = 0x000145B6;
const char32_t ANATOLIAN_HIEROGLYPH_A387 = 0x000145B7;
const char32_t ANATOLIAN_HIEROGLYPH_A388 = 0x000145B8;
const char32_t ANATOLIAN_HIEROGLYPH_A389 = 0x000145B9;
const char32_t ANATOLIAN_HIEROGLYPH_A390 = 0x000145BA;
const char32_t ANATOLIAN_HIEROGLYPH_A391 = 0x000145BB;
const char32_t ANATOLIAN_HIEROGLYPH_A392 = 0x000145BC;
const char32_t ANATOLIAN_HIEROGLYPH_A393_EIGHT = 0x000145BD;
const char32_t ANATOLIAN_HIEROGLYPH_A394 = 0x000145BE;
const char32_t ANATOLIAN_HIEROGLYPH_A395 = 0x000145BF;
const char32_t ANATOLIAN_HIEROGLYPH_A396 = 0x000145C0;
const char32_t ANATOLIAN_HIEROGLYPH_A397 = 0x000145C1;
const char32_t ANATOLIAN_HIEROGLYPH_A398 = 0x000145C2;
const char32_t ANATOLIAN_HIEROGLYPH_A399 = 0x000145C3;
const char32_t ANATOLIAN_HIEROGLYPH_A400 = 0x000145C4;
const char32_t ANATOLIAN_HIEROGLYPH_A401 = 0x000145C5;
const char32_t ANATOLIAN_HIEROGLYPH_A402 = 0x000145C6;
const char32_t ANATOLIAN_HIEROGLYPH_A403 = 0x000145C7;
const char32_t ANATOLIAN_HIEROGLYPH_A404 = 0x000145C8;
const char32_t ANATOLIAN_HIEROGLYPH_A405 = 0x000145C9;
const char32_t ANATOLIAN_HIEROGLYPH_A406 = 0x000145CA;
const char32_t ANATOLIAN_HIEROGLYPH_A407 = 0x000145CB;
const char32_t ANATOLIAN_HIEROGLYPH_A408 = 0x000145CC;
const char32_t ANATOLIAN_HIEROGLYPH_A409 = 0x000145CD;
const char32_t ANATOLIAN_HIEROGLYPH_A410_BEGIN_LOGOGRAM_MARK = 0x000145CE;
const char32_t ANATOLIAN_HIEROGLYPH_A410A_END_LOGOGRAM_MARK = 0x000145CF;
const char32_t ANATOLIAN_HIEROGLYPH_A411 = 0x000145D0;
const char32_t ANATOLIAN_HIEROGLYPH_A412 = 0x000145D1;
const char32_t ANATOLIAN_HIEROGLYPH_A413 = 0x000145D2;
const char32_t ANATOLIAN_HIEROGLYPH_A414 = 0x000145D3;
const char32_t ANATOLIAN_HIEROGLYPH_A415 = 0x000145D4;
const char32_t ANATOLIAN_HIEROGLYPH_A416 = 0x000145D5;
const char32_t ANATOLIAN_HIEROGLYPH_A417 = 0x000145D6;
const char32_t ANATOLIAN_HIEROGLYPH_A418 = 0x000145D7;
const char32_t ANATOLIAN_HIEROGLYPH_A419 = 0x000145D8;
const char32_t ANATOLIAN_HIEROGLYPH_A420 = 0x000145D9;
const char32_t ANATOLIAN_HIEROGLYPH_A421 = 0x000145DA;
const char32_t ANATOLIAN_HIEROGLYPH_A422 = 0x000145DB;
const char32_t ANATOLIAN_HIEROGLYPH_A423 = 0x000145DC;
const char32_t ANATOLIAN_HIEROGLYPH_A424 = 0x000145DD;
const char32_t ANATOLIAN_HIEROGLYPH_A425 = 0x000145DE;
const char32_t ANATOLIAN_HIEROGLYPH_A426 = 0x000145DF;
const char32_t ANATOLIAN_HIEROGLYPH_A427 = 0x000145E0;
const char32_t ANATOLIAN_HIEROGLYPH_A428 = 0x000145E1;
const char32_t ANATOLIAN_HIEROGLYPH_A429 = 0x000145E2;
const char32_t ANATOLIAN_HIEROGLYPH_A430 = 0x000145E3;
const char32_t ANATOLIAN_HIEROGLYPH_A431 = 0x000145E4;
const char32_t ANATOLIAN_HIEROGLYPH_A432 = 0x000145E5;
const char32_t ANATOLIAN_HIEROGLYPH_A433 = 0x000145E6;
const char32_t ANATOLIAN_HIEROGLYPH_A434 = 0x000145E7;
const char32_t ANATOLIAN_HIEROGLYPH_A435 = 0x000145E8;
const char32_t ANATOLIAN_HIEROGLYPH_A436 = 0x000145E9;
const char32_t ANATOLIAN_HIEROGLYPH_A437 = 0x000145EA;
const char32_t ANATOLIAN_HIEROGLYPH_A438 = 0x000145EB;
const char32_t ANATOLIAN_HIEROGLYPH_A439 = 0x000145EC;
const char32_t ANATOLIAN_HIEROGLYPH_A440 = 0x000145ED;
const char32_t ANATOLIAN_HIEROGLYPH_A441 = 0x000145EE;
const char32_t ANATOLIAN_HIEROGLYPH_A442 = 0x000145EF;
const char32_t ANATOLIAN_HIEROGLYPH_A443 = 0x000145F0;
const char32_t ANATOLIAN_HIEROGLYPH_A444 = 0x000145F1;
const char32_t ANATOLIAN_HIEROGLYPH_A445 = 0x000145F2;
const char32_t ANATOLIAN_HIEROGLYPH_A446 = 0x000145F3;
const char32_t ANATOLIAN_HIEROGLYPH_A447 = 0x000145F4;
const char32_t ANATOLIAN_HIEROGLYPH_A448 = 0x000145F5;
const char32_t ANATOLIAN_HIEROGLYPH_A449 = 0x000145F6;
const char32_t ANATOLIAN_HIEROGLYPH_A450 = 0x000145F7;
const char32_t ANATOLIAN_HIEROGLYPH_A450A = 0x000145F8;
const char32_t ANATOLIAN_HIEROGLYPH_A451 = 0x000145F9;
const char32_t ANATOLIAN_HIEROGLYPH_A452 = 0x000145FA;
const char32_t ANATOLIAN_HIEROGLYPH_A453 = 0x000145FB;
const char32_t ANATOLIAN_HIEROGLYPH_A454 = 0x000145FC;
const char32_t ANATOLIAN_HIEROGLYPH_A455 = 0x000145FD;
const char32_t ANATOLIAN_HIEROGLYPH_A456 = 0x000145FE;
const char32_t ANATOLIAN_HIEROGLYPH_A457 = 0x000145FF;
const char32_t ANATOLIAN_HIEROGLYPH_A457A = 0x00014600;
const char32_t ANATOLIAN_HIEROGLYPH_A458 = 0x00014601;
const char32_t ANATOLIAN_HIEROGLYPH_A459 = 0x00014602;
const char32_t ANATOLIAN_HIEROGLYPH_A460 = 0x00014603;
const char32_t ANATOLIAN_HIEROGLYPH_A461 = 0x00014604;
const char32_t ANATOLIAN_HIEROGLYPH_A462 = 0x00014605;
const char32_t ANATOLIAN_HIEROGLYPH_A463 = 0x00014606;
const char32_t ANATOLIAN_HIEROGLYPH_A464 = 0x00014607;
const char32_t ANATOLIAN_HIEROGLYPH_A465 = 0x00014608;
const char32_t ANATOLIAN_HIEROGLYPH_A466 = 0x00014609;
const char32_t ANATOLIAN_HIEROGLYPH_A467 = 0x0001460A;
const char32_t ANATOLIAN_HIEROGLYPH_A468 = 0x0001460B;
const char32_t ANATOLIAN_HIEROGLYPH_A469 = 0x0001460C;
const char32_t ANATOLIAN_HIEROGLYPH_A470 = 0x0001460D;
const char32_t ANATOLIAN_HIEROGLYPH_A471 = 0x0001460E;
const char32_t ANATOLIAN_HIEROGLYPH_A472 = 0x0001460F;
const char32_t ANATOLIAN_HIEROGLYPH_A473 = 0x00014610;
const char32_t ANATOLIAN_HIEROGLYPH_A474 = 0x00014611;
const char32_t ANATOLIAN_HIEROGLYPH_A475 = 0x00014612;
const char32_t ANATOLIAN_HIEROGLYPH_A476 = 0x00014613;
const char32_t ANATOLIAN_HIEROGLYPH_A477 = 0x00014614;
const char32_t ANATOLIAN_HIEROGLYPH_A478 = 0x00014615;
const char32_t ANATOLIAN_HIEROGLYPH_A479 = 0x00014616;
const char32_t ANATOLIAN_HIEROGLYPH_A480 = 0x00014617;
const char32_t ANATOLIAN_HIEROGLYPH_A481 = 0x00014618;
const char32_t ANATOLIAN_HIEROGLYPH_A482 = 0x00014619;
const char32_t ANATOLIAN_HIEROGLYPH_A483 = 0x0001461A;
const char32_t ANATOLIAN_HIEROGLYPH_A484 = 0x0001461B;
const char32_t ANATOLIAN_HIEROGLYPH_A485 = 0x0001461C;
const char32_t ANATOLIAN_HIEROGLYPH_A486 = 0x0001461D;
const char32_t ANATOLIAN_HIEROGLYPH_A487 = 0x0001461E;
const char32_t ANATOLIAN_HIEROGLYPH_A488 = 0x0001461F;
const char32_t ANATOLIAN_HIEROGLYPH_A489 = 0x00014620;
const char32_t ANATOLIAN_HIEROGLYPH_A490 = 0x00014621;
const char32_t ANATOLIAN_HIEROGLYPH_A491 = 0x00014622;
const char32_t ANATOLIAN_HIEROGLYPH_A492 = 0x00014623;
const char32_t ANATOLIAN_HIEROGLYPH_A493 = 0x00014624;
const char32_t ANATOLIAN_HIEROGLYPH_A494 = 0x00014625;
const char32_t ANATOLIAN_HIEROGLYPH_A495 = 0x00014626;
const char32_t ANATOLIAN_HIEROGLYPH_A496 = 0x00014627;
const char32_t ANATOLIAN_HIEROGLYPH_A497 = 0x00014628;
const char32_t ANATOLIAN_HIEROGLYPH_A501 = 0x00014629;
const char32_t ANATOLIAN_HIEROGLYPH_A502 = 0x0001462A;
const char32_t ANATOLIAN_HIEROGLYPH_A503 = 0x0001462B;
const char32_t ANATOLIAN_HIEROGLYPH_A504 = 0x0001462C;
const char32_t ANATOLIAN_HIEROGLYPH_A505 = 0x0001462D;
const char32_t ANATOLIAN_HIEROGLYPH_A506 = 0x0001462E;
const char32_t ANATOLIAN_HIEROGLYPH_A507 = 0x0001462F;
const char32_t ANATOLIAN_HIEROGLYPH_A508 = 0x00014630;
const char32_t ANATOLIAN_HIEROGLYPH_A509 = 0x00014631;
const char32_t ANATOLIAN_HIEROGLYPH_A510 = 0x00014632;
const char32_t ANATOLIAN_HIEROGLYPH_A511 = 0x00014633;
const char32_t ANATOLIAN_HIEROGLYPH_A512 = 0x00014634;
const char32_t ANATOLIAN_HIEROGLYPH_A513 = 0x00014635;
const char32_t ANATOLIAN_HIEROGLYPH_A514 = 0x00014636;
const char32_t ANATOLIAN_HIEROGLYPH_A515 = 0x00014637;
const char32_t ANATOLIAN_HIEROGLYPH_A516 = 0x00014638;
const char32_t ANATOLIAN_HIEROGLYPH_A517 = 0x00014639;
const char32_t ANATOLIAN_HIEROGLYPH_A518 = 0x0001463A;
const char32_t ANATOLIAN_HIEROGLYPH_A519 = 0x0001463B;
const char32_t ANATOLIAN_HIEROGLYPH_A520 = 0x0001463C;
const char32_t ANATOLIAN_HIEROGLYPH_A521 = 0x0001463D;
const char32_t ANATOLIAN_HIEROGLYPH_A522 = 0x0001463E;
const char32_t ANATOLIAN_HIEROGLYPH_A523 = 0x0001463F;
const char32_t ANATOLIAN_HIEROGLYPH_A524 = 0x00014640;
const char32_t ANATOLIAN_HIEROGLYPH_A525 = 0x00014641;
const char32_t ANATOLIAN_HIEROGLYPH_A526 = 0x00014642;
const char32_t ANATOLIAN_HIEROGLYPH_A527 = 0x00014643;
const char32_t ANATOLIAN_HIEROGLYPH_A528 = 0x00014644;
const char32_t ANATOLIAN_HIEROGLYPH_A529 = 0x00014645;
const char32_t ANATOLIAN_HIEROGLYPH_A530 = 0x00014646;
const char32_t GURUNG_KHEMA_LETTER_A = 0x00016100;
const char32_t GURUNG_KHEMA_LETTER_KA = 0x00016101;
const char32_t GURUNG_KHEMA_LETTER_KHA = 0x00016102;
const char32_t GURUNG_KHEMA_LETTER_GA = 0x00016103;
const char32_t GURUNG_KHEMA_LETTER_GHA = 0x00016104;
const char32_t GURUNG_KHEMA_LETTER_NGA = 0x00016105;
const char32_t GURUNG_KHEMA_LETTER_CA = 0x00016106;
const char32_t GURUNG_KHEMA_LETTER_CHA = 0x00016107;
const char32_t GURUNG_KHEMA_LETTER_JA = 0x00016108;
const char32_t GURUNG_KHEMA_LETTER_JHA = 0x00016109;
const char32_t GURUNG_KHEMA_LETTER_HA = 0x0001610A;
const char32_t GURUNG_KHEMA_LETTER_TTA = 0x0001610B;
const char32_t GURUNG_KHEMA_LETTER_TTHA = 0x0001610C;
const char32_t GURUNG_KHEMA_LETTER_DDA = 0x0001610D;
const char32_t GURUNG_KHEMA_LETTER_DDHA = 0x0001610E;
const char32_t GURUNG_KHEMA_LETTER_VA = 0x0001610F;
const char32_t GURUNG_KHEMA_LETTER_TA = 0x00016110;
const char32_t GURUNG_KHEMA_LETTER_THA = 0x00016111;
const char32_t GURUNG_KHEMA_LETTER_DA = 0x00016112;
const char32_t GURUNG_KHEMA_LETTER_DHA = 0x00016113;
const char32_t GURUNG_KHEMA_LETTER_NA = 0x00016114;
const char32_t GURUNG_KHEMA_LETTER_PA = 0x00016115;
const char32_t GURUNG_KHEMA_LETTER_PHA = 0x00016116;
const char32_t GURUNG_KHEMA_LETTER_BA = 0x00016117;
const char32_t GURUNG_KHEMA_LETTER_BHA = 0x00016118;
const char32_t GURUNG_KHEMA_LETTER_MA = 0x00016119;
const char32_t GURUNG_KHEMA_LETTER_YA = 0x0001611A;
const char32_t GURUNG_KHEMA_LETTER_RA = 0x0001611B;
const char32_t GURUNG_KHEMA_LETTER_LA = 0x0001611C;
const char32_t GURUNG_KHEMA_LETTER_SA = 0x0001611D;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_AA = 0x0001611E;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_I = 0x0001611F;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_II = 0x00016120;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_U = 0x00016121;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_UU = 0x00016122;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_E = 0x00016123;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_EE = 0x00016124;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_AI = 0x00016125;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_O = 0x00016126;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_OO = 0x00016127;
const char32_t GURUNG_KHEMA_VOWEL_SIGN_AU = 0x00016128;
const char32_t GURUNG_KHEMA_VOWEL_LENGTH_MARK = 0x00016129;
const char32_t GURUNG_KHEMA_CONSONANT_SIGN_MEDIAL_YA = 0x0001612A;
const char32_t GURUNG_KHEMA_CONSONANT_SIGN_MEDIAL_VA = 0x0001612B;
const char32_t GURUNG_KHEMA_CONSONANT_SIGN_MEDIAL_HA = 0x0001612C;
const char32_t GURUNG_KHEMA_SIGN_ANUSVARA = 0x0001612D;
const char32_t GURUNG_KHEMA_CONSONANT_SIGN_MEDIAL_RA = 0x0001612E;
const char32_t GURUNG_KHEMA_SIGN_THOLHOMA = 0x0001612F;
const char32_t GURUNG_KHEMA_DIGIT_ZERO = 0x00016130;
const char32_t GURUNG_KHEMA_DIGIT_ONE = 0x00016131;
const char32_t GURUNG_KHEMA_DIGIT_TWO = 0x00016132;
const char32_t GURUNG_KHEMA_DIGIT_THREE = 0x00016133;
const char32_t GURUNG_KHEMA_DIGIT_FOUR = 0x00016134;
const char32_t GURUNG_KHEMA_DIGIT_FIVE = 0x00016135;
const char32_t GURUNG_KHEMA_DIGIT_SIX = 0x00016136;
const char32_t GURUNG_KHEMA_DIGIT_SEVEN = 0x00016137;
const char32_t GURUNG_KHEMA_DIGIT_EIGHT = 0x00016138;
const char32_t GURUNG_KHEMA_DIGIT_NINE = 0x00016139;
const char32_t BAMUM_LETTER_PHASE_A_NGKUE_MFON = 0x00016800;
const char32_t BAMUM_LETTER_PHASE_A_GBIEE_FON = 0x00016801;
const char32_t BAMUM_LETTER_PHASE_A_PON_MFON_PIPAEMGBIEE = 0x00016802;
const char32_t BAMUM_LETTER_PHASE_A_PON_MFON_PIPAEMBA = 0x00016803;
const char32_t BAMUM_LETTER_PHASE_A_NAA_MFON = 0x00016804;
const char32_t BAMUM_LETTER_PHASE_A_SHUENSHUET = 0x00016805;
const char32_t BAMUM_LETTER_PHASE_A_TITA_MFON = 0x00016806;
const char32_t BAMUM_LETTER_PHASE_A_NZA_MFON = 0x00016807;
const char32_t BAMUM_LETTER_PHASE_A_SHINDA_PA_NJI = 0x00016808;
const char32_t BAMUM_LETTER_PHASE_A_PON_PA_NJI_PIPAEMGBIEE = 0x00016809;
const char32_t BAMUM_LETTER_PHASE_A_PON_PA_NJI_PIPAEMBA = 0x0001680A;
const char32_t BAMUM_LETTER_PHASE_A_MAEMBGBIEE = 0x0001680B;
const char32_t BAMUM_LETTER_PHASE_A_TU_MAEMBA = 0x0001680C;
const char32_t BAMUM_LETTER_PHASE_A_NGANGU = 0x0001680D;
const char32_t BAMUM_LETTER_PHASE_A_MAEMVEUX = 0x0001680E;
const char32_t BAMUM_LETTER_PHASE_A_MANSUAE = 0x0001680F;
const char32_t BAMUM_LETTER_PHASE_A_MVEUAENGAM = 0x00016810;
const char32_t BAMUM_LETTER_PHASE_A_SEUNYAM = 0x00016811;
const char32_t BAMUM_LETTER_PHASE_A_NTOQPEN = 0x00016812;
const char32_t BAMUM_LETTER_PHASE_A_KEUKEUTNDA = 0x00016813;
const char32_t BAMUM_LETTER_PHASE_A_NKINDI = 0x00016814;
const char32_t BAMUM_LETTER_PHASE_A_SUU = 0x00016815;
const char32_t BAMUM_LETTER_PHASE_A_NGKUENZEUM = 0x00016816;
const char32_t BAMUM_LETTER_PHASE_A_LAPAQ = 0x00016817;
const char32_t BAMUM_LETTER_PHASE_A_LET_KUT = 0x00016818;
const char32_t BAMUM_LETTER_PHASE_A_NTAP_MFAA = 0x00016819;
const char32_t BAMUM_LETTER_PHASE_A_MAEKEUP = 0x0001681A;
const char32_t BAMUM_LETTER_PHASE_A_PASHAE = 0x0001681B;
const char32_t BAMUM_LETTER_PHASE_A_GHEUAERAE = 0x0001681C;
const char32_t BAMUM_LETTER_PHASE_A_PAMSHAE = 0x0001681D;
const char32_t BAMUM_LETTER_PHASE_A_MON_NGGEUAET = 0x0001681E;
const char32_t BAMUM_LETTER_PHASE_A_NZUN_MEUT = 0x0001681F;
const char32_t BAMUM_LETTER_PHASE_A_U_YUQ_NAE = 0x00016820;
const char32_t BAMUM_LETTER_PHASE_A_GHEUAEGHEUAE = 0x00016821;
const char32_t BAMUM_LETTER_PHASE_A_NTAP_NTAA = 0x00016822;
const char32_t BAMUM_LETTER_PHASE_A_SISA = 0x00016823;
const char32_t BAMUM_LETTER_PHASE_A_MGBASA = 0x00016824;
const char32_t BAMUM_LETTER_PHASE_A_MEUNJOMNDEUQ = 0x00016825;
const char32_t BAMUM_LETTER_PHASE_A_MOOMPUQ = 0x00016826;
const char32_t BAMUM_LETTER_PHASE_A_KAFA = 0x00016827;
const char32_t BAMUM_LETTER_PHASE_A_PA_LEERAEWA = 0x00016828;
const char32_t BAMUM_LETTER_PHASE_A_NDA_LEERAEWA = 0x00016829;
const char32_t BAMUM_LETTER_PHASE_A_PET = 0x0001682A;
const char32_t BAMUM_LETTER_PHASE_A_MAEMKPEN = 0x0001682B;
const char32_t BAMUM_LETTER_PHASE_A_NIKA = 0x0001682C;
const char32_t BAMUM_LETTER_PHASE_A_PUP = 0x0001682D;
const char32_t BAMUM_LETTER_PHASE_A_TUAEP = 0x0001682E;
const char32_t BAMUM_LETTER_PHASE_A_LUAEP = 0x0001682F;
const char32_t BAMUM_LETTER_PHASE_A_SONJAM = 0x00016830;
const char32_t BAMUM_LETTER_PHASE_A_TEUTEUWEN = 0x00016831;
const char32_t BAMUM_LETTER_PHASE_A_MAENYI = 0x00016832;
const char32_t BAMUM_LETTER_PHASE_A_KET = 0x00016833;
const char32_t BAMUM_LETTER_PHASE_A_NDAANGGEUAET = 0x00016834;
const char32_t BAMUM_LETTER_PHASE_A_KUOQ = 0x00016835;
const char32_t BAMUM_LETTER_PHASE_A_MOOMEUT = 0x00016836;
const char32_t BAMUM_LETTER_PHASE_A_SHUM = 0x00016837;
const char32_t BAMUM_LETTER_PHASE_A_LOMMAE = 0x00016838;
const char32_t BAMUM_LETTER_PHASE_A_FIRI = 0x00016839;
const char32_t BAMUM_LETTER_PHASE_A_ROM = 0x0001683A;
const char32_t BAMUM_LETTER_PHASE_A_KPOQ = 0x0001683B;
const char32_t BAMUM_LETTER_PHASE_A_SOQ = 0x0001683C;
const char32_t BAMUM_LETTER_PHASE_A_MAP_PIEET = 0x0001683D;
const char32_t BAMUM_LETTER_PHASE_A_SHIRAE = 0x0001683E;
const char32_t BAMUM_LETTER_PHASE_A_NTAP = 0x0001683F;
const char32_t BAMUM_LETTER_PHASE_A_SHOQ_NSHUT_YUM = 0x00016840;
const char32_t BAMUM_LETTER_PHASE_A_NYIT_MONGKEUAEQ = 0x00016841;
const char32_t BAMUM_LETTER_PHASE_A_PAARAE = 0x00016842;
const char32_t BAMUM_LETTER_PHASE_A_NKAARAE = 0x00016843;
const char32_t BAMUM_LETTER_PHASE_A_UNKNOWN = 0x00016844;
const char32_t BAMUM_LETTER_PHASE_A_NGGEN = 0x00016845;
const char32_t BAMUM_LETTER_PHASE_A_MAESI = 0x00016846;
const char32_t BAMUM_LETTER_PHASE_A_NJAM = 0x00016847;
const char32_t BAMUM_LETTER_PHASE_A_MBANYI = 0x00016848;
const char32_t BAMUM_LETTER_PHASE_A_NYET = 0x00016849;
const char32_t BAMUM_LETTER_PHASE_A_TEUAEN = 0x0001684A;
const char32_t BAMUM_LETTER_PHASE_A_SOT = 0x0001684B;
const char32_t BAMUM_LETTER_PHASE_A_PAAM = 0x0001684C;
const char32_t BAMUM_LETTER_PHASE_A_NSHIEE = 0x0001684D;
const char32_t BAMUM_LETTER_PHASE_A_MAEM = 0x0001684E;
const char32_t BAMUM_LETTER_PHASE_A_NYI = 0x0001684F;
const char32_t BAMUM_LETTER_PHASE_A_KAQ = 0x00016850;
const char32_t BAMUM_LETTER_PHASE_A_NSHA = 0x00016851;
const char32_t BAMUM_LETTER_PHASE_A_VEE = 0x00016852;
const char32_t BAMUM_LETTER_PHASE_A_LU = 0x00016853;
const char32_t BAMUM_LETTER_PHASE_A_NEN = 0x00016854;
const char32_t BAMUM_LETTER_PHASE_A_NAQ = 0x00016855;
const char32_t BAMUM_LETTER_PHASE_A_MBAQ = 0x00016856;
const char32_t BAMUM_LETTER_PHASE_B_NSHUET = 0x00016857;
const char32_t BAMUM_LETTER_PHASE_B_TU_MAEMGBIEE = 0x00016858;
const char32_t BAMUM_LETTER_PHASE_B_SIEE = 0x00016859;
const char32_t BAMUM_LETTER_PHASE_B_SET_TU = 0x0001685A;
const char32_t BAMUM_LETTER_PHASE_B_LOM_NTEUM = 0x0001685B;
const char32_t BAMUM_LETTER_PHASE_B_MBA_MAELEE = 0x0001685C;
const char32_t BAMUM_LETTER_PHASE_B_KIEEM = 0x0001685D;
const char32_t BAMUM_LETTER_PHASE_B_YEURAE = 0x0001685E;
const char32_t BAMUM_LETTER_PHASE_B_MBAARAE = 0x0001685F;
const char32_t BAMUM_LETTER_PHASE_B_KAM = 0x00016860;
const char32_t BAMUM_LETTER_PHASE_B_PEESHI = 0x00016861;
const char32_t BAMUM_LETTER_PHASE_B_YAFU_LEERAEWA = 0x00016862;
const char32_t BAMUM_LETTER_PHASE_B_LAM_NSHUT_NYAM = 0x00016863;
const char32_t BAMUM_LETTER_PHASE_B_NTIEE_SHEUOQ = 0x00016864;
const char32_t BAMUM_LETTER_PHASE_B_NDU_NJAA = 0x00016865;
const char32_t BAMUM_LETTER_PHASE_B_GHEUGHEUAEM = 0x00016866;
const char32_t BAMUM_LETTER_PHASE_B_PIT = 0x00016867;
const char32_t BAMUM_LETTER_PHASE_B_TU_NSIEE = 0x00016868;
const char32_t BAMUM_LETTER_PHASE_B_SHET_NJAQ = 0x00016869;
const char32_t BAMUM_LETTER_PHASE_B_SHEUAEQTU = 0x0001686A;
const char32_t BAMUM_LETTER_PHASE_B_MFON_TEUAEQ = 0x0001686B;
const char32_t BAMUM_LETTER_PHASE_B_MBIT_MBAAKET = 0x0001686C;
const char32_t BAMUM_LETTER_PHASE_B_NYI_NTEUM = 0x0001686D;
const char32_t BAMUM_LETTER_PHASE_B_KEUPUQ = 0x0001686E;
const char32_t BAMUM_LETTER_PHASE_B_GHEUGHEN = 0x0001686F;
const char32_t BAMUM_LETTER_PHASE_B_KEUYEUX = 0x00016870;
const char32_t BAMUM_LETTER_PHASE_B_LAANAE = 0x00016871;
const char32_t BAMUM_LETTER_PHASE_B_PARUM = 0x00016872;
const char32_t BAMUM_LETTER_PHASE_B_VEUM = 0x00016873;
const char32_t BAMUM_LETTER_PHASE_B_NGKINDI_MVOP = 0x00016874;
const char32_t BAMUM_LETTER_PHASE_B_NGGEU_MBU = 0x00016875;
const char32_t BAMUM_LETTER_PHASE_B_WUAET = 0x00016876;
const char32_t BAMUM_LETTER_PHASE_B_SAKEUAE = 0x00016877;
const char32_t BAMUM_LETTER_PHASE_B_TAAM = 0x00016878;
const char32_t BAMUM_LETTER_PHASE_B_MEUQ = 0x00016879;
const char32_t BAMUM_LETTER_PHASE_B_NGGUOQ = 0x0001687A;
const char32_t BAMUM_LETTER_PHASE_B_NGGUOQ_LARGE = 0x0001687B;
const char32_t BAMUM_LETTER_PHASE_B_MFIYAQ = 0x0001687C;
const char32_t BAMUM_LETTER_PHASE_B_SUE = 0x0001687D;
const char32_t BAMUM_LETTER_PHASE_B_MBEURI = 0x0001687E;
const char32_t BAMUM_LETTER_PHASE_B_MONTIEEN = 0x0001687F;
const char32_t BAMUM_LETTER_PHASE_B_NYAEMAE = 0x00016880;
const char32_t BAMUM_LETTER_PHASE_B_PUNGAAM = 0x00016881;
const char32_t BAMUM_LETTER_PHASE_B_MEUT_NGGEET = 0x00016882;
const char32_t BAMUM_LETTER_PHASE_B_FEUX = 0x00016883;
const char32_t BAMUM_LETTER_PHASE_B_MBUOQ = 0x00016884;
const char32_t BAMUM_LETTER_PHASE_B_FEE = 0x00016885;
const char32_t BAMUM_LETTER_PHASE_B_KEUAEM = 0x00016886;
const char32_t BAMUM_LETTER_PHASE_B_MA_NJEUAENA = 0x00016887;
const char32_t BAMUM_LETTER_PHASE_B_MA_NJUQA = 0x00016888;
const char32_t BAMUM_LETTER_PHASE_B_LET = 0x00016889;
const char32_t BAMUM_LETTER_PHASE_B_NGGAAM = 0x0001688A;
const char32_t BAMUM_LETTER_PHASE_B_NSEN = 0x0001688B;
const char32_t BAMUM_LETTER_PHASE_B_MA = 0x0001688C;
const char32_t BAMUM_LETTER_PHASE_B_KIQ = 0x0001688D;
const char32_t BAMUM_LETTER_PHASE_B_NGOM = 0x0001688E;
const char32_t BAMUM_LETTER_PHASE_C_NGKUE_MAEMBA = 0x0001688F;
const char32_t BAMUM_LETTER_PHASE_C_NZA = 0x00016890;
const char32_t BAMUM_LETTER_PHASE_C_YUM = 0x00016891;
const char32_t BAMUM_LETTER_PHASE_C_WANGKUOQ = 0x00016892;
const char32_t BAMUM_LETTER_PHASE_C_NGGEN = 0x00016893;
const char32_t BAMUM_LETTER_PHASE_C_NDEUAEREE = 0x00016894;
const char32_t BAMUM_LETTER_PHASE_C_NGKAQ = 0x00016895;
const char32_t BAMUM_LETTER_PHASE_C_GHARAE = 0x00016896;
const char32_t BAMUM_LETTER_PHASE_C_MBEEKEET = 0x00016897;
const char32_t BAMUM_LETTER_PHASE_C_GBAYI = 0x00016898;
const char32_t BAMUM_LETTER_PHASE_C_NYIR_MKPARAQ_MEUN = 0x00016899;
const char32_t BAMUM_LETTER_PHASE_C_NTU_MBIT = 0x0001689A;
const char32_t BAMUM_LETTER_PHASE_C_MBEUM = 0x0001689B;
const char32_t BAMUM_LETTER_PHASE_C_PIRIEEN = 0x0001689C;
const char32_t BAMUM_LETTER_PHASE_C_NDOMBU = 0x0001689D;
const char32_t BAMUM_LETTER_PHASE_C_MBAA_CABBAGE_TREE = 0x0001689E;
const char32_t BAMUM_LETTER_PHASE_C_KEUSHEUAEP = 0x0001689F;
const char32_t BAMUM_LETTER_PHASE_C_GHAP = 0x000168A0;
const char32_t BAMUM_LETTER_PHASE_C_KEUKAQ = 0x000168A1;
const char32_t BAMUM_LETTER_PHASE_C_YU_MUOMAE = 0x000168A2;
const char32_t BAMUM_LETTER_PHASE_C_NZEUM = 0x000168A3;
const char32_t BAMUM_LETTER_PHASE_C_MBUE = 0x000168A4;
const char32_t BAMUM_LETTER_PHASE_C_NSEUAEN = 0x000168A5;
const char32_t BAMUM_LETTER_PHASE_C_MBIT = 0x000168A6;
const char32_t BAMUM_LETTER_PHASE_C_YEUQ = 0x000168A7;
const char32_t BAMUM_LETTER_PHASE_C_KPARAQ = 0x000168A8;
const char32_t BAMUM_LETTER_PHASE_C_KAA = 0x000168A9;
const char32_t BAMUM_LETTER_PHASE_C_SEUX = 0x000168AA;
const char32_t BAMUM_LETTER_PHASE_C_NDIDA = 0x000168AB;
const char32_t BAMUM_LETTER_PHASE_C_TAASHAE = 0x000168AC;
const char32_t BAMUM_LETTER_PHASE_C_NJUEQ = 0x000168AD;
const char32_t BAMUM_LETTER_PHASE_C_TITA_YUE = 0x000168AE;
const char32_t BAMUM_LETTER_PHASE_C_SUAET = 0x000168AF;
const char32_t BAMUM_LETTER_PHASE_C_NGGUAEN_NYAM = 0x000168B0;
const char32_t BAMUM_LETTER_PHASE_C_VEUX = 0x000168B1;
const char32_t BAMUM_LETTER_PHASE_C_NANSANAQ = 0x000168B2;
const char32_t BAMUM_LETTER_PHASE_C_MA_KEUAERI = 0x000168B3;
const char32_t BAMUM_LETTER_PHASE_C_NTAA = 0x000168B4;
const char32_t BAMUM_LETTER_PHASE_C_NGGUON = 0x000168B5;
const char32_t BAMUM_LETTER_PHASE_C_LAP = 0x000168B6;
const char32_t BAMUM_LETTER_PHASE_C_MBIRIEEN = 0x000168B7;
const char32_t BAMUM_LETTER_PHASE_C_MGBASAQ = 0x000168B8;
const char32_t BAMUM_LETTER_PHASE_C_NTEUNGBA = 0x000168B9;
const char32_t BAMUM_LETTER_PHASE_C_TEUTEUX = 0x000168BA;
const char32_t BAMUM_LETTER_PHASE_C_NGGUM = 0x000168BB;
const char32_t BAMUM_LETTER_PHASE_C_FUE = 0x000168BC;
const char32_t BAMUM_LETTER_PHASE_C_NDEUT = 0x000168BD;
const char32_t BAMUM_LETTER_PHASE_C_NSA = 0x000168BE;
const char32_t BAMUM_LETTER_PHASE_C_NSHAQ = 0x000168BF;
const char32_t BAMUM_LETTER_PHASE_C_BUNG = 0x000168C0;
const char32_t BAMUM_LETTER_PHASE_C_VEUAEPEN = 0x000168C1;
const char32_t BAMUM_LETTER_PHASE_C_MBERAE = 0x000168C2;
const char32_t BAMUM_LETTER_PHASE_C_RU = 0x000168C3;
const char32_t BAMUM_LETTER_PHASE_C_NJAEM = 0x000168C4;
const char32_t BAMUM_LETTER_PHASE_C_LAM = 0x000168C5;
const char32_t BAMUM_LETTER_PHASE_C_TITUAEP = 0x000168C6;
const char32_t BAMUM_LETTER_PHASE_C_NSUOT_NGOM = 0x000168C7;
const char32_t BAMUM_LETTER_PHASE_C_NJEEEE = 0x000168C8;
const char32_t BAMUM_LETTER_PHASE_C_KET = 0x000168C9;
const char32_t BAMUM_LETTER_PHASE_C_NGGU = 0x000168CA;
const char32_t BAMUM_LETTER_PHASE_C_MAESI = 0x000168CB;
const char32_t BAMUM_LETTER_PHASE_C_MBUAEM = 0x000168CC;
const char32_t BAMUM_LETTER_PHASE_C_LU = 0x000168CD;
const char32_t BAMUM_LETTER_PHASE_C_KUT = 0x000168CE;
const char32_t BAMUM_LETTER_PHASE_C_NJAM = 0x000168CF;
const char32_t BAMUM_LETTER_PHASE_C_NGOM = 0x000168D0;
const char32_t BAMUM_LETTER_PHASE_C_WUP = 0x000168D1;
const char32_t BAMUM_LETTER_PHASE_C_NGGUEET = 0x000168D2;
const char32_t BAMUM_LETTER_PHASE_C_NSOM = 0x000168D3;
const char32_t BAMUM_LETTER_PHASE_C_NTEN = 0x000168D4;
const char32_t BAMUM_LETTER_PHASE_C_KUOP_NKAARAE = 0x000168D5;
const char32_t BAMUM_LETTER_PHASE_C_NSUN = 0x000168D6;
const char32_t BAMUM_LETTER_PHASE_C_NDAM = 0x000168D7;
const char32_t BAMUM_LETTER_PHASE_C_MA_NSIEE = 0x000168D8;
const char32_t BAMUM_LETTER_PHASE_C_YAA = 0x000168D9;
const char32_t BAMUM_LETTER_PHASE_C_NDAP = 0x000168DA;
const char32_t BAMUM_LETTER_PHASE_C_SHUEQ = 0x000168DB;
const char32_t BAMUM_LETTER_PHASE_C_SETFON = 0x000168DC;
const char32_t BAMUM_LETTER_PHASE_C_MBI = 0x000168DD;
const char32_t BAMUM_LETTER_PHASE_C_MAEMBA = 0x000168DE;
const char32_t BAMUM_LETTER_PHASE_C_MBANYI = 0x000168DF;
const char32_t BAMUM_LETTER_PHASE_C_KEUSEUX = 0x000168E0;
const char32_t BAMUM_LETTER_PHASE_C_MBEUX = 0x000168E1;
const char32_t BAMUM_LETTER_PHASE_C_KEUM = 0x000168E2;
const char32_t BAMUM_LETTER_PHASE_C_MBAA_PICKET = 0x000168E3;
const char32_t BAMUM_LETTER_PHASE_C_YUWOQ = 0x000168E4;
const char32_t BAMUM_LETTER_PHASE_C_NJEUX = 0x000168E5;
const char32_t BAMUM_LETTER_PHASE_C_MIEE = 0x000168E6;
const char32_t BAMUM_LETTER_PHASE_C_MUAE = 0x000168E7;
const char32_t BAMUM_LETTER_PHASE_C_SHIQ = 0x000168E8;
const char32_t BAMUM_LETTER_PHASE_C_KEN_LAW = 0x000168E9;
const char32_t BAMUM_LETTER_PHASE_C_KEN_FATIGUE = 0x000168EA;
const char32_t BAMUM_LETTER_PHASE_C_NGAQ = 0x000168EB;
const char32_t BAMUM_LETTER_PHASE_C_NAQ = 0x000168EC;
const char32_t BAMUM_LETTER_PHASE_C_LIQ = 0x000168ED;
const char32_t BAMUM_LETTER_PHASE_C_PIN = 0x000168EE;
const char32_t BAMUM_LETTER_PHASE_C_PEN = 0x000168EF;
const char32_t BAMUM_LETTER_PHASE_C_TET = 0x000168F0;
const char32_t BAMUM_LETTER_PHASE_D_MBUO = 0x000168F1;
const char32_t BAMUM_LETTER_PHASE_D_WAP = 0x000168F2;
const char32_t BAMUM_LETTER_PHASE_D_NJI = 0x000168F3;
const char32_t BAMUM_LETTER_PHASE_D_MFON = 0x000168F4;
const char32_t BAMUM_LETTER_PHASE_D_NJIEE = 0x000168F5;
const char32_t BAMUM_LETTER_PHASE_D_LIEE = 0x000168F6;
const char32_t BAMUM_LETTER_PHASE_D_NJEUT = 0x000168F7;
const char32_t BAMUM_LETTER_PHASE_D_NSHEE = 0x000168F8;
const char32_t BAMUM_LETTER_PHASE_D_NGGAAMAE = 0x000168F9;
const char32_t BAMUM_LETTER_PHASE_D_NYAM = 0x000168FA;
const char32_t BAMUM_LETTER_PHASE_D_WUAEN = 0x000168FB;
const char32_t BAMUM_LETTER_PHASE_D_NGKUN = 0x000168FC;
const char32_t BAMUM_LETTER_PHASE_D_SHEE = 0x000168FD;
const char32_t BAMUM_LETTER_PHASE_D_NGKAP = 0x000168FE;
const char32_t BAMUM_LETTER_PHASE_D_KEUAETMEUN = 0x000168FF;
const char32_t BAMUM_LETTER_PHASE_D_TEUT = 0x00016900;
const char32_t BAMUM_LETTER_PHASE_D_SHEUAE = 0x00016901;
const char32_t BAMUM_LETTER_PHASE_D_NJAP = 0x00016902;
const char32_t BAMUM_LETTER_PHASE_D_SUE = 0x00016903;
const char32_t BAMUM_LETTER_PHASE_D_KET = 0x00016904;
const char32_t BAMUM_LETTER_PHASE_D_YAEMMAE = 0x00016905;
const char32_t BAMUM_LETTER_PHASE_D_KUOM = 0x00016906;
const char32_t BAMUM_LETTER_PHASE_D_SAP = 0x00016907;
const char32_t BAMUM_LETTER_PHASE_D_MFEUT = 0x00016908;
const char32_t BAMUM_LETTER_PHASE_D_NDEUX = 0x00016909;
const char32_t BAMUM_LETTER_PHASE_D_MALEERI = 0x0001690A;
const char32_t BAMUM_LETTER_PHASE_D_MEUT = 0x0001690B;
const char32_t BAMUM_LETTER_PHASE_D_SEUAEQ = 0x0001690C;
const char32_t BAMUM_LETTER_PHASE_D_YEN = 0x0001690D;
const char32_t BAMUM_LETTER_PHASE_D_NJEUAEM = 0x0001690E;
const char32_t BAMUM_LETTER_PHASE_D_KEUOT_MBUAE = 0x0001690F;
const char32_t BAMUM_LETTER_PHASE_D_NGKEURI = 0x00016910;
const char32_t BAMUM_LETTER_PHASE_D_TU = 0x00016911;
const char32_t BAMUM_LETTER_PHASE_D_GHAA = 0x00016912;
const char32_t BAMUM_LETTER_PHASE_D_NGKYEE = 0x00016913;
const char32_t BAMUM_LETTER_PHASE_D_FEUFEUAET = 0x00016914;
const char32_t BAMUM_LETTER_PHASE_D_NDEE = 0x00016915;
const char32_t BAMUM_LETTER_PHASE_D_MGBOFUM = 0x00016916;
const char32_t BAMUM_LETTER_PHASE_D_LEUAEP = 0x00016917;
const char32_t BAMUM_LETTER_PHASE_D_NDON = 0x00016918;
const char32_t BAMUM_LETTER_PHASE_D_MONI = 0x00016919;
const char32_t BAMUM_LETTER_PHASE_D_MGBEUN = 0x0001691A;
const char32_t BAMUM_LETTER_PHASE_D_PUUT = 0x0001691B;
const char32_t BAMUM_LETTER_PHASE_D_MGBIEE = 0x0001691C;
const char32_t BAMUM_LETTER_PHASE_D_MFO = 0x0001691D;
const char32_t BAMUM_LETTER_PHASE_D_LUM = 0x0001691E;
const char32_t BAMUM_LETTER_PHASE_D_NSIEEP = 0x0001691F;
const char32_t BAMUM_LETTER_PHASE_D_MBAA = 0x00016920;
const char32_t BAMUM_LETTER_PHASE_D_KWAET = 0x00016921;
const char32_t BAMUM_LETTER_PHASE_D_NYET = 0x00016922;
const char32_t BAMUM_LETTER_PHASE_D_TEUAEN = 0x00016923;
const char32_t BAMUM_LETTER_PHASE_D_SOT = 0x00016924;
const char32_t BAMUM_LETTER_PHASE_D_YUWOQ = 0x00016925;
const char32_t BAMUM_LETTER_PHASE_D_KEUM = 0x00016926;
const char32_t BAMUM_LETTER_PHASE_D_RAEM = 0x00016927;
const char32_t BAMUM_LETTER_PHASE_D_TEEEE = 0x00016928;
const char32_t BAMUM_LETTER_PHASE_D_NGKEUAEQ = 0x00016929;
const char32_t BAMUM_LETTER_PHASE_D_MFEUAE = 0x0001692A;
const char32_t BAMUM_LETTER_PHASE_D_NSIEET = 0x0001692B;
const char32_t BAMUM_LETTER_PHASE_D_KEUP = 0x0001692C;
const char32_t BAMUM_LETTER_PHASE_D_PIP = 0x0001692D;
const char32_t BAMUM_LETTER_PHASE_D_PEUTAE = 0x0001692E;
const char32_t BAMUM_LETTER_PHASE_D_NYUE = 0x0001692F;
const char32_t BAMUM_LETTER_PHASE_D_LET = 0x00016930;
const char32_t BAMUM_LETTER_PHASE_D_NGGAAM = 0x00016931;
const char32_t BAMUM_LETTER_PHASE_D_MFIEE = 0x00016932;
const char32_t BAMUM_LETTER_PHASE_D_NGGWAEN = 0x00016933;
const char32_t BAMUM_LETTER_PHASE_D_YUOM = 0x00016934;
const char32_t BAMUM_LETTER_PHASE_D_PAP = 0x00016935;
const char32_t BAMUM_LETTER_PHASE_D_YUOP = 0x00016936;
const char32_t BAMUM_LETTER_PHASE_D_NDAM = 0x00016937;
const char32_t BAMUM_LETTER_PHASE_D_NTEUM = 0x00016938;
const char32_t BAMUM_LETTER_PHASE_D_SUAE = 0x00016939;
const char32_t BAMUM_LETTER_PHASE_D_KUN = 0x0001693A;
const char32_t BAMUM_LETTER_PHASE_D_NGGEUX = 0x0001693B;
const char32_t BAMUM_LETTER_PHASE_D_NGKIEE = 0x0001693C;
const char32_t BAMUM_LETTER_PHASE_D_TUOT = 0x0001693D;
const char32_t BAMUM_LETTER_PHASE_D_MEUN = 0x0001693E;
const char32_t BAMUM_LETTER_PHASE_D_KUQ = 0x0001693F;
const char32_t BAMUM_LETTER_PHASE_D_NSUM = 0x00016940;
const char32_t BAMUM_LETTER_PHASE_D_TEUN = 0x00016941;
const char32_t BAMUM_LETTER_PHASE_D_MAENJET = 0x00016942;
const char32_t BAMUM_LETTER_PHASE_D_NGGAP = 0x00016943;
const char32_t BAMUM_LETTER_PHASE_D_LEUM = 0x00016944;
const char32_t BAMUM_LETTER_PHASE_D_NGGUOM = 0x00016945;
const char32_t BAMUM_LETTER_PHASE_D_NSHUT = 0x00016946;
const char32_t BAMUM_LETTER_PHASE_D_NJUEQ = 0x00016947;
const char32_t BAMUM_LETTER_PHASE_D_GHEUAE = 0x00016948;
const char32_t BAMUM_LETTER_PHASE_D_KU = 0x00016949;
const char32_t BAMUM_LETTER_PHASE_D_REN_OLD = 0x0001694A;
const char32_t BAMUM_LETTER_PHASE_D_TAE = 0x0001694B;
const char32_t BAMUM_LETTER_PHASE_D_TOQ = 0x0001694C;
const char32_t BAMUM_LETTER_PHASE_D_NYI = 0x0001694D;
const char32_t BAMUM_LETTER_PHASE_D_RII = 0x0001694E;
const char32_t BAMUM_LETTER_PHASE_D_LEEEE = 0x0001694F;
const char32_t BAMUM_LETTER_PHASE_D_MEEEE = 0x00016950;
const char32_t BAMUM_LETTER_PHASE_D_M = 0x00016951;
const char32_t BAMUM_LETTER_PHASE_D_SUU = 0x00016952;
const char32_t BAMUM_LETTER_PHASE_D_MU = 0x00016953;
const char32_t BAMUM_LETTER_PHASE_D_SHII = 0x00016954;
const char32_t BAMUM_LETTER_PHASE_D_SHEUX = 0x00016955;
const char32_t BAMUM_LETTER_PHASE_D_KYEE = 0x00016956;
const char32_t BAMUM_LETTER_PHASE_D_NU = 0x00016957;
const char32_t BAMUM_LETTER_PHASE_D_SHU = 0x00016958;
const char32_t BAMUM_LETTER_PHASE_D_NTEE = 0x00016959;
const char32_t BAMUM_LETTER_PHASE_D_PEE = 0x0001695A;
const char32_t BAMUM_LETTER_PHASE_D_NI = 0x0001695B;
const char32_t BAMUM_LETTER_PHASE_D_SHOQ = 0x0001695C;
const char32_t BAMUM_LETTER_PHASE_D_PUQ = 0x0001695D;
const char32_t BAMUM_LETTER_PHASE_D_MVOP = 0x0001695E;
const char32_t BAMUM_LETTER_PHASE_D_LOQ = 0x0001695F;
const char32_t BAMUM_LETTER_PHASE_D_REN_MUCH = 0x00016960;
const char32_t BAMUM_LETTER_PHASE_D_TI = 0x00016961;
const char32_t BAMUM_LETTER_PHASE_D_NTUU = 0x00016962;
const char32_t BAMUM_LETTER_PHASE_D_MBAA_SEVEN = 0x00016963;
const char32_t BAMUM_LETTER_PHASE_D_SAQ = 0x00016964;
const char32_t BAMUM_LETTER_PHASE_D_FAA = 0x00016965;
const char32_t BAMUM_LETTER_PHASE_E_NDAP = 0x00016966;
const char32_t BAMUM_LETTER_PHASE_E_TOON = 0x00016967;
const char32_t BAMUM_LETTER_PHASE_E_MBEUM = 0x00016968;
const char32_t BAMUM_LETTER_PHASE_E_LAP = 0x00016969;
const char32_t BAMUM_LETTER_PHASE_E_VOM = 0x0001696A;
const char32_t BAMUM_LETTER_PHASE_E_LOON = 0x0001696B;
const char32_t BAMUM_LETTER_PHASE_E_PAA = 0x0001696C;
const char32_t BAMUM_LETTER_PHASE_E_SOM = 0x0001696D;
const char32_t BAMUM_LETTER_PHASE_E_RAQ = 0x0001696E;
const char32_t BAMUM_LETTER_PHASE_E_NSHUOP = 0x0001696F;
const char32_t BAMUM_LETTER_PHASE_E_NDUN = 0x00016970;
const char32_t BAMUM_LETTER_PHASE_E_PUAE = 0x00016971;
const char32_t BAMUM_LETTER_PHASE_E_TAM = 0x00016972;
const char32_t BAMUM_LETTER_PHASE_E_NGKA = 0x00016973;
const char32_t BAMUM_LETTER_PHASE_E_KPEUX = 0x00016974;
const char32_t BAMUM_LETTER_PHASE_E_WUO = 0x00016975;
const char32_t BAMUM_LETTER_PHASE_E_SEE = 0x00016976;
const char32_t BAMUM_LETTER_PHASE_E_NGGEUAET = 0x00016977;
const char32_t BAMUM_LETTER_PHASE_E_PAAM = 0x00016978;
const char32_t BAMUM_LETTER_PHASE_E_TOO = 0x00016979;
const char32_t BAMUM_LETTER_PHASE_E_KUOP = 0x0001697A;
const char32_t BAMUM_LETTER_PHASE_E_LOM = 0x0001697B;
const char32_t BAMUM_LETTER_PHASE_E_NSHIEE = 0x0001697C;
const char32_t BAMUM_LETTER_PHASE_E_NGOP = 0x0001697D;
const char32_t BAMUM_LETTER_PHASE_E_MAEM = 0x0001697E;
const char32_t BAMUM_LETTER_PHASE_E_NGKEUX = 0x0001697F;
const char32_t BAMUM_LETTER_PHASE_E_NGOQ = 0x00016980;
const char32_t BAMUM_LETTER_PHASE_E_NSHUE = 0x00016981;
const char32_t BAMUM_LETTER_PHASE_E_RIMGBA = 0x00016982;
const char32_t BAMUM_LETTER_PHASE_E_NJEUX = 0x00016983;
const char32_t BAMUM_LETTER_PHASE_E_PEEM = 0x00016984;
const char32_t BAMUM_LETTER_PHASE_E_SAA = 0x00016985;
const char32_t BAMUM_LETTER_PHASE_E_NGGURAE = 0x00016986;
const char32_t BAMUM_LETTER_PHASE_E_MGBA = 0x00016987;
const char32_t BAMUM_LETTER_PHASE_E_GHEUX = 0x00016988;
const char32_t BAMUM_LETTER_PHASE_E_NGKEUAEM = 0x00016989;
const char32_t BAMUM_LETTER_PHASE_E_NJAEMLI = 0x0001698A;
const char32_t BAMUM_LETTER_PHASE_E_MAP = 0x0001698B;
const char32_t BAMUM_LETTER_PHASE_E_LOOT = 0x0001698C;
const char32_t BAMUM_LETTER_PHASE_E_NGGEEEE = 0x0001698D;
const char32_t BAMUM_LETTER_PHASE_E_NDIQ = 0x0001698E;
const char32_t BAMUM_LETTER_PHASE_E_TAEN_NTEUM = 0x0001698F;
const char32_t BAMUM_LETTER_PHASE_E_SET = 0x00016990;
const char32_t BAMUM_LETTER_PHASE_E_PUM = 0x00016991;
const char32_t BAMUM_LETTER_PHASE_E_NDAA_SOFTNESS = 0x00016992;
const char32_t BAMUM_LETTER_PHASE_E_NGGUAESHAE_NYAM = 0x00016993;
const char32_t BAMUM_LETTER_PHASE_E_YIEE = 0x00016994;
const char32_t BAMUM_LETTER_PHASE_E_GHEUN = 0x00016995;
const char32_t BAMUM_LETTER_PHASE_E_TUAE = 0x00016996;
const char32_t BAMUM_LETTER_PHASE_E_YEUAE = 0x00016997;
const char32_t BAMUM_LETTER_PHASE_E_PO = 0x00016998;
const char32_t BAMUM_LETTER_PHASE_E_TUMAE = 0x00016999;
const char32_t BAMUM_LETTER_PHASE_E_KEUAE = 0x0001699A;
const char32_t BAMUM_LETTER_PHASE_E_SUAEN = 0x0001699B;
const char32_t BAMUM_LETTER_PHASE_E_TEUAEQ = 0x0001699C;
const char32_t BAMUM_LETTER_PHASE_E_VEUAE = 0x0001699D;
const char32_t BAMUM_LETTER_PHASE_E_WEUX = 0x0001699E;
const char32_t BAMUM_LETTER_PHASE_E_LAAM = 0x0001699F;
const char32_t BAMUM_LETTER_PHASE_E_PU = 0x000169A0;
const char32_t BAMUM_LETTER_PHASE_E_TAAQ = 0x000169A1;
const char32_t BAMUM_LETTER_PHASE_E_GHAAMAE = 0x000169A2;
const char32_t BAMUM_LETTER_PHASE_E_NGEUREUT = 0x000169A3;
const char32_t BAMUM_LETTER_PHASE_E_SHEUAEQ = 0x000169A4;
const char32_t BAMUM_LETTER_PHASE_E_MGBEN = 0x000169A5;
const char32_t BAMUM_LETTER_PHASE_E_MBEE = 0x000169A6;
const char32_t BAMUM_LETTER_PHASE_E_NZAQ = 0x000169A7;
const char32_t BAMUM_LETTER_PHASE_E_NKOM = 0x000169A8;
const char32_t BAMUM_LETTER_PHASE_E_GBET = 0x000169A9;
const char32_t BAMUM_LETTER_PHASE_E_TUM = 0x000169AA;
const char32_t BAMUM_LETTER_PHASE_E_KUET = 0x000169AB;
const char32_t BAMUM_LETTER_PHASE_E_YAP = 0x000169AC;
const char32_t BAMUM_LETTER_PHASE_E_NYI_CLEAVER = 0x000169AD;
const char32_t BAMUM_LETTER_PHASE_E_YIT = 0x000169AE;
const char32_t BAMUM_LETTER_PHASE_E_MFEUQ = 0x000169AF;
const char32_t BAMUM_LETTER_PHASE_E_NDIAQ = 0x000169B0;
const char32_t BAMUM_LETTER_PHASE_E_PIEEQ = 0x000169B1;
const char32_t BAMUM_LETTER_PHASE_E_YUEQ = 0x000169B2;
const char32_t BAMUM_LETTER_PHASE_E_LEUAEM = 0x000169B3;
const char32_t BAMUM_LETTER_PHASE_E_FUE = 0x000169B4;
const char32_t BAMUM_LETTER_PHASE_E_GBEUX = 0x000169B5;
const char32_t BAMUM_LETTER_PHASE_E_NGKUP = 0x000169B6;
const char32_t BAMUM_LETTER_PHASE_E_KET = 0x000169B7;
const char32_t BAMUM_LETTER_PHASE_E_MAE = 0x000169B8;
const char32_t BAMUM_LETTER_PHASE_E_NGKAAMI = 0x000169B9;
const char32_t BAMUM_LETTER_PHASE_E_GHET = 0x000169BA;
const char32_t BAMUM_LETTER_PHASE_E_FA = 0x000169BB;
const char32_t BAMUM_LETTER_PHASE_E_NTUM = 0x000169BC;
const char32_t BAMUM_LETTER_PHASE_E_PEUT = 0x000169BD;
const char32_t BAMUM_LETTER_PHASE_E_YEUM = 0x000169BE;
const char32_t BAMUM_LETTER_PHASE_E_NGGEUAE = 0x000169BF;
const char32_t BAMUM_LETTER_PHASE_E_NYI_BETWEEN = 0x000169C0;
const char32_t BAMUM_LETTER_PHASE_E_NZUQ = 0x000169C1;
const char32_t BAMUM_LETTER_PHASE_E_POON = 0x000169C2;
const char32_t BAMUM_LETTER_PHASE_E_MIEE = 0x000169C3;
const char32_t BAMUM_LETTER_PHASE_E_FUET = 0x000169C4;
const char32_t BAMUM_LETTER_PHASE_E_NAE = 0x000169C5;
const char32_t BAMUM_LETTER_PHASE_E_MUAE = 0x000169C6;
const char32_t BAMUM_LETTER_PHASE_E_GHEUAE = 0x000169C7;
const char32_t BAMUM_LETTER_PHASE_E_FU_I = 0x000169C8;
const char32_t BAMUM_LETTER_PHASE_E_MVI = 0x000169C9;
const char32_t BAMUM_LETTER_PHASE_E_PUAQ = 0x000169CA;
const char32_t BAMUM_LETTER_PHASE_E_NGKUM = 0x000169CB;
const char32_t BAMUM_LETTER_PHASE_E_KUT = 0x000169CC;
const char32_t BAMUM_LETTER_PHASE_E_PIET = 0x000169CD;
const char32_t BAMUM_LETTER_PHASE_E_NTAP = 0x000169CE;
const char32_t BAMUM_LETTER_PHASE_E_YEUAET = 0x000169CF;
const char32_t BAMUM_LETTER_PHASE_E_NGGUP = 0x000169D0;
const char32_t BAMUM_LETTER_PHASE_E_PA_PEOPLE = 0x000169D1;
const char32_t BAMUM_LETTER_PHASE_E_FU_CALL = 0x000169D2;
const char32_t BAMUM_LETTER_PHASE_E_FOM = 0x000169D3;
const char32_t BAMUM_LETTER_PHASE_E_NJEE = 0x000169D4;
const char32_t BAMUM_LETTER_PHASE_E_A = 0x000169D5;
const char32_t BAMUM_LETTER_PHASE_E_TOQ = 0x000169D6;
const char32_t BAMUM_LETTER_PHASE_E_O = 0x000169D7;
const char32_t BAMUM_LETTER_PHASE_E_I = 0x000169D8;
const char32_t BAMUM_LETTER_PHASE_E_LAQ = 0x000169D9;
const char32_t BAMUM_LETTER_PHASE_E_PA_PLURAL = 0x000169DA;
const char32_t BAMUM_LETTER_PHASE_E_TAA = 0x000169DB;
const char32_t BAMUM_LETTER_PHASE_E_TAQ = 0x000169DC;
const char32_t BAMUM_LETTER_PHASE_E_NDAA_MY_HOUSE = 0x000169DD;
const char32_t BAMUM_LETTER_PHASE_E_SHIQ = 0x000169DE;
const char32_t BAMUM_LETTER_PHASE_E_YEUX = 0x000169DF;
const char32_t BAMUM_LETTER_PHASE_E_NGUAE = 0x000169E0;
const char32_t BAMUM_LETTER_PHASE_E_YUAEN = 0x000169E1;
const char32_t BAMUM_LETTER_PHASE_E_YOQ_SWIMMING = 0x000169E2;
const char32_t BAMUM_LETTER_PHASE_E_YOQ_COVER = 0x000169E3;
const char32_t BAMUM_LETTER_PHASE_E_YUQ = 0x000169E4;
const char32_t BAMUM_LETTER_PHASE_E_YUN = 0x000169E5;
const char32_t BAMUM_LETTER_PHASE_E_KEUX = 0x000169E6;
const char32_t BAMUM_LETTER_PHASE_E_PEUX = 0x000169E7;
const char32_t BAMUM_LETTER_PHASE_E_NJEE_EPOCH = 0x000169E8;
const char32_t BAMUM_LETTER_PHASE_E_PUE = 0x000169E9;
const char32_t BAMUM_LETTER_PHASE_E_WUE = 0x000169EA;
const char32_t BAMUM_LETTER_PHASE_E_FEE = 0x000169EB;
const char32_t BAMUM_LETTER_PHASE_E_VEE = 0x000169EC;
const char32_t BAMUM_LETTER_PHASE_E_LU = 0x000169ED;
const char32_t BAMUM_LETTER_PHASE_E_MI = 0x000169EE;
const char32_t BAMUM_LETTER_PHASE_E_REUX = 0x000169EF;
const char32_t BAMUM_LETTER_PHASE_E_RAE = 0x000169F0;
const char32_t BAMUM_LETTER_PHASE_E_NGUAET = 0x000169F1;
const char32_t BAMUM_LETTER_PHASE_E_NGA = 0x000169F2;
const char32_t BAMUM_LETTER_PHASE_E_SHO = 0x000169F3;
const char32_t BAMUM_LETTER_PHASE_E_SHOQ = 0x000169F4;
const char32_t BAMUM_LETTER_PHASE_E_FU_REMEDY = 0x000169F5;
const char32_t BAMUM_LETTER_PHASE_E_NA = 0x000169F6;
const char32_t BAMUM_LETTER_PHASE_E_PI = 0x000169F7;
const char32_t BAMUM_LETTER_PHASE_E_LOQ = 0x000169F8;
const char32_t BAMUM_LETTER_PHASE_E_KO = 0x000169F9;
const char32_t BAMUM_LETTER_PHASE_E_MEN = 0x000169FA;
const char32_t BAMUM_LETTER_PHASE_E_MA = 0x000169FB;
const char32_t BAMUM_LETTER_PHASE_E_MAQ = 0x000169FC;
const char32_t BAMUM_LETTER_PHASE_E_TEU = 0x000169FD;
const char32_t BAMUM_LETTER_PHASE_E_KI = 0x000169FE;
const char32_t BAMUM_LETTER_PHASE_E_MON = 0x000169FF;
const char32_t BAMUM_LETTER_PHASE_E_TEN = 0x00016A00;
const char32_t BAMUM_LETTER_PHASE_E_FAQ = 0x00016A01;
const char32_t BAMUM_LETTER_PHASE_E_GHOM = 0x00016A02;
const char32_t BAMUM_LETTER_PHASE_F_KA = 0x00016A03;
const char32_t BAMUM_LETTER_PHASE_F_U = 0x00016A04;
const char32_t BAMUM_LETTER_PHASE_F_KU = 0x00016A05;
const char32_t BAMUM_LETTER_PHASE_F_EE = 0x00016A06;
const char32_t BAMUM_LETTER_PHASE_F_REE = 0x00016A07;
const char32_t BAMUM_LETTER_PHASE_F_TAE = 0x00016A08;
const char32_t BAMUM_LETTER_PHASE_F_NYI = 0x00016A09;
const char32_t BAMUM_LETTER_PHASE_F_LA = 0x00016A0A;
const char32_t BAMUM_LETTER_PHASE_F_RII = 0x00016A0B;
const char32_t BAMUM_LETTER_PHASE_F_RIEE = 0x00016A0C;
const char32_t BAMUM_LETTER_PHASE_F_MEEEE = 0x00016A0D;
const char32_t BAMUM_LETTER_PHASE_F_TAA = 0x00016A0E;
const char32_t BAMUM_LETTER_PHASE_F_NDAA = 0x00016A0F;
const char32_t BAMUM_LETTER_PHASE_F_NJAEM = 0x00016A10;
const char32_t BAMUM_LETTER_PHASE_F_M = 0x00016A11;
const char32_t BAMUM_LETTER_PHASE_F_SUU = 0x00016A12;
const char32_t BAMUM_LETTER_PHASE_F_SHII = 0x00016A13;
const char32_t BAMUM_LETTER_PHASE_F_SI = 0x00016A14;
const char32_t BAMUM_LETTER_PHASE_F_SEUX = 0x00016A15;
const char32_t BAMUM_LETTER_PHASE_F_KYEE = 0x00016A16;
const char32_t BAMUM_LETTER_PHASE_F_KET = 0x00016A17;
const char32_t BAMUM_LETTER_PHASE_F_NUAE = 0x00016A18;
const char32_t BAMUM_LETTER_PHASE_F_NU = 0x00016A19;
const char32_t BAMUM_LETTER_PHASE_F_NJUAE = 0x00016A1A;
const char32_t BAMUM_LETTER_PHASE_F_YOQ = 0x00016A1B;
const char32_t BAMUM_LETTER_PHASE_F_SHU = 0x00016A1C;
const char32_t BAMUM_LETTER_PHASE_F_YA = 0x00016A1D;
const char32_t BAMUM_LETTER_PHASE_F_NSHA = 0x00016A1E;
const char32_t BAMUM_LETTER_PHASE_F_PEUX = 0x00016A1F;
const char32_t BAMUM_LETTER_PHASE_F_NTEE = 0x00016A20;
const char32_t BAMUM_LETTER_PHASE_F_WUE = 0x00016A21;
const char32_t BAMUM_LETTER_PHASE_F_PEE = 0x00016A22;
const char32_t BAMUM_LETTER_PHASE_F_RU = 0x00016A23;
const char32_t BAMUM_LETTER_PHASE_F_NI = 0x00016A24;
const char32_t BAMUM_LETTER_PHASE_F_REUX = 0x00016A25;
const char32_t BAMUM_LETTER_PHASE_F_KEN = 0x00016A26;
const char32_t BAMUM_LETTER_PHASE_F_NGKWAEN = 0x00016A27;
const char32_t BAMUM_LETTER_PHASE_F_NGGA = 0x00016A28;
const char32_t BAMUM_LETTER_PHASE_F_SHO = 0x00016A29;
const char32_t BAMUM_LETTER_PHASE_F_PUAE = 0x00016A2A;
const char32_t BAMUM_LETTER_PHASE_F_FOM = 0x00016A2B;
const char32_t BAMUM_LETTER_PHASE_F_WA = 0x00016A2C;
const char32_t BAMUM_LETTER_PHASE_F_LI = 0x00016A2D;
const char32_t BAMUM_LETTER_PHASE_F_LOQ = 0x00016A2E;
const char32_t BAMUM_LETTER_PHASE_F_KO = 0x00016A2F;
const char32_t BAMUM_LETTER_PHASE_F_MBEN = 0x00016A30;
const char32_t BAMUM_LETTER_PHASE_F_REN = 0x00016A31;
const char32_t BAMUM_LETTER_PHASE_F_MA = 0x00016A32;
const char32_t BAMUM_LETTER_PHASE_F_MO = 0x00016A33;
const char32_t BAMUM_LETTER_PHASE_F_MBAA = 0x00016A34;
const char32_t BAMUM_LETTER_PHASE_F_TET = 0x00016A35;
const char32_t BAMUM_LETTER_PHASE_F_KPA = 0x00016A36;
const char32_t BAMUM_LETTER_PHASE_F_SAMBA = 0x00016A37;
const char32_t BAMUM_LETTER_PHASE_F_VUEQ = 0x00016A38;
const char32_t MRO_LETTER_TA = 0x00016A40;
const char32_t MRO_LETTER_NGI = 0x00016A41;
const char32_t MRO_LETTER_YO = 0x00016A42;
const char32_t MRO_LETTER_MIM = 0x00016A43;
const char32_t MRO_LETTER_BA = 0x00016A44;
const char32_t MRO_LETTER_DA = 0x00016A45;
const char32_t MRO_LETTER_A = 0x00016A46;
const char32_t MRO_LETTER_PHI = 0x00016A47;
const char32_t MRO_LETTER_KHAI = 0x00016A48;
const char32_t MRO_LETTER_HAO = 0x00016A49;
const char32_t MRO_LETTER_DAI = 0x00016A4A;
const char32_t MRO_LETTER_CHU = 0x00016A4B;
const char32_t MRO_LETTER_KEAAE = 0x00016A4C;
const char32_t MRO_LETTER_OL = 0x00016A4D;
const char32_t MRO_LETTER_MAEM = 0x00016A4E;
const char32_t MRO_LETTER_NIN = 0x00016A4F;
const char32_t MRO_LETTER_PA = 0x00016A50;
const char32_t MRO_LETTER_OO = 0x00016A51;
const char32_t MRO_LETTER_O = 0x00016A52;
const char32_t MRO_LETTER_RO = 0x00016A53;
const char32_t MRO_LETTER_SHI = 0x00016A54;
const char32_t MRO_LETTER_THEA = 0x00016A55;
const char32_t MRO_LETTER_EA = 0x00016A56;
const char32_t MRO_LETTER_WA = 0x00016A57;
const char32_t MRO_LETTER_E = 0x00016A58;
const char32_t MRO_LETTER_KO = 0x00016A59;
const char32_t MRO_LETTER_LAN = 0x00016A5A;
const char32_t MRO_LETTER_LA = 0x00016A5B;
const char32_t MRO_LETTER_HAI = 0x00016A5C;
const char32_t MRO_LETTER_RI = 0x00016A5D;
const char32_t MRO_LETTER_TEK = 0x00016A5E;
const char32_t MRO_DIGIT_ZERO = 0x00016A60;
const char32_t MRO_DIGIT_ONE = 0x00016A61;
const char32_t MRO_DIGIT_TWO = 0x00016A62;
const char32_t MRO_DIGIT_THREE = 0x00016A63;
const char32_t MRO_DIGIT_FOUR = 0x00016A64;
const char32_t MRO_DIGIT_FIVE = 0x00016A65;
const char32_t MRO_DIGIT_SIX = 0x00016A66;
const char32_t MRO_DIGIT_SEVEN = 0x00016A67;
const char32_t MRO_DIGIT_EIGHT = 0x00016A68;
const char32_t MRO_DIGIT_NINE = 0x00016A69;
const char32_t MRO_DANDA = 0x00016A6E;
const char32_t MRO_DOUBLE_DANDA = 0x00016A6F;
const char32_t TANGSA_LETTER_OZ = 0x00016A70;
const char32_t TANGSA_LETTER_OC = 0x00016A71;
const char32_t TANGSA_LETTER_OQ = 0x00016A72;
const char32_t TANGSA_LETTER_OX = 0x00016A73;
const char32_t TANGSA_LETTER_AZ = 0x00016A74;
const char32_t TANGSA_LETTER_AC = 0x00016A75;
const char32_t TANGSA_LETTER_AQ = 0x00016A76;
const char32_t TANGSA_LETTER_AX = 0x00016A77;
const char32_t TANGSA_LETTER_VZ = 0x00016A78;
const char32_t TANGSA_LETTER_VC = 0x00016A79;
const char32_t TANGSA_LETTER_VQ = 0x00016A7A;
const char32_t TANGSA_LETTER_VX = 0x00016A7B;
const char32_t TANGSA_LETTER_EZ = 0x00016A7C;
const char32_t TANGSA_LETTER_EC = 0x00016A7D;
const char32_t TANGSA_LETTER_EQ = 0x00016A7E;
const char32_t TANGSA_LETTER_EX = 0x00016A7F;
const char32_t TANGSA_LETTER_IZ = 0x00016A80;
const char32_t TANGSA_LETTER_IC = 0x00016A81;
const char32_t TANGSA_LETTER_IQ = 0x00016A82;
const char32_t TANGSA_LETTER_IX = 0x00016A83;
const char32_t TANGSA_LETTER_UZ = 0x00016A84;
const char32_t TANGSA_LETTER_UC = 0x00016A85;
const char32_t TANGSA_LETTER_UQ = 0x00016A86;
const char32_t TANGSA_LETTER_UX = 0x00016A87;
const char32_t TANGSA_LETTER_AWZ = 0x00016A88;
const char32_t TANGSA_LETTER_AWC = 0x00016A89;
const char32_t TANGSA_LETTER_AWQ = 0x00016A8A;
const char32_t TANGSA_LETTER_AWX = 0x00016A8B;
const char32_t TANGSA_LETTER_UIZ = 0x00016A8C;
const char32_t TANGSA_LETTER_UIC = 0x00016A8D;
const char32_t TANGSA_LETTER_UIQ = 0x00016A8E;
const char32_t TANGSA_LETTER_UIX = 0x00016A8F;
const char32_t TANGSA_LETTER_FINAL_NG = 0x00016A90;
const char32_t TANGSA_LETTER_LONG_UEX = 0x00016A91;
const char32_t TANGSA_LETTER_SHORT_UEZ = 0x00016A92;
const char32_t TANGSA_LETTER_SHORT_AWX = 0x00016A93;
const char32_t TANGSA_LETTER_UEC = 0x00016A94;
const char32_t TANGSA_LETTER_UEZ = 0x00016A95;
const char32_t TANGSA_LETTER_UEQ = 0x00016A96;
const char32_t TANGSA_LETTER_UEX = 0x00016A97;
const char32_t TANGSA_LETTER_UIUZ = 0x00016A98;
const char32_t TANGSA_LETTER_UIUC = 0x00016A99;
const char32_t TANGSA_LETTER_UIUQ = 0x00016A9A;
const char32_t TANGSA_LETTER_UIUX = 0x00016A9B;
const char32_t TANGSA_LETTER_MZ = 0x00016A9C;
const char32_t TANGSA_LETTER_MC = 0x00016A9D;
const char32_t TANGSA_LETTER_MQ = 0x00016A9E;
const char32_t TANGSA_LETTER_MX = 0x00016A9F;
const char32_t TANGSA_LETTER_KA = 0x00016AA0;
const char32_t TANGSA_LETTER_KHA = 0x00016AA1;
const char32_t TANGSA_LETTER_GA = 0x00016AA2;
const char32_t TANGSA_LETTER_NGA = 0x00016AA3;
const char32_t TANGSA_LETTER_SA = 0x00016AA4;
const char32_t TANGSA_LETTER_YA = 0x00016AA5;
const char32_t TANGSA_LETTER_WA = 0x00016AA6;
const char32_t TANGSA_LETTER_PA = 0x00016AA7;
const char32_t TANGSA_LETTER_NYA = 0x00016AA8;
const char32_t TANGSA_LETTER_PHA = 0x00016AA9;
const char32_t TANGSA_LETTER_BA = 0x00016AAA;
const char32_t TANGSA_LETTER_MA = 0x00016AAB;
const char32_t TANGSA_LETTER_NA = 0x00016AAC;
const char32_t TANGSA_LETTER_HA = 0x00016AAD;
const char32_t TANGSA_LETTER_LA = 0x00016AAE;
const char32_t TANGSA_LETTER_HTA = 0x00016AAF;
const char32_t TANGSA_LETTER_TA = 0x00016AB0;
const char32_t TANGSA_LETTER_DA = 0x00016AB1;
const char32_t TANGSA_LETTER_RA = 0x00016AB2;
const char32_t TANGSA_LETTER_NHA = 0x00016AB3;
const char32_t TANGSA_LETTER_SHA = 0x00016AB4;
const char32_t TANGSA_LETTER_CA = 0x00016AB5;
const char32_t TANGSA_LETTER_TSA = 0x00016AB6;
const char32_t TANGSA_LETTER_GHA = 0x00016AB7;
const char32_t TANGSA_LETTER_HTTA = 0x00016AB8;
const char32_t TANGSA_LETTER_THA = 0x00016AB9;
const char32_t TANGSA_LETTER_XA = 0x00016ABA;
const char32_t TANGSA_LETTER_FA = 0x00016ABB;
const char32_t TANGSA_LETTER_DHA = 0x00016ABC;
const char32_t TANGSA_LETTER_CHA = 0x00016ABD;
const char32_t TANGSA_LETTER_ZA = 0x00016ABE;
const char32_t TANGSA_DIGIT_ZERO = 0x00016AC0;
const char32_t TANGSA_DIGIT_ONE = 0x00016AC1;
const char32_t TANGSA_DIGIT_TWO = 0x00016AC2;
const char32_t TANGSA_DIGIT_THREE = 0x00016AC3;
const char32_t TANGSA_DIGIT_FOUR = 0x00016AC4;
const char32_t TANGSA_DIGIT_FIVE = 0x00016AC5;
const char32_t TANGSA_DIGIT_SIX = 0x00016AC6;
const char32_t TANGSA_DIGIT_SEVEN = 0x00016AC7;
const char32_t TANGSA_DIGIT_EIGHT = 0x00016AC8;
const char32_t TANGSA_DIGIT_NINE = 0x00016AC9;
const char32_t BASSA_VAH_LETTER_ENNI = 0x00016AD0;
const char32_t BASSA_VAH_LETTER_KA = 0x00016AD1;
const char32_t BASSA_VAH_LETTER_SE = 0x00016AD2;
const char32_t BASSA_VAH_LETTER_FA = 0x00016AD3;
const char32_t BASSA_VAH_LETTER_MBE = 0x00016AD4;
const char32_t BASSA_VAH_LETTER_YIE = 0x00016AD5;
const char32_t BASSA_VAH_LETTER_GAH = 0x00016AD6;
const char32_t BASSA_VAH_LETTER_DHII = 0x00016AD7;
const char32_t BASSA_VAH_LETTER_KPAH = 0x00016AD8;
const char32_t BASSA_VAH_LETTER_JO = 0x00016AD9;
const char32_t BASSA_VAH_LETTER_HWAH = 0x00016ADA;
const char32_t BASSA_VAH_LETTER_WA = 0x00016ADB;
const char32_t BASSA_VAH_LETTER_ZO = 0x00016ADC;
const char32_t BASSA_VAH_LETTER_GBU = 0x00016ADD;
const char32_t BASSA_VAH_LETTER_DO = 0x00016ADE;
const char32_t BASSA_VAH_LETTER_CE = 0x00016ADF;
const char32_t BASSA_VAH_LETTER_UWU = 0x00016AE0;
const char32_t BASSA_VAH_LETTER_TO = 0x00016AE1;
const char32_t BASSA_VAH_LETTER_BA = 0x00016AE2;
const char32_t BASSA_VAH_LETTER_VU = 0x00016AE3;
const char32_t BASSA_VAH_LETTER_YEIN = 0x00016AE4;
const char32_t BASSA_VAH_LETTER_PA = 0x00016AE5;
const char32_t BASSA_VAH_LETTER_WADDA = 0x00016AE6;
const char32_t BASSA_VAH_LETTER_A = 0x00016AE7;
const char32_t BASSA_VAH_LETTER_O = 0x00016AE8;
const char32_t BASSA_VAH_LETTER_OO = 0x00016AE9;
const char32_t BASSA_VAH_LETTER_U = 0x00016AEA;
const char32_t BASSA_VAH_LETTER_EE = 0x00016AEB;
const char32_t BASSA_VAH_LETTER_E = 0x00016AEC;
const char32_t BASSA_VAH_LETTER_I = 0x00016AED;
const char32_t BASSA_VAH_COMBINING_HIGH_TONE = 0x00016AF0;
const char32_t BASSA_VAH_COMBINING_LOW_TONE = 0x00016AF1;
const char32_t BASSA_VAH_COMBINING_MID_TONE = 0x00016AF2;
const char32_t BASSA_VAH_COMBINING_LOW_MID_TONE = 0x00016AF3;
const char32_t BASSA_VAH_COMBINING_HIGH_LOW_TONE = 0x00016AF4;
const char32_t BASSA_VAH_FULL_STOP = 0x00016AF5;
const char32_t PAHAWH_HMONG_VOWEL_KEEB = 0x00016B00;
const char32_t PAHAWH_HMONG_VOWEL_KEEV = 0x00016B01;
const char32_t PAHAWH_HMONG_VOWEL_KIB = 0x00016B02;
const char32_t PAHAWH_HMONG_VOWEL_KIV = 0x00016B03;
const char32_t PAHAWH_HMONG_VOWEL_KAUB = 0x00016B04;
const char32_t PAHAWH_HMONG_VOWEL_KAUV = 0x00016B05;
const char32_t PAHAWH_HMONG_VOWEL_KUB = 0x00016B06;
const char32_t PAHAWH_HMONG_VOWEL_KUV = 0x00016B07;
const char32_t PAHAWH_HMONG_VOWEL_KEB = 0x00016B08;
const char32_t PAHAWH_HMONG_VOWEL_KEV = 0x00016B09;
const char32_t PAHAWH_HMONG_VOWEL_KAIB = 0x00016B0A;
const char32_t PAHAWH_HMONG_VOWEL_KAIV = 0x00016B0B;
const char32_t PAHAWH_HMONG_VOWEL_KOOB = 0x00016B0C;
const char32_t PAHAWH_HMONG_VOWEL_KOOV = 0x00016B0D;
const char32_t PAHAWH_HMONG_VOWEL_KAWB = 0x00016B0E;
const char32_t PAHAWH_HMONG_VOWEL_KAWV = 0x00016B0F;
const char32_t PAHAWH_HMONG_VOWEL_KUAB = 0x00016B10;
const char32_t PAHAWH_HMONG_VOWEL_KUAV = 0x00016B11;
const char32_t PAHAWH_HMONG_VOWEL_KOB = 0x00016B12;
const char32_t PAHAWH_HMONG_VOWEL_KOV = 0x00016B13;
const char32_t PAHAWH_HMONG_VOWEL_KIAB = 0x00016B14;
const char32_t PAHAWH_HMONG_VOWEL_KIAV = 0x00016B15;
const char32_t PAHAWH_HMONG_VOWEL_KAB = 0x00016B16;
const char32_t PAHAWH_HMONG_VOWEL_KAV = 0x00016B17;
const char32_t PAHAWH_HMONG_VOWEL_KWB = 0x00016B18;
const char32_t PAHAWH_HMONG_VOWEL_KWV = 0x00016B19;
const char32_t PAHAWH_HMONG_VOWEL_KAAB = 0x00016B1A;
const char32_t PAHAWH_HMONG_VOWEL_KAAV = 0x00016B1B;
const char32_t PAHAWH_HMONG_CONSONANT_VAU = 0x00016B1C;
const char32_t PAHAWH_HMONG_CONSONANT_NTSAU = 0x00016B1D;
const char32_t PAHAWH_HMONG_CONSONANT_LAU = 0x00016B1E;
const char32_t PAHAWH_HMONG_CONSONANT_HAU = 0x00016B1F;
const char32_t PAHAWH_HMONG_CONSONANT_NLAU = 0x00016B20;
const char32_t PAHAWH_HMONG_CONSONANT_RAU = 0x00016B21;
const char32_t PAHAWH_HMONG_CONSONANT_NKAU = 0x00016B22;
const char32_t PAHAWH_HMONG_CONSONANT_QHAU = 0x00016B23;
const char32_t PAHAWH_HMONG_CONSONANT_YAU = 0x00016B24;
const char32_t PAHAWH_HMONG_CONSONANT_HLAU = 0x00016B25;
const char32_t PAHAWH_HMONG_CONSONANT_MAU = 0x00016B26;
const char32_t PAHAWH_HMONG_CONSONANT_CHAU = 0x00016B27;
const char32_t PAHAWH_HMONG_CONSONANT_NCHAU = 0x00016B28;
const char32_t PAHAWH_HMONG_CONSONANT_HNAU = 0x00016B29;
const char32_t PAHAWH_HMONG_CONSONANT_PLHAU = 0x00016B2A;
const char32_t PAHAWH_HMONG_CONSONANT_NTHAU = 0x00016B2B;
const char32_t PAHAWH_HMONG_CONSONANT_NAU = 0x00016B2C;
const char32_t PAHAWH_HMONG_CONSONANT_AU = 0x00016B2D;
const char32_t PAHAWH_HMONG_CONSONANT_XAU = 0x00016B2E;
const char32_t PAHAWH_HMONG_CONSONANT_CAU = 0x00016B2F;
const char32_t PAHAWH_HMONG_MARK_CIM_TUB = 0x00016B30;
const char32_t PAHAWH_HMONG_MARK_CIM_SO = 0x00016B31;
const char32_t PAHAWH_HMONG_MARK_CIM_KES = 0x00016B32;
const char32_t PAHAWH_HMONG_MARK_CIM_KHAV = 0x00016B33;
const char32_t PAHAWH_HMONG_MARK_CIM_SUAM = 0x00016B34;
const char32_t PAHAWH_HMONG_MARK_CIM_HOM = 0x00016B35;
const char32_t PAHAWH_HMONG_MARK_CIM_TAUM = 0x00016B36;
const char32_t PAHAWH_HMONG_SIGN_VOS_THOM = 0x00016B37;
const char32_t PAHAWH_HMONG_SIGN_VOS_TSHAB_CEEB = 0x00016B38;
const char32_t PAHAWH_HMONG_SIGN_CIM_CHEEM = 0x00016B39;
const char32_t PAHAWH_HMONG_SIGN_VOS_THIAB = 0x00016B3A;
const char32_t PAHAWH_HMONG_SIGN_VOS_FEEM = 0x00016B3B;
const char32_t PAHAWH_HMONG_SIGN_XYEEM_NTXIV = 0x00016B3C;
const char32_t PAHAWH_HMONG_SIGN_XYEEM_RHO = 0x00016B3D;
const char32_t PAHAWH_HMONG_SIGN_XYEEM_TOV = 0x00016B3E;
const char32_t PAHAWH_HMONG_SIGN_XYEEM_FAIB = 0x00016B3F;
const char32_t PAHAWH_HMONG_SIGN_VOS_SEEV = 0x00016B40;
const char32_t PAHAWH_HMONG_SIGN_MEEJ_SUAB = 0x00016B41;
const char32_t PAHAWH_HMONG_SIGN_VOS_NRUA = 0x00016B42;
const char32_t PAHAWH_HMONG_SIGN_IB_YAM = 0x00016B43;
const char32_t PAHAWH_HMONG_SIGN_XAUS = 0x00016B44;
const char32_t PAHAWH_HMONG_SIGN_CIM_TSOV_ROG = 0x00016B45;
const char32_t PAHAWH_HMONG_DIGIT_ZERO = 0x00016B50;
const char32_t PAHAWH_HMONG_DIGIT_ONE = 0x00016B51;
const char32_t PAHAWH_HMONG_DIGIT_TWO = 0x00016B52;
const char32_t PAHAWH_HMONG_DIGIT_THREE = 0x00016B53;
const char32_t PAHAWH_HMONG_DIGIT_FOUR = 0x00016B54;
const char32_t PAHAWH_HMONG_DIGIT_FIVE = 0x00016B55;
const char32_t PAHAWH_HMONG_DIGIT_SIX = 0x00016B56;
const char32_t PAHAWH_HMONG_DIGIT_SEVEN = 0x00016B57;
const char32_t PAHAWH_HMONG_DIGIT_EIGHT = 0x00016B58;
const char32_t PAHAWH_HMONG_DIGIT_NINE = 0x00016B59;
const char32_t PAHAWH_HMONG_NUMBER_TENS = 0x00016B5B;
const char32_t PAHAWH_HMONG_NUMBER_HUNDREDS = 0x00016B5C;
const char32_t PAHAWH_HMONG_NUMBER_TEN_THOUSANDS = 0x00016B5D;
const char32_t PAHAWH_HMONG_NUMBER_MILLIONS = 0x00016B5E;
const char32_t PAHAWH_HMONG_NUMBER_HUNDRED_MILLIONS = 0x00016B5F;
const char32_t PAHAWH_HMONG_NUMBER_TEN_BILLIONS = 0x00016B60;
const char32_t PAHAWH_HMONG_NUMBER_TRILLIONS = 0x00016B61;
const char32_t PAHAWH_HMONG_SIGN_VOS_LUB = 0x00016B63;
const char32_t PAHAWH_HMONG_SIGN_XYOO = 0x00016B64;
const char32_t PAHAWH_HMONG_SIGN_HLI = 0x00016B65;
const char32_t PAHAWH_HMONG_SIGN_THIRD_STAGE_HLI = 0x00016B66;
const char32_t PAHAWH_HMONG_SIGN_ZWJ_THAJ = 0x00016B67;
const char32_t PAHAWH_HMONG_SIGN_HNUB = 0x00016B68;
const char32_t PAHAWH_HMONG_SIGN_NQIG = 0x00016B69;
const char32_t PAHAWH_HMONG_SIGN_XIAB = 0x00016B6A;
const char32_t PAHAWH_HMONG_SIGN_NTUJ = 0x00016B6B;
const char32_t PAHAWH_HMONG_SIGN_AV = 0x00016B6C;
const char32_t PAHAWH_HMONG_SIGN_TXHEEJ_CEEV = 0x00016B6D;
const char32_t PAHAWH_HMONG_SIGN_MEEJ_TSEEB = 0x00016B6E;
const char32_t PAHAWH_HMONG_SIGN_TAU = 0x00016B6F;
const char32_t PAHAWH_HMONG_SIGN_LOS = 0x00016B70;
const char32_t PAHAWH_HMONG_SIGN_MUS = 0x00016B71;
const char32_t PAHAWH_HMONG_SIGN_CIM_HAIS_LUS_NTOG_NTOG = 0x00016B72;
const char32_t PAHAWH_HMONG_SIGN_CIM_CUAM_TSHOOJ = 0x00016B73;
const char32_t PAHAWH_HMONG_SIGN_CIM_TXWV = 0x00016B74;
const char32_t PAHAWH_HMONG_SIGN_CIM_TXWV_CHWV = 0x00016B75;
const char32_t PAHAWH_HMONG_SIGN_CIM_PUB_DAWB = 0x00016B76;
const char32_t PAHAWH_HMONG_SIGN_CIM_NRES_TOS = 0x00016B77;
const char32_t PAHAWH_HMONG_CLAN_SIGN_TSHEEJ = 0x00016B7D;
const char32_t PAHAWH_HMONG_CLAN_SIGN_YEEG = 0x00016B7E;
const char32_t PAHAWH_HMONG_CLAN_SIGN_LIS = 0x00016B7F;
const char32_t PAHAWH_HMONG_CLAN_SIGN_LAUJ = 0x00016B80;
const char32_t PAHAWH_HMONG_CLAN_SIGN_XYOOJ = 0x00016B81;
const char32_t PAHAWH_HMONG_CLAN_SIGN_KOO = 0x00016B82;
const char32_t PAHAWH_HMONG_CLAN_SIGN_HAWJ = 0x00016B83;
const char32_t PAHAWH_HMONG_CLAN_SIGN_MUAS = 0x00016B84;
const char32_t PAHAWH_HMONG_CLAN_SIGN_THOJ = 0x00016B85;
const char32_t PAHAWH_HMONG_CLAN_SIGN_TSAB = 0x00016B86;
const char32_t PAHAWH_HMONG_CLAN_SIGN_PHAB = 0x00016B87;
const char32_t PAHAWH_HMONG_CLAN_SIGN_KHAB = 0x00016B88;
const char32_t PAHAWH_HMONG_CLAN_SIGN_HAM = 0x00016B89;
const char32_t PAHAWH_HMONG_CLAN_SIGN_VAJ = 0x00016B8A;
const char32_t PAHAWH_HMONG_CLAN_SIGN_FAJ = 0x00016B8B;
const char32_t PAHAWH_HMONG_CLAN_SIGN_YAJ = 0x00016B8C;
const char32_t PAHAWH_HMONG_CLAN_SIGN_TSWB = 0x00016B8D;
const char32_t PAHAWH_HMONG_CLAN_SIGN_KWM = 0x00016B8E;
const char32_t PAHAWH_HMONG_CLAN_SIGN_VWJ = 0x00016B8F;
const char32_t KIRAT_RAI_SIGN_ANUSVARA = 0x00016D40;
const char32_t KIRAT_RAI_SIGN_TONPI = 0x00016D41;
const char32_t KIRAT_RAI_SIGN_VISARGA = 0x00016D42;
const char32_t KIRAT_RAI_LETTER_A = 0x00016D43;
const char32_t KIRAT_RAI_LETTER_KA = 0x00016D44;
const char32_t KIRAT_RAI_LETTER_KHA = 0x00016D45;
const char32_t KIRAT_RAI_LETTER_GA = 0x00016D46;
const char32_t KIRAT_RAI_LETTER_GHA = 0x00016D47;
const char32_t KIRAT_RAI_LETTER_NGA = 0x00016D48;
const char32_t KIRAT_RAI_LETTER_CA = 0x00016D49;
const char32_t KIRAT_RAI_LETTER_CHA = 0x00016D4A;
const char32_t KIRAT_RAI_LETTER_JA = 0x00016D4B;
const char32_t KIRAT_RAI_LETTER_JHA = 0x00016D4C;
const char32_t KIRAT_RAI_LETTER_NYA = 0x00016D4D;
const char32_t KIRAT_RAI_LETTER_TTA = 0x00016D4E;
const char32_t KIRAT_RAI_LETTER_TTHA = 0x00016D4F;
const char32_t KIRAT_RAI_LETTER_DDA = 0x00016D50;
const char32_t KIRAT_RAI_LETTER_DDHA = 0x00016D51;
const char32_t KIRAT_RAI_LETTER_TA = 0x00016D52;
const char32_t KIRAT_RAI_LETTER_THA = 0x00016D53;
const char32_t KIRAT_RAI_LETTER_DA = 0x00016D54;
const char32_t KIRAT_RAI_LETTER_DHA = 0x00016D55;
const char32_t KIRAT_RAI_LETTER_NA = 0x00016D56;
const char32_t KIRAT_RAI_LETTER_PA = 0x00016D57;
const char32_t KIRAT_RAI_LETTER_PHA = 0x00016D58;
const char32_t KIRAT_RAI_LETTER_BA = 0x00016D59;
const char32_t KIRAT_RAI_LETTER_BHA = 0x00016D5A;
const char32_t KIRAT_RAI_LETTER_MA = 0x00016D5B;
const char32_t KIRAT_RAI_LETTER_YA = 0x00016D5C;
const char32_t KIRAT_RAI_LETTER_RA = 0x00016D5D;
const char32_t KIRAT_RAI_LETTER_LA = 0x00016D5E;
const char32_t KIRAT_RAI_LETTER_VA = 0x00016D5F;
const char32_t KIRAT_RAI_LETTER_SA = 0x00016D60;
const char32_t KIRAT_RAI_LETTER_SHA = 0x00016D61;
const char32_t KIRAT_RAI_LETTER_HA = 0x00016D62;
const char32_t KIRAT_RAI_VOWEL_SIGN_AA = 0x00016D63;
const char32_t KIRAT_RAI_VOWEL_SIGN_I = 0x00016D64;
const char32_t KIRAT_RAI_VOWEL_SIGN_U = 0x00016D65;
const char32_t KIRAT_RAI_VOWEL_SIGN_UE = 0x00016D66;
const char32_t KIRAT_RAI_VOWEL_SIGN_E = 0x00016D67;
const char32_t KIRAT_RAI_VOWEL_SIGN_AI = 0x00016D68;
const char32_t KIRAT_RAI_VOWEL_SIGN_O = 0x00016D69;
const char32_t KIRAT_RAI_VOWEL_SIGN_AU = 0x00016D6A;
const char32_t KIRAT_RAI_SIGN_VIRAMA = 0x00016D6B;
const char32_t KIRAT_RAI_SIGN_SAAT = 0x00016D6C;
const char32_t KIRAT_RAI_SIGN_YUPI = 0x00016D6D;
const char32_t KIRAT_RAI_DANDA = 0x00016D6E;
const char32_t KIRAT_RAI_DOUBLE_DANDA = 0x00016D6F;
const char32_t KIRAT_RAI_DIGIT_ZERO = 0x00016D70;
const char32_t KIRAT_RAI_DIGIT_ONE = 0x00016D71;
const char32_t KIRAT_RAI_DIGIT_TWO = 0x00016D72;
const char32_t KIRAT_RAI_DIGIT_THREE = 0x00016D73;
const char32_t KIRAT_RAI_DIGIT_FOUR = 0x00016D74;
const char32_t KIRAT_RAI_DIGIT_FIVE = 0x00016D75;
const char32_t KIRAT_RAI_DIGIT_SIX = 0x00016D76;
const char32_t KIRAT_RAI_DIGIT_SEVEN = 0x00016D77;
const char32_t KIRAT_RAI_DIGIT_EIGHT = 0x00016D78;
const char32_t KIRAT_RAI_DIGIT_NINE = 0x00016D79;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_M = 0x00016E40;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_S = 0x00016E41;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_V = 0x00016E42;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_W = 0x00016E43;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_ATIU = 0x00016E44;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_Z = 0x00016E45;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_KP = 0x00016E46;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_P = 0x00016E47;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_T = 0x00016E48;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_G = 0x00016E49;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_F = 0x00016E4A;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_I = 0x00016E4B;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_K = 0x00016E4C;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_A = 0x00016E4D;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_J = 0x00016E4E;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_E = 0x00016E4F;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_B = 0x00016E50;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_C = 0x00016E51;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_U = 0x00016E52;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_YU = 0x00016E53;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_L = 0x00016E54;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_Q = 0x00016E55;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_HP = 0x00016E56;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_NY = 0x00016E57;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_X = 0x00016E58;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_D = 0x00016E59;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_OE = 0x00016E5A;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_N = 0x00016E5B;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_R = 0x00016E5C;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_O = 0x00016E5D;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_AI = 0x00016E5E;
const char32_t MEDEFAIDRIN_CAPITAL_LETTER_Y = 0x00016E5F;
const char32_t MEDEFAIDRIN_SMALL_LETTER_M = 0x00016E60;
const char32_t MEDEFAIDRIN_SMALL_LETTER_S = 0x00016E61;
const char32_t MEDEFAIDRIN_SMALL_LETTER_V = 0x00016E62;
const char32_t MEDEFAIDRIN_SMALL_LETTER_W = 0x00016E63;
const char32_t MEDEFAIDRIN_SMALL_LETTER_ATIU = 0x00016E64;
const char32_t MEDEFAIDRIN_SMALL_LETTER_Z = 0x00016E65;
const char32_t MEDEFAIDRIN_SMALL_LETTER_KP = 0x00016E66;
const char32_t MEDEFAIDRIN_SMALL_LETTER_P = 0x00016E67;
const char32_t MEDEFAIDRIN_SMALL_LETTER_T = 0x00016E68;
const char32_t MEDEFAIDRIN_SMALL_LETTER_G = 0x00016E69;
const char32_t MEDEFAIDRIN_SMALL_LETTER_F = 0x00016E6A;
const char32_t MEDEFAIDRIN_SMALL_LETTER_I = 0x00016E6B;
const char32_t MEDEFAIDRIN_SMALL_LETTER_K = 0x00016E6C;
const char32_t MEDEFAIDRIN_SMALL_LETTER_A = 0x00016E6D;
const char32_t MEDEFAIDRIN_SMALL_LETTER_J = 0x00016E6E;
const char32_t MEDEFAIDRIN_SMALL_LETTER_E = 0x00016E6F;
const char32_t MEDEFAIDRIN_SMALL_LETTER_B = 0x00016E70;
const char32_t MEDEFAIDRIN_SMALL_LETTER_C = 0x00016E71;
const char32_t MEDEFAIDRIN_SMALL_LETTER_U = 0x00016E72;
const char32_t MEDEFAIDRIN_SMALL_LETTER_YU = 0x00016E73;
const char32_t MEDEFAIDRIN_SMALL_LETTER_L = 0x00016E74;
const char32_t MEDEFAIDRIN_SMALL_LETTER_Q = 0x00016E75;
const char32_t MEDEFAIDRIN_SMALL_LETTER_HP = 0x00016E76;
const char32_t MEDEFAIDRIN_SMALL_LETTER_NY = 0x00016E77;
const char32_t MEDEFAIDRIN_SMALL_LETTER_X = 0x00016E78;
const char32_t MEDEFAIDRIN_SMALL_LETTER_D = 0x00016E79;
const char32_t MEDEFAIDRIN_SMALL_LETTER_OE = 0x00016E7A;
const char32_t MEDEFAIDRIN_SMALL_LETTER_N = 0x00016E7B;
const char32_t MEDEFAIDRIN_SMALL_LETTER_R = 0x00016E7C;
const char32_t MEDEFAIDRIN_SMALL_LETTER_O = 0x00016E7D;
const char32_t MEDEFAIDRIN_SMALL_LETTER_AI = 0x00016E7E;
const char32_t MEDEFAIDRIN_SMALL_LETTER_Y = 0x00016E7F;
const char32_t MEDEFAIDRIN_DIGIT_ZERO = 0x00016E80;
const char32_t MEDEFAIDRIN_DIGIT_ONE = 0x00016E81;
const char32_t MEDEFAIDRIN_DIGIT_TWO = 0x00016E82;
const char32_t MEDEFAIDRIN_DIGIT_THREE = 0x00016E83;
const char32_t MEDEFAIDRIN_DIGIT_FOUR = 0x00016E84;
const char32_t MEDEFAIDRIN_DIGIT_FIVE = 0x00016E85;
const char32_t MEDEFAIDRIN_DIGIT_SIX = 0x00016E86;
const char32_t MEDEFAIDRIN_DIGIT_SEVEN = 0x00016E87;
const char32_t MEDEFAIDRIN_DIGIT_EIGHT = 0x00016E88;
const char32_t MEDEFAIDRIN_DIGIT_NINE = 0x00016E89;
const char32_t MEDEFAIDRIN_NUMBER_TEN = 0x00016E8A;
const char32_t MEDEFAIDRIN_NUMBER_ELEVEN = 0x00016E8B;
const char32_t MEDEFAIDRIN_NUMBER_TWELVE = 0x00016E8C;
const char32_t MEDEFAIDRIN_NUMBER_THIRTEEN = 0x00016E8D;
const char32_t MEDEFAIDRIN_NUMBER_FOURTEEN = 0x00016E8E;
const char32_t MEDEFAIDRIN_NUMBER_FIFTEEN = 0x00016E8F;
const char32_t MEDEFAIDRIN_NUMBER_SIXTEEN = 0x00016E90;
const char32_t MEDEFAIDRIN_NUMBER_SEVENTEEN = 0x00016E91;
const char32_t MEDEFAIDRIN_NUMBER_EIGHTEEN = 0x00016E92;
const char32_t MEDEFAIDRIN_NUMBER_NINETEEN = 0x00016E93;
const char32_t MEDEFAIDRIN_DIGIT_ONE_ALTERNATE_FORM = 0x00016E94;
const char32_t MEDEFAIDRIN_DIGIT_TWO_ALTERNATE_FORM = 0x00016E95;
const char32_t MEDEFAIDRIN_DIGIT_THREE_ALTERNATE_FORM = 0x00016E96;
const char32_t MEDEFAIDRIN_COMMA = 0x00016E97;
const char32_t MEDEFAIDRIN_FULL_STOP = 0x00016E98;
const char32_t MEDEFAIDRIN_SYMBOL_AIVA = 0x00016E99;
const char32_t MEDEFAIDRIN_EXCLAMATION_OH = 0x00016E9A;
const char32_t BERIA_ERFE_CAPITAL_LETTER_ARKAB = 0x00016EA0;
const char32_t BERIA_ERFE_CAPITAL_LETTER_BASIGNA = 0x00016EA1;
const char32_t BERIA_ERFE_CAPITAL_LETTER_DARBAI = 0x00016EA2;
const char32_t BERIA_ERFE_CAPITAL_LETTER_EH = 0x00016EA3;
const char32_t BERIA_ERFE_CAPITAL_LETTER_FITKO = 0x00016EA4;
const char32_t BERIA_ERFE_CAPITAL_LETTER_GOWAY = 0x00016EA5;
const char32_t BERIA_ERFE_CAPITAL_LETTER_HIRDEABO = 0x00016EA6;
const char32_t BERIA_ERFE_CAPITAL_LETTER_I = 0x00016EA7;
const char32_t BERIA_ERFE_CAPITAL_LETTER_DJAI = 0x00016EA8;
const char32_t BERIA_ERFE_CAPITAL_LETTER_KOBO = 0x00016EA9;
const char32_t BERIA_ERFE_CAPITAL_LETTER_LAKKO = 0x00016EAA;
const char32_t BERIA_ERFE_CAPITAL_LETTER_MERI = 0x00016EAB;
const char32_t BERIA_ERFE_CAPITAL_LETTER_NINI = 0x00016EAC;
const char32_t BERIA_ERFE_CAPITAL_LETTER_GNA = 0x00016EAD;
const char32_t BERIA_ERFE_CAPITAL_LETTER_NGAY = 0x00016EAE;
const char32_t BERIA_ERFE_CAPITAL_LETTER_OI = 0x00016EAF;
const char32_t BERIA_ERFE_CAPITAL_LETTER_PI = 0x00016EB0;
const char32_t BERIA_ERFE_CAPITAL_LETTER_ERIGO = 0x00016EB1;
const char32_t BERIA_ERFE_CAPITAL_LETTER_ERIGO_TAMURA = 0x00016EB2;
const char32_t BERIA_ERFE_CAPITAL_LETTER_SERI = 0x00016EB3;
const char32_t BERIA_ERFE_CAPITAL_LETTER_SHEP = 0x00016EB4;
const char32_t BERIA_ERFE_CAPITAL_LETTER_TATASOUE = 0x00016EB5;
const char32_t BERIA_ERFE_CAPITAL_LETTER_UI = 0x00016EB6;
const char32_t BERIA_ERFE_CAPITAL_LETTER_WASSE = 0x00016EB7;
const char32_t BERIA_ERFE_CAPITAL_LETTER_AY = 0x00016EB8;
const char32_t BERIA_ERFE_SMALL_LETTER_ARKAB = 0x00016EBB;
const char32_t BERIA_ERFE_SMALL_LETTER_BASIGNA = 0x00016EBC;
const char32_t BERIA_ERFE_SMALL_LETTER_DARBAI = 0x00016EBD;
const char32_t BERIA_ERFE_SMALL_LETTER_EH = 0x00016EBE;
const char32_t BERIA_ERFE_SMALL_LETTER_FITKO = 0x00016EBF;
const char32_t BERIA_ERFE_SMALL_LETTER_GOWAY = 0x00016EC0;
const char32_t BERIA_ERFE_SMALL_LETTER_HIRDEABO = 0x00016EC1;
const char32_t BERIA_ERFE_SMALL_LETTER_I = 0x00016EC2;
const char32_t BERIA_ERFE_SMALL_LETTER_DJAI = 0x00016EC3;
const char32_t BERIA_ERFE_SMALL_LETTER_KOBO = 0x00016EC4;
const char32_t BERIA_ERFE_SMALL_LETTER_LAKKO = 0x00016EC5;
const char32_t BERIA_ERFE_SMALL_LETTER_MERI = 0x00016EC6;
const char32_t BERIA_ERFE_SMALL_LETTER_NINI = 0x00016EC7;
const char32_t BERIA_ERFE_SMALL_LETTER_GNA = 0x00016EC8;
const char32_t BERIA_ERFE_SMALL_LETTER_NGAY = 0x00016EC9;
const char32_t BERIA_ERFE_SMALL_LETTER_OI = 0x00016ECA;
const char32_t BERIA_ERFE_SMALL_LETTER_PI = 0x00016ECB;
const char32_t BERIA_ERFE_SMALL_LETTER_ERIGO = 0x00016ECC;
const char32_t BERIA_ERFE_SMALL_LETTER_ERIGO_TAMURA = 0x00016ECD;
const char32_t BERIA_ERFE_SMALL_LETTER_SERI = 0x00016ECE;
const char32_t BERIA_ERFE_SMALL_LETTER_SHEP = 0x00016ECF;
const char32_t BERIA_ERFE_SMALL_LETTER_TATASOUE = 0x00016ED0;
const char32_t BERIA_ERFE_SMALL_LETTER_UI = 0x00016ED1;
const char32_t BERIA_ERFE_SMALL_LETTER_WASSE = 0x00016ED2;
const char32_t BERIA_ERFE_SMALL_LETTER_AY = 0x00016ED3;
const char32_t MIAO_LETTER_PA = 0x00016F00;
const char32_t MIAO_LETTER_BA = 0x00016F01;
const char32_t MIAO_LETTER_YI_PA = 0x00016F02;
const char32_t MIAO_LETTER_PLA = 0x00016F03;
const char32_t MIAO_LETTER_MA = 0x00016F04;
const char32_t MIAO_LETTER_MHA = 0x00016F05;
const char32_t MIAO_LETTER_ARCHAIC_MA = 0x00016F06;
const char32_t MIAO_LETTER_FA = 0x00016F07;
const char32_t MIAO_LETTER_VA = 0x00016F08;
const char32_t MIAO_LETTER_VFA = 0x00016F09;
const char32_t MIAO_LETTER_TA = 0x00016F0A;
const char32_t MIAO_LETTER_DA = 0x00016F0B;
const char32_t MIAO_LETTER_YI_TTA = 0x00016F0C;
const char32_t MIAO_LETTER_YI_TA = 0x00016F0D;
const char32_t MIAO_LETTER_TTA = 0x00016F0E;
const char32_t MIAO_LETTER_DDA = 0x00016F0F;
const char32_t MIAO_LETTER_NA = 0x00016F10;
const char32_t MIAO_LETTER_NHA = 0x00016F11;
const char32_t MIAO_LETTER_YI_NNA = 0x00016F12;
const char32_t MIAO_LETTER_ARCHAIC_NA = 0x00016F13;
const char32_t MIAO_LETTER_NNA = 0x00016F14;
const char32_t MIAO_LETTER_NNHA = 0x00016F15;
const char32_t MIAO_LETTER_LA = 0x00016F16;
const char32_t MIAO_LETTER_LYA = 0x00016F17;
const char32_t MIAO_LETTER_LHA = 0x00016F18;
const char32_t MIAO_LETTER_LHYA = 0x00016F19;
const char32_t MIAO_LETTER_TLHA = 0x00016F1A;
const char32_t MIAO_LETTER_DLHA = 0x00016F1B;
const char32_t MIAO_LETTER_TLHYA = 0x00016F1C;
const char32_t MIAO_LETTER_DLHYA = 0x00016F1D;
const char32_t MIAO_LETTER_KA = 0x00016F1E;
const char32_t MIAO_LETTER_GA = 0x00016F1F;
const char32_t MIAO_LETTER_YI_KA = 0x00016F20;
const char32_t MIAO_LETTER_QA = 0x00016F21;
const char32_t MIAO_LETTER_QGA = 0x00016F22;
const char32_t MIAO_LETTER_NGA = 0x00016F23;
const char32_t MIAO_LETTER_NGHA = 0x00016F24;
const char32_t MIAO_LETTER_ARCHAIC_NGA = 0x00016F25;
const char32_t MIAO_LETTER_HA = 0x00016F26;
const char32_t MIAO_LETTER_XA = 0x00016F27;
const char32_t MIAO_LETTER_GHA = 0x00016F28;
const char32_t MIAO_LETTER_GHHA = 0x00016F29;
const char32_t MIAO_LETTER_TSSA = 0x00016F2A;
const char32_t MIAO_LETTER_DZZA = 0x00016F2B;
const char32_t MIAO_LETTER_NYA = 0x00016F2C;
const char32_t MIAO_LETTER_NYHA = 0x00016F2D;
const char32_t MIAO_LETTER_TSHA = 0x00016F2E;
const char32_t MIAO_LETTER_DZHA = 0x00016F2F;
const char32_t MIAO_LETTER_YI_TSHA = 0x00016F30;
const char32_t MIAO_LETTER_YI_DZHA = 0x00016F31;
const char32_t MIAO_LETTER_REFORMED_TSHA = 0x00016F32;
const char32_t MIAO_LETTER_SHA = 0x00016F33;
const char32_t MIAO_LETTER_SSA = 0x00016F34;
const char32_t MIAO_LETTER_ZHA = 0x00016F35;
const char32_t MIAO_LETTER_ZSHA = 0x00016F36;
const char32_t MIAO_LETTER_TSA = 0x00016F37;
const char32_t MIAO_LETTER_DZA = 0x00016F38;
const char32_t MIAO_LETTER_YI_TSA = 0x00016F39;
const char32_t MIAO_LETTER_SA = 0x00016F3A;
const char32_t MIAO_LETTER_ZA = 0x00016F3B;
const char32_t MIAO_LETTER_ZSA = 0x00016F3C;
const char32_t MIAO_LETTER_ZZA = 0x00016F3D;
const char32_t MIAO_LETTER_ZZSA = 0x00016F3E;
const char32_t MIAO_LETTER_ARCHAIC_ZZA = 0x00016F3F;
const char32_t MIAO_LETTER_ZZYA = 0x00016F40;
const char32_t MIAO_LETTER_ZZSYA = 0x00016F41;
const char32_t MIAO_LETTER_WA = 0x00016F42;
const char32_t MIAO_LETTER_AH = 0x00016F43;
const char32_t MIAO_LETTER_HHA = 0x00016F44;
const char32_t MIAO_LETTER_BRI = 0x00016F45;
const char32_t MIAO_LETTER_SYI = 0x00016F46;
const char32_t MIAO_LETTER_DZYI = 0x00016F47;
const char32_t MIAO_LETTER_TE = 0x00016F48;
const char32_t MIAO_LETTER_TSE = 0x00016F49;
const char32_t MIAO_LETTER_RTE = 0x00016F4A;
const char32_t MIAO_SIGN_CONSONANT_MODIFIER_BAR = 0x00016F4F;
const char32_t MIAO_LETTER_NASALIZATION = 0x00016F50;
const char32_t MIAO_SIGN_ASPIRATION = 0x00016F51;
const char32_t MIAO_SIGN_REFORMED_VOICING = 0x00016F52;
const char32_t MIAO_SIGN_REFORMED_ASPIRATION = 0x00016F53;
const char32_t MIAO_VOWEL_SIGN_A = 0x00016F54;
const char32_t MIAO_VOWEL_SIGN_AA = 0x00016F55;
const char32_t MIAO_VOWEL_SIGN_AHH = 0x00016F56;
const char32_t MIAO_VOWEL_SIGN_AN = 0x00016F57;
const char32_t MIAO_VOWEL_SIGN_ANG = 0x00016F58;
const char32_t MIAO_VOWEL_SIGN_O = 0x00016F59;
const char32_t MIAO_VOWEL_SIGN_OO = 0x00016F5A;
const char32_t MIAO_VOWEL_SIGN_WO = 0x00016F5B;
const char32_t MIAO_VOWEL_SIGN_W = 0x00016F5C;
const char32_t MIAO_VOWEL_SIGN_E = 0x00016F5D;
const char32_t MIAO_VOWEL_SIGN_EN = 0x00016F5E;
const char32_t MIAO_VOWEL_SIGN_ENG = 0x00016F5F;
const char32_t MIAO_VOWEL_SIGN_OEY = 0x00016F60;
const char32_t MIAO_VOWEL_SIGN_I = 0x00016F61;
const char32_t MIAO_VOWEL_SIGN_IA = 0x00016F62;
const char32_t MIAO_VOWEL_SIGN_IAN = 0x00016F63;
const char32_t MIAO_VOWEL_SIGN_IANG = 0x00016F64;
const char32_t MIAO_VOWEL_SIGN_IO = 0x00016F65;
const char32_t MIAO_VOWEL_SIGN_IE = 0x00016F66;
const char32_t MIAO_VOWEL_SIGN_II = 0x00016F67;
const char32_t MIAO_VOWEL_SIGN_IU = 0x00016F68;
const char32_t MIAO_VOWEL_SIGN_ING = 0x00016F69;
const char32_t MIAO_VOWEL_SIGN_U = 0x00016F6A;
const char32_t MIAO_VOWEL_SIGN_UA = 0x00016F6B;
const char32_t MIAO_VOWEL_SIGN_UAN = 0x00016F6C;
const char32_t MIAO_VOWEL_SIGN_UANG = 0x00016F6D;
const char32_t MIAO_VOWEL_SIGN_UU = 0x00016F6E;
const char32_t MIAO_VOWEL_SIGN_UEI = 0x00016F6F;
const char32_t MIAO_VOWEL_SIGN_UNG = 0x00016F70;
const char32_t MIAO_VOWEL_SIGN_Y = 0x00016F71;
const char32_t MIAO_VOWEL_SIGN_YI = 0x00016F72;
const char32_t MIAO_VOWEL_SIGN_AE = 0x00016F73;
const char32_t MIAO_VOWEL_SIGN_AEE = 0x00016F74;
const char32_t MIAO_VOWEL_SIGN_ERR = 0x00016F75;
const char32_t MIAO_VOWEL_SIGN_ROUNDED_ERR = 0x00016F76;
const char32_t MIAO_VOWEL_SIGN_ER = 0x00016F77;
const char32_t MIAO_VOWEL_SIGN_ROUNDED_ER = 0x00016F78;
const char32_t MIAO_VOWEL_SIGN_AI = 0x00016F79;
const char32_t MIAO_VOWEL_SIGN_EI = 0x00016F7A;
const char32_t MIAO_VOWEL_SIGN_AU = 0x00016F7B;
const char32_t MIAO_VOWEL_SIGN_OU = 0x00016F7C;
const char32_t MIAO_VOWEL_SIGN_N = 0x00016F7D;
const char32_t MIAO_VOWEL_SIGN_NG = 0x00016F7E;
const char32_t MIAO_VOWEL_SIGN_UOG = 0x00016F7F;
const char32_t MIAO_VOWEL_SIGN_YUI = 0x00016F80;
const char32_t MIAO_VOWEL_SIGN_OG = 0x00016F81;
const char32_t MIAO_VOWEL_SIGN_OER = 0x00016F82;
const char32_t MIAO_VOWEL_SIGN_VW = 0x00016F83;
const char32_t MIAO_VOWEL_SIGN_IG = 0x00016F84;
const char32_t MIAO_VOWEL_SIGN_EA = 0x00016F85;
const char32_t MIAO_VOWEL_SIGN_IONG = 0x00016F86;
const char32_t MIAO_VOWEL_SIGN_UI = 0x00016F87;
const char32_t MIAO_TONE_RIGHT = 0x00016F8F;
const char32_t MIAO_TONE_TOP_RIGHT = 0x00016F90;
const char32_t MIAO_TONE_ABOVE = 0x00016F91;
const char32_t MIAO_TONE_BELOW = 0x00016F92;
const char32_t MIAO_LETTER_TONE_2 = 0x00016F93;
const char32_t MIAO_LETTER_TONE_3 = 0x00016F94;
const char32_t MIAO_LETTER_TONE_4 = 0x00016F95;
const char32_t MIAO_LETTER_TONE_5 = 0x00016F96;
const char32_t MIAO_LETTER_TONE_6 = 0x00016F97;
const char32_t MIAO_LETTER_TONE_7 = 0x00016F98;
const char32_t MIAO_LETTER_TONE_8 = 0x00016F99;
const char32_t MIAO_LETTER_REFORMED_TONE_1 = 0x00016F9A;
const char32_t MIAO_LETTER_REFORMED_TONE_2 = 0x00016F9B;
const char32_t MIAO_LETTER_REFORMED_TONE_4 = 0x00016F9C;
const char32_t MIAO_LETTER_REFORMED_TONE_5 = 0x00016F9D;
const char32_t MIAO_LETTER_REFORMED_TONE_6 = 0x00016F9E;
const char32_t MIAO_LETTER_REFORMED_TONE_8 = 0x00016F9F;
const char32_t TANGUT_ITERATION_MARK = 0x00016FE0;
const char32_t NUSHU_ITERATION_MARK = 0x00016FE1;
const char32_t OLD_CHINESE_HOOK_MARK = 0x00016FE2;
const char32_t OLD_CHINESE_ITERATION_MARK = 0x00016FE3;
const char32_t KHITAN_SMALL_SCRIPT_FILLER = 0x00016FE4;
const char32_t VIETNAMESE_ALTERNATE_READING_MARK_CA = 0x00016FF0;
const char32_t VIETNAMESE_ALTERNATE_READING_MARK_NHAY = 0x00016FF1;
const char32_t CHINESE_SMALL_SIMPLIFIED_ER = 0x00016FF2;
const char32_t CHINESE_SMALL_TRADITIONAL_ER = 0x00016FF3;
const char32_t YANGQIN_SIGN_SLOW_ONE_BEAT = 0x00016FF4;
const char32_t YANGQIN_SIGN_SLOW_THREE_HALF_BEATS = 0x00016FF5;
const char32_t YANGQIN_SIGN_SLOW_TWO_BEATS = 0x00016FF6;
const char32_t TANGUT_IDEOGRAPH_FIRST = 0x00017000;
const char32_t TANGUT_IDEOGRAPH_LAST = 0x000187FF;
const char32_t TANGUT_COMPONENT_001 = 0x00018800;
const char32_t TANGUT_COMPONENT_002 = 0x00018801;
const char32_t TANGUT_COMPONENT_003 = 0x00018802;
const char32_t TANGUT_COMPONENT_004 = 0x00018803;
const char32_t TANGUT_COMPONENT_005 = 0x00018804;
const char32_t TANGUT_COMPONENT_006 = 0x00018805;
const char32_t TANGUT_COMPONENT_007 = 0x00018806;
const char32_t TANGUT_COMPONENT_008 = 0x00018807;
const char32_t TANGUT_COMPONENT_009 = 0x00018808;
const char32_t TANGUT_COMPONENT_010 = 0x00018809;
const char32_t TANGUT_COMPONENT_011 = 0x0001880A;
const char32_t TANGUT_COMPONENT_012 = 0x0001880B;
const char32_t TANGUT_COMPONENT_013 = 0x0001880C;
const char32_t TANGUT_COMPONENT_014 = 0x0001880D;
const char32_t TANGUT_COMPONENT_015 = 0x0001880E;
const char32_t TANGUT_COMPONENT_016 = 0x0001880F;
const char32_t TANGUT_COMPONENT_017 = 0x00018810;
const char32_t TANGUT_COMPONENT_018 = 0x00018811;
const char32_t TANGUT_COMPONENT_019 = 0x00018812;
const char32_t TANGUT_COMPONENT_020 = 0x00018813;
const char32_t TANGUT_COMPONENT_021 = 0x00018814;
const char32_t TANGUT_COMPONENT_022 = 0x00018815;
const char32_t TANGUT_COMPONENT_023 = 0x00018816;
const char32_t TANGUT_COMPONENT_024 = 0x00018817;
const char32_t TANGUT_COMPONENT_025 = 0x00018818;
const char32_t TANGUT_COMPONENT_026 = 0x00018819;
const char32_t TANGUT_COMPONENT_027 = 0x0001881A;
const char32_t TANGUT_COMPONENT_028 = 0x0001881B;
const char32_t TANGUT_COMPONENT_029 = 0x0001881C;
const char32_t TANGUT_COMPONENT_030 = 0x0001881D;
const char32_t TANGUT_COMPONENT_031 = 0x0001881E;
const char32_t TANGUT_COMPONENT_032 = 0x0001881F;
const char32_t TANGUT_COMPONENT_033 = 0x00018820;
const char32_t TANGUT_COMPONENT_034 = 0x00018821;
const char32_t TANGUT_COMPONENT_035 = 0x00018822;
const char32_t TANGUT_COMPONENT_036 = 0x00018823;
const char32_t TANGUT_COMPONENT_037 = 0x00018824;
const char32_t TANGUT_COMPONENT_038 = 0x00018825;
const char32_t TANGUT_COMPONENT_039 = 0x00018826;
const char32_t TANGUT_COMPONENT_040 = 0x00018827;
const char32_t TANGUT_COMPONENT_041 = 0x00018828;
const char32_t TANGUT_COMPONENT_042 = 0x00018829;
const char32_t TANGUT_COMPONENT_043 = 0x0001882A;
const char32_t TANGUT_COMPONENT_044 = 0x0001882B;
const char32_t TANGUT_COMPONENT_045 = 0x0001882C;
const char32_t TANGUT_COMPONENT_046 = 0x0001882D;
const char32_t TANGUT_COMPONENT_047 = 0x0001882E;
const char32_t TANGUT_COMPONENT_048 = 0x0001882F;
const char32_t TANGUT_COMPONENT_049 = 0x00018830;
const char32_t TANGUT_COMPONENT_050 = 0x00018831;
const char32_t TANGUT_COMPONENT_051 = 0x00018832;
const char32_t TANGUT_COMPONENT_052 = 0x00018833;
const char32_t TANGUT_COMPONENT_053 = 0x00018834;
const char32_t TANGUT_COMPONENT_054 = 0x00018835;
const char32_t TANGUT_COMPONENT_055 = 0x00018836;
const char32_t TANGUT_COMPONENT_056 = 0x00018837;
const char32_t TANGUT_COMPONENT_057 = 0x00018838;
const char32_t TANGUT_COMPONENT_058 = 0x00018839;
const char32_t TANGUT_COMPONENT_059 = 0x0001883A;
const char32_t TANGUT_COMPONENT_060 = 0x0001883B;
const char32_t TANGUT_COMPONENT_061 = 0x0001883C;
const char32_t TANGUT_COMPONENT_062 = 0x0001883D;
const char32_t TANGUT_COMPONENT_063 = 0x0001883E;
const char32_t TANGUT_COMPONENT_064 = 0x0001883F;
const char32_t TANGUT_COMPONENT_065 = 0x00018840;
const char32_t TANGUT_COMPONENT_066 = 0x00018841;
const char32_t TANGUT_COMPONENT_067 = 0x00018842;
const char32_t TANGUT_COMPONENT_068 = 0x00018843;
const char32_t TANGUT_COMPONENT_069 = 0x00018844;
const char32_t TANGUT_COMPONENT_070 = 0x00018845;
const char32_t TANGUT_COMPONENT_071 = 0x00018846;
const char32_t TANGUT_COMPONENT_072 = 0x00018847;
const char32_t TANGUT_COMPONENT_073 = 0x00018848;
const char32_t TANGUT_COMPONENT_074 = 0x00018849;
const char32_t TANGUT_COMPONENT_075 = 0x0001884A;
const char32_t TANGUT_COMPONENT_076 = 0x0001884B;
const char32_t TANGUT_COMPONENT_077 = 0x0001884C;
const char32_t TANGUT_COMPONENT_078 = 0x0001884D;
const char32_t TANGUT_COMPONENT_079 = 0x0001884E;
const char32_t TANGUT_COMPONENT_080 = 0x0001884F;
const char32_t TANGUT_COMPONENT_081 = 0x00018850;
const char32_t TANGUT_COMPONENT_082 = 0x00018851;
const char32_t TANGUT_COMPONENT_083 = 0x00018852;
const char32_t TANGUT_COMPONENT_084 = 0x00018853;
const char32_t TANGUT_COMPONENT_085 = 0x00018854;
const char32_t TANGUT_COMPONENT_086 = 0x00018855;
const char32_t TANGUT_COMPONENT_087 = 0x00018856;
const char32_t TANGUT_COMPONENT_088 = 0x00018857;
const char32_t TANGUT_COMPONENT_089 = 0x00018858;
const char32_t TANGUT_COMPONENT_090 = 0x00018859;
const char32_t TANGUT_COMPONENT_091 = 0x0001885A;
const char32_t TANGUT_COMPONENT_092 = 0x0001885B;
const char32_t TANGUT_COMPONENT_093 = 0x0001885C;
const char32_t TANGUT_COMPONENT_094 = 0x0001885D;
const char32_t TANGUT_COMPONENT_095 = 0x0001885E;
const char32_t TANGUT_COMPONENT_096 = 0x0001885F;
const char32_t TANGUT_COMPONENT_097 = 0x00018860;
const char32_t TANGUT_COMPONENT_098 = 0x00018861;
const char32_t TANGUT_COMPONENT_099 = 0x00018862;
const char32_t TANGUT_COMPONENT_100 = 0x00018863;
const char32_t TANGUT_COMPONENT_101 = 0x00018864;
const char32_t TANGUT_COMPONENT_102 = 0x00018865;
const char32_t TANGUT_COMPONENT_103 = 0x00018866;
const char32_t TANGUT_COMPONENT_104 = 0x00018867;
const char32_t TANGUT_COMPONENT_105 = 0x00018868;
const char32_t TANGUT_COMPONENT_106 = 0x00018869;
const char32_t TANGUT_COMPONENT_107 = 0x0001886A;
const char32_t TANGUT_COMPONENT_108 = 0x0001886B;
const char32_t TANGUT_COMPONENT_109 = 0x0001886C;
const char32_t TANGUT_COMPONENT_110 = 0x0001886D;
const char32_t TANGUT_COMPONENT_111 = 0x0001886E;
const char32_t TANGUT_COMPONENT_112 = 0x0001886F;
const char32_t TANGUT_COMPONENT_113 = 0x00018870;
const char32_t TANGUT_COMPONENT_114 = 0x00018871;
const char32_t TANGUT_COMPONENT_115 = 0x00018872;
const char32_t TANGUT_COMPONENT_116 = 0x00018873;
const char32_t TANGUT_COMPONENT_117 = 0x00018874;
const char32_t TANGUT_COMPONENT_118 = 0x00018875;
const char32_t TANGUT_COMPONENT_119 = 0x00018876;
const char32_t TANGUT_COMPONENT_120 = 0x00018877;
const char32_t TANGUT_COMPONENT_121 = 0x00018878;
const char32_t TANGUT_COMPONENT_122 = 0x00018879;
const char32_t TANGUT_COMPONENT_123 = 0x0001887A;
const char32_t TANGUT_COMPONENT_124 = 0x0001887B;
const char32_t TANGUT_COMPONENT_125 = 0x0001887C;
const char32_t TANGUT_COMPONENT_126 = 0x0001887D;
const char32_t TANGUT_COMPONENT_127 = 0x0001887E;
const char32_t TANGUT_COMPONENT_128 = 0x0001887F;
const char32_t TANGUT_COMPONENT_129 = 0x00018880;
const char32_t TANGUT_COMPONENT_130 = 0x00018881;
const char32_t TANGUT_COMPONENT_131 = 0x00018882;
const char32_t TANGUT_COMPONENT_132 = 0x00018883;
const char32_t TANGUT_COMPONENT_133 = 0x00018884;
const char32_t TANGUT_COMPONENT_134 = 0x00018885;
const char32_t TANGUT_COMPONENT_135 = 0x00018886;
const char32_t TANGUT_COMPONENT_136 = 0x00018887;
const char32_t TANGUT_COMPONENT_137 = 0x00018888;
const char32_t TANGUT_COMPONENT_138 = 0x00018889;
const char32_t TANGUT_COMPONENT_139 = 0x0001888A;
const char32_t TANGUT_COMPONENT_140 = 0x0001888B;
const char32_t TANGUT_COMPONENT_141 = 0x0001888C;
const char32_t TANGUT_COMPONENT_142 = 0x0001888D;
const char32_t TANGUT_COMPONENT_143 = 0x0001888E;
const char32_t TANGUT_COMPONENT_144 = 0x0001888F;
const char32_t TANGUT_COMPONENT_145 = 0x00018890;
const char32_t TANGUT_COMPONENT_146 = 0x00018891;
const char32_t TANGUT_COMPONENT_147 = 0x00018892;
const char32_t TANGUT_COMPONENT_148 = 0x00018893;
const char32_t TANGUT_COMPONENT_149 = 0x00018894;
const char32_t TANGUT_COMPONENT_150 = 0x00018895;
const char32_t TANGUT_COMPONENT_151 = 0x00018896;
const char32_t TANGUT_COMPONENT_152 = 0x00018897;
const char32_t TANGUT_COMPONENT_153 = 0x00018898;
const char32_t TANGUT_COMPONENT_154 = 0x00018899;
const char32_t TANGUT_COMPONENT_155 = 0x0001889A;
const char32_t TANGUT_COMPONENT_156 = 0x0001889B;
const char32_t TANGUT_COMPONENT_157 = 0x0001889C;
const char32_t TANGUT_COMPONENT_158 = 0x0001889D;
const char32_t TANGUT_COMPONENT_159 = 0x0001889E;
const char32_t TANGUT_COMPONENT_160 = 0x0001889F;
const char32_t TANGUT_COMPONENT_161 = 0x000188A0;
const char32_t TANGUT_COMPONENT_162 = 0x000188A1;
const char32_t TANGUT_COMPONENT_163 = 0x000188A2;
const char32_t TANGUT_COMPONENT_164 = 0x000188A3;
const char32_t TANGUT_COMPONENT_165 = 0x000188A4;
const char32_t TANGUT_COMPONENT_166 = 0x000188A5;
const char32_t TANGUT_COMPONENT_167 = 0x000188A6;
const char32_t TANGUT_COMPONENT_168 = 0x000188A7;
const char32_t TANGUT_COMPONENT_169 = 0x000188A8;
const char32_t TANGUT_COMPONENT_170 = 0x000188A9;
const char32_t TANGUT_COMPONENT_171 = 0x000188AA;
const char32_t TANGUT_COMPONENT_172 = 0x000188AB;
const char32_t TANGUT_COMPONENT_173 = 0x000188AC;
const char32_t TANGUT_COMPONENT_174 = 0x000188AD;
const char32_t TANGUT_COMPONENT_175 = 0x000188AE;
const char32_t TANGUT_COMPONENT_176 = 0x000188AF;
const char32_t TANGUT_COMPONENT_177 = 0x000188B0;
const char32_t TANGUT_COMPONENT_178 = 0x000188B1;
const char32_t TANGUT_COMPONENT_179 = 0x000188B2;
const char32_t TANGUT_COMPONENT_180 = 0x000188B3;
const char32_t TANGUT_COMPONENT_181 = 0x000188B4;
const char32_t TANGUT_COMPONENT_182 = 0x000188B5;
const char32_t TANGUT_COMPONENT_183 = 0x000188B6;
const char32_t TANGUT_COMPONENT_184 = 0x000188B7;
const char32_t TANGUT_COMPONENT_185 = 0x000188B8;
const char32_t TANGUT_COMPONENT_186 = 0x000188B9;
const char32_t TANGUT_COMPONENT_187 = 0x000188BA;
const char32_t TANGUT_COMPONENT_188 = 0x000188BB;
const char32_t TANGUT_COMPONENT_189 = 0x000188BC;
const char32_t TANGUT_COMPONENT_190 = 0x000188BD;
const char32_t TANGUT_COMPONENT_191 = 0x000188BE;
const char32_t TANGUT_COMPONENT_192 = 0x000188BF;
const char32_t TANGUT_COMPONENT_193 = 0x000188C0;
const char32_t TANGUT_COMPONENT_194 = 0x000188C1;
const char32_t TANGUT_COMPONENT_195 = 0x000188C2;
const char32_t TANGUT_COMPONENT_196 = 0x000188C3;
const char32_t TANGUT_COMPONENT_197 = 0x000188C4;
const char32_t TANGUT_COMPONENT_198 = 0x000188C5;
const char32_t TANGUT_COMPONENT_199 = 0x000188C6;
const char32_t TANGUT_COMPONENT_200 = 0x000188C7;
const char32_t TANGUT_COMPONENT_201 = 0x000188C8;
const char32_t TANGUT_COMPONENT_202 = 0x000188C9;
const char32_t TANGUT_COMPONENT_203 = 0x000188CA;
const char32_t TANGUT_COMPONENT_204 = 0x000188CB;
const char32_t TANGUT_COMPONENT_205 = 0x000188CC;
const char32_t TANGUT_COMPONENT_206 = 0x000188CD;
const char32_t TANGUT_COMPONENT_207 = 0x000188CE;
const char32_t TANGUT_COMPONENT_208 = 0x000188CF;
const char32_t TANGUT_COMPONENT_209 = 0x000188D0;
const char32_t TANGUT_COMPONENT_210 = 0x000188D1;
const char32_t TANGUT_COMPONENT_211 = 0x000188D2;
const char32_t TANGUT_COMPONENT_212 = 0x000188D3;
const char32_t TANGUT_COMPONENT_213 = 0x000188D4;
const char32_t TANGUT_COMPONENT_214 = 0x000188D5;
const char32_t TANGUT_COMPONENT_215 = 0x000188D6;
const char32_t TANGUT_COMPONENT_216 = 0x000188D7;
const char32_t TANGUT_COMPONENT_217 = 0x000188D8;
const char32_t TANGUT_COMPONENT_218 = 0x000188D9;
const char32_t TANGUT_COMPONENT_219 = 0x000188DA;
const char32_t TANGUT_COMPONENT_220 = 0x000188DB;
const char32_t TANGUT_COMPONENT_221 = 0x000188DC;
const char32_t TANGUT_COMPONENT_222 = 0x000188DD;
const char32_t TANGUT_COMPONENT_223 = 0x000188DE;
const char32_t TANGUT_COMPONENT_224 = 0x000188DF;
const char32_t TANGUT_COMPONENT_225 = 0x000188E0;
const char32_t TANGUT_COMPONENT_226 = 0x000188E1;
const char32_t TANGUT_COMPONENT_227 = 0x000188E2;
const char32_t TANGUT_COMPONENT_228 = 0x000188E3;
const char32_t TANGUT_COMPONENT_229 = 0x000188E4;
const char32_t TANGUT_COMPONENT_230 = 0x000188E5;
const char32_t TANGUT_COMPONENT_231 = 0x000188E6;
const char32_t TANGUT_COMPONENT_232 = 0x000188E7;
const char32_t TANGUT_COMPONENT_233 = 0x000188E8;
const char32_t TANGUT_COMPONENT_234 = 0x000188E9;
const char32_t TANGUT_COMPONENT_235 = 0x000188EA;
const char32_t TANGUT_COMPONENT_236 = 0x000188EB;
const char32_t TANGUT_COMPONENT_237 = 0x000188EC;
const char32_t TANGUT_COMPONENT_238 = 0x000188ED;
const char32_t TANGUT_COMPONENT_239 = 0x000188EE;
const char32_t TANGUT_COMPONENT_240 = 0x000188EF;
const char32_t TANGUT_COMPONENT_241 = 0x000188F0;
const char32_t TANGUT_COMPONENT_242 = 0x000188F1;
const char32_t TANGUT_COMPONENT_243 = 0x000188F2;
const char32_t TANGUT_COMPONENT_244 = 0x000188F3;
const char32_t TANGUT_COMPONENT_245 = 0x000188F4;
const char32_t TANGUT_COMPONENT_246 = 0x000188F5;
const char32_t TANGUT_COMPONENT_247 = 0x000188F6;
const char32_t TANGUT_COMPONENT_248 = 0x000188F7;
const char32_t TANGUT_COMPONENT_249 = 0x000188F8;
const char32_t TANGUT_COMPONENT_250 = 0x000188F9;
const char32_t TANGUT_COMPONENT_251 = 0x000188FA;
const char32_t TANGUT_COMPONENT_252 = 0x000188FB;
const char32_t TANGUT_COMPONENT_253 = 0x000188FC;
const char32_t TANGUT_COMPONENT_254 = 0x000188FD;
const char32_t TANGUT_COMPONENT_255 = 0x000188FE;
const char32_t TANGUT_COMPONENT_256 = 0x000188FF;
const char32_t TANGUT_COMPONENT_257 = 0x00018900;
const char32_t TANGUT_COMPONENT_258 = 0x00018901;
const char32_t TANGUT_COMPONENT_259 = 0x00018902;
const char32_t TANGUT_COMPONENT_260 = 0x00018903;
const char32_t TANGUT_COMPONENT_261 = 0x00018904;
const char32_t TANGUT_COMPONENT_262 = 0x00018905;
const char32_t TANGUT_COMPONENT_263 = 0x00018906;
const char32_t TANGUT_COMPONENT_264 = 0x00018907;
const char32_t TANGUT_COMPONENT_265 = 0x00018908;
const char32_t TANGUT_COMPONENT_266 = 0x00018909;
const char32_t TANGUT_COMPONENT_267 = 0x0001890A;
const char32_t TANGUT_COMPONENT_268 = 0x0001890B;
const char32_t TANGUT_COMPONENT_269 = 0x0001890C;
const char32_t TANGUT_COMPONENT_270 = 0x0001890D;
const char32_t TANGUT_COMPONENT_271 = 0x0001890E;
const char32_t TANGUT_COMPONENT_272 = 0x0001890F;
const char32_t TANGUT_COMPONENT_273 = 0x00018910;
const char32_t TANGUT_COMPONENT_274 = 0x00018911;
const char32_t TANGUT_COMPONENT_275 = 0x00018912;
const char32_t TANGUT_COMPONENT_276 = 0x00018913;
const char32_t TANGUT_COMPONENT_277 = 0x00018914;
const char32_t TANGUT_COMPONENT_278 = 0x00018915;
const char32_t TANGUT_COMPONENT_279 = 0x00018916;
const char32_t TANGUT_COMPONENT_280 = 0x00018917;
const char32_t TANGUT_COMPONENT_281 = 0x00018918;
const char32_t TANGUT_COMPONENT_282 = 0x00018919;
const char32_t TANGUT_COMPONENT_283 = 0x0001891A;
const char32_t TANGUT_COMPONENT_284 = 0x0001891B;
const char32_t TANGUT_COMPONENT_285 = 0x0001891C;
const char32_t TANGUT_COMPONENT_286 = 0x0001891D;
const char32_t TANGUT_COMPONENT_287 = 0x0001891E;
const char32_t TANGUT_COMPONENT_288 = 0x0001891F;
const char32_t TANGUT_COMPONENT_289 = 0x00018920;
const char32_t TANGUT_COMPONENT_290 = 0x00018921;
const char32_t TANGUT_COMPONENT_291 = 0x00018922;
const char32_t TANGUT_COMPONENT_292 = 0x00018923;
const char32_t TANGUT_COMPONENT_293 = 0x00018924;
const char32_t TANGUT_COMPONENT_294 = 0x00018925;
const char32_t TANGUT_COMPONENT_295 = 0x00018926;
const char32_t TANGUT_COMPONENT_296 = 0x00018927;
const char32_t TANGUT_COMPONENT_297 = 0x00018928;
const char32_t TANGUT_COMPONENT_298 = 0x00018929;
const char32_t TANGUT_COMPONENT_299 = 0x0001892A;
const char32_t TANGUT_COMPONENT_300 = 0x0001892B;
const char32_t TANGUT_COMPONENT_301 = 0x0001892C;
const char32_t TANGUT_COMPONENT_302 = 0x0001892D;
const char32_t TANGUT_COMPONENT_303 = 0x0001892E;
const char32_t TANGUT_COMPONENT_304 = 0x0001892F;
const char32_t TANGUT_COMPONENT_305 = 0x00018930;
const char32_t TANGUT_COMPONENT_306 = 0x00018931;
const char32_t TANGUT_COMPONENT_307 = 0x00018932;
const char32_t TANGUT_COMPONENT_308 = 0x00018933;
const char32_t TANGUT_COMPONENT_309 = 0x00018934;
const char32_t TANGUT_COMPONENT_310 = 0x00018935;
const char32_t TANGUT_COMPONENT_311 = 0x00018936;
const char32_t TANGUT_COMPONENT_312 = 0x00018937;
const char32_t TANGUT_COMPONENT_313 = 0x00018938;
const char32_t TANGUT_COMPONENT_314 = 0x00018939;
const char32_t TANGUT_COMPONENT_315 = 0x0001893A;
const char32_t TANGUT_COMPONENT_316 = 0x0001893B;
const char32_t TANGUT_COMPONENT_317 = 0x0001893C;
const char32_t TANGUT_COMPONENT_318 = 0x0001893D;
const char32_t TANGUT_COMPONENT_319 = 0x0001893E;
const char32_t TANGUT_COMPONENT_320 = 0x0001893F;
const char32_t TANGUT_COMPONENT_321 = 0x00018940;
const char32_t TANGUT_COMPONENT_322 = 0x00018941;
const char32_t TANGUT_COMPONENT_323 = 0x00018942;
const char32_t TANGUT_COMPONENT_324 = 0x00018943;
const char32_t TANGUT_COMPONENT_325 = 0x00018944;
const char32_t TANGUT_COMPONENT_326 = 0x00018945;
const char32_t TANGUT_COMPONENT_327 = 0x00018946;
const char32_t TANGUT_COMPONENT_328 = 0x00018947;
const char32_t TANGUT_COMPONENT_329 = 0x00018948;
const char32_t TANGUT_COMPONENT_330 = 0x00018949;
const char32_t TANGUT_COMPONENT_331 = 0x0001894A;
const char32_t TANGUT_COMPONENT_332 = 0x0001894B;
const char32_t TANGUT_COMPONENT_333 = 0x0001894C;
const char32_t TANGUT_COMPONENT_334 = 0x0001894D;
const char32_t TANGUT_COMPONENT_335 = 0x0001894E;
const char32_t TANGUT_COMPONENT_336 = 0x0001894F;
const char32_t TANGUT_COMPONENT_337 = 0x00018950;
const char32_t TANGUT_COMPONENT_338 = 0x00018951;
const char32_t TANGUT_COMPONENT_339 = 0x00018952;
const char32_t TANGUT_COMPONENT_340 = 0x00018953;
const char32_t TANGUT_COMPONENT_341 = 0x00018954;
const char32_t TANGUT_COMPONENT_342 = 0x00018955;
const char32_t TANGUT_COMPONENT_343 = 0x00018956;
const char32_t TANGUT_COMPONENT_344 = 0x00018957;
const char32_t TANGUT_COMPONENT_345 = 0x00018958;
const char32_t TANGUT_COMPONENT_346 = 0x00018959;
const char32_t TANGUT_COMPONENT_347 = 0x0001895A;
const char32_t TANGUT_COMPONENT_348 = 0x0001895B;
const char32_t TANGUT_COMPONENT_349 = 0x0001895C;
const char32_t TANGUT_COMPONENT_350 = 0x0001895D;
const char32_t TANGUT_COMPONENT_351 = 0x0001895E;
const char32_t TANGUT_COMPONENT_352 = 0x0001895F;
const char32_t TANGUT_COMPONENT_353 = 0x00018960;
const char32_t TANGUT_COMPONENT_354 = 0x00018961;
const char32_t TANGUT_COMPONENT_355 = 0x00018962;
const char32_t TANGUT_COMPONENT_356 = 0x00018963;
const char32_t TANGUT_COMPONENT_357 = 0x00018964;
const char32_t TANGUT_COMPONENT_358 = 0x00018965;
const char32_t TANGUT_COMPONENT_359 = 0x00018966;
const char32_t TANGUT_COMPONENT_360 = 0x00018967;
const char32_t TANGUT_COMPONENT_361 = 0x00018968;
const char32_t TANGUT_COMPONENT_362 = 0x00018969;
const char32_t TANGUT_COMPONENT_363 = 0x0001896A;
const char32_t TANGUT_COMPONENT_364 = 0x0001896B;
const char32_t TANGUT_COMPONENT_365 = 0x0001896C;
const char32_t TANGUT_COMPONENT_366 = 0x0001896D;
const char32_t TANGUT_COMPONENT_367 = 0x0001896E;
const char32_t TANGUT_COMPONENT_368 = 0x0001896F;
const char32_t TANGUT_COMPONENT_369 = 0x00018970;
const char32_t TANGUT_COMPONENT_370 = 0x00018971;
const char32_t TANGUT_COMPONENT_371 = 0x00018972;
const char32_t TANGUT_COMPONENT_372 = 0x00018973;
const char32_t TANGUT_COMPONENT_373 = 0x00018974;
const char32_t TANGUT_COMPONENT_374 = 0x00018975;
const char32_t TANGUT_COMPONENT_375 = 0x00018976;
const char32_t TANGUT_COMPONENT_376 = 0x00018977;
const char32_t TANGUT_COMPONENT_377 = 0x00018978;
const char32_t TANGUT_COMPONENT_378 = 0x00018979;
const char32_t TANGUT_COMPONENT_379 = 0x0001897A;
const char32_t TANGUT_COMPONENT_380 = 0x0001897B;
const char32_t TANGUT_COMPONENT_381 = 0x0001897C;
const char32_t TANGUT_COMPONENT_382 = 0x0001897D;
const char32_t TANGUT_COMPONENT_383 = 0x0001897E;
const char32_t TANGUT_COMPONENT_384 = 0x0001897F;
const char32_t TANGUT_COMPONENT_385 = 0x00018980;
const char32_t TANGUT_COMPONENT_386 = 0x00018981;
const char32_t TANGUT_COMPONENT_387 = 0x00018982;
const char32_t TANGUT_COMPONENT_388 = 0x00018983;
const char32_t TANGUT_COMPONENT_389 = 0x00018984;
const char32_t TANGUT_COMPONENT_390 = 0x00018985;
const char32_t TANGUT_COMPONENT_391 = 0x00018986;
const char32_t TANGUT_COMPONENT_392 = 0x00018987;
const char32_t TANGUT_COMPONENT_393 = 0x00018988;
const char32_t TANGUT_COMPONENT_394 = 0x00018989;
const char32_t TANGUT_COMPONENT_395 = 0x0001898A;
const char32_t TANGUT_COMPONENT_396 = 0x0001898B;
const char32_t TANGUT_COMPONENT_397 = 0x0001898C;
const char32_t TANGUT_COMPONENT_398 = 0x0001898D;
const char32_t TANGUT_COMPONENT_399 = 0x0001898E;
const char32_t TANGUT_COMPONENT_400 = 0x0001898F;
const char32_t TANGUT_COMPONENT_401 = 0x00018990;
const char32_t TANGUT_COMPONENT_402 = 0x00018991;
const char32_t TANGUT_COMPONENT_403 = 0x00018992;
const char32_t TANGUT_COMPONENT_404 = 0x00018993;
const char32_t TANGUT_COMPONENT_405 = 0x00018994;
const char32_t TANGUT_COMPONENT_406 = 0x00018995;
const char32_t TANGUT_COMPONENT_407 = 0x00018996;
const char32_t TANGUT_COMPONENT_408 = 0x00018997;
const char32_t TANGUT_COMPONENT_409 = 0x00018998;
const char32_t TANGUT_COMPONENT_410 = 0x00018999;
const char32_t TANGUT_COMPONENT_411 = 0x0001899A;
const char32_t TANGUT_COMPONENT_412 = 0x0001899B;
const char32_t TANGUT_COMPONENT_413 = 0x0001899C;
const char32_t TANGUT_COMPONENT_414 = 0x0001899D;
const char32_t TANGUT_COMPONENT_415 = 0x0001899E;
const char32_t TANGUT_COMPONENT_416 = 0x0001899F;
const char32_t TANGUT_COMPONENT_417 = 0x000189A0;
const char32_t TANGUT_COMPONENT_418 = 0x000189A1;
const char32_t TANGUT_COMPONENT_419 = 0x000189A2;
const char32_t TANGUT_COMPONENT_420 = 0x000189A3;
const char32_t TANGUT_COMPONENT_421 = 0x000189A4;
const char32_t TANGUT_COMPONENT_422 = 0x000189A5;
const char32_t TANGUT_COMPONENT_423 = 0x000189A6;
const char32_t TANGUT_COMPONENT_424 = 0x000189A7;
const char32_t TANGUT_COMPONENT_425 = 0x000189A8;
const char32_t TANGUT_COMPONENT_426 = 0x000189A9;
const char32_t TANGUT_COMPONENT_427 = 0x000189AA;
const char32_t TANGUT_COMPONENT_428 = 0x000189AB;
const char32_t TANGUT_COMPONENT_429 = 0x000189AC;
const char32_t TANGUT_COMPONENT_430 = 0x000189AD;
const char32_t TANGUT_COMPONENT_431 = 0x000189AE;
const char32_t TANGUT_COMPONENT_432 = 0x000189AF;
const char32_t TANGUT_COMPONENT_433 = 0x000189B0;
const char32_t TANGUT_COMPONENT_434 = 0x000189B1;
const char32_t TANGUT_COMPONENT_435 = 0x000189B2;
const char32_t TANGUT_COMPONENT_436 = 0x000189B3;
const char32_t TANGUT_COMPONENT_437 = 0x000189B4;
const char32_t TANGUT_COMPONENT_438 = 0x000189B5;
const char32_t TANGUT_COMPONENT_439 = 0x000189B6;
const char32_t TANGUT_COMPONENT_440 = 0x000189B7;
const char32_t TANGUT_COMPONENT_441 = 0x000189B8;
const char32_t TANGUT_COMPONENT_442 = 0x000189B9;
const char32_t TANGUT_COMPONENT_443 = 0x000189BA;
const char32_t TANGUT_COMPONENT_444 = 0x000189BB;
const char32_t TANGUT_COMPONENT_445 = 0x000189BC;
const char32_t TANGUT_COMPONENT_446 = 0x000189BD;
const char32_t TANGUT_COMPONENT_447 = 0x000189BE;
const char32_t TANGUT_COMPONENT_448 = 0x000189BF;
const char32_t TANGUT_COMPONENT_449 = 0x000189C0;
const char32_t TANGUT_COMPONENT_450 = 0x000189C1;
const char32_t TANGUT_COMPONENT_451 = 0x000189C2;
const char32_t TANGUT_COMPONENT_452 = 0x000189C3;
const char32_t TANGUT_COMPONENT_453 = 0x000189C4;
const char32_t TANGUT_COMPONENT_454 = 0x000189C5;
const char32_t TANGUT_COMPONENT_455 = 0x000189C6;
const char32_t TANGUT_COMPONENT_456 = 0x000189C7;
const char32_t TANGUT_COMPONENT_457 = 0x000189C8;
const char32_t TANGUT_COMPONENT_458 = 0x000189C9;
const char32_t TANGUT_COMPONENT_459 = 0x000189CA;
const char32_t TANGUT_COMPONENT_460 = 0x000189CB;
const char32_t TANGUT_COMPONENT_461 = 0x000189CC;
const char32_t TANGUT_COMPONENT_462 = 0x000189CD;
const char32_t TANGUT_COMPONENT_463 = 0x000189CE;
const char32_t TANGUT_COMPONENT_464 = 0x000189CF;
const char32_t TANGUT_COMPONENT_465 = 0x000189D0;
const char32_t TANGUT_COMPONENT_466 = 0x000189D1;
const char32_t TANGUT_COMPONENT_467 = 0x000189D2;
const char32_t TANGUT_COMPONENT_468 = 0x000189D3;
const char32_t TANGUT_COMPONENT_469 = 0x000189D4;
const char32_t TANGUT_COMPONENT_470 = 0x000189D5;
const char32_t TANGUT_COMPONENT_471 = 0x000189D6;
const char32_t TANGUT_COMPONENT_472 = 0x000189D7;
const char32_t TANGUT_COMPONENT_473 = 0x000189D8;
const char32_t TANGUT_COMPONENT_474 = 0x000189D9;
const char32_t TANGUT_COMPONENT_475 = 0x000189DA;
const char32_t TANGUT_COMPONENT_476 = 0x000189DB;
const char32_t TANGUT_COMPONENT_477 = 0x000189DC;
const char32_t TANGUT_COMPONENT_478 = 0x000189DD;
const char32_t TANGUT_COMPONENT_479 = 0x000189DE;
const char32_t TANGUT_COMPONENT_480 = 0x000189DF;
const char32_t TANGUT_COMPONENT_481 = 0x000189E0;
const char32_t TANGUT_COMPONENT_482 = 0x000189E1;
const char32_t TANGUT_COMPONENT_483 = 0x000189E2;
const char32_t TANGUT_COMPONENT_484 = 0x000189E3;
const char32_t TANGUT_COMPONENT_485 = 0x000189E4;
const char32_t TANGUT_COMPONENT_486 = 0x000189E5;
const char32_t TANGUT_COMPONENT_487 = 0x000189E6;
const char32_t TANGUT_COMPONENT_488 = 0x000189E7;
const char32_t TANGUT_COMPONENT_489 = 0x000189E8;
const char32_t TANGUT_COMPONENT_490 = 0x000189E9;
const char32_t TANGUT_COMPONENT_491 = 0x000189EA;
const char32_t TANGUT_COMPONENT_492 = 0x000189EB;
const char32_t TANGUT_COMPONENT_493 = 0x000189EC;
const char32_t TANGUT_COMPONENT_494 = 0x000189ED;
const char32_t TANGUT_COMPONENT_495 = 0x000189EE;
const char32_t TANGUT_COMPONENT_496 = 0x000189EF;
const char32_t TANGUT_COMPONENT_497 = 0x000189F0;
const char32_t TANGUT_COMPONENT_498 = 0x000189F1;
const char32_t TANGUT_COMPONENT_499 = 0x000189F2;
const char32_t TANGUT_COMPONENT_500 = 0x000189F3;
const char32_t TANGUT_COMPONENT_501 = 0x000189F4;
const char32_t TANGUT_COMPONENT_502 = 0x000189F5;
const char32_t TANGUT_COMPONENT_503 = 0x000189F6;
const char32_t TANGUT_COMPONENT_504 = 0x000189F7;
const char32_t TANGUT_COMPONENT_505 = 0x000189F8;
const char32_t TANGUT_COMPONENT_506 = 0x000189F9;
const char32_t TANGUT_COMPONENT_507 = 0x000189FA;
const char32_t TANGUT_COMPONENT_508 = 0x000189FB;
const char32_t TANGUT_COMPONENT_509 = 0x000189FC;
const char32_t TANGUT_COMPONENT_510 = 0x000189FD;
const char32_t TANGUT_COMPONENT_511 = 0x000189FE;
const char32_t TANGUT_COMPONENT_512 = 0x000189FF;
const char32_t TANGUT_COMPONENT_513 = 0x00018A00;
const char32_t TANGUT_COMPONENT_514 = 0x00018A01;
const char32_t TANGUT_COMPONENT_515 = 0x00018A02;
const char32_t TANGUT_COMPONENT_516 = 0x00018A03;
const char32_t TANGUT_COMPONENT_517 = 0x00018A04;
const char32_t TANGUT_COMPONENT_518 = 0x00018A05;
const char32_t TANGUT_COMPONENT_519 = 0x00018A06;
const char32_t TANGUT_COMPONENT_520 = 0x00018A07;
const char32_t TANGUT_COMPONENT_521 = 0x00018A08;
const char32_t TANGUT_COMPONENT_522 = 0x00018A09;
const char32_t TANGUT_COMPONENT_523 = 0x00018A0A;
const char32_t TANGUT_COMPONENT_524 = 0x00018A0B;
const char32_t TANGUT_COMPONENT_525 = 0x00018A0C;
const char32_t TANGUT_COMPONENT_526 = 0x00018A0D;
const char32_t TANGUT_COMPONENT_527 = 0x00018A0E;
const char32_t TANGUT_COMPONENT_528 = 0x00018A0F;
const char32_t TANGUT_COMPONENT_529 = 0x00018A10;
const char32_t TANGUT_COMPONENT_530 = 0x00018A11;
const char32_t TANGUT_COMPONENT_531 = 0x00018A12;
const char32_t TANGUT_COMPONENT_532 = 0x00018A13;
const char32_t TANGUT_COMPONENT_533 = 0x00018A14;
const char32_t TANGUT_COMPONENT_534 = 0x00018A15;
const char32_t TANGUT_COMPONENT_535 = 0x00018A16;
const char32_t TANGUT_COMPONENT_536 = 0x00018A17;
const char32_t TANGUT_COMPONENT_537 = 0x00018A18;
const char32_t TANGUT_COMPONENT_538 = 0x00018A19;
const char32_t TANGUT_COMPONENT_539 = 0x00018A1A;
const char32_t TANGUT_COMPONENT_540 = 0x00018A1B;
const char32_t TANGUT_COMPONENT_541 = 0x00018A1C;
const char32_t TANGUT_COMPONENT_542 = 0x00018A1D;
const char32_t TANGUT_COMPONENT_543 = 0x00018A1E;
const char32_t TANGUT_COMPONENT_544 = 0x00018A1F;
const char32_t TANGUT_COMPONENT_545 = 0x00018A20;
const char32_t TANGUT_COMPONENT_546 = 0x00018A21;
const char32_t TANGUT_COMPONENT_547 = 0x00018A22;
const char32_t TANGUT_COMPONENT_548 = 0x00018A23;
const char32_t TANGUT_COMPONENT_549 = 0x00018A24;
const char32_t TANGUT_COMPONENT_550 = 0x00018A25;
const char32_t TANGUT_COMPONENT_551 = 0x00018A26;
const char32_t TANGUT_COMPONENT_552 = 0x00018A27;
const char32_t TANGUT_COMPONENT_553 = 0x00018A28;
const char32_t TANGUT_COMPONENT_554 = 0x00018A29;
const char32_t TANGUT_COMPONENT_555 = 0x00018A2A;
const char32_t TANGUT_COMPONENT_556 = 0x00018A2B;
const char32_t TANGUT_COMPONENT_557 = 0x00018A2C;
const char32_t TANGUT_COMPONENT_558 = 0x00018A2D;
const char32_t TANGUT_COMPONENT_559 = 0x00018A2E;
const char32_t TANGUT_COMPONENT_560 = 0x00018A2F;
const char32_t TANGUT_COMPONENT_561 = 0x00018A30;
const char32_t TANGUT_COMPONENT_562 = 0x00018A31;
const char32_t TANGUT_COMPONENT_563 = 0x00018A32;
const char32_t TANGUT_COMPONENT_564 = 0x00018A33;
const char32_t TANGUT_COMPONENT_565 = 0x00018A34;
const char32_t TANGUT_COMPONENT_566 = 0x00018A35;
const char32_t TANGUT_COMPONENT_567 = 0x00018A36;
const char32_t TANGUT_COMPONENT_568 = 0x00018A37;
const char32_t TANGUT_COMPONENT_569 = 0x00018A38;
const char32_t TANGUT_COMPONENT_570 = 0x00018A39;
const char32_t TANGUT_COMPONENT_571 = 0x00018A3A;
const char32_t TANGUT_COMPONENT_572 = 0x00018A3B;
const char32_t TANGUT_COMPONENT_573 = 0x00018A3C;
const char32_t TANGUT_COMPONENT_574 = 0x00018A3D;
const char32_t TANGUT_COMPONENT_575 = 0x00018A3E;
const char32_t TANGUT_COMPONENT_576 = 0x00018A3F;
const char32_t TANGUT_COMPONENT_577 = 0x00018A40;
const char32_t TANGUT_COMPONENT_578 = 0x00018A41;
const char32_t TANGUT_COMPONENT_579 = 0x00018A42;
const char32_t TANGUT_COMPONENT_580 = 0x00018A43;
const char32_t TANGUT_COMPONENT_581 = 0x00018A44;
const char32_t TANGUT_COMPONENT_582 = 0x00018A45;
const char32_t TANGUT_COMPONENT_583 = 0x00018A46;
const char32_t TANGUT_COMPONENT_584 = 0x00018A47;
const char32_t TANGUT_COMPONENT_585 = 0x00018A48;
const char32_t TANGUT_COMPONENT_586 = 0x00018A49;
const char32_t TANGUT_COMPONENT_587 = 0x00018A4A;
const char32_t TANGUT_COMPONENT_588 = 0x00018A4B;
const char32_t TANGUT_COMPONENT_589 = 0x00018A4C;
const char32_t TANGUT_COMPONENT_590 = 0x00018A4D;
const char32_t TANGUT_COMPONENT_591 = 0x00018A4E;
const char32_t TANGUT_COMPONENT_592 = 0x00018A4F;
const char32_t TANGUT_COMPONENT_593 = 0x00018A50;
const char32_t TANGUT_COMPONENT_594 = 0x00018A51;
const char32_t TANGUT_COMPONENT_595 = 0x00018A52;
const char32_t TANGUT_COMPONENT_596 = 0x00018A53;
const char32_t TANGUT_COMPONENT_597 = 0x00018A54;
const char32_t TANGUT_COMPONENT_598 = 0x00018A55;
const char32_t TANGUT_COMPONENT_599 = 0x00018A56;
const char32_t TANGUT_COMPONENT_600 = 0x00018A57;
const char32_t TANGUT_COMPONENT_601 = 0x00018A58;
const char32_t TANGUT_COMPONENT_602 = 0x00018A59;
const char32_t TANGUT_COMPONENT_603 = 0x00018A5A;
const char32_t TANGUT_COMPONENT_604 = 0x00018A5B;
const char32_t TANGUT_COMPONENT_605 = 0x00018A5C;
const char32_t TANGUT_COMPONENT_606 = 0x00018A5D;
const char32_t TANGUT_COMPONENT_607 = 0x00018A5E;
const char32_t TANGUT_COMPONENT_608 = 0x00018A5F;
const char32_t TANGUT_COMPONENT_609 = 0x00018A60;
const char32_t TANGUT_COMPONENT_610 = 0x00018A61;
const char32_t TANGUT_COMPONENT_611 = 0x00018A62;
const char32_t TANGUT_COMPONENT_612 = 0x00018A63;
const char32_t TANGUT_COMPONENT_613 = 0x00018A64;
const char32_t TANGUT_COMPONENT_614 = 0x00018A65;
const char32_t TANGUT_COMPONENT_615 = 0x00018A66;
const char32_t TANGUT_COMPONENT_616 = 0x00018A67;
const char32_t TANGUT_COMPONENT_617 = 0x00018A68;
const char32_t TANGUT_COMPONENT_618 = 0x00018A69;
const char32_t TANGUT_COMPONENT_619 = 0x00018A6A;
const char32_t TANGUT_COMPONENT_620 = 0x00018A6B;
const char32_t TANGUT_COMPONENT_621 = 0x00018A6C;
const char32_t TANGUT_COMPONENT_622 = 0x00018A6D;
const char32_t TANGUT_COMPONENT_623 = 0x00018A6E;
const char32_t TANGUT_COMPONENT_624 = 0x00018A6F;
const char32_t TANGUT_COMPONENT_625 = 0x00018A70;
const char32_t TANGUT_COMPONENT_626 = 0x00018A71;
const char32_t TANGUT_COMPONENT_627 = 0x00018A72;
const char32_t TANGUT_COMPONENT_628 = 0x00018A73;
const char32_t TANGUT_COMPONENT_629 = 0x00018A74;
const char32_t TANGUT_COMPONENT_630 = 0x00018A75;
const char32_t TANGUT_COMPONENT_631 = 0x00018A76;
const char32_t TANGUT_COMPONENT_632 = 0x00018A77;
const char32_t TANGUT_COMPONENT_633 = 0x00018A78;
const char32_t TANGUT_COMPONENT_634 = 0x00018A79;
const char32_t TANGUT_COMPONENT_635 = 0x00018A7A;
const char32_t TANGUT_COMPONENT_636 = 0x00018A7B;
const char32_t TANGUT_COMPONENT_637 = 0x00018A7C;
const char32_t TANGUT_COMPONENT_638 = 0x00018A7D;
const char32_t TANGUT_COMPONENT_639 = 0x00018A7E;
const char32_t TANGUT_COMPONENT_640 = 0x00018A7F;
const char32_t TANGUT_COMPONENT_641 = 0x00018A80;
const char32_t TANGUT_COMPONENT_642 = 0x00018A81;
const char32_t TANGUT_COMPONENT_643 = 0x00018A82;
const char32_t TANGUT_COMPONENT_644 = 0x00018A83;
const char32_t TANGUT_COMPONENT_645 = 0x00018A84;
const char32_t TANGUT_COMPONENT_646 = 0x00018A85;
const char32_t TANGUT_COMPONENT_647 = 0x00018A86;
const char32_t TANGUT_COMPONENT_648 = 0x00018A87;
const char32_t TANGUT_COMPONENT_649 = 0x00018A88;
const char32_t TANGUT_COMPONENT_650 = 0x00018A89;
const char32_t TANGUT_COMPONENT_651 = 0x00018A8A;
const char32_t TANGUT_COMPONENT_652 = 0x00018A8B;
const char32_t TANGUT_COMPONENT_653 = 0x00018A8C;
const char32_t TANGUT_COMPONENT_654 = 0x00018A8D;
const char32_t TANGUT_COMPONENT_655 = 0x00018A8E;
const char32_t TANGUT_COMPONENT_656 = 0x00018A8F;
const char32_t TANGUT_COMPONENT_657 = 0x00018A90;
const char32_t TANGUT_COMPONENT_658 = 0x00018A91;
const char32_t TANGUT_COMPONENT_659 = 0x00018A92;
const char32_t TANGUT_COMPONENT_660 = 0x00018A93;
const char32_t TANGUT_COMPONENT_661 = 0x00018A94;
const char32_t TANGUT_COMPONENT_662 = 0x00018A95;
const char32_t TANGUT_COMPONENT_663 = 0x00018A96;
const char32_t TANGUT_COMPONENT_664 = 0x00018A97;
const char32_t TANGUT_COMPONENT_665 = 0x00018A98;
const char32_t TANGUT_COMPONENT_666 = 0x00018A99;
const char32_t TANGUT_COMPONENT_667 = 0x00018A9A;
const char32_t TANGUT_COMPONENT_668 = 0x00018A9B;
const char32_t TANGUT_COMPONENT_669 = 0x00018A9C;
const char32_t TANGUT_COMPONENT_670 = 0x00018A9D;
const char32_t TANGUT_COMPONENT_671 = 0x00018A9E;
const char32_t TANGUT_COMPONENT_672 = 0x00018A9F;
const char32_t TANGUT_COMPONENT_673 = 0x00018AA0;
const char32_t TANGUT_COMPONENT_674 = 0x00018AA1;
const char32_t TANGUT_COMPONENT_675 = 0x00018AA2;
const char32_t TANGUT_COMPONENT_676 = 0x00018AA3;
const char32_t TANGUT_COMPONENT_677 = 0x00018AA4;
const char32_t TANGUT_COMPONENT_678 = 0x00018AA5;
const char32_t TANGUT_COMPONENT_679 = 0x00018AA6;
const char32_t TANGUT_COMPONENT_680 = 0x00018AA7;
const char32_t TANGUT_COMPONENT_681 = 0x00018AA8;
const char32_t TANGUT_COMPONENT_682 = 0x00018AA9;
const char32_t TANGUT_COMPONENT_683 = 0x00018AAA;
const char32_t TANGUT_COMPONENT_684 = 0x00018AAB;
const char32_t TANGUT_COMPONENT_685 = 0x00018AAC;
const char32_t TANGUT_COMPONENT_686 = 0x00018AAD;
const char32_t TANGUT_COMPONENT_687 = 0x00018AAE;
const char32_t TANGUT_COMPONENT_688 = 0x00018AAF;
const char32_t TANGUT_COMPONENT_689 = 0x00018AB0;
const char32_t TANGUT_COMPONENT_690 = 0x00018AB1;
const char32_t TANGUT_COMPONENT_691 = 0x00018AB2;
const char32_t TANGUT_COMPONENT_692 = 0x00018AB3;
const char32_t TANGUT_COMPONENT_693 = 0x00018AB4;
const char32_t TANGUT_COMPONENT_694 = 0x00018AB5;
const char32_t TANGUT_COMPONENT_695 = 0x00018AB6;
const char32_t TANGUT_COMPONENT_696 = 0x00018AB7;
const char32_t TANGUT_COMPONENT_697 = 0x00018AB8;
const char32_t TANGUT_COMPONENT_698 = 0x00018AB9;
const char32_t TANGUT_COMPONENT_699 = 0x00018ABA;
const char32_t TANGUT_COMPONENT_700 = 0x00018ABB;
const char32_t TANGUT_COMPONENT_701 = 0x00018ABC;
const char32_t TANGUT_COMPONENT_702 = 0x00018ABD;
const char32_t TANGUT_COMPONENT_703 = 0x00018ABE;
const char32_t TANGUT_COMPONENT_704 = 0x00018ABF;
const char32_t TANGUT_COMPONENT_705 = 0x00018AC0;
const char32_t TANGUT_COMPONENT_706 = 0x00018AC1;
const char32_t TANGUT_COMPONENT_707 = 0x00018AC2;
const char32_t TANGUT_COMPONENT_708 = 0x00018AC3;
const char32_t TANGUT_COMPONENT_709 = 0x00018AC4;
const char32_t TANGUT_COMPONENT_710 = 0x00018AC5;
const char32_t TANGUT_COMPONENT_711 = 0x00018AC6;
const char32_t TANGUT_COMPONENT_712 = 0x00018AC7;
const char32_t TANGUT_COMPONENT_713 = 0x00018AC8;
const char32_t TANGUT_COMPONENT_714 = 0x00018AC9;
const char32_t TANGUT_COMPONENT_715 = 0x00018ACA;
const char32_t TANGUT_COMPONENT_716 = 0x00018ACB;
const char32_t TANGUT_COMPONENT_717 = 0x00018ACC;
const char32_t TANGUT_COMPONENT_718 = 0x00018ACD;
const char32_t TANGUT_COMPONENT_719 = 0x00018ACE;
const char32_t TANGUT_COMPONENT_720 = 0x00018ACF;
const char32_t TANGUT_COMPONENT_721 = 0x00018AD0;
const char32_t TANGUT_COMPONENT_722 = 0x00018AD1;
const char32_t TANGUT_COMPONENT_723 = 0x00018AD2;
const char32_t TANGUT_COMPONENT_724 = 0x00018AD3;
const char32_t TANGUT_COMPONENT_725 = 0x00018AD4;
const char32_t TANGUT_COMPONENT_726 = 0x00018AD5;
const char32_t TANGUT_COMPONENT_727 = 0x00018AD6;
const char32_t TANGUT_COMPONENT_728 = 0x00018AD7;
const char32_t TANGUT_COMPONENT_729 = 0x00018AD8;
const char32_t TANGUT_COMPONENT_730 = 0x00018AD9;
const char32_t TANGUT_COMPONENT_731 = 0x00018ADA;
const char32_t TANGUT_COMPONENT_732 = 0x00018ADB;
const char32_t TANGUT_COMPONENT_733 = 0x00018ADC;
const char32_t TANGUT_COMPONENT_734 = 0x00018ADD;
const char32_t TANGUT_COMPONENT_735 = 0x00018ADE;
const char32_t TANGUT_COMPONENT_736 = 0x00018ADF;
const char32_t TANGUT_COMPONENT_737 = 0x00018AE0;
const char32_t TANGUT_COMPONENT_738 = 0x00018AE1;
const char32_t TANGUT_COMPONENT_739 = 0x00018AE2;
const char32_t TANGUT_COMPONENT_740 = 0x00018AE3;
const char32_t TANGUT_COMPONENT_741 = 0x00018AE4;
const char32_t TANGUT_COMPONENT_742 = 0x00018AE5;
const char32_t TANGUT_COMPONENT_743 = 0x00018AE6;
const char32_t TANGUT_COMPONENT_744 = 0x00018AE7;
const char32_t TANGUT_COMPONENT_745 = 0x00018AE8;
const char32_t TANGUT_COMPONENT_746 = 0x00018AE9;
const char32_t TANGUT_COMPONENT_747 = 0x00018AEA;
const char32_t TANGUT_COMPONENT_748 = 0x00018AEB;
const char32_t TANGUT_COMPONENT_749 = 0x00018AEC;
const char32_t TANGUT_COMPONENT_750 = 0x00018AED;
const char32_t TANGUT_COMPONENT_751 = 0x00018AEE;
const char32_t TANGUT_COMPONENT_752 = 0x00018AEF;
const char32_t TANGUT_COMPONENT_753 = 0x00018AF0;
const char32_t TANGUT_COMPONENT_754 = 0x00018AF1;
const char32_t TANGUT_COMPONENT_755 = 0x00018AF2;
const char32_t TANGUT_COMPONENT_756 = 0x00018AF3;
const char32_t TANGUT_COMPONENT_757 = 0x00018AF4;
const char32_t TANGUT_COMPONENT_758 = 0x00018AF5;
const char32_t TANGUT_COMPONENT_759 = 0x00018AF6;
const char32_t TANGUT_COMPONENT_760 = 0x00018AF7;
const char32_t TANGUT_COMPONENT_761 = 0x00018AF8;
const char32_t TANGUT_COMPONENT_762 = 0x00018AF9;
const char32_t TANGUT_COMPONENT_763 = 0x00018AFA;
const char32_t TANGUT_COMPONENT_764 = 0x00018AFB;
const char32_t TANGUT_COMPONENT_765 = 0x00018AFC;
const char32_t TANGUT_COMPONENT_766 = 0x00018AFD;
const char32_t TANGUT_COMPONENT_767 = 0x00018AFE;
const char32_t TANGUT_COMPONENT_768 = 0x00018AFF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B00 = 0x00018B00;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B01 = 0x00018B01;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B02 = 0x00018B02;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B03 = 0x00018B03;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B04 = 0x00018B04;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B05 = 0x00018B05;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B06 = 0x00018B06;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B07 = 0x00018B07;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B08 = 0x00018B08;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B09 = 0x00018B09;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B0A = 0x00018B0A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B0B = 0x00018B0B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B0C = 0x00018B0C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B0D = 0x00018B0D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B0E = 0x00018B0E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B0F = 0x00018B0F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B10 = 0x00018B10;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B11 = 0x00018B11;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B12 = 0x00018B12;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B13 = 0x00018B13;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B14 = 0x00018B14;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B15 = 0x00018B15;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B16 = 0x00018B16;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B17 = 0x00018B17;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B18 = 0x00018B18;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B19 = 0x00018B19;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B1A = 0x00018B1A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B1B = 0x00018B1B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B1C = 0x00018B1C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B1D = 0x00018B1D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B1E = 0x00018B1E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B1F = 0x00018B1F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B20 = 0x00018B20;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B21 = 0x00018B21;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B22 = 0x00018B22;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B23 = 0x00018B23;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B24 = 0x00018B24;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B25 = 0x00018B25;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B26 = 0x00018B26;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B27 = 0x00018B27;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B28 = 0x00018B28;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B29 = 0x00018B29;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B2A = 0x00018B2A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B2B = 0x00018B2B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B2C = 0x00018B2C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B2D = 0x00018B2D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B2E = 0x00018B2E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B2F = 0x00018B2F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B30 = 0x00018B30;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B31 = 0x00018B31;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B32 = 0x00018B32;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B33 = 0x00018B33;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B34 = 0x00018B34;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B35 = 0x00018B35;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B36 = 0x00018B36;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B37 = 0x00018B37;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B38 = 0x00018B38;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B39 = 0x00018B39;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B3A = 0x00018B3A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B3B = 0x00018B3B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B3C = 0x00018B3C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B3D = 0x00018B3D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B3E = 0x00018B3E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B3F = 0x00018B3F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B40 = 0x00018B40;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B41 = 0x00018B41;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B42 = 0x00018B42;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B43 = 0x00018B43;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B44 = 0x00018B44;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B45 = 0x00018B45;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B46 = 0x00018B46;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B47 = 0x00018B47;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B48 = 0x00018B48;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B49 = 0x00018B49;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B4A = 0x00018B4A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B4B = 0x00018B4B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B4C = 0x00018B4C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B4D = 0x00018B4D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B4E = 0x00018B4E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B4F = 0x00018B4F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B50 = 0x00018B50;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B51 = 0x00018B51;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B52 = 0x00018B52;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B53 = 0x00018B53;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B54 = 0x00018B54;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B55 = 0x00018B55;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B56 = 0x00018B56;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B57 = 0x00018B57;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B58 = 0x00018B58;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B59 = 0x00018B59;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B5A = 0x00018B5A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B5B = 0x00018B5B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B5C = 0x00018B5C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B5D = 0x00018B5D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B5E = 0x00018B5E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B5F = 0x00018B5F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B60 = 0x00018B60;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B61 = 0x00018B61;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B62 = 0x00018B62;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B63 = 0x00018B63;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B64 = 0x00018B64;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B65 = 0x00018B65;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B66 = 0x00018B66;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B67 = 0x00018B67;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B68 = 0x00018B68;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B69 = 0x00018B69;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B6A = 0x00018B6A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B6B = 0x00018B6B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B6C = 0x00018B6C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B6D = 0x00018B6D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B6E = 0x00018B6E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B6F = 0x00018B6F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B70 = 0x00018B70;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B71 = 0x00018B71;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B72 = 0x00018B72;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B73 = 0x00018B73;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B74 = 0x00018B74;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B75 = 0x00018B75;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B76 = 0x00018B76;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B77 = 0x00018B77;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B78 = 0x00018B78;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B79 = 0x00018B79;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B7A = 0x00018B7A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B7B = 0x00018B7B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B7C = 0x00018B7C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B7D = 0x00018B7D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B7E = 0x00018B7E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B7F = 0x00018B7F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B80 = 0x00018B80;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B81 = 0x00018B81;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B82 = 0x00018B82;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B83 = 0x00018B83;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B84 = 0x00018B84;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B85 = 0x00018B85;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B86 = 0x00018B86;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B87 = 0x00018B87;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B88 = 0x00018B88;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B89 = 0x00018B89;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B8A = 0x00018B8A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B8B = 0x00018B8B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B8C = 0x00018B8C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B8D = 0x00018B8D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B8E = 0x00018B8E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B8F = 0x00018B8F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B90 = 0x00018B90;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B91 = 0x00018B91;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B92 = 0x00018B92;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B93 = 0x00018B93;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B94 = 0x00018B94;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B95 = 0x00018B95;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B96 = 0x00018B96;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B97 = 0x00018B97;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B98 = 0x00018B98;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B99 = 0x00018B99;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B9A = 0x00018B9A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B9B = 0x00018B9B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B9C = 0x00018B9C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B9D = 0x00018B9D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B9E = 0x00018B9E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18B9F = 0x00018B9F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA0 = 0x00018BA0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA1 = 0x00018BA1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA2 = 0x00018BA2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA3 = 0x00018BA3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA4 = 0x00018BA4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA5 = 0x00018BA5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA6 = 0x00018BA6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA7 = 0x00018BA7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA8 = 0x00018BA8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BA9 = 0x00018BA9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BAA = 0x00018BAA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BAB = 0x00018BAB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BAC = 0x00018BAC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BAD = 0x00018BAD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BAE = 0x00018BAE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BAF = 0x00018BAF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB0 = 0x00018BB0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB1 = 0x00018BB1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB2 = 0x00018BB2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB3 = 0x00018BB3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB4 = 0x00018BB4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB5 = 0x00018BB5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB6 = 0x00018BB6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB7 = 0x00018BB7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB8 = 0x00018BB8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BB9 = 0x00018BB9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BBA = 0x00018BBA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BBB = 0x00018BBB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BBC = 0x00018BBC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BBD = 0x00018BBD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BBE = 0x00018BBE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BBF = 0x00018BBF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC0 = 0x00018BC0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC1 = 0x00018BC1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC2 = 0x00018BC2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC3 = 0x00018BC3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC4 = 0x00018BC4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC5 = 0x00018BC5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC6 = 0x00018BC6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC7 = 0x00018BC7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC8 = 0x00018BC8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BC9 = 0x00018BC9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BCA = 0x00018BCA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BCB = 0x00018BCB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BCC = 0x00018BCC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BCD = 0x00018BCD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BCE = 0x00018BCE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BCF = 0x00018BCF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD0 = 0x00018BD0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD1 = 0x00018BD1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD2 = 0x00018BD2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD3 = 0x00018BD3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD4 = 0x00018BD4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD5 = 0x00018BD5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD6 = 0x00018BD6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD7 = 0x00018BD7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD8 = 0x00018BD8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BD9 = 0x00018BD9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BDA = 0x00018BDA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BDB = 0x00018BDB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BDC = 0x00018BDC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BDD = 0x00018BDD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BDE = 0x00018BDE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BDF = 0x00018BDF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE0 = 0x00018BE0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE1 = 0x00018BE1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE2 = 0x00018BE2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE3 = 0x00018BE3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE4 = 0x00018BE4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE5 = 0x00018BE5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE6 = 0x00018BE6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE7 = 0x00018BE7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE8 = 0x00018BE8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BE9 = 0x00018BE9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BEA = 0x00018BEA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BEB = 0x00018BEB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BEC = 0x00018BEC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BED = 0x00018BED;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BEE = 0x00018BEE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BEF = 0x00018BEF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF0 = 0x00018BF0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF1 = 0x00018BF1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF2 = 0x00018BF2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF3 = 0x00018BF3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF4 = 0x00018BF4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF5 = 0x00018BF5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF6 = 0x00018BF6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF7 = 0x00018BF7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF8 = 0x00018BF8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BF9 = 0x00018BF9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BFA = 0x00018BFA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BFB = 0x00018BFB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BFC = 0x00018BFC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BFD = 0x00018BFD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BFE = 0x00018BFE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18BFF = 0x00018BFF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C00 = 0x00018C00;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C01 = 0x00018C01;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C02 = 0x00018C02;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C03 = 0x00018C03;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C04 = 0x00018C04;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C05 = 0x00018C05;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C06 = 0x00018C06;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C07 = 0x00018C07;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C08 = 0x00018C08;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C09 = 0x00018C09;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C0A = 0x00018C0A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C0B = 0x00018C0B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C0C = 0x00018C0C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C0D = 0x00018C0D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C0E = 0x00018C0E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C0F = 0x00018C0F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C10 = 0x00018C10;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C11 = 0x00018C11;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C12 = 0x00018C12;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C13 = 0x00018C13;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C14 = 0x00018C14;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C15 = 0x00018C15;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C16 = 0x00018C16;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C17 = 0x00018C17;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C18 = 0x00018C18;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C19 = 0x00018C19;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C1A = 0x00018C1A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C1B = 0x00018C1B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C1C = 0x00018C1C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C1D = 0x00018C1D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C1E = 0x00018C1E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C1F = 0x00018C1F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C20 = 0x00018C20;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C21 = 0x00018C21;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C22 = 0x00018C22;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C23 = 0x00018C23;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C24 = 0x00018C24;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C25 = 0x00018C25;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C26 = 0x00018C26;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C27 = 0x00018C27;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C28 = 0x00018C28;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C29 = 0x00018C29;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C2A = 0x00018C2A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C2B = 0x00018C2B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C2C = 0x00018C2C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C2D = 0x00018C2D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C2E = 0x00018C2E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C2F = 0x00018C2F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C30 = 0x00018C30;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C31 = 0x00018C31;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C32 = 0x00018C32;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C33 = 0x00018C33;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C34 = 0x00018C34;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C35 = 0x00018C35;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C36 = 0x00018C36;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C37 = 0x00018C37;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C38 = 0x00018C38;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C39 = 0x00018C39;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C3A = 0x00018C3A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C3B = 0x00018C3B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C3C = 0x00018C3C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C3D = 0x00018C3D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C3E = 0x00018C3E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C3F = 0x00018C3F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C40 = 0x00018C40;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C41 = 0x00018C41;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C42 = 0x00018C42;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C43 = 0x00018C43;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C44 = 0x00018C44;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C45 = 0x00018C45;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C46 = 0x00018C46;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C47 = 0x00018C47;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C48 = 0x00018C48;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C49 = 0x00018C49;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C4A = 0x00018C4A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C4B = 0x00018C4B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C4C = 0x00018C4C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C4D = 0x00018C4D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C4E = 0x00018C4E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C4F = 0x00018C4F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C50 = 0x00018C50;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C51 = 0x00018C51;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C52 = 0x00018C52;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C53 = 0x00018C53;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C54 = 0x00018C54;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C55 = 0x00018C55;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C56 = 0x00018C56;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C57 = 0x00018C57;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C58 = 0x00018C58;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C59 = 0x00018C59;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C5A = 0x00018C5A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C5B = 0x00018C5B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C5C = 0x00018C5C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C5D = 0x00018C5D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C5E = 0x00018C5E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C5F = 0x00018C5F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C60 = 0x00018C60;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C61 = 0x00018C61;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C62 = 0x00018C62;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C63 = 0x00018C63;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C64 = 0x00018C64;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C65 = 0x00018C65;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C66 = 0x00018C66;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C67 = 0x00018C67;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C68 = 0x00018C68;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C69 = 0x00018C69;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C6A = 0x00018C6A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C6B = 0x00018C6B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C6C = 0x00018C6C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C6D = 0x00018C6D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C6E = 0x00018C6E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C6F = 0x00018C6F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C70 = 0x00018C70;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C71 = 0x00018C71;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C72 = 0x00018C72;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C73 = 0x00018C73;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C74 = 0x00018C74;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C75 = 0x00018C75;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C76 = 0x00018C76;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C77 = 0x00018C77;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C78 = 0x00018C78;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C79 = 0x00018C79;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C7A = 0x00018C7A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C7B = 0x00018C7B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C7C = 0x00018C7C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C7D = 0x00018C7D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C7E = 0x00018C7E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C7F = 0x00018C7F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C80 = 0x00018C80;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C81 = 0x00018C81;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C82 = 0x00018C82;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C83 = 0x00018C83;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C84 = 0x00018C84;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C85 = 0x00018C85;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C86 = 0x00018C86;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C87 = 0x00018C87;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C88 = 0x00018C88;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C89 = 0x00018C89;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C8A = 0x00018C8A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C8B = 0x00018C8B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C8C = 0x00018C8C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C8D = 0x00018C8D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C8E = 0x00018C8E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C8F = 0x00018C8F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C90 = 0x00018C90;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C91 = 0x00018C91;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C92 = 0x00018C92;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C93 = 0x00018C93;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C94 = 0x00018C94;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C95 = 0x00018C95;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C96 = 0x00018C96;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C97 = 0x00018C97;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C98 = 0x00018C98;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C99 = 0x00018C99;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C9A = 0x00018C9A;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C9B = 0x00018C9B;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C9C = 0x00018C9C;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C9D = 0x00018C9D;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C9E = 0x00018C9E;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18C9F = 0x00018C9F;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA0 = 0x00018CA0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA1 = 0x00018CA1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA2 = 0x00018CA2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA3 = 0x00018CA3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA4 = 0x00018CA4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA5 = 0x00018CA5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA6 = 0x00018CA6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA7 = 0x00018CA7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA8 = 0x00018CA8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CA9 = 0x00018CA9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CAA = 0x00018CAA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CAB = 0x00018CAB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CAC = 0x00018CAC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CAD = 0x00018CAD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CAE = 0x00018CAE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CAF = 0x00018CAF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB0 = 0x00018CB0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB1 = 0x00018CB1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB2 = 0x00018CB2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB3 = 0x00018CB3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB4 = 0x00018CB4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB5 = 0x00018CB5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB6 = 0x00018CB6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB7 = 0x00018CB7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB8 = 0x00018CB8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CB9 = 0x00018CB9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CBA = 0x00018CBA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CBB = 0x00018CBB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CBC = 0x00018CBC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CBD = 0x00018CBD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CBE = 0x00018CBE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CBF = 0x00018CBF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC0 = 0x00018CC0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC1 = 0x00018CC1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC2 = 0x00018CC2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC3 = 0x00018CC3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC4 = 0x00018CC4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC5 = 0x00018CC5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC6 = 0x00018CC6;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC7 = 0x00018CC7;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC8 = 0x00018CC8;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CC9 = 0x00018CC9;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CCA = 0x00018CCA;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CCB = 0x00018CCB;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CCC = 0x00018CCC;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CCD = 0x00018CCD;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CCE = 0x00018CCE;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CCF = 0x00018CCF;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CD0 = 0x00018CD0;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CD1 = 0x00018CD1;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CD2 = 0x00018CD2;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CD3 = 0x00018CD3;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CD4 = 0x00018CD4;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CD5 = 0x00018CD5;
const char32_t KHITAN_SMALL_SCRIPT_CHARACTER_18CFF = 0x00018CFF;
const char32_t TANGUT_IDEOGRAPH_SUPPLEMENT_FIRST = 0x00018D00;
const char32_t TANGUT_IDEOGRAPH_SUPPLEMENT_LAST = 0x00018D1E;
const char32_t TANGUT_COMPONENT_769 = 0x00018D80;
const char32_t TANGUT_COMPONENT_770 = 0x00018D81;
const char32_t TANGUT_COMPONENT_771 = 0x00018D82;
const char32_t TANGUT_COMPONENT_772 = 0x00018D83;
const char32_t TANGUT_COMPONENT_773 = 0x00018D84;
const char32_t TANGUT_COMPONENT_774 = 0x00018D85;
const char32_t TANGUT_COMPONENT_775 = 0x00018D86;
const char32_t TANGUT_COMPONENT_776 = 0x00018D87;
const char32_t TANGUT_COMPONENT_777 = 0x00018D88;
const char32_t TANGUT_COMPONENT_778 = 0x00018D89;
const char32_t TANGUT_COMPONENT_779 = 0x00018D8A;
const char32_t TANGUT_COMPONENT_780 = 0x00018D8B;
const char32_t TANGUT_COMPONENT_781 = 0x00018D8C;
const char32_t TANGUT_COMPONENT_782 = 0x00018D8D;
const char32_t TANGUT_COMPONENT_783 = 0x00018D8E;
const char32_t TANGUT_COMPONENT_784 = 0x00018D8F;
const char32_t TANGUT_COMPONENT_785 = 0x00018D90;
const char32_t TANGUT_COMPONENT_786 = 0x00018D91;
const char32_t TANGUT_COMPONENT_787 = 0x00018D92;
const char32_t TANGUT_COMPONENT_788 = 0x00018D93;
const char32_t TANGUT_COMPONENT_789 = 0x00018D94;
const char32_t TANGUT_COMPONENT_790 = 0x00018D95;
const char32_t TANGUT_COMPONENT_791 = 0x00018D96;
const char32_t TANGUT_COMPONENT_792 = 0x00018D97;
const char32_t TANGUT_COMPONENT_793 = 0x00018D98;
const char32_t TANGUT_COMPONENT_794 = 0x00018D99;
const char32_t TANGUT_COMPONENT_795 = 0x00018D9A;
const char32_t TANGUT_COMPONENT_796 = 0x00018D9B;
const char32_t TANGUT_COMPONENT_797 = 0x00018D9C;
const char32_t TANGUT_COMPONENT_798 = 0x00018D9D;
const char32_t TANGUT_COMPONENT_799 = 0x00018D9E;
const char32_t TANGUT_COMPONENT_800 = 0x00018D9F;
const char32_t TANGUT_COMPONENT_801 = 0x00018DA0;
const char32_t TANGUT_COMPONENT_802 = 0x00018DA1;
const char32_t TANGUT_COMPONENT_803 = 0x00018DA2;
const char32_t TANGUT_COMPONENT_804 = 0x00018DA3;
const char32_t TANGUT_COMPONENT_805 = 0x00018DA4;
const char32_t TANGUT_COMPONENT_806 = 0x00018DA5;
const char32_t TANGUT_COMPONENT_807 = 0x00018DA6;
const char32_t TANGUT_COMPONENT_808 = 0x00018DA7;
const char32_t TANGUT_COMPONENT_809 = 0x00018DA8;
const char32_t TANGUT_COMPONENT_810 = 0x00018DA9;
const char32_t TANGUT_COMPONENT_811 = 0x00018DAA;
const char32_t TANGUT_COMPONENT_812 = 0x00018DAB;
const char32_t TANGUT_COMPONENT_813 = 0x00018DAC;
const char32_t TANGUT_COMPONENT_814 = 0x00018DAD;
const char32_t TANGUT_COMPONENT_815 = 0x00018DAE;
const char32_t TANGUT_COMPONENT_816 = 0x00018DAF;
const char32_t TANGUT_COMPONENT_817 = 0x00018DB0;
const char32_t TANGUT_COMPONENT_818 = 0x00018DB1;
const char32_t TANGUT_COMPONENT_819 = 0x00018DB2;
const char32_t TANGUT_COMPONENT_820 = 0x00018DB3;
const char32_t TANGUT_COMPONENT_821 = 0x00018DB4;
const char32_t TANGUT_COMPONENT_822 = 0x00018DB5;
const char32_t TANGUT_COMPONENT_823 = 0x00018DB6;
const char32_t TANGUT_COMPONENT_824 = 0x00018DB7;
const char32_t TANGUT_COMPONENT_825 = 0x00018DB8;
const char32_t TANGUT_COMPONENT_826 = 0x00018DB9;
const char32_t TANGUT_COMPONENT_827 = 0x00018DBA;
const char32_t TANGUT_COMPONENT_828 = 0x00018DBB;
const char32_t TANGUT_COMPONENT_829 = 0x00018DBC;
const char32_t TANGUT_COMPONENT_830 = 0x00018DBD;
const char32_t TANGUT_COMPONENT_831 = 0x00018DBE;
const char32_t TANGUT_COMPONENT_832 = 0x00018DBF;
const char32_t TANGUT_COMPONENT_833 = 0x00018DC0;
const char32_t TANGUT_COMPONENT_834 = 0x00018DC1;
const char32_t TANGUT_COMPONENT_835 = 0x00018DC2;
const char32_t TANGUT_COMPONENT_836 = 0x00018DC3;
const char32_t TANGUT_COMPONENT_837 = 0x00018DC4;
const char32_t TANGUT_COMPONENT_838 = 0x00018DC5;
const char32_t TANGUT_COMPONENT_839 = 0x00018DC6;
const char32_t TANGUT_COMPONENT_840 = 0x00018DC7;
const char32_t TANGUT_COMPONENT_841 = 0x00018DC8;
const char32_t TANGUT_COMPONENT_842 = 0x00018DC9;
const char32_t TANGUT_COMPONENT_843 = 0x00018DCA;
const char32_t TANGUT_COMPONENT_844 = 0x00018DCB;
const char32_t TANGUT_COMPONENT_845 = 0x00018DCC;
const char32_t TANGUT_COMPONENT_846 = 0x00018DCD;
const char32_t TANGUT_COMPONENT_847 = 0x00018DCE;
const char32_t TANGUT_COMPONENT_848 = 0x00018DCF;
const char32_t TANGUT_COMPONENT_849 = 0x00018DD0;
const char32_t TANGUT_COMPONENT_850 = 0x00018DD1;
const char32_t TANGUT_COMPONENT_851 = 0x00018DD2;
const char32_t TANGUT_COMPONENT_852 = 0x00018DD3;
const char32_t TANGUT_COMPONENT_853 = 0x00018DD4;
const char32_t TANGUT_COMPONENT_854 = 0x00018DD5;
const char32_t TANGUT_COMPONENT_855 = 0x00018DD6;
const char32_t TANGUT_COMPONENT_856 = 0x00018DD7;
const char32_t TANGUT_COMPONENT_857 = 0x00018DD8;
const char32_t TANGUT_COMPONENT_858 = 0x00018DD9;
const char32_t TANGUT_COMPONENT_859 = 0x00018DDA;
const char32_t TANGUT_COMPONENT_860 = 0x00018DDB;
const char32_t TANGUT_COMPONENT_861 = 0x00018DDC;
const char32_t TANGUT_COMPONENT_862 = 0x00018DDD;
const char32_t TANGUT_COMPONENT_863 = 0x00018DDE;
const char32_t TANGUT_COMPONENT_864 = 0x00018DDF;
const char32_t TANGUT_COMPONENT_865 = 0x00018DE0;
const char32_t TANGUT_COMPONENT_866 = 0x00018DE1;
const char32_t TANGUT_COMPONENT_867 = 0x00018DE2;
const char32_t TANGUT_COMPONENT_868 = 0x00018DE3;
const char32_t TANGUT_COMPONENT_869 = 0x00018DE4;
const char32_t TANGUT_COMPONENT_870 = 0x00018DE5;
const char32_t TANGUT_COMPONENT_871 = 0x00018DE6;
const char32_t TANGUT_COMPONENT_872 = 0x00018DE7;
const char32_t TANGUT_COMPONENT_873 = 0x00018DE8;
const char32_t TANGUT_COMPONENT_874 = 0x00018DE9;
const char32_t TANGUT_COMPONENT_875 = 0x00018DEA;
const char32_t TANGUT_COMPONENT_876 = 0x00018DEB;
const char32_t TANGUT_COMPONENT_877 = 0x00018DEC;
const char32_t TANGUT_COMPONENT_878 = 0x00018DED;
const char32_t TANGUT_COMPONENT_879 = 0x00018DEE;
const char32_t TANGUT_COMPONENT_880 = 0x00018DEF;
const char32_t TANGUT_COMPONENT_881 = 0x00018DF0;
const char32_t TANGUT_COMPONENT_882 = 0x00018DF1;
const char32_t TANGUT_COMPONENT_883 = 0x00018DF2;
const char32_t KATAKANA_LETTER_MINNAN_TONE_2 = 0x0001AFF0;
const char32_t KATAKANA_LETTER_MINNAN_TONE_3 = 0x0001AFF1;
const char32_t KATAKANA_LETTER_MINNAN_TONE_4 = 0x0001AFF2;
const char32_t KATAKANA_LETTER_MINNAN_TONE_5 = 0x0001AFF3;
const char32_t KATAKANA_LETTER_MINNAN_TONE_7 = 0x0001AFF5;
const char32_t KATAKANA_LETTER_MINNAN_TONE_8 = 0x0001AFF6;
const char32_t KATAKANA_LETTER_MINNAN_NASALIZED_TONE_1 = 0x0001AFF7;
const char32_t KATAKANA_LETTER_MINNAN_NASALIZED_TONE_2 = 0x0001AFF8;
const char32_t KATAKANA_LETTER_MINNAN_NASALIZED_TONE_3 = 0x0001AFF9;
const char32_t KATAKANA_LETTER_MINNAN_NASALIZED_TONE_4 = 0x0001AFFA;
const char32_t KATAKANA_LETTER_MINNAN_NASALIZED_TONE_5 = 0x0001AFFB;
const char32_t KATAKANA_LETTER_MINNAN_NASALIZED_TONE_7 = 0x0001AFFD;
const char32_t KATAKANA_LETTER_MINNAN_NASALIZED_TONE_8 = 0x0001AFFE;
const char32_t KATAKANA_LETTER_ARCHAIC_E = 0x0001B000;
const char32_t HIRAGANA_LETTER_ARCHAIC_YE = 0x0001B001;
const char32_t HENTAIGANA_LETTER_A_1 = 0x0001B002;
const char32_t HENTAIGANA_LETTER_A_2 = 0x0001B003;
const char32_t HENTAIGANA_LETTER_A_3 = 0x0001B004;
const char32_t HENTAIGANA_LETTER_A_WO = 0x0001B005;
const char32_t HENTAIGANA_LETTER_I_1 = 0x0001B006;
const char32_t HENTAIGANA_LETTER_I_2 = 0x0001B007;
const char32_t HENTAIGANA_LETTER_I_3 = 0x0001B008;
const char32_t HENTAIGANA_LETTER_I_4 = 0x0001B009;
const char32_t HENTAIGANA_LETTER_U_1 = 0x0001B00A;
const char32_t HENTAIGANA_LETTER_U_2 = 0x0001B00B;
const char32_t HENTAIGANA_LETTER_U_3 = 0x0001B00C;
const char32_t HENTAIGANA_LETTER_U_4 = 0x0001B00D;
const char32_t HENTAIGANA_LETTER_U_5 = 0x0001B00E;
const char32_t HENTAIGANA_LETTER_E_2 = 0x0001B00F;
const char32_t HENTAIGANA_LETTER_E_3 = 0x0001B010;
const char32_t HENTAIGANA_LETTER_E_4 = 0x0001B011;
const char32_t HENTAIGANA_LETTER_E_5 = 0x0001B012;
const char32_t HENTAIGANA_LETTER_E_6 = 0x0001B013;
const char32_t HENTAIGANA_LETTER_O_1 = 0x0001B014;
const char32_t HENTAIGANA_LETTER_O_2 = 0x0001B015;
const char32_t HENTAIGANA_LETTER_O_3 = 0x0001B016;
const char32_t HENTAIGANA_LETTER_KA_1 = 0x0001B017;
const char32_t HENTAIGANA_LETTER_KA_2 = 0x0001B018;
const char32_t HENTAIGANA_LETTER_KA_3 = 0x0001B019;
const char32_t HENTAIGANA_LETTER_KA_4 = 0x0001B01A;
const char32_t HENTAIGANA_LETTER_KA_5 = 0x0001B01B;
const char32_t HENTAIGANA_LETTER_KA_6 = 0x0001B01C;
const char32_t HENTAIGANA_LETTER_KA_7 = 0x0001B01D;
const char32_t HENTAIGANA_LETTER_KA_8 = 0x0001B01E;
const char32_t HENTAIGANA_LETTER_KA_9 = 0x0001B01F;
const char32_t HENTAIGANA_LETTER_KA_10 = 0x0001B020;
const char32_t HENTAIGANA_LETTER_KA_11 = 0x0001B021;
const char32_t HENTAIGANA_LETTER_KA_KE = 0x0001B022;
const char32_t HENTAIGANA_LETTER_KI_1 = 0x0001B023;
const char32_t HENTAIGANA_LETTER_KI_2 = 0x0001B024;
const char32_t HENTAIGANA_LETTER_KI_3 = 0x0001B025;
const char32_t HENTAIGANA_LETTER_KI_4 = 0x0001B026;
const char32_t HENTAIGANA_LETTER_KI_5 = 0x0001B027;
const char32_t HENTAIGANA_LETTER_KI_6 = 0x0001B028;
const char32_t HENTAIGANA_LETTER_KI_7 = 0x0001B029;
const char32_t HENTAIGANA_LETTER_KI_8 = 0x0001B02A;
const char32_t HENTAIGANA_LETTER_KU_1 = 0x0001B02B;
const char32_t HENTAIGANA_LETTER_KU_2 = 0x0001B02C;
const char32_t HENTAIGANA_LETTER_KU_3 = 0x0001B02D;
const char32_t HENTAIGANA_LETTER_KU_4 = 0x0001B02E;
const char32_t HENTAIGANA_LETTER_KU_5 = 0x0001B02F;
const char32_t HENTAIGANA_LETTER_KU_6 = 0x0001B030;
const char32_t HENTAIGANA_LETTER_KU_7 = 0x0001B031;
const char32_t HENTAIGANA_LETTER_KE_1 = 0x0001B032;
const char32_t HENTAIGANA_LETTER_KE_2 = 0x0001B033;
const char32_t HENTAIGANA_LETTER_KE_3 = 0x0001B034;
const char32_t HENTAIGANA_LETTER_KE_4 = 0x0001B035;
const char32_t HENTAIGANA_LETTER_KE_5 = 0x0001B036;
const char32_t HENTAIGANA_LETTER_KE_6 = 0x0001B037;
const char32_t HENTAIGANA_LETTER_KO_1 = 0x0001B038;
const char32_t HENTAIGANA_LETTER_KO_2 = 0x0001B039;
const char32_t HENTAIGANA_LETTER_KO_3 = 0x0001B03A;
const char32_t HENTAIGANA_LETTER_KO_KI = 0x0001B03B;
const char32_t HENTAIGANA_LETTER_SA_1 = 0x0001B03C;
const char32_t HENTAIGANA_LETTER_SA_2 = 0x0001B03D;
const char32_t HENTAIGANA_LETTER_SA_3 = 0x0001B03E;
const char32_t HENTAIGANA_LETTER_SA_4 = 0x0001B03F;
const char32_t HENTAIGANA_LETTER_SA_5 = 0x0001B040;
const char32_t HENTAIGANA_LETTER_SA_6 = 0x0001B041;
const char32_t HENTAIGANA_LETTER_SA_7 = 0x0001B042;
const char32_t HENTAIGANA_LETTER_SA_8 = 0x0001B043;
const char32_t HENTAIGANA_LETTER_SI_1 = 0x0001B044;
const char32_t HENTAIGANA_LETTER_SI_2 = 0x0001B045;
const char32_t HENTAIGANA_LETTER_SI_3 = 0x0001B046;
const char32_t HENTAIGANA_LETTER_SI_4 = 0x0001B047;
const char32_t HENTAIGANA_LETTER_SI_5 = 0x0001B048;
const char32_t HENTAIGANA_LETTER_SI_6 = 0x0001B049;
const char32_t HENTAIGANA_LETTER_SU_1 = 0x0001B04A;
const char32_t HENTAIGANA_LETTER_SU_2 = 0x0001B04B;
const char32_t HENTAIGANA_LETTER_SU_3 = 0x0001B04C;
const char32_t HENTAIGANA_LETTER_SU_4 = 0x0001B04D;
const char32_t HENTAIGANA_LETTER_SU_5 = 0x0001B04E;
const char32_t HENTAIGANA_LETTER_SU_6 = 0x0001B04F;
const char32_t HENTAIGANA_LETTER_SU_7 = 0x0001B050;
const char32_t HENTAIGANA_LETTER_SU_8 = 0x0001B051;
const char32_t HENTAIGANA_LETTER_SE_1 = 0x0001B052;
const char32_t HENTAIGANA_LETTER_SE_2 = 0x0001B053;
const char32_t HENTAIGANA_LETTER_SE_3 = 0x0001B054;
const char32_t HENTAIGANA_LETTER_SE_4 = 0x0001B055;
const char32_t HENTAIGANA_LETTER_SE_5 = 0x0001B056;
const char32_t HENTAIGANA_LETTER_SO_1 = 0x0001B057;
const char32_t HENTAIGANA_LETTER_SO_2 = 0x0001B058;
const char32_t HENTAIGANA_LETTER_SO_3 = 0x0001B059;
const char32_t HENTAIGANA_LETTER_SO_4 = 0x0001B05A;
const char32_t HENTAIGANA_LETTER_SO_5 = 0x0001B05B;
const char32_t HENTAIGANA_LETTER_SO_6 = 0x0001B05C;
const char32_t HENTAIGANA_LETTER_SO_7 = 0x0001B05D;
const char32_t HENTAIGANA_LETTER_TA_1 = 0x0001B05E;
const char32_t HENTAIGANA_LETTER_TA_2 = 0x0001B05F;
const char32_t HENTAIGANA_LETTER_TA_3 = 0x0001B060;
const char32_t HENTAIGANA_LETTER_TA_4 = 0x0001B061;
const char32_t HENTAIGANA_LETTER_TI_1 = 0x0001B062;
const char32_t HENTAIGANA_LETTER_TI_2 = 0x0001B063;
const char32_t HENTAIGANA_LETTER_TI_3 = 0x0001B064;
const char32_t HENTAIGANA_LETTER_TI_4 = 0x0001B065;
const char32_t HENTAIGANA_LETTER_TI_5 = 0x0001B066;
const char32_t HENTAIGANA_LETTER_TI_6 = 0x0001B067;
const char32_t HENTAIGANA_LETTER_TI_7 = 0x0001B068;
const char32_t HENTAIGANA_LETTER_TU_1 = 0x0001B069;
const char32_t HENTAIGANA_LETTER_TU_2 = 0x0001B06A;
const char32_t HENTAIGANA_LETTER_TU_3 = 0x0001B06B;
const char32_t HENTAIGANA_LETTER_TU_4 = 0x0001B06C;
const char32_t HENTAIGANA_LETTER_TU_TO = 0x0001B06D;
const char32_t HENTAIGANA_LETTER_TE_1 = 0x0001B06E;
const char32_t HENTAIGANA_LETTER_TE_2 = 0x0001B06F;
const char32_t HENTAIGANA_LETTER_TE_3 = 0x0001B070;
const char32_t HENTAIGANA_LETTER_TE_4 = 0x0001B071;
const char32_t HENTAIGANA_LETTER_TE_5 = 0x0001B072;
const char32_t HENTAIGANA_LETTER_TE_6 = 0x0001B073;
const char32_t HENTAIGANA_LETTER_TE_7 = 0x0001B074;
const char32_t HENTAIGANA_LETTER_TE_8 = 0x0001B075;
const char32_t HENTAIGANA_LETTER_TE_9 = 0x0001B076;
const char32_t HENTAIGANA_LETTER_TO_1 = 0x0001B077;
const char32_t HENTAIGANA_LETTER_TO_2 = 0x0001B078;
const char32_t HENTAIGANA_LETTER_TO_3 = 0x0001B079;
const char32_t HENTAIGANA_LETTER_TO_4 = 0x0001B07A;
const char32_t HENTAIGANA_LETTER_TO_5 = 0x0001B07B;
const char32_t HENTAIGANA_LETTER_TO_6 = 0x0001B07C;
const char32_t HENTAIGANA_LETTER_TO_RA = 0x0001B07D;
const char32_t HENTAIGANA_LETTER_NA_1 = 0x0001B07E;
const char32_t HENTAIGANA_LETTER_NA_2 = 0x0001B07F;
const char32_t HENTAIGANA_LETTER_NA_3 = 0x0001B080;
const char32_t HENTAIGANA_LETTER_NA_4 = 0x0001B081;
const char32_t HENTAIGANA_LETTER_NA_5 = 0x0001B082;
const char32_t HENTAIGANA_LETTER_NA_6 = 0x0001B083;
const char32_t HENTAIGANA_LETTER_NA_7 = 0x0001B084;
const char32_t HENTAIGANA_LETTER_NA_8 = 0x0001B085;
const char32_t HENTAIGANA_LETTER_NA_9 = 0x0001B086;
const char32_t HENTAIGANA_LETTER_NI_1 = 0x0001B087;
const char32_t HENTAIGANA_LETTER_NI_2 = 0x0001B088;
const char32_t HENTAIGANA_LETTER_NI_3 = 0x0001B089;
const char32_t HENTAIGANA_LETTER_NI_4 = 0x0001B08A;
const char32_t HENTAIGANA_LETTER_NI_5 = 0x0001B08B;
const char32_t HENTAIGANA_LETTER_NI_6 = 0x0001B08C;
const char32_t HENTAIGANA_LETTER_NI_7 = 0x0001B08D;
const char32_t HENTAIGANA_LETTER_NI_TE = 0x0001B08E;
const char32_t HENTAIGANA_LETTER_NU_1 = 0x0001B08F;
const char32_t HENTAIGANA_LETTER_NU_2 = 0x0001B090;
const char32_t HENTAIGANA_LETTER_NU_3 = 0x0001B091;
const char32_t HENTAIGANA_LETTER_NE_1 = 0x0001B092;
const char32_t HENTAIGANA_LETTER_NE_2 = 0x0001B093;
const char32_t HENTAIGANA_LETTER_NE_3 = 0x0001B094;
const char32_t HENTAIGANA_LETTER_NE_4 = 0x0001B095;
const char32_t HENTAIGANA_LETTER_NE_5 = 0x0001B096;
const char32_t HENTAIGANA_LETTER_NE_6 = 0x0001B097;
const char32_t HENTAIGANA_LETTER_NE_KO = 0x0001B098;
const char32_t HENTAIGANA_LETTER_NO_1 = 0x0001B099;
const char32_t HENTAIGANA_LETTER_NO_2 = 0x0001B09A;
const char32_t HENTAIGANA_LETTER_NO_3 = 0x0001B09B;
const char32_t HENTAIGANA_LETTER_NO_4 = 0x0001B09C;
const char32_t HENTAIGANA_LETTER_NO_5 = 0x0001B09D;
const char32_t HENTAIGANA_LETTER_HA_1 = 0x0001B09E;
const char32_t HENTAIGANA_LETTER_HA_2 = 0x0001B09F;
const char32_t HENTAIGANA_LETTER_HA_3 = 0x0001B0A0;
const char32_t HENTAIGANA_LETTER_HA_4 = 0x0001B0A1;
const char32_t HENTAIGANA_LETTER_HA_5 = 0x0001B0A2;
const char32_t HENTAIGANA_LETTER_HA_6 = 0x0001B0A3;
const char32_t HENTAIGANA_LETTER_HA_7 = 0x0001B0A4;
const char32_t HENTAIGANA_LETTER_HA_8 = 0x0001B0A5;
const char32_t HENTAIGANA_LETTER_HA_9 = 0x0001B0A6;
const char32_t HENTAIGANA_LETTER_HA_10 = 0x0001B0A7;
const char32_t HENTAIGANA_LETTER_HA_11 = 0x0001B0A8;
const char32_t HENTAIGANA_LETTER_HI_1 = 0x0001B0A9;
const char32_t HENTAIGANA_LETTER_HI_2 = 0x0001B0AA;
const char32_t HENTAIGANA_LETTER_HI_3 = 0x0001B0AB;
const char32_t HENTAIGANA_LETTER_HI_4 = 0x0001B0AC;
const char32_t HENTAIGANA_LETTER_HI_5 = 0x0001B0AD;
const char32_t HENTAIGANA_LETTER_HI_6 = 0x0001B0AE;
const char32_t HENTAIGANA_LETTER_HI_7 = 0x0001B0AF;
const char32_t HENTAIGANA_LETTER_HU_1 = 0x0001B0B0;
const char32_t HENTAIGANA_LETTER_HU_2 = 0x0001B0B1;
const char32_t HENTAIGANA_LETTER_HU_3 = 0x0001B0B2;
const char32_t HENTAIGANA_LETTER_HE_1 = 0x0001B0B3;
const char32_t HENTAIGANA_LETTER_HE_2 = 0x0001B0B4;
const char32_t HENTAIGANA_LETTER_HE_3 = 0x0001B0B5;
const char32_t HENTAIGANA_LETTER_HE_4 = 0x0001B0B6;
const char32_t HENTAIGANA_LETTER_HE_5 = 0x0001B0B7;
const char32_t HENTAIGANA_LETTER_HE_6 = 0x0001B0B8;
const char32_t HENTAIGANA_LETTER_HE_7 = 0x0001B0B9;
const char32_t HENTAIGANA_LETTER_HO_1 = 0x0001B0BA;
const char32_t HENTAIGANA_LETTER_HO_2 = 0x0001B0BB;
const char32_t HENTAIGANA_LETTER_HO_3 = 0x0001B0BC;
const char32_t HENTAIGANA_LETTER_HO_4 = 0x0001B0BD;
const char32_t HENTAIGANA_LETTER_HO_5 = 0x0001B0BE;
const char32_t HENTAIGANA_LETTER_HO_6 = 0x0001B0BF;
const char32_t HENTAIGANA_LETTER_HO_7 = 0x0001B0C0;
const char32_t HENTAIGANA_LETTER_HO_8 = 0x0001B0C1;
const char32_t HENTAIGANA_LETTER_MA_1 = 0x0001B0C2;
const char32_t HENTAIGANA_LETTER_MA_2 = 0x0001B0C3;
const char32_t HENTAIGANA_LETTER_MA_3 = 0x0001B0C4;
const char32_t HENTAIGANA_LETTER_MA_4 = 0x0001B0C5;
const char32_t HENTAIGANA_LETTER_MA_5 = 0x0001B0C6;
const char32_t HENTAIGANA_LETTER_MA_6 = 0x0001B0C7;
const char32_t HENTAIGANA_LETTER_MA_7 = 0x0001B0C8;
const char32_t HENTAIGANA_LETTER_MI_1 = 0x0001B0C9;
const char32_t HENTAIGANA_LETTER_MI_2 = 0x0001B0CA;
const char32_t HENTAIGANA_LETTER_MI_3 = 0x0001B0CB;
const char32_t HENTAIGANA_LETTER_MI_4 = 0x0001B0CC;
const char32_t HENTAIGANA_LETTER_MI_5 = 0x0001B0CD;
const char32_t HENTAIGANA_LETTER_MI_6 = 0x0001B0CE;
const char32_t HENTAIGANA_LETTER_MI_7 = 0x0001B0CF;
const char32_t HENTAIGANA_LETTER_MU_1 = 0x0001B0D0;
const char32_t HENTAIGANA_LETTER_MU_2 = 0x0001B0D1;
const char32_t HENTAIGANA_LETTER_MU_3 = 0x0001B0D2;
const char32_t HENTAIGANA_LETTER_MU_4 = 0x0001B0D3;
const char32_t HENTAIGANA_LETTER_ME_1 = 0x0001B0D4;
const char32_t HENTAIGANA_LETTER_ME_2 = 0x0001B0D5;
const char32_t HENTAIGANA_LETTER_ME_MA = 0x0001B0D6;
const char32_t HENTAIGANA_LETTER_MO_1 = 0x0001B0D7;
const char32_t HENTAIGANA_LETTER_MO_2 = 0x0001B0D8;
const char32_t HENTAIGANA_LETTER_MO_3 = 0x0001B0D9;
const char32_t HENTAIGANA_LETTER_MO_4 = 0x0001B0DA;
const char32_t HENTAIGANA_LETTER_MO_5 = 0x0001B0DB;
const char32_t HENTAIGANA_LETTER_MO_6 = 0x0001B0DC;
const char32_t HENTAIGANA_LETTER_YA_1 = 0x0001B0DD;
const char32_t HENTAIGANA_LETTER_YA_2 = 0x0001B0DE;
const char32_t HENTAIGANA_LETTER_YA_3 = 0x0001B0DF;
const char32_t HENTAIGANA_LETTER_YA_4 = 0x0001B0E0;
const char32_t HENTAIGANA_LETTER_YA_5 = 0x0001B0E1;
const char32_t HENTAIGANA_LETTER_YA_YO = 0x0001B0E2;
const char32_t HENTAIGANA_LETTER_YU_1 = 0x0001B0E3;
const char32_t HENTAIGANA_LETTER_YU_2 = 0x0001B0E4;
const char32_t HENTAIGANA_LETTER_YU_3 = 0x0001B0E5;
const char32_t HENTAIGANA_LETTER_YU_4 = 0x0001B0E6;
const char32_t HENTAIGANA_LETTER_YO_1 = 0x0001B0E7;
const char32_t HENTAIGANA_LETTER_YO_2 = 0x0001B0E8;
const char32_t HENTAIGANA_LETTER_YO_3 = 0x0001B0E9;
const char32_t HENTAIGANA_LETTER_YO_4 = 0x0001B0EA;
const char32_t HENTAIGANA_LETTER_YO_5 = 0x0001B0EB;
const char32_t HENTAIGANA_LETTER_YO_6 = 0x0001B0EC;
const char32_t HENTAIGANA_LETTER_RA_1 = 0x0001B0ED;
const char32_t HENTAIGANA_LETTER_RA_2 = 0x0001B0EE;
const char32_t HENTAIGANA_LETTER_RA_3 = 0x0001B0EF;
const char32_t HENTAIGANA_LETTER_RA_4 = 0x0001B0F0;
const char32_t HENTAIGANA_LETTER_RI_1 = 0x0001B0F1;
const char32_t HENTAIGANA_LETTER_RI_2 = 0x0001B0F2;
const char32_t HENTAIGANA_LETTER_RI_3 = 0x0001B0F3;
const char32_t HENTAIGANA_LETTER_RI_4 = 0x0001B0F4;
const char32_t HENTAIGANA_LETTER_RI_5 = 0x0001B0F5;
const char32_t HENTAIGANA_LETTER_RI_6 = 0x0001B0F6;
const char32_t HENTAIGANA_LETTER_RI_7 = 0x0001B0F7;
const char32_t HENTAIGANA_LETTER_RU_1 = 0x0001B0F8;
const char32_t HENTAIGANA_LETTER_RU_2 = 0x0001B0F9;
const char32_t HENTAIGANA_LETTER_RU_3 = 0x0001B0FA;
const char32_t HENTAIGANA_LETTER_RU_4 = 0x0001B0FB;
const char32_t HENTAIGANA_LETTER_RU_5 = 0x0001B0FC;
const char32_t HENTAIGANA_LETTER_RU_6 = 0x0001B0FD;
const char32_t HENTAIGANA_LETTER_RE_1 = 0x0001B0FE;
const char32_t HENTAIGANA_LETTER_RE_2 = 0x0001B0FF;
const char32_t HENTAIGANA_LETTER_RE_3 = 0x0001B100;
const char32_t HENTAIGANA_LETTER_RE_4 = 0x0001B101;
const char32_t HENTAIGANA_LETTER_RO_1 = 0x0001B102;
const char32_t HENTAIGANA_LETTER_RO_2 = 0x0001B103;
const char32_t HENTAIGANA_LETTER_RO_3 = 0x0001B104;
const char32_t HENTAIGANA_LETTER_RO_4 = 0x0001B105;
const char32_t HENTAIGANA_LETTER_RO_5 = 0x0001B106;
const char32_t HENTAIGANA_LETTER_RO_6 = 0x0001B107;
const char32_t HENTAIGANA_LETTER_WA_1 = 0x0001B108;
const char32_t HENTAIGANA_LETTER_WA_2 = 0x0001B109;
const char32_t HENTAIGANA_LETTER_WA_3 = 0x0001B10A;
const char32_t HENTAIGANA_LETTER_WA_4 = 0x0001B10B;
const char32_t HENTAIGANA_LETTER_WA_5 = 0x0001B10C;
const char32_t HENTAIGANA_LETTER_WI_1 = 0x0001B10D;
const char32_t HENTAIGANA_LETTER_WI_2 = 0x0001B10E;
const char32_t HENTAIGANA_LETTER_WI_3 = 0x0001B10F;
const char32_t HENTAIGANA_LETTER_WI_4 = 0x0001B110;
const char32_t HENTAIGANA_LETTER_WI_5 = 0x0001B111;
const char32_t HENTAIGANA_LETTER_WE_1 = 0x0001B112;
const char32_t HENTAIGANA_LETTER_WE_2 = 0x0001B113;
const char32_t HENTAIGANA_LETTER_WE_3 = 0x0001B114;
const char32_t HENTAIGANA_LETTER_WE_4 = 0x0001B115;
const char32_t HENTAIGANA_LETTER_WO_1 = 0x0001B116;
const char32_t HENTAIGANA_LETTER_WO_2 = 0x0001B117;
const char32_t HENTAIGANA_LETTER_WO_3 = 0x0001B118;
const char32_t HENTAIGANA_LETTER_WO_4 = 0x0001B119;
const char32_t HENTAIGANA_LETTER_WO_5 = 0x0001B11A;
const char32_t HENTAIGANA_LETTER_WO_6 = 0x0001B11B;
const char32_t HENTAIGANA_LETTER_WO_7 = 0x0001B11C;
const char32_t HENTAIGANA_LETTER_N_MU_MO_1 = 0x0001B11D;
const char32_t HENTAIGANA_LETTER_N_MU_MO_2 = 0x0001B11E;
const char32_t HIRAGANA_LETTER_ARCHAIC_WU = 0x0001B11F;
const char32_t KATAKANA_LETTER_ARCHAIC_YI = 0x0001B120;
const char32_t KATAKANA_LETTER_ARCHAIC_YE = 0x0001B121;
const char32_t KATAKANA_LETTER_ARCHAIC_WU = 0x0001B122;
const char32_t HIRAGANA_LETTER_SMALL_KO = 0x0001B132;
const char32_t HIRAGANA_LETTER_SMALL_WI = 0x0001B150;
const char32_t HIRAGANA_LETTER_SMALL_WE = 0x0001B151;
const char32_t HIRAGANA_LETTER_SMALL_WO = 0x0001B152;
const char32_t KATAKANA_LETTER_SMALL_KO = 0x0001B155;
const char32_t KATAKANA_LETTER_SMALL_WI = 0x0001B164;
const char32_t KATAKANA_LETTER_SMALL_WE = 0x0001B165;
const char32_t KATAKANA_LETTER_SMALL_WO = 0x0001B166;
const char32_t KATAKANA_LETTER_SMALL_N = 0x0001B167;
const char32_t NUSHU_CHARACTER_1B170 = 0x0001B170;
const char32_t NUSHU_CHARACTER_1B171 = 0x0001B171;
const char32_t NUSHU_CHARACTER_1B172 = 0x0001B172;
const char32_t NUSHU_CHARACTER_1B173 = 0x0001B173;
const char32_t NUSHU_CHARACTER_1B174 = 0x0001B174;
const char32_t NUSHU_CHARACTER_1B175 = 0x0001B175;
const char32_t NUSHU_CHARACTER_1B176 = 0x0001B176;
const char32_t NUSHU_CHARACTER_1B177 = 0x0001B177;
const char32_t NUSHU_CHARACTER_1B178 = 0x0001B178;
const char32_t NUSHU_CHARACTER_1B179 = 0x0001B179;
const char32_t NUSHU_CHARACTER_1B17A = 0x0001B17A;
const char32_t NUSHU_CHARACTER_1B17B = 0x0001B17B;
const char32_t NUSHU_CHARACTER_1B17C = 0x0001B17C;
const char32_t NUSHU_CHARACTER_1B17D = 0x0001B17D;
const char32_t NUSHU_CHARACTER_1B17E = 0x0001B17E;
const char32_t NUSHU_CHARACTER_1B17F = 0x0001B17F;
const char32_t NUSHU_CHARACTER_1B180 = 0x0001B180;
const char32_t NUSHU_CHARACTER_1B181 = 0x0001B181;
const char32_t NUSHU_CHARACTER_1B182 = 0x0001B182;
const char32_t NUSHU_CHARACTER_1B183 = 0x0001B183;
const char32_t NUSHU_CHARACTER_1B184 = 0x0001B184;
const char32_t NUSHU_CHARACTER_1B185 = 0x0001B185;
const char32_t NUSHU_CHARACTER_1B186 = 0x0001B186;
const char32_t NUSHU_CHARACTER_1B187 = 0x0001B187;
const char32_t NUSHU_CHARACTER_1B188 = 0x0001B188;
const char32_t NUSHU_CHARACTER_1B189 = 0x0001B189;
const char32_t NUSHU_CHARACTER_1B18A = 0x0001B18A;
const char32_t NUSHU_CHARACTER_1B18B = 0x0001B18B;
const char32_t NUSHU_CHARACTER_1B18C = 0x0001B18C;
const char32_t NUSHU_CHARACTER_1B18D = 0x0001B18D;
const char32_t NUSHU_CHARACTER_1B18E = 0x0001B18E;
const char32_t NUSHU_CHARACTER_1B18F = 0x0001B18F;
const char32_t NUSHU_CHARACTER_1B190 = 0x0001B190;
const char32_t NUSHU_CHARACTER_1B191 = 0x0001B191;
const char32_t NUSHU_CHARACTER_1B192 = 0x0001B192;
const char32_t NUSHU_CHARACTER_1B193 = 0x0001B193;
const char32_t NUSHU_CHARACTER_1B194 = 0x0001B194;
const char32_t NUSHU_CHARACTER_1B195 = 0x0001B195;
const char32_t NUSHU_CHARACTER_1B196 = 0x0001B196;
const char32_t NUSHU_CHARACTER_1B197 = 0x0001B197;
const char32_t NUSHU_CHARACTER_1B198 = 0x0001B198;
const char32_t NUSHU_CHARACTER_1B199 = 0x0001B199;
const char32_t NUSHU_CHARACTER_1B19A = 0x0001B19A;
const char32_t NUSHU_CHARACTER_1B19B = 0x0001B19B;
const char32_t NUSHU_CHARACTER_1B19C = 0x0001B19C;
const char32_t NUSHU_CHARACTER_1B19D = 0x0001B19D;
const char32_t NUSHU_CHARACTER_1B19E = 0x0001B19E;
const char32_t NUSHU_CHARACTER_1B19F = 0x0001B19F;
const char32_t NUSHU_CHARACTER_1B1A0 = 0x0001B1A0;
const char32_t NUSHU_CHARACTER_1B1A1 = 0x0001B1A1;
const char32_t NUSHU_CHARACTER_1B1A2 = 0x0001B1A2;
const char32_t NUSHU_CHARACTER_1B1A3 = 0x0001B1A3;
const char32_t NUSHU_CHARACTER_1B1A4 = 0x0001B1A4;
const char32_t NUSHU_CHARACTER_1B1A5 = 0x0001B1A5;
const char32_t NUSHU_CHARACTER_1B1A6 = 0x0001B1A6;
const char32_t NUSHU_CHARACTER_1B1A7 = 0x0001B1A7;
const char32_t NUSHU_CHARACTER_1B1A8 = 0x0001B1A8;
const char32_t NUSHU_CHARACTER_1B1A9 = 0x0001B1A9;
const char32_t NUSHU_CHARACTER_1B1AA = 0x0001B1AA;
const char32_t NUSHU_CHARACTER_1B1AB = 0x0001B1AB;
const char32_t NUSHU_CHARACTER_1B1AC = 0x0001B1AC;
const char32_t NUSHU_CHARACTER_1B1AD = 0x0001B1AD;
const char32_t NUSHU_CHARACTER_1B1AE = 0x0001B1AE;
const char32_t NUSHU_CHARACTER_1B1AF = 0x0001B1AF;
const char32_t NUSHU_CHARACTER_1B1B0 = 0x0001B1B0;
const char32_t NUSHU_CHARACTER_1B1B1 = 0x0001B1B1;
const char32_t NUSHU_CHARACTER_1B1B2 = 0x0001B1B2;
const char32_t NUSHU_CHARACTER_1B1B3 = 0x0001B1B3;
const char32_t NUSHU_CHARACTER_1B1B4 = 0x0001B1B4;
const char32_t NUSHU_CHARACTER_1B1B5 = 0x0001B1B5;
const char32_t NUSHU_CHARACTER_1B1B6 = 0x0001B1B6;
const char32_t NUSHU_CHARACTER_1B1B7 = 0x0001B1B7;
const char32_t NUSHU_CHARACTER_1B1B8 = 0x0001B1B8;
const char32_t NUSHU_CHARACTER_1B1B9 = 0x0001B1B9;
const char32_t NUSHU_CHARACTER_1B1BA = 0x0001B1BA;
const char32_t NUSHU_CHARACTER_1B1BB = 0x0001B1BB;
const char32_t NUSHU_CHARACTER_1B1BC = 0x0001B1BC;
const char32_t NUSHU_CHARACTER_1B1BD = 0x0001B1BD;
const char32_t NUSHU_CHARACTER_1B1BE = 0x0001B1BE;
const char32_t NUSHU_CHARACTER_1B1BF = 0x0001B1BF;
const char32_t NUSHU_CHARACTER_1B1C0 = 0x0001B1C0;
const char32_t NUSHU_CHARACTER_1B1C1 = 0x0001B1C1;
const char32_t NUSHU_CHARACTER_1B1C2 = 0x0001B1C2;
const char32_t NUSHU_CHARACTER_1B1C3 = 0x0001B1C3;
const char32_t NUSHU_CHARACTER_1B1C4 = 0x0001B1C4;
const char32_t NUSHU_CHARACTER_1B1C5 = 0x0001B1C5;
const char32_t NUSHU_CHARACTER_1B1C6 = 0x0001B1C6;
const char32_t NUSHU_CHARACTER_1B1C7 = 0x0001B1C7;
const char32_t NUSHU_CHARACTER_1B1C8 = 0x0001B1C8;
const char32_t NUSHU_CHARACTER_1B1C9 = 0x0001B1C9;
const char32_t NUSHU_CHARACTER_1B1CA = 0x0001B1CA;
const char32_t NUSHU_CHARACTER_1B1CB = 0x0001B1CB;
const char32_t NUSHU_CHARACTER_1B1CC = 0x0001B1CC;
const char32_t NUSHU_CHARACTER_1B1CD = 0x0001B1CD;
const char32_t NUSHU_CHARACTER_1B1CE = 0x0001B1CE;
const char32_t NUSHU_CHARACTER_1B1CF = 0x0001B1CF;
const char32_t NUSHU_CHARACTER_1B1D0 = 0x0001B1D0;
const char32_t NUSHU_CHARACTER_1B1D1 = 0x0001B1D1;
const char32_t NUSHU_CHARACTER_1B1D2 = 0x0001B1D2;
const char32_t NUSHU_CHARACTER_1B1D3 = 0x0001B1D3;
const char32_t NUSHU_CHARACTER_1B1D4 = 0x0001B1D4;
const char32_t NUSHU_CHARACTER_1B1D5 = 0x0001B1D5;
const char32_t NUSHU_CHARACTER_1B1D6 = 0x0001B1D6;
const char32_t NUSHU_CHARACTER_1B1D7 = 0x0001B1D7;
const char32_t NUSHU_CHARACTER_1B1D8 = 0x0001B1D8;
const char32_t NUSHU_CHARACTER_1B1D9 = 0x0001B1D9;
const char32_t NUSHU_CHARACTER_1B1DA = 0x0001B1DA;
const char32_t NUSHU_CHARACTER_1B1DB = 0x0001B1DB;
const char32_t NUSHU_CHARACTER_1B1DC = 0x0001B1DC;
const char32_t NUSHU_CHARACTER_1B1DD = 0x0001B1DD;
const char32_t NUSHU_CHARACTER_1B1DE = 0x0001B1DE;
const char32_t NUSHU_CHARACTER_1B1DF = 0x0001B1DF;
const char32_t NUSHU_CHARACTER_1B1E0 = 0x0001B1E0;
const char32_t NUSHU_CHARACTER_1B1E1 = 0x0001B1E1;
const char32_t NUSHU_CHARACTER_1B1E2 = 0x0001B1E2;
const char32_t NUSHU_CHARACTER_1B1E3 = 0x0001B1E3;
const char32_t NUSHU_CHARACTER_1B1E4 = 0x0001B1E4;
const char32_t NUSHU_CHARACTER_1B1E5 = 0x0001B1E5;
const char32_t NUSHU_CHARACTER_1B1E6 = 0x0001B1E6;
const char32_t NUSHU_CHARACTER_1B1E7 = 0x0001B1E7;
const char32_t NUSHU_CHARACTER_1B1E8 = 0x0001B1E8;
const char32_t NUSHU_CHARACTER_1B1E9 = 0x0001B1E9;
const char32_t NUSHU_CHARACTER_1B1EA = 0x0001B1EA;
const char32_t NUSHU_CHARACTER_1B1EB = 0x0001B1EB;
const char32_t NUSHU_CHARACTER_1B1EC = 0x0001B1EC;
const char32_t NUSHU_CHARACTER_1B1ED = 0x0001B1ED;
const char32_t NUSHU_CHARACTER_1B1EE = 0x0001B1EE;
const char32_t NUSHU_CHARACTER_1B1EF = 0x0001B1EF;
const char32_t NUSHU_CHARACTER_1B1F0 = 0x0001B1F0;
const char32_t NUSHU_CHARACTER_1B1F1 = 0x0001B1F1;
const char32_t NUSHU_CHARACTER_1B1F2 = 0x0001B1F2;
const char32_t NUSHU_CHARACTER_1B1F3 = 0x0001B1F3;
const char32_t NUSHU_CHARACTER_1B1F4 = 0x0001B1F4;
const char32_t NUSHU_CHARACTER_1B1F5 = 0x0001B1F5;
const char32_t NUSHU_CHARACTER_1B1F6 = 0x0001B1F6;
const char32_t NUSHU_CHARACTER_1B1F7 = 0x0001B1F7;
const char32_t NUSHU_CHARACTER_1B1F8 = 0x0001B1F8;
const char32_t NUSHU_CHARACTER_1B1F9 = 0x0001B1F9;
const char32_t NUSHU_CHARACTER_1B1FA = 0x0001B1FA;
const char32_t NUSHU_CHARACTER_1B1FB = 0x0001B1FB;
const char32_t NUSHU_CHARACTER_1B1FC = 0x0001B1FC;
const char32_t NUSHU_CHARACTER_1B1FD = 0x0001B1FD;
const char32_t NUSHU_CHARACTER_1B1FE = 0x0001B1FE;
const char32_t NUSHU_CHARACTER_1B1FF = 0x0001B1FF;
const char32_t NUSHU_CHARACTER_1B200 = 0x0001B200;
const char32_t NUSHU_CHARACTER_1B201 = 0x0001B201;
const char32_t NUSHU_CHARACTER_1B202 = 0x0001B202;
const char32_t NUSHU_CHARACTER_1B203 = 0x0001B203;
const char32_t NUSHU_CHARACTER_1B204 = 0x0001B204;
const char32_t NUSHU_CHARACTER_1B205 = 0x0001B205;
const char32_t NUSHU_CHARACTER_1B206 = 0x0001B206;
const char32_t NUSHU_CHARACTER_1B207 = 0x0001B207;
const char32_t NUSHU_CHARACTER_1B208 = 0x0001B208;
const char32_t NUSHU_CHARACTER_1B209 = 0x0001B209;
const char32_t NUSHU_CHARACTER_1B20A = 0x0001B20A;
const char32_t NUSHU_CHARACTER_1B20B = 0x0001B20B;
const char32_t NUSHU_CHARACTER_1B20C = 0x0001B20C;
const char32_t NUSHU_CHARACTER_1B20D = 0x0001B20D;
const char32_t NUSHU_CHARACTER_1B20E = 0x0001B20E;
const char32_t NUSHU_CHARACTER_1B20F = 0x0001B20F;
const char32_t NUSHU_CHARACTER_1B210 = 0x0001B210;
const char32_t NUSHU_CHARACTER_1B211 = 0x0001B211;
const char32_t NUSHU_CHARACTER_1B212 = 0x0001B212;
const char32_t NUSHU_CHARACTER_1B213 = 0x0001B213;
const char32_t NUSHU_CHARACTER_1B214 = 0x0001B214;
const char32_t NUSHU_CHARACTER_1B215 = 0x0001B215;
const char32_t NUSHU_CHARACTER_1B216 = 0x0001B216;
const char32_t NUSHU_CHARACTER_1B217 = 0x0001B217;
const char32_t NUSHU_CHARACTER_1B218 = 0x0001B218;
const char32_t NUSHU_CHARACTER_1B219 = 0x0001B219;
const char32_t NUSHU_CHARACTER_1B21A = 0x0001B21A;
const char32_t NUSHU_CHARACTER_1B21B = 0x0001B21B;
const char32_t NUSHU_CHARACTER_1B21C = 0x0001B21C;
const char32_t NUSHU_CHARACTER_1B21D = 0x0001B21D;
const char32_t NUSHU_CHARACTER_1B21E = 0x0001B21E;
const char32_t NUSHU_CHARACTER_1B21F = 0x0001B21F;
const char32_t NUSHU_CHARACTER_1B220 = 0x0001B220;
const char32_t NUSHU_CHARACTER_1B221 = 0x0001B221;
const char32_t NUSHU_CHARACTER_1B222 = 0x0001B222;
const char32_t NUSHU_CHARACTER_1B223 = 0x0001B223;
const char32_t NUSHU_CHARACTER_1B224 = 0x0001B224;
const char32_t NUSHU_CHARACTER_1B225 = 0x0001B225;
const char32_t NUSHU_CHARACTER_1B226 = 0x0001B226;
const char32_t NUSHU_CHARACTER_1B227 = 0x0001B227;
const char32_t NUSHU_CHARACTER_1B228 = 0x0001B228;
const char32_t NUSHU_CHARACTER_1B229 = 0x0001B229;
const char32_t NUSHU_CHARACTER_1B22A = 0x0001B22A;
const char32_t NUSHU_CHARACTER_1B22B = 0x0001B22B;
const char32_t NUSHU_CHARACTER_1B22C = 0x0001B22C;
const char32_t NUSHU_CHARACTER_1B22D = 0x0001B22D;
const char32_t NUSHU_CHARACTER_1B22E = 0x0001B22E;
const char32_t NUSHU_CHARACTER_1B22F = 0x0001B22F;
const char32_t NUSHU_CHARACTER_1B230 = 0x0001B230;
const char32_t NUSHU_CHARACTER_1B231 = 0x0001B231;
const char32_t NUSHU_CHARACTER_1B232 = 0x0001B232;
const char32_t NUSHU_CHARACTER_1B233 = 0x0001B233;
const char32_t NUSHU_CHARACTER_1B234 = 0x0001B234;
const char32_t NUSHU_CHARACTER_1B235 = 0x0001B235;
const char32_t NUSHU_CHARACTER_1B236 = 0x0001B236;
const char32_t NUSHU_CHARACTER_1B237 = 0x0001B237;
const char32_t NUSHU_CHARACTER_1B238 = 0x0001B238;
const char32_t NUSHU_CHARACTER_1B239 = 0x0001B239;
const char32_t NUSHU_CHARACTER_1B23A = 0x0001B23A;
const char32_t NUSHU_CHARACTER_1B23B = 0x0001B23B;
const char32_t NUSHU_CHARACTER_1B23C = 0x0001B23C;
const char32_t NUSHU_CHARACTER_1B23D = 0x0001B23D;
const char32_t NUSHU_CHARACTER_1B23E = 0x0001B23E;
const char32_t NUSHU_CHARACTER_1B23F = 0x0001B23F;
const char32_t NUSHU_CHARACTER_1B240 = 0x0001B240;
const char32_t NUSHU_CHARACTER_1B241 = 0x0001B241;
const char32_t NUSHU_CHARACTER_1B242 = 0x0001B242;
const char32_t NUSHU_CHARACTER_1B243 = 0x0001B243;
const char32_t NUSHU_CHARACTER_1B244 = 0x0001B244;
const char32_t NUSHU_CHARACTER_1B245 = 0x0001B245;
const char32_t NUSHU_CHARACTER_1B246 = 0x0001B246;
const char32_t NUSHU_CHARACTER_1B247 = 0x0001B247;
const char32_t NUSHU_CHARACTER_1B248 = 0x0001B248;
const char32_t NUSHU_CHARACTER_1B249 = 0x0001B249;
const char32_t NUSHU_CHARACTER_1B24A = 0x0001B24A;
const char32_t NUSHU_CHARACTER_1B24B = 0x0001B24B;
const char32_t NUSHU_CHARACTER_1B24C = 0x0001B24C;
const char32_t NUSHU_CHARACTER_1B24D = 0x0001B24D;
const char32_t NUSHU_CHARACTER_1B24E = 0x0001B24E;
const char32_t NUSHU_CHARACTER_1B24F = 0x0001B24F;
const char32_t NUSHU_CHARACTER_1B250 = 0x0001B250;
const char32_t NUSHU_CHARACTER_1B251 = 0x0001B251;
const char32_t NUSHU_CHARACTER_1B252 = 0x0001B252;
const char32_t NUSHU_CHARACTER_1B253 = 0x0001B253;
const char32_t NUSHU_CHARACTER_1B254 = 0x0001B254;
const char32_t NUSHU_CHARACTER_1B255 = 0x0001B255;
const char32_t NUSHU_CHARACTER_1B256 = 0x0001B256;
const char32_t NUSHU_CHARACTER_1B257 = 0x0001B257;
const char32_t NUSHU_CHARACTER_1B258 = 0x0001B258;
const char32_t NUSHU_CHARACTER_1B259 = 0x0001B259;
const char32_t NUSHU_CHARACTER_1B25A = 0x0001B25A;
const char32_t NUSHU_CHARACTER_1B25B = 0x0001B25B;
const char32_t NUSHU_CHARACTER_1B25C = 0x0001B25C;
const char32_t NUSHU_CHARACTER_1B25D = 0x0001B25D;
const char32_t NUSHU_CHARACTER_1B25E = 0x0001B25E;
const char32_t NUSHU_CHARACTER_1B25F = 0x0001B25F;
const char32_t NUSHU_CHARACTER_1B260 = 0x0001B260;
const char32_t NUSHU_CHARACTER_1B261 = 0x0001B261;
const char32_t NUSHU_CHARACTER_1B262 = 0x0001B262;
const char32_t NUSHU_CHARACTER_1B263 = 0x0001B263;
const char32_t NUSHU_CHARACTER_1B264 = 0x0001B264;
const char32_t NUSHU_CHARACTER_1B265 = 0x0001B265;
const char32_t NUSHU_CHARACTER_1B266 = 0x0001B266;
const char32_t NUSHU_CHARACTER_1B267 = 0x0001B267;
const char32_t NUSHU_CHARACTER_1B268 = 0x0001B268;
const char32_t NUSHU_CHARACTER_1B269 = 0x0001B269;
const char32_t NUSHU_CHARACTER_1B26A = 0x0001B26A;
const char32_t NUSHU_CHARACTER_1B26B = 0x0001B26B;
const char32_t NUSHU_CHARACTER_1B26C = 0x0001B26C;
const char32_t NUSHU_CHARACTER_1B26D = 0x0001B26D;
const char32_t NUSHU_CHARACTER_1B26E = 0x0001B26E;
const char32_t NUSHU_CHARACTER_1B26F = 0x0001B26F;
const char32_t NUSHU_CHARACTER_1B270 = 0x0001B270;
const char32_t NUSHU_CHARACTER_1B271 = 0x0001B271;
const char32_t NUSHU_CHARACTER_1B272 = 0x0001B272;
const char32_t NUSHU_CHARACTER_1B273 = 0x0001B273;
const char32_t NUSHU_CHARACTER_1B274 = 0x0001B274;
const char32_t NUSHU_CHARACTER_1B275 = 0x0001B275;
const char32_t NUSHU_CHARACTER_1B276 = 0x0001B276;
const char32_t NUSHU_CHARACTER_1B277 = 0x0001B277;
const char32_t NUSHU_CHARACTER_1B278 = 0x0001B278;
const char32_t NUSHU_CHARACTER_1B279 = 0x0001B279;
const char32_t NUSHU_CHARACTER_1B27A = 0x0001B27A;
const char32_t NUSHU_CHARACTER_1B27B = 0x0001B27B;
const char32_t NUSHU_CHARACTER_1B27C = 0x0001B27C;
const char32_t NUSHU_CHARACTER_1B27D = 0x0001B27D;
const char32_t NUSHU_CHARACTER_1B27E = 0x0001B27E;
const char32_t NUSHU_CHARACTER_1B27F = 0x0001B27F;
const char32_t NUSHU_CHARACTER_1B280 = 0x0001B280;
const char32_t NUSHU_CHARACTER_1B281 = 0x0001B281;
const char32_t NUSHU_CHARACTER_1B282 = 0x0001B282;
const char32_t NUSHU_CHARACTER_1B283 = 0x0001B283;
const char32_t NUSHU_CHARACTER_1B284 = 0x0001B284;
const char32_t NUSHU_CHARACTER_1B285 = 0x0001B285;
const char32_t NUSHU_CHARACTER_1B286 = 0x0001B286;
const char32_t NUSHU_CHARACTER_1B287 = 0x0001B287;
const char32_t NUSHU_CHARACTER_1B288 = 0x0001B288;
const char32_t NUSHU_CHARACTER_1B289 = 0x0001B289;
const char32_t NUSHU_CHARACTER_1B28A = 0x0001B28A;
const char32_t NUSHU_CHARACTER_1B28B = 0x0001B28B;
const char32_t NUSHU_CHARACTER_1B28C = 0x0001B28C;
const char32_t NUSHU_CHARACTER_1B28D = 0x0001B28D;
const char32_t NUSHU_CHARACTER_1B28E = 0x0001B28E;
const char32_t NUSHU_CHARACTER_1B28F = 0x0001B28F;
const char32_t NUSHU_CHARACTER_1B290 = 0x0001B290;
const char32_t NUSHU_CHARACTER_1B291 = 0x0001B291;
const char32_t NUSHU_CHARACTER_1B292 = 0x0001B292;
const char32_t NUSHU_CHARACTER_1B293 = 0x0001B293;
const char32_t NUSHU_CHARACTER_1B294 = 0x0001B294;
const char32_t NUSHU_CHARACTER_1B295 = 0x0001B295;
const char32_t NUSHU_CHARACTER_1B296 = 0x0001B296;
const char32_t NUSHU_CHARACTER_1B297 = 0x0001B297;
const char32_t NUSHU_CHARACTER_1B298 = 0x0001B298;
const char32_t NUSHU_CHARACTER_1B299 = 0x0001B299;
const char32_t NUSHU_CHARACTER_1B29A = 0x0001B29A;
const char32_t NUSHU_CHARACTER_1B29B = 0x0001B29B;
const char32_t NUSHU_CHARACTER_1B29C = 0x0001B29C;
const char32_t NUSHU_CHARACTER_1B29D = 0x0001B29D;
const char32_t NUSHU_CHARACTER_1B29E = 0x0001B29E;
const char32_t NUSHU_CHARACTER_1B29F = 0x0001B29F;
const char32_t NUSHU_CHARACTER_1B2A0 = 0x0001B2A0;
const char32_t NUSHU_CHARACTER_1B2A1 = 0x0001B2A1;
const char32_t NUSHU_CHARACTER_1B2A2 = 0x0001B2A2;
const char32_t NUSHU_CHARACTER_1B2A3 = 0x0001B2A3;
const char32_t NUSHU_CHARACTER_1B2A4 = 0x0001B2A4;
const char32_t NUSHU_CHARACTER_1B2A5 = 0x0001B2A5;
const char32_t NUSHU_CHARACTER_1B2A6 = 0x0001B2A6;
const char32_t NUSHU_CHARACTER_1B2A7 = 0x0001B2A7;
const char32_t NUSHU_CHARACTER_1B2A8 = 0x0001B2A8;
const char32_t NUSHU_CHARACTER_1B2A9 = 0x0001B2A9;
const char32_t NUSHU_CHARACTER_1B2AA = 0x0001B2AA;
const char32_t NUSHU_CHARACTER_1B2AB = 0x0001B2AB;
const char32_t NUSHU_CHARACTER_1B2AC = 0x0001B2AC;
const char32_t NUSHU_CHARACTER_1B2AD = 0x0001B2AD;
const char32_t NUSHU_CHARACTER_1B2AE = 0x0001B2AE;
const char32_t NUSHU_CHARACTER_1B2AF = 0x0001B2AF;
const char32_t NUSHU_CHARACTER_1B2B0 = 0x0001B2B0;
const char32_t NUSHU_CHARACTER_1B2B1 = 0x0001B2B1;
const char32_t NUSHU_CHARACTER_1B2B2 = 0x0001B2B2;
const char32_t NUSHU_CHARACTER_1B2B3 = 0x0001B2B3;
const char32_t NUSHU_CHARACTER_1B2B4 = 0x0001B2B4;
const char32_t NUSHU_CHARACTER_1B2B5 = 0x0001B2B5;
const char32_t NUSHU_CHARACTER_1B2B6 = 0x0001B2B6;
const char32_t NUSHU_CHARACTER_1B2B7 = 0x0001B2B7;
const char32_t NUSHU_CHARACTER_1B2B8 = 0x0001B2B8;
const char32_t NUSHU_CHARACTER_1B2B9 = 0x0001B2B9;
const char32_t NUSHU_CHARACTER_1B2BA = 0x0001B2BA;
const char32_t NUSHU_CHARACTER_1B2BB = 0x0001B2BB;
const char32_t NUSHU_CHARACTER_1B2BC = 0x0001B2BC;
const char32_t NUSHU_CHARACTER_1B2BD = 0x0001B2BD;
const char32_t NUSHU_CHARACTER_1B2BE = 0x0001B2BE;
const char32_t NUSHU_CHARACTER_1B2BF = 0x0001B2BF;
const char32_t NUSHU_CHARACTER_1B2C0 = 0x0001B2C0;
const char32_t NUSHU_CHARACTER_1B2C1 = 0x0001B2C1;
const char32_t NUSHU_CHARACTER_1B2C2 = 0x0001B2C2;
const char32_t NUSHU_CHARACTER_1B2C3 = 0x0001B2C3;
const char32_t NUSHU_CHARACTER_1B2C4 = 0x0001B2C4;
const char32_t NUSHU_CHARACTER_1B2C5 = 0x0001B2C5;
const char32_t NUSHU_CHARACTER_1B2C6 = 0x0001B2C6;
const char32_t NUSHU_CHARACTER_1B2C7 = 0x0001B2C7;
const char32_t NUSHU_CHARACTER_1B2C8 = 0x0001B2C8;
const char32_t NUSHU_CHARACTER_1B2C9 = 0x0001B2C9;
const char32_t NUSHU_CHARACTER_1B2CA = 0x0001B2CA;
const char32_t NUSHU_CHARACTER_1B2CB = 0x0001B2CB;
const char32_t NUSHU_CHARACTER_1B2CC = 0x0001B2CC;
const char32_t NUSHU_CHARACTER_1B2CD = 0x0001B2CD;
const char32_t NUSHU_CHARACTER_1B2CE = 0x0001B2CE;
const char32_t NUSHU_CHARACTER_1B2CF = 0x0001B2CF;
const char32_t NUSHU_CHARACTER_1B2D0 = 0x0001B2D0;
const char32_t NUSHU_CHARACTER_1B2D1 = 0x0001B2D1;
const char32_t NUSHU_CHARACTER_1B2D2 = 0x0001B2D2;
const char32_t NUSHU_CHARACTER_1B2D3 = 0x0001B2D3;
const char32_t NUSHU_CHARACTER_1B2D4 = 0x0001B2D4;
const char32_t NUSHU_CHARACTER_1B2D5 = 0x0001B2D5;
const char32_t NUSHU_CHARACTER_1B2D6 = 0x0001B2D6;
const char32_t NUSHU_CHARACTER_1B2D7 = 0x0001B2D7;
const char32_t NUSHU_CHARACTER_1B2D8 = 0x0001B2D8;
const char32_t NUSHU_CHARACTER_1B2D9 = 0x0001B2D9;
const char32_t NUSHU_CHARACTER_1B2DA = 0x0001B2DA;
const char32_t NUSHU_CHARACTER_1B2DB = 0x0001B2DB;
const char32_t NUSHU_CHARACTER_1B2DC = 0x0001B2DC;
const char32_t NUSHU_CHARACTER_1B2DD = 0x0001B2DD;
const char32_t NUSHU_CHARACTER_1B2DE = 0x0001B2DE;
const char32_t NUSHU_CHARACTER_1B2DF = 0x0001B2DF;
const char32_t NUSHU_CHARACTER_1B2E0 = 0x0001B2E0;
const char32_t NUSHU_CHARACTER_1B2E1 = 0x0001B2E1;
const char32_t NUSHU_CHARACTER_1B2E2 = 0x0001B2E2;
const char32_t NUSHU_CHARACTER_1B2E3 = 0x0001B2E3;
const char32_t NUSHU_CHARACTER_1B2E4 = 0x0001B2E4;
const char32_t NUSHU_CHARACTER_1B2E5 = 0x0001B2E5;
const char32_t NUSHU_CHARACTER_1B2E6 = 0x0001B2E6;
const char32_t NUSHU_CHARACTER_1B2E7 = 0x0001B2E7;
const char32_t NUSHU_CHARACTER_1B2E8 = 0x0001B2E8;
const char32_t NUSHU_CHARACTER_1B2E9 = 0x0001B2E9;
const char32_t NUSHU_CHARACTER_1B2EA = 0x0001B2EA;
const char32_t NUSHU_CHARACTER_1B2EB = 0x0001B2EB;
const char32_t NUSHU_CHARACTER_1B2EC = 0x0001B2EC;
const char32_t NUSHU_CHARACTER_1B2ED = 0x0001B2ED;
const char32_t NUSHU_CHARACTER_1B2EE = 0x0001B2EE;
const char32_t NUSHU_CHARACTER_1B2EF = 0x0001B2EF;
const char32_t NUSHU_CHARACTER_1B2F0 = 0x0001B2F0;
const char32_t NUSHU_CHARACTER_1B2F1 = 0x0001B2F1;
const char32_t NUSHU_CHARACTER_1B2F2 = 0x0001B2F2;
const char32_t NUSHU_CHARACTER_1B2F3 = 0x0001B2F3;
const char32_t NUSHU_CHARACTER_1B2F4 = 0x0001B2F4;
const char32_t NUSHU_CHARACTER_1B2F5 = 0x0001B2F5;
const char32_t NUSHU_CHARACTER_1B2F6 = 0x0001B2F6;
const char32_t NUSHU_CHARACTER_1B2F7 = 0x0001B2F7;
const char32_t NUSHU_CHARACTER_1B2F8 = 0x0001B2F8;
const char32_t NUSHU_CHARACTER_1B2F9 = 0x0001B2F9;
const char32_t NUSHU_CHARACTER_1B2FA = 0x0001B2FA;
const char32_t NUSHU_CHARACTER_1B2FB = 0x0001B2FB;
const char32_t DUPLOYAN_LETTER_H = 0x0001BC00;
const char32_t DUPLOYAN_LETTER_X = 0x0001BC01;
const char32_t DUPLOYAN_LETTER_P = 0x0001BC02;
const char32_t DUPLOYAN_LETTER_T = 0x0001BC03;
const char32_t DUPLOYAN_LETTER_F = 0x0001BC04;
const char32_t DUPLOYAN_LETTER_K = 0x0001BC05;
const char32_t DUPLOYAN_LETTER_L = 0x0001BC06;
const char32_t DUPLOYAN_LETTER_B = 0x0001BC07;
const char32_t DUPLOYAN_LETTER_D = 0x0001BC08;
const char32_t DUPLOYAN_LETTER_V = 0x0001BC09;
const char32_t DUPLOYAN_LETTER_G = 0x0001BC0A;
const char32_t DUPLOYAN_LETTER_R = 0x0001BC0B;
const char32_t DUPLOYAN_LETTER_P_N = 0x0001BC0C;
const char32_t DUPLOYAN_LETTER_D_S = 0x0001BC0D;
const char32_t DUPLOYAN_LETTER_F_N = 0x0001BC0E;
const char32_t DUPLOYAN_LETTER_K_M = 0x0001BC0F;
const char32_t DUPLOYAN_LETTER_R_S = 0x0001BC10;
const char32_t DUPLOYAN_LETTER_TH = 0x0001BC11;
const char32_t DUPLOYAN_LETTER_SLOAN_DH = 0x0001BC12;
const char32_t DUPLOYAN_LETTER_DH = 0x0001BC13;
const char32_t DUPLOYAN_LETTER_KK = 0x0001BC14;
const char32_t DUPLOYAN_LETTER_SLOAN_J = 0x0001BC15;
const char32_t DUPLOYAN_LETTER_HL = 0x0001BC16;
const char32_t DUPLOYAN_LETTER_LH = 0x0001BC17;
const char32_t DUPLOYAN_LETTER_RH = 0x0001BC18;
const char32_t DUPLOYAN_LETTER_M = 0x0001BC19;
const char32_t DUPLOYAN_LETTER_N = 0x0001BC1A;
const char32_t DUPLOYAN_LETTER_J = 0x0001BC1B;
const char32_t DUPLOYAN_LETTER_S = 0x0001BC1C;
const char32_t DUPLOYAN_LETTER_M_N = 0x0001BC1D;
const char32_t DUPLOYAN_LETTER_N_M = 0x0001BC1E;
const char32_t DUPLOYAN_LETTER_J_M = 0x0001BC1F;
const char32_t DUPLOYAN_LETTER_S_J = 0x0001BC20;
const char32_t DUPLOYAN_LETTER_M_WITH_DOT = 0x0001BC21;
const char32_t DUPLOYAN_LETTER_N_WITH_DOT = 0x0001BC22;
const char32_t DUPLOYAN_LETTER_J_WITH_DOT = 0x0001BC23;
const char32_t DUPLOYAN_LETTER_J_WITH_DOTS_INSIDE_AND_ABOVE = 0x0001BC24;
const char32_t DUPLOYAN_LETTER_S_WITH_DOT = 0x0001BC25;
const char32_t DUPLOYAN_LETTER_S_WITH_DOT_BELOW = 0x0001BC26;
const char32_t DUPLOYAN_LETTER_M_S = 0x0001BC27;
const char32_t DUPLOYAN_LETTER_N_S = 0x0001BC28;
const char32_t DUPLOYAN_LETTER_J_S = 0x0001BC29;
const char32_t DUPLOYAN_LETTER_S_S = 0x0001BC2A;
const char32_t DUPLOYAN_LETTER_M_N_S = 0x0001BC2B;
const char32_t DUPLOYAN_LETTER_N_M_S = 0x0001BC2C;
const char32_t DUPLOYAN_LETTER_J_M_S = 0x0001BC2D;
const char32_t DUPLOYAN_LETTER_S_J_S = 0x0001BC2E;
const char32_t DUPLOYAN_LETTER_J_S_WITH_DOT = 0x0001BC2F;
const char32_t DUPLOYAN_LETTER_J_N = 0x0001BC30;
const char32_t DUPLOYAN_LETTER_J_N_S = 0x0001BC31;
const char32_t DUPLOYAN_LETTER_S_T = 0x0001BC32;
const char32_t DUPLOYAN_LETTER_S_T_R = 0x0001BC33;
const char32_t DUPLOYAN_LETTER_S_P = 0x0001BC34;
const char32_t DUPLOYAN_LETTER_S_P_R = 0x0001BC35;
const char32_t DUPLOYAN_LETTER_T_S = 0x0001BC36;
const char32_t DUPLOYAN_LETTER_T_R_S = 0x0001BC37;
const char32_t DUPLOYAN_LETTER_W = 0x0001BC38;
const char32_t DUPLOYAN_LETTER_WH = 0x0001BC39;
const char32_t DUPLOYAN_LETTER_W_R = 0x0001BC3A;
const char32_t DUPLOYAN_LETTER_S_N = 0x0001BC3B;
const char32_t DUPLOYAN_LETTER_S_M = 0x0001BC3C;
const char32_t DUPLOYAN_LETTER_K_R_S = 0x0001BC3D;
const char32_t DUPLOYAN_LETTER_G_R_S = 0x0001BC3E;
const char32_t DUPLOYAN_LETTER_S_K = 0x0001BC3F;
const char32_t DUPLOYAN_LETTER_S_K_R = 0x0001BC40;
const char32_t DUPLOYAN_LETTER_A = 0x0001BC41;
const char32_t DUPLOYAN_LETTER_SLOAN_OW = 0x0001BC42;
const char32_t DUPLOYAN_LETTER_OA = 0x0001BC43;
const char32_t DUPLOYAN_LETTER_O = 0x0001BC44;
const char32_t DUPLOYAN_LETTER_AOU = 0x0001BC45;
const char32_t DUPLOYAN_LETTER_I = 0x0001BC46;
const char32_t DUPLOYAN_LETTER_E = 0x0001BC47;
const char32_t DUPLOYAN_LETTER_IE = 0x0001BC48;
const char32_t DUPLOYAN_LETTER_SHORT_I = 0x0001BC49;
const char32_t DUPLOYAN_LETTER_UI = 0x0001BC4A;
const char32_t DUPLOYAN_LETTER_EE = 0x0001BC4B;
const char32_t DUPLOYAN_LETTER_SLOAN_EH = 0x0001BC4C;
const char32_t DUPLOYAN_LETTER_ROMANIAN_I = 0x0001BC4D;
const char32_t DUPLOYAN_LETTER_SLOAN_EE = 0x0001BC4E;
const char32_t DUPLOYAN_LETTER_LONG_I = 0x0001BC4F;
const char32_t DUPLOYAN_LETTER_YE = 0x0001BC50;
const char32_t DUPLOYAN_LETTER_U = 0x0001BC51;
const char32_t DUPLOYAN_LETTER_EU = 0x0001BC52;
const char32_t DUPLOYAN_LETTER_XW = 0x0001BC53;
const char32_t DUPLOYAN_LETTER_U_N = 0x0001BC54;
const char32_t DUPLOYAN_LETTER_LONG_U = 0x0001BC55;
const char32_t DUPLOYAN_LETTER_ROMANIAN_U = 0x0001BC56;
const char32_t DUPLOYAN_LETTER_UH = 0x0001BC57;
const char32_t DUPLOYAN_LETTER_SLOAN_U = 0x0001BC58;
const char32_t DUPLOYAN_LETTER_OOH = 0x0001BC59;
const char32_t DUPLOYAN_LETTER_OW = 0x0001BC5A;
const char32_t DUPLOYAN_LETTER_OU = 0x0001BC5B;
const char32_t DUPLOYAN_LETTER_WA = 0x0001BC5C;
const char32_t DUPLOYAN_LETTER_WO = 0x0001BC5D;
const char32_t DUPLOYAN_LETTER_WI = 0x0001BC5E;
const char32_t DUPLOYAN_LETTER_WEI = 0x0001BC5F;
const char32_t DUPLOYAN_LETTER_WOW = 0x0001BC60;
const char32_t DUPLOYAN_LETTER_NASAL_U = 0x0001BC61;
const char32_t DUPLOYAN_LETTER_NASAL_O = 0x0001BC62;
const char32_t DUPLOYAN_LETTER_NASAL_I = 0x0001BC63;
const char32_t DUPLOYAN_LETTER_NASAL_A = 0x0001BC64;
const char32_t DUPLOYAN_LETTER_PERNIN_AN = 0x0001BC65;
const char32_t DUPLOYAN_LETTER_PERNIN_AM = 0x0001BC66;
const char32_t DUPLOYAN_LETTER_SLOAN_EN = 0x0001BC67;
const char32_t DUPLOYAN_LETTER_SLOAN_AN = 0x0001BC68;
const char32_t DUPLOYAN_LETTER_SLOAN_ON = 0x0001BC69;
const char32_t DUPLOYAN_LETTER_VOCALIC_M = 0x0001BC6A;
const char32_t DUPLOYAN_AFFIX_LEFT_HORIZONTAL_SECANT = 0x0001BC70;
const char32_t DUPLOYAN_AFFIX_MID_HORIZONTAL_SECANT = 0x0001BC71;
const char32_t DUPLOYAN_AFFIX_RIGHT_HORIZONTAL_SECANT = 0x0001BC72;
const char32_t DUPLOYAN_AFFIX_LOW_VERTICAL_SECANT = 0x0001BC73;
const char32_t DUPLOYAN_AFFIX_MID_VERTICAL_SECANT = 0x0001BC74;
const char32_t DUPLOYAN_AFFIX_HIGH_VERTICAL_SECANT = 0x0001BC75;
const char32_t DUPLOYAN_AFFIX_ATTACHED_SECANT = 0x0001BC76;
const char32_t DUPLOYAN_AFFIX_ATTACHED_LEFT_TO_RIGHT_SECANT = 0x0001BC77;
const char32_t DUPLOYAN_AFFIX_ATTACHED_TANGENT = 0x0001BC78;
const char32_t DUPLOYAN_AFFIX_ATTACHED_TAIL = 0x0001BC79;
const char32_t DUPLOYAN_AFFIX_ATTACHED_E_HOOK = 0x0001BC7A;
const char32_t DUPLOYAN_AFFIX_ATTACHED_I_HOOK = 0x0001BC7B;
const char32_t DUPLOYAN_AFFIX_ATTACHED_TANGENT_HOOK = 0x0001BC7C;
const char32_t DUPLOYAN_AFFIX_HIGH_ACUTE = 0x0001BC80;
const char32_t DUPLOYAN_AFFIX_HIGH_TIGHT_ACUTE = 0x0001BC81;
const char32_t DUPLOYAN_AFFIX_HIGH_GRAVE = 0x0001BC82;
const char32_t DUPLOYAN_AFFIX_HIGH_LONG_GRAVE = 0x0001BC83;
const char32_t DUPLOYAN_AFFIX_HIGH_DOT = 0x0001BC84;
const char32_t DUPLOYAN_AFFIX_HIGH_CIRCLE = 0x0001BC85;
const char32_t DUPLOYAN_AFFIX_HIGH_LINE = 0x0001BC86;
const char32_t DUPLOYAN_AFFIX_HIGH_WAVE = 0x0001BC87;
const char32_t DUPLOYAN_AFFIX_HIGH_VERTICAL = 0x0001BC88;
const char32_t DUPLOYAN_AFFIX_LOW_ACUTE = 0x0001BC90;
const char32_t DUPLOYAN_AFFIX_LOW_TIGHT_ACUTE = 0x0001BC91;
const char32_t DUPLOYAN_AFFIX_LOW_GRAVE = 0x0001BC92;
const char32_t DUPLOYAN_AFFIX_LOW_LONG_GRAVE = 0x0001BC93;
const char32_t DUPLOYAN_AFFIX_LOW_DOT = 0x0001BC94;
const char32_t DUPLOYAN_AFFIX_LOW_CIRCLE = 0x0001BC95;
const char32_t DUPLOYAN_AFFIX_LOW_LINE = 0x0001BC96;
const char32_t DUPLOYAN_AFFIX_LOW_WAVE = 0x0001BC97;
const char32_t DUPLOYAN_AFFIX_LOW_VERTICAL = 0x0001BC98;
const char32_t DUPLOYAN_AFFIX_LOW_ARROW = 0x0001BC99;
const char32_t DUPLOYAN_SIGN_O_WITH_CROSS = 0x0001BC9C;
const char32_t DUPLOYAN_THICK_LETTER_SELECTOR = 0x0001BC9D;
const char32_t DUPLOYAN_DOUBLE_MARK = 0x0001BC9E;
const char32_t DUPLOYAN_PUNCTUATION_CHINOOK_FULL_STOP = 0x0001BC9F;
const char32_t SHORTHAND_FORMAT_LETTER_OVERLAP = 0x0001BCA0;
const char32_t SHORTHAND_FORMAT_CONTINUING_OVERLAP = 0x0001BCA1;
const char32_t SHORTHAND_FORMAT_DOWN_STEP = 0x0001BCA2;
const char32_t SHORTHAND_FORMAT_UP_STEP = 0x0001BCA3;
const char32_t UP_POINTING_GO_KART = 0x0001CC00;
const char32_t RIGHT_POINTING_GO_KART = 0x0001CC01;
const char32_t LEFT_POINTING_STICK_FIGURE = 0x0001CC02;
const char32_t RIGHT_POINTING_STICK_FIGURE = 0x0001CC03;
const char32_t DOWN_POINTING_STICK_FIGURE = 0x0001CC04;
const char32_t LOWER_HORIZONTAL_RULER_SEGMENT = 0x0001CC05;
const char32_t RIGHT_VERTICAL_RULER_SEGMENT = 0x0001CC06;
const char32_t LOWER_RIGHT_RULER_SEGMENT = 0x0001CC07;
const char32_t ANTENNA = 0x0001CC08;
const char32_t HORIZONTAL_RESISTOR_SEGMENT = 0x0001CC09;
const char32_t VERTICAL_RESISTOR_SEGMENT = 0x0001CC0A;
const char32_t LEFT_THIRD_INDUCTOR = 0x0001CC0B;
const char32_t MIDDLE_THIRD_INDUCTOR = 0x0001CC0C;
const char32_t RIGHT_THIRD_INDUCTOR = 0x0001CC0D;
const char32_t LEFT_POINTING_DIODE = 0x0001CC0E;
const char32_t RIGHT_POINTING_DIODE = 0x0001CC0F;
const char32_t NPN_TRANSISTOR = 0x0001CC10;
const char32_t PNP_TRANSISTOR = 0x0001CC11;
const char32_t RECEPTACLE = 0x0001CC12;
const char32_t HORIZONTAL_CAPACITOR = 0x0001CC13;
const char32_t VERTICAL_CAPACITOR = 0x0001CC14;
const char32_t LOGIC_GATE_OR = 0x0001CC15;
const char32_t LOGIC_GATE_AND = 0x0001CC16;
const char32_t LOGIC_GATE_INVERTED_INPUTS = 0x0001CC17;
const char32_t LOGIC_GATE_INVERTED_OUTPUT = 0x0001CC18;
const char32_t LOGIC_GATE_BUFFER = 0x0001CC19;
const char32_t LOGIC_GATE_BUFFER_WITH_INVERTED_INPUT = 0x0001CC1A;
const char32_t BOX_DRAWINGS_LIGHT_HORIZONTAL_AND_UPPER_RIGHT = 0x0001CC1B;
const char32_t BOX_DRAWINGS_LIGHT_HORIZONTAL_AND_LOWER_RIGHT = 0x0001CC1C;
const char32_t BOX_DRAWINGS_LIGHT_TOP_AND_UPPER_LEFT = 0x0001CC1D;
const char32_t BOX_DRAWINGS_LIGHT_BOTTOM_AND_LOWER_LEFT = 0x0001CC1E;
const char32_t BOX_DRAWINGS_DOUBLE_DIAGONAL_UPPER_RIGHT_TO_LOWER_LEFT = 0x0001CC1F;
const char32_t BOX_DRAWINGS_DOUBLE_DIAGONAL_UPPER_LEFT_TO_LOWER_RIGHT = 0x0001CC20;
const char32_t SEPARATED_BLOCK_QUADRANT_1 = 0x0001CC21;
const char32_t SEPARATED_BLOCK_QUADRANT_2 = 0x0001CC22;
const char32_t SEPARATED_BLOCK_QUADRANT_12 = 0x0001CC23;
const char32_t SEPARATED_BLOCK_QUADRANT_3 = 0x0001CC24;
const char32_t SEPARATED_BLOCK_QUADRANT_13 = 0x0001CC25;
const char32_t SEPARATED_BLOCK_QUADRANT_23 = 0x0001CC26;
const char32_t SEPARATED_BLOCK_QUADRANT_123 = 0x0001CC27;
const char32_t SEPARATED_BLOCK_QUADRANT_4 = 0x0001CC28;
const char32_t SEPARATED_BLOCK_QUADRANT_14 = 0x0001CC29;
const char32_t SEPARATED_BLOCK_QUADRANT_24 = 0x0001CC2A;
const char32_t SEPARATED_BLOCK_QUADRANT_124 = 0x0001CC2B;
const char32_t SEPARATED_BLOCK_QUADRANT_34 = 0x0001CC2C;
const char32_t SEPARATED_BLOCK_QUADRANT_134 = 0x0001CC2D;
const char32_t SEPARATED_BLOCK_QUADRANT_234 = 0x0001CC2E;
const char32_t SEPARATED_BLOCK_QUADRANT_1234 = 0x0001CC2F;
const char32_t UPPER_LEFT_TWELFTH_CIRCLE = 0x0001CC30;
const char32_t UPPER_CENTRE_LEFT_TWELFTH_CIRCLE = 0x0001CC31;
const char32_t UPPER_CENTRE_RIGHT_TWELFTH_CIRCLE = 0x0001CC32;
const char32_t UPPER_RIGHT_TWELFTH_CIRCLE = 0x0001CC33;
const char32_t UPPER_MIDDLE_LEFT_TWELFTH_CIRCLE = 0x0001CC34;
const char32_t UPPER_LEFT_QUARTER_CIRCLE = 0x0001CC35;
const char32_t UPPER_RIGHT_QUARTER_CIRCLE = 0x0001CC36;
const char32_t UPPER_MIDDLE_RIGHT_TWELFTH_CIRCLE = 0x0001CC37;
const char32_t LOWER_MIDDLE_LEFT_TWELFTH_CIRCLE = 0x0001CC38;
const char32_t LOWER_LEFT_QUARTER_CIRCLE = 0x0001CC39;
const char32_t LOWER_RIGHT_QUARTER_CIRCLE = 0x0001CC3A;
const char32_t LOWER_MIDDLE_RIGHT_TWELFTH_CIRCLE = 0x0001CC3B;
const char32_t LOWER_LEFT_TWELFTH_CIRCLE = 0x0001CC3C;
const char32_t LOWER_CENTRE_LEFT_TWELFTH_CIRCLE = 0x0001CC3D;
const char32_t LOWER_CENTRE_RIGHT_TWELFTH_CIRCLE = 0x0001CC3E;
const char32_t LOWER_RIGHT_TWELFTH_CIRCLE = 0x0001CC3F;
const char32_t SPARSE_HORIZONTAL_FILL = 0x0001CC40;
const char32_t SPARSE_VERTICAL_FILL = 0x0001CC41;
const char32_t ORTHOGONAL_CROSSHATCH_FILL = 0x0001CC42;
const char32_t DIAGONAL_CROSSHATCH_FILL = 0x0001CC43;
const char32_t DENSE_VERTICAL_FILL = 0x0001CC44;
const char32_t DENSE_HORIZONTAL_FILL = 0x0001CC45;
const char32_t SPECKLE_FILL_FRAME_1 = 0x0001CC46;
const char32_t SPECKLE_FILL_FRAME_2 = 0x0001CC47;
const char32_t LEFT_FACING_BASSINET = 0x0001CC48;
const char32_t RIGHT_FACING_BASSINET = 0x0001CC49;
const char32_t FLYING_SAUCER_WITH_BEAMS = 0x0001CC4A;
const char32_t FLYING_SAUCER_WITHOUT_BEAMS = 0x0001CC4B;
const char32_t ALIEN_MONSTER_OPEN_JAWS = 0x0001CC4C;
const char32_t ALIEN_MONSTER_CLOSED_JAWS = 0x0001CC4D;
const char32_t ALIEN_SQUID_OPEN_TENTACLES = 0x0001CC4E;
const char32_t ALIEN_SQUID_CLOSED_TENTACLES = 0x0001CC4F;
const char32_t ALIEN_CRAB_STEPPING_RIGHT = 0x0001CC50;
const char32_t ALIEN_CRAB_STEPPING_LEFT = 0x0001CC51;
const char32_t ALIEN_SPIDER_CROUCHING = 0x0001CC52;
const char32_t ALIEN_SPIDER_SPREAD = 0x0001CC53;
const char32_t ALIEN_MONSTER_STEP_1 = 0x0001CC54;
const char32_t ALIEN_MONSTER_STEP_2 = 0x0001CC55;
const char32_t LEFT_POINTING_ROCKET_SHIP = 0x0001CC56;
const char32_t UP_POINTING_ROCKET_SHIP = 0x0001CC57;
const char32_t RIGHT_POINTING_ROCKET_SHIP = 0x0001CC58;
const char32_t DOWN_POINTING_ROCKET_SHIP = 0x0001CC59;
const char32_t TOP_HALF_LEFT_FACING_ROBOT = 0x0001CC5A;
const char32_t TOP_HALF_FORWARD_FACING_ROBOT = 0x0001CC5B;
const char32_t TOP_HALF_RIGHT_FACING_ROBOT = 0x0001CC5C;
const char32_t BOTTOM_HALF_LEFT_FACING_ROBOT = 0x0001CC5D;
const char32_t BOTTOM_HALF_FORWARD_FACING_ROBOT = 0x0001CC5E;
const char32_t BOTTOM_HALF_RIGHT_FACING_ROBOT = 0x0001CC5F;
const char32_t LEFT_POINTING_ATOMIC_BOMB = 0x0001CC60;
const char32_t UP_POINTING_ATOMIC_BOMB = 0x0001CC61;
const char32_t RIGHT_POINTING_ATOMIC_BOMB = 0x0001CC62;
const char32_t DOWN_POINTING_ATOMIC_BOMB = 0x0001CC63;
const char32_t MUSHROOM_CLOUD = 0x0001CC64;
const char32_t LEFT_POINTING_RIFLE = 0x0001CC65;
const char32_t UP_POINTING_RIFLE = 0x0001CC66;
const char32_t RIGHT_POINTING_RIFLE = 0x0001CC67;
const char32_t DOWN_POINTING_RIFLE = 0x0001CC68;
const char32_t EIGHT_RAYS_INWARD = 0x0001CC69;
const char32_t EIGHT_RAYS_OUTWARD = 0x0001CC6A;
const char32_t BLACK_LARGE_CIRCLE_MINUS_LEFT_QUARTER_SECTION = 0x0001CC6B;
const char32_t BLACK_LARGE_CIRCLE_MINUS_UPPER_QUARTER_SECTION = 0x0001CC6C;
const char32_t BLACK_LARGE_CIRCLE_MINUS_RIGHT_QUARTER_SECTION = 0x0001CC6D;
const char32_t BLACK_LARGE_CIRCLE_MINUS_LOWER_QUARTER_SECTION = 0x0001CC6E;
const char32_t BLACK_NEUTRAL_FACE = 0x0001CC6F;
const char32_t LEFT_FACING_SNAKE_HEAD_WITH_OPEN_MOUTH = 0x0001CC70;
const char32_t UP_FACING_SNAKE_HEAD_WITH_OPEN_MOUTH = 0x0001CC71;
const char32_t RIGHT_FACING_SNAKE_HEAD_WITH_OPEN_MOUTH = 0x0001CC72;
const char32_t DOWN_FACING_SNAKE_HEAD_WITH_OPEN_MOUTH = 0x0001CC73;
const char32_t LEFT_FACING_SNAKE_HEAD_WITH_CLOSED_MOUTH = 0x0001CC74;
const char32_t UP_FACING_SNAKE_HEAD_WITH_CLOSED_MOUTH = 0x0001CC75;
const char32_t RIGHT_FACING_SNAKE_HEAD_WITH_CLOSED_MOUTH = 0x0001CC76;
const char32_t DOWN_FACING_SNAKE_HEAD_WITH_CLOSED_MOUTH = 0x0001CC77;
const char32_t LEFT_POINTING_ENERGY_WAVE = 0x0001CC78;
const char32_t UP_POINTING_ENERGY_WAVE = 0x0001CC79;
const char32_t RIGHT_POINTING_ENERGY_WAVE = 0x0001CC7A;
const char32_t DOWN_POINTING_ENERGY_WAVE = 0x0001CC7B;
const char32_t SQUARE_SPIRAL_FROM_TOP_LEFT = 0x0001CC7C;
const char32_t SQUARE_SPIRAL_FROM_TOP_RIGHT = 0x0001CC7D;
const char32_t SQUARE_SPIRAL_FROM_BOTTOM_RIGHT = 0x0001CC7E;
const char32_t SQUARE_SPIRAL_FROM_BOTTOM_LEFT = 0x0001CC7F;
const char32_t STRIPED_LEFT_POINTING_TRIANGLE = 0x0001CC80;
const char32_t STRIPED_UP_POINTING_TRIANGLE = 0x0001CC81;
const char32_t STRIPED_RIGHT_POINTING_TRIANGLE = 0x0001CC82;
const char32_t STRIPED_DOWN_POINTING_TRIANGLE = 0x0001CC83;
const char32_t VERTICAL_LADDER = 0x0001CC84;
const char32_t HORIZONTAL_LADDER = 0x0001CC85;
const char32_t WHITE_LOWER_LEFT_POINTER = 0x0001CC86;
const char32_t WHITE_LOWER_RIGHT_POINTER = 0x0001CC87;
const char32_t TWO_RINGS_ALIGNED_HORIZONTALLY = 0x0001CC88;
const char32_t SQUARE_FOUR_CORNER_SALTIRES = 0x0001CC89;
const char32_t SQUARE_FOUR_CORNER_DIAGONALS = 0x0001CC8A;
const char32_t SQUARE_FOUR_CORNER_BLACK_TRIANGLES = 0x0001CC8B;
const char32_t SQUARE_APERTURE = 0x0001CC8C;
const char32_t INVERSE_BLACK_DIAMOND = 0x0001CC8D;
const char32_t LEFT_AND_UPPER_ONE_EIGHTH_BLOCK_CONTAINING_BLACK_SMALL_SQUARE = 0x0001CC8E;
const char32_t INVERSE_BLACK_SMALL_SQUARE = 0x0001CC8F;
const char32_t VERTICAL_LINE_WITH_FOUR_TICK_MARKS = 0x0001CC90;
const char32_t HORIZONTAL_LINE_WITH_FOUR_TICK_MARKS = 0x0001CC91;
const char32_t LEFT_FACING_FISH = 0x0001CC92;
const char32_t RIGHT_FACING_FISH = 0x0001CC93;
const char32_t LEFT_FACING_FISH_WITH_OPEN_MOUTH = 0x0001CC94;
const char32_t RIGHT_FACING_FISH_WITH_OPEN_MOUTH = 0x0001CC95;
const char32_t FLAPPING_BIRD = 0x0001CC96;
const char32_t LEFT_POINTING_RACING_CAR = 0x0001CC97;
const char32_t UP_POINTING_RACING_CAR = 0x0001CC98;
const char32_t RIGHT_POINTING_RACING_CAR = 0x0001CC99;
const char32_t DOWN_POINTING_RACING_CAR = 0x0001CC9A;
const char32_t HORIZONTAL_RACING_CAR = 0x0001CC9B;
const char32_t VERTICAL_RACING_CAR = 0x0001CC9C;
const char32_t VERTICAL_GO_KART = 0x0001CC9D;
const char32_t LEFT_POINTING_TANK = 0x0001CC9E;
const char32_t RIGHT_POINTING_TANK = 0x0001CC9F;
const char32_t LEFT_POINTING_ROCKET_BOOSTER = 0x0001CCA0;
const char32_t RIGHT_POINTING_ROCKET_BOOSTER = 0x0001CCA1;
const char32_t LEFT_POINTING_ROLLER_COASTER_CAR = 0x0001CCA2;
const char32_t RIGHT_POINTING_ROLLER_COASTER_CAR = 0x0001CCA3;
const char32_t LEFT_HALF_FLYING_SAUCER = 0x0001CCA4;
const char32_t RIGHT_HALF_FLYING_SAUCER = 0x0001CCA5;
const char32_t UPPER_LEFT_QUADRANT_FACE_WITH_OPEN_EYES = 0x0001CCA6;
const char32_t UPPER_RIGHT_QUADRANT_FACE_WITH_OPEN_EYES = 0x0001CCA7;
const char32_t UPPER_LEFT_QUADRANT_FACE_WITH_CLOSED_EYES = 0x0001CCA8;
const char32_t UPPER_RIGHT_QUADRANT_FACE_WITH_CLOSED_EYES = 0x0001CCA9;
const char32_t LOWER_LEFT_QUADRANT_SMILING_FACE = 0x0001CCAA;
const char32_t LOWER_RIGHT_QUADRANT_SMILING_FACE = 0x0001CCAB;
const char32_t LOWER_LEFT_QUADRANT_NEUTRAL_FACE = 0x0001CCAC;
const char32_t LOWER_RIGHT_QUADRANT_NEUTRAL_FACE = 0x0001CCAD;
const char32_t LOWER_LEFT_QUADRANT_FACE_WITH_OPEN_MOUTH = 0x0001CCAE;
const char32_t LOWER_RIGHT_QUADRANT_FACE_WITH_OPEN_MOUTH = 0x0001CCAF;
const char32_t LOWER_LEFT_QUADRANT_FROWNING_FACE = 0x0001CCB0;
const char32_t LOWER_RIGHT_QUADRANT_FROWNING_FACE = 0x0001CCB1;
const char32_t UPPER_LEFT_QUADRANT_TELEVISION = 0x0001CCB2;
const char32_t UPPER_RIGHT_QUADRANT_TELEVISION = 0x0001CCB3;
const char32_t LOWER_LEFT_QUADRANT_TELEVISION = 0x0001CCB4;
const char32_t LOWER_RIGHT_QUADRANT_TELEVISION = 0x0001CCB5;
const char32_t UPPER_LEFT_QUADRANT_MICROCOMPUTER = 0x0001CCB6;
const char32_t UPPER_RIGHT_QUADRANT_MICROCOMPUTER = 0x0001CCB7;
const char32_t LOWER_LEFT_QUADRANT_MICROCOMPUTER = 0x0001CCB8;
const char32_t LOWER_RIGHT_QUADRANT_MICROCOMPUTER = 0x0001CCB9;
const char32_t UPPER_LEFT_QUADRANT_CHESS_KING = 0x0001CCBA;
const char32_t UPPER_RIGHT_QUADRANT_CHESS_KING = 0x0001CCBB;
const char32_t LOWER_LEFT_QUADRANT_CHESS_KING = 0x0001CCBC;
const char32_t LOWER_RIGHT_QUADRANT_CHESS_KING = 0x0001CCBD;
const char32_t UPPER_LEFT_QUADRANT_CHESS_QUEEN = 0x0001CCBE;
const char32_t UPPER_RIGHT_QUADRANT_CHESS_QUEEN = 0x0001CCBF;
const char32_t LOWER_LEFT_QUADRANT_CHESS_QUEEN = 0x0001CCC0;
const char32_t LOWER_RIGHT_QUADRANT_CHESS_QUEEN = 0x0001CCC1;
const char32_t UPPER_LEFT_QUADRANT_CHESS_ROOK = 0x0001CCC2;
const char32_t UPPER_RIGHT_QUADRANT_CHESS_ROOK = 0x0001CCC3;
const char32_t LOWER_LEFT_QUADRANT_CHESS_ROOK = 0x0001CCC4;
const char32_t LOWER_RIGHT_QUADRANT_CHESS_ROOK = 0x0001CCC5;
const char32_t UPPER_LEFT_QUADRANT_CHESS_BISHOP = 0x0001CCC6;
const char32_t UPPER_RIGHT_QUADRANT_CHESS_BISHOP = 0x0001CCC7;
const char32_t LOWER_LEFT_QUADRANT_CHESS_BISHOP = 0x0001CCC8;
const char32_t LOWER_RIGHT_QUADRANT_CHESS_BISHOP = 0x0001CCC9;
const char32_t UPPER_LEFT_QUADRANT_CHESS_KNIGHT = 0x0001CCCA;
const char32_t UPPER_RIGHT_QUADRANT_CHESS_KNIGHT = 0x0001CCCB;
const char32_t LOWER_LEFT_QUADRANT_CHESS_KNIGHT = 0x0001CCCC;
const char32_t LOWER_RIGHT_QUADRANT_CHESS_KNIGHT = 0x0001CCCD;
const char32_t UPPER_LEFT_QUADRANT_CHESS_PAWN = 0x0001CCCE;
const char32_t UPPER_RIGHT_QUADRANT_CHESS_PAWN = 0x0001CCCF;
const char32_t LOWER_LEFT_QUADRANT_CHESS_PAWN = 0x0001CCD0;
const char32_t LOWER_RIGHT_QUADRANT_CHESS_PAWN = 0x0001CCD1;
const char32_t UPPER_LEFT_QUADRANT_STANDING_KNIGHT = 0x0001CCD2;
const char32_t UPPER_RIGHT_QUADRANT_STANDING_KNIGHT = 0x0001CCD3;
const char32_t LOWER_LEFT_QUADRANT_STANDING_KNIGHT = 0x0001CCD4;
const char32_t LOWER_RIGHT_QUADRANT_STANDING_KNIGHT = 0x0001CCD5;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_A = 0x0001CCD6;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_B = 0x0001CCD7;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_C = 0x0001CCD8;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_D = 0x0001CCD9;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_E = 0x0001CCDA;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_F = 0x0001CCDB;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_G = 0x0001CCDC;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_H = 0x0001CCDD;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_I = 0x0001CCDE;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_J = 0x0001CCDF;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_K = 0x0001CCE0;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_L = 0x0001CCE1;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_M = 0x0001CCE2;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_N = 0x0001CCE3;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_O = 0x0001CCE4;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_P = 0x0001CCE5;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_Q = 0x0001CCE6;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_R = 0x0001CCE7;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_S = 0x0001CCE8;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_T = 0x0001CCE9;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_U = 0x0001CCEA;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_V = 0x0001CCEB;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_W = 0x0001CCEC;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_X = 0x0001CCED;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_Y = 0x0001CCEE;
const char32_t OUTLINED_LATIN_CAPITAL_LETTER_Z = 0x0001CCEF;
const char32_t OUTLINED_DIGIT_ZERO = 0x0001CCF0;
const char32_t OUTLINED_DIGIT_ONE = 0x0001CCF1;
const char32_t OUTLINED_DIGIT_TWO = 0x0001CCF2;
const char32_t OUTLINED_DIGIT_THREE = 0x0001CCF3;
const char32_t OUTLINED_DIGIT_FOUR = 0x0001CCF4;
const char32_t OUTLINED_DIGIT_FIVE = 0x0001CCF5;
const char32_t OUTLINED_DIGIT_SIX = 0x0001CCF6;
const char32_t OUTLINED_DIGIT_SEVEN = 0x0001CCF7;
const char32_t OUTLINED_DIGIT_EIGHT = 0x0001CCF8;
const char32_t OUTLINED_DIGIT_NINE = 0x0001CCF9;
const char32_t SNAKE_SYMBOL = 0x0001CCFA;
const char32_t FLYING_SAUCER_SYMBOL = 0x0001CCFB;
const char32_t NOSE_SYMBOL = 0x0001CCFC;
const char32_t BLOCK_OCTANT_3 = 0x0001CD00;
const char32_t BLOCK_OCTANT_23 = 0x0001CD01;
const char32_t BLOCK_OCTANT_123 = 0x0001CD02;
const char32_t BLOCK_OCTANT_4 = 0x0001CD03;
const char32_t BLOCK_OCTANT_14 = 0x0001CD04;
const char32_t BLOCK_OCTANT_124 = 0x0001CD05;
const char32_t BLOCK_OCTANT_34 = 0x0001CD06;
const char32_t BLOCK_OCTANT_134 = 0x0001CD07;
const char32_t BLOCK_OCTANT_234 = 0x0001CD08;
const char32_t BLOCK_OCTANT_5 = 0x0001CD09;
const char32_t BLOCK_OCTANT_15 = 0x0001CD0A;
const char32_t BLOCK_OCTANT_25 = 0x0001CD0B;
const char32_t BLOCK_OCTANT_125 = 0x0001CD0C;
const char32_t BLOCK_OCTANT_135 = 0x0001CD0D;
const char32_t BLOCK_OCTANT_235 = 0x0001CD0E;
const char32_t BLOCK_OCTANT_1235 = 0x0001CD0F;
const char32_t BLOCK_OCTANT_45 = 0x0001CD10;
const char32_t BLOCK_OCTANT_145 = 0x0001CD11;
const char32_t BLOCK_OCTANT_245 = 0x0001CD12;
const char32_t BLOCK_OCTANT_1245 = 0x0001CD13;
const char32_t BLOCK_OCTANT_345 = 0x0001CD14;
const char32_t BLOCK_OCTANT_1345 = 0x0001CD15;
const char32_t BLOCK_OCTANT_2345 = 0x0001CD16;
const char32_t BLOCK_OCTANT_12345 = 0x0001CD17;
const char32_t BLOCK_OCTANT_6 = 0x0001CD18;
const char32_t BLOCK_OCTANT_16 = 0x0001CD19;
const char32_t BLOCK_OCTANT_26 = 0x0001CD1A;
const char32_t BLOCK_OCTANT_126 = 0x0001CD1B;
const char32_t BLOCK_OCTANT_36 = 0x0001CD1C;
const char32_t BLOCK_OCTANT_136 = 0x0001CD1D;
const char32_t BLOCK_OCTANT_236 = 0x0001CD1E;
const char32_t BLOCK_OCTANT_1236 = 0x0001CD1F;
const char32_t BLOCK_OCTANT_146 = 0x0001CD20;
const char32_t BLOCK_OCTANT_246 = 0x0001CD21;
const char32_t BLOCK_OCTANT_1246 = 0x0001CD22;
const char32_t BLOCK_OCTANT_346 = 0x0001CD23;
const char32_t BLOCK_OCTANT_1346 = 0x0001CD24;
const char32_t BLOCK_OCTANT_2346 = 0x0001CD25;
const char32_t BLOCK_OCTANT_12346 = 0x0001CD26;
const char32_t BLOCK_OCTANT_56 = 0x0001CD27;
const char32_t BLOCK_OCTANT_156 = 0x0001CD28;
const char32_t BLOCK_OCTANT_256 = 0x0001CD29;
const char32_t BLOCK_OCTANT_1256 = 0x0001CD2A;
const char32_t BLOCK_OCTANT_356 = 0x0001CD2B;
const char32_t BLOCK_OCTANT_1356 = 0x0001CD2C;
const char32_t BLOCK_OCTANT_2356 = 0x0001CD2D;
const char32_t BLOCK_OCTANT_12356 = 0x0001CD2E;
const char32_t BLOCK_OCTANT_456 = 0x0001CD2F;
const char32_t BLOCK_OCTANT_1456 = 0x0001CD30;
const char32_t BLOCK_OCTANT_2456 = 0x0001CD31;
const char32_t BLOCK_OCTANT_12456 = 0x0001CD32;
const char32_t BLOCK_OCTANT_3456 = 0x0001CD33;
const char32_t BLOCK_OCTANT_13456 = 0x0001CD34;
const char32_t BLOCK_OCTANT_23456 = 0x0001CD35;
const char32_t BLOCK_OCTANT_17 = 0x0001CD36;
const char32_t BLOCK_OCTANT_27 = 0x0001CD37;
const char32_t BLOCK_OCTANT_127 = 0x0001CD38;
const char32_t BLOCK_OCTANT_37 = 0x0001CD39;
const char32_t BLOCK_OCTANT_137 = 0x0001CD3A;
const char32_t BLOCK_OCTANT_237 = 0x0001CD3B;
const char32_t BLOCK_OCTANT_1237 = 0x0001CD3C;
const char32_t BLOCK_OCTANT_47 = 0x0001CD3D;
const char32_t BLOCK_OCTANT_147 = 0x0001CD3E;
const char32_t BLOCK_OCTANT_247 = 0x0001CD3F;
const char32_t BLOCK_OCTANT_1247 = 0x0001CD40;
const char32_t BLOCK_OCTANT_347 = 0x0001CD41;
const char32_t BLOCK_OCTANT_1347 = 0x0001CD42;
const char32_t BLOCK_OCTANT_2347 = 0x0001CD43;
const char32_t BLOCK_OCTANT_12347 = 0x0001CD44;
const char32_t BLOCK_OCTANT_157 = 0x0001CD45;
const char32_t BLOCK_OCTANT_257 = 0x0001CD46;
const char32_t BLOCK_OCTANT_1257 = 0x0001CD47;
const char32_t BLOCK_OCTANT_357 = 0x0001CD48;
const char32_t BLOCK_OCTANT_2357 = 0x0001CD49;
const char32_t BLOCK_OCTANT_12357 = 0x0001CD4A;
const char32_t BLOCK_OCTANT_457 = 0x0001CD4B;
const char32_t BLOCK_OCTANT_1457 = 0x0001CD4C;
const char32_t BLOCK_OCTANT_12457 = 0x0001CD4D;
const char32_t BLOCK_OCTANT_3457 = 0x0001CD4E;
const char32_t BLOCK_OCTANT_13457 = 0x0001CD4F;
const char32_t BLOCK_OCTANT_23457 = 0x0001CD50;
const char32_t BLOCK_OCTANT_67 = 0x0001CD51;
const char32_t BLOCK_OCTANT_167 = 0x0001CD52;
const char32_t BLOCK_OCTANT_267 = 0x0001CD53;
const char32_t BLOCK_OCTANT_1267 = 0x0001CD54;
const char32_t BLOCK_OCTANT_367 = 0x0001CD55;
const char32_t BLOCK_OCTANT_1367 = 0x0001CD56;
const char32_t BLOCK_OCTANT_2367 = 0x0001CD57;
const char32_t BLOCK_OCTANT_12367 = 0x0001CD58;
const char32_t BLOCK_OCTANT_467 = 0x0001CD59;
const char32_t BLOCK_OCTANT_1467 = 0x0001CD5A;
const char32_t BLOCK_OCTANT_2467 = 0x0001CD5B;
const char32_t BLOCK_OCTANT_12467 = 0x0001CD5C;
const char32_t BLOCK_OCTANT_3467 = 0x0001CD5D;
const char32_t BLOCK_OCTANT_13467 = 0x0001CD5E;
const char32_t BLOCK_OCTANT_23467 = 0x0001CD5F;
const char32_t BLOCK_OCTANT_123467 = 0x0001CD60;
const char32_t BLOCK_OCTANT_567 = 0x0001CD61;
const char32_t BLOCK_OCTANT_1567 = 0x0001CD62;
const char32_t BLOCK_OCTANT_2567 = 0x0001CD63;
const char32_t BLOCK_OCTANT_12567 = 0x0001CD64;
const char32_t BLOCK_OCTANT_3567 = 0x0001CD65;
const char32_t BLOCK_OCTANT_13567 = 0x0001CD66;
const char32_t BLOCK_OCTANT_23567 = 0x0001CD67;
const char32_t BLOCK_OCTANT_123567 = 0x0001CD68;
const char32_t BLOCK_OCTANT_4567 = 0x0001CD69;
const char32_t BLOCK_OCTANT_14567 = 0x0001CD6A;
const char32_t BLOCK_OCTANT_24567 = 0x0001CD6B;
const char32_t BLOCK_OCTANT_124567 = 0x0001CD6C;
const char32_t BLOCK_OCTANT_34567 = 0x0001CD6D;
const char32_t BLOCK_OCTANT_134567 = 0x0001CD6E;
const char32_t BLOCK_OCTANT_234567 = 0x0001CD6F;
const char32_t BLOCK_OCTANT_1234567 = 0x0001CD70;
const char32_t BLOCK_OCTANT_18 = 0x0001CD71;
const char32_t BLOCK_OCTANT_28 = 0x0001CD72;
const char32_t BLOCK_OCTANT_128 = 0x0001CD73;
const char32_t BLOCK_OCTANT_38 = 0x0001CD74;
const char32_t BLOCK_OCTANT_138 = 0x0001CD75;
const char32_t BLOCK_OCTANT_238 = 0x0001CD76;
const char32_t BLOCK_OCTANT_1238 = 0x0001CD77;
const char32_t BLOCK_OCTANT_48 = 0x0001CD78;
const char32_t BLOCK_OCTANT_148 = 0x0001CD79;
const char32_t BLOCK_OCTANT_248 = 0x0001CD7A;
const char32_t BLOCK_OCTANT_1248 = 0x0001CD7B;
const char32_t BLOCK_OCTANT_348 = 0x0001CD7C;
const char32_t BLOCK_OCTANT_1348 = 0x0001CD7D;
const char32_t BLOCK_OCTANT_2348 = 0x0001CD7E;
const char32_t BLOCK_OCTANT_12348 = 0x0001CD7F;
const char32_t BLOCK_OCTANT_58 = 0x0001CD80;
const char32_t BLOCK_OCTANT_158 = 0x0001CD81;
const char32_t BLOCK_OCTANT_258 = 0x0001CD82;
const char32_t BLOCK_OCTANT_1258 = 0x0001CD83;
const char32_t BLOCK_OCTANT_358 = 0x0001CD84;
const char32_t BLOCK_OCTANT_1358 = 0x0001CD85;
const char32_t BLOCK_OCTANT_2358 = 0x0001CD86;
const char32_t BLOCK_OCTANT_12358 = 0x0001CD87;
const char32_t BLOCK_OCTANT_458 = 0x0001CD88;
const char32_t BLOCK_OCTANT_1458 = 0x0001CD89;
const char32_t BLOCK_OCTANT_2458 = 0x0001CD8A;
const char32_t BLOCK_OCTANT_12458 = 0x0001CD8B;
const char32_t BLOCK_OCTANT_3458 = 0x0001CD8C;
const char32_t BLOCK_OCTANT_13458 = 0x0001CD8D;
const char32_t BLOCK_OCTANT_23458 = 0x0001CD8E;
const char32_t BLOCK_OCTANT_123458 = 0x0001CD8F;
const char32_t BLOCK_OCTANT_168 = 0x0001CD90;
const char32_t BLOCK_OCTANT_268 = 0x0001CD91;
const char32_t BLOCK_OCTANT_1268 = 0x0001CD92;
const char32_t BLOCK_OCTANT_368 = 0x0001CD93;
const char32_t BLOCK_OCTANT_2368 = 0x0001CD94;
const char32_t BLOCK_OCTANT_12368 = 0x0001CD95;
const char32_t BLOCK_OCTANT_468 = 0x0001CD96;
const char32_t BLOCK_OCTANT_1468 = 0x0001CD97;
const char32_t BLOCK_OCTANT_12468 = 0x0001CD98;
const char32_t BLOCK_OCTANT_3468 = 0x0001CD99;
const char32_t BLOCK_OCTANT_13468 = 0x0001CD9A;
const char32_t BLOCK_OCTANT_23468 = 0x0001CD9B;
const char32_t BLOCK_OCTANT_568 = 0x0001CD9C;
const char32_t BLOCK_OCTANT_1568 = 0x0001CD9D;
const char32_t BLOCK_OCTANT_2568 = 0x0001CD9E;
const char32_t BLOCK_OCTANT_12568 = 0x0001CD9F;
const char32_t BLOCK_OCTANT_3568 = 0x0001CDA0;
const char32_t BLOCK_OCTANT_13568 = 0x0001CDA1;
const char32_t BLOCK_OCTANT_23568 = 0x0001CDA2;
const char32_t BLOCK_OCTANT_123568 = 0x0001CDA3;
const char32_t BLOCK_OCTANT_4568 = 0x0001CDA4;
const char32_t BLOCK_OCTANT_14568 = 0x0001CDA5;
const char32_t BLOCK_OCTANT_24568 = 0x0001CDA6;
const char32_t BLOCK_OCTANT_124568 = 0x0001CDA7;
const char32_t BLOCK_OCTANT_34568 = 0x0001CDA8;
const char32_t BLOCK_OCTANT_134568 = 0x0001CDA9;
const char32_t BLOCK_OCTANT_234568 = 0x0001CDAA;
const char32_t BLOCK_OCTANT_1234568 = 0x0001CDAB;
const char32_t BLOCK_OCTANT_178 = 0x0001CDAC;
const char32_t BLOCK_OCTANT_278 = 0x0001CDAD;
const char32_t BLOCK_OCTANT_1278 = 0x0001CDAE;
const char32_t BLOCK_OCTANT_378 = 0x0001CDAF;
const char32_t BLOCK_OCTANT_1378 = 0x0001CDB0;
const char32_t BLOCK_OCTANT_2378 = 0x0001CDB1;
const char32_t BLOCK_OCTANT_12378 = 0x0001CDB2;
const char32_t BLOCK_OCTANT_478 = 0x0001CDB3;
const char32_t BLOCK_OCTANT_1478 = 0x0001CDB4;
const char32_t BLOCK_OCTANT_2478 = 0x0001CDB5;
const char32_t BLOCK_OCTANT_12478 = 0x0001CDB6;
const char32_t BLOCK_OCTANT_3478 = 0x0001CDB7;
const char32_t BLOCK_OCTANT_13478 = 0x0001CDB8;
const char32_t BLOCK_OCTANT_23478 = 0x0001CDB9;
const char32_t BLOCK_OCTANT_123478 = 0x0001CDBA;
const char32_t BLOCK_OCTANT_578 = 0x0001CDBB;
const char32_t BLOCK_OCTANT_1578 = 0x0001CDBC;
const char32_t BLOCK_OCTANT_2578 = 0x0001CDBD;
const char32_t BLOCK_OCTANT_12578 = 0x0001CDBE;
const char32_t BLOCK_OCTANT_3578 = 0x0001CDBF;
const char32_t BLOCK_OCTANT_13578 = 0x0001CDC0;
const char32_t BLOCK_OCTANT_23578 = 0x0001CDC1;
const char32_t BLOCK_OCTANT_123578 = 0x0001CDC2;
const char32_t BLOCK_OCTANT_4578 = 0x0001CDC3;
const char32_t BLOCK_OCTANT_14578 = 0x0001CDC4;
const char32_t BLOCK_OCTANT_24578 = 0x0001CDC5;
const char32_t BLOCK_OCTANT_124578 = 0x0001CDC6;
const char32_t BLOCK_OCTANT_34578 = 0x0001CDC7;
const char32_t BLOCK_OCTANT_134578 = 0x0001CDC8;
const char32_t BLOCK_OCTANT_234578 = 0x0001CDC9;
const char32_t BLOCK_OCTANT_1234578 = 0x0001CDCA;
const char32_t BLOCK_OCTANT_678 = 0x0001CDCB;
const char32_t BLOCK_OCTANT_1678 = 0x0001CDCC;
const char32_t BLOCK_OCTANT_2678 = 0x0001CDCD;
const char32_t BLOCK_OCTANT_12678 = 0x0001CDCE;
const char32_t BLOCK_OCTANT_3678 = 0x0001CDCF;
const char32_t BLOCK_OCTANT_13678 = 0x0001CDD0;
const char32_t BLOCK_OCTANT_23678 = 0x0001CDD1;
const char32_t BLOCK_OCTANT_123678 = 0x0001CDD2;
const char32_t BLOCK_OCTANT_4678 = 0x0001CDD3;
const char32_t BLOCK_OCTANT_14678 = 0x0001CDD4;
const char32_t BLOCK_OCTANT_24678 = 0x0001CDD5;
const char32_t BLOCK_OCTANT_124678 = 0x0001CDD6;
const char32_t BLOCK_OCTANT_34678 = 0x0001CDD7;
const char32_t BLOCK_OCTANT_134678 = 0x0001CDD8;
const char32_t BLOCK_OCTANT_234678 = 0x0001CDD9;
const char32_t BLOCK_OCTANT_1234678 = 0x0001CDDA;
const char32_t BLOCK_OCTANT_15678 = 0x0001CDDB;
const char32_t BLOCK_OCTANT_25678 = 0x0001CDDC;
const char32_t BLOCK_OCTANT_125678 = 0x0001CDDD;
const char32_t BLOCK_OCTANT_35678 = 0x0001CDDE;
const char32_t BLOCK_OCTANT_235678 = 0x0001CDDF;
const char32_t BLOCK_OCTANT_1235678 = 0x0001CDE0;
const char32_t BLOCK_OCTANT_45678 = 0x0001CDE1;
const char32_t BLOCK_OCTANT_145678 = 0x0001CDE2;
const char32_t BLOCK_OCTANT_1245678 = 0x0001CDE3;
const char32_t BLOCK_OCTANT_1345678 = 0x0001CDE4;
const char32_t BLOCK_OCTANT_2345678 = 0x0001CDE5;
const char32_t TOP_HALF_STANDING_PERSON = 0x0001CDE6;
const char32_t BOTTOM_HALF_STANDING_PERSON = 0x0001CDE7;
const char32_t TOP_HALF_RIGHT_FACING_RUNNER_FRAME_1 = 0x0001CDE8;
const char32_t BOTTOM_HALF_RIGHT_FACING_RUNNER_FRAME_1 = 0x0001CDE9;
const char32_t TOP_HALF_RIGHT_FACING_RUNNER_FRAME_2 = 0x0001CDEA;
const char32_t BOTTOM_HALF_RIGHT_FACING_RUNNER_FRAME_2 = 0x0001CDEB;
const char32_t TOP_HALF_LEFT_FACING_RUNNER_FRAME_1 = 0x0001CDEC;
const char32_t BOTTOM_HALF_LEFT_FACING_RUNNER_FRAME_1 = 0x0001CDED;
const char32_t TOP_HALF_LEFT_FACING_RUNNER_FRAME_2 = 0x0001CDEE;
const char32_t BOTTOM_HALF_LEFT_FACING_RUNNER_FRAME_2 = 0x0001CDEF;
const char32_t TOP_HALF_FORWARD_FACING_RUNNER = 0x0001CDF0;
const char32_t BOTTOM_HALF_FORWARD_FACING_RUNNER_FRAME_1 = 0x0001CDF1;
const char32_t BOTTOM_HALF_FORWARD_FACING_RUNNER_FRAME_2 = 0x0001CDF2;
const char32_t BOTTOM_HALF_FORWARD_FACING_RUNNER_FRAME_3 = 0x0001CDF3;
const char32_t BOTTOM_HALF_FORWARD_FACING_RUNNER_FRAME_4 = 0x0001CDF4;
const char32_t MOON_LANDER = 0x0001CDF5;
const char32_t TOP_HALF_FLAILING_ROBOT_FRAME_1 = 0x0001CDF6;
const char32_t TOP_HALF_FLAILING_ROBOT_FRAME_2 = 0x0001CDF7;
const char32_t DOWN_POINTING_AIRPLANE = 0x0001CDF8;
const char32_t LEFT_POINTING_AIRPLANE = 0x0001CDF9;
const char32_t SMALL_UP_POINTING_AIRPLANE = 0x0001CDFA;
const char32_t UP_POINTING_FROG = 0x0001CDFB;
const char32_t DOWN_POINTING_FROG = 0x0001CDFC;
const char32_t EXPLOSION_FRAME_1 = 0x0001CDFD;
const char32_t EXPLOSION_FRAME_2 = 0x0001CDFE;
const char32_t EXPLOSION_FRAME_3 = 0x0001CDFF;
const char32_t RIGHT_HALF_AND_LEFT_HALF_WHITE_CIRCLE = 0x0001CE00;
const char32_t LOWER_HALF_AND_UPPER_HALF_WHITE_CIRCLE = 0x0001CE01;
const char32_t EXPLOSION_AT_HORIZON = 0x0001CE02;
const char32_t UPPER_HALF_HEAVY_WHITE_SQUARE = 0x0001CE03;
const char32_t LOWER_HALF_HEAVY_WHITE_SQUARE = 0x0001CE04;
const char32_t HEAVY_WHITE_SQUARE_CONTAINING_BLACK_VERY_SMALL_SQUARE = 0x0001CE05;
const char32_t WHITE_VERTICAL_RECTANGLE_WITH_HORIZONTAL_BAR = 0x0001CE06;
const char32_t TOP_LEFT_BLACK_LEFT_POINTING_SMALL_TRIANGLE = 0x0001CE07;
const char32_t FUNNEL = 0x0001CE08;
const char32_t BOX_DRAWINGS_DOUBLE_DIAGONAL_LOWER_LEFT_TO_MIDDLE_CENTRE_TO_LOWER_RIGHT = 0x0001CE09;
const char32_t BOX_DRAWINGS_DOUBLE_DIAGONAL_UPPER_LEFT_TO_MIDDLE_CENTRE_TO_UPPER_RIGHT = 0x0001CE0A;
const char32_t LEFT_HALF_WHITE_ELLIPSE = 0x0001CE0B;
const char32_t RIGHT_HALF_WHITE_ELLIPSE = 0x0001CE0C;
const char32_t LEFT_HALF_TRIPLE_DASH_HORIZONTAL = 0x0001CE0D;
const char32_t RIGHT_HALF_TRIPLE_DASH_HORIZONTAL = 0x0001CE0E;
const char32_t HORIZONTAL_LINE_WITH_TICK_MARK = 0x0001CE0F;
const char32_t LEFT_HALF_HORIZONTAL_LINE_WITH_THREE_TICK_MARKS = 0x0001CE10;
const char32_t RIGHT_HALF_HORIZONTAL_LINE_WITH_THREE_TICK_MARKS = 0x0001CE11;
const char32_t HORIZONTAL_LINE_WITH_THREE_TICK_MARKS = 0x0001CE12;
const char32_t LOWER_HALF_VERTICAL_LINE_WITH_THREE_TICK_MARKS = 0x0001CE13;
const char32_t UPPER_HALF_VERTICAL_LINE_WITH_THREE_TICK_MARKS = 0x0001CE14;
const char32_t VERTICAL_LINE_WITH_THREE_TICK_MARKS = 0x0001CE15;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL_AND_TOP_RIGHT = 0x0001CE16;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL_AND_BOTTOM_RIGHT = 0x0001CE17;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL_AND_TOP_LEFT = 0x0001CE18;
const char32_t BOX_DRAWINGS_LIGHT_VERTICAL_AND_BOTTOM_LEFT = 0x0001CE19;
const char32_t LARGE_TYPE_PIECE_UPPER_LEFT_ARC = 0x0001CE1A;
const char32_t LARGE_TYPE_PIECE_UPPER_LEFT_CORNER = 0x0001CE1B;
const char32_t LARGE_TYPE_PIECE_UPPER_TERMINAL = 0x0001CE1C;
const char32_t LARGE_TYPE_PIECE_UPPER_LEFT_CROTCH = 0x0001CE1D;
const char32_t LARGE_TYPE_PIECE_LEFT_ARM = 0x0001CE1E;
const char32_t LARGE_TYPE_PIECE_CROSSBAR = 0x0001CE1F;
const char32_t LARGE_TYPE_PIECE_CROSSBAR_WITH_LOWER_STEM = 0x0001CE20;
const char32_t LARGE_TYPE_PIECE_UPPER_HALF_VERTEX_OF_M = 0x0001CE21;
const char32_t LARGE_TYPE_PIECE_DIAGONAL_LOWER_LEFT = 0x0001CE22;
const char32_t LARGE_TYPE_PIECE_SHORT_UPPER_TERMINAL = 0x0001CE23;
const char32_t LARGE_TYPE_PIECE_UPPER_RIGHT_ARC = 0x0001CE24;
const char32_t LARGE_TYPE_PIECE_RIGHT_ARM = 0x0001CE25;
const char32_t LARGE_TYPE_PIECE_UPPER_RIGHT_CROTCH = 0x0001CE26;
const char32_t LARGE_TYPE_PIECE_UPPER_RIGHT_CORNER = 0x0001CE27;
const char32_t LARGE_TYPE_PIECE_STEM_WITH_RIGHT_CROSSBAR = 0x0001CE28;
const char32_t LARGE_TYPE_PIECE_STEM = 0x0001CE29;
const char32_t LARGE_TYPE_PIECE_DIAGONAL_UPPER_RIGHT_AND_LOWER_RIGHT = 0x0001CE2A;
const char32_t LARGE_TYPE_PIECE_DIAGONAL_UPPER_RIGHT = 0x0001CE2B;
const char32_t LARGE_TYPE_PIECE_DIAGONAL_LOWER_RIGHT = 0x0001CE2C;
const char32_t LARGE_TYPE_PIECE_SHORT_LOWER_TERMINAL = 0x0001CE2D;
const char32_t LARGE_TYPE_PIECE_LOWER_LEFT_AND_UPPER_LEFT_ARC = 0x0001CE2E;
const char32_t LARGE_TYPE_PIECE_CENTRE_OF_K = 0x0001CE2F;
const char32_t LARGE_TYPE_PIECE_LOWER_HALF_VERTEX_OF_M = 0x0001CE30;
const char32_t LARGE_TYPE_PIECE_UPPER_HALF_VERTEX_OF_W = 0x0001CE31;
const char32_t LARGE_TYPE_PIECE_CENTRE_OF_X = 0x0001CE32;
const char32_t LARGE_TYPE_PIECE_CENTRE_OF_Y = 0x0001CE33;
const char32_t LARGE_TYPE_PIECE_CENTRE_OF_Z_WITH_CROSSBAR = 0x0001CE34;
const char32_t LARGE_TYPE_PIECE_RAISED_UPPER_LEFT_ARC = 0x0001CE35;
const char32_t LARGE_TYPE_PIECE_STEM_WITH_LEFT_CROSSBAR = 0x0001CE36;
const char32_t LARGE_TYPE_PIECE_LOWER_RIGHT_AND_UPPER_RIGHT_ARC = 0x0001CE37;
const char32_t LARGE_TYPE_PIECE_DIAGONAL_UPPER_LEFT_AND_LOWER_LEFT = 0x0001CE38;
const char32_t LARGE_TYPE_PIECE_STEM_WITH_LEFT_JOINT = 0x0001CE39;
const char32_t LARGE_TYPE_PIECE_STEM_WITH_CROSSBAR = 0x0001CE3A;
const char32_t LARGE_TYPE_PIECE_DIAGONAL_UPPER_LEFT = 0x0001CE3B;
const char32_t LARGE_TYPE_PIECE_LOWER_TERMINAL = 0x0001CE3C;
const char32_t LARGE_TYPE_PIECE_LOWER_LEFT_CORNER = 0x0001CE3D;
const char32_t LARGE_TYPE_PIECE_LOWER_LEFT_ARC = 0x0001CE3E;
const char32_t LARGE_TYPE_PIECE_LOWER_LEFT_CROTCH = 0x0001CE3F;
const char32_t LARGE_TYPE_PIECE_CROSSBAR_WITH_UPPER_STEM = 0x0001CE40;
const char32_t LARGE_TYPE_PIECE_VERTEX_OF_V = 0x0001CE41;
const char32_t LARGE_TYPE_PIECE_LOWER_HALF_VERTEX_OF_W = 0x0001CE42;
const char32_t LARGE_TYPE_PIECE_LOWER_RIGHT_ARC = 0x0001CE43;
const char32_t LARGE_TYPE_PIECE_LOWER_RIGHT_CORNER = 0x0001CE44;
const char32_t LARGE_TYPE_PIECE_LOWER_RIGHT_ARC_WITH_TAIL = 0x0001CE45;
const char32_t LARGE_TYPE_PIECE_LOWER_RIGHT_CROTCH = 0x0001CE46;
const char32_t LARGE_TYPE_PIECE_STEM_45 = 0x0001CE47;
const char32_t LARGE_TYPE_PIECE_STEM_2345 = 0x0001CE48;
const char32_t LARGE_TYPE_PIECE_STEM_4 = 0x0001CE49;
const char32_t LARGE_TYPE_PIECE_STEM_34 = 0x0001CE4A;
const char32_t LARGE_TYPE_PIECE_STEM_234 = 0x0001CE4B;
const char32_t LARGE_TYPE_PIECE_STEM_1234 = 0x0001CE4C;
const char32_t LARGE_TYPE_PIECE_STEM_3 = 0x0001CE4D;
const char32_t LARGE_TYPE_PIECE_STEM_23 = 0x0001CE4E;
const char32_t LARGE_TYPE_PIECE_STEM_2 = 0x0001CE4F;
const char32_t LARGE_TYPE_PIECE_STEM_12 = 0x0001CE50;
const char32_t SEPARATED_BLOCK_SEXTANT_1 = 0x0001CE51;
const char32_t SEPARATED_BLOCK_SEXTANT_2 = 0x0001CE52;
const char32_t SEPARATED_BLOCK_SEXTANT_12 = 0x0001CE53;
const char32_t SEPARATED_BLOCK_SEXTANT_3 = 0x0001CE54;
const char32_t SEPARATED_BLOCK_SEXTANT_13 = 0x0001CE55;
const char32_t SEPARATED_BLOCK_SEXTANT_23 = 0x0001CE56;
const char32_t SEPARATED_BLOCK_SEXTANT_123 = 0x0001CE57;
const char32_t SEPARATED_BLOCK_SEXTANT_4 = 0x0001CE58;
const char32_t SEPARATED_BLOCK_SEXTANT_14 = 0x0001CE59;
const char32_t SEPARATED_BLOCK_SEXTANT_24 = 0x0001CE5A;
const char32_t SEPARATED_BLOCK_SEXTANT_124 = 0x0001CE5B;
const char32_t SEPARATED_BLOCK_SEXTANT_34 = 0x0001CE5C;
const char32_t SEPARATED_BLOCK_SEXTANT_134 = 0x0001CE5D;
const char32_t SEPARATED_BLOCK_SEXTANT_234 = 0x0001CE5E;
const char32_t SEPARATED_BLOCK_SEXTANT_1234 = 0x0001CE5F;
const char32_t SEPARATED_BLOCK_SEXTANT_5 = 0x0001CE60;
const char32_t SEPARATED_BLOCK_SEXTANT_15 = 0x0001CE61;
const char32_t SEPARATED_BLOCK_SEXTANT_25 = 0x0001CE62;
const char32_t SEPARATED_BLOCK_SEXTANT_125 = 0x0001CE63;
const char32_t SEPARATED_BLOCK_SEXTANT_35 = 0x0001CE64;
const char32_t SEPARATED_BLOCK_SEXTANT_135 = 0x0001CE65;
const char32_t SEPARATED_BLOCK_SEXTANT_235 = 0x0001CE66;
const char32_t SEPARATED_BLOCK_SEXTANT_1235 = 0x0001CE67;
const char32_t SEPARATED_BLOCK_SEXTANT_45 = 0x0001CE68;
const char32_t SEPARATED_BLOCK_SEXTANT_145 = 0x0001CE69;
const char32_t SEPARATED_BLOCK_SEXTANT_245 = 0x0001CE6A;
const char32_t SEPARATED_BLOCK_SEXTANT_1245 = 0x0001CE6B;
const char32_t SEPARATED_BLOCK_SEXTANT_345 = 0x0001CE6C;
const char32_t SEPARATED_BLOCK_SEXTANT_1345 = 0x0001CE6D;
const char32_t SEPARATED_BLOCK_SEXTANT_2345 = 0x0001CE6E;
const char32_t SEPARATED_BLOCK_SEXTANT_12345 = 0x0001CE6F;
const char32_t SEPARATED_BLOCK_SEXTANT_6 = 0x0001CE70;
const char32_t SEPARATED_BLOCK_SEXTANT_16 = 0x0001CE71;
const char32_t SEPARATED_BLOCK_SEXTANT_26 = 0x0001CE72;
const char32_t SEPARATED_BLOCK_SEXTANT_126 = 0x0001CE73;
const char32_t SEPARATED_BLOCK_SEXTANT_36 = 0x0001CE74;
const char32_t SEPARATED_BLOCK_SEXTANT_136 = 0x0001CE75;
const char32_t SEPARATED_BLOCK_SEXTANT_236 = 0x0001CE76;
const char32_t SEPARATED_BLOCK_SEXTANT_1236 = 0x0001CE77;
const char32_t SEPARATED_BLOCK_SEXTANT_46 = 0x0001CE78;
const char32_t SEPARATED_BLOCK_SEXTANT_146 = 0x0001CE79;
const char32_t SEPARATED_BLOCK_SEXTANT_246 = 0x0001CE7A;
const char32_t SEPARATED_BLOCK_SEXTANT_1246 = 0x0001CE7B;
const char32_t SEPARATED_BLOCK_SEXTANT_346 = 0x0001CE7C;
const char32_t SEPARATED_BLOCK_SEXTANT_1346 = 0x0001CE7D;
const char32_t SEPARATED_BLOCK_SEXTANT_2346 = 0x0001CE7E;
const char32_t SEPARATED_BLOCK_SEXTANT_12346 = 0x0001CE7F;
const char32_t SEPARATED_BLOCK_SEXTANT_56 = 0x0001CE80;
const char32_t SEPARATED_BLOCK_SEXTANT_156 = 0x0001CE81;
const char32_t SEPARATED_BLOCK_SEXTANT_256 = 0x0001CE82;
const char32_t SEPARATED_BLOCK_SEXTANT_1256 = 0x0001CE83;
const char32_t SEPARATED_BLOCK_SEXTANT_356 = 0x0001CE84;
const char32_t SEPARATED_BLOCK_SEXTANT_1356 = 0x0001CE85;
const char32_t SEPARATED_BLOCK_SEXTANT_2356 = 0x0001CE86;
const char32_t SEPARATED_BLOCK_SEXTANT_12356 = 0x0001CE87;
const char32_t SEPARATED_BLOCK_SEXTANT_456 = 0x0001CE88;
const char32_t SEPARATED_BLOCK_SEXTANT_1456 = 0x0001CE89;
const char32_t SEPARATED_BLOCK_SEXTANT_2456 = 0x0001CE8A;
const char32_t SEPARATED_BLOCK_SEXTANT_12456 = 0x0001CE8B;
const char32_t SEPARATED_BLOCK_SEXTANT_3456 = 0x0001CE8C;
const char32_t SEPARATED_BLOCK_SEXTANT_13456 = 0x0001CE8D;
const char32_t SEPARATED_BLOCK_SEXTANT_23456 = 0x0001CE8E;
const char32_t SEPARATED_BLOCK_SEXTANT_123456 = 0x0001CE8F;
const char32_t UPPER_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE90;
const char32_t UPPER_CENTRE_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE91;
const char32_t UPPER_CENTRE_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE92;
const char32_t UPPER_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE93;
const char32_t UPPER_MIDDLE_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE94;
const char32_t UPPER_MIDDLE_CENTRE_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE95;
const char32_t UPPER_MIDDLE_CENTRE_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE96;
const char32_t UPPER_MIDDLE_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE97;
const char32_t LOWER_MIDDLE_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE98;
const char32_t LOWER_MIDDLE_CENTRE_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE99;
const char32_t LOWER_MIDDLE_CENTRE_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE9A;
const char32_t LOWER_MIDDLE_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE9B;
const char32_t LOWER_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE9C;
const char32_t LOWER_CENTRE_LEFT_ONE_SIXTEENTH_BLOCK = 0x0001CE9D;
const char32_t LOWER_CENTRE_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE9E;
const char32_t LOWER_RIGHT_ONE_SIXTEENTH_BLOCK = 0x0001CE9F;
const char32_t RIGHT_HALF_LOWER_ONE_QUARTER_BLOCK = 0x0001CEA0;
const char32_t RIGHT_THREE_QUARTERS_LOWER_ONE_QUARTER_BLOCK = 0x0001CEA1;
const char32_t LEFT_THREE_QUARTERS_LOWER_ONE_QUARTER_BLOCK = 0x0001CEA2;
const char32_t LEFT_HALF_LOWER_ONE_QUARTER_BLOCK = 0x0001CEA3;
const char32_t LOWER_HALF_LEFT_ONE_QUARTER_BLOCK = 0x0001CEA4;
const char32_t LOWER_THREE_QUARTERS_LEFT_ONE_QUARTER_BLOCK = 0x0001CEA5;
const char32_t UPPER_THREE_QUARTERS_LEFT_ONE_QUARTER_BLOCK = 0x0001CEA6;
const char32_t UPPER_HALF_LEFT_ONE_QUARTER_BLOCK = 0x0001CEA7;
const char32_t LEFT_HALF_UPPER_ONE_QUARTER_BLOCK = 0x0001CEA8;
const char32_t LEFT_THREE_QUARTERS_UPPER_ONE_QUARTER_BLOCK = 0x0001CEA9;
const char32_t RIGHT_THREE_QUARTERS_UPPER_ONE_QUARTER_BLOCK = 0x0001CEAA;
const char32_t RIGHT_HALF_UPPER_ONE_QUARTER_BLOCK = 0x0001CEAB;
const char32_t UPPER_HALF_RIGHT_ONE_QUARTER_BLOCK = 0x0001CEAC;
const char32_t UPPER_THREE_QUARTERS_RIGHT_ONE_QUARTER_BLOCK = 0x0001CEAD;
const char32_t LOWER_THREE_QUARTERS_RIGHT_ONE_QUARTER_BLOCK = 0x0001CEAE;
const char32_t LOWER_HALF_RIGHT_ONE_QUARTER_BLOCK = 0x0001CEAF;
const char32_t HORIZONTAL_ZIGZAG_LINE = 0x0001CEB0;
const char32_t KEYHOLE = 0x0001CEB1;
const char32_t OLD_PERSONAL_COMPUTER_WITH_MONITOR_IN_PORTRAIT_ORIENTATION = 0x0001CEB2;
const char32_t BLACK_RIGHT_TRIANGLE_CARET = 0x0001CEB3;
const char32_t FRAGILE_SYMBOL = 0x0001CEBA;
const char32_t OFFICE_BUILDING_SYMBOL = 0x0001CEBB;
const char32_t TREE_SYMBOL = 0x0001CEBC;
const char32_t APPLE_SYMBOL = 0x0001CEBD;
const char32_t CHERRY_SYMBOL = 0x0001CEBE;
const char32_t STRAWBERRY_SYMBOL = 0x0001CEBF;
const char32_t HEBE = 0x0001CEC0;
const char32_t IRIS = 0x0001CEC1;
const char32_t FLORA = 0x0001CEC2;
const char32_t METIS = 0x0001CEC3;
const char32_t PARTHENOPE = 0x0001CEC4;
const char32_t VICTORIA = 0x0001CEC5;
const char32_t EGERIA = 0x0001CEC6;
const char32_t IRENE = 0x0001CEC7;
const char32_t EUNOMIA = 0x0001CEC8;
const char32_t PSYCHE = 0x0001CEC9;
const char32_t THETIS = 0x0001CECA;
const char32_t MELPOMENE = 0x0001CECB;
const char32_t FORTUNA = 0x0001CECC;
const char32_t ASTRONOMICAL_SYMBOL_FOR_ASTEROID_PROSERPINA = 0x0001CECD;
const char32_t BELLONA = 0x0001CECE;
const char32_t AMPHITRITE = 0x0001CECF;
const char32_t LEUKOTHEA = 0x0001CED0;
const char32_t GEOMANTIC_FIGURE_POPULUS = 0x0001CEE0;
const char32_t GEOMANTIC_FIGURE_TRISTITIA = 0x0001CEE1;
const char32_t GEOMANTIC_FIGURE_ALBUS = 0x0001CEE2;
const char32_t GEOMANTIC_FIGURE_FORTUNA_MAJOR = 0x0001CEE3;
const char32_t GEOMANTIC_FIGURE_RUBEUS = 0x0001CEE4;
const char32_t GEOMANTIC_FIGURE_ACQUISITIO = 0x0001CEE5;
const char32_t GEOMANTIC_FIGURE_CONJUNCTIO = 0x0001CEE6;
const char32_t GEOMANTIC_FIGURE_CAPUT_DRACONIS = 0x0001CEE7;
const char32_t GEOMANTIC_FIGURE_LAETITIA = 0x0001CEE8;
const char32_t GEOMANTIC_FIGURE_CARCER = 0x0001CEE9;
const char32_t GEOMANTIC_FIGURE_AMISSIO = 0x0001CEEA;
const char32_t GEOMANTIC_FIGURE_PUELLA = 0x0001CEEB;
const char32_t GEOMANTIC_FIGURE_FORTUNA_MINOR = 0x0001CEEC;
const char32_t GEOMANTIC_FIGURE_PUER = 0x0001CEED;
const char32_t GEOMANTIC_FIGURE_CAUDA_DRACONIS = 0x0001CEEE;
const char32_t GEOMANTIC_FIGURE_VIA = 0x0001CEEF;
const char32_t MEDIUM_SMALL_WHITE_CIRCLE_WITH_HORIZONTAL_BAR = 0x0001CEF0;
const char32_t ZNAMENNY_COMBINING_MARK_GORAZDO_NIZKO_S_KRYZHEM_ON_LEFT = 0x0001CF00;
const char32_t ZNAMENNY_COMBINING_MARK_NIZKO_S_KRYZHEM_ON_LEFT = 0x0001CF01;
const char32_t ZNAMENNY_COMBINING_MARK_TSATA_ON_LEFT = 0x0001CF02;
const char32_t ZNAMENNY_COMBINING_MARK_GORAZDO_NIZKO_ON_LEFT = 0x0001CF03;
const char32_t ZNAMENNY_COMBINING_MARK_NIZKO_ON_LEFT = 0x0001CF04;
const char32_t ZNAMENNY_COMBINING_MARK_SREDNE_ON_LEFT = 0x0001CF05;
const char32_t ZNAMENNY_COMBINING_MARK_MALO_POVYSHE_ON_LEFT = 0x0001CF06;
const char32_t ZNAMENNY_COMBINING_MARK_POVYSHE_ON_LEFT = 0x0001CF07;
const char32_t ZNAMENNY_COMBINING_MARK_VYSOKO_ON_LEFT = 0x0001CF08;
const char32_t ZNAMENNY_COMBINING_MARK_MALO_POVYSHE_S_KHOKHLOM_ON_LEFT = 0x0001CF09;
const char32_t ZNAMENNY_COMBINING_MARK_POVYSHE_S_KHOKHLOM_ON_LEFT = 0x0001CF0A;
const char32_t ZNAMENNY_COMBINING_MARK_VYSOKO_S_KHOKHLOM_ON_LEFT = 0x0001CF0B;
const char32_t ZNAMENNY_COMBINING_MARK_GORAZDO_NIZKO_S_KRYZHEM_ON_RIGHT = 0x0001CF0C;
const char32_t ZNAMENNY_COMBINING_MARK_NIZKO_S_KRYZHEM_ON_RIGHT = 0x0001CF0D;
const char32_t ZNAMENNY_COMBINING_MARK_TSATA_ON_RIGHT = 0x0001CF0E;
const char32_t ZNAMENNY_COMBINING_MARK_GORAZDO_NIZKO_ON_RIGHT = 0x0001CF0F;
const char32_t ZNAMENNY_COMBINING_MARK_NIZKO_ON_RIGHT = 0x0001CF10;
const char32_t ZNAMENNY_COMBINING_MARK_SREDNE_ON_RIGHT = 0x0001CF11;
const char32_t ZNAMENNY_COMBINING_MARK_MALO_POVYSHE_ON_RIGHT = 0x0001CF12;
const char32_t ZNAMENNY_COMBINING_MARK_POVYSHE_ON_RIGHT = 0x0001CF13;
const char32_t ZNAMENNY_COMBINING_MARK_VYSOKO_ON_RIGHT = 0x0001CF14;
const char32_t ZNAMENNY_COMBINING_MARK_MALO_POVYSHE_S_KHOKHLOM_ON_RIGHT = 0x0001CF15;
const char32_t ZNAMENNY_COMBINING_MARK_POVYSHE_S_KHOKHLOM_ON_RIGHT = 0x0001CF16;
const char32_t ZNAMENNY_COMBINING_MARK_VYSOKO_S_KHOKHLOM_ON_RIGHT = 0x0001CF17;
const char32_t ZNAMENNY_COMBINING_MARK_TSATA_S_KRYZHEM = 0x0001CF18;
const char32_t ZNAMENNY_COMBINING_MARK_MALO_POVYSHE_S_KRYZHEM = 0x0001CF19;
const char32_t ZNAMENNY_COMBINING_MARK_STRANNO_MALO_POVYSHE = 0x0001CF1A;
const char32_t ZNAMENNY_COMBINING_MARK_POVYSHE_S_KRYZHEM = 0x0001CF1B;
const char32_t ZNAMENNY_COMBINING_MARK_POVYSHE_STRANNO = 0x0001CF1C;
const char32_t ZNAMENNY_COMBINING_MARK_VYSOKO_S_KRYZHEM = 0x0001CF1D;
const char32_t ZNAMENNY_COMBINING_MARK_MALO_POVYSHE_STRANNO = 0x0001CF1E;
const char32_t ZNAMENNY_COMBINING_MARK_GORAZDO_VYSOKO = 0x0001CF1F;
const char32_t ZNAMENNY_COMBINING_MARK_ZELO = 0x0001CF20;
const char32_t ZNAMENNY_COMBINING_MARK_ON = 0x0001CF21;
const char32_t ZNAMENNY_COMBINING_MARK_RAVNO = 0x0001CF22;
const char32_t ZNAMENNY_COMBINING_MARK_TIKHAYA = 0x0001CF23;
const char32_t ZNAMENNY_COMBINING_MARK_BORZAYA = 0x0001CF24;
const char32_t ZNAMENNY_COMBINING_MARK_UDARKA = 0x0001CF25;
const char32_t ZNAMENNY_COMBINING_MARK_PODVERTKA = 0x0001CF26;
const char32_t ZNAMENNY_COMBINING_MARK_LOMKA = 0x0001CF27;
const char32_t ZNAMENNY_COMBINING_MARK_KUPNAYA = 0x0001CF28;
const char32_t ZNAMENNY_COMBINING_MARK_KACHKA = 0x0001CF29;
const char32_t ZNAMENNY_COMBINING_MARK_ZEVOK = 0x0001CF2A;
const char32_t ZNAMENNY_COMBINING_MARK_SKOBA = 0x0001CF2B;
const char32_t ZNAMENNY_COMBINING_MARK_RAZSEKA = 0x0001CF2C;
const char32_t ZNAMENNY_COMBINING_MARK_KRYZH_ON_LEFT = 0x0001CF2D;
const char32_t ZNAMENNY_COMBINING_TONAL_RANGE_MARK_MRACHNO = 0x0001CF30;
const char32_t ZNAMENNY_COMBINING_TONAL_RANGE_MARK_SVETLO = 0x0001CF31;
const char32_t ZNAMENNY_COMBINING_TONAL_RANGE_MARK_TRESVETLO = 0x0001CF32;
const char32_t ZNAMENNY_COMBINING_MARK_ZADERZHKA = 0x0001CF33;
const char32_t ZNAMENNY_COMBINING_MARK_DEMESTVENNY_ZADERZHKA = 0x0001CF34;
const char32_t ZNAMENNY_COMBINING_MARK_OTSECHKA = 0x0001CF35;
const char32_t ZNAMENNY_COMBINING_MARK_PODCHASHIE = 0x0001CF36;
const char32_t ZNAMENNY_COMBINING_MARK_PODCHASHIE_WITH_VERTICAL_STROKE = 0x0001CF37;
const char32_t ZNAMENNY_COMBINING_MARK_CHASHKA = 0x0001CF38;
const char32_t ZNAMENNY_COMBINING_MARK_CHASHKA_POLNAYA = 0x0001CF39;
const char32_t ZNAMENNY_COMBINING_MARK_OBLACHKO = 0x0001CF3A;
const char32_t ZNAMENNY_COMBINING_MARK_SOROCHYA_NOZHKA = 0x0001CF3B;
const char32_t ZNAMENNY_COMBINING_MARK_TOCHKA = 0x0001CF3C;
const char32_t ZNAMENNY_COMBINING_MARK_DVOETOCHIE = 0x0001CF3D;
const char32_t ZNAMENNY_COMBINING_ATTACHING_VERTICAL_OMET = 0x0001CF3E;
const char32_t ZNAMENNY_COMBINING_MARK_CURVED_OMET = 0x0001CF3F;
const char32_t ZNAMENNY_COMBINING_MARK_KRYZH = 0x0001CF40;
const char32_t ZNAMENNY_COMBINING_LOWER_TONAL_RANGE_INDICATOR = 0x0001CF41;
const char32_t ZNAMENNY_PRIZNAK_MODIFIER_LEVEL_2 = 0x0001CF42;
const char32_t ZNAMENNY_PRIZNAK_MODIFIER_LEVEL_3 = 0x0001CF43;
const char32_t ZNAMENNY_PRIZNAK_MODIFIER_DIRECTION_FLIP = 0x0001CF44;
const char32_t ZNAMENNY_PRIZNAK_MODIFIER_KRYZH = 0x0001CF45;
const char32_t ZNAMENNY_PRIZNAK_MODIFIER_ROG = 0x0001CF46;
const char32_t ZNAMENNY_NEUME_KRYUK = 0x0001CF50;
const char32_t ZNAMENNY_NEUME_KRYUK_TIKHY = 0x0001CF51;
const char32_t ZNAMENNY_NEUME_PARAKLIT = 0x0001CF52;
const char32_t ZNAMENNY_NEUME_DVA_V_CHELNU = 0x0001CF53;
const char32_t ZNAMENNY_NEUME_KLYUCH = 0x0001CF54;
const char32_t ZNAMENNY_NEUME_ZANOZHEK = 0x0001CF55;
const char32_t ZNAMENNY_NEUME_STOPITSA = 0x0001CF56;
const char32_t ZNAMENNY_NEUME_STOPITSA_S_OCHKOM = 0x0001CF57;
const char32_t ZNAMENNY_NEUME_PEREVODKA = 0x0001CF58;
const char32_t ZNAMENNY_NEUME_PEREVODKA_NEPOSTOYANNAYA = 0x0001CF59;
const char32_t ZNAMENNY_NEUME_STOPITSA_WITH_SOROCHYA_NOZHKA = 0x0001CF5A;
const char32_t ZNAMENNY_NEUME_CHELYUSTKA = 0x0001CF5B;
const char32_t ZNAMENNY_NEUME_PALKA = 0x0001CF5C;
const char32_t ZNAMENNY_NEUME_ZAPYATAYA = 0x0001CF5D;
const char32_t ZNAMENNY_NEUME_GOLUBCHIK_BORZY = 0x0001CF5E;
const char32_t ZNAMENNY_NEUME_GOLUBCHIK_TIKHY = 0x0001CF5F;
const char32_t ZNAMENNY_NEUME_GOLUBCHIK_MRACHNY = 0x0001CF60;
const char32_t ZNAMENNY_NEUME_GOLUBCHIK_SVETLY = 0x0001CF61;
const char32_t ZNAMENNY_NEUME_GOLUBCHIK_TRESVETLY = 0x0001CF62;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_PROSTAYA = 0x0001CF63;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_MRACHNAYA = 0x0001CF64;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_SVETLAYA = 0x0001CF65;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_TRESVETLAYA = 0x0001CF66;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_KLYUCHEVAYA_PROSTAYA = 0x0001CF67;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_KLYUCHEVAYA_MRACHNAYA = 0x0001CF68;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_KLYUCHEVAYA_SVETLAYA = 0x0001CF69;
const char32_t ZNAMENNY_NEUME_VRAKHIYA_KLYUCHEVAYA_TRESVETLAYA = 0x0001CF6A;
const char32_t ZNAMENNY_NEUME_DOUBLE_ZAPYATAYA = 0x0001CF6B;
const char32_t ZNAMENNY_NEUME_REVERSED_CHELYUSTKA = 0x0001CF6C;
const char32_t ZNAMENNY_NEUME_DERBITSA = 0x0001CF6D;
const char32_t ZNAMENNY_NEUME_KHAMILO = 0x0001CF6E;
const char32_t ZNAMENNY_NEUME_CHASHKA = 0x0001CF6F;
const char32_t ZNAMENNY_NEUME_PODCHASHIE = 0x0001CF70;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_MRACHNAYA = 0x0001CF71;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_SVETLAYA = 0x0001CF72;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_TRESVETLAYA = 0x0001CF73;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_TIKHAYA = 0x0001CF74;
const char32_t ZNAMENNY_NEUME_DEMESTVENNY_KLYUCH = 0x0001CF75;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_KLYUCHEVAYA_SVETLAYA = 0x0001CF76;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_KLYUCHENEPOSTOYANNAYA = 0x0001CF77;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_KLYUCHEVAYA_TIKHAYA = 0x0001CF78;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_DVOECHELNAYA_PROSTAYA = 0x0001CF79;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_DVOECHELNAYA_SVETLAYA = 0x0001CF7A;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_DVOECHELNAYA_NEPOSTOYANNAYA = 0x0001CF7B;
const char32_t ZNAMENNY_NEUME_SKAMEYTSA_DVOECHELNAYA_KLYUCHEVAYA = 0x0001CF7C;
const char32_t ZNAMENNY_NEUME_SLOZHITIE = 0x0001CF7D;
const char32_t ZNAMENNY_NEUME_SLOZHITIE_S_ZAPYATOY = 0x0001CF7E;
const char32_t ZNAMENNY_NEUME_SLOZHITIE_ZAKRYTOE = 0x0001CF7F;
const char32_t ZNAMENNY_NEUME_SLOZHITIE_S_KRYZHEM = 0x0001CF80;
const char32_t ZNAMENNY_NEUME_KRYZH = 0x0001CF81;
const char32_t ZNAMENNY_NEUME_ROG = 0x0001CF82;
const char32_t ZNAMENNY_NEUME_FITA = 0x0001CF83;
const char32_t ZNAMENNY_NEUME_KOBYLA = 0x0001CF84;
const char32_t ZNAMENNY_NEUME_ZMEYTSA = 0x0001CF85;
const char32_t ZNAMENNY_NEUME_STATYA = 0x0001CF86;
const char32_t ZNAMENNY_NEUME_STATYA_S_ZAPYATOY = 0x0001CF87;
const char32_t ZNAMENNY_NEUME_STATYA_S_KRYZHEM = 0x0001CF88;
const char32_t ZNAMENNY_NEUME_STATYA_S_ZAPYATOY_I_KRYZHEM = 0x0001CF89;
const char32_t ZNAMENNY_NEUME_STATYA_S_KRYZHEM_I_ZAPYATOY = 0x0001CF8A;
const char32_t ZNAMENNY_NEUME_STATYA_ZAKRYTAYA = 0x0001CF8B;
const char32_t ZNAMENNY_NEUME_STATYA_ZAKRYTAYA_S_ZAPYATOY = 0x0001CF8C;
const char32_t ZNAMENNY_NEUME_STATYA_S_ROGOM = 0x0001CF8D;
const char32_t ZNAMENNY_NEUME_STATYA_S_DVUMYA_ZAPYATYMI = 0x0001CF8E;
const char32_t ZNAMENNY_NEUME_STATYA_S_ZAPYATOY_I_PODCHASHIEM = 0x0001CF8F;
const char32_t ZNAMENNY_NEUME_POLKULIZMY = 0x0001CF90;
const char32_t ZNAMENNY_NEUME_STATYA_NEPOSTOYANNAYA = 0x0001CF91;
const char32_t ZNAMENNY_NEUME_STRELA_PROSTAYA = 0x0001CF92;
const char32_t ZNAMENNY_NEUME_STRELA_MRACHNOTIKHAYA = 0x0001CF93;
const char32_t ZNAMENNY_NEUME_STRELA_KRYZHEVAYA = 0x0001CF94;
const char32_t ZNAMENNY_NEUME_STRELA_POLUPOVODNAYA = 0x0001CF95;
const char32_t ZNAMENNY_NEUME_STRELA_POVODNAYA = 0x0001CF96;
const char32_t ZNAMENNY_NEUME_STRELA_NEPOSTOYANNAYA = 0x0001CF97;
const char32_t ZNAMENNY_NEUME_STRELA_KLYUCHEPOVODNAYA = 0x0001CF98;
const char32_t ZNAMENNY_NEUME_STRELA_KLYUCHENEPOSTOYANNAYA = 0x0001CF99;
const char32_t ZNAMENNY_NEUME_STRELA_TIKHAYA_PUTNAYA = 0x0001CF9A;
const char32_t ZNAMENNY_NEUME_STRELA_DVOECHELNAYA = 0x0001CF9B;
const char32_t ZNAMENNY_NEUME_STRELA_DVOECHELNOKRYZHEVAYA = 0x0001CF9C;
const char32_t ZNAMENNY_NEUME_STRELA_DVOECHELNOPOVODNAYA = 0x0001CF9D;
const char32_t ZNAMENNY_NEUME_STRELA_DVOECHELNAYA_KLYUCHEVAYA = 0x0001CF9E;
const char32_t ZNAMENNY_NEUME_STRELA_DVOECHELNOPOVODNAYA_KLYUCHEVAYA = 0x0001CF9F;
const char32_t ZNAMENNY_NEUME_STRELA_GROMNAYA_WITH_SINGLE_ZAPYATAYA = 0x0001CFA0;
const char32_t ZNAMENNY_NEUME_STRELA_GROMOPOVODNAYA_WITH_SINGLE_ZAPYATAYA = 0x0001CFA1;
const char32_t ZNAMENNY_NEUME_STRELA_GROMNAYA = 0x0001CFA2;
const char32_t ZNAMENNY_NEUME_STRELA_GROMOPOVODNAYA = 0x0001CFA3;
const char32_t ZNAMENNY_NEUME_STRELA_GROMOPOVODNAYA_WITH_DOUBLE_ZAPYATAYA = 0x0001CFA4;
const char32_t ZNAMENNY_NEUME_STRELA_GROMOKRYZHEVAYA = 0x0001CFA5;
const char32_t ZNAMENNY_NEUME_STRELA_GROMOKRYZHEVAYA_POVODNAYA = 0x0001CFA6;
const char32_t ZNAMENNY_NEUME_MECHIK = 0x0001CFA7;
const char32_t ZNAMENNY_NEUME_MECHIK_POVODNY = 0x0001CFA8;
const char32_t ZNAMENNY_NEUME_MECHIK_KLYUCHEVOY = 0x0001CFA9;
const char32_t ZNAMENNY_NEUME_MECHIK_KLYUCHEPOVODNY = 0x0001CFAA;
const char32_t ZNAMENNY_NEUME_MECHIK_KLYUCHENEPOSTOYANNY = 0x0001CFAB;
const char32_t ZNAMENNY_NEUME_STRELA_TRYASOGLASNAYA = 0x0001CFAC;
const char32_t ZNAMENNY_NEUME_STRELA_TRYASOPOVODNAYA = 0x0001CFAD;
const char32_t ZNAMENNY_NEUME_STRELA_TRYASOSTRELNAYA = 0x0001CFAE;
const char32_t ZNAMENNY_NEUME_OSOKA = 0x0001CFAF;
const char32_t ZNAMENNY_NEUME_OSOKA_SVETLAYA = 0x0001CFB0;
const char32_t ZNAMENNY_NEUME_OSOKA_TRESVETLAYA = 0x0001CFB1;
const char32_t ZNAMENNY_NEUME_OSOKA_KRYUKOVAYA_SVETLAYA = 0x0001CFB2;
const char32_t ZNAMENNY_NEUME_OSOKA_KLYUCHEVAYA_SVETLAYA = 0x0001CFB3;
const char32_t ZNAMENNY_NEUME_OSOKA_KLYUCHEVAYA_NEPOSTOYANNAYA = 0x0001CFB4;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA = 0x0001CFB5;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_POVODNAYA = 0x0001CFB6;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_GROMNAYA_WITH_SINGLE_ZAPYATAYA = 0x0001CFB7;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_GROMOPOVODNAYA_WITH_SINGLE_ZAPYATAYA = 0x0001CFB8;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_GROMNAYA = 0x0001CFB9;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_GROMOPOVODNAYA = 0x0001CFBA;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_GROMOPOVODNAYA_WITH_DOUBLE_ZAPYATAYA = 0x0001CFBB;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_GROMOKRYZHEVAYA = 0x0001CFBC;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_GROMOKRYZHEVAYA_POVODNAYA = 0x0001CFBD;
const char32_t ZNAMENNY_NEUME_STRELA_KRYUKOVAYA_TRYASKA = 0x0001CFBE;
const char32_t ZNAMENNY_NEUME_KUFISMA = 0x0001CFBF;
const char32_t ZNAMENNY_NEUME_OBLAKO = 0x0001CFC0;
const char32_t ZNAMENNY_NEUME_DUDA = 0x0001CFC1;
const char32_t ZNAMENNY_NEUME_NEMKA = 0x0001CFC2;
const char32_t ZNAMENNY_NEUME_PAUK = 0x0001CFC3;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PSILI = 0x0001D000;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DASEIA = 0x0001D001;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PERISPOMENI = 0x0001D002;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OXEIA_EKFONITIKON = 0x0001D003;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OXEIA_DIPLI = 0x0001D004;
const char32_t BYZANTINE_MUSICAL_SYMBOL_VAREIA_EKFONITIKON = 0x0001D005;
const char32_t BYZANTINE_MUSICAL_SYMBOL_VAREIA_DIPLI = 0x0001D006;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KATHISTI = 0x0001D007;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SYRMATIKI = 0x0001D008;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PARAKLITIKI = 0x0001D009;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YPOKRISIS = 0x0001D00A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YPOKRISIS_DIPLI = 0x0001D00B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KREMASTI = 0x0001D00C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APESO_EKFONITIKON = 0x0001D00D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_EXO_EKFONITIKON = 0x0001D00E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TELEIA = 0x0001D00F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMATA = 0x0001D010;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APOSTROFOS = 0x0001D011;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APOSTROFOS_DIPLI = 0x0001D012;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SYNEVMA = 0x0001D013;
const char32_t BYZANTINE_MUSICAL_SYMBOL_THITA = 0x0001D014;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OLIGON_ARCHAION = 0x0001D015;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORGON_ARCHAION = 0x0001D016;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PSILON = 0x0001D017;
const char32_t BYZANTINE_MUSICAL_SYMBOL_CHAMILON = 0x0001D018;
const char32_t BYZANTINE_MUSICAL_SYMBOL_VATHY = 0x0001D019;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ISON_ARCHAION = 0x0001D01A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMA_ARCHAION = 0x0001D01B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMATA_ARCHAION = 0x0001D01C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SAXIMATA = 0x0001D01D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PARICHON = 0x0001D01E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_STAVROS_APODEXIA = 0x0001D01F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OXEIAI_ARCHAION = 0x0001D020;
const char32_t BYZANTINE_MUSICAL_SYMBOL_VAREIAI_ARCHAION = 0x0001D021;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APODERMA_ARCHAION = 0x0001D022;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APOTHEMA = 0x0001D023;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KLASMA = 0x0001D024;
const char32_t BYZANTINE_MUSICAL_SYMBOL_REVMA = 0x0001D025;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PIASMA_ARCHAION = 0x0001D026;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TINAGMA = 0x0001D027;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ANATRICHISMA = 0x0001D028;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SEISMA = 0x0001D029;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SYNAGMA_ARCHAION = 0x0001D02A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SYNAGMA_META_STAVROU = 0x0001D02B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OYRANISMA_ARCHAION = 0x0001D02C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_THEMA = 0x0001D02D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_LEMOI = 0x0001D02E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DYO = 0x0001D02F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TRIA = 0x0001D030;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TESSERA = 0x0001D031;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KRATIMATA = 0x0001D032;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APESO_EXO_NEO = 0x0001D033;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_ARCHAION = 0x0001D034;
const char32_t BYZANTINE_MUSICAL_SYMBOL_IMIFTHORA = 0x0001D035;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TROMIKON_ARCHAION = 0x0001D036;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KATAVA_TROMIKON = 0x0001D037;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PELASTON = 0x0001D038;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PSIFISTON = 0x0001D039;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KONTEVMA = 0x0001D03A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_CHOREVMA_ARCHAION = 0x0001D03B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_RAPISMA = 0x0001D03C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PARAKALESMA_ARCHAION = 0x0001D03D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PARAKLITIKI_ARCHAION = 0x0001D03E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ICHADIN = 0x0001D03F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_NANA = 0x0001D040;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PETASMA = 0x0001D041;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KONTEVMA_ALLO = 0x0001D042;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TROMIKON_ALLO = 0x0001D043;
const char32_t BYZANTINE_MUSICAL_SYMBOL_STRAGGISMATA = 0x0001D044;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GRONTHISMATA = 0x0001D045;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ISON_NEO = 0x0001D046;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OLIGON_NEO = 0x0001D047;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OXEIA_NEO = 0x0001D048;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PETASTI = 0x0001D049;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KOUFISMA = 0x0001D04A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PETASTOKOUFISMA = 0x0001D04B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KRATIMOKOUFISMA = 0x0001D04C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PELASTON_NEO = 0x0001D04D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMATA_NEO_ANO = 0x0001D04E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMA_NEO_ANO = 0x0001D04F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YPSILI = 0x0001D050;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APOSTROFOS_NEO = 0x0001D051;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APOSTROFOI_SYNDESMOS_NEO = 0x0001D052;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YPORROI = 0x0001D053;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KRATIMOYPORROON = 0x0001D054;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ELAFRON = 0x0001D055;
const char32_t BYZANTINE_MUSICAL_SYMBOL_CHAMILI = 0x0001D056;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MIKRON_ISON = 0x0001D057;
const char32_t BYZANTINE_MUSICAL_SYMBOL_VAREIA_NEO = 0x0001D058;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PIASMA_NEO = 0x0001D059;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PSIFISTON_NEO = 0x0001D05A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OMALON = 0x0001D05B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ANTIKENOMA = 0x0001D05C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_LYGISMA = 0x0001D05D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PARAKLITIKI_NEO = 0x0001D05E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PARAKALESMA_NEO = 0x0001D05F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ETERON_PARAKALESMA = 0x0001D060;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KYLISMA = 0x0001D061;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ANTIKENOKYLISMA = 0x0001D062;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TROMIKON_NEO = 0x0001D063;
const char32_t BYZANTINE_MUSICAL_SYMBOL_EKSTREPTON = 0x0001D064;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SYNAGMA_NEO = 0x0001D065;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SYRMA = 0x0001D066;
const char32_t BYZANTINE_MUSICAL_SYMBOL_CHOREVMA_NEO = 0x0001D067;
const char32_t BYZANTINE_MUSICAL_SYMBOL_EPEGERMA = 0x0001D068;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SEISMA_NEO = 0x0001D069;
const char32_t BYZANTINE_MUSICAL_SYMBOL_XIRON_KLASMA = 0x0001D06A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TROMIKOPSIFISTON = 0x0001D06B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PSIFISTOLYGISMA = 0x0001D06C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TROMIKOLYGISMA = 0x0001D06D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TROMIKOPARAKALESMA = 0x0001D06E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PSIFISTOPARAKALESMA = 0x0001D06F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TROMIKOSYNAGMA = 0x0001D070;
const char32_t BYZANTINE_MUSICAL_SYMBOL_PSIFISTOSYNAGMA = 0x0001D071;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORGOSYNTHETON = 0x0001D072;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARGOSYNTHETON = 0x0001D073;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ETERON_ARGOSYNTHETON = 0x0001D074;
const char32_t BYZANTINE_MUSICAL_SYMBOL_OYRANISMA_NEO = 0x0001D075;
const char32_t BYZANTINE_MUSICAL_SYMBOL_THEMATISMOS_ESO = 0x0001D076;
const char32_t BYZANTINE_MUSICAL_SYMBOL_THEMATISMOS_EXO = 0x0001D077;
const char32_t BYZANTINE_MUSICAL_SYMBOL_THEMA_APLOUN = 0x0001D078;
const char32_t BYZANTINE_MUSICAL_SYMBOL_THES_KAI_APOTHES = 0x0001D079;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KATAVASMA = 0x0001D07A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ENDOFONON = 0x0001D07B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YFEN_KATO = 0x0001D07C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YFEN_ANO = 0x0001D07D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_STAVROS = 0x0001D07E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KLASMA_ANO = 0x0001D07F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIPLI_ARCHAION = 0x0001D080;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KRATIMA_ARCHAION = 0x0001D081;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KRATIMA_ALLO = 0x0001D082;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KRATIMA_NEO = 0x0001D083;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APODERMA_NEO = 0x0001D084;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APLI = 0x0001D085;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIPLI = 0x0001D086;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TRIPLI = 0x0001D087;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TETRAPLI = 0x0001D088;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KORONIS = 0x0001D089;
const char32_t BYZANTINE_MUSICAL_SYMBOL_LEIMMA_ENOS_CHRONOU = 0x0001D08A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_LEIMMA_DYO_CHRONON = 0x0001D08B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_LEIMMA_TRION_CHRONON = 0x0001D08C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_LEIMMA_TESSARON_CHRONON = 0x0001D08D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_LEIMMA_IMISEOS_CHRONOU = 0x0001D08E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORGON_NEO_ANO = 0x0001D08F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORGON_PARESTIGMENON_ARISTERA = 0x0001D090;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORGON_PARESTIGMENON_DEXIA = 0x0001D091;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIGORGON = 0x0001D092;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIGORGON_PARESTIGMENON_ARISTERA_KATO = 0x0001D093;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIGORGON_PARESTIGMENON_ARISTERA_ANO = 0x0001D094;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIGORGON_PARESTIGMENON_DEXIA = 0x0001D095;
const char32_t BYZANTINE_MUSICAL_SYMBOL_TRIGORGON = 0x0001D096;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARGON = 0x0001D097;
const char32_t BYZANTINE_MUSICAL_SYMBOL_IMIDIARGON = 0x0001D098;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIARGON = 0x0001D099;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_POLI_ARGI = 0x0001D09A;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_ARGOTERI = 0x0001D09B;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_ARGI = 0x0001D09C;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_METRIA = 0x0001D09D;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_MESI = 0x0001D09E;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_GORGI = 0x0001D09F;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_GORGOTERI = 0x0001D0A0;
const char32_t BYZANTINE_MUSICAL_SYMBOL_AGOGI_POLI_GORGI = 0x0001D0A1;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_PROTOS_ICHOS = 0x0001D0A2;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_ALLI_PROTOS_ICHOS = 0x0001D0A3;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_DEYTEROS_ICHOS = 0x0001D0A4;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_ALLI_DEYTEROS_ICHOS = 0x0001D0A5;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_TRITOS_ICHOS = 0x0001D0A6;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_TRIFONIAS = 0x0001D0A7;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_TETARTOS_ICHOS = 0x0001D0A8;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_TETARTOS_LEGETOS_ICHOS = 0x0001D0A9;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_LEGETOS_ICHOS = 0x0001D0AA;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_PLAGIOS_ICHOS = 0x0001D0AB;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ISAKIA_TELOUS_ICHIMATOS = 0x0001D0AC;
const char32_t BYZANTINE_MUSICAL_SYMBOL_APOSTROFOI_TELOUS_ICHIMATOS = 0x0001D0AD;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FANEROSIS_TETRAFONIAS = 0x0001D0AE;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FANEROSIS_MONOFONIAS = 0x0001D0AF;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FANEROSIS_DIFONIAS = 0x0001D0B0;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_VARYS_ICHOS = 0x0001D0B1;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_PROTOVARYS_ICHOS = 0x0001D0B2;
const char32_t BYZANTINE_MUSICAL_SYMBOL_MARTYRIA_PLAGIOS_TETARTOS_ICHOS = 0x0001D0B3;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORTHMIKON_N_APLOUN = 0x0001D0B4;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORTHMIKON_N_DIPLOUN = 0x0001D0B5;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ENARXIS_KAI_FTHORA_VOU = 0x0001D0B6;
const char32_t BYZANTINE_MUSICAL_SYMBOL_IMIFONON = 0x0001D0B7;
const char32_t BYZANTINE_MUSICAL_SYMBOL_IMIFTHORON = 0x0001D0B8;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_ARCHAION_DEYTEROU_ICHOU = 0x0001D0B9;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_DIATONIKI_PA = 0x0001D0BA;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_DIATONIKI_NANA = 0x0001D0BB;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_NAOS_ICHOS = 0x0001D0BC;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_DIATONIKI_DI = 0x0001D0BD;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_SKLIRON_DIATONON_DI = 0x0001D0BE;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_DIATONIKI_KE = 0x0001D0BF;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_DIATONIKI_ZO = 0x0001D0C0;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_DIATONIKI_NI_KATO = 0x0001D0C1;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_DIATONIKI_NI_ANO = 0x0001D0C2;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_MALAKON_CHROMA_DIFONIAS = 0x0001D0C3;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_MALAKON_CHROMA_MONOFONIAS = 0x0001D0C4;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FHTORA_SKLIRON_CHROMA_VASIS = 0x0001D0C5;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_SKLIRON_CHROMA_SYNAFI = 0x0001D0C6;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_NENANO = 0x0001D0C7;
const char32_t BYZANTINE_MUSICAL_SYMBOL_CHROA_ZYGOS = 0x0001D0C8;
const char32_t BYZANTINE_MUSICAL_SYMBOL_CHROA_KLITON = 0x0001D0C9;
const char32_t BYZANTINE_MUSICAL_SYMBOL_CHROA_SPATHI = 0x0001D0CA;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_I_YFESIS_TETARTIMORION = 0x0001D0CB;
const char32_t BYZANTINE_MUSICAL_SYMBOL_FTHORA_ENARMONIOS_ANTIFONIA = 0x0001D0CC;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YFESIS_TRITIMORION = 0x0001D0CD;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIESIS_TRITIMORION = 0x0001D0CE;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIESIS_TETARTIMORION = 0x0001D0CF;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIESIS_APLI_DYO_DODEKATA = 0x0001D0D0;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIESIS_MONOGRAMMOS_TESSERA_DODEKATA = 0x0001D0D1;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIESIS_DIGRAMMOS_EX_DODEKATA = 0x0001D0D2;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIESIS_TRIGRAMMOS_OKTO_DODEKATA = 0x0001D0D3;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YFESIS_APLI_DYO_DODEKATA = 0x0001D0D4;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YFESIS_MONOGRAMMOS_TESSERA_DODEKATA = 0x0001D0D5;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YFESIS_DIGRAMMOS_EX_DODEKATA = 0x0001D0D6;
const char32_t BYZANTINE_MUSICAL_SYMBOL_YFESIS_TRIGRAMMOS_OKTO_DODEKATA = 0x0001D0D7;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GENIKI_DIESIS = 0x0001D0D8;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GENIKI_YFESIS = 0x0001D0D9;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIASTOLI_APLI_MIKRI = 0x0001D0DA;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIASTOLI_APLI_MEGALI = 0x0001D0DB;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIASTOLI_DIPLI = 0x0001D0DC;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIASTOLI_THESEOS = 0x0001D0DD;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_THESEOS = 0x0001D0DE;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_THESEOS_DISIMOU = 0x0001D0DF;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_THESEOS_TRISIMOU = 0x0001D0E0;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_THESEOS_TETRASIMOU = 0x0001D0E1;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_ARSEOS = 0x0001D0E2;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_ARSEOS_DISIMOU = 0x0001D0E3;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_ARSEOS_TRISIMOU = 0x0001D0E4;
const char32_t BYZANTINE_MUSICAL_SYMBOL_SIMANSIS_ARSEOS_TETRASIMOU = 0x0001D0E5;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIGRAMMA_GG = 0x0001D0E6;
const char32_t BYZANTINE_MUSICAL_SYMBOL_DIFTOGGOS_OU = 0x0001D0E7;
const char32_t BYZANTINE_MUSICAL_SYMBOL_STIGMA = 0x0001D0E8;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARKTIKO_PA = 0x0001D0E9;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARKTIKO_VOU = 0x0001D0EA;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARKTIKO_GA = 0x0001D0EB;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARKTIKO_DI = 0x0001D0EC;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARKTIKO_KE = 0x0001D0ED;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARKTIKO_ZO = 0x0001D0EE;
const char32_t BYZANTINE_MUSICAL_SYMBOL_ARKTIKO_NI = 0x0001D0EF;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMATA_NEO_MESO = 0x0001D0F0;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMA_NEO_MESO = 0x0001D0F1;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMATA_NEO_KATO = 0x0001D0F2;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KENTIMA_NEO_KATO = 0x0001D0F3;
const char32_t BYZANTINE_MUSICAL_SYMBOL_KLASMA_KATO = 0x0001D0F4;
const char32_t BYZANTINE_MUSICAL_SYMBOL_GORGON_NEO_KATO = 0x0001D0F5;
const char32_t MUSICAL_SYMBOL_SINGLE_BARLINE = 0x0001D100;
const char32_t MUSICAL_SYMBOL_DOUBLE_BARLINE = 0x0001D101;
const char32_t MUSICAL_SYMBOL_FINAL_BARLINE = 0x0001D102;
const char32_t MUSICAL_SYMBOL_REVERSE_FINAL_BARLINE = 0x0001D103;
const char32_t MUSICAL_SYMBOL_DASHED_BARLINE = 0x0001D104;
const char32_t MUSICAL_SYMBOL_SHORT_BARLINE = 0x0001D105;
const char32_t MUSICAL_SYMBOL_LEFT_REPEAT_SIGN = 0x0001D106;
const char32_t MUSICAL_SYMBOL_RIGHT_REPEAT_SIGN = 0x0001D107;
const char32_t MUSICAL_SYMBOL_REPEAT_DOTS = 0x0001D108;
const char32_t MUSICAL_SYMBOL_DAL_SEGNO = 0x0001D109;
const char32_t MUSICAL_SYMBOL_DA_CAPO = 0x0001D10A;
const char32_t MUSICAL_SYMBOL_SEGNO = 0x0001D10B;
const char32_t MUSICAL_SYMBOL_CODA = 0x0001D10C;
const char32_t MUSICAL_SYMBOL_REPEATED_FIGURE_1 = 0x0001D10D;
const char32_t MUSICAL_SYMBOL_REPEATED_FIGURE_2 = 0x0001D10E;
const char32_t MUSICAL_SYMBOL_REPEATED_FIGURE_3 = 0x0001D10F;
const char32_t MUSICAL_SYMBOL_FERMATA = 0x0001D110;
const char32_t MUSICAL_SYMBOL_FERMATA_BELOW = 0x0001D111;
const char32_t MUSICAL_SYMBOL_BREATH_MARK = 0x0001D112;
const char32_t MUSICAL_SYMBOL_CAESURA = 0x0001D113;
const char32_t MUSICAL_SYMBOL_BRACE = 0x0001D114;
const char32_t MUSICAL_SYMBOL_BRACKET = 0x0001D115;
const char32_t MUSICAL_SYMBOL_ONE_LINE_STAFF = 0x0001D116;
const char32_t MUSICAL_SYMBOL_TWO_LINE_STAFF = 0x0001D117;
const char32_t MUSICAL_SYMBOL_THREE_LINE_STAFF = 0x0001D118;
const char32_t MUSICAL_SYMBOL_FOUR_LINE_STAFF = 0x0001D119;
const char32_t MUSICAL_SYMBOL_FIVE_LINE_STAFF = 0x0001D11A;
const char32_t MUSICAL_SYMBOL_SIX_LINE_STAFF = 0x0001D11B;
const char32_t MUSICAL_SYMBOL_SIX_STRING_FRETBOARD = 0x0001D11C;
const char32_t MUSICAL_SYMBOL_FOUR_STRING_FRETBOARD = 0x0001D11D;
const char32_t MUSICAL_SYMBOL_G_CLEF = 0x0001D11E;
const char32_t MUSICAL_SYMBOL_G_CLEF_OTTAVA_ALTA = 0x0001D11F;
const char32_t MUSICAL_SYMBOL_G_CLEF_OTTAVA_BASSA = 0x0001D120;
const char32_t MUSICAL_SYMBOL_C_CLEF = 0x0001D121;
const char32_t MUSICAL_SYMBOL_F_CLEF = 0x0001D122;
const char32_t MUSICAL_SYMBOL_F_CLEF_OTTAVA_ALTA = 0x0001D123;
const char32_t MUSICAL_SYMBOL_F_CLEF_OTTAVA_BASSA = 0x0001D124;
const char32_t MUSICAL_SYMBOL_DRUM_CLEF_1 = 0x0001D125;
const char32_t MUSICAL_SYMBOL_DRUM_CLEF_2 = 0x0001D126;
const char32_t MUSICAL_SYMBOL_MULTIPLE_MEASURE_REST = 0x0001D129;
const char32_t MUSICAL_SYMBOL_DOUBLE_SHARP = 0x0001D12A;
const char32_t MUSICAL_SYMBOL_DOUBLE_FLAT = 0x0001D12B;
const char32_t MUSICAL_SYMBOL_FLAT_UP = 0x0001D12C;
const char32_t MUSICAL_SYMBOL_FLAT_DOWN = 0x0001D12D;
const char32_t MUSICAL_SYMBOL_NATURAL_UP = 0x0001D12E;
const char32_t MUSICAL_SYMBOL_NATURAL_DOWN = 0x0001D12F;
const char32_t MUSICAL_SYMBOL_SHARP_UP = 0x0001D130;
const char32_t MUSICAL_SYMBOL_SHARP_DOWN = 0x0001D131;
const char32_t MUSICAL_SYMBOL_QUARTER_TONE_SHARP = 0x0001D132;
const char32_t MUSICAL_SYMBOL_QUARTER_TONE_FLAT = 0x0001D133;
const char32_t MUSICAL_SYMBOL_COMMON_TIME = 0x0001D134;
const char32_t MUSICAL_SYMBOL_CUT_TIME = 0x0001D135;
const char32_t MUSICAL_SYMBOL_OTTAVA_ALTA = 0x0001D136;
const char32_t MUSICAL_SYMBOL_OTTAVA_BASSA = 0x0001D137;
const char32_t MUSICAL_SYMBOL_QUINDICESIMA_ALTA = 0x0001D138;
const char32_t MUSICAL_SYMBOL_QUINDICESIMA_BASSA = 0x0001D139;
const char32_t MUSICAL_SYMBOL_MULTI_REST = 0x0001D13A;
const char32_t MUSICAL_SYMBOL_WHOLE_REST = 0x0001D13B;
const char32_t MUSICAL_SYMBOL_HALF_REST = 0x0001D13C;
const char32_t MUSICAL_SYMBOL_QUARTER_REST = 0x0001D13D;
const char32_t MUSICAL_SYMBOL_EIGHTH_REST = 0x0001D13E;
const char32_t MUSICAL_SYMBOL_SIXTEENTH_REST = 0x0001D13F;
const char32_t MUSICAL_SYMBOL_THIRTY_SECOND_REST = 0x0001D140;
const char32_t MUSICAL_SYMBOL_SIXTY_FOURTH_REST = 0x0001D141;
const char32_t MUSICAL_SYMBOL_ONE_HUNDRED_TWENTY_EIGHTH_REST = 0x0001D142;
const char32_t MUSICAL_SYMBOL_X_NOTEHEAD = 0x0001D143;
const char32_t MUSICAL_SYMBOL_PLUS_NOTEHEAD = 0x0001D144;
const char32_t MUSICAL_SYMBOL_CIRCLE_X_NOTEHEAD = 0x0001D145;
const char32_t MUSICAL_SYMBOL_SQUARE_NOTEHEAD_WHITE = 0x0001D146;
const char32_t MUSICAL_SYMBOL_SQUARE_NOTEHEAD_BLACK = 0x0001D147;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_UP_WHITE = 0x0001D148;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_UP_BLACK = 0x0001D149;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_LEFT_WHITE = 0x0001D14A;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_LEFT_BLACK = 0x0001D14B;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_RIGHT_WHITE = 0x0001D14C;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_RIGHT_BLACK = 0x0001D14D;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_DOWN_WHITE = 0x0001D14E;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_DOWN_BLACK = 0x0001D14F;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_UP_RIGHT_WHITE = 0x0001D150;
const char32_t MUSICAL_SYMBOL_TRIANGLE_NOTEHEAD_UP_RIGHT_BLACK = 0x0001D151;
const char32_t MUSICAL_SYMBOL_MOON_NOTEHEAD_WHITE = 0x0001D152;
const char32_t MUSICAL_SYMBOL_MOON_NOTEHEAD_BLACK = 0x0001D153;
const char32_t MUSICAL_SYMBOL_TRIANGLE_ROUND_NOTEHEAD_DOWN_WHITE = 0x0001D154;
const char32_t MUSICAL_SYMBOL_TRIANGLE_ROUND_NOTEHEAD_DOWN_BLACK = 0x0001D155;
const char32_t MUSICAL_SYMBOL_PARENTHESIS_NOTEHEAD = 0x0001D156;
const char32_t MUSICAL_SYMBOL_VOID_NOTEHEAD = 0x0001D157;
const char32_t MUSICAL_SYMBOL_NOTEHEAD_BLACK = 0x0001D158;
const char32_t MUSICAL_SYMBOL_NULL_NOTEHEAD = 0x0001D159;
const char32_t MUSICAL_SYMBOL_CLUSTER_NOTEHEAD_WHITE = 0x0001D15A;
const char32_t MUSICAL_SYMBOL_CLUSTER_NOTEHEAD_BLACK = 0x0001D15B;
const char32_t MUSICAL_SYMBOL_BREVE = 0x0001D15C;
const char32_t MUSICAL_SYMBOL_WHOLE_NOTE = 0x0001D15D;
const char32_t MUSICAL_SYMBOL_HALF_NOTE = 0x0001D15E;
const char32_t MUSICAL_SYMBOL_QUARTER_NOTE = 0x0001D15F;
const char32_t MUSICAL_SYMBOL_EIGHTH_NOTE = 0x0001D160;
const char32_t MUSICAL_SYMBOL_SIXTEENTH_NOTE = 0x0001D161;
const char32_t MUSICAL_SYMBOL_THIRTY_SECOND_NOTE = 0x0001D162;
const char32_t MUSICAL_SYMBOL_SIXTY_FOURTH_NOTE = 0x0001D163;
const char32_t MUSICAL_SYMBOL_ONE_HUNDRED_TWENTY_EIGHTH_NOTE = 0x0001D164;
const char32_t MUSICAL_SYMBOL_COMBINING_STEM = 0x0001D165;
const char32_t MUSICAL_SYMBOL_COMBINING_SPRECHGESANG_STEM = 0x0001D166;
const char32_t MUSICAL_SYMBOL_COMBINING_TREMOLO_1 = 0x0001D167;
const char32_t MUSICAL_SYMBOL_COMBINING_TREMOLO_2 = 0x0001D168;
const char32_t MUSICAL_SYMBOL_COMBINING_TREMOLO_3 = 0x0001D169;
const char32_t MUSICAL_SYMBOL_FINGERED_TREMOLO_1 = 0x0001D16A;
const char32_t MUSICAL_SYMBOL_FINGERED_TREMOLO_2 = 0x0001D16B;
const char32_t MUSICAL_SYMBOL_FINGERED_TREMOLO_3 = 0x0001D16C;
const char32_t MUSICAL_SYMBOL_COMBINING_AUGMENTATION_DOT = 0x0001D16D;
const char32_t MUSICAL_SYMBOL_COMBINING_FLAG_1 = 0x0001D16E;
const char32_t MUSICAL_SYMBOL_COMBINING_FLAG_2 = 0x0001D16F;
const char32_t MUSICAL_SYMBOL_COMBINING_FLAG_3 = 0x0001D170;
const char32_t MUSICAL_SYMBOL_COMBINING_FLAG_4 = 0x0001D171;
const char32_t MUSICAL_SYMBOL_COMBINING_FLAG_5 = 0x0001D172;
const char32_t MUSICAL_SYMBOL_BEGIN_BEAM = 0x0001D173;
const char32_t MUSICAL_SYMBOL_END_BEAM = 0x0001D174;
const char32_t MUSICAL_SYMBOL_BEGIN_TIE = 0x0001D175;
const char32_t MUSICAL_SYMBOL_END_TIE = 0x0001D176;
const char32_t MUSICAL_SYMBOL_BEGIN_SLUR = 0x0001D177;
const char32_t MUSICAL_SYMBOL_END_SLUR = 0x0001D178;
const char32_t MUSICAL_SYMBOL_BEGIN_PHRASE = 0x0001D179;
const char32_t MUSICAL_SYMBOL_END_PHRASE = 0x0001D17A;
const char32_t MUSICAL_SYMBOL_COMBINING_ACCENT = 0x0001D17B;
const char32_t MUSICAL_SYMBOL_COMBINING_STACCATO = 0x0001D17C;
const char32_t MUSICAL_SYMBOL_COMBINING_TENUTO = 0x0001D17D;
const char32_t MUSICAL_SYMBOL_COMBINING_STACCATISSIMO = 0x0001D17E;
const char32_t MUSICAL_SYMBOL_COMBINING_MARCATO = 0x0001D17F;
const char32_t MUSICAL_SYMBOL_COMBINING_MARCATO_STACCATO = 0x0001D180;
const char32_t MUSICAL_SYMBOL_COMBINING_ACCENT_STACCATO = 0x0001D181;
const char32_t MUSICAL_SYMBOL_COMBINING_LOURE = 0x0001D182;
const char32_t MUSICAL_SYMBOL_ARPEGGIATO_UP = 0x0001D183;
const char32_t MUSICAL_SYMBOL_ARPEGGIATO_DOWN = 0x0001D184;
const char32_t MUSICAL_SYMBOL_COMBINING_DOIT = 0x0001D185;
const char32_t MUSICAL_SYMBOL_COMBINING_RIP = 0x0001D186;
const char32_t MUSICAL_SYMBOL_COMBINING_FLIP = 0x0001D187;
const char32_t MUSICAL_SYMBOL_COMBINING_SMEAR = 0x0001D188;
const char32_t MUSICAL_SYMBOL_COMBINING_BEND = 0x0001D189;
const char32_t MUSICAL_SYMBOL_COMBINING_DOUBLE_TONGUE = 0x0001D18A;
const char32_t MUSICAL_SYMBOL_COMBINING_TRIPLE_TONGUE = 0x0001D18B;
const char32_t MUSICAL_SYMBOL_RINFORZANDO = 0x0001D18C;
const char32_t MUSICAL_SYMBOL_SUBITO = 0x0001D18D;
const char32_t MUSICAL_SYMBOL_Z = 0x0001D18E;
const char32_t MUSICAL_SYMBOL_PIANO = 0x0001D18F;
const char32_t MUSICAL_SYMBOL_MEZZO = 0x0001D190;
const char32_t MUSICAL_SYMBOL_FORTE = 0x0001D191;
const char32_t MUSICAL_SYMBOL_CRESCENDO = 0x0001D192;
const char32_t MUSICAL_SYMBOL_DECRESCENDO = 0x0001D193;
const char32_t MUSICAL_SYMBOL_GRACE_NOTE_SLASH = 0x0001D194;
const char32_t MUSICAL_SYMBOL_GRACE_NOTE_NO_SLASH = 0x0001D195;
const char32_t MUSICAL_SYMBOL_TR = 0x0001D196;
const char32_t MUSICAL_SYMBOL_TURN = 0x0001D197;
const char32_t MUSICAL_SYMBOL_INVERTED_TURN = 0x0001D198;
const char32_t MUSICAL_SYMBOL_TURN_SLASH = 0x0001D199;
const char32_t MUSICAL_SYMBOL_TURN_UP = 0x0001D19A;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_1 = 0x0001D19B;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_2 = 0x0001D19C;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_3 = 0x0001D19D;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_4 = 0x0001D19E;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_5 = 0x0001D19F;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_6 = 0x0001D1A0;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_7 = 0x0001D1A1;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_8 = 0x0001D1A2;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_9 = 0x0001D1A3;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_10 = 0x0001D1A4;
const char32_t MUSICAL_SYMBOL_ORNAMENT_STROKE_11 = 0x0001D1A5;
const char32_t MUSICAL_SYMBOL_HAUPTSTIMME = 0x0001D1A6;
const char32_t MUSICAL_SYMBOL_NEBENSTIMME = 0x0001D1A7;
const char32_t MUSICAL_SYMBOL_END_OF_STIMME = 0x0001D1A8;
const char32_t MUSICAL_SYMBOL_DEGREE_SLASH = 0x0001D1A9;
const char32_t MUSICAL_SYMBOL_COMBINING_DOWN_BOW = 0x0001D1AA;
const char32_t MUSICAL_SYMBOL_COMBINING_UP_BOW = 0x0001D1AB;
const char32_t MUSICAL_SYMBOL_COMBINING_HARMONIC = 0x0001D1AC;
const char32_t MUSICAL_SYMBOL_COMBINING_SNAP_PIZZICATO = 0x0001D1AD;
const char32_t MUSICAL_SYMBOL_PEDAL_MARK = 0x0001D1AE;
const char32_t MUSICAL_SYMBOL_PEDAL_UP_MARK = 0x0001D1AF;
const char32_t MUSICAL_SYMBOL_HALF_PEDAL_MARK = 0x0001D1B0;
const char32_t MUSICAL_SYMBOL_GLISSANDO_UP = 0x0001D1B1;
const char32_t MUSICAL_SYMBOL_GLISSANDO_DOWN = 0x0001D1B2;
const char32_t MUSICAL_SYMBOL_WITH_FINGERNAILS = 0x0001D1B3;
const char32_t MUSICAL_SYMBOL_DAMP = 0x0001D1B4;
const char32_t MUSICAL_SYMBOL_DAMP_ALL = 0x0001D1B5;
const char32_t MUSICAL_SYMBOL_MAXIMA = 0x0001D1B6;
const char32_t MUSICAL_SYMBOL_LONGA = 0x0001D1B7;
const char32_t MUSICAL_SYMBOL_BREVIS = 0x0001D1B8;
const char32_t MUSICAL_SYMBOL_SEMIBREVIS_WHITE = 0x0001D1B9;
const char32_t MUSICAL_SYMBOL_SEMIBREVIS_BLACK = 0x0001D1BA;
const char32_t MUSICAL_SYMBOL_MINIMA = 0x0001D1BB;
const char32_t MUSICAL_SYMBOL_MINIMA_BLACK = 0x0001D1BC;
const char32_t MUSICAL_SYMBOL_SEMIMINIMA_WHITE = 0x0001D1BD;
const char32_t MUSICAL_SYMBOL_SEMIMINIMA_BLACK = 0x0001D1BE;
const char32_t MUSICAL_SYMBOL_FUSA_WHITE = 0x0001D1BF;
const char32_t MUSICAL_SYMBOL_FUSA_BLACK = 0x0001D1C0;
const char32_t MUSICAL_SYMBOL_LONGA_PERFECTA_REST = 0x0001D1C1;
const char32_t MUSICAL_SYMBOL_LONGA_IMPERFECTA_REST = 0x0001D1C2;
const char32_t MUSICAL_SYMBOL_BREVIS_REST = 0x0001D1C3;
const char32_t MUSICAL_SYMBOL_SEMIBREVIS_REST = 0x0001D1C4;
const char32_t MUSICAL_SYMBOL_MINIMA_REST = 0x0001D1C5;
const char32_t MUSICAL_SYMBOL_SEMIMINIMA_REST = 0x0001D1C6;
const char32_t MUSICAL_SYMBOL_TEMPUS_PERFECTUM_CUM_PROLATIONE_PERFECTA = 0x0001D1C7;
const char32_t MUSICAL_SYMBOL_TEMPUS_PERFECTUM_CUM_PROLATIONE_IMPERFECTA = 0x0001D1C8;
const char32_t MUSICAL_SYMBOL_TEMPUS_PERFECTUM_CUM_PROLATIONE_PERFECTA_DIMINUTION_1 = 0x0001D1C9;
const char32_t MUSICAL_SYMBOL_TEMPUS_IMPERFECTUM_CUM_PROLATIONE_PERFECTA = 0x0001D1CA;
const char32_t MUSICAL_SYMBOL_TEMPUS_IMPERFECTUM_CUM_PROLATIONE_IMPERFECTA = 0x0001D1CB;
const char32_t MUSICAL_SYMBOL_TEMPUS_IMPERFECTUM_CUM_PROLATIONE_IMPERFECTA_DIMINUTION_1 = 0x0001D1CC;
const char32_t MUSICAL_SYMBOL_TEMPUS_IMPERFECTUM_CUM_PROLATIONE_IMPERFECTA_DIMINUTION_2 = 0x0001D1CD;
const char32_t MUSICAL_SYMBOL_TEMPUS_IMPERFECTUM_CUM_PROLATIONE_IMPERFECTA_DIMINUTION_3 = 0x0001D1CE;
const char32_t MUSICAL_SYMBOL_CROIX = 0x0001D1CF;
const char32_t MUSICAL_SYMBOL_GREGORIAN_C_CLEF = 0x0001D1D0;
const char32_t MUSICAL_SYMBOL_GREGORIAN_F_CLEF = 0x0001D1D1;
const char32_t MUSICAL_SYMBOL_SQUARE_B = 0x0001D1D2;
const char32_t MUSICAL_SYMBOL_VIRGA = 0x0001D1D3;
const char32_t MUSICAL_SYMBOL_PODATUS = 0x0001D1D4;
const char32_t MUSICAL_SYMBOL_CLIVIS = 0x0001D1D5;
const char32_t MUSICAL_SYMBOL_SCANDICUS = 0x0001D1D6;
const char32_t MUSICAL_SYMBOL_CLIMACUS = 0x0001D1D7;
const char32_t MUSICAL_SYMBOL_TORCULUS = 0x0001D1D8;
const char32_t MUSICAL_SYMBOL_PORRECTUS = 0x0001D1D9;
const char32_t MUSICAL_SYMBOL_PORRECTUS_FLEXUS = 0x0001D1DA;
const char32_t MUSICAL_SYMBOL_SCANDICUS_FLEXUS = 0x0001D1DB;
const char32_t MUSICAL_SYMBOL_TORCULUS_RESUPINUS = 0x0001D1DC;
const char32_t MUSICAL_SYMBOL_PES_SUBPUNCTIS = 0x0001D1DD;
const char32_t MUSICAL_SYMBOL_KIEVAN_C_CLEF = 0x0001D1DE;
const char32_t MUSICAL_SYMBOL_KIEVAN_END_OF_PIECE = 0x0001D1DF;
const char32_t MUSICAL_SYMBOL_KIEVAN_FINAL_NOTE = 0x0001D1E0;
const char32_t MUSICAL_SYMBOL_KIEVAN_RECITATIVE_MARK = 0x0001D1E1;
const char32_t MUSICAL_SYMBOL_KIEVAN_WHOLE_NOTE = 0x0001D1E2;
const char32_t MUSICAL_SYMBOL_KIEVAN_HALF_NOTE = 0x0001D1E3;
const char32_t MUSICAL_SYMBOL_KIEVAN_QUARTER_NOTE_STEM_DOWN = 0x0001D1E4;
const char32_t MUSICAL_SYMBOL_KIEVAN_QUARTER_NOTE_STEM_UP = 0x0001D1E5;
const char32_t MUSICAL_SYMBOL_KIEVAN_EIGHTH_NOTE_STEM_DOWN = 0x0001D1E6;
const char32_t MUSICAL_SYMBOL_KIEVAN_EIGHTH_NOTE_STEM_UP = 0x0001D1E7;
const char32_t MUSICAL_SYMBOL_KIEVAN_FLAT_SIGN = 0x0001D1E8;
const char32_t MUSICAL_SYMBOL_SORI = 0x0001D1E9;
const char32_t MUSICAL_SYMBOL_KORON = 0x0001D1EA;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_1 = 0x0001D200;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_2 = 0x0001D201;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_3 = 0x0001D202;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_4 = 0x0001D203;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_5 = 0x0001D204;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_6 = 0x0001D205;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_7 = 0x0001D206;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_8 = 0x0001D207;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_9 = 0x0001D208;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_10 = 0x0001D209;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_11 = 0x0001D20A;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_12 = 0x0001D20B;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_13 = 0x0001D20C;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_14 = 0x0001D20D;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_15 = 0x0001D20E;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_16 = 0x0001D20F;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_17 = 0x0001D210;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_18 = 0x0001D211;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_19 = 0x0001D212;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_20 = 0x0001D213;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_21 = 0x0001D214;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_22 = 0x0001D215;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_23 = 0x0001D216;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_24 = 0x0001D217;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_50 = 0x0001D218;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_51 = 0x0001D219;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_52 = 0x0001D21A;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_53 = 0x0001D21B;
const char32_t GREEK_VOCAL_NOTATION_SYMBOL_54 = 0x0001D21C;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_1 = 0x0001D21D;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_2 = 0x0001D21E;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_4 = 0x0001D21F;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_5 = 0x0001D220;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_7 = 0x0001D221;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_8 = 0x0001D222;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_11 = 0x0001D223;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_12 = 0x0001D224;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_13 = 0x0001D225;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_14 = 0x0001D226;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_17 = 0x0001D227;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_18 = 0x0001D228;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_19 = 0x0001D229;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_23 = 0x0001D22A;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_24 = 0x0001D22B;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_25 = 0x0001D22C;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_26 = 0x0001D22D;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_27 = 0x0001D22E;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_29 = 0x0001D22F;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_30 = 0x0001D230;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_32 = 0x0001D231;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_36 = 0x0001D232;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_37 = 0x0001D233;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_38 = 0x0001D234;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_39 = 0x0001D235;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_40 = 0x0001D236;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_42 = 0x0001D237;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_43 = 0x0001D238;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_45 = 0x0001D239;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_47 = 0x0001D23A;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_48 = 0x0001D23B;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_49 = 0x0001D23C;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_50 = 0x0001D23D;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_51 = 0x0001D23E;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_52 = 0x0001D23F;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_53 = 0x0001D240;
const char32_t GREEK_INSTRUMENTAL_NOTATION_SYMBOL_54 = 0x0001D241;
const char32_t COMBINING_GREEK_MUSICAL_TRISEME = 0x0001D242;
const char32_t COMBINING_GREEK_MUSICAL_TETRASEME = 0x0001D243;
const char32_t COMBINING_GREEK_MUSICAL_PENTASEME = 0x0001D244;
const char32_t GREEK_MUSICAL_LEIMMA = 0x0001D245;
const char32_t KAKTOVIK_NUMERAL_ZERO = 0x0001D2C0;
const char32_t KAKTOVIK_NUMERAL_ONE = 0x0001D2C1;
const char32_t KAKTOVIK_NUMERAL_TWO = 0x0001D2C2;
const char32_t KAKTOVIK_NUMERAL_THREE = 0x0001D2C3;
const char32_t KAKTOVIK_NUMERAL_FOUR = 0x0001D2C4;
const char32_t KAKTOVIK_NUMERAL_FIVE = 0x0001D2C5;
const char32_t KAKTOVIK_NUMERAL_SIX = 0x0001D2C6;
const char32_t KAKTOVIK_NUMERAL_SEVEN = 0x0001D2C7;
const char32_t KAKTOVIK_NUMERAL_EIGHT = 0x0001D2C8;
const char32_t KAKTOVIK_NUMERAL_NINE = 0x0001D2C9;
const char32_t KAKTOVIK_NUMERAL_TEN = 0x0001D2CA;
const char32_t KAKTOVIK_NUMERAL_ELEVEN = 0x0001D2CB;
const char32_t KAKTOVIK_NUMERAL_TWELVE = 0x0001D2CC;
const char32_t KAKTOVIK_NUMERAL_THIRTEEN = 0x0001D2CD;
const char32_t KAKTOVIK_NUMERAL_FOURTEEN = 0x0001D2CE;
const char32_t KAKTOVIK_NUMERAL_FIFTEEN = 0x0001D2CF;
const char32_t KAKTOVIK_NUMERAL_SIXTEEN = 0x0001D2D0;
const char32_t KAKTOVIK_NUMERAL_SEVENTEEN = 0x0001D2D1;
const char32_t KAKTOVIK_NUMERAL_EIGHTEEN = 0x0001D2D2;
const char32_t KAKTOVIK_NUMERAL_NINETEEN = 0x0001D2D3;
const char32_t MAYAN_NUMERAL_ZERO = 0x0001D2E0;
const char32_t MAYAN_NUMERAL_ONE = 0x0001D2E1;
const char32_t MAYAN_NUMERAL_TWO = 0x0001D2E2;
const char32_t MAYAN_NUMERAL_THREE = 0x0001D2E3;
const char32_t MAYAN_NUMERAL_FOUR = 0x0001D2E4;
const char32_t MAYAN_NUMERAL_FIVE = 0x0001D2E5;
const char32_t MAYAN_NUMERAL_SIX = 0x0001D2E6;
const char32_t MAYAN_NUMERAL_SEVEN = 0x0001D2E7;
const char32_t MAYAN_NUMERAL_EIGHT = 0x0001D2E8;
const char32_t MAYAN_NUMERAL_NINE = 0x0001D2E9;
const char32_t MAYAN_NUMERAL_TEN = 0x0001D2EA;
const char32_t MAYAN_NUMERAL_ELEVEN = 0x0001D2EB;
const char32_t MAYAN_NUMERAL_TWELVE = 0x0001D2EC;
const char32_t MAYAN_NUMERAL_THIRTEEN = 0x0001D2ED;
const char32_t MAYAN_NUMERAL_FOURTEEN = 0x0001D2EE;
const char32_t MAYAN_NUMERAL_FIFTEEN = 0x0001D2EF;
const char32_t MAYAN_NUMERAL_SIXTEEN = 0x0001D2F0;
const char32_t MAYAN_NUMERAL_SEVENTEEN = 0x0001D2F1;
const char32_t MAYAN_NUMERAL_EIGHTEEN = 0x0001D2F2;
const char32_t MAYAN_NUMERAL_NINETEEN = 0x0001D2F3;
const char32_t MONOGRAM_FOR_EARTH = 0x0001D300;
const char32_t DIGRAM_FOR_HEAVENLY_EARTH = 0x0001D301;
const char32_t DIGRAM_FOR_HUMAN_EARTH = 0x0001D302;
const char32_t DIGRAM_FOR_EARTHLY_HEAVEN = 0x0001D303;
const char32_t DIGRAM_FOR_EARTHLY_HUMAN = 0x0001D304;
const char32_t DIGRAM_FOR_EARTH = 0x0001D305;
const char32_t TETRAGRAM_FOR_CENTRE = 0x0001D306;
const char32_t TETRAGRAM_FOR_FULL_CIRCLE = 0x0001D307;
const char32_t TETRAGRAM_FOR_MIRED = 0x0001D308;
const char32_t TETRAGRAM_FOR_BARRIER = 0x0001D309;
const char32_t TETRAGRAM_FOR_KEEPING_SMALL = 0x0001D30A;
const char32_t TETRAGRAM_FOR_CONTRARIETY = 0x0001D30B;
const char32_t TETRAGRAM_FOR_ASCENT = 0x0001D30C;
const char32_t TETRAGRAM_FOR_OPPOSITION = 0x0001D30D;
const char32_t TETRAGRAM_FOR_BRANCHING_OUT = 0x0001D30E;
const char32_t TETRAGRAM_FOR_DEFECTIVENESS_OR_DISTORTION = 0x0001D30F;
const char32_t TETRAGRAM_FOR_DIVERGENCE = 0x0001D310;
const char32_t TETRAGRAM_FOR_YOUTHFULNESS = 0x0001D311;
const char32_t TETRAGRAM_FOR_INCREASE = 0x0001D312;
const char32_t TETRAGRAM_FOR_PENETRATION = 0x0001D313;
const char32_t TETRAGRAM_FOR_REACH = 0x0001D314;
const char32_t TETRAGRAM_FOR_CONTACT = 0x0001D315;
const char32_t TETRAGRAM_FOR_HOLDING_BACK = 0x0001D316;
const char32_t TETRAGRAM_FOR_WAITING = 0x0001D317;
const char32_t TETRAGRAM_FOR_FOLLOWING = 0x0001D318;
const char32_t TETRAGRAM_FOR_ADVANCE = 0x0001D319;
const char32_t TETRAGRAM_FOR_RELEASE = 0x0001D31A;
const char32_t TETRAGRAM_FOR_RESISTANCE = 0x0001D31B;
const char32_t TETRAGRAM_FOR_EASE = 0x0001D31C;
const char32_t TETRAGRAM_FOR_JOY = 0x0001D31D;
const char32_t TETRAGRAM_FOR_CONTENTION = 0x0001D31E;
const char32_t TETRAGRAM_FOR_ENDEAVOUR = 0x0001D31F;
const char32_t TETRAGRAM_FOR_DUTIES = 0x0001D320;
const char32_t TETRAGRAM_FOR_CHANGE = 0x0001D321;
const char32_t TETRAGRAM_FOR_DECISIVENESS = 0x0001D322;
const char32_t TETRAGRAM_FOR_BOLD_RESOLUTION = 0x0001D323;
const char32_t TETRAGRAM_FOR_PACKING = 0x0001D324;
const char32_t TETRAGRAM_FOR_LEGION = 0x0001D325;
const char32_t TETRAGRAM_FOR_CLOSENESS = 0x0001D326;
const char32_t TETRAGRAM_FOR_KINSHIP = 0x0001D327;
const char32_t TETRAGRAM_FOR_GATHERING = 0x0001D328;
const char32_t TETRAGRAM_FOR_STRENGTH = 0x0001D329;
const char32_t TETRAGRAM_FOR_PURITY = 0x0001D32A;
const char32_t TETRAGRAM_FOR_FULLNESS = 0x0001D32B;
const char32_t TETRAGRAM_FOR_RESIDENCE = 0x0001D32C;
const char32_t TETRAGRAM_FOR_LAW_OR_MODEL = 0x0001D32D;
const char32_t TETRAGRAM_FOR_RESPONSE = 0x0001D32E;
const char32_t TETRAGRAM_FOR_GOING_TO_MEET = 0x0001D32F;
const char32_t TETRAGRAM_FOR_ENCOUNTERS = 0x0001D330;
const char32_t TETRAGRAM_FOR_STOVE = 0x0001D331;
const char32_t TETRAGRAM_FOR_GREATNESS = 0x0001D332;
const char32_t TETRAGRAM_FOR_ENLARGEMENT = 0x0001D333;
const char32_t TETRAGRAM_FOR_PATTERN = 0x0001D334;
const char32_t TETRAGRAM_FOR_RITUAL = 0x0001D335;
const char32_t TETRAGRAM_FOR_FLIGHT = 0x0001D336;
const char32_t TETRAGRAM_FOR_VASTNESS_OR_WASTING = 0x0001D337;
const char32_t TETRAGRAM_FOR_CONSTANCY = 0x0001D338;
const char32_t TETRAGRAM_FOR_MEASURE = 0x0001D339;
const char32_t TETRAGRAM_FOR_ETERNITY = 0x0001D33A;
const char32_t TETRAGRAM_FOR_UNITY = 0x0001D33B;
const char32_t TETRAGRAM_FOR_DIMINISHMENT = 0x0001D33C;
const char32_t TETRAGRAM_FOR_CLOSED_MOUTH = 0x0001D33D;
const char32_t TETRAGRAM_FOR_GUARDEDNESS = 0x0001D33E;
const char32_t TETRAGRAM_FOR_GATHERING_IN = 0x0001D33F;
const char32_t TETRAGRAM_FOR_MASSING = 0x0001D340;
const char32_t TETRAGRAM_FOR_ACCUMULATION = 0x0001D341;
const char32_t TETRAGRAM_FOR_EMBELLISHMENT = 0x0001D342;
const char32_t TETRAGRAM_FOR_DOUBT = 0x0001D343;
const char32_t TETRAGRAM_FOR_WATCH = 0x0001D344;
const char32_t TETRAGRAM_FOR_SINKING = 0x0001D345;
const char32_t TETRAGRAM_FOR_INNER = 0x0001D346;
const char32_t TETRAGRAM_FOR_DEPARTURE = 0x0001D347;
const char32_t TETRAGRAM_FOR_DARKENING = 0x0001D348;
const char32_t TETRAGRAM_FOR_DIMMING = 0x0001D349;
const char32_t TETRAGRAM_FOR_EXHAUSTION = 0x0001D34A;
const char32_t TETRAGRAM_FOR_SEVERANCE = 0x0001D34B;
const char32_t TETRAGRAM_FOR_STOPPAGE = 0x0001D34C;
const char32_t TETRAGRAM_FOR_HARDNESS = 0x0001D34D;
const char32_t TETRAGRAM_FOR_COMPLETION = 0x0001D34E;
const char32_t TETRAGRAM_FOR_CLOSURE = 0x0001D34F;
const char32_t TETRAGRAM_FOR_FAILURE = 0x0001D350;
const char32_t TETRAGRAM_FOR_AGGRAVATION = 0x0001D351;
const char32_t TETRAGRAM_FOR_COMPLIANCE = 0x0001D352;
const char32_t TETRAGRAM_FOR_ON_THE_VERGE = 0x0001D353;
const char32_t TETRAGRAM_FOR_DIFFICULTIES = 0x0001D354;
const char32_t TETRAGRAM_FOR_LABOURING = 0x0001D355;
const char32_t TETRAGRAM_FOR_FOSTERING = 0x0001D356;
const char32_t COUNTING_ROD_UNIT_DIGIT_ONE = 0x0001D360;
const char32_t COUNTING_ROD_UNIT_DIGIT_TWO = 0x0001D361;
const char32_t COUNTING_ROD_UNIT_DIGIT_THREE = 0x0001D362;
const char32_t COUNTING_ROD_UNIT_DIGIT_FOUR = 0x0001D363;
const char32_t COUNTING_ROD_UNIT_DIGIT_FIVE = 0x0001D364;
const char32_t COUNTING_ROD_UNIT_DIGIT_SIX = 0x0001D365;
const char32_t COUNTING_ROD_UNIT_DIGIT_SEVEN = 0x0001D366;
const char32_t COUNTING_ROD_UNIT_DIGIT_EIGHT = 0x0001D367;
const char32_t COUNTING_ROD_UNIT_DIGIT_NINE = 0x0001D368;
const char32_t COUNTING_ROD_TENS_DIGIT_ONE = 0x0001D369;
const char32_t COUNTING_ROD_TENS_DIGIT_TWO = 0x0001D36A;
const char32_t COUNTING_ROD_TENS_DIGIT_THREE = 0x0001D36B;
const char32_t COUNTING_ROD_TENS_DIGIT_FOUR = 0x0001D36C;
const char32_t COUNTING_ROD_TENS_DIGIT_FIVE = 0x0001D36D;
const char32_t COUNTING_ROD_TENS_DIGIT_SIX = 0x0001D36E;
const char32_t COUNTING_ROD_TENS_DIGIT_SEVEN = 0x0001D36F;
const char32_t COUNTING_ROD_TENS_DIGIT_EIGHT = 0x0001D370;
const char32_t COUNTING_ROD_TENS_DIGIT_NINE = 0x0001D371;
const char32_t IDEOGRAPHIC_TALLY_MARK_ONE = 0x0001D372;
const char32_t IDEOGRAPHIC_TALLY_MARK_TWO = 0x0001D373;
const char32_t IDEOGRAPHIC_TALLY_MARK_THREE = 0x0001D374;
const char32_t IDEOGRAPHIC_TALLY_MARK_FOUR = 0x0001D375;
const char32_t IDEOGRAPHIC_TALLY_MARK_FIVE = 0x0001D376;
const char32_t TALLY_MARK_ONE = 0x0001D377;
const char32_t TALLY_MARK_FIVE = 0x0001D378;
const char32_t MATHEMATICAL_BOLD_CAPITAL_A = 0x0001D400;
const char32_t MATHEMATICAL_BOLD_CAPITAL_B = 0x0001D401;
const char32_t MATHEMATICAL_BOLD_CAPITAL_C = 0x0001D402;
const char32_t MATHEMATICAL_BOLD_CAPITAL_D = 0x0001D403;
const char32_t MATHEMATICAL_BOLD_CAPITAL_E = 0x0001D404;
const char32_t MATHEMATICAL_BOLD_CAPITAL_F = 0x0001D405;
const char32_t MATHEMATICAL_BOLD_CAPITAL_G = 0x0001D406;
const char32_t MATHEMATICAL_BOLD_CAPITAL_H = 0x0001D407;
const char32_t MATHEMATICAL_BOLD_CAPITAL_I = 0x0001D408;
const char32_t MATHEMATICAL_BOLD_CAPITAL_J = 0x0001D409;
const char32_t MATHEMATICAL_BOLD_CAPITAL_K = 0x0001D40A;
const char32_t MATHEMATICAL_BOLD_CAPITAL_L = 0x0001D40B;
const char32_t MATHEMATICAL_BOLD_CAPITAL_M = 0x0001D40C;
const char32_t MATHEMATICAL_BOLD_CAPITAL_N = 0x0001D40D;
const char32_t MATHEMATICAL_BOLD_CAPITAL_O = 0x0001D40E;
const char32_t MATHEMATICAL_BOLD_CAPITAL_P = 0x0001D40F;
const char32_t MATHEMATICAL_BOLD_CAPITAL_Q = 0x0001D410;
const char32_t MATHEMATICAL_BOLD_CAPITAL_R = 0x0001D411;
const char32_t MATHEMATICAL_BOLD_CAPITAL_S = 0x0001D412;
const char32_t MATHEMATICAL_BOLD_CAPITAL_T = 0x0001D413;
const char32_t MATHEMATICAL_BOLD_CAPITAL_U = 0x0001D414;
const char32_t MATHEMATICAL_BOLD_CAPITAL_V = 0x0001D415;
const char32_t MATHEMATICAL_BOLD_CAPITAL_W = 0x0001D416;
const char32_t MATHEMATICAL_BOLD_CAPITAL_X = 0x0001D417;
const char32_t MATHEMATICAL_BOLD_CAPITAL_Y = 0x0001D418;
const char32_t MATHEMATICAL_BOLD_CAPITAL_Z = 0x0001D419;
const char32_t MATHEMATICAL_BOLD_SMALL_A = 0x0001D41A;
const char32_t MATHEMATICAL_BOLD_SMALL_B = 0x0001D41B;
const char32_t MATHEMATICAL_BOLD_SMALL_C = 0x0001D41C;
const char32_t MATHEMATICAL_BOLD_SMALL_D = 0x0001D41D;
const char32_t MATHEMATICAL_BOLD_SMALL_E = 0x0001D41E;
const char32_t MATHEMATICAL_BOLD_SMALL_F = 0x0001D41F;
const char32_t MATHEMATICAL_BOLD_SMALL_G = 0x0001D420;
const char32_t MATHEMATICAL_BOLD_SMALL_H = 0x0001D421;
const char32_t MATHEMATICAL_BOLD_SMALL_I = 0x0001D422;
const char32_t MATHEMATICAL_BOLD_SMALL_J = 0x0001D423;
const char32_t MATHEMATICAL_BOLD_SMALL_K = 0x0001D424;
const char32_t MATHEMATICAL_BOLD_SMALL_L = 0x0001D425;
const char32_t MATHEMATICAL_BOLD_SMALL_M = 0x0001D426;
const char32_t MATHEMATICAL_BOLD_SMALL_N = 0x0001D427;
const char32_t MATHEMATICAL_BOLD_SMALL_O = 0x0001D428;
const char32_t MATHEMATICAL_BOLD_SMALL_P = 0x0001D429;
const char32_t MATHEMATICAL_BOLD_SMALL_Q = 0x0001D42A;
const char32_t MATHEMATICAL_BOLD_SMALL_R = 0x0001D42B;
const char32_t MATHEMATICAL_BOLD_SMALL_S = 0x0001D42C;
const char32_t MATHEMATICAL_BOLD_SMALL_T = 0x0001D42D;
const char32_t MATHEMATICAL_BOLD_SMALL_U = 0x0001D42E;
const char32_t MATHEMATICAL_BOLD_SMALL_V = 0x0001D42F;
const char32_t MATHEMATICAL_BOLD_SMALL_W = 0x0001D430;
const char32_t MATHEMATICAL_BOLD_SMALL_X = 0x0001D431;
const char32_t MATHEMATICAL_BOLD_SMALL_Y = 0x0001D432;
const char32_t MATHEMATICAL_BOLD_SMALL_Z = 0x0001D433;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_A = 0x0001D434;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_B = 0x0001D435;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_C = 0x0001D436;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_D = 0x0001D437;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_E = 0x0001D438;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_F = 0x0001D439;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_G = 0x0001D43A;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_H = 0x0001D43B;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_I = 0x0001D43C;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_J = 0x0001D43D;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_K = 0x0001D43E;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_L = 0x0001D43F;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_M = 0x0001D440;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_N = 0x0001D441;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_O = 0x0001D442;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_P = 0x0001D443;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_Q = 0x0001D444;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_R = 0x0001D445;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_S = 0x0001D446;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_T = 0x0001D447;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_U = 0x0001D448;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_V = 0x0001D449;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_W = 0x0001D44A;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_X = 0x0001D44B;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_Y = 0x0001D44C;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_Z = 0x0001D44D;
const char32_t MATHEMATICAL_ITALIC_SMALL_A = 0x0001D44E;
const char32_t MATHEMATICAL_ITALIC_SMALL_B = 0x0001D44F;
const char32_t MATHEMATICAL_ITALIC_SMALL_C = 0x0001D450;
const char32_t MATHEMATICAL_ITALIC_SMALL_D = 0x0001D451;
const char32_t MATHEMATICAL_ITALIC_SMALL_E = 0x0001D452;
const char32_t MATHEMATICAL_ITALIC_SMALL_F = 0x0001D453;
const char32_t MATHEMATICAL_ITALIC_SMALL_G = 0x0001D454;
const char32_t MATHEMATICAL_ITALIC_SMALL_I = 0x0001D456;
const char32_t MATHEMATICAL_ITALIC_SMALL_J = 0x0001D457;
const char32_t MATHEMATICAL_ITALIC_SMALL_K = 0x0001D458;
const char32_t MATHEMATICAL_ITALIC_SMALL_L = 0x0001D459;
const char32_t MATHEMATICAL_ITALIC_SMALL_M = 0x0001D45A;
const char32_t MATHEMATICAL_ITALIC_SMALL_N = 0x0001D45B;
const char32_t MATHEMATICAL_ITALIC_SMALL_O = 0x0001D45C;
const char32_t MATHEMATICAL_ITALIC_SMALL_P = 0x0001D45D;
const char32_t MATHEMATICAL_ITALIC_SMALL_Q = 0x0001D45E;
const char32_t MATHEMATICAL_ITALIC_SMALL_R = 0x0001D45F;
const char32_t MATHEMATICAL_ITALIC_SMALL_S = 0x0001D460;
const char32_t MATHEMATICAL_ITALIC_SMALL_T = 0x0001D461;
const char32_t MATHEMATICAL_ITALIC_SMALL_U = 0x0001D462;
const char32_t MATHEMATICAL_ITALIC_SMALL_V = 0x0001D463;
const char32_t MATHEMATICAL_ITALIC_SMALL_W = 0x0001D464;
const char32_t MATHEMATICAL_ITALIC_SMALL_X = 0x0001D465;
const char32_t MATHEMATICAL_ITALIC_SMALL_Y = 0x0001D466;
const char32_t MATHEMATICAL_ITALIC_SMALL_Z = 0x0001D467;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_A = 0x0001D468;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_B = 0x0001D469;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_C = 0x0001D46A;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_D = 0x0001D46B;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_E = 0x0001D46C;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_F = 0x0001D46D;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_G = 0x0001D46E;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_H = 0x0001D46F;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_I = 0x0001D470;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_J = 0x0001D471;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_K = 0x0001D472;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_L = 0x0001D473;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_M = 0x0001D474;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_N = 0x0001D475;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_O = 0x0001D476;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_P = 0x0001D477;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_Q = 0x0001D478;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_R = 0x0001D479;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_S = 0x0001D47A;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_T = 0x0001D47B;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_U = 0x0001D47C;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_V = 0x0001D47D;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_W = 0x0001D47E;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_X = 0x0001D47F;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_Y = 0x0001D480;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_Z = 0x0001D481;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_A = 0x0001D482;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_B = 0x0001D483;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_C = 0x0001D484;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_D = 0x0001D485;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_E = 0x0001D486;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_F = 0x0001D487;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_G = 0x0001D488;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_H = 0x0001D489;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_I = 0x0001D48A;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_J = 0x0001D48B;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_K = 0x0001D48C;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_L = 0x0001D48D;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_M = 0x0001D48E;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_N = 0x0001D48F;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_O = 0x0001D490;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_P = 0x0001D491;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_Q = 0x0001D492;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_R = 0x0001D493;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_S = 0x0001D494;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_T = 0x0001D495;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_U = 0x0001D496;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_V = 0x0001D497;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_W = 0x0001D498;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_X = 0x0001D499;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_Y = 0x0001D49A;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_Z = 0x0001D49B;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_A = 0x0001D49C;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_C = 0x0001D49E;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_D = 0x0001D49F;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_G = 0x0001D4A2;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_J = 0x0001D4A5;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_K = 0x0001D4A6;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_N = 0x0001D4A9;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_O = 0x0001D4AA;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_P = 0x0001D4AB;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_Q = 0x0001D4AC;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_S = 0x0001D4AE;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_T = 0x0001D4AF;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_U = 0x0001D4B0;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_V = 0x0001D4B1;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_W = 0x0001D4B2;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_X = 0x0001D4B3;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_Y = 0x0001D4B4;
const char32_t MATHEMATICAL_SCRIPT_CAPITAL_Z = 0x0001D4B5;
const char32_t MATHEMATICAL_SCRIPT_SMALL_A = 0x0001D4B6;
const char32_t MATHEMATICAL_SCRIPT_SMALL_B = 0x0001D4B7;
const char32_t MATHEMATICAL_SCRIPT_SMALL_C = 0x0001D4B8;
const char32_t MATHEMATICAL_SCRIPT_SMALL_D = 0x0001D4B9;
const char32_t MATHEMATICAL_SCRIPT_SMALL_F = 0x0001D4BB;
const char32_t MATHEMATICAL_SCRIPT_SMALL_H = 0x0001D4BD;
const char32_t MATHEMATICAL_SCRIPT_SMALL_I = 0x0001D4BE;
const char32_t MATHEMATICAL_SCRIPT_SMALL_J = 0x0001D4BF;
const char32_t MATHEMATICAL_SCRIPT_SMALL_K = 0x0001D4C0;
const char32_t MATHEMATICAL_SCRIPT_SMALL_L = 0x0001D4C1;
const char32_t MATHEMATICAL_SCRIPT_SMALL_M = 0x0001D4C2;
const char32_t MATHEMATICAL_SCRIPT_SMALL_N = 0x0001D4C3;
const char32_t MATHEMATICAL_SCRIPT_SMALL_P = 0x0001D4C5;
const char32_t MATHEMATICAL_SCRIPT_SMALL_Q = 0x0001D4C6;
const char32_t MATHEMATICAL_SCRIPT_SMALL_R = 0x0001D4C7;
const char32_t MATHEMATICAL_SCRIPT_SMALL_S = 0x0001D4C8;
const char32_t MATHEMATICAL_SCRIPT_SMALL_T = 0x0001D4C9;
const char32_t MATHEMATICAL_SCRIPT_SMALL_U = 0x0001D4CA;
const char32_t MATHEMATICAL_SCRIPT_SMALL_V = 0x0001D4CB;
const char32_t MATHEMATICAL_SCRIPT_SMALL_W = 0x0001D4CC;
const char32_t MATHEMATICAL_SCRIPT_SMALL_X = 0x0001D4CD;
const char32_t MATHEMATICAL_SCRIPT_SMALL_Y = 0x0001D4CE;
const char32_t MATHEMATICAL_SCRIPT_SMALL_Z = 0x0001D4CF;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_A = 0x0001D4D0;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_B = 0x0001D4D1;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_C = 0x0001D4D2;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_D = 0x0001D4D3;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_E = 0x0001D4D4;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_F = 0x0001D4D5;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_G = 0x0001D4D6;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_H = 0x0001D4D7;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_I = 0x0001D4D8;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_J = 0x0001D4D9;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_K = 0x0001D4DA;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_L = 0x0001D4DB;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_M = 0x0001D4DC;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_N = 0x0001D4DD;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_O = 0x0001D4DE;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_P = 0x0001D4DF;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_Q = 0x0001D4E0;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_R = 0x0001D4E1;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_S = 0x0001D4E2;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_T = 0x0001D4E3;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_U = 0x0001D4E4;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_V = 0x0001D4E5;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_W = 0x0001D4E6;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_X = 0x0001D4E7;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_Y = 0x0001D4E8;
const char32_t MATHEMATICAL_BOLD_SCRIPT_CAPITAL_Z = 0x0001D4E9;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_A = 0x0001D4EA;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_B = 0x0001D4EB;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_C = 0x0001D4EC;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_D = 0x0001D4ED;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_E = 0x0001D4EE;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_F = 0x0001D4EF;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_G = 0x0001D4F0;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_H = 0x0001D4F1;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_I = 0x0001D4F2;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_J = 0x0001D4F3;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_K = 0x0001D4F4;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_L = 0x0001D4F5;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_M = 0x0001D4F6;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_N = 0x0001D4F7;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_O = 0x0001D4F8;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_P = 0x0001D4F9;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_Q = 0x0001D4FA;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_R = 0x0001D4FB;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_S = 0x0001D4FC;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_T = 0x0001D4FD;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_U = 0x0001D4FE;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_V = 0x0001D4FF;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_W = 0x0001D500;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_X = 0x0001D501;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_Y = 0x0001D502;
const char32_t MATHEMATICAL_BOLD_SCRIPT_SMALL_Z = 0x0001D503;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_A = 0x0001D504;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_B = 0x0001D505;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_D = 0x0001D507;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_E = 0x0001D508;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_F = 0x0001D509;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_G = 0x0001D50A;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_J = 0x0001D50D;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_K = 0x0001D50E;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_L = 0x0001D50F;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_M = 0x0001D510;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_N = 0x0001D511;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_O = 0x0001D512;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_P = 0x0001D513;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_Q = 0x0001D514;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_S = 0x0001D516;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_T = 0x0001D517;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_U = 0x0001D518;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_V = 0x0001D519;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_W = 0x0001D51A;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_X = 0x0001D51B;
const char32_t MATHEMATICAL_FRAKTUR_CAPITAL_Y = 0x0001D51C;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_A = 0x0001D51E;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_B = 0x0001D51F;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_C = 0x0001D520;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_D = 0x0001D521;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_E = 0x0001D522;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_F = 0x0001D523;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_G = 0x0001D524;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_H = 0x0001D525;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_I = 0x0001D526;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_J = 0x0001D527;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_K = 0x0001D528;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_L = 0x0001D529;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_M = 0x0001D52A;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_N = 0x0001D52B;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_O = 0x0001D52C;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_P = 0x0001D52D;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_Q = 0x0001D52E;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_R = 0x0001D52F;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_S = 0x0001D530;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_T = 0x0001D531;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_U = 0x0001D532;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_V = 0x0001D533;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_W = 0x0001D534;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_X = 0x0001D535;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_Y = 0x0001D536;
const char32_t MATHEMATICAL_FRAKTUR_SMALL_Z = 0x0001D537;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_A = 0x0001D538;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_B = 0x0001D539;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_D = 0x0001D53B;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_E = 0x0001D53C;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_F = 0x0001D53D;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_G = 0x0001D53E;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_I = 0x0001D540;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_J = 0x0001D541;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_K = 0x0001D542;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_L = 0x0001D543;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_M = 0x0001D544;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_O = 0x0001D546;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_S = 0x0001D54A;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_T = 0x0001D54B;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_U = 0x0001D54C;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_V = 0x0001D54D;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_W = 0x0001D54E;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_X = 0x0001D54F;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_CAPITAL_Y = 0x0001D550;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_A = 0x0001D552;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_B = 0x0001D553;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_C = 0x0001D554;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_D = 0x0001D555;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_E = 0x0001D556;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_F = 0x0001D557;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_G = 0x0001D558;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_H = 0x0001D559;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_I = 0x0001D55A;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_J = 0x0001D55B;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_K = 0x0001D55C;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_L = 0x0001D55D;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_M = 0x0001D55E;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_N = 0x0001D55F;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_O = 0x0001D560;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_P = 0x0001D561;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_Q = 0x0001D562;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_R = 0x0001D563;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_S = 0x0001D564;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_T = 0x0001D565;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_U = 0x0001D566;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_V = 0x0001D567;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_W = 0x0001D568;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_X = 0x0001D569;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_Y = 0x0001D56A;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_SMALL_Z = 0x0001D56B;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_A = 0x0001D56C;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_B = 0x0001D56D;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_C = 0x0001D56E;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_D = 0x0001D56F;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_E = 0x0001D570;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_F = 0x0001D571;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_G = 0x0001D572;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_H = 0x0001D573;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_I = 0x0001D574;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_J = 0x0001D575;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_K = 0x0001D576;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_L = 0x0001D577;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_M = 0x0001D578;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_N = 0x0001D579;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_O = 0x0001D57A;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_P = 0x0001D57B;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_Q = 0x0001D57C;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_R = 0x0001D57D;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_S = 0x0001D57E;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_T = 0x0001D57F;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_U = 0x0001D580;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_V = 0x0001D581;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_W = 0x0001D582;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_X = 0x0001D583;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_Y = 0x0001D584;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_CAPITAL_Z = 0x0001D585;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_A = 0x0001D586;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_B = 0x0001D587;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_C = 0x0001D588;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_D = 0x0001D589;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_E = 0x0001D58A;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_F = 0x0001D58B;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_G = 0x0001D58C;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_H = 0x0001D58D;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_I = 0x0001D58E;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_J = 0x0001D58F;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_K = 0x0001D590;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_L = 0x0001D591;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_M = 0x0001D592;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_N = 0x0001D593;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_O = 0x0001D594;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_P = 0x0001D595;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_Q = 0x0001D596;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_R = 0x0001D597;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_S = 0x0001D598;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_T = 0x0001D599;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_U = 0x0001D59A;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_V = 0x0001D59B;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_W = 0x0001D59C;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_X = 0x0001D59D;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_Y = 0x0001D59E;
const char32_t MATHEMATICAL_BOLD_FRAKTUR_SMALL_Z = 0x0001D59F;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_A = 0x0001D5A0;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_B = 0x0001D5A1;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_C = 0x0001D5A2;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_D = 0x0001D5A3;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_E = 0x0001D5A4;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_F = 0x0001D5A5;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_G = 0x0001D5A6;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_H = 0x0001D5A7;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_I = 0x0001D5A8;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_J = 0x0001D5A9;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_K = 0x0001D5AA;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_L = 0x0001D5AB;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_M = 0x0001D5AC;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_N = 0x0001D5AD;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_O = 0x0001D5AE;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_P = 0x0001D5AF;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_Q = 0x0001D5B0;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_R = 0x0001D5B1;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_S = 0x0001D5B2;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_T = 0x0001D5B3;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_U = 0x0001D5B4;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_V = 0x0001D5B5;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_W = 0x0001D5B6;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_X = 0x0001D5B7;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_Y = 0x0001D5B8;
const char32_t MATHEMATICAL_SANS_SERIF_CAPITAL_Z = 0x0001D5B9;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_A = 0x0001D5BA;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_B = 0x0001D5BB;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_C = 0x0001D5BC;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_D = 0x0001D5BD;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_E = 0x0001D5BE;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_F = 0x0001D5BF;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_G = 0x0001D5C0;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_H = 0x0001D5C1;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_I = 0x0001D5C2;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_J = 0x0001D5C3;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_K = 0x0001D5C4;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_L = 0x0001D5C5;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_M = 0x0001D5C6;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_N = 0x0001D5C7;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_O = 0x0001D5C8;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_P = 0x0001D5C9;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_Q = 0x0001D5CA;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_R = 0x0001D5CB;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_S = 0x0001D5CC;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_T = 0x0001D5CD;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_U = 0x0001D5CE;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_V = 0x0001D5CF;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_W = 0x0001D5D0;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_X = 0x0001D5D1;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_Y = 0x0001D5D2;
const char32_t MATHEMATICAL_SANS_SERIF_SMALL_Z = 0x0001D5D3;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_A = 0x0001D5D4;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_B = 0x0001D5D5;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_C = 0x0001D5D6;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_D = 0x0001D5D7;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_E = 0x0001D5D8;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_F = 0x0001D5D9;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_G = 0x0001D5DA;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_H = 0x0001D5DB;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_I = 0x0001D5DC;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_J = 0x0001D5DD;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_K = 0x0001D5DE;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_L = 0x0001D5DF;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_M = 0x0001D5E0;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_N = 0x0001D5E1;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_O = 0x0001D5E2;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_P = 0x0001D5E3;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_Q = 0x0001D5E4;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_R = 0x0001D5E5;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_S = 0x0001D5E6;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_T = 0x0001D5E7;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_U = 0x0001D5E8;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_V = 0x0001D5E9;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_W = 0x0001D5EA;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_X = 0x0001D5EB;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_Y = 0x0001D5EC;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_Z = 0x0001D5ED;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_A = 0x0001D5EE;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_B = 0x0001D5EF;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_C = 0x0001D5F0;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_D = 0x0001D5F1;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_E = 0x0001D5F2;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_F = 0x0001D5F3;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_G = 0x0001D5F4;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_H = 0x0001D5F5;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_I = 0x0001D5F6;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_J = 0x0001D5F7;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_K = 0x0001D5F8;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_L = 0x0001D5F9;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_M = 0x0001D5FA;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_N = 0x0001D5FB;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_O = 0x0001D5FC;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_P = 0x0001D5FD;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_Q = 0x0001D5FE;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_R = 0x0001D5FF;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_S = 0x0001D600;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_T = 0x0001D601;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_U = 0x0001D602;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_V = 0x0001D603;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_W = 0x0001D604;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_X = 0x0001D605;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_Y = 0x0001D606;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_Z = 0x0001D607;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_A = 0x0001D608;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_B = 0x0001D609;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_C = 0x0001D60A;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_D = 0x0001D60B;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_E = 0x0001D60C;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_F = 0x0001D60D;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_G = 0x0001D60E;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_H = 0x0001D60F;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_I = 0x0001D610;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_J = 0x0001D611;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_K = 0x0001D612;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_L = 0x0001D613;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_M = 0x0001D614;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_N = 0x0001D615;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_O = 0x0001D616;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_P = 0x0001D617;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_Q = 0x0001D618;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_R = 0x0001D619;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_S = 0x0001D61A;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_T = 0x0001D61B;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_U = 0x0001D61C;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_V = 0x0001D61D;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_W = 0x0001D61E;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_X = 0x0001D61F;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_Y = 0x0001D620;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_CAPITAL_Z = 0x0001D621;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_A = 0x0001D622;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_B = 0x0001D623;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_C = 0x0001D624;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_D = 0x0001D625;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_E = 0x0001D626;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_F = 0x0001D627;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_G = 0x0001D628;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_H = 0x0001D629;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_I = 0x0001D62A;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_J = 0x0001D62B;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_K = 0x0001D62C;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_L = 0x0001D62D;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_M = 0x0001D62E;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_N = 0x0001D62F;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_O = 0x0001D630;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_P = 0x0001D631;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_Q = 0x0001D632;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_R = 0x0001D633;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_S = 0x0001D634;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_T = 0x0001D635;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_U = 0x0001D636;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_V = 0x0001D637;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_W = 0x0001D638;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_X = 0x0001D639;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_Y = 0x0001D63A;
const char32_t MATHEMATICAL_SANS_SERIF_ITALIC_SMALL_Z = 0x0001D63B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_A = 0x0001D63C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_B = 0x0001D63D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_C = 0x0001D63E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_D = 0x0001D63F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_E = 0x0001D640;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_F = 0x0001D641;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_G = 0x0001D642;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_H = 0x0001D643;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_I = 0x0001D644;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_J = 0x0001D645;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_K = 0x0001D646;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_L = 0x0001D647;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_M = 0x0001D648;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_N = 0x0001D649;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_O = 0x0001D64A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_P = 0x0001D64B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_Q = 0x0001D64C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_R = 0x0001D64D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_S = 0x0001D64E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_T = 0x0001D64F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_U = 0x0001D650;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_V = 0x0001D651;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_W = 0x0001D652;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_X = 0x0001D653;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_Y = 0x0001D654;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_Z = 0x0001D655;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_A = 0x0001D656;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_B = 0x0001D657;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_C = 0x0001D658;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_D = 0x0001D659;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_E = 0x0001D65A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_F = 0x0001D65B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_G = 0x0001D65C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_H = 0x0001D65D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_I = 0x0001D65E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_J = 0x0001D65F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_K = 0x0001D660;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_L = 0x0001D661;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_M = 0x0001D662;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_N = 0x0001D663;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_O = 0x0001D664;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_P = 0x0001D665;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_Q = 0x0001D666;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_R = 0x0001D667;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_S = 0x0001D668;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_T = 0x0001D669;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_U = 0x0001D66A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_V = 0x0001D66B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_W = 0x0001D66C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_X = 0x0001D66D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_Y = 0x0001D66E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_Z = 0x0001D66F;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_A = 0x0001D670;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_B = 0x0001D671;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_C = 0x0001D672;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_D = 0x0001D673;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_E = 0x0001D674;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_F = 0x0001D675;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_G = 0x0001D676;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_H = 0x0001D677;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_I = 0x0001D678;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_J = 0x0001D679;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_K = 0x0001D67A;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_L = 0x0001D67B;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_M = 0x0001D67C;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_N = 0x0001D67D;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_O = 0x0001D67E;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_P = 0x0001D67F;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_Q = 0x0001D680;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_R = 0x0001D681;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_S = 0x0001D682;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_T = 0x0001D683;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_U = 0x0001D684;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_V = 0x0001D685;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_W = 0x0001D686;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_X = 0x0001D687;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_Y = 0x0001D688;
const char32_t MATHEMATICAL_MONOSPACE_CAPITAL_Z = 0x0001D689;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_A = 0x0001D68A;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_B = 0x0001D68B;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_C = 0x0001D68C;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_D = 0x0001D68D;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_E = 0x0001D68E;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_F = 0x0001D68F;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_G = 0x0001D690;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_H = 0x0001D691;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_I = 0x0001D692;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_J = 0x0001D693;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_K = 0x0001D694;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_L = 0x0001D695;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_M = 0x0001D696;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_N = 0x0001D697;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_O = 0x0001D698;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_P = 0x0001D699;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_Q = 0x0001D69A;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_R = 0x0001D69B;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_S = 0x0001D69C;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_T = 0x0001D69D;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_U = 0x0001D69E;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_V = 0x0001D69F;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_W = 0x0001D6A0;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_X = 0x0001D6A1;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_Y = 0x0001D6A2;
const char32_t MATHEMATICAL_MONOSPACE_SMALL_Z = 0x0001D6A3;
const char32_t MATHEMATICAL_ITALIC_SMALL_DOTLESS_I = 0x0001D6A4;
const char32_t MATHEMATICAL_ITALIC_SMALL_DOTLESS_J = 0x0001D6A5;
const char32_t MATHEMATICAL_BOLD_CAPITAL_ALPHA = 0x0001D6A8;
const char32_t MATHEMATICAL_BOLD_CAPITAL_BETA = 0x0001D6A9;
const char32_t MATHEMATICAL_BOLD_CAPITAL_GAMMA = 0x0001D6AA;
const char32_t MATHEMATICAL_BOLD_CAPITAL_DELTA = 0x0001D6AB;
const char32_t MATHEMATICAL_BOLD_CAPITAL_EPSILON = 0x0001D6AC;
const char32_t MATHEMATICAL_BOLD_CAPITAL_ZETA = 0x0001D6AD;
const char32_t MATHEMATICAL_BOLD_CAPITAL_ETA = 0x0001D6AE;
const char32_t MATHEMATICAL_BOLD_CAPITAL_THETA = 0x0001D6AF;
const char32_t MATHEMATICAL_BOLD_CAPITAL_IOTA = 0x0001D6B0;
const char32_t MATHEMATICAL_BOLD_CAPITAL_KAPPA = 0x0001D6B1;
const char32_t MATHEMATICAL_BOLD_CAPITAL_LAMDA = 0x0001D6B2;
const char32_t MATHEMATICAL_BOLD_CAPITAL_MU = 0x0001D6B3;
const char32_t MATHEMATICAL_BOLD_CAPITAL_NU = 0x0001D6B4;
const char32_t MATHEMATICAL_BOLD_CAPITAL_XI = 0x0001D6B5;
const char32_t MATHEMATICAL_BOLD_CAPITAL_OMICRON = 0x0001D6B6;
const char32_t MATHEMATICAL_BOLD_CAPITAL_PI = 0x0001D6B7;
const char32_t MATHEMATICAL_BOLD_CAPITAL_RHO = 0x0001D6B8;
const char32_t MATHEMATICAL_BOLD_CAPITAL_THETA_SYMBOL = 0x0001D6B9;
const char32_t MATHEMATICAL_BOLD_CAPITAL_SIGMA = 0x0001D6BA;
const char32_t MATHEMATICAL_BOLD_CAPITAL_TAU = 0x0001D6BB;
const char32_t MATHEMATICAL_BOLD_CAPITAL_UPSILON = 0x0001D6BC;
const char32_t MATHEMATICAL_BOLD_CAPITAL_PHI = 0x0001D6BD;
const char32_t MATHEMATICAL_BOLD_CAPITAL_CHI = 0x0001D6BE;
const char32_t MATHEMATICAL_BOLD_CAPITAL_PSI = 0x0001D6BF;
const char32_t MATHEMATICAL_BOLD_CAPITAL_OMEGA = 0x0001D6C0;
const char32_t MATHEMATICAL_BOLD_NABLA = 0x0001D6C1;
const char32_t MATHEMATICAL_BOLD_SMALL_ALPHA = 0x0001D6C2;
const char32_t MATHEMATICAL_BOLD_SMALL_BETA = 0x0001D6C3;
const char32_t MATHEMATICAL_BOLD_SMALL_GAMMA = 0x0001D6C4;
const char32_t MATHEMATICAL_BOLD_SMALL_DELTA = 0x0001D6C5;
const char32_t MATHEMATICAL_BOLD_SMALL_EPSILON = 0x0001D6C6;
const char32_t MATHEMATICAL_BOLD_SMALL_ZETA = 0x0001D6C7;
const char32_t MATHEMATICAL_BOLD_SMALL_ETA = 0x0001D6C8;
const char32_t MATHEMATICAL_BOLD_SMALL_THETA = 0x0001D6C9;
const char32_t MATHEMATICAL_BOLD_SMALL_IOTA = 0x0001D6CA;
const char32_t MATHEMATICAL_BOLD_SMALL_KAPPA = 0x0001D6CB;
const char32_t MATHEMATICAL_BOLD_SMALL_LAMDA = 0x0001D6CC;
const char32_t MATHEMATICAL_BOLD_SMALL_MU = 0x0001D6CD;
const char32_t MATHEMATICAL_BOLD_SMALL_NU = 0x0001D6CE;
const char32_t MATHEMATICAL_BOLD_SMALL_XI = 0x0001D6CF;
const char32_t MATHEMATICAL_BOLD_SMALL_OMICRON = 0x0001D6D0;
const char32_t MATHEMATICAL_BOLD_SMALL_PI = 0x0001D6D1;
const char32_t MATHEMATICAL_BOLD_SMALL_RHO = 0x0001D6D2;
const char32_t MATHEMATICAL_BOLD_SMALL_FINAL_SIGMA = 0x0001D6D3;
const char32_t MATHEMATICAL_BOLD_SMALL_SIGMA = 0x0001D6D4;
const char32_t MATHEMATICAL_BOLD_SMALL_TAU = 0x0001D6D5;
const char32_t MATHEMATICAL_BOLD_SMALL_UPSILON = 0x0001D6D6;
const char32_t MATHEMATICAL_BOLD_SMALL_PHI = 0x0001D6D7;
const char32_t MATHEMATICAL_BOLD_SMALL_CHI = 0x0001D6D8;
const char32_t MATHEMATICAL_BOLD_SMALL_PSI = 0x0001D6D9;
const char32_t MATHEMATICAL_BOLD_SMALL_OMEGA = 0x0001D6DA;
const char32_t MATHEMATICAL_BOLD_PARTIAL_DIFFERENTIAL = 0x0001D6DB;
const char32_t MATHEMATICAL_BOLD_EPSILON_SYMBOL = 0x0001D6DC;
const char32_t MATHEMATICAL_BOLD_THETA_SYMBOL = 0x0001D6DD;
const char32_t MATHEMATICAL_BOLD_KAPPA_SYMBOL = 0x0001D6DE;
const char32_t MATHEMATICAL_BOLD_PHI_SYMBOL = 0x0001D6DF;
const char32_t MATHEMATICAL_BOLD_RHO_SYMBOL = 0x0001D6E0;
const char32_t MATHEMATICAL_BOLD_PI_SYMBOL = 0x0001D6E1;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_ALPHA = 0x0001D6E2;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_BETA = 0x0001D6E3;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_GAMMA = 0x0001D6E4;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_DELTA = 0x0001D6E5;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_EPSILON = 0x0001D6E6;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_ZETA = 0x0001D6E7;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_ETA = 0x0001D6E8;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_THETA = 0x0001D6E9;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_IOTA = 0x0001D6EA;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_KAPPA = 0x0001D6EB;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_LAMDA = 0x0001D6EC;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_MU = 0x0001D6ED;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_NU = 0x0001D6EE;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_XI = 0x0001D6EF;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_OMICRON = 0x0001D6F0;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_PI = 0x0001D6F1;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_RHO = 0x0001D6F2;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_THETA_SYMBOL = 0x0001D6F3;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_SIGMA = 0x0001D6F4;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_TAU = 0x0001D6F5;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_UPSILON = 0x0001D6F6;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_PHI = 0x0001D6F7;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_CHI = 0x0001D6F8;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_PSI = 0x0001D6F9;
const char32_t MATHEMATICAL_ITALIC_CAPITAL_OMEGA = 0x0001D6FA;
const char32_t MATHEMATICAL_ITALIC_NABLA = 0x0001D6FB;
const char32_t MATHEMATICAL_ITALIC_SMALL_ALPHA = 0x0001D6FC;
const char32_t MATHEMATICAL_ITALIC_SMALL_BETA = 0x0001D6FD;
const char32_t MATHEMATICAL_ITALIC_SMALL_GAMMA = 0x0001D6FE;
const char32_t MATHEMATICAL_ITALIC_SMALL_DELTA = 0x0001D6FF;
const char32_t MATHEMATICAL_ITALIC_SMALL_EPSILON = 0x0001D700;
const char32_t MATHEMATICAL_ITALIC_SMALL_ZETA = 0x0001D701;
const char32_t MATHEMATICAL_ITALIC_SMALL_ETA = 0x0001D702;
const char32_t MATHEMATICAL_ITALIC_SMALL_THETA = 0x0001D703;
const char32_t MATHEMATICAL_ITALIC_SMALL_IOTA = 0x0001D704;
const char32_t MATHEMATICAL_ITALIC_SMALL_KAPPA = 0x0001D705;
const char32_t MATHEMATICAL_ITALIC_SMALL_LAMDA = 0x0001D706;
const char32_t MATHEMATICAL_ITALIC_SMALL_MU = 0x0001D707;
const char32_t MATHEMATICAL_ITALIC_SMALL_NU = 0x0001D708;
const char32_t MATHEMATICAL_ITALIC_SMALL_XI = 0x0001D709;
const char32_t MATHEMATICAL_ITALIC_SMALL_OMICRON = 0x0001D70A;
const char32_t MATHEMATICAL_ITALIC_SMALL_PI = 0x0001D70B;
const char32_t MATHEMATICAL_ITALIC_SMALL_RHO = 0x0001D70C;
const char32_t MATHEMATICAL_ITALIC_SMALL_FINAL_SIGMA = 0x0001D70D;
const char32_t MATHEMATICAL_ITALIC_SMALL_SIGMA = 0x0001D70E;
const char32_t MATHEMATICAL_ITALIC_SMALL_TAU = 0x0001D70F;
const char32_t MATHEMATICAL_ITALIC_SMALL_UPSILON = 0x0001D710;
const char32_t MATHEMATICAL_ITALIC_SMALL_PHI = 0x0001D711;
const char32_t MATHEMATICAL_ITALIC_SMALL_CHI = 0x0001D712;
const char32_t MATHEMATICAL_ITALIC_SMALL_PSI = 0x0001D713;
const char32_t MATHEMATICAL_ITALIC_SMALL_OMEGA = 0x0001D714;
const char32_t MATHEMATICAL_ITALIC_PARTIAL_DIFFERENTIAL = 0x0001D715;
const char32_t MATHEMATICAL_ITALIC_EPSILON_SYMBOL = 0x0001D716;
const char32_t MATHEMATICAL_ITALIC_THETA_SYMBOL = 0x0001D717;
const char32_t MATHEMATICAL_ITALIC_KAPPA_SYMBOL = 0x0001D718;
const char32_t MATHEMATICAL_ITALIC_PHI_SYMBOL = 0x0001D719;
const char32_t MATHEMATICAL_ITALIC_RHO_SYMBOL = 0x0001D71A;
const char32_t MATHEMATICAL_ITALIC_PI_SYMBOL = 0x0001D71B;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_ALPHA = 0x0001D71C;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_BETA = 0x0001D71D;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_GAMMA = 0x0001D71E;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_DELTA = 0x0001D71F;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_EPSILON = 0x0001D720;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_ZETA = 0x0001D721;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_ETA = 0x0001D722;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_THETA = 0x0001D723;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_IOTA = 0x0001D724;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_KAPPA = 0x0001D725;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_LAMDA = 0x0001D726;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_MU = 0x0001D727;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_NU = 0x0001D728;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_XI = 0x0001D729;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_OMICRON = 0x0001D72A;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_PI = 0x0001D72B;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_RHO = 0x0001D72C;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_THETA_SYMBOL = 0x0001D72D;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_SIGMA = 0x0001D72E;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_TAU = 0x0001D72F;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_UPSILON = 0x0001D730;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_PHI = 0x0001D731;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_CHI = 0x0001D732;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_PSI = 0x0001D733;
const char32_t MATHEMATICAL_BOLD_ITALIC_CAPITAL_OMEGA = 0x0001D734;
const char32_t MATHEMATICAL_BOLD_ITALIC_NABLA = 0x0001D735;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_ALPHA = 0x0001D736;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_BETA = 0x0001D737;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_GAMMA = 0x0001D738;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_DELTA = 0x0001D739;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_EPSILON = 0x0001D73A;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_ZETA = 0x0001D73B;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_ETA = 0x0001D73C;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_THETA = 0x0001D73D;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_IOTA = 0x0001D73E;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_KAPPA = 0x0001D73F;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_LAMDA = 0x0001D740;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_MU = 0x0001D741;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_NU = 0x0001D742;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_XI = 0x0001D743;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_OMICRON = 0x0001D744;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_PI = 0x0001D745;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_RHO = 0x0001D746;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_FINAL_SIGMA = 0x0001D747;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_SIGMA = 0x0001D748;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_TAU = 0x0001D749;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_UPSILON = 0x0001D74A;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_PHI = 0x0001D74B;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_CHI = 0x0001D74C;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_PSI = 0x0001D74D;
const char32_t MATHEMATICAL_BOLD_ITALIC_SMALL_OMEGA = 0x0001D74E;
const char32_t MATHEMATICAL_BOLD_ITALIC_PARTIAL_DIFFERENTIAL = 0x0001D74F;
const char32_t MATHEMATICAL_BOLD_ITALIC_EPSILON_SYMBOL = 0x0001D750;
const char32_t MATHEMATICAL_BOLD_ITALIC_THETA_SYMBOL = 0x0001D751;
const char32_t MATHEMATICAL_BOLD_ITALIC_KAPPA_SYMBOL = 0x0001D752;
const char32_t MATHEMATICAL_BOLD_ITALIC_PHI_SYMBOL = 0x0001D753;
const char32_t MATHEMATICAL_BOLD_ITALIC_RHO_SYMBOL = 0x0001D754;
const char32_t MATHEMATICAL_BOLD_ITALIC_PI_SYMBOL = 0x0001D755;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_ALPHA = 0x0001D756;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_BETA = 0x0001D757;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_GAMMA = 0x0001D758;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_DELTA = 0x0001D759;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_EPSILON = 0x0001D75A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_ZETA = 0x0001D75B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_ETA = 0x0001D75C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_THETA = 0x0001D75D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_IOTA = 0x0001D75E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_KAPPA = 0x0001D75F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_LAMDA = 0x0001D760;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_MU = 0x0001D761;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_NU = 0x0001D762;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_XI = 0x0001D763;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_OMICRON = 0x0001D764;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_PI = 0x0001D765;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_RHO = 0x0001D766;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_THETA_SYMBOL = 0x0001D767;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_SIGMA = 0x0001D768;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_TAU = 0x0001D769;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_UPSILON = 0x0001D76A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_PHI = 0x0001D76B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_CHI = 0x0001D76C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_PSI = 0x0001D76D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_CAPITAL_OMEGA = 0x0001D76E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_NABLA = 0x0001D76F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_ALPHA = 0x0001D770;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_BETA = 0x0001D771;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_GAMMA = 0x0001D772;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_DELTA = 0x0001D773;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_EPSILON = 0x0001D774;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_ZETA = 0x0001D775;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_ETA = 0x0001D776;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_THETA = 0x0001D777;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_IOTA = 0x0001D778;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_KAPPA = 0x0001D779;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_LAMDA = 0x0001D77A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_MU = 0x0001D77B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_NU = 0x0001D77C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_XI = 0x0001D77D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_OMICRON = 0x0001D77E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_PI = 0x0001D77F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_RHO = 0x0001D780;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_FINAL_SIGMA = 0x0001D781;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_SIGMA = 0x0001D782;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_TAU = 0x0001D783;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_UPSILON = 0x0001D784;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_PHI = 0x0001D785;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_CHI = 0x0001D786;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_PSI = 0x0001D787;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_SMALL_OMEGA = 0x0001D788;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_PARTIAL_DIFFERENTIAL = 0x0001D789;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_EPSILON_SYMBOL = 0x0001D78A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_THETA_SYMBOL = 0x0001D78B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_KAPPA_SYMBOL = 0x0001D78C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_PHI_SYMBOL = 0x0001D78D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_RHO_SYMBOL = 0x0001D78E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_PI_SYMBOL = 0x0001D78F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_ALPHA = 0x0001D790;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_BETA = 0x0001D791;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_GAMMA = 0x0001D792;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_DELTA = 0x0001D793;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_EPSILON = 0x0001D794;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_ZETA = 0x0001D795;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_ETA = 0x0001D796;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_THETA = 0x0001D797;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_IOTA = 0x0001D798;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_KAPPA = 0x0001D799;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_LAMDA = 0x0001D79A;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_MU = 0x0001D79B;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_NU = 0x0001D79C;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_XI = 0x0001D79D;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_OMICRON = 0x0001D79E;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_PI = 0x0001D79F;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_RHO = 0x0001D7A0;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_THETA_SYMBOL = 0x0001D7A1;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_SIGMA = 0x0001D7A2;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_TAU = 0x0001D7A3;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_UPSILON = 0x0001D7A4;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_PHI = 0x0001D7A5;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_CHI = 0x0001D7A6;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_PSI = 0x0001D7A7;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_CAPITAL_OMEGA = 0x0001D7A8;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_NABLA = 0x0001D7A9;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_ALPHA = 0x0001D7AA;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_BETA = 0x0001D7AB;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_GAMMA = 0x0001D7AC;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_DELTA = 0x0001D7AD;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_EPSILON = 0x0001D7AE;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_ZETA = 0x0001D7AF;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_ETA = 0x0001D7B0;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_THETA = 0x0001D7B1;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_IOTA = 0x0001D7B2;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_KAPPA = 0x0001D7B3;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_LAMDA = 0x0001D7B4;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_MU = 0x0001D7B5;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_NU = 0x0001D7B6;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_XI = 0x0001D7B7;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_OMICRON = 0x0001D7B8;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_PI = 0x0001D7B9;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_RHO = 0x0001D7BA;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_FINAL_SIGMA = 0x0001D7BB;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_SIGMA = 0x0001D7BC;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_TAU = 0x0001D7BD;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_UPSILON = 0x0001D7BE;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_PHI = 0x0001D7BF;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_CHI = 0x0001D7C0;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_PSI = 0x0001D7C1;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_SMALL_OMEGA = 0x0001D7C2;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_PARTIAL_DIFFERENTIAL = 0x0001D7C3;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_EPSILON_SYMBOL = 0x0001D7C4;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_THETA_SYMBOL = 0x0001D7C5;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_KAPPA_SYMBOL = 0x0001D7C6;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_PHI_SYMBOL = 0x0001D7C7;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_RHO_SYMBOL = 0x0001D7C8;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_ITALIC_PI_SYMBOL = 0x0001D7C9;
const char32_t MATHEMATICAL_BOLD_CAPITAL_DIGAMMA = 0x0001D7CA;
const char32_t MATHEMATICAL_BOLD_SMALL_DIGAMMA = 0x0001D7CB;
const char32_t MATHEMATICAL_BOLD_DIGIT_ZERO = 0x0001D7CE;
const char32_t MATHEMATICAL_BOLD_DIGIT_ONE = 0x0001D7CF;
const char32_t MATHEMATICAL_BOLD_DIGIT_TWO = 0x0001D7D0;
const char32_t MATHEMATICAL_BOLD_DIGIT_THREE = 0x0001D7D1;
const char32_t MATHEMATICAL_BOLD_DIGIT_FOUR = 0x0001D7D2;
const char32_t MATHEMATICAL_BOLD_DIGIT_FIVE = 0x0001D7D3;
const char32_t MATHEMATICAL_BOLD_DIGIT_SIX = 0x0001D7D4;
const char32_t MATHEMATICAL_BOLD_DIGIT_SEVEN = 0x0001D7D5;
const char32_t MATHEMATICAL_BOLD_DIGIT_EIGHT = 0x0001D7D6;
const char32_t MATHEMATICAL_BOLD_DIGIT_NINE = 0x0001D7D7;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_ZERO = 0x0001D7D8;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_ONE = 0x0001D7D9;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_TWO = 0x0001D7DA;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_THREE = 0x0001D7DB;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_FOUR = 0x0001D7DC;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_FIVE = 0x0001D7DD;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_SIX = 0x0001D7DE;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_SEVEN = 0x0001D7DF;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_EIGHT = 0x0001D7E0;
const char32_t MATHEMATICAL_DOUBLE_STRUCK_DIGIT_NINE = 0x0001D7E1;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_ZERO = 0x0001D7E2;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_ONE = 0x0001D7E3;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_TWO = 0x0001D7E4;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_THREE = 0x0001D7E5;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_FOUR = 0x0001D7E6;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_FIVE = 0x0001D7E7;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_SIX = 0x0001D7E8;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_SEVEN = 0x0001D7E9;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_EIGHT = 0x0001D7EA;
const char32_t MATHEMATICAL_SANS_SERIF_DIGIT_NINE = 0x0001D7EB;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_ZERO = 0x0001D7EC;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_ONE = 0x0001D7ED;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_TWO = 0x0001D7EE;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_THREE = 0x0001D7EF;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_FOUR = 0x0001D7F0;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_FIVE = 0x0001D7F1;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_SIX = 0x0001D7F2;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_SEVEN = 0x0001D7F3;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_EIGHT = 0x0001D7F4;
const char32_t MATHEMATICAL_SANS_SERIF_BOLD_DIGIT_NINE = 0x0001D7F5;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_ZERO = 0x0001D7F6;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_ONE = 0x0001D7F7;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_TWO = 0x0001D7F8;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_THREE = 0x0001D7F9;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_FOUR = 0x0001D7FA;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_FIVE = 0x0001D7FB;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_SIX = 0x0001D7FC;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_SEVEN = 0x0001D7FD;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_EIGHT = 0x0001D7FE;
const char32_t MATHEMATICAL_MONOSPACE_DIGIT_NINE = 0x0001D7FF;
const char32_t SIGNWRITING_HAND_FIST_INDEX = 0x0001D800;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX = 0x0001D801;
const char32_t SIGNWRITING_HAND_CUP_INDEX = 0x0001D802;
const char32_t SIGNWRITING_HAND_OVAL_INDEX = 0x0001D803;
const char32_t SIGNWRITING_HAND_HINGE_INDEX = 0x0001D804;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX = 0x0001D805;
const char32_t SIGNWRITING_HAND_FIST_INDEX_BENT = 0x0001D806;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_BENT = 0x0001D807;
const char32_t SIGNWRITING_HAND_FIST_THUMB_UNDER_INDEX_BENT = 0x0001D808;
const char32_t SIGNWRITING_HAND_FIST_INDEX_RAISED_KNUCKLE = 0x0001D809;
const char32_t SIGNWRITING_HAND_FIST_INDEX_CUPPED = 0x0001D80A;
const char32_t SIGNWRITING_HAND_FIST_INDEX_HINGED = 0x0001D80B;
const char32_t SIGNWRITING_HAND_FIST_INDEX_HINGED_LOW = 0x0001D80C;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_HINGE = 0x0001D80D;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE = 0x0001D80E;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_MIDDLE = 0x0001D80F;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_BENT = 0x0001D810;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_RAISED_KNUCKLES = 0x0001D811;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_HINGED = 0x0001D812;
const char32_t SIGNWRITING_HAND_FIST_INDEX_UP_MIDDLE_HINGED = 0x0001D813;
const char32_t SIGNWRITING_HAND_FIST_INDEX_HINGED_MIDDLE_UP = 0x0001D814;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED = 0x0001D815;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_INDEX_BENT = 0x0001D816;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_MIDDLE_BENT = 0x0001D817;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_CUPPED = 0x0001D818;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_HINGED = 0x0001D819;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CROSSED = 0x0001D81A;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_MIDDLE_CROSSED = 0x0001D81B;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_BENT_OVER_INDEX = 0x0001D81C;
const char32_t SIGNWRITING_HAND_FIST_INDEX_BENT_OVER_MIDDLE = 0x0001D81D;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB = 0x0001D81E;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_MIDDLE_THUMB = 0x0001D81F;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_STRAIGHT_THUMB_BENT = 0x0001D820;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_BENT_THUMB_STRAIGHT = 0x0001D821;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_BENT = 0x0001D822;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_HINGED_SPREAD_THUMB_SIDE = 0x0001D823;
const char32_t SIGNWRITING_HAND_FIST_INDEX_UP_MIDDLE_HINGED_THUMB_SIDE = 0x0001D824;
const char32_t SIGNWRITING_HAND_FIST_INDEX_UP_MIDDLE_HINGED_THUMB_CONJOINED = 0x0001D825;
const char32_t SIGNWRITING_HAND_FIST_INDEX_HINGED_MIDDLE_UP_THUMB_SIDE = 0x0001D826;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_UP_SPREAD_THUMB_FORWARD = 0x0001D827;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_CUPPED = 0x0001D828;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_CIRCLED = 0x0001D829;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_HOOKED = 0x0001D82A;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_HINGED = 0x0001D82B;
const char32_t SIGNWRITING_HAND_FIST_THUMB_BETWEEN_INDEX_MIDDLE_STRAIGHT = 0x0001D82C;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_THUMB_SIDE = 0x0001D82D;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_THUMB_SIDE_CONJOINED = 0x0001D82E;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_THUMB_SIDE_BENT = 0x0001D82F;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_HOOKED_INDEX_UP = 0x0001D830;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_HOOKED_MIDDLE_UP = 0x0001D831;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_HINGED_THUMB_SIDE = 0x0001D832;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CROSSED_THUMB_SIDE = 0x0001D833;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_THUMB_FORWARD = 0x0001D834;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CONJOINED_CUPPED_THUMB_FORWARD = 0x0001D835;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_CUPPED_INDEX_UP = 0x0001D836;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_CUPPED_MIDDLE_UP = 0x0001D837;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_CIRCLED_INDEX_UP = 0x0001D838;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_CIRCLED_INDEX_HINGED = 0x0001D839;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_ANGLED_OUT_MIDDLE_UP = 0x0001D83A;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_ANGLED_IN_MIDDLE_UP = 0x0001D83B;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_CIRCLED_MIDDLE_UP = 0x0001D83C;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_CONJOINED_HINGED = 0x0001D83D;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_ANGLED_OUT = 0x0001D83E;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_THUMB_ANGLED = 0x0001D83F;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_ANGLED_OUT_INDEX_UP = 0x0001D840;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_ANGLED_OUT_INDEX_CROSSED = 0x0001D841;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_ANGLED_INDEX_UP = 0x0001D842;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_HOOKED_MIDDLE_HINGED = 0x0001D843;
const char32_t SIGNWRITING_HAND_FLAT_FOUR_FINGERS = 0x0001D844;
const char32_t SIGNWRITING_HAND_FLAT_FOUR_FINGERS_BENT = 0x0001D845;
const char32_t SIGNWRITING_HAND_FLAT_FOUR_FINGERS_HINGED = 0x0001D846;
const char32_t SIGNWRITING_HAND_FLAT_FOUR_FINGERS_CONJOINED = 0x0001D847;
const char32_t SIGNWRITING_HAND_FLAT_FOUR_FINGERS_CONJOINED_SPLIT = 0x0001D848;
const char32_t SIGNWRITING_HAND_CLAW_FOUR_FINGERS_CONJOINED = 0x0001D849;
const char32_t SIGNWRITING_HAND_FIST_FOUR_FINGERS_CONJOINED_BENT = 0x0001D84A;
const char32_t SIGNWRITING_HAND_HINGE_FOUR_FINGERS_CONJOINED = 0x0001D84B;
const char32_t SIGNWRITING_HAND_FLAT_FIVE_FINGERS_SPREAD = 0x0001D84C;
const char32_t SIGNWRITING_HAND_FLAT_HEEL_FIVE_FINGERS_SPREAD = 0x0001D84D;
const char32_t SIGNWRITING_HAND_FLAT_FIVE_FINGERS_SPREAD_FOUR_BENT = 0x0001D84E;
const char32_t SIGNWRITING_HAND_FLAT_HEEL_FIVE_FINGERS_SPREAD_FOUR_BENT = 0x0001D84F;
const char32_t SIGNWRITING_HAND_FLAT_FIVE_FINGERS_SPREAD_BENT = 0x0001D850;
const char32_t SIGNWRITING_HAND_FLAT_HEEL_FIVE_FINGERS_SPREAD_BENT = 0x0001D851;
const char32_t SIGNWRITING_HAND_FLAT_FIVE_FINGERS_SPREAD_THUMB_FORWARD = 0x0001D852;
const char32_t SIGNWRITING_HAND_CUP_FIVE_FINGERS_SPREAD = 0x0001D853;
const char32_t SIGNWRITING_HAND_CUP_FIVE_FINGERS_SPREAD_OPEN = 0x0001D854;
const char32_t SIGNWRITING_HAND_HINGE_FIVE_FINGERS_SPREAD_OPEN = 0x0001D855;
const char32_t SIGNWRITING_HAND_OVAL_FIVE_FINGERS_SPREAD = 0x0001D856;
const char32_t SIGNWRITING_HAND_FLAT_FIVE_FINGERS_SPREAD_HINGED = 0x0001D857;
const char32_t SIGNWRITING_HAND_FLAT_FIVE_FINGERS_SPREAD_HINGED_THUMB_SIDE = 0x0001D858;
const char32_t SIGNWRITING_HAND_FLAT_FIVE_FINGERS_SPREAD_HINGED_NO_THUMB = 0x0001D859;
const char32_t SIGNWRITING_HAND_FLAT = 0x0001D85A;
const char32_t SIGNWRITING_HAND_FLAT_BETWEEN_PALM_FACINGS = 0x0001D85B;
const char32_t SIGNWRITING_HAND_FLAT_HEEL = 0x0001D85C;
const char32_t SIGNWRITING_HAND_FLAT_THUMB_SIDE = 0x0001D85D;
const char32_t SIGNWRITING_HAND_FLAT_HEEL_THUMB_SIDE = 0x0001D85E;
const char32_t SIGNWRITING_HAND_FLAT_THUMB_BENT = 0x0001D85F;
const char32_t SIGNWRITING_HAND_FLAT_THUMB_FORWARD = 0x0001D860;
const char32_t SIGNWRITING_HAND_FLAT_SPLIT_INDEX_THUMB_SIDE = 0x0001D861;
const char32_t SIGNWRITING_HAND_FLAT_SPLIT_CENTRE = 0x0001D862;
const char32_t SIGNWRITING_HAND_FLAT_SPLIT_CENTRE_THUMB_SIDE = 0x0001D863;
const char32_t SIGNWRITING_HAND_FLAT_SPLIT_CENTRE_THUMB_SIDE_BENT = 0x0001D864;
const char32_t SIGNWRITING_HAND_FLAT_SPLIT_LITTLE = 0x0001D865;
const char32_t SIGNWRITING_HAND_CLAW = 0x0001D866;
const char32_t SIGNWRITING_HAND_CLAW_THUMB_SIDE = 0x0001D867;
const char32_t SIGNWRITING_HAND_CLAW_NO_THUMB = 0x0001D868;
const char32_t SIGNWRITING_HAND_CLAW_THUMB_FORWARD = 0x0001D869;
const char32_t SIGNWRITING_HAND_HOOK_CURLICUE = 0x0001D86A;
const char32_t SIGNWRITING_HAND_HOOK = 0x0001D86B;
const char32_t SIGNWRITING_HAND_CUP_OPEN = 0x0001D86C;
const char32_t SIGNWRITING_HAND_CUP = 0x0001D86D;
const char32_t SIGNWRITING_HAND_CUP_OPEN_THUMB_SIDE = 0x0001D86E;
const char32_t SIGNWRITING_HAND_CUP_THUMB_SIDE = 0x0001D86F;
const char32_t SIGNWRITING_HAND_CUP_OPEN_NO_THUMB = 0x0001D870;
const char32_t SIGNWRITING_HAND_CUP_NO_THUMB = 0x0001D871;
const char32_t SIGNWRITING_HAND_CUP_OPEN_THUMB_FORWARD = 0x0001D872;
const char32_t SIGNWRITING_HAND_CUP_THUMB_FORWARD = 0x0001D873;
const char32_t SIGNWRITING_HAND_CURLICUE_OPEN = 0x0001D874;
const char32_t SIGNWRITING_HAND_CURLICUE = 0x0001D875;
const char32_t SIGNWRITING_HAND_CIRCLE = 0x0001D876;
const char32_t SIGNWRITING_HAND_OVAL = 0x0001D877;
const char32_t SIGNWRITING_HAND_OVAL_THUMB_SIDE = 0x0001D878;
const char32_t SIGNWRITING_HAND_OVAL_NO_THUMB = 0x0001D879;
const char32_t SIGNWRITING_HAND_OVAL_THUMB_FORWARD = 0x0001D87A;
const char32_t SIGNWRITING_HAND_HINGE_OPEN = 0x0001D87B;
const char32_t SIGNWRITING_HAND_HINGE_OPEN_THUMB_FORWARD = 0x0001D87C;
const char32_t SIGNWRITING_HAND_HINGE = 0x0001D87D;
const char32_t SIGNWRITING_HAND_HINGE_SMALL = 0x0001D87E;
const char32_t SIGNWRITING_HAND_HINGE_OPEN_THUMB_SIDE = 0x0001D87F;
const char32_t SIGNWRITING_HAND_HINGE_THUMB_SIDE = 0x0001D880;
const char32_t SIGNWRITING_HAND_HINGE_OPEN_NO_THUMB = 0x0001D881;
const char32_t SIGNWRITING_HAND_HINGE_NO_THUMB = 0x0001D882;
const char32_t SIGNWRITING_HAND_HINGE_THUMB_SIDE_TOUCHING_INDEX = 0x0001D883;
const char32_t SIGNWRITING_HAND_HINGE_THUMB_BETWEEN_MIDDLE_RING = 0x0001D884;
const char32_t SIGNWRITING_HAND_ANGLE = 0x0001D885;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_RING = 0x0001D886;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_MIDDLE_RING = 0x0001D887;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_MIDDLE_RING = 0x0001D888;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX_MIDDLE_RING = 0x0001D889;
const char32_t SIGNWRITING_HAND_HINGE_LITTLE = 0x0001D88A;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_RING_BENT = 0x0001D88B;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_RING_CONJOINED = 0x0001D88C;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_MIDDLE_RING_CONJOINED = 0x0001D88D;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_DOWN = 0x0001D88E;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_DOWN_RIPPLE_STRAIGHT = 0x0001D88F;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_DOWN_RIPPLE_CURVED = 0x0001D890;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_DOWN_OTHERS_CIRCLED = 0x0001D891;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_UP = 0x0001D892;
const char32_t SIGNWRITING_HAND_FIST_THUMB_UNDER_LITTLE_UP = 0x0001D893;
const char32_t SIGNWRITING_HAND_CIRCLE_LITTLE_UP = 0x0001D894;
const char32_t SIGNWRITING_HAND_OVAL_LITTLE_UP = 0x0001D895;
const char32_t SIGNWRITING_HAND_ANGLE_LITTLE_UP = 0x0001D896;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_RAISED_KNUCKLE = 0x0001D897;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_BENT = 0x0001D898;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_TOUCHES_THUMB = 0x0001D899;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_THUMB = 0x0001D89A;
const char32_t SIGNWRITING_HAND_HINGE_LITTLE_THUMB = 0x0001D89B;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_INDEX_THUMB = 0x0001D89C;
const char32_t SIGNWRITING_HAND_HINGE_LITTLE_INDEX_THUMB = 0x0001D89D;
const char32_t SIGNWRITING_HAND_ANGLE_LITTLE_INDEX_THUMB_INDEX_THUMB_OUT = 0x0001D89E;
const char32_t SIGNWRITING_HAND_ANGLE_LITTLE_INDEX_THUMB_INDEX_THUMB = 0x0001D89F;
const char32_t SIGNWRITING_HAND_FIST_LITTLE_INDEX = 0x0001D8A0;
const char32_t SIGNWRITING_HAND_CIRCLE_LITTLE_INDEX = 0x0001D8A1;
const char32_t SIGNWRITING_HAND_HINGE_LITTLE_INDEX = 0x0001D8A2;
const char32_t SIGNWRITING_HAND_ANGLE_LITTLE_INDEX = 0x0001D8A3;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_LITTLE = 0x0001D8A4;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_MIDDLE_LITTLE = 0x0001D8A5;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_MIDDLE_LITTLE = 0x0001D8A6;
const char32_t SIGNWRITING_HAND_HINGE_RING = 0x0001D8A7;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX_MIDDLE_LITTLE = 0x0001D8A8;
const char32_t SIGNWRITING_HAND_FIST_INDEX_MIDDLE_CROSS_LITTLE = 0x0001D8A9;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_MIDDLE_CROSS_LITTLE = 0x0001D8AA;
const char32_t SIGNWRITING_HAND_FIST_RING_DOWN = 0x0001D8AB;
const char32_t SIGNWRITING_HAND_HINGE_RING_DOWN_INDEX_THUMB_HOOK_MIDDLE = 0x0001D8AC;
const char32_t SIGNWRITING_HAND_ANGLE_RING_DOWN_MIDDLE_THUMB_INDEX_CROSS = 0x0001D8AD;
const char32_t SIGNWRITING_HAND_FIST_RING_UP = 0x0001D8AE;
const char32_t SIGNWRITING_HAND_FIST_RING_RAISED_KNUCKLE = 0x0001D8AF;
const char32_t SIGNWRITING_HAND_FIST_RING_LITTLE = 0x0001D8B0;
const char32_t SIGNWRITING_HAND_CIRCLE_RING_LITTLE = 0x0001D8B1;
const char32_t SIGNWRITING_HAND_OVAL_RING_LITTLE = 0x0001D8B2;
const char32_t SIGNWRITING_HAND_ANGLE_RING_LITTLE = 0x0001D8B3;
const char32_t SIGNWRITING_HAND_FIST_RING_MIDDLE = 0x0001D8B4;
const char32_t SIGNWRITING_HAND_FIST_RING_MIDDLE_CONJOINED = 0x0001D8B5;
const char32_t SIGNWRITING_HAND_FIST_RING_MIDDLE_RAISED_KNUCKLES = 0x0001D8B6;
const char32_t SIGNWRITING_HAND_FIST_RING_INDEX = 0x0001D8B7;
const char32_t SIGNWRITING_HAND_FIST_RING_THUMB = 0x0001D8B8;
const char32_t SIGNWRITING_HAND_HOOK_RING_THUMB = 0x0001D8B9;
const char32_t SIGNWRITING_HAND_FIST_INDEX_RING_LITTLE = 0x0001D8BA;
const char32_t SIGNWRITING_HAND_CIRCLE_INDEX_RING_LITTLE = 0x0001D8BB;
const char32_t SIGNWRITING_HAND_CURLICUE_INDEX_RING_LITTLE_ON = 0x0001D8BC;
const char32_t SIGNWRITING_HAND_HOOK_INDEX_RING_LITTLE_OUT = 0x0001D8BD;
const char32_t SIGNWRITING_HAND_HOOK_INDEX_RING_LITTLE_IN = 0x0001D8BE;
const char32_t SIGNWRITING_HAND_HOOK_INDEX_RING_LITTLE_UNDER = 0x0001D8BF;
const char32_t SIGNWRITING_HAND_CUP_INDEX_RING_LITTLE = 0x0001D8C0;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_RING_LITTLE = 0x0001D8C1;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX_RING_LITTLE_OUT = 0x0001D8C2;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX_RING_LITTLE = 0x0001D8C3;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_DOWN = 0x0001D8C4;
const char32_t SIGNWRITING_HAND_HINGE_MIDDLE = 0x0001D8C5;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_UP = 0x0001D8C6;
const char32_t SIGNWRITING_HAND_CIRCLE_MIDDLE_UP = 0x0001D8C7;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_RAISED_KNUCKLE = 0x0001D8C8;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_UP_THUMB_SIDE = 0x0001D8C9;
const char32_t SIGNWRITING_HAND_HOOK_MIDDLE_THUMB = 0x0001D8CA;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_THUMB_LITTLE = 0x0001D8CB;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_LITTLE = 0x0001D8CC;
const char32_t SIGNWRITING_HAND_FIST_MIDDLE_RING_LITTLE = 0x0001D8CD;
const char32_t SIGNWRITING_HAND_CIRCLE_MIDDLE_RING_LITTLE = 0x0001D8CE;
const char32_t SIGNWRITING_HAND_CURLICUE_MIDDLE_RING_LITTLE_ON = 0x0001D8CF;
const char32_t SIGNWRITING_HAND_CUP_MIDDLE_RING_LITTLE = 0x0001D8D0;
const char32_t SIGNWRITING_HAND_HINGE_MIDDLE_RING_LITTLE = 0x0001D8D1;
const char32_t SIGNWRITING_HAND_ANGLE_MIDDLE_RING_LITTLE_OUT = 0x0001D8D2;
const char32_t SIGNWRITING_HAND_ANGLE_MIDDLE_RING_LITTLE_IN = 0x0001D8D3;
const char32_t SIGNWRITING_HAND_ANGLE_MIDDLE_RING_LITTLE = 0x0001D8D4;
const char32_t SIGNWRITING_HAND_CIRCLE_MIDDLE_RING_LITTLE_BENT = 0x0001D8D5;
const char32_t SIGNWRITING_HAND_CLAW_MIDDLE_RING_LITTLE_CONJOINED = 0x0001D8D6;
const char32_t SIGNWRITING_HAND_CLAW_MIDDLE_RING_LITTLE_CONJOINED_SIDE = 0x0001D8D7;
const char32_t SIGNWRITING_HAND_HOOK_MIDDLE_RING_LITTLE_CONJOINED_OUT = 0x0001D8D8;
const char32_t SIGNWRITING_HAND_HOOK_MIDDLE_RING_LITTLE_CONJOINED_IN = 0x0001D8D9;
const char32_t SIGNWRITING_HAND_HOOK_MIDDLE_RING_LITTLE_CONJOINED = 0x0001D8DA;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_HINGED = 0x0001D8DB;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_SIDE = 0x0001D8DC;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_THUMB_SIDE = 0x0001D8DD;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_SIDE_THUMB_DIAGONAL = 0x0001D8DE;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_SIDE_THUMB_CONJOINED = 0x0001D8DF;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_SIDE_THUMB_BENT = 0x0001D8E0;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_SIDE_INDEX_BENT = 0x0001D8E1;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_SIDE_BOTH_BENT = 0x0001D8E2;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_SIDE_INDEX_HINGE = 0x0001D8E3;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_FORWARD_INDEX_STRAIGHT = 0x0001D8E4;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_FORWARD_INDEX_BENT = 0x0001D8E5;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_HOOK = 0x0001D8E6;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_CURLICUE = 0x0001D8E7;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_CURVE_THUMB_INSIDE = 0x0001D8E8;
const char32_t SIGNWRITING_HAND_CLAW_INDEX_THUMB_CURVE_THUMB_INSIDE = 0x0001D8E9;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_CURVE_THUMB_UNDER = 0x0001D8EA;
const char32_t SIGNWRITING_HAND_FIST_INDEX_THUMB_CIRCLE = 0x0001D8EB;
const char32_t SIGNWRITING_HAND_CUP_INDEX_THUMB = 0x0001D8EC;
const char32_t SIGNWRITING_HAND_CUP_INDEX_THUMB_OPEN = 0x0001D8ED;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_THUMB_OPEN = 0x0001D8EE;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_THUMB_LARGE = 0x0001D8EF;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_THUMB = 0x0001D8F0;
const char32_t SIGNWRITING_HAND_HINGE_INDEX_THUMB_SMALL = 0x0001D8F1;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX_THUMB_OUT = 0x0001D8F2;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX_THUMB_IN = 0x0001D8F3;
const char32_t SIGNWRITING_HAND_ANGLE_INDEX_THUMB = 0x0001D8F4;
const char32_t SIGNWRITING_HAND_FIST_THUMB = 0x0001D8F5;
const char32_t SIGNWRITING_HAND_FIST_THUMB_HEEL = 0x0001D8F6;
const char32_t SIGNWRITING_HAND_FIST_THUMB_SIDE_DIAGONAL = 0x0001D8F7;
const char32_t SIGNWRITING_HAND_FIST_THUMB_SIDE_CONJOINED = 0x0001D8F8;
const char32_t SIGNWRITING_HAND_FIST_THUMB_SIDE_BENT = 0x0001D8F9;
const char32_t SIGNWRITING_HAND_FIST_THUMB_FORWARD = 0x0001D8FA;
const char32_t SIGNWRITING_HAND_FIST_THUMB_BETWEEN_INDEX_MIDDLE = 0x0001D8FB;
const char32_t SIGNWRITING_HAND_FIST_THUMB_BETWEEN_MIDDLE_RING = 0x0001D8FC;
const char32_t SIGNWRITING_HAND_FIST_THUMB_BETWEEN_RING_LITTLE = 0x0001D8FD;
const char32_t SIGNWRITING_HAND_FIST_THUMB_UNDER_TWO_FINGERS = 0x0001D8FE;
const char32_t SIGNWRITING_HAND_FIST_THUMB_OVER_TWO_FINGERS = 0x0001D8FF;
const char32_t SIGNWRITING_HAND_FIST_THUMB_UNDER_THREE_FINGERS = 0x0001D900;
const char32_t SIGNWRITING_HAND_FIST_THUMB_UNDER_FOUR_FINGERS = 0x0001D901;
const char32_t SIGNWRITING_HAND_FIST_THUMB_OVER_FOUR_RAISED_KNUCKLES = 0x0001D902;
const char32_t SIGNWRITING_HAND_FIST = 0x0001D903;
const char32_t SIGNWRITING_HAND_FIST_HEEL = 0x0001D904;
const char32_t SIGNWRITING_TOUCH_SINGLE = 0x0001D905;
const char32_t SIGNWRITING_TOUCH_MULTIPLE = 0x0001D906;
const char32_t SIGNWRITING_TOUCH_BETWEEN = 0x0001D907;
const char32_t SIGNWRITING_GRASP_SINGLE = 0x0001D908;
const char32_t SIGNWRITING_GRASP_MULTIPLE = 0x0001D909;
const char32_t SIGNWRITING_GRASP_BETWEEN = 0x0001D90A;
const char32_t SIGNWRITING_STRIKE_SINGLE = 0x0001D90B;
const char32_t SIGNWRITING_STRIKE_MULTIPLE = 0x0001D90C;
const char32_t SIGNWRITING_STRIKE_BETWEEN = 0x0001D90D;
const char32_t SIGNWRITING_BRUSH_SINGLE = 0x0001D90E;
const char32_t SIGNWRITING_BRUSH_MULTIPLE = 0x0001D90F;
const char32_t SIGNWRITING_BRUSH_BETWEEN = 0x0001D910;
const char32_t SIGNWRITING_RUB_SINGLE = 0x0001D911;
const char32_t SIGNWRITING_RUB_MULTIPLE = 0x0001D912;
const char32_t SIGNWRITING_RUB_BETWEEN = 0x0001D913;
const char32_t SIGNWRITING_SURFACE_SYMBOLS = 0x0001D914;
const char32_t SIGNWRITING_SURFACE_BETWEEN = 0x0001D915;
const char32_t SIGNWRITING_SQUEEZE_LARGE_SINGLE = 0x0001D916;
const char32_t SIGNWRITING_SQUEEZE_SMALL_SINGLE = 0x0001D917;
const char32_t SIGNWRITING_SQUEEZE_LARGE_MULTIPLE = 0x0001D918;
const char32_t SIGNWRITING_SQUEEZE_SMALL_MULTIPLE = 0x0001D919;
const char32_t SIGNWRITING_SQUEEZE_SEQUENTIAL = 0x0001D91A;
const char32_t SIGNWRITING_FLICK_LARGE_SINGLE = 0x0001D91B;
const char32_t SIGNWRITING_FLICK_SMALL_SINGLE = 0x0001D91C;
const char32_t SIGNWRITING_FLICK_LARGE_MULTIPLE = 0x0001D91D;
const char32_t SIGNWRITING_FLICK_SMALL_MULTIPLE = 0x0001D91E;
const char32_t SIGNWRITING_FLICK_SEQUENTIAL = 0x0001D91F;
const char32_t SIGNWRITING_SQUEEZE_FLICK_ALTERNATING = 0x0001D920;
const char32_t SIGNWRITING_MOVEMENT_HINGE_UP_DOWN_LARGE = 0x0001D921;
const char32_t SIGNWRITING_MOVEMENT_HINGE_UP_DOWN_SMALL = 0x0001D922;
const char32_t SIGNWRITING_MOVEMENT_HINGE_UP_SEQUENTIAL = 0x0001D923;
const char32_t SIGNWRITING_MOVEMENT_HINGE_DOWN_SEQUENTIAL = 0x0001D924;
const char32_t SIGNWRITING_MOVEMENT_HINGE_UP_DOWN_ALTERNATING_LARGE = 0x0001D925;
const char32_t SIGNWRITING_MOVEMENT_HINGE_UP_DOWN_ALTERNATING_SMALL = 0x0001D926;
const char32_t SIGNWRITING_MOVEMENT_HINGE_SIDE_TO_SIDE_SCISSORS = 0x0001D927;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_FINGER_CONTACT = 0x0001D928;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_FINGER_CONTACT = 0x0001D929;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_SINGLE_STRAIGHT_SMALL = 0x0001D92A;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_SINGLE_STRAIGHT_MEDIUM = 0x0001D92B;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_SINGLE_STRAIGHT_LARGE = 0x0001D92C;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_SINGLE_STRAIGHT_LARGEST = 0x0001D92D;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_SINGLE_WRIST_FLEX = 0x0001D92E;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_DOUBLE_STRAIGHT = 0x0001D92F;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_DOUBLE_WRIST_FLEX = 0x0001D930;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_DOUBLE_ALTERNATING = 0x0001D931;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_DOUBLE_ALTERNATING_WRIST_FLEX = 0x0001D932;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CROSS = 0x0001D933;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_TRIPLE_STRAIGHT_MOVEMENT = 0x0001D934;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_TRIPLE_WRIST_FLEX = 0x0001D935;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_TRIPLE_ALTERNATING = 0x0001D936;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_TRIPLE_ALTERNATING_WRIST_FLEX = 0x0001D937;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_BEND_SMALL = 0x0001D938;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_BEND_MEDIUM = 0x0001D939;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_BEND_LARGE = 0x0001D93A;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CORNER_SMALL = 0x0001D93B;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CORNER_MEDIUM = 0x0001D93C;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CORNER_LARGE = 0x0001D93D;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CORNER_ROTATION = 0x0001D93E;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CHECK_SMALL = 0x0001D93F;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CHECK_MEDIUM = 0x0001D940;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CHECK_LARGE = 0x0001D941;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_BOX_SMALL = 0x0001D942;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_BOX_MEDIUM = 0x0001D943;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_BOX_LARGE = 0x0001D944;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_ZIGZAG_SMALL = 0x0001D945;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_ZIGZAG_MEDIUM = 0x0001D946;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_ZIGZAG_LARGE = 0x0001D947;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_PEAKS_SMALL = 0x0001D948;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_PEAKS_MEDIUM = 0x0001D949;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_PEAKS_LARGE = 0x0001D94A;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ROTATION_WALLPLANE_SINGLE = 0x0001D94B;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ROTATION_WALLPLANE_DOUBLE = 0x0001D94C;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ROTATION_WALLPLANE_ALTERNATING = 0x0001D94D;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ROTATION_FLOORPLANE_SINGLE = 0x0001D94E;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ROTATION_FLOORPLANE_DOUBLE = 0x0001D94F;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ROTATION_FLOORPLANE_ALTERNATING = 0x0001D950;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_SHAKING = 0x0001D951;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ARM_SPIRAL_SINGLE = 0x0001D952;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ARM_SPIRAL_DOUBLE = 0x0001D953;
const char32_t SIGNWRITING_TRAVEL_WALLPLANE_ARM_SPIRAL_TRIPLE = 0x0001D954;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_AWAY_SMALL = 0x0001D955;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_AWAY_MEDIUM = 0x0001D956;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_AWAY_LARGE = 0x0001D957;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_AWAY_LARGEST = 0x0001D958;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_TOWARDS_SMALL = 0x0001D959;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_TOWARDS_MEDIUM = 0x0001D95A;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_TOWARDS_LARGE = 0x0001D95B;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_TOWARDS_LARGEST = 0x0001D95C;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_AWAY_SMALL = 0x0001D95D;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_AWAY_MEDIUM = 0x0001D95E;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_AWAY_LARGE = 0x0001D95F;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_AWAY_LARGEST = 0x0001D960;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_TOWARDS_SMALL = 0x0001D961;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_TOWARDS_MEDIUM = 0x0001D962;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_TOWARDS_LARGE = 0x0001D963;
const char32_t SIGNWRITING_MOVEMENT_DIAGONAL_BETWEEN_TOWARDS_LARGEST = 0x0001D964;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_SINGLE_STRAIGHT_SMALL = 0x0001D965;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_SINGLE_STRAIGHT_MEDIUM = 0x0001D966;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_SINGLE_STRAIGHT_LARGE = 0x0001D967;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_SINGLE_STRAIGHT_LARGEST = 0x0001D968;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_SINGLE_WRIST_FLEX = 0x0001D969;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_DOUBLE_STRAIGHT = 0x0001D96A;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_DOUBLE_WRIST_FLEX = 0x0001D96B;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_DOUBLE_ALTERNATING = 0x0001D96C;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_DOUBLE_ALTERNATING_WRIST_FLEX = 0x0001D96D;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CROSS = 0x0001D96E;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_TRIPLE_STRAIGHT_MOVEMENT = 0x0001D96F;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_TRIPLE_WRIST_FLEX = 0x0001D970;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_TRIPLE_ALTERNATING_MOVEMENT = 0x0001D971;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_TRIPLE_ALTERNATING_WRIST_FLEX = 0x0001D972;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_BEND = 0x0001D973;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CORNER_SMALL = 0x0001D974;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CORNER_MEDIUM = 0x0001D975;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CORNER_LARGE = 0x0001D976;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CHECK = 0x0001D977;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_BOX_SMALL = 0x0001D978;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_BOX_MEDIUM = 0x0001D979;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_BOX_LARGE = 0x0001D97A;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ZIGZAG_SMALL = 0x0001D97B;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ZIGZAG_MEDIUM = 0x0001D97C;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ZIGZAG_LARGE = 0x0001D97D;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_PEAKS_SMALL = 0x0001D97E;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_PEAKS_MEDIUM = 0x0001D97F;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_PEAKS_LARGE = 0x0001D980;
const char32_t SIGNWRITING_TRAVEL_FLOORPLANE_ROTATION_FLOORPLANE_SINGLE = 0x0001D981;
const char32_t SIGNWRITING_TRAVEL_FLOORPLANE_ROTATION_FLOORPLANE_DOUBLE = 0x0001D982;
const char32_t SIGNWRITING_TRAVEL_FLOORPLANE_ROTATION_FLOORPLANE_ALTERNATING = 0x0001D983;
const char32_t SIGNWRITING_TRAVEL_FLOORPLANE_ROTATION_WALLPLANE_SINGLE = 0x0001D984;
const char32_t SIGNWRITING_TRAVEL_FLOORPLANE_ROTATION_WALLPLANE_DOUBLE = 0x0001D985;
const char32_t SIGNWRITING_TRAVEL_FLOORPLANE_ROTATION_WALLPLANE_ALTERNATING = 0x0001D986;
const char32_t SIGNWRITING_TRAVEL_FLOORPLANE_SHAKING = 0x0001D987;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_QUARTER_SMALL = 0x0001D988;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_QUARTER_MEDIUM = 0x0001D989;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_QUARTER_LARGE = 0x0001D98A;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_QUARTER_LARGEST = 0x0001D98B;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_HALF_CIRCLE_SMALL = 0x0001D98C;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_HALF_CIRCLE_MEDIUM = 0x0001D98D;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_HALF_CIRCLE_LARGE = 0x0001D98E;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_HALF_CIRCLE_LARGEST = 0x0001D98F;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_THREE_QUARTER_CIRCLE_SMALL = 0x0001D990;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_THREE_QUARTER_CIRCLE_MEDIUM = 0x0001D991;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_HUMP_SMALL = 0x0001D992;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_HUMP_MEDIUM = 0x0001D993;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_HUMP_LARGE = 0x0001D994;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_LOOP_SMALL = 0x0001D995;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_LOOP_MEDIUM = 0x0001D996;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_LOOP_LARGE = 0x0001D997;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_LOOP_SMALL_DOUBLE = 0x0001D998;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_CURVE_DOUBLE_SMALL = 0x0001D999;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_CURVE_DOUBLE_MEDIUM = 0x0001D99A;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_CURVE_DOUBLE_LARGE = 0x0001D99B;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_CURVE_TRIPLE_SMALL = 0x0001D99C;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_CURVE_TRIPLE_MEDIUM = 0x0001D99D;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_CURVE_TRIPLE_LARGE = 0x0001D99E;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_THEN_STRAIGHT = 0x0001D99F;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVED_CROSS_SMALL = 0x0001D9A0;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVED_CROSS_MEDIUM = 0x0001D9A1;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_SINGLE = 0x0001D9A2;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_DOUBLE = 0x0001D9A3;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_ALTERNATE = 0x0001D9A4;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_SHAKING = 0x0001D9A5;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_HITTING_FRONT_WALL = 0x0001D9A6;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_HUMP_HITTING_FRONT_WALL = 0x0001D9A7;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_LOOP_HITTING_FRONT_WALL = 0x0001D9A8;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_HITTING_FRONT_WALL = 0x0001D9A9;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_SINGLE_HITTING_FRONT_WALL = 0x0001D9AA;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_DOUBLE_HITTING_FRONT_WALL = 0x0001D9AB;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_ALTERNATING_HITTING_FRONT_WALL = 0x0001D9AC;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_CURVE_HITTING_CHEST = 0x0001D9AD;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_HUMP_HITTING_CHEST = 0x0001D9AE;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_LOOP_HITTING_CHEST = 0x0001D9AF;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_HITTING_CHEST = 0x0001D9B0;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_SINGLE_HITTING_CHEST = 0x0001D9B1;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_DOUBLE_HITTING_CHEST = 0x0001D9B2;
const char32_t SIGNWRITING_ROTATION_WALLPLANE_ALTERNATING_HITTING_CHEST = 0x0001D9B3;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_DIAGONAL_PATH_SMALL = 0x0001D9B4;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_DIAGONAL_PATH_MEDIUM = 0x0001D9B5;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WAVE_DIAGONAL_PATH_LARGE = 0x0001D9B6;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_HITTING_CEILING_SMALL = 0x0001D9B7;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_HITTING_CEILING_LARGE = 0x0001D9B8;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_CEILING_SMALL_DOUBLE = 0x0001D9B9;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_CEILING_LARGE_DOUBLE = 0x0001D9BA;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_CEILING_SMALL_TRIPLE = 0x0001D9BB;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_CEILING_LARGE_TRIPLE = 0x0001D9BC;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_CEILING_SMALL_SINGLE = 0x0001D9BD;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_CEILING_LARGE_SINGLE = 0x0001D9BE;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_CEILING_SMALL_DOUBLE = 0x0001D9BF;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_CEILING_LARGE_DOUBLE = 0x0001D9C0;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WAVE_HITTING_CEILING_SMALL = 0x0001D9C1;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WAVE_HITTING_CEILING_LARGE = 0x0001D9C2;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_SINGLE_HITTING_CEILING = 0x0001D9C3;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_DOUBLE_HITTING_CEILING = 0x0001D9C4;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_ALTERNATING_HITTING_CEILING = 0x0001D9C5;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_HITTING_FLOOR_SMALL = 0x0001D9C6;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_HITTING_FLOOR_LARGE = 0x0001D9C7;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_FLOOR_SMALL_DOUBLE = 0x0001D9C8;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_FLOOR_LARGE_DOUBLE = 0x0001D9C9;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_FLOOR_TRIPLE_SMALL_TRIPLE = 0x0001D9CA;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_HITTING_FLOOR_TRIPLE_LARGE_TRIPLE = 0x0001D9CB;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_FLOOR_SMALL_SINGLE = 0x0001D9CC;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_FLOOR_LARGE_SINGLE = 0x0001D9CD;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_FLOOR_SMALL_DOUBLE = 0x0001D9CE;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_HITTING_FLOOR_LARGE_DOUBLE = 0x0001D9CF;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WAVE_HITTING_FLOOR_SMALL = 0x0001D9D0;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WAVE_HITTING_FLOOR_LARGE = 0x0001D9D1;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_SINGLE_HITTING_FLOOR = 0x0001D9D2;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_DOUBLE_HITTING_FLOOR = 0x0001D9D3;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_ALTERNATING_HITTING_FLOOR = 0x0001D9D4;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_SMALL = 0x0001D9D5;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_MEDIUM = 0x0001D9D6;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_LARGE = 0x0001D9D7;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_LARGEST = 0x0001D9D8;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_CURVE_COMBINED = 0x0001D9D9;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_HUMP_SMALL = 0x0001D9DA;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_LOOP_SMALL = 0x0001D9DB;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WAVE_SNAKE = 0x0001D9DC;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WAVE_SMALL = 0x0001D9DD;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WAVE_LARGE = 0x0001D9DE;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_SINGLE = 0x0001D9DF;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_DOUBLE = 0x0001D9E0;
const char32_t SIGNWRITING_ROTATION_FLOORPLANE_ALTERNATING = 0x0001D9E1;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_SHAKING_PARALLEL = 0x0001D9E2;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_ARM_CIRCLE_SMALL_SINGLE = 0x0001D9E3;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_ARM_CIRCLE_MEDIUM_SINGLE = 0x0001D9E4;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_ARM_CIRCLE_SMALL_DOUBLE = 0x0001D9E5;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_ARM_CIRCLE_MEDIUM_DOUBLE = 0x0001D9E6;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ARM_CIRCLE_HITTING_WALL_SMALL_SINGLE = 0x0001D9E7;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ARM_CIRCLE_HITTING_WALL_MEDIUM_SINGLE = 0x0001D9E8;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ARM_CIRCLE_HITTING_WALL_LARGE_SINGLE = 0x0001D9E9;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ARM_CIRCLE_HITTING_WALL_SMALL_DOUBLE = 0x0001D9EA;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ARM_CIRCLE_HITTING_WALL_MEDIUM_DOUBLE = 0x0001D9EB;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_ARM_CIRCLE_HITTING_WALL_LARGE_DOUBLE = 0x0001D9EC;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WRIST_CIRCLE_FRONT_SINGLE = 0x0001D9ED;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_WRIST_CIRCLE_FRONT_DOUBLE = 0x0001D9EE;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WRIST_CIRCLE_HITTING_WALL_SINGLE = 0x0001D9EF;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_WRIST_CIRCLE_HITTING_WALL_DOUBLE = 0x0001D9F0;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_FINGER_CIRCLES_SINGLE = 0x0001D9F1;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_FINGER_CIRCLES_DOUBLE = 0x0001D9F2;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_FINGER_CIRCLES_HITTING_WALL_SINGLE = 0x0001D9F3;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_FINGER_CIRCLES_HITTING_WALL_DOUBLE = 0x0001D9F4;
const char32_t SIGNWRITING_DYNAMIC_ARROWHEAD_SMALL = 0x0001D9F5;
const char32_t SIGNWRITING_DYNAMIC_ARROWHEAD_LARGE = 0x0001D9F6;
const char32_t SIGNWRITING_DYNAMIC_FAST = 0x0001D9F7;
const char32_t SIGNWRITING_DYNAMIC_SLOW = 0x0001D9F8;
const char32_t SIGNWRITING_DYNAMIC_TENSE = 0x0001D9F9;
const char32_t SIGNWRITING_DYNAMIC_RELAXED = 0x0001D9FA;
const char32_t SIGNWRITING_DYNAMIC_SIMULTANEOUS = 0x0001D9FB;
const char32_t SIGNWRITING_DYNAMIC_SIMULTANEOUS_ALTERNATING = 0x0001D9FC;
const char32_t SIGNWRITING_DYNAMIC_EVERY_OTHER_TIME = 0x0001D9FD;
const char32_t SIGNWRITING_DYNAMIC_GRADUAL = 0x0001D9FE;
const char32_t SIGNWRITING_HEAD = 0x0001D9FF;
const char32_t SIGNWRITING_HEAD_RIM = 0x0001DA00;
const char32_t SIGNWRITING_HEAD_MOVEMENT_WALLPLANE_STRAIGHT = 0x0001DA01;
const char32_t SIGNWRITING_HEAD_MOVEMENT_WALLPLANE_TILT = 0x0001DA02;
const char32_t SIGNWRITING_HEAD_MOVEMENT_FLOORPLANE_STRAIGHT = 0x0001DA03;
const char32_t SIGNWRITING_HEAD_MOVEMENT_WALLPLANE_CURVE = 0x0001DA04;
const char32_t SIGNWRITING_HEAD_MOVEMENT_FLOORPLANE_CURVE = 0x0001DA05;
const char32_t SIGNWRITING_HEAD_MOVEMENT_CIRCLE = 0x0001DA06;
const char32_t SIGNWRITING_FACE_DIRECTION_POSITION_NOSE_FORWARD_TILTING = 0x0001DA07;
const char32_t SIGNWRITING_FACE_DIRECTION_POSITION_NOSE_UP_OR_DOWN = 0x0001DA08;
const char32_t SIGNWRITING_FACE_DIRECTION_POSITION_NOSE_UP_OR_DOWN_TILTING = 0x0001DA09;
const char32_t SIGNWRITING_EYEBROWS_STRAIGHT_UP = 0x0001DA0A;
const char32_t SIGNWRITING_EYEBROWS_STRAIGHT_NEUTRAL = 0x0001DA0B;
const char32_t SIGNWRITING_EYEBROWS_STRAIGHT_DOWN = 0x0001DA0C;
const char32_t SIGNWRITING_DREAMY_EYEBROWS_NEUTRAL_DOWN = 0x0001DA0D;
const char32_t SIGNWRITING_DREAMY_EYEBROWS_DOWN_NEUTRAL = 0x0001DA0E;
const char32_t SIGNWRITING_DREAMY_EYEBROWS_UP_NEUTRAL = 0x0001DA0F;
const char32_t SIGNWRITING_DREAMY_EYEBROWS_NEUTRAL_UP = 0x0001DA10;
const char32_t SIGNWRITING_FOREHEAD_NEUTRAL = 0x0001DA11;
const char32_t SIGNWRITING_FOREHEAD_CONTACT = 0x0001DA12;
const char32_t SIGNWRITING_FOREHEAD_WRINKLED = 0x0001DA13;
const char32_t SIGNWRITING_EYES_OPEN = 0x0001DA14;
const char32_t SIGNWRITING_EYES_SQUEEZED = 0x0001DA15;
const char32_t SIGNWRITING_EYES_CLOSED = 0x0001DA16;
const char32_t SIGNWRITING_EYE_BLINK_SINGLE = 0x0001DA17;
const char32_t SIGNWRITING_EYE_BLINK_MULTIPLE = 0x0001DA18;
const char32_t SIGNWRITING_EYES_HALF_OPEN = 0x0001DA19;
const char32_t SIGNWRITING_EYES_WIDE_OPEN = 0x0001DA1A;
const char32_t SIGNWRITING_EYES_HALF_CLOSED = 0x0001DA1B;
const char32_t SIGNWRITING_EYES_WIDENING_MOVEMENT = 0x0001DA1C;
const char32_t SIGNWRITING_EYE_WINK = 0x0001DA1D;
const char32_t SIGNWRITING_EYELASHES_UP = 0x0001DA1E;
const char32_t SIGNWRITING_EYELASHES_DOWN = 0x0001DA1F;
const char32_t SIGNWRITING_EYELASHES_FLUTTERING = 0x0001DA20;
const char32_t SIGNWRITING_EYEGAZE_WALLPLANE_STRAIGHT = 0x0001DA21;
const char32_t SIGNWRITING_EYEGAZE_WALLPLANE_STRAIGHT_DOUBLE = 0x0001DA22;
const char32_t SIGNWRITING_EYEGAZE_WALLPLANE_STRAIGHT_ALTERNATING = 0x0001DA23;
const char32_t SIGNWRITING_EYEGAZE_FLOORPLANE_STRAIGHT = 0x0001DA24;
const char32_t SIGNWRITING_EYEGAZE_FLOORPLANE_STRAIGHT_DOUBLE = 0x0001DA25;
const char32_t SIGNWRITING_EYEGAZE_FLOORPLANE_STRAIGHT_ALTERNATING = 0x0001DA26;
const char32_t SIGNWRITING_EYEGAZE_WALLPLANE_CURVED = 0x0001DA27;
const char32_t SIGNWRITING_EYEGAZE_FLOORPLANE_CURVED = 0x0001DA28;
const char32_t SIGNWRITING_EYEGAZE_WALLPLANE_CIRCLING = 0x0001DA29;
const char32_t SIGNWRITING_CHEEKS_PUFFED = 0x0001DA2A;
const char32_t SIGNWRITING_CHEEKS_NEUTRAL = 0x0001DA2B;
const char32_t SIGNWRITING_CHEEKS_SUCKED = 0x0001DA2C;
const char32_t SIGNWRITING_TENSE_CHEEKS_HIGH = 0x0001DA2D;
const char32_t SIGNWRITING_TENSE_CHEEKS_MIDDLE = 0x0001DA2E;
const char32_t SIGNWRITING_TENSE_CHEEKS_LOW = 0x0001DA2F;
const char32_t SIGNWRITING_EARS = 0x0001DA30;
const char32_t SIGNWRITING_NOSE_NEUTRAL = 0x0001DA31;
const char32_t SIGNWRITING_NOSE_CONTACT = 0x0001DA32;
const char32_t SIGNWRITING_NOSE_WRINKLES = 0x0001DA33;
const char32_t SIGNWRITING_NOSE_WIGGLES = 0x0001DA34;
const char32_t SIGNWRITING_AIR_BLOWING_OUT = 0x0001DA35;
const char32_t SIGNWRITING_AIR_SUCKING_IN = 0x0001DA36;
const char32_t SIGNWRITING_AIR_BLOW_SMALL_ROTATIONS = 0x0001DA37;
const char32_t SIGNWRITING_AIR_SUCK_SMALL_ROTATIONS = 0x0001DA38;
const char32_t SIGNWRITING_BREATH_INHALE = 0x0001DA39;
const char32_t SIGNWRITING_BREATH_EXHALE = 0x0001DA3A;
const char32_t SIGNWRITING_MOUTH_CLOSED_NEUTRAL = 0x0001DA3B;
const char32_t SIGNWRITING_MOUTH_CLOSED_FORWARD = 0x0001DA3C;
const char32_t SIGNWRITING_MOUTH_CLOSED_CONTACT = 0x0001DA3D;
const char32_t SIGNWRITING_MOUTH_SMILE = 0x0001DA3E;
const char32_t SIGNWRITING_MOUTH_SMILE_WRINKLED = 0x0001DA3F;
const char32_t SIGNWRITING_MOUTH_SMILE_OPEN = 0x0001DA40;
const char32_t SIGNWRITING_MOUTH_FROWN = 0x0001DA41;
const char32_t SIGNWRITING_MOUTH_FROWN_WRINKLED = 0x0001DA42;
const char32_t SIGNWRITING_MOUTH_FROWN_OPEN = 0x0001DA43;
const char32_t SIGNWRITING_MOUTH_OPEN_CIRCLE = 0x0001DA44;
const char32_t SIGNWRITING_MOUTH_OPEN_FORWARD = 0x0001DA45;
const char32_t SIGNWRITING_MOUTH_OPEN_WRINKLED = 0x0001DA46;
const char32_t SIGNWRITING_MOUTH_OPEN_OVAL = 0x0001DA47;
const char32_t SIGNWRITING_MOUTH_OPEN_OVAL_WRINKLED = 0x0001DA48;
const char32_t SIGNWRITING_MOUTH_OPEN_OVAL_YAWN = 0x0001DA49;
const char32_t SIGNWRITING_MOUTH_OPEN_RECTANGLE = 0x0001DA4A;
const char32_t SIGNWRITING_MOUTH_OPEN_RECTANGLE_WRINKLED = 0x0001DA4B;
const char32_t SIGNWRITING_MOUTH_OPEN_RECTANGLE_YAWN = 0x0001DA4C;
const char32_t SIGNWRITING_MOUTH_KISS = 0x0001DA4D;
const char32_t SIGNWRITING_MOUTH_KISS_FORWARD = 0x0001DA4E;
const char32_t SIGNWRITING_MOUTH_KISS_WRINKLED = 0x0001DA4F;
const char32_t SIGNWRITING_MOUTH_TENSE = 0x0001DA50;
const char32_t SIGNWRITING_MOUTH_TENSE_FORWARD = 0x0001DA51;
const char32_t SIGNWRITING_MOUTH_TENSE_SUCKED = 0x0001DA52;
const char32_t SIGNWRITING_LIPS_PRESSED_TOGETHER = 0x0001DA53;
const char32_t SIGNWRITING_LIP_LOWER_OVER_UPPER = 0x0001DA54;
const char32_t SIGNWRITING_LIP_UPPER_OVER_LOWER = 0x0001DA55;
const char32_t SIGNWRITING_MOUTH_CORNERS = 0x0001DA56;
const char32_t SIGNWRITING_MOUTH_WRINKLES_SINGLE = 0x0001DA57;
const char32_t SIGNWRITING_MOUTH_WRINKLES_DOUBLE = 0x0001DA58;
const char32_t SIGNWRITING_TONGUE_STICKING_OUT_FAR = 0x0001DA59;
const char32_t SIGNWRITING_TONGUE_LICKING_LIPS = 0x0001DA5A;
const char32_t SIGNWRITING_TONGUE_TIP_BETWEEN_LIPS = 0x0001DA5B;
const char32_t SIGNWRITING_TONGUE_TIP_TOUCHING_INSIDE_MOUTH = 0x0001DA5C;
const char32_t SIGNWRITING_TONGUE_INSIDE_MOUTH_RELAXED = 0x0001DA5D;
const char32_t SIGNWRITING_TONGUE_MOVES_AGAINST_CHEEK = 0x0001DA5E;
const char32_t SIGNWRITING_TONGUE_CENTRE_STICKING_OUT = 0x0001DA5F;
const char32_t SIGNWRITING_TONGUE_CENTRE_INSIDE_MOUTH = 0x0001DA60;
const char32_t SIGNWRITING_TEETH = 0x0001DA61;
const char32_t SIGNWRITING_TEETH_MOVEMENT = 0x0001DA62;
const char32_t SIGNWRITING_TEETH_ON_TONGUE = 0x0001DA63;
const char32_t SIGNWRITING_TEETH_ON_TONGUE_MOVEMENT = 0x0001DA64;
const char32_t SIGNWRITING_TEETH_ON_LIPS = 0x0001DA65;
const char32_t SIGNWRITING_TEETH_ON_LIPS_MOVEMENT = 0x0001DA66;
const char32_t SIGNWRITING_TEETH_BITE_LIPS = 0x0001DA67;
const char32_t SIGNWRITING_MOVEMENT_WALLPLANE_JAW = 0x0001DA68;
const char32_t SIGNWRITING_MOVEMENT_FLOORPLANE_JAW = 0x0001DA69;
const char32_t SIGNWRITING_NECK = 0x0001DA6A;
const char32_t SIGNWRITING_HAIR = 0x0001DA6B;
const char32_t SIGNWRITING_EXCITEMENT = 0x0001DA6C;
const char32_t SIGNWRITING_SHOULDER_HIP_SPINE = 0x0001DA6D;
const char32_t SIGNWRITING_SHOULDER_HIP_POSITIONS = 0x0001DA6E;
const char32_t SIGNWRITING_WALLPLANE_SHOULDER_HIP_MOVE = 0x0001DA6F;
const char32_t SIGNWRITING_FLOORPLANE_SHOULDER_HIP_MOVE = 0x0001DA70;
const char32_t SIGNWRITING_SHOULDER_TILTING_FROM_WAIST = 0x0001DA71;
const char32_t SIGNWRITING_TORSO_WALLPLANE_STRAIGHT_STRETCH = 0x0001DA72;
const char32_t SIGNWRITING_TORSO_WALLPLANE_CURVED_BEND = 0x0001DA73;
const char32_t SIGNWRITING_TORSO_FLOORPLANE_TWISTING = 0x0001DA74;
const char32_t SIGNWRITING_UPPER_BODY_TILTING_FROM_HIP_JOINTS = 0x0001DA75;
const char32_t SIGNWRITING_LIMB_COMBINATION = 0x0001DA76;
const char32_t SIGNWRITING_LIMB_LENGTH_1 = 0x0001DA77;
const char32_t SIGNWRITING_LIMB_LENGTH_2 = 0x0001DA78;
const char32_t SIGNWRITING_LIMB_LENGTH_3 = 0x0001DA79;
const char32_t SIGNWRITING_LIMB_LENGTH_4 = 0x0001DA7A;
const char32_t SIGNWRITING_LIMB_LENGTH_5 = 0x0001DA7B;
const char32_t SIGNWRITING_LIMB_LENGTH_6 = 0x0001DA7C;
const char32_t SIGNWRITING_LIMB_LENGTH_7 = 0x0001DA7D;
const char32_t SIGNWRITING_FINGER = 0x0001DA7E;
const char32_t SIGNWRITING_LOCATION_WALLPLANE_SPACE = 0x0001DA7F;
const char32_t SIGNWRITING_LOCATION_FLOORPLANE_SPACE = 0x0001DA80;
const char32_t SIGNWRITING_LOCATION_HEIGHT = 0x0001DA81;
const char32_t SIGNWRITING_LOCATION_WIDTH = 0x0001DA82;
const char32_t SIGNWRITING_LOCATION_DEPTH = 0x0001DA83;
const char32_t SIGNWRITING_LOCATION_HEAD_NECK = 0x0001DA84;
const char32_t SIGNWRITING_LOCATION_TORSO = 0x0001DA85;
const char32_t SIGNWRITING_LOCATION_LIMBS_DIGITS = 0x0001DA86;
const char32_t SIGNWRITING_COMMA = 0x0001DA87;
const char32_t SIGNWRITING_FULL_STOP = 0x0001DA88;
const char32_t SIGNWRITING_SEMICOLON = 0x0001DA89;
const char32_t SIGNWRITING_COLON = 0x0001DA8A;
const char32_t SIGNWRITING_PARENTHESIS = 0x0001DA8B;
const char32_t SIGNWRITING_FILL_MODIFIER_2 = 0x0001DA9B;
const char32_t SIGNWRITING_FILL_MODIFIER_3 = 0x0001DA9C;
const char32_t SIGNWRITING_FILL_MODIFIER_4 = 0x0001DA9D;
const char32_t SIGNWRITING_FILL_MODIFIER_5 = 0x0001DA9E;
const char32_t SIGNWRITING_FILL_MODIFIER_6 = 0x0001DA9F;
const char32_t SIGNWRITING_ROTATION_MODIFIER_2 = 0x0001DAA1;
const char32_t SIGNWRITING_ROTATION_MODIFIER_3 = 0x0001DAA2;
const char32_t SIGNWRITING_ROTATION_MODIFIER_4 = 0x0001DAA3;
const char32_t SIGNWRITING_ROTATION_MODIFIER_5 = 0x0001DAA4;
const char32_t SIGNWRITING_ROTATION_MODIFIER_6 = 0x0001DAA5;
const char32_t SIGNWRITING_ROTATION_MODIFIER_7 = 0x0001DAA6;
const char32_t SIGNWRITING_ROTATION_MODIFIER_8 = 0x0001DAA7;
const char32_t SIGNWRITING_ROTATION_MODIFIER_9 = 0x0001DAA8;
const char32_t SIGNWRITING_ROTATION_MODIFIER_10 = 0x0001DAA9;
const char32_t SIGNWRITING_ROTATION_MODIFIER_11 = 0x0001DAAA;
const char32_t SIGNWRITING_ROTATION_MODIFIER_12 = 0x0001DAAB;
const char32_t SIGNWRITING_ROTATION_MODIFIER_13 = 0x0001DAAC;
const char32_t SIGNWRITING_ROTATION_MODIFIER_14 = 0x0001DAAD;
const char32_t SIGNWRITING_ROTATION_MODIFIER_15 = 0x0001DAAE;
const char32_t SIGNWRITING_ROTATION_MODIFIER_16 = 0x0001DAAF;
const char32_t LATIN_SMALL_LETTER_FENG_DIGRAPH_WITH_TRILL = 0x0001DF00;
const char32_t LATIN_SMALL_LETTER_REVERSED_SCRIPT_G = 0x0001DF01;
const char32_t LATIN_LETTER_SMALL_CAPITAL_TURNED_G = 0x0001DF02;
const char32_t LATIN_SMALL_LETTER_REVERSED_K = 0x0001DF03;
const char32_t LATIN_LETTER_SMALL_CAPITAL_L_WITH_BELT = 0x0001DF04;
const char32_t LATIN_SMALL_LETTER_LEZH_WITH_RETROFLEX_HOOK = 0x0001DF05;
const char32_t LATIN_SMALL_LETTER_TURNED_Y_WITH_BELT = 0x0001DF06;
const char32_t LATIN_SMALL_LETTER_REVERSED_ENG = 0x0001DF07;
const char32_t LATIN_SMALL_LETTER_TURNED_R_WITH_LONG_LEG_AND_RETROFLEX_HOOK = 0x0001DF08;
const char32_t LATIN_SMALL_LETTER_T_WITH_HOOK_AND_RETROFLEX_HOOK = 0x0001DF09;
const char32_t LATIN_LETTER_RETROFLEX_CLICK_WITH_RETROFLEX_HOOK = 0x0001DF0A;
const char32_t LATIN_SMALL_LETTER_ESH_WITH_DOUBLE_BAR = 0x0001DF0B;
const char32_t LATIN_SMALL_LETTER_ESH_WITH_DOUBLE_BAR_AND_CURL = 0x0001DF0C;
const char32_t LATIN_SMALL_LETTER_TURNED_T_WITH_CURL = 0x0001DF0D;
const char32_t LATIN_LETTER_INVERTED_GLOTTAL_STOP_WITH_CURL = 0x0001DF0E;
const char32_t LATIN_LETTER_STRETCHED_C_WITH_CURL = 0x0001DF0F;
const char32_t LATIN_LETTER_SMALL_CAPITAL_TURNED_K = 0x0001DF10;
const char32_t LATIN_SMALL_LETTER_L_WITH_FISHHOOK = 0x0001DF11;
const char32_t LATIN_SMALL_LETTER_DEZH_DIGRAPH_WITH_PALATAL_HOOK = 0x0001DF12;
const char32_t LATIN_SMALL_LETTER_L_WITH_BELT_AND_PALATAL_HOOK = 0x0001DF13;
const char32_t LATIN_SMALL_LETTER_ENG_WITH_PALATAL_HOOK = 0x0001DF14;
const char32_t LATIN_SMALL_LETTER_TURNED_R_WITH_PALATAL_HOOK = 0x0001DF15;
const char32_t LATIN_SMALL_LETTER_R_WITH_FISHHOOK_AND_PALATAL_HOOK = 0x0001DF16;
const char32_t LATIN_SMALL_LETTER_TESH_DIGRAPH_WITH_PALATAL_HOOK = 0x0001DF17;
const char32_t LATIN_SMALL_LETTER_EZH_WITH_PALATAL_HOOK = 0x0001DF18;
const char32_t LATIN_SMALL_LETTER_DEZH_DIGRAPH_WITH_RETROFLEX_HOOK = 0x0001DF19;
const char32_t LATIN_SMALL_LETTER_I_WITH_STROKE_AND_RETROFLEX_HOOK = 0x0001DF1A;
const char32_t LATIN_SMALL_LETTER_O_WITH_RETROFLEX_HOOK = 0x0001DF1B;
const char32_t LATIN_SMALL_LETTER_TESH_DIGRAPH_WITH_RETROFLEX_HOOK = 0x0001DF1C;
const char32_t LATIN_SMALL_LETTER_C_WITH_RETROFLEX_HOOK = 0x0001DF1D;
const char32_t LATIN_SMALL_LETTER_S_WITH_CURL = 0x0001DF1E;
const char32_t LATIN_SMALL_LETTER_D_WITH_MID_HEIGHT_LEFT_HOOK = 0x0001DF25;
const char32_t LATIN_SMALL_LETTER_L_WITH_MID_HEIGHT_LEFT_HOOK = 0x0001DF26;
const char32_t LATIN_SMALL_LETTER_N_WITH_MID_HEIGHT_LEFT_HOOK = 0x0001DF27;
const char32_t LATIN_SMALL_LETTER_R_WITH_MID_HEIGHT_LEFT_HOOK = 0x0001DF28;
const char32_t LATIN_SMALL_LETTER_S_WITH_MID_HEIGHT_LEFT_HOOK = 0x0001DF29;
const char32_t LATIN_SMALL_LETTER_T_WITH_MID_HEIGHT_LEFT_HOOK = 0x0001DF2A;
const char32_t COMBINING_GLAGOLITIC_LETTER_AZU = 0x0001E000;
const char32_t COMBINING_GLAGOLITIC_LETTER_BUKY = 0x0001E001;
const char32_t COMBINING_GLAGOLITIC_LETTER_VEDE = 0x0001E002;
const char32_t COMBINING_GLAGOLITIC_LETTER_GLAGOLI = 0x0001E003;
const char32_t COMBINING_GLAGOLITIC_LETTER_DOBRO = 0x0001E004;
const char32_t COMBINING_GLAGOLITIC_LETTER_YESTU = 0x0001E005;
const char32_t COMBINING_GLAGOLITIC_LETTER_ZHIVETE = 0x0001E006;
const char32_t COMBINING_GLAGOLITIC_LETTER_ZEMLJA = 0x0001E008;
const char32_t COMBINING_GLAGOLITIC_LETTER_IZHE = 0x0001E009;
const char32_t COMBINING_GLAGOLITIC_LETTER_INITIAL_IZHE = 0x0001E00A;
const char32_t COMBINING_GLAGOLITIC_LETTER_I = 0x0001E00B;
const char32_t COMBINING_GLAGOLITIC_LETTER_DJERVI = 0x0001E00C;
const char32_t COMBINING_GLAGOLITIC_LETTER_KAKO = 0x0001E00D;
const char32_t COMBINING_GLAGOLITIC_LETTER_LJUDIJE = 0x0001E00E;
const char32_t COMBINING_GLAGOLITIC_LETTER_MYSLITE = 0x0001E00F;
const char32_t COMBINING_GLAGOLITIC_LETTER_NASHI = 0x0001E010;
const char32_t COMBINING_GLAGOLITIC_LETTER_ONU = 0x0001E011;
const char32_t COMBINING_GLAGOLITIC_LETTER_POKOJI = 0x0001E012;
const char32_t COMBINING_GLAGOLITIC_LETTER_RITSI = 0x0001E013;
const char32_t COMBINING_GLAGOLITIC_LETTER_SLOVO = 0x0001E014;
const char32_t COMBINING_GLAGOLITIC_LETTER_TVRIDO = 0x0001E015;
const char32_t COMBINING_GLAGOLITIC_LETTER_UKU = 0x0001E016;
const char32_t COMBINING_GLAGOLITIC_LETTER_FRITU = 0x0001E017;
const char32_t COMBINING_GLAGOLITIC_LETTER_HERU = 0x0001E018;
const char32_t COMBINING_GLAGOLITIC_LETTER_SHTA = 0x0001E01B;
const char32_t COMBINING_GLAGOLITIC_LETTER_TSI = 0x0001E01C;
const char32_t COMBINING_GLAGOLITIC_LETTER_CHRIVI = 0x0001E01D;
const char32_t COMBINING_GLAGOLITIC_LETTER_SHA = 0x0001E01E;
const char32_t COMBINING_GLAGOLITIC_LETTER_YERU = 0x0001E01F;
const char32_t COMBINING_GLAGOLITIC_LETTER_YERI = 0x0001E020;
const char32_t COMBINING_GLAGOLITIC_LETTER_YATI = 0x0001E021;
const char32_t COMBINING_GLAGOLITIC_LETTER_YU = 0x0001E023;
const char32_t COMBINING_GLAGOLITIC_LETTER_SMALL_YUS = 0x0001E024;
const char32_t COMBINING_GLAGOLITIC_LETTER_YO = 0x0001E026;
const char32_t COMBINING_GLAGOLITIC_LETTER_IOTATED_SMALL_YUS = 0x0001E027;
const char32_t COMBINING_GLAGOLITIC_LETTER_BIG_YUS = 0x0001E028;
const char32_t COMBINING_GLAGOLITIC_LETTER_IOTATED_BIG_YUS = 0x0001E029;
const char32_t COMBINING_GLAGOLITIC_LETTER_FITA = 0x0001E02A;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_A = 0x0001E030;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_BE = 0x0001E031;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_VE = 0x0001E032;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_GHE = 0x0001E033;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_DE = 0x0001E034;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_IE = 0x0001E035;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_ZHE = 0x0001E036;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_ZE = 0x0001E037;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_I = 0x0001E038;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_KA = 0x0001E039;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_EL = 0x0001E03A;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_EM = 0x0001E03B;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_O = 0x0001E03C;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_PE = 0x0001E03D;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_ER = 0x0001E03E;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_ES = 0x0001E03F;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_TE = 0x0001E040;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_U = 0x0001E041;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_EF = 0x0001E042;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_HA = 0x0001E043;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_TSE = 0x0001E044;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_CHE = 0x0001E045;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_SHA = 0x0001E046;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_YERU = 0x0001E047;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_E = 0x0001E048;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_YU = 0x0001E049;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_DZZE = 0x0001E04A;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_SCHWA = 0x0001E04B;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_BYELORUSSIAN_UKRAINIAN_I = 0x0001E04C;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_JE = 0x0001E04D;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_BARRED_O = 0x0001E04E;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_STRAIGHT_U = 0x0001E04F;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_PALOCHKA = 0x0001E050;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_A = 0x0001E051;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_BE = 0x0001E052;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_VE = 0x0001E053;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_GHE = 0x0001E054;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_DE = 0x0001E055;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_IE = 0x0001E056;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_ZHE = 0x0001E057;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_ZE = 0x0001E058;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_I = 0x0001E059;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_KA = 0x0001E05A;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_EL = 0x0001E05B;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_O = 0x0001E05C;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_PE = 0x0001E05D;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_ES = 0x0001E05E;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_U = 0x0001E05F;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_EF = 0x0001E060;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_HA = 0x0001E061;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_TSE = 0x0001E062;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_CHE = 0x0001E063;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_SHA = 0x0001E064;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_HARD_SIGN = 0x0001E065;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_YERU = 0x0001E066;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_GHE_WITH_UPTURN = 0x0001E067;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_BYELORUSSIAN_UKRAINIAN_I = 0x0001E068;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_DZE = 0x0001E069;
const char32_t CYRILLIC_SUBSCRIPT_SMALL_LETTER_DZHE = 0x0001E06A;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_ES_WITH_DESCENDER = 0x0001E06B;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_YERU_WITH_BACK_YER = 0x0001E06C;
const char32_t MODIFIER_LETTER_CYRILLIC_SMALL_STRAIGHT_U_WITH_STROKE = 0x0001E06D;
const char32_t COMBINING_CYRILLIC_SMALL_LETTER_BYELORUSSIAN_UKRAINIAN_I = 0x0001E08F;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_MA = 0x0001E100;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_TSA = 0x0001E101;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NTA = 0x0001E102;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_TA = 0x0001E103;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_HA = 0x0001E104;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NA = 0x0001E105;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_XA = 0x0001E106;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NKA = 0x0001E107;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_CA = 0x0001E108;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_LA = 0x0001E109;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_SA = 0x0001E10A;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_ZA = 0x0001E10B;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NCA = 0x0001E10C;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NTSA = 0x0001E10D;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_KA = 0x0001E10E;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_DA = 0x0001E10F;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NYA = 0x0001E110;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NRA = 0x0001E111;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_VA = 0x0001E112;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NTXA = 0x0001E113;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_TXA = 0x0001E114;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_FA = 0x0001E115;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_RA = 0x0001E116;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_QA = 0x0001E117;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_YA = 0x0001E118;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NQA = 0x0001E119;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_PA = 0x0001E11A;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_XYA = 0x0001E11B;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NPA = 0x0001E11C;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_DLA = 0x0001E11D;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_NPLA = 0x0001E11E;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_HAH = 0x0001E11F;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_MLA = 0x0001E120;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_PLA = 0x0001E121;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_GA = 0x0001E122;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_RRA = 0x0001E123;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_A = 0x0001E124;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_AA = 0x0001E125;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_I = 0x0001E126;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_U = 0x0001E127;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_O = 0x0001E128;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_OO = 0x0001E129;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_E = 0x0001E12A;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_EE = 0x0001E12B;
const char32_t NYIAKENG_PUACHUE_HMONG_LETTER_W = 0x0001E12C;
const char32_t NYIAKENG_PUACHUE_HMONG_TONE_B = 0x0001E130;
const char32_t NYIAKENG_PUACHUE_HMONG_TONE_M = 0x0001E131;
const char32_t NYIAKENG_PUACHUE_HMONG_TONE_J = 0x0001E132;
const char32_t NYIAKENG_PUACHUE_HMONG_TONE_V = 0x0001E133;
const char32_t NYIAKENG_PUACHUE_HMONG_TONE_S = 0x0001E134;
const char32_t NYIAKENG_PUACHUE_HMONG_TONE_G = 0x0001E135;
const char32_t NYIAKENG_PUACHUE_HMONG_TONE_D = 0x0001E136;
const char32_t NYIAKENG_PUACHUE_HMONG_SIGN_FOR_PERSON = 0x0001E137;
const char32_t NYIAKENG_PUACHUE_HMONG_SIGN_FOR_THING = 0x0001E138;
const char32_t NYIAKENG_PUACHUE_HMONG_SIGN_FOR_LOCATION = 0x0001E139;
const char32_t NYIAKENG_PUACHUE_HMONG_SIGN_FOR_ANIMAL = 0x0001E13A;
const char32_t NYIAKENG_PUACHUE_HMONG_SIGN_FOR_INVERTEBRATE = 0x0001E13B;
const char32_t NYIAKENG_PUACHUE_HMONG_SIGN_XW_XW = 0x0001E13C;
const char32_t NYIAKENG_PUACHUE_HMONG_SYLLABLE_LENGTHENER = 0x0001E13D;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_ZERO = 0x0001E140;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_ONE = 0x0001E141;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_TWO = 0x0001E142;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_THREE = 0x0001E143;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_FOUR = 0x0001E144;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_FIVE = 0x0001E145;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_SIX = 0x0001E146;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_SEVEN = 0x0001E147;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_EIGHT = 0x0001E148;
const char32_t NYIAKENG_PUACHUE_HMONG_DIGIT_NINE = 0x0001E149;
const char32_t NYIAKENG_PUACHUE_HMONG_LOGOGRAM_NYAJ = 0x0001E14E;
const char32_t NYIAKENG_PUACHUE_HMONG_CIRCLED_CA = 0x0001E14F;
const char32_t TOTO_LETTER_PA = 0x0001E290;
const char32_t TOTO_LETTER_BA = 0x0001E291;
const char32_t TOTO_LETTER_TA = 0x0001E292;
const char32_t TOTO_LETTER_DA = 0x0001E293;
const char32_t TOTO_LETTER_KA = 0x0001E294;
const char32_t TOTO_LETTER_GA = 0x0001E295;
const char32_t TOTO_LETTER_MA = 0x0001E296;
const char32_t TOTO_LETTER_NA = 0x0001E297;
const char32_t TOTO_LETTER_NGA = 0x0001E298;
const char32_t TOTO_LETTER_SA = 0x0001E299;
const char32_t TOTO_LETTER_CHA = 0x0001E29A;
const char32_t TOTO_LETTER_YA = 0x0001E29B;
const char32_t TOTO_LETTER_WA = 0x0001E29C;
const char32_t TOTO_LETTER_JA = 0x0001E29D;
const char32_t TOTO_LETTER_HA = 0x0001E29E;
const char32_t TOTO_LETTER_RA = 0x0001E29F;
const char32_t TOTO_LETTER_LA = 0x0001E2A0;
const char32_t TOTO_LETTER_I = 0x0001E2A1;
const char32_t TOTO_LETTER_BREATHY_I = 0x0001E2A2;
const char32_t TOTO_LETTER_IU = 0x0001E2A3;
const char32_t TOTO_LETTER_BREATHY_IU = 0x0001E2A4;
const char32_t TOTO_LETTER_U = 0x0001E2A5;
const char32_t TOTO_LETTER_E = 0x0001E2A6;
const char32_t TOTO_LETTER_BREATHY_E = 0x0001E2A7;
const char32_t TOTO_LETTER_EO = 0x0001E2A8;
const char32_t TOTO_LETTER_BREATHY_EO = 0x0001E2A9;
const char32_t TOTO_LETTER_O = 0x0001E2AA;
const char32_t TOTO_LETTER_AE = 0x0001E2AB;
const char32_t TOTO_LETTER_BREATHY_AE = 0x0001E2AC;
const char32_t TOTO_LETTER_A = 0x0001E2AD;
const char32_t TOTO_SIGN_RISING_TONE = 0x0001E2AE;
const char32_t WANCHO_LETTER_AA = 0x0001E2C0;
const char32_t WANCHO_LETTER_A = 0x0001E2C1;
const char32_t WANCHO_LETTER_BA = 0x0001E2C2;
const char32_t WANCHO_LETTER_CA = 0x0001E2C3;
const char32_t WANCHO_LETTER_DA = 0x0001E2C4;
const char32_t WANCHO_LETTER_GA = 0x0001E2C5;
const char32_t WANCHO_LETTER_YA = 0x0001E2C6;
const char32_t WANCHO_LETTER_PHA = 0x0001E2C7;
const char32_t WANCHO_LETTER_LA = 0x0001E2C8;
const char32_t WANCHO_LETTER_NA = 0x0001E2C9;
const char32_t WANCHO_LETTER_PA = 0x0001E2CA;
const char32_t WANCHO_LETTER_TA = 0x0001E2CB;
const char32_t WANCHO_LETTER_THA = 0x0001E2CC;
const char32_t WANCHO_LETTER_FA = 0x0001E2CD;
const char32_t WANCHO_LETTER_SA = 0x0001E2CE;
const char32_t WANCHO_LETTER_SHA = 0x0001E2CF;
const char32_t WANCHO_LETTER_JA = 0x0001E2D0;
const char32_t WANCHO_LETTER_ZA = 0x0001E2D1;
const char32_t WANCHO_LETTER_WA = 0x0001E2D2;
const char32_t WANCHO_LETTER_VA = 0x0001E2D3;
const char32_t WANCHO_LETTER_KA = 0x0001E2D4;
const char32_t WANCHO_LETTER_O = 0x0001E2D5;
const char32_t WANCHO_LETTER_AU = 0x0001E2D6;
const char32_t WANCHO_LETTER_RA = 0x0001E2D7;
const char32_t WANCHO_LETTER_MA = 0x0001E2D8;
const char32_t WANCHO_LETTER_KHA = 0x0001E2D9;
const char32_t WANCHO_LETTER_HA = 0x0001E2DA;
const char32_t WANCHO_LETTER_E = 0x0001E2DB;
const char32_t WANCHO_LETTER_I = 0x0001E2DC;
const char32_t WANCHO_LETTER_NGA = 0x0001E2DD;
const char32_t WANCHO_LETTER_U = 0x0001E2DE;
const char32_t WANCHO_LETTER_LLHA = 0x0001E2DF;
const char32_t WANCHO_LETTER_TSA = 0x0001E2E0;
const char32_t WANCHO_LETTER_TRA = 0x0001E2E1;
const char32_t WANCHO_LETTER_ONG = 0x0001E2E2;
const char32_t WANCHO_LETTER_AANG = 0x0001E2E3;
const char32_t WANCHO_LETTER_ANG = 0x0001E2E4;
const char32_t WANCHO_LETTER_ING = 0x0001E2E5;
const char32_t WANCHO_LETTER_ON = 0x0001E2E6;
const char32_t WANCHO_LETTER_EN = 0x0001E2E7;
const char32_t WANCHO_LETTER_AAN = 0x0001E2E8;
const char32_t WANCHO_LETTER_NYA = 0x0001E2E9;
const char32_t WANCHO_LETTER_UEN = 0x0001E2EA;
const char32_t WANCHO_LETTER_YIH = 0x0001E2EB;
const char32_t WANCHO_TONE_TUP = 0x0001E2EC;
const char32_t WANCHO_TONE_TUPNI = 0x0001E2ED;
const char32_t WANCHO_TONE_KOI = 0x0001E2EE;
const char32_t WANCHO_TONE_KOINI = 0x0001E2EF;
const char32_t WANCHO_DIGIT_ZERO = 0x0001E2F0;
const char32_t WANCHO_DIGIT_ONE = 0x0001E2F1;
const char32_t WANCHO_DIGIT_TWO = 0x0001E2F2;
const char32_t WANCHO_DIGIT_THREE = 0x0001E2F3;
const char32_t WANCHO_DIGIT_FOUR = 0x0001E2F4;
const char32_t WANCHO_DIGIT_FIVE = 0x0001E2F5;
const char32_t WANCHO_DIGIT_SIX = 0x0001E2F6;
const char32_t WANCHO_DIGIT_SEVEN = 0x0001E2F7;
const char32_t WANCHO_DIGIT_EIGHT = 0x0001E2F8;
const char32_t WANCHO_DIGIT_NINE = 0x0001E2F9;
const char32_t WANCHO_NGUN_SIGN = 0x0001E2FF;
const char32_t NAG_MUNDARI_LETTER_O = 0x0001E4D0;
const char32_t NAG_MUNDARI_LETTER_OP = 0x0001E4D1;
const char32_t NAG_MUNDARI_LETTER_OL = 0x0001E4D2;
const char32_t NAG_MUNDARI_LETTER_OY = 0x0001E4D3;
const char32_t NAG_MUNDARI_LETTER_ONG = 0x0001E4D4;
const char32_t NAG_MUNDARI_LETTER_A = 0x0001E4D5;
const char32_t NAG_MUNDARI_LETTER_AJ = 0x0001E4D6;
const char32_t NAG_MUNDARI_LETTER_AB = 0x0001E4D7;
const char32_t NAG_MUNDARI_LETTER_ANY = 0x0001E4D8;
const char32_t NAG_MUNDARI_LETTER_AH = 0x0001E4D9;
const char32_t NAG_MUNDARI_LETTER_I = 0x0001E4DA;
const char32_t NAG_MUNDARI_LETTER_IS = 0x0001E4DB;
const char32_t NAG_MUNDARI_LETTER_IDD = 0x0001E4DC;
const char32_t NAG_MUNDARI_LETTER_IT = 0x0001E4DD;
const char32_t NAG_MUNDARI_LETTER_IH = 0x0001E4DE;
const char32_t NAG_MUNDARI_LETTER_U = 0x0001E4DF;
const char32_t NAG_MUNDARI_LETTER_UC = 0x0001E4E0;
const char32_t NAG_MUNDARI_LETTER_UD = 0x0001E4E1;
const char32_t NAG_MUNDARI_LETTER_UK = 0x0001E4E2;
const char32_t NAG_MUNDARI_LETTER_UR = 0x0001E4E3;
const char32_t NAG_MUNDARI_LETTER_E = 0x0001E4E4;
const char32_t NAG_MUNDARI_LETTER_ENN = 0x0001E4E5;
const char32_t NAG_MUNDARI_LETTER_EG = 0x0001E4E6;
const char32_t NAG_MUNDARI_LETTER_EM = 0x0001E4E7;
const char32_t NAG_MUNDARI_LETTER_EN = 0x0001E4E8;
const char32_t NAG_MUNDARI_LETTER_ETT = 0x0001E4E9;
const char32_t NAG_MUNDARI_LETTER_ELL = 0x0001E4EA;
const char32_t NAG_MUNDARI_SIGN_OJOD = 0x0001E4EB;
const char32_t NAG_MUNDARI_SIGN_MUHOR = 0x0001E4EC;
const char32_t NAG_MUNDARI_SIGN_TOYOR = 0x0001E4ED;
const char32_t NAG_MUNDARI_SIGN_IKIR = 0x0001E4EE;
const char32_t NAG_MUNDARI_SIGN_SUTUH = 0x0001E4EF;
const char32_t NAG_MUNDARI_DIGIT_ZERO = 0x0001E4F0;
const char32_t NAG_MUNDARI_DIGIT_ONE = 0x0001E4F1;
const char32_t NAG_MUNDARI_DIGIT_TWO = 0x0001E4F2;
const char32_t NAG_MUNDARI_DIGIT_THREE = 0x0001E4F3;
const char32_t NAG_MUNDARI_DIGIT_FOUR = 0x0001E4F4;
const char32_t NAG_MUNDARI_DIGIT_FIVE = 0x0001E4F5;
const char32_t NAG_MUNDARI_DIGIT_SIX = 0x0001E4F6;
const char32_t NAG_MUNDARI_DIGIT_SEVEN = 0x0001E4F7;
const char32_t NAG_MUNDARI_DIGIT_EIGHT = 0x0001E4F8;
const char32_t NAG_MUNDARI_DIGIT_NINE = 0x0001E4F9;
const char32_t OL_ONAL_LETTER_O = 0x0001E5D0;
const char32_t OL_ONAL_LETTER_OM = 0x0001E5D1;
const char32_t OL_ONAL_LETTER_ONG = 0x0001E5D2;
const char32_t OL_ONAL_LETTER_ORR = 0x0001E5D3;
const char32_t OL_ONAL_LETTER_OO = 0x0001E5D4;
const char32_t OL_ONAL_LETTER_OY = 0x0001E5D5;
const char32_t OL_ONAL_LETTER_A = 0x0001E5D6;
const char32_t OL_ONAL_LETTER_AD = 0x0001E5D7;
const char32_t OL_ONAL_LETTER_AB = 0x0001E5D8;
const char32_t OL_ONAL_LETTER_AH = 0x0001E5D9;
const char32_t OL_ONAL_LETTER_AL = 0x0001E5DA;
const char32_t OL_ONAL_LETTER_AW = 0x0001E5DB;
const char32_t OL_ONAL_LETTER_I = 0x0001E5DC;
const char32_t OL_ONAL_LETTER_IT = 0x0001E5DD;
const char32_t OL_ONAL_LETTER_IP = 0x0001E5DE;
const char32_t OL_ONAL_LETTER_ITT = 0x0001E5DF;
const char32_t OL_ONAL_LETTER_ID = 0x0001E5E0;
const char32_t OL_ONAL_LETTER_IN = 0x0001E5E1;
const char32_t OL_ONAL_LETTER_U = 0x0001E5E2;
const char32_t OL_ONAL_LETTER_UK = 0x0001E5E3;
const char32_t OL_ONAL_LETTER_UDD = 0x0001E5E4;
const char32_t OL_ONAL_LETTER_UJ = 0x0001E5E5;
const char32_t OL_ONAL_LETTER_UNY = 0x0001E5E6;
const char32_t OL_ONAL_LETTER_UR = 0x0001E5E7;
const char32_t OL_ONAL_LETTER_E = 0x0001E5E8;
const char32_t OL_ONAL_LETTER_ES = 0x0001E5E9;
const char32_t OL_ONAL_LETTER_EH = 0x0001E5EA;
const char32_t OL_ONAL_LETTER_EC = 0x0001E5EB;
const char32_t OL_ONAL_LETTER_ENN = 0x0001E5EC;
const char32_t OL_ONAL_LETTER_EG = 0x0001E5ED;
const char32_t OL_ONAL_SIGN_MU = 0x0001E5EE;
const char32_t OL_ONAL_SIGN_IKIR = 0x0001E5EF;
const char32_t OL_ONAL_SIGN_HODDOND = 0x0001E5F0;
const char32_t OL_ONAL_DIGIT_ZERO = 0x0001E5F1;
const char32_t OL_ONAL_DIGIT_ONE = 0x0001E5F2;
const char32_t OL_ONAL_DIGIT_TWO = 0x0001E5F3;
const char32_t OL_ONAL_DIGIT_THREE = 0x0001E5F4;
const char32_t OL_ONAL_DIGIT_FOUR = 0x0001E5F5;
const char32_t OL_ONAL_DIGIT_FIVE = 0x0001E5F6;
const char32_t OL_ONAL_DIGIT_SIX = 0x0001E5F7;
const char32_t OL_ONAL_DIGIT_SEVEN = 0x0001E5F8;
const char32_t OL_ONAL_DIGIT_EIGHT = 0x0001E5F9;
const char32_t OL_ONAL_DIGIT_NINE = 0x0001E5FA;
const char32_t OL_ONAL_ABBREVIATION_SIGN = 0x0001E5FF;
const char32_t TAI_YO_LETTER_LOW_KO = 0x0001E6C0;
const char32_t TAI_YO_LETTER_HIGH_KO = 0x0001E6C1;
const char32_t TAI_YO_LETTER_LOW_KHO = 0x0001E6C2;
const char32_t TAI_YO_LETTER_HIGH_KHO = 0x0001E6C3;
const char32_t TAI_YO_LETTER_GO = 0x0001E6C4;
const char32_t TAI_YO_LETTER_NGO = 0x0001E6C5;
const char32_t TAI_YO_LETTER_CO = 0x0001E6C6;
const char32_t TAI_YO_LETTER_LOW_XO = 0x0001E6C7;
const char32_t TAI_YO_LETTER_HIGH_XO = 0x0001E6C8;
const char32_t TAI_YO_LETTER_LOW_NYO = 0x0001E6C9;
const char32_t TAI_YO_LETTER_HIGH_NYO = 0x0001E6CA;
const char32_t TAI_YO_LETTER_DO = 0x0001E6CB;
const char32_t TAI_YO_LETTER_LOW_TO = 0x0001E6CC;
const char32_t TAI_YO_LETTER_HIGH_TO = 0x0001E6CD;
const char32_t TAI_YO_LETTER_THO = 0x0001E6CE;
const char32_t TAI_YO_LETTER_NO = 0x0001E6CF;
const char32_t TAI_YO_LETTER_BO = 0x0001E6D0;
const char32_t TAI_YO_LETTER_LOW_PO = 0x0001E6D1;
const char32_t TAI_YO_LETTER_HIGH_PO = 0x0001E6D2;
const char32_t TAI_YO_LETTER_PHO = 0x0001E6D3;
const char32_t TAI_YO_LETTER_LOW_FO = 0x0001E6D4;
const char32_t TAI_YO_LETTER_HIGH_FO = 0x0001E6D5;
const char32_t TAI_YO_LETTER_MO = 0x0001E6D6;
const char32_t TAI_YO_LETTER_YO = 0x0001E6D7;
const char32_t TAI_YO_LETTER_LO = 0x0001E6D8;
const char32_t TAI_YO_LETTER_VO = 0x0001E6D9;
const char32_t TAI_YO_LETTER_LOW_HO = 0x0001E6DA;
const char32_t TAI_YO_LETTER_HIGH_HO = 0x0001E6DB;
const char32_t TAI_YO_LETTER_QO = 0x0001E6DC;
const char32_t TAI_YO_LETTER_LOW_KVO = 0x0001E6DD;
const char32_t TAI_YO_LETTER_HIGH_KVO = 0x0001E6DE;
const char32_t TAI_YO_LETTER_AA = 0x0001E6E0;
const char32_t TAI_YO_LETTER_I = 0x0001E6E1;
const char32_t TAI_YO_LETTER_UE = 0x0001E6E2;
const char32_t TAI_YO_SIGN_UE = 0x0001E6E3;
const char32_t TAI_YO_LETTER_U = 0x0001E6E4;
const char32_t TAI_YO_LETTER_AE = 0x0001E6E5;
const char32_t TAI_YO_SIGN_AU = 0x0001E6E6;
const char32_t TAI_YO_LETTER_O = 0x0001E6E7;
const char32_t TAI_YO_LETTER_E = 0x0001E6E8;
const char32_t TAI_YO_LETTER_IA = 0x0001E6E9;
const char32_t TAI_YO_LETTER_UEA = 0x0001E6EA;
const char32_t TAI_YO_LETTER_UA = 0x0001E6EB;
const char32_t TAI_YO_LETTER_OO = 0x0001E6EC;
const char32_t TAI_YO_LETTER_AUE = 0x0001E6ED;
const char32_t TAI_YO_SIGN_AY = 0x0001E6EE;
const char32_t TAI_YO_SIGN_ANG = 0x0001E6EF;
const char32_t TAI_YO_LETTER_AN = 0x0001E6F0;
const char32_t TAI_YO_LETTER_AM = 0x0001E6F1;
const char32_t TAI_YO_LETTER_AK = 0x0001E6F2;
const char32_t TAI_YO_LETTER_AT = 0x0001E6F3;
const char32_t TAI_YO_LETTER_AP = 0x0001E6F4;
const char32_t TAI_YO_SIGN_OM = 0x0001E6F5;
const char32_t TAI_YO_SYMBOL_MUEANG = 0x0001E6FE;
const char32_t TAI_YO_XAM_LAI = 0x0001E6FF;
const char32_t ETHIOPIC_SYLLABLE_HHYA = 0x0001E7E0;
const char32_t ETHIOPIC_SYLLABLE_HHYU = 0x0001E7E1;
const char32_t ETHIOPIC_SYLLABLE_HHYI = 0x0001E7E2;
const char32_t ETHIOPIC_SYLLABLE_HHYAA = 0x0001E7E3;
const char32_t ETHIOPIC_SYLLABLE_HHYEE = 0x0001E7E4;
const char32_t ETHIOPIC_SYLLABLE_HHYE = 0x0001E7E5;
const char32_t ETHIOPIC_SYLLABLE_HHYO = 0x0001E7E6;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_HHWA = 0x0001E7E8;
const char32_t ETHIOPIC_SYLLABLE_HHWI = 0x0001E7E9;
const char32_t ETHIOPIC_SYLLABLE_HHWEE = 0x0001E7EA;
const char32_t ETHIOPIC_SYLLABLE_HHWE = 0x0001E7EB;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_MWI = 0x0001E7ED;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_MWEE = 0x0001E7EE;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_QWI = 0x0001E7F0;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_QWEE = 0x0001E7F1;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_QWE = 0x0001E7F2;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_BWI = 0x0001E7F3;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_BWEE = 0x0001E7F4;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_KWI = 0x0001E7F5;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_KWEE = 0x0001E7F6;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_KWE = 0x0001E7F7;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_GWI = 0x0001E7F8;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_GWEE = 0x0001E7F9;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_GWE = 0x0001E7FA;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_FWI = 0x0001E7FB;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_FWEE = 0x0001E7FC;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_PWI = 0x0001E7FD;
const char32_t ETHIOPIC_SYLLABLE_GURAGE_PWEE = 0x0001E7FE;
const char32_t MENDE_KIKAKUI_SYLLABLE_M001_KI = 0x0001E800;
const char32_t MENDE_KIKAKUI_SYLLABLE_M002_KA = 0x0001E801;
const char32_t MENDE_KIKAKUI_SYLLABLE_M003_KU = 0x0001E802;
const char32_t MENDE_KIKAKUI_SYLLABLE_M065_KEE = 0x0001E803;
const char32_t MENDE_KIKAKUI_SYLLABLE_M095_KE = 0x0001E804;
const char32_t MENDE_KIKAKUI_SYLLABLE_M076_KOO = 0x0001E805;
const char32_t MENDE_KIKAKUI_SYLLABLE_M048_KO = 0x0001E806;
const char32_t MENDE_KIKAKUI_SYLLABLE_M179_KUA = 0x0001E807;
const char32_t MENDE_KIKAKUI_SYLLABLE_M004_WI = 0x0001E808;
const char32_t MENDE_KIKAKUI_SYLLABLE_M005_WA = 0x0001E809;
const char32_t MENDE_KIKAKUI_SYLLABLE_M006_WU = 0x0001E80A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M126_WEE = 0x0001E80B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M118_WE = 0x0001E80C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M114_WOO = 0x0001E80D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M045_WO = 0x0001E80E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M194_WUI = 0x0001E80F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M143_WEI = 0x0001E810;
const char32_t MENDE_KIKAKUI_SYLLABLE_M061_WVI = 0x0001E811;
const char32_t MENDE_KIKAKUI_SYLLABLE_M049_WVA = 0x0001E812;
const char32_t MENDE_KIKAKUI_SYLLABLE_M139_WVE = 0x0001E813;
const char32_t MENDE_KIKAKUI_SYLLABLE_M007_MIN = 0x0001E814;
const char32_t MENDE_KIKAKUI_SYLLABLE_M008_MAN = 0x0001E815;
const char32_t MENDE_KIKAKUI_SYLLABLE_M009_MUN = 0x0001E816;
const char32_t MENDE_KIKAKUI_SYLLABLE_M059_MEN = 0x0001E817;
const char32_t MENDE_KIKAKUI_SYLLABLE_M094_MON = 0x0001E818;
const char32_t MENDE_KIKAKUI_SYLLABLE_M154_MUAN = 0x0001E819;
const char32_t MENDE_KIKAKUI_SYLLABLE_M189_MUEN = 0x0001E81A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M010_BI = 0x0001E81B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M011_BA = 0x0001E81C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M012_BU = 0x0001E81D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M150_BEE = 0x0001E81E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M097_BE = 0x0001E81F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M103_BOO = 0x0001E820;
const char32_t MENDE_KIKAKUI_SYLLABLE_M138_BO = 0x0001E821;
const char32_t MENDE_KIKAKUI_SYLLABLE_M013_I = 0x0001E822;
const char32_t MENDE_KIKAKUI_SYLLABLE_M014_A = 0x0001E823;
const char32_t MENDE_KIKAKUI_SYLLABLE_M015_U = 0x0001E824;
const char32_t MENDE_KIKAKUI_SYLLABLE_M163_EE = 0x0001E825;
const char32_t MENDE_KIKAKUI_SYLLABLE_M100_E = 0x0001E826;
const char32_t MENDE_KIKAKUI_SYLLABLE_M165_OO = 0x0001E827;
const char32_t MENDE_KIKAKUI_SYLLABLE_M147_O = 0x0001E828;
const char32_t MENDE_KIKAKUI_SYLLABLE_M137_EI = 0x0001E829;
const char32_t MENDE_KIKAKUI_SYLLABLE_M131_IN = 0x0001E82A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M135_IN = 0x0001E82B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M195_AN = 0x0001E82C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M178_EN = 0x0001E82D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M019_SI = 0x0001E82E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M020_SA = 0x0001E82F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M021_SU = 0x0001E830;
const char32_t MENDE_KIKAKUI_SYLLABLE_M162_SEE = 0x0001E831;
const char32_t MENDE_KIKAKUI_SYLLABLE_M116_SE = 0x0001E832;
const char32_t MENDE_KIKAKUI_SYLLABLE_M136_SOO = 0x0001E833;
const char32_t MENDE_KIKAKUI_SYLLABLE_M079_SO = 0x0001E834;
const char32_t MENDE_KIKAKUI_SYLLABLE_M196_SIA = 0x0001E835;
const char32_t MENDE_KIKAKUI_SYLLABLE_M025_LI = 0x0001E836;
const char32_t MENDE_KIKAKUI_SYLLABLE_M026_LA = 0x0001E837;
const char32_t MENDE_KIKAKUI_SYLLABLE_M027_LU = 0x0001E838;
const char32_t MENDE_KIKAKUI_SYLLABLE_M084_LEE = 0x0001E839;
const char32_t MENDE_KIKAKUI_SYLLABLE_M073_LE = 0x0001E83A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M054_LOO = 0x0001E83B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M153_LO = 0x0001E83C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M110_LONG_LE = 0x0001E83D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M016_DI = 0x0001E83E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M017_DA = 0x0001E83F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M018_DU = 0x0001E840;
const char32_t MENDE_KIKAKUI_SYLLABLE_M089_DEE = 0x0001E841;
const char32_t MENDE_KIKAKUI_SYLLABLE_M180_DOO = 0x0001E842;
const char32_t MENDE_KIKAKUI_SYLLABLE_M181_DO = 0x0001E843;
const char32_t MENDE_KIKAKUI_SYLLABLE_M022_TI = 0x0001E844;
const char32_t MENDE_KIKAKUI_SYLLABLE_M023_TA = 0x0001E845;
const char32_t MENDE_KIKAKUI_SYLLABLE_M024_TU = 0x0001E846;
const char32_t MENDE_KIKAKUI_SYLLABLE_M091_TEE = 0x0001E847;
const char32_t MENDE_KIKAKUI_SYLLABLE_M055_TE = 0x0001E848;
const char32_t MENDE_KIKAKUI_SYLLABLE_M104_TOO = 0x0001E849;
const char32_t MENDE_KIKAKUI_SYLLABLE_M069_TO = 0x0001E84A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M028_JI = 0x0001E84B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M029_JA = 0x0001E84C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M030_JU = 0x0001E84D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M157_JEE = 0x0001E84E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M113_JE = 0x0001E84F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M160_JOO = 0x0001E850;
const char32_t MENDE_KIKAKUI_SYLLABLE_M063_JO = 0x0001E851;
const char32_t MENDE_KIKAKUI_SYLLABLE_M175_LONG_JO = 0x0001E852;
const char32_t MENDE_KIKAKUI_SYLLABLE_M031_YI = 0x0001E853;
const char32_t MENDE_KIKAKUI_SYLLABLE_M032_YA = 0x0001E854;
const char32_t MENDE_KIKAKUI_SYLLABLE_M033_YU = 0x0001E855;
const char32_t MENDE_KIKAKUI_SYLLABLE_M109_YEE = 0x0001E856;
const char32_t MENDE_KIKAKUI_SYLLABLE_M080_YE = 0x0001E857;
const char32_t MENDE_KIKAKUI_SYLLABLE_M141_YOO = 0x0001E858;
const char32_t MENDE_KIKAKUI_SYLLABLE_M121_YO = 0x0001E859;
const char32_t MENDE_KIKAKUI_SYLLABLE_M034_FI = 0x0001E85A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M035_FA = 0x0001E85B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M036_FU = 0x0001E85C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M078_FEE = 0x0001E85D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M075_FE = 0x0001E85E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M133_FOO = 0x0001E85F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M088_FO = 0x0001E860;
const char32_t MENDE_KIKAKUI_SYLLABLE_M197_FUA = 0x0001E861;
const char32_t MENDE_KIKAKUI_SYLLABLE_M101_FAN = 0x0001E862;
const char32_t MENDE_KIKAKUI_SYLLABLE_M037_NIN = 0x0001E863;
const char32_t MENDE_KIKAKUI_SYLLABLE_M038_NAN = 0x0001E864;
const char32_t MENDE_KIKAKUI_SYLLABLE_M039_NUN = 0x0001E865;
const char32_t MENDE_KIKAKUI_SYLLABLE_M117_NEN = 0x0001E866;
const char32_t MENDE_KIKAKUI_SYLLABLE_M169_NON = 0x0001E867;
const char32_t MENDE_KIKAKUI_SYLLABLE_M176_HI = 0x0001E868;
const char32_t MENDE_KIKAKUI_SYLLABLE_M041_HA = 0x0001E869;
const char32_t MENDE_KIKAKUI_SYLLABLE_M186_HU = 0x0001E86A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M040_HEE = 0x0001E86B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M096_HE = 0x0001E86C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M042_HOO = 0x0001E86D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M140_HO = 0x0001E86E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M083_HEEI = 0x0001E86F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M128_HOOU = 0x0001E870;
const char32_t MENDE_KIKAKUI_SYLLABLE_M053_HIN = 0x0001E871;
const char32_t MENDE_KIKAKUI_SYLLABLE_M130_HAN = 0x0001E872;
const char32_t MENDE_KIKAKUI_SYLLABLE_M087_HUN = 0x0001E873;
const char32_t MENDE_KIKAKUI_SYLLABLE_M052_HEN = 0x0001E874;
const char32_t MENDE_KIKAKUI_SYLLABLE_M193_HON = 0x0001E875;
const char32_t MENDE_KIKAKUI_SYLLABLE_M046_HUAN = 0x0001E876;
const char32_t MENDE_KIKAKUI_SYLLABLE_M090_NGGI = 0x0001E877;
const char32_t MENDE_KIKAKUI_SYLLABLE_M043_NGGA = 0x0001E878;
const char32_t MENDE_KIKAKUI_SYLLABLE_M082_NGGU = 0x0001E879;
const char32_t MENDE_KIKAKUI_SYLLABLE_M115_NGGEE = 0x0001E87A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M146_NGGE = 0x0001E87B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M156_NGGOO = 0x0001E87C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M120_NGGO = 0x0001E87D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M159_NGGAA = 0x0001E87E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M127_NGGUA = 0x0001E87F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M086_LONG_NGGE = 0x0001E880;
const char32_t MENDE_KIKAKUI_SYLLABLE_M106_LONG_NGGOO = 0x0001E881;
const char32_t MENDE_KIKAKUI_SYLLABLE_M183_LONG_NGGO = 0x0001E882;
const char32_t MENDE_KIKAKUI_SYLLABLE_M155_GI = 0x0001E883;
const char32_t MENDE_KIKAKUI_SYLLABLE_M111_GA = 0x0001E884;
const char32_t MENDE_KIKAKUI_SYLLABLE_M168_GU = 0x0001E885;
const char32_t MENDE_KIKAKUI_SYLLABLE_M190_GEE = 0x0001E886;
const char32_t MENDE_KIKAKUI_SYLLABLE_M166_GUEI = 0x0001E887;
const char32_t MENDE_KIKAKUI_SYLLABLE_M167_GUAN = 0x0001E888;
const char32_t MENDE_KIKAKUI_SYLLABLE_M184_NGEN = 0x0001E889;
const char32_t MENDE_KIKAKUI_SYLLABLE_M057_NGON = 0x0001E88A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M177_NGUAN = 0x0001E88B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M068_PI = 0x0001E88C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M099_PA = 0x0001E88D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M050_PU = 0x0001E88E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M081_PEE = 0x0001E88F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M051_PE = 0x0001E890;
const char32_t MENDE_KIKAKUI_SYLLABLE_M102_POO = 0x0001E891;
const char32_t MENDE_KIKAKUI_SYLLABLE_M066_PO = 0x0001E892;
const char32_t MENDE_KIKAKUI_SYLLABLE_M145_MBI = 0x0001E893;
const char32_t MENDE_KIKAKUI_SYLLABLE_M062_MBA = 0x0001E894;
const char32_t MENDE_KIKAKUI_SYLLABLE_M122_MBU = 0x0001E895;
const char32_t MENDE_KIKAKUI_SYLLABLE_M047_MBEE = 0x0001E896;
const char32_t MENDE_KIKAKUI_SYLLABLE_M188_MBEE = 0x0001E897;
const char32_t MENDE_KIKAKUI_SYLLABLE_M072_MBE = 0x0001E898;
const char32_t MENDE_KIKAKUI_SYLLABLE_M172_MBOO = 0x0001E899;
const char32_t MENDE_KIKAKUI_SYLLABLE_M174_MBO = 0x0001E89A;
const char32_t MENDE_KIKAKUI_SYLLABLE_M187_MBUU = 0x0001E89B;
const char32_t MENDE_KIKAKUI_SYLLABLE_M161_LONG_MBE = 0x0001E89C;
const char32_t MENDE_KIKAKUI_SYLLABLE_M105_LONG_MBOO = 0x0001E89D;
const char32_t MENDE_KIKAKUI_SYLLABLE_M142_LONG_MBO = 0x0001E89E;
const char32_t MENDE_KIKAKUI_SYLLABLE_M132_KPI = 0x0001E89F;
const char32_t MENDE_KIKAKUI_SYLLABLE_M092_KPA = 0x0001E8A0;
const char32_t MENDE_KIKAKUI_SYLLABLE_M074_KPU = 0x0001E8A1;
const char32_t MENDE_KIKAKUI_SYLLABLE_M044_KPEE = 0x0001E8A2;
const char32_t MENDE_KIKAKUI_SYLLABLE_M108_KPE = 0x0001E8A3;
const char32_t MENDE_KIKAKUI_SYLLABLE_M112_KPOO = 0x0001E8A4;
const char32_t MENDE_KIKAKUI_SYLLABLE_M158_KPO = 0x0001E8A5;
const char32_t MENDE_KIKAKUI_SYLLABLE_M124_GBI = 0x0001E8A6;
const char32_t MENDE_KIKAKUI_SYLLABLE_M056_GBA = 0x0001E8A7;
const char32_t MENDE_KIKAKUI_SYLLABLE_M148_GBU = 0x0001E8A8;
const char32_t MENDE_KIKAKUI_SYLLABLE_M093_GBEE = 0x0001E8A9;
const char32_t MENDE_KIKAKUI_SYLLABLE_M107_GBE = 0x0001E8AA;
const char32_t MENDE_KIKAKUI_SYLLABLE_M071_GBOO = 0x0001E8AB;
const char32_t MENDE_KIKAKUI_SYLLABLE_M070_GBO = 0x0001E8AC;
const char32_t MENDE_KIKAKUI_SYLLABLE_M171_RA = 0x0001E8AD;
const char32_t MENDE_KIKAKUI_SYLLABLE_M123_NDI = 0x0001E8AE;
const char32_t MENDE_KIKAKUI_SYLLABLE_M129_NDA = 0x0001E8AF;
const char32_t MENDE_KIKAKUI_SYLLABLE_M125_NDU = 0x0001E8B0;
const char32_t MENDE_KIKAKUI_SYLLABLE_M191_NDEE = 0x0001E8B1;
const char32_t MENDE_KIKAKUI_SYLLABLE_M119_NDE = 0x0001E8B2;
const char32_t MENDE_KIKAKUI_SYLLABLE_M067_NDOO = 0x0001E8B3;
const char32_t MENDE_KIKAKUI_SYLLABLE_M064_NDO = 0x0001E8B4;
const char32_t MENDE_KIKAKUI_SYLLABLE_M152_NJA = 0x0001E8B5;
const char32_t MENDE_KIKAKUI_SYLLABLE_M192_NJU = 0x0001E8B6;
const char32_t MENDE_KIKAKUI_SYLLABLE_M149_NJEE = 0x0001E8B7;
const char32_t MENDE_KIKAKUI_SYLLABLE_M134_NJOO = 0x0001E8B8;
const char32_t MENDE_KIKAKUI_SYLLABLE_M182_VI = 0x0001E8B9;
const char32_t MENDE_KIKAKUI_SYLLABLE_M185_VA = 0x0001E8BA;
const char32_t MENDE_KIKAKUI_SYLLABLE_M151_VU = 0x0001E8BB;
const char32_t MENDE_KIKAKUI_SYLLABLE_M173_VEE = 0x0001E8BC;
const char32_t MENDE_KIKAKUI_SYLLABLE_M085_VE = 0x0001E8BD;
const char32_t MENDE_KIKAKUI_SYLLABLE_M144_VOO = 0x0001E8BE;
const char32_t MENDE_KIKAKUI_SYLLABLE_M077_VO = 0x0001E8BF;
const char32_t MENDE_KIKAKUI_SYLLABLE_M164_NYIN = 0x0001E8C0;
const char32_t MENDE_KIKAKUI_SYLLABLE_M058_NYAN = 0x0001E8C1;
const char32_t MENDE_KIKAKUI_SYLLABLE_M170_NYUN = 0x0001E8C2;
const char32_t MENDE_KIKAKUI_SYLLABLE_M098_NYEN = 0x0001E8C3;
const char32_t MENDE_KIKAKUI_SYLLABLE_M060_NYON = 0x0001E8C4;
const char32_t MENDE_KIKAKUI_DIGIT_ONE = 0x0001E8C7;
const char32_t MENDE_KIKAKUI_DIGIT_TWO = 0x0001E8C8;
const char32_t MENDE_KIKAKUI_DIGIT_THREE = 0x0001E8C9;
const char32_t MENDE_KIKAKUI_DIGIT_FOUR = 0x0001E8CA;
const char32_t MENDE_KIKAKUI_DIGIT_FIVE = 0x0001E8CB;
const char32_t MENDE_KIKAKUI_DIGIT_SIX = 0x0001E8CC;
const char32_t MENDE_KIKAKUI_DIGIT_SEVEN = 0x0001E8CD;
const char32_t MENDE_KIKAKUI_DIGIT_EIGHT = 0x0001E8CE;
const char32_t MENDE_KIKAKUI_DIGIT_NINE = 0x0001E8CF;
const char32_t MENDE_KIKAKUI_COMBINING_NUMBER_TEENS = 0x0001E8D0;
const char32_t MENDE_KIKAKUI_COMBINING_NUMBER_TENS = 0x0001E8D1;
const char32_t MENDE_KIKAKUI_COMBINING_NUMBER_HUNDREDS = 0x0001E8D2;
const char32_t MENDE_KIKAKUI_COMBINING_NUMBER_THOUSANDS = 0x0001E8D3;
const char32_t MENDE_KIKAKUI_COMBINING_NUMBER_TEN_THOUSANDS = 0x0001E8D4;
const char32_t MENDE_KIKAKUI_COMBINING_NUMBER_HUNDRED_THOUSANDS = 0x0001E8D5;
const char32_t MENDE_KIKAKUI_COMBINING_NUMBER_MILLIONS = 0x0001E8D6;
const char32_t ADLAM_CAPITAL_LETTER_ALIF = 0x0001E900;
const char32_t ADLAM_CAPITAL_LETTER_DAALI = 0x0001E901;
const char32_t ADLAM_CAPITAL_LETTER_LAAM = 0x0001E902;
const char32_t ADLAM_CAPITAL_LETTER_MIIM = 0x0001E903;
const char32_t ADLAM_CAPITAL_LETTER_BA = 0x0001E904;
const char32_t ADLAM_CAPITAL_LETTER_SINNYIIYHE = 0x0001E905;
const char32_t ADLAM_CAPITAL_LETTER_PE = 0x0001E906;
const char32_t ADLAM_CAPITAL_LETTER_BHE = 0x0001E907;
const char32_t ADLAM_CAPITAL_LETTER_RA = 0x0001E908;
const char32_t ADLAM_CAPITAL_LETTER_E = 0x0001E909;
const char32_t ADLAM_CAPITAL_LETTER_FA = 0x0001E90A;
const char32_t ADLAM_CAPITAL_LETTER_I = 0x0001E90B;
const char32_t ADLAM_CAPITAL_LETTER_O = 0x0001E90C;
const char32_t ADLAM_CAPITAL_LETTER_DHA = 0x0001E90D;
const char32_t ADLAM_CAPITAL_LETTER_YHE = 0x0001E90E;
const char32_t ADLAM_CAPITAL_LETTER_WAW = 0x0001E90F;
const char32_t ADLAM_CAPITAL_LETTER_NUN = 0x0001E910;
const char32_t ADLAM_CAPITAL_LETTER_KAF = 0x0001E911;
const char32_t ADLAM_CAPITAL_LETTER_YA = 0x0001E912;
const char32_t ADLAM_CAPITAL_LETTER_U = 0x0001E913;
const char32_t ADLAM_CAPITAL_LETTER_JIIM = 0x0001E914;
const char32_t ADLAM_CAPITAL_LETTER_CHI = 0x0001E915;
const char32_t ADLAM_CAPITAL_LETTER_HA = 0x0001E916;
const char32_t ADLAM_CAPITAL_LETTER_QAAF = 0x0001E917;
const char32_t ADLAM_CAPITAL_LETTER_GA = 0x0001E918;
const char32_t ADLAM_CAPITAL_LETTER_NYA = 0x0001E919;
const char32_t ADLAM_CAPITAL_LETTER_TU = 0x0001E91A;
const char32_t ADLAM_CAPITAL_LETTER_NHA = 0x0001E91B;
const char32_t ADLAM_CAPITAL_LETTER_VA = 0x0001E91C;
const char32_t ADLAM_CAPITAL_LETTER_KHA = 0x0001E91D;
const char32_t ADLAM_CAPITAL_LETTER_GBE = 0x0001E91E;
const char32_t ADLAM_CAPITAL_LETTER_ZAL = 0x0001E91F;
const char32_t ADLAM_CAPITAL_LETTER_KPO = 0x0001E920;
const char32_t ADLAM_CAPITAL_LETTER_SHA = 0x0001E921;
const char32_t ADLAM_SMALL_LETTER_ALIF = 0x0001E922;
const char32_t ADLAM_SMALL_LETTER_DAALI = 0x0001E923;
const char32_t ADLAM_SMALL_LETTER_LAAM = 0x0001E924;
const char32_t ADLAM_SMALL_LETTER_MIIM = 0x0001E925;
const char32_t ADLAM_SMALL_LETTER_BA = 0x0001E926;
const char32_t ADLAM_SMALL_LETTER_SINNYIIYHE = 0x0001E927;
const char32_t ADLAM_SMALL_LETTER_PE = 0x0001E928;
const char32_t ADLAM_SMALL_LETTER_BHE = 0x0001E929;
const char32_t ADLAM_SMALL_LETTER_RA = 0x0001E92A;
const char32_t ADLAM_SMALL_LETTER_E = 0x0001E92B;
const char32_t ADLAM_SMALL_LETTER_FA = 0x0001E92C;
const char32_t ADLAM_SMALL_LETTER_I = 0x0001E92D;
const char32_t ADLAM_SMALL_LETTER_O = 0x0001E92E;
const char32_t ADLAM_SMALL_LETTER_DHA = 0x0001E92F;
const char32_t ADLAM_SMALL_LETTER_YHE = 0x0001E930;
const char32_t ADLAM_SMALL_LETTER_WAW = 0x0001E931;
const char32_t ADLAM_SMALL_LETTER_NUN = 0x0001E932;
const char32_t ADLAM_SMALL_LETTER_KAF = 0x0001E933;
const char32_t ADLAM_SMALL_LETTER_YA = 0x0001E934;
const char32_t ADLAM_SMALL_LETTER_U = 0x0001E935;
const char32_t ADLAM_SMALL_LETTER_JIIM = 0x0001E936;
const char32_t ADLAM_SMALL_LETTER_CHI = 0x0001E937;
const char32_t ADLAM_SMALL_LETTER_HA = 0x0001E938;
const char32_t ADLAM_SMALL_LETTER_QAAF = 0x0001E939;
const char32_t ADLAM_SMALL_LETTER_GA = 0x0001E93A;
const char32_t ADLAM_SMALL_LETTER_NYA = 0x0001E93B;
const char32_t ADLAM_SMALL_LETTER_TU = 0x0001E93C;
const char32_t ADLAM_SMALL_LETTER_NHA = 0x0001E93D;
const char32_t ADLAM_SMALL_LETTER_VA = 0x0001E93E;
const char32_t ADLAM_SMALL_LETTER_KHA = 0x0001E93F;
const char32_t ADLAM_SMALL_LETTER_GBE = 0x0001E940;
const char32_t ADLAM_SMALL_LETTER_ZAL = 0x0001E941;
const char32_t ADLAM_SMALL_LETTER_KPO = 0x0001E942;
const char32_t ADLAM_SMALL_LETTER_SHA = 0x0001E943;
const char32_t ADLAM_ALIF_LENGTHENER = 0x0001E944;
const char32_t ADLAM_VOWEL_LENGTHENER = 0x0001E945;
const char32_t ADLAM_GEMINATION_MARK = 0x0001E946;
const char32_t ADLAM_HAMZA = 0x0001E947;
const char32_t ADLAM_CONSONANT_MODIFIER = 0x0001E948;
const char32_t ADLAM_GEMINATE_CONSONANT_MODIFIER = 0x0001E949;
const char32_t ADLAM_NUKTA = 0x0001E94A;
const char32_t ADLAM_NASALIZATION_MARK = 0x0001E94B;
const char32_t ADLAM_DIGIT_ZERO = 0x0001E950;
const char32_t ADLAM_DIGIT_ONE = 0x0001E951;
const char32_t ADLAM_DIGIT_TWO = 0x0001E952;
const char32_t ADLAM_DIGIT_THREE = 0x0001E953;
const char32_t ADLAM_DIGIT_FOUR = 0x0001E954;
const char32_t ADLAM_DIGIT_FIVE = 0x0001E955;
const char32_t ADLAM_DIGIT_SIX = 0x0001E956;
const char32_t ADLAM_DIGIT_SEVEN = 0x0001E957;
const char32_t ADLAM_DIGIT_EIGHT = 0x0001E958;
const char32_t ADLAM_DIGIT_NINE = 0x0001E959;
const char32_t ADLAM_INITIAL_EXCLAMATION_MARK = 0x0001E95E;
const char32_t ADLAM_INITIAL_QUESTION_MARK = 0x0001E95F;
const char32_t INDIC_SIYAQ_NUMBER_ONE = 0x0001EC71;
const char32_t INDIC_SIYAQ_NUMBER_TWO = 0x0001EC72;
const char32_t INDIC_SIYAQ_NUMBER_THREE = 0x0001EC73;
const char32_t INDIC_SIYAQ_NUMBER_FOUR = 0x0001EC74;
const char32_t INDIC_SIYAQ_NUMBER_FIVE = 0x0001EC75;
const char32_t INDIC_SIYAQ_NUMBER_SIX = 0x0001EC76;
const char32_t INDIC_SIYAQ_NUMBER_SEVEN = 0x0001EC77;
const char32_t INDIC_SIYAQ_NUMBER_EIGHT = 0x0001EC78;
const char32_t INDIC_SIYAQ_NUMBER_NINE = 0x0001EC79;
const char32_t INDIC_SIYAQ_NUMBER_TEN = 0x0001EC7A;
const char32_t INDIC_SIYAQ_NUMBER_TWENTY = 0x0001EC7B;
const char32_t INDIC_SIYAQ_NUMBER_THIRTY = 0x0001EC7C;
const char32_t INDIC_SIYAQ_NUMBER_FORTY = 0x0001EC7D;
const char32_t INDIC_SIYAQ_NUMBER_FIFTY = 0x0001EC7E;
const char32_t INDIC_SIYAQ_NUMBER_SIXTY = 0x0001EC7F;
const char32_t INDIC_SIYAQ_NUMBER_SEVENTY = 0x0001EC80;
const char32_t INDIC_SIYAQ_NUMBER_EIGHTY = 0x0001EC81;
const char32_t INDIC_SIYAQ_NUMBER_NINETY = 0x0001EC82;
const char32_t INDIC_SIYAQ_NUMBER_ONE_HUNDRED = 0x0001EC83;
const char32_t INDIC_SIYAQ_NUMBER_TWO_HUNDRED = 0x0001EC84;
const char32_t INDIC_SIYAQ_NUMBER_THREE_HUNDRED = 0x0001EC85;
const char32_t INDIC_SIYAQ_NUMBER_FOUR_HUNDRED = 0x0001EC86;
const char32_t INDIC_SIYAQ_NUMBER_FIVE_HUNDRED = 0x0001EC87;
const char32_t INDIC_SIYAQ_NUMBER_SIX_HUNDRED = 0x0001EC88;
const char32_t INDIC_SIYAQ_NUMBER_SEVEN_HUNDRED = 0x0001EC89;
const char32_t INDIC_SIYAQ_NUMBER_EIGHT_HUNDRED = 0x0001EC8A;
const char32_t INDIC_SIYAQ_NUMBER_NINE_HUNDRED = 0x0001EC8B;
const char32_t INDIC_SIYAQ_NUMBER_ONE_THOUSAND = 0x0001EC8C;
const char32_t INDIC_SIYAQ_NUMBER_TWO_THOUSAND = 0x0001EC8D;
const char32_t INDIC_SIYAQ_NUMBER_THREE_THOUSAND = 0x0001EC8E;
const char32_t INDIC_SIYAQ_NUMBER_FOUR_THOUSAND = 0x0001EC8F;
const char32_t INDIC_SIYAQ_NUMBER_FIVE_THOUSAND = 0x0001EC90;
const char32_t INDIC_SIYAQ_NUMBER_SIX_THOUSAND = 0x0001EC91;
const char32_t INDIC_SIYAQ_NUMBER_SEVEN_THOUSAND = 0x0001EC92;
const char32_t INDIC_SIYAQ_NUMBER_EIGHT_THOUSAND = 0x0001EC93;
const char32_t INDIC_SIYAQ_NUMBER_NINE_THOUSAND = 0x0001EC94;
const char32_t INDIC_SIYAQ_NUMBER_TEN_THOUSAND = 0x0001EC95;
const char32_t INDIC_SIYAQ_NUMBER_TWENTY_THOUSAND = 0x0001EC96;
const char32_t INDIC_SIYAQ_NUMBER_THIRTY_THOUSAND = 0x0001EC97;
const char32_t INDIC_SIYAQ_NUMBER_FORTY_THOUSAND = 0x0001EC98;
const char32_t INDIC_SIYAQ_NUMBER_FIFTY_THOUSAND = 0x0001EC99;
const char32_t INDIC_SIYAQ_NUMBER_SIXTY_THOUSAND = 0x0001EC9A;
const char32_t INDIC_SIYAQ_NUMBER_SEVENTY_THOUSAND = 0x0001EC9B;
const char32_t INDIC_SIYAQ_NUMBER_EIGHTY_THOUSAND = 0x0001EC9C;
const char32_t INDIC_SIYAQ_NUMBER_NINETY_THOUSAND = 0x0001EC9D;
const char32_t INDIC_SIYAQ_NUMBER_LAKH = 0x0001EC9E;
const char32_t INDIC_SIYAQ_NUMBER_LAKHAN = 0x0001EC9F;
const char32_t INDIC_SIYAQ_LAKH_MARK = 0x0001ECA0;
const char32_t INDIC_SIYAQ_NUMBER_KAROR = 0x0001ECA1;
const char32_t INDIC_SIYAQ_NUMBER_KARORAN = 0x0001ECA2;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_ONE = 0x0001ECA3;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_TWO = 0x0001ECA4;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_THREE = 0x0001ECA5;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_FOUR = 0x0001ECA6;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_FIVE = 0x0001ECA7;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_SIX = 0x0001ECA8;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_SEVEN = 0x0001ECA9;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_EIGHT = 0x0001ECAA;
const char32_t INDIC_SIYAQ_NUMBER_PREFIXED_NINE = 0x0001ECAB;
const char32_t INDIC_SIYAQ_PLACEHOLDER = 0x0001ECAC;
const char32_t INDIC_SIYAQ_FRACTION_ONE_QUARTER = 0x0001ECAD;
const char32_t INDIC_SIYAQ_FRACTION_ONE_HALF = 0x0001ECAE;
const char32_t INDIC_SIYAQ_FRACTION_THREE_QUARTERS = 0x0001ECAF;
const char32_t INDIC_SIYAQ_RUPEE_MARK = 0x0001ECB0;
const char32_t INDIC_SIYAQ_NUMBER_ALTERNATE_ONE = 0x0001ECB1;
const char32_t INDIC_SIYAQ_NUMBER_ALTERNATE_TWO = 0x0001ECB2;
const char32_t INDIC_SIYAQ_NUMBER_ALTERNATE_TEN_THOUSAND = 0x0001ECB3;
const char32_t INDIC_SIYAQ_ALTERNATE_LAKH_MARK = 0x0001ECB4;
const char32_t OTTOMAN_SIYAQ_NUMBER_ONE = 0x0001ED01;
const char32_t OTTOMAN_SIYAQ_NUMBER_TWO = 0x0001ED02;
const char32_t OTTOMAN_SIYAQ_NUMBER_THREE = 0x0001ED03;
const char32_t OTTOMAN_SIYAQ_NUMBER_FOUR = 0x0001ED04;
const char32_t OTTOMAN_SIYAQ_NUMBER_FIVE = 0x0001ED05;
const char32_t OTTOMAN_SIYAQ_NUMBER_SIX = 0x0001ED06;
const char32_t OTTOMAN_SIYAQ_NUMBER_SEVEN = 0x0001ED07;
const char32_t OTTOMAN_SIYAQ_NUMBER_EIGHT = 0x0001ED08;
const char32_t OTTOMAN_SIYAQ_NUMBER_NINE = 0x0001ED09;
const char32_t OTTOMAN_SIYAQ_NUMBER_TEN = 0x0001ED0A;
const char32_t OTTOMAN_SIYAQ_NUMBER_TWENTY = 0x0001ED0B;
const char32_t OTTOMAN_SIYAQ_NUMBER_THIRTY = 0x0001ED0C;
const char32_t OTTOMAN_SIYAQ_NUMBER_FORTY = 0x0001ED0D;
const char32_t OTTOMAN_SIYAQ_NUMBER_FIFTY = 0x0001ED0E;
const char32_t OTTOMAN_SIYAQ_NUMBER_SIXTY = 0x0001ED0F;
const char32_t OTTOMAN_SIYAQ_NUMBER_SEVENTY = 0x0001ED10;
const char32_t OTTOMAN_SIYAQ_NUMBER_EIGHTY = 0x0001ED11;
const char32_t OTTOMAN_SIYAQ_NUMBER_NINETY = 0x0001ED12;
const char32_t OTTOMAN_SIYAQ_NUMBER_ONE_HUNDRED = 0x0001ED13;
const char32_t OTTOMAN_SIYAQ_NUMBER_TWO_HUNDRED = 0x0001ED14;
const char32_t OTTOMAN_SIYAQ_NUMBER_THREE_HUNDRED = 0x0001ED15;
const char32_t OTTOMAN_SIYAQ_NUMBER_FOUR_HUNDRED = 0x0001ED16;
const char32_t OTTOMAN_SIYAQ_NUMBER_FIVE_HUNDRED = 0x0001ED17;
const char32_t OTTOMAN_SIYAQ_NUMBER_SIX_HUNDRED = 0x0001ED18;
const char32_t OTTOMAN_SIYAQ_NUMBER_SEVEN_HUNDRED = 0x0001ED19;
const char32_t OTTOMAN_SIYAQ_NUMBER_EIGHT_HUNDRED = 0x0001ED1A;
const char32_t OTTOMAN_SIYAQ_NUMBER_NINE_HUNDRED = 0x0001ED1B;
const char32_t OTTOMAN_SIYAQ_NUMBER_ONE_THOUSAND = 0x0001ED1C;
const char32_t OTTOMAN_SIYAQ_NUMBER_TWO_THOUSAND = 0x0001ED1D;
const char32_t OTTOMAN_SIYAQ_NUMBER_THREE_THOUSAND = 0x0001ED1E;
const char32_t OTTOMAN_SIYAQ_NUMBER_FOUR_THOUSAND = 0x0001ED1F;
const char32_t OTTOMAN_SIYAQ_NUMBER_FIVE_THOUSAND = 0x0001ED20;
const char32_t OTTOMAN_SIYAQ_NUMBER_SIX_THOUSAND = 0x0001ED21;
const char32_t OTTOMAN_SIYAQ_NUMBER_SEVEN_THOUSAND = 0x0001ED22;
const char32_t OTTOMAN_SIYAQ_NUMBER_EIGHT_THOUSAND = 0x0001ED23;
const char32_t OTTOMAN_SIYAQ_NUMBER_NINE_THOUSAND = 0x0001ED24;
const char32_t OTTOMAN_SIYAQ_NUMBER_TEN_THOUSAND = 0x0001ED25;
const char32_t OTTOMAN_SIYAQ_NUMBER_TWENTY_THOUSAND = 0x0001ED26;
const char32_t OTTOMAN_SIYAQ_NUMBER_THIRTY_THOUSAND = 0x0001ED27;
const char32_t OTTOMAN_SIYAQ_NUMBER_FORTY_THOUSAND = 0x0001ED28;
const char32_t OTTOMAN_SIYAQ_NUMBER_FIFTY_THOUSAND = 0x0001ED29;
const char32_t OTTOMAN_SIYAQ_NUMBER_SIXTY_THOUSAND = 0x0001ED2A;
const char32_t OTTOMAN_SIYAQ_NUMBER_SEVENTY_THOUSAND = 0x0001ED2B;
const char32_t OTTOMAN_SIYAQ_NUMBER_EIGHTY_THOUSAND = 0x0001ED2C;
const char32_t OTTOMAN_SIYAQ_NUMBER_NINETY_THOUSAND = 0x0001ED2D;
const char32_t OTTOMAN_SIYAQ_MARRATAN = 0x0001ED2E;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_TWO = 0x0001ED2F;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_THREE = 0x0001ED30;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_FOUR = 0x0001ED31;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_FIVE = 0x0001ED32;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_SIX = 0x0001ED33;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_SEVEN = 0x0001ED34;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_EIGHT = 0x0001ED35;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_NINE = 0x0001ED36;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_TEN = 0x0001ED37;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_FOUR_HUNDRED = 0x0001ED38;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_SIX_HUNDRED = 0x0001ED39;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_TWO_THOUSAND = 0x0001ED3A;
const char32_t OTTOMAN_SIYAQ_ALTERNATE_NUMBER_TEN_THOUSAND = 0x0001ED3B;
const char32_t OTTOMAN_SIYAQ_FRACTION_ONE_HALF = 0x0001ED3C;
const char32_t OTTOMAN_SIYAQ_FRACTION_ONE_SIXTH = 0x0001ED3D;
const char32_t ARABIC_MATHEMATICAL_ALEF = 0x0001EE00;
const char32_t ARABIC_MATHEMATICAL_BEH = 0x0001EE01;
const char32_t ARABIC_MATHEMATICAL_JEEM = 0x0001EE02;
const char32_t ARABIC_MATHEMATICAL_DAL = 0x0001EE03;
const char32_t ARABIC_MATHEMATICAL_WAW = 0x0001EE05;
const char32_t ARABIC_MATHEMATICAL_ZAIN = 0x0001EE06;
const char32_t ARABIC_MATHEMATICAL_HAH = 0x0001EE07;
const char32_t ARABIC_MATHEMATICAL_TAH = 0x0001EE08;
const char32_t ARABIC_MATHEMATICAL_YEH = 0x0001EE09;
const char32_t ARABIC_MATHEMATICAL_KAF = 0x0001EE0A;
const char32_t ARABIC_MATHEMATICAL_LAM = 0x0001EE0B;
const char32_t ARABIC_MATHEMATICAL_MEEM = 0x0001EE0C;
const char32_t ARABIC_MATHEMATICAL_NOON = 0x0001EE0D;
const char32_t ARABIC_MATHEMATICAL_SEEN = 0x0001EE0E;
const char32_t ARABIC_MATHEMATICAL_AIN = 0x0001EE0F;
const char32_t ARABIC_MATHEMATICAL_FEH = 0x0001EE10;
const char32_t ARABIC_MATHEMATICAL_SAD = 0x0001EE11;
const char32_t ARABIC_MATHEMATICAL_QAF = 0x0001EE12;
const char32_t ARABIC_MATHEMATICAL_REH = 0x0001EE13;
const char32_t ARABIC_MATHEMATICAL_SHEEN = 0x0001EE14;
const char32_t ARABIC_MATHEMATICAL_TEH = 0x0001EE15;
const char32_t ARABIC_MATHEMATICAL_THEH = 0x0001EE16;
const char32_t ARABIC_MATHEMATICAL_KHAH = 0x0001EE17;
const char32_t ARABIC_MATHEMATICAL_THAL = 0x0001EE18;
const char32_t ARABIC_MATHEMATICAL_DAD = 0x0001EE19;
const char32_t ARABIC_MATHEMATICAL_ZAH = 0x0001EE1A;
const char32_t ARABIC_MATHEMATICAL_GHAIN = 0x0001EE1B;
const char32_t ARABIC_MATHEMATICAL_DOTLESS_BEH = 0x0001EE1C;
const char32_t ARABIC_MATHEMATICAL_DOTLESS_NOON = 0x0001EE1D;
const char32_t ARABIC_MATHEMATICAL_DOTLESS_FEH = 0x0001EE1E;
const char32_t ARABIC_MATHEMATICAL_DOTLESS_QAF = 0x0001EE1F;
const char32_t ARABIC_MATHEMATICAL_INITIAL_BEH = 0x0001EE21;
const char32_t ARABIC_MATHEMATICAL_INITIAL_JEEM = 0x0001EE22;
const char32_t ARABIC_MATHEMATICAL_INITIAL_HEH = 0x0001EE24;
const char32_t ARABIC_MATHEMATICAL_INITIAL_HAH = 0x0001EE27;
const char32_t ARABIC_MATHEMATICAL_INITIAL_YEH = 0x0001EE29;
const char32_t ARABIC_MATHEMATICAL_INITIAL_KAF = 0x0001EE2A;
const char32_t ARABIC_MATHEMATICAL_INITIAL_LAM = 0x0001EE2B;
const char32_t ARABIC_MATHEMATICAL_INITIAL_MEEM = 0x0001EE2C;
const char32_t ARABIC_MATHEMATICAL_INITIAL_NOON = 0x0001EE2D;
const char32_t ARABIC_MATHEMATICAL_INITIAL_SEEN = 0x0001EE2E;
const char32_t ARABIC_MATHEMATICAL_INITIAL_AIN = 0x0001EE2F;
const char32_t ARABIC_MATHEMATICAL_INITIAL_FEH = 0x0001EE30;
const char32_t ARABIC_MATHEMATICAL_INITIAL_SAD = 0x0001EE31;
const char32_t ARABIC_MATHEMATICAL_INITIAL_QAF = 0x0001EE32;
const char32_t ARABIC_MATHEMATICAL_INITIAL_SHEEN = 0x0001EE34;
const char32_t ARABIC_MATHEMATICAL_INITIAL_TEH = 0x0001EE35;
const char32_t ARABIC_MATHEMATICAL_INITIAL_THEH = 0x0001EE36;
const char32_t ARABIC_MATHEMATICAL_INITIAL_KHAH = 0x0001EE37;
const char32_t ARABIC_MATHEMATICAL_INITIAL_DAD = 0x0001EE39;
const char32_t ARABIC_MATHEMATICAL_INITIAL_GHAIN = 0x0001EE3B;
const char32_t ARABIC_MATHEMATICAL_TAILED_JEEM = 0x0001EE42;
const char32_t ARABIC_MATHEMATICAL_TAILED_HAH = 0x0001EE47;
const char32_t ARABIC_MATHEMATICAL_TAILED_YEH = 0x0001EE49;
const char32_t ARABIC_MATHEMATICAL_TAILED_LAM = 0x0001EE4B;
const char32_t ARABIC_MATHEMATICAL_TAILED_NOON = 0x0001EE4D;
const char32_t ARABIC_MATHEMATICAL_TAILED_SEEN = 0x0001EE4E;
const char32_t ARABIC_MATHEMATICAL_TAILED_AIN = 0x0001EE4F;
const char32_t ARABIC_MATHEMATICAL_TAILED_SAD = 0x0001EE51;
const char32_t ARABIC_MATHEMATICAL_TAILED_QAF = 0x0001EE52;
const char32_t ARABIC_MATHEMATICAL_TAILED_SHEEN = 0x0001EE54;
const char32_t ARABIC_MATHEMATICAL_TAILED_KHAH = 0x0001EE57;
const char32_t ARABIC_MATHEMATICAL_TAILED_DAD = 0x0001EE59;
const char32_t ARABIC_MATHEMATICAL_TAILED_GHAIN = 0x0001EE5B;
const char32_t ARABIC_MATHEMATICAL_TAILED_DOTLESS_NOON = 0x0001EE5D;
const char32_t ARABIC_MATHEMATICAL_TAILED_DOTLESS_QAF = 0x0001EE5F;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_BEH = 0x0001EE61;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_JEEM = 0x0001EE62;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_HEH = 0x0001EE64;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_HAH = 0x0001EE67;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_TAH = 0x0001EE68;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_YEH = 0x0001EE69;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_KAF = 0x0001EE6A;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_MEEM = 0x0001EE6C;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_NOON = 0x0001EE6D;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_SEEN = 0x0001EE6E;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_AIN = 0x0001EE6F;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_FEH = 0x0001EE70;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_SAD = 0x0001EE71;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_QAF = 0x0001EE72;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_SHEEN = 0x0001EE74;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_TEH = 0x0001EE75;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_THEH = 0x0001EE76;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_KHAH = 0x0001EE77;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_DAD = 0x0001EE79;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_ZAH = 0x0001EE7A;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_GHAIN = 0x0001EE7B;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_DOTLESS_BEH = 0x0001EE7C;
const char32_t ARABIC_MATHEMATICAL_STRETCHED_DOTLESS_FEH = 0x0001EE7E;
const char32_t ARABIC_MATHEMATICAL_LOOPED_ALEF = 0x0001EE80;
const char32_t ARABIC_MATHEMATICAL_LOOPED_BEH = 0x0001EE81;
const char32_t ARABIC_MATHEMATICAL_LOOPED_JEEM = 0x0001EE82;
const char32_t ARABIC_MATHEMATICAL_LOOPED_DAL = 0x0001EE83;
const char32_t ARABIC_MATHEMATICAL_LOOPED_HEH = 0x0001EE84;
const char32_t ARABIC_MATHEMATICAL_LOOPED_WAW = 0x0001EE85;
const char32_t ARABIC_MATHEMATICAL_LOOPED_ZAIN = 0x0001EE86;
const char32_t ARABIC_MATHEMATICAL_LOOPED_HAH = 0x0001EE87;
const char32_t ARABIC_MATHEMATICAL_LOOPED_TAH = 0x0001EE88;
const char32_t ARABIC_MATHEMATICAL_LOOPED_YEH = 0x0001EE89;
const char32_t ARABIC_MATHEMATICAL_LOOPED_LAM = 0x0001EE8B;
const char32_t ARABIC_MATHEMATICAL_LOOPED_MEEM = 0x0001EE8C;
const char32_t ARABIC_MATHEMATICAL_LOOPED_NOON = 0x0001EE8D;
const char32_t ARABIC_MATHEMATICAL_LOOPED_SEEN = 0x0001EE8E;
const char32_t ARABIC_MATHEMATICAL_LOOPED_AIN = 0x0001EE8F;
const char32_t ARABIC_MATHEMATICAL_LOOPED_FEH = 0x0001EE90;
const char32_t ARABIC_MATHEMATICAL_LOOPED_SAD = 0x0001EE91;
const char32_t ARABIC_MATHEMATICAL_LOOPED_QAF = 0x0001EE92;
const char32_t ARABIC_MATHEMATICAL_LOOPED_REH = 0x0001EE93;
const char32_t ARABIC_MATHEMATICAL_LOOPED_SHEEN = 0x0001EE94;
const char32_t ARABIC_MATHEMATICAL_LOOPED_TEH = 0x0001EE95;
const char32_t ARABIC_MATHEMATICAL_LOOPED_THEH = 0x0001EE96;
const char32_t ARABIC_MATHEMATICAL_LOOPED_KHAH = 0x0001EE97;
const char32_t ARABIC_MATHEMATICAL_LOOPED_THAL = 0x0001EE98;
const char32_t ARABIC_MATHEMATICAL_LOOPED_DAD = 0x0001EE99;
const char32_t ARABIC_MATHEMATICAL_LOOPED_ZAH = 0x0001EE9A;
const char32_t ARABIC_MATHEMATICAL_LOOPED_GHAIN = 0x0001EE9B;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_BEH = 0x0001EEA1;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_JEEM = 0x0001EEA2;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_DAL = 0x0001EEA3;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_WAW = 0x0001EEA5;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_ZAIN = 0x0001EEA6;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_HAH = 0x0001EEA7;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_TAH = 0x0001EEA8;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_YEH = 0x0001EEA9;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_LAM = 0x0001EEAB;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_MEEM = 0x0001EEAC;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_NOON = 0x0001EEAD;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_SEEN = 0x0001EEAE;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_AIN = 0x0001EEAF;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_FEH = 0x0001EEB0;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_SAD = 0x0001EEB1;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_QAF = 0x0001EEB2;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_REH = 0x0001EEB3;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_SHEEN = 0x0001EEB4;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_TEH = 0x0001EEB5;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_THEH = 0x0001EEB6;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_KHAH = 0x0001EEB7;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_THAL = 0x0001EEB8;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_DAD = 0x0001EEB9;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_ZAH = 0x0001EEBA;
const char32_t ARABIC_MATHEMATICAL_DOUBLE_STRUCK_GHAIN = 0x0001EEBB;
const char32_t ARABIC_MATHEMATICAL_OPERATOR_MEEM_WITH_HAH_WITH_TATWEEL = 0x0001EEF0;
const char32_t ARABIC_MATHEMATICAL_OPERATOR_HAH_WITH_DAL = 0x0001EEF1;
const char32_t MAHJONG_TILE_EAST_WIND = 0x0001F000;
const char32_t MAHJONG_TILE_SOUTH_WIND = 0x0001F001;
const char32_t MAHJONG_TILE_WEST_WIND = 0x0001F002;
const char32_t MAHJONG_TILE_NORTH_WIND = 0x0001F003;
const char32_t MAHJONG_TILE_RED_DRAGON = 0x0001F004;
const char32_t MAHJONG_TILE_GREEN_DRAGON = 0x0001F005;
const char32_t MAHJONG_TILE_WHITE_DRAGON = 0x0001F006;
const char32_t MAHJONG_TILE_ONE_OF_CHARACTERS = 0x0001F007;
const char32_t MAHJONG_TILE_TWO_OF_CHARACTERS = 0x0001F008;
const char32_t MAHJONG_TILE_THREE_OF_CHARACTERS = 0x0001F009;
const char32_t MAHJONG_TILE_FOUR_OF_CHARACTERS = 0x0001F00A;
const char32_t MAHJONG_TILE_FIVE_OF_CHARACTERS = 0x0001F00B;
const char32_t MAHJONG_TILE_SIX_OF_CHARACTERS = 0x0001F00C;
const char32_t MAHJONG_TILE_SEVEN_OF_CHARACTERS = 0x0001F00D;
const char32_t MAHJONG_TILE_EIGHT_OF_CHARACTERS = 0x0001F00E;
const char32_t MAHJONG_TILE_NINE_OF_CHARACTERS = 0x0001F00F;
const char32_t MAHJONG_TILE_ONE_OF_BAMBOOS = 0x0001F010;
const char32_t MAHJONG_TILE_TWO_OF_BAMBOOS = 0x0001F011;
const char32_t MAHJONG_TILE_THREE_OF_BAMBOOS = 0x0001F012;
const char32_t MAHJONG_TILE_FOUR_OF_BAMBOOS = 0x0001F013;
const char32_t MAHJONG_TILE_FIVE_OF_BAMBOOS = 0x0001F014;
const char32_t MAHJONG_TILE_SIX_OF_BAMBOOS = 0x0001F015;
const char32_t MAHJONG_TILE_SEVEN_OF_BAMBOOS = 0x0001F016;
const char32_t MAHJONG_TILE_EIGHT_OF_BAMBOOS = 0x0001F017;
const char32_t MAHJONG_TILE_NINE_OF_BAMBOOS = 0x0001F018;
const char32_t MAHJONG_TILE_ONE_OF_CIRCLES = 0x0001F019;
const char32_t MAHJONG_TILE_TWO_OF_CIRCLES = 0x0001F01A;
const char32_t MAHJONG_TILE_THREE_OF_CIRCLES = 0x0001F01B;
const char32_t MAHJONG_TILE_FOUR_OF_CIRCLES = 0x0001F01C;
const char32_t MAHJONG_TILE_FIVE_OF_CIRCLES = 0x0001F01D;
const char32_t MAHJONG_TILE_SIX_OF_CIRCLES = 0x0001F01E;
const char32_t MAHJONG_TILE_SEVEN_OF_CIRCLES = 0x0001F01F;
const char32_t MAHJONG_TILE_EIGHT_OF_CIRCLES = 0x0001F020;
const char32_t MAHJONG_TILE_NINE_OF_CIRCLES = 0x0001F021;
const char32_t MAHJONG_TILE_PLUM = 0x0001F022;
const char32_t MAHJONG_TILE_ORCHID = 0x0001F023;
const char32_t MAHJONG_TILE_BAMBOO = 0x0001F024;
const char32_t MAHJONG_TILE_CHRYSANTHEMUM = 0x0001F025;
const char32_t MAHJONG_TILE_SPRING = 0x0001F026;
const char32_t MAHJONG_TILE_SUMMER = 0x0001F027;
const char32_t MAHJONG_TILE_AUTUMN = 0x0001F028;
const char32_t MAHJONG_TILE_WINTER = 0x0001F029;
const char32_t MAHJONG_TILE_JOKER = 0x0001F02A;
const char32_t MAHJONG_TILE_BACK = 0x0001F02B;
const char32_t DOMINO_TILE_HORIZONTAL_BACK = 0x0001F030;
const char32_t DOMINO_TILE_HORIZONTAL_00_00 = 0x0001F031;
const char32_t DOMINO_TILE_HORIZONTAL_00_01 = 0x0001F032;
const char32_t DOMINO_TILE_HORIZONTAL_00_02 = 0x0001F033;
const char32_t DOMINO_TILE_HORIZONTAL_00_03 = 0x0001F034;
const char32_t DOMINO_TILE_HORIZONTAL_00_04 = 0x0001F035;
const char32_t DOMINO_TILE_HORIZONTAL_00_05 = 0x0001F036;
const char32_t DOMINO_TILE_HORIZONTAL_00_06 = 0x0001F037;
const char32_t DOMINO_TILE_HORIZONTAL_01_00 = 0x0001F038;
const char32_t DOMINO_TILE_HORIZONTAL_01_01 = 0x0001F039;
const char32_t DOMINO_TILE_HORIZONTAL_01_02 = 0x0001F03A;
const char32_t DOMINO_TILE_HORIZONTAL_01_03 = 0x0001F03B;
const char32_t DOMINO_TILE_HORIZONTAL_01_04 = 0x0001F03C;
const char32_t DOMINO_TILE_HORIZONTAL_01_05 = 0x0001F03D;
const char32_t DOMINO_TILE_HORIZONTAL_01_06 = 0x0001F03E;
const char32_t DOMINO_TILE_HORIZONTAL_02_00 = 0x0001F03F;
const char32_t DOMINO_TILE_HORIZONTAL_02_01 = 0x0001F040;
const char32_t DOMINO_TILE_HORIZONTAL_02_02 = 0x0001F041;
const char32_t DOMINO_TILE_HORIZONTAL_02_03 = 0x0001F042;
const char32_t DOMINO_TILE_HORIZONTAL_02_04 = 0x0001F043;
const char32_t DOMINO_TILE_HORIZONTAL_02_05 = 0x0001F044;
const char32_t DOMINO_TILE_HORIZONTAL_02_06 = 0x0001F045;
const char32_t DOMINO_TILE_HORIZONTAL_03_00 = 0x0001F046;
const char32_t DOMINO_TILE_HORIZONTAL_03_01 = 0x0001F047;
const char32_t DOMINO_TILE_HORIZONTAL_03_02 = 0x0001F048;
const char32_t DOMINO_TILE_HORIZONTAL_03_03 = 0x0001F049;
const char32_t DOMINO_TILE_HORIZONTAL_03_04 = 0x0001F04A;
const char32_t DOMINO_TILE_HORIZONTAL_03_05 = 0x0001F04B;
const char32_t DOMINO_TILE_HORIZONTAL_03_06 = 0x0001F04C;
const char32_t DOMINO_TILE_HORIZONTAL_04_00 = 0x0001F04D;
const char32_t DOMINO_TILE_HORIZONTAL_04_01 = 0x0001F04E;
const char32_t DOMINO_TILE_HORIZONTAL_04_02 = 0x0001F04F;
const char32_t DOMINO_TILE_HORIZONTAL_04_03 = 0x0001F050;
const char32_t DOMINO_TILE_HORIZONTAL_04_04 = 0x0001F051;
const char32_t DOMINO_TILE_HORIZONTAL_04_05 = 0x0001F052;
const char32_t DOMINO_TILE_HORIZONTAL_04_06 = 0x0001F053;
const char32_t DOMINO_TILE_HORIZONTAL_05_00 = 0x0001F054;
const char32_t DOMINO_TILE_HORIZONTAL_05_01 = 0x0001F055;
const char32_t DOMINO_TILE_HORIZONTAL_05_02 = 0x0001F056;
const char32_t DOMINO_TILE_HORIZONTAL_05_03 = 0x0001F057;
const char32_t DOMINO_TILE_HORIZONTAL_05_04 = 0x0001F058;
const char32_t DOMINO_TILE_HORIZONTAL_05_05 = 0x0001F059;
const char32_t DOMINO_TILE_HORIZONTAL_05_06 = 0x0001F05A;
const char32_t DOMINO_TILE_HORIZONTAL_06_00 = 0x0001F05B;
const char32_t DOMINO_TILE_HORIZONTAL_06_01 = 0x0001F05C;
const char32_t DOMINO_TILE_HORIZONTAL_06_02 = 0x0001F05D;
const char32_t DOMINO_TILE_HORIZONTAL_06_03 = 0x0001F05E;
const char32_t DOMINO_TILE_HORIZONTAL_06_04 = 0x0001F05F;
const char32_t DOMINO_TILE_HORIZONTAL_06_05 = 0x0001F060;
const char32_t DOMINO_TILE_HORIZONTAL_06_06 = 0x0001F061;
const char32_t DOMINO_TILE_VERTICAL_BACK = 0x0001F062;
const char32_t DOMINO_TILE_VERTICAL_00_00 = 0x0001F063;
const char32_t DOMINO_TILE_VERTICAL_00_01 = 0x0001F064;
const char32_t DOMINO_TILE_VERTICAL_00_02 = 0x0001F065;
const char32_t DOMINO_TILE_VERTICAL_00_03 = 0x0001F066;
const char32_t DOMINO_TILE_VERTICAL_00_04 = 0x0001F067;
const char32_t DOMINO_TILE_VERTICAL_00_05 = 0x0001F068;
const char32_t DOMINO_TILE_VERTICAL_00_06 = 0x0001F069;
const char32_t DOMINO_TILE_VERTICAL_01_00 = 0x0001F06A;
const char32_t DOMINO_TILE_VERTICAL_01_01 = 0x0001F06B;
const char32_t DOMINO_TILE_VERTICAL_01_02 = 0x0001F06C;
const char32_t DOMINO_TILE_VERTICAL_01_03 = 0x0001F06D;
const char32_t DOMINO_TILE_VERTICAL_01_04 = 0x0001F06E;
const char32_t DOMINO_TILE_VERTICAL_01_05 = 0x0001F06F;
const char32_t DOMINO_TILE_VERTICAL_01_06 = 0x0001F070;
const char32_t DOMINO_TILE_VERTICAL_02_00 = 0x0001F071;
const char32_t DOMINO_TILE_VERTICAL_02_01 = 0x0001F072;
const char32_t DOMINO_TILE_VERTICAL_02_02 = 0x0001F073;
const char32_t DOMINO_TILE_VERTICAL_02_03 = 0x0001F074;
const char32_t DOMINO_TILE_VERTICAL_02_04 = 0x0001F075;
const char32_t DOMINO_TILE_VERTICAL_02_05 = 0x0001F076;
const char32_t DOMINO_TILE_VERTICAL_02_06 = 0x0001F077;
const char32_t DOMINO_TILE_VERTICAL_03_00 = 0x0001F078;
const char32_t DOMINO_TILE_VERTICAL_03_01 = 0x0001F079;
const char32_t DOMINO_TILE_VERTICAL_03_02 = 0x0001F07A;
const char32_t DOMINO_TILE_VERTICAL_03_03 = 0x0001F07B;
const char32_t DOMINO_TILE_VERTICAL_03_04 = 0x0001F07C;
const char32_t DOMINO_TILE_VERTICAL_03_05 = 0x0001F07D;
const char32_t DOMINO_TILE_VERTICAL_03_06 = 0x0001F07E;
const char32_t DOMINO_TILE_VERTICAL_04_00 = 0x0001F07F;
const char32_t DOMINO_TILE_VERTICAL_04_01 = 0x0001F080;
const char32_t DOMINO_TILE_VERTICAL_04_02 = 0x0001F081;
const char32_t DOMINO_TILE_VERTICAL_04_03 = 0x0001F082;
const char32_t DOMINO_TILE_VERTICAL_04_04 = 0x0001F083;
const char32_t DOMINO_TILE_VERTICAL_04_05 = 0x0001F084;
const char32_t DOMINO_TILE_VERTICAL_04_06 = 0x0001F085;
const char32_t DOMINO_TILE_VERTICAL_05_00 = 0x0001F086;
const char32_t DOMINO_TILE_VERTICAL_05_01 = 0x0001F087;
const char32_t DOMINO_TILE_VERTICAL_05_02 = 0x0001F088;
const char32_t DOMINO_TILE_VERTICAL_05_03 = 0x0001F089;
const char32_t DOMINO_TILE_VERTICAL_05_04 = 0x0001F08A;
const char32_t DOMINO_TILE_VERTICAL_05_05 = 0x0001F08B;
const char32_t DOMINO_TILE_VERTICAL_05_06 = 0x0001F08C;
const char32_t DOMINO_TILE_VERTICAL_06_00 = 0x0001F08D;
const char32_t DOMINO_TILE_VERTICAL_06_01 = 0x0001F08E;
const char32_t DOMINO_TILE_VERTICAL_06_02 = 0x0001F08F;
const char32_t DOMINO_TILE_VERTICAL_06_03 = 0x0001F090;
const char32_t DOMINO_TILE_VERTICAL_06_04 = 0x0001F091;
const char32_t DOMINO_TILE_VERTICAL_06_05 = 0x0001F092;
const char32_t DOMINO_TILE_VERTICAL_06_06 = 0x0001F093;
const char32_t PLAYING_CARD_BACK = 0x0001F0A0;
const char32_t PLAYING_CARD_ACE_OF_SPADES = 0x0001F0A1;
const char32_t PLAYING_CARD_TWO_OF_SPADES = 0x0001F0A2;
const char32_t PLAYING_CARD_THREE_OF_SPADES = 0x0001F0A3;
const char32_t PLAYING_CARD_FOUR_OF_SPADES = 0x0001F0A4;
const char32_t PLAYING_CARD_FIVE_OF_SPADES = 0x0001F0A5;
const char32_t PLAYING_CARD_SIX_OF_SPADES = 0x0001F0A6;
const char32_t PLAYING_CARD_SEVEN_OF_SPADES = 0x0001F0A7;
const char32_t PLAYING_CARD_EIGHT_OF_SPADES = 0x0001F0A8;
const char32_t PLAYING_CARD_NINE_OF_SPADES = 0x0001F0A9;
const char32_t PLAYING_CARD_TEN_OF_SPADES = 0x0001F0AA;
const char32_t PLAYING_CARD_JACK_OF_SPADES = 0x0001F0AB;
const char32_t PLAYING_CARD_KNIGHT_OF_SPADES = 0x0001F0AC;
const char32_t PLAYING_CARD_QUEEN_OF_SPADES = 0x0001F0AD;
const char32_t PLAYING_CARD_KING_OF_SPADES = 0x0001F0AE;
const char32_t PLAYING_CARD_ACE_OF_HEARTS = 0x0001F0B1;
const char32_t PLAYING_CARD_TWO_OF_HEARTS = 0x0001F0B2;
const char32_t PLAYING_CARD_THREE_OF_HEARTS = 0x0001F0B3;
const char32_t PLAYING_CARD_FOUR_OF_HEARTS = 0x0001F0B4;
const char32_t PLAYING_CARD_FIVE_OF_HEARTS = 0x0001F0B5;
const char32_t PLAYING_CARD_SIX_OF_HEARTS = 0x0001F0B6;
const char32_t PLAYING_CARD_SEVEN_OF_HEARTS = 0x0001F0B7;
const char32_t PLAYING_CARD_EIGHT_OF_HEARTS = 0x0001F0B8;
const char32_t PLAYING_CARD_NINE_OF_HEARTS = 0x0001F0B9;
const char32_t PLAYING_CARD_TEN_OF_HEARTS = 0x0001F0BA;
const char32_t PLAYING_CARD_JACK_OF_HEARTS = 0x0001F0BB;
const char32_t PLAYING_CARD_KNIGHT_OF_HEARTS = 0x0001F0BC;
const char32_t PLAYING_CARD_QUEEN_OF_HEARTS = 0x0001F0BD;
const char32_t PLAYING_CARD_KING_OF_HEARTS = 0x0001F0BE;
const char32_t PLAYING_CARD_RED_JOKER = 0x0001F0BF;
const char32_t PLAYING_CARD_ACE_OF_DIAMONDS = 0x0001F0C1;
const char32_t PLAYING_CARD_TWO_OF_DIAMONDS = 0x0001F0C2;
const char32_t PLAYING_CARD_THREE_OF_DIAMONDS = 0x0001F0C3;
const char32_t PLAYING_CARD_FOUR_OF_DIAMONDS = 0x0001F0C4;
const char32_t PLAYING_CARD_FIVE_OF_DIAMONDS = 0x0001F0C5;
const char32_t PLAYING_CARD_SIX_OF_DIAMONDS = 0x0001F0C6;
const char32_t PLAYING_CARD_SEVEN_OF_DIAMONDS = 0x0001F0C7;
const char32_t PLAYING_CARD_EIGHT_OF_DIAMONDS = 0x0001F0C8;
const char32_t PLAYING_CARD_NINE_OF_DIAMONDS = 0x0001F0C9;
const char32_t PLAYING_CARD_TEN_OF_DIAMONDS = 0x0001F0CA;
const char32_t PLAYING_CARD_JACK_OF_DIAMONDS = 0x0001F0CB;
const char32_t PLAYING_CARD_KNIGHT_OF_DIAMONDS = 0x0001F0CC;
const char32_t PLAYING_CARD_QUEEN_OF_DIAMONDS = 0x0001F0CD;
const char32_t PLAYING_CARD_KING_OF_DIAMONDS = 0x0001F0CE;
const char32_t PLAYING_CARD_BLACK_JOKER = 0x0001F0CF;
const char32_t PLAYING_CARD_ACE_OF_CLUBS = 0x0001F0D1;
const char32_t PLAYING_CARD_TWO_OF_CLUBS = 0x0001F0D2;
const char32_t PLAYING_CARD_THREE_OF_CLUBS = 0x0001F0D3;
const char32_t PLAYING_CARD_FOUR_OF_CLUBS = 0x0001F0D4;
const char32_t PLAYING_CARD_FIVE_OF_CLUBS = 0x0001F0D5;
const char32_t PLAYING_CARD_SIX_OF_CLUBS = 0x0001F0D6;
const char32_t PLAYING_CARD_SEVEN_OF_CLUBS = 0x0001F0D7;
const char32_t PLAYING_CARD_EIGHT_OF_CLUBS = 0x0001F0D8;
const char32_t PLAYING_CARD_NINE_OF_CLUBS = 0x0001F0D9;
const char32_t PLAYING_CARD_TEN_OF_CLUBS = 0x0001F0DA;
const char32_t PLAYING_CARD_JACK_OF_CLUBS = 0x0001F0DB;
const char32_t PLAYING_CARD_KNIGHT_OF_CLUBS = 0x0001F0DC;
const char32_t PLAYING_CARD_QUEEN_OF_CLUBS = 0x0001F0DD;
const char32_t PLAYING_CARD_KING_OF_CLUBS = 0x0001F0DE;
const char32_t PLAYING_CARD_WHITE_JOKER = 0x0001F0DF;
const char32_t PLAYING_CARD_FOOL = 0x0001F0E0;
const char32_t PLAYING_CARD_TRUMP_1 = 0x0001F0E1;
const char32_t PLAYING_CARD_TRUMP_2 = 0x0001F0E2;
const char32_t PLAYING_CARD_TRUMP_3 = 0x0001F0E3;
const char32_t PLAYING_CARD_TRUMP_4 = 0x0001F0E4;
const char32_t PLAYING_CARD_TRUMP_5 = 0x0001F0E5;
const char32_t PLAYING_CARD_TRUMP_6 = 0x0001F0E6;
const char32_t PLAYING_CARD_TRUMP_7 = 0x0001F0E7;
const char32_t PLAYING_CARD_TRUMP_8 = 0x0001F0E8;
const char32_t PLAYING_CARD_TRUMP_9 = 0x0001F0E9;
const char32_t PLAYING_CARD_TRUMP_10 = 0x0001F0EA;
const char32_t PLAYING_CARD_TRUMP_11 = 0x0001F0EB;
const char32_t PLAYING_CARD_TRUMP_12 = 0x0001F0EC;
const char32_t PLAYING_CARD_TRUMP_13 = 0x0001F0ED;
const char32_t PLAYING_CARD_TRUMP_14 = 0x0001F0EE;
const char32_t PLAYING_CARD_TRUMP_15 = 0x0001F0EF;
const char32_t PLAYING_CARD_TRUMP_16 = 0x0001F0F0;
const char32_t PLAYING_CARD_TRUMP_17 = 0x0001F0F1;
const char32_t PLAYING_CARD_TRUMP_18 = 0x0001F0F2;
const char32_t PLAYING_CARD_TRUMP_19 = 0x0001F0F3;
const char32_t PLAYING_CARD_TRUMP_20 = 0x0001F0F4;
const char32_t PLAYING_CARD_TRUMP_21 = 0x0001F0F5;
const char32_t DIGIT_ZERO_FULL_STOP = 0x0001F100;
const char32_t DIGIT_ZERO_COMMA = 0x0001F101;
const char32_t DIGIT_ONE_COMMA = 0x0001F102;
const char32_t DIGIT_TWO_COMMA = 0x0001F103;
const char32_t DIGIT_THREE_COMMA = 0x0001F104;
const char32_t DIGIT_FOUR_COMMA = 0x0001F105;
const char32_t DIGIT_FIVE_COMMA = 0x0001F106;
const char32_t DIGIT_SIX_COMMA = 0x0001F107;
const char32_t DIGIT_SEVEN_COMMA = 0x0001F108;
const char32_t DIGIT_EIGHT_COMMA = 0x0001F109;
const char32_t DIGIT_NINE_COMMA = 0x0001F10A;
const char32_t DINGBAT_CIRCLED_SANS_SERIF_DIGIT_ZERO = 0x0001F10B;
const char32_t DINGBAT_NEGATIVE_CIRCLED_SANS_SERIF_DIGIT_ZERO = 0x0001F10C;
const char32_t CIRCLED_ZERO_WITH_SLASH = 0x0001F10D;
const char32_t CIRCLED_ANTICLOCKWISE_ARROW = 0x0001F10E;
const char32_t CIRCLED_DOLLAR_SIGN_WITH_OVERLAID_BACKSLASH = 0x0001F10F;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_A = 0x0001F110;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_B = 0x0001F111;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_C = 0x0001F112;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_D = 0x0001F113;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_E = 0x0001F114;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_F = 0x0001F115;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_G = 0x0001F116;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_H = 0x0001F117;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_I = 0x0001F118;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_J = 0x0001F119;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_K = 0x0001F11A;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_L = 0x0001F11B;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_M = 0x0001F11C;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_N = 0x0001F11D;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_O = 0x0001F11E;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_P = 0x0001F11F;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_Q = 0x0001F120;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_R = 0x0001F121;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_S = 0x0001F122;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_T = 0x0001F123;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_U = 0x0001F124;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_V = 0x0001F125;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_W = 0x0001F126;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_X = 0x0001F127;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_Y = 0x0001F128;
const char32_t PARENTHESIZED_LATIN_CAPITAL_LETTER_Z = 0x0001F129;
const char32_t TORTOISE_SHELL_BRACKETED_LATIN_CAPITAL_LETTER_S = 0x0001F12A;
const char32_t CIRCLED_ITALIC_LATIN_CAPITAL_LETTER_C = 0x0001F12B;
const char32_t CIRCLED_ITALIC_LATIN_CAPITAL_LETTER_R = 0x0001F12C;
const char32_t CIRCLED_CD = 0x0001F12D;
const char32_t CIRCLED_WZ = 0x0001F12E;
const char32_t COPYLEFT_SYMBOL = 0x0001F12F;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_A = 0x0001F130;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_B = 0x0001F131;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_C = 0x0001F132;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_D = 0x0001F133;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_E = 0x0001F134;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_F = 0x0001F135;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_G = 0x0001F136;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_H = 0x0001F137;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_I = 0x0001F138;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_J = 0x0001F139;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_K = 0x0001F13A;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_L = 0x0001F13B;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_M = 0x0001F13C;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_N = 0x0001F13D;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_O = 0x0001F13E;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_P = 0x0001F13F;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_Q = 0x0001F140;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_R = 0x0001F141;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_S = 0x0001F142;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_T = 0x0001F143;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_U = 0x0001F144;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_V = 0x0001F145;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_W = 0x0001F146;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_X = 0x0001F147;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_Y = 0x0001F148;
const char32_t SQUARED_LATIN_CAPITAL_LETTER_Z = 0x0001F149;
const char32_t SQUARED_HV = 0x0001F14A;
const char32_t SQUARED_MV = 0x0001F14B;
const char32_t SQUARED_SD = 0x0001F14C;
const char32_t SQUARED_SS = 0x0001F14D;
const char32_t SQUARED_PPV = 0x0001F14E;
const char32_t SQUARED_WC = 0x0001F14F;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_A = 0x0001F150;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_B = 0x0001F151;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_C = 0x0001F152;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_D = 0x0001F153;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_E = 0x0001F154;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_F = 0x0001F155;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_G = 0x0001F156;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_H = 0x0001F157;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_I = 0x0001F158;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_J = 0x0001F159;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_K = 0x0001F15A;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_L = 0x0001F15B;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_M = 0x0001F15C;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_N = 0x0001F15D;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_O = 0x0001F15E;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_P = 0x0001F15F;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_Q = 0x0001F160;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_R = 0x0001F161;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_S = 0x0001F162;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_T = 0x0001F163;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_U = 0x0001F164;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_V = 0x0001F165;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_W = 0x0001F166;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_X = 0x0001F167;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_Y = 0x0001F168;
const char32_t NEGATIVE_CIRCLED_LATIN_CAPITAL_LETTER_Z = 0x0001F169;
const char32_t RAISED_MC_SIGN = 0x0001F16A;
const char32_t RAISED_MD_SIGN = 0x0001F16B;
const char32_t RAISED_MR_SIGN = 0x0001F16C;
const char32_t CIRCLED_CC = 0x0001F16D;
const char32_t CIRCLED_C_WITH_OVERLAID_BACKSLASH = 0x0001F16E;
const char32_t CIRCLED_HUMAN_FIGURE = 0x0001F16F;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_A = 0x0001F170;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_B = 0x0001F171;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_C = 0x0001F172;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_D = 0x0001F173;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_E = 0x0001F174;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_F = 0x0001F175;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_G = 0x0001F176;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_H = 0x0001F177;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_I = 0x0001F178;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_J = 0x0001F179;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_K = 0x0001F17A;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_L = 0x0001F17B;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_M = 0x0001F17C;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_N = 0x0001F17D;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_O = 0x0001F17E;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_P = 0x0001F17F;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_Q = 0x0001F180;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_R = 0x0001F181;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_S = 0x0001F182;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_T = 0x0001F183;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_U = 0x0001F184;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_V = 0x0001F185;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_W = 0x0001F186;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_X = 0x0001F187;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_Y = 0x0001F188;
const char32_t NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_Z = 0x0001F189;
const char32_t CROSSED_NEGATIVE_SQUARED_LATIN_CAPITAL_LETTER_P = 0x0001F18A;
const char32_t NEGATIVE_SQUARED_IC = 0x0001F18B;
const char32_t NEGATIVE_SQUARED_PA = 0x0001F18C;
const char32_t NEGATIVE_SQUARED_SA = 0x0001F18D;
const char32_t NEGATIVE_SQUARED_AB = 0x0001F18E;
const char32_t NEGATIVE_SQUARED_WC = 0x0001F18F;
const char32_t SQUARE_DJ = 0x0001F190;
const char32_t SQUARED_CL = 0x0001F191;
const char32_t SQUARED_COOL = 0x0001F192;
const char32_t SQUARED_FREE = 0x0001F193;
const char32_t SQUARED_ID = 0x0001F194;
const char32_t SQUARED_NEW = 0x0001F195;
const char32_t SQUARED_NG = 0x0001F196;
const char32_t SQUARED_OK = 0x0001F197;
const char32_t SQUARED_SOS = 0x0001F198;
const char32_t SQUARED_UP_WITH_EXCLAMATION_MARK = 0x0001F199;
const char32_t SQUARED_VS = 0x0001F19A;
const char32_t SQUARED_THREE_D = 0x0001F19B;
const char32_t SQUARED_SECOND_SCREEN = 0x0001F19C;
const char32_t SQUARED_TWO_K = 0x0001F19D;
const char32_t SQUARED_FOUR_K = 0x0001F19E;
const char32_t SQUARED_EIGHT_K = 0x0001F19F;
const char32_t SQUARED_FIVE_POINT_ONE = 0x0001F1A0;
const char32_t SQUARED_SEVEN_POINT_ONE = 0x0001F1A1;
const char32_t SQUARED_TWENTY_TWO_POINT_TWO = 0x0001F1A2;
const char32_t SQUARED_SIXTY_P = 0x0001F1A3;
const char32_t SQUARED_ONE_HUNDRED_TWENTY_P = 0x0001F1A4;
const char32_t SQUARED_LATIN_SMALL_LETTER_D = 0x0001F1A5;
const char32_t SQUARED_HC = 0x0001F1A6;
const char32_t SQUARED_HDR = 0x0001F1A7;
const char32_t SQUARED_HI_RES = 0x0001F1A8;
const char32_t SQUARED_LOSSLESS = 0x0001F1A9;
const char32_t SQUARED_SHV = 0x0001F1AA;
const char32_t SQUARED_UHD = 0x0001F1AB;
const char32_t SQUARED_VOD = 0x0001F1AC;
const char32_t MASK_WORK_SYMBOL = 0x0001F1AD;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_A = 0x0001F1E6;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_B = 0x0001F1E7;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_C = 0x0001F1E8;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_D = 0x0001F1E9;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_E = 0x0001F1EA;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_F = 0x0001F1EB;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_G = 0x0001F1EC;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_H = 0x0001F1ED;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_I = 0x0001F1EE;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_J = 0x0001F1EF;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_K = 0x0001F1F0;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_L = 0x0001F1F1;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_M = 0x0001F1F2;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_N = 0x0001F1F3;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_O = 0x0001F1F4;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_P = 0x0001F1F5;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_Q = 0x0001F1F6;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_R = 0x0001F1F7;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_S = 0x0001F1F8;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_T = 0x0001F1F9;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_U = 0x0001F1FA;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_V = 0x0001F1FB;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_W = 0x0001F1FC;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_X = 0x0001F1FD;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_Y = 0x0001F1FE;
const char32_t REGIONAL_INDICATOR_SYMBOL_LETTER_Z = 0x0001F1FF;
const char32_t SQUARE_HIRAGANA_HOKA = 0x0001F200;
const char32_t SQUARED_KATAKANA_KOKO = 0x0001F201;
const char32_t SQUARED_KATAKANA_SA = 0x0001F202;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_624B = 0x0001F210;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_5B57 = 0x0001F211;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_53CC = 0x0001F212;
const char32_t SQUARED_KATAKANA_DE = 0x0001F213;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_4E8C = 0x0001F214;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_591A = 0x0001F215;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_89E3 = 0x0001F216;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_5929 = 0x0001F217;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_4EA4 = 0x0001F218;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6620 = 0x0001F219;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_7121 = 0x0001F21A;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6599 = 0x0001F21B;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_524D = 0x0001F21C;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_5F8C = 0x0001F21D;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_518D = 0x0001F21E;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_65B0 = 0x0001F21F;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_521D = 0x0001F220;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_7D42 = 0x0001F221;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_751F = 0x0001F222;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_8CA9 = 0x0001F223;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_58F0 = 0x0001F224;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_5439 = 0x0001F225;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6F14 = 0x0001F226;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6295 = 0x0001F227;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6355 = 0x0001F228;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_4E00 = 0x0001F229;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_4E09 = 0x0001F22A;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_904A = 0x0001F22B;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_5DE6 = 0x0001F22C;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_4E2D = 0x0001F22D;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_53F3 = 0x0001F22E;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6307 = 0x0001F22F;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_8D70 = 0x0001F230;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6253 = 0x0001F231;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_7981 = 0x0001F232;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_7A7A = 0x0001F233;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_5408 = 0x0001F234;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6E80 = 0x0001F235;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6709 = 0x0001F236;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_6708 = 0x0001F237;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_7533 = 0x0001F238;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_5272 = 0x0001F239;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_55B6 = 0x0001F23A;
const char32_t SQUARED_CJK_UNIFIED_IDEOGRAPH_914D = 0x0001F23B;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_672C = 0x0001F240;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_4E09 = 0x0001F241;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_4E8C = 0x0001F242;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_5B89 = 0x0001F243;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_70B9 = 0x0001F244;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_6253 = 0x0001F245;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_76D7 = 0x0001F246;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_52DD = 0x0001F247;
const char32_t TORTOISE_SHELL_BRACKETED_CJK_UNIFIED_IDEOGRAPH_6557 = 0x0001F248;
const char32_t CIRCLED_IDEOGRAPH_ADVANTAGE = 0x0001F250;
const char32_t CIRCLED_IDEOGRAPH_ACCEPT = 0x0001F251;
const char32_t ROUNDED_SYMBOL_FOR_FU = 0x0001F260;
const char32_t ROUNDED_SYMBOL_FOR_LU = 0x0001F261;
const char32_t ROUNDED_SYMBOL_FOR_SHOU = 0x0001F262;
const char32_t ROUNDED_SYMBOL_FOR_XI = 0x0001F263;
const char32_t ROUNDED_SYMBOL_FOR_SHUANGXI = 0x0001F264;
const char32_t ROUNDED_SYMBOL_FOR_CAI = 0x0001F265;
const char32_t CYCLONE = 0x0001F300;
const char32_t FOGGY = 0x0001F301;
const char32_t CLOSED_UMBRELLA = 0x0001F302;
const char32_t NIGHT_WITH_STARS = 0x0001F303;
const char32_t SUNRISE_OVER_MOUNTAINS = 0x0001F304;
const char32_t SUNRISE = 0x0001F305;
const char32_t CITYSCAPE_AT_DUSK = 0x0001F306;
const char32_t SUNSET_OVER_BUILDINGS = 0x0001F307;
const char32_t RAINBOW = 0x0001F308;
const char32_t BRIDGE_AT_NIGHT = 0x0001F309;
const char32_t WATER_WAVE = 0x0001F30A;
const char32_t VOLCANO = 0x0001F30B;
const char32_t MILKY_WAY = 0x0001F30C;
const char32_t EARTH_GLOBE_EUROPE_AFRICA = 0x0001F30D;
const char32_t EARTH_GLOBE_AMERICAS = 0x0001F30E;
const char32_t EARTH_GLOBE_ASIA_AUSTRALIA = 0x0001F30F;
const char32_t GLOBE_WITH_MERIDIANS = 0x0001F310;
const char32_t NEW_MOON_SYMBOL = 0x0001F311;
const char32_t WAXING_CRESCENT_MOON_SYMBOL = 0x0001F312;
const char32_t FIRST_QUARTER_MOON_SYMBOL = 0x0001F313;
const char32_t WAXING_GIBBOUS_MOON_SYMBOL = 0x0001F314;
const char32_t FULL_MOON_SYMBOL = 0x0001F315;
const char32_t WANING_GIBBOUS_MOON_SYMBOL = 0x0001F316;
const char32_t LAST_QUARTER_MOON_SYMBOL = 0x0001F317;
const char32_t WANING_CRESCENT_MOON_SYMBOL = 0x0001F318;
const char32_t CRESCENT_MOON = 0x0001F319;
const char32_t NEW_MOON_WITH_FACE = 0x0001F31A;
const char32_t FIRST_QUARTER_MOON_WITH_FACE = 0x0001F31B;
const char32_t LAST_QUARTER_MOON_WITH_FACE = 0x0001F31C;
const char32_t FULL_MOON_WITH_FACE = 0x0001F31D;
const char32_t SUN_WITH_FACE = 0x0001F31E;
const char32_t GLOWING_STAR = 0x0001F31F;
const char32_t SHOOTING_STAR = 0x0001F320;
const char32_t THERMOMETER = 0x0001F321;
const char32_t BLACK_DROPLET = 0x0001F322;
const char32_t WHITE_SUN = 0x0001F323;
const char32_t WHITE_SUN_WITH_SMALL_CLOUD = 0x0001F324;
const char32_t WHITE_SUN_BEHIND_CLOUD = 0x0001F325;
const char32_t WHITE_SUN_BEHIND_CLOUD_WITH_RAIN = 0x0001F326;
const char32_t CLOUD_WITH_RAIN = 0x0001F327;
const char32_t CLOUD_WITH_SNOW = 0x0001F328;
const char32_t CLOUD_WITH_LIGHTNING = 0x0001F329;
const char32_t CLOUD_WITH_TORNADO = 0x0001F32A;
const char32_t FOG = 0x0001F32B;
const char32_t WIND_BLOWING_FACE = 0x0001F32C;
const char32_t HOT_DOG = 0x0001F32D;
const char32_t TACO = 0x0001F32E;
const char32_t BURRITO = 0x0001F32F;
const char32_t CHESTNUT = 0x0001F330;
const char32_t SEEDLING = 0x0001F331;
const char32_t EVERGREEN_TREE = 0x0001F332;
const char32_t DECIDUOUS_TREE = 0x0001F333;
const char32_t PALM_TREE = 0x0001F334;
const char32_t CACTUS = 0x0001F335;
const char32_t HOT_PEPPER = 0x0001F336;
const char32_t TULIP = 0x0001F337;
const char32_t CHERRY_BLOSSOM = 0x0001F338;
const char32_t ROSE = 0x0001F339;
const char32_t HIBISCUS = 0x0001F33A;
const char32_t SUNFLOWER = 0x0001F33B;
const char32_t BLOSSOM = 0x0001F33C;
const char32_t EAR_OF_MAIZE = 0x0001F33D;
const char32_t EAR_OF_RICE = 0x0001F33E;
const char32_t HERB = 0x0001F33F;
const char32_t FOUR_LEAF_CLOVER = 0x0001F340;
const char32_t MAPLE_LEAF = 0x0001F341;
const char32_t FALLEN_LEAF = 0x0001F342;
const char32_t LEAF_FLUTTERING_IN_WIND = 0x0001F343;
const char32_t MUSHROOM = 0x0001F344;
const char32_t TOMATO = 0x0001F345;
const char32_t AUBERGINE = 0x0001F346;
const char32_t GRAPES = 0x0001F347;
const char32_t MELON = 0x0001F348;
const char32_t WATERMELON = 0x0001F349;
const char32_t TANGERINE = 0x0001F34A;
const char32_t LEMON = 0x0001F34B;
const char32_t BANANA = 0x0001F34C;
const char32_t PINEAPPLE = 0x0001F34D;
const char32_t RED_APPLE = 0x0001F34E;
const char32_t GREEN_APPLE = 0x0001F34F;
const char32_t PEAR = 0x0001F350;
const char32_t PEACH = 0x0001F351;
const char32_t CHERRIES = 0x0001F352;
const char32_t STRAWBERRY = 0x0001F353;
const char32_t HAMBURGER = 0x0001F354;
const char32_t SLICE_OF_PIZZA = 0x0001F355;
const char32_t MEAT_ON_BONE = 0x0001F356;
const char32_t POULTRY_LEG = 0x0001F357;
const char32_t RICE_CRACKER = 0x0001F358;
const char32_t RICE_BALL = 0x0001F359;
const char32_t COOKED_RICE = 0x0001F35A;
const char32_t CURRY_AND_RICE = 0x0001F35B;
const char32_t STEAMING_BOWL = 0x0001F35C;
const char32_t SPAGHETTI = 0x0001F35D;
const char32_t BREAD = 0x0001F35E;
const char32_t FRENCH_FRIES = 0x0001F35F;
const char32_t ROASTED_SWEET_POTATO = 0x0001F360;
const char32_t DANGO = 0x0001F361;
const char32_t ODEN = 0x0001F362;
const char32_t SUSHI = 0x0001F363;
const char32_t FRIED_SHRIMP = 0x0001F364;
const char32_t FISH_CAKE_WITH_SWIRL_DESIGN = 0x0001F365;
const char32_t SOFT_ICE_CREAM = 0x0001F366;
const char32_t SHAVED_ICE = 0x0001F367;
const char32_t ICE_CREAM = 0x0001F368;
const char32_t DOUGHNUT = 0x0001F369;
const char32_t COOKIE = 0x0001F36A;
const char32_t CHOCOLATE_BAR = 0x0001F36B;
const char32_t CANDY = 0x0001F36C;
const char32_t LOLLIPOP = 0x0001F36D;
const char32_t CUSTARD = 0x0001F36E;
const char32_t HONEY_POT = 0x0001F36F;
const char32_t SHORTCAKE = 0x0001F370;
const char32_t BENTO_BOX = 0x0001F371;
const char32_t POT_OF_FOOD = 0x0001F372;
const char32_t COOKING = 0x0001F373;
const char32_t FORK_AND_KNIFE = 0x0001F374;
const char32_t TEACUP_WITHOUT_HANDLE = 0x0001F375;
const char32_t SAKE_BOTTLE_AND_CUP = 0x0001F376;
const char32_t WINE_GLASS = 0x0001F377;
const char32_t COCKTAIL_GLASS = 0x0001F378;
const char32_t TROPICAL_DRINK = 0x0001F379;
const char32_t BEER_MUG = 0x0001F37A;
const char32_t CLINKING_BEER_MUGS = 0x0001F37B;
const char32_t BABY_BOTTLE = 0x0001F37C;
const char32_t FORK_AND_KNIFE_WITH_PLATE = 0x0001F37D;
const char32_t BOTTLE_WITH_POPPING_CORK = 0x0001F37E;
const char32_t POPCORN = 0x0001F37F;
const char32_t RIBBON = 0x0001F380;
const char32_t WRAPPED_PRESENT = 0x0001F381;
const char32_t BIRTHDAY_CAKE = 0x0001F382;
const char32_t JACK_O_LANTERN = 0x0001F383;
const char32_t CHRISTMAS_TREE = 0x0001F384;
const char32_t FATHER_CHRISTMAS = 0x0001F385;
const char32_t FIREWORKS = 0x0001F386;
const char32_t FIREWORK_SPARKLER = 0x0001F387;
const char32_t BALLOON = 0x0001F388;
const char32_t PARTY_POPPER = 0x0001F389;
const char32_t CONFETTI_BALL = 0x0001F38A;
const char32_t TANABATA_TREE = 0x0001F38B;
const char32_t CROSSED_FLAGS = 0x0001F38C;
const char32_t PINE_DECORATION = 0x0001F38D;
const char32_t JAPANESE_DOLLS = 0x0001F38E;
const char32_t CARP_STREAMER = 0x0001F38F;
const char32_t WIND_CHIME = 0x0001F390;
const char32_t MOON_VIEWING_CEREMONY = 0x0001F391;
const char32_t SCHOOL_SATCHEL = 0x0001F392;
const char32_t GRADUATION_CAP = 0x0001F393;
const char32_t HEART_WITH_TIP_ON_THE_LEFT = 0x0001F394;
const char32_t BOUQUET_OF_FLOWERS = 0x0001F395;
const char32_t MILITARY_MEDAL = 0x0001F396;
const char32_t REMINDER_RIBBON = 0x0001F397;
const char32_t MUSICAL_KEYBOARD_WITH_JACKS = 0x0001F398;
const char32_t STUDIO_MICROPHONE = 0x0001F399;
const char32_t LEVEL_SLIDER = 0x0001F39A;
const char32_t CONTROL_KNOBS = 0x0001F39B;
const char32_t BEAMED_ASCENDING_MUSICAL_NOTES = 0x0001F39C;
const char32_t BEAMED_DESCENDING_MUSICAL_NOTES = 0x0001F39D;
const char32_t FILM_FRAMES = 0x0001F39E;
const char32_t ADMISSION_TICKETS = 0x0001F39F;
const char32_t CAROUSEL_HORSE = 0x0001F3A0;
const char32_t FERRIS_WHEEL = 0x0001F3A1;
const char32_t ROLLER_COASTER = 0x0001F3A2;
const char32_t FISHING_POLE_AND_FISH = 0x0001F3A3;
const char32_t MICROPHONE = 0x0001F3A4;
const char32_t MOVIE_CAMERA = 0x0001F3A5;
const char32_t CINEMA = 0x0001F3A6;
const char32_t HEADPHONE = 0x0001F3A7;
const char32_t ARTIST_PALETTE = 0x0001F3A8;
const char32_t TOP_HAT = 0x0001F3A9;
const char32_t CIRCUS_TENT = 0x0001F3AA;
const char32_t TICKET = 0x0001F3AB;
const char32_t CLAPPER_BOARD = 0x0001F3AC;
const char32_t PERFORMING_ARTS = 0x0001F3AD;
const char32_t VIDEO_GAME = 0x0001F3AE;
const char32_t DIRECT_HIT = 0x0001F3AF;
const char32_t SLOT_MACHINE = 0x0001F3B0;
const char32_t BILLIARDS = 0x0001F3B1;
const char32_t GAME_DIE = 0x0001F3B2;
const char32_t BOWLING = 0x0001F3B3;
const char32_t FLOWER_PLAYING_CARDS = 0x0001F3B4;
const char32_t MUSICAL_NOTE = 0x0001F3B5;
const char32_t MULTIPLE_MUSICAL_NOTES = 0x0001F3B6;
const char32_t SAXOPHONE = 0x0001F3B7;
const char32_t GUITAR = 0x0001F3B8;
const char32_t MUSICAL_KEYBOARD = 0x0001F3B9;
const char32_t TRUMPET = 0x0001F3BA;
const char32_t VIOLIN = 0x0001F3BB;
const char32_t MUSICAL_SCORE = 0x0001F3BC;
const char32_t RUNNING_SHIRT_WITH_SASH = 0x0001F3BD;
const char32_t TENNIS_RACQUET_AND_BALL = 0x0001F3BE;
const char32_t SKI_AND_SKI_BOOT = 0x0001F3BF;
const char32_t BASKETBALL_AND_HOOP = 0x0001F3C0;
const char32_t CHEQUERED_FLAG = 0x0001F3C1;
const char32_t SNOWBOARDER = 0x0001F3C2;
const char32_t RUNNER = 0x0001F3C3;
const char32_t SURFER = 0x0001F3C4;
const char32_t SPORTS_MEDAL = 0x0001F3C5;
const char32_t TROPHY = 0x0001F3C6;
const char32_t HORSE_RACING = 0x0001F3C7;
const char32_t AMERICAN_FOOTBALL = 0x0001F3C8;
const char32_t RUGBY_FOOTBALL = 0x0001F3C9;
const char32_t SWIMMER = 0x0001F3CA;
const char32_t WEIGHT_LIFTER = 0x0001F3CB;
const char32_t GOLFER = 0x0001F3CC;
const char32_t RACING_MOTORCYCLE = 0x0001F3CD;
const char32_t RACING_CAR = 0x0001F3CE;
const char32_t CRICKET_BAT_AND_BALL = 0x0001F3CF;
const char32_t VOLLEYBALL = 0x0001F3D0;
const char32_t FIELD_HOCKEY_STICK_AND_BALL = 0x0001F3D1;
const char32_t ICE_HOCKEY_STICK_AND_PUCK = 0x0001F3D2;
const char32_t TABLE_TENNIS_PADDLE_AND_BALL = 0x0001F3D3;
const char32_t SNOW_CAPPED_MOUNTAIN = 0x0001F3D4;
const char32_t CAMPING = 0x0001F3D5;
const char32_t BEACH_WITH_UMBRELLA = 0x0001F3D6;
const char32_t BUILDING_CONSTRUCTION = 0x0001F3D7;
const char32_t HOUSE_BUILDINGS = 0x0001F3D8;
const char32_t CITYSCAPE = 0x0001F3D9;
const char32_t DERELICT_HOUSE_BUILDING = 0x0001F3DA;
const char32_t CLASSICAL_BUILDING = 0x0001F3DB;
const char32_t DESERT = 0x0001F3DC;
const char32_t DESERT_ISLAND = 0x0001F3DD;
const char32_t NATIONAL_PARK = 0x0001F3DE;
const char32_t STADIUM = 0x0001F3DF;
const char32_t HOUSE_BUILDING = 0x0001F3E0;
const char32_t HOUSE_WITH_GARDEN = 0x0001F3E1;
const char32_t OFFICE_BUILDING = 0x0001F3E2;
const char32_t JAPANESE_POST_OFFICE = 0x0001F3E3;
const char32_t EUROPEAN_POST_OFFICE = 0x0001F3E4;
const char32_t HOSPITAL = 0x0001F3E5;
const char32_t BANK = 0x0001F3E6;
const char32_t AUTOMATED_TELLER_MACHINE = 0x0001F3E7;
const char32_t HOTEL = 0x0001F3E8;
const char32_t LOVE_HOTEL = 0x0001F3E9;
const char32_t CONVENIENCE_STORE = 0x0001F3EA;
const char32_t SCHOOL = 0x0001F3EB;
const char32_t DEPARTMENT_STORE = 0x0001F3EC;
const char32_t FACTORY = 0x0001F3ED;
const char32_t IZAKAYA_LANTERN = 0x0001F3EE;
const char32_t JAPANESE_CASTLE = 0x0001F3EF;
const char32_t EUROPEAN_CASTLE = 0x0001F3F0;
const char32_t WHITE_PENNANT = 0x0001F3F1;
const char32_t BLACK_PENNANT = 0x0001F3F2;
const char32_t WAVING_WHITE_FLAG = 0x0001F3F3;
const char32_t WAVING_BLACK_FLAG = 0x0001F3F4;
const char32_t ROSETTE = 0x0001F3F5;
const char32_t BLACK_ROSETTE = 0x0001F3F6;
const char32_t LABEL = 0x0001F3F7;
const char32_t BADMINTON_RACQUET_AND_SHUTTLECOCK = 0x0001F3F8;
const char32_t BOW_AND_ARROW = 0x0001F3F9;
const char32_t AMPHORA = 0x0001F3FA;
const char32_t EMOJI_MODIFIER_FITZPATRICK_TYPE_1_2 = 0x0001F3FB;
const char32_t EMOJI_MODIFIER_FITZPATRICK_TYPE_3 = 0x0001F3FC;
const char32_t EMOJI_MODIFIER_FITZPATRICK_TYPE_4 = 0x0001F3FD;
const char32_t EMOJI_MODIFIER_FITZPATRICK_TYPE_5 = 0x0001F3FE;
const char32_t EMOJI_MODIFIER_FITZPATRICK_TYPE_6 = 0x0001F3FF;
const char32_t RAT = 0x0001F400;
const char32_t MOUSE = 0x0001F401;
const char32_t OX = 0x0001F402;
const char32_t WATER_BUFFALO = 0x0001F403;
const char32_t COW = 0x0001F404;
const char32_t TIGER = 0x0001F405;
const char32_t LEOPARD = 0x0001F406;
const char32_t RABBIT = 0x0001F407;
const char32_t CAT = 0x0001F408;
const char32_t DRAGON = 0x0001F409;
const char32_t CROCODILE = 0x0001F40A;
const char32_t WHALE = 0x0001F40B;
const char32_t SNAIL = 0x0001F40C;
const char32_t SNAKE = 0x0001F40D;
const char32_t HORSE = 0x0001F40E;
const char32_t RAM = 0x0001F40F;
const char32_t GOAT = 0x0001F410;
const char32_t SHEEP = 0x0001F411;
const char32_t MONKEY = 0x0001F412;
const char32_t ROOSTER = 0x0001F413;
const char32_t CHICKEN = 0x0001F414;
const char32_t DOG = 0x0001F415;
const char32_t PIG = 0x0001F416;
const char32_t BOAR = 0x0001F417;
const char32_t ELEPHANT = 0x0001F418;
const char32_t OCTOPUS = 0x0001F419;
const char32_t SPIRAL_SHELL = 0x0001F41A;
const char32_t BUG = 0x0001F41B;
const char32_t ANT = 0x0001F41C;
const char32_t HONEYBEE = 0x0001F41D;
const char32_t LADY_BEETLE = 0x0001F41E;
const char32_t FISH = 0x0001F41F;
const char32_t TROPICAL_FISH = 0x0001F420;
const char32_t BLOWFISH = 0x0001F421;
const char32_t TURTLE = 0x0001F422;
const char32_t HATCHING_CHICK = 0x0001F423;
const char32_t BABY_CHICK = 0x0001F424;
const char32_t FRONT_FACING_BABY_CHICK = 0x0001F425;
const char32_t BIRD = 0x0001F426;
const char32_t PENGUIN = 0x0001F427;
const char32_t KOALA = 0x0001F428;
const char32_t POODLE = 0x0001F429;
const char32_t DROMEDARY_CAMEL = 0x0001F42A;
const char32_t BACTRIAN_CAMEL = 0x0001F42B;
const char32_t DOLPHIN = 0x0001F42C;
const char32_t MOUSE_FACE = 0x0001F42D;
const char32_t COW_FACE = 0x0001F42E;
const char32_t TIGER_FACE = 0x0001F42F;
const char32_t RABBIT_FACE = 0x0001F430;
const char32_t CAT_FACE = 0x0001F431;
const char32_t DRAGON_FACE = 0x0001F432;
const char32_t SPOUTING_WHALE = 0x0001F433;
const char32_t HORSE_FACE = 0x0001F434;
const char32_t MONKEY_FACE = 0x0001F435;
const char32_t DOG_FACE = 0x0001F436;
const char32_t PIG_FACE = 0x0001F437;
const char32_t FROG_FACE = 0x0001F438;
const char32_t HAMSTER_FACE = 0x0001F439;
const char32_t WOLF_FACE = 0x0001F43A;
const char32_t BEAR_FACE = 0x0001F43B;
const char32_t PANDA_FACE = 0x0001F43C;
const char32_t PIG_NOSE = 0x0001F43D;
const char32_t PAW_PRINTS = 0x0001F43E;
const char32_t CHIPMUNK = 0x0001F43F;
const char32_t EYES = 0x0001F440;
const char32_t EYE = 0x0001F441;
const char32_t EAR = 0x0001F442;
const char32_t NOSE = 0x0001F443;
const char32_t MOUTH = 0x0001F444;
const char32_t TONGUE = 0x0001F445;
const char32_t WHITE_UP_POINTING_BACKHAND_INDEX = 0x0001F446;
const char32_t WHITE_DOWN_POINTING_BACKHAND_INDEX = 0x0001F447;
const char32_t WHITE_LEFT_POINTING_BACKHAND_INDEX = 0x0001F448;
const char32_t WHITE_RIGHT_POINTING_BACKHAND_INDEX = 0x0001F449;
const char32_t FISTED_HAND_SIGN = 0x0001F44A;
const char32_t WAVING_HAND_SIGN = 0x0001F44B;
const char32_t OK_HAND_SIGN = 0x0001F44C;
const char32_t THUMBS_UP_SIGN = 0x0001F44D;
const char32_t THUMBS_DOWN_SIGN = 0x0001F44E;
const char32_t CLAPPING_HANDS_SIGN = 0x0001F44F;
const char32_t OPEN_HANDS_SIGN = 0x0001F450;
const char32_t CROWN = 0x0001F451;
const char32_t WOMANS_HAT = 0x0001F452;
const char32_t EYEGLASSES = 0x0001F453;
const char32_t NECKTIE = 0x0001F454;
const char32_t T_SHIRT = 0x0001F455;
const char32_t JEANS = 0x0001F456;
const char32_t DRESS = 0x0001F457;
const char32_t KIMONO = 0x0001F458;
const char32_t BIKINI = 0x0001F459;
const char32_t WOMANS_CLOTHES = 0x0001F45A;
const char32_t PURSE = 0x0001F45B;
const char32_t HANDBAG = 0x0001F45C;
const char32_t POUCH = 0x0001F45D;
const char32_t MANS_SHOE = 0x0001F45E;
const char32_t ATHLETIC_SHOE = 0x0001F45F;
const char32_t HIGH_HEELED_SHOE = 0x0001F460;
const char32_t WOMANS_SANDAL = 0x0001F461;
const char32_t WOMANS_BOOTS = 0x0001F462;
const char32_t FOOTPRINTS = 0x0001F463;
const char32_t BUST_IN_SILHOUETTE = 0x0001F464;
const char32_t BUSTS_IN_SILHOUETTE = 0x0001F465;
const char32_t BOY = 0x0001F466;
const char32_t GIRL = 0x0001F467;
const char32_t MAN = 0x0001F468;
const char32_t WOMAN = 0x0001F469;
const char32_t FAMILY = 0x0001F46A;
const char32_t MAN_AND_WOMAN_HOLDING_HANDS = 0x0001F46B;
const char32_t TWO_MEN_HOLDING_HANDS = 0x0001F46C;
const char32_t TWO_WOMEN_HOLDING_HANDS = 0x0001F46D;
const char32_t POLICE_OFFICER = 0x0001F46E;
const char32_t WOMAN_WITH_BUNNY_EARS = 0x0001F46F;
const char32_t BRIDE_WITH_VEIL = 0x0001F470;
const char32_t PERSON_WITH_BLOND_HAIR = 0x0001F471;
const char32_t MAN_WITH_GUA_PI_MAO = 0x0001F472;
const char32_t MAN_WITH_TURBAN = 0x0001F473;
const char32_t OLDER_MAN = 0x0001F474;
const char32_t OLDER_WOMAN = 0x0001F475;
const char32_t BABY = 0x0001F476;
const char32_t CONSTRUCTION_WORKER = 0x0001F477;
const char32_t PRINCESS = 0x0001F478;
const char32_t JAPANESE_OGRE = 0x0001F479;
const char32_t JAPANESE_GOBLIN = 0x0001F47A;
const char32_t GHOST = 0x0001F47B;
const char32_t BABY_ANGEL = 0x0001F47C;
const char32_t EXTRATERRESTRIAL_ALIEN = 0x0001F47D;
const char32_t ALIEN_MONSTER = 0x0001F47E;
const char32_t IMP = 0x0001F47F;
const char32_t SKULL = 0x0001F480;
const char32_t INFORMATION_DESK_PERSON = 0x0001F481;
const char32_t GUARDSMAN = 0x0001F482;
const char32_t DANCER = 0x0001F483;
const char32_t LIPSTICK = 0x0001F484;
const char32_t NAIL_POLISH = 0x0001F485;
const char32_t FACE_MASSAGE = 0x0001F486;
const char32_t HAIRCUT = 0x0001F487;
const char32_t BARBER_POLE = 0x0001F488;
const char32_t SYRINGE = 0x0001F489;
const char32_t PILL = 0x0001F48A;
const char32_t KISS_MARK = 0x0001F48B;
const char32_t LOVE_LETTER = 0x0001F48C;
const char32_t RING = 0x0001F48D;
const char32_t GEM_STONE = 0x0001F48E;
const char32_t KISS = 0x0001F48F;
const char32_t BOUQUET = 0x0001F490;
const char32_t COUPLE_WITH_HEART = 0x0001F491;
const char32_t WEDDING = 0x0001F492;
const char32_t BEATING_HEART = 0x0001F493;
const char32_t BROKEN_HEART = 0x0001F494;
const char32_t TWO_HEARTS = 0x0001F495;
const char32_t SPARKLING_HEART = 0x0001F496;
const char32_t GROWING_HEART = 0x0001F497;
const char32_t HEART_WITH_ARROW = 0x0001F498;
const char32_t BLUE_HEART = 0x0001F499;
const char32_t GREEN_HEART = 0x0001F49A;
const char32_t YELLOW_HEART = 0x0001F49B;
const char32_t PURPLE_HEART = 0x0001F49C;
const char32_t HEART_WITH_RIBBON = 0x0001F49D;
const char32_t REVOLVING_HEARTS = 0x0001F49E;
const char32_t HEART_DECORATION = 0x0001F49F;
const char32_t DIAMOND_SHAPE_WITH_A_DOT_INSIDE = 0x0001F4A0;
const char32_t ELECTRIC_LIGHT_BULB = 0x0001F4A1;
const char32_t ANGER_SYMBOL = 0x0001F4A2;
const char32_t BOMB = 0x0001F4A3;
const char32_t SLEEPING_SYMBOL = 0x0001F4A4;
const char32_t COLLISION_SYMBOL = 0x0001F4A5;
const char32_t SPLASHING_SWEAT_SYMBOL = 0x0001F4A6;
const char32_t DROPLET = 0x0001F4A7;
const char32_t DASH_SYMBOL = 0x0001F4A8;
const char32_t PILE_OF_POO = 0x0001F4A9;
const char32_t FLEXED_BICEPS = 0x0001F4AA;
const char32_t DIZZY_SYMBOL = 0x0001F4AB;
const char32_t SPEECH_BALLOON = 0x0001F4AC;
const char32_t THOUGHT_BALLOON = 0x0001F4AD;
const char32_t WHITE_FLOWER = 0x0001F4AE;
const char32_t HUNDRED_POINTS_SYMBOL = 0x0001F4AF;
const char32_t MONEY_BAG = 0x0001F4B0;
const char32_t CURRENCY_EXCHANGE = 0x0001F4B1;
const char32_t HEAVY_DOLLAR_SIGN = 0x0001F4B2;
const char32_t CREDIT_CARD = 0x0001F4B3;
const char32_t BANKNOTE_WITH_YEN_SIGN = 0x0001F4B4;
const char32_t BANKNOTE_WITH_DOLLAR_SIGN = 0x0001F4B5;
const char32_t BANKNOTE_WITH_EURO_SIGN = 0x0001F4B6;
const char32_t BANKNOTE_WITH_POUND_SIGN = 0x0001F4B7;
const char32_t MONEY_WITH_WINGS = 0x0001F4B8;
const char32_t CHART_WITH_UPWARDS_TREND_AND_YEN_SIGN = 0x0001F4B9;
const char32_t SEAT = 0x0001F4BA;
const char32_t PERSONAL_COMPUTER = 0x0001F4BB;
const char32_t BRIEFCASE = 0x0001F4BC;
const char32_t MINIDISC = 0x0001F4BD;
const char32_t FLOPPY_DISK = 0x0001F4BE;
const char32_t OPTICAL_DISC = 0x0001F4BF;
const char32_t DVD = 0x0001F4C0;
const char32_t FILE_FOLDER = 0x0001F4C1;
const char32_t OPEN_FILE_FOLDER = 0x0001F4C2;
const char32_t PAGE_WITH_CURL = 0x0001F4C3;
const char32_t PAGE_FACING_UP = 0x0001F4C4;
const char32_t CALENDAR = 0x0001F4C5;
const char32_t TEAR_OFF_CALENDAR = 0x0001F4C6;
const char32_t CARD_INDEX = 0x0001F4C7;
const char32_t CHART_WITH_UPWARDS_TREND = 0x0001F4C8;
const char32_t CHART_WITH_DOWNWARDS_TREND = 0x0001F4C9;
const char32_t BAR_CHART = 0x0001F4CA;
const char32_t CLIPBOARD = 0x0001F4CB;
const char32_t PUSHPIN = 0x0001F4CC;
const char32_t ROUND_PUSHPIN = 0x0001F4CD;
const char32_t PAPERCLIP = 0x0001F4CE;
const char32_t STRAIGHT_RULER = 0x0001F4CF;
const char32_t TRIANGULAR_RULER = 0x0001F4D0;
const char32_t BOOKMARK_TABS = 0x0001F4D1;
const char32_t LEDGER = 0x0001F4D2;
const char32_t NOTEBOOK = 0x0001F4D3;
const char32_t NOTEBOOK_WITH_DECORATIVE_COVER = 0x0001F4D4;
const char32_t CLOSED_BOOK = 0x0001F4D5;
const char32_t OPEN_BOOK = 0x0001F4D6;
const char32_t GREEN_BOOK = 0x0001F4D7;
const char32_t BLUE_BOOK = 0x0001F4D8;
const char32_t ORANGE_BOOK = 0x0001F4D9;
const char32_t BOOKS = 0x0001F4DA;
const char32_t NAME_BADGE = 0x0001F4DB;
const char32_t SCROLL = 0x0001F4DC;
const char32_t MEMO = 0x0001F4DD;
const char32_t TELEPHONE_RECEIVER = 0x0001F4DE;
const char32_t PAGER = 0x0001F4DF;
const char32_t FAX_MACHINE = 0x0001F4E0;
const char32_t SATELLITE_ANTENNA = 0x0001F4E1;
const char32_t PUBLIC_ADDRESS_LOUDSPEAKER = 0x0001F4E2;
const char32_t CHEERING_MEGAPHONE = 0x0001F4E3;
const char32_t OUTBOX_TRAY = 0x0001F4E4;
const char32_t INBOX_TRAY = 0x0001F4E5;
const char32_t PACKAGE = 0x0001F4E6;
const char32_t E_MAIL_SYMBOL = 0x0001F4E7;
const char32_t INCOMING_ENVELOPE = 0x0001F4E8;
const char32_t ENVELOPE_WITH_DOWNWARDS_ARROW_ABOVE = 0x0001F4E9;
const char32_t CLOSED_MAILBOX_WITH_LOWERED_FLAG = 0x0001F4EA;
const char32_t CLOSED_MAILBOX_WITH_RAISED_FLAG = 0x0001F4EB;
const char32_t OPEN_MAILBOX_WITH_RAISED_FLAG = 0x0001F4EC;
const char32_t OPEN_MAILBOX_WITH_LOWERED_FLAG = 0x0001F4ED;
const char32_t POSTBOX = 0x0001F4EE;
const char32_t POSTAL_HORN = 0x0001F4EF;
const char32_t NEWSPAPER = 0x0001F4F0;
const char32_t MOBILE_PHONE = 0x0001F4F1;
const char32_t MOBILE_PHONE_WITH_RIGHTWARDS_ARROW_AT_LEFT = 0x0001F4F2;
const char32_t VIBRATION_MODE = 0x0001F4F3;
const char32_t MOBILE_PHONE_OFF = 0x0001F4F4;
const char32_t NO_MOBILE_PHONES = 0x0001F4F5;
const char32_t ANTENNA_WITH_BARS = 0x0001F4F6;
const char32_t CAMERA = 0x0001F4F7;
const char32_t CAMERA_WITH_FLASH = 0x0001F4F8;
const char32_t VIDEO_CAMERA = 0x0001F4F9;
const char32_t TELEVISION = 0x0001F4FA;
const char32_t RADIO = 0x0001F4FB;
const char32_t VIDEOCASSETTE = 0x0001F4FC;
const char32_t FILM_PROJECTOR = 0x0001F4FD;
const char32_t PORTABLE_STEREO = 0x0001F4FE;
const char32_t PRAYER_BEADS = 0x0001F4FF;
const char32_t TWISTED_RIGHTWARDS_ARROWS = 0x0001F500;
const char32_t CLOCKWISE_RIGHTWARDS_AND_LEFTWARDS_OPEN_CIRCLE_ARROWS = 0x0001F501;
const char32_t CLOCKWISE_RIGHTWARDS_AND_LEFTWARDS_OPEN_CIRCLE_ARROWS_WITH_CIRCLED_ONE_OVERLAY = 0x0001F502;
const char32_t CLOCKWISE_DOWNWARDS_AND_UPWARDS_OPEN_CIRCLE_ARROWS = 0x0001F503;
const char32_t ANTICLOCKWISE_DOWNWARDS_AND_UPWARDS_OPEN_CIRCLE_ARROWS = 0x0001F504;
const char32_t LOW_BRIGHTNESS_SYMBOL = 0x0001F505;
const char32_t HIGH_BRIGHTNESS_SYMBOL = 0x0001F506;
const char32_t SPEAKER_WITH_CANCELLATION_STROKE = 0x0001F507;
const char32_t SPEAKER = 0x0001F508;
const char32_t SPEAKER_WITH_ONE_SOUND_WAVE = 0x0001F509;
const char32_t SPEAKER_WITH_THREE_SOUND_WAVES = 0x0001F50A;
const char32_t BATTERY = 0x0001F50B;
const char32_t ELECTRIC_PLUG = 0x0001F50C;
const char32_t LEFT_POINTING_MAGNIFYING_GLASS = 0x0001F50D;
const char32_t RIGHT_POINTING_MAGNIFYING_GLASS = 0x0001F50E;
const char32_t LOCK_WITH_INK_PEN = 0x0001F50F;
const char32_t CLOSED_LOCK_WITH_KEY = 0x0001F510;
const char32_t KEY = 0x0001F511;
const char32_t LOCK = 0x0001F512;
const char32_t OPEN_LOCK = 0x0001F513;
const char32_t BELL = 0x0001F514;
const char32_t BELL_WITH_CANCELLATION_STROKE = 0x0001F515;
const char32_t BOOKMARK = 0x0001F516;
const char32_t LINK_SYMBOL = 0x0001F517;
const char32_t RADIO_BUTTON = 0x0001F518;
const char32_t BACK_WITH_LEFTWARDS_ARROW_ABOVE = 0x0001F519;
const char32_t END_WITH_LEFTWARDS_ARROW_ABOVE = 0x0001F51A;
const char32_t ON_WITH_EXCLAMATION_MARK_WITH_LEFT_RIGHT_ARROW_ABOVE = 0x0001F51B;
const char32_t SOON_WITH_RIGHTWARDS_ARROW_ABOVE = 0x0001F51C;
const char32_t TOP_WITH_UPWARDS_ARROW_ABOVE = 0x0001F51D;
const char32_t NO_ONE_UNDER_EIGHTEEN_SYMBOL = 0x0001F51E;
const char32_t KEYCAP_TEN = 0x0001F51F;
const char32_t INPUT_SYMBOL_FOR_LATIN_CAPITAL_LETTERS = 0x0001F520;
const char32_t INPUT_SYMBOL_FOR_LATIN_SMALL_LETTERS = 0x0001F521;
const char32_t INPUT_SYMBOL_FOR_NUMBERS = 0x0001F522;
const char32_t INPUT_SYMBOL_FOR_SYMBOLS = 0x0001F523;
const char32_t INPUT_SYMBOL_FOR_LATIN_LETTERS = 0x0001F524;
const char32_t FIRE = 0x0001F525;
const char32_t ELECTRIC_TORCH = 0x0001F526;
const char32_t WRENCH = 0x0001F527;
const char32_t HAMMER = 0x0001F528;
const char32_t NUT_AND_BOLT = 0x0001F529;
const char32_t HOCHO = 0x0001F52A;
const char32_t PISTOL = 0x0001F52B;
const char32_t MICROSCOPE = 0x0001F52C;
const char32_t TELESCOPE = 0x0001F52D;
const char32_t CRYSTAL_BALL = 0x0001F52E;
const char32_t SIX_POINTED_STAR_WITH_MIDDLE_DOT = 0x0001F52F;
const char32_t JAPANESE_SYMBOL_FOR_BEGINNER = 0x0001F530;
const char32_t TRIDENT_EMBLEM = 0x0001F531;
const char32_t BLACK_SQUARE_BUTTON = 0x0001F532;
const char32_t WHITE_SQUARE_BUTTON = 0x0001F533;
const char32_t LARGE_RED_CIRCLE = 0x0001F534;
const char32_t LARGE_BLUE_CIRCLE = 0x0001F535;
const char32_t LARGE_ORANGE_DIAMOND = 0x0001F536;
const char32_t LARGE_BLUE_DIAMOND = 0x0001F537;
const char32_t SMALL_ORANGE_DIAMOND = 0x0001F538;
const char32_t SMALL_BLUE_DIAMOND = 0x0001F539;
const char32_t UP_POINTING_RED_TRIANGLE = 0x0001F53A;
const char32_t DOWN_POINTING_RED_TRIANGLE = 0x0001F53B;
const char32_t UP_POINTING_SMALL_RED_TRIANGLE = 0x0001F53C;
const char32_t DOWN_POINTING_SMALL_RED_TRIANGLE = 0x0001F53D;
const char32_t LOWER_RIGHT_SHADOWED_WHITE_CIRCLE = 0x0001F53E;
const char32_t UPPER_RIGHT_SHADOWED_WHITE_CIRCLE = 0x0001F53F;
const char32_t CIRCLED_CROSS_POMMEE = 0x0001F540;
const char32_t CROSS_POMMEE_WITH_HALF_CIRCLE_BELOW = 0x0001F541;
const char32_t CROSS_POMMEE = 0x0001F542;
const char32_t NOTCHED_LEFT_SEMICIRCLE_WITH_THREE_DOTS = 0x0001F543;
const char32_t NOTCHED_RIGHT_SEMICIRCLE_WITH_THREE_DOTS = 0x0001F544;
const char32_t SYMBOL_FOR_MARKS_CHAPTER = 0x0001F545;
const char32_t WHITE_LATIN_CROSS = 0x0001F546;
const char32_t HEAVY_LATIN_CROSS = 0x0001F547;
const char32_t CELTIC_CROSS = 0x0001F548;
const char32_t OM_SYMBOL = 0x0001F549;
const char32_t DOVE_OF_PEACE = 0x0001F54A;
const char32_t KAABA = 0x0001F54B;
const char32_t MOSQUE = 0x0001F54C;
const char32_t SYNAGOGUE = 0x0001F54D;
const char32_t MENORAH_WITH_NINE_BRANCHES = 0x0001F54E;
const char32_t BOWL_OF_HYGIEIA = 0x0001F54F;
const char32_t CLOCK_FACE_ONE_OCLOCK = 0x0001F550;
const char32_t CLOCK_FACE_TWO_OCLOCK = 0x0001F551;
const char32_t CLOCK_FACE_THREE_OCLOCK = 0x0001F552;
const char32_t CLOCK_FACE_FOUR_OCLOCK = 0x0001F553;
const char32_t CLOCK_FACE_FIVE_OCLOCK = 0x0001F554;
const char32_t CLOCK_FACE_SIX_OCLOCK = 0x0001F555;
const char32_t CLOCK_FACE_SEVEN_OCLOCK = 0x0001F556;
const char32_t CLOCK_FACE_EIGHT_OCLOCK = 0x0001F557;
const char32_t CLOCK_FACE_NINE_OCLOCK = 0x0001F558;
const char32_t CLOCK_FACE_TEN_OCLOCK = 0x0001F559;
const char32_t CLOCK_FACE_ELEVEN_OCLOCK = 0x0001F55A;
const char32_t CLOCK_FACE_TWELVE_OCLOCK = 0x0001F55B;
const char32_t CLOCK_FACE_ONE_THIRTY = 0x0001F55C;
const char32_t CLOCK_FACE_TWO_THIRTY = 0x0001F55D;
const char32_t CLOCK_FACE_THREE_THIRTY = 0x0001F55E;
const char32_t CLOCK_FACE_FOUR_THIRTY = 0x0001F55F;
const char32_t CLOCK_FACE_FIVE_THIRTY = 0x0001F560;
const char32_t CLOCK_FACE_SIX_THIRTY = 0x0001F561;
const char32_t CLOCK_FACE_SEVEN_THIRTY = 0x0001F562;
const char32_t CLOCK_FACE_EIGHT_THIRTY = 0x0001F563;
const char32_t CLOCK_FACE_NINE_THIRTY = 0x0001F564;
const char32_t CLOCK_FACE_TEN_THIRTY = 0x0001F565;
const char32_t CLOCK_FACE_ELEVEN_THIRTY = 0x0001F566;
const char32_t CLOCK_FACE_TWELVE_THIRTY = 0x0001F567;
const char32_t RIGHT_SPEAKER = 0x0001F568;
const char32_t RIGHT_SPEAKER_WITH_ONE_SOUND_WAVE = 0x0001F569;
const char32_t RIGHT_SPEAKER_WITH_THREE_SOUND_WAVES = 0x0001F56A;
const char32_t BULLHORN = 0x0001F56B;
const char32_t BULLHORN_WITH_SOUND_WAVES = 0x0001F56C;
const char32_t RINGING_BELL = 0x0001F56D;
const char32_t BOOK = 0x0001F56E;
const char32_t CANDLE = 0x0001F56F;
const char32_t MANTELPIECE_CLOCK = 0x0001F570;
const char32_t BLACK_SKULL_AND_CROSSBONES = 0x0001F571;
const char32_t NO_PIRACY = 0x0001F572;
const char32_t HOLE = 0x0001F573;
const char32_t MAN_IN_BUSINESS_SUIT_LEVITATING = 0x0001F574;
const char32_t SLEUTH_OR_SPY = 0x0001F575;
const char32_t DARK_SUNGLASSES = 0x0001F576;
const char32_t SPIDER = 0x0001F577;
const char32_t SPIDER_WEB = 0x0001F578;
const char32_t JOYSTICK = 0x0001F579;
const char32_t MAN_DANCING = 0x0001F57A;
const char32_t LEFT_HAND_TELEPHONE_RECEIVER = 0x0001F57B;
const char32_t TELEPHONE_RECEIVER_WITH_PAGE = 0x0001F57C;
const char32_t RIGHT_HAND_TELEPHONE_RECEIVER = 0x0001F57D;
const char32_t WHITE_TOUCHTONE_TELEPHONE = 0x0001F57E;
const char32_t BLACK_TOUCHTONE_TELEPHONE = 0x0001F57F;
const char32_t TELEPHONE_ON_TOP_OF_MODEM = 0x0001F580;
const char32_t CLAMSHELL_MOBILE_PHONE = 0x0001F581;
const char32_t BACK_OF_ENVELOPE = 0x0001F582;
const char32_t STAMPED_ENVELOPE = 0x0001F583;
const char32_t ENVELOPE_WITH_LIGHTNING = 0x0001F584;
const char32_t FLYING_ENVELOPE = 0x0001F585;
const char32_t PEN_OVER_STAMPED_ENVELOPE = 0x0001F586;
const char32_t LINKED_PAPERCLIPS = 0x0001F587;
const char32_t BLACK_PUSHPIN = 0x0001F588;
const char32_t LOWER_LEFT_PENCIL = 0x0001F589;
const char32_t LOWER_LEFT_BALLPOINT_PEN = 0x0001F58A;
const char32_t LOWER_LEFT_FOUNTAIN_PEN = 0x0001F58B;
const char32_t LOWER_LEFT_PAINTBRUSH = 0x0001F58C;
const char32_t LOWER_LEFT_CRAYON = 0x0001F58D;
const char32_t LEFT_WRITING_HAND = 0x0001F58E;
const char32_t TURNED_OK_HAND_SIGN = 0x0001F58F;
const char32_t RAISED_HAND_WITH_FINGERS_SPLAYED = 0x0001F590;
const char32_t REVERSED_RAISED_HAND_WITH_FINGERS_SPLAYED = 0x0001F591;
const char32_t REVERSED_THUMBS_UP_SIGN = 0x0001F592;
const char32_t REVERSED_THUMBS_DOWN_SIGN = 0x0001F593;
const char32_t REVERSED_VICTORY_HAND = 0x0001F594;
const char32_t REVERSED_HAND_WITH_MIDDLE_FINGER_EXTENDED = 0x0001F595;
const char32_t RAISED_HAND_WITH_PART_BETWEEN_MIDDLE_AND_RING_FINGERS = 0x0001F596;
const char32_t WHITE_DOWN_POINTING_LEFT_HAND_INDEX = 0x0001F597;
const char32_t SIDEWAYS_WHITE_LEFT_POINTING_INDEX = 0x0001F598;
const char32_t SIDEWAYS_WHITE_RIGHT_POINTING_INDEX = 0x0001F599;
const char32_t SIDEWAYS_BLACK_LEFT_POINTING_INDEX = 0x0001F59A;
const char32_t SIDEWAYS_BLACK_RIGHT_POINTING_INDEX = 0x0001F59B;
const char32_t BLACK_LEFT_POINTING_BACKHAND_INDEX = 0x0001F59C;
const char32_t BLACK_RIGHT_POINTING_BACKHAND_INDEX = 0x0001F59D;
const char32_t SIDEWAYS_WHITE_UP_POINTING_INDEX = 0x0001F59E;
const char32_t SIDEWAYS_WHITE_DOWN_POINTING_INDEX = 0x0001F59F;
const char32_t SIDEWAYS_BLACK_UP_POINTING_INDEX = 0x0001F5A0;
const char32_t SIDEWAYS_BLACK_DOWN_POINTING_INDEX = 0x0001F5A1;
const char32_t BLACK_UP_POINTING_BACKHAND_INDEX = 0x0001F5A2;
const char32_t BLACK_DOWN_POINTING_BACKHAND_INDEX = 0x0001F5A3;
const char32_t BLACK_HEART = 0x0001F5A4;
const char32_t DESKTOP_COMPUTER = 0x0001F5A5;
const char32_t KEYBOARD_AND_MOUSE = 0x0001F5A6;
const char32_t THREE_NETWORKED_COMPUTERS = 0x0001F5A7;
const char32_t PRINTER = 0x0001F5A8;
const char32_t POCKET_CALCULATOR = 0x0001F5A9;
const char32_t BLACK_HARD_SHELL_FLOPPY_DISK = 0x0001F5AA;
const char32_t WHITE_HARD_SHELL_FLOPPY_DISK = 0x0001F5AB;
const char32_t SOFT_SHELL_FLOPPY_DISK = 0x0001F5AC;
const char32_t TAPE_CARTRIDGE = 0x0001F5AD;
const char32_t WIRED_KEYBOARD = 0x0001F5AE;
const char32_t ONE_BUTTON_MOUSE = 0x0001F5AF;
const char32_t TWO_BUTTON_MOUSE = 0x0001F5B0;
const char32_t THREE_BUTTON_MOUSE = 0x0001F5B1;
const char32_t TRACKBALL = 0x0001F5B2;
const char32_t OLD_PERSONAL_COMPUTER = 0x0001F5B3;
const char32_t HARD_DISK = 0x0001F5B4;
const char32_t SCREEN = 0x0001F5B5;
const char32_t PRINTER_ICON = 0x0001F5B6;
const char32_t FAX_ICON = 0x0001F5B7;
const char32_t OPTICAL_DISC_ICON = 0x0001F5B8;
const char32_t DOCUMENT_WITH_TEXT = 0x0001F5B9;
const char32_t DOCUMENT_WITH_TEXT_AND_PICTURE = 0x0001F5BA;
const char32_t DOCUMENT_WITH_PICTURE = 0x0001F5BB;
const char32_t FRAME_WITH_PICTURE = 0x0001F5BC;
const char32_t FRAME_WITH_TILES = 0x0001F5BD;
const char32_t FRAME_WITH_AN_X = 0x0001F5BE;
const char32_t BLACK_FOLDER = 0x0001F5BF;
const char32_t FOLDER = 0x0001F5C0;
const char32_t OPEN_FOLDER = 0x0001F5C1;
const char32_t CARD_INDEX_DIVIDERS = 0x0001F5C2;
const char32_t CARD_FILE_BOX = 0x0001F5C3;
const char32_t FILE_CABINET = 0x0001F5C4;
const char32_t EMPTY_NOTE = 0x0001F5C5;
const char32_t EMPTY_NOTE_PAGE = 0x0001F5C6;
const char32_t EMPTY_NOTE_PAD = 0x0001F5C7;
const char32_t NOTE = 0x0001F5C8;
const char32_t NOTE_PAGE = 0x0001F5C9;
const char32_t NOTE_PAD = 0x0001F5CA;
const char32_t EMPTY_DOCUMENT = 0x0001F5CB;
const char32_t EMPTY_PAGE = 0x0001F5CC;
const char32_t EMPTY_PAGES = 0x0001F5CD;
const char32_t DOCUMENT = 0x0001F5CE;
const char32_t PAGE = 0x0001F5CF;
const char32_t PAGES = 0x0001F5D0;
const char32_t WASTEBASKET = 0x0001F5D1;
const char32_t SPIRAL_NOTE_PAD = 0x0001F5D2;
const char32_t SPIRAL_CALENDAR_PAD = 0x0001F5D3;
const char32_t DESKTOP_WINDOW = 0x0001F5D4;
const char32_t MINIMIZE = 0x0001F5D5;
const char32_t MAXIMIZE = 0x0001F5D6;
const char32_t OVERLAP = 0x0001F5D7;
const char32_t CLOCKWISE_RIGHT_AND_LEFT_SEMICIRCLE_ARROWS = 0x0001F5D8;
const char32_t CANCELLATION_X = 0x0001F5D9;
const char32_t INCREASE_FONT_SIZE_SYMBOL = 0x0001F5DA;
const char32_t DECREASE_FONT_SIZE_SYMBOL = 0x0001F5DB;
const char32_t COMPRESSION = 0x0001F5DC;
const char32_t OLD_KEY = 0x0001F5DD;
const char32_t ROLLED_UP_NEWSPAPER = 0x0001F5DE;
const char32_t PAGE_WITH_CIRCLED_TEXT = 0x0001F5DF;
const char32_t STOCK_CHART = 0x0001F5E0;
const char32_t DAGGER_KNIFE = 0x0001F5E1;
const char32_t LIPS = 0x0001F5E2;
const char32_t SPEAKING_HEAD_IN_SILHOUETTE = 0x0001F5E3;
const char32_t THREE_RAYS_ABOVE = 0x0001F5E4;
const char32_t THREE_RAYS_BELOW = 0x0001F5E5;
const char32_t THREE_RAYS_LEFT = 0x0001F5E6;
const char32_t THREE_RAYS_RIGHT = 0x0001F5E7;
const char32_t LEFT_SPEECH_BUBBLE = 0x0001F5E8;
const char32_t RIGHT_SPEECH_BUBBLE = 0x0001F5E9;
const char32_t TWO_SPEECH_BUBBLES = 0x0001F5EA;
const char32_t THREE_SPEECH_BUBBLES = 0x0001F5EB;
const char32_t LEFT_THOUGHT_BUBBLE = 0x0001F5EC;
const char32_t RIGHT_THOUGHT_BUBBLE = 0x0001F5ED;
const char32_t LEFT_ANGER_BUBBLE = 0x0001F5EE;
const char32_t RIGHT_ANGER_BUBBLE = 0x0001F5EF;
const char32_t MOOD_BUBBLE = 0x0001F5F0;
const char32_t LIGHTNING_MOOD_BUBBLE = 0x0001F5F1;
const char32_t LIGHTNING_MOOD = 0x0001F5F2;
const char32_t BALLOT_BOX_WITH_BALLOT = 0x0001F5F3;
const char32_t BALLOT_SCRIPT_X = 0x0001F5F4;
const char32_t BALLOT_BOX_WITH_SCRIPT_X = 0x0001F5F5;
const char32_t BALLOT_BOLD_SCRIPT_X = 0x0001F5F6;
const char32_t BALLOT_BOX_WITH_BOLD_SCRIPT_X = 0x0001F5F7;
const char32_t LIGHT_CHECK_MARK = 0x0001F5F8;
const char32_t BALLOT_BOX_WITH_BOLD_CHECK = 0x0001F5F9;
const char32_t WORLD_MAP = 0x0001F5FA;
const char32_t MOUNT_FUJI = 0x0001F5FB;
const char32_t TOKYO_TOWER = 0x0001F5FC;
const char32_t STATUE_OF_LIBERTY = 0x0001F5FD;
const char32_t SILHOUETTE_OF_JAPAN = 0x0001F5FE;
const char32_t MOYAI = 0x0001F5FF;
const char32_t GRINNING_FACE = 0x0001F600;
const char32_t GRINNING_FACE_WITH_SMILING_EYES = 0x0001F601;
const char32_t FACE_WITH_TEARS_OF_JOY = 0x0001F602;
const char32_t SMILING_FACE_WITH_OPEN_MOUTH = 0x0001F603;
const char32_t SMILING_FACE_WITH_OPEN_MOUTH_AND_SMILING_EYES = 0x0001F604;
const char32_t SMILING_FACE_WITH_OPEN_MOUTH_AND_COLD_SWEAT = 0x0001F605;
const char32_t SMILING_FACE_WITH_OPEN_MOUTH_AND_TIGHTLY_CLOSED_EYES = 0x0001F606;
const char32_t SMILING_FACE_WITH_HALO = 0x0001F607;
const char32_t SMILING_FACE_WITH_HORNS = 0x0001F608;
const char32_t WINKING_FACE = 0x0001F609;
const char32_t SMILING_FACE_WITH_SMILING_EYES = 0x0001F60A;
const char32_t FACE_SAVOURING_DELICIOUS_FOOD = 0x0001F60B;
const char32_t RELIEVED_FACE = 0x0001F60C;
const char32_t SMILING_FACE_WITH_HEART_SHAPED_EYES = 0x0001F60D;
const char32_t SMILING_FACE_WITH_SUNGLASSES = 0x0001F60E;
const char32_t SMIRKING_FACE = 0x0001F60F;
const char32_t NEUTRAL_FACE = 0x0001F610;
const char32_t EXPRESSIONLESS_FACE = 0x0001F611;
const char32_t UNAMUSED_FACE = 0x0001F612;
const char32_t FACE_WITH_COLD_SWEAT = 0x0001F613;
const char32_t PENSIVE_FACE = 0x0001F614;
const char32_t CONFUSED_FACE = 0x0001F615;
const char32_t CONFOUNDED_FACE = 0x0001F616;
const char32_t KISSING_FACE = 0x0001F617;
const char32_t FACE_THROWING_A_KISS = 0x0001F618;
const char32_t KISSING_FACE_WITH_SMILING_EYES = 0x0001F619;
const char32_t KISSING_FACE_WITH_CLOSED_EYES = 0x0001F61A;
const char32_t FACE_WITH_STUCK_OUT_TONGUE = 0x0001F61B;
const char32_t FACE_WITH_STUCK_OUT_TONGUE_AND_WINKING_EYE = 0x0001F61C;
const char32_t FACE_WITH_STUCK_OUT_TONGUE_AND_TIGHTLY_CLOSED_EYES = 0x0001F61D;
const char32_t DISAPPOINTED_FACE = 0x0001F61E;
const char32_t WORRIED_FACE = 0x0001F61F;
const char32_t ANGRY_FACE = 0x0001F620;
const char32_t POUTING_FACE = 0x0001F621;
const char32_t CRYING_FACE = 0x0001F622;
const char32_t PERSEVERING_FACE = 0x0001F623;
const char32_t FACE_WITH_LOOK_OF_TRIUMPH = 0x0001F624;
const char32_t DISAPPOINTED_BUT_RELIEVED_FACE = 0x0001F625;
const char32_t FROWNING_FACE_WITH_OPEN_MOUTH = 0x0001F626;
const char32_t ANGUISHED_FACE = 0x0001F627;
const char32_t FEARFUL_FACE = 0x0001F628;
const char32_t WEARY_FACE = 0x0001F629;
const char32_t SLEEPY_FACE = 0x0001F62A;
const char32_t TIRED_FACE = 0x0001F62B;
const char32_t GRIMACING_FACE = 0x0001F62C;
const char32_t LOUDLY_CRYING_FACE = 0x0001F62D;
const char32_t FACE_WITH_OPEN_MOUTH = 0x0001F62E;
const char32_t HUSHED_FACE = 0x0001F62F;
const char32_t FACE_WITH_OPEN_MOUTH_AND_COLD_SWEAT = 0x0001F630;
const char32_t FACE_SCREAMING_IN_FEAR = 0x0001F631;
const char32_t ASTONISHED_FACE = 0x0001F632;
const char32_t FLUSHED_FACE = 0x0001F633;
const char32_t SLEEPING_FACE = 0x0001F634;
const char32_t DIZZY_FACE = 0x0001F635;
const char32_t FACE_WITHOUT_MOUTH = 0x0001F636;
const char32_t FACE_WITH_MEDICAL_MASK = 0x0001F637;
const char32_t GRINNING_CAT_FACE_WITH_SMILING_EYES = 0x0001F638;
const char32_t CAT_FACE_WITH_TEARS_OF_JOY = 0x0001F639;
const char32_t SMILING_CAT_FACE_WITH_OPEN_MOUTH = 0x0001F63A;
const char32_t SMILING_CAT_FACE_WITH_HEART_SHAPED_EYES = 0x0001F63B;
const char32_t CAT_FACE_WITH_WRY_SMILE = 0x0001F63C;
const char32_t KISSING_CAT_FACE_WITH_CLOSED_EYES = 0x0001F63D;
const char32_t POUTING_CAT_FACE = 0x0001F63E;
const char32_t CRYING_CAT_FACE = 0x0001F63F;
const char32_t WEARY_CAT_FACE = 0x0001F640;
const char32_t SLIGHTLY_FROWNING_FACE = 0x0001F641;
const char32_t SLIGHTLY_SMILING_FACE = 0x0001F642;
const char32_t UPSIDE_DOWN_FACE = 0x0001F643;
const char32_t FACE_WITH_ROLLING_EYES = 0x0001F644;
const char32_t FACE_WITH_NO_GOOD_GESTURE = 0x0001F645;
const char32_t FACE_WITH_OK_GESTURE = 0x0001F646;
const char32_t PERSON_BOWING_DEEPLY = 0x0001F647;
const char32_t SEE_NO_EVIL_MONKEY = 0x0001F648;
const char32_t HEAR_NO_EVIL_MONKEY = 0x0001F649;
const char32_t SPEAK_NO_EVIL_MONKEY = 0x0001F64A;
const char32_t HAPPY_PERSON_RAISING_ONE_HAND = 0x0001F64B;
const char32_t PERSON_RAISING_BOTH_HANDS_IN_CELEBRATION = 0x0001F64C;
const char32_t PERSON_FROWNING = 0x0001F64D;
const char32_t PERSON_WITH_POUTING_FACE = 0x0001F64E;
const char32_t PERSON_WITH_FOLDED_HANDS = 0x0001F64F;
const char32_t NORTH_WEST_POINTING_LEAF = 0x0001F650;
const char32_t SOUTH_WEST_POINTING_LEAF = 0x0001F651;
const char32_t NORTH_EAST_POINTING_LEAF = 0x0001F652;
const char32_t SOUTH_EAST_POINTING_LEAF = 0x0001F653;
const char32_t TURNED_NORTH_WEST_POINTING_LEAF = 0x0001F654;
const char32_t TURNED_SOUTH_WEST_POINTING_LEAF = 0x0001F655;
const char32_t TURNED_NORTH_EAST_POINTING_LEAF = 0x0001F656;
const char32_t TURNED_SOUTH_EAST_POINTING_LEAF = 0x0001F657;
const char32_t NORTH_WEST_POINTING_VINE_LEAF = 0x0001F658;
const char32_t SOUTH_WEST_POINTING_VINE_LEAF = 0x0001F659;
const char32_t NORTH_EAST_POINTING_VINE_LEAF = 0x0001F65A;
const char32_t SOUTH_EAST_POINTING_VINE_LEAF = 0x0001F65B;
const char32_t HEAVY_NORTH_WEST_POINTING_VINE_LEAF = 0x0001F65C;
const char32_t HEAVY_SOUTH_WEST_POINTING_VINE_LEAF = 0x0001F65D;
const char32_t HEAVY_NORTH_EAST_POINTING_VINE_LEAF = 0x0001F65E;
const char32_t HEAVY_SOUTH_EAST_POINTING_VINE_LEAF = 0x0001F65F;
const char32_t NORTH_WEST_POINTING_BUD = 0x0001F660;
const char32_t SOUTH_WEST_POINTING_BUD = 0x0001F661;
const char32_t NORTH_EAST_POINTING_BUD = 0x0001F662;
const char32_t SOUTH_EAST_POINTING_BUD = 0x0001F663;
const char32_t HEAVY_NORTH_WEST_POINTING_BUD = 0x0001F664;
const char32_t HEAVY_SOUTH_WEST_POINTING_BUD = 0x0001F665;
const char32_t HEAVY_NORTH_EAST_POINTING_BUD = 0x0001F666;
const char32_t HEAVY_SOUTH_EAST_POINTING_BUD = 0x0001F667;
const char32_t HOLLOW_QUILT_SQUARE_ORNAMENT = 0x0001F668;
const char32_t HOLLOW_QUILT_SQUARE_ORNAMENT_IN_BLACK_SQUARE = 0x0001F669;
const char32_t SOLID_QUILT_SQUARE_ORNAMENT = 0x0001F66A;
const char32_t SOLID_QUILT_SQUARE_ORNAMENT_IN_BLACK_SQUARE = 0x0001F66B;
const char32_t LEFTWARDS_ROCKET = 0x0001F66C;
const char32_t UPWARDS_ROCKET = 0x0001F66D;
const char32_t RIGHTWARDS_ROCKET = 0x0001F66E;
const char32_t DOWNWARDS_ROCKET = 0x0001F66F;
const char32_t SCRIPT_LIGATURE_ET_ORNAMENT = 0x0001F670;
const char32_t HEAVY_SCRIPT_LIGATURE_ET_ORNAMENT = 0x0001F671;
const char32_t LIGATURE_OPEN_ET_ORNAMENT = 0x0001F672;
const char32_t HEAVY_LIGATURE_OPEN_ET_ORNAMENT = 0x0001F673;
const char32_t HEAVY_AMPERSAND_ORNAMENT = 0x0001F674;
const char32_t SWASH_AMPERSAND_ORNAMENT = 0x0001F675;
const char32_t SANS_SERIF_HEAVY_DOUBLE_TURNED_COMMA_QUOTATION_MARK_ORNAMENT = 0x0001F676;
const char32_t SANS_SERIF_HEAVY_DOUBLE_COMMA_QUOTATION_MARK_ORNAMENT = 0x0001F677;
const char32_t SANS_SERIF_HEAVY_LOW_DOUBLE_COMMA_QUOTATION_MARK_ORNAMENT = 0x0001F678;
const char32_t HEAVY_INTERROBANG_ORNAMENT = 0x0001F679;
const char32_t SANS_SERIF_INTERROBANG_ORNAMENT = 0x0001F67A;
const char32_t HEAVY_SANS_SERIF_INTERROBANG_ORNAMENT = 0x0001F67B;
const char32_t VERY_HEAVY_SOLIDUS = 0x0001F67C;
const char32_t VERY_HEAVY_REVERSE_SOLIDUS = 0x0001F67D;
const char32_t CHECKER_BOARD = 0x0001F67E;
const char32_t REVERSE_CHECKER_BOARD = 0x0001F67F;
const char32_t ROCKET = 0x0001F680;
const char32_t HELICOPTER = 0x0001F681;
const char32_t STEAM_LOCOMOTIVE = 0x0001F682;
const char32_t RAILWAY_CAR = 0x0001F683;
const char32_t HIGH_SPEED_TRAIN = 0x0001F684;
const char32_t HIGH_SPEED_TRAIN_WITH_BULLET_NOSE = 0x0001F685;
const char32_t TRAIN = 0x0001F686;
const char32_t METRO = 0x0001F687;
const char32_t LIGHT_RAIL = 0x0001F688;
const char32_t STATION = 0x0001F689;
const char32_t TRAM = 0x0001F68A;
const char32_t TRAM_CAR = 0x0001F68B;
const char32_t BUS = 0x0001F68C;
const char32_t ONCOMING_BUS = 0x0001F68D;
const char32_t TROLLEYBUS = 0x0001F68E;
const char32_t BUS_STOP = 0x0001F68F;
const char32_t MINIBUS = 0x0001F690;
const char32_t AMBULANCE = 0x0001F691;
const char32_t FIRE_ENGINE = 0x0001F692;
const char32_t POLICE_CAR = 0x0001F693;
const char32_t ONCOMING_POLICE_CAR = 0x0001F694;
const char32_t TAXI = 0x0001F695;
const char32_t ONCOMING_TAXI = 0x0001F696;
const char32_t AUTOMOBILE = 0x0001F697;
const char32_t ONCOMING_AUTOMOBILE = 0x0001F698;
const char32_t RECREATIONAL_VEHICLE = 0x0001F699;
const char32_t DELIVERY_TRUCK = 0x0001F69A;
const char32_t ARTICULATED_LORRY = 0x0001F69B;
const char32_t TRACTOR = 0x0001F69C;
const char32_t MONORAIL = 0x0001F69D;
const char32_t MOUNTAIN_RAILWAY = 0x0001F69E;
const char32_t SUSPENSION_RAILWAY = 0x0001F69F;
const char32_t MOUNTAIN_CABLEWAY = 0x0001F6A0;
const char32_t AERIAL_TRAMWAY = 0x0001F6A1;
const char32_t SHIP = 0x0001F6A2;
const char32_t ROWBOAT = 0x0001F6A3;
const char32_t SPEEDBOAT = 0x0001F6A4;
const char32_t HORIZONTAL_TRAFFIC_LIGHT = 0x0001F6A5;
const char32_t VERTICAL_TRAFFIC_LIGHT = 0x0001F6A6;
const char32_t CONSTRUCTION_SIGN = 0x0001F6A7;
const char32_t POLICE_CARS_REVOLVING_LIGHT = 0x0001F6A8;
const char32_t TRIANGULAR_FLAG_ON_POST = 0x0001F6A9;
const char32_t DOOR = 0x0001F6AA;
const char32_t NO_ENTRY_SIGN = 0x0001F6AB;
const char32_t SMOKING_SYMBOL = 0x0001F6AC;
const char32_t NO_SMOKING_SYMBOL = 0x0001F6AD;
const char32_t PUT_LITTER_IN_ITS_PLACE_SYMBOL = 0x0001F6AE;
const char32_t DO_NOT_LITTER_SYMBOL = 0x0001F6AF;
const char32_t POTABLE_WATER_SYMBOL = 0x0001F6B0;
const char32_t NON_POTABLE_WATER_SYMBOL = 0x0001F6B1;
const char32_t BICYCLE = 0x0001F6B2;
const char32_t NO_BICYCLES = 0x0001F6B3;
const char32_t BICYCLIST = 0x0001F6B4;
const char32_t MOUNTAIN_BICYCLIST = 0x0001F6B5;
const char32_t PEDESTRIAN = 0x0001F6B6;
const char32_t NO_PEDESTRIANS = 0x0001F6B7;
const char32_t CHILDREN_CROSSING = 0x0001F6B8;
const char32_t MENS_SYMBOL = 0x0001F6B9;
const char32_t WOMENS_SYMBOL = 0x0001F6BA;
const char32_t RESTROOM = 0x0001F6BB;
const char32_t BABY_SYMBOL = 0x0001F6BC;
const char32_t TOILET = 0x0001F6BD;
const char32_t WATER_CLOSET = 0x0001F6BE;
const char32_t SHOWER = 0x0001F6BF;
const char32_t BATH = 0x0001F6C0;
const char32_t BATHTUB = 0x0001F6C1;
const char32_t PASSPORT_CONTROL = 0x0001F6C2;
const char32_t CUSTOMS = 0x0001F6C3;
const char32_t BAGGAGE_CLAIM = 0x0001F6C4;
const char32_t LEFT_LUGGAGE = 0x0001F6C5;
const char32_t TRIANGLE_WITH_ROUNDED_CORNERS = 0x0001F6C6;
const char32_t PROHIBITED_SIGN = 0x0001F6C7;
const char32_t CIRCLED_INFORMATION_SOURCE = 0x0001F6C8;
const char32_t BOYS_SYMBOL = 0x0001F6C9;
const char32_t GIRLS_SYMBOL = 0x0001F6CA;
const char32_t COUCH_AND_LAMP = 0x0001F6CB;
const char32_t SLEEPING_ACCOMMODATION = 0x0001F6CC;
const char32_t SHOPPING_BAGS = 0x0001F6CD;
const char32_t BELLHOP_BELL = 0x0001F6CE;
const char32_t BED = 0x0001F6CF;
const char32_t PLACE_OF_WORSHIP = 0x0001F6D0;
const char32_t OCTAGONAL_SIGN = 0x0001F6D1;
const char32_t SHOPPING_TROLLEY = 0x0001F6D2;
const char32_t STUPA = 0x0001F6D3;
const char32_t PAGODA = 0x0001F6D4;
const char32_t HINDU_TEMPLE = 0x0001F6D5;
const char32_t HUT = 0x0001F6D6;
const char32_t ELEVATOR = 0x0001F6D7;
const char32_t LANDSLIDE = 0x0001F6D8;
const char32_t WIRELESS = 0x0001F6DC;
const char32_t PLAYGROUND_SLIDE = 0x0001F6DD;
const char32_t WHEEL = 0x0001F6DE;
const char32_t RING_BUOY = 0x0001F6DF;
const char32_t HAMMER_AND_WRENCH = 0x0001F6E0;
const char32_t SHIELD = 0x0001F6E1;
const char32_t OIL_DRUM = 0x0001F6E2;
const char32_t MOTORWAY = 0x0001F6E3;
const char32_t RAILWAY_TRACK = 0x0001F6E4;
const char32_t MOTOR_BOAT = 0x0001F6E5;
const char32_t UP_POINTING_MILITARY_AIRPLANE = 0x0001F6E6;
const char32_t UP_POINTING_AIRPLANE = 0x0001F6E7;
const char32_t UP_POINTING_SMALL_AIRPLANE = 0x0001F6E8;
const char32_t SMALL_AIRPLANE = 0x0001F6E9;
const char32_t NORTHEAST_POINTING_AIRPLANE = 0x0001F6EA;
const char32_t AIRPLANE_DEPARTURE = 0x0001F6EB;
const char32_t AIRPLANE_ARRIVING = 0x0001F6EC;
const char32_t SATELLITE = 0x0001F6F0;
const char32_t ONCOMING_FIRE_ENGINE = 0x0001F6F1;
const char32_t DIESEL_LOCOMOTIVE = 0x0001F6F2;
const char32_t PASSENGER_SHIP = 0x0001F6F3;
const char32_t SCOOTER = 0x0001F6F4;
const char32_t MOTOR_SCOOTER = 0x0001F6F5;
const char32_t CANOE = 0x0001F6F6;
const char32_t SLED = 0x0001F6F7;
const char32_t FLYING_SAUCER = 0x0001F6F8;
const char32_t SKATEBOARD = 0x0001F6F9;
const char32_t AUTO_RICKSHAW = 0x0001F6FA;
const char32_t PICKUP_TRUCK = 0x0001F6FB;
const char32_t ROLLER_SKATE = 0x0001F6FC;
const char32_t ALCHEMICAL_SYMBOL_FOR_QUINTESSENCE = 0x0001F700;
const char32_t ALCHEMICAL_SYMBOL_FOR_AIR = 0x0001F701;
const char32_t ALCHEMICAL_SYMBOL_FOR_FIRE = 0x0001F702;
const char32_t ALCHEMICAL_SYMBOL_FOR_EARTH = 0x0001F703;
const char32_t ALCHEMICAL_SYMBOL_FOR_WATER = 0x0001F704;
const char32_t ALCHEMICAL_SYMBOL_FOR_AQUAFORTIS = 0x0001F705;
const char32_t ALCHEMICAL_SYMBOL_FOR_AQUA_REGIA = 0x0001F706;
const char32_t ALCHEMICAL_SYMBOL_FOR_AQUA_REGIA_2 = 0x0001F707;
const char32_t ALCHEMICAL_SYMBOL_FOR_AQUA_VITAE = 0x0001F708;
const char32_t ALCHEMICAL_SYMBOL_FOR_AQUA_VITAE_2 = 0x0001F709;
const char32_t ALCHEMICAL_SYMBOL_FOR_VINEGAR = 0x0001F70A;
const char32_t ALCHEMICAL_SYMBOL_FOR_VINEGAR_2 = 0x0001F70B;
const char32_t ALCHEMICAL_SYMBOL_FOR_VINEGAR_3 = 0x0001F70C;
const char32_t ALCHEMICAL_SYMBOL_FOR_SULFUR = 0x0001F70D;
const char32_t ALCHEMICAL_SYMBOL_FOR_PHILOSOPHERS_SULFUR = 0x0001F70E;
const char32_t ALCHEMICAL_SYMBOL_FOR_BLACK_SULFUR = 0x0001F70F;
const char32_t ALCHEMICAL_SYMBOL_FOR_MERCURY_SUBLIMATE = 0x0001F710;
const char32_t ALCHEMICAL_SYMBOL_FOR_MERCURY_SUBLIMATE_2 = 0x0001F711;
const char32_t ALCHEMICAL_SYMBOL_FOR_MERCURY_SUBLIMATE_3 = 0x0001F712;
const char32_t ALCHEMICAL_SYMBOL_FOR_CINNABAR = 0x0001F713;
const char32_t ALCHEMICAL_SYMBOL_FOR_SALT = 0x0001F714;
const char32_t ALCHEMICAL_SYMBOL_FOR_NITRE = 0x0001F715;
const char32_t ALCHEMICAL_SYMBOL_FOR_VITRIOL = 0x0001F716;
const char32_t ALCHEMICAL_SYMBOL_FOR_VITRIOL_2 = 0x0001F717;
const char32_t ALCHEMICAL_SYMBOL_FOR_ROCK_SALT = 0x0001F718;
const char32_t ALCHEMICAL_SYMBOL_FOR_ROCK_SALT_2 = 0x0001F719;
const char32_t ALCHEMICAL_SYMBOL_FOR_GOLD = 0x0001F71A;
const char32_t ALCHEMICAL_SYMBOL_FOR_SILVER = 0x0001F71B;
const char32_t ALCHEMICAL_SYMBOL_FOR_IRON_ORE = 0x0001F71C;
const char32_t ALCHEMICAL_SYMBOL_FOR_IRON_ORE_2 = 0x0001F71D;
const char32_t ALCHEMICAL_SYMBOL_FOR_CROCUS_OF_IRON = 0x0001F71E;
const char32_t ALCHEMICAL_SYMBOL_FOR_REGULUS_OF_IRON = 0x0001F71F;
const char32_t ALCHEMICAL_SYMBOL_FOR_COPPER_ORE = 0x0001F720;
const char32_t ALCHEMICAL_SYMBOL_FOR_IRON_COPPER_ORE = 0x0001F721;
const char32_t ALCHEMICAL_SYMBOL_FOR_SUBLIMATE_OF_COPPER = 0x0001F722;
const char32_t ALCHEMICAL_SYMBOL_FOR_CROCUS_OF_COPPER = 0x0001F723;
const char32_t ALCHEMICAL_SYMBOL_FOR_CROCUS_OF_COPPER_2 = 0x0001F724;
const char32_t ALCHEMICAL_SYMBOL_FOR_COPPER_ANTIMONIATE = 0x0001F725;
const char32_t ALCHEMICAL_SYMBOL_FOR_SALT_OF_COPPER_ANTIMONIATE = 0x0001F726;
const char32_t ALCHEMICAL_SYMBOL_FOR_SUBLIMATE_OF_SALT_OF_COPPER = 0x0001F727;
const char32_t ALCHEMICAL_SYMBOL_FOR_VERDIGRIS = 0x0001F728;
const char32_t ALCHEMICAL_SYMBOL_FOR_TIN_ORE = 0x0001F729;
const char32_t ALCHEMICAL_SYMBOL_FOR_LEAD_ORE = 0x0001F72A;
const char32_t ALCHEMICAL_SYMBOL_FOR_ANTIMONY_ORE = 0x0001F72B;
const char32_t ALCHEMICAL_SYMBOL_FOR_SUBLIMATE_OF_ANTIMONY = 0x0001F72C;
const char32_t ALCHEMICAL_SYMBOL_FOR_SALT_OF_ANTIMONY = 0x0001F72D;
const char32_t ALCHEMICAL_SYMBOL_FOR_SUBLIMATE_OF_SALT_OF_ANTIMONY = 0x0001F72E;
const char32_t ALCHEMICAL_SYMBOL_FOR_VINEGAR_OF_ANTIMONY = 0x0001F72F;
const char32_t ALCHEMICAL_SYMBOL_FOR_REGULUS_OF_ANTIMONY = 0x0001F730;
const char32_t ALCHEMICAL_SYMBOL_FOR_REGULUS_OF_ANTIMONY_2 = 0x0001F731;
const char32_t ALCHEMICAL_SYMBOL_FOR_REGULUS = 0x0001F732;
const char32_t ALCHEMICAL_SYMBOL_FOR_REGULUS_2 = 0x0001F733;
const char32_t ALCHEMICAL_SYMBOL_FOR_REGULUS_3 = 0x0001F734;
const char32_t ALCHEMICAL_SYMBOL_FOR_REGULUS_4 = 0x0001F735;
const char32_t ALCHEMICAL_SYMBOL_FOR_ALKALI = 0x0001F736;
const char32_t ALCHEMICAL_SYMBOL_FOR_ALKALI_2 = 0x0001F737;
const char32_t ALCHEMICAL_SYMBOL_FOR_MARCASITE = 0x0001F738;
const char32_t ALCHEMICAL_SYMBOL_FOR_SAL_AMMONIAC = 0x0001F739;
const char32_t ALCHEMICAL_SYMBOL_FOR_ARSENIC = 0x0001F73A;
const char32_t ALCHEMICAL_SYMBOL_FOR_REALGAR = 0x0001F73B;
const char32_t ALCHEMICAL_SYMBOL_FOR_REALGAR_2 = 0x0001F73C;
const char32_t ALCHEMICAL_SYMBOL_FOR_AURIPIGMENT = 0x0001F73D;
const char32_t ALCHEMICAL_SYMBOL_FOR_BISMUTH_ORE = 0x0001F73E;
const char32_t ALCHEMICAL_SYMBOL_FOR_TARTAR = 0x0001F73F;
const char32_t ALCHEMICAL_SYMBOL_FOR_TARTAR_2 = 0x0001F740;
const char32_t ALCHEMICAL_SYMBOL_FOR_QUICK_LIME = 0x0001F741;
const char32_t ALCHEMICAL_SYMBOL_FOR_BORAX = 0x0001F742;
const char32_t ALCHEMICAL_SYMBOL_FOR_BORAX_2 = 0x0001F743;
const char32_t ALCHEMICAL_SYMBOL_FOR_BORAX_3 = 0x0001F744;
const char32_t ALCHEMICAL_SYMBOL_FOR_ALUM = 0x0001F745;
const char32_t ALCHEMICAL_SYMBOL_FOR_OIL = 0x0001F746;
const char32_t ALCHEMICAL_SYMBOL_FOR_SPIRIT = 0x0001F747;
const char32_t ALCHEMICAL_SYMBOL_FOR_TINCTURE = 0x0001F748;
const char32_t ALCHEMICAL_SYMBOL_FOR_GUM = 0x0001F749;
const char32_t ALCHEMICAL_SYMBOL_FOR_WAX = 0x0001F74A;
const char32_t ALCHEMICAL_SYMBOL_FOR_POWDER = 0x0001F74B;
const char32_t ALCHEMICAL_SYMBOL_FOR_CALX = 0x0001F74C;
const char32_t ALCHEMICAL_SYMBOL_FOR_TUTTY = 0x0001F74D;
const char32_t ALCHEMICAL_SYMBOL_FOR_CAPUT_MORTUUM = 0x0001F74E;
const char32_t ALCHEMICAL_SYMBOL_FOR_SCEPTER_OF_JOVE = 0x0001F74F;
const char32_t ALCHEMICAL_SYMBOL_FOR_CADUCEUS = 0x0001F750;
const char32_t ALCHEMICAL_SYMBOL_FOR_TRIDENT = 0x0001F751;
const char32_t ALCHEMICAL_SYMBOL_FOR_STARRED_TRIDENT = 0x0001F752;
const char32_t ALCHEMICAL_SYMBOL_FOR_LODESTONE = 0x0001F753;
const char32_t ALCHEMICAL_SYMBOL_FOR_SOAP = 0x0001F754;
const char32_t ALCHEMICAL_SYMBOL_FOR_URINE = 0x0001F755;
const char32_t ALCHEMICAL_SYMBOL_FOR_HORSE_DUNG = 0x0001F756;
const char32_t ALCHEMICAL_SYMBOL_FOR_ASHES = 0x0001F757;
const char32_t ALCHEMICAL_SYMBOL_FOR_POT_ASHES = 0x0001F758;
const char32_t ALCHEMICAL_SYMBOL_FOR_BRICK = 0x0001F759;
const char32_t ALCHEMICAL_SYMBOL_FOR_POWDERED_BRICK = 0x0001F75A;
const char32_t ALCHEMICAL_SYMBOL_FOR_AMALGAM = 0x0001F75B;
const char32_t ALCHEMICAL_SYMBOL_FOR_STRATUM_SUPER_STRATUM = 0x0001F75C;
const char32_t ALCHEMICAL_SYMBOL_FOR_STRATUM_SUPER_STRATUM_2 = 0x0001F75D;
const char32_t ALCHEMICAL_SYMBOL_FOR_SUBLIMATION = 0x0001F75E;
const char32_t ALCHEMICAL_SYMBOL_FOR_PRECIPITATE = 0x0001F75F;
const char32_t ALCHEMICAL_SYMBOL_FOR_DISTILL = 0x0001F760;
const char32_t ALCHEMICAL_SYMBOL_FOR_DISSOLVE = 0x0001F761;
const char32_t ALCHEMICAL_SYMBOL_FOR_DISSOLVE_2 = 0x0001F762;
const char32_t ALCHEMICAL_SYMBOL_FOR_PURIFY = 0x0001F763;
const char32_t ALCHEMICAL_SYMBOL_FOR_PUTREFACTION = 0x0001F764;
const char32_t ALCHEMICAL_SYMBOL_FOR_CRUCIBLE = 0x0001F765;
const char32_t ALCHEMICAL_SYMBOL_FOR_CRUCIBLE_2 = 0x0001F766;
const char32_t ALCHEMICAL_SYMBOL_FOR_CRUCIBLE_3 = 0x0001F767;
const char32_t ALCHEMICAL_SYMBOL_FOR_CRUCIBLE_4 = 0x0001F768;
const char32_t ALCHEMICAL_SYMBOL_FOR_CRUCIBLE_5 = 0x0001F769;
const char32_t ALCHEMICAL_SYMBOL_FOR_ALEMBIC = 0x0001F76A;
const char32_t ALCHEMICAL_SYMBOL_FOR_BATH_OF_MARY = 0x0001F76B;
const char32_t ALCHEMICAL_SYMBOL_FOR_BATH_OF_VAPOURS = 0x0001F76C;
const char32_t ALCHEMICAL_SYMBOL_FOR_RETORT = 0x0001F76D;
const char32_t ALCHEMICAL_SYMBOL_FOR_HOUR = 0x0001F76E;
const char32_t ALCHEMICAL_SYMBOL_FOR_NIGHT = 0x0001F76F;
const char32_t ALCHEMICAL_SYMBOL_FOR_DAY_NIGHT = 0x0001F770;
const char32_t ALCHEMICAL_SYMBOL_FOR_MONTH = 0x0001F771;
const char32_t ALCHEMICAL_SYMBOL_FOR_HALF_DRAM = 0x0001F772;
const char32_t ALCHEMICAL_SYMBOL_FOR_HALF_OUNCE = 0x0001F773;
const char32_t LOT_OF_FORTUNE = 0x0001F774;
const char32_t OCCULTATION = 0x0001F775;
const char32_t LUNAR_ECLIPSE = 0x0001F776;
const char32_t VESTA_FORM_TWO = 0x0001F777;
const char32_t ASTRAEA_FORM_TWO = 0x0001F778;
const char32_t HYGIEA_FORM_TWO = 0x0001F779;
const char32_t PARTHENOPE_FORM_TWO = 0x0001F77A;
const char32_t HAUMEA = 0x0001F77B;
const char32_t MAKEMAKE = 0x0001F77C;
const char32_t GONGGONG = 0x0001F77D;
const char32_t QUAOAR = 0x0001F77E;
const char32_t ORCUS = 0x0001F77F;
const char32_t BLACK_LEFT_POINTING_ISOSCELES_RIGHT_TRIANGLE = 0x0001F780;
const char32_t BLACK_UP_POINTING_ISOSCELES_RIGHT_TRIANGLE = 0x0001F781;
const char32_t BLACK_RIGHT_POINTING_ISOSCELES_RIGHT_TRIANGLE = 0x0001F782;
const char32_t BLACK_DOWN_POINTING_ISOSCELES_RIGHT_TRIANGLE = 0x0001F783;
const char32_t BLACK_SLIGHTLY_SMALL_CIRCLE = 0x0001F784;
const char32_t MEDIUM_BOLD_WHITE_CIRCLE = 0x0001F785;
const char32_t BOLD_WHITE_CIRCLE = 0x0001F786;
const char32_t HEAVY_WHITE_CIRCLE = 0x0001F787;
const char32_t VERY_HEAVY_WHITE_CIRCLE = 0x0001F788;
const char32_t EXTREMELY_HEAVY_WHITE_CIRCLE = 0x0001F789;
const char32_t WHITE_CIRCLE_CONTAINING_BLACK_SMALL_CIRCLE = 0x0001F78A;
const char32_t ROUND_TARGET = 0x0001F78B;
const char32_t BLACK_TINY_SQUARE = 0x0001F78C;
const char32_t BLACK_SLIGHTLY_SMALL_SQUARE = 0x0001F78D;
const char32_t LIGHT_WHITE_SQUARE = 0x0001F78E;
const char32_t MEDIUM_WHITE_SQUARE = 0x0001F78F;
const char32_t BOLD_WHITE_SQUARE = 0x0001F790;
const char32_t HEAVY_WHITE_SQUARE = 0x0001F791;
const char32_t VERY_HEAVY_WHITE_SQUARE = 0x0001F792;
const char32_t EXTREMELY_HEAVY_WHITE_SQUARE = 0x0001F793;
const char32_t WHITE_SQUARE_CONTAINING_BLACK_VERY_SMALL_SQUARE = 0x0001F794;
const char32_t WHITE_SQUARE_CONTAINING_BLACK_MEDIUM_SQUARE = 0x0001F795;
const char32_t SQUARE_TARGET = 0x0001F796;
const char32_t BLACK_TINY_DIAMOND = 0x0001F797;
const char32_t BLACK_VERY_SMALL_DIAMOND = 0x0001F798;
const char32_t BLACK_MEDIUM_SMALL_DIAMOND = 0x0001F799;
const char32_t WHITE_DIAMOND_CONTAINING_BLACK_VERY_SMALL_DIAMOND = 0x0001F79A;
const char32_t WHITE_DIAMOND_CONTAINING_BLACK_MEDIUM_DIAMOND = 0x0001F79B;
const char32_t DIAMOND_TARGET = 0x0001F79C;
const char32_t BLACK_TINY_LOZENGE = 0x0001F79D;
const char32_t BLACK_VERY_SMALL_LOZENGE = 0x0001F79E;
const char32_t BLACK_MEDIUM_SMALL_LOZENGE = 0x0001F79F;
const char32_t WHITE_LOZENGE_CONTAINING_BLACK_SMALL_LOZENGE = 0x0001F7A0;
const char32_t THIN_GREEK_CROSS = 0x0001F7A1;
const char32_t LIGHT_GREEK_CROSS = 0x0001F7A2;
const char32_t MEDIUM_GREEK_CROSS = 0x0001F7A3;
const char32_t BOLD_GREEK_CROSS = 0x0001F7A4;
const char32_t VERY_BOLD_GREEK_CROSS = 0x0001F7A5;
const char32_t VERY_HEAVY_GREEK_CROSS = 0x0001F7A6;
const char32_t EXTREMELY_HEAVY_GREEK_CROSS = 0x0001F7A7;
const char32_t THIN_SALTIRE = 0x0001F7A8;
const char32_t LIGHT_SALTIRE = 0x0001F7A9;
const char32_t MEDIUM_SALTIRE = 0x0001F7AA;
const char32_t BOLD_SALTIRE = 0x0001F7AB;
const char32_t HEAVY_SALTIRE = 0x0001F7AC;
const char32_t VERY_HEAVY_SALTIRE = 0x0001F7AD;
const char32_t EXTREMELY_HEAVY_SALTIRE = 0x0001F7AE;
const char32_t LIGHT_FIVE_SPOKED_ASTERISK = 0x0001F7AF;
const char32_t MEDIUM_FIVE_SPOKED_ASTERISK = 0x0001F7B0;
const char32_t BOLD_FIVE_SPOKED_ASTERISK = 0x0001F7B1;
const char32_t HEAVY_FIVE_SPOKED_ASTERISK = 0x0001F7B2;
const char32_t VERY_HEAVY_FIVE_SPOKED_ASTERISK = 0x0001F7B3;
const char32_t EXTREMELY_HEAVY_FIVE_SPOKED_ASTERISK = 0x0001F7B4;
const char32_t LIGHT_SIX_SPOKED_ASTERISK = 0x0001F7B5;
const char32_t MEDIUM_SIX_SPOKED_ASTERISK = 0x0001F7B6;
const char32_t BOLD_SIX_SPOKED_ASTERISK = 0x0001F7B7;
const char32_t HEAVY_SIX_SPOKED_ASTERISK = 0x0001F7B8;
const char32_t VERY_HEAVY_SIX_SPOKED_ASTERISK = 0x0001F7B9;
const char32_t EXTREMELY_HEAVY_SIX_SPOKED_ASTERISK = 0x0001F7BA;
const char32_t LIGHT_EIGHT_SPOKED_ASTERISK = 0x0001F7BB;
const char32_t MEDIUM_EIGHT_SPOKED_ASTERISK = 0x0001F7BC;
const char32_t BOLD_EIGHT_SPOKED_ASTERISK = 0x0001F7BD;
const char32_t HEAVY_EIGHT_SPOKED_ASTERISK = 0x0001F7BE;
const char32_t VERY_HEAVY_EIGHT_SPOKED_ASTERISK = 0x0001F7BF;
const char32_t LIGHT_THREE_POINTED_BLACK_STAR = 0x0001F7C0;
const char32_t MEDIUM_THREE_POINTED_BLACK_STAR = 0x0001F7C1;
const char32_t THREE_POINTED_BLACK_STAR = 0x0001F7C2;
const char32_t MEDIUM_THREE_POINTED_PINWHEEL_STAR = 0x0001F7C3;
const char32_t LIGHT_FOUR_POINTED_BLACK_STAR = 0x0001F7C4;
const char32_t MEDIUM_FOUR_POINTED_BLACK_STAR = 0x0001F7C5;
const char32_t FOUR_POINTED_BLACK_STAR = 0x0001F7C6;
const char32_t MEDIUM_FOUR_POINTED_PINWHEEL_STAR = 0x0001F7C7;
const char32_t REVERSE_LIGHT_FOUR_POINTED_PINWHEEL_STAR = 0x0001F7C8;
const char32_t LIGHT_FIVE_POINTED_BLACK_STAR = 0x0001F7C9;
const char32_t HEAVY_FIVE_POINTED_BLACK_STAR = 0x0001F7CA;
const char32_t MEDIUM_SIX_POINTED_BLACK_STAR = 0x0001F7CB;
const char32_t HEAVY_SIX_POINTED_BLACK_STAR = 0x0001F7CC;
const char32_t SIX_POINTED_PINWHEEL_STAR = 0x0001F7CD;
const char32_t MEDIUM_EIGHT_POINTED_BLACK_STAR = 0x0001F7CE;
const char32_t HEAVY_EIGHT_POINTED_BLACK_STAR = 0x0001F7CF;
const char32_t VERY_HEAVY_EIGHT_POINTED_BLACK_STAR = 0x0001F7D0;
const char32_t HEAVY_EIGHT_POINTED_PINWHEEL_STAR = 0x0001F7D1;
const char32_t LIGHT_TWELVE_POINTED_BLACK_STAR = 0x0001F7D2;
const char32_t HEAVY_TWELVE_POINTED_BLACK_STAR = 0x0001F7D3;
const char32_t HEAVY_TWELVE_POINTED_PINWHEEL_STAR = 0x0001F7D4;
const char32_t CIRCLED_TRIANGLE = 0x0001F7D5;
const char32_t NEGATIVE_CIRCLED_TRIANGLE = 0x0001F7D6;
const char32_t CIRCLED_SQUARE = 0x0001F7D7;
const char32_t NEGATIVE_CIRCLED_SQUARE = 0x0001F7D8;
const char32_t NINE_POINTED_WHITE_STAR = 0x0001F7D9;
const char32_t LARGE_ORANGE_CIRCLE = 0x0001F7E0;
const char32_t LARGE_YELLOW_CIRCLE = 0x0001F7E1;
const char32_t LARGE_GREEN_CIRCLE = 0x0001F7E2;
const char32_t LARGE_PURPLE_CIRCLE = 0x0001F7E3;
const char32_t LARGE_BROWN_CIRCLE = 0x0001F7E4;
const char32_t LARGE_RED_SQUARE = 0x0001F7E5;
const char32_t LARGE_BLUE_SQUARE = 0x0001F7E6;
const char32_t LARGE_ORANGE_SQUARE = 0x0001F7E7;
const char32_t LARGE_YELLOW_SQUARE = 0x0001F7E8;
const char32_t LARGE_GREEN_SQUARE = 0x0001F7E9;
const char32_t LARGE_PURPLE_SQUARE = 0x0001F7EA;
const char32_t LARGE_BROWN_SQUARE = 0x0001F7EB;
const char32_t HEAVY_EQUALS_SIGN = 0x0001F7F0;
const char32_t LEFTWARDS_ARROW_WITH_SMALL_TRIANGLE_ARROWHEAD = 0x0001F800;
const char32_t UPWARDS_ARROW_WITH_SMALL_TRIANGLE_ARROWHEAD = 0x0001F801;
const char32_t RIGHTWARDS_ARROW_WITH_SMALL_TRIANGLE_ARROWHEAD = 0x0001F802;
const char32_t DOWNWARDS_ARROW_WITH_SMALL_TRIANGLE_ARROWHEAD = 0x0001F803;
const char32_t LEFTWARDS_ARROW_WITH_MEDIUM_TRIANGLE_ARROWHEAD = 0x0001F804;
const char32_t UPWARDS_ARROW_WITH_MEDIUM_TRIANGLE_ARROWHEAD = 0x0001F805;
const char32_t RIGHTWARDS_ARROW_WITH_MEDIUM_TRIANGLE_ARROWHEAD = 0x0001F806;
const char32_t DOWNWARDS_ARROW_WITH_MEDIUM_TRIANGLE_ARROWHEAD = 0x0001F807;
const char32_t LEFTWARDS_ARROW_WITH_LARGE_TRIANGLE_ARROWHEAD = 0x0001F808;
const char32_t UPWARDS_ARROW_WITH_LARGE_TRIANGLE_ARROWHEAD = 0x0001F809;
const char32_t RIGHTWARDS_ARROW_WITH_LARGE_TRIANGLE_ARROWHEAD = 0x0001F80A;
const char32_t DOWNWARDS_ARROW_WITH_LARGE_TRIANGLE_ARROWHEAD = 0x0001F80B;
const char32_t LEFTWARDS_ARROW_WITH_SMALL_EQUILATERAL_ARROWHEAD = 0x0001F810;
const char32_t UPWARDS_ARROW_WITH_SMALL_EQUILATERAL_ARROWHEAD = 0x0001F811;
const char32_t RIGHTWARDS_ARROW_WITH_SMALL_EQUILATERAL_ARROWHEAD = 0x0001F812;
const char32_t DOWNWARDS_ARROW_WITH_SMALL_EQUILATERAL_ARROWHEAD = 0x0001F813;
const char32_t LEFTWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F814;
const char32_t UPWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F815;
const char32_t RIGHTWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F816;
const char32_t DOWNWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F817;
const char32_t HEAVY_LEFTWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F818;
const char32_t HEAVY_UPWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F819;
const char32_t HEAVY_RIGHTWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F81A;
const char32_t HEAVY_DOWNWARDS_ARROW_WITH_EQUILATERAL_ARROWHEAD = 0x0001F81B;
const char32_t HEAVY_LEFTWARDS_ARROW_WITH_LARGE_EQUILATERAL_ARROWHEAD = 0x0001F81C;
const char32_t HEAVY_UPWARDS_ARROW_WITH_LARGE_EQUILATERAL_ARROWHEAD = 0x0001F81D;
const char32_t HEAVY_RIGHTWARDS_ARROW_WITH_LARGE_EQUILATERAL_ARROWHEAD = 0x0001F81E;
const char32_t HEAVY_DOWNWARDS_ARROW_WITH_LARGE_EQUILATERAL_ARROWHEAD = 0x0001F81F;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_NARROW_SHAFT = 0x0001F820;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_NARROW_SHAFT = 0x0001F821;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_NARROW_SHAFT = 0x0001F822;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_NARROW_SHAFT = 0x0001F823;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_MEDIUM_SHAFT = 0x0001F824;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_MEDIUM_SHAFT = 0x0001F825;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_MEDIUM_SHAFT = 0x0001F826;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_MEDIUM_SHAFT = 0x0001F827;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_BOLD_SHAFT = 0x0001F828;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_BOLD_SHAFT = 0x0001F829;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_BOLD_SHAFT = 0x0001F82A;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_BOLD_SHAFT = 0x0001F82B;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_HEAVY_SHAFT = 0x0001F82C;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_HEAVY_SHAFT = 0x0001F82D;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_HEAVY_SHAFT = 0x0001F82E;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_HEAVY_SHAFT = 0x0001F82F;
const char32_t LEFTWARDS_TRIANGLE_HEADED_ARROW_WITH_VERY_HEAVY_SHAFT = 0x0001F830;
const char32_t UPWARDS_TRIANGLE_HEADED_ARROW_WITH_VERY_HEAVY_SHAFT = 0x0001F831;
const char32_t RIGHTWARDS_TRIANGLE_HEADED_ARROW_WITH_VERY_HEAVY_SHAFT = 0x0001F832;
const char32_t DOWNWARDS_TRIANGLE_HEADED_ARROW_WITH_VERY_HEAVY_SHAFT = 0x0001F833;
const char32_t LEFTWARDS_FINGER_POST_ARROW = 0x0001F834;
const char32_t UPWARDS_FINGER_POST_ARROW = 0x0001F835;
const char32_t RIGHTWARDS_FINGER_POST_ARROW = 0x0001F836;
const char32_t DOWNWARDS_FINGER_POST_ARROW = 0x0001F837;
const char32_t LEFTWARDS_SQUARED_ARROW = 0x0001F838;
const char32_t UPWARDS_SQUARED_ARROW = 0x0001F839;
const char32_t RIGHTWARDS_SQUARED_ARROW = 0x0001F83A;
const char32_t DOWNWARDS_SQUARED_ARROW = 0x0001F83B;
const char32_t LEFTWARDS_COMPRESSED_ARROW = 0x0001F83C;
const char32_t UPWARDS_COMPRESSED_ARROW = 0x0001F83D;
const char32_t RIGHTWARDS_COMPRESSED_ARROW = 0x0001F83E;
const char32_t DOWNWARDS_COMPRESSED_ARROW = 0x0001F83F;
const char32_t LEFTWARDS_HEAVY_COMPRESSED_ARROW = 0x0001F840;
const char32_t UPWARDS_HEAVY_COMPRESSED_ARROW = 0x0001F841;
const char32_t RIGHTWARDS_HEAVY_COMPRESSED_ARROW = 0x0001F842;
const char32_t DOWNWARDS_HEAVY_COMPRESSED_ARROW = 0x0001F843;
const char32_t LEFTWARDS_HEAVY_ARROW = 0x0001F844;
const char32_t UPWARDS_HEAVY_ARROW = 0x0001F845;
const char32_t RIGHTWARDS_HEAVY_ARROW = 0x0001F846;
const char32_t DOWNWARDS_HEAVY_ARROW = 0x0001F847;
const char32_t LEFTWARDS_SANS_SERIF_ARROW = 0x0001F850;
const char32_t UPWARDS_SANS_SERIF_ARROW = 0x0001F851;
const char32_t RIGHTWARDS_SANS_SERIF_ARROW = 0x0001F852;
const char32_t DOWNWARDS_SANS_SERIF_ARROW = 0x0001F853;
const char32_t NORTH_WEST_SANS_SERIF_ARROW = 0x0001F854;
const char32_t NORTH_EAST_SANS_SERIF_ARROW = 0x0001F855;
const char32_t SOUTH_EAST_SANS_SERIF_ARROW = 0x0001F856;
const char32_t SOUTH_WEST_SANS_SERIF_ARROW = 0x0001F857;
const char32_t LEFT_RIGHT_SANS_SERIF_ARROW = 0x0001F858;
const char32_t UP_DOWN_SANS_SERIF_ARROW = 0x0001F859;
const char32_t WIDE_HEADED_LEFTWARDS_LIGHT_BARB_ARROW = 0x0001F860;
const char32_t WIDE_HEADED_UPWARDS_LIGHT_BARB_ARROW = 0x0001F861;
const char32_t WIDE_HEADED_RIGHTWARDS_LIGHT_BARB_ARROW = 0x0001F862;
const char32_t WIDE_HEADED_DOWNWARDS_LIGHT_BARB_ARROW = 0x0001F863;
const char32_t WIDE_HEADED_NORTH_WEST_LIGHT_BARB_ARROW = 0x0001F864;
const char32_t WIDE_HEADED_NORTH_EAST_LIGHT_BARB_ARROW = 0x0001F865;
const char32_t WIDE_HEADED_SOUTH_EAST_LIGHT_BARB_ARROW = 0x0001F866;
const char32_t WIDE_HEADED_SOUTH_WEST_LIGHT_BARB_ARROW = 0x0001F867;
const char32_t WIDE_HEADED_LEFTWARDS_BARB_ARROW = 0x0001F868;
const char32_t WIDE_HEADED_UPWARDS_BARB_ARROW = 0x0001F869;
const char32_t WIDE_HEADED_RIGHTWARDS_BARB_ARROW = 0x0001F86A;
const char32_t WIDE_HEADED_DOWNWARDS_BARB_ARROW = 0x0001F86B;
const char32_t WIDE_HEADED_NORTH_WEST_BARB_ARROW = 0x0001F86C;
const char32_t WIDE_HEADED_NORTH_EAST_BARB_ARROW = 0x0001F86D;
const char32_t WIDE_HEADED_SOUTH_EAST_BARB_ARROW = 0x0001F86E;
const char32_t WIDE_HEADED_SOUTH_WEST_BARB_ARROW = 0x0001F86F;
const char32_t WIDE_HEADED_LEFTWARDS_MEDIUM_BARB_ARROW = 0x0001F870;
const char32_t WIDE_HEADED_UPWARDS_MEDIUM_BARB_ARROW = 0x0001F871;
const char32_t WIDE_HEADED_RIGHTWARDS_MEDIUM_BARB_ARROW = 0x0001F872;
const char32_t WIDE_HEADED_DOWNWARDS_MEDIUM_BARB_ARROW = 0x0001F873;
const char32_t WIDE_HEADED_NORTH_WEST_MEDIUM_BARB_ARROW = 0x0001F874;
const char32_t WIDE_HEADED_NORTH_EAST_MEDIUM_BARB_ARROW = 0x0001F875;
const char32_t WIDE_HEADED_SOUTH_EAST_MEDIUM_BARB_ARROW = 0x0001F876;
const char32_t WIDE_HEADED_SOUTH_WEST_MEDIUM_BARB_ARROW = 0x0001F877;
const char32_t WIDE_HEADED_LEFTWARDS_HEAVY_BARB_ARROW = 0x0001F878;
const char32_t WIDE_HEADED_UPWARDS_HEAVY_BARB_ARROW = 0x0001F879;
const char32_t WIDE_HEADED_RIGHTWARDS_HEAVY_BARB_ARROW = 0x0001F87A;
const char32_t WIDE_HEADED_DOWNWARDS_HEAVY_BARB_ARROW = 0x0001F87B;
const char32_t WIDE_HEADED_NORTH_WEST_HEAVY_BARB_ARROW = 0x0001F87C;
const char32_t WIDE_HEADED_NORTH_EAST_HEAVY_BARB_ARROW = 0x0001F87D;
const char32_t WIDE_HEADED_SOUTH_EAST_HEAVY_BARB_ARROW = 0x0001F87E;
const char32_t WIDE_HEADED_SOUTH_WEST_HEAVY_BARB_ARROW = 0x0001F87F;
const char32_t WIDE_HEADED_LEFTWARDS_VERY_HEAVY_BARB_ARROW = 0x0001F880;
const char32_t WIDE_HEADED_UPWARDS_VERY_HEAVY_BARB_ARROW = 0x0001F881;
const char32_t WIDE_HEADED_RIGHTWARDS_VERY_HEAVY_BARB_ARROW = 0x0001F882;
const char32_t WIDE_HEADED_DOWNWARDS_VERY_HEAVY_BARB_ARROW = 0x0001F883;
const char32_t WIDE_HEADED_NORTH_WEST_VERY_HEAVY_BARB_ARROW = 0x0001F884;
const char32_t WIDE_HEADED_NORTH_EAST_VERY_HEAVY_BARB_ARROW = 0x0001F885;
const char32_t WIDE_HEADED_SOUTH_EAST_VERY_HEAVY_BARB_ARROW = 0x0001F886;
const char32_t WIDE_HEADED_SOUTH_WEST_VERY_HEAVY_BARB_ARROW = 0x0001F887;
const char32_t LEFTWARDS_TRIANGLE_ARROWHEAD = 0x0001F890;
const char32_t UPWARDS_TRIANGLE_ARROWHEAD = 0x0001F891;
const char32_t RIGHTWARDS_TRIANGLE_ARROWHEAD = 0x0001F892;
const char32_t DOWNWARDS_TRIANGLE_ARROWHEAD = 0x0001F893;
const char32_t LEFTWARDS_WHITE_ARROW_WITHIN_TRIANGLE_ARROWHEAD = 0x0001F894;
const char32_t UPWARDS_WHITE_ARROW_WITHIN_TRIANGLE_ARROWHEAD = 0x0001F895;
const char32_t RIGHTWARDS_WHITE_ARROW_WITHIN_TRIANGLE_ARROWHEAD = 0x0001F896;
const char32_t DOWNWARDS_WHITE_ARROW_WITHIN_TRIANGLE_ARROWHEAD = 0x0001F897;
const char32_t LEFTWARDS_ARROW_WITH_NOTCHED_TAIL = 0x0001F898;
const char32_t UPWARDS_ARROW_WITH_NOTCHED_TAIL = 0x0001F899;
const char32_t RIGHTWARDS_ARROW_WITH_NOTCHED_TAIL = 0x0001F89A;
const char32_t DOWNWARDS_ARROW_WITH_NOTCHED_TAIL = 0x0001F89B;
const char32_t HEAVY_ARROW_SHAFT_WIDTH_ONE = 0x0001F89C;
const char32_t HEAVY_ARROW_SHAFT_WIDTH_TWO_THIRDS = 0x0001F89D;
const char32_t HEAVY_ARROW_SHAFT_WIDTH_ONE_HALF = 0x0001F89E;
const char32_t HEAVY_ARROW_SHAFT_WIDTH_ONE_THIRD = 0x0001F89F;
const char32_t LEFTWARDS_BOTTOM_SHADED_WHITE_ARROW = 0x0001F8A0;
const char32_t RIGHTWARDS_BOTTOM_SHADED_WHITE_ARROW = 0x0001F8A1;
const char32_t LEFTWARDS_TOP_SHADED_WHITE_ARROW = 0x0001F8A2;
const char32_t RIGHTWARDS_TOP_SHADED_WHITE_ARROW = 0x0001F8A3;
const char32_t LEFTWARDS_LEFT_SHADED_WHITE_ARROW = 0x0001F8A4;
const char32_t RIGHTWARDS_RIGHT_SHADED_WHITE_ARROW = 0x0001F8A5;
const char32_t LEFTWARDS_RIGHT_SHADED_WHITE_ARROW = 0x0001F8A6;
const char32_t RIGHTWARDS_LEFT_SHADED_WHITE_ARROW = 0x0001F8A7;
const char32_t LEFTWARDS_BACK_TILTED_SHADOWED_WHITE_ARROW = 0x0001F8A8;
const char32_t RIGHTWARDS_BACK_TILTED_SHADOWED_WHITE_ARROW = 0x0001F8A9;
const char32_t LEFTWARDS_FRONT_TILTED_SHADOWED_WHITE_ARROW = 0x0001F8AA;
const char32_t RIGHTWARDS_FRONT_TILTED_SHADOWED_WHITE_ARROW = 0x0001F8AB;
const char32_t WHITE_ARROW_SHAFT_WIDTH_ONE = 0x0001F8AC;
const char32_t WHITE_ARROW_SHAFT_WIDTH_TWO_THIRDS = 0x0001F8AD;
const char32_t ARROW_POINTING_UPWARDS_THEN_NORTH_WEST = 0x0001F8B0;
const char32_t ARROW_POINTING_RIGHTWARDS_THEN_CURVING_SOUTH_WEST = 0x0001F8B1;
const char32_t RIGHTWARDS_ARROW_WITH_LOWER_HOOK = 0x0001F8B2;
const char32_t DOWNWARDS_BLACK_ARROW_TO_BAR = 0x0001F8B3;
const char32_t NEGATIVE_SQUARED_LEFTWARDS_ARROW = 0x0001F8B4;
const char32_t NEGATIVE_SQUARED_UPWARDS_ARROW = 0x0001F8B5;
const char32_t NEGATIVE_SQUARED_RIGHTWARDS_ARROW = 0x0001F8B6;
const char32_t NEGATIVE_SQUARED_DOWNWARDS_ARROW = 0x0001F8B7;
const char32_t NORTH_WEST_ARROW_FROM_BAR = 0x0001F8B8;
const char32_t NORTH_EAST_ARROW_FROM_BAR = 0x0001F8B9;
const char32_t SOUTH_EAST_ARROW_FROM_BAR = 0x0001F8BA;
const char32_t SOUTH_WEST_ARROW_FROM_BAR = 0x0001F8BB;
const char32_t LEFTWARDS_ARROW_FROM_DOWNWARDS_ARROW = 0x0001F8C0;
const char32_t RIGHTWARDS_ARROW_FROM_DOWNWARDS_ARROW = 0x0001F8C1;
const char32_t LONG_RIGHTWARDS_ARROW_OVER_LONG_LEFTWARDS_ARROW = 0x0001F8D0;
const char32_t LONG_RIGHTWARDS_HARPOON_OVER_LONG_LEFTWARDS_HARPOON = 0x0001F8D1;
const char32_t LONG_RIGHTWARDS_HARPOON_ABOVE_SHORT_LEFTWARDS_HARPOON = 0x0001F8D2;
const char32_t SHORT_RIGHTWARDS_HARPOON_ABOVE_LONG_LEFTWARDS_HARPOON = 0x0001F8D3;
const char32_t LONG_LEFTWARDS_HARPOON_ABOVE_SHORT_RIGHTWARDS_HARPOON = 0x0001F8D4;
const char32_t SHORT_LEFTWARDS_HARPOON_ABOVE_LONG_RIGHTWARDS_HARPOON = 0x0001F8D5;
const char32_t LONG_RIGHTWARDS_ARROW_THROUGH_X = 0x0001F8D6;
const char32_t LONG_RIGHTWARDS_ARROW_WITH_DOUBLE_SLASH = 0x0001F8D7;
const char32_t LONG_LEFT_RIGHT_ARROW_WITH_DEPENDENT_LOBE = 0x0001F8D8;
const char32_t CIRCLED_CROSS_FORMEE_WITH_FOUR_DOTS = 0x0001F900;
const char32_t CIRCLED_CROSS_FORMEE_WITH_TWO_DOTS = 0x0001F901;
const char32_t CIRCLED_CROSS_FORMEE = 0x0001F902;
const char32_t LEFT_HALF_CIRCLE_WITH_FOUR_DOTS = 0x0001F903;
const char32_t LEFT_HALF_CIRCLE_WITH_THREE_DOTS = 0x0001F904;
const char32_t LEFT_HALF_CIRCLE_WITH_TWO_DOTS = 0x0001F905;
const char32_t LEFT_HALF_CIRCLE_WITH_DOT = 0x0001F906;
const char32_t LEFT_HALF_CIRCLE = 0x0001F907;
const char32_t DOWNWARD_FACING_HOOK = 0x0001F908;
const char32_t DOWNWARD_FACING_NOTCHED_HOOK = 0x0001F909;
const char32_t DOWNWARD_FACING_HOOK_WITH_DOT = 0x0001F90A;
const char32_t DOWNWARD_FACING_NOTCHED_HOOK_WITH_DOT = 0x0001F90B;
const char32_t PINCHED_FINGERS = 0x0001F90C;
const char32_t WHITE_HEART = 0x0001F90D;
const char32_t BROWN_HEART = 0x0001F90E;
const char32_t PINCHING_HAND = 0x0001F90F;
const char32_t ZIPPER_MOUTH_FACE = 0x0001F910;
const char32_t MONEY_MOUTH_FACE = 0x0001F911;
const char32_t FACE_WITH_THERMOMETER = 0x0001F912;
const char32_t NERD_FACE = 0x0001F913;
const char32_t THINKING_FACE = 0x0001F914;
const char32_t FACE_WITH_HEAD_BANDAGE = 0x0001F915;
const char32_t ROBOT_FACE = 0x0001F916;
const char32_t HUGGING_FACE = 0x0001F917;
const char32_t SIGN_OF_THE_HORNS = 0x0001F918;
const char32_t CALL_ME_HAND = 0x0001F919;
const char32_t RAISED_BACK_OF_HAND = 0x0001F91A;
const char32_t LEFT_FACING_FIST = 0x0001F91B;
const char32_t RIGHT_FACING_FIST = 0x0001F91C;
const char32_t HANDSHAKE = 0x0001F91D;
const char32_t HAND_WITH_INDEX_AND_MIDDLE_FINGERS_CROSSED = 0x0001F91E;
const char32_t I_LOVE_YOU_HAND_SIGN = 0x0001F91F;
const char32_t FACE_WITH_COWBOY_HAT = 0x0001F920;
const char32_t CLOWN_FACE = 0x0001F921;
const char32_t NAUSEATED_FACE = 0x0001F922;
const char32_t ROLLING_ON_THE_FLOOR_LAUGHING = 0x0001F923;
const char32_t DROOLING_FACE = 0x0001F924;
const char32_t LYING_FACE = 0x0001F925;
const char32_t FACE_PALM = 0x0001F926;
const char32_t SNEEZING_FACE = 0x0001F927;
const char32_t FACE_WITH_ONE_EYEBROW_RAISED = 0x0001F928;
const char32_t GRINNING_FACE_WITH_STAR_EYES = 0x0001F929;
const char32_t GRINNING_FACE_WITH_ONE_LARGE_AND_ONE_SMALL_EYE = 0x0001F92A;
const char32_t FACE_WITH_FINGER_COVERING_CLOSED_LIPS = 0x0001F92B;
const char32_t SERIOUS_FACE_WITH_SYMBOLS_COVERING_MOUTH = 0x0001F92C;
const char32_t SMILING_FACE_WITH_SMILING_EYES_AND_HAND_COVERING_MOUTH = 0x0001F92D;
const char32_t FACE_WITH_OPEN_MOUTH_VOMITING = 0x0001F92E;
const char32_t SHOCKED_FACE_WITH_EXPLODING_HEAD = 0x0001F92F;
const char32_t PREGNANT_WOMAN = 0x0001F930;
const char32_t BREAST_FEEDING = 0x0001F931;
const char32_t PALMS_UP_TOGETHER = 0x0001F932;
const char32_t SELFIE = 0x0001F933;
const char32_t PRINCE = 0x0001F934;
const char32_t MAN_IN_TUXEDO = 0x0001F935;
const char32_t MOTHER_CHRISTMAS = 0x0001F936;
const char32_t SHRUG = 0x0001F937;
const char32_t PERSON_DOING_CARTWHEEL = 0x0001F938;
const char32_t JUGGLING = 0x0001F939;
const char32_t FENCER = 0x0001F93A;
const char32_t MODERN_PENTATHLON = 0x0001F93B;
const char32_t WRESTLERS = 0x0001F93C;
const char32_t WATER_POLO = 0x0001F93D;
const char32_t HANDBALL = 0x0001F93E;
const char32_t DIVING_MASK = 0x0001F93F;
const char32_t WILTED_FLOWER = 0x0001F940;
const char32_t DRUM_WITH_DRUMSTICKS = 0x0001F941;
const char32_t CLINKING_GLASSES = 0x0001F942;
const char32_t TUMBLER_GLASS = 0x0001F943;
const char32_t SPOON = 0x0001F944;
const char32_t GOAL_NET = 0x0001F945;
const char32_t RIFLE = 0x0001F946;
const char32_t FIRST_PLACE_MEDAL = 0x0001F947;
const char32_t SECOND_PLACE_MEDAL = 0x0001F948;
const char32_t THIRD_PLACE_MEDAL = 0x0001F949;
const char32_t BOXING_GLOVE = 0x0001F94A;
const char32_t MARTIAL_ARTS_UNIFORM = 0x0001F94B;
const char32_t CURLING_STONE = 0x0001F94C;
const char32_t LACROSSE_STICK_AND_BALL = 0x0001F94D;
const char32_t SOFTBALL = 0x0001F94E;
const char32_t FLYING_DISC = 0x0001F94F;
const char32_t CROISSANT = 0x0001F950;
const char32_t AVOCADO = 0x0001F951;
const char32_t CUCUMBER = 0x0001F952;
const char32_t BACON = 0x0001F953;
const char32_t POTATO = 0x0001F954;
const char32_t CARROT = 0x0001F955;
const char32_t BAGUETTE_BREAD = 0x0001F956;
const char32_t GREEN_SALAD = 0x0001F957;
const char32_t SHALLOW_PAN_OF_FOOD = 0x0001F958;
const char32_t STUFFED_FLATBREAD = 0x0001F959;
const char32_t EGG = 0x0001F95A;
const char32_t GLASS_OF_MILK = 0x0001F95B;
const char32_t PEANUTS = 0x0001F95C;
const char32_t KIWIFRUIT = 0x0001F95D;
const char32_t PANCAKES = 0x0001F95E;
const char32_t DUMPLING = 0x0001F95F;
const char32_t FORTUNE_COOKIE = 0x0001F960;
const char32_t TAKEOUT_BOX = 0x0001F961;
const char32_t CHOPSTICKS = 0x0001F962;
const char32_t BOWL_WITH_SPOON = 0x0001F963;
const char32_t CUP_WITH_STRAW = 0x0001F964;
const char32_t COCONUT = 0x0001F965;
const char32_t BROCCOLI = 0x0001F966;
const char32_t PIE = 0x0001F967;
const char32_t PRETZEL = 0x0001F968;
const char32_t CUT_OF_MEAT = 0x0001F969;
const char32_t SANDWICH = 0x0001F96A;
const char32_t CANNED_FOOD = 0x0001F96B;
const char32_t LEAFY_GREEN = 0x0001F96C;
const char32_t MANGO = 0x0001F96D;
const char32_t MOON_CAKE = 0x0001F96E;
const char32_t BAGEL = 0x0001F96F;
const char32_t SMILING_FACE_WITH_SMILING_EYES_AND_THREE_HEARTS = 0x0001F970;
const char32_t YAWNING_FACE = 0x0001F971;
const char32_t SMILING_FACE_WITH_TEAR = 0x0001F972;
const char32_t FACE_WITH_PARTY_HORN_AND_PARTY_HAT = 0x0001F973;
const char32_t FACE_WITH_UNEVEN_EYES_AND_WAVY_MOUTH = 0x0001F974;
const char32_t OVERHEATED_FACE = 0x0001F975;
const char32_t FREEZING_FACE = 0x0001F976;
const char32_t NINJA = 0x0001F977;
const char32_t DISGUISED_FACE = 0x0001F978;
const char32_t FACE_HOLDING_BACK_TEARS = 0x0001F979;
const char32_t FACE_WITH_PLEADING_EYES = 0x0001F97A;
const char32_t SARI = 0x0001F97B;
const char32_t LAB_COAT = 0x0001F97C;
const char32_t GOGGLES = 0x0001F97D;
const char32_t HIKING_BOOT = 0x0001F97E;
const char32_t FLAT_SHOE = 0x0001F97F;
const char32_t CRAB = 0x0001F980;
const char32_t LION_FACE = 0x0001F981;
const char32_t SCORPION = 0x0001F982;
const char32_t TURKEY = 0x0001F983;
const char32_t UNICORN_FACE = 0x0001F984;
const char32_t EAGLE = 0x0001F985;
const char32_t DUCK = 0x0001F986;
const char32_t BAT = 0x0001F987;
const char32_t SHARK = 0x0001F988;
const char32_t OWL = 0x0001F989;
const char32_t FOX_FACE = 0x0001F98A;
const char32_t BUTTERFLY = 0x0001F98B;
const char32_t DEER = 0x0001F98C;
const char32_t GORILLA = 0x0001F98D;
const char32_t LIZARD = 0x0001F98E;
const char32_t RHINOCEROS = 0x0001F98F;
const char32_t SHRIMP = 0x0001F990;
const char32_t SQUID = 0x0001F991;
const char32_t GIRAFFE_FACE = 0x0001F992;
const char32_t ZEBRA_FACE = 0x0001F993;
const char32_t HEDGEHOG = 0x0001F994;
const char32_t SAUROPOD = 0x0001F995;
const char32_t T_REX = 0x0001F996;
const char32_t CRICKET = 0x0001F997;
const char32_t KANGAROO = 0x0001F998;
const char32_t LLAMA = 0x0001F999;
const char32_t PEACOCK = 0x0001F99A;
const char32_t HIPPOPOTAMUS = 0x0001F99B;
const char32_t PARROT = 0x0001F99C;
const char32_t RACCOON = 0x0001F99D;
const char32_t LOBSTER = 0x0001F99E;
const char32_t MOSQUITO = 0x0001F99F;
const char32_t MICROBE = 0x0001F9A0;
const char32_t BADGER = 0x0001F9A1;
const char32_t SWAN = 0x0001F9A2;
const char32_t MAMMOTH = 0x0001F9A3;
const char32_t DODO = 0x0001F9A4;
const char32_t SLOTH = 0x0001F9A5;
const char32_t OTTER = 0x0001F9A6;
const char32_t ORANGUTAN = 0x0001F9A7;
const char32_t SKUNK = 0x0001F9A8;
const char32_t FLAMINGO = 0x0001F9A9;
const char32_t OYSTER = 0x0001F9AA;
const char32_t BEAVER = 0x0001F9AB;
const char32_t BISON = 0x0001F9AC;
const char32_t SEAL = 0x0001F9AD;
const char32_t GUIDE_DOG = 0x0001F9AE;
const char32_t PROBING_CANE = 0x0001F9AF;
const char32_t EMOJI_COMPONENT_RED_HAIR = 0x0001F9B0;
const char32_t EMOJI_COMPONENT_CURLY_HAIR = 0x0001F9B1;
const char32_t EMOJI_COMPONENT_BALD = 0x0001F9B2;
const char32_t EMOJI_COMPONENT_WHITE_HAIR = 0x0001F9B3;
const char32_t BONE = 0x0001F9B4;
const char32_t LEG = 0x0001F9B5;
const char32_t FOOT = 0x0001F9B6;
const char32_t TOOTH = 0x0001F9B7;
const char32_t SUPERHERO = 0x0001F9B8;
const char32_t SUPERVILLAIN = 0x0001F9B9;
const char32_t SAFETY_VEST = 0x0001F9BA;
const char32_t EAR_WITH_HEARING_AID = 0x0001F9BB;
const char32_t MOTORIZED_WHEELCHAIR = 0x0001F9BC;
const char32_t MANUAL_WHEELCHAIR = 0x0001F9BD;
const char32_t MECHANICAL_ARM = 0x0001F9BE;
const char32_t MECHANICAL_LEG = 0x0001F9BF;
const char32_t CHEESE_WEDGE = 0x0001F9C0;
const char32_t CUPCAKE = 0x0001F9C1;
const char32_t SALT_SHAKER = 0x0001F9C2;
const char32_t BEVERAGE_BOX = 0x0001F9C3;
const char32_t GARLIC = 0x0001F9C4;
const char32_t ONION = 0x0001F9C5;
const char32_t FALAFEL = 0x0001F9C6;
const char32_t WAFFLE = 0x0001F9C7;
const char32_t BUTTER = 0x0001F9C8;
const char32_t MATE_DRINK = 0x0001F9C9;
const char32_t ICE_CUBE = 0x0001F9CA;
const char32_t BUBBLE_TEA = 0x0001F9CB;
const char32_t TROLL = 0x0001F9CC;
const char32_t STANDING_PERSON = 0x0001F9CD;
const char32_t KNEELING_PERSON = 0x0001F9CE;
const char32_t DEAF_PERSON = 0x0001F9CF;
const char32_t FACE_WITH_MONOCLE = 0x0001F9D0;
const char32_t ADULT = 0x0001F9D1;
const char32_t CHILD = 0x0001F9D2;
const char32_t OLDER_ADULT = 0x0001F9D3;
const char32_t BEARDED_PERSON = 0x0001F9D4;
const char32_t PERSON_WITH_HEADSCARF = 0x0001F9D5;
const char32_t PERSON_IN_STEAMY_ROOM = 0x0001F9D6;
const char32_t PERSON_CLIMBING = 0x0001F9D7;
const char32_t PERSON_IN_LOTUS_POSITION = 0x0001F9D8;
const char32_t MAGE = 0x0001F9D9;
const char32_t FAIRY = 0x0001F9DA;
const char32_t VAMPIRE = 0x0001F9DB;
const char32_t MERPERSON = 0x0001F9DC;
const char32_t ELF = 0x0001F9DD;
const char32_t GENIE = 0x0001F9DE;
const char32_t ZOMBIE = 0x0001F9DF;
const char32_t BRAIN = 0x0001F9E0;
const char32_t ORANGE_HEART = 0x0001F9E1;
const char32_t BILLED_CAP = 0x0001F9E2;
const char32_t SCARF = 0x0001F9E3;
const char32_t GLOVES = 0x0001F9E4;
const char32_t COAT = 0x0001F9E5;
const char32_t SOCKS = 0x0001F9E6;
const char32_t RED_GIFT_ENVELOPE = 0x0001F9E7;
const char32_t FIRECRACKER = 0x0001F9E8;
const char32_t JIGSAW_PUZZLE_PIECE = 0x0001F9E9;
const char32_t TEST_TUBE = 0x0001F9EA;
const char32_t PETRI_DISH = 0x0001F9EB;
const char32_t DNA_DOUBLE_HELIX = 0x0001F9EC;
const char32_t COMPASS = 0x0001F9ED;
const char32_t ABACUS = 0x0001F9EE;
const char32_t FIRE_EXTINGUISHER = 0x0001F9EF;
const char32_t TOOLBOX = 0x0001F9F0;
const char32_t BRICK = 0x0001F9F1;
const char32_t MAGNET = 0x0001F9F2;
const char32_t LUGGAGE = 0x0001F9F3;
const char32_t LOTION_BOTTLE = 0x0001F9F4;
const char32_t SPOOL_OF_THREAD = 0x0001F9F5;
const char32_t BALL_OF_YARN = 0x0001F9F6;
const char32_t SAFETY_PIN = 0x0001F9F7;
const char32_t TEDDY_BEAR = 0x0001F9F8;
const char32_t BROOM = 0x0001F9F9;
const char32_t BASKET = 0x0001F9FA;
const char32_t ROLL_OF_PAPER = 0x0001F9FB;
const char32_t BAR_OF_SOAP = 0x0001F9FC;
const char32_t SPONGE = 0x0001F9FD;
const char32_t RECEIPT = 0x0001F9FE;
const char32_t NAZAR_AMULET = 0x0001F9FF;
const char32_t NEUTRAL_CHESS_KING = 0x0001FA00;
const char32_t NEUTRAL_CHESS_QUEEN = 0x0001FA01;
const char32_t NEUTRAL_CHESS_ROOK = 0x0001FA02;
const char32_t NEUTRAL_CHESS_BISHOP = 0x0001FA03;
const char32_t NEUTRAL_CHESS_KNIGHT = 0x0001FA04;
const char32_t NEUTRAL_CHESS_PAWN = 0x0001FA05;
const char32_t WHITE_CHESS_KNIGHT_ROTATED_FORTY_FIVE_DEGREES = 0x0001FA06;
const char32_t BLACK_CHESS_KNIGHT_ROTATED_FORTY_FIVE_DEGREES = 0x0001FA07;
const char32_t NEUTRAL_CHESS_KNIGHT_ROTATED_FORTY_FIVE_DEGREES = 0x0001FA08;
const char32_t WHITE_CHESS_KING_ROTATED_NINETY_DEGREES = 0x0001FA09;
const char32_t WHITE_CHESS_QUEEN_ROTATED_NINETY_DEGREES = 0x0001FA0A;
const char32_t WHITE_CHESS_ROOK_ROTATED_NINETY_DEGREES = 0x0001FA0B;
const char32_t WHITE_CHESS_BISHOP_ROTATED_NINETY_DEGREES = 0x0001FA0C;
const char32_t WHITE_CHESS_KNIGHT_ROTATED_NINETY_DEGREES = 0x0001FA0D;
const char32_t WHITE_CHESS_PAWN_ROTATED_NINETY_DEGREES = 0x0001FA0E;
const char32_t BLACK_CHESS_KING_ROTATED_NINETY_DEGREES = 0x0001FA0F;
const char32_t BLACK_CHESS_QUEEN_ROTATED_NINETY_DEGREES = 0x0001FA10;
const char32_t BLACK_CHESS_ROOK_ROTATED_NINETY_DEGREES = 0x0001FA11;
const char32_t BLACK_CHESS_BISHOP_ROTATED_NINETY_DEGREES = 0x0001FA12;
const char32_t BLACK_CHESS_KNIGHT_ROTATED_NINETY_DEGREES = 0x0001FA13;
const char32_t BLACK_CHESS_PAWN_ROTATED_NINETY_DEGREES = 0x0001FA14;
const char32_t NEUTRAL_CHESS_KING_ROTATED_NINETY_DEGREES = 0x0001FA15;
const char32_t NEUTRAL_CHESS_QUEEN_ROTATED_NINETY_DEGREES = 0x0001FA16;
const char32_t NEUTRAL_CHESS_ROOK_ROTATED_NINETY_DEGREES = 0x0001FA17;
const char32_t NEUTRAL_CHESS_BISHOP_ROTATED_NINETY_DEGREES = 0x0001FA18;
const char32_t NEUTRAL_CHESS_KNIGHT_ROTATED_NINETY_DEGREES = 0x0001FA19;
const char32_t NEUTRAL_CHESS_PAWN_ROTATED_NINETY_DEGREES = 0x0001FA1A;
const char32_t WHITE_CHESS_KNIGHT_ROTATED_ONE_HUNDRED_THIRTY_FIVE_DEGREES = 0x0001FA1B;
const char32_t BLACK_CHESS_KNIGHT_ROTATED_ONE_HUNDRED_THIRTY_FIVE_DEGREES = 0x0001FA1C;
const char32_t NEUTRAL_CHESS_KNIGHT_ROTATED_ONE_HUNDRED_THIRTY_FIVE_DEGREES = 0x0001FA1D;
const char32_t WHITE_CHESS_TURNED_KING = 0x0001FA1E;
const char32_t WHITE_CHESS_TURNED_QUEEN = 0x0001FA1F;
const char32_t WHITE_CHESS_TURNED_ROOK = 0x0001FA20;
const char32_t WHITE_CHESS_TURNED_BISHOP = 0x0001FA21;
const char32_t WHITE_CHESS_TURNED_KNIGHT = 0x0001FA22;
const char32_t WHITE_CHESS_TURNED_PAWN = 0x0001FA23;
const char32_t BLACK_CHESS_TURNED_KING = 0x0001FA24;
const char32_t BLACK_CHESS_TURNED_QUEEN = 0x0001FA25;
const char32_t BLACK_CHESS_TURNED_ROOK = 0x0001FA26;
const char32_t BLACK_CHESS_TURNED_BISHOP = 0x0001FA27;
const char32_t BLACK_CHESS_TURNED_KNIGHT = 0x0001FA28;
const char32_t BLACK_CHESS_TURNED_PAWN = 0x0001FA29;
const char32_t NEUTRAL_CHESS_TURNED_KING = 0x0001FA2A;
const char32_t NEUTRAL_CHESS_TURNED_QUEEN = 0x0001FA2B;
const char32_t NEUTRAL_CHESS_TURNED_ROOK = 0x0001FA2C;
const char32_t NEUTRAL_CHESS_TURNED_BISHOP = 0x0001FA2D;
const char32_t NEUTRAL_CHESS_TURNED_KNIGHT = 0x0001FA2E;
const char32_t NEUTRAL_CHESS_TURNED_PAWN = 0x0001FA2F;
const char32_t WHITE_CHESS_KNIGHT_ROTATED_TWO_HUNDRED_TWENTY_FIVE_DEGREES = 0x0001FA30;
const char32_t BLACK_CHESS_KNIGHT_ROTATED_TWO_HUNDRED_TWENTY_FIVE_DEGREES = 0x0001FA31;
const char32_t NEUTRAL_CHESS_KNIGHT_ROTATED_TWO_HUNDRED_TWENTY_FIVE_DEGREES = 0x0001FA32;
const char32_t WHITE_CHESS_KING_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA33;
const char32_t WHITE_CHESS_QUEEN_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA34;
const char32_t WHITE_CHESS_ROOK_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA35;
const char32_t WHITE_CHESS_BISHOP_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA36;
const char32_t WHITE_CHESS_KNIGHT_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA37;
const char32_t WHITE_CHESS_PAWN_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA38;
const char32_t BLACK_CHESS_KING_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA39;
const char32_t BLACK_CHESS_QUEEN_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA3A;
const char32_t BLACK_CHESS_ROOK_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA3B;
const char32_t BLACK_CHESS_BISHOP_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA3C;
const char32_t BLACK_CHESS_KNIGHT_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA3D;
const char32_t BLACK_CHESS_PAWN_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA3E;
const char32_t NEUTRAL_CHESS_KING_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA3F;
const char32_t NEUTRAL_CHESS_QUEEN_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA40;
const char32_t NEUTRAL_CHESS_ROOK_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA41;
const char32_t NEUTRAL_CHESS_BISHOP_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA42;
const char32_t NEUTRAL_CHESS_KNIGHT_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA43;
const char32_t NEUTRAL_CHESS_PAWN_ROTATED_TWO_HUNDRED_SEVENTY_DEGREES = 0x0001FA44;
const char32_t WHITE_CHESS_KNIGHT_ROTATED_THREE_HUNDRED_FIFTEEN_DEGREES = 0x0001FA45;
const char32_t BLACK_CHESS_KNIGHT_ROTATED_THREE_HUNDRED_FIFTEEN_DEGREES = 0x0001FA46;
const char32_t NEUTRAL_CHESS_KNIGHT_ROTATED_THREE_HUNDRED_FIFTEEN_DEGREES = 0x0001FA47;
const char32_t WHITE_CHESS_EQUIHOPPER = 0x0001FA48;
const char32_t BLACK_CHESS_EQUIHOPPER = 0x0001FA49;
const char32_t NEUTRAL_CHESS_EQUIHOPPER = 0x0001FA4A;
const char32_t WHITE_CHESS_EQUIHOPPER_ROTATED_NINETY_DEGREES = 0x0001FA4B;
const char32_t BLACK_CHESS_EQUIHOPPER_ROTATED_NINETY_DEGREES = 0x0001FA4C;
const char32_t NEUTRAL_CHESS_EQUIHOPPER_ROTATED_NINETY_DEGREES = 0x0001FA4D;
const char32_t WHITE_CHESS_KNIGHT_QUEEN = 0x0001FA4E;
const char32_t WHITE_CHESS_KNIGHT_ROOK = 0x0001FA4F;
const char32_t WHITE_CHESS_KNIGHT_BISHOP = 0x0001FA50;
const char32_t BLACK_CHESS_KNIGHT_QUEEN = 0x0001FA51;
const char32_t BLACK_CHESS_KNIGHT_ROOK = 0x0001FA52;
const char32_t BLACK_CHESS_KNIGHT_BISHOP = 0x0001FA53;
const char32_t WHITE_CHESS_FERZ = 0x0001FA54;
const char32_t WHITE_CHESS_ALFIL = 0x0001FA55;
const char32_t BLACK_CHESS_FERZ = 0x0001FA56;
const char32_t BLACK_CHESS_ALFIL = 0x0001FA57;
const char32_t XIANGQI_RED_GENERAL = 0x0001FA60;
const char32_t XIANGQI_RED_MANDARIN = 0x0001FA61;
const char32_t XIANGQI_RED_ELEPHANT = 0x0001FA62;
const char32_t XIANGQI_RED_HORSE = 0x0001FA63;
const char32_t XIANGQI_RED_CHARIOT = 0x0001FA64;
const char32_t XIANGQI_RED_CANNON = 0x0001FA65;
const char32_t XIANGQI_RED_SOLDIER = 0x0001FA66;
const char32_t XIANGQI_BLACK_GENERAL = 0x0001FA67;
const char32_t XIANGQI_BLACK_MANDARIN = 0x0001FA68;
const char32_t XIANGQI_BLACK_ELEPHANT = 0x0001FA69;
const char32_t XIANGQI_BLACK_HORSE = 0x0001FA6A;
const char32_t XIANGQI_BLACK_CHARIOT = 0x0001FA6B;
const char32_t XIANGQI_BLACK_CANNON = 0x0001FA6C;
const char32_t XIANGQI_BLACK_SOLDIER = 0x0001FA6D;
const char32_t BALLET_SHOES = 0x0001FA70;
const char32_t ONE_PIECE_SWIMSUIT = 0x0001FA71;
const char32_t BRIEFS = 0x0001FA72;
const char32_t SHORTS = 0x0001FA73;
const char32_t THONG_SANDAL = 0x0001FA74;
const char32_t LIGHT_BLUE_HEART = 0x0001FA75;
const char32_t GREY_HEART = 0x0001FA76;
const char32_t PINK_HEART = 0x0001FA77;
const char32_t DROP_OF_BLOOD = 0x0001FA78;
const char32_t ADHESIVE_BANDAGE = 0x0001FA79;
const char32_t STETHOSCOPE = 0x0001FA7A;
const char32_t X_RAY = 0x0001FA7B;
const char32_t CRUTCH = 0x0001FA7C;
const char32_t YO_YO = 0x0001FA80;
const char32_t KITE = 0x0001FA81;
const char32_t PARACHUTE = 0x0001FA82;
const char32_t BOOMERANG = 0x0001FA83;
const char32_t MAGIC_WAND = 0x0001FA84;
const char32_t PINATA = 0x0001FA85;
const char32_t NESTING_DOLLS = 0x0001FA86;
const char32_t MARACAS = 0x0001FA87;
const char32_t FLUTE = 0x0001FA88;
const char32_t HARP = 0x0001FA89;
const char32_t TROMBONE = 0x0001FA8A;
const char32_t TREASURE_CHEST = 0x0001FA8E;
const char32_t SHOVEL = 0x0001FA8F;
const char32_t RINGED_PLANET = 0x0001FA90;
const char32_t CHAIR = 0x0001FA91;
const char32_t RAZOR = 0x0001FA92;
const char32_t AXE = 0x0001FA93;
const char32_t DIYA_LAMP = 0x0001FA94;
const char32_t BANJO = 0x0001FA95;
const char32_t MILITARY_HELMET = 0x0001FA96;
const char32_t ACCORDION = 0x0001FA97;
const char32_t LONG_DRUM = 0x0001FA98;
const char32_t COIN = 0x0001FA99;
const char32_t CARPENTRY_SAW = 0x0001FA9A;
const char32_t SCREWDRIVER = 0x0001FA9B;
const char32_t LADDER = 0x0001FA9C;
const char32_t HOOK = 0x0001FA9D;
const char32_t MIRROR = 0x0001FA9E;
const char32_t WINDOW = 0x0001FA9F;
const char32_t PLUNGER = 0x0001FAA0;
const char32_t SEWING_NEEDLE = 0x0001FAA1;
const char32_t KNOT = 0x0001FAA2;
const char32_t BUCKET = 0x0001FAA3;
const char32_t MOUSE_TRAP = 0x0001FAA4;
const char32_t TOOTHBRUSH = 0x0001FAA5;
const char32_t HEADSTONE = 0x0001FAA6;
const char32_t PLACARD = 0x0001FAA7;
const char32_t ROCK = 0x0001FAA8;
const char32_t MIRROR_BALL = 0x0001FAA9;
const char32_t IDENTIFICATION_CARD = 0x0001FAAA;
const char32_t LOW_BATTERY = 0x0001FAAB;
const char32_t HAMSA = 0x0001FAAC;
const char32_t FOLDING_HAND_FAN = 0x0001FAAD;
const char32_t HAIR_PICK = 0x0001FAAE;
const char32_t KHANDA = 0x0001FAAF;
const char32_t FLY = 0x0001FAB0;
const char32_t WORM = 0x0001FAB1;
const char32_t BEETLE = 0x0001FAB2;
const char32_t COCKROACH = 0x0001FAB3;
const char32_t POTTED_PLANT = 0x0001FAB4;
const char32_t WOOD = 0x0001FAB5;
const char32_t FEATHER = 0x0001FAB6;
const char32_t LOTUS = 0x0001FAB7;
const char32_t CORAL = 0x0001FAB8;
const char32_t EMPTY_NEST = 0x0001FAB9;
const char32_t NEST_WITH_EGGS = 0x0001FABA;
const char32_t HYACINTH = 0x0001FABB;
const char32_t JELLYFISH = 0x0001FABC;
const char32_t WING = 0x0001FABD;
const char32_t LEAFLESS_TREE = 0x0001FABE;
const char32_t GOOSE = 0x0001FABF;
const char32_t ANATOMICAL_HEART = 0x0001FAC0;
const char32_t LUNGS = 0x0001FAC1;
const char32_t PEOPLE_HUGGING = 0x0001FAC2;
const char32_t PREGNANT_MAN = 0x0001FAC3;
const char32_t PREGNANT_PERSON = 0x0001FAC4;
const char32_t PERSON_WITH_CROWN = 0x0001FAC5;
const char32_t FINGERPRINT = 0x0001FAC6;
const char32_t HAIRY_CREATURE = 0x0001FAC8;
const char32_t ORCA = 0x0001FACD;
const char32_t MOOSE = 0x0001FACE;
const char32_t DONKEY = 0x0001FACF;
const char32_t BLUEBERRIES = 0x0001FAD0;
const char32_t BELL_PEPPER = 0x0001FAD1;
const char32_t OLIVE = 0x0001FAD2;
const char32_t FLATBREAD = 0x0001FAD3;
const char32_t TAMALE = 0x0001FAD4;
const char32_t FONDUE = 0x0001FAD5;
const char32_t TEAPOT = 0x0001FAD6;
const char32_t POURING_LIQUID = 0x0001FAD7;
const char32_t BEANS = 0x0001FAD8;
const char32_t JAR = 0x0001FAD9;
const char32_t GINGER_ROOT = 0x0001FADA;
const char32_t PEA_POD = 0x0001FADB;
const char32_t ROOT_VEGETABLE = 0x0001FADC;
const char32_t SPLATTER = 0x0001FADF;
const char32_t MELTING_FACE = 0x0001FAE0;
const char32_t SALUTING_FACE = 0x0001FAE1;
const char32_t FACE_WITH_OPEN_EYES_AND_HAND_OVER_MOUTH = 0x0001FAE2;
const char32_t FACE_WITH_PEEKING_EYE = 0x0001FAE3;
const char32_t FACE_WITH_DIAGONAL_MOUTH = 0x0001FAE4;
const char32_t DOTTED_LINE_FACE = 0x0001FAE5;
const char32_t BITING_LIP = 0x0001FAE6;
const char32_t BUBBLES = 0x0001FAE7;
const char32_t SHAKING_FACE = 0x0001FAE8;
const char32_t FACE_WITH_BAGS_UNDER_EYES = 0x0001FAE9;
const char32_t DISTORTED_FACE = 0x0001FAEA;
const char32_t FIGHT_CLOUD = 0x0001FAEF;
const char32_t HAND_WITH_INDEX_FINGER_AND_THUMB_CROSSED = 0x0001FAF0;
const char32_t RIGHTWARDS_HAND = 0x0001FAF1;
const char32_t LEFTWARDS_HAND = 0x0001FAF2;
const char32_t PALM_DOWN_HAND = 0x0001FAF3;
const char32_t PALM_UP_HAND = 0x0001FAF4;
const char32_t INDEX_POINTING_AT_THE_VIEWER = 0x0001FAF5;
const char32_t HEART_HANDS = 0x0001FAF6;
const char32_t LEFTWARDS_PUSHING_HAND = 0x0001FAF7;
const char32_t RIGHTWARDS_PUSHING_HAND = 0x0001FAF8;
const char32_t BLOCK_SEXTANT_1 = 0x0001FB00;
const char32_t BLOCK_SEXTANT_2 = 0x0001FB01;
const char32_t BLOCK_SEXTANT_12 = 0x0001FB02;
const char32_t BLOCK_SEXTANT_3 = 0x0001FB03;
const char32_t BLOCK_SEXTANT_13 = 0x0001FB04;
const char32_t BLOCK_SEXTANT_23 = 0x0001FB05;
const char32_t BLOCK_SEXTANT_123 = 0x0001FB06;
const char32_t BLOCK_SEXTANT_4 = 0x0001FB07;
const char32_t BLOCK_SEXTANT_14 = 0x0001FB08;
const char32_t BLOCK_SEXTANT_24 = 0x0001FB09;
const char32_t BLOCK_SEXTANT_124 = 0x0001FB0A;
const char32_t BLOCK_SEXTANT_34 = 0x0001FB0B;
const char32_t BLOCK_SEXTANT_134 = 0x0001FB0C;
const char32_t BLOCK_SEXTANT_234 = 0x0001FB0D;
const char32_t BLOCK_SEXTANT_1234 = 0x0001FB0E;
const char32_t BLOCK_SEXTANT_5 = 0x0001FB0F;
const char32_t BLOCK_SEXTANT_15 = 0x0001FB10;
const char32_t BLOCK_SEXTANT_25 = 0x0001FB11;
const char32_t BLOCK_SEXTANT_125 = 0x0001FB12;
const char32_t BLOCK_SEXTANT_35 = 0x0001FB13;
const char32_t BLOCK_SEXTANT_235 = 0x0001FB14;
const char32_t BLOCK_SEXTANT_1235 = 0x0001FB15;
const char32_t BLOCK_SEXTANT_45 = 0x0001FB16;
const char32_t BLOCK_SEXTANT_145 = 0x0001FB17;
const char32_t BLOCK_SEXTANT_245 = 0x0001FB18;
const char32_t BLOCK_SEXTANT_1245 = 0x0001FB19;
const char32_t BLOCK_SEXTANT_345 = 0x0001FB1A;
const char32_t BLOCK_SEXTANT_1345 = 0x0001FB1B;
const char32_t BLOCK_SEXTANT_2345 = 0x0001FB1C;
const char32_t BLOCK_SEXTANT_12345 = 0x0001FB1D;
const char32_t BLOCK_SEXTANT_6 = 0x0001FB1E;
const char32_t BLOCK_SEXTANT_16 = 0x0001FB1F;
const char32_t BLOCK_SEXTANT_26 = 0x0001FB20;
const char32_t BLOCK_SEXTANT_126 = 0x0001FB21;
const char32_t BLOCK_SEXTANT_36 = 0x0001FB22;
const char32_t BLOCK_SEXTANT_136 = 0x0001FB23;
const char32_t BLOCK_SEXTANT_236 = 0x0001FB24;
const char32_t BLOCK_SEXTANT_1236 = 0x0001FB25;
const char32_t BLOCK_SEXTANT_46 = 0x0001FB26;
const char32_t BLOCK_SEXTANT_146 = 0x0001FB27;
const char32_t BLOCK_SEXTANT_1246 = 0x0001FB28;
const char32_t BLOCK_SEXTANT_346 = 0x0001FB29;
const char32_t BLOCK_SEXTANT_1346 = 0x0001FB2A;
const char32_t BLOCK_SEXTANT_2346 = 0x0001FB2B;
const char32_t BLOCK_SEXTANT_12346 = 0x0001FB2C;
const char32_t BLOCK_SEXTANT_56 = 0x0001FB2D;
const char32_t BLOCK_SEXTANT_156 = 0x0001FB2E;
const char32_t BLOCK_SEXTANT_256 = 0x0001FB2F;
const char32_t BLOCK_SEXTANT_1256 = 0x0001FB30;
const char32_t BLOCK_SEXTANT_356 = 0x0001FB31;
const char32_t BLOCK_SEXTANT_1356 = 0x0001FB32;
const char32_t BLOCK_SEXTANT_2356 = 0x0001FB33;
const char32_t BLOCK_SEXTANT_12356 = 0x0001FB34;
const char32_t BLOCK_SEXTANT_456 = 0x0001FB35;
const char32_t BLOCK_SEXTANT_1456 = 0x0001FB36;
const char32_t BLOCK_SEXTANT_2456 = 0x0001FB37;
const char32_t BLOCK_SEXTANT_12456 = 0x0001FB38;
const char32_t BLOCK_SEXTANT_3456 = 0x0001FB39;
const char32_t BLOCK_SEXTANT_13456 = 0x0001FB3A;
const char32_t BLOCK_SEXTANT_23456 = 0x0001FB3B;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FB3C;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_LOWER_RIGHT = 0x0001FB3D;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FB3E;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_LOWER_RIGHT = 0x0001FB3F;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_LEFT_TO_LOWER_CENTRE = 0x0001FB40;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_UPPER_CENTRE = 0x0001FB41;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_UPPER_RIGHT = 0x0001FB42;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_UPPER_CENTRE = 0x0001FB43;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_UPPER_RIGHT = 0x0001FB44;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_LEFT_TO_UPPER_CENTRE = 0x0001FB45;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_UPPER_MIDDLE_RIGHT = 0x0001FB46;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_CENTRE_TO_LOWER_MIDDLE_RIGHT = 0x0001FB47;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_LEFT_TO_LOWER_MIDDLE_RIGHT = 0x0001FB48;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_CENTRE_TO_UPPER_MIDDLE_RIGHT = 0x0001FB49;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_LEFT_TO_UPPER_MIDDLE_RIGHT = 0x0001FB4A;
const char32_t LOWER_RIGHT_BLOCK_DIAGONAL_LOWER_CENTRE_TO_UPPER_RIGHT = 0x0001FB4B;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_CENTRE_TO_UPPER_MIDDLE_RIGHT = 0x0001FB4C;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_LEFT_TO_UPPER_MIDDLE_RIGHT = 0x0001FB4D;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_CENTRE_TO_LOWER_MIDDLE_RIGHT = 0x0001FB4E;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_LEFT_TO_LOWER_MIDDLE_RIGHT = 0x0001FB4F;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_CENTRE_TO_LOWER_RIGHT = 0x0001FB50;
const char32_t LOWER_LEFT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_LOWER_MIDDLE_RIGHT = 0x0001FB51;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FB52;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_LOWER_RIGHT = 0x0001FB53;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FB54;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_LOWER_RIGHT = 0x0001FB55;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_LEFT_TO_LOWER_CENTRE = 0x0001FB56;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_UPPER_CENTRE = 0x0001FB57;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_UPPER_RIGHT = 0x0001FB58;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_UPPER_CENTRE = 0x0001FB59;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_UPPER_RIGHT = 0x0001FB5A;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_LEFT_TO_UPPER_CENTRE = 0x0001FB5B;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_MIDDLE_LEFT_TO_UPPER_MIDDLE_RIGHT = 0x0001FB5C;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_CENTRE_TO_LOWER_MIDDLE_RIGHT = 0x0001FB5D;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_LEFT_TO_LOWER_MIDDLE_RIGHT = 0x0001FB5E;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_CENTRE_TO_UPPER_MIDDLE_RIGHT = 0x0001FB5F;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_LEFT_TO_UPPER_MIDDLE_RIGHT = 0x0001FB60;
const char32_t UPPER_LEFT_BLOCK_DIAGONAL_LOWER_CENTRE_TO_UPPER_RIGHT = 0x0001FB61;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_CENTRE_TO_UPPER_MIDDLE_RIGHT = 0x0001FB62;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_LEFT_TO_UPPER_MIDDLE_RIGHT = 0x0001FB63;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_CENTRE_TO_LOWER_MIDDLE_RIGHT = 0x0001FB64;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_LEFT_TO_LOWER_MIDDLE_RIGHT = 0x0001FB65;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_CENTRE_TO_LOWER_RIGHT = 0x0001FB66;
const char32_t UPPER_RIGHT_BLOCK_DIAGONAL_UPPER_MIDDLE_LEFT_TO_LOWER_MIDDLE_RIGHT = 0x0001FB67;
const char32_t UPPER_AND_RIGHT_AND_LOWER_TRIANGULAR_THREE_QUARTERS_BLOCK = 0x0001FB68;
const char32_t LEFT_AND_LOWER_AND_RIGHT_TRIANGULAR_THREE_QUARTERS_BLOCK = 0x0001FB69;
const char32_t UPPER_AND_LEFT_AND_LOWER_TRIANGULAR_THREE_QUARTERS_BLOCK = 0x0001FB6A;
const char32_t LEFT_AND_UPPER_AND_RIGHT_TRIANGULAR_THREE_QUARTERS_BLOCK = 0x0001FB6B;
const char32_t LEFT_TRIANGULAR_ONE_QUARTER_BLOCK = 0x0001FB6C;
const char32_t UPPER_TRIANGULAR_ONE_QUARTER_BLOCK = 0x0001FB6D;
const char32_t RIGHT_TRIANGULAR_ONE_QUARTER_BLOCK = 0x0001FB6E;
const char32_t LOWER_TRIANGULAR_ONE_QUARTER_BLOCK = 0x0001FB6F;
const char32_t VERTICAL_ONE_EIGHTH_BLOCK_2 = 0x0001FB70;
const char32_t VERTICAL_ONE_EIGHTH_BLOCK_3 = 0x0001FB71;
const char32_t VERTICAL_ONE_EIGHTH_BLOCK_4 = 0x0001FB72;
const char32_t VERTICAL_ONE_EIGHTH_BLOCK_5 = 0x0001FB73;
const char32_t VERTICAL_ONE_EIGHTH_BLOCK_6 = 0x0001FB74;
const char32_t VERTICAL_ONE_EIGHTH_BLOCK_7 = 0x0001FB75;
const char32_t HORIZONTAL_ONE_EIGHTH_BLOCK_2 = 0x0001FB76;
const char32_t HORIZONTAL_ONE_EIGHTH_BLOCK_3 = 0x0001FB77;
const char32_t HORIZONTAL_ONE_EIGHTH_BLOCK_4 = 0x0001FB78;
const char32_t HORIZONTAL_ONE_EIGHTH_BLOCK_5 = 0x0001FB79;
const char32_t HORIZONTAL_ONE_EIGHTH_BLOCK_6 = 0x0001FB7A;
const char32_t HORIZONTAL_ONE_EIGHTH_BLOCK_7 = 0x0001FB7B;
const char32_t LEFT_AND_LOWER_ONE_EIGHTH_BLOCK = 0x0001FB7C;
const char32_t LEFT_AND_UPPER_ONE_EIGHTH_BLOCK = 0x0001FB7D;
const char32_t RIGHT_AND_UPPER_ONE_EIGHTH_BLOCK = 0x0001FB7E;
const char32_t RIGHT_AND_LOWER_ONE_EIGHTH_BLOCK = 0x0001FB7F;
const char32_t UPPER_AND_LOWER_ONE_EIGHTH_BLOCK = 0x0001FB80;
const char32_t HORIZONTAL_ONE_EIGHTH_BLOCK_1358 = 0x0001FB81;
const char32_t UPPER_ONE_QUARTER_BLOCK = 0x0001FB82;
const char32_t UPPER_THREE_EIGHTHS_BLOCK = 0x0001FB83;
const char32_t UPPER_FIVE_EIGHTHS_BLOCK = 0x0001FB84;
const char32_t UPPER_THREE_QUARTERS_BLOCK = 0x0001FB85;
const char32_t UPPER_SEVEN_EIGHTHS_BLOCK = 0x0001FB86;
const char32_t RIGHT_ONE_QUARTER_BLOCK = 0x0001FB87;
const char32_t RIGHT_THREE_EIGHTHS_BLOCK = 0x0001FB88;
const char32_t RIGHT_FIVE_EIGHTHS_BLOCK = 0x0001FB89;
const char32_t RIGHT_THREE_QUARTERS_BLOCK = 0x0001FB8A;
const char32_t RIGHT_SEVEN_EIGHTHS_BLOCK = 0x0001FB8B;
const char32_t LEFT_HALF_MEDIUM_SHADE = 0x0001FB8C;
const char32_t RIGHT_HALF_MEDIUM_SHADE = 0x0001FB8D;
const char32_t UPPER_HALF_MEDIUM_SHADE = 0x0001FB8E;
const char32_t LOWER_HALF_MEDIUM_SHADE = 0x0001FB8F;
const char32_t INVERSE_MEDIUM_SHADE = 0x0001FB90;
const char32_t UPPER_HALF_BLOCK_AND_LOWER_HALF_INVERSE_MEDIUM_SHADE = 0x0001FB91;
const char32_t UPPER_HALF_INVERSE_MEDIUM_SHADE_AND_LOWER_HALF_BLOCK = 0x0001FB92;
const char32_t LEFT_HALF_INVERSE_MEDIUM_SHADE_AND_RIGHT_HALF_BLOCK = 0x0001FB94;
const char32_t CHECKER_BOARD_FILL = 0x0001FB95;
const char32_t INVERSE_CHECKER_BOARD_FILL = 0x0001FB96;
const char32_t HEAVY_HORIZONTAL_FILL = 0x0001FB97;
const char32_t UPPER_LEFT_TO_LOWER_RIGHT_FILL = 0x0001FB98;
const char32_t UPPER_RIGHT_TO_LOWER_LEFT_FILL = 0x0001FB99;
const char32_t UPPER_AND_LOWER_TRIANGULAR_HALF_BLOCK = 0x0001FB9A;
const char32_t LEFT_AND_RIGHT_TRIANGULAR_HALF_BLOCK = 0x0001FB9B;
const char32_t UPPER_LEFT_TRIANGULAR_MEDIUM_SHADE = 0x0001FB9C;
const char32_t UPPER_RIGHT_TRIANGULAR_MEDIUM_SHADE = 0x0001FB9D;
const char32_t LOWER_RIGHT_TRIANGULAR_MEDIUM_SHADE = 0x0001FB9E;
const char32_t LOWER_LEFT_TRIANGULAR_MEDIUM_SHADE = 0x0001FB9F;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_LEFT = 0x0001FBA0;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_RIGHT = 0x0001FBA1;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FBA2;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_RIGHT_TO_LOWER_CENTRE = 0x0001FBA3;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FBA4;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_RIGHT_TO_LOWER_CENTRE = 0x0001FBA5;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_LEFT_TO_LOWER_CENTRE_TO_MIDDLE_RIGHT = 0x0001FBA6;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_LEFT_TO_UPPER_CENTRE_TO_MIDDLE_RIGHT = 0x0001FBA7;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_LEFT_AND_MIDDLE_RIGHT_TO_LOWER_CENTRE = 0x0001FBA8;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_RIGHT_AND_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FBA9;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_RIGHT_TO_LOWER_CENTRE_TO_MIDDLE_LEFT = 0x0001FBAA;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_MIDDLE_LEFT_TO_LOWER_CENTRE_TO_MIDDLE_RIGHT = 0x0001FBAB;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_LEFT_TO_UPPER_CENTRE_TO_MIDDLE_RIGHT_TO_LOWER_CENTRE = 0x0001FBAC;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_RIGHT_TO_UPPER_CENTRE_TO_MIDDLE_LEFT_TO_LOWER_CENTRE = 0x0001FBAD;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_DIAMOND = 0x0001FBAE;
const char32_t BOX_DRAWINGS_LIGHT_HORIZONTAL_WITH_VERTICAL_STROKE = 0x0001FBAF;
const char32_t ARROWHEAD_SHAPED_POINTER = 0x0001FBB0;
const char32_t INVERSE_CHECK_MARK = 0x0001FBB1;
const char32_t LEFT_HALF_RUNNING_MAN = 0x0001FBB2;
const char32_t RIGHT_HALF_RUNNING_MAN = 0x0001FBB3;
const char32_t INVERSE_DOWNWARDS_ARROW_WITH_TIP_LEFTWARDS = 0x0001FBB4;
const char32_t LEFTWARDS_ARROW_AND_UPPER_AND_LOWER_ONE_EIGHTH_BLOCK = 0x0001FBB5;
const char32_t RIGHTWARDS_ARROW_AND_UPPER_AND_LOWER_ONE_EIGHTH_BLOCK = 0x0001FBB6;
const char32_t DOWNWARDS_ARROW_AND_RIGHT_ONE_EIGHTH_BLOCK = 0x0001FBB7;
const char32_t UPWARDS_ARROW_AND_RIGHT_ONE_EIGHTH_BLOCK = 0x0001FBB8;
const char32_t LEFT_HALF_FOLDER = 0x0001FBB9;
const char32_t RIGHT_HALF_FOLDER = 0x0001FBBA;
const char32_t VOIDED_GREEK_CROSS = 0x0001FBBB;
const char32_t RIGHT_OPEN_SQUARED_DOT = 0x0001FBBC;
const char32_t NEGATIVE_DIAGONAL_CROSS = 0x0001FBBD;
const char32_t NEGATIVE_DIAGONAL_MIDDLE_RIGHT_TO_LOWER_CENTRE = 0x0001FBBE;
const char32_t NEGATIVE_DIAGONAL_DIAMOND = 0x0001FBBF;
const char32_t WHITE_HEAVY_SALTIRE_WITH_ROUNDED_CORNERS = 0x0001FBC0;
const char32_t LEFT_THIRD_WHITE_RIGHT_POINTING_INDEX = 0x0001FBC1;
const char32_t MIDDLE_THIRD_WHITE_RIGHT_POINTING_INDEX = 0x0001FBC2;
const char32_t RIGHT_THIRD_WHITE_RIGHT_POINTING_INDEX = 0x0001FBC3;
const char32_t NEGATIVE_SQUARED_QUESTION_MARK = 0x0001FBC4;
const char32_t STICK_FIGURE = 0x0001FBC5;
const char32_t STICK_FIGURE_WITH_ARMS_RAISED = 0x0001FBC6;
const char32_t STICK_FIGURE_LEANING_LEFT = 0x0001FBC7;
const char32_t STICK_FIGURE_LEANING_RIGHT = 0x0001FBC8;
const char32_t STICK_FIGURE_WITH_DRESS = 0x0001FBC9;
const char32_t WHITE_UP_POINTING_CHEVRON = 0x0001FBCA;
const char32_t WHITE_CROSS_MARK = 0x0001FBCB;
const char32_t RAISED_SMALL_LEFT_SQUARE_BRACKET = 0x0001FBCC;
const char32_t BLACK_SMALL_UP_POINTING_CHEVRON = 0x0001FBCD;
const char32_t LEFT_TWO_THIRDS_BLOCK = 0x0001FBCE;
const char32_t LEFT_ONE_THIRD_BLOCK = 0x0001FBCF;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_RIGHT_TO_LOWER_LEFT = 0x0001FBD0;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_RIGHT_TO_MIDDLE_LEFT = 0x0001FBD1;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_LEFT_TO_MIDDLE_RIGHT = 0x0001FBD2;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_MIDDLE_LEFT_TO_LOWER_RIGHT = 0x0001FBD3;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_LEFT_TO_LOWER_CENTRE = 0x0001FBD4;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_LOWER_RIGHT = 0x0001FBD5;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_RIGHT_TO_LOWER_CENTRE = 0x0001FBD6;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_CENTRE_TO_LOWER_LEFT = 0x0001FBD7;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_LEFT_TO_MIDDLE_CENTRE_TO_UPPER_RIGHT = 0x0001FBD8;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_RIGHT_TO_MIDDLE_CENTRE_TO_LOWER_RIGHT = 0x0001FBD9;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_LOWER_LEFT_TO_MIDDLE_CENTRE_TO_LOWER_RIGHT = 0x0001FBDA;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_LEFT_TO_MIDDLE_CENTRE_TO_LOWER_LEFT = 0x0001FBDB;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_LEFT_TO_LOWER_CENTRE_TO_UPPER_RIGHT = 0x0001FBDC;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_RIGHT_TO_MIDDLE_LEFT_TO_LOWER_RIGHT = 0x0001FBDD;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_LOWER_LEFT_TO_UPPER_CENTRE_TO_LOWER_RIGHT = 0x0001FBDE;
const char32_t BOX_DRAWINGS_LIGHT_DIAGONAL_UPPER_LEFT_TO_MIDDLE_RIGHT_TO_LOWER_LEFT = 0x0001FBDF;
const char32_t TOP_JUSTIFIED_LOWER_HALF_WHITE_CIRCLE = 0x0001FBE0;
const char32_t RIGHT_JUSTIFIED_LEFT_HALF_WHITE_CIRCLE = 0x0001FBE1;
const char32_t BOTTOM_JUSTIFIED_UPPER_HALF_WHITE_CIRCLE = 0x0001FBE2;
const char32_t LEFT_JUSTIFIED_RIGHT_HALF_WHITE_CIRCLE = 0x0001FBE3;
const char32_t UPPER_CENTRE_ONE_QUARTER_BLOCK = 0x0001FBE4;
const char32_t LOWER_CENTRE_ONE_QUARTER_BLOCK = 0x0001FBE5;
const char32_t MIDDLE_LEFT_ONE_QUARTER_BLOCK = 0x0001FBE6;
const char32_t MIDDLE_RIGHT_ONE_QUARTER_BLOCK = 0x0001FBE7;
const char32_t TOP_JUSTIFIED_LOWER_HALF_BLACK_CIRCLE = 0x0001FBE8;
const char32_t RIGHT_JUSTIFIED_LEFT_HALF_BLACK_CIRCLE = 0x0001FBE9;
const char32_t BOTTOM_JUSTIFIED_UPPER_HALF_BLACK_CIRCLE = 0x0001FBEA;
const char32_t LEFT_JUSTIFIED_RIGHT_HALF_BLACK_CIRCLE = 0x0001FBEB;
const char32_t TOP_RIGHT_JUSTIFIED_LOWER_LEFT_QUARTER_BLACK_CIRCLE = 0x0001FBEC;
const char32_t BOTTOM_LEFT_JUSTIFIED_UPPER_RIGHT_QUARTER_BLACK_CIRCLE = 0x0001FBED;
const char32_t BOTTOM_RIGHT_JUSTIFIED_UPPER_LEFT_QUARTER_BLACK_CIRCLE = 0x0001FBEE;
const char32_t TOP_LEFT_JUSTIFIED_LOWER_RIGHT_QUARTER_BLACK_CIRCLE = 0x0001FBEF;
const char32_t SEGMENTED_DIGIT_ZERO = 0x0001FBF0;
const char32_t SEGMENTED_DIGIT_ONE = 0x0001FBF1;
const char32_t SEGMENTED_DIGIT_TWO = 0x0001FBF2;
const char32_t SEGMENTED_DIGIT_THREE = 0x0001FBF3;
const char32_t SEGMENTED_DIGIT_FOUR = 0x0001FBF4;
const char32_t SEGMENTED_DIGIT_FIVE = 0x0001FBF5;
const char32_t SEGMENTED_DIGIT_SIX = 0x0001FBF6;
const char32_t SEGMENTED_DIGIT_SEVEN = 0x0001FBF7;
const char32_t SEGMENTED_DIGIT_EIGHT = 0x0001FBF8;
const char32_t SEGMENTED_DIGIT_NINE = 0x0001FBF9;
const char32_t ALARM_BELL_SYMBOL = 0x0001FBFA;
const char32_t CJK_IDEOGRAPH_EXTENSION_B_FIRST = 0x00020000;
const char32_t CJK_IDEOGRAPH_EXTENSION_B_LAST = 0x0002A6DF;
const char32_t CJK_IDEOGRAPH_EXTENSION_C_FIRST = 0x0002A700;
const char32_t CJK_IDEOGRAPH_EXTENSION_C_LAST = 0x0002B73F;
const char32_t CJK_IDEOGRAPH_EXTENSION_D_FIRST = 0x0002B740;
const char32_t CJK_IDEOGRAPH_EXTENSION_D_LAST = 0x0002B81D;
const char32_t CJK_IDEOGRAPH_EXTENSION_E_FIRST = 0x0002B820;
const char32_t CJK_IDEOGRAPH_EXTENSION_E_LAST = 0x0002CEAD;
const char32_t CJK_IDEOGRAPH_EXTENSION_F_FIRST = 0x0002CEB0;
const char32_t CJK_IDEOGRAPH_EXTENSION_F_LAST = 0x0002EBE0;
const char32_t CJK_IDEOGRAPH_EXTENSION_I_FIRST = 0x0002EBF0;
const char32_t CJK_IDEOGRAPH_EXTENSION_I_LAST = 0x0002EE5D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F800 = 0x0002F800;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F801 = 0x0002F801;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F802 = 0x0002F802;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F803 = 0x0002F803;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F804 = 0x0002F804;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F805 = 0x0002F805;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F806 = 0x0002F806;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F807 = 0x0002F807;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F808 = 0x0002F808;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F809 = 0x0002F809;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F80A = 0x0002F80A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F80B = 0x0002F80B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F80C = 0x0002F80C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F80D = 0x0002F80D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F80E = 0x0002F80E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F80F = 0x0002F80F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F810 = 0x0002F810;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F811 = 0x0002F811;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F812 = 0x0002F812;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F813 = 0x0002F813;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F814 = 0x0002F814;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F815 = 0x0002F815;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F816 = 0x0002F816;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F817 = 0x0002F817;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F818 = 0x0002F818;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F819 = 0x0002F819;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F81A = 0x0002F81A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F81B = 0x0002F81B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F81C = 0x0002F81C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F81D = 0x0002F81D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F81E = 0x0002F81E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F81F = 0x0002F81F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F820 = 0x0002F820;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F821 = 0x0002F821;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F822 = 0x0002F822;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F823 = 0x0002F823;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F824 = 0x0002F824;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F825 = 0x0002F825;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F826 = 0x0002F826;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F827 = 0x0002F827;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F828 = 0x0002F828;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F829 = 0x0002F829;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F82A = 0x0002F82A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F82B = 0x0002F82B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F82C = 0x0002F82C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F82D = 0x0002F82D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F82E = 0x0002F82E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F82F = 0x0002F82F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F830 = 0x0002F830;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F831 = 0x0002F831;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F832 = 0x0002F832;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F833 = 0x0002F833;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F834 = 0x0002F834;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F835 = 0x0002F835;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F836 = 0x0002F836;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F837 = 0x0002F837;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F838 = 0x0002F838;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F839 = 0x0002F839;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F83A = 0x0002F83A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F83B = 0x0002F83B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F83C = 0x0002F83C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F83D = 0x0002F83D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F83E = 0x0002F83E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F83F = 0x0002F83F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F840 = 0x0002F840;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F841 = 0x0002F841;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F842 = 0x0002F842;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F843 = 0x0002F843;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F844 = 0x0002F844;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F845 = 0x0002F845;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F846 = 0x0002F846;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F847 = 0x0002F847;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F848 = 0x0002F848;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F849 = 0x0002F849;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F84A = 0x0002F84A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F84B = 0x0002F84B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F84C = 0x0002F84C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F84D = 0x0002F84D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F84E = 0x0002F84E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F84F = 0x0002F84F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F850 = 0x0002F850;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F851 = 0x0002F851;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F852 = 0x0002F852;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F853 = 0x0002F853;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F854 = 0x0002F854;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F855 = 0x0002F855;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F856 = 0x0002F856;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F857 = 0x0002F857;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F858 = 0x0002F858;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F859 = 0x0002F859;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F85A = 0x0002F85A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F85B = 0x0002F85B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F85C = 0x0002F85C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F85D = 0x0002F85D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F85E = 0x0002F85E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F85F = 0x0002F85F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F860 = 0x0002F860;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F861 = 0x0002F861;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F862 = 0x0002F862;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F863 = 0x0002F863;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F864 = 0x0002F864;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F865 = 0x0002F865;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F866 = 0x0002F866;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F867 = 0x0002F867;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F868 = 0x0002F868;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F869 = 0x0002F869;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F86A = 0x0002F86A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F86B = 0x0002F86B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F86C = 0x0002F86C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F86D = 0x0002F86D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F86E = 0x0002F86E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F86F = 0x0002F86F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F870 = 0x0002F870;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F871 = 0x0002F871;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F872 = 0x0002F872;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F873 = 0x0002F873;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F874 = 0x0002F874;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F875 = 0x0002F875;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F876 = 0x0002F876;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F877 = 0x0002F877;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F878 = 0x0002F878;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F879 = 0x0002F879;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F87A = 0x0002F87A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F87B = 0x0002F87B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F87C = 0x0002F87C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F87D = 0x0002F87D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F87E = 0x0002F87E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F87F = 0x0002F87F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F880 = 0x0002F880;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F881 = 0x0002F881;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F882 = 0x0002F882;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F883 = 0x0002F883;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F884 = 0x0002F884;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F885 = 0x0002F885;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F886 = 0x0002F886;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F887 = 0x0002F887;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F888 = 0x0002F888;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F889 = 0x0002F889;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F88A = 0x0002F88A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F88B = 0x0002F88B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F88C = 0x0002F88C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F88D = 0x0002F88D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F88E = 0x0002F88E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F88F = 0x0002F88F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F890 = 0x0002F890;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F891 = 0x0002F891;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F892 = 0x0002F892;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F893 = 0x0002F893;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F894 = 0x0002F894;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F895 = 0x0002F895;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F896 = 0x0002F896;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F897 = 0x0002F897;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F898 = 0x0002F898;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F899 = 0x0002F899;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F89A = 0x0002F89A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F89B = 0x0002F89B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F89C = 0x0002F89C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F89D = 0x0002F89D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F89E = 0x0002F89E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F89F = 0x0002F89F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A0 = 0x0002F8A0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A1 = 0x0002F8A1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A2 = 0x0002F8A2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A3 = 0x0002F8A3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A4 = 0x0002F8A4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A5 = 0x0002F8A5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A6 = 0x0002F8A6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A7 = 0x0002F8A7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A8 = 0x0002F8A8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8A9 = 0x0002F8A9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8AA = 0x0002F8AA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8AB = 0x0002F8AB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8AC = 0x0002F8AC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8AD = 0x0002F8AD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8AE = 0x0002F8AE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8AF = 0x0002F8AF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B0 = 0x0002F8B0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B1 = 0x0002F8B1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B2 = 0x0002F8B2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B3 = 0x0002F8B3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B4 = 0x0002F8B4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B5 = 0x0002F8B5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B6 = 0x0002F8B6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B7 = 0x0002F8B7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B8 = 0x0002F8B8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8B9 = 0x0002F8B9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8BA = 0x0002F8BA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8BB = 0x0002F8BB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8BC = 0x0002F8BC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8BD = 0x0002F8BD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8BE = 0x0002F8BE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8BF = 0x0002F8BF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C0 = 0x0002F8C0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C1 = 0x0002F8C1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C2 = 0x0002F8C2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C3 = 0x0002F8C3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C4 = 0x0002F8C4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C5 = 0x0002F8C5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C6 = 0x0002F8C6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C7 = 0x0002F8C7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C8 = 0x0002F8C8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8C9 = 0x0002F8C9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8CA = 0x0002F8CA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8CB = 0x0002F8CB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8CC = 0x0002F8CC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8CD = 0x0002F8CD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8CE = 0x0002F8CE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8CF = 0x0002F8CF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D0 = 0x0002F8D0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D1 = 0x0002F8D1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D2 = 0x0002F8D2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D3 = 0x0002F8D3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D4 = 0x0002F8D4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D5 = 0x0002F8D5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D6 = 0x0002F8D6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D7 = 0x0002F8D7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D8 = 0x0002F8D8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8D9 = 0x0002F8D9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8DA = 0x0002F8DA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8DB = 0x0002F8DB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8DC = 0x0002F8DC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8DD = 0x0002F8DD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8DE = 0x0002F8DE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8DF = 0x0002F8DF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E0 = 0x0002F8E0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E1 = 0x0002F8E1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E2 = 0x0002F8E2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E3 = 0x0002F8E3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E4 = 0x0002F8E4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E5 = 0x0002F8E5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E6 = 0x0002F8E6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E7 = 0x0002F8E7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E8 = 0x0002F8E8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8E9 = 0x0002F8E9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8EA = 0x0002F8EA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8EB = 0x0002F8EB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8EC = 0x0002F8EC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8ED = 0x0002F8ED;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8EE = 0x0002F8EE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8EF = 0x0002F8EF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F0 = 0x0002F8F0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F1 = 0x0002F8F1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F2 = 0x0002F8F2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F3 = 0x0002F8F3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F4 = 0x0002F8F4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F5 = 0x0002F8F5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F6 = 0x0002F8F6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F7 = 0x0002F8F7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F8 = 0x0002F8F8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8F9 = 0x0002F8F9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8FA = 0x0002F8FA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8FB = 0x0002F8FB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8FC = 0x0002F8FC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8FD = 0x0002F8FD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8FE = 0x0002F8FE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F8FF = 0x0002F8FF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F900 = 0x0002F900;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F901 = 0x0002F901;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F902 = 0x0002F902;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F903 = 0x0002F903;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F904 = 0x0002F904;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F905 = 0x0002F905;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F906 = 0x0002F906;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F907 = 0x0002F907;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F908 = 0x0002F908;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F909 = 0x0002F909;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F90A = 0x0002F90A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F90B = 0x0002F90B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F90C = 0x0002F90C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F90D = 0x0002F90D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F90E = 0x0002F90E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F90F = 0x0002F90F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F910 = 0x0002F910;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F911 = 0x0002F911;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F912 = 0x0002F912;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F913 = 0x0002F913;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F914 = 0x0002F914;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F915 = 0x0002F915;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F916 = 0x0002F916;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F917 = 0x0002F917;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F918 = 0x0002F918;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F919 = 0x0002F919;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F91A = 0x0002F91A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F91B = 0x0002F91B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F91C = 0x0002F91C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F91D = 0x0002F91D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F91E = 0x0002F91E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F91F = 0x0002F91F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F920 = 0x0002F920;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F921 = 0x0002F921;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F922 = 0x0002F922;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F923 = 0x0002F923;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F924 = 0x0002F924;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F925 = 0x0002F925;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F926 = 0x0002F926;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F927 = 0x0002F927;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F928 = 0x0002F928;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F929 = 0x0002F929;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F92A = 0x0002F92A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F92B = 0x0002F92B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F92C = 0x0002F92C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F92D = 0x0002F92D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F92E = 0x0002F92E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F92F = 0x0002F92F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F930 = 0x0002F930;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F931 = 0x0002F931;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F932 = 0x0002F932;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F933 = 0x0002F933;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F934 = 0x0002F934;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F935 = 0x0002F935;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F936 = 0x0002F936;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F937 = 0x0002F937;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F938 = 0x0002F938;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F939 = 0x0002F939;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F93A = 0x0002F93A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F93B = 0x0002F93B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F93C = 0x0002F93C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F93D = 0x0002F93D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F93E = 0x0002F93E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F93F = 0x0002F93F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F940 = 0x0002F940;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F941 = 0x0002F941;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F942 = 0x0002F942;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F943 = 0x0002F943;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F944 = 0x0002F944;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F945 = 0x0002F945;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F946 = 0x0002F946;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F947 = 0x0002F947;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F948 = 0x0002F948;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F949 = 0x0002F949;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F94A = 0x0002F94A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F94B = 0x0002F94B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F94C = 0x0002F94C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F94D = 0x0002F94D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F94E = 0x0002F94E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F94F = 0x0002F94F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F950 = 0x0002F950;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F951 = 0x0002F951;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F952 = 0x0002F952;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F953 = 0x0002F953;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F954 = 0x0002F954;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F955 = 0x0002F955;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F956 = 0x0002F956;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F957 = 0x0002F957;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F958 = 0x0002F958;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F959 = 0x0002F959;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F95A = 0x0002F95A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F95B = 0x0002F95B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F95C = 0x0002F95C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F95D = 0x0002F95D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F95E = 0x0002F95E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F95F = 0x0002F95F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F960 = 0x0002F960;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F961 = 0x0002F961;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F962 = 0x0002F962;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F963 = 0x0002F963;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F964 = 0x0002F964;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F965 = 0x0002F965;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F966 = 0x0002F966;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F967 = 0x0002F967;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F968 = 0x0002F968;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F969 = 0x0002F969;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F96A = 0x0002F96A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F96B = 0x0002F96B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F96C = 0x0002F96C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F96D = 0x0002F96D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F96E = 0x0002F96E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F96F = 0x0002F96F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F970 = 0x0002F970;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F971 = 0x0002F971;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F972 = 0x0002F972;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F973 = 0x0002F973;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F974 = 0x0002F974;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F975 = 0x0002F975;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F976 = 0x0002F976;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F977 = 0x0002F977;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F978 = 0x0002F978;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F979 = 0x0002F979;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F97A = 0x0002F97A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F97B = 0x0002F97B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F97C = 0x0002F97C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F97D = 0x0002F97D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F97E = 0x0002F97E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F97F = 0x0002F97F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F980 = 0x0002F980;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F981 = 0x0002F981;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F982 = 0x0002F982;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F983 = 0x0002F983;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F984 = 0x0002F984;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F985 = 0x0002F985;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F986 = 0x0002F986;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F987 = 0x0002F987;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F988 = 0x0002F988;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F989 = 0x0002F989;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F98A = 0x0002F98A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F98B = 0x0002F98B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F98C = 0x0002F98C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F98D = 0x0002F98D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F98E = 0x0002F98E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F98F = 0x0002F98F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F990 = 0x0002F990;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F991 = 0x0002F991;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F992 = 0x0002F992;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F993 = 0x0002F993;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F994 = 0x0002F994;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F995 = 0x0002F995;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F996 = 0x0002F996;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F997 = 0x0002F997;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F998 = 0x0002F998;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F999 = 0x0002F999;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F99A = 0x0002F99A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F99B = 0x0002F99B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F99C = 0x0002F99C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F99D = 0x0002F99D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F99E = 0x0002F99E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F99F = 0x0002F99F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A0 = 0x0002F9A0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A1 = 0x0002F9A1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A2 = 0x0002F9A2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A3 = 0x0002F9A3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A4 = 0x0002F9A4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A5 = 0x0002F9A5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A6 = 0x0002F9A6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A7 = 0x0002F9A7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A8 = 0x0002F9A8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9A9 = 0x0002F9A9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9AA = 0x0002F9AA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9AB = 0x0002F9AB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9AC = 0x0002F9AC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9AD = 0x0002F9AD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9AE = 0x0002F9AE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9AF = 0x0002F9AF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B0 = 0x0002F9B0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B1 = 0x0002F9B1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B2 = 0x0002F9B2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B3 = 0x0002F9B3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B4 = 0x0002F9B4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B5 = 0x0002F9B5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B6 = 0x0002F9B6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B7 = 0x0002F9B7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B8 = 0x0002F9B8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9B9 = 0x0002F9B9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9BA = 0x0002F9BA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9BB = 0x0002F9BB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9BC = 0x0002F9BC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9BD = 0x0002F9BD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9BE = 0x0002F9BE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9BF = 0x0002F9BF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C0 = 0x0002F9C0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C1 = 0x0002F9C1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C2 = 0x0002F9C2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C3 = 0x0002F9C3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C4 = 0x0002F9C4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C5 = 0x0002F9C5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C6 = 0x0002F9C6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C7 = 0x0002F9C7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C8 = 0x0002F9C8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9C9 = 0x0002F9C9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9CA = 0x0002F9CA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9CB = 0x0002F9CB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9CC = 0x0002F9CC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9CD = 0x0002F9CD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9CE = 0x0002F9CE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9CF = 0x0002F9CF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D0 = 0x0002F9D0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D1 = 0x0002F9D1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D2 = 0x0002F9D2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D3 = 0x0002F9D3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D4 = 0x0002F9D4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D5 = 0x0002F9D5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D6 = 0x0002F9D6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D7 = 0x0002F9D7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D8 = 0x0002F9D8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9D9 = 0x0002F9D9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9DA = 0x0002F9DA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9DB = 0x0002F9DB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9DC = 0x0002F9DC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9DD = 0x0002F9DD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9DE = 0x0002F9DE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9DF = 0x0002F9DF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E0 = 0x0002F9E0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E1 = 0x0002F9E1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E2 = 0x0002F9E2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E3 = 0x0002F9E3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E4 = 0x0002F9E4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E5 = 0x0002F9E5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E6 = 0x0002F9E6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E7 = 0x0002F9E7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E8 = 0x0002F9E8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9E9 = 0x0002F9E9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9EA = 0x0002F9EA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9EB = 0x0002F9EB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9EC = 0x0002F9EC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9ED = 0x0002F9ED;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9EE = 0x0002F9EE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9EF = 0x0002F9EF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F0 = 0x0002F9F0;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F1 = 0x0002F9F1;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F2 = 0x0002F9F2;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F3 = 0x0002F9F3;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F4 = 0x0002F9F4;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F5 = 0x0002F9F5;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F6 = 0x0002F9F6;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F7 = 0x0002F9F7;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F8 = 0x0002F9F8;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9F9 = 0x0002F9F9;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9FA = 0x0002F9FA;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9FB = 0x0002F9FB;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9FC = 0x0002F9FC;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9FD = 0x0002F9FD;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9FE = 0x0002F9FE;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2F9FF = 0x0002F9FF;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA00 = 0x0002FA00;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA01 = 0x0002FA01;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA02 = 0x0002FA02;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA03 = 0x0002FA03;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA04 = 0x0002FA04;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA05 = 0x0002FA05;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA06 = 0x0002FA06;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA07 = 0x0002FA07;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA08 = 0x0002FA08;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA09 = 0x0002FA09;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA0A = 0x0002FA0A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA0B = 0x0002FA0B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA0C = 0x0002FA0C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA0D = 0x0002FA0D;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA0E = 0x0002FA0E;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA0F = 0x0002FA0F;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA10 = 0x0002FA10;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA11 = 0x0002FA11;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA12 = 0x0002FA12;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA13 = 0x0002FA13;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA14 = 0x0002FA14;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA15 = 0x0002FA15;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA16 = 0x0002FA16;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA17 = 0x0002FA17;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA18 = 0x0002FA18;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA19 = 0x0002FA19;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA1A = 0x0002FA1A;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA1B = 0x0002FA1B;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA1C = 0x0002FA1C;
const char32_t CJK_COMPATIBILITY_IDEOGRAPH_2FA1D = 0x0002FA1D;
const char32_t CJK_IDEOGRAPH_EXTENSION_G_FIRST = 0x00030000;
const char32_t CJK_IDEOGRAPH_EXTENSION_G_LAST = 0x0003134A;
const char32_t CJK_IDEOGRAPH_EXTENSION_H_FIRST = 0x00031350;
const char32_t CJK_IDEOGRAPH_EXTENSION_H_LAST = 0x000323AF;
const char32_t CJK_IDEOGRAPH_EXTENSION_J_FIRST = 0x000323B0;
const char32_t CJK_IDEOGRAPH_EXTENSION_J_LAST = 0x00033479;
const char32_t LANGUAGE_TAG = 0x000E0001;
const char32_t TAG_SPACE = 0x000E0020;
const char32_t TAG_EXCLAMATION_MARK = 0x000E0021;
const char32_t TAG_QUOTATION_MARK = 0x000E0022;
const char32_t TAG_NUMBER_SIGN = 0x000E0023;
const char32_t TAG_DOLLAR_SIGN = 0x000E0024;
const char32_t TAG_PERCENT_SIGN = 0x000E0025;
const char32_t TAG_AMPERSAND = 0x000E0026;
const char32_t TAG_APOSTROPHE = 0x000E0027;
const char32_t TAG_LEFT_PARENTHESIS = 0x000E0028;
const char32_t TAG_RIGHT_PARENTHESIS = 0x000E0029;
const char32_t TAG_ASTERISK = 0x000E002A;
const char32_t TAG_PLUS_SIGN = 0x000E002B;
const char32_t TAG_COMMA = 0x000E002C;
const char32_t TAG_HYPHEN_MINUS = 0x000E002D;
const char32_t TAG_FULL_STOP = 0x000E002E;
const char32_t TAG_SOLIDUS = 0x000E002F;
const char32_t TAG_DIGIT_ZERO = 0x000E0030;
const char32_t TAG_DIGIT_ONE = 0x000E0031;
const char32_t TAG_DIGIT_TWO = 0x000E0032;
const char32_t TAG_DIGIT_THREE = 0x000E0033;
const char32_t TAG_DIGIT_FOUR = 0x000E0034;
const char32_t TAG_DIGIT_FIVE = 0x000E0035;
const char32_t TAG_DIGIT_SIX = 0x000E0036;
const char32_t TAG_DIGIT_SEVEN = 0x000E0037;
const char32_t TAG_DIGIT_EIGHT = 0x000E0038;
const char32_t TAG_DIGIT_NINE = 0x000E0039;
const char32_t TAG_COLON = 0x000E003A;
const char32_t TAG_SEMICOLON = 0x000E003B;
const char32_t TAG_LESS_THAN_SIGN = 0x000E003C;
const char32_t TAG_EQUALS_SIGN = 0x000E003D;
const char32_t TAG_GREATER_THAN_SIGN = 0x000E003E;
const char32_t TAG_QUESTION_MARK = 0x000E003F;
const char32_t TAG_COMMERCIAL_AT = 0x000E0040;
const char32_t TAG_LATIN_CAPITAL_LETTER_A = 0x000E0041;
const char32_t TAG_LATIN_CAPITAL_LETTER_B = 0x000E0042;
const char32_t TAG_LATIN_CAPITAL_LETTER_C = 0x000E0043;
const char32_t TAG_LATIN_CAPITAL_LETTER_D = 0x000E0044;
const char32_t TAG_LATIN_CAPITAL_LETTER_E = 0x000E0045;
const char32_t TAG_LATIN_CAPITAL_LETTER_F = 0x000E0046;
const char32_t TAG_LATIN_CAPITAL_LETTER_G = 0x000E0047;
const char32_t TAG_LATIN_CAPITAL_LETTER_H = 0x000E0048;
const char32_t TAG_LATIN_CAPITAL_LETTER_I = 0x000E0049;
const char32_t TAG_LATIN_CAPITAL_LETTER_J = 0x000E004A;
const char32_t TAG_LATIN_CAPITAL_LETTER_K = 0x000E004B;
const char32_t TAG_LATIN_CAPITAL_LETTER_L = 0x000E004C;
const char32_t TAG_LATIN_CAPITAL_LETTER_M = 0x000E004D;
const char32_t TAG_LATIN_CAPITAL_LETTER_N = 0x000E004E;
const char32_t TAG_LATIN_CAPITAL_LETTER_O = 0x000E004F;
const char32_t TAG_LATIN_CAPITAL_LETTER_P = 0x000E0050;
const char32_t TAG_LATIN_CAPITAL_LETTER_Q = 0x000E0051;
const char32_t TAG_LATIN_CAPITAL_LETTER_R = 0x000E0052;
const char32_t TAG_LATIN_CAPITAL_LETTER_S = 0x000E0053;
const char32_t TAG_LATIN_CAPITAL_LETTER_T = 0x000E0054;
const char32_t TAG_LATIN_CAPITAL_LETTER_U = 0x000E0055;
const char32_t TAG_LATIN_CAPITAL_LETTER_V = 0x000E0056;
const char32_t TAG_LATIN_CAPITAL_LETTER_W = 0x000E0057;
const char32_t TAG_LATIN_CAPITAL_LETTER_X = 0x000E0058;
const char32_t TAG_LATIN_CAPITAL_LETTER_Y = 0x000E0059;
const char32_t TAG_LATIN_CAPITAL_LETTER_Z = 0x000E005A;
const char32_t TAG_LEFT_SQUARE_BRACKET = 0x000E005B;
const char32_t TAG_REVERSE_SOLIDUS = 0x000E005C;
const char32_t TAG_RIGHT_SQUARE_BRACKET = 0x000E005D;
const char32_t TAG_CIRCUMFLEX_ACCENT = 0x000E005E;
const char32_t TAG_LOW_LINE = 0x000E005F;
const char32_t TAG_GRAVE_ACCENT = 0x000E0060;
const char32_t TAG_LATIN_SMALL_LETTER_A = 0x000E0061;
const char32_t TAG_LATIN_SMALL_LETTER_B = 0x000E0062;
const char32_t TAG_LATIN_SMALL_LETTER_C = 0x000E0063;
const char32_t TAG_LATIN_SMALL_LETTER_D = 0x000E0064;
const char32_t TAG_LATIN_SMALL_LETTER_E = 0x000E0065;
const char32_t TAG_LATIN_SMALL_LETTER_F = 0x000E0066;
const char32_t TAG_LATIN_SMALL_LETTER_G = 0x000E0067;
const char32_t TAG_LATIN_SMALL_LETTER_H = 0x000E0068;
const char32_t TAG_LATIN_SMALL_LETTER_I = 0x000E0069;
const char32_t TAG_LATIN_SMALL_LETTER_J = 0x000E006A;
const char32_t TAG_LATIN_SMALL_LETTER_K = 0x000E006B;
const char32_t TAG_LATIN_SMALL_LETTER_L = 0x000E006C;
const char32_t TAG_LATIN_SMALL_LETTER_M = 0x000E006D;
const char32_t TAG_LATIN_SMALL_LETTER_N = 0x000E006E;
const char32_t TAG_LATIN_SMALL_LETTER_O = 0x000E006F;
const char32_t TAG_LATIN_SMALL_LETTER_P = 0x000E0070;
const char32_t TAG_LATIN_SMALL_LETTER_Q = 0x000E0071;
const char32_t TAG_LATIN_SMALL_LETTER_R = 0x000E0072;
const char32_t TAG_LATIN_SMALL_LETTER_S = 0x000E0073;
const char32_t TAG_LATIN_SMALL_LETTER_T = 0x000E0074;
const char32_t TAG_LATIN_SMALL_LETTER_U = 0x000E0075;
const char32_t TAG_LATIN_SMALL_LETTER_V = 0x000E0076;
const char32_t TAG_LATIN_SMALL_LETTER_W = 0x000E0077;
const char32_t TAG_LATIN_SMALL_LETTER_X = 0x000E0078;
const char32_t TAG_LATIN_SMALL_LETTER_Y = 0x000E0079;
const char32_t TAG_LATIN_SMALL_LETTER_Z = 0x000E007A;
const char32_t TAG_LEFT_CURLY_BRACKET = 0x000E007B;
const char32_t TAG_VERTICAL_LINE = 0x000E007C;
const char32_t TAG_RIGHT_CURLY_BRACKET = 0x000E007D;
const char32_t TAG_TILDE = 0x000E007E;
const char32_t CANCEL_TAG = 0x000E007F;
const char32_t VARIATION_SELECTOR_17 = 0x000E0100;
const char32_t VARIATION_SELECTOR_18 = 0x000E0101;
const char32_t VARIATION_SELECTOR_19 = 0x000E0102;
const char32_t VARIATION_SELECTOR_20 = 0x000E0103;
const char32_t VARIATION_SELECTOR_21 = 0x000E0104;
const char32_t VARIATION_SELECTOR_22 = 0x000E0105;
const char32_t VARIATION_SELECTOR_23 = 0x000E0106;
const char32_t VARIATION_SELECTOR_24 = 0x000E0107;
const char32_t VARIATION_SELECTOR_25 = 0x000E0108;
const char32_t VARIATION_SELECTOR_26 = 0x000E0109;
const char32_t VARIATION_SELECTOR_27 = 0x000E010A;
const char32_t VARIATION_SELECTOR_28 = 0x000E010B;
const char32_t VARIATION_SELECTOR_29 = 0x000E010C;
const char32_t VARIATION_SELECTOR_30 = 0x000E010D;
const char32_t VARIATION_SELECTOR_31 = 0x000E010E;
const char32_t VARIATION_SELECTOR_32 = 0x000E010F;
const char32_t VARIATION_SELECTOR_33 = 0x000E0110;
const char32_t VARIATION_SELECTOR_34 = 0x000E0111;
const char32_t VARIATION_SELECTOR_35 = 0x000E0112;
const char32_t VARIATION_SELECTOR_36 = 0x000E0113;
const char32_t VARIATION_SELECTOR_37 = 0x000E0114;
const char32_t VARIATION_SELECTOR_38 = 0x000E0115;
const char32_t VARIATION_SELECTOR_39 = 0x000E0116;
const char32_t VARIATION_SELECTOR_40 = 0x000E0117;
const char32_t VARIATION_SELECTOR_41 = 0x000E0118;
const char32_t VARIATION_SELECTOR_42 = 0x000E0119;
const char32_t VARIATION_SELECTOR_43 = 0x000E011A;
const char32_t VARIATION_SELECTOR_44 = 0x000E011B;
const char32_t VARIATION_SELECTOR_45 = 0x000E011C;
const char32_t VARIATION_SELECTOR_46 = 0x000E011D;
const char32_t VARIATION_SELECTOR_47 = 0x000E011E;
const char32_t VARIATION_SELECTOR_48 = 0x000E011F;
const char32_t VARIATION_SELECTOR_49 = 0x000E0120;
const char32_t VARIATION_SELECTOR_50 = 0x000E0121;
const char32_t VARIATION_SELECTOR_51 = 0x000E0122;
const char32_t VARIATION_SELECTOR_52 = 0x000E0123;
const char32_t VARIATION_SELECTOR_53 = 0x000E0124;
const char32_t VARIATION_SELECTOR_54 = 0x000E0125;
const char32_t VARIATION_SELECTOR_55 = 0x000E0126;
const char32_t VARIATION_SELECTOR_56 = 0x000E0127;
const char32_t VARIATION_SELECTOR_57 = 0x000E0128;
const char32_t VARIATION_SELECTOR_58 = 0x000E0129;
const char32_t VARIATION_SELECTOR_59 = 0x000E012A;
const char32_t VARIATION_SELECTOR_60 = 0x000E012B;
const char32_t VARIATION_SELECTOR_61 = 0x000E012C;
const char32_t VARIATION_SELECTOR_62 = 0x000E012D;
const char32_t VARIATION_SELECTOR_63 = 0x000E012E;
const char32_t VARIATION_SELECTOR_64 = 0x000E012F;
const char32_t VARIATION_SELECTOR_65 = 0x000E0130;
const char32_t VARIATION_SELECTOR_66 = 0x000E0131;
const char32_t VARIATION_SELECTOR_67 = 0x000E0132;
const char32_t VARIATION_SELECTOR_68 = 0x000E0133;
const char32_t VARIATION_SELECTOR_69 = 0x000E0134;
const char32_t VARIATION_SELECTOR_70 = 0x000E0135;
const char32_t VARIATION_SELECTOR_71 = 0x000E0136;
const char32_t VARIATION_SELECTOR_72 = 0x000E0137;
const char32_t VARIATION_SELECTOR_73 = 0x000E0138;
const char32_t VARIATION_SELECTOR_74 = 0x000E0139;
const char32_t VARIATION_SELECTOR_75 = 0x000E013A;
const char32_t VARIATION_SELECTOR_76 = 0x000E013B;
const char32_t VARIATION_SELECTOR_77 = 0x000E013C;
const char32_t VARIATION_SELECTOR_78 = 0x000E013D;
const char32_t VARIATION_SELECTOR_79 = 0x000E013E;
const char32_t VARIATION_SELECTOR_80 = 0x000E013F;
const char32_t VARIATION_SELECTOR_81 = 0x000E0140;
const char32_t VARIATION_SELECTOR_82 = 0x000E0141;
const char32_t VARIATION_SELECTOR_83 = 0x000E0142;
const char32_t VARIATION_SELECTOR_84 = 0x000E0143;
const char32_t VARIATION_SELECTOR_85 = 0x000E0144;
const char32_t VARIATION_SELECTOR_86 = 0x000E0145;
const char32_t VARIATION_SELECTOR_87 = 0x000E0146;
const char32_t VARIATION_SELECTOR_88 = 0x000E0147;
const char32_t VARIATION_SELECTOR_89 = 0x000E0148;
const char32_t VARIATION_SELECTOR_90 = 0x000E0149;
const char32_t VARIATION_SELECTOR_91 = 0x000E014A;
const char32_t VARIATION_SELECTOR_92 = 0x000E014B;
const char32_t VARIATION_SELECTOR_93 = 0x000E014C;
const char32_t VARIATION_SELECTOR_94 = 0x000E014D;
const char32_t VARIATION_SELECTOR_95 = 0x000E014E;
const char32_t VARIATION_SELECTOR_96 = 0x000E014F;
const char32_t VARIATION_SELECTOR_97 = 0x000E0150;
const char32_t VARIATION_SELECTOR_98 = 0x000E0151;
const char32_t VARIATION_SELECTOR_99 = 0x000E0152;
const char32_t VARIATION_SELECTOR_100 = 0x000E0153;
const char32_t VARIATION_SELECTOR_101 = 0x000E0154;
const char32_t VARIATION_SELECTOR_102 = 0x000E0155;
const char32_t VARIATION_SELECTOR_103 = 0x000E0156;
const char32_t VARIATION_SELECTOR_104 = 0x000E0157;
const char32_t VARIATION_SELECTOR_105 = 0x000E0158;
const char32_t VARIATION_SELECTOR_106 = 0x000E0159;
const char32_t VARIATION_SELECTOR_107 = 0x000E015A;
const char32_t VARIATION_SELECTOR_108 = 0x000E015B;
const char32_t VARIATION_SELECTOR_109 = 0x000E015C;
const char32_t VARIATION_SELECTOR_110 = 0x000E015D;
const char32_t VARIATION_SELECTOR_111 = 0x000E015E;
const char32_t VARIATION_SELECTOR_112 = 0x000E015F;
const char32_t VARIATION_SELECTOR_113 = 0x000E0160;
const char32_t VARIATION_SELECTOR_114 = 0x000E0161;
const char32_t VARIATION_SELECTOR_115 = 0x000E0162;
const char32_t VARIATION_SELECTOR_116 = 0x000E0163;
const char32_t VARIATION_SELECTOR_117 = 0x000E0164;
const char32_t VARIATION_SELECTOR_118 = 0x000E0165;
const char32_t VARIATION_SELECTOR_119 = 0x000E0166;
const char32_t VARIATION_SELECTOR_120 = 0x000E0167;
const char32_t VARIATION_SELECTOR_121 = 0x000E0168;
const char32_t VARIATION_SELECTOR_122 = 0x000E0169;
const char32_t VARIATION_SELECTOR_123 = 0x000E016A;
const char32_t VARIATION_SELECTOR_124 = 0x000E016B;
const char32_t VARIATION_SELECTOR_125 = 0x000E016C;
const char32_t VARIATION_SELECTOR_126 = 0x000E016D;
const char32_t VARIATION_SELECTOR_127 = 0x000E016E;
const char32_t VARIATION_SELECTOR_128 = 0x000E016F;
const char32_t VARIATION_SELECTOR_129 = 0x000E0170;
const char32_t VARIATION_SELECTOR_130 = 0x000E0171;
const char32_t VARIATION_SELECTOR_131 = 0x000E0172;
const char32_t VARIATION_SELECTOR_132 = 0x000E0173;
const char32_t VARIATION_SELECTOR_133 = 0x000E0174;
const char32_t VARIATION_SELECTOR_134 = 0x000E0175;
const char32_t VARIATION_SELECTOR_135 = 0x000E0176;
const char32_t VARIATION_SELECTOR_136 = 0x000E0177;
const char32_t VARIATION_SELECTOR_137 = 0x000E0178;
const char32_t VARIATION_SELECTOR_138 = 0x000E0179;
const char32_t VARIATION_SELECTOR_139 = 0x000E017A;
const char32_t VARIATION_SELECTOR_140 = 0x000E017B;
const char32_t VARIATION_SELECTOR_141 = 0x000E017C;
const char32_t VARIATION_SELECTOR_142 = 0x000E017D;
const char32_t VARIATION_SELECTOR_143 = 0x000E017E;
const char32_t VARIATION_SELECTOR_144 = 0x000E017F;
const char32_t VARIATION_SELECTOR_145 = 0x000E0180;
const char32_t VARIATION_SELECTOR_146 = 0x000E0181;
const char32_t VARIATION_SELECTOR_147 = 0x000E0182;
const char32_t VARIATION_SELECTOR_148 = 0x000E0183;
const char32_t VARIATION_SELECTOR_149 = 0x000E0184;
const char32_t VARIATION_SELECTOR_150 = 0x000E0185;
const char32_t VARIATION_SELECTOR_151 = 0x000E0186;
const char32_t VARIATION_SELECTOR_152 = 0x000E0187;
const char32_t VARIATION_SELECTOR_153 = 0x000E0188;
const char32_t VARIATION_SELECTOR_154 = 0x000E0189;
const char32_t VARIATION_SELECTOR_155 = 0x000E018A;
const char32_t VARIATION_SELECTOR_156 = 0x000E018B;
const char32_t VARIATION_SELECTOR_157 = 0x000E018C;
const char32_t VARIATION_SELECTOR_158 = 0x000E018D;
const char32_t VARIATION_SELECTOR_159 = 0x000E018E;
const char32_t VARIATION_SELECTOR_160 = 0x000E018F;
const char32_t VARIATION_SELECTOR_161 = 0x000E0190;
const char32_t VARIATION_SELECTOR_162 = 0x000E0191;
const char32_t VARIATION_SELECTOR_163 = 0x000E0192;
const char32_t VARIATION_SELECTOR_164 = 0x000E0193;
const char32_t VARIATION_SELECTOR_165 = 0x000E0194;
const char32_t VARIATION_SELECTOR_166 = 0x000E0195;
const char32_t VARIATION_SELECTOR_167 = 0x000E0196;
const char32_t VARIATION_SELECTOR_168 = 0x000E0197;
const char32_t VARIATION_SELECTOR_169 = 0x000E0198;
const char32_t VARIATION_SELECTOR_170 = 0x000E0199;
const char32_t VARIATION_SELECTOR_171 = 0x000E019A;
const char32_t VARIATION_SELECTOR_172 = 0x000E019B;
const char32_t VARIATION_SELECTOR_173 = 0x000E019C;
const char32_t VARIATION_SELECTOR_174 = 0x000E019D;
const char32_t VARIATION_SELECTOR_175 = 0x000E019E;
const char32_t VARIATION_SELECTOR_176 = 0x000E019F;
const char32_t VARIATION_SELECTOR_177 = 0x000E01A0;
const char32_t VARIATION_SELECTOR_178 = 0x000E01A1;
const char32_t VARIATION_SELECTOR_179 = 0x000E01A2;
const char32_t VARIATION_SELECTOR_180 = 0x000E01A3;
const char32_t VARIATION_SELECTOR_181 = 0x000E01A4;
const char32_t VARIATION_SELECTOR_182 = 0x000E01A5;
const char32_t VARIATION_SELECTOR_183 = 0x000E01A6;
const char32_t VARIATION_SELECTOR_184 = 0x000E01A7;
const char32_t VARIATION_SELECTOR_185 = 0x000E01A8;
const char32_t VARIATION_SELECTOR_186 = 0x000E01A9;
const char32_t VARIATION_SELECTOR_187 = 0x000E01AA;
const char32_t VARIATION_SELECTOR_188 = 0x000E01AB;
const char32_t VARIATION_SELECTOR_189 = 0x000E01AC;
const char32_t VARIATION_SELECTOR_190 = 0x000E01AD;
const char32_t VARIATION_SELECTOR_191 = 0x000E01AE;
const char32_t VARIATION_SELECTOR_192 = 0x000E01AF;
const char32_t VARIATION_SELECTOR_193 = 0x000E01B0;
const char32_t VARIATION_SELECTOR_194 = 0x000E01B1;
const char32_t VARIATION_SELECTOR_195 = 0x000E01B2;
const char32_t VARIATION_SELECTOR_196 = 0x000E01B3;
const char32_t VARIATION_SELECTOR_197 = 0x000E01B4;
const char32_t VARIATION_SELECTOR_198 = 0x000E01B5;
const char32_t VARIATION_SELECTOR_199 = 0x000E01B6;
const char32_t VARIATION_SELECTOR_200 = 0x000E01B7;
const char32_t VARIATION_SELECTOR_201 = 0x000E01B8;
const char32_t VARIATION_SELECTOR_202 = 0x000E01B9;
const char32_t VARIATION_SELECTOR_203 = 0x000E01BA;
const char32_t VARIATION_SELECTOR_204 = 0x000E01BB;
const char32_t VARIATION_SELECTOR_205 = 0x000E01BC;
const char32_t VARIATION_SELECTOR_206 = 0x000E01BD;
const char32_t VARIATION_SELECTOR_207 = 0x000E01BE;
const char32_t VARIATION_SELECTOR_208 = 0x000E01BF;
const char32_t VARIATION_SELECTOR_209 = 0x000E01C0;
const char32_t VARIATION_SELECTOR_210 = 0x000E01C1;
const char32_t VARIATION_SELECTOR_211 = 0x000E01C2;
const char32_t VARIATION_SELECTOR_212 = 0x000E01C3;
const char32_t VARIATION_SELECTOR_213 = 0x000E01C4;
const char32_t VARIATION_SELECTOR_214 = 0x000E01C5;
const char32_t VARIATION_SELECTOR_215 = 0x000E01C6;
const char32_t VARIATION_SELECTOR_216 = 0x000E01C7;
const char32_t VARIATION_SELECTOR_217 = 0x000E01C8;
const char32_t VARIATION_SELECTOR_218 = 0x000E01C9;
const char32_t VARIATION_SELECTOR_219 = 0x000E01CA;
const char32_t VARIATION_SELECTOR_220 = 0x000E01CB;
const char32_t VARIATION_SELECTOR_221 = 0x000E01CC;
const char32_t VARIATION_SELECTOR_222 = 0x000E01CD;
const char32_t VARIATION_SELECTOR_223 = 0x000E01CE;
const char32_t VARIATION_SELECTOR_224 = 0x000E01CF;
const char32_t VARIATION_SELECTOR_225 = 0x000E01D0;
const char32_t VARIATION_SELECTOR_226 = 0x000E01D1;
const char32_t VARIATION_SELECTOR_227 = 0x000E01D2;
const char32_t VARIATION_SELECTOR_228 = 0x000E01D3;
const char32_t VARIATION_SELECTOR_229 = 0x000E01D4;
const char32_t VARIATION_SELECTOR_230 = 0x000E01D5;
const char32_t VARIATION_SELECTOR_231 = 0x000E01D6;
const char32_t VARIATION_SELECTOR_232 = 0x000E01D7;
const char32_t VARIATION_SELECTOR_233 = 0x000E01D8;
const char32_t VARIATION_SELECTOR_234 = 0x000E01D9;
const char32_t VARIATION_SELECTOR_235 = 0x000E01DA;
const char32_t VARIATION_SELECTOR_236 = 0x000E01DB;
const char32_t VARIATION_SELECTOR_237 = 0x000E01DC;
const char32_t VARIATION_SELECTOR_238 = 0x000E01DD;
const char32_t VARIATION_SELECTOR_239 = 0x000E01DE;
const char32_t VARIATION_SELECTOR_240 = 0x000E01DF;
const char32_t VARIATION_SELECTOR_241 = 0x000E01E0;
const char32_t VARIATION_SELECTOR_242 = 0x000E01E1;
const char32_t VARIATION_SELECTOR_243 = 0x000E01E2;
const char32_t VARIATION_SELECTOR_244 = 0x000E01E3;
const char32_t VARIATION_SELECTOR_245 = 0x000E01E4;
const char32_t VARIATION_SELECTOR_246 = 0x000E01E5;
const char32_t VARIATION_SELECTOR_247 = 0x000E01E6;
const char32_t VARIATION_SELECTOR_248 = 0x000E01E7;
const char32_t VARIATION_SELECTOR_249 = 0x000E01E8;
const char32_t VARIATION_SELECTOR_250 = 0x000E01E9;
const char32_t VARIATION_SELECTOR_251 = 0x000E01EA;
const char32_t VARIATION_SELECTOR_252 = 0x000E01EB;
const char32_t VARIATION_SELECTOR_253 = 0x000E01EC;
const char32_t VARIATION_SELECTOR_254 = 0x000E01ED;
const char32_t VARIATION_SELECTOR_255 = 0x000E01EE;
const char32_t VARIATION_SELECTOR_256 = 0x000E01EF;
const char32_t PLANE_15_PRIVATE_USE_FIRST = 0x000F0000;
const char32_t PLANE_15_PRIVATE_USE_LAST = 0x000FFFFD;
const char32_t PLANE_16_PRIVATE_USE_FIRST = 0x00100000;
const char32_t PLANE_16_PRIVATE_USE_LAST = 0x0010FFFD;

} // namespace unicode

